__all__ = [
    'HResult',
]

from enum import Enum
from typing import Final, Optional

from ntstatus._thirtytwobits import ThirtyTwoBits

HRESULT_LOOKUP: dict[int, ThirtyTwoBits] = dict()

def cached_hresult(code: int, name: str) -> ThirtyTwoBits:
    if code in HRESULT_LOOKUP:
        return HRESULT_LOOKUP[code]

    bits = ThirtyTwoBits(code, name)
    HRESULT_LOOKUP[code] = bits
    return bits

class HResult:
    """
    This class defines all named [HRESULT values](https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-erref/0642cb2f-2075-4469-918c-4441e69c548a) as attributes.

    Constants are represented using the `ntstatus.ThirtyTwoBits` class which can be equality-compared to integers; the comparision will return True for both the signed and unsigned interpretation of the underlying 32 bits.

    Static methods:
    - `HResult.decode(code)`: Looks up an HRESULT constant by its numeric value and returns it as an `ntstatus.ThirtyTwoBits` object, with its name set to the the name of the constant. Returns `None` if argument is not a known HRESULT constant.
    - `HResult.decode_name(code, default='')`: Returns the name of the HRESULT value associated with the first argument if there is such a value. Otherwise, returns the second argument.
    - `HResult.make(code)`: Looks up an HRESULT constant by its numeric value and returns it as an `ntstatus.ThirtyTwoBits` object, with its name set to the the name of the constant. Raises `ValueError` if argument is not a known HRESULT value.

    The complete list of HRESULT values can be found [here](https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-erref/705fb797-2175-4a90-b5a3-3918024b10b8). `HResult` encodes all of them, plus the HRESULT version of all NTSTATUS values and Win32 Error Codes exposed by `ntstatus.NtStatus` and `ntstatus.Win32Error`.
    """

    @staticmethod
    def decode(code: int) -> Optional[ThirtyTwoBits]:
        """
        Looks up an HRESULT constant by its numeric value and returns it as an `ntstatus.ThirtyTwoBits` object, with its name set to the the name of the constant. Returns `None` if argument is not a known HRESULT constant.
        """

        if not isinstance(code, int):
            raise TypeError(f'Not a valid int value: {repr(code)}')

        return HRESULT_LOOKUP.get(code, None)

    @staticmethod
    def decode_name(code: int, default: str = '') -> str:
        """
        Returns the name of the HRESULT value associated with the first argument if there is such a value. Otherwise, returns the second argument.
        """

        hres = HResult.decode(code)
        if hres is not None:
            return hres.name
        else:
            return default

    @staticmethod
    def make(code: int) -> ThirtyTwoBits:
        """
        Looks up an HRESULT constant by its numeric value and returns it as an `ntstatus.ThirtyTwoBits` object, with its name set to the the name of the constant. Raises `ValueError` if argument is not a known HRESULT value.
        """

        hres = HResult.decode(code)
        if hres is not None:
            return hres
        else:
            raise ValueError(f'Unknown HRESULT value: {code}')

    # AUTOGENERATED FROM HERE, DO NOT TOUCH

    E_PENDING: Final[ThirtyTwoBits] = cached_hresult(0x8000000A, 'E_PENDING')
    E_BOUNDS: Final[ThirtyTwoBits] = cached_hresult(0x8000000B, 'E_BOUNDS')
    E_CHANGED_STATE: Final[ThirtyTwoBits] = cached_hresult(0x8000000C, 'E_CHANGED_STATE')
    E_ILLEGAL_STATE_CHANGE: Final[ThirtyTwoBits] = cached_hresult(0x8000000D, 'E_ILLEGAL_STATE_CHANGE')
    E_ILLEGAL_METHOD_CALL: Final[ThirtyTwoBits] = cached_hresult(0x8000000E, 'E_ILLEGAL_METHOD_CALL')
    RO_E_METADATA_NAME_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8000000F, 'RO_E_METADATA_NAME_NOT_FOUND')
    RO_E_METADATA_NAME_IS_NAMESPACE: Final[ThirtyTwoBits] = cached_hresult(0x80000010, 'RO_E_METADATA_NAME_IS_NAMESPACE')
    RO_E_METADATA_INVALID_TYPE_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0x80000011, 'RO_E_METADATA_INVALID_TYPE_FORMAT')
    RO_E_INVALID_METADATA_FILE: Final[ThirtyTwoBits] = cached_hresult(0x80000012, 'RO_E_INVALID_METADATA_FILE')
    RO_E_CLOSED: Final[ThirtyTwoBits] = cached_hresult(0x80000013, 'RO_E_CLOSED')
    RO_E_EXCLUSIVE_WRITE: Final[ThirtyTwoBits] = cached_hresult(0x80000014, 'RO_E_EXCLUSIVE_WRITE')
    RO_E_CHANGE_NOTIFICATION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x80000015, 'RO_E_CHANGE_NOTIFICATION_IN_PROGRESS')
    RO_E_ERROR_STRING_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80000016, 'RO_E_ERROR_STRING_NOT_FOUND')
    E_STRING_NOT_NULL_TERMINATED: Final[ThirtyTwoBits] = cached_hresult(0x80000017, 'E_STRING_NOT_NULL_TERMINATED')
    E_ILLEGAL_DELEGATE_ASSIGNMENT: Final[ThirtyTwoBits] = cached_hresult(0x80000018, 'E_ILLEGAL_DELEGATE_ASSIGNMENT')
    E_ASYNC_OPERATION_NOT_STARTED: Final[ThirtyTwoBits] = cached_hresult(0x80000019, 'E_ASYNC_OPERATION_NOT_STARTED')
    E_APPLICATION_EXITING: Final[ThirtyTwoBits] = cached_hresult(0x8000001A, 'E_APPLICATION_EXITING')
    E_APPLICATION_VIEW_EXITING: Final[ThirtyTwoBits] = cached_hresult(0x8000001B, 'E_APPLICATION_VIEW_EXITING')
    RO_E_MUST_BE_AGILE: Final[ThirtyTwoBits] = cached_hresult(0x8000001C, 'RO_E_MUST_BE_AGILE')
    RO_E_UNSUPPORTED_FROM_MTA: Final[ThirtyTwoBits] = cached_hresult(0x8000001D, 'RO_E_UNSUPPORTED_FROM_MTA')
    RO_E_COMMITTED: Final[ThirtyTwoBits] = cached_hresult(0x8000001E, 'RO_E_COMMITTED')
    RO_E_BLOCKED_CROSS_ASTA_CALL: Final[ThirtyTwoBits] = cached_hresult(0x8000001F, 'RO_E_BLOCKED_CROSS_ASTA_CALL')
    RO_E_CANNOT_ACTIVATE_FULL_TRUST_SERVER: Final[ThirtyTwoBits] = cached_hresult(0x80000020, 'RO_E_CANNOT_ACTIVATE_FULL_TRUST_SERVER')
    RO_E_CANNOT_ACTIVATE_UNIVERSAL_APPLICATION_SERVER: Final[ThirtyTwoBits] = cached_hresult(0x80000021, 'RO_E_CANNOT_ACTIVATE_UNIVERSAL_APPLICATION_SERVER')
    E_NOTIMPL: Final[ThirtyTwoBits] = cached_hresult(0x80004001, 'E_NOTIMPL')
    E_NOINTERFACE: Final[ThirtyTwoBits] = cached_hresult(0x80004002, 'E_NOINTERFACE')
    E_POINTER: Final[ThirtyTwoBits] = cached_hresult(0x80004003, 'E_POINTER')
    E_ABORT: Final[ThirtyTwoBits] = cached_hresult(0x80004004, 'E_ABORT')
    E_FAIL: Final[ThirtyTwoBits] = cached_hresult(0x80004005, 'E_FAIL')
    CO_E_INIT_TLS: Final[ThirtyTwoBits] = cached_hresult(0x80004006, 'CO_E_INIT_TLS')
    CO_E_INIT_SHARED_ALLOCATOR: Final[ThirtyTwoBits] = cached_hresult(0x80004007, 'CO_E_INIT_SHARED_ALLOCATOR')
    CO_E_INIT_MEMORY_ALLOCATOR: Final[ThirtyTwoBits] = cached_hresult(0x80004008, 'CO_E_INIT_MEMORY_ALLOCATOR')
    CO_E_INIT_CLASS_CACHE: Final[ThirtyTwoBits] = cached_hresult(0x80004009, 'CO_E_INIT_CLASS_CACHE')
    CO_E_INIT_RPC_CHANNEL: Final[ThirtyTwoBits] = cached_hresult(0x8000400A, 'CO_E_INIT_RPC_CHANNEL')
    CO_E_INIT_TLS_SET_CHANNEL_CONTROL: Final[ThirtyTwoBits] = cached_hresult(0x8000400B, 'CO_E_INIT_TLS_SET_CHANNEL_CONTROL')
    CO_E_INIT_TLS_CHANNEL_CONTROL: Final[ThirtyTwoBits] = cached_hresult(0x8000400C, 'CO_E_INIT_TLS_CHANNEL_CONTROL')
    CO_E_INIT_UNACCEPTED_USER_ALLOCATOR: Final[ThirtyTwoBits] = cached_hresult(0x8000400D, 'CO_E_INIT_UNACCEPTED_USER_ALLOCATOR')
    CO_E_INIT_SCM_MUTEX_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x8000400E, 'CO_E_INIT_SCM_MUTEX_EXISTS')
    CO_E_INIT_SCM_FILE_MAPPING_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x8000400F, 'CO_E_INIT_SCM_FILE_MAPPING_EXISTS')
    CO_E_INIT_SCM_MAP_VIEW_OF_FILE: Final[ThirtyTwoBits] = cached_hresult(0x80004010, 'CO_E_INIT_SCM_MAP_VIEW_OF_FILE')
    CO_E_INIT_SCM_EXEC_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80004011, 'CO_E_INIT_SCM_EXEC_FAILURE')
    CO_E_INIT_ONLY_SINGLE_THREADED: Final[ThirtyTwoBits] = cached_hresult(0x80004012, 'CO_E_INIT_ONLY_SINGLE_THREADED')
    CO_E_CANT_REMOTE: Final[ThirtyTwoBits] = cached_hresult(0x80004013, 'CO_E_CANT_REMOTE')
    CO_E_BAD_SERVER_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80004014, 'CO_E_BAD_SERVER_NAME')
    CO_E_WRONG_SERVER_IDENTITY: Final[ThirtyTwoBits] = cached_hresult(0x80004015, 'CO_E_WRONG_SERVER_IDENTITY')
    CO_E_OLE1DDE_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80004016, 'CO_E_OLE1DDE_DISABLED')
    CO_E_RUNAS_SYNTAX: Final[ThirtyTwoBits] = cached_hresult(0x80004017, 'CO_E_RUNAS_SYNTAX')
    CO_E_CREATEPROCESS_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80004018, 'CO_E_CREATEPROCESS_FAILURE')
    CO_E_RUNAS_CREATEPROCESS_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80004019, 'CO_E_RUNAS_CREATEPROCESS_FAILURE')
    CO_E_RUNAS_LOGON_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8000401A, 'CO_E_RUNAS_LOGON_FAILURE')
    CO_E_LAUNCH_PERMSSION_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x8000401B, 'CO_E_LAUNCH_PERMSSION_DENIED')
    CO_E_START_SERVICE_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8000401C, 'CO_E_START_SERVICE_FAILURE')
    CO_E_REMOTE_COMMUNICATION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8000401D, 'CO_E_REMOTE_COMMUNICATION_FAILURE')
    CO_E_SERVER_START_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x8000401E, 'CO_E_SERVER_START_TIMEOUT')
    CO_E_CLSREG_INCONSISTENT: Final[ThirtyTwoBits] = cached_hresult(0x8000401F, 'CO_E_CLSREG_INCONSISTENT')
    CO_E_IIDREG_INCONSISTENT: Final[ThirtyTwoBits] = cached_hresult(0x80004020, 'CO_E_IIDREG_INCONSISTENT')
    CO_E_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80004021, 'CO_E_NOT_SUPPORTED')
    CO_E_RELOAD_DLL: Final[ThirtyTwoBits] = cached_hresult(0x80004022, 'CO_E_RELOAD_DLL')
    CO_E_MSI_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80004023, 'CO_E_MSI_ERROR')
    CO_E_ATTEMPT_TO_CREATE_OUTSIDE_CLIENT_CONTEXT: Final[ThirtyTwoBits] = cached_hresult(0x80004024, 'CO_E_ATTEMPT_TO_CREATE_OUTSIDE_CLIENT_CONTEXT')
    CO_E_SERVER_PAUSED: Final[ThirtyTwoBits] = cached_hresult(0x80004025, 'CO_E_SERVER_PAUSED')
    CO_E_SERVER_NOT_PAUSED: Final[ThirtyTwoBits] = cached_hresult(0x80004026, 'CO_E_SERVER_NOT_PAUSED')
    CO_E_CLASS_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80004027, 'CO_E_CLASS_DISABLED')
    CO_E_CLRNOTAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80004028, 'CO_E_CLRNOTAVAILABLE')
    CO_E_ASYNC_WORK_REJECTED: Final[ThirtyTwoBits] = cached_hresult(0x80004029, 'CO_E_ASYNC_WORK_REJECTED')
    CO_E_SERVER_INIT_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x8000402A, 'CO_E_SERVER_INIT_TIMEOUT')
    CO_E_NO_SECCTX_IN_ACTIVATE: Final[ThirtyTwoBits] = cached_hresult(0x8000402B, 'CO_E_NO_SECCTX_IN_ACTIVATE')
    CO_E_TRACKER_CONFIG: Final[ThirtyTwoBits] = cached_hresult(0x80004030, 'CO_E_TRACKER_CONFIG')
    CO_E_THREADPOOL_CONFIG: Final[ThirtyTwoBits] = cached_hresult(0x80004031, 'CO_E_THREADPOOL_CONFIG')
    CO_E_SXS_CONFIG: Final[ThirtyTwoBits] = cached_hresult(0x80004032, 'CO_E_SXS_CONFIG')
    CO_E_MALFORMED_SPN: Final[ThirtyTwoBits] = cached_hresult(0x80004033, 'CO_E_MALFORMED_SPN')
    CO_E_UNREVOKED_REGISTRATION_ON_APARTMENT_SHUTDOWN: Final[ThirtyTwoBits] = cached_hresult(0x80004034, 'CO_E_UNREVOKED_REGISTRATION_ON_APARTMENT_SHUTDOWN')
    CO_E_PREMATURE_STUB_RUNDOWN: Final[ThirtyTwoBits] = cached_hresult(0x80004035, 'CO_E_PREMATURE_STUB_RUNDOWN')
    E_UNEXPECTED: Final[ThirtyTwoBits] = cached_hresult(0x8000FFFF, 'E_UNEXPECTED')
    RPC_E_CALL_REJECTED: Final[ThirtyTwoBits] = cached_hresult(0x80010001, 'RPC_E_CALL_REJECTED')
    RPC_E_CALL_CANCELED: Final[ThirtyTwoBits] = cached_hresult(0x80010002, 'RPC_E_CALL_CANCELED')
    RPC_E_CANTPOST_INSENDCALL: Final[ThirtyTwoBits] = cached_hresult(0x80010003, 'RPC_E_CANTPOST_INSENDCALL')
    RPC_E_CANTCALLOUT_INASYNCCALL: Final[ThirtyTwoBits] = cached_hresult(0x80010004, 'RPC_E_CANTCALLOUT_INASYNCCALL')
    RPC_E_CANTCALLOUT_INEXTERNALCALL: Final[ThirtyTwoBits] = cached_hresult(0x80010005, 'RPC_E_CANTCALLOUT_INEXTERNALCALL')
    RPC_E_CONNECTION_TERMINATED: Final[ThirtyTwoBits] = cached_hresult(0x80010006, 'RPC_E_CONNECTION_TERMINATED')
    RPC_E_SERVER_DIED: Final[ThirtyTwoBits] = cached_hresult(0x80010007, 'RPC_E_SERVER_DIED')
    RPC_E_CLIENT_DIED: Final[ThirtyTwoBits] = cached_hresult(0x80010008, 'RPC_E_CLIENT_DIED')
    RPC_E_INVALID_DATAPACKET: Final[ThirtyTwoBits] = cached_hresult(0x80010009, 'RPC_E_INVALID_DATAPACKET')
    RPC_E_CANTTRANSMIT_CALL: Final[ThirtyTwoBits] = cached_hresult(0x8001000A, 'RPC_E_CANTTRANSMIT_CALL')
    RPC_E_CLIENT_CANTMARSHAL_DATA: Final[ThirtyTwoBits] = cached_hresult(0x8001000B, 'RPC_E_CLIENT_CANTMARSHAL_DATA')
    RPC_E_CLIENT_CANTUNMARSHAL_DATA: Final[ThirtyTwoBits] = cached_hresult(0x8001000C, 'RPC_E_CLIENT_CANTUNMARSHAL_DATA')
    RPC_E_SERVER_CANTMARSHAL_DATA: Final[ThirtyTwoBits] = cached_hresult(0x8001000D, 'RPC_E_SERVER_CANTMARSHAL_DATA')
    RPC_E_SERVER_CANTUNMARSHAL_DATA: Final[ThirtyTwoBits] = cached_hresult(0x8001000E, 'RPC_E_SERVER_CANTUNMARSHAL_DATA')
    RPC_E_INVALID_DATA: Final[ThirtyTwoBits] = cached_hresult(0x8001000F, 'RPC_E_INVALID_DATA')
    RPC_E_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0x80010010, 'RPC_E_INVALID_PARAMETER')
    RPC_E_CANTCALLOUT_AGAIN: Final[ThirtyTwoBits] = cached_hresult(0x80010011, 'RPC_E_CANTCALLOUT_AGAIN')
    RPC_E_SERVER_DIED_DNE: Final[ThirtyTwoBits] = cached_hresult(0x80010012, 'RPC_E_SERVER_DIED_DNE')
    RPC_E_SYS_CALL_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80010100, 'RPC_E_SYS_CALL_FAILED')
    RPC_E_OUT_OF_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0x80010101, 'RPC_E_OUT_OF_RESOURCES')
    RPC_E_ATTEMPTED_MULTITHREAD: Final[ThirtyTwoBits] = cached_hresult(0x80010102, 'RPC_E_ATTEMPTED_MULTITHREAD')
    RPC_E_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x80010103, 'RPC_E_NOT_REGISTERED')
    RPC_E_FAULT: Final[ThirtyTwoBits] = cached_hresult(0x80010104, 'RPC_E_FAULT')
    RPC_E_SERVERFAULT: Final[ThirtyTwoBits] = cached_hresult(0x80010105, 'RPC_E_SERVERFAULT')
    RPC_E_CHANGED_MODE: Final[ThirtyTwoBits] = cached_hresult(0x80010106, 'RPC_E_CHANGED_MODE')
    RPC_E_INVALIDMETHOD: Final[ThirtyTwoBits] = cached_hresult(0x80010107, 'RPC_E_INVALIDMETHOD')
    RPC_E_DISCONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x80010108, 'RPC_E_DISCONNECTED')
    RPC_E_RETRY: Final[ThirtyTwoBits] = cached_hresult(0x80010109, 'RPC_E_RETRY')
    RPC_E_SERVERCALL_RETRYLATER: Final[ThirtyTwoBits] = cached_hresult(0x8001010A, 'RPC_E_SERVERCALL_RETRYLATER')
    RPC_E_SERVERCALL_REJECTED: Final[ThirtyTwoBits] = cached_hresult(0x8001010B, 'RPC_E_SERVERCALL_REJECTED')
    RPC_E_INVALID_CALLDATA: Final[ThirtyTwoBits] = cached_hresult(0x8001010C, 'RPC_E_INVALID_CALLDATA')
    RPC_E_CANTCALLOUT_ININPUTSYNCCALL: Final[ThirtyTwoBits] = cached_hresult(0x8001010D, 'RPC_E_CANTCALLOUT_ININPUTSYNCCALL')
    RPC_E_WRONG_THREAD: Final[ThirtyTwoBits] = cached_hresult(0x8001010E, 'RPC_E_WRONG_THREAD')
    RPC_E_THREAD_NOT_INIT: Final[ThirtyTwoBits] = cached_hresult(0x8001010F, 'RPC_E_THREAD_NOT_INIT')
    RPC_E_VERSION_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80010110, 'RPC_E_VERSION_MISMATCH')
    RPC_E_INVALID_HEADER: Final[ThirtyTwoBits] = cached_hresult(0x80010111, 'RPC_E_INVALID_HEADER')
    RPC_E_INVALID_EXTENSION: Final[ThirtyTwoBits] = cached_hresult(0x80010112, 'RPC_E_INVALID_EXTENSION')
    RPC_E_INVALID_IPID: Final[ThirtyTwoBits] = cached_hresult(0x80010113, 'RPC_E_INVALID_IPID')
    RPC_E_INVALID_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x80010114, 'RPC_E_INVALID_OBJECT')
    RPC_S_CALLPENDING: Final[ThirtyTwoBits] = cached_hresult(0x80010115, 'RPC_S_CALLPENDING')
    RPC_S_WAITONTIMER: Final[ThirtyTwoBits] = cached_hresult(0x80010116, 'RPC_S_WAITONTIMER')
    RPC_E_CALL_COMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x80010117, 'RPC_E_CALL_COMPLETE')
    RPC_E_UNSECURE_CALL: Final[ThirtyTwoBits] = cached_hresult(0x80010118, 'RPC_E_UNSECURE_CALL')
    RPC_E_TOO_LATE: Final[ThirtyTwoBits] = cached_hresult(0x80010119, 'RPC_E_TOO_LATE')
    RPC_E_NO_GOOD_SECURITY_PACKAGES: Final[ThirtyTwoBits] = cached_hresult(0x8001011A, 'RPC_E_NO_GOOD_SECURITY_PACKAGES')
    RPC_E_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x8001011B, 'RPC_E_ACCESS_DENIED')
    RPC_E_REMOTE_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x8001011C, 'RPC_E_REMOTE_DISABLED')
    RPC_E_INVALID_OBJREF: Final[ThirtyTwoBits] = cached_hresult(0x8001011D, 'RPC_E_INVALID_OBJREF')
    RPC_E_NO_CONTEXT: Final[ThirtyTwoBits] = cached_hresult(0x8001011E, 'RPC_E_NO_CONTEXT')
    RPC_E_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x8001011F, 'RPC_E_TIMEOUT')
    RPC_E_NO_SYNC: Final[ThirtyTwoBits] = cached_hresult(0x80010120, 'RPC_E_NO_SYNC')
    RPC_E_FULLSIC_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80010121, 'RPC_E_FULLSIC_REQUIRED')
    RPC_E_INVALID_STD_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80010122, 'RPC_E_INVALID_STD_NAME')
    CO_E_FAILEDTOIMPERSONATE: Final[ThirtyTwoBits] = cached_hresult(0x80010123, 'CO_E_FAILEDTOIMPERSONATE')
    CO_E_FAILEDTOGETSECCTX: Final[ThirtyTwoBits] = cached_hresult(0x80010124, 'CO_E_FAILEDTOGETSECCTX')
    CO_E_FAILEDTOOPENTHREADTOKEN: Final[ThirtyTwoBits] = cached_hresult(0x80010125, 'CO_E_FAILEDTOOPENTHREADTOKEN')
    CO_E_FAILEDTOGETTOKENINFO: Final[ThirtyTwoBits] = cached_hresult(0x80010126, 'CO_E_FAILEDTOGETTOKENINFO')
    CO_E_TRUSTEEDOESNTMATCHCLIENT: Final[ThirtyTwoBits] = cached_hresult(0x80010127, 'CO_E_TRUSTEEDOESNTMATCHCLIENT')
    CO_E_FAILEDTOQUERYCLIENTBLANKET: Final[ThirtyTwoBits] = cached_hresult(0x80010128, 'CO_E_FAILEDTOQUERYCLIENTBLANKET')
    CO_E_FAILEDTOSETDACL: Final[ThirtyTwoBits] = cached_hresult(0x80010129, 'CO_E_FAILEDTOSETDACL')
    CO_E_ACCESSCHECKFAILED: Final[ThirtyTwoBits] = cached_hresult(0x8001012A, 'CO_E_ACCESSCHECKFAILED')
    CO_E_NETACCESSAPIFAILED: Final[ThirtyTwoBits] = cached_hresult(0x8001012B, 'CO_E_NETACCESSAPIFAILED')
    CO_E_WRONGTRUSTEENAMESYNTAX: Final[ThirtyTwoBits] = cached_hresult(0x8001012C, 'CO_E_WRONGTRUSTEENAMESYNTAX')
    CO_E_INVALIDSID: Final[ThirtyTwoBits] = cached_hresult(0x8001012D, 'CO_E_INVALIDSID')
    CO_E_CONVERSIONFAILED: Final[ThirtyTwoBits] = cached_hresult(0x8001012E, 'CO_E_CONVERSIONFAILED')
    CO_E_NOMATCHINGSIDFOUND: Final[ThirtyTwoBits] = cached_hresult(0x8001012F, 'CO_E_NOMATCHINGSIDFOUND')
    CO_E_LOOKUPACCSIDFAILED: Final[ThirtyTwoBits] = cached_hresult(0x80010130, 'CO_E_LOOKUPACCSIDFAILED')
    CO_E_NOMATCHINGNAMEFOUND: Final[ThirtyTwoBits] = cached_hresult(0x80010131, 'CO_E_NOMATCHINGNAMEFOUND')
    CO_E_LOOKUPACCNAMEFAILED: Final[ThirtyTwoBits] = cached_hresult(0x80010132, 'CO_E_LOOKUPACCNAMEFAILED')
    CO_E_SETSERLHNDLFAILED: Final[ThirtyTwoBits] = cached_hresult(0x80010133, 'CO_E_SETSERLHNDLFAILED')
    CO_E_FAILEDTOGETWINDIR: Final[ThirtyTwoBits] = cached_hresult(0x80010134, 'CO_E_FAILEDTOGETWINDIR')
    CO_E_PATHTOOLONG: Final[ThirtyTwoBits] = cached_hresult(0x80010135, 'CO_E_PATHTOOLONG')
    CO_E_FAILEDTOGENUUID: Final[ThirtyTwoBits] = cached_hresult(0x80010136, 'CO_E_FAILEDTOGENUUID')
    CO_E_FAILEDTOCREATEFILE: Final[ThirtyTwoBits] = cached_hresult(0x80010137, 'CO_E_FAILEDTOCREATEFILE')
    CO_E_FAILEDTOCLOSEHANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80010138, 'CO_E_FAILEDTOCLOSEHANDLE')
    CO_E_EXCEEDSYSACLLIMIT: Final[ThirtyTwoBits] = cached_hresult(0x80010139, 'CO_E_EXCEEDSYSACLLIMIT')
    CO_E_ACESINWRONGORDER: Final[ThirtyTwoBits] = cached_hresult(0x8001013A, 'CO_E_ACESINWRONGORDER')
    CO_E_INCOMPATIBLESTREAMVERSION: Final[ThirtyTwoBits] = cached_hresult(0x8001013B, 'CO_E_INCOMPATIBLESTREAMVERSION')
    CO_E_FAILEDTOOPENPROCESSTOKEN: Final[ThirtyTwoBits] = cached_hresult(0x8001013C, 'CO_E_FAILEDTOOPENPROCESSTOKEN')
    CO_E_DECODEFAILED: Final[ThirtyTwoBits] = cached_hresult(0x8001013D, 'CO_E_DECODEFAILED')
    CO_E_ACNOTINITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0x8001013F, 'CO_E_ACNOTINITIALIZED')
    CO_E_CANCEL_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80010140, 'CO_E_CANCEL_DISABLED')
    RPC_E_UNEXPECTED: Final[ThirtyTwoBits] = cached_hresult(0x8001FFFF, 'RPC_E_UNEXPECTED')
    DISP_E_UNKNOWNINTERFACE: Final[ThirtyTwoBits] = cached_hresult(0x80020001, 'DISP_E_UNKNOWNINTERFACE')
    DISP_E_MEMBERNOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x80020003, 'DISP_E_MEMBERNOTFOUND')
    DISP_E_PARAMNOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x80020004, 'DISP_E_PARAMNOTFOUND')
    DISP_E_TYPEMISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80020005, 'DISP_E_TYPEMISMATCH')
    DISP_E_UNKNOWNNAME: Final[ThirtyTwoBits] = cached_hresult(0x80020006, 'DISP_E_UNKNOWNNAME')
    DISP_E_NONAMEDARGS: Final[ThirtyTwoBits] = cached_hresult(0x80020007, 'DISP_E_NONAMEDARGS')
    DISP_E_BADVARTYPE: Final[ThirtyTwoBits] = cached_hresult(0x80020008, 'DISP_E_BADVARTYPE')
    DISP_E_EXCEPTION: Final[ThirtyTwoBits] = cached_hresult(0x80020009, 'DISP_E_EXCEPTION')
    DISP_E_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0x8002000A, 'DISP_E_OVERFLOW')
    DISP_E_BADINDEX: Final[ThirtyTwoBits] = cached_hresult(0x8002000B, 'DISP_E_BADINDEX')
    DISP_E_UNKNOWNLCID: Final[ThirtyTwoBits] = cached_hresult(0x8002000C, 'DISP_E_UNKNOWNLCID')
    DISP_E_ARRAYISLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x8002000D, 'DISP_E_ARRAYISLOCKED')
    DISP_E_BADPARAMCOUNT: Final[ThirtyTwoBits] = cached_hresult(0x8002000E, 'DISP_E_BADPARAMCOUNT')
    DISP_E_PARAMNOTOPTIONAL: Final[ThirtyTwoBits] = cached_hresult(0x8002000F, 'DISP_E_PARAMNOTOPTIONAL')
    DISP_E_BADCALLEE: Final[ThirtyTwoBits] = cached_hresult(0x80020010, 'DISP_E_BADCALLEE')
    DISP_E_NOTACOLLECTION: Final[ThirtyTwoBits] = cached_hresult(0x80020011, 'DISP_E_NOTACOLLECTION')
    DISP_E_DIVBYZERO: Final[ThirtyTwoBits] = cached_hresult(0x80020012, 'DISP_E_DIVBYZERO')
    DISP_E_BUFFERTOOSMALL: Final[ThirtyTwoBits] = cached_hresult(0x80020013, 'DISP_E_BUFFERTOOSMALL')
    TYPE_E_BUFFERTOOSMALL: Final[ThirtyTwoBits] = cached_hresult(0x80028016, 'TYPE_E_BUFFERTOOSMALL')
    TYPE_E_FIELDNOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x80028017, 'TYPE_E_FIELDNOTFOUND')
    TYPE_E_INVDATAREAD: Final[ThirtyTwoBits] = cached_hresult(0x80028018, 'TYPE_E_INVDATAREAD')
    TYPE_E_UNSUPFORMAT: Final[ThirtyTwoBits] = cached_hresult(0x80028019, 'TYPE_E_UNSUPFORMAT')
    TYPE_E_REGISTRYACCESS: Final[ThirtyTwoBits] = cached_hresult(0x8002801C, 'TYPE_E_REGISTRYACCESS')
    TYPE_E_LIBNOTREGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x8002801D, 'TYPE_E_LIBNOTREGISTERED')
    TYPE_E_UNDEFINEDTYPE: Final[ThirtyTwoBits] = cached_hresult(0x80028027, 'TYPE_E_UNDEFINEDTYPE')
    TYPE_E_QUALIFIEDNAMEDISALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80028028, 'TYPE_E_QUALIFIEDNAMEDISALLOWED')
    TYPE_E_INVALIDSTATE: Final[ThirtyTwoBits] = cached_hresult(0x80028029, 'TYPE_E_INVALIDSTATE')
    TYPE_E_WRONGTYPEKIND: Final[ThirtyTwoBits] = cached_hresult(0x8002802A, 'TYPE_E_WRONGTYPEKIND')
    TYPE_E_ELEMENTNOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x8002802B, 'TYPE_E_ELEMENTNOTFOUND')
    TYPE_E_AMBIGUOUSNAME: Final[ThirtyTwoBits] = cached_hresult(0x8002802C, 'TYPE_E_AMBIGUOUSNAME')
    TYPE_E_NAMECONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x8002802D, 'TYPE_E_NAMECONFLICT')
    TYPE_E_UNKNOWNLCID: Final[ThirtyTwoBits] = cached_hresult(0x8002802E, 'TYPE_E_UNKNOWNLCID')
    TYPE_E_DLLFUNCTIONNOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x8002802F, 'TYPE_E_DLLFUNCTIONNOTFOUND')
    TYPE_E_BADMODULEKIND: Final[ThirtyTwoBits] = cached_hresult(0x800288BD, 'TYPE_E_BADMODULEKIND')
    TYPE_E_SIZETOOBIG: Final[ThirtyTwoBits] = cached_hresult(0x800288C5, 'TYPE_E_SIZETOOBIG')
    TYPE_E_DUPLICATEID: Final[ThirtyTwoBits] = cached_hresult(0x800288C6, 'TYPE_E_DUPLICATEID')
    TYPE_E_INVALIDID: Final[ThirtyTwoBits] = cached_hresult(0x800288CF, 'TYPE_E_INVALIDID')
    TYPE_E_TYPEMISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80028CA0, 'TYPE_E_TYPEMISMATCH')
    TYPE_E_OUTOFBOUNDS: Final[ThirtyTwoBits] = cached_hresult(0x80028CA1, 'TYPE_E_OUTOFBOUNDS')
    TYPE_E_IOERROR: Final[ThirtyTwoBits] = cached_hresult(0x80028CA2, 'TYPE_E_IOERROR')
    TYPE_E_CANTCREATETMPFILE: Final[ThirtyTwoBits] = cached_hresult(0x80028CA3, 'TYPE_E_CANTCREATETMPFILE')
    TYPE_E_CANTLOADLIBRARY: Final[ThirtyTwoBits] = cached_hresult(0x80029C4A, 'TYPE_E_CANTLOADLIBRARY')
    TYPE_E_INCONSISTENTPROPFUNCS: Final[ThirtyTwoBits] = cached_hresult(0x80029C83, 'TYPE_E_INCONSISTENTPROPFUNCS')
    TYPE_E_CIRCULARTYPE: Final[ThirtyTwoBits] = cached_hresult(0x80029C84, 'TYPE_E_CIRCULARTYPE')
    STG_E_INVALIDFUNCTION: Final[ThirtyTwoBits] = cached_hresult(0x80030001, 'STG_E_INVALIDFUNCTION')
    STG_E_FILENOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x80030002, 'STG_E_FILENOTFOUND')
    STG_E_PATHNOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x80030003, 'STG_E_PATHNOTFOUND')
    STG_E_TOOMANYOPENFILES: Final[ThirtyTwoBits] = cached_hresult(0x80030004, 'STG_E_TOOMANYOPENFILES')
    STG_E_ACCESSDENIED: Final[ThirtyTwoBits] = cached_hresult(0x80030005, 'STG_E_ACCESSDENIED')
    STG_E_INVALIDHANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80030006, 'STG_E_INVALIDHANDLE')
    STG_E_INSUFFICIENTMEMORY: Final[ThirtyTwoBits] = cached_hresult(0x80030008, 'STG_E_INSUFFICIENTMEMORY')
    STG_E_INVALIDPOINTER: Final[ThirtyTwoBits] = cached_hresult(0x80030009, 'STG_E_INVALIDPOINTER')
    STG_E_NOMOREFILES: Final[ThirtyTwoBits] = cached_hresult(0x80030012, 'STG_E_NOMOREFILES')
    STG_E_DISKISWRITEPROTECTED: Final[ThirtyTwoBits] = cached_hresult(0x80030013, 'STG_E_DISKISWRITEPROTECTED')
    STG_E_SEEKERROR: Final[ThirtyTwoBits] = cached_hresult(0x80030019, 'STG_E_SEEKERROR')
    STG_E_WRITEFAULT: Final[ThirtyTwoBits] = cached_hresult(0x8003001D, 'STG_E_WRITEFAULT')
    STG_E_READFAULT: Final[ThirtyTwoBits] = cached_hresult(0x8003001E, 'STG_E_READFAULT')
    STG_E_SHAREVIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x80030020, 'STG_E_SHAREVIOLATION')
    STG_E_LOCKVIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x80030021, 'STG_E_LOCKVIOLATION')
    STG_E_FILEALREADYEXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80030050, 'STG_E_FILEALREADYEXISTS')
    STG_E_INVALIDPARAMETER: Final[ThirtyTwoBits] = cached_hresult(0x80030057, 'STG_E_INVALIDPARAMETER')
    STG_E_MEDIUMFULL: Final[ThirtyTwoBits] = cached_hresult(0x80030070, 'STG_E_MEDIUMFULL')
    STG_E_PROPSETMISMATCHED: Final[ThirtyTwoBits] = cached_hresult(0x800300F0, 'STG_E_PROPSETMISMATCHED')
    STG_E_ABNORMALAPIEXIT: Final[ThirtyTwoBits] = cached_hresult(0x800300FA, 'STG_E_ABNORMALAPIEXIT')
    STG_E_INVALIDHEADER: Final[ThirtyTwoBits] = cached_hresult(0x800300FB, 'STG_E_INVALIDHEADER')
    STG_E_INVALIDNAME: Final[ThirtyTwoBits] = cached_hresult(0x800300FC, 'STG_E_INVALIDNAME')
    STG_E_UNKNOWN: Final[ThirtyTwoBits] = cached_hresult(0x800300FD, 'STG_E_UNKNOWN')
    STG_E_UNIMPLEMENTEDFUNCTION: Final[ThirtyTwoBits] = cached_hresult(0x800300FE, 'STG_E_UNIMPLEMENTEDFUNCTION')
    STG_E_INVALIDFLAG: Final[ThirtyTwoBits] = cached_hresult(0x800300FF, 'STG_E_INVALIDFLAG')
    STG_E_INUSE: Final[ThirtyTwoBits] = cached_hresult(0x80030100, 'STG_E_INUSE')
    STG_E_NOTCURRENT: Final[ThirtyTwoBits] = cached_hresult(0x80030101, 'STG_E_NOTCURRENT')
    STG_E_REVERTED: Final[ThirtyTwoBits] = cached_hresult(0x80030102, 'STG_E_REVERTED')
    STG_E_CANTSAVE: Final[ThirtyTwoBits] = cached_hresult(0x80030103, 'STG_E_CANTSAVE')
    STG_E_OLDFORMAT: Final[ThirtyTwoBits] = cached_hresult(0x80030104, 'STG_E_OLDFORMAT')
    STG_E_OLDDLL: Final[ThirtyTwoBits] = cached_hresult(0x80030105, 'STG_E_OLDDLL')
    STG_E_SHAREREQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80030106, 'STG_E_SHAREREQUIRED')
    STG_E_NOTFILEBASEDSTORAGE: Final[ThirtyTwoBits] = cached_hresult(0x80030107, 'STG_E_NOTFILEBASEDSTORAGE')
    STG_E_EXTANTMARSHALLINGS: Final[ThirtyTwoBits] = cached_hresult(0x80030108, 'STG_E_EXTANTMARSHALLINGS')
    STG_E_DOCFILECORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x80030109, 'STG_E_DOCFILECORRUPT')
    STG_E_BADBASEADDRESS: Final[ThirtyTwoBits] = cached_hresult(0x80030110, 'STG_E_BADBASEADDRESS')
    STG_E_DOCFILETOOLARGE: Final[ThirtyTwoBits] = cached_hresult(0x80030111, 'STG_E_DOCFILETOOLARGE')
    STG_E_NOTSIMPLEFORMAT: Final[ThirtyTwoBits] = cached_hresult(0x80030112, 'STG_E_NOTSIMPLEFORMAT')
    STG_S_CONVERTED: Final[ThirtyTwoBits] = cached_hresult(0x00030200, 'STG_S_CONVERTED')
    STG_S_BLOCK: Final[ThirtyTwoBits] = cached_hresult(0x00030201, 'STG_S_BLOCK')
    STG_E_INCOMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x80030201, 'STG_E_INCOMPLETE')
    STG_S_RETRYNOW: Final[ThirtyTwoBits] = cached_hresult(0x00030202, 'STG_S_RETRYNOW')
    STG_E_TERMINATED: Final[ThirtyTwoBits] = cached_hresult(0x80030202, 'STG_E_TERMINATED')
    STG_S_MONITORING: Final[ThirtyTwoBits] = cached_hresult(0x00030203, 'STG_S_MONITORING')
    STG_S_MULTIPLEOPENS: Final[ThirtyTwoBits] = cached_hresult(0x00030204, 'STG_S_MULTIPLEOPENS')
    STG_S_CONSOLIDATIONFAILED: Final[ThirtyTwoBits] = cached_hresult(0x00030205, 'STG_S_CONSOLIDATIONFAILED')
    STG_S_CANNOTCONSOLIDATE: Final[ThirtyTwoBits] = cached_hresult(0x00030206, 'STG_S_CANNOTCONSOLIDATE')
    STG_S_POWER_CYCLE_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x00030207, 'STG_S_POWER_CYCLE_REQUIRED')
    STG_E_FIRMWARE_SLOT_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80030208, 'STG_E_FIRMWARE_SLOT_INVALID')
    STG_E_FIRMWARE_IMAGE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80030209, 'STG_E_FIRMWARE_IMAGE_INVALID')
    STG_E_DEVICE_UNRESPONSIVE: Final[ThirtyTwoBits] = cached_hresult(0x8003020A, 'STG_E_DEVICE_UNRESPONSIVE')
    STG_E_STATUS_COPY_PROTECTION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80030305, 'STG_E_STATUS_COPY_PROTECTION_FAILURE')
    STG_E_CSS_AUTHENTICATION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80030306, 'STG_E_CSS_AUTHENTICATION_FAILURE')
    STG_E_CSS_KEY_NOT_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0x80030307, 'STG_E_CSS_KEY_NOT_PRESENT')
    STG_E_CSS_KEY_NOT_ESTABLISHED: Final[ThirtyTwoBits] = cached_hresult(0x80030308, 'STG_E_CSS_KEY_NOT_ESTABLISHED')
    STG_E_CSS_SCRAMBLED_SECTOR: Final[ThirtyTwoBits] = cached_hresult(0x80030309, 'STG_E_CSS_SCRAMBLED_SECTOR')
    STG_E_CSS_REGION_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x8003030A, 'STG_E_CSS_REGION_MISMATCH')
    STG_E_RESETS_EXHAUSTED: Final[ThirtyTwoBits] = cached_hresult(0x8003030B, 'STG_E_RESETS_EXHAUSTED')
    OLE_S_USEREG: Final[ThirtyTwoBits] = cached_hresult(0x00040000, 'OLE_S_USEREG')
    OLE_E_OLEVERB: Final[ThirtyTwoBits] = cached_hresult(0x80040000, 'OLE_E_OLEVERB')
    OLE_S_STATIC: Final[ThirtyTwoBits] = cached_hresult(0x00040001, 'OLE_S_STATIC')
    OLE_E_ADVF: Final[ThirtyTwoBits] = cached_hresult(0x80040001, 'OLE_E_ADVF')
    OLE_S_MAC_CLIPFORMAT: Final[ThirtyTwoBits] = cached_hresult(0x00040002, 'OLE_S_MAC_CLIPFORMAT')
    OLE_E_ENUM_NOMORE: Final[ThirtyTwoBits] = cached_hresult(0x80040002, 'OLE_E_ENUM_NOMORE')
    OLE_E_ADVISENOTSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80040003, 'OLE_E_ADVISENOTSUPPORTED')
    OLE_E_NOCONNECTION: Final[ThirtyTwoBits] = cached_hresult(0x80040004, 'OLE_E_NOCONNECTION')
    OLE_E_NOTRUNNING: Final[ThirtyTwoBits] = cached_hresult(0x80040005, 'OLE_E_NOTRUNNING')
    OLE_E_NOCACHE: Final[ThirtyTwoBits] = cached_hresult(0x80040006, 'OLE_E_NOCACHE')
    OLE_E_BLANK: Final[ThirtyTwoBits] = cached_hresult(0x80040007, 'OLE_E_BLANK')
    OLE_E_CLASSDIFF: Final[ThirtyTwoBits] = cached_hresult(0x80040008, 'OLE_E_CLASSDIFF')
    OLE_E_CANT_GETMONIKER: Final[ThirtyTwoBits] = cached_hresult(0x80040009, 'OLE_E_CANT_GETMONIKER')
    OLE_E_CANT_BINDTOSOURCE: Final[ThirtyTwoBits] = cached_hresult(0x8004000A, 'OLE_E_CANT_BINDTOSOURCE')
    OLE_E_STATIC: Final[ThirtyTwoBits] = cached_hresult(0x8004000B, 'OLE_E_STATIC')
    OLE_E_PROMPTSAVECANCELLED: Final[ThirtyTwoBits] = cached_hresult(0x8004000C, 'OLE_E_PROMPTSAVECANCELLED')
    OLE_E_INVALIDRECT: Final[ThirtyTwoBits] = cached_hresult(0x8004000D, 'OLE_E_INVALIDRECT')
    OLE_E_WRONGCOMPOBJ: Final[ThirtyTwoBits] = cached_hresult(0x8004000E, 'OLE_E_WRONGCOMPOBJ')
    OLE_E_INVALIDHWND: Final[ThirtyTwoBits] = cached_hresult(0x8004000F, 'OLE_E_INVALIDHWND')
    OLE_E_NOT_INPLACEACTIVE: Final[ThirtyTwoBits] = cached_hresult(0x80040010, 'OLE_E_NOT_INPLACEACTIVE')
    OLE_E_CANTCONVERT: Final[ThirtyTwoBits] = cached_hresult(0x80040011, 'OLE_E_CANTCONVERT')
    OLE_E_NOSTORAGE: Final[ThirtyTwoBits] = cached_hresult(0x80040012, 'OLE_E_NOSTORAGE')
    DV_E_FORMATETC: Final[ThirtyTwoBits] = cached_hresult(0x80040064, 'DV_E_FORMATETC')
    DV_E_DVTARGETDEVICE: Final[ThirtyTwoBits] = cached_hresult(0x80040065, 'DV_E_DVTARGETDEVICE')
    DV_E_STGMEDIUM: Final[ThirtyTwoBits] = cached_hresult(0x80040066, 'DV_E_STGMEDIUM')
    DV_E_STATDATA: Final[ThirtyTwoBits] = cached_hresult(0x80040067, 'DV_E_STATDATA')
    DV_E_LINDEX: Final[ThirtyTwoBits] = cached_hresult(0x80040068, 'DV_E_LINDEX')
    DV_E_TYMED: Final[ThirtyTwoBits] = cached_hresult(0x80040069, 'DV_E_TYMED')
    DV_E_CLIPFORMAT: Final[ThirtyTwoBits] = cached_hresult(0x8004006A, 'DV_E_CLIPFORMAT')
    DV_E_DVASPECT: Final[ThirtyTwoBits] = cached_hresult(0x8004006B, 'DV_E_DVASPECT')
    DV_E_DVTARGETDEVICE_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x8004006C, 'DV_E_DVTARGETDEVICE_SIZE')
    DV_E_NOIVIEWOBJECT: Final[ThirtyTwoBits] = cached_hresult(0x8004006D, 'DV_E_NOIVIEWOBJECT')
    DRAGDROP_S_DROP: Final[ThirtyTwoBits] = cached_hresult(0x00040100, 'DRAGDROP_S_DROP')
    DRAGDROP_E_NOTREGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x80040100, 'DRAGDROP_E_NOTREGISTERED')
    DRAGDROP_S_CANCEL: Final[ThirtyTwoBits] = cached_hresult(0x00040101, 'DRAGDROP_S_CANCEL')
    DRAGDROP_E_ALREADYREGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x80040101, 'DRAGDROP_E_ALREADYREGISTERED')
    DRAGDROP_S_USEDEFAULTCURSORS: Final[ThirtyTwoBits] = cached_hresult(0x00040102, 'DRAGDROP_S_USEDEFAULTCURSORS')
    DRAGDROP_E_INVALIDHWND: Final[ThirtyTwoBits] = cached_hresult(0x80040102, 'DRAGDROP_E_INVALIDHWND')
    DRAGDROP_E_CONCURRENT_DRAG_ATTEMPTED: Final[ThirtyTwoBits] = cached_hresult(0x80040103, 'DRAGDROP_E_CONCURRENT_DRAG_ATTEMPTED')
    CLASS_E_NOAGGREGATION: Final[ThirtyTwoBits] = cached_hresult(0x80040110, 'CLASS_E_NOAGGREGATION')
    CLASS_E_CLASSNOTAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80040111, 'CLASS_E_CLASSNOTAVAILABLE')
    CLASS_E_NOTLICENSED: Final[ThirtyTwoBits] = cached_hresult(0x80040112, 'CLASS_E_NOTLICENSED')
    DATA_S_SAMEFORMATETC: Final[ThirtyTwoBits] = cached_hresult(0x00040130, 'DATA_S_SAMEFORMATETC')
    VIEW_S_ALREADY_FROZEN: Final[ThirtyTwoBits] = cached_hresult(0x00040140, 'VIEW_S_ALREADY_FROZEN')
    VIEW_E_DRAW: Final[ThirtyTwoBits] = cached_hresult(0x80040140, 'VIEW_E_DRAW')
    REGDB_E_READREGDB: Final[ThirtyTwoBits] = cached_hresult(0x80040150, 'REGDB_E_READREGDB')
    REGDB_E_WRITEREGDB: Final[ThirtyTwoBits] = cached_hresult(0x80040151, 'REGDB_E_WRITEREGDB')
    REGDB_E_KEYMISSING: Final[ThirtyTwoBits] = cached_hresult(0x80040152, 'REGDB_E_KEYMISSING')
    REGDB_E_INVALIDVALUE: Final[ThirtyTwoBits] = cached_hresult(0x80040153, 'REGDB_E_INVALIDVALUE')
    REGDB_E_CLASSNOTREG: Final[ThirtyTwoBits] = cached_hresult(0x80040154, 'REGDB_E_CLASSNOTREG')
    REGDB_E_IIDNOTREG: Final[ThirtyTwoBits] = cached_hresult(0x80040155, 'REGDB_E_IIDNOTREG')
    REGDB_E_BADTHREADINGMODEL: Final[ThirtyTwoBits] = cached_hresult(0x80040156, 'REGDB_E_BADTHREADINGMODEL')
    REGDB_E_PACKAGEPOLICYVIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x80040157, 'REGDB_E_PACKAGEPOLICYVIOLATION')
    CAT_E_CATIDNOEXIST: Final[ThirtyTwoBits] = cached_hresult(0x80040160, 'CAT_E_CATIDNOEXIST')
    CAT_E_NODESCRIPTION: Final[ThirtyTwoBits] = cached_hresult(0x80040161, 'CAT_E_NODESCRIPTION')
    CS_E_PACKAGE_NOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x80040164, 'CS_E_PACKAGE_NOTFOUND')
    CS_E_NOT_DELETABLE: Final[ThirtyTwoBits] = cached_hresult(0x80040165, 'CS_E_NOT_DELETABLE')
    CS_E_CLASS_NOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x80040166, 'CS_E_CLASS_NOTFOUND')
    CS_E_INVALID_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80040167, 'CS_E_INVALID_VERSION')
    CS_E_NO_CLASSSTORE: Final[ThirtyTwoBits] = cached_hresult(0x80040168, 'CS_E_NO_CLASSSTORE')
    CS_E_OBJECT_NOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x80040169, 'CS_E_OBJECT_NOTFOUND')
    CS_E_OBJECT_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x8004016A, 'CS_E_OBJECT_ALREADY_EXISTS')
    CS_E_INVALID_PATH: Final[ThirtyTwoBits] = cached_hresult(0x8004016B, 'CS_E_INVALID_PATH')
    CS_E_NETWORK_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8004016C, 'CS_E_NETWORK_ERROR')
    CS_E_ADMIN_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8004016D, 'CS_E_ADMIN_LIMIT_EXCEEDED')
    CS_E_SCHEMA_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x8004016E, 'CS_E_SCHEMA_MISMATCH')
    CS_E_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8004016F, 'CS_E_INTERNAL_ERROR')
    CACHE_S_FORMATETC_NOTSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x00040170, 'CACHE_S_FORMATETC_NOTSUPPORTED')
    CACHE_E_NOCACHE_UPDATED: Final[ThirtyTwoBits] = cached_hresult(0x80040170, 'CACHE_E_NOCACHE_UPDATED')
    CACHE_S_SAMECACHE: Final[ThirtyTwoBits] = cached_hresult(0x00040171, 'CACHE_S_SAMECACHE')
    CACHE_S_SOMECACHES_NOTUPDATED: Final[ThirtyTwoBits] = cached_hresult(0x00040172, 'CACHE_S_SOMECACHES_NOTUPDATED')
    OLEOBJ_S_INVALIDVERB: Final[ThirtyTwoBits] = cached_hresult(0x00040180, 'OLEOBJ_S_INVALIDVERB')
    OLEOBJ_E_NOVERBS: Final[ThirtyTwoBits] = cached_hresult(0x80040180, 'OLEOBJ_E_NOVERBS')
    OLEOBJ_S_CANNOT_DOVERB_NOW: Final[ThirtyTwoBits] = cached_hresult(0x00040181, 'OLEOBJ_S_CANNOT_DOVERB_NOW')
    OLEOBJ_E_INVALIDVERB: Final[ThirtyTwoBits] = cached_hresult(0x80040181, 'OLEOBJ_E_INVALIDVERB')
    OLEOBJ_S_INVALIDHWND: Final[ThirtyTwoBits] = cached_hresult(0x00040182, 'OLEOBJ_S_INVALIDHWND')
    INPLACE_S_TRUNCATED: Final[ThirtyTwoBits] = cached_hresult(0x000401A0, 'INPLACE_S_TRUNCATED')
    INPLACE_E_NOTUNDOABLE: Final[ThirtyTwoBits] = cached_hresult(0x800401A0, 'INPLACE_E_NOTUNDOABLE')
    INPLACE_E_NOTOOLSPACE: Final[ThirtyTwoBits] = cached_hresult(0x800401A1, 'INPLACE_E_NOTOOLSPACE')
    CONVERT10_S_NO_PRESENTATION: Final[ThirtyTwoBits] = cached_hresult(0x000401C0, 'CONVERT10_S_NO_PRESENTATION')
    CONVERT10_E_OLESTREAM_GET: Final[ThirtyTwoBits] = cached_hresult(0x800401C0, 'CONVERT10_E_OLESTREAM_GET')
    CONVERT10_E_OLESTREAM_PUT: Final[ThirtyTwoBits] = cached_hresult(0x800401C1, 'CONVERT10_E_OLESTREAM_PUT')
    CONVERT10_E_OLESTREAM_FMT: Final[ThirtyTwoBits] = cached_hresult(0x800401C2, 'CONVERT10_E_OLESTREAM_FMT')
    CONVERT10_E_OLESTREAM_BITMAP_TO_DIB: Final[ThirtyTwoBits] = cached_hresult(0x800401C3, 'CONVERT10_E_OLESTREAM_BITMAP_TO_DIB')
    CONVERT10_E_STG_FMT: Final[ThirtyTwoBits] = cached_hresult(0x800401C4, 'CONVERT10_E_STG_FMT')
    CONVERT10_E_STG_NO_STD_STREAM: Final[ThirtyTwoBits] = cached_hresult(0x800401C5, 'CONVERT10_E_STG_NO_STD_STREAM')
    CONVERT10_E_STG_DIB_TO_BITMAP: Final[ThirtyTwoBits] = cached_hresult(0x800401C6, 'CONVERT10_E_STG_DIB_TO_BITMAP')
    CONVERT10_E_OLELINK_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x800401C7, 'CONVERT10_E_OLELINK_DISABLED')
    CLIPBRD_E_CANT_OPEN: Final[ThirtyTwoBits] = cached_hresult(0x800401D0, 'CLIPBRD_E_CANT_OPEN')
    CLIPBRD_E_CANT_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0x800401D1, 'CLIPBRD_E_CANT_EMPTY')
    CLIPBRD_E_CANT_SET: Final[ThirtyTwoBits] = cached_hresult(0x800401D2, 'CLIPBRD_E_CANT_SET')
    CLIPBRD_E_BAD_DATA: Final[ThirtyTwoBits] = cached_hresult(0x800401D3, 'CLIPBRD_E_BAD_DATA')
    CLIPBRD_E_CANT_CLOSE: Final[ThirtyTwoBits] = cached_hresult(0x800401D4, 'CLIPBRD_E_CANT_CLOSE')
    MK_E_CONNECTMANUALLY: Final[ThirtyTwoBits] = cached_hresult(0x800401E0, 'MK_E_CONNECTMANUALLY')
    MK_E_EXCEEDEDDEADLINE: Final[ThirtyTwoBits] = cached_hresult(0x800401E1, 'MK_E_EXCEEDEDDEADLINE')
    MK_S_REDUCED_TO_SELF: Final[ThirtyTwoBits] = cached_hresult(0x000401E2, 'MK_S_REDUCED_TO_SELF')
    MK_E_NEEDGENERIC: Final[ThirtyTwoBits] = cached_hresult(0x800401E2, 'MK_E_NEEDGENERIC')
    MK_E_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x800401E3, 'MK_E_UNAVAILABLE')
    MK_S_ME: Final[ThirtyTwoBits] = cached_hresult(0x000401E4, 'MK_S_ME')
    MK_E_SYNTAX: Final[ThirtyTwoBits] = cached_hresult(0x800401E4, 'MK_E_SYNTAX')
    MK_S_HIM: Final[ThirtyTwoBits] = cached_hresult(0x000401E5, 'MK_S_HIM')
    MK_E_NOOBJECT: Final[ThirtyTwoBits] = cached_hresult(0x800401E5, 'MK_E_NOOBJECT')
    MK_S_US: Final[ThirtyTwoBits] = cached_hresult(0x000401E6, 'MK_S_US')
    MK_E_INVALIDEXTENSION: Final[ThirtyTwoBits] = cached_hresult(0x800401E6, 'MK_E_INVALIDEXTENSION')
    MK_S_MONIKERALREADYREGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x000401E7, 'MK_S_MONIKERALREADYREGISTERED')
    MK_E_INTERMEDIATEINTERFACENOTSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x800401E7, 'MK_E_INTERMEDIATEINTERFACENOTSUPPORTED')
    MK_E_NOTBINDABLE: Final[ThirtyTwoBits] = cached_hresult(0x800401E8, 'MK_E_NOTBINDABLE')
    MK_E_NOTBOUND: Final[ThirtyTwoBits] = cached_hresult(0x800401E9, 'MK_E_NOTBOUND')
    MK_E_CANTOPENFILE: Final[ThirtyTwoBits] = cached_hresult(0x800401EA, 'MK_E_CANTOPENFILE')
    MK_E_MUSTBOTHERUSER: Final[ThirtyTwoBits] = cached_hresult(0x800401EB, 'MK_E_MUSTBOTHERUSER')
    MK_E_NOINVERSE: Final[ThirtyTwoBits] = cached_hresult(0x800401EC, 'MK_E_NOINVERSE')
    MK_E_NOSTORAGE: Final[ThirtyTwoBits] = cached_hresult(0x800401ED, 'MK_E_NOSTORAGE')
    MK_E_NOPREFIX: Final[ThirtyTwoBits] = cached_hresult(0x800401EE, 'MK_E_NOPREFIX')
    MK_E_ENUMERATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800401EF, 'MK_E_ENUMERATION_FAILED')
    CO_E_NOTINITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0x800401F0, 'CO_E_NOTINITIALIZED')
    CO_E_ALREADYINITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0x800401F1, 'CO_E_ALREADYINITIALIZED')
    CO_E_CANTDETERMINECLASS: Final[ThirtyTwoBits] = cached_hresult(0x800401F2, 'CO_E_CANTDETERMINECLASS')
    CO_E_CLASSSTRING: Final[ThirtyTwoBits] = cached_hresult(0x800401F3, 'CO_E_CLASSSTRING')
    CO_E_IIDSTRING: Final[ThirtyTwoBits] = cached_hresult(0x800401F4, 'CO_E_IIDSTRING')
    CO_E_APPNOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x800401F5, 'CO_E_APPNOTFOUND')
    CO_E_APPSINGLEUSE: Final[ThirtyTwoBits] = cached_hresult(0x800401F6, 'CO_E_APPSINGLEUSE')
    CO_E_ERRORINAPP: Final[ThirtyTwoBits] = cached_hresult(0x800401F7, 'CO_E_ERRORINAPP')
    CO_E_DLLNOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x800401F8, 'CO_E_DLLNOTFOUND')
    CO_E_ERRORINDLL: Final[ThirtyTwoBits] = cached_hresult(0x800401F9, 'CO_E_ERRORINDLL')
    CO_E_WRONGOSFORAPP: Final[ThirtyTwoBits] = cached_hresult(0x800401FA, 'CO_E_WRONGOSFORAPP')
    CO_E_OBJNOTREG: Final[ThirtyTwoBits] = cached_hresult(0x800401FB, 'CO_E_OBJNOTREG')
    CO_E_OBJISREG: Final[ThirtyTwoBits] = cached_hresult(0x800401FC, 'CO_E_OBJISREG')
    CO_E_OBJNOTCONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x800401FD, 'CO_E_OBJNOTCONNECTED')
    CO_E_APPDIDNTREG: Final[ThirtyTwoBits] = cached_hresult(0x800401FE, 'CO_E_APPDIDNTREG')
    CO_E_RELEASED: Final[ThirtyTwoBits] = cached_hresult(0x800401FF, 'CO_E_RELEASED')
    EVENT_S_SOME_SUBSCRIBERS_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x00040200, 'EVENT_S_SOME_SUBSCRIBERS_FAILED')
    EVENT_E_ALL_SUBSCRIBERS_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80040201, 'EVENT_E_ALL_SUBSCRIBERS_FAILED')
    EVENT_S_NOSUBSCRIBERS: Final[ThirtyTwoBits] = cached_hresult(0x00040202, 'EVENT_S_NOSUBSCRIBERS')
    EVENT_E_QUERYSYNTAX: Final[ThirtyTwoBits] = cached_hresult(0x80040203, 'EVENT_E_QUERYSYNTAX')
    EVENT_E_QUERYFIELD: Final[ThirtyTwoBits] = cached_hresult(0x80040204, 'EVENT_E_QUERYFIELD')
    EVENT_E_INTERNALEXCEPTION: Final[ThirtyTwoBits] = cached_hresult(0x80040205, 'EVENT_E_INTERNALEXCEPTION')
    EVENT_E_INTERNALERROR: Final[ThirtyTwoBits] = cached_hresult(0x80040206, 'EVENT_E_INTERNALERROR')
    EVENT_E_INVALID_PER_USER_SID: Final[ThirtyTwoBits] = cached_hresult(0x80040207, 'EVENT_E_INVALID_PER_USER_SID')
    EVENT_E_USER_EXCEPTION: Final[ThirtyTwoBits] = cached_hresult(0x80040208, 'EVENT_E_USER_EXCEPTION')
    EVENT_E_TOO_MANY_METHODS: Final[ThirtyTwoBits] = cached_hresult(0x80040209, 'EVENT_E_TOO_MANY_METHODS')
    EVENT_E_MISSING_EVENTCLASS: Final[ThirtyTwoBits] = cached_hresult(0x8004020A, 'EVENT_E_MISSING_EVENTCLASS')
    EVENT_E_NOT_ALL_REMOVED: Final[ThirtyTwoBits] = cached_hresult(0x8004020B, 'EVENT_E_NOT_ALL_REMOVED')
    EVENT_E_COMPLUS_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x8004020C, 'EVENT_E_COMPLUS_NOT_INSTALLED')
    EVENT_E_CANT_MODIFY_OR_DELETE_UNCONFIGURED_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x8004020D, 'EVENT_E_CANT_MODIFY_OR_DELETE_UNCONFIGURED_OBJECT')
    EVENT_E_CANT_MODIFY_OR_DELETE_CONFIGURED_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x8004020E, 'EVENT_E_CANT_MODIFY_OR_DELETE_CONFIGURED_OBJECT')
    EVENT_E_INVALID_EVENT_CLASS_PARTITION: Final[ThirtyTwoBits] = cached_hresult(0x8004020F, 'EVENT_E_INVALID_EVENT_CLASS_PARTITION')
    EVENT_E_PER_USER_SID_NOT_LOGGED_ON: Final[ThirtyTwoBits] = cached_hresult(0x80040210, 'EVENT_E_PER_USER_SID_NOT_LOGGED_ON')
    TPC_E_NO_DEFAULT_TABLET: Final[ThirtyTwoBits] = cached_hresult(0x80040212, 'TPC_E_NO_DEFAULT_TABLET')
    TPC_E_INVALID_INPUT_RECT: Final[ThirtyTwoBits] = cached_hresult(0x80040219, 'TPC_E_INVALID_INPUT_RECT')
    TPC_E_UNKNOWN_PROPERTY: Final[ThirtyTwoBits] = cached_hresult(0x8004021B, 'TPC_E_UNKNOWN_PROPERTY')
    TPC_E_INVALID_STROKE: Final[ThirtyTwoBits] = cached_hresult(0x80040222, 'TPC_E_INVALID_STROKE')
    TPC_E_INITIALIZE_FAIL: Final[ThirtyTwoBits] = cached_hresult(0x80040223, 'TPC_E_INITIALIZE_FAIL')
    TPC_E_NOT_RELEVANT: Final[ThirtyTwoBits] = cached_hresult(0x80040232, 'TPC_E_NOT_RELEVANT')
    TPC_E_INVALID_PACKET_DESCRIPTION: Final[ThirtyTwoBits] = cached_hresult(0x80040233, 'TPC_E_INVALID_PACKET_DESCRIPTION')
    TPC_E_RECOGNIZER_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x80040235, 'TPC_E_RECOGNIZER_NOT_REGISTERED')
    TPC_E_INVALID_RIGHTS: Final[ThirtyTwoBits] = cached_hresult(0x80040236, 'TPC_E_INVALID_RIGHTS')
    TPC_E_OUT_OF_ORDER_CALL: Final[ThirtyTwoBits] = cached_hresult(0x80040237, 'TPC_E_OUT_OF_ORDER_CALL')
    TPC_E_QUEUE_FULL: Final[ThirtyTwoBits] = cached_hresult(0x80040238, 'TPC_E_QUEUE_FULL')
    TPC_E_INVALID_CONFIGURATION: Final[ThirtyTwoBits] = cached_hresult(0x80040239, 'TPC_E_INVALID_CONFIGURATION')
    TPC_E_INVALID_DATA_FROM_RECOGNIZER: Final[ThirtyTwoBits] = cached_hresult(0x8004023A, 'TPC_E_INVALID_DATA_FROM_RECOGNIZER')
    TPC_E_INVALID_PROPERTY: Final[ThirtyTwoBits] = cached_hresult(0x80040241, 'TPC_E_INVALID_PROPERTY')
    TPC_S_TRUNCATED: Final[ThirtyTwoBits] = cached_hresult(0x00040252, 'TPC_S_TRUNCATED')
    TPC_S_INTERRUPTED: Final[ThirtyTwoBits] = cached_hresult(0x00040253, 'TPC_S_INTERRUPTED')
    TPC_S_NO_DATA_TO_PROCESS: Final[ThirtyTwoBits] = cached_hresult(0x00040254, 'TPC_S_NO_DATA_TO_PROCESS')
    SCHED_S_TASK_READY: Final[ThirtyTwoBits] = cached_hresult(0x00041300, 'SCHED_S_TASK_READY')
    SCHED_S_TASK_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x00041301, 'SCHED_S_TASK_RUNNING')
    SCHED_S_TASK_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x00041302, 'SCHED_S_TASK_DISABLED')
    SCHED_S_TASK_HAS_NOT_RUN: Final[ThirtyTwoBits] = cached_hresult(0x00041303, 'SCHED_S_TASK_HAS_NOT_RUN')
    SCHED_S_TASK_NO_MORE_RUNS: Final[ThirtyTwoBits] = cached_hresult(0x00041304, 'SCHED_S_TASK_NO_MORE_RUNS')
    SCHED_S_TASK_NOT_SCHEDULED: Final[ThirtyTwoBits] = cached_hresult(0x00041305, 'SCHED_S_TASK_NOT_SCHEDULED')
    SCHED_S_TASK_TERMINATED: Final[ThirtyTwoBits] = cached_hresult(0x00041306, 'SCHED_S_TASK_TERMINATED')
    SCHED_S_TASK_NO_VALID_TRIGGERS: Final[ThirtyTwoBits] = cached_hresult(0x00041307, 'SCHED_S_TASK_NO_VALID_TRIGGERS')
    SCHED_S_EVENT_TRIGGER: Final[ThirtyTwoBits] = cached_hresult(0x00041308, 'SCHED_S_EVENT_TRIGGER')
    SCHED_E_TRIGGER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80041309, 'SCHED_E_TRIGGER_NOT_FOUND')
    SCHED_E_TASK_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0x8004130A, 'SCHED_E_TASK_NOT_READY')
    SCHED_E_TASK_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x8004130B, 'SCHED_E_TASK_NOT_RUNNING')
    SCHED_E_SERVICE_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x8004130C, 'SCHED_E_SERVICE_NOT_INSTALLED')
    SCHED_E_CANNOT_OPEN_TASK: Final[ThirtyTwoBits] = cached_hresult(0x8004130D, 'SCHED_E_CANNOT_OPEN_TASK')
    SCHED_E_INVALID_TASK: Final[ThirtyTwoBits] = cached_hresult(0x8004130E, 'SCHED_E_INVALID_TASK')
    SCHED_E_ACCOUNT_INFORMATION_NOT_SET: Final[ThirtyTwoBits] = cached_hresult(0x8004130F, 'SCHED_E_ACCOUNT_INFORMATION_NOT_SET')
    SCHED_E_ACCOUNT_NAME_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80041310, 'SCHED_E_ACCOUNT_NAME_NOT_FOUND')
    SCHED_E_ACCOUNT_DBASE_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x80041311, 'SCHED_E_ACCOUNT_DBASE_CORRUPT')
    SCHED_E_NO_SECURITY_SERVICES: Final[ThirtyTwoBits] = cached_hresult(0x80041312, 'SCHED_E_NO_SECURITY_SERVICES')
    SCHED_E_UNKNOWN_OBJECT_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80041313, 'SCHED_E_UNKNOWN_OBJECT_VERSION')
    SCHED_E_UNSUPPORTED_ACCOUNT_OPTION: Final[ThirtyTwoBits] = cached_hresult(0x80041314, 'SCHED_E_UNSUPPORTED_ACCOUNT_OPTION')
    SCHED_E_SERVICE_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x80041315, 'SCHED_E_SERVICE_NOT_RUNNING')
    SCHED_E_UNEXPECTEDNODE: Final[ThirtyTwoBits] = cached_hresult(0x80041316, 'SCHED_E_UNEXPECTEDNODE')
    SCHED_E_NAMESPACE: Final[ThirtyTwoBits] = cached_hresult(0x80041317, 'SCHED_E_NAMESPACE')
    SCHED_E_INVALIDVALUE: Final[ThirtyTwoBits] = cached_hresult(0x80041318, 'SCHED_E_INVALIDVALUE')
    SCHED_E_MISSINGNODE: Final[ThirtyTwoBits] = cached_hresult(0x80041319, 'SCHED_E_MISSINGNODE')
    SCHED_E_MALFORMEDXML: Final[ThirtyTwoBits] = cached_hresult(0x8004131A, 'SCHED_E_MALFORMEDXML')
    SCHED_S_SOME_TRIGGERS_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x0004131B, 'SCHED_S_SOME_TRIGGERS_FAILED')
    SCHED_S_BATCH_LOGON_PROBLEM: Final[ThirtyTwoBits] = cached_hresult(0x0004131C, 'SCHED_S_BATCH_LOGON_PROBLEM')
    SCHED_E_TOO_MANY_NODES: Final[ThirtyTwoBits] = cached_hresult(0x8004131D, 'SCHED_E_TOO_MANY_NODES')
    SCHED_E_PAST_END_BOUNDARY: Final[ThirtyTwoBits] = cached_hresult(0x8004131E, 'SCHED_E_PAST_END_BOUNDARY')
    SCHED_E_ALREADY_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x8004131F, 'SCHED_E_ALREADY_RUNNING')
    SCHED_E_USER_NOT_LOGGED_ON: Final[ThirtyTwoBits] = cached_hresult(0x80041320, 'SCHED_E_USER_NOT_LOGGED_ON')
    SCHED_E_INVALID_TASK_HASH: Final[ThirtyTwoBits] = cached_hresult(0x80041321, 'SCHED_E_INVALID_TASK_HASH')
    SCHED_E_SERVICE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80041322, 'SCHED_E_SERVICE_NOT_AVAILABLE')
    SCHED_E_SERVICE_TOO_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x80041323, 'SCHED_E_SERVICE_TOO_BUSY')
    SCHED_E_TASK_ATTEMPTED: Final[ThirtyTwoBits] = cached_hresult(0x80041324, 'SCHED_E_TASK_ATTEMPTED')
    SCHED_S_TASK_QUEUED: Final[ThirtyTwoBits] = cached_hresult(0x00041325, 'SCHED_S_TASK_QUEUED')
    SCHED_E_TASK_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80041326, 'SCHED_E_TASK_DISABLED')
    SCHED_E_TASK_NOT_V1_COMPAT: Final[ThirtyTwoBits] = cached_hresult(0x80041327, 'SCHED_E_TASK_NOT_V1_COMPAT')
    SCHED_E_START_ON_DEMAND: Final[ThirtyTwoBits] = cached_hresult(0x80041328, 'SCHED_E_START_ON_DEMAND')
    SCHED_E_TASK_NOT_UBPM_COMPAT: Final[ThirtyTwoBits] = cached_hresult(0x80041329, 'SCHED_E_TASK_NOT_UBPM_COMPAT')
    SCHED_E_DEPRECATED_FEATURE_USED: Final[ThirtyTwoBits] = cached_hresult(0x80041330, 'SCHED_E_DEPRECATED_FEATURE_USED')
    XACT_S_ASYNC: Final[ThirtyTwoBits] = cached_hresult(0x0004D000, 'XACT_S_ASYNC')
    XACT_E_ALREADYOTHERSINGLEPHASE: Final[ThirtyTwoBits] = cached_hresult(0x8004D000, 'XACT_E_ALREADYOTHERSINGLEPHASE')
    XACT_S_DEFECT: Final[ThirtyTwoBits] = cached_hresult(0x0004D001, 'XACT_S_DEFECT')
    XACT_E_CANTRETAIN: Final[ThirtyTwoBits] = cached_hresult(0x8004D001, 'XACT_E_CANTRETAIN')
    XACT_S_READONLY: Final[ThirtyTwoBits] = cached_hresult(0x0004D002, 'XACT_S_READONLY')
    XACT_E_COMMITFAILED: Final[ThirtyTwoBits] = cached_hresult(0x8004D002, 'XACT_E_COMMITFAILED')
    XACT_S_SOMENORETAIN: Final[ThirtyTwoBits] = cached_hresult(0x0004D003, 'XACT_S_SOMENORETAIN')
    XACT_E_COMMITPREVENTED: Final[ThirtyTwoBits] = cached_hresult(0x8004D003, 'XACT_E_COMMITPREVENTED')
    XACT_S_OKINFORM: Final[ThirtyTwoBits] = cached_hresult(0x0004D004, 'XACT_S_OKINFORM')
    XACT_E_HEURISTICABORT: Final[ThirtyTwoBits] = cached_hresult(0x8004D004, 'XACT_E_HEURISTICABORT')
    XACT_S_MADECHANGESCONTENT: Final[ThirtyTwoBits] = cached_hresult(0x0004D005, 'XACT_S_MADECHANGESCONTENT')
    XACT_E_HEURISTICCOMMIT: Final[ThirtyTwoBits] = cached_hresult(0x8004D005, 'XACT_E_HEURISTICCOMMIT')
    XACT_S_MADECHANGESINFORM: Final[ThirtyTwoBits] = cached_hresult(0x0004D006, 'XACT_S_MADECHANGESINFORM')
    XACT_E_HEURISTICDAMAGE: Final[ThirtyTwoBits] = cached_hresult(0x8004D006, 'XACT_E_HEURISTICDAMAGE')
    XACT_S_ALLNORETAIN: Final[ThirtyTwoBits] = cached_hresult(0x0004D007, 'XACT_S_ALLNORETAIN')
    XACT_E_HEURISTICDANGER: Final[ThirtyTwoBits] = cached_hresult(0x8004D007, 'XACT_E_HEURISTICDANGER')
    XACT_S_ABORTING: Final[ThirtyTwoBits] = cached_hresult(0x0004D008, 'XACT_S_ABORTING')
    XACT_E_ISOLATIONLEVEL: Final[ThirtyTwoBits] = cached_hresult(0x8004D008, 'XACT_E_ISOLATIONLEVEL')
    XACT_S_SINGLEPHASE: Final[ThirtyTwoBits] = cached_hresult(0x0004D009, 'XACT_S_SINGLEPHASE')
    XACT_E_NOASYNC: Final[ThirtyTwoBits] = cached_hresult(0x8004D009, 'XACT_E_NOASYNC')
    XACT_S_LOCALLY_OK: Final[ThirtyTwoBits] = cached_hresult(0x0004D00A, 'XACT_S_LOCALLY_OK')
    XACT_E_NOENLIST: Final[ThirtyTwoBits] = cached_hresult(0x8004D00A, 'XACT_E_NOENLIST')
    XACT_E_NOISORETAIN: Final[ThirtyTwoBits] = cached_hresult(0x8004D00B, 'XACT_E_NOISORETAIN')
    XACT_E_NORESOURCE: Final[ThirtyTwoBits] = cached_hresult(0x8004D00C, 'XACT_E_NORESOURCE')
    XACT_E_NOTCURRENT: Final[ThirtyTwoBits] = cached_hresult(0x8004D00D, 'XACT_E_NOTCURRENT')
    XACT_E_NOTRANSACTION: Final[ThirtyTwoBits] = cached_hresult(0x8004D00E, 'XACT_E_NOTRANSACTION')
    XACT_E_NOTSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8004D00F, 'XACT_E_NOTSUPPORTED')
    XACT_S_LASTRESOURCEMANAGER: Final[ThirtyTwoBits] = cached_hresult(0x0004D010, 'XACT_S_LASTRESOURCEMANAGER')
    XACT_E_UNKNOWNRMGRID: Final[ThirtyTwoBits] = cached_hresult(0x8004D010, 'XACT_E_UNKNOWNRMGRID')
    XACT_E_WRONGSTATE: Final[ThirtyTwoBits] = cached_hresult(0x8004D011, 'XACT_E_WRONGSTATE')
    XACT_E_WRONGUOW: Final[ThirtyTwoBits] = cached_hresult(0x8004D012, 'XACT_E_WRONGUOW')
    XACT_E_XTIONEXISTS: Final[ThirtyTwoBits] = cached_hresult(0x8004D013, 'XACT_E_XTIONEXISTS')
    XACT_E_NOIMPORTOBJECT: Final[ThirtyTwoBits] = cached_hresult(0x8004D014, 'XACT_E_NOIMPORTOBJECT')
    XACT_E_INVALIDCOOKIE: Final[ThirtyTwoBits] = cached_hresult(0x8004D015, 'XACT_E_INVALIDCOOKIE')
    XACT_E_INDOUBT: Final[ThirtyTwoBits] = cached_hresult(0x8004D016, 'XACT_E_INDOUBT')
    XACT_E_NOTIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x8004D017, 'XACT_E_NOTIMEOUT')
    XACT_E_ALREADYINPROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x8004D018, 'XACT_E_ALREADYINPROGRESS')
    XACT_E_ABORTED: Final[ThirtyTwoBits] = cached_hresult(0x8004D019, 'XACT_E_ABORTED')
    XACT_E_LOGFULL: Final[ThirtyTwoBits] = cached_hresult(0x8004D01A, 'XACT_E_LOGFULL')
    XACT_E_TMNOTAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x8004D01B, 'XACT_E_TMNOTAVAILABLE')
    XACT_E_CONNECTION_DOWN: Final[ThirtyTwoBits] = cached_hresult(0x8004D01C, 'XACT_E_CONNECTION_DOWN')
    XACT_E_CONNECTION_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x8004D01D, 'XACT_E_CONNECTION_DENIED')
    XACT_E_REENLISTTIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x8004D01E, 'XACT_E_REENLISTTIMEOUT')
    XACT_E_TIP_CONNECT_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8004D01F, 'XACT_E_TIP_CONNECT_FAILED')
    XACT_E_TIP_PROTOCOL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8004D020, 'XACT_E_TIP_PROTOCOL_ERROR')
    XACT_E_TIP_PULL_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8004D021, 'XACT_E_TIP_PULL_FAILED')
    XACT_E_DEST_TMNOTAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x8004D022, 'XACT_E_DEST_TMNOTAVAILABLE')
    XACT_E_TIP_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x8004D023, 'XACT_E_TIP_DISABLED')
    XACT_E_NETWORK_TX_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x8004D024, 'XACT_E_NETWORK_TX_DISABLED')
    XACT_E_PARTNER_NETWORK_TX_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x8004D025, 'XACT_E_PARTNER_NETWORK_TX_DISABLED')
    XACT_E_XA_TX_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x8004D026, 'XACT_E_XA_TX_DISABLED')
    XACT_E_UNABLE_TO_READ_DTC_CONFIG: Final[ThirtyTwoBits] = cached_hresult(0x8004D027, 'XACT_E_UNABLE_TO_READ_DTC_CONFIG')
    XACT_E_UNABLE_TO_LOAD_DTC_PROXY: Final[ThirtyTwoBits] = cached_hresult(0x8004D028, 'XACT_E_UNABLE_TO_LOAD_DTC_PROXY')
    XACT_E_ABORTING: Final[ThirtyTwoBits] = cached_hresult(0x8004D029, 'XACT_E_ABORTING')
    XACT_E_PUSH_COMM_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8004D02A, 'XACT_E_PUSH_COMM_FAILURE')
    XACT_E_PULL_COMM_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8004D02B, 'XACT_E_PULL_COMM_FAILURE')
    XACT_E_LU_TX_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x8004D02C, 'XACT_E_LU_TX_DISABLED')
    XACT_E_CLERKNOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x8004D080, 'XACT_E_CLERKNOTFOUND')
    XACT_E_CLERKEXISTS: Final[ThirtyTwoBits] = cached_hresult(0x8004D081, 'XACT_E_CLERKEXISTS')
    XACT_E_RECOVERYINPROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x8004D082, 'XACT_E_RECOVERYINPROGRESS')
    XACT_E_TRANSACTIONCLOSED: Final[ThirtyTwoBits] = cached_hresult(0x8004D083, 'XACT_E_TRANSACTIONCLOSED')
    XACT_E_INVALIDLSN: Final[ThirtyTwoBits] = cached_hresult(0x8004D084, 'XACT_E_INVALIDLSN')
    XACT_E_REPLAYREQUEST: Final[ThirtyTwoBits] = cached_hresult(0x8004D085, 'XACT_E_REPLAYREQUEST')
    CONTEXT_E_ABORTED: Final[ThirtyTwoBits] = cached_hresult(0x8004E002, 'CONTEXT_E_ABORTED')
    CONTEXT_E_ABORTING: Final[ThirtyTwoBits] = cached_hresult(0x8004E003, 'CONTEXT_E_ABORTING')
    CONTEXT_E_NOCONTEXT: Final[ThirtyTwoBits] = cached_hresult(0x8004E004, 'CONTEXT_E_NOCONTEXT')
    CONTEXT_E_WOULD_DEADLOCK: Final[ThirtyTwoBits] = cached_hresult(0x8004E005, 'CONTEXT_E_WOULD_DEADLOCK')
    CONTEXT_E_SYNCH_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x8004E006, 'CONTEXT_E_SYNCH_TIMEOUT')
    CONTEXT_E_OLDREF: Final[ThirtyTwoBits] = cached_hresult(0x8004E007, 'CONTEXT_E_OLDREF')
    CONTEXT_E_ROLENOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x8004E00C, 'CONTEXT_E_ROLENOTFOUND')
    CONTEXT_E_TMNOTAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x8004E00F, 'CONTEXT_E_TMNOTAVAILABLE')
    CO_E_ACTIVATIONFAILED: Final[ThirtyTwoBits] = cached_hresult(0x8004E021, 'CO_E_ACTIVATIONFAILED')
    CO_E_ACTIVATIONFAILED_EVENTLOGGED: Final[ThirtyTwoBits] = cached_hresult(0x8004E022, 'CO_E_ACTIVATIONFAILED_EVENTLOGGED')
    CO_E_ACTIVATIONFAILED_CATALOGERROR: Final[ThirtyTwoBits] = cached_hresult(0x8004E023, 'CO_E_ACTIVATIONFAILED_CATALOGERROR')
    CO_E_ACTIVATIONFAILED_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x8004E024, 'CO_E_ACTIVATIONFAILED_TIMEOUT')
    CO_E_INITIALIZATIONFAILED: Final[ThirtyTwoBits] = cached_hresult(0x8004E025, 'CO_E_INITIALIZATIONFAILED')
    CONTEXT_E_NOJIT: Final[ThirtyTwoBits] = cached_hresult(0x8004E026, 'CONTEXT_E_NOJIT')
    CONTEXT_E_NOTRANSACTION: Final[ThirtyTwoBits] = cached_hresult(0x8004E027, 'CONTEXT_E_NOTRANSACTION')
    CO_E_THREADINGMODEL_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0x8004E028, 'CO_E_THREADINGMODEL_CHANGED')
    CO_E_NOIISINTRINSICS: Final[ThirtyTwoBits] = cached_hresult(0x8004E029, 'CO_E_NOIISINTRINSICS')
    CO_E_NOCOOKIES: Final[ThirtyTwoBits] = cached_hresult(0x8004E02A, 'CO_E_NOCOOKIES')
    CO_E_DBERROR: Final[ThirtyTwoBits] = cached_hresult(0x8004E02B, 'CO_E_DBERROR')
    CO_E_NOTPOOLED: Final[ThirtyTwoBits] = cached_hresult(0x8004E02C, 'CO_E_NOTPOOLED')
    CO_E_NOTCONSTRUCTED: Final[ThirtyTwoBits] = cached_hresult(0x8004E02D, 'CO_E_NOTCONSTRUCTED')
    CO_E_NOSYNCHRONIZATION: Final[ThirtyTwoBits] = cached_hresult(0x8004E02E, 'CO_E_NOSYNCHRONIZATION')
    CO_E_ISOLEVELMISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x8004E02F, 'CO_E_ISOLEVELMISMATCH')
    CO_E_CALL_OUT_OF_TX_SCOPE_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x8004E030, 'CO_E_CALL_OUT_OF_TX_SCOPE_NOT_ALLOWED')
    CO_E_EXIT_TRANSACTION_SCOPE_NOT_CALLED: Final[ThirtyTwoBits] = cached_hresult(0x8004E031, 'CO_E_EXIT_TRANSACTION_SCOPE_NOT_CALLED')
    E_ACCESSDENIED: Final[ThirtyTwoBits] = cached_hresult(0x80070005, 'E_ACCESSDENIED')
    E_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80070006, 'E_HANDLE')
    E_OUTOFMEMORY: Final[ThirtyTwoBits] = cached_hresult(0x8007000E, 'E_OUTOFMEMORY')
    E_INVALIDARG: Final[ThirtyTwoBits] = cached_hresult(0x80070057, 'E_INVALIDARG')
    WININET_E_OUT_OF_HANDLES: Final[ThirtyTwoBits] = cached_hresult(0x80072EE1, 'WININET_E_OUT_OF_HANDLES')
    WININET_E_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x80072EE2, 'WININET_E_TIMEOUT')
    WININET_E_EXTENDED_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80072EE3, 'WININET_E_EXTENDED_ERROR')
    WININET_E_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80072EE4, 'WININET_E_INTERNAL_ERROR')
    WININET_E_INVALID_URL: Final[ThirtyTwoBits] = cached_hresult(0x80072EE5, 'WININET_E_INVALID_URL')
    WININET_E_UNRECOGNIZED_SCHEME: Final[ThirtyTwoBits] = cached_hresult(0x80072EE6, 'WININET_E_UNRECOGNIZED_SCHEME')
    WININET_E_NAME_NOT_RESOLVED: Final[ThirtyTwoBits] = cached_hresult(0x80072EE7, 'WININET_E_NAME_NOT_RESOLVED')
    WININET_E_PROTOCOL_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80072EE8, 'WININET_E_PROTOCOL_NOT_FOUND')
    WININET_E_INVALID_OPTION: Final[ThirtyTwoBits] = cached_hresult(0x80072EE9, 'WININET_E_INVALID_OPTION')
    WININET_E_BAD_OPTION_LENGTH: Final[ThirtyTwoBits] = cached_hresult(0x80072EEA, 'WININET_E_BAD_OPTION_LENGTH')
    WININET_E_OPTION_NOT_SETTABLE: Final[ThirtyTwoBits] = cached_hresult(0x80072EEB, 'WININET_E_OPTION_NOT_SETTABLE')
    WININET_E_SHUTDOWN: Final[ThirtyTwoBits] = cached_hresult(0x80072EEC, 'WININET_E_SHUTDOWN')
    WININET_E_INCORRECT_USER_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80072EED, 'WININET_E_INCORRECT_USER_NAME')
    WININET_E_INCORRECT_PASSWORD: Final[ThirtyTwoBits] = cached_hresult(0x80072EEE, 'WININET_E_INCORRECT_PASSWORD')
    WININET_E_LOGIN_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80072EEF, 'WININET_E_LOGIN_FAILURE')
    WININET_E_INVALID_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x80072EF0, 'WININET_E_INVALID_OPERATION')
    WININET_E_OPERATION_CANCELLED: Final[ThirtyTwoBits] = cached_hresult(0x80072EF1, 'WININET_E_OPERATION_CANCELLED')
    WININET_E_INCORRECT_HANDLE_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x80072EF2, 'WININET_E_INCORRECT_HANDLE_TYPE')
    WININET_E_INCORRECT_HANDLE_STATE: Final[ThirtyTwoBits] = cached_hresult(0x80072EF3, 'WININET_E_INCORRECT_HANDLE_STATE')
    WININET_E_NOT_PROXY_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0x80072EF4, 'WININET_E_NOT_PROXY_REQUEST')
    WININET_E_REGISTRY_VALUE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80072EF5, 'WININET_E_REGISTRY_VALUE_NOT_FOUND')
    WININET_E_BAD_REGISTRY_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0x80072EF6, 'WININET_E_BAD_REGISTRY_PARAMETER')
    WININET_E_NO_DIRECT_ACCESS: Final[ThirtyTwoBits] = cached_hresult(0x80072EF7, 'WININET_E_NO_DIRECT_ACCESS')
    WININET_E_NO_CONTEXT: Final[ThirtyTwoBits] = cached_hresult(0x80072EF8, 'WININET_E_NO_CONTEXT')
    WININET_E_NO_CALLBACK: Final[ThirtyTwoBits] = cached_hresult(0x80072EF9, 'WININET_E_NO_CALLBACK')
    WININET_E_REQUEST_PENDING: Final[ThirtyTwoBits] = cached_hresult(0x80072EFA, 'WININET_E_REQUEST_PENDING')
    WININET_E_INCORRECT_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0x80072EFB, 'WININET_E_INCORRECT_FORMAT')
    WININET_E_ITEM_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80072EFC, 'WININET_E_ITEM_NOT_FOUND')
    WININET_E_CANNOT_CONNECT: Final[ThirtyTwoBits] = cached_hresult(0x80072EFD, 'WININET_E_CANNOT_CONNECT')
    WININET_E_CONNECTION_ABORTED: Final[ThirtyTwoBits] = cached_hresult(0x80072EFE, 'WININET_E_CONNECTION_ABORTED')
    WININET_E_CONNECTION_RESET: Final[ThirtyTwoBits] = cached_hresult(0x80072EFF, 'WININET_E_CONNECTION_RESET')
    WININET_E_FORCE_RETRY: Final[ThirtyTwoBits] = cached_hresult(0x80072F00, 'WININET_E_FORCE_RETRY')
    WININET_E_INVALID_PROXY_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0x80072F01, 'WININET_E_INVALID_PROXY_REQUEST')
    WININET_E_NEED_UI: Final[ThirtyTwoBits] = cached_hresult(0x80072F02, 'WININET_E_NEED_UI')
    WININET_E_HANDLE_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80072F04, 'WININET_E_HANDLE_EXISTS')
    WININET_E_SEC_CERT_DATE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80072F05, 'WININET_E_SEC_CERT_DATE_INVALID')
    WININET_E_SEC_CERT_CN_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80072F06, 'WININET_E_SEC_CERT_CN_INVALID')
    WININET_E_HTTP_TO_HTTPS_ON_REDIR: Final[ThirtyTwoBits] = cached_hresult(0x80072F07, 'WININET_E_HTTP_TO_HTTPS_ON_REDIR')
    WININET_E_HTTPS_TO_HTTP_ON_REDIR: Final[ThirtyTwoBits] = cached_hresult(0x80072F08, 'WININET_E_HTTPS_TO_HTTP_ON_REDIR')
    WININET_E_MIXED_SECURITY: Final[ThirtyTwoBits] = cached_hresult(0x80072F09, 'WININET_E_MIXED_SECURITY')
    WININET_E_CHG_POST_IS_NON_SECURE: Final[ThirtyTwoBits] = cached_hresult(0x80072F0A, 'WININET_E_CHG_POST_IS_NON_SECURE')
    WININET_E_POST_IS_NON_SECURE: Final[ThirtyTwoBits] = cached_hresult(0x80072F0B, 'WININET_E_POST_IS_NON_SECURE')
    WININET_E_CLIENT_AUTH_CERT_NEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80072F0C, 'WININET_E_CLIENT_AUTH_CERT_NEEDED')
    WININET_E_INVALID_CA: Final[ThirtyTwoBits] = cached_hresult(0x80072F0D, 'WININET_E_INVALID_CA')
    WININET_E_CLIENT_AUTH_NOT_SETUP: Final[ThirtyTwoBits] = cached_hresult(0x80072F0E, 'WININET_E_CLIENT_AUTH_NOT_SETUP')
    WININET_E_ASYNC_THREAD_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80072F0F, 'WININET_E_ASYNC_THREAD_FAILED')
    WININET_E_REDIRECT_SCHEME_CHANGE: Final[ThirtyTwoBits] = cached_hresult(0x80072F10, 'WININET_E_REDIRECT_SCHEME_CHANGE')
    WININET_E_DIALOG_PENDING: Final[ThirtyTwoBits] = cached_hresult(0x80072F11, 'WININET_E_DIALOG_PENDING')
    WININET_E_RETRY_DIALOG: Final[ThirtyTwoBits] = cached_hresult(0x80072F12, 'WININET_E_RETRY_DIALOG')
    WININET_E_NO_NEW_CONTAINERS: Final[ThirtyTwoBits] = cached_hresult(0x80072F13, 'WININET_E_NO_NEW_CONTAINERS')
    WININET_E_HTTPS_HTTP_SUBMIT_REDIR: Final[ThirtyTwoBits] = cached_hresult(0x80072F14, 'WININET_E_HTTPS_HTTP_SUBMIT_REDIR')
    WININET_E_SEC_CERT_ERRORS: Final[ThirtyTwoBits] = cached_hresult(0x80072F17, 'WININET_E_SEC_CERT_ERRORS')
    WININET_E_SEC_CERT_REV_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80072F19, 'WININET_E_SEC_CERT_REV_FAILED')
    WININET_E_HEADER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80072F76, 'WININET_E_HEADER_NOT_FOUND')
    WININET_E_DOWNLEVEL_SERVER: Final[ThirtyTwoBits] = cached_hresult(0x80072F77, 'WININET_E_DOWNLEVEL_SERVER')
    WININET_E_INVALID_SERVER_RESPONSE: Final[ThirtyTwoBits] = cached_hresult(0x80072F78, 'WININET_E_INVALID_SERVER_RESPONSE')
    WININET_E_INVALID_HEADER: Final[ThirtyTwoBits] = cached_hresult(0x80072F79, 'WININET_E_INVALID_HEADER')
    WININET_E_INVALID_QUERY_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0x80072F7A, 'WININET_E_INVALID_QUERY_REQUEST')
    WININET_E_HEADER_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80072F7B, 'WININET_E_HEADER_ALREADY_EXISTS')
    WININET_E_REDIRECT_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80072F7C, 'WININET_E_REDIRECT_FAILED')
    WININET_E_SECURITY_CHANNEL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80072F7D, 'WININET_E_SECURITY_CHANNEL_ERROR')
    WININET_E_UNABLE_TO_CACHE_FILE: Final[ThirtyTwoBits] = cached_hresult(0x80072F7E, 'WININET_E_UNABLE_TO_CACHE_FILE')
    WININET_E_TCPIP_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x80072F7F, 'WININET_E_TCPIP_NOT_INSTALLED')
    WININET_E_NOT_REDIRECTED: Final[ThirtyTwoBits] = cached_hresult(0x80072F80, 'WININET_E_NOT_REDIRECTED')
    WININET_E_COOKIE_NEEDS_CONFIRMATION: Final[ThirtyTwoBits] = cached_hresult(0x80072F81, 'WININET_E_COOKIE_NEEDS_CONFIRMATION')
    WININET_E_COOKIE_DECLINED: Final[ThirtyTwoBits] = cached_hresult(0x80072F82, 'WININET_E_COOKIE_DECLINED')
    WININET_E_DISCONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x80072F83, 'WININET_E_DISCONNECTED')
    WININET_E_SERVER_UNREACHABLE: Final[ThirtyTwoBits] = cached_hresult(0x80072F84, 'WININET_E_SERVER_UNREACHABLE')
    WININET_E_PROXY_SERVER_UNREACHABLE: Final[ThirtyTwoBits] = cached_hresult(0x80072F85, 'WININET_E_PROXY_SERVER_UNREACHABLE')
    WININET_E_BAD_AUTO_PROXY_SCRIPT: Final[ThirtyTwoBits] = cached_hresult(0x80072F86, 'WININET_E_BAD_AUTO_PROXY_SCRIPT')
    WININET_E_UNABLE_TO_DOWNLOAD_SCRIPT: Final[ThirtyTwoBits] = cached_hresult(0x80072F87, 'WININET_E_UNABLE_TO_DOWNLOAD_SCRIPT')
    WININET_E_REDIRECT_NEEDS_CONFIRMATION: Final[ThirtyTwoBits] = cached_hresult(0x80072F88, 'WININET_E_REDIRECT_NEEDS_CONFIRMATION')
    WININET_E_SEC_INVALID_CERT: Final[ThirtyTwoBits] = cached_hresult(0x80072F89, 'WININET_E_SEC_INVALID_CERT')
    WININET_E_SEC_CERT_REVOKED: Final[ThirtyTwoBits] = cached_hresult(0x80072F8A, 'WININET_E_SEC_CERT_REVOKED')
    WININET_E_FAILED_DUETOSECURITYCHECK: Final[ThirtyTwoBits] = cached_hresult(0x80072F8B, 'WININET_E_FAILED_DUETOSECURITYCHECK')
    WININET_E_NOT_INITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0x80072F8C, 'WININET_E_NOT_INITIALIZED')
    WININET_E_LOGIN_FAILURE_DISPLAY_ENTITY_BODY: Final[ThirtyTwoBits] = cached_hresult(0x80072F8E, 'WININET_E_LOGIN_FAILURE_DISPLAY_ENTITY_BODY')
    WININET_E_DECODING_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80072F8F, 'WININET_E_DECODING_FAILED')
    CO_E_CLASS_CREATE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80080001, 'CO_E_CLASS_CREATE_FAILED')
    CO_E_SCM_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80080002, 'CO_E_SCM_ERROR')
    CO_E_SCM_RPC_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80080003, 'CO_E_SCM_RPC_FAILURE')
    CO_E_BAD_PATH: Final[ThirtyTwoBits] = cached_hresult(0x80080004, 'CO_E_BAD_PATH')
    CO_E_SERVER_EXEC_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80080005, 'CO_E_SERVER_EXEC_FAILURE')
    CO_E_OBJSRV_RPC_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80080006, 'CO_E_OBJSRV_RPC_FAILURE')
    MK_E_NO_NORMALIZED: Final[ThirtyTwoBits] = cached_hresult(0x80080007, 'MK_E_NO_NORMALIZED')
    CO_E_SERVER_STOPPING: Final[ThirtyTwoBits] = cached_hresult(0x80080008, 'CO_E_SERVER_STOPPING')
    MEM_E_INVALID_ROOT: Final[ThirtyTwoBits] = cached_hresult(0x80080009, 'MEM_E_INVALID_ROOT')
    MEM_E_INVALID_LINK: Final[ThirtyTwoBits] = cached_hresult(0x80080010, 'MEM_E_INVALID_LINK')
    MEM_E_INVALID_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x80080011, 'MEM_E_INVALID_SIZE')
    CO_S_NOTALLINTERFACES: Final[ThirtyTwoBits] = cached_hresult(0x00080012, 'CO_S_NOTALLINTERFACES')
    CO_S_MACHINENAMENOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x00080013, 'CO_S_MACHINENAMENOTFOUND')
    CO_E_MISSING_DISPLAYNAME: Final[ThirtyTwoBits] = cached_hresult(0x80080015, 'CO_E_MISSING_DISPLAYNAME')
    CO_E_RUNAS_VALUE_MUST_BE_AAA: Final[ThirtyTwoBits] = cached_hresult(0x80080016, 'CO_E_RUNAS_VALUE_MUST_BE_AAA')
    CO_E_ELEVATION_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80080017, 'CO_E_ELEVATION_DISABLED')
    APPX_E_PACKAGING_INTERNAL: Final[ThirtyTwoBits] = cached_hresult(0x80080200, 'APPX_E_PACKAGING_INTERNAL')
    APPX_E_INTERLEAVING_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80080201, 'APPX_E_INTERLEAVING_NOT_ALLOWED')
    APPX_E_RELATIONSHIPS_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80080202, 'APPX_E_RELATIONSHIPS_NOT_ALLOWED')
    APPX_E_MISSING_REQUIRED_FILE: Final[ThirtyTwoBits] = cached_hresult(0x80080203, 'APPX_E_MISSING_REQUIRED_FILE')
    APPX_E_INVALID_MANIFEST: Final[ThirtyTwoBits] = cached_hresult(0x80080204, 'APPX_E_INVALID_MANIFEST')
    APPX_E_INVALID_BLOCKMAP: Final[ThirtyTwoBits] = cached_hresult(0x80080205, 'APPX_E_INVALID_BLOCKMAP')
    APPX_E_CORRUPT_CONTENT: Final[ThirtyTwoBits] = cached_hresult(0x80080206, 'APPX_E_CORRUPT_CONTENT')
    APPX_E_BLOCK_HASH_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80080207, 'APPX_E_BLOCK_HASH_INVALID')
    APPX_E_REQUESTED_RANGE_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0x80080208, 'APPX_E_REQUESTED_RANGE_TOO_LARGE')
    APPX_E_INVALID_SIP_CLIENT_DATA: Final[ThirtyTwoBits] = cached_hresult(0x80080209, 'APPX_E_INVALID_SIP_CLIENT_DATA')
    APPX_E_INVALID_KEY_INFO: Final[ThirtyTwoBits] = cached_hresult(0x8008020A, 'APPX_E_INVALID_KEY_INFO')
    APPX_E_INVALID_CONTENTGROUPMAP: Final[ThirtyTwoBits] = cached_hresult(0x8008020B, 'APPX_E_INVALID_CONTENTGROUPMAP')
    APPX_E_INVALID_APPINSTALLER: Final[ThirtyTwoBits] = cached_hresult(0x8008020C, 'APPX_E_INVALID_APPINSTALLER')
    APPX_E_DELTA_BASELINE_VERSION_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x8008020D, 'APPX_E_DELTA_BASELINE_VERSION_MISMATCH')
    APPX_E_DELTA_PACKAGE_MISSING_FILE: Final[ThirtyTwoBits] = cached_hresult(0x8008020E, 'APPX_E_DELTA_PACKAGE_MISSING_FILE')
    APPX_E_INVALID_DELTA_PACKAGE: Final[ThirtyTwoBits] = cached_hresult(0x8008020F, 'APPX_E_INVALID_DELTA_PACKAGE')
    APPX_E_DELTA_APPENDED_PACKAGE_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80080210, 'APPX_E_DELTA_APPENDED_PACKAGE_NOT_ALLOWED')
    APPX_E_INVALID_PACKAGING_LAYOUT: Final[ThirtyTwoBits] = cached_hresult(0x80080211, 'APPX_E_INVALID_PACKAGING_LAYOUT')
    APPX_E_INVALID_PACKAGESIGNCONFIG: Final[ThirtyTwoBits] = cached_hresult(0x80080212, 'APPX_E_INVALID_PACKAGESIGNCONFIG')
    APPX_E_RESOURCESPRI_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80080213, 'APPX_E_RESOURCESPRI_NOT_ALLOWED')
    APPX_E_FILE_COMPRESSION_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80080214, 'APPX_E_FILE_COMPRESSION_MISMATCH')
    APPX_E_INVALID_PAYLOAD_PACKAGE_EXTENSION: Final[ThirtyTwoBits] = cached_hresult(0x80080215, 'APPX_E_INVALID_PAYLOAD_PACKAGE_EXTENSION')
    APPX_E_INVALID_ENCRYPTION_EXCLUSION_FILE_LIST: Final[ThirtyTwoBits] = cached_hresult(0x80080216, 'APPX_E_INVALID_ENCRYPTION_EXCLUSION_FILE_LIST')
    APPX_E_INVALID_PACKAGE_FOLDER_ACLS: Final[ThirtyTwoBits] = cached_hresult(0x80080217, 'APPX_E_INVALID_PACKAGE_FOLDER_ACLS')
    APPX_E_INVALID_PUBLISHER_BRIDGING: Final[ThirtyTwoBits] = cached_hresult(0x80080218, 'APPX_E_INVALID_PUBLISHER_BRIDGING')
    APPX_E_DIGEST_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80080219, 'APPX_E_DIGEST_MISMATCH')
    BT_E_SPURIOUS_ACTIVATION: Final[ThirtyTwoBits] = cached_hresult(0x80080300, 'BT_E_SPURIOUS_ACTIVATION')
    NTE_BAD_UID: Final[ThirtyTwoBits] = cached_hresult(0x80090001, 'NTE_BAD_UID')
    ERROR_AUDITING_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xC0090001, 'ERROR_AUDITING_DISABLED')
    NTE_BAD_HASH: Final[ThirtyTwoBits] = cached_hresult(0x80090002, 'NTE_BAD_HASH')
    ERROR_ALL_SIDS_FILTERED: Final[ThirtyTwoBits] = cached_hresult(0xC0090002, 'ERROR_ALL_SIDS_FILTERED')
    NTE_BAD_KEY: Final[ThirtyTwoBits] = cached_hresult(0x80090003, 'NTE_BAD_KEY')
    ERROR_BIZRULES_NOT_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0xC0090003, 'ERROR_BIZRULES_NOT_ENABLED')
    NTE_BAD_LEN: Final[ThirtyTwoBits] = cached_hresult(0x80090004, 'NTE_BAD_LEN')
    NTE_BAD_DATA: Final[ThirtyTwoBits] = cached_hresult(0x80090005, 'NTE_BAD_DATA')
    NTE_BAD_SIGNATURE: Final[ThirtyTwoBits] = cached_hresult(0x80090006, 'NTE_BAD_SIGNATURE')
    NTE_BAD_VER: Final[ThirtyTwoBits] = cached_hresult(0x80090007, 'NTE_BAD_VER')
    NTE_BAD_ALGID: Final[ThirtyTwoBits] = cached_hresult(0x80090008, 'NTE_BAD_ALGID')
    NTE_BAD_FLAGS: Final[ThirtyTwoBits] = cached_hresult(0x80090009, 'NTE_BAD_FLAGS')
    NTE_BAD_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x8009000A, 'NTE_BAD_TYPE')
    NTE_BAD_KEY_STATE: Final[ThirtyTwoBits] = cached_hresult(0x8009000B, 'NTE_BAD_KEY_STATE')
    NTE_BAD_HASH_STATE: Final[ThirtyTwoBits] = cached_hresult(0x8009000C, 'NTE_BAD_HASH_STATE')
    NTE_NO_KEY: Final[ThirtyTwoBits] = cached_hresult(0x8009000D, 'NTE_NO_KEY')
    NTE_NO_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0x8009000E, 'NTE_NO_MEMORY')
    NTE_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x8009000F, 'NTE_EXISTS')
    NTE_PERM: Final[ThirtyTwoBits] = cached_hresult(0x80090010, 'NTE_PERM')
    NTE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80090011, 'NTE_NOT_FOUND')
    NTE_DOUBLE_ENCRYPT: Final[ThirtyTwoBits] = cached_hresult(0x80090012, 'NTE_DOUBLE_ENCRYPT')
    NTE_BAD_PROVIDER: Final[ThirtyTwoBits] = cached_hresult(0x80090013, 'NTE_BAD_PROVIDER')
    NTE_BAD_PROV_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x80090014, 'NTE_BAD_PROV_TYPE')
    NTE_BAD_PUBLIC_KEY: Final[ThirtyTwoBits] = cached_hresult(0x80090015, 'NTE_BAD_PUBLIC_KEY')
    NTE_BAD_KEYSET: Final[ThirtyTwoBits] = cached_hresult(0x80090016, 'NTE_BAD_KEYSET')
    NTE_PROV_TYPE_NOT_DEF: Final[ThirtyTwoBits] = cached_hresult(0x80090017, 'NTE_PROV_TYPE_NOT_DEF')
    NTE_PROV_TYPE_ENTRY_BAD: Final[ThirtyTwoBits] = cached_hresult(0x80090018, 'NTE_PROV_TYPE_ENTRY_BAD')
    NTE_KEYSET_NOT_DEF: Final[ThirtyTwoBits] = cached_hresult(0x80090019, 'NTE_KEYSET_NOT_DEF')
    NTE_KEYSET_ENTRY_BAD: Final[ThirtyTwoBits] = cached_hresult(0x8009001A, 'NTE_KEYSET_ENTRY_BAD')
    NTE_PROV_TYPE_NO_MATCH: Final[ThirtyTwoBits] = cached_hresult(0x8009001B, 'NTE_PROV_TYPE_NO_MATCH')
    NTE_SIGNATURE_FILE_BAD: Final[ThirtyTwoBits] = cached_hresult(0x8009001C, 'NTE_SIGNATURE_FILE_BAD')
    NTE_PROVIDER_DLL_FAIL: Final[ThirtyTwoBits] = cached_hresult(0x8009001D, 'NTE_PROVIDER_DLL_FAIL')
    NTE_PROV_DLL_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8009001E, 'NTE_PROV_DLL_NOT_FOUND')
    NTE_BAD_KEYSET_PARAM: Final[ThirtyTwoBits] = cached_hresult(0x8009001F, 'NTE_BAD_KEYSET_PARAM')
    NTE_FAIL: Final[ThirtyTwoBits] = cached_hresult(0x80090020, 'NTE_FAIL')
    NTE_SYS_ERR: Final[ThirtyTwoBits] = cached_hresult(0x80090021, 'NTE_SYS_ERR')
    NTE_SILENT_CONTEXT: Final[ThirtyTwoBits] = cached_hresult(0x80090022, 'NTE_SILENT_CONTEXT')
    NTE_TOKEN_KEYSET_STORAGE_FULL: Final[ThirtyTwoBits] = cached_hresult(0x80090023, 'NTE_TOKEN_KEYSET_STORAGE_FULL')
    NTE_TEMPORARY_PROFILE: Final[ThirtyTwoBits] = cached_hresult(0x80090024, 'NTE_TEMPORARY_PROFILE')
    NTE_FIXEDPARAMETER: Final[ThirtyTwoBits] = cached_hresult(0x80090025, 'NTE_FIXEDPARAMETER')
    NTE_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80090026, 'NTE_INVALID_HANDLE')
    NTE_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0x80090027, 'NTE_INVALID_PARAMETER')
    NTE_BUFFER_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0x80090028, 'NTE_BUFFER_TOO_SMALL')
    NTE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80090029, 'NTE_NOT_SUPPORTED')
    NTE_NO_MORE_ITEMS: Final[ThirtyTwoBits] = cached_hresult(0x8009002A, 'NTE_NO_MORE_ITEMS')
    NTE_BUFFERS_OVERLAP: Final[ThirtyTwoBits] = cached_hresult(0x8009002B, 'NTE_BUFFERS_OVERLAP')
    NTE_DECRYPTION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8009002C, 'NTE_DECRYPTION_FAILURE')
    NTE_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8009002D, 'NTE_INTERNAL_ERROR')
    NTE_UI_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8009002E, 'NTE_UI_REQUIRED')
    NTE_HMAC_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8009002F, 'NTE_HMAC_NOT_SUPPORTED')
    NTE_DEVICE_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0x80090030, 'NTE_DEVICE_NOT_READY')
    NTE_AUTHENTICATION_IGNORED: Final[ThirtyTwoBits] = cached_hresult(0x80090031, 'NTE_AUTHENTICATION_IGNORED')
    NTE_VALIDATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80090032, 'NTE_VALIDATION_FAILED')
    NTE_INCORRECT_PASSWORD: Final[ThirtyTwoBits] = cached_hresult(0x80090033, 'NTE_INCORRECT_PASSWORD')
    NTE_ENCRYPTION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80090034, 'NTE_ENCRYPTION_FAILURE')
    NTE_DEVICE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80090035, 'NTE_DEVICE_NOT_FOUND')
    NTE_USER_CANCELLED: Final[ThirtyTwoBits] = cached_hresult(0x80090036, 'NTE_USER_CANCELLED')
    NTE_PASSWORD_CHANGE_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80090037, 'NTE_PASSWORD_CHANGE_REQUIRED')
    NTE_NOT_ACTIVE_CONSOLE: Final[ThirtyTwoBits] = cached_hresult(0x80090038, 'NTE_NOT_ACTIVE_CONSOLE')
    SEC_E_INSUFFICIENT_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0x80090300, 'SEC_E_INSUFFICIENT_MEMORY')
    SEC_E_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80090301, 'SEC_E_INVALID_HANDLE')
    SEC_E_UNSUPPORTED_FUNCTION: Final[ThirtyTwoBits] = cached_hresult(0x80090302, 'SEC_E_UNSUPPORTED_FUNCTION')
    SEC_E_TARGET_UNKNOWN: Final[ThirtyTwoBits] = cached_hresult(0x80090303, 'SEC_E_TARGET_UNKNOWN')
    SEC_E_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80090304, 'SEC_E_INTERNAL_ERROR')
    SEC_E_SECPKG_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80090305, 'SEC_E_SECPKG_NOT_FOUND')
    SEC_E_NOT_OWNER: Final[ThirtyTwoBits] = cached_hresult(0x80090306, 'SEC_E_NOT_OWNER')
    SEC_E_CANNOT_INSTALL: Final[ThirtyTwoBits] = cached_hresult(0x80090307, 'SEC_E_CANNOT_INSTALL')
    SEC_E_INVALID_TOKEN: Final[ThirtyTwoBits] = cached_hresult(0x80090308, 'SEC_E_INVALID_TOKEN')
    SEC_E_CANNOT_PACK: Final[ThirtyTwoBits] = cached_hresult(0x80090309, 'SEC_E_CANNOT_PACK')
    SEC_E_QOP_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8009030A, 'SEC_E_QOP_NOT_SUPPORTED')
    SEC_E_NO_IMPERSONATION: Final[ThirtyTwoBits] = cached_hresult(0x8009030B, 'SEC_E_NO_IMPERSONATION')
    SEC_E_LOGON_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x8009030C, 'SEC_E_LOGON_DENIED')
    SEC_E_UNKNOWN_CREDENTIALS: Final[ThirtyTwoBits] = cached_hresult(0x8009030D, 'SEC_E_UNKNOWN_CREDENTIALS')
    SEC_E_NO_CREDENTIALS: Final[ThirtyTwoBits] = cached_hresult(0x8009030E, 'SEC_E_NO_CREDENTIALS')
    SEC_E_MESSAGE_ALTERED: Final[ThirtyTwoBits] = cached_hresult(0x8009030F, 'SEC_E_MESSAGE_ALTERED')
    SEC_E_OUT_OF_SEQUENCE: Final[ThirtyTwoBits] = cached_hresult(0x80090310, 'SEC_E_OUT_OF_SEQUENCE')
    SEC_E_NO_AUTHENTICATING_AUTHORITY: Final[ThirtyTwoBits] = cached_hresult(0x80090311, 'SEC_E_NO_AUTHENTICATING_AUTHORITY')
    SEC_I_CONTINUE_NEEDED: Final[ThirtyTwoBits] = cached_hresult(0x00090312, 'SEC_I_CONTINUE_NEEDED')
    SEC_I_COMPLETE_NEEDED: Final[ThirtyTwoBits] = cached_hresult(0x00090313, 'SEC_I_COMPLETE_NEEDED')
    SEC_I_COMPLETE_AND_CONTINUE: Final[ThirtyTwoBits] = cached_hresult(0x00090314, 'SEC_I_COMPLETE_AND_CONTINUE')
    SEC_I_LOCAL_LOGON: Final[ThirtyTwoBits] = cached_hresult(0x00090315, 'SEC_I_LOCAL_LOGON')
    SEC_I_GENERIC_EXTENSION_RECEIVED: Final[ThirtyTwoBits] = cached_hresult(0x00090316, 'SEC_I_GENERIC_EXTENSION_RECEIVED')
    SEC_E_BAD_PKGID: Final[ThirtyTwoBits] = cached_hresult(0x80090316, 'SEC_E_BAD_PKGID')
    SEC_I_CONTEXT_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x00090317, 'SEC_I_CONTEXT_EXPIRED')
    SEC_E_CONTEXT_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x80090317, 'SEC_E_CONTEXT_EXPIRED')
    SEC_E_INCOMPLETE_MESSAGE: Final[ThirtyTwoBits] = cached_hresult(0x80090318, 'SEC_E_INCOMPLETE_MESSAGE')
    SEC_I_INCOMPLETE_CREDENTIALS: Final[ThirtyTwoBits] = cached_hresult(0x00090320, 'SEC_I_INCOMPLETE_CREDENTIALS')
    SEC_E_INCOMPLETE_CREDENTIALS: Final[ThirtyTwoBits] = cached_hresult(0x80090320, 'SEC_E_INCOMPLETE_CREDENTIALS')
    SEC_I_RENEGOTIATE: Final[ThirtyTwoBits] = cached_hresult(0x00090321, 'SEC_I_RENEGOTIATE')
    SEC_E_BUFFER_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0x80090321, 'SEC_E_BUFFER_TOO_SMALL')
    SEC_E_WRONG_PRINCIPAL: Final[ThirtyTwoBits] = cached_hresult(0x80090322, 'SEC_E_WRONG_PRINCIPAL')
    SEC_I_NO_LSA_CONTEXT: Final[ThirtyTwoBits] = cached_hresult(0x00090323, 'SEC_I_NO_LSA_CONTEXT')
    SEC_E_TIME_SKEW: Final[ThirtyTwoBits] = cached_hresult(0x80090324, 'SEC_E_TIME_SKEW')
    SEC_E_UNTRUSTED_ROOT: Final[ThirtyTwoBits] = cached_hresult(0x80090325, 'SEC_E_UNTRUSTED_ROOT')
    SEC_E_ILLEGAL_MESSAGE: Final[ThirtyTwoBits] = cached_hresult(0x80090326, 'SEC_E_ILLEGAL_MESSAGE')
    SEC_E_CERT_UNKNOWN: Final[ThirtyTwoBits] = cached_hresult(0x80090327, 'SEC_E_CERT_UNKNOWN')
    SEC_E_CERT_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x80090328, 'SEC_E_CERT_EXPIRED')
    SEC_E_ENCRYPT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80090329, 'SEC_E_ENCRYPT_FAILURE')
    SEC_E_DECRYPT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80090330, 'SEC_E_DECRYPT_FAILURE')
    SEC_E_ALGORITHM_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80090331, 'SEC_E_ALGORITHM_MISMATCH')
    SEC_E_SECURITY_QOS_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80090332, 'SEC_E_SECURITY_QOS_FAILED')
    SEC_E_UNFINISHED_CONTEXT_DELETED: Final[ThirtyTwoBits] = cached_hresult(0x80090333, 'SEC_E_UNFINISHED_CONTEXT_DELETED')
    SEC_E_NO_TGT_REPLY: Final[ThirtyTwoBits] = cached_hresult(0x80090334, 'SEC_E_NO_TGT_REPLY')
    SEC_E_NO_IP_ADDRESSES: Final[ThirtyTwoBits] = cached_hresult(0x80090335, 'SEC_E_NO_IP_ADDRESSES')
    SEC_E_WRONG_CREDENTIAL_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80090336, 'SEC_E_WRONG_CREDENTIAL_HANDLE')
    SEC_E_CRYPTO_SYSTEM_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80090337, 'SEC_E_CRYPTO_SYSTEM_INVALID')
    SEC_E_MAX_REFERRALS_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80090338, 'SEC_E_MAX_REFERRALS_EXCEEDED')
    SEC_E_MUST_BE_KDC: Final[ThirtyTwoBits] = cached_hresult(0x80090339, 'SEC_E_MUST_BE_KDC')
    SEC_E_STRONG_CRYPTO_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8009033A, 'SEC_E_STRONG_CRYPTO_NOT_SUPPORTED')
    SEC_E_TOO_MANY_PRINCIPALS: Final[ThirtyTwoBits] = cached_hresult(0x8009033B, 'SEC_E_TOO_MANY_PRINCIPALS')
    SEC_E_NO_PA_DATA: Final[ThirtyTwoBits] = cached_hresult(0x8009033C, 'SEC_E_NO_PA_DATA')
    SEC_E_PKINIT_NAME_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x8009033D, 'SEC_E_PKINIT_NAME_MISMATCH')
    SEC_E_SMARTCARD_LOGON_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8009033E, 'SEC_E_SMARTCARD_LOGON_REQUIRED')
    SEC_E_SHUTDOWN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x8009033F, 'SEC_E_SHUTDOWN_IN_PROGRESS')
    SEC_E_KDC_INVALID_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0x80090340, 'SEC_E_KDC_INVALID_REQUEST')
    SEC_E_KDC_UNABLE_TO_REFER: Final[ThirtyTwoBits] = cached_hresult(0x80090341, 'SEC_E_KDC_UNABLE_TO_REFER')
    SEC_E_KDC_UNKNOWN_ETYPE: Final[ThirtyTwoBits] = cached_hresult(0x80090342, 'SEC_E_KDC_UNKNOWN_ETYPE')
    SEC_E_UNSUPPORTED_PREAUTH: Final[ThirtyTwoBits] = cached_hresult(0x80090343, 'SEC_E_UNSUPPORTED_PREAUTH')
    SEC_E_DELEGATION_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80090345, 'SEC_E_DELEGATION_REQUIRED')
    SEC_E_BAD_BINDINGS: Final[ThirtyTwoBits] = cached_hresult(0x80090346, 'SEC_E_BAD_BINDINGS')
    SEC_E_MULTIPLE_ACCOUNTS: Final[ThirtyTwoBits] = cached_hresult(0x80090347, 'SEC_E_MULTIPLE_ACCOUNTS')
    SEC_E_NO_KERB_KEY: Final[ThirtyTwoBits] = cached_hresult(0x80090348, 'SEC_E_NO_KERB_KEY')
    SEC_E_CERT_WRONG_USAGE: Final[ThirtyTwoBits] = cached_hresult(0x80090349, 'SEC_E_CERT_WRONG_USAGE')
    SEC_E_DOWNGRADE_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0x80090350, 'SEC_E_DOWNGRADE_DETECTED')
    SEC_E_SMARTCARD_CERT_REVOKED: Final[ThirtyTwoBits] = cached_hresult(0x80090351, 'SEC_E_SMARTCARD_CERT_REVOKED')
    SEC_E_ISSUING_CA_UNTRUSTED: Final[ThirtyTwoBits] = cached_hresult(0x80090352, 'SEC_E_ISSUING_CA_UNTRUSTED')
    SEC_E_REVOCATION_OFFLINE_C: Final[ThirtyTwoBits] = cached_hresult(0x80090353, 'SEC_E_REVOCATION_OFFLINE_C')
    SEC_E_PKINIT_CLIENT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80090354, 'SEC_E_PKINIT_CLIENT_FAILURE')
    SEC_E_SMARTCARD_CERT_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x80090355, 'SEC_E_SMARTCARD_CERT_EXPIRED')
    SEC_E_NO_S4U_PROT_SUPPORT: Final[ThirtyTwoBits] = cached_hresult(0x80090356, 'SEC_E_NO_S4U_PROT_SUPPORT')
    SEC_E_CROSSREALM_DELEGATION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80090357, 'SEC_E_CROSSREALM_DELEGATION_FAILURE')
    SEC_E_REVOCATION_OFFLINE_KDC: Final[ThirtyTwoBits] = cached_hresult(0x80090358, 'SEC_E_REVOCATION_OFFLINE_KDC')
    SEC_E_ISSUING_CA_UNTRUSTED_KDC: Final[ThirtyTwoBits] = cached_hresult(0x80090359, 'SEC_E_ISSUING_CA_UNTRUSTED_KDC')
    SEC_E_KDC_CERT_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x8009035A, 'SEC_E_KDC_CERT_EXPIRED')
    SEC_E_KDC_CERT_REVOKED: Final[ThirtyTwoBits] = cached_hresult(0x8009035B, 'SEC_E_KDC_CERT_REVOKED')
    SEC_I_SIGNATURE_NEEDED: Final[ThirtyTwoBits] = cached_hresult(0x0009035C, 'SEC_I_SIGNATURE_NEEDED')
    SEC_E_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0x8009035D, 'SEC_E_INVALID_PARAMETER')
    SEC_E_DELEGATION_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x8009035E, 'SEC_E_DELEGATION_POLICY')
    SEC_E_POLICY_NLTM_ONLY: Final[ThirtyTwoBits] = cached_hresult(0x8009035F, 'SEC_E_POLICY_NLTM_ONLY')
    SEC_I_NO_RENEGOTIATION: Final[ThirtyTwoBits] = cached_hresult(0x00090360, 'SEC_I_NO_RENEGOTIATION')
    SEC_E_NO_CONTEXT: Final[ThirtyTwoBits] = cached_hresult(0x80090361, 'SEC_E_NO_CONTEXT')
    SEC_E_PKU2U_CERT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80090362, 'SEC_E_PKU2U_CERT_FAILURE')
    SEC_E_MUTUAL_AUTH_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80090363, 'SEC_E_MUTUAL_AUTH_FAILED')
    SEC_I_MESSAGE_FRAGMENT: Final[ThirtyTwoBits] = cached_hresult(0x00090364, 'SEC_I_MESSAGE_FRAGMENT')
    SEC_E_ONLY_HTTPS_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80090365, 'SEC_E_ONLY_HTTPS_ALLOWED')
    SEC_I_CONTINUE_NEEDED_MESSAGE_OK: Final[ThirtyTwoBits] = cached_hresult(0x00090366, 'SEC_I_CONTINUE_NEEDED_MESSAGE_OK')
    SEC_E_APPLICATION_PROTOCOL_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80090367, 'SEC_E_APPLICATION_PROTOCOL_MISMATCH')
    SEC_I_ASYNC_CALL_PENDING: Final[ThirtyTwoBits] = cached_hresult(0x00090368, 'SEC_I_ASYNC_CALL_PENDING')
    SEC_E_INVALID_UPN_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80090369, 'SEC_E_INVALID_UPN_NAME')
    SEC_E_EXT_BUFFER_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0x8009036A, 'SEC_E_EXT_BUFFER_TOO_SMALL')
    SEC_E_INSUFFICIENT_BUFFERS: Final[ThirtyTwoBits] = cached_hresult(0x8009036B, 'SEC_E_INSUFFICIENT_BUFFERS')
    CRYPT_E_MSG_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80091001, 'CRYPT_E_MSG_ERROR')
    CRYPT_E_UNKNOWN_ALGO: Final[ThirtyTwoBits] = cached_hresult(0x80091002, 'CRYPT_E_UNKNOWN_ALGO')
    CRYPT_E_OID_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0x80091003, 'CRYPT_E_OID_FORMAT')
    CRYPT_E_INVALID_MSG_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x80091004, 'CRYPT_E_INVALID_MSG_TYPE')
    CRYPT_E_UNEXPECTED_ENCODING: Final[ThirtyTwoBits] = cached_hresult(0x80091005, 'CRYPT_E_UNEXPECTED_ENCODING')
    CRYPT_E_AUTH_ATTR_MISSING: Final[ThirtyTwoBits] = cached_hresult(0x80091006, 'CRYPT_E_AUTH_ATTR_MISSING')
    CRYPT_E_HASH_VALUE: Final[ThirtyTwoBits] = cached_hresult(0x80091007, 'CRYPT_E_HASH_VALUE')
    CRYPT_E_INVALID_INDEX: Final[ThirtyTwoBits] = cached_hresult(0x80091008, 'CRYPT_E_INVALID_INDEX')
    CRYPT_E_ALREADY_DECRYPTED: Final[ThirtyTwoBits] = cached_hresult(0x80091009, 'CRYPT_E_ALREADY_DECRYPTED')
    CRYPT_E_NOT_DECRYPTED: Final[ThirtyTwoBits] = cached_hresult(0x8009100A, 'CRYPT_E_NOT_DECRYPTED')
    CRYPT_E_RECIPIENT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8009100B, 'CRYPT_E_RECIPIENT_NOT_FOUND')
    CRYPT_E_CONTROL_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x8009100C, 'CRYPT_E_CONTROL_TYPE')
    CRYPT_E_ISSUER_SERIALNUMBER: Final[ThirtyTwoBits] = cached_hresult(0x8009100D, 'CRYPT_E_ISSUER_SERIALNUMBER')
    CRYPT_E_SIGNER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8009100E, 'CRYPT_E_SIGNER_NOT_FOUND')
    CRYPT_E_ATTRIBUTES_MISSING: Final[ThirtyTwoBits] = cached_hresult(0x8009100F, 'CRYPT_E_ATTRIBUTES_MISSING')
    CRYPT_E_STREAM_MSG_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0x80091010, 'CRYPT_E_STREAM_MSG_NOT_READY')
    CRYPT_E_STREAM_INSUFFICIENT_DATA: Final[ThirtyTwoBits] = cached_hresult(0x80091011, 'CRYPT_E_STREAM_INSUFFICIENT_DATA')
    CRYPT_I_NEW_PROTECTION_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x00091012, 'CRYPT_I_NEW_PROTECTION_REQUIRED')
    CRYPT_E_BAD_LEN: Final[ThirtyTwoBits] = cached_hresult(0x80092001, 'CRYPT_E_BAD_LEN')
    CRYPT_E_BAD_ENCODE: Final[ThirtyTwoBits] = cached_hresult(0x80092002, 'CRYPT_E_BAD_ENCODE')
    CRYPT_E_FILE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80092003, 'CRYPT_E_FILE_ERROR')
    CRYPT_E_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80092004, 'CRYPT_E_NOT_FOUND')
    CRYPT_E_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80092005, 'CRYPT_E_EXISTS')
    CRYPT_E_NO_PROVIDER: Final[ThirtyTwoBits] = cached_hresult(0x80092006, 'CRYPT_E_NO_PROVIDER')
    CRYPT_E_SELF_SIGNED: Final[ThirtyTwoBits] = cached_hresult(0x80092007, 'CRYPT_E_SELF_SIGNED')
    CRYPT_E_DELETED_PREV: Final[ThirtyTwoBits] = cached_hresult(0x80092008, 'CRYPT_E_DELETED_PREV')
    CRYPT_E_NO_MATCH: Final[ThirtyTwoBits] = cached_hresult(0x80092009, 'CRYPT_E_NO_MATCH')
    CRYPT_E_UNEXPECTED_MSG_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x8009200A, 'CRYPT_E_UNEXPECTED_MSG_TYPE')
    CRYPT_E_NO_KEY_PROPERTY: Final[ThirtyTwoBits] = cached_hresult(0x8009200B, 'CRYPT_E_NO_KEY_PROPERTY')
    CRYPT_E_NO_DECRYPT_CERT: Final[ThirtyTwoBits] = cached_hresult(0x8009200C, 'CRYPT_E_NO_DECRYPT_CERT')
    CRYPT_E_BAD_MSG: Final[ThirtyTwoBits] = cached_hresult(0x8009200D, 'CRYPT_E_BAD_MSG')
    CRYPT_E_NO_SIGNER: Final[ThirtyTwoBits] = cached_hresult(0x8009200E, 'CRYPT_E_NO_SIGNER')
    CRYPT_E_PENDING_CLOSE: Final[ThirtyTwoBits] = cached_hresult(0x8009200F, 'CRYPT_E_PENDING_CLOSE')
    CRYPT_E_REVOKED: Final[ThirtyTwoBits] = cached_hresult(0x80092010, 'CRYPT_E_REVOKED')
    CRYPT_E_NO_REVOCATION_DLL: Final[ThirtyTwoBits] = cached_hresult(0x80092011, 'CRYPT_E_NO_REVOCATION_DLL')
    CRYPT_E_NO_REVOCATION_CHECK: Final[ThirtyTwoBits] = cached_hresult(0x80092012, 'CRYPT_E_NO_REVOCATION_CHECK')
    CRYPT_E_REVOCATION_OFFLINE: Final[ThirtyTwoBits] = cached_hresult(0x80092013, 'CRYPT_E_REVOCATION_OFFLINE')
    CRYPT_E_NOT_IN_REVOCATION_DATABASE: Final[ThirtyTwoBits] = cached_hresult(0x80092014, 'CRYPT_E_NOT_IN_REVOCATION_DATABASE')
    CRYPT_E_INVALID_NUMERIC_STRING: Final[ThirtyTwoBits] = cached_hresult(0x80092020, 'CRYPT_E_INVALID_NUMERIC_STRING')
    CRYPT_E_INVALID_PRINTABLE_STRING: Final[ThirtyTwoBits] = cached_hresult(0x80092021, 'CRYPT_E_INVALID_PRINTABLE_STRING')
    CRYPT_E_INVALID_IA5_STRING: Final[ThirtyTwoBits] = cached_hresult(0x80092022, 'CRYPT_E_INVALID_IA5_STRING')
    CRYPT_E_INVALID_X500_STRING: Final[ThirtyTwoBits] = cached_hresult(0x80092023, 'CRYPT_E_INVALID_X500_STRING')
    CRYPT_E_NOT_CHAR_STRING: Final[ThirtyTwoBits] = cached_hresult(0x80092024, 'CRYPT_E_NOT_CHAR_STRING')
    CRYPT_E_FILERESIZED: Final[ThirtyTwoBits] = cached_hresult(0x80092025, 'CRYPT_E_FILERESIZED')
    CRYPT_E_SECURITY_SETTINGS: Final[ThirtyTwoBits] = cached_hresult(0x80092026, 'CRYPT_E_SECURITY_SETTINGS')
    CRYPT_E_NO_VERIFY_USAGE_DLL: Final[ThirtyTwoBits] = cached_hresult(0x80092027, 'CRYPT_E_NO_VERIFY_USAGE_DLL')
    CRYPT_E_NO_VERIFY_USAGE_CHECK: Final[ThirtyTwoBits] = cached_hresult(0x80092028, 'CRYPT_E_NO_VERIFY_USAGE_CHECK')
    CRYPT_E_VERIFY_USAGE_OFFLINE: Final[ThirtyTwoBits] = cached_hresult(0x80092029, 'CRYPT_E_VERIFY_USAGE_OFFLINE')
    CRYPT_E_NOT_IN_CTL: Final[ThirtyTwoBits] = cached_hresult(0x8009202A, 'CRYPT_E_NOT_IN_CTL')
    CRYPT_E_NO_TRUSTED_SIGNER: Final[ThirtyTwoBits] = cached_hresult(0x8009202B, 'CRYPT_E_NO_TRUSTED_SIGNER')
    CRYPT_E_MISSING_PUBKEY_PARA: Final[ThirtyTwoBits] = cached_hresult(0x8009202C, 'CRYPT_E_MISSING_PUBKEY_PARA')
    CRYPT_E_OBJECT_LOCATOR_OBJECT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8009202D, 'CRYPT_E_OBJECT_LOCATOR_OBJECT_NOT_FOUND')
    CRYPT_E_OSS_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80093000, 'CRYPT_E_OSS_ERROR')
    OSS_MORE_BUF: Final[ThirtyTwoBits] = cached_hresult(0x80093001, 'OSS_MORE_BUF')
    OSS_NEGATIVE_UINTEGER: Final[ThirtyTwoBits] = cached_hresult(0x80093002, 'OSS_NEGATIVE_UINTEGER')
    OSS_PDU_RANGE: Final[ThirtyTwoBits] = cached_hresult(0x80093003, 'OSS_PDU_RANGE')
    OSS_MORE_INPUT: Final[ThirtyTwoBits] = cached_hresult(0x80093004, 'OSS_MORE_INPUT')
    OSS_DATA_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80093005, 'OSS_DATA_ERROR')
    OSS_BAD_ARG: Final[ThirtyTwoBits] = cached_hresult(0x80093006, 'OSS_BAD_ARG')
    OSS_BAD_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80093007, 'OSS_BAD_VERSION')
    OSS_OUT_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0x80093008, 'OSS_OUT_MEMORY')
    OSS_PDU_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80093009, 'OSS_PDU_MISMATCH')
    OSS_LIMITED: Final[ThirtyTwoBits] = cached_hresult(0x8009300A, 'OSS_LIMITED')
    OSS_BAD_PTR: Final[ThirtyTwoBits] = cached_hresult(0x8009300B, 'OSS_BAD_PTR')
    OSS_BAD_TIME: Final[ThirtyTwoBits] = cached_hresult(0x8009300C, 'OSS_BAD_TIME')
    OSS_INDEFINITE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8009300D, 'OSS_INDEFINITE_NOT_SUPPORTED')
    OSS_MEM_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8009300E, 'OSS_MEM_ERROR')
    OSS_BAD_TABLE: Final[ThirtyTwoBits] = cached_hresult(0x8009300F, 'OSS_BAD_TABLE')
    OSS_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0x80093010, 'OSS_TOO_LONG')
    OSS_CONSTRAINT_VIOLATED: Final[ThirtyTwoBits] = cached_hresult(0x80093011, 'OSS_CONSTRAINT_VIOLATED')
    OSS_FATAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80093012, 'OSS_FATAL_ERROR')
    OSS_ACCESS_SERIALIZATION_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80093013, 'OSS_ACCESS_SERIALIZATION_ERROR')
    OSS_NULL_TBL: Final[ThirtyTwoBits] = cached_hresult(0x80093014, 'OSS_NULL_TBL')
    OSS_NULL_FCN: Final[ThirtyTwoBits] = cached_hresult(0x80093015, 'OSS_NULL_FCN')
    OSS_BAD_ENCRULES: Final[ThirtyTwoBits] = cached_hresult(0x80093016, 'OSS_BAD_ENCRULES')
    OSS_UNAVAIL_ENCRULES: Final[ThirtyTwoBits] = cached_hresult(0x80093017, 'OSS_UNAVAIL_ENCRULES')
    OSS_CANT_OPEN_TRACE_WINDOW: Final[ThirtyTwoBits] = cached_hresult(0x80093018, 'OSS_CANT_OPEN_TRACE_WINDOW')
    OSS_UNIMPLEMENTED: Final[ThirtyTwoBits] = cached_hresult(0x80093019, 'OSS_UNIMPLEMENTED')
    OSS_OID_DLL_NOT_LINKED: Final[ThirtyTwoBits] = cached_hresult(0x8009301A, 'OSS_OID_DLL_NOT_LINKED')
    OSS_CANT_OPEN_TRACE_FILE: Final[ThirtyTwoBits] = cached_hresult(0x8009301B, 'OSS_CANT_OPEN_TRACE_FILE')
    OSS_TRACE_FILE_ALREADY_OPEN: Final[ThirtyTwoBits] = cached_hresult(0x8009301C, 'OSS_TRACE_FILE_ALREADY_OPEN')
    OSS_TABLE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x8009301D, 'OSS_TABLE_MISMATCH')
    OSS_TYPE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8009301E, 'OSS_TYPE_NOT_SUPPORTED')
    OSS_REAL_DLL_NOT_LINKED: Final[ThirtyTwoBits] = cached_hresult(0x8009301F, 'OSS_REAL_DLL_NOT_LINKED')
    OSS_REAL_CODE_NOT_LINKED: Final[ThirtyTwoBits] = cached_hresult(0x80093020, 'OSS_REAL_CODE_NOT_LINKED')
    OSS_OUT_OF_RANGE: Final[ThirtyTwoBits] = cached_hresult(0x80093021, 'OSS_OUT_OF_RANGE')
    OSS_COPIER_DLL_NOT_LINKED: Final[ThirtyTwoBits] = cached_hresult(0x80093022, 'OSS_COPIER_DLL_NOT_LINKED')
    OSS_CONSTRAINT_DLL_NOT_LINKED: Final[ThirtyTwoBits] = cached_hresult(0x80093023, 'OSS_CONSTRAINT_DLL_NOT_LINKED')
    OSS_COMPARATOR_DLL_NOT_LINKED: Final[ThirtyTwoBits] = cached_hresult(0x80093024, 'OSS_COMPARATOR_DLL_NOT_LINKED')
    OSS_COMPARATOR_CODE_NOT_LINKED: Final[ThirtyTwoBits] = cached_hresult(0x80093025, 'OSS_COMPARATOR_CODE_NOT_LINKED')
    OSS_MEM_MGR_DLL_NOT_LINKED: Final[ThirtyTwoBits] = cached_hresult(0x80093026, 'OSS_MEM_MGR_DLL_NOT_LINKED')
    OSS_PDV_DLL_NOT_LINKED: Final[ThirtyTwoBits] = cached_hresult(0x80093027, 'OSS_PDV_DLL_NOT_LINKED')
    OSS_PDV_CODE_NOT_LINKED: Final[ThirtyTwoBits] = cached_hresult(0x80093028, 'OSS_PDV_CODE_NOT_LINKED')
    OSS_API_DLL_NOT_LINKED: Final[ThirtyTwoBits] = cached_hresult(0x80093029, 'OSS_API_DLL_NOT_LINKED')
    OSS_BERDER_DLL_NOT_LINKED: Final[ThirtyTwoBits] = cached_hresult(0x8009302A, 'OSS_BERDER_DLL_NOT_LINKED')
    OSS_PER_DLL_NOT_LINKED: Final[ThirtyTwoBits] = cached_hresult(0x8009302B, 'OSS_PER_DLL_NOT_LINKED')
    OSS_OPEN_TYPE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8009302C, 'OSS_OPEN_TYPE_ERROR')
    OSS_MUTEX_NOT_CREATED: Final[ThirtyTwoBits] = cached_hresult(0x8009302D, 'OSS_MUTEX_NOT_CREATED')
    OSS_CANT_CLOSE_TRACE_FILE: Final[ThirtyTwoBits] = cached_hresult(0x8009302E, 'OSS_CANT_CLOSE_TRACE_FILE')
    CRYPT_E_ASN1_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80093100, 'CRYPT_E_ASN1_ERROR')
    CRYPT_E_ASN1_INTERNAL: Final[ThirtyTwoBits] = cached_hresult(0x80093101, 'CRYPT_E_ASN1_INTERNAL')
    CRYPT_E_ASN1_EOD: Final[ThirtyTwoBits] = cached_hresult(0x80093102, 'CRYPT_E_ASN1_EOD')
    CRYPT_E_ASN1_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x80093103, 'CRYPT_E_ASN1_CORRUPT')
    CRYPT_E_ASN1_LARGE: Final[ThirtyTwoBits] = cached_hresult(0x80093104, 'CRYPT_E_ASN1_LARGE')
    CRYPT_E_ASN1_CONSTRAINT: Final[ThirtyTwoBits] = cached_hresult(0x80093105, 'CRYPT_E_ASN1_CONSTRAINT')
    CRYPT_E_ASN1_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0x80093106, 'CRYPT_E_ASN1_MEMORY')
    CRYPT_E_ASN1_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0x80093107, 'CRYPT_E_ASN1_OVERFLOW')
    CRYPT_E_ASN1_BADPDU: Final[ThirtyTwoBits] = cached_hresult(0x80093108, 'CRYPT_E_ASN1_BADPDU')
    CRYPT_E_ASN1_BADARGS: Final[ThirtyTwoBits] = cached_hresult(0x80093109, 'CRYPT_E_ASN1_BADARGS')
    CRYPT_E_ASN1_BADREAL: Final[ThirtyTwoBits] = cached_hresult(0x8009310A, 'CRYPT_E_ASN1_BADREAL')
    CRYPT_E_ASN1_BADTAG: Final[ThirtyTwoBits] = cached_hresult(0x8009310B, 'CRYPT_E_ASN1_BADTAG')
    CRYPT_E_ASN1_CHOICE: Final[ThirtyTwoBits] = cached_hresult(0x8009310C, 'CRYPT_E_ASN1_CHOICE')
    CRYPT_E_ASN1_RULE: Final[ThirtyTwoBits] = cached_hresult(0x8009310D, 'CRYPT_E_ASN1_RULE')
    CRYPT_E_ASN1_UTF8: Final[ThirtyTwoBits] = cached_hresult(0x8009310E, 'CRYPT_E_ASN1_UTF8')
    CRYPT_E_ASN1_PDU_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x80093133, 'CRYPT_E_ASN1_PDU_TYPE')
    CRYPT_E_ASN1_NYI: Final[ThirtyTwoBits] = cached_hresult(0x80093134, 'CRYPT_E_ASN1_NYI')
    CRYPT_E_ASN1_EXTENDED: Final[ThirtyTwoBits] = cached_hresult(0x80093201, 'CRYPT_E_ASN1_EXTENDED')
    CRYPT_E_ASN1_NOEOD: Final[ThirtyTwoBits] = cached_hresult(0x80093202, 'CRYPT_E_ASN1_NOEOD')
    CERTSRV_E_BAD_REQUESTSUBJECT: Final[ThirtyTwoBits] = cached_hresult(0x80094001, 'CERTSRV_E_BAD_REQUESTSUBJECT')
    CERTSRV_E_NO_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0x80094002, 'CERTSRV_E_NO_REQUEST')
    CERTSRV_E_BAD_REQUESTSTATUS: Final[ThirtyTwoBits] = cached_hresult(0x80094003, 'CERTSRV_E_BAD_REQUESTSTATUS')
    CERTSRV_E_PROPERTY_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0x80094004, 'CERTSRV_E_PROPERTY_EMPTY')
    CERTSRV_E_INVALID_CA_CERTIFICATE: Final[ThirtyTwoBits] = cached_hresult(0x80094005, 'CERTSRV_E_INVALID_CA_CERTIFICATE')
    CERTSRV_E_SERVER_SUSPENDED: Final[ThirtyTwoBits] = cached_hresult(0x80094006, 'CERTSRV_E_SERVER_SUSPENDED')
    CERTSRV_E_ENCODING_LENGTH: Final[ThirtyTwoBits] = cached_hresult(0x80094007, 'CERTSRV_E_ENCODING_LENGTH')
    CERTSRV_E_ROLECONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x80094008, 'CERTSRV_E_ROLECONFLICT')
    CERTSRV_E_RESTRICTEDOFFICER: Final[ThirtyTwoBits] = cached_hresult(0x80094009, 'CERTSRV_E_RESTRICTEDOFFICER')
    CERTSRV_E_KEY_ARCHIVAL_NOT_CONFIGURED: Final[ThirtyTwoBits] = cached_hresult(0x8009400A, 'CERTSRV_E_KEY_ARCHIVAL_NOT_CONFIGURED')
    CERTSRV_E_NO_VALID_KRA: Final[ThirtyTwoBits] = cached_hresult(0x8009400B, 'CERTSRV_E_NO_VALID_KRA')
    CERTSRV_E_BAD_REQUEST_KEY_ARCHIVAL: Final[ThirtyTwoBits] = cached_hresult(0x8009400C, 'CERTSRV_E_BAD_REQUEST_KEY_ARCHIVAL')
    CERTSRV_E_NO_CAADMIN_DEFINED: Final[ThirtyTwoBits] = cached_hresult(0x8009400D, 'CERTSRV_E_NO_CAADMIN_DEFINED')
    CERTSRV_E_BAD_RENEWAL_CERT_ATTRIBUTE: Final[ThirtyTwoBits] = cached_hresult(0x8009400E, 'CERTSRV_E_BAD_RENEWAL_CERT_ATTRIBUTE')
    CERTSRV_E_NO_DB_SESSIONS: Final[ThirtyTwoBits] = cached_hresult(0x8009400F, 'CERTSRV_E_NO_DB_SESSIONS')
    CERTSRV_E_ALIGNMENT_FAULT: Final[ThirtyTwoBits] = cached_hresult(0x80094010, 'CERTSRV_E_ALIGNMENT_FAULT')
    CERTSRV_E_ENROLL_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x80094011, 'CERTSRV_E_ENROLL_DENIED')
    CERTSRV_E_TEMPLATE_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x80094012, 'CERTSRV_E_TEMPLATE_DENIED')
    CERTSRV_E_DOWNLEVEL_DC_SSL_OR_UPGRADE: Final[ThirtyTwoBits] = cached_hresult(0x80094013, 'CERTSRV_E_DOWNLEVEL_DC_SSL_OR_UPGRADE')
    CERTSRV_E_ADMIN_DENIED_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0x80094014, 'CERTSRV_E_ADMIN_DENIED_REQUEST')
    CERTSRV_E_NO_POLICY_SERVER: Final[ThirtyTwoBits] = cached_hresult(0x80094015, 'CERTSRV_E_NO_POLICY_SERVER')
    CERTSRV_E_WEAK_SIGNATURE_OR_KEY: Final[ThirtyTwoBits] = cached_hresult(0x80094016, 'CERTSRV_E_WEAK_SIGNATURE_OR_KEY')
    CERTSRV_E_KEY_ATTESTATION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80094017, 'CERTSRV_E_KEY_ATTESTATION_NOT_SUPPORTED')
    CERTSRV_E_ENCRYPTION_CERT_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80094018, 'CERTSRV_E_ENCRYPTION_CERT_REQUIRED')
    CERTSRV_E_UNSUPPORTED_CERT_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x80094800, 'CERTSRV_E_UNSUPPORTED_CERT_TYPE')
    CERTSRV_E_NO_CERT_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x80094801, 'CERTSRV_E_NO_CERT_TYPE')
    CERTSRV_E_TEMPLATE_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x80094802, 'CERTSRV_E_TEMPLATE_CONFLICT')
    CERTSRV_E_SUBJECT_ALT_NAME_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80094803, 'CERTSRV_E_SUBJECT_ALT_NAME_REQUIRED')
    CERTSRV_E_ARCHIVED_KEY_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80094804, 'CERTSRV_E_ARCHIVED_KEY_REQUIRED')
    CERTSRV_E_SMIME_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80094805, 'CERTSRV_E_SMIME_REQUIRED')
    CERTSRV_E_BAD_RENEWAL_SUBJECT: Final[ThirtyTwoBits] = cached_hresult(0x80094806, 'CERTSRV_E_BAD_RENEWAL_SUBJECT')
    CERTSRV_E_BAD_TEMPLATE_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80094807, 'CERTSRV_E_BAD_TEMPLATE_VERSION')
    CERTSRV_E_TEMPLATE_POLICY_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80094808, 'CERTSRV_E_TEMPLATE_POLICY_REQUIRED')
    CERTSRV_E_SIGNATURE_POLICY_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80094809, 'CERTSRV_E_SIGNATURE_POLICY_REQUIRED')
    CERTSRV_E_SIGNATURE_COUNT: Final[ThirtyTwoBits] = cached_hresult(0x8009480A, 'CERTSRV_E_SIGNATURE_COUNT')
    CERTSRV_E_SIGNATURE_REJECTED: Final[ThirtyTwoBits] = cached_hresult(0x8009480B, 'CERTSRV_E_SIGNATURE_REJECTED')
    CERTSRV_E_ISSUANCE_POLICY_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8009480C, 'CERTSRV_E_ISSUANCE_POLICY_REQUIRED')
    CERTSRV_E_SUBJECT_UPN_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8009480D, 'CERTSRV_E_SUBJECT_UPN_REQUIRED')
    CERTSRV_E_SUBJECT_DIRECTORY_GUID_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8009480E, 'CERTSRV_E_SUBJECT_DIRECTORY_GUID_REQUIRED')
    CERTSRV_E_SUBJECT_DNS_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8009480F, 'CERTSRV_E_SUBJECT_DNS_REQUIRED')
    CERTSRV_E_ARCHIVED_KEY_UNEXPECTED: Final[ThirtyTwoBits] = cached_hresult(0x80094810, 'CERTSRV_E_ARCHIVED_KEY_UNEXPECTED')
    CERTSRV_E_KEY_LENGTH: Final[ThirtyTwoBits] = cached_hresult(0x80094811, 'CERTSRV_E_KEY_LENGTH')
    CERTSRV_E_SUBJECT_EMAIL_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80094812, 'CERTSRV_E_SUBJECT_EMAIL_REQUIRED')
    CERTSRV_E_UNKNOWN_CERT_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x80094813, 'CERTSRV_E_UNKNOWN_CERT_TYPE')
    CERTSRV_E_CERT_TYPE_OVERLAP: Final[ThirtyTwoBits] = cached_hresult(0x80094814, 'CERTSRV_E_CERT_TYPE_OVERLAP')
    CERTSRV_E_TOO_MANY_SIGNATURES: Final[ThirtyTwoBits] = cached_hresult(0x80094815, 'CERTSRV_E_TOO_MANY_SIGNATURES')
    CERTSRV_E_RENEWAL_BAD_PUBLIC_KEY: Final[ThirtyTwoBits] = cached_hresult(0x80094816, 'CERTSRV_E_RENEWAL_BAD_PUBLIC_KEY')
    CERTSRV_E_INVALID_EK: Final[ThirtyTwoBits] = cached_hresult(0x80094817, 'CERTSRV_E_INVALID_EK')
    CERTSRV_E_INVALID_IDBINDING: Final[ThirtyTwoBits] = cached_hresult(0x80094818, 'CERTSRV_E_INVALID_IDBINDING')
    CERTSRV_E_INVALID_ATTESTATION: Final[ThirtyTwoBits] = cached_hresult(0x80094819, 'CERTSRV_E_INVALID_ATTESTATION')
    CERTSRV_E_KEY_ATTESTATION: Final[ThirtyTwoBits] = cached_hresult(0x8009481A, 'CERTSRV_E_KEY_ATTESTATION')
    CERTSRV_E_CORRUPT_KEY_ATTESTATION: Final[ThirtyTwoBits] = cached_hresult(0x8009481B, 'CERTSRV_E_CORRUPT_KEY_ATTESTATION')
    CERTSRV_E_EXPIRED_CHALLENGE: Final[ThirtyTwoBits] = cached_hresult(0x8009481C, 'CERTSRV_E_EXPIRED_CHALLENGE')
    CERTSRV_E_INVALID_RESPONSE: Final[ThirtyTwoBits] = cached_hresult(0x8009481D, 'CERTSRV_E_INVALID_RESPONSE')
    CERTSRV_E_INVALID_REQUESTID: Final[ThirtyTwoBits] = cached_hresult(0x8009481E, 'CERTSRV_E_INVALID_REQUESTID')
    CERTSRV_E_REQUEST_PRECERTIFICATE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x8009481F, 'CERTSRV_E_REQUEST_PRECERTIFICATE_MISMATCH')
    CERTSRV_E_PENDING_CLIENT_RESPONSE: Final[ThirtyTwoBits] = cached_hresult(0x80094820, 'CERTSRV_E_PENDING_CLIENT_RESPONSE')
    CERTSRV_E_SEC_EXT_DIRECTORY_SID_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80094821, 'CERTSRV_E_SEC_EXT_DIRECTORY_SID_REQUIRED')
    XENROLL_E_KEY_NOT_EXPORTABLE: Final[ThirtyTwoBits] = cached_hresult(0x80095000, 'XENROLL_E_KEY_NOT_EXPORTABLE')
    XENROLL_E_CANNOT_ADD_ROOT_CERT: Final[ThirtyTwoBits] = cached_hresult(0x80095001, 'XENROLL_E_CANNOT_ADD_ROOT_CERT')
    XENROLL_E_RESPONSE_KA_HASH_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80095002, 'XENROLL_E_RESPONSE_KA_HASH_NOT_FOUND')
    XENROLL_E_RESPONSE_UNEXPECTED_KA_HASH: Final[ThirtyTwoBits] = cached_hresult(0x80095003, 'XENROLL_E_RESPONSE_UNEXPECTED_KA_HASH')
    XENROLL_E_RESPONSE_KA_HASH_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80095004, 'XENROLL_E_RESPONSE_KA_HASH_MISMATCH')
    XENROLL_E_KEYSPEC_SMIME_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80095005, 'XENROLL_E_KEYSPEC_SMIME_MISMATCH')
    TRUST_E_SYSTEM_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80096001, 'TRUST_E_SYSTEM_ERROR')
    TRUST_E_NO_SIGNER_CERT: Final[ThirtyTwoBits] = cached_hresult(0x80096002, 'TRUST_E_NO_SIGNER_CERT')
    TRUST_E_COUNTER_SIGNER: Final[ThirtyTwoBits] = cached_hresult(0x80096003, 'TRUST_E_COUNTER_SIGNER')
    TRUST_E_CERT_SIGNATURE: Final[ThirtyTwoBits] = cached_hresult(0x80096004, 'TRUST_E_CERT_SIGNATURE')
    TRUST_E_TIME_STAMP: Final[ThirtyTwoBits] = cached_hresult(0x80096005, 'TRUST_E_TIME_STAMP')
    TRUST_E_BAD_DIGEST: Final[ThirtyTwoBits] = cached_hresult(0x80096010, 'TRUST_E_BAD_DIGEST')
    TRUST_E_MALFORMED_SIGNATURE: Final[ThirtyTwoBits] = cached_hresult(0x80096011, 'TRUST_E_MALFORMED_SIGNATURE')
    TRUST_E_BASIC_CONSTRAINTS: Final[ThirtyTwoBits] = cached_hresult(0x80096019, 'TRUST_E_BASIC_CONSTRAINTS')
    TRUST_E_FINANCIAL_CRITERIA: Final[ThirtyTwoBits] = cached_hresult(0x8009601E, 'TRUST_E_FINANCIAL_CRITERIA')
    MSSIPOTF_E_OUTOFMEMRANGE: Final[ThirtyTwoBits] = cached_hresult(0x80097001, 'MSSIPOTF_E_OUTOFMEMRANGE')
    MSSIPOTF_E_CANTGETOBJECT: Final[ThirtyTwoBits] = cached_hresult(0x80097002, 'MSSIPOTF_E_CANTGETOBJECT')
    MSSIPOTF_E_NOHEADTABLE: Final[ThirtyTwoBits] = cached_hresult(0x80097003, 'MSSIPOTF_E_NOHEADTABLE')
    MSSIPOTF_E_BAD_MAGICNUMBER: Final[ThirtyTwoBits] = cached_hresult(0x80097004, 'MSSIPOTF_E_BAD_MAGICNUMBER')
    MSSIPOTF_E_BAD_OFFSET_TABLE: Final[ThirtyTwoBits] = cached_hresult(0x80097005, 'MSSIPOTF_E_BAD_OFFSET_TABLE')
    MSSIPOTF_E_TABLE_TAGORDER: Final[ThirtyTwoBits] = cached_hresult(0x80097006, 'MSSIPOTF_E_TABLE_TAGORDER')
    MSSIPOTF_E_TABLE_LONGWORD: Final[ThirtyTwoBits] = cached_hresult(0x80097007, 'MSSIPOTF_E_TABLE_LONGWORD')
    MSSIPOTF_E_BAD_FIRST_TABLE_PLACEMENT: Final[ThirtyTwoBits] = cached_hresult(0x80097008, 'MSSIPOTF_E_BAD_FIRST_TABLE_PLACEMENT')
    MSSIPOTF_E_TABLES_OVERLAP: Final[ThirtyTwoBits] = cached_hresult(0x80097009, 'MSSIPOTF_E_TABLES_OVERLAP')
    MSSIPOTF_E_TABLE_PADBYTES: Final[ThirtyTwoBits] = cached_hresult(0x8009700A, 'MSSIPOTF_E_TABLE_PADBYTES')
    MSSIPOTF_E_FILETOOSMALL: Final[ThirtyTwoBits] = cached_hresult(0x8009700B, 'MSSIPOTF_E_FILETOOSMALL')
    MSSIPOTF_E_TABLE_CHECKSUM: Final[ThirtyTwoBits] = cached_hresult(0x8009700C, 'MSSIPOTF_E_TABLE_CHECKSUM')
    MSSIPOTF_E_FILE_CHECKSUM: Final[ThirtyTwoBits] = cached_hresult(0x8009700D, 'MSSIPOTF_E_FILE_CHECKSUM')
    MSSIPOTF_E_FAILED_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x80097010, 'MSSIPOTF_E_FAILED_POLICY')
    MSSIPOTF_E_FAILED_HINTS_CHECK: Final[ThirtyTwoBits] = cached_hresult(0x80097011, 'MSSIPOTF_E_FAILED_HINTS_CHECK')
    MSSIPOTF_E_NOT_OPENTYPE: Final[ThirtyTwoBits] = cached_hresult(0x80097012, 'MSSIPOTF_E_NOT_OPENTYPE')
    MSSIPOTF_E_FILE: Final[ThirtyTwoBits] = cached_hresult(0x80097013, 'MSSIPOTF_E_FILE')
    MSSIPOTF_E_CRYPT: Final[ThirtyTwoBits] = cached_hresult(0x80097014, 'MSSIPOTF_E_CRYPT')
    MSSIPOTF_E_BADVERSION: Final[ThirtyTwoBits] = cached_hresult(0x80097015, 'MSSIPOTF_E_BADVERSION')
    MSSIPOTF_E_DSIG_STRUCTURE: Final[ThirtyTwoBits] = cached_hresult(0x80097016, 'MSSIPOTF_E_DSIG_STRUCTURE')
    MSSIPOTF_E_PCONST_CHECK: Final[ThirtyTwoBits] = cached_hresult(0x80097017, 'MSSIPOTF_E_PCONST_CHECK')
    MSSIPOTF_E_STRUCTURE: Final[ThirtyTwoBits] = cached_hresult(0x80097018, 'MSSIPOTF_E_STRUCTURE')
    ERROR_CRED_REQUIRES_CONFIRMATION: Final[ThirtyTwoBits] = cached_hresult(0x80097019, 'ERROR_CRED_REQUIRES_CONFIRMATION')
    TRUST_E_PROVIDER_UNKNOWN: Final[ThirtyTwoBits] = cached_hresult(0x800B0001, 'TRUST_E_PROVIDER_UNKNOWN')
    TRUST_E_ACTION_UNKNOWN: Final[ThirtyTwoBits] = cached_hresult(0x800B0002, 'TRUST_E_ACTION_UNKNOWN')
    TRUST_E_SUBJECT_FORM_UNKNOWN: Final[ThirtyTwoBits] = cached_hresult(0x800B0003, 'TRUST_E_SUBJECT_FORM_UNKNOWN')
    TRUST_E_SUBJECT_NOT_TRUSTED: Final[ThirtyTwoBits] = cached_hresult(0x800B0004, 'TRUST_E_SUBJECT_NOT_TRUSTED')
    DIGSIG_E_ENCODE: Final[ThirtyTwoBits] = cached_hresult(0x800B0005, 'DIGSIG_E_ENCODE')
    DIGSIG_E_DECODE: Final[ThirtyTwoBits] = cached_hresult(0x800B0006, 'DIGSIG_E_DECODE')
    DIGSIG_E_EXTENSIBILITY: Final[ThirtyTwoBits] = cached_hresult(0x800B0007, 'DIGSIG_E_EXTENSIBILITY')
    DIGSIG_E_CRYPTO: Final[ThirtyTwoBits] = cached_hresult(0x800B0008, 'DIGSIG_E_CRYPTO')
    PERSIST_E_SIZEDEFINITE: Final[ThirtyTwoBits] = cached_hresult(0x800B0009, 'PERSIST_E_SIZEDEFINITE')
    PERSIST_E_SIZEINDEFINITE: Final[ThirtyTwoBits] = cached_hresult(0x800B000A, 'PERSIST_E_SIZEINDEFINITE')
    PERSIST_E_NOTSELFSIZING: Final[ThirtyTwoBits] = cached_hresult(0x800B000B, 'PERSIST_E_NOTSELFSIZING')
    TRUST_E_NOSIGNATURE: Final[ThirtyTwoBits] = cached_hresult(0x800B0100, 'TRUST_E_NOSIGNATURE')
    CERT_E_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x800B0101, 'CERT_E_EXPIRED')
    CERT_E_VALIDITYPERIODNESTING: Final[ThirtyTwoBits] = cached_hresult(0x800B0102, 'CERT_E_VALIDITYPERIODNESTING')
    CERT_E_ROLE: Final[ThirtyTwoBits] = cached_hresult(0x800B0103, 'CERT_E_ROLE')
    CERT_E_PATHLENCONST: Final[ThirtyTwoBits] = cached_hresult(0x800B0104, 'CERT_E_PATHLENCONST')
    CERT_E_CRITICAL: Final[ThirtyTwoBits] = cached_hresult(0x800B0105, 'CERT_E_CRITICAL')
    CERT_E_PURPOSE: Final[ThirtyTwoBits] = cached_hresult(0x800B0106, 'CERT_E_PURPOSE')
    CERT_E_ISSUERCHAINING: Final[ThirtyTwoBits] = cached_hresult(0x800B0107, 'CERT_E_ISSUERCHAINING')
    CERT_E_MALFORMED: Final[ThirtyTwoBits] = cached_hresult(0x800B0108, 'CERT_E_MALFORMED')
    CERT_E_UNTRUSTEDROOT: Final[ThirtyTwoBits] = cached_hresult(0x800B0109, 'CERT_E_UNTRUSTEDROOT')
    CERT_E_CHAINING: Final[ThirtyTwoBits] = cached_hresult(0x800B010A, 'CERT_E_CHAINING')
    TRUST_E_FAIL: Final[ThirtyTwoBits] = cached_hresult(0x800B010B, 'TRUST_E_FAIL')
    CERT_E_REVOKED: Final[ThirtyTwoBits] = cached_hresult(0x800B010C, 'CERT_E_REVOKED')
    CERT_E_UNTRUSTEDTESTROOT: Final[ThirtyTwoBits] = cached_hresult(0x800B010D, 'CERT_E_UNTRUSTEDTESTROOT')
    CERT_E_REVOCATION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x800B010E, 'CERT_E_REVOCATION_FAILURE')
    CERT_E_CN_NO_MATCH: Final[ThirtyTwoBits] = cached_hresult(0x800B010F, 'CERT_E_CN_NO_MATCH')
    CERT_E_WRONG_USAGE: Final[ThirtyTwoBits] = cached_hresult(0x800B0110, 'CERT_E_WRONG_USAGE')
    TRUST_E_EXPLICIT_DISTRUST: Final[ThirtyTwoBits] = cached_hresult(0x800B0111, 'TRUST_E_EXPLICIT_DISTRUST')
    CERT_E_UNTRUSTEDCA: Final[ThirtyTwoBits] = cached_hresult(0x800B0112, 'CERT_E_UNTRUSTEDCA')
    CERT_E_INVALID_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x800B0113, 'CERT_E_INVALID_POLICY')
    CERT_E_INVALID_NAME: Final[ThirtyTwoBits] = cached_hresult(0x800B0114, 'CERT_E_INVALID_NAME')
    INET_E_INVALID_URL: Final[ThirtyTwoBits] = cached_hresult(0x800C0002, 'INET_E_INVALID_URL')
    INET_E_NO_SESSION: Final[ThirtyTwoBits] = cached_hresult(0x800C0003, 'INET_E_NO_SESSION')
    INET_E_CANNOT_CONNECT: Final[ThirtyTwoBits] = cached_hresult(0x800C0004, 'INET_E_CANNOT_CONNECT')
    INET_E_RESOURCE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800C0005, 'INET_E_RESOURCE_NOT_FOUND')
    INET_E_OBJECT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800C0006, 'INET_E_OBJECT_NOT_FOUND')
    INET_E_DATA_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x800C0007, 'INET_E_DATA_NOT_AVAILABLE')
    INET_E_DOWNLOAD_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x800C0008, 'INET_E_DOWNLOAD_FAILURE')
    INET_E_AUTHENTICATION_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x800C0009, 'INET_E_AUTHENTICATION_REQUIRED')
    INET_E_NO_VALID_MEDIA: Final[ThirtyTwoBits] = cached_hresult(0x800C000A, 'INET_E_NO_VALID_MEDIA')
    INET_E_CONNECTION_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x800C000B, 'INET_E_CONNECTION_TIMEOUT')
    INET_E_INVALID_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0x800C000C, 'INET_E_INVALID_REQUEST')
    INET_E_UNKNOWN_PROTOCOL: Final[ThirtyTwoBits] = cached_hresult(0x800C000D, 'INET_E_UNKNOWN_PROTOCOL')
    INET_E_SECURITY_PROBLEM: Final[ThirtyTwoBits] = cached_hresult(0x800C000E, 'INET_E_SECURITY_PROBLEM')
    INET_E_CANNOT_LOAD_DATA: Final[ThirtyTwoBits] = cached_hresult(0x800C000F, 'INET_E_CANNOT_LOAD_DATA')
    INET_E_CANNOT_INSTANTIATE_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x800C0010, 'INET_E_CANNOT_INSTANTIATE_OBJECT')
    INET_E_REDIRECT_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800C0014, 'INET_E_REDIRECT_FAILED')
    INET_E_REDIRECT_TO_DIR: Final[ThirtyTwoBits] = cached_hresult(0x800C0015, 'INET_E_REDIRECT_TO_DIR')
    INET_E_INVALID_CERTIFICATE: Final[ThirtyTwoBits] = cached_hresult(0x800C0019, 'INET_E_INVALID_CERTIFICATE')
    SPAPI_E_EXPECTED_SECTION_NAME: Final[ThirtyTwoBits] = cached_hresult(0x800F0000, 'SPAPI_E_EXPECTED_SECTION_NAME')
    SPAPI_E_BAD_SECTION_NAME_LINE: Final[ThirtyTwoBits] = cached_hresult(0x800F0001, 'SPAPI_E_BAD_SECTION_NAME_LINE')
    SPAPI_E_SECTION_NAME_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0x800F0002, 'SPAPI_E_SECTION_NAME_TOO_LONG')
    SPAPI_E_GENERAL_SYNTAX: Final[ThirtyTwoBits] = cached_hresult(0x800F0003, 'SPAPI_E_GENERAL_SYNTAX')
    SPAPI_E_WRONG_INF_STYLE: Final[ThirtyTwoBits] = cached_hresult(0x800F0100, 'SPAPI_E_WRONG_INF_STYLE')
    SPAPI_E_SECTION_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800F0101, 'SPAPI_E_SECTION_NOT_FOUND')
    SPAPI_E_LINE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800F0102, 'SPAPI_E_LINE_NOT_FOUND')
    SPAPI_E_NO_BACKUP: Final[ThirtyTwoBits] = cached_hresult(0x800F0103, 'SPAPI_E_NO_BACKUP')
    SPAPI_E_NO_ASSOCIATED_CLASS: Final[ThirtyTwoBits] = cached_hresult(0x800F0200, 'SPAPI_E_NO_ASSOCIATED_CLASS')
    SPAPI_E_CLASS_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x800F0201, 'SPAPI_E_CLASS_MISMATCH')
    SPAPI_E_DUPLICATE_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800F0202, 'SPAPI_E_DUPLICATE_FOUND')
    SPAPI_E_NO_DRIVER_SELECTED: Final[ThirtyTwoBits] = cached_hresult(0x800F0203, 'SPAPI_E_NO_DRIVER_SELECTED')
    SPAPI_E_KEY_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x800F0204, 'SPAPI_E_KEY_DOES_NOT_EXIST')
    SPAPI_E_INVALID_DEVINST_NAME: Final[ThirtyTwoBits] = cached_hresult(0x800F0205, 'SPAPI_E_INVALID_DEVINST_NAME')
    SPAPI_E_INVALID_CLASS: Final[ThirtyTwoBits] = cached_hresult(0x800F0206, 'SPAPI_E_INVALID_CLASS')
    SPAPI_E_DEVINST_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800F0207, 'SPAPI_E_DEVINST_ALREADY_EXISTS')
    SPAPI_E_DEVINFO_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x800F0208, 'SPAPI_E_DEVINFO_NOT_REGISTERED')
    SPAPI_E_INVALID_REG_PROPERTY: Final[ThirtyTwoBits] = cached_hresult(0x800F0209, 'SPAPI_E_INVALID_REG_PROPERTY')
    SPAPI_E_NO_INF: Final[ThirtyTwoBits] = cached_hresult(0x800F020A, 'SPAPI_E_NO_INF')
    SPAPI_E_NO_SUCH_DEVINST: Final[ThirtyTwoBits] = cached_hresult(0x800F020B, 'SPAPI_E_NO_SUCH_DEVINST')
    SPAPI_E_CANT_LOAD_CLASS_ICON: Final[ThirtyTwoBits] = cached_hresult(0x800F020C, 'SPAPI_E_CANT_LOAD_CLASS_ICON')
    SPAPI_E_INVALID_CLASS_INSTALLER: Final[ThirtyTwoBits] = cached_hresult(0x800F020D, 'SPAPI_E_INVALID_CLASS_INSTALLER')
    SPAPI_E_DI_DO_DEFAULT: Final[ThirtyTwoBits] = cached_hresult(0x800F020E, 'SPAPI_E_DI_DO_DEFAULT')
    SPAPI_E_DI_NOFILECOPY: Final[ThirtyTwoBits] = cached_hresult(0x800F020F, 'SPAPI_E_DI_NOFILECOPY')
    SPAPI_E_INVALID_HWPROFILE: Final[ThirtyTwoBits] = cached_hresult(0x800F0210, 'SPAPI_E_INVALID_HWPROFILE')
    SPAPI_E_NO_DEVICE_SELECTED: Final[ThirtyTwoBits] = cached_hresult(0x800F0211, 'SPAPI_E_NO_DEVICE_SELECTED')
    SPAPI_E_DEVINFO_LIST_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x800F0212, 'SPAPI_E_DEVINFO_LIST_LOCKED')
    SPAPI_E_DEVINFO_DATA_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x800F0213, 'SPAPI_E_DEVINFO_DATA_LOCKED')
    SPAPI_E_DI_BAD_PATH: Final[ThirtyTwoBits] = cached_hresult(0x800F0214, 'SPAPI_E_DI_BAD_PATH')
    SPAPI_E_NO_CLASSINSTALL_PARAMS: Final[ThirtyTwoBits] = cached_hresult(0x800F0215, 'SPAPI_E_NO_CLASSINSTALL_PARAMS')
    SPAPI_E_FILEQUEUE_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x800F0216, 'SPAPI_E_FILEQUEUE_LOCKED')
    SPAPI_E_BAD_SERVICE_INSTALLSECT: Final[ThirtyTwoBits] = cached_hresult(0x800F0217, 'SPAPI_E_BAD_SERVICE_INSTALLSECT')
    SPAPI_E_NO_CLASS_DRIVER_LIST: Final[ThirtyTwoBits] = cached_hresult(0x800F0218, 'SPAPI_E_NO_CLASS_DRIVER_LIST')
    SPAPI_E_NO_ASSOCIATED_SERVICE: Final[ThirtyTwoBits] = cached_hresult(0x800F0219, 'SPAPI_E_NO_ASSOCIATED_SERVICE')
    SPAPI_E_NO_DEFAULT_DEVICE_INTERFACE: Final[ThirtyTwoBits] = cached_hresult(0x800F021A, 'SPAPI_E_NO_DEFAULT_DEVICE_INTERFACE')
    SPAPI_E_DEVICE_INTERFACE_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0x800F021B, 'SPAPI_E_DEVICE_INTERFACE_ACTIVE')
    SPAPI_E_DEVICE_INTERFACE_REMOVED: Final[ThirtyTwoBits] = cached_hresult(0x800F021C, 'SPAPI_E_DEVICE_INTERFACE_REMOVED')
    SPAPI_E_BAD_INTERFACE_INSTALLSECT: Final[ThirtyTwoBits] = cached_hresult(0x800F021D, 'SPAPI_E_BAD_INTERFACE_INSTALLSECT')
    SPAPI_E_NO_SUCH_INTERFACE_CLASS: Final[ThirtyTwoBits] = cached_hresult(0x800F021E, 'SPAPI_E_NO_SUCH_INTERFACE_CLASS')
    SPAPI_E_INVALID_REFERENCE_STRING: Final[ThirtyTwoBits] = cached_hresult(0x800F021F, 'SPAPI_E_INVALID_REFERENCE_STRING')
    SPAPI_E_INVALID_MACHINENAME: Final[ThirtyTwoBits] = cached_hresult(0x800F0220, 'SPAPI_E_INVALID_MACHINENAME')
    SPAPI_E_REMOTE_COMM_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x800F0221, 'SPAPI_E_REMOTE_COMM_FAILURE')
    SPAPI_E_MACHINE_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x800F0222, 'SPAPI_E_MACHINE_UNAVAILABLE')
    SPAPI_E_NO_CONFIGMGR_SERVICES: Final[ThirtyTwoBits] = cached_hresult(0x800F0223, 'SPAPI_E_NO_CONFIGMGR_SERVICES')
    SPAPI_E_INVALID_PROPPAGE_PROVIDER: Final[ThirtyTwoBits] = cached_hresult(0x800F0224, 'SPAPI_E_INVALID_PROPPAGE_PROVIDER')
    SPAPI_E_NO_SUCH_DEVICE_INTERFACE: Final[ThirtyTwoBits] = cached_hresult(0x800F0225, 'SPAPI_E_NO_SUCH_DEVICE_INTERFACE')
    SPAPI_E_DI_POSTPROCESSING_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x800F0226, 'SPAPI_E_DI_POSTPROCESSING_REQUIRED')
    SPAPI_E_INVALID_COINSTALLER: Final[ThirtyTwoBits] = cached_hresult(0x800F0227, 'SPAPI_E_INVALID_COINSTALLER')
    SPAPI_E_NO_COMPAT_DRIVERS: Final[ThirtyTwoBits] = cached_hresult(0x800F0228, 'SPAPI_E_NO_COMPAT_DRIVERS')
    SPAPI_E_NO_DEVICE_ICON: Final[ThirtyTwoBits] = cached_hresult(0x800F0229, 'SPAPI_E_NO_DEVICE_ICON')
    SPAPI_E_INVALID_INF_LOGCONFIG: Final[ThirtyTwoBits] = cached_hresult(0x800F022A, 'SPAPI_E_INVALID_INF_LOGCONFIG')
    SPAPI_E_DI_DONT_INSTALL: Final[ThirtyTwoBits] = cached_hresult(0x800F022B, 'SPAPI_E_DI_DONT_INSTALL')
    SPAPI_E_INVALID_FILTER_DRIVER: Final[ThirtyTwoBits] = cached_hresult(0x800F022C, 'SPAPI_E_INVALID_FILTER_DRIVER')
    SPAPI_E_NON_WINDOWS_NT_DRIVER: Final[ThirtyTwoBits] = cached_hresult(0x800F022D, 'SPAPI_E_NON_WINDOWS_NT_DRIVER')
    SPAPI_E_NON_WINDOWS_DRIVER: Final[ThirtyTwoBits] = cached_hresult(0x800F022E, 'SPAPI_E_NON_WINDOWS_DRIVER')
    SPAPI_E_NO_CATALOG_FOR_OEM_INF: Final[ThirtyTwoBits] = cached_hresult(0x800F022F, 'SPAPI_E_NO_CATALOG_FOR_OEM_INF')
    SPAPI_E_DEVINSTALL_QUEUE_NONNATIVE: Final[ThirtyTwoBits] = cached_hresult(0x800F0230, 'SPAPI_E_DEVINSTALL_QUEUE_NONNATIVE')
    SPAPI_E_NOT_DISABLEABLE: Final[ThirtyTwoBits] = cached_hresult(0x800F0231, 'SPAPI_E_NOT_DISABLEABLE')
    SPAPI_E_CANT_REMOVE_DEVINST: Final[ThirtyTwoBits] = cached_hresult(0x800F0232, 'SPAPI_E_CANT_REMOVE_DEVINST')
    SPAPI_E_INVALID_TARGET: Final[ThirtyTwoBits] = cached_hresult(0x800F0233, 'SPAPI_E_INVALID_TARGET')
    SPAPI_E_DRIVER_NONNATIVE: Final[ThirtyTwoBits] = cached_hresult(0x800F0234, 'SPAPI_E_DRIVER_NONNATIVE')
    SPAPI_E_IN_WOW64: Final[ThirtyTwoBits] = cached_hresult(0x800F0235, 'SPAPI_E_IN_WOW64')
    SPAPI_E_SET_SYSTEM_RESTORE_POINT: Final[ThirtyTwoBits] = cached_hresult(0x800F0236, 'SPAPI_E_SET_SYSTEM_RESTORE_POINT')
    SPAPI_E_INCORRECTLY_COPIED_INF: Final[ThirtyTwoBits] = cached_hresult(0x800F0237, 'SPAPI_E_INCORRECTLY_COPIED_INF')
    SPAPI_E_SCE_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x800F0238, 'SPAPI_E_SCE_DISABLED')
    SPAPI_E_UNKNOWN_EXCEPTION: Final[ThirtyTwoBits] = cached_hresult(0x800F0239, 'SPAPI_E_UNKNOWN_EXCEPTION')
    SPAPI_E_PNP_REGISTRY_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800F023A, 'SPAPI_E_PNP_REGISTRY_ERROR')
    SPAPI_E_REMOTE_REQUEST_UNSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x800F023B, 'SPAPI_E_REMOTE_REQUEST_UNSUPPORTED')
    SPAPI_E_NOT_AN_INSTALLED_OEM_INF: Final[ThirtyTwoBits] = cached_hresult(0x800F023C, 'SPAPI_E_NOT_AN_INSTALLED_OEM_INF')
    SPAPI_E_INF_IN_USE_BY_DEVICES: Final[ThirtyTwoBits] = cached_hresult(0x800F023D, 'SPAPI_E_INF_IN_USE_BY_DEVICES')
    SPAPI_E_DI_FUNCTION_OBSOLETE: Final[ThirtyTwoBits] = cached_hresult(0x800F023E, 'SPAPI_E_DI_FUNCTION_OBSOLETE')
    SPAPI_E_NO_AUTHENTICODE_CATALOG: Final[ThirtyTwoBits] = cached_hresult(0x800F023F, 'SPAPI_E_NO_AUTHENTICODE_CATALOG')
    SPAPI_E_AUTHENTICODE_DISALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x800F0240, 'SPAPI_E_AUTHENTICODE_DISALLOWED')
    SPAPI_E_AUTHENTICODE_TRUSTED_PUBLISHER: Final[ThirtyTwoBits] = cached_hresult(0x800F0241, 'SPAPI_E_AUTHENTICODE_TRUSTED_PUBLISHER')
    SPAPI_E_AUTHENTICODE_TRUST_NOT_ESTABLISHED: Final[ThirtyTwoBits] = cached_hresult(0x800F0242, 'SPAPI_E_AUTHENTICODE_TRUST_NOT_ESTABLISHED')
    SPAPI_E_AUTHENTICODE_PUBLISHER_NOT_TRUSTED: Final[ThirtyTwoBits] = cached_hresult(0x800F0243, 'SPAPI_E_AUTHENTICODE_PUBLISHER_NOT_TRUSTED')
    SPAPI_E_SIGNATURE_OSATTRIBUTE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x800F0244, 'SPAPI_E_SIGNATURE_OSATTRIBUTE_MISMATCH')
    SPAPI_E_ONLY_VALIDATE_VIA_AUTHENTICODE: Final[ThirtyTwoBits] = cached_hresult(0x800F0245, 'SPAPI_E_ONLY_VALIDATE_VIA_AUTHENTICODE')
    SPAPI_E_DEVICE_INSTALLER_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0x800F0246, 'SPAPI_E_DEVICE_INSTALLER_NOT_READY')
    SPAPI_E_DRIVER_STORE_ADD_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800F0247, 'SPAPI_E_DRIVER_STORE_ADD_FAILED')
    SPAPI_E_DEVICE_INSTALL_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x800F0248, 'SPAPI_E_DEVICE_INSTALL_BLOCKED')
    SPAPI_E_DRIVER_INSTALL_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x800F0249, 'SPAPI_E_DRIVER_INSTALL_BLOCKED')
    SPAPI_E_WRONG_INF_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x800F024A, 'SPAPI_E_WRONG_INF_TYPE')
    SPAPI_E_FILE_HASH_NOT_IN_CATALOG: Final[ThirtyTwoBits] = cached_hresult(0x800F024B, 'SPAPI_E_FILE_HASH_NOT_IN_CATALOG')
    SPAPI_E_DRIVER_STORE_DELETE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800F024C, 'SPAPI_E_DRIVER_STORE_DELETE_FAILED')
    SPAPI_E_UNRECOVERABLE_STACK_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0x800F0300, 'SPAPI_E_UNRECOVERABLE_STACK_OVERFLOW')
    SPAPI_E_ERROR_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x800F1000, 'SPAPI_E_ERROR_NOT_INSTALLED')
    SCARD_F_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80100001, 'SCARD_F_INTERNAL_ERROR')
    SCARD_E_CANCELLED: Final[ThirtyTwoBits] = cached_hresult(0x80100002, 'SCARD_E_CANCELLED')
    SCARD_E_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80100003, 'SCARD_E_INVALID_HANDLE')
    SCARD_E_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0x80100004, 'SCARD_E_INVALID_PARAMETER')
    SCARD_E_INVALID_TARGET: Final[ThirtyTwoBits] = cached_hresult(0x80100005, 'SCARD_E_INVALID_TARGET')
    SCARD_E_NO_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0x80100006, 'SCARD_E_NO_MEMORY')
    SCARD_F_WAITED_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0x80100007, 'SCARD_F_WAITED_TOO_LONG')
    SCARD_E_INSUFFICIENT_BUFFER: Final[ThirtyTwoBits] = cached_hresult(0x80100008, 'SCARD_E_INSUFFICIENT_BUFFER')
    SCARD_E_UNKNOWN_READER: Final[ThirtyTwoBits] = cached_hresult(0x80100009, 'SCARD_E_UNKNOWN_READER')
    SCARD_E_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x8010000A, 'SCARD_E_TIMEOUT')
    SCARD_E_SHARING_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x8010000B, 'SCARD_E_SHARING_VIOLATION')
    SCARD_E_NO_SMARTCARD: Final[ThirtyTwoBits] = cached_hresult(0x8010000C, 'SCARD_E_NO_SMARTCARD')
    SCARD_E_UNKNOWN_CARD: Final[ThirtyTwoBits] = cached_hresult(0x8010000D, 'SCARD_E_UNKNOWN_CARD')
    SCARD_E_CANT_DISPOSE: Final[ThirtyTwoBits] = cached_hresult(0x8010000E, 'SCARD_E_CANT_DISPOSE')
    SCARD_E_PROTO_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x8010000F, 'SCARD_E_PROTO_MISMATCH')
    SCARD_E_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0x80100010, 'SCARD_E_NOT_READY')
    SCARD_E_INVALID_VALUE: Final[ThirtyTwoBits] = cached_hresult(0x80100011, 'SCARD_E_INVALID_VALUE')
    SCARD_E_SYSTEM_CANCELLED: Final[ThirtyTwoBits] = cached_hresult(0x80100012, 'SCARD_E_SYSTEM_CANCELLED')
    SCARD_F_COMM_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80100013, 'SCARD_F_COMM_ERROR')
    SCARD_F_UNKNOWN_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80100014, 'SCARD_F_UNKNOWN_ERROR')
    SCARD_E_INVALID_ATR: Final[ThirtyTwoBits] = cached_hresult(0x80100015, 'SCARD_E_INVALID_ATR')
    SCARD_E_NOT_TRANSACTED: Final[ThirtyTwoBits] = cached_hresult(0x80100016, 'SCARD_E_NOT_TRANSACTED')
    SCARD_E_READER_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80100017, 'SCARD_E_READER_UNAVAILABLE')
    SCARD_P_SHUTDOWN: Final[ThirtyTwoBits] = cached_hresult(0x80100018, 'SCARD_P_SHUTDOWN')
    SCARD_E_PCI_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0x80100019, 'SCARD_E_PCI_TOO_SMALL')
    SCARD_E_READER_UNSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8010001A, 'SCARD_E_READER_UNSUPPORTED')
    SCARD_E_DUPLICATE_READER: Final[ThirtyTwoBits] = cached_hresult(0x8010001B, 'SCARD_E_DUPLICATE_READER')
    SCARD_E_CARD_UNSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8010001C, 'SCARD_E_CARD_UNSUPPORTED')
    SCARD_E_NO_SERVICE: Final[ThirtyTwoBits] = cached_hresult(0x8010001D, 'SCARD_E_NO_SERVICE')
    SCARD_E_SERVICE_STOPPED: Final[ThirtyTwoBits] = cached_hresult(0x8010001E, 'SCARD_E_SERVICE_STOPPED')
    SCARD_E_UNEXPECTED: Final[ThirtyTwoBits] = cached_hresult(0x8010001F, 'SCARD_E_UNEXPECTED')
    SCARD_E_ICC_INSTALLATION: Final[ThirtyTwoBits] = cached_hresult(0x80100020, 'SCARD_E_ICC_INSTALLATION')
    SCARD_E_ICC_CREATEORDER: Final[ThirtyTwoBits] = cached_hresult(0x80100021, 'SCARD_E_ICC_CREATEORDER')
    SCARD_E_UNSUPPORTED_FEATURE: Final[ThirtyTwoBits] = cached_hresult(0x80100022, 'SCARD_E_UNSUPPORTED_FEATURE')
    SCARD_E_DIR_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80100023, 'SCARD_E_DIR_NOT_FOUND')
    SCARD_E_FILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80100024, 'SCARD_E_FILE_NOT_FOUND')
    SCARD_E_NO_DIR: Final[ThirtyTwoBits] = cached_hresult(0x80100025, 'SCARD_E_NO_DIR')
    SCARD_E_NO_FILE: Final[ThirtyTwoBits] = cached_hresult(0x80100026, 'SCARD_E_NO_FILE')
    SCARD_E_NO_ACCESS: Final[ThirtyTwoBits] = cached_hresult(0x80100027, 'SCARD_E_NO_ACCESS')
    SCARD_E_WRITE_TOO_MANY: Final[ThirtyTwoBits] = cached_hresult(0x80100028, 'SCARD_E_WRITE_TOO_MANY')
    SCARD_E_BAD_SEEK: Final[ThirtyTwoBits] = cached_hresult(0x80100029, 'SCARD_E_BAD_SEEK')
    SCARD_E_INVALID_CHV: Final[ThirtyTwoBits] = cached_hresult(0x8010002A, 'SCARD_E_INVALID_CHV')
    SCARD_E_UNKNOWN_RES_MNG: Final[ThirtyTwoBits] = cached_hresult(0x8010002B, 'SCARD_E_UNKNOWN_RES_MNG')
    SCARD_E_NO_SUCH_CERTIFICATE: Final[ThirtyTwoBits] = cached_hresult(0x8010002C, 'SCARD_E_NO_SUCH_CERTIFICATE')
    SCARD_E_CERTIFICATE_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x8010002D, 'SCARD_E_CERTIFICATE_UNAVAILABLE')
    SCARD_E_NO_READERS_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x8010002E, 'SCARD_E_NO_READERS_AVAILABLE')
    SCARD_E_COMM_DATA_LOST: Final[ThirtyTwoBits] = cached_hresult(0x8010002F, 'SCARD_E_COMM_DATA_LOST')
    SCARD_E_NO_KEY_CONTAINER: Final[ThirtyTwoBits] = cached_hresult(0x80100030, 'SCARD_E_NO_KEY_CONTAINER')
    SCARD_E_SERVER_TOO_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x80100031, 'SCARD_E_SERVER_TOO_BUSY')
    SCARD_E_PIN_CACHE_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x80100032, 'SCARD_E_PIN_CACHE_EXPIRED')
    SCARD_E_NO_PIN_CACHE: Final[ThirtyTwoBits] = cached_hresult(0x80100033, 'SCARD_E_NO_PIN_CACHE')
    SCARD_E_READ_ONLY_CARD: Final[ThirtyTwoBits] = cached_hresult(0x80100034, 'SCARD_E_READ_ONLY_CARD')
    SCARD_W_UNSUPPORTED_CARD: Final[ThirtyTwoBits] = cached_hresult(0x80100065, 'SCARD_W_UNSUPPORTED_CARD')
    SCARD_W_UNRESPONSIVE_CARD: Final[ThirtyTwoBits] = cached_hresult(0x80100066, 'SCARD_W_UNRESPONSIVE_CARD')
    SCARD_W_UNPOWERED_CARD: Final[ThirtyTwoBits] = cached_hresult(0x80100067, 'SCARD_W_UNPOWERED_CARD')
    SCARD_W_RESET_CARD: Final[ThirtyTwoBits] = cached_hresult(0x80100068, 'SCARD_W_RESET_CARD')
    SCARD_W_REMOVED_CARD: Final[ThirtyTwoBits] = cached_hresult(0x80100069, 'SCARD_W_REMOVED_CARD')
    SCARD_W_SECURITY_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x8010006A, 'SCARD_W_SECURITY_VIOLATION')
    SCARD_W_WRONG_CHV: Final[ThirtyTwoBits] = cached_hresult(0x8010006B, 'SCARD_W_WRONG_CHV')
    SCARD_W_CHV_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x8010006C, 'SCARD_W_CHV_BLOCKED')
    SCARD_W_EOF: Final[ThirtyTwoBits] = cached_hresult(0x8010006D, 'SCARD_W_EOF')
    SCARD_W_CANCELLED_BY_USER: Final[ThirtyTwoBits] = cached_hresult(0x8010006E, 'SCARD_W_CANCELLED_BY_USER')
    SCARD_W_CARD_NOT_AUTHENTICATED: Final[ThirtyTwoBits] = cached_hresult(0x8010006F, 'SCARD_W_CARD_NOT_AUTHENTICATED')
    SCARD_W_CACHE_ITEM_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80100070, 'SCARD_W_CACHE_ITEM_NOT_FOUND')
    SCARD_W_CACHE_ITEM_STALE: Final[ThirtyTwoBits] = cached_hresult(0x80100071, 'SCARD_W_CACHE_ITEM_STALE')
    SCARD_W_CACHE_ITEM_TOO_BIG: Final[ThirtyTwoBits] = cached_hresult(0x80100072, 'SCARD_W_CACHE_ITEM_TOO_BIG')
    COMADMIN_E_OBJECTERRORS: Final[ThirtyTwoBits] = cached_hresult(0x80110401, 'COMADMIN_E_OBJECTERRORS')
    COMADMIN_E_OBJECTINVALID: Final[ThirtyTwoBits] = cached_hresult(0x80110402, 'COMADMIN_E_OBJECTINVALID')
    COMADMIN_E_KEYMISSING: Final[ThirtyTwoBits] = cached_hresult(0x80110403, 'COMADMIN_E_KEYMISSING')
    COMADMIN_E_ALREADYINSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x80110404, 'COMADMIN_E_ALREADYINSTALLED')
    COMADMIN_E_APP_FILE_WRITEFAIL: Final[ThirtyTwoBits] = cached_hresult(0x80110407, 'COMADMIN_E_APP_FILE_WRITEFAIL')
    COMADMIN_E_APP_FILE_READFAIL: Final[ThirtyTwoBits] = cached_hresult(0x80110408, 'COMADMIN_E_APP_FILE_READFAIL')
    COMADMIN_E_APP_FILE_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80110409, 'COMADMIN_E_APP_FILE_VERSION')
    COMADMIN_E_BADPATH: Final[ThirtyTwoBits] = cached_hresult(0x8011040A, 'COMADMIN_E_BADPATH')
    COMADMIN_E_APPLICATIONEXISTS: Final[ThirtyTwoBits] = cached_hresult(0x8011040B, 'COMADMIN_E_APPLICATIONEXISTS')
    COMADMIN_E_ROLEEXISTS: Final[ThirtyTwoBits] = cached_hresult(0x8011040C, 'COMADMIN_E_ROLEEXISTS')
    COMADMIN_E_CANTCOPYFILE: Final[ThirtyTwoBits] = cached_hresult(0x8011040D, 'COMADMIN_E_CANTCOPYFILE')
    COMADMIN_E_NOUSER: Final[ThirtyTwoBits] = cached_hresult(0x8011040F, 'COMADMIN_E_NOUSER')
    COMADMIN_E_INVALIDUSERIDS: Final[ThirtyTwoBits] = cached_hresult(0x80110410, 'COMADMIN_E_INVALIDUSERIDS')
    COMADMIN_E_NOREGISTRYCLSID: Final[ThirtyTwoBits] = cached_hresult(0x80110411, 'COMADMIN_E_NOREGISTRYCLSID')
    COMADMIN_E_BADREGISTRYPROGID: Final[ThirtyTwoBits] = cached_hresult(0x80110412, 'COMADMIN_E_BADREGISTRYPROGID')
    COMADMIN_E_AUTHENTICATIONLEVEL: Final[ThirtyTwoBits] = cached_hresult(0x80110413, 'COMADMIN_E_AUTHENTICATIONLEVEL')
    COMADMIN_E_USERPASSWDNOTVALID: Final[ThirtyTwoBits] = cached_hresult(0x80110414, 'COMADMIN_E_USERPASSWDNOTVALID')
    COMADMIN_E_CLSIDORIIDMISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80110418, 'COMADMIN_E_CLSIDORIIDMISMATCH')
    COMADMIN_E_REMOTEINTERFACE: Final[ThirtyTwoBits] = cached_hresult(0x80110419, 'COMADMIN_E_REMOTEINTERFACE')
    COMADMIN_E_DLLREGISTERSERVER: Final[ThirtyTwoBits] = cached_hresult(0x8011041A, 'COMADMIN_E_DLLREGISTERSERVER')
    COMADMIN_E_NOSERVERSHARE: Final[ThirtyTwoBits] = cached_hresult(0x8011041B, 'COMADMIN_E_NOSERVERSHARE')
    COMADMIN_E_DLLLOADFAILED: Final[ThirtyTwoBits] = cached_hresult(0x8011041D, 'COMADMIN_E_DLLLOADFAILED')
    COMADMIN_E_BADREGISTRYLIBID: Final[ThirtyTwoBits] = cached_hresult(0x8011041E, 'COMADMIN_E_BADREGISTRYLIBID')
    COMADMIN_E_APPDIRNOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x8011041F, 'COMADMIN_E_APPDIRNOTFOUND')
    COMADMIN_E_REGISTRARFAILED: Final[ThirtyTwoBits] = cached_hresult(0x80110423, 'COMADMIN_E_REGISTRARFAILED')
    COMADMIN_E_COMPFILE_DOESNOTEXIST: Final[ThirtyTwoBits] = cached_hresult(0x80110424, 'COMADMIN_E_COMPFILE_DOESNOTEXIST')
    COMADMIN_E_COMPFILE_LOADDLLFAIL: Final[ThirtyTwoBits] = cached_hresult(0x80110425, 'COMADMIN_E_COMPFILE_LOADDLLFAIL')
    COMADMIN_E_COMPFILE_GETCLASSOBJ: Final[ThirtyTwoBits] = cached_hresult(0x80110426, 'COMADMIN_E_COMPFILE_GETCLASSOBJ')
    COMADMIN_E_COMPFILE_CLASSNOTAVAIL: Final[ThirtyTwoBits] = cached_hresult(0x80110427, 'COMADMIN_E_COMPFILE_CLASSNOTAVAIL')
    COMADMIN_E_COMPFILE_BADTLB: Final[ThirtyTwoBits] = cached_hresult(0x80110428, 'COMADMIN_E_COMPFILE_BADTLB')
    COMADMIN_E_COMPFILE_NOTINSTALLABLE: Final[ThirtyTwoBits] = cached_hresult(0x80110429, 'COMADMIN_E_COMPFILE_NOTINSTALLABLE')
    COMADMIN_E_NOTCHANGEABLE: Final[ThirtyTwoBits] = cached_hresult(0x8011042A, 'COMADMIN_E_NOTCHANGEABLE')
    COMADMIN_E_NOTDELETEABLE: Final[ThirtyTwoBits] = cached_hresult(0x8011042B, 'COMADMIN_E_NOTDELETEABLE')
    COMADMIN_E_SESSION: Final[ThirtyTwoBits] = cached_hresult(0x8011042C, 'COMADMIN_E_SESSION')
    COMADMIN_E_COMP_MOVE_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x8011042D, 'COMADMIN_E_COMP_MOVE_LOCKED')
    COMADMIN_E_COMP_MOVE_BAD_DEST: Final[ThirtyTwoBits] = cached_hresult(0x8011042E, 'COMADMIN_E_COMP_MOVE_BAD_DEST')
    COMADMIN_E_REGISTERTLB: Final[ThirtyTwoBits] = cached_hresult(0x80110430, 'COMADMIN_E_REGISTERTLB')
    COMADMIN_E_SYSTEMAPP: Final[ThirtyTwoBits] = cached_hresult(0x80110433, 'COMADMIN_E_SYSTEMAPP')
    COMADMIN_E_COMPFILE_NOREGISTRAR: Final[ThirtyTwoBits] = cached_hresult(0x80110434, 'COMADMIN_E_COMPFILE_NOREGISTRAR')
    COMADMIN_E_COREQCOMPINSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x80110435, 'COMADMIN_E_COREQCOMPINSTALLED')
    COMADMIN_E_SERVICENOTINSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x80110436, 'COMADMIN_E_SERVICENOTINSTALLED')
    COMADMIN_E_PROPERTYSAVEFAILED: Final[ThirtyTwoBits] = cached_hresult(0x80110437, 'COMADMIN_E_PROPERTYSAVEFAILED')
    COMADMIN_E_OBJECTEXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80110438, 'COMADMIN_E_OBJECTEXISTS')
    COMADMIN_E_COMPONENTEXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80110439, 'COMADMIN_E_COMPONENTEXISTS')
    COMADMIN_E_REGFILE_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x8011043B, 'COMADMIN_E_REGFILE_CORRUPT')
    COMADMIN_E_PROPERTY_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0x8011043C, 'COMADMIN_E_PROPERTY_OVERFLOW')
    COMADMIN_E_NOTINREGISTRY: Final[ThirtyTwoBits] = cached_hresult(0x8011043E, 'COMADMIN_E_NOTINREGISTRY')
    COMADMIN_E_OBJECTNOTPOOLABLE: Final[ThirtyTwoBits] = cached_hresult(0x8011043F, 'COMADMIN_E_OBJECTNOTPOOLABLE')
    COMADMIN_E_APPLID_MATCHES_CLSID: Final[ThirtyTwoBits] = cached_hresult(0x80110446, 'COMADMIN_E_APPLID_MATCHES_CLSID')
    COMADMIN_E_ROLE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x80110447, 'COMADMIN_E_ROLE_DOES_NOT_EXIST')
    COMADMIN_E_START_APP_NEEDS_COMPONENTS: Final[ThirtyTwoBits] = cached_hresult(0x80110448, 'COMADMIN_E_START_APP_NEEDS_COMPONENTS')
    COMADMIN_E_REQUIRES_DIFFERENT_PLATFORM: Final[ThirtyTwoBits] = cached_hresult(0x80110449, 'COMADMIN_E_REQUIRES_DIFFERENT_PLATFORM')
    COMADMIN_E_CAN_NOT_EXPORT_APP_PROXY: Final[ThirtyTwoBits] = cached_hresult(0x8011044A, 'COMADMIN_E_CAN_NOT_EXPORT_APP_PROXY')
    COMADMIN_E_CAN_NOT_START_APP: Final[ThirtyTwoBits] = cached_hresult(0x8011044B, 'COMADMIN_E_CAN_NOT_START_APP')
    COMADMIN_E_CAN_NOT_EXPORT_SYS_APP: Final[ThirtyTwoBits] = cached_hresult(0x8011044C, 'COMADMIN_E_CAN_NOT_EXPORT_SYS_APP')
    COMADMIN_E_CANT_SUBSCRIBE_TO_COMPONENT: Final[ThirtyTwoBits] = cached_hresult(0x8011044D, 'COMADMIN_E_CANT_SUBSCRIBE_TO_COMPONENT')
    COMADMIN_E_EVENTCLASS_CANT_BE_SUBSCRIBER: Final[ThirtyTwoBits] = cached_hresult(0x8011044E, 'COMADMIN_E_EVENTCLASS_CANT_BE_SUBSCRIBER')
    COMADMIN_E_LIB_APP_PROXY_INCOMPATIBLE: Final[ThirtyTwoBits] = cached_hresult(0x8011044F, 'COMADMIN_E_LIB_APP_PROXY_INCOMPATIBLE')
    COMADMIN_E_BASE_PARTITION_ONLY: Final[ThirtyTwoBits] = cached_hresult(0x80110450, 'COMADMIN_E_BASE_PARTITION_ONLY')
    COMADMIN_E_START_APP_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80110451, 'COMADMIN_E_START_APP_DISABLED')
    COMADMIN_E_CAT_DUPLICATE_PARTITION_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80110457, 'COMADMIN_E_CAT_DUPLICATE_PARTITION_NAME')
    COMADMIN_E_CAT_INVALID_PARTITION_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80110458, 'COMADMIN_E_CAT_INVALID_PARTITION_NAME')
    COMADMIN_E_CAT_PARTITION_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x80110459, 'COMADMIN_E_CAT_PARTITION_IN_USE')
    COMADMIN_E_FILE_PARTITION_DUPLICATE_FILES: Final[ThirtyTwoBits] = cached_hresult(0x8011045A, 'COMADMIN_E_FILE_PARTITION_DUPLICATE_FILES')
    COMADMIN_E_CAT_IMPORTED_COMPONENTS_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x8011045B, 'COMADMIN_E_CAT_IMPORTED_COMPONENTS_NOT_ALLOWED')
    COMADMIN_E_AMBIGUOUS_APPLICATION_NAME: Final[ThirtyTwoBits] = cached_hresult(0x8011045C, 'COMADMIN_E_AMBIGUOUS_APPLICATION_NAME')
    COMADMIN_E_AMBIGUOUS_PARTITION_NAME: Final[ThirtyTwoBits] = cached_hresult(0x8011045D, 'COMADMIN_E_AMBIGUOUS_PARTITION_NAME')
    COMADMIN_E_REGDB_NOTINITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0x80110472, 'COMADMIN_E_REGDB_NOTINITIALIZED')
    COMADMIN_E_REGDB_NOTOPEN: Final[ThirtyTwoBits] = cached_hresult(0x80110473, 'COMADMIN_E_REGDB_NOTOPEN')
    COMADMIN_E_REGDB_SYSTEMERR: Final[ThirtyTwoBits] = cached_hresult(0x80110474, 'COMADMIN_E_REGDB_SYSTEMERR')
    COMADMIN_E_REGDB_ALREADYRUNNING: Final[ThirtyTwoBits] = cached_hresult(0x80110475, 'COMADMIN_E_REGDB_ALREADYRUNNING')
    COMADMIN_E_MIG_VERSIONNOTSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80110480, 'COMADMIN_E_MIG_VERSIONNOTSUPPORTED')
    COMADMIN_E_MIG_SCHEMANOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x80110481, 'COMADMIN_E_MIG_SCHEMANOTFOUND')
    COMADMIN_E_CAT_BITNESSMISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80110482, 'COMADMIN_E_CAT_BITNESSMISMATCH')
    COMADMIN_E_CAT_UNACCEPTABLEBITNESS: Final[ThirtyTwoBits] = cached_hresult(0x80110483, 'COMADMIN_E_CAT_UNACCEPTABLEBITNESS')
    COMADMIN_E_CAT_WRONGAPPBITNESS: Final[ThirtyTwoBits] = cached_hresult(0x80110484, 'COMADMIN_E_CAT_WRONGAPPBITNESS')
    COMADMIN_E_CAT_PAUSE_RESUME_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80110485, 'COMADMIN_E_CAT_PAUSE_RESUME_NOT_SUPPORTED')
    COMADMIN_E_CAT_SERVERFAULT: Final[ThirtyTwoBits] = cached_hresult(0x80110486, 'COMADMIN_E_CAT_SERVERFAULT')
    COMQC_E_APPLICATION_NOT_QUEUED: Final[ThirtyTwoBits] = cached_hresult(0x80110600, 'COMQC_E_APPLICATION_NOT_QUEUED')
    COMQC_E_NO_QUEUEABLE_INTERFACES: Final[ThirtyTwoBits] = cached_hresult(0x80110601, 'COMQC_E_NO_QUEUEABLE_INTERFACES')
    COMQC_E_QUEUING_SERVICE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80110602, 'COMQC_E_QUEUING_SERVICE_NOT_AVAILABLE')
    COMQC_E_NO_IPERSISTSTREAM: Final[ThirtyTwoBits] = cached_hresult(0x80110603, 'COMQC_E_NO_IPERSISTSTREAM')
    COMQC_E_BAD_MESSAGE: Final[ThirtyTwoBits] = cached_hresult(0x80110604, 'COMQC_E_BAD_MESSAGE')
    COMQC_E_UNAUTHENTICATED: Final[ThirtyTwoBits] = cached_hresult(0x80110605, 'COMQC_E_UNAUTHENTICATED')
    COMQC_E_UNTRUSTED_ENQUEUER: Final[ThirtyTwoBits] = cached_hresult(0x80110606, 'COMQC_E_UNTRUSTED_ENQUEUER')
    MSDTC_E_DUPLICATE_RESOURCE: Final[ThirtyTwoBits] = cached_hresult(0x80110701, 'MSDTC_E_DUPLICATE_RESOURCE')
    COMADMIN_E_OBJECT_PARENT_MISSING: Final[ThirtyTwoBits] = cached_hresult(0x80110808, 'COMADMIN_E_OBJECT_PARENT_MISSING')
    COMADMIN_E_OBJECT_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x80110809, 'COMADMIN_E_OBJECT_DOES_NOT_EXIST')
    COMADMIN_E_APP_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x8011080A, 'COMADMIN_E_APP_NOT_RUNNING')
    COMADMIN_E_INVALID_PARTITION: Final[ThirtyTwoBits] = cached_hresult(0x8011080B, 'COMADMIN_E_INVALID_PARTITION')
    COMADMIN_E_SVCAPP_NOT_POOLABLE_OR_RECYCLABLE: Final[ThirtyTwoBits] = cached_hresult(0x8011080D, 'COMADMIN_E_SVCAPP_NOT_POOLABLE_OR_RECYCLABLE')
    COMADMIN_E_USER_IN_SET: Final[ThirtyTwoBits] = cached_hresult(0x8011080E, 'COMADMIN_E_USER_IN_SET')
    COMADMIN_E_CANTRECYCLELIBRARYAPPS: Final[ThirtyTwoBits] = cached_hresult(0x8011080F, 'COMADMIN_E_CANTRECYCLELIBRARYAPPS')
    COMADMIN_E_CANTRECYCLESERVICEAPPS: Final[ThirtyTwoBits] = cached_hresult(0x80110811, 'COMADMIN_E_CANTRECYCLESERVICEAPPS')
    COMADMIN_E_PROCESSALREADYRECYCLED: Final[ThirtyTwoBits] = cached_hresult(0x80110812, 'COMADMIN_E_PROCESSALREADYRECYCLED')
    COMADMIN_E_PAUSEDPROCESSMAYNOTBERECYCLED: Final[ThirtyTwoBits] = cached_hresult(0x80110813, 'COMADMIN_E_PAUSEDPROCESSMAYNOTBERECYCLED')
    COMADMIN_E_CANTMAKEINPROCSERVICE: Final[ThirtyTwoBits] = cached_hresult(0x80110814, 'COMADMIN_E_CANTMAKEINPROCSERVICE')
    COMADMIN_E_PROGIDINUSEBYCLSID: Final[ThirtyTwoBits] = cached_hresult(0x80110815, 'COMADMIN_E_PROGIDINUSEBYCLSID')
    COMADMIN_E_DEFAULT_PARTITION_NOT_IN_SET: Final[ThirtyTwoBits] = cached_hresult(0x80110816, 'COMADMIN_E_DEFAULT_PARTITION_NOT_IN_SET')
    COMADMIN_E_RECYCLEDPROCESSMAYNOTBEPAUSED: Final[ThirtyTwoBits] = cached_hresult(0x80110817, 'COMADMIN_E_RECYCLEDPROCESSMAYNOTBEPAUSED')
    COMADMIN_E_PARTITION_ACCESSDENIED: Final[ThirtyTwoBits] = cached_hresult(0x80110818, 'COMADMIN_E_PARTITION_ACCESSDENIED')
    COMADMIN_E_PARTITION_MSI_ONLY: Final[ThirtyTwoBits] = cached_hresult(0x80110819, 'COMADMIN_E_PARTITION_MSI_ONLY')
    COMADMIN_E_LEGACYCOMPS_NOT_ALLOWED_IN_1_0_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0x8011081A, 'COMADMIN_E_LEGACYCOMPS_NOT_ALLOWED_IN_1_0_FORMAT')
    COMADMIN_E_LEGACYCOMPS_NOT_ALLOWED_IN_NONBASE_PARTITIONS: Final[ThirtyTwoBits] = cached_hresult(0x8011081B, 'COMADMIN_E_LEGACYCOMPS_NOT_ALLOWED_IN_NONBASE_PARTITIONS')
    COMADMIN_E_COMP_MOVE_SOURCE: Final[ThirtyTwoBits] = cached_hresult(0x8011081C, 'COMADMIN_E_COMP_MOVE_SOURCE')
    COMADMIN_E_COMP_MOVE_DEST: Final[ThirtyTwoBits] = cached_hresult(0x8011081D, 'COMADMIN_E_COMP_MOVE_DEST')
    COMADMIN_E_COMP_MOVE_PRIVATE: Final[ThirtyTwoBits] = cached_hresult(0x8011081E, 'COMADMIN_E_COMP_MOVE_PRIVATE')
    COMADMIN_E_BASEPARTITION_REQUIRED_IN_SET: Final[ThirtyTwoBits] = cached_hresult(0x8011081F, 'COMADMIN_E_BASEPARTITION_REQUIRED_IN_SET')
    COMADMIN_E_CANNOT_ALIAS_EVENTCLASS: Final[ThirtyTwoBits] = cached_hresult(0x80110820, 'COMADMIN_E_CANNOT_ALIAS_EVENTCLASS')
    COMADMIN_E_PRIVATE_ACCESSDENIED: Final[ThirtyTwoBits] = cached_hresult(0x80110821, 'COMADMIN_E_PRIVATE_ACCESSDENIED')
    COMADMIN_E_SAFERINVALID: Final[ThirtyTwoBits] = cached_hresult(0x80110822, 'COMADMIN_E_SAFERINVALID')
    COMADMIN_E_REGISTRY_ACCESSDENIED: Final[ThirtyTwoBits] = cached_hresult(0x80110823, 'COMADMIN_E_REGISTRY_ACCESSDENIED')
    COMADMIN_E_PARTITIONS_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80110824, 'COMADMIN_E_PARTITIONS_DISABLED')
    MENROLL_E_DEVICE_MESSAGE_FORMAT_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80180001, 'MENROLL_E_DEVICE_MESSAGE_FORMAT_ERROR')
    MENROLL_E_DEVICE_AUTHENTICATION_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80180002, 'MENROLL_E_DEVICE_AUTHENTICATION_ERROR')
    MENROLL_E_DEVICE_AUTHORIZATION_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80180003, 'MENROLL_E_DEVICE_AUTHORIZATION_ERROR')
    MENROLL_E_DEVICE_CERTIFICATEREQUEST_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80180004, 'MENROLL_E_DEVICE_CERTIFICATEREQUEST_ERROR')
    MENROLL_E_DEVICE_CONFIGMGRSERVER_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80180005, 'MENROLL_E_DEVICE_CONFIGMGRSERVER_ERROR')
    MENROLL_E_DEVICE_INTERNALSERVICE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80180006, 'MENROLL_E_DEVICE_INTERNALSERVICE_ERROR')
    MENROLL_E_DEVICE_INVALIDSECURITY_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80180007, 'MENROLL_E_DEVICE_INVALIDSECURITY_ERROR')
    MENROLL_E_DEVICE_UNKNOWN_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80180008, 'MENROLL_E_DEVICE_UNKNOWN_ERROR')
    MENROLL_E_ENROLLMENT_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x80180009, 'MENROLL_E_ENROLLMENT_IN_PROGRESS')
    MENROLL_E_DEVICE_ALREADY_ENROLLED: Final[ThirtyTwoBits] = cached_hresult(0x8018000A, 'MENROLL_E_DEVICE_ALREADY_ENROLLED')
    MENROLL_E_DISCOVERY_SEC_CERT_DATE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x8018000D, 'MENROLL_E_DISCOVERY_SEC_CERT_DATE_INVALID')
    MENROLL_E_PASSWORD_NEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8018000E, 'MENROLL_E_PASSWORD_NEEDED')
    MENROLL_E_WAB_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8018000F, 'MENROLL_E_WAB_ERROR')
    MENROLL_E_CONNECTIVITY: Final[ThirtyTwoBits] = cached_hresult(0x80180010, 'MENROLL_E_CONNECTIVITY')
    MENROLL_S_ENROLLMENT_SUSPENDED: Final[ThirtyTwoBits] = cached_hresult(0x00180011, 'MENROLL_S_ENROLLMENT_SUSPENDED')
    MENROLL_E_INVALIDSSLCERT: Final[ThirtyTwoBits] = cached_hresult(0x80180012, 'MENROLL_E_INVALIDSSLCERT')
    MENROLL_E_DEVICECAPREACHED: Final[ThirtyTwoBits] = cached_hresult(0x80180013, 'MENROLL_E_DEVICECAPREACHED')
    MENROLL_E_DEVICENOTSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80180014, 'MENROLL_E_DEVICENOTSUPPORTED')
    MENROLL_E_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80180015, 'MENROLL_E_NOT_SUPPORTED')
    MENROLL_E_NOTELIGIBLETORENEW: Final[ThirtyTwoBits] = cached_hresult(0x80180016, 'MENROLL_E_NOTELIGIBLETORENEW')
    MENROLL_E_INMAINTENANCE: Final[ThirtyTwoBits] = cached_hresult(0x80180017, 'MENROLL_E_INMAINTENANCE')
    MENROLL_E_USER_LICENSE: Final[ThirtyTwoBits] = cached_hresult(0x80180018, 'MENROLL_E_USER_LICENSE')
    MENROLL_E_ENROLLMENTDATAINVALID: Final[ThirtyTwoBits] = cached_hresult(0x80180019, 'MENROLL_E_ENROLLMENTDATAINVALID')
    MENROLL_E_INSECUREREDIRECT: Final[ThirtyTwoBits] = cached_hresult(0x8018001A, 'MENROLL_E_INSECUREREDIRECT')
    MENROLL_E_PLATFORM_WRONG_STATE: Final[ThirtyTwoBits] = cached_hresult(0x8018001B, 'MENROLL_E_PLATFORM_WRONG_STATE')
    MENROLL_E_PLATFORM_LICENSE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8018001C, 'MENROLL_E_PLATFORM_LICENSE_ERROR')
    MENROLL_E_PLATFORM_UNKNOWN_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8018001D, 'MENROLL_E_PLATFORM_UNKNOWN_ERROR')
    MENROLL_E_PROV_CSP_CERTSTORE: Final[ThirtyTwoBits] = cached_hresult(0x8018001E, 'MENROLL_E_PROV_CSP_CERTSTORE')
    MENROLL_E_PROV_CSP_W7: Final[ThirtyTwoBits] = cached_hresult(0x8018001F, 'MENROLL_E_PROV_CSP_W7')
    MENROLL_E_PROV_CSP_DMCLIENT: Final[ThirtyTwoBits] = cached_hresult(0x80180020, 'MENROLL_E_PROV_CSP_DMCLIENT')
    MENROLL_E_PROV_CSP_PFW: Final[ThirtyTwoBits] = cached_hresult(0x80180021, 'MENROLL_E_PROV_CSP_PFW')
    MENROLL_E_PROV_CSP_MISC: Final[ThirtyTwoBits] = cached_hresult(0x80180022, 'MENROLL_E_PROV_CSP_MISC')
    MENROLL_E_PROV_UNKNOWN: Final[ThirtyTwoBits] = cached_hresult(0x80180023, 'MENROLL_E_PROV_UNKNOWN')
    MENROLL_E_PROV_SSLCERTNOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x80180024, 'MENROLL_E_PROV_SSLCERTNOTFOUND')
    MENROLL_E_PROV_CSP_APPMGMT: Final[ThirtyTwoBits] = cached_hresult(0x80180025, 'MENROLL_E_PROV_CSP_APPMGMT')
    MENROLL_E_DEVICE_MANAGEMENT_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x80180026, 'MENROLL_E_DEVICE_MANAGEMENT_BLOCKED')
    MENROLL_E_CERTPOLICY_PRIVATEKEYCREATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80180027, 'MENROLL_E_CERTPOLICY_PRIVATEKEYCREATION_FAILED')
    MENROLL_E_CERTAUTH_FAILED_TO_FIND_CERT: Final[ThirtyTwoBits] = cached_hresult(0x80180028, 'MENROLL_E_CERTAUTH_FAILED_TO_FIND_CERT')
    MENROLL_E_EMPTY_MESSAGE: Final[ThirtyTwoBits] = cached_hresult(0x80180029, 'MENROLL_E_EMPTY_MESSAGE')
    MENROLL_E_USER_CANCELLED: Final[ThirtyTwoBits] = cached_hresult(0x80180030, 'MENROLL_E_USER_CANCELLED')
    MENROLL_E_MDM_NOT_CONFIGURED: Final[ThirtyTwoBits] = cached_hresult(0x80180031, 'MENROLL_E_MDM_NOT_CONFIGURED')
    MENROLL_E_CUSTOMSERVERERROR: Final[ThirtyTwoBits] = cached_hresult(0x80180032, 'MENROLL_E_CUSTOMSERVERERROR')
    HTTP_E_STATUS_UNEXPECTED: Final[ThirtyTwoBits] = cached_hresult(0x80190001, 'HTTP_E_STATUS_UNEXPECTED')
    HTTP_E_STATUS_UNEXPECTED_REDIRECTION: Final[ThirtyTwoBits] = cached_hresult(0x80190003, 'HTTP_E_STATUS_UNEXPECTED_REDIRECTION')
    HTTP_E_STATUS_UNEXPECTED_CLIENT_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80190004, 'HTTP_E_STATUS_UNEXPECTED_CLIENT_ERROR')
    HTTP_E_STATUS_UNEXPECTED_SERVER_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80190005, 'HTTP_E_STATUS_UNEXPECTED_SERVER_ERROR')
    HTTP_E_STATUS_AMBIGUOUS: Final[ThirtyTwoBits] = cached_hresult(0x8019012C, 'HTTP_E_STATUS_AMBIGUOUS')
    HTTP_E_STATUS_MOVED: Final[ThirtyTwoBits] = cached_hresult(0x8019012D, 'HTTP_E_STATUS_MOVED')
    HTTP_E_STATUS_REDIRECT: Final[ThirtyTwoBits] = cached_hresult(0x8019012E, 'HTTP_E_STATUS_REDIRECT')
    HTTP_E_STATUS_REDIRECT_METHOD: Final[ThirtyTwoBits] = cached_hresult(0x8019012F, 'HTTP_E_STATUS_REDIRECT_METHOD')
    HTTP_E_STATUS_NOT_MODIFIED: Final[ThirtyTwoBits] = cached_hresult(0x80190130, 'HTTP_E_STATUS_NOT_MODIFIED')
    HTTP_E_STATUS_USE_PROXY: Final[ThirtyTwoBits] = cached_hresult(0x80190131, 'HTTP_E_STATUS_USE_PROXY')
    HTTP_E_STATUS_REDIRECT_KEEP_VERB: Final[ThirtyTwoBits] = cached_hresult(0x80190133, 'HTTP_E_STATUS_REDIRECT_KEEP_VERB')
    HTTP_E_STATUS_BAD_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0x80190190, 'HTTP_E_STATUS_BAD_REQUEST')
    HTTP_E_STATUS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x80190191, 'HTTP_E_STATUS_DENIED')
    HTTP_E_STATUS_PAYMENT_REQ: Final[ThirtyTwoBits] = cached_hresult(0x80190192, 'HTTP_E_STATUS_PAYMENT_REQ')
    HTTP_E_STATUS_FORBIDDEN: Final[ThirtyTwoBits] = cached_hresult(0x80190193, 'HTTP_E_STATUS_FORBIDDEN')
    HTTP_E_STATUS_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80190194, 'HTTP_E_STATUS_NOT_FOUND')
    HTTP_E_STATUS_BAD_METHOD: Final[ThirtyTwoBits] = cached_hresult(0x80190195, 'HTTP_E_STATUS_BAD_METHOD')
    HTTP_E_STATUS_NONE_ACCEPTABLE: Final[ThirtyTwoBits] = cached_hresult(0x80190196, 'HTTP_E_STATUS_NONE_ACCEPTABLE')
    HTTP_E_STATUS_PROXY_AUTH_REQ: Final[ThirtyTwoBits] = cached_hresult(0x80190197, 'HTTP_E_STATUS_PROXY_AUTH_REQ')
    HTTP_E_STATUS_REQUEST_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x80190198, 'HTTP_E_STATUS_REQUEST_TIMEOUT')
    HTTP_E_STATUS_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x80190199, 'HTTP_E_STATUS_CONFLICT')
    HTTP_E_STATUS_GONE: Final[ThirtyTwoBits] = cached_hresult(0x8019019A, 'HTTP_E_STATUS_GONE')
    HTTP_E_STATUS_LENGTH_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8019019B, 'HTTP_E_STATUS_LENGTH_REQUIRED')
    HTTP_E_STATUS_PRECOND_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8019019C, 'HTTP_E_STATUS_PRECOND_FAILED')
    HTTP_E_STATUS_REQUEST_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0x8019019D, 'HTTP_E_STATUS_REQUEST_TOO_LARGE')
    HTTP_E_STATUS_URI_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0x8019019E, 'HTTP_E_STATUS_URI_TOO_LONG')
    HTTP_E_STATUS_UNSUPPORTED_MEDIA: Final[ThirtyTwoBits] = cached_hresult(0x8019019F, 'HTTP_E_STATUS_UNSUPPORTED_MEDIA')
    HTTP_E_STATUS_RANGE_NOT_SATISFIABLE: Final[ThirtyTwoBits] = cached_hresult(0x801901A0, 'HTTP_E_STATUS_RANGE_NOT_SATISFIABLE')
    HTTP_E_STATUS_EXPECTATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x801901A1, 'HTTP_E_STATUS_EXPECTATION_FAILED')
    HTTP_E_STATUS_SERVER_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x801901F4, 'HTTP_E_STATUS_SERVER_ERROR')
    HTTP_E_STATUS_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x801901F5, 'HTTP_E_STATUS_NOT_SUPPORTED')
    HTTP_E_STATUS_BAD_GATEWAY: Final[ThirtyTwoBits] = cached_hresult(0x801901F6, 'HTTP_E_STATUS_BAD_GATEWAY')
    HTTP_E_STATUS_SERVICE_UNAVAIL: Final[ThirtyTwoBits] = cached_hresult(0x801901F7, 'HTTP_E_STATUS_SERVICE_UNAVAIL')
    HTTP_E_STATUS_GATEWAY_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x801901F8, 'HTTP_E_STATUS_GATEWAY_TIMEOUT')
    HTTP_E_STATUS_VERSION_NOT_SUP: Final[ThirtyTwoBits] = cached_hresult(0x801901F9, 'HTTP_E_STATUS_VERSION_NOT_SUP')
    WER_S_REPORT_DEBUG: Final[ThirtyTwoBits] = cached_hresult(0x001B0000, 'WER_S_REPORT_DEBUG')
    WER_S_REPORT_UPLOADED: Final[ThirtyTwoBits] = cached_hresult(0x001B0001, 'WER_S_REPORT_UPLOADED')
    WER_S_REPORT_QUEUED: Final[ThirtyTwoBits] = cached_hresult(0x001B0002, 'WER_S_REPORT_QUEUED')
    WER_S_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x001B0003, 'WER_S_DISABLED')
    WER_S_SUSPENDED_UPLOAD: Final[ThirtyTwoBits] = cached_hresult(0x001B0004, 'WER_S_SUSPENDED_UPLOAD')
    WER_S_DISABLED_QUEUE: Final[ThirtyTwoBits] = cached_hresult(0x001B0005, 'WER_S_DISABLED_QUEUE')
    WER_S_DISABLED_ARCHIVE: Final[ThirtyTwoBits] = cached_hresult(0x001B0006, 'WER_S_DISABLED_ARCHIVE')
    WER_S_REPORT_ASYNC: Final[ThirtyTwoBits] = cached_hresult(0x001B0007, 'WER_S_REPORT_ASYNC')
    WER_S_IGNORE_ASSERT_INSTANCE: Final[ThirtyTwoBits] = cached_hresult(0x001B0008, 'WER_S_IGNORE_ASSERT_INSTANCE')
    WER_S_IGNORE_ALL_ASSERTS: Final[ThirtyTwoBits] = cached_hresult(0x001B0009, 'WER_S_IGNORE_ALL_ASSERTS')
    WER_S_ASSERT_CONTINUE: Final[ThirtyTwoBits] = cached_hresult(0x001B000A, 'WER_S_ASSERT_CONTINUE')
    WER_S_THROTTLED: Final[ThirtyTwoBits] = cached_hresult(0x001B000B, 'WER_S_THROTTLED')
    WER_S_REPORT_UPLOADED_CAB: Final[ThirtyTwoBits] = cached_hresult(0x001B000C, 'WER_S_REPORT_UPLOADED_CAB')
    WER_E_CRASH_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x801B8000, 'WER_E_CRASH_FAILURE')
    WER_E_CANCELED: Final[ThirtyTwoBits] = cached_hresult(0x801B8001, 'WER_E_CANCELED')
    WER_E_NETWORK_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x801B8002, 'WER_E_NETWORK_FAILURE')
    WER_E_NOT_INITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0x801B8003, 'WER_E_NOT_INITIALIZED')
    WER_E_ALREADY_REPORTING: Final[ThirtyTwoBits] = cached_hresult(0x801B8004, 'WER_E_ALREADY_REPORTING')
    WER_E_DUMP_THROTTLED: Final[ThirtyTwoBits] = cached_hresult(0x801B8005, 'WER_E_DUMP_THROTTLED')
    WER_E_INSUFFICIENT_CONSENT: Final[ThirtyTwoBits] = cached_hresult(0x801B8006, 'WER_E_INSUFFICIENT_CONSENT')
    WER_E_TOO_HEAVY: Final[ThirtyTwoBits] = cached_hresult(0x801B8007, 'WER_E_TOO_HEAVY')
    ERROR_FLT_IO_COMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x001F0001, 'ERROR_FLT_IO_COMPLETE')
    ERROR_FLT_NO_HANDLER_DEFINED: Final[ThirtyTwoBits] = cached_hresult(0x801F0001, 'ERROR_FLT_NO_HANDLER_DEFINED')
    ERROR_FLT_CONTEXT_ALREADY_DEFINED: Final[ThirtyTwoBits] = cached_hresult(0x801F0002, 'ERROR_FLT_CONTEXT_ALREADY_DEFINED')
    ERROR_FLT_INVALID_ASYNCHRONOUS_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0x801F0003, 'ERROR_FLT_INVALID_ASYNCHRONOUS_REQUEST')
    ERROR_FLT_DISALLOW_FAST_IO: Final[ThirtyTwoBits] = cached_hresult(0x801F0004, 'ERROR_FLT_DISALLOW_FAST_IO')
    ERROR_FLT_INVALID_NAME_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0x801F0005, 'ERROR_FLT_INVALID_NAME_REQUEST')
    ERROR_FLT_NOT_SAFE_TO_POST_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x801F0006, 'ERROR_FLT_NOT_SAFE_TO_POST_OPERATION')
    ERROR_FLT_NOT_INITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0x801F0007, 'ERROR_FLT_NOT_INITIALIZED')
    ERROR_FLT_FILTER_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0x801F0008, 'ERROR_FLT_FILTER_NOT_READY')
    ERROR_FLT_POST_OPERATION_CLEANUP: Final[ThirtyTwoBits] = cached_hresult(0x801F0009, 'ERROR_FLT_POST_OPERATION_CLEANUP')
    ERROR_FLT_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x801F000A, 'ERROR_FLT_INTERNAL_ERROR')
    ERROR_FLT_DELETING_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x801F000B, 'ERROR_FLT_DELETING_OBJECT')
    ERROR_FLT_MUST_BE_NONPAGED_POOL: Final[ThirtyTwoBits] = cached_hresult(0x801F000C, 'ERROR_FLT_MUST_BE_NONPAGED_POOL')
    ERROR_FLT_DUPLICATE_ENTRY: Final[ThirtyTwoBits] = cached_hresult(0x801F000D, 'ERROR_FLT_DUPLICATE_ENTRY')
    ERROR_FLT_CBDQ_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x801F000E, 'ERROR_FLT_CBDQ_DISABLED')
    ERROR_FLT_DO_NOT_ATTACH: Final[ThirtyTwoBits] = cached_hresult(0x801F000F, 'ERROR_FLT_DO_NOT_ATTACH')
    ERROR_FLT_DO_NOT_DETACH: Final[ThirtyTwoBits] = cached_hresult(0x801F0010, 'ERROR_FLT_DO_NOT_DETACH')
    ERROR_FLT_INSTANCE_ALTITUDE_COLLISION: Final[ThirtyTwoBits] = cached_hresult(0x801F0011, 'ERROR_FLT_INSTANCE_ALTITUDE_COLLISION')
    ERROR_FLT_INSTANCE_NAME_COLLISION: Final[ThirtyTwoBits] = cached_hresult(0x801F0012, 'ERROR_FLT_INSTANCE_NAME_COLLISION')
    ERROR_FLT_FILTER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x801F0013, 'ERROR_FLT_FILTER_NOT_FOUND')
    ERROR_FLT_VOLUME_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x801F0014, 'ERROR_FLT_VOLUME_NOT_FOUND')
    ERROR_FLT_INSTANCE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x801F0015, 'ERROR_FLT_INSTANCE_NOT_FOUND')
    ERROR_FLT_CONTEXT_ALLOCATION_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x801F0016, 'ERROR_FLT_CONTEXT_ALLOCATION_NOT_FOUND')
    ERROR_FLT_INVALID_CONTEXT_REGISTRATION: Final[ThirtyTwoBits] = cached_hresult(0x801F0017, 'ERROR_FLT_INVALID_CONTEXT_REGISTRATION')
    ERROR_FLT_NAME_CACHE_MISS: Final[ThirtyTwoBits] = cached_hresult(0x801F0018, 'ERROR_FLT_NAME_CACHE_MISS')
    ERROR_FLT_NO_DEVICE_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x801F0019, 'ERROR_FLT_NO_DEVICE_OBJECT')
    ERROR_FLT_VOLUME_ALREADY_MOUNTED: Final[ThirtyTwoBits] = cached_hresult(0x801F001A, 'ERROR_FLT_VOLUME_ALREADY_MOUNTED')
    ERROR_FLT_ALREADY_ENLISTED: Final[ThirtyTwoBits] = cached_hresult(0x801F001B, 'ERROR_FLT_ALREADY_ENLISTED')
    ERROR_FLT_CONTEXT_ALREADY_LINKED: Final[ThirtyTwoBits] = cached_hresult(0x801F001C, 'ERROR_FLT_CONTEXT_ALREADY_LINKED')
    ERROR_FLT_NO_WAITER_FOR_REPLY: Final[ThirtyTwoBits] = cached_hresult(0x801F0020, 'ERROR_FLT_NO_WAITER_FOR_REPLY')
    ERROR_FLT_REGISTRATION_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x801F0023, 'ERROR_FLT_REGISTRATION_BUSY')
    ERROR_FLT_WCOS_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x801F0024, 'ERROR_FLT_WCOS_NOT_SUPPORTED')
    ERROR_HUNG_DISPLAY_DRIVER_THREAD: Final[ThirtyTwoBits] = cached_hresult(0x80260001, 'ERROR_HUNG_DISPLAY_DRIVER_THREAD')
    ERROR_GRAPHICS_SESSION_TYPE_CHANGE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xC02605E8, 'ERROR_GRAPHICS_SESSION_TYPE_CHANGE_IN_PROGRESS')
    ERROR_MONITOR_NO_DESCRIPTOR: Final[ThirtyTwoBits] = cached_hresult(0x00261001, 'ERROR_MONITOR_NO_DESCRIPTOR')
    ERROR_MONITOR_UNKNOWN_DESCRIPTOR_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0x00261002, 'ERROR_MONITOR_UNKNOWN_DESCRIPTOR_FORMAT')
    ERROR_MONITOR_INVALID_DESCRIPTOR_CHECKSUM: Final[ThirtyTwoBits] = cached_hresult(0xC0261003, 'ERROR_MONITOR_INVALID_DESCRIPTOR_CHECKSUM')
    ERROR_MONITOR_INVALID_STANDARD_TIMING_BLOCK: Final[ThirtyTwoBits] = cached_hresult(0xC0261004, 'ERROR_MONITOR_INVALID_STANDARD_TIMING_BLOCK')
    ERROR_MONITOR_WMI_DATABLOCK_REGISTRATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xC0261005, 'ERROR_MONITOR_WMI_DATABLOCK_REGISTRATION_FAILED')
    ERROR_MONITOR_INVALID_SERIAL_NUMBER_MONDSC_BLOCK: Final[ThirtyTwoBits] = cached_hresult(0xC0261006, 'ERROR_MONITOR_INVALID_SERIAL_NUMBER_MONDSC_BLOCK')
    ERROR_MONITOR_INVALID_USER_FRIENDLY_MONDSC_BLOCK: Final[ThirtyTwoBits] = cached_hresult(0xC0261007, 'ERROR_MONITOR_INVALID_USER_FRIENDLY_MONDSC_BLOCK')
    ERROR_MONITOR_NO_MORE_DESCRIPTOR_DATA: Final[ThirtyTwoBits] = cached_hresult(0xC0261008, 'ERROR_MONITOR_NO_MORE_DESCRIPTOR_DATA')
    ERROR_MONITOR_INVALID_DETAILED_TIMING_BLOCK: Final[ThirtyTwoBits] = cached_hresult(0xC0261009, 'ERROR_MONITOR_INVALID_DETAILED_TIMING_BLOCK')
    ERROR_MONITOR_INVALID_MANUFACTURE_DATE: Final[ThirtyTwoBits] = cached_hresult(0xC026100A, 'ERROR_MONITOR_INVALID_MANUFACTURE_DATE')
    ERROR_GRAPHICS_NOT_EXCLUSIVE_MODE_OWNER: Final[ThirtyTwoBits] = cached_hresult(0xC0262000, 'ERROR_GRAPHICS_NOT_EXCLUSIVE_MODE_OWNER')
    ERROR_GRAPHICS_INSUFFICIENT_DMA_BUFFER: Final[ThirtyTwoBits] = cached_hresult(0xC0262001, 'ERROR_GRAPHICS_INSUFFICIENT_DMA_BUFFER')
    ERROR_GRAPHICS_INVALID_DISPLAY_ADAPTER: Final[ThirtyTwoBits] = cached_hresult(0xC0262002, 'ERROR_GRAPHICS_INVALID_DISPLAY_ADAPTER')
    ERROR_GRAPHICS_ADAPTER_WAS_RESET: Final[ThirtyTwoBits] = cached_hresult(0xC0262003, 'ERROR_GRAPHICS_ADAPTER_WAS_RESET')
    ERROR_GRAPHICS_INVALID_DRIVER_MODEL: Final[ThirtyTwoBits] = cached_hresult(0xC0262004, 'ERROR_GRAPHICS_INVALID_DRIVER_MODEL')
    ERROR_GRAPHICS_PRESENT_MODE_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0xC0262005, 'ERROR_GRAPHICS_PRESENT_MODE_CHANGED')
    ERROR_GRAPHICS_PRESENT_OCCLUDED: Final[ThirtyTwoBits] = cached_hresult(0xC0262006, 'ERROR_GRAPHICS_PRESENT_OCCLUDED')
    ERROR_GRAPHICS_PRESENT_DENIED: Final[ThirtyTwoBits] = cached_hresult(0xC0262007, 'ERROR_GRAPHICS_PRESENT_DENIED')
    ERROR_GRAPHICS_CANNOTCOLORCONVERT: Final[ThirtyTwoBits] = cached_hresult(0xC0262008, 'ERROR_GRAPHICS_CANNOTCOLORCONVERT')
    ERROR_GRAPHICS_DRIVER_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xC0262009, 'ERROR_GRAPHICS_DRIVER_MISMATCH')
    ERROR_GRAPHICS_PARTIAL_DATA_POPULATED: Final[ThirtyTwoBits] = cached_hresult(0x4026200A, 'ERROR_GRAPHICS_PARTIAL_DATA_POPULATED')
    ERROR_GRAPHICS_PRESENT_REDIRECTION_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xC026200B, 'ERROR_GRAPHICS_PRESENT_REDIRECTION_DISABLED')
    ERROR_GRAPHICS_PRESENT_UNOCCLUDED: Final[ThirtyTwoBits] = cached_hresult(0xC026200C, 'ERROR_GRAPHICS_PRESENT_UNOCCLUDED')
    ERROR_GRAPHICS_WINDOWDC_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xC026200D, 'ERROR_GRAPHICS_WINDOWDC_NOT_AVAILABLE')
    ERROR_GRAPHICS_WINDOWLESS_PRESENT_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xC026200E, 'ERROR_GRAPHICS_WINDOWLESS_PRESENT_DISABLED')
    ERROR_GRAPHICS_PRESENT_INVALID_WINDOW: Final[ThirtyTwoBits] = cached_hresult(0xC026200F, 'ERROR_GRAPHICS_PRESENT_INVALID_WINDOW')
    ERROR_GRAPHICS_PRESENT_BUFFER_NOT_BOUND: Final[ThirtyTwoBits] = cached_hresult(0xC0262010, 'ERROR_GRAPHICS_PRESENT_BUFFER_NOT_BOUND')
    ERROR_GRAPHICS_VAIL_STATE_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0xC0262011, 'ERROR_GRAPHICS_VAIL_STATE_CHANGED')
    ERROR_GRAPHICS_INDIRECT_DISPLAY_ABANDON_SWAPCHAIN: Final[ThirtyTwoBits] = cached_hresult(0xC0262012, 'ERROR_GRAPHICS_INDIRECT_DISPLAY_ABANDON_SWAPCHAIN')
    ERROR_GRAPHICS_INDIRECT_DISPLAY_DEVICE_STOPPED: Final[ThirtyTwoBits] = cached_hresult(0xC0262013, 'ERROR_GRAPHICS_INDIRECT_DISPLAY_DEVICE_STOPPED')
    ERROR_GRAPHICS_VAIL_FAILED_TO_SEND_CREATE_SUPERWETINK_MESSAGE: Final[ThirtyTwoBits] = cached_hresult(0xC0262014, 'ERROR_GRAPHICS_VAIL_FAILED_TO_SEND_CREATE_SUPERWETINK_MESSAGE')
    ERROR_GRAPHICS_VAIL_FAILED_TO_SEND_DESTROY_SUPERWETINK_MESSAGE: Final[ThirtyTwoBits] = cached_hresult(0xC0262015, 'ERROR_GRAPHICS_VAIL_FAILED_TO_SEND_DESTROY_SUPERWETINK_MESSAGE')
    ERROR_GRAPHICS_VAIL_FAILED_TO_SEND_COMPOSITION_WINDOW_DPI_MESSAGE: Final[ThirtyTwoBits] = cached_hresult(0xC0262016, 'ERROR_GRAPHICS_VAIL_FAILED_TO_SEND_COMPOSITION_WINDOW_DPI_MESSAGE')
    ERROR_GRAPHICS_LINK_CONFIGURATION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xC0262017, 'ERROR_GRAPHICS_LINK_CONFIGURATION_IN_PROGRESS')
    ERROR_GRAPHICS_MPO_ALLOCATION_UNPINNED: Final[ThirtyTwoBits] = cached_hresult(0xC0262018, 'ERROR_GRAPHICS_MPO_ALLOCATION_UNPINNED')
    ERROR_GRAPHICS_NO_VIDEO_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0xC0262100, 'ERROR_GRAPHICS_NO_VIDEO_MEMORY')
    ERROR_GRAPHICS_CANT_LOCK_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0xC0262101, 'ERROR_GRAPHICS_CANT_LOCK_MEMORY')
    ERROR_GRAPHICS_ALLOCATION_BUSY: Final[ThirtyTwoBits] = cached_hresult(0xC0262102, 'ERROR_GRAPHICS_ALLOCATION_BUSY')
    ERROR_GRAPHICS_TOO_MANY_REFERENCES: Final[ThirtyTwoBits] = cached_hresult(0xC0262103, 'ERROR_GRAPHICS_TOO_MANY_REFERENCES')
    ERROR_GRAPHICS_TRY_AGAIN_LATER: Final[ThirtyTwoBits] = cached_hresult(0xC0262104, 'ERROR_GRAPHICS_TRY_AGAIN_LATER')
    ERROR_GRAPHICS_TRY_AGAIN_NOW: Final[ThirtyTwoBits] = cached_hresult(0xC0262105, 'ERROR_GRAPHICS_TRY_AGAIN_NOW')
    ERROR_GRAPHICS_ALLOCATION_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xC0262106, 'ERROR_GRAPHICS_ALLOCATION_INVALID')
    ERROR_GRAPHICS_UNSWIZZLING_APERTURE_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xC0262107, 'ERROR_GRAPHICS_UNSWIZZLING_APERTURE_UNAVAILABLE')
    ERROR_GRAPHICS_UNSWIZZLING_APERTURE_UNSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xC0262108, 'ERROR_GRAPHICS_UNSWIZZLING_APERTURE_UNSUPPORTED')
    ERROR_GRAPHICS_CANT_EVICT_PINNED_ALLOCATION: Final[ThirtyTwoBits] = cached_hresult(0xC0262109, 'ERROR_GRAPHICS_CANT_EVICT_PINNED_ALLOCATION')
    ERROR_GRAPHICS_INVALID_ALLOCATION_USAGE: Final[ThirtyTwoBits] = cached_hresult(0xC0262110, 'ERROR_GRAPHICS_INVALID_ALLOCATION_USAGE')
    ERROR_GRAPHICS_CANT_RENDER_LOCKED_ALLOCATION: Final[ThirtyTwoBits] = cached_hresult(0xC0262111, 'ERROR_GRAPHICS_CANT_RENDER_LOCKED_ALLOCATION')
    ERROR_GRAPHICS_ALLOCATION_CLOSED: Final[ThirtyTwoBits] = cached_hresult(0xC0262112, 'ERROR_GRAPHICS_ALLOCATION_CLOSED')
    ERROR_GRAPHICS_INVALID_ALLOCATION_INSTANCE: Final[ThirtyTwoBits] = cached_hresult(0xC0262113, 'ERROR_GRAPHICS_INVALID_ALLOCATION_INSTANCE')
    ERROR_GRAPHICS_INVALID_ALLOCATION_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0xC0262114, 'ERROR_GRAPHICS_INVALID_ALLOCATION_HANDLE')
    ERROR_GRAPHICS_WRONG_ALLOCATION_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0xC0262115, 'ERROR_GRAPHICS_WRONG_ALLOCATION_DEVICE')
    ERROR_GRAPHICS_ALLOCATION_CONTENT_LOST: Final[ThirtyTwoBits] = cached_hresult(0xC0262116, 'ERROR_GRAPHICS_ALLOCATION_CONTENT_LOST')
    ERROR_GRAPHICS_GPU_EXCEPTION_ON_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0xC0262200, 'ERROR_GRAPHICS_GPU_EXCEPTION_ON_DEVICE')
    ERROR_GRAPHICS_SKIP_ALLOCATION_PREPARATION: Final[ThirtyTwoBits] = cached_hresult(0x40262201, 'ERROR_GRAPHICS_SKIP_ALLOCATION_PREPARATION')
    ERROR_GRAPHICS_INVALID_VIDPN_TOPOLOGY: Final[ThirtyTwoBits] = cached_hresult(0xC0262300, 'ERROR_GRAPHICS_INVALID_VIDPN_TOPOLOGY')
    ERROR_GRAPHICS_VIDPN_TOPOLOGY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xC0262301, 'ERROR_GRAPHICS_VIDPN_TOPOLOGY_NOT_SUPPORTED')
    ERROR_GRAPHICS_VIDPN_TOPOLOGY_CURRENTLY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xC0262302, 'ERROR_GRAPHICS_VIDPN_TOPOLOGY_CURRENTLY_NOT_SUPPORTED')
    ERROR_GRAPHICS_INVALID_VIDPN: Final[ThirtyTwoBits] = cached_hresult(0xC0262303, 'ERROR_GRAPHICS_INVALID_VIDPN')
    ERROR_GRAPHICS_INVALID_VIDEO_PRESENT_SOURCE: Final[ThirtyTwoBits] = cached_hresult(0xC0262304, 'ERROR_GRAPHICS_INVALID_VIDEO_PRESENT_SOURCE')
    ERROR_GRAPHICS_INVALID_VIDEO_PRESENT_TARGET: Final[ThirtyTwoBits] = cached_hresult(0xC0262305, 'ERROR_GRAPHICS_INVALID_VIDEO_PRESENT_TARGET')
    ERROR_GRAPHICS_VIDPN_MODALITY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xC0262306, 'ERROR_GRAPHICS_VIDPN_MODALITY_NOT_SUPPORTED')
    ERROR_GRAPHICS_MODE_NOT_PINNED: Final[ThirtyTwoBits] = cached_hresult(0x00262307, 'ERROR_GRAPHICS_MODE_NOT_PINNED')
    ERROR_GRAPHICS_INVALID_VIDPN_SOURCEMODESET: Final[ThirtyTwoBits] = cached_hresult(0xC0262308, 'ERROR_GRAPHICS_INVALID_VIDPN_SOURCEMODESET')
    ERROR_GRAPHICS_INVALID_VIDPN_TARGETMODESET: Final[ThirtyTwoBits] = cached_hresult(0xC0262309, 'ERROR_GRAPHICS_INVALID_VIDPN_TARGETMODESET')
    ERROR_GRAPHICS_INVALID_FREQUENCY: Final[ThirtyTwoBits] = cached_hresult(0xC026230A, 'ERROR_GRAPHICS_INVALID_FREQUENCY')
    ERROR_GRAPHICS_INVALID_ACTIVE_REGION: Final[ThirtyTwoBits] = cached_hresult(0xC026230B, 'ERROR_GRAPHICS_INVALID_ACTIVE_REGION')
    ERROR_GRAPHICS_INVALID_TOTAL_REGION: Final[ThirtyTwoBits] = cached_hresult(0xC026230C, 'ERROR_GRAPHICS_INVALID_TOTAL_REGION')
    ERROR_GRAPHICS_INVALID_VIDEO_PRESENT_SOURCE_MODE: Final[ThirtyTwoBits] = cached_hresult(0xC0262310, 'ERROR_GRAPHICS_INVALID_VIDEO_PRESENT_SOURCE_MODE')
    ERROR_GRAPHICS_INVALID_VIDEO_PRESENT_TARGET_MODE: Final[ThirtyTwoBits] = cached_hresult(0xC0262311, 'ERROR_GRAPHICS_INVALID_VIDEO_PRESENT_TARGET_MODE')
    ERROR_GRAPHICS_PINNED_MODE_MUST_REMAIN_IN_SET: Final[ThirtyTwoBits] = cached_hresult(0xC0262312, 'ERROR_GRAPHICS_PINNED_MODE_MUST_REMAIN_IN_SET')
    ERROR_GRAPHICS_PATH_ALREADY_IN_TOPOLOGY: Final[ThirtyTwoBits] = cached_hresult(0xC0262313, 'ERROR_GRAPHICS_PATH_ALREADY_IN_TOPOLOGY')
    ERROR_GRAPHICS_MODE_ALREADY_IN_MODESET: Final[ThirtyTwoBits] = cached_hresult(0xC0262314, 'ERROR_GRAPHICS_MODE_ALREADY_IN_MODESET')
    ERROR_GRAPHICS_INVALID_VIDEOPRESENTSOURCESET: Final[ThirtyTwoBits] = cached_hresult(0xC0262315, 'ERROR_GRAPHICS_INVALID_VIDEOPRESENTSOURCESET')
    ERROR_GRAPHICS_INVALID_VIDEOPRESENTTARGETSET: Final[ThirtyTwoBits] = cached_hresult(0xC0262316, 'ERROR_GRAPHICS_INVALID_VIDEOPRESENTTARGETSET')
    ERROR_GRAPHICS_SOURCE_ALREADY_IN_SET: Final[ThirtyTwoBits] = cached_hresult(0xC0262317, 'ERROR_GRAPHICS_SOURCE_ALREADY_IN_SET')
    ERROR_GRAPHICS_TARGET_ALREADY_IN_SET: Final[ThirtyTwoBits] = cached_hresult(0xC0262318, 'ERROR_GRAPHICS_TARGET_ALREADY_IN_SET')
    ERROR_GRAPHICS_INVALID_VIDPN_PRESENT_PATH: Final[ThirtyTwoBits] = cached_hresult(0xC0262319, 'ERROR_GRAPHICS_INVALID_VIDPN_PRESENT_PATH')
    ERROR_GRAPHICS_NO_RECOMMENDED_VIDPN_TOPOLOGY: Final[ThirtyTwoBits] = cached_hresult(0xC026231A, 'ERROR_GRAPHICS_NO_RECOMMENDED_VIDPN_TOPOLOGY')
    ERROR_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGESET: Final[ThirtyTwoBits] = cached_hresult(0xC026231B, 'ERROR_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGESET')
    ERROR_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGE: Final[ThirtyTwoBits] = cached_hresult(0xC026231C, 'ERROR_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGE')
    ERROR_GRAPHICS_FREQUENCYRANGE_NOT_IN_SET: Final[ThirtyTwoBits] = cached_hresult(0xC026231D, 'ERROR_GRAPHICS_FREQUENCYRANGE_NOT_IN_SET')
    ERROR_GRAPHICS_NO_PREFERRED_MODE: Final[ThirtyTwoBits] = cached_hresult(0x0026231E, 'ERROR_GRAPHICS_NO_PREFERRED_MODE')
    ERROR_GRAPHICS_FREQUENCYRANGE_ALREADY_IN_SET: Final[ThirtyTwoBits] = cached_hresult(0xC026231F, 'ERROR_GRAPHICS_FREQUENCYRANGE_ALREADY_IN_SET')
    ERROR_GRAPHICS_STALE_MODESET: Final[ThirtyTwoBits] = cached_hresult(0xC0262320, 'ERROR_GRAPHICS_STALE_MODESET')
    ERROR_GRAPHICS_INVALID_MONITOR_SOURCEMODESET: Final[ThirtyTwoBits] = cached_hresult(0xC0262321, 'ERROR_GRAPHICS_INVALID_MONITOR_SOURCEMODESET')
    ERROR_GRAPHICS_INVALID_MONITOR_SOURCE_MODE: Final[ThirtyTwoBits] = cached_hresult(0xC0262322, 'ERROR_GRAPHICS_INVALID_MONITOR_SOURCE_MODE')
    ERROR_GRAPHICS_NO_RECOMMENDED_FUNCTIONAL_VIDPN: Final[ThirtyTwoBits] = cached_hresult(0xC0262323, 'ERROR_GRAPHICS_NO_RECOMMENDED_FUNCTIONAL_VIDPN')
    ERROR_GRAPHICS_MODE_ID_MUST_BE_UNIQUE: Final[ThirtyTwoBits] = cached_hresult(0xC0262324, 'ERROR_GRAPHICS_MODE_ID_MUST_BE_UNIQUE')
    ERROR_GRAPHICS_EMPTY_ADAPTER_MONITOR_MODE_SUPPORT_INTERSECTION: Final[ThirtyTwoBits] = cached_hresult(0xC0262325, 'ERROR_GRAPHICS_EMPTY_ADAPTER_MONITOR_MODE_SUPPORT_INTERSECTION')
    ERROR_GRAPHICS_VIDEO_PRESENT_TARGETS_LESS_THAN_SOURCES: Final[ThirtyTwoBits] = cached_hresult(0xC0262326, 'ERROR_GRAPHICS_VIDEO_PRESENT_TARGETS_LESS_THAN_SOURCES')
    ERROR_GRAPHICS_PATH_NOT_IN_TOPOLOGY: Final[ThirtyTwoBits] = cached_hresult(0xC0262327, 'ERROR_GRAPHICS_PATH_NOT_IN_TOPOLOGY')
    ERROR_GRAPHICS_ADAPTER_MUST_HAVE_AT_LEAST_ONE_SOURCE: Final[ThirtyTwoBits] = cached_hresult(0xC0262328, 'ERROR_GRAPHICS_ADAPTER_MUST_HAVE_AT_LEAST_ONE_SOURCE')
    ERROR_GRAPHICS_ADAPTER_MUST_HAVE_AT_LEAST_ONE_TARGET: Final[ThirtyTwoBits] = cached_hresult(0xC0262329, 'ERROR_GRAPHICS_ADAPTER_MUST_HAVE_AT_LEAST_ONE_TARGET')
    ERROR_GRAPHICS_INVALID_MONITORDESCRIPTORSET: Final[ThirtyTwoBits] = cached_hresult(0xC026232A, 'ERROR_GRAPHICS_INVALID_MONITORDESCRIPTORSET')
    ERROR_GRAPHICS_INVALID_MONITORDESCRIPTOR: Final[ThirtyTwoBits] = cached_hresult(0xC026232B, 'ERROR_GRAPHICS_INVALID_MONITORDESCRIPTOR')
    ERROR_GRAPHICS_MONITORDESCRIPTOR_NOT_IN_SET: Final[ThirtyTwoBits] = cached_hresult(0xC026232C, 'ERROR_GRAPHICS_MONITORDESCRIPTOR_NOT_IN_SET')
    ERROR_GRAPHICS_MONITORDESCRIPTOR_ALREADY_IN_SET: Final[ThirtyTwoBits] = cached_hresult(0xC026232D, 'ERROR_GRAPHICS_MONITORDESCRIPTOR_ALREADY_IN_SET')
    ERROR_GRAPHICS_MONITORDESCRIPTOR_ID_MUST_BE_UNIQUE: Final[ThirtyTwoBits] = cached_hresult(0xC026232E, 'ERROR_GRAPHICS_MONITORDESCRIPTOR_ID_MUST_BE_UNIQUE')
    ERROR_GRAPHICS_INVALID_VIDPN_TARGET_SUBSET_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xC026232F, 'ERROR_GRAPHICS_INVALID_VIDPN_TARGET_SUBSET_TYPE')
    ERROR_GRAPHICS_RESOURCES_NOT_RELATED: Final[ThirtyTwoBits] = cached_hresult(0xC0262330, 'ERROR_GRAPHICS_RESOURCES_NOT_RELATED')
    ERROR_GRAPHICS_SOURCE_ID_MUST_BE_UNIQUE: Final[ThirtyTwoBits] = cached_hresult(0xC0262331, 'ERROR_GRAPHICS_SOURCE_ID_MUST_BE_UNIQUE')
    ERROR_GRAPHICS_TARGET_ID_MUST_BE_UNIQUE: Final[ThirtyTwoBits] = cached_hresult(0xC0262332, 'ERROR_GRAPHICS_TARGET_ID_MUST_BE_UNIQUE')
    ERROR_GRAPHICS_NO_AVAILABLE_VIDPN_TARGET: Final[ThirtyTwoBits] = cached_hresult(0xC0262333, 'ERROR_GRAPHICS_NO_AVAILABLE_VIDPN_TARGET')
    ERROR_GRAPHICS_MONITOR_COULD_NOT_BE_ASSOCIATED_WITH_ADAPTER: Final[ThirtyTwoBits] = cached_hresult(0xC0262334, 'ERROR_GRAPHICS_MONITOR_COULD_NOT_BE_ASSOCIATED_WITH_ADAPTER')
    ERROR_GRAPHICS_NO_VIDPNMGR: Final[ThirtyTwoBits] = cached_hresult(0xC0262335, 'ERROR_GRAPHICS_NO_VIDPNMGR')
    ERROR_GRAPHICS_NO_ACTIVE_VIDPN: Final[ThirtyTwoBits] = cached_hresult(0xC0262336, 'ERROR_GRAPHICS_NO_ACTIVE_VIDPN')
    ERROR_GRAPHICS_STALE_VIDPN_TOPOLOGY: Final[ThirtyTwoBits] = cached_hresult(0xC0262337, 'ERROR_GRAPHICS_STALE_VIDPN_TOPOLOGY')
    ERROR_GRAPHICS_MONITOR_NOT_CONNECTED: Final[ThirtyTwoBits] = cached_hresult(0xC0262338, 'ERROR_GRAPHICS_MONITOR_NOT_CONNECTED')
    ERROR_GRAPHICS_SOURCE_NOT_IN_TOPOLOGY: Final[ThirtyTwoBits] = cached_hresult(0xC0262339, 'ERROR_GRAPHICS_SOURCE_NOT_IN_TOPOLOGY')
    ERROR_GRAPHICS_INVALID_PRIMARYSURFACE_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xC026233A, 'ERROR_GRAPHICS_INVALID_PRIMARYSURFACE_SIZE')
    ERROR_GRAPHICS_INVALID_VISIBLEREGION_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xC026233B, 'ERROR_GRAPHICS_INVALID_VISIBLEREGION_SIZE')
    ERROR_GRAPHICS_INVALID_STRIDE: Final[ThirtyTwoBits] = cached_hresult(0xC026233C, 'ERROR_GRAPHICS_INVALID_STRIDE')
    ERROR_GRAPHICS_INVALID_PIXELFORMAT: Final[ThirtyTwoBits] = cached_hresult(0xC026233D, 'ERROR_GRAPHICS_INVALID_PIXELFORMAT')
    ERROR_GRAPHICS_INVALID_COLORBASIS: Final[ThirtyTwoBits] = cached_hresult(0xC026233E, 'ERROR_GRAPHICS_INVALID_COLORBASIS')
    ERROR_GRAPHICS_INVALID_PIXELVALUEACCESSMODE: Final[ThirtyTwoBits] = cached_hresult(0xC026233F, 'ERROR_GRAPHICS_INVALID_PIXELVALUEACCESSMODE')
    ERROR_GRAPHICS_TARGET_NOT_IN_TOPOLOGY: Final[ThirtyTwoBits] = cached_hresult(0xC0262340, 'ERROR_GRAPHICS_TARGET_NOT_IN_TOPOLOGY')
    ERROR_GRAPHICS_NO_DISPLAY_MODE_MANAGEMENT_SUPPORT: Final[ThirtyTwoBits] = cached_hresult(0xC0262341, 'ERROR_GRAPHICS_NO_DISPLAY_MODE_MANAGEMENT_SUPPORT')
    ERROR_GRAPHICS_VIDPN_SOURCE_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0xC0262342, 'ERROR_GRAPHICS_VIDPN_SOURCE_IN_USE')
    ERROR_GRAPHICS_CANT_ACCESS_ACTIVE_VIDPN: Final[ThirtyTwoBits] = cached_hresult(0xC0262343, 'ERROR_GRAPHICS_CANT_ACCESS_ACTIVE_VIDPN')
    ERROR_GRAPHICS_INVALID_PATH_IMPORTANCE_ORDINAL: Final[ThirtyTwoBits] = cached_hresult(0xC0262344, 'ERROR_GRAPHICS_INVALID_PATH_IMPORTANCE_ORDINAL')
    ERROR_GRAPHICS_INVALID_PATH_CONTENT_GEOMETRY_TRANSFORMATION: Final[ThirtyTwoBits] = cached_hresult(0xC0262345, 'ERROR_GRAPHICS_INVALID_PATH_CONTENT_GEOMETRY_TRANSFORMATION')
    ERROR_GRAPHICS_PATH_CONTENT_GEOMETRY_TRANSFORMATION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xC0262346, 'ERROR_GRAPHICS_PATH_CONTENT_GEOMETRY_TRANSFORMATION_NOT_SUPPORTED')
    ERROR_GRAPHICS_INVALID_GAMMA_RAMP: Final[ThirtyTwoBits] = cached_hresult(0xC0262347, 'ERROR_GRAPHICS_INVALID_GAMMA_RAMP')
    ERROR_GRAPHICS_GAMMA_RAMP_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xC0262348, 'ERROR_GRAPHICS_GAMMA_RAMP_NOT_SUPPORTED')
    ERROR_GRAPHICS_MULTISAMPLING_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xC0262349, 'ERROR_GRAPHICS_MULTISAMPLING_NOT_SUPPORTED')
    ERROR_GRAPHICS_MODE_NOT_IN_MODESET: Final[ThirtyTwoBits] = cached_hresult(0xC026234A, 'ERROR_GRAPHICS_MODE_NOT_IN_MODESET')
    ERROR_GRAPHICS_DATASET_IS_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0x0026234B, 'ERROR_GRAPHICS_DATASET_IS_EMPTY')
    ERROR_GRAPHICS_NO_MORE_ELEMENTS_IN_DATASET: Final[ThirtyTwoBits] = cached_hresult(0x0026234C, 'ERROR_GRAPHICS_NO_MORE_ELEMENTS_IN_DATASET')
    ERROR_GRAPHICS_INVALID_VIDPN_TOPOLOGY_RECOMMENDATION_REASON: Final[ThirtyTwoBits] = cached_hresult(0xC026234D, 'ERROR_GRAPHICS_INVALID_VIDPN_TOPOLOGY_RECOMMENDATION_REASON')
    ERROR_GRAPHICS_INVALID_PATH_CONTENT_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xC026234E, 'ERROR_GRAPHICS_INVALID_PATH_CONTENT_TYPE')
    ERROR_GRAPHICS_INVALID_COPYPROTECTION_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xC026234F, 'ERROR_GRAPHICS_INVALID_COPYPROTECTION_TYPE')
    ERROR_GRAPHICS_UNASSIGNED_MODESET_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xC0262350, 'ERROR_GRAPHICS_UNASSIGNED_MODESET_ALREADY_EXISTS')
    ERROR_GRAPHICS_PATH_CONTENT_GEOMETRY_TRANSFORMATION_NOT_PINNED: Final[ThirtyTwoBits] = cached_hresult(0x00262351, 'ERROR_GRAPHICS_PATH_CONTENT_GEOMETRY_TRANSFORMATION_NOT_PINNED')
    ERROR_GRAPHICS_INVALID_SCANLINE_ORDERING: Final[ThirtyTwoBits] = cached_hresult(0xC0262352, 'ERROR_GRAPHICS_INVALID_SCANLINE_ORDERING')
    ERROR_GRAPHICS_TOPOLOGY_CHANGES_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0xC0262353, 'ERROR_GRAPHICS_TOPOLOGY_CHANGES_NOT_ALLOWED')
    ERROR_GRAPHICS_NO_AVAILABLE_IMPORTANCE_ORDINALS: Final[ThirtyTwoBits] = cached_hresult(0xC0262354, 'ERROR_GRAPHICS_NO_AVAILABLE_IMPORTANCE_ORDINALS')
    ERROR_GRAPHICS_INCOMPATIBLE_PRIVATE_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0xC0262355, 'ERROR_GRAPHICS_INCOMPATIBLE_PRIVATE_FORMAT')
    ERROR_GRAPHICS_INVALID_MODE_PRUNING_ALGORITHM: Final[ThirtyTwoBits] = cached_hresult(0xC0262356, 'ERROR_GRAPHICS_INVALID_MODE_PRUNING_ALGORITHM')
    ERROR_GRAPHICS_INVALID_MONITOR_CAPABILITY_ORIGIN: Final[ThirtyTwoBits] = cached_hresult(0xC0262357, 'ERROR_GRAPHICS_INVALID_MONITOR_CAPABILITY_ORIGIN')
    ERROR_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGE_CONSTRAINT: Final[ThirtyTwoBits] = cached_hresult(0xC0262358, 'ERROR_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGE_CONSTRAINT')
    ERROR_GRAPHICS_MAX_NUM_PATHS_REACHED: Final[ThirtyTwoBits] = cached_hresult(0xC0262359, 'ERROR_GRAPHICS_MAX_NUM_PATHS_REACHED')
    ERROR_GRAPHICS_CANCEL_VIDPN_TOPOLOGY_AUGMENTATION: Final[ThirtyTwoBits] = cached_hresult(0xC026235A, 'ERROR_GRAPHICS_CANCEL_VIDPN_TOPOLOGY_AUGMENTATION')
    ERROR_GRAPHICS_INVALID_CLIENT_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xC026235B, 'ERROR_GRAPHICS_INVALID_CLIENT_TYPE')
    ERROR_GRAPHICS_CLIENTVIDPN_NOT_SET: Final[ThirtyTwoBits] = cached_hresult(0xC026235C, 'ERROR_GRAPHICS_CLIENTVIDPN_NOT_SET')
    ERROR_GRAPHICS_SPECIFIED_CHILD_ALREADY_CONNECTED: Final[ThirtyTwoBits] = cached_hresult(0xC0262400, 'ERROR_GRAPHICS_SPECIFIED_CHILD_ALREADY_CONNECTED')
    ERROR_GRAPHICS_CHILD_DESCRIPTOR_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xC0262401, 'ERROR_GRAPHICS_CHILD_DESCRIPTOR_NOT_SUPPORTED')
    ERROR_GRAPHICS_UNKNOWN_CHILD_STATUS: Final[ThirtyTwoBits] = cached_hresult(0x4026242F, 'ERROR_GRAPHICS_UNKNOWN_CHILD_STATUS')
    ERROR_GRAPHICS_NOT_A_LINKED_ADAPTER: Final[ThirtyTwoBits] = cached_hresult(0xC0262430, 'ERROR_GRAPHICS_NOT_A_LINKED_ADAPTER')
    ERROR_GRAPHICS_LEADLINK_NOT_ENUMERATED: Final[ThirtyTwoBits] = cached_hresult(0xC0262431, 'ERROR_GRAPHICS_LEADLINK_NOT_ENUMERATED')
    ERROR_GRAPHICS_CHAINLINKS_NOT_ENUMERATED: Final[ThirtyTwoBits] = cached_hresult(0xC0262432, 'ERROR_GRAPHICS_CHAINLINKS_NOT_ENUMERATED')
    ERROR_GRAPHICS_ADAPTER_CHAIN_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0xC0262433, 'ERROR_GRAPHICS_ADAPTER_CHAIN_NOT_READY')
    ERROR_GRAPHICS_CHAINLINKS_NOT_STARTED: Final[ThirtyTwoBits] = cached_hresult(0xC0262434, 'ERROR_GRAPHICS_CHAINLINKS_NOT_STARTED')
    ERROR_GRAPHICS_CHAINLINKS_NOT_POWERED_ON: Final[ThirtyTwoBits] = cached_hresult(0xC0262435, 'ERROR_GRAPHICS_CHAINLINKS_NOT_POWERED_ON')
    ERROR_GRAPHICS_INCONSISTENT_DEVICE_LINK_STATE: Final[ThirtyTwoBits] = cached_hresult(0xC0262436, 'ERROR_GRAPHICS_INCONSISTENT_DEVICE_LINK_STATE')
    ERROR_GRAPHICS_LEADLINK_START_DEFERRED: Final[ThirtyTwoBits] = cached_hresult(0x40262437, 'ERROR_GRAPHICS_LEADLINK_START_DEFERRED')
    ERROR_GRAPHICS_NOT_POST_DEVICE_DRIVER: Final[ThirtyTwoBits] = cached_hresult(0xC0262438, 'ERROR_GRAPHICS_NOT_POST_DEVICE_DRIVER')
    ERROR_GRAPHICS_POLLING_TOO_FREQUENTLY: Final[ThirtyTwoBits] = cached_hresult(0x40262439, 'ERROR_GRAPHICS_POLLING_TOO_FREQUENTLY')
    ERROR_GRAPHICS_START_DEFERRED: Final[ThirtyTwoBits] = cached_hresult(0x4026243A, 'ERROR_GRAPHICS_START_DEFERRED')
    ERROR_GRAPHICS_ADAPTER_ACCESS_NOT_EXCLUDED: Final[ThirtyTwoBits] = cached_hresult(0xC026243B, 'ERROR_GRAPHICS_ADAPTER_ACCESS_NOT_EXCLUDED')
    ERROR_GRAPHICS_DEPENDABLE_CHILD_STATUS: Final[ThirtyTwoBits] = cached_hresult(0x4026243C, 'ERROR_GRAPHICS_DEPENDABLE_CHILD_STATUS')
    ERROR_GRAPHICS_OPM_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xC0262500, 'ERROR_GRAPHICS_OPM_NOT_SUPPORTED')
    ERROR_GRAPHICS_COPP_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xC0262501, 'ERROR_GRAPHICS_COPP_NOT_SUPPORTED')
    ERROR_GRAPHICS_UAB_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xC0262502, 'ERROR_GRAPHICS_UAB_NOT_SUPPORTED')
    ERROR_GRAPHICS_OPM_INVALID_ENCRYPTED_PARAMETERS: Final[ThirtyTwoBits] = cached_hresult(0xC0262503, 'ERROR_GRAPHICS_OPM_INVALID_ENCRYPTED_PARAMETERS')
    ERROR_GRAPHICS_OPM_NO_VIDEO_OUTPUTS_EXIST: Final[ThirtyTwoBits] = cached_hresult(0xC0262505, 'ERROR_GRAPHICS_OPM_NO_VIDEO_OUTPUTS_EXIST')
    ERROR_GRAPHICS_OPM_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xC026250B, 'ERROR_GRAPHICS_OPM_INTERNAL_ERROR')
    ERROR_GRAPHICS_OPM_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0xC026250C, 'ERROR_GRAPHICS_OPM_INVALID_HANDLE')
    ERROR_GRAPHICS_PVP_INVALID_CERTIFICATE_LENGTH: Final[ThirtyTwoBits] = cached_hresult(0xC026250E, 'ERROR_GRAPHICS_PVP_INVALID_CERTIFICATE_LENGTH')
    ERROR_GRAPHICS_OPM_SPANNING_MODE_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0xC026250F, 'ERROR_GRAPHICS_OPM_SPANNING_MODE_ENABLED')
    ERROR_GRAPHICS_OPM_THEATER_MODE_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0xC0262510, 'ERROR_GRAPHICS_OPM_THEATER_MODE_ENABLED')
    ERROR_GRAPHICS_PVP_HFS_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xC0262511, 'ERROR_GRAPHICS_PVP_HFS_FAILED')
    ERROR_GRAPHICS_OPM_INVALID_SRM: Final[ThirtyTwoBits] = cached_hresult(0xC0262512, 'ERROR_GRAPHICS_OPM_INVALID_SRM')
    ERROR_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_HDCP: Final[ThirtyTwoBits] = cached_hresult(0xC0262513, 'ERROR_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_HDCP')
    ERROR_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_ACP: Final[ThirtyTwoBits] = cached_hresult(0xC0262514, 'ERROR_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_ACP')
    ERROR_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_CGMSA: Final[ThirtyTwoBits] = cached_hresult(0xC0262515, 'ERROR_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_CGMSA')
    ERROR_GRAPHICS_OPM_HDCP_SRM_NEVER_SET: Final[ThirtyTwoBits] = cached_hresult(0xC0262516, 'ERROR_GRAPHICS_OPM_HDCP_SRM_NEVER_SET')
    ERROR_GRAPHICS_OPM_RESOLUTION_TOO_HIGH: Final[ThirtyTwoBits] = cached_hresult(0xC0262517, 'ERROR_GRAPHICS_OPM_RESOLUTION_TOO_HIGH')
    ERROR_GRAPHICS_OPM_ALL_HDCP_HARDWARE_ALREADY_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0xC0262518, 'ERROR_GRAPHICS_OPM_ALL_HDCP_HARDWARE_ALREADY_IN_USE')
    ERROR_GRAPHICS_OPM_VIDEO_OUTPUT_NO_LONGER_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xC026251A, 'ERROR_GRAPHICS_OPM_VIDEO_OUTPUT_NO_LONGER_EXISTS')
    ERROR_GRAPHICS_OPM_SESSION_TYPE_CHANGE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xC026251B, 'ERROR_GRAPHICS_OPM_SESSION_TYPE_CHANGE_IN_PROGRESS')
    ERROR_GRAPHICS_OPM_VIDEO_OUTPUT_DOES_NOT_HAVE_COPP_SEMANTICS: Final[ThirtyTwoBits] = cached_hresult(0xC026251C, 'ERROR_GRAPHICS_OPM_VIDEO_OUTPUT_DOES_NOT_HAVE_COPP_SEMANTICS')
    ERROR_GRAPHICS_OPM_INVALID_INFORMATION_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0xC026251D, 'ERROR_GRAPHICS_OPM_INVALID_INFORMATION_REQUEST')
    ERROR_GRAPHICS_OPM_DRIVER_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xC026251E, 'ERROR_GRAPHICS_OPM_DRIVER_INTERNAL_ERROR')
    ERROR_GRAPHICS_OPM_VIDEO_OUTPUT_DOES_NOT_HAVE_OPM_SEMANTICS: Final[ThirtyTwoBits] = cached_hresult(0xC026251F, 'ERROR_GRAPHICS_OPM_VIDEO_OUTPUT_DOES_NOT_HAVE_OPM_SEMANTICS')
    ERROR_GRAPHICS_OPM_SIGNALING_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xC0262520, 'ERROR_GRAPHICS_OPM_SIGNALING_NOT_SUPPORTED')
    ERROR_GRAPHICS_OPM_INVALID_CONFIGURATION_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0xC0262521, 'ERROR_GRAPHICS_OPM_INVALID_CONFIGURATION_REQUEST')
    ERROR_GRAPHICS_I2C_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xC0262580, 'ERROR_GRAPHICS_I2C_NOT_SUPPORTED')
    ERROR_GRAPHICS_I2C_DEVICE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0xC0262581, 'ERROR_GRAPHICS_I2C_DEVICE_DOES_NOT_EXIST')
    ERROR_GRAPHICS_I2C_ERROR_TRANSMITTING_DATA: Final[ThirtyTwoBits] = cached_hresult(0xC0262582, 'ERROR_GRAPHICS_I2C_ERROR_TRANSMITTING_DATA')
    ERROR_GRAPHICS_I2C_ERROR_RECEIVING_DATA: Final[ThirtyTwoBits] = cached_hresult(0xC0262583, 'ERROR_GRAPHICS_I2C_ERROR_RECEIVING_DATA')
    ERROR_GRAPHICS_DDCCI_VCP_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xC0262584, 'ERROR_GRAPHICS_DDCCI_VCP_NOT_SUPPORTED')
    ERROR_GRAPHICS_DDCCI_INVALID_DATA: Final[ThirtyTwoBits] = cached_hresult(0xC0262585, 'ERROR_GRAPHICS_DDCCI_INVALID_DATA')
    ERROR_GRAPHICS_DDCCI_MONITOR_RETURNED_INVALID_TIMING_STATUS_BYTE: Final[ThirtyTwoBits] = cached_hresult(0xC0262586, 'ERROR_GRAPHICS_DDCCI_MONITOR_RETURNED_INVALID_TIMING_STATUS_BYTE')
    ERROR_GRAPHICS_MCA_INVALID_CAPABILITIES_STRING: Final[ThirtyTwoBits] = cached_hresult(0xC0262587, 'ERROR_GRAPHICS_MCA_INVALID_CAPABILITIES_STRING')
    ERROR_GRAPHICS_MCA_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xC0262588, 'ERROR_GRAPHICS_MCA_INTERNAL_ERROR')
    ERROR_GRAPHICS_DDCCI_INVALID_MESSAGE_COMMAND: Final[ThirtyTwoBits] = cached_hresult(0xC0262589, 'ERROR_GRAPHICS_DDCCI_INVALID_MESSAGE_COMMAND')
    ERROR_GRAPHICS_DDCCI_INVALID_MESSAGE_LENGTH: Final[ThirtyTwoBits] = cached_hresult(0xC026258A, 'ERROR_GRAPHICS_DDCCI_INVALID_MESSAGE_LENGTH')
    ERROR_GRAPHICS_DDCCI_INVALID_MESSAGE_CHECKSUM: Final[ThirtyTwoBits] = cached_hresult(0xC026258B, 'ERROR_GRAPHICS_DDCCI_INVALID_MESSAGE_CHECKSUM')
    ERROR_GRAPHICS_INVALID_PHYSICAL_MONITOR_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0xC026258C, 'ERROR_GRAPHICS_INVALID_PHYSICAL_MONITOR_HANDLE')
    ERROR_GRAPHICS_MONITOR_NO_LONGER_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xC026258D, 'ERROR_GRAPHICS_MONITOR_NO_LONGER_EXISTS')
    ERROR_GRAPHICS_DDCCI_CURRENT_CURRENT_VALUE_GREATER_THAN_MAXIMUM_VALUE: Final[ThirtyTwoBits] = cached_hresult(0xC02625D8, 'ERROR_GRAPHICS_DDCCI_CURRENT_CURRENT_VALUE_GREATER_THAN_MAXIMUM_VALUE')
    ERROR_GRAPHICS_MCA_INVALID_VCP_VERSION: Final[ThirtyTwoBits] = cached_hresult(0xC02625D9, 'ERROR_GRAPHICS_MCA_INVALID_VCP_VERSION')
    ERROR_GRAPHICS_MCA_MONITOR_VIOLATES_MCCS_SPECIFICATION: Final[ThirtyTwoBits] = cached_hresult(0xC02625DA, 'ERROR_GRAPHICS_MCA_MONITOR_VIOLATES_MCCS_SPECIFICATION')
    ERROR_GRAPHICS_MCA_MCCS_VERSION_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xC02625DB, 'ERROR_GRAPHICS_MCA_MCCS_VERSION_MISMATCH')
    ERROR_GRAPHICS_MCA_UNSUPPORTED_MCCS_VERSION: Final[ThirtyTwoBits] = cached_hresult(0xC02625DC, 'ERROR_GRAPHICS_MCA_UNSUPPORTED_MCCS_VERSION')
    ERROR_GRAPHICS_MCA_INVALID_TECHNOLOGY_TYPE_RETURNED: Final[ThirtyTwoBits] = cached_hresult(0xC02625DE, 'ERROR_GRAPHICS_MCA_INVALID_TECHNOLOGY_TYPE_RETURNED')
    ERROR_GRAPHICS_MCA_UNSUPPORTED_COLOR_TEMPERATURE: Final[ThirtyTwoBits] = cached_hresult(0xC02625DF, 'ERROR_GRAPHICS_MCA_UNSUPPORTED_COLOR_TEMPERATURE')
    ERROR_GRAPHICS_ONLY_CONSOLE_SESSION_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xC02625E0, 'ERROR_GRAPHICS_ONLY_CONSOLE_SESSION_SUPPORTED')
    ERROR_GRAPHICS_NO_DISPLAY_DEVICE_CORRESPONDS_TO_NAME: Final[ThirtyTwoBits] = cached_hresult(0xC02625E1, 'ERROR_GRAPHICS_NO_DISPLAY_DEVICE_CORRESPONDS_TO_NAME')
    ERROR_GRAPHICS_DISPLAY_DEVICE_NOT_ATTACHED_TO_DESKTOP: Final[ThirtyTwoBits] = cached_hresult(0xC02625E2, 'ERROR_GRAPHICS_DISPLAY_DEVICE_NOT_ATTACHED_TO_DESKTOP')
    ERROR_GRAPHICS_MIRRORING_DEVICES_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xC02625E3, 'ERROR_GRAPHICS_MIRRORING_DEVICES_NOT_SUPPORTED')
    ERROR_GRAPHICS_INVALID_POINTER: Final[ThirtyTwoBits] = cached_hresult(0xC02625E4, 'ERROR_GRAPHICS_INVALID_POINTER')
    ERROR_GRAPHICS_NO_MONITORS_CORRESPOND_TO_DISPLAY_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0xC02625E5, 'ERROR_GRAPHICS_NO_MONITORS_CORRESPOND_TO_DISPLAY_DEVICE')
    ERROR_GRAPHICS_PARAMETER_ARRAY_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0xC02625E6, 'ERROR_GRAPHICS_PARAMETER_ARRAY_TOO_SMALL')
    ERROR_GRAPHICS_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xC02625E7, 'ERROR_GRAPHICS_INTERNAL_ERROR')
    DWM_E_COMPOSITIONDISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80263001, 'DWM_E_COMPOSITIONDISABLED')
    DWM_E_REMOTING_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80263002, 'DWM_E_REMOTING_NOT_SUPPORTED')
    DWM_E_NO_REDIRECTION_SURFACE_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80263003, 'DWM_E_NO_REDIRECTION_SURFACE_AVAILABLE')
    DWM_E_NOT_QUEUING_PRESENTS: Final[ThirtyTwoBits] = cached_hresult(0x80263004, 'DWM_E_NOT_QUEUING_PRESENTS')
    DWM_S_GDI_REDIRECTION_SURFACE: Final[ThirtyTwoBits] = cached_hresult(0x00263005, 'DWM_S_GDI_REDIRECTION_SURFACE')
    DWM_E_ADAPTER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80263005, 'DWM_E_ADAPTER_NOT_FOUND')
    DWM_E_TEXTURE_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0x80263007, 'DWM_E_TEXTURE_TOO_LARGE')
    DWM_S_GDI_REDIRECTION_SURFACE_BLT_VIA_GDI: Final[ThirtyTwoBits] = cached_hresult(0x00263008, 'DWM_S_GDI_REDIRECTION_SURFACE_BLT_VIA_GDI')
    NAP_E_INVALID_PACKET: Final[ThirtyTwoBits] = cached_hresult(0x80270001, 'NAP_E_INVALID_PACKET')
    NAP_E_MISSING_SOH: Final[ThirtyTwoBits] = cached_hresult(0x80270002, 'NAP_E_MISSING_SOH')
    NAP_E_CONFLICTING_ID: Final[ThirtyTwoBits] = cached_hresult(0x80270003, 'NAP_E_CONFLICTING_ID')
    NAP_E_NO_CACHED_SOH: Final[ThirtyTwoBits] = cached_hresult(0x80270004, 'NAP_E_NO_CACHED_SOH')
    NAP_E_STILL_BOUND: Final[ThirtyTwoBits] = cached_hresult(0x80270005, 'NAP_E_STILL_BOUND')
    NAP_E_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x80270006, 'NAP_E_NOT_REGISTERED')
    NAP_E_NOT_INITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0x80270007, 'NAP_E_NOT_INITIALIZED')
    NAP_E_MISMATCHED_ID: Final[ThirtyTwoBits] = cached_hresult(0x80270008, 'NAP_E_MISMATCHED_ID')
    NAP_E_NOT_PENDING: Final[ThirtyTwoBits] = cached_hresult(0x80270009, 'NAP_E_NOT_PENDING')
    NAP_E_ID_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8027000A, 'NAP_E_ID_NOT_FOUND')
    NAP_E_MAXSIZE_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0x8027000B, 'NAP_E_MAXSIZE_TOO_SMALL')
    NAP_E_SERVICE_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x8027000C, 'NAP_E_SERVICE_NOT_RUNNING')
    NAP_S_CERT_ALREADY_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0x0027000D, 'NAP_S_CERT_ALREADY_PRESENT')
    NAP_E_ENTITY_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x8027000E, 'NAP_E_ENTITY_DISABLED')
    NAP_E_NETSH_GROUPPOLICY_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8027000F, 'NAP_E_NETSH_GROUPPOLICY_ERROR')
    NAP_E_TOO_MANY_CALLS: Final[ThirtyTwoBits] = cached_hresult(0x80270010, 'NAP_E_TOO_MANY_CALLS')
    NAP_E_SHV_CONFIG_EXISTED: Final[ThirtyTwoBits] = cached_hresult(0x80270011, 'NAP_E_SHV_CONFIG_EXISTED')
    NAP_E_SHV_CONFIG_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80270012, 'NAP_E_SHV_CONFIG_NOT_FOUND')
    NAP_E_SHV_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x80270013, 'NAP_E_SHV_TIMEOUT')
    FA_E_MAX_PERSISTED_ITEMS_REACHED: Final[ThirtyTwoBits] = cached_hresult(0x80270220, 'FA_E_MAX_PERSISTED_ITEMS_REACHED')
    FA_E_HOMEGROUP_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80270222, 'FA_E_HOMEGROUP_NOT_AVAILABLE')
    E_MONITOR_RESOLUTION_TOO_LOW: Final[ThirtyTwoBits] = cached_hresult(0x80270250, 'E_MONITOR_RESOLUTION_TOO_LOW')
    E_ELEVATED_ACTIVATION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80270251, 'E_ELEVATED_ACTIVATION_NOT_SUPPORTED')
    E_UAC_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80270252, 'E_UAC_DISABLED')
    E_FULL_ADMIN_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80270253, 'E_FULL_ADMIN_NOT_SUPPORTED')
    E_APPLICATION_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x80270254, 'E_APPLICATION_NOT_REGISTERED')
    E_MULTIPLE_EXTENSIONS_FOR_APPLICATION: Final[ThirtyTwoBits] = cached_hresult(0x80270255, 'E_MULTIPLE_EXTENSIONS_FOR_APPLICATION')
    E_MULTIPLE_PACKAGES_FOR_FAMILY: Final[ThirtyTwoBits] = cached_hresult(0x80270256, 'E_MULTIPLE_PACKAGES_FOR_FAMILY')
    E_APPLICATION_MANAGER_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x80270257, 'E_APPLICATION_MANAGER_NOT_RUNNING')
    S_STORE_LAUNCHED_FOR_REMEDIATION: Final[ThirtyTwoBits] = cached_hresult(0x00270258, 'S_STORE_LAUNCHED_FOR_REMEDIATION')
    S_APPLICATION_ACTIVATION_ERROR_HANDLED_BY_DIALOG: Final[ThirtyTwoBits] = cached_hresult(0x00270259, 'S_APPLICATION_ACTIVATION_ERROR_HANDLED_BY_DIALOG')
    E_APPLICATION_ACTIVATION_TIMED_OUT: Final[ThirtyTwoBits] = cached_hresult(0x8027025A, 'E_APPLICATION_ACTIVATION_TIMED_OUT')
    E_APPLICATION_ACTIVATION_EXEC_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8027025B, 'E_APPLICATION_ACTIVATION_EXEC_FAILURE')
    E_APPLICATION_TEMPORARY_LICENSE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8027025C, 'E_APPLICATION_TEMPORARY_LICENSE_ERROR')
    E_APPLICATION_TRIAL_LICENSE_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x8027025D, 'E_APPLICATION_TRIAL_LICENSE_EXPIRED')
    E_SKYDRIVE_ROOT_TARGET_FILE_SYSTEM_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80270260, 'E_SKYDRIVE_ROOT_TARGET_FILE_SYSTEM_NOT_SUPPORTED')
    E_SKYDRIVE_ROOT_TARGET_OVERLAP: Final[ThirtyTwoBits] = cached_hresult(0x80270261, 'E_SKYDRIVE_ROOT_TARGET_OVERLAP')
    E_SKYDRIVE_ROOT_TARGET_CANNOT_INDEX: Final[ThirtyTwoBits] = cached_hresult(0x80270262, 'E_SKYDRIVE_ROOT_TARGET_CANNOT_INDEX')
    E_SKYDRIVE_FILE_NOT_UPLOADED: Final[ThirtyTwoBits] = cached_hresult(0x80270263, 'E_SKYDRIVE_FILE_NOT_UPLOADED')
    E_SKYDRIVE_UPDATE_AVAILABILITY_FAIL: Final[ThirtyTwoBits] = cached_hresult(0x80270264, 'E_SKYDRIVE_UPDATE_AVAILABILITY_FAIL')
    E_SKYDRIVE_ROOT_TARGET_VOLUME_ROOT_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80270265, 'E_SKYDRIVE_ROOT_TARGET_VOLUME_ROOT_NOT_SUPPORTED')
    TPM_E_ERROR_MASK: Final[ThirtyTwoBits] = cached_hresult(0x80280000, 'TPM_E_ERROR_MASK')
    TPM_E_AUTHFAIL: Final[ThirtyTwoBits] = cached_hresult(0x80280001, 'TPM_E_AUTHFAIL')
    TPM_E_BADINDEX: Final[ThirtyTwoBits] = cached_hresult(0x80280002, 'TPM_E_BADINDEX')
    TPM_E_BAD_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0x80280003, 'TPM_E_BAD_PARAMETER')
    TPM_E_AUDITFAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80280004, 'TPM_E_AUDITFAILURE')
    TPM_E_CLEAR_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80280005, 'TPM_E_CLEAR_DISABLED')
    TPM_E_DEACTIVATED: Final[ThirtyTwoBits] = cached_hresult(0x80280006, 'TPM_E_DEACTIVATED')
    TPM_E_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80280007, 'TPM_E_DISABLED')
    TPM_E_DISABLED_CMD: Final[ThirtyTwoBits] = cached_hresult(0x80280008, 'TPM_E_DISABLED_CMD')
    TPM_E_FAIL: Final[ThirtyTwoBits] = cached_hresult(0x80280009, 'TPM_E_FAIL')
    TPM_E_BAD_ORDINAL: Final[ThirtyTwoBits] = cached_hresult(0x8028000A, 'TPM_E_BAD_ORDINAL')
    TPM_E_INSTALL_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x8028000B, 'TPM_E_INSTALL_DISABLED')
    TPM_E_INVALID_KEYHANDLE: Final[ThirtyTwoBits] = cached_hresult(0x8028000C, 'TPM_E_INVALID_KEYHANDLE')
    TPM_E_KEYNOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x8028000D, 'TPM_E_KEYNOTFOUND')
    TPM_E_INAPPROPRIATE_ENC: Final[ThirtyTwoBits] = cached_hresult(0x8028000E, 'TPM_E_INAPPROPRIATE_ENC')
    TPM_E_MIGRATEFAIL: Final[ThirtyTwoBits] = cached_hresult(0x8028000F, 'TPM_E_MIGRATEFAIL')
    TPM_E_INVALID_PCR_INFO: Final[ThirtyTwoBits] = cached_hresult(0x80280010, 'TPM_E_INVALID_PCR_INFO')
    TPM_E_NOSPACE: Final[ThirtyTwoBits] = cached_hresult(0x80280011, 'TPM_E_NOSPACE')
    TPM_E_NOSRK: Final[ThirtyTwoBits] = cached_hresult(0x80280012, 'TPM_E_NOSRK')
    TPM_E_NOTSEALED_BLOB: Final[ThirtyTwoBits] = cached_hresult(0x80280013, 'TPM_E_NOTSEALED_BLOB')
    TPM_E_OWNER_SET: Final[ThirtyTwoBits] = cached_hresult(0x80280014, 'TPM_E_OWNER_SET')
    TPM_E_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0x80280015, 'TPM_E_RESOURCES')
    TPM_E_SHORTRANDOM: Final[ThirtyTwoBits] = cached_hresult(0x80280016, 'TPM_E_SHORTRANDOM')
    TPM_E_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x80280017, 'TPM_E_SIZE')
    TPM_E_WRONGPCRVAL: Final[ThirtyTwoBits] = cached_hresult(0x80280018, 'TPM_E_WRONGPCRVAL')
    TPM_E_BAD_PARAM_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x80280019, 'TPM_E_BAD_PARAM_SIZE')
    TPM_E_SHA_THREAD: Final[ThirtyTwoBits] = cached_hresult(0x8028001A, 'TPM_E_SHA_THREAD')
    TPM_E_SHA_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8028001B, 'TPM_E_SHA_ERROR')
    TPM_E_FAILEDSELFTEST: Final[ThirtyTwoBits] = cached_hresult(0x8028001C, 'TPM_E_FAILEDSELFTEST')
    TPM_E_AUTH2FAIL: Final[ThirtyTwoBits] = cached_hresult(0x8028001D, 'TPM_E_AUTH2FAIL')
    TPM_E_BADTAG: Final[ThirtyTwoBits] = cached_hresult(0x8028001E, 'TPM_E_BADTAG')
    TPM_E_IOERROR: Final[ThirtyTwoBits] = cached_hresult(0x8028001F, 'TPM_E_IOERROR')
    TPM_E_ENCRYPT_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80280020, 'TPM_E_ENCRYPT_ERROR')
    TPM_E_DECRYPT_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80280021, 'TPM_E_DECRYPT_ERROR')
    TPM_E_INVALID_AUTHHANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80280022, 'TPM_E_INVALID_AUTHHANDLE')
    TPM_E_NO_ENDORSEMENT: Final[ThirtyTwoBits] = cached_hresult(0x80280023, 'TPM_E_NO_ENDORSEMENT')
    TPM_E_INVALID_KEYUSAGE: Final[ThirtyTwoBits] = cached_hresult(0x80280024, 'TPM_E_INVALID_KEYUSAGE')
    TPM_E_WRONG_ENTITYTYPE: Final[ThirtyTwoBits] = cached_hresult(0x80280025, 'TPM_E_WRONG_ENTITYTYPE')
    TPM_E_INVALID_POSTINIT: Final[ThirtyTwoBits] = cached_hresult(0x80280026, 'TPM_E_INVALID_POSTINIT')
    TPM_E_INAPPROPRIATE_SIG: Final[ThirtyTwoBits] = cached_hresult(0x80280027, 'TPM_E_INAPPROPRIATE_SIG')
    TPM_E_BAD_KEY_PROPERTY: Final[ThirtyTwoBits] = cached_hresult(0x80280028, 'TPM_E_BAD_KEY_PROPERTY')
    TPM_E_BAD_MIGRATION: Final[ThirtyTwoBits] = cached_hresult(0x80280029, 'TPM_E_BAD_MIGRATION')
    TPM_E_BAD_SCHEME: Final[ThirtyTwoBits] = cached_hresult(0x8028002A, 'TPM_E_BAD_SCHEME')
    TPM_E_BAD_DATASIZE: Final[ThirtyTwoBits] = cached_hresult(0x8028002B, 'TPM_E_BAD_DATASIZE')
    TPM_E_BAD_MODE: Final[ThirtyTwoBits] = cached_hresult(0x8028002C, 'TPM_E_BAD_MODE')
    TPM_E_BAD_PRESENCE: Final[ThirtyTwoBits] = cached_hresult(0x8028002D, 'TPM_E_BAD_PRESENCE')
    TPM_E_BAD_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x8028002E, 'TPM_E_BAD_VERSION')
    TPM_E_NO_WRAP_TRANSPORT: Final[ThirtyTwoBits] = cached_hresult(0x8028002F, 'TPM_E_NO_WRAP_TRANSPORT')
    TPM_E_AUDITFAIL_UNSUCCESSFUL: Final[ThirtyTwoBits] = cached_hresult(0x80280030, 'TPM_E_AUDITFAIL_UNSUCCESSFUL')
    TPM_E_AUDITFAIL_SUCCESSFUL: Final[ThirtyTwoBits] = cached_hresult(0x80280031, 'TPM_E_AUDITFAIL_SUCCESSFUL')
    TPM_E_NOTRESETABLE: Final[ThirtyTwoBits] = cached_hresult(0x80280032, 'TPM_E_NOTRESETABLE')
    TPM_E_NOTLOCAL: Final[ThirtyTwoBits] = cached_hresult(0x80280033, 'TPM_E_NOTLOCAL')
    TPM_E_BAD_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x80280034, 'TPM_E_BAD_TYPE')
    TPM_E_INVALID_RESOURCE: Final[ThirtyTwoBits] = cached_hresult(0x80280035, 'TPM_E_INVALID_RESOURCE')
    TPM_E_NOTFIPS: Final[ThirtyTwoBits] = cached_hresult(0x80280036, 'TPM_E_NOTFIPS')
    TPM_E_INVALID_FAMILY: Final[ThirtyTwoBits] = cached_hresult(0x80280037, 'TPM_E_INVALID_FAMILY')
    TPM_E_NO_NV_PERMISSION: Final[ThirtyTwoBits] = cached_hresult(0x80280038, 'TPM_E_NO_NV_PERMISSION')
    TPM_E_REQUIRES_SIGN: Final[ThirtyTwoBits] = cached_hresult(0x80280039, 'TPM_E_REQUIRES_SIGN')
    TPM_E_KEY_NOTSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8028003A, 'TPM_E_KEY_NOTSUPPORTED')
    TPM_E_AUTH_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x8028003B, 'TPM_E_AUTH_CONFLICT')
    TPM_E_AREA_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x8028003C, 'TPM_E_AREA_LOCKED')
    TPM_E_BAD_LOCALITY: Final[ThirtyTwoBits] = cached_hresult(0x8028003D, 'TPM_E_BAD_LOCALITY')
    TPM_E_READ_ONLY: Final[ThirtyTwoBits] = cached_hresult(0x8028003E, 'TPM_E_READ_ONLY')
    TPM_E_PER_NOWRITE: Final[ThirtyTwoBits] = cached_hresult(0x8028003F, 'TPM_E_PER_NOWRITE')
    TPM_E_FAMILYCOUNT: Final[ThirtyTwoBits] = cached_hresult(0x80280040, 'TPM_E_FAMILYCOUNT')
    TPM_E_WRITE_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x80280041, 'TPM_E_WRITE_LOCKED')
    TPM_E_BAD_ATTRIBUTES: Final[ThirtyTwoBits] = cached_hresult(0x80280042, 'TPM_E_BAD_ATTRIBUTES')
    TPM_E_INVALID_STRUCTURE: Final[ThirtyTwoBits] = cached_hresult(0x80280043, 'TPM_E_INVALID_STRUCTURE')
    TPM_E_KEY_OWNER_CONTROL: Final[ThirtyTwoBits] = cached_hresult(0x80280044, 'TPM_E_KEY_OWNER_CONTROL')
    TPM_E_BAD_COUNTER: Final[ThirtyTwoBits] = cached_hresult(0x80280045, 'TPM_E_BAD_COUNTER')
    TPM_E_NOT_FULLWRITE: Final[ThirtyTwoBits] = cached_hresult(0x80280046, 'TPM_E_NOT_FULLWRITE')
    TPM_E_CONTEXT_GAP: Final[ThirtyTwoBits] = cached_hresult(0x80280047, 'TPM_E_CONTEXT_GAP')
    TPM_E_MAXNVWRITES: Final[ThirtyTwoBits] = cached_hresult(0x80280048, 'TPM_E_MAXNVWRITES')
    TPM_E_NOOPERATOR: Final[ThirtyTwoBits] = cached_hresult(0x80280049, 'TPM_E_NOOPERATOR')
    TPM_E_RESOURCEMISSING: Final[ThirtyTwoBits] = cached_hresult(0x8028004A, 'TPM_E_RESOURCEMISSING')
    TPM_E_DELEGATE_LOCK: Final[ThirtyTwoBits] = cached_hresult(0x8028004B, 'TPM_E_DELEGATE_LOCK')
    TPM_E_DELEGATE_FAMILY: Final[ThirtyTwoBits] = cached_hresult(0x8028004C, 'TPM_E_DELEGATE_FAMILY')
    TPM_E_DELEGATE_ADMIN: Final[ThirtyTwoBits] = cached_hresult(0x8028004D, 'TPM_E_DELEGATE_ADMIN')
    TPM_E_TRANSPORT_NOTEXCLUSIVE: Final[ThirtyTwoBits] = cached_hresult(0x8028004E, 'TPM_E_TRANSPORT_NOTEXCLUSIVE')
    TPM_E_OWNER_CONTROL: Final[ThirtyTwoBits] = cached_hresult(0x8028004F, 'TPM_E_OWNER_CONTROL')
    TPM_E_DAA_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0x80280050, 'TPM_E_DAA_RESOURCES')
    TPM_E_DAA_INPUT_DATA0: Final[ThirtyTwoBits] = cached_hresult(0x80280051, 'TPM_E_DAA_INPUT_DATA0')
    TPM_E_DAA_INPUT_DATA1: Final[ThirtyTwoBits] = cached_hresult(0x80280052, 'TPM_E_DAA_INPUT_DATA1')
    TPM_E_DAA_ISSUER_SETTINGS: Final[ThirtyTwoBits] = cached_hresult(0x80280053, 'TPM_E_DAA_ISSUER_SETTINGS')
    TPM_E_DAA_TPM_SETTINGS: Final[ThirtyTwoBits] = cached_hresult(0x80280054, 'TPM_E_DAA_TPM_SETTINGS')
    TPM_E_DAA_STAGE: Final[ThirtyTwoBits] = cached_hresult(0x80280055, 'TPM_E_DAA_STAGE')
    TPM_E_DAA_ISSUER_VALIDITY: Final[ThirtyTwoBits] = cached_hresult(0x80280056, 'TPM_E_DAA_ISSUER_VALIDITY')
    TPM_E_DAA_WRONG_W: Final[ThirtyTwoBits] = cached_hresult(0x80280057, 'TPM_E_DAA_WRONG_W')
    TPM_E_BAD_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80280058, 'TPM_E_BAD_HANDLE')
    TPM_E_BAD_DELEGATE: Final[ThirtyTwoBits] = cached_hresult(0x80280059, 'TPM_E_BAD_DELEGATE')
    TPM_E_BADCONTEXT: Final[ThirtyTwoBits] = cached_hresult(0x8028005A, 'TPM_E_BADCONTEXT')
    TPM_E_TOOMANYCONTEXTS: Final[ThirtyTwoBits] = cached_hresult(0x8028005B, 'TPM_E_TOOMANYCONTEXTS')
    TPM_E_MA_TICKET_SIGNATURE: Final[ThirtyTwoBits] = cached_hresult(0x8028005C, 'TPM_E_MA_TICKET_SIGNATURE')
    TPM_E_MA_DESTINATION: Final[ThirtyTwoBits] = cached_hresult(0x8028005D, 'TPM_E_MA_DESTINATION')
    TPM_E_MA_SOURCE: Final[ThirtyTwoBits] = cached_hresult(0x8028005E, 'TPM_E_MA_SOURCE')
    TPM_E_MA_AUTHORITY: Final[ThirtyTwoBits] = cached_hresult(0x8028005F, 'TPM_E_MA_AUTHORITY')
    TPM_E_PERMANENTEK: Final[ThirtyTwoBits] = cached_hresult(0x80280061, 'TPM_E_PERMANENTEK')
    TPM_E_BAD_SIGNATURE: Final[ThirtyTwoBits] = cached_hresult(0x80280062, 'TPM_E_BAD_SIGNATURE')
    TPM_E_NOCONTEXTSPACE: Final[ThirtyTwoBits] = cached_hresult(0x80280063, 'TPM_E_NOCONTEXTSPACE')
    TPM_20_E_ASYMMETRIC: Final[ThirtyTwoBits] = cached_hresult(0x80280081, 'TPM_20_E_ASYMMETRIC')
    TPM_20_E_ATTRIBUTES: Final[ThirtyTwoBits] = cached_hresult(0x80280082, 'TPM_20_E_ATTRIBUTES')
    TPM_20_E_HASH: Final[ThirtyTwoBits] = cached_hresult(0x80280083, 'TPM_20_E_HASH')
    TPM_20_E_VALUE: Final[ThirtyTwoBits] = cached_hresult(0x80280084, 'TPM_20_E_VALUE')
    TPM_20_E_HIERARCHY: Final[ThirtyTwoBits] = cached_hresult(0x80280085, 'TPM_20_E_HIERARCHY')
    TPM_20_E_KEY_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x80280087, 'TPM_20_E_KEY_SIZE')
    TPM_20_E_MGF: Final[ThirtyTwoBits] = cached_hresult(0x80280088, 'TPM_20_E_MGF')
    TPM_20_E_MODE: Final[ThirtyTwoBits] = cached_hresult(0x80280089, 'TPM_20_E_MODE')
    TPM_20_E_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x8028008A, 'TPM_20_E_TYPE')
    TPM_20_E_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x8028008B, 'TPM_20_E_HANDLE')
    TPM_20_E_KDF: Final[ThirtyTwoBits] = cached_hresult(0x8028008C, 'TPM_20_E_KDF')
    TPM_20_E_RANGE: Final[ThirtyTwoBits] = cached_hresult(0x8028008D, 'TPM_20_E_RANGE')
    TPM_20_E_AUTH_FAIL: Final[ThirtyTwoBits] = cached_hresult(0x8028008E, 'TPM_20_E_AUTH_FAIL')
    TPM_20_E_NONCE: Final[ThirtyTwoBits] = cached_hresult(0x8028008F, 'TPM_20_E_NONCE')
    TPM_20_E_PP: Final[ThirtyTwoBits] = cached_hresult(0x80280090, 'TPM_20_E_PP')
    TPM_20_E_SCHEME: Final[ThirtyTwoBits] = cached_hresult(0x80280092, 'TPM_20_E_SCHEME')
    TPM_20_E_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x80280095, 'TPM_20_E_SIZE')
    TPM_20_E_SYMMETRIC: Final[ThirtyTwoBits] = cached_hresult(0x80280096, 'TPM_20_E_SYMMETRIC')
    TPM_20_E_TAG: Final[ThirtyTwoBits] = cached_hresult(0x80280097, 'TPM_20_E_TAG')
    TPM_20_E_SELECTOR: Final[ThirtyTwoBits] = cached_hresult(0x80280098, 'TPM_20_E_SELECTOR')
    TPM_20_E_INSUFFICIENT: Final[ThirtyTwoBits] = cached_hresult(0x8028009A, 'TPM_20_E_INSUFFICIENT')
    TPM_20_E_SIGNATURE: Final[ThirtyTwoBits] = cached_hresult(0x8028009B, 'TPM_20_E_SIGNATURE')
    TPM_20_E_KEY: Final[ThirtyTwoBits] = cached_hresult(0x8028009C, 'TPM_20_E_KEY')
    TPM_20_E_POLICY_FAIL: Final[ThirtyTwoBits] = cached_hresult(0x8028009D, 'TPM_20_E_POLICY_FAIL')
    TPM_20_E_INTEGRITY: Final[ThirtyTwoBits] = cached_hresult(0x8028009F, 'TPM_20_E_INTEGRITY')
    TPM_20_E_TICKET: Final[ThirtyTwoBits] = cached_hresult(0x802800A0, 'TPM_20_E_TICKET')
    TPM_20_E_RESERVED_BITS: Final[ThirtyTwoBits] = cached_hresult(0x802800A1, 'TPM_20_E_RESERVED_BITS')
    TPM_20_E_BAD_AUTH: Final[ThirtyTwoBits] = cached_hresult(0x802800A2, 'TPM_20_E_BAD_AUTH')
    TPM_20_E_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x802800A3, 'TPM_20_E_EXPIRED')
    TPM_20_E_POLICY_CC: Final[ThirtyTwoBits] = cached_hresult(0x802800A4, 'TPM_20_E_POLICY_CC')
    TPM_20_E_BINDING: Final[ThirtyTwoBits] = cached_hresult(0x802800A5, 'TPM_20_E_BINDING')
    TPM_20_E_CURVE: Final[ThirtyTwoBits] = cached_hresult(0x802800A6, 'TPM_20_E_CURVE')
    TPM_20_E_ECC_POINT: Final[ThirtyTwoBits] = cached_hresult(0x802800A7, 'TPM_20_E_ECC_POINT')
    TPM_20_E_INITIALIZE: Final[ThirtyTwoBits] = cached_hresult(0x80280100, 'TPM_20_E_INITIALIZE')
    TPM_20_E_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80280101, 'TPM_20_E_FAILURE')
    TPM_20_E_SEQUENCE: Final[ThirtyTwoBits] = cached_hresult(0x80280103, 'TPM_20_E_SEQUENCE')
    TPM_20_E_PRIVATE: Final[ThirtyTwoBits] = cached_hresult(0x8028010B, 'TPM_20_E_PRIVATE')
    TPM_20_E_HMAC: Final[ThirtyTwoBits] = cached_hresult(0x80280119, 'TPM_20_E_HMAC')
    TPM_20_E_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80280120, 'TPM_20_E_DISABLED')
    TPM_20_E_EXCLUSIVE: Final[ThirtyTwoBits] = cached_hresult(0x80280121, 'TPM_20_E_EXCLUSIVE')
    TPM_20_E_ECC_CURVE: Final[ThirtyTwoBits] = cached_hresult(0x80280123, 'TPM_20_E_ECC_CURVE')
    TPM_20_E_AUTH_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x80280124, 'TPM_20_E_AUTH_TYPE')
    TPM_20_E_AUTH_MISSING: Final[ThirtyTwoBits] = cached_hresult(0x80280125, 'TPM_20_E_AUTH_MISSING')
    TPM_20_E_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x80280126, 'TPM_20_E_POLICY')
    TPM_20_E_PCR: Final[ThirtyTwoBits] = cached_hresult(0x80280127, 'TPM_20_E_PCR')
    TPM_20_E_PCR_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0x80280128, 'TPM_20_E_PCR_CHANGED')
    TPM_20_E_UPGRADE: Final[ThirtyTwoBits] = cached_hresult(0x8028012D, 'TPM_20_E_UPGRADE')
    TPM_20_E_TOO_MANY_CONTEXTS: Final[ThirtyTwoBits] = cached_hresult(0x8028012E, 'TPM_20_E_TOO_MANY_CONTEXTS')
    TPM_20_E_AUTH_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x8028012F, 'TPM_20_E_AUTH_UNAVAILABLE')
    TPM_20_E_REBOOT: Final[ThirtyTwoBits] = cached_hresult(0x80280130, 'TPM_20_E_REBOOT')
    TPM_20_E_UNBALANCED: Final[ThirtyTwoBits] = cached_hresult(0x80280131, 'TPM_20_E_UNBALANCED')
    TPM_20_E_COMMAND_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x80280142, 'TPM_20_E_COMMAND_SIZE')
    TPM_20_E_COMMAND_CODE: Final[ThirtyTwoBits] = cached_hresult(0x80280143, 'TPM_20_E_COMMAND_CODE')
    TPM_20_E_AUTHSIZE: Final[ThirtyTwoBits] = cached_hresult(0x80280144, 'TPM_20_E_AUTHSIZE')
    TPM_20_E_AUTH_CONTEXT: Final[ThirtyTwoBits] = cached_hresult(0x80280145, 'TPM_20_E_AUTH_CONTEXT')
    TPM_20_E_NV_RANGE: Final[ThirtyTwoBits] = cached_hresult(0x80280146, 'TPM_20_E_NV_RANGE')
    TPM_20_E_NV_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x80280147, 'TPM_20_E_NV_SIZE')
    TPM_20_E_NV_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x80280148, 'TPM_20_E_NV_LOCKED')
    TPM_20_E_NV_AUTHORIZATION: Final[ThirtyTwoBits] = cached_hresult(0x80280149, 'TPM_20_E_NV_AUTHORIZATION')
    TPM_20_E_NV_UNINITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0x8028014A, 'TPM_20_E_NV_UNINITIALIZED')
    TPM_20_E_NV_SPACE: Final[ThirtyTwoBits] = cached_hresult(0x8028014B, 'TPM_20_E_NV_SPACE')
    TPM_20_E_NV_DEFINED: Final[ThirtyTwoBits] = cached_hresult(0x8028014C, 'TPM_20_E_NV_DEFINED')
    TPM_20_E_BAD_CONTEXT: Final[ThirtyTwoBits] = cached_hresult(0x80280150, 'TPM_20_E_BAD_CONTEXT')
    TPM_20_E_CPHASH: Final[ThirtyTwoBits] = cached_hresult(0x80280151, 'TPM_20_E_CPHASH')
    TPM_20_E_PARENT: Final[ThirtyTwoBits] = cached_hresult(0x80280152, 'TPM_20_E_PARENT')
    TPM_20_E_NEEDS_TEST: Final[ThirtyTwoBits] = cached_hresult(0x80280153, 'TPM_20_E_NEEDS_TEST')
    TPM_20_E_NO_RESULT: Final[ThirtyTwoBits] = cached_hresult(0x80280154, 'TPM_20_E_NO_RESULT')
    TPM_20_E_SENSITIVE: Final[ThirtyTwoBits] = cached_hresult(0x80280155, 'TPM_20_E_SENSITIVE')
    TPM_E_COMMAND_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x80280400, 'TPM_E_COMMAND_BLOCKED')
    TPM_E_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80280401, 'TPM_E_INVALID_HANDLE')
    TPM_E_DUPLICATE_VHANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80280402, 'TPM_E_DUPLICATE_VHANDLE')
    TPM_E_EMBEDDED_COMMAND_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x80280403, 'TPM_E_EMBEDDED_COMMAND_BLOCKED')
    TPM_E_EMBEDDED_COMMAND_UNSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80280404, 'TPM_E_EMBEDDED_COMMAND_UNSUPPORTED')
    TPM_E_RETRY: Final[ThirtyTwoBits] = cached_hresult(0x80280800, 'TPM_E_RETRY')
    TPM_E_NEEDS_SELFTEST: Final[ThirtyTwoBits] = cached_hresult(0x80280801, 'TPM_E_NEEDS_SELFTEST')
    TPM_E_DOING_SELFTEST: Final[ThirtyTwoBits] = cached_hresult(0x80280802, 'TPM_E_DOING_SELFTEST')
    TPM_E_DEFEND_LOCK_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x80280803, 'TPM_E_DEFEND_LOCK_RUNNING')
    TPM_20_E_CONTEXT_GAP: Final[ThirtyTwoBits] = cached_hresult(0x80280901, 'TPM_20_E_CONTEXT_GAP')
    TPM_20_E_OBJECT_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0x80280902, 'TPM_20_E_OBJECT_MEMORY')
    TPM_20_E_SESSION_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0x80280903, 'TPM_20_E_SESSION_MEMORY')
    TPM_20_E_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0x80280904, 'TPM_20_E_MEMORY')
    TPM_20_E_SESSION_HANDLES: Final[ThirtyTwoBits] = cached_hresult(0x80280905, 'TPM_20_E_SESSION_HANDLES')
    TPM_20_E_OBJECT_HANDLES: Final[ThirtyTwoBits] = cached_hresult(0x80280906, 'TPM_20_E_OBJECT_HANDLES')
    TPM_20_E_LOCALITY: Final[ThirtyTwoBits] = cached_hresult(0x80280907, 'TPM_20_E_LOCALITY')
    TPM_20_E_YIELDED: Final[ThirtyTwoBits] = cached_hresult(0x80280908, 'TPM_20_E_YIELDED')
    TPM_20_E_CANCELED: Final[ThirtyTwoBits] = cached_hresult(0x80280909, 'TPM_20_E_CANCELED')
    TPM_20_E_TESTING: Final[ThirtyTwoBits] = cached_hresult(0x8028090A, 'TPM_20_E_TESTING')
    TPM_20_E_NV_RATE: Final[ThirtyTwoBits] = cached_hresult(0x80280920, 'TPM_20_E_NV_RATE')
    TPM_20_E_LOCKOUT: Final[ThirtyTwoBits] = cached_hresult(0x80280921, 'TPM_20_E_LOCKOUT')
    TPM_20_E_RETRY: Final[ThirtyTwoBits] = cached_hresult(0x80280922, 'TPM_20_E_RETRY')
    TPM_20_E_NV_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80280923, 'TPM_20_E_NV_UNAVAILABLE')
    TBS_E_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80284001, 'TBS_E_INTERNAL_ERROR')
    TBS_E_BAD_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0x80284002, 'TBS_E_BAD_PARAMETER')
    TBS_E_INVALID_OUTPUT_POINTER: Final[ThirtyTwoBits] = cached_hresult(0x80284003, 'TBS_E_INVALID_OUTPUT_POINTER')
    TBS_E_INVALID_CONTEXT: Final[ThirtyTwoBits] = cached_hresult(0x80284004, 'TBS_E_INVALID_CONTEXT')
    TBS_E_INSUFFICIENT_BUFFER: Final[ThirtyTwoBits] = cached_hresult(0x80284005, 'TBS_E_INSUFFICIENT_BUFFER')
    TBS_E_IOERROR: Final[ThirtyTwoBits] = cached_hresult(0x80284006, 'TBS_E_IOERROR')
    TBS_E_INVALID_CONTEXT_PARAM: Final[ThirtyTwoBits] = cached_hresult(0x80284007, 'TBS_E_INVALID_CONTEXT_PARAM')
    TBS_E_SERVICE_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x80284008, 'TBS_E_SERVICE_NOT_RUNNING')
    TBS_E_TOO_MANY_TBS_CONTEXTS: Final[ThirtyTwoBits] = cached_hresult(0x80284009, 'TBS_E_TOO_MANY_TBS_CONTEXTS')
    TBS_E_TOO_MANY_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0x8028400A, 'TBS_E_TOO_MANY_RESOURCES')
    TBS_E_SERVICE_START_PENDING: Final[ThirtyTwoBits] = cached_hresult(0x8028400B, 'TBS_E_SERVICE_START_PENDING')
    TBS_E_PPI_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8028400C, 'TBS_E_PPI_NOT_SUPPORTED')
    TBS_E_COMMAND_CANCELED: Final[ThirtyTwoBits] = cached_hresult(0x8028400D, 'TBS_E_COMMAND_CANCELED')
    TBS_E_BUFFER_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0x8028400E, 'TBS_E_BUFFER_TOO_LARGE')
    TBS_E_TPM_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8028400F, 'TBS_E_TPM_NOT_FOUND')
    TBS_E_SERVICE_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80284010, 'TBS_E_SERVICE_DISABLED')
    TBS_E_NO_EVENT_LOG: Final[ThirtyTwoBits] = cached_hresult(0x80284011, 'TBS_E_NO_EVENT_LOG')
    TBS_E_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x80284012, 'TBS_E_ACCESS_DENIED')
    TBS_E_PROVISIONING_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80284013, 'TBS_E_PROVISIONING_NOT_ALLOWED')
    TBS_E_PPI_FUNCTION_UNSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80284014, 'TBS_E_PPI_FUNCTION_UNSUPPORTED')
    TBS_E_OWNERAUTH_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80284015, 'TBS_E_OWNERAUTH_NOT_FOUND')
    TBS_E_PROVISIONING_INCOMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x80284016, 'TBS_E_PROVISIONING_INCOMPLETE')
    TPMAPI_E_INVALID_STATE: Final[ThirtyTwoBits] = cached_hresult(0x80290100, 'TPMAPI_E_INVALID_STATE')
    TPMAPI_E_NOT_ENOUGH_DATA: Final[ThirtyTwoBits] = cached_hresult(0x80290101, 'TPMAPI_E_NOT_ENOUGH_DATA')
    TPMAPI_E_TOO_MUCH_DATA: Final[ThirtyTwoBits] = cached_hresult(0x80290102, 'TPMAPI_E_TOO_MUCH_DATA')
    TPMAPI_E_INVALID_OUTPUT_POINTER: Final[ThirtyTwoBits] = cached_hresult(0x80290103, 'TPMAPI_E_INVALID_OUTPUT_POINTER')
    TPMAPI_E_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0x80290104, 'TPMAPI_E_INVALID_PARAMETER')
    TPMAPI_E_OUT_OF_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0x80290105, 'TPMAPI_E_OUT_OF_MEMORY')
    TPMAPI_E_BUFFER_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0x80290106, 'TPMAPI_E_BUFFER_TOO_SMALL')
    TPMAPI_E_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80290107, 'TPMAPI_E_INTERNAL_ERROR')
    TPMAPI_E_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x80290108, 'TPMAPI_E_ACCESS_DENIED')
    TPMAPI_E_AUTHORIZATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80290109, 'TPMAPI_E_AUTHORIZATION_FAILED')
    TPMAPI_E_INVALID_CONTEXT_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x8029010A, 'TPMAPI_E_INVALID_CONTEXT_HANDLE')
    TPMAPI_E_TBS_COMMUNICATION_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8029010B, 'TPMAPI_E_TBS_COMMUNICATION_ERROR')
    TPMAPI_E_TPM_COMMAND_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8029010C, 'TPMAPI_E_TPM_COMMAND_ERROR')
    TPMAPI_E_MESSAGE_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0x8029010D, 'TPMAPI_E_MESSAGE_TOO_LARGE')
    TPMAPI_E_INVALID_ENCODING: Final[ThirtyTwoBits] = cached_hresult(0x8029010E, 'TPMAPI_E_INVALID_ENCODING')
    TPMAPI_E_INVALID_KEY_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x8029010F, 'TPMAPI_E_INVALID_KEY_SIZE')
    TPMAPI_E_ENCRYPTION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80290110, 'TPMAPI_E_ENCRYPTION_FAILED')
    TPMAPI_E_INVALID_KEY_PARAMS: Final[ThirtyTwoBits] = cached_hresult(0x80290111, 'TPMAPI_E_INVALID_KEY_PARAMS')
    TPMAPI_E_INVALID_MIGRATION_AUTHORIZATION_BLOB: Final[ThirtyTwoBits] = cached_hresult(0x80290112, 'TPMAPI_E_INVALID_MIGRATION_AUTHORIZATION_BLOB')
    TPMAPI_E_INVALID_PCR_INDEX: Final[ThirtyTwoBits] = cached_hresult(0x80290113, 'TPMAPI_E_INVALID_PCR_INDEX')
    TPMAPI_E_INVALID_DELEGATE_BLOB: Final[ThirtyTwoBits] = cached_hresult(0x80290114, 'TPMAPI_E_INVALID_DELEGATE_BLOB')
    TPMAPI_E_INVALID_CONTEXT_PARAMS: Final[ThirtyTwoBits] = cached_hresult(0x80290115, 'TPMAPI_E_INVALID_CONTEXT_PARAMS')
    TPMAPI_E_INVALID_KEY_BLOB: Final[ThirtyTwoBits] = cached_hresult(0x80290116, 'TPMAPI_E_INVALID_KEY_BLOB')
    TPMAPI_E_INVALID_PCR_DATA: Final[ThirtyTwoBits] = cached_hresult(0x80290117, 'TPMAPI_E_INVALID_PCR_DATA')
    TPMAPI_E_INVALID_OWNER_AUTH: Final[ThirtyTwoBits] = cached_hresult(0x80290118, 'TPMAPI_E_INVALID_OWNER_AUTH')
    TPMAPI_E_FIPS_RNG_CHECK_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80290119, 'TPMAPI_E_FIPS_RNG_CHECK_FAILED')
    TPMAPI_E_EMPTY_TCG_LOG: Final[ThirtyTwoBits] = cached_hresult(0x8029011A, 'TPMAPI_E_EMPTY_TCG_LOG')
    TPMAPI_E_INVALID_TCG_LOG_ENTRY: Final[ThirtyTwoBits] = cached_hresult(0x8029011B, 'TPMAPI_E_INVALID_TCG_LOG_ENTRY')
    TPMAPI_E_TCG_SEPARATOR_ABSENT: Final[ThirtyTwoBits] = cached_hresult(0x8029011C, 'TPMAPI_E_TCG_SEPARATOR_ABSENT')
    TPMAPI_E_TCG_INVALID_DIGEST_ENTRY: Final[ThirtyTwoBits] = cached_hresult(0x8029011D, 'TPMAPI_E_TCG_INVALID_DIGEST_ENTRY')
    TPMAPI_E_POLICY_DENIES_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x8029011E, 'TPMAPI_E_POLICY_DENIES_OPERATION')
    TPMAPI_E_NV_BITS_NOT_DEFINED: Final[ThirtyTwoBits] = cached_hresult(0x8029011F, 'TPMAPI_E_NV_BITS_NOT_DEFINED')
    TPMAPI_E_NV_BITS_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0x80290120, 'TPMAPI_E_NV_BITS_NOT_READY')
    TPMAPI_E_SEALING_KEY_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80290121, 'TPMAPI_E_SEALING_KEY_NOT_AVAILABLE')
    TPMAPI_E_NO_AUTHORIZATION_CHAIN_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80290122, 'TPMAPI_E_NO_AUTHORIZATION_CHAIN_FOUND')
    TPMAPI_E_SVN_COUNTER_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80290123, 'TPMAPI_E_SVN_COUNTER_NOT_AVAILABLE')
    TPMAPI_E_OWNER_AUTH_NOT_NULL: Final[ThirtyTwoBits] = cached_hresult(0x80290124, 'TPMAPI_E_OWNER_AUTH_NOT_NULL')
    TPMAPI_E_ENDORSEMENT_AUTH_NOT_NULL: Final[ThirtyTwoBits] = cached_hresult(0x80290125, 'TPMAPI_E_ENDORSEMENT_AUTH_NOT_NULL')
    TPMAPI_E_AUTHORIZATION_REVOKED: Final[ThirtyTwoBits] = cached_hresult(0x80290126, 'TPMAPI_E_AUTHORIZATION_REVOKED')
    TPMAPI_E_MALFORMED_AUTHORIZATION_KEY: Final[ThirtyTwoBits] = cached_hresult(0x80290127, 'TPMAPI_E_MALFORMED_AUTHORIZATION_KEY')
    TPMAPI_E_AUTHORIZING_KEY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80290128, 'TPMAPI_E_AUTHORIZING_KEY_NOT_SUPPORTED')
    TPMAPI_E_INVALID_AUTHORIZATION_SIGNATURE: Final[ThirtyTwoBits] = cached_hresult(0x80290129, 'TPMAPI_E_INVALID_AUTHORIZATION_SIGNATURE')
    TPMAPI_E_MALFORMED_AUTHORIZATION_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x8029012A, 'TPMAPI_E_MALFORMED_AUTHORIZATION_POLICY')
    TPMAPI_E_MALFORMED_AUTHORIZATION_OTHER: Final[ThirtyTwoBits] = cached_hresult(0x8029012B, 'TPMAPI_E_MALFORMED_AUTHORIZATION_OTHER')
    TPMAPI_E_SEALING_KEY_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0x8029012C, 'TPMAPI_E_SEALING_KEY_CHANGED')
    TPMAPI_E_INVALID_TPM_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x8029012D, 'TPMAPI_E_INVALID_TPM_VERSION')
    TPMAPI_E_INVALID_POLICYAUTH_BLOB_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x8029012E, 'TPMAPI_E_INVALID_POLICYAUTH_BLOB_TYPE')
    TBSIMP_E_BUFFER_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0x80290200, 'TBSIMP_E_BUFFER_TOO_SMALL')
    TBSIMP_E_CLEANUP_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80290201, 'TBSIMP_E_CLEANUP_FAILED')
    TBSIMP_E_INVALID_CONTEXT_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80290202, 'TBSIMP_E_INVALID_CONTEXT_HANDLE')
    TBSIMP_E_INVALID_CONTEXT_PARAM: Final[ThirtyTwoBits] = cached_hresult(0x80290203, 'TBSIMP_E_INVALID_CONTEXT_PARAM')
    TBSIMP_E_TPM_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80290204, 'TBSIMP_E_TPM_ERROR')
    TBSIMP_E_HASH_BAD_KEY: Final[ThirtyTwoBits] = cached_hresult(0x80290205, 'TBSIMP_E_HASH_BAD_KEY')
    TBSIMP_E_DUPLICATE_VHANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80290206, 'TBSIMP_E_DUPLICATE_VHANDLE')
    TBSIMP_E_INVALID_OUTPUT_POINTER: Final[ThirtyTwoBits] = cached_hresult(0x80290207, 'TBSIMP_E_INVALID_OUTPUT_POINTER')
    TBSIMP_E_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0x80290208, 'TBSIMP_E_INVALID_PARAMETER')
    TBSIMP_E_RPC_INIT_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80290209, 'TBSIMP_E_RPC_INIT_FAILED')
    TBSIMP_E_SCHEDULER_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x8029020A, 'TBSIMP_E_SCHEDULER_NOT_RUNNING')
    TBSIMP_E_COMMAND_CANCELED: Final[ThirtyTwoBits] = cached_hresult(0x8029020B, 'TBSIMP_E_COMMAND_CANCELED')
    TBSIMP_E_OUT_OF_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0x8029020C, 'TBSIMP_E_OUT_OF_MEMORY')
    TBSIMP_E_LIST_NO_MORE_ITEMS: Final[ThirtyTwoBits] = cached_hresult(0x8029020D, 'TBSIMP_E_LIST_NO_MORE_ITEMS')
    TBSIMP_E_LIST_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8029020E, 'TBSIMP_E_LIST_NOT_FOUND')
    TBSIMP_E_NOT_ENOUGH_SPACE: Final[ThirtyTwoBits] = cached_hresult(0x8029020F, 'TBSIMP_E_NOT_ENOUGH_SPACE')
    TBSIMP_E_NOT_ENOUGH_TPM_CONTEXTS: Final[ThirtyTwoBits] = cached_hresult(0x80290210, 'TBSIMP_E_NOT_ENOUGH_TPM_CONTEXTS')
    TBSIMP_E_COMMAND_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80290211, 'TBSIMP_E_COMMAND_FAILED')
    TBSIMP_E_UNKNOWN_ORDINAL: Final[ThirtyTwoBits] = cached_hresult(0x80290212, 'TBSIMP_E_UNKNOWN_ORDINAL')
    TBSIMP_E_RESOURCE_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x80290213, 'TBSIMP_E_RESOURCE_EXPIRED')
    TBSIMP_E_INVALID_RESOURCE: Final[ThirtyTwoBits] = cached_hresult(0x80290214, 'TBSIMP_E_INVALID_RESOURCE')
    TBSIMP_E_NOTHING_TO_UNLOAD: Final[ThirtyTwoBits] = cached_hresult(0x80290215, 'TBSIMP_E_NOTHING_TO_UNLOAD')
    TBSIMP_E_HASH_TABLE_FULL: Final[ThirtyTwoBits] = cached_hresult(0x80290216, 'TBSIMP_E_HASH_TABLE_FULL')
    TBSIMP_E_TOO_MANY_TBS_CONTEXTS: Final[ThirtyTwoBits] = cached_hresult(0x80290217, 'TBSIMP_E_TOO_MANY_TBS_CONTEXTS')
    TBSIMP_E_TOO_MANY_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0x80290218, 'TBSIMP_E_TOO_MANY_RESOURCES')
    TBSIMP_E_PPI_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80290219, 'TBSIMP_E_PPI_NOT_SUPPORTED')
    TBSIMP_E_TPM_INCOMPATIBLE: Final[ThirtyTwoBits] = cached_hresult(0x8029021A, 'TBSIMP_E_TPM_INCOMPATIBLE')
    TBSIMP_E_NO_EVENT_LOG: Final[ThirtyTwoBits] = cached_hresult(0x8029021B, 'TBSIMP_E_NO_EVENT_LOG')
    TPM_E_PPI_ACPI_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80290300, 'TPM_E_PPI_ACPI_FAILURE')
    TPM_E_PPI_USER_ABORT: Final[ThirtyTwoBits] = cached_hresult(0x80290301, 'TPM_E_PPI_USER_ABORT')
    TPM_E_PPI_BIOS_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80290302, 'TPM_E_PPI_BIOS_FAILURE')
    TPM_E_PPI_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80290303, 'TPM_E_PPI_NOT_SUPPORTED')
    TPM_E_PPI_BLOCKED_IN_BIOS: Final[ThirtyTwoBits] = cached_hresult(0x80290304, 'TPM_E_PPI_BLOCKED_IN_BIOS')
    TPM_E_PCP_ERROR_MASK: Final[ThirtyTwoBits] = cached_hresult(0x80290400, 'TPM_E_PCP_ERROR_MASK')
    TPM_E_PCP_DEVICE_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0x80290401, 'TPM_E_PCP_DEVICE_NOT_READY')
    TPM_E_PCP_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80290402, 'TPM_E_PCP_INVALID_HANDLE')
    TPM_E_PCP_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0x80290403, 'TPM_E_PCP_INVALID_PARAMETER')
    TPM_E_PCP_FLAG_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80290404, 'TPM_E_PCP_FLAG_NOT_SUPPORTED')
    TPM_E_PCP_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80290405, 'TPM_E_PCP_NOT_SUPPORTED')
    TPM_E_PCP_BUFFER_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0x80290406, 'TPM_E_PCP_BUFFER_TOO_SMALL')
    TPM_E_PCP_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80290407, 'TPM_E_PCP_INTERNAL_ERROR')
    TPM_E_PCP_AUTHENTICATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80290408, 'TPM_E_PCP_AUTHENTICATION_FAILED')
    TPM_E_PCP_AUTHENTICATION_IGNORED: Final[ThirtyTwoBits] = cached_hresult(0x80290409, 'TPM_E_PCP_AUTHENTICATION_IGNORED')
    TPM_E_PCP_POLICY_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8029040A, 'TPM_E_PCP_POLICY_NOT_FOUND')
    TPM_E_PCP_PROFILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8029040B, 'TPM_E_PCP_PROFILE_NOT_FOUND')
    TPM_E_PCP_VALIDATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8029040C, 'TPM_E_PCP_VALIDATION_FAILED')
    TPM_E_PCP_WRONG_PARENT: Final[ThirtyTwoBits] = cached_hresult(0x8029040E, 'TPM_E_PCP_WRONG_PARENT')
    TPM_E_KEY_NOT_LOADED: Final[ThirtyTwoBits] = cached_hresult(0x8029040F, 'TPM_E_KEY_NOT_LOADED')
    TPM_E_NO_KEY_CERTIFICATION: Final[ThirtyTwoBits] = cached_hresult(0x80290410, 'TPM_E_NO_KEY_CERTIFICATION')
    TPM_E_KEY_NOT_FINALIZED: Final[ThirtyTwoBits] = cached_hresult(0x80290411, 'TPM_E_KEY_NOT_FINALIZED')
    TPM_E_ATTESTATION_CHALLENGE_NOT_SET: Final[ThirtyTwoBits] = cached_hresult(0x80290412, 'TPM_E_ATTESTATION_CHALLENGE_NOT_SET')
    TPM_E_NOT_PCR_BOUND: Final[ThirtyTwoBits] = cached_hresult(0x80290413, 'TPM_E_NOT_PCR_BOUND')
    TPM_E_KEY_ALREADY_FINALIZED: Final[ThirtyTwoBits] = cached_hresult(0x80290414, 'TPM_E_KEY_ALREADY_FINALIZED')
    TPM_E_KEY_USAGE_POLICY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80290415, 'TPM_E_KEY_USAGE_POLICY_NOT_SUPPORTED')
    TPM_E_KEY_USAGE_POLICY_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80290416, 'TPM_E_KEY_USAGE_POLICY_INVALID')
    TPM_E_SOFT_KEY_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80290417, 'TPM_E_SOFT_KEY_ERROR')
    TPM_E_KEY_NOT_AUTHENTICATED: Final[ThirtyTwoBits] = cached_hresult(0x80290418, 'TPM_E_KEY_NOT_AUTHENTICATED')
    TPM_E_PCP_KEY_NOT_AIK: Final[ThirtyTwoBits] = cached_hresult(0x80290419, 'TPM_E_PCP_KEY_NOT_AIK')
    TPM_E_KEY_NOT_SIGNING_KEY: Final[ThirtyTwoBits] = cached_hresult(0x8029041A, 'TPM_E_KEY_NOT_SIGNING_KEY')
    TPM_E_LOCKED_OUT: Final[ThirtyTwoBits] = cached_hresult(0x8029041B, 'TPM_E_LOCKED_OUT')
    TPM_E_CLAIM_TYPE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8029041C, 'TPM_E_CLAIM_TYPE_NOT_SUPPORTED')
    TPM_E_VERSION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8029041D, 'TPM_E_VERSION_NOT_SUPPORTED')
    TPM_E_BUFFER_LENGTH_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x8029041E, 'TPM_E_BUFFER_LENGTH_MISMATCH')
    TPM_E_PCP_IFX_RSA_KEY_CREATION_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x8029041F, 'TPM_E_PCP_IFX_RSA_KEY_CREATION_BLOCKED')
    TPM_E_PCP_TICKET_MISSING: Final[ThirtyTwoBits] = cached_hresult(0x80290420, 'TPM_E_PCP_TICKET_MISSING')
    TPM_E_PCP_RAW_POLICY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80290421, 'TPM_E_PCP_RAW_POLICY_NOT_SUPPORTED')
    TPM_E_PCP_KEY_HANDLE_INVALIDATED: Final[ThirtyTwoBits] = cached_hresult(0x80290422, 'TPM_E_PCP_KEY_HANDLE_INVALIDATED')
    TPM_E_PCP_UNSUPPORTED_PSS_SALT: Final[ThirtyTwoBits] = cached_hresult(0x40290423, 'TPM_E_PCP_UNSUPPORTED_PSS_SALT')
    TPM_E_PCP_PLATFORM_CLAIM_MAY_BE_OUTDATED: Final[ThirtyTwoBits] = cached_hresult(0x40290424, 'TPM_E_PCP_PLATFORM_CLAIM_MAY_BE_OUTDATED')
    TPM_E_PCP_PLATFORM_CLAIM_OUTDATED: Final[ThirtyTwoBits] = cached_hresult(0x40290425, 'TPM_E_PCP_PLATFORM_CLAIM_OUTDATED')
    TPM_E_PCP_PLATFORM_CLAIM_REBOOT: Final[ThirtyTwoBits] = cached_hresult(0x40290426, 'TPM_E_PCP_PLATFORM_CLAIM_REBOOT')
    TPM_E_ZERO_EXHAUST_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0x80290500, 'TPM_E_ZERO_EXHAUST_ENABLED')
    TPM_E_PROVISIONING_INCOMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x80290600, 'TPM_E_PROVISIONING_INCOMPLETE')
    TPM_E_INVALID_OWNER_AUTH: Final[ThirtyTwoBits] = cached_hresult(0x80290601, 'TPM_E_INVALID_OWNER_AUTH')
    TPM_E_TOO_MUCH_DATA: Final[ThirtyTwoBits] = cached_hresult(0x80290602, 'TPM_E_TOO_MUCH_DATA')
    TPM_E_TPM_GENERATED_EPS: Final[ThirtyTwoBits] = cached_hresult(0x80290603, 'TPM_E_TPM_GENERATED_EPS')
    UI_E_CREATE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x802A0001, 'UI_E_CREATE_FAILED')
    UI_E_SHUTDOWN_CALLED: Final[ThirtyTwoBits] = cached_hresult(0x802A0002, 'UI_E_SHUTDOWN_CALLED')
    UI_E_ILLEGAL_REENTRANCY: Final[ThirtyTwoBits] = cached_hresult(0x802A0003, 'UI_E_ILLEGAL_REENTRANCY')
    UI_E_OBJECT_SEALED: Final[ThirtyTwoBits] = cached_hresult(0x802A0004, 'UI_E_OBJECT_SEALED')
    UI_E_VALUE_NOT_SET: Final[ThirtyTwoBits] = cached_hresult(0x802A0005, 'UI_E_VALUE_NOT_SET')
    UI_E_VALUE_NOT_DETERMINED: Final[ThirtyTwoBits] = cached_hresult(0x802A0006, 'UI_E_VALUE_NOT_DETERMINED')
    UI_E_INVALID_OUTPUT: Final[ThirtyTwoBits] = cached_hresult(0x802A0007, 'UI_E_INVALID_OUTPUT')
    UI_E_BOOLEAN_EXPECTED: Final[ThirtyTwoBits] = cached_hresult(0x802A0008, 'UI_E_BOOLEAN_EXPECTED')
    UI_E_DIFFERENT_OWNER: Final[ThirtyTwoBits] = cached_hresult(0x802A0009, 'UI_E_DIFFERENT_OWNER')
    UI_E_AMBIGUOUS_MATCH: Final[ThirtyTwoBits] = cached_hresult(0x802A000A, 'UI_E_AMBIGUOUS_MATCH')
    UI_E_FP_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0x802A000B, 'UI_E_FP_OVERFLOW')
    UI_E_WRONG_THREAD: Final[ThirtyTwoBits] = cached_hresult(0x802A000C, 'UI_E_WRONG_THREAD')
    UI_E_STORYBOARD_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0x802A0101, 'UI_E_STORYBOARD_ACTIVE')
    UI_E_STORYBOARD_NOT_PLAYING: Final[ThirtyTwoBits] = cached_hresult(0x802A0102, 'UI_E_STORYBOARD_NOT_PLAYING')
    UI_E_START_KEYFRAME_AFTER_END: Final[ThirtyTwoBits] = cached_hresult(0x802A0103, 'UI_E_START_KEYFRAME_AFTER_END')
    UI_E_END_KEYFRAME_NOT_DETERMINED: Final[ThirtyTwoBits] = cached_hresult(0x802A0104, 'UI_E_END_KEYFRAME_NOT_DETERMINED')
    UI_E_LOOPS_OVERLAP: Final[ThirtyTwoBits] = cached_hresult(0x802A0105, 'UI_E_LOOPS_OVERLAP')
    UI_E_TRANSITION_ALREADY_USED: Final[ThirtyTwoBits] = cached_hresult(0x802A0106, 'UI_E_TRANSITION_ALREADY_USED')
    UI_E_TRANSITION_NOT_IN_STORYBOARD: Final[ThirtyTwoBits] = cached_hresult(0x802A0107, 'UI_E_TRANSITION_NOT_IN_STORYBOARD')
    UI_E_TRANSITION_ECLIPSED: Final[ThirtyTwoBits] = cached_hresult(0x802A0108, 'UI_E_TRANSITION_ECLIPSED')
    UI_E_TIME_BEFORE_LAST_UPDATE: Final[ThirtyTwoBits] = cached_hresult(0x802A0109, 'UI_E_TIME_BEFORE_LAST_UPDATE')
    UI_E_TIMER_CLIENT_ALREADY_CONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x802A010A, 'UI_E_TIMER_CLIENT_ALREADY_CONNECTED')
    UI_E_INVALID_DIMENSION: Final[ThirtyTwoBits] = cached_hresult(0x802A010B, 'UI_E_INVALID_DIMENSION')
    UI_E_PRIMITIVE_OUT_OF_BOUNDS: Final[ThirtyTwoBits] = cached_hresult(0x802A010C, 'UI_E_PRIMITIVE_OUT_OF_BOUNDS')
    UI_E_WINDOW_CLOSED: Final[ThirtyTwoBits] = cached_hresult(0x802A0201, 'UI_E_WINDOW_CLOSED')
    PLA_E_DCS_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80300002, 'PLA_E_DCS_NOT_FOUND')
    PLA_E_TOO_MANY_FOLDERS: Final[ThirtyTwoBits] = cached_hresult(0x80300045, 'PLA_E_TOO_MANY_FOLDERS')
    PLA_E_NO_MIN_DISK: Final[ThirtyTwoBits] = cached_hresult(0x80300070, 'PLA_E_NO_MIN_DISK')
    PLA_E_DCS_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x803000AA, 'PLA_E_DCS_IN_USE')
    PLA_E_DCS_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x803000B7, 'PLA_E_DCS_ALREADY_EXISTS')
    PLA_S_PROPERTY_IGNORED: Final[ThirtyTwoBits] = cached_hresult(0x00300100, 'PLA_S_PROPERTY_IGNORED')
    PLA_E_PROPERTY_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x80300101, 'PLA_E_PROPERTY_CONFLICT')
    PLA_E_DCS_SINGLETON_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80300102, 'PLA_E_DCS_SINGLETON_REQUIRED')
    PLA_E_CREDENTIALS_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80300103, 'PLA_E_CREDENTIALS_REQUIRED')
    PLA_E_DCS_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x80300104, 'PLA_E_DCS_NOT_RUNNING')
    PLA_E_CONFLICT_INCL_EXCL_API: Final[ThirtyTwoBits] = cached_hresult(0x80300105, 'PLA_E_CONFLICT_INCL_EXCL_API')
    PLA_E_NETWORK_EXE_NOT_VALID: Final[ThirtyTwoBits] = cached_hresult(0x80300106, 'PLA_E_NETWORK_EXE_NOT_VALID')
    PLA_E_EXE_ALREADY_CONFIGURED: Final[ThirtyTwoBits] = cached_hresult(0x80300107, 'PLA_E_EXE_ALREADY_CONFIGURED')
    PLA_E_EXE_PATH_NOT_VALID: Final[ThirtyTwoBits] = cached_hresult(0x80300108, 'PLA_E_EXE_PATH_NOT_VALID')
    PLA_E_DC_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80300109, 'PLA_E_DC_ALREADY_EXISTS')
    PLA_E_DCS_START_WAIT_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x8030010A, 'PLA_E_DCS_START_WAIT_TIMEOUT')
    PLA_E_DC_START_WAIT_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x8030010B, 'PLA_E_DC_START_WAIT_TIMEOUT')
    PLA_E_REPORT_WAIT_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x8030010C, 'PLA_E_REPORT_WAIT_TIMEOUT')
    PLA_E_NO_DUPLICATES: Final[ThirtyTwoBits] = cached_hresult(0x8030010D, 'PLA_E_NO_DUPLICATES')
    PLA_E_EXE_FULL_PATH_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8030010E, 'PLA_E_EXE_FULL_PATH_REQUIRED')
    PLA_E_INVALID_SESSION_NAME: Final[ThirtyTwoBits] = cached_hresult(0x8030010F, 'PLA_E_INVALID_SESSION_NAME')
    PLA_E_PLA_CHANNEL_NOT_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0x80300110, 'PLA_E_PLA_CHANNEL_NOT_ENABLED')
    PLA_E_TASKSCHED_CHANNEL_NOT_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0x80300111, 'PLA_E_TASKSCHED_CHANNEL_NOT_ENABLED')
    PLA_E_RULES_MANAGER_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80300112, 'PLA_E_RULES_MANAGER_FAILED')
    PLA_E_CABAPI_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80300113, 'PLA_E_CABAPI_FAILURE')
    FVE_E_LOCKED_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0x80310000, 'FVE_E_LOCKED_VOLUME')
    FVE_E_NOT_ENCRYPTED: Final[ThirtyTwoBits] = cached_hresult(0x80310001, 'FVE_E_NOT_ENCRYPTED')
    FVE_E_NO_TPM_BIOS: Final[ThirtyTwoBits] = cached_hresult(0x80310002, 'FVE_E_NO_TPM_BIOS')
    FVE_E_NO_MBR_METRIC: Final[ThirtyTwoBits] = cached_hresult(0x80310003, 'FVE_E_NO_MBR_METRIC')
    FVE_E_NO_BOOTSECTOR_METRIC: Final[ThirtyTwoBits] = cached_hresult(0x80310004, 'FVE_E_NO_BOOTSECTOR_METRIC')
    FVE_E_NO_BOOTMGR_METRIC: Final[ThirtyTwoBits] = cached_hresult(0x80310005, 'FVE_E_NO_BOOTMGR_METRIC')
    FVE_E_WRONG_BOOTMGR: Final[ThirtyTwoBits] = cached_hresult(0x80310006, 'FVE_E_WRONG_BOOTMGR')
    FVE_E_SECURE_KEY_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80310007, 'FVE_E_SECURE_KEY_REQUIRED')
    FVE_E_NOT_ACTIVATED: Final[ThirtyTwoBits] = cached_hresult(0x80310008, 'FVE_E_NOT_ACTIVATED')
    FVE_E_ACTION_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80310009, 'FVE_E_ACTION_NOT_ALLOWED')
    FVE_E_AD_SCHEMA_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x8031000A, 'FVE_E_AD_SCHEMA_NOT_INSTALLED')
    FVE_E_AD_INVALID_DATATYPE: Final[ThirtyTwoBits] = cached_hresult(0x8031000B, 'FVE_E_AD_INVALID_DATATYPE')
    FVE_E_AD_INVALID_DATASIZE: Final[ThirtyTwoBits] = cached_hresult(0x8031000C, 'FVE_E_AD_INVALID_DATASIZE')
    FVE_E_AD_NO_VALUES: Final[ThirtyTwoBits] = cached_hresult(0x8031000D, 'FVE_E_AD_NO_VALUES')
    FVE_E_AD_ATTR_NOT_SET: Final[ThirtyTwoBits] = cached_hresult(0x8031000E, 'FVE_E_AD_ATTR_NOT_SET')
    FVE_E_AD_GUID_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8031000F, 'FVE_E_AD_GUID_NOT_FOUND')
    FVE_E_BAD_INFORMATION: Final[ThirtyTwoBits] = cached_hresult(0x80310010, 'FVE_E_BAD_INFORMATION')
    FVE_E_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0x80310011, 'FVE_E_TOO_SMALL')
    FVE_E_SYSTEM_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0x80310012, 'FVE_E_SYSTEM_VOLUME')
    FVE_E_FAILED_WRONG_FS: Final[ThirtyTwoBits] = cached_hresult(0x80310013, 'FVE_E_FAILED_WRONG_FS')
    FVE_E_BAD_PARTITION_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x80310014, 'FVE_E_BAD_PARTITION_SIZE')
    FVE_E_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80310015, 'FVE_E_NOT_SUPPORTED')
    FVE_E_BAD_DATA: Final[ThirtyTwoBits] = cached_hresult(0x80310016, 'FVE_E_BAD_DATA')
    FVE_E_VOLUME_NOT_BOUND: Final[ThirtyTwoBits] = cached_hresult(0x80310017, 'FVE_E_VOLUME_NOT_BOUND')
    FVE_E_TPM_NOT_OWNED: Final[ThirtyTwoBits] = cached_hresult(0x80310018, 'FVE_E_TPM_NOT_OWNED')
    FVE_E_NOT_DATA_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0x80310019, 'FVE_E_NOT_DATA_VOLUME')
    FVE_E_AD_INSUFFICIENT_BUFFER: Final[ThirtyTwoBits] = cached_hresult(0x8031001A, 'FVE_E_AD_INSUFFICIENT_BUFFER')
    FVE_E_CONV_READ: Final[ThirtyTwoBits] = cached_hresult(0x8031001B, 'FVE_E_CONV_READ')
    FVE_E_CONV_WRITE: Final[ThirtyTwoBits] = cached_hresult(0x8031001C, 'FVE_E_CONV_WRITE')
    FVE_E_KEY_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8031001D, 'FVE_E_KEY_REQUIRED')
    FVE_E_CLUSTERING_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8031001E, 'FVE_E_CLUSTERING_NOT_SUPPORTED')
    FVE_E_VOLUME_BOUND_ALREADY: Final[ThirtyTwoBits] = cached_hresult(0x8031001F, 'FVE_E_VOLUME_BOUND_ALREADY')
    FVE_E_OS_NOT_PROTECTED: Final[ThirtyTwoBits] = cached_hresult(0x80310020, 'FVE_E_OS_NOT_PROTECTED')
    FVE_E_PROTECTION_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80310021, 'FVE_E_PROTECTION_DISABLED')
    FVE_E_RECOVERY_KEY_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80310022, 'FVE_E_RECOVERY_KEY_REQUIRED')
    FVE_E_FOREIGN_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0x80310023, 'FVE_E_FOREIGN_VOLUME')
    FVE_E_OVERLAPPED_UPDATE: Final[ThirtyTwoBits] = cached_hresult(0x80310024, 'FVE_E_OVERLAPPED_UPDATE')
    FVE_E_TPM_SRK_AUTH_NOT_ZERO: Final[ThirtyTwoBits] = cached_hresult(0x80310025, 'FVE_E_TPM_SRK_AUTH_NOT_ZERO')
    FVE_E_FAILED_SECTOR_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x80310026, 'FVE_E_FAILED_SECTOR_SIZE')
    FVE_E_FAILED_AUTHENTICATION: Final[ThirtyTwoBits] = cached_hresult(0x80310027, 'FVE_E_FAILED_AUTHENTICATION')
    FVE_E_NOT_OS_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0x80310028, 'FVE_E_NOT_OS_VOLUME')
    FVE_E_AUTOUNLOCK_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0x80310029, 'FVE_E_AUTOUNLOCK_ENABLED')
    FVE_E_WRONG_BOOTSECTOR: Final[ThirtyTwoBits] = cached_hresult(0x8031002A, 'FVE_E_WRONG_BOOTSECTOR')
    FVE_E_WRONG_SYSTEM_FS: Final[ThirtyTwoBits] = cached_hresult(0x8031002B, 'FVE_E_WRONG_SYSTEM_FS')
    FVE_E_POLICY_PASSWORD_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8031002C, 'FVE_E_POLICY_PASSWORD_REQUIRED')
    FVE_E_CANNOT_SET_FVEK_ENCRYPTED: Final[ThirtyTwoBits] = cached_hresult(0x8031002D, 'FVE_E_CANNOT_SET_FVEK_ENCRYPTED')
    FVE_E_CANNOT_ENCRYPT_NO_KEY: Final[ThirtyTwoBits] = cached_hresult(0x8031002E, 'FVE_E_CANNOT_ENCRYPT_NO_KEY')
    FVE_E_BOOTABLE_CDDVD: Final[ThirtyTwoBits] = cached_hresult(0x80310030, 'FVE_E_BOOTABLE_CDDVD')
    FVE_E_PROTECTOR_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80310031, 'FVE_E_PROTECTOR_EXISTS')
    FVE_E_RELATIVE_PATH: Final[ThirtyTwoBits] = cached_hresult(0x80310032, 'FVE_E_RELATIVE_PATH')
    FVE_E_PROTECTOR_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80310033, 'FVE_E_PROTECTOR_NOT_FOUND')
    FVE_E_INVALID_KEY_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0x80310034, 'FVE_E_INVALID_KEY_FORMAT')
    FVE_E_INVALID_PASSWORD_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0x80310035, 'FVE_E_INVALID_PASSWORD_FORMAT')
    FVE_E_FIPS_RNG_CHECK_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80310036, 'FVE_E_FIPS_RNG_CHECK_FAILED')
    FVE_E_FIPS_PREVENTS_RECOVERY_PASSWORD: Final[ThirtyTwoBits] = cached_hresult(0x80310037, 'FVE_E_FIPS_PREVENTS_RECOVERY_PASSWORD')
    FVE_E_FIPS_PREVENTS_EXTERNAL_KEY_EXPORT: Final[ThirtyTwoBits] = cached_hresult(0x80310038, 'FVE_E_FIPS_PREVENTS_EXTERNAL_KEY_EXPORT')
    FVE_E_NOT_DECRYPTED: Final[ThirtyTwoBits] = cached_hresult(0x80310039, 'FVE_E_NOT_DECRYPTED')
    FVE_E_INVALID_PROTECTOR_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x8031003A, 'FVE_E_INVALID_PROTECTOR_TYPE')
    FVE_E_NO_PROTECTORS_TO_TEST: Final[ThirtyTwoBits] = cached_hresult(0x8031003B, 'FVE_E_NO_PROTECTORS_TO_TEST')
    FVE_E_KEYFILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8031003C, 'FVE_E_KEYFILE_NOT_FOUND')
    FVE_E_KEYFILE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x8031003D, 'FVE_E_KEYFILE_INVALID')
    FVE_E_KEYFILE_NO_VMK: Final[ThirtyTwoBits] = cached_hresult(0x8031003E, 'FVE_E_KEYFILE_NO_VMK')
    FVE_E_TPM_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x8031003F, 'FVE_E_TPM_DISABLED')
    FVE_E_NOT_ALLOWED_IN_SAFE_MODE: Final[ThirtyTwoBits] = cached_hresult(0x80310040, 'FVE_E_NOT_ALLOWED_IN_SAFE_MODE')
    FVE_E_TPM_INVALID_PCR: Final[ThirtyTwoBits] = cached_hresult(0x80310041, 'FVE_E_TPM_INVALID_PCR')
    FVE_E_TPM_NO_VMK: Final[ThirtyTwoBits] = cached_hresult(0x80310042, 'FVE_E_TPM_NO_VMK')
    FVE_E_PIN_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80310043, 'FVE_E_PIN_INVALID')
    FVE_E_AUTH_INVALID_APPLICATION: Final[ThirtyTwoBits] = cached_hresult(0x80310044, 'FVE_E_AUTH_INVALID_APPLICATION')
    FVE_E_AUTH_INVALID_CONFIG: Final[ThirtyTwoBits] = cached_hresult(0x80310045, 'FVE_E_AUTH_INVALID_CONFIG')
    FVE_E_FIPS_DISABLE_PROTECTION_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80310046, 'FVE_E_FIPS_DISABLE_PROTECTION_NOT_ALLOWED')
    FVE_E_FS_NOT_EXTENDED: Final[ThirtyTwoBits] = cached_hresult(0x80310047, 'FVE_E_FS_NOT_EXTENDED')
    FVE_E_FIRMWARE_TYPE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80310048, 'FVE_E_FIRMWARE_TYPE_NOT_SUPPORTED')
    FVE_E_NO_LICENSE: Final[ThirtyTwoBits] = cached_hresult(0x80310049, 'FVE_E_NO_LICENSE')
    FVE_E_NOT_ON_STACK: Final[ThirtyTwoBits] = cached_hresult(0x8031004A, 'FVE_E_NOT_ON_STACK')
    FVE_E_FS_MOUNTED: Final[ThirtyTwoBits] = cached_hresult(0x8031004B, 'FVE_E_FS_MOUNTED')
    FVE_E_TOKEN_NOT_IMPERSONATED: Final[ThirtyTwoBits] = cached_hresult(0x8031004C, 'FVE_E_TOKEN_NOT_IMPERSONATED')
    FVE_E_DRY_RUN_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8031004D, 'FVE_E_DRY_RUN_FAILED')
    FVE_E_REBOOT_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8031004E, 'FVE_E_REBOOT_REQUIRED')
    FVE_E_DEBUGGER_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0x8031004F, 'FVE_E_DEBUGGER_ENABLED')
    FVE_E_RAW_ACCESS: Final[ThirtyTwoBits] = cached_hresult(0x80310050, 'FVE_E_RAW_ACCESS')
    FVE_E_RAW_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x80310051, 'FVE_E_RAW_BLOCKED')
    FVE_E_BCD_APPLICATIONS_PATH_INCORRECT: Final[ThirtyTwoBits] = cached_hresult(0x80310052, 'FVE_E_BCD_APPLICATIONS_PATH_INCORRECT')
    FVE_E_NOT_ALLOWED_IN_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80310053, 'FVE_E_NOT_ALLOWED_IN_VERSION')
    FVE_E_NO_AUTOUNLOCK_MASTER_KEY: Final[ThirtyTwoBits] = cached_hresult(0x80310054, 'FVE_E_NO_AUTOUNLOCK_MASTER_KEY')
    FVE_E_MOR_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80310055, 'FVE_E_MOR_FAILED')
    FVE_E_HIDDEN_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0x80310056, 'FVE_E_HIDDEN_VOLUME')
    FVE_E_TRANSIENT_STATE: Final[ThirtyTwoBits] = cached_hresult(0x80310057, 'FVE_E_TRANSIENT_STATE')
    FVE_E_PUBKEY_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80310058, 'FVE_E_PUBKEY_NOT_ALLOWED')
    FVE_E_VOLUME_HANDLE_OPEN: Final[ThirtyTwoBits] = cached_hresult(0x80310059, 'FVE_E_VOLUME_HANDLE_OPEN')
    FVE_E_NO_FEATURE_LICENSE: Final[ThirtyTwoBits] = cached_hresult(0x8031005A, 'FVE_E_NO_FEATURE_LICENSE')
    FVE_E_INVALID_STARTUP_OPTIONS: Final[ThirtyTwoBits] = cached_hresult(0x8031005B, 'FVE_E_INVALID_STARTUP_OPTIONS')
    FVE_E_POLICY_RECOVERY_PASSWORD_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x8031005C, 'FVE_E_POLICY_RECOVERY_PASSWORD_NOT_ALLOWED')
    FVE_E_POLICY_RECOVERY_PASSWORD_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8031005D, 'FVE_E_POLICY_RECOVERY_PASSWORD_REQUIRED')
    FVE_E_POLICY_RECOVERY_KEY_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x8031005E, 'FVE_E_POLICY_RECOVERY_KEY_NOT_ALLOWED')
    FVE_E_POLICY_RECOVERY_KEY_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8031005F, 'FVE_E_POLICY_RECOVERY_KEY_REQUIRED')
    FVE_E_POLICY_STARTUP_PIN_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80310060, 'FVE_E_POLICY_STARTUP_PIN_NOT_ALLOWED')
    FVE_E_POLICY_STARTUP_PIN_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80310061, 'FVE_E_POLICY_STARTUP_PIN_REQUIRED')
    FVE_E_POLICY_STARTUP_KEY_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80310062, 'FVE_E_POLICY_STARTUP_KEY_NOT_ALLOWED')
    FVE_E_POLICY_STARTUP_KEY_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80310063, 'FVE_E_POLICY_STARTUP_KEY_REQUIRED')
    FVE_E_POLICY_STARTUP_PIN_KEY_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80310064, 'FVE_E_POLICY_STARTUP_PIN_KEY_NOT_ALLOWED')
    FVE_E_POLICY_STARTUP_PIN_KEY_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80310065, 'FVE_E_POLICY_STARTUP_PIN_KEY_REQUIRED')
    FVE_E_POLICY_STARTUP_TPM_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80310066, 'FVE_E_POLICY_STARTUP_TPM_NOT_ALLOWED')
    FVE_E_POLICY_STARTUP_TPM_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80310067, 'FVE_E_POLICY_STARTUP_TPM_REQUIRED')
    FVE_E_POLICY_INVALID_PIN_LENGTH: Final[ThirtyTwoBits] = cached_hresult(0x80310068, 'FVE_E_POLICY_INVALID_PIN_LENGTH')
    FVE_E_KEY_PROTECTOR_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80310069, 'FVE_E_KEY_PROTECTOR_NOT_SUPPORTED')
    FVE_E_POLICY_PASSPHRASE_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x8031006A, 'FVE_E_POLICY_PASSPHRASE_NOT_ALLOWED')
    FVE_E_POLICY_PASSPHRASE_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8031006B, 'FVE_E_POLICY_PASSPHRASE_REQUIRED')
    FVE_E_FIPS_PREVENTS_PASSPHRASE: Final[ThirtyTwoBits] = cached_hresult(0x8031006C, 'FVE_E_FIPS_PREVENTS_PASSPHRASE')
    FVE_E_OS_VOLUME_PASSPHRASE_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x8031006D, 'FVE_E_OS_VOLUME_PASSPHRASE_NOT_ALLOWED')
    FVE_E_INVALID_BITLOCKER_OID: Final[ThirtyTwoBits] = cached_hresult(0x8031006E, 'FVE_E_INVALID_BITLOCKER_OID')
    FVE_E_VOLUME_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0x8031006F, 'FVE_E_VOLUME_TOO_SMALL')
    FVE_E_DV_NOT_SUPPORTED_ON_FS: Final[ThirtyTwoBits] = cached_hresult(0x80310070, 'FVE_E_DV_NOT_SUPPORTED_ON_FS')
    FVE_E_DV_NOT_ALLOWED_BY_GP: Final[ThirtyTwoBits] = cached_hresult(0x80310071, 'FVE_E_DV_NOT_ALLOWED_BY_GP')
    FVE_E_POLICY_USER_CERTIFICATE_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80310072, 'FVE_E_POLICY_USER_CERTIFICATE_NOT_ALLOWED')
    FVE_E_POLICY_USER_CERTIFICATE_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80310073, 'FVE_E_POLICY_USER_CERTIFICATE_REQUIRED')
    FVE_E_POLICY_USER_CERT_MUST_BE_HW: Final[ThirtyTwoBits] = cached_hresult(0x80310074, 'FVE_E_POLICY_USER_CERT_MUST_BE_HW')
    FVE_E_POLICY_USER_CONFIGURE_FDV_AUTOUNLOCK_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80310075, 'FVE_E_POLICY_USER_CONFIGURE_FDV_AUTOUNLOCK_NOT_ALLOWED')
    FVE_E_POLICY_USER_CONFIGURE_RDV_AUTOUNLOCK_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80310076, 'FVE_E_POLICY_USER_CONFIGURE_RDV_AUTOUNLOCK_NOT_ALLOWED')
    FVE_E_POLICY_USER_CONFIGURE_RDV_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80310077, 'FVE_E_POLICY_USER_CONFIGURE_RDV_NOT_ALLOWED')
    FVE_E_POLICY_USER_ENABLE_RDV_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80310078, 'FVE_E_POLICY_USER_ENABLE_RDV_NOT_ALLOWED')
    FVE_E_POLICY_USER_DISABLE_RDV_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80310079, 'FVE_E_POLICY_USER_DISABLE_RDV_NOT_ALLOWED')
    FVE_E_POLICY_INVALID_PASSPHRASE_LENGTH: Final[ThirtyTwoBits] = cached_hresult(0x80310080, 'FVE_E_POLICY_INVALID_PASSPHRASE_LENGTH')
    FVE_E_POLICY_PASSPHRASE_TOO_SIMPLE: Final[ThirtyTwoBits] = cached_hresult(0x80310081, 'FVE_E_POLICY_PASSPHRASE_TOO_SIMPLE')
    FVE_E_RECOVERY_PARTITION: Final[ThirtyTwoBits] = cached_hresult(0x80310082, 'FVE_E_RECOVERY_PARTITION')
    FVE_E_POLICY_CONFLICT_FDV_RK_OFF_AUK_ON: Final[ThirtyTwoBits] = cached_hresult(0x80310083, 'FVE_E_POLICY_CONFLICT_FDV_RK_OFF_AUK_ON')
    FVE_E_POLICY_CONFLICT_RDV_RK_OFF_AUK_ON: Final[ThirtyTwoBits] = cached_hresult(0x80310084, 'FVE_E_POLICY_CONFLICT_RDV_RK_OFF_AUK_ON')
    FVE_E_NON_BITLOCKER_OID: Final[ThirtyTwoBits] = cached_hresult(0x80310085, 'FVE_E_NON_BITLOCKER_OID')
    FVE_E_POLICY_PROHIBITS_SELFSIGNED: Final[ThirtyTwoBits] = cached_hresult(0x80310086, 'FVE_E_POLICY_PROHIBITS_SELFSIGNED')
    FVE_E_POLICY_CONFLICT_RO_AND_STARTUP_KEY_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80310087, 'FVE_E_POLICY_CONFLICT_RO_AND_STARTUP_KEY_REQUIRED')
    FVE_E_CONV_RECOVERY_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80310088, 'FVE_E_CONV_RECOVERY_FAILED')
    FVE_E_VIRTUALIZED_SPACE_TOO_BIG: Final[ThirtyTwoBits] = cached_hresult(0x80310089, 'FVE_E_VIRTUALIZED_SPACE_TOO_BIG')
    FVE_E_POLICY_CONFLICT_OSV_RP_OFF_ADB_ON: Final[ThirtyTwoBits] = cached_hresult(0x80310090, 'FVE_E_POLICY_CONFLICT_OSV_RP_OFF_ADB_ON')
    FVE_E_POLICY_CONFLICT_FDV_RP_OFF_ADB_ON: Final[ThirtyTwoBits] = cached_hresult(0x80310091, 'FVE_E_POLICY_CONFLICT_FDV_RP_OFF_ADB_ON')
    FVE_E_POLICY_CONFLICT_RDV_RP_OFF_ADB_ON: Final[ThirtyTwoBits] = cached_hresult(0x80310092, 'FVE_E_POLICY_CONFLICT_RDV_RP_OFF_ADB_ON')
    FVE_E_NON_BITLOCKER_KU: Final[ThirtyTwoBits] = cached_hresult(0x80310093, 'FVE_E_NON_BITLOCKER_KU')
    FVE_E_PRIVATEKEY_AUTH_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80310094, 'FVE_E_PRIVATEKEY_AUTH_FAILED')
    FVE_E_REMOVAL_OF_DRA_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80310095, 'FVE_E_REMOVAL_OF_DRA_FAILED')
    FVE_E_OPERATION_NOT_SUPPORTED_ON_VISTA_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0x80310096, 'FVE_E_OPERATION_NOT_SUPPORTED_ON_VISTA_VOLUME')
    FVE_E_CANT_LOCK_AUTOUNLOCK_ENABLED_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0x80310097, 'FVE_E_CANT_LOCK_AUTOUNLOCK_ENABLED_VOLUME')
    FVE_E_FIPS_HASH_KDF_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80310098, 'FVE_E_FIPS_HASH_KDF_NOT_ALLOWED')
    FVE_E_ENH_PIN_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80310099, 'FVE_E_ENH_PIN_INVALID')
    FVE_E_INVALID_PIN_CHARS: Final[ThirtyTwoBits] = cached_hresult(0x8031009A, 'FVE_E_INVALID_PIN_CHARS')
    FVE_E_INVALID_DATUM_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x8031009B, 'FVE_E_INVALID_DATUM_TYPE')
    FVE_E_EFI_ONLY: Final[ThirtyTwoBits] = cached_hresult(0x8031009C, 'FVE_E_EFI_ONLY')
    FVE_E_MULTIPLE_NKP_CERTS: Final[ThirtyTwoBits] = cached_hresult(0x8031009D, 'FVE_E_MULTIPLE_NKP_CERTS')
    FVE_E_REMOVAL_OF_NKP_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8031009E, 'FVE_E_REMOVAL_OF_NKP_FAILED')
    FVE_E_INVALID_NKP_CERT: Final[ThirtyTwoBits] = cached_hresult(0x8031009F, 'FVE_E_INVALID_NKP_CERT')
    FVE_E_NO_EXISTING_PIN: Final[ThirtyTwoBits] = cached_hresult(0x803100A0, 'FVE_E_NO_EXISTING_PIN')
    FVE_E_PROTECTOR_CHANGE_PIN_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x803100A1, 'FVE_E_PROTECTOR_CHANGE_PIN_MISMATCH')
    FVE_E_PIN_PROTECTOR_CHANGE_BY_STD_USER_DISALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x803100A2, 'FVE_E_PIN_PROTECTOR_CHANGE_BY_STD_USER_DISALLOWED')
    FVE_E_PROTECTOR_CHANGE_MAX_PIN_CHANGE_ATTEMPTS_REACHED: Final[ThirtyTwoBits] = cached_hresult(0x803100A3, 'FVE_E_PROTECTOR_CHANGE_MAX_PIN_CHANGE_ATTEMPTS_REACHED')
    FVE_E_POLICY_PASSPHRASE_REQUIRES_ASCII: Final[ThirtyTwoBits] = cached_hresult(0x803100A4, 'FVE_E_POLICY_PASSPHRASE_REQUIRES_ASCII')
    FVE_E_FULL_ENCRYPTION_NOT_ALLOWED_ON_TP_STORAGE: Final[ThirtyTwoBits] = cached_hresult(0x803100A5, 'FVE_E_FULL_ENCRYPTION_NOT_ALLOWED_ON_TP_STORAGE')
    FVE_E_WIPE_NOT_ALLOWED_ON_TP_STORAGE: Final[ThirtyTwoBits] = cached_hresult(0x803100A6, 'FVE_E_WIPE_NOT_ALLOWED_ON_TP_STORAGE')
    FVE_E_KEY_LENGTH_NOT_SUPPORTED_BY_EDRIVE: Final[ThirtyTwoBits] = cached_hresult(0x803100A7, 'FVE_E_KEY_LENGTH_NOT_SUPPORTED_BY_EDRIVE')
    FVE_E_NO_EXISTING_PASSPHRASE: Final[ThirtyTwoBits] = cached_hresult(0x803100A8, 'FVE_E_NO_EXISTING_PASSPHRASE')
    FVE_E_PROTECTOR_CHANGE_PASSPHRASE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x803100A9, 'FVE_E_PROTECTOR_CHANGE_PASSPHRASE_MISMATCH')
    FVE_E_PASSPHRASE_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0x803100AA, 'FVE_E_PASSPHRASE_TOO_LONG')
    FVE_E_NO_PASSPHRASE_WITH_TPM: Final[ThirtyTwoBits] = cached_hresult(0x803100AB, 'FVE_E_NO_PASSPHRASE_WITH_TPM')
    FVE_E_NO_TPM_WITH_PASSPHRASE: Final[ThirtyTwoBits] = cached_hresult(0x803100AC, 'FVE_E_NO_TPM_WITH_PASSPHRASE')
    FVE_E_NOT_ALLOWED_ON_CSV_STACK: Final[ThirtyTwoBits] = cached_hresult(0x803100AD, 'FVE_E_NOT_ALLOWED_ON_CSV_STACK')
    FVE_E_NOT_ALLOWED_ON_CLUSTER: Final[ThirtyTwoBits] = cached_hresult(0x803100AE, 'FVE_E_NOT_ALLOWED_ON_CLUSTER')
    FVE_E_EDRIVE_NO_FAILOVER_TO_SW: Final[ThirtyTwoBits] = cached_hresult(0x803100AF, 'FVE_E_EDRIVE_NO_FAILOVER_TO_SW')
    FVE_E_EDRIVE_BAND_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x803100B0, 'FVE_E_EDRIVE_BAND_IN_USE')
    FVE_E_EDRIVE_DISALLOWED_BY_GP: Final[ThirtyTwoBits] = cached_hresult(0x803100B1, 'FVE_E_EDRIVE_DISALLOWED_BY_GP')
    FVE_E_EDRIVE_INCOMPATIBLE_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0x803100B2, 'FVE_E_EDRIVE_INCOMPATIBLE_VOLUME')
    FVE_E_NOT_ALLOWED_TO_UPGRADE_WHILE_CONVERTING: Final[ThirtyTwoBits] = cached_hresult(0x803100B3, 'FVE_E_NOT_ALLOWED_TO_UPGRADE_WHILE_CONVERTING')
    FVE_E_EDRIVE_DV_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x803100B4, 'FVE_E_EDRIVE_DV_NOT_SUPPORTED')
    FVE_E_NO_PREBOOT_KEYBOARD_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0x803100B5, 'FVE_E_NO_PREBOOT_KEYBOARD_DETECTED')
    FVE_E_NO_PREBOOT_KEYBOARD_OR_WINRE_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0x803100B6, 'FVE_E_NO_PREBOOT_KEYBOARD_OR_WINRE_DETECTED')
    FVE_E_POLICY_REQUIRES_STARTUP_PIN_ON_TOUCH_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0x803100B7, 'FVE_E_POLICY_REQUIRES_STARTUP_PIN_ON_TOUCH_DEVICE')
    FVE_E_POLICY_REQUIRES_RECOVERY_PASSWORD_ON_TOUCH_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0x803100B8, 'FVE_E_POLICY_REQUIRES_RECOVERY_PASSWORD_ON_TOUCH_DEVICE')
    FVE_E_WIPE_CANCEL_NOT_APPLICABLE: Final[ThirtyTwoBits] = cached_hresult(0x803100B9, 'FVE_E_WIPE_CANCEL_NOT_APPLICABLE')
    FVE_E_SECUREBOOT_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x803100BA, 'FVE_E_SECUREBOOT_DISABLED')
    FVE_E_SECUREBOOT_CONFIGURATION_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x803100BB, 'FVE_E_SECUREBOOT_CONFIGURATION_INVALID')
    FVE_E_EDRIVE_DRY_RUN_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x803100BC, 'FVE_E_EDRIVE_DRY_RUN_FAILED')
    FVE_E_SHADOW_COPY_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0x803100BD, 'FVE_E_SHADOW_COPY_PRESENT')
    FVE_E_POLICY_INVALID_ENHANCED_BCD_SETTINGS: Final[ThirtyTwoBits] = cached_hresult(0x803100BE, 'FVE_E_POLICY_INVALID_ENHANCED_BCD_SETTINGS')
    FVE_E_EDRIVE_INCOMPATIBLE_FIRMWARE: Final[ThirtyTwoBits] = cached_hresult(0x803100BF, 'FVE_E_EDRIVE_INCOMPATIBLE_FIRMWARE')
    FVE_E_PROTECTOR_CHANGE_MAX_PASSPHRASE_CHANGE_ATTEMPTS_REACHED: Final[ThirtyTwoBits] = cached_hresult(0x803100C0, 'FVE_E_PROTECTOR_CHANGE_MAX_PASSPHRASE_CHANGE_ATTEMPTS_REACHED')
    FVE_E_PASSPHRASE_PROTECTOR_CHANGE_BY_STD_USER_DISALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x803100C1, 'FVE_E_PASSPHRASE_PROTECTOR_CHANGE_BY_STD_USER_DISALLOWED')
    FVE_E_LIVEID_ACCOUNT_SUSPENDED: Final[ThirtyTwoBits] = cached_hresult(0x803100C2, 'FVE_E_LIVEID_ACCOUNT_SUSPENDED')
    FVE_E_LIVEID_ACCOUNT_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x803100C3, 'FVE_E_LIVEID_ACCOUNT_BLOCKED')
    FVE_E_NOT_PROVISIONED_ON_ALL_VOLUMES: Final[ThirtyTwoBits] = cached_hresult(0x803100C4, 'FVE_E_NOT_PROVISIONED_ON_ALL_VOLUMES')
    FVE_E_DE_FIXED_DATA_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x803100C5, 'FVE_E_DE_FIXED_DATA_NOT_SUPPORTED')
    FVE_E_DE_HARDWARE_NOT_COMPLIANT: Final[ThirtyTwoBits] = cached_hresult(0x803100C6, 'FVE_E_DE_HARDWARE_NOT_COMPLIANT')
    FVE_E_DE_WINRE_NOT_CONFIGURED: Final[ThirtyTwoBits] = cached_hresult(0x803100C7, 'FVE_E_DE_WINRE_NOT_CONFIGURED')
    FVE_E_DE_PROTECTION_SUSPENDED: Final[ThirtyTwoBits] = cached_hresult(0x803100C8, 'FVE_E_DE_PROTECTION_SUSPENDED')
    FVE_E_DE_OS_VOLUME_NOT_PROTECTED: Final[ThirtyTwoBits] = cached_hresult(0x803100C9, 'FVE_E_DE_OS_VOLUME_NOT_PROTECTED')
    FVE_E_DE_DEVICE_LOCKEDOUT: Final[ThirtyTwoBits] = cached_hresult(0x803100CA, 'FVE_E_DE_DEVICE_LOCKEDOUT')
    FVE_E_DE_PROTECTION_NOT_YET_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0x803100CB, 'FVE_E_DE_PROTECTION_NOT_YET_ENABLED')
    FVE_E_INVALID_PIN_CHARS_DETAILED: Final[ThirtyTwoBits] = cached_hresult(0x803100CC, 'FVE_E_INVALID_PIN_CHARS_DETAILED')
    FVE_E_DEVICE_LOCKOUT_COUNTER_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x803100CD, 'FVE_E_DEVICE_LOCKOUT_COUNTER_UNAVAILABLE')
    FVE_E_DEVICELOCKOUT_COUNTER_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x803100CE, 'FVE_E_DEVICELOCKOUT_COUNTER_MISMATCH')
    FVE_E_BUFFER_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0x803100CF, 'FVE_E_BUFFER_TOO_LARGE')
    FVE_E_NO_SUCH_CAPABILITY_ON_TARGET: Final[ThirtyTwoBits] = cached_hresult(0x803100D0, 'FVE_E_NO_SUCH_CAPABILITY_ON_TARGET')
    FVE_E_DE_PREVENTED_FOR_OS: Final[ThirtyTwoBits] = cached_hresult(0x803100D1, 'FVE_E_DE_PREVENTED_FOR_OS')
    FVE_E_DE_VOLUME_OPTED_OUT: Final[ThirtyTwoBits] = cached_hresult(0x803100D2, 'FVE_E_DE_VOLUME_OPTED_OUT')
    FVE_E_DE_VOLUME_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x803100D3, 'FVE_E_DE_VOLUME_NOT_SUPPORTED')
    FVE_E_EOW_NOT_SUPPORTED_IN_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x803100D4, 'FVE_E_EOW_NOT_SUPPORTED_IN_VERSION')
    FVE_E_ADBACKUP_NOT_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0x803100D5, 'FVE_E_ADBACKUP_NOT_ENABLED')
    FVE_E_VOLUME_EXTEND_PREVENTS_EOW_DECRYPT: Final[ThirtyTwoBits] = cached_hresult(0x803100D6, 'FVE_E_VOLUME_EXTEND_PREVENTS_EOW_DECRYPT')
    FVE_E_NOT_DE_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0x803100D7, 'FVE_E_NOT_DE_VOLUME')
    FVE_E_PROTECTION_CANNOT_BE_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x803100D8, 'FVE_E_PROTECTION_CANNOT_BE_DISABLED')
    FVE_E_OSV_KSR_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x803100D9, 'FVE_E_OSV_KSR_NOT_ALLOWED')
    FVE_E_AD_BACKUP_REQUIRED_POLICY_NOT_SET_OS_DRIVE: Final[ThirtyTwoBits] = cached_hresult(0x803100DA, 'FVE_E_AD_BACKUP_REQUIRED_POLICY_NOT_SET_OS_DRIVE')
    FVE_E_AD_BACKUP_REQUIRED_POLICY_NOT_SET_FIXED_DRIVE: Final[ThirtyTwoBits] = cached_hresult(0x803100DB, 'FVE_E_AD_BACKUP_REQUIRED_POLICY_NOT_SET_FIXED_DRIVE')
    FVE_E_AD_BACKUP_REQUIRED_POLICY_NOT_SET_REMOVABLE_DRIVE: Final[ThirtyTwoBits] = cached_hresult(0x803100DC, 'FVE_E_AD_BACKUP_REQUIRED_POLICY_NOT_SET_REMOVABLE_DRIVE')
    FVE_E_KEY_ROTATION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x803100DD, 'FVE_E_KEY_ROTATION_NOT_SUPPORTED')
    FVE_E_EXECUTE_REQUEST_SENT_TOO_SOON: Final[ThirtyTwoBits] = cached_hresult(0x803100DE, 'FVE_E_EXECUTE_REQUEST_SENT_TOO_SOON')
    FVE_E_KEY_ROTATION_NOT_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0x803100DF, 'FVE_E_KEY_ROTATION_NOT_ENABLED')
    FVE_E_DEVICE_NOT_JOINED: Final[ThirtyTwoBits] = cached_hresult(0x803100E0, 'FVE_E_DEVICE_NOT_JOINED')
    FVE_E_AAD_ENDPOINT_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x803100E1, 'FVE_E_AAD_ENDPOINT_BUSY')
    FVE_E_INVALID_NBP_CERT: Final[ThirtyTwoBits] = cached_hresult(0x803100E2, 'FVE_E_INVALID_NBP_CERT')
    FVE_E_EDRIVE_BAND_ENUMERATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x803100E3, 'FVE_E_EDRIVE_BAND_ENUMERATION_FAILED')
    FVE_E_POLICY_ON_RDV_EXCLUSION_LIST: Final[ThirtyTwoBits] = cached_hresult(0x803100E4, 'FVE_E_POLICY_ON_RDV_EXCLUSION_LIST')
    FVE_E_PREDICTED_TPM_PROTECTOR_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x803100E5, 'FVE_E_PREDICTED_TPM_PROTECTOR_NOT_SUPPORTED')
    FVE_E_SETUP_TPM_CALLBACK_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x803100E6, 'FVE_E_SETUP_TPM_CALLBACK_NOT_SUPPORTED')
    FVE_E_TPM_CONTEXT_SETUP_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x803100E7, 'FVE_E_TPM_CONTEXT_SETUP_NOT_SUPPORTED')
    FVE_E_UPDATE_INVALID_CONFIG: Final[ThirtyTwoBits] = cached_hresult(0x803100E8, 'FVE_E_UPDATE_INVALID_CONFIG')
    FVE_E_AAD_SERVER_FAIL_RETRY_AFTER: Final[ThirtyTwoBits] = cached_hresult(0x803100E9, 'FVE_E_AAD_SERVER_FAIL_RETRY_AFTER')
    FVE_E_AAD_SERVER_FAIL_BACKOFF: Final[ThirtyTwoBits] = cached_hresult(0x803100EA, 'FVE_E_AAD_SERVER_FAIL_BACKOFF')
    FVE_E_DATASET_FULL: Final[ThirtyTwoBits] = cached_hresult(0x803100EB, 'FVE_E_DATASET_FULL')
    FVE_E_METADATA_FULL: Final[ThirtyTwoBits] = cached_hresult(0x803100EC, 'FVE_E_METADATA_FULL')
    FWP_E_CALLOUT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80320001, 'FWP_E_CALLOUT_NOT_FOUND')
    FWP_E_CONDITION_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80320002, 'FWP_E_CONDITION_NOT_FOUND')
    FWP_E_FILTER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80320003, 'FWP_E_FILTER_NOT_FOUND')
    FWP_E_LAYER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80320004, 'FWP_E_LAYER_NOT_FOUND')
    FWP_E_PROVIDER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80320005, 'FWP_E_PROVIDER_NOT_FOUND')
    FWP_E_PROVIDER_CONTEXT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80320006, 'FWP_E_PROVIDER_CONTEXT_NOT_FOUND')
    FWP_E_SUBLAYER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80320007, 'FWP_E_SUBLAYER_NOT_FOUND')
    FWP_E_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80320008, 'FWP_E_NOT_FOUND')
    FWP_E_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80320009, 'FWP_E_ALREADY_EXISTS')
    FWP_E_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x8032000A, 'FWP_E_IN_USE')
    FWP_E_DYNAMIC_SESSION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x8032000B, 'FWP_E_DYNAMIC_SESSION_IN_PROGRESS')
    FWP_E_WRONG_SESSION: Final[ThirtyTwoBits] = cached_hresult(0x8032000C, 'FWP_E_WRONG_SESSION')
    FWP_E_NO_TXN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x8032000D, 'FWP_E_NO_TXN_IN_PROGRESS')
    FWP_E_TXN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x8032000E, 'FWP_E_TXN_IN_PROGRESS')
    FWP_E_TXN_ABORTED: Final[ThirtyTwoBits] = cached_hresult(0x8032000F, 'FWP_E_TXN_ABORTED')
    FWP_E_SESSION_ABORTED: Final[ThirtyTwoBits] = cached_hresult(0x80320010, 'FWP_E_SESSION_ABORTED')
    FWP_E_INCOMPATIBLE_TXN: Final[ThirtyTwoBits] = cached_hresult(0x80320011, 'FWP_E_INCOMPATIBLE_TXN')
    FWP_E_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x80320012, 'FWP_E_TIMEOUT')
    FWP_E_NET_EVENTS_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80320013, 'FWP_E_NET_EVENTS_DISABLED')
    FWP_E_INCOMPATIBLE_LAYER: Final[ThirtyTwoBits] = cached_hresult(0x80320014, 'FWP_E_INCOMPATIBLE_LAYER')
    FWP_E_KM_CLIENTS_ONLY: Final[ThirtyTwoBits] = cached_hresult(0x80320015, 'FWP_E_KM_CLIENTS_ONLY')
    FWP_E_LIFETIME_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80320016, 'FWP_E_LIFETIME_MISMATCH')
    FWP_E_BUILTIN_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x80320017, 'FWP_E_BUILTIN_OBJECT')
    FWP_E_TOO_MANY_CALLOUTS: Final[ThirtyTwoBits] = cached_hresult(0x80320018, 'FWP_E_TOO_MANY_CALLOUTS')
    FWP_E_NOTIFICATION_DROPPED: Final[ThirtyTwoBits] = cached_hresult(0x80320019, 'FWP_E_NOTIFICATION_DROPPED')
    FWP_E_TRAFFIC_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x8032001A, 'FWP_E_TRAFFIC_MISMATCH')
    FWP_E_INCOMPATIBLE_SA_STATE: Final[ThirtyTwoBits] = cached_hresult(0x8032001B, 'FWP_E_INCOMPATIBLE_SA_STATE')
    FWP_E_NULL_POINTER: Final[ThirtyTwoBits] = cached_hresult(0x8032001C, 'FWP_E_NULL_POINTER')
    FWP_E_INVALID_ENUMERATOR: Final[ThirtyTwoBits] = cached_hresult(0x8032001D, 'FWP_E_INVALID_ENUMERATOR')
    FWP_E_INVALID_FLAGS: Final[ThirtyTwoBits] = cached_hresult(0x8032001E, 'FWP_E_INVALID_FLAGS')
    FWP_E_INVALID_NET_MASK: Final[ThirtyTwoBits] = cached_hresult(0x8032001F, 'FWP_E_INVALID_NET_MASK')
    FWP_E_INVALID_RANGE: Final[ThirtyTwoBits] = cached_hresult(0x80320020, 'FWP_E_INVALID_RANGE')
    FWP_E_INVALID_INTERVAL: Final[ThirtyTwoBits] = cached_hresult(0x80320021, 'FWP_E_INVALID_INTERVAL')
    FWP_E_ZERO_LENGTH_ARRAY: Final[ThirtyTwoBits] = cached_hresult(0x80320022, 'FWP_E_ZERO_LENGTH_ARRAY')
    FWP_E_NULL_DISPLAY_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80320023, 'FWP_E_NULL_DISPLAY_NAME')
    FWP_E_INVALID_ACTION_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x80320024, 'FWP_E_INVALID_ACTION_TYPE')
    FWP_E_INVALID_WEIGHT: Final[ThirtyTwoBits] = cached_hresult(0x80320025, 'FWP_E_INVALID_WEIGHT')
    FWP_E_MATCH_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80320026, 'FWP_E_MATCH_TYPE_MISMATCH')
    FWP_E_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80320027, 'FWP_E_TYPE_MISMATCH')
    FWP_E_OUT_OF_BOUNDS: Final[ThirtyTwoBits] = cached_hresult(0x80320028, 'FWP_E_OUT_OF_BOUNDS')
    FWP_E_RESERVED: Final[ThirtyTwoBits] = cached_hresult(0x80320029, 'FWP_E_RESERVED')
    FWP_E_DUPLICATE_CONDITION: Final[ThirtyTwoBits] = cached_hresult(0x8032002A, 'FWP_E_DUPLICATE_CONDITION')
    FWP_E_DUPLICATE_KEYMOD: Final[ThirtyTwoBits] = cached_hresult(0x8032002B, 'FWP_E_DUPLICATE_KEYMOD')
    FWP_E_ACTION_INCOMPATIBLE_WITH_LAYER: Final[ThirtyTwoBits] = cached_hresult(0x8032002C, 'FWP_E_ACTION_INCOMPATIBLE_WITH_LAYER')
    FWP_E_ACTION_INCOMPATIBLE_WITH_SUBLAYER: Final[ThirtyTwoBits] = cached_hresult(0x8032002D, 'FWP_E_ACTION_INCOMPATIBLE_WITH_SUBLAYER')
    FWP_E_CONTEXT_INCOMPATIBLE_WITH_LAYER: Final[ThirtyTwoBits] = cached_hresult(0x8032002E, 'FWP_E_CONTEXT_INCOMPATIBLE_WITH_LAYER')
    FWP_E_CONTEXT_INCOMPATIBLE_WITH_CALLOUT: Final[ThirtyTwoBits] = cached_hresult(0x8032002F, 'FWP_E_CONTEXT_INCOMPATIBLE_WITH_CALLOUT')
    FWP_E_INCOMPATIBLE_AUTH_METHOD: Final[ThirtyTwoBits] = cached_hresult(0x80320030, 'FWP_E_INCOMPATIBLE_AUTH_METHOD')
    FWP_E_INCOMPATIBLE_DH_GROUP: Final[ThirtyTwoBits] = cached_hresult(0x80320031, 'FWP_E_INCOMPATIBLE_DH_GROUP')
    FWP_E_EM_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80320032, 'FWP_E_EM_NOT_SUPPORTED')
    FWP_E_NEVER_MATCH: Final[ThirtyTwoBits] = cached_hresult(0x80320033, 'FWP_E_NEVER_MATCH')
    FWP_E_PROVIDER_CONTEXT_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80320034, 'FWP_E_PROVIDER_CONTEXT_MISMATCH')
    FWP_E_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0x80320035, 'FWP_E_INVALID_PARAMETER')
    FWP_E_TOO_MANY_SUBLAYERS: Final[ThirtyTwoBits] = cached_hresult(0x80320036, 'FWP_E_TOO_MANY_SUBLAYERS')
    FWP_E_CALLOUT_NOTIFICATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80320037, 'FWP_E_CALLOUT_NOTIFICATION_FAILED')
    FWP_E_INVALID_AUTH_TRANSFORM: Final[ThirtyTwoBits] = cached_hresult(0x80320038, 'FWP_E_INVALID_AUTH_TRANSFORM')
    FWP_E_INVALID_CIPHER_TRANSFORM: Final[ThirtyTwoBits] = cached_hresult(0x80320039, 'FWP_E_INVALID_CIPHER_TRANSFORM')
    FWP_E_INCOMPATIBLE_CIPHER_TRANSFORM: Final[ThirtyTwoBits] = cached_hresult(0x8032003A, 'FWP_E_INCOMPATIBLE_CIPHER_TRANSFORM')
    FWP_E_INVALID_TRANSFORM_COMBINATION: Final[ThirtyTwoBits] = cached_hresult(0x8032003B, 'FWP_E_INVALID_TRANSFORM_COMBINATION')
    FWP_E_DUPLICATE_AUTH_METHOD: Final[ThirtyTwoBits] = cached_hresult(0x8032003C, 'FWP_E_DUPLICATE_AUTH_METHOD')
    FWP_E_INVALID_TUNNEL_ENDPOINT: Final[ThirtyTwoBits] = cached_hresult(0x8032003D, 'FWP_E_INVALID_TUNNEL_ENDPOINT')
    FWP_E_L2_DRIVER_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0x8032003E, 'FWP_E_L2_DRIVER_NOT_READY')
    FWP_E_KEY_DICTATOR_ALREADY_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x8032003F, 'FWP_E_KEY_DICTATOR_ALREADY_REGISTERED')
    FWP_E_KEY_DICTATION_INVALID_KEYING_MATERIAL: Final[ThirtyTwoBits] = cached_hresult(0x80320040, 'FWP_E_KEY_DICTATION_INVALID_KEYING_MATERIAL')
    FWP_E_CONNECTIONS_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80320041, 'FWP_E_CONNECTIONS_DISABLED')
    FWP_E_INVALID_DNS_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80320042, 'FWP_E_INVALID_DNS_NAME')
    FWP_E_STILL_ON: Final[ThirtyTwoBits] = cached_hresult(0x80320043, 'FWP_E_STILL_ON')
    FWP_E_IKEEXT_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x80320044, 'FWP_E_IKEEXT_NOT_RUNNING')
    FWP_E_DROP_NOICMP: Final[ThirtyTwoBits] = cached_hresult(0x80320104, 'FWP_E_DROP_NOICMP')
    HCS_E_TERMINATED_DURING_START: Final[ThirtyTwoBits] = cached_hresult(0x80370100, 'HCS_E_TERMINATED_DURING_START')
    HCS_E_IMAGE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80370101, 'HCS_E_IMAGE_MISMATCH')
    HCS_E_HYPERV_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x80370102, 'HCS_E_HYPERV_NOT_INSTALLED')
    HCS_E_INVALID_STATE: Final[ThirtyTwoBits] = cached_hresult(0x80370105, 'HCS_E_INVALID_STATE')
    HCS_E_UNEXPECTED_EXIT: Final[ThirtyTwoBits] = cached_hresult(0x80370106, 'HCS_E_UNEXPECTED_EXIT')
    HCS_E_TERMINATED: Final[ThirtyTwoBits] = cached_hresult(0x80370107, 'HCS_E_TERMINATED')
    HCS_E_CONNECT_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80370108, 'HCS_E_CONNECT_FAILED')
    HCS_E_CONNECTION_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x80370109, 'HCS_E_CONNECTION_TIMEOUT')
    HCS_E_CONNECTION_CLOSED: Final[ThirtyTwoBits] = cached_hresult(0x8037010A, 'HCS_E_CONNECTION_CLOSED')
    HCS_E_UNKNOWN_MESSAGE: Final[ThirtyTwoBits] = cached_hresult(0x8037010B, 'HCS_E_UNKNOWN_MESSAGE')
    HCS_E_UNSUPPORTED_PROTOCOL_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x8037010C, 'HCS_E_UNSUPPORTED_PROTOCOL_VERSION')
    HCS_E_INVALID_JSON: Final[ThirtyTwoBits] = cached_hresult(0x8037010D, 'HCS_E_INVALID_JSON')
    HCS_E_SYSTEM_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8037010E, 'HCS_E_SYSTEM_NOT_FOUND')
    HCS_E_SYSTEM_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x8037010F, 'HCS_E_SYSTEM_ALREADY_EXISTS')
    HCS_E_SYSTEM_ALREADY_STOPPED: Final[ThirtyTwoBits] = cached_hresult(0x80370110, 'HCS_E_SYSTEM_ALREADY_STOPPED')
    HCS_E_PROTOCOL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80370111, 'HCS_E_PROTOCOL_ERROR')
    HCS_E_INVALID_LAYER: Final[ThirtyTwoBits] = cached_hresult(0x80370112, 'HCS_E_INVALID_LAYER')
    HCS_E_WINDOWS_INSIDER_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80370113, 'HCS_E_WINDOWS_INSIDER_REQUIRED')
    HCS_E_SERVICE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80370114, 'HCS_E_SERVICE_NOT_AVAILABLE')
    HCS_E_OPERATION_NOT_STARTED: Final[ThirtyTwoBits] = cached_hresult(0x80370115, 'HCS_E_OPERATION_NOT_STARTED')
    HCS_E_OPERATION_ALREADY_STARTED: Final[ThirtyTwoBits] = cached_hresult(0x80370116, 'HCS_E_OPERATION_ALREADY_STARTED')
    HCS_E_OPERATION_PENDING: Final[ThirtyTwoBits] = cached_hresult(0x80370117, 'HCS_E_OPERATION_PENDING')
    HCS_E_OPERATION_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x80370118, 'HCS_E_OPERATION_TIMEOUT')
    HCS_E_OPERATION_SYSTEM_CALLBACK_ALREADY_SET: Final[ThirtyTwoBits] = cached_hresult(0x80370119, 'HCS_E_OPERATION_SYSTEM_CALLBACK_ALREADY_SET')
    HCS_E_OPERATION_RESULT_ALLOCATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8037011A, 'HCS_E_OPERATION_RESULT_ALLOCATION_FAILED')
    HCS_E_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x8037011B, 'HCS_E_ACCESS_DENIED')
    HCS_E_GUEST_CRITICAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8037011C, 'HCS_E_GUEST_CRITICAL_ERROR')
    HCS_E_PROCESS_INFO_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x8037011D, 'HCS_E_PROCESS_INFO_NOT_AVAILABLE')
    HCS_E_SERVICE_DISCONNECT: Final[ThirtyTwoBits] = cached_hresult(0x8037011E, 'HCS_E_SERVICE_DISCONNECT')
    HCS_E_PROCESS_ALREADY_STOPPED: Final[ThirtyTwoBits] = cached_hresult(0x8037011F, 'HCS_E_PROCESS_ALREADY_STOPPED')
    HCS_E_SYSTEM_NOT_CONFIGURED_FOR_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x80370120, 'HCS_E_SYSTEM_NOT_CONFIGURED_FOR_OPERATION')
    HCS_E_OPERATION_ALREADY_CANCELLED: Final[ThirtyTwoBits] = cached_hresult(0x80370121, 'HCS_E_OPERATION_ALREADY_CANCELLED')
    WHV_E_UNKNOWN_CAPABILITY: Final[ThirtyTwoBits] = cached_hresult(0x80370300, 'WHV_E_UNKNOWN_CAPABILITY')
    WHV_E_INSUFFICIENT_BUFFER: Final[ThirtyTwoBits] = cached_hresult(0x80370301, 'WHV_E_INSUFFICIENT_BUFFER')
    WHV_E_UNKNOWN_PROPERTY: Final[ThirtyTwoBits] = cached_hresult(0x80370302, 'WHV_E_UNKNOWN_PROPERTY')
    WHV_E_UNSUPPORTED_HYPERVISOR_CONFIG: Final[ThirtyTwoBits] = cached_hresult(0x80370303, 'WHV_E_UNSUPPORTED_HYPERVISOR_CONFIG')
    WHV_E_INVALID_PARTITION_CONFIG: Final[ThirtyTwoBits] = cached_hresult(0x80370304, 'WHV_E_INVALID_PARTITION_CONFIG')
    WHV_E_GPA_RANGE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80370305, 'WHV_E_GPA_RANGE_NOT_FOUND')
    WHV_E_VP_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80370306, 'WHV_E_VP_ALREADY_EXISTS')
    WHV_E_VP_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x80370307, 'WHV_E_VP_DOES_NOT_EXIST')
    WHV_E_INVALID_VP_STATE: Final[ThirtyTwoBits] = cached_hresult(0x80370308, 'WHV_E_INVALID_VP_STATE')
    WHV_E_INVALID_VP_REGISTER_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80370309, 'WHV_E_INVALID_VP_REGISTER_NAME')
    WHV_E_UNSUPPORTED_PROCESSOR_CONFIG: Final[ThirtyTwoBits] = cached_hresult(0x80370310, 'WHV_E_UNSUPPORTED_PROCESSOR_CONFIG')
    VM_SAVED_STATE_DUMP_E_PARTITION_STATE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xC0370500, 'VM_SAVED_STATE_DUMP_E_PARTITION_STATE_NOT_FOUND')
    VM_SAVED_STATE_DUMP_E_GUEST_MEMORY_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xC0370501, 'VM_SAVED_STATE_DUMP_E_GUEST_MEMORY_NOT_FOUND')
    VM_SAVED_STATE_DUMP_E_NO_VP_FOUND_IN_PARTITION_STATE: Final[ThirtyTwoBits] = cached_hresult(0xC0370502, 'VM_SAVED_STATE_DUMP_E_NO_VP_FOUND_IN_PARTITION_STATE')
    VM_SAVED_STATE_DUMP_E_NESTED_VIRTUALIZATION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xC0370503, 'VM_SAVED_STATE_DUMP_E_NESTED_VIRTUALIZATION_NOT_SUPPORTED')
    VM_SAVED_STATE_DUMP_E_WINDOWS_KERNEL_IMAGE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xC0370504, 'VM_SAVED_STATE_DUMP_E_WINDOWS_KERNEL_IMAGE_NOT_FOUND')
    VM_SAVED_STATE_DUMP_E_VA_NOT_MAPPED: Final[ThirtyTwoBits] = cached_hresult(0xC0370505, 'VM_SAVED_STATE_DUMP_E_VA_NOT_MAPPED')
    VM_SAVED_STATE_DUMP_E_INVALID_VP_STATE: Final[ThirtyTwoBits] = cached_hresult(0xC0370506, 'VM_SAVED_STATE_DUMP_E_INVALID_VP_STATE')
    VM_SAVED_STATE_DUMP_E_VP_VTL_NOT_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0xC0370509, 'VM_SAVED_STATE_DUMP_E_VP_VTL_NOT_ENABLED')
    ERROR_DM_OPERATION_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xC0370600, 'ERROR_DM_OPERATION_LIMIT_EXCEEDED')
    HCN_E_NETWORK_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x803B0001, 'HCN_E_NETWORK_NOT_FOUND')
    HCN_E_ENDPOINT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x803B0002, 'HCN_E_ENDPOINT_NOT_FOUND')
    HCN_E_LAYER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x803B0003, 'HCN_E_LAYER_NOT_FOUND')
    HCN_E_SWITCH_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x803B0004, 'HCN_E_SWITCH_NOT_FOUND')
    HCN_E_SUBNET_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x803B0005, 'HCN_E_SUBNET_NOT_FOUND')
    HCN_E_ADAPTER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x803B0006, 'HCN_E_ADAPTER_NOT_FOUND')
    HCN_E_PORT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x803B0007, 'HCN_E_PORT_NOT_FOUND')
    HCN_E_POLICY_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x803B0008, 'HCN_E_POLICY_NOT_FOUND')
    HCN_E_VFP_PORTSETTING_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x803B0009, 'HCN_E_VFP_PORTSETTING_NOT_FOUND')
    HCN_E_INVALID_NETWORK: Final[ThirtyTwoBits] = cached_hresult(0x803B000A, 'HCN_E_INVALID_NETWORK')
    HCN_E_INVALID_NETWORK_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x803B000B, 'HCN_E_INVALID_NETWORK_TYPE')
    HCN_E_INVALID_ENDPOINT: Final[ThirtyTwoBits] = cached_hresult(0x803B000C, 'HCN_E_INVALID_ENDPOINT')
    HCN_E_INVALID_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x803B000D, 'HCN_E_INVALID_POLICY')
    HCN_E_INVALID_POLICY_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x803B000E, 'HCN_E_INVALID_POLICY_TYPE')
    HCN_E_INVALID_REMOTE_ENDPOINT_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x803B000F, 'HCN_E_INVALID_REMOTE_ENDPOINT_OPERATION')
    HCN_E_NETWORK_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x803B0010, 'HCN_E_NETWORK_ALREADY_EXISTS')
    HCN_E_LAYER_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x803B0011, 'HCN_E_LAYER_ALREADY_EXISTS')
    HCN_E_POLICY_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x803B0012, 'HCN_E_POLICY_ALREADY_EXISTS')
    HCN_E_PORT_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x803B0013, 'HCN_E_PORT_ALREADY_EXISTS')
    HCN_E_ENDPOINT_ALREADY_ATTACHED: Final[ThirtyTwoBits] = cached_hresult(0x803B0014, 'HCN_E_ENDPOINT_ALREADY_ATTACHED')
    HCN_E_REQUEST_UNSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x803B0015, 'HCN_E_REQUEST_UNSUPPORTED')
    HCN_E_MAPPING_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x803B0016, 'HCN_E_MAPPING_NOT_SUPPORTED')
    HCN_E_DEGRADED_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x803B0017, 'HCN_E_DEGRADED_OPERATION')
    HCN_E_SHARED_SWITCH_MODIFICATION: Final[ThirtyTwoBits] = cached_hresult(0x803B0018, 'HCN_E_SHARED_SWITCH_MODIFICATION')
    HCN_E_GUID_CONVERSION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x803B0019, 'HCN_E_GUID_CONVERSION_FAILURE')
    HCN_E_REGKEY_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x803B001A, 'HCN_E_REGKEY_FAILURE')
    HCN_E_INVALID_JSON: Final[ThirtyTwoBits] = cached_hresult(0x803B001B, 'HCN_E_INVALID_JSON')
    HCN_E_INVALID_JSON_REFERENCE: Final[ThirtyTwoBits] = cached_hresult(0x803B001C, 'HCN_E_INVALID_JSON_REFERENCE')
    HCN_E_ENDPOINT_SHARING_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x803B001D, 'HCN_E_ENDPOINT_SHARING_DISABLED')
    HCN_E_INVALID_IP: Final[ThirtyTwoBits] = cached_hresult(0x803B001E, 'HCN_E_INVALID_IP')
    HCN_E_SWITCH_EXTENSION_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x803B001F, 'HCN_E_SWITCH_EXTENSION_NOT_FOUND')
    HCN_E_MANAGER_STOPPED: Final[ThirtyTwoBits] = cached_hresult(0x803B0020, 'HCN_E_MANAGER_STOPPED')
    GCN_E_MODULE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x803B0021, 'GCN_E_MODULE_NOT_FOUND')
    GCN_E_NO_REQUEST_HANDLERS: Final[ThirtyTwoBits] = cached_hresult(0x803B0022, 'GCN_E_NO_REQUEST_HANDLERS')
    GCN_E_REQUEST_UNSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x803B0023, 'GCN_E_REQUEST_UNSUPPORTED')
    GCN_E_RUNTIMEKEYS_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x803B0024, 'GCN_E_RUNTIMEKEYS_FAILED')
    GCN_E_NETADAPTER_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x803B0025, 'GCN_E_NETADAPTER_TIMEOUT')
    GCN_E_NETADAPTER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x803B0026, 'GCN_E_NETADAPTER_NOT_FOUND')
    GCN_E_NETCOMPARTMENT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x803B0027, 'GCN_E_NETCOMPARTMENT_NOT_FOUND')
    GCN_E_NETINTERFACE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x803B0028, 'GCN_E_NETINTERFACE_NOT_FOUND')
    GCN_E_DEFAULTNAMESPACE_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x803B0029, 'GCN_E_DEFAULTNAMESPACE_EXISTS')
    HCN_E_ICS_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x803B002A, 'HCN_E_ICS_DISABLED')
    HCN_E_ENDPOINT_NAMESPACE_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x803B002B, 'HCN_E_ENDPOINT_NAMESPACE_ALREADY_EXISTS')
    HCN_E_ENTITY_HAS_REFERENCES: Final[ThirtyTwoBits] = cached_hresult(0x803B002C, 'HCN_E_ENTITY_HAS_REFERENCES')
    HCN_E_INVALID_INTERNAL_PORT: Final[ThirtyTwoBits] = cached_hresult(0x803B002D, 'HCN_E_INVALID_INTERNAL_PORT')
    HCN_E_NAMESPACE_ATTACH_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x803B002E, 'HCN_E_NAMESPACE_ATTACH_FAILED')
    HCN_E_ADDR_INVALID_OR_RESERVED: Final[ThirtyTwoBits] = cached_hresult(0x803B002F, 'HCN_E_ADDR_INVALID_OR_RESERVED')
    HCN_E_INVALID_PREFIX: Final[ThirtyTwoBits] = cached_hresult(0x803B0030, 'HCN_E_INVALID_PREFIX')
    HCN_E_OBJECT_USED_AFTER_UNLOAD: Final[ThirtyTwoBits] = cached_hresult(0x803B0031, 'HCN_E_OBJECT_USED_AFTER_UNLOAD')
    HCN_E_INVALID_SUBNET: Final[ThirtyTwoBits] = cached_hresult(0x803B0032, 'HCN_E_INVALID_SUBNET')
    HCN_E_INVALID_IP_SUBNET: Final[ThirtyTwoBits] = cached_hresult(0x803B0033, 'HCN_E_INVALID_IP_SUBNET')
    HCN_E_ENDPOINT_NOT_ATTACHED: Final[ThirtyTwoBits] = cached_hresult(0x803B0034, 'HCN_E_ENDPOINT_NOT_ATTACHED')
    HCN_E_ENDPOINT_NOT_LOCAL: Final[ThirtyTwoBits] = cached_hresult(0x803B0035, 'HCN_E_ENDPOINT_NOT_LOCAL')
    HCN_INTERFACEPARAMETERS_ALREADY_APPLIED: Final[ThirtyTwoBits] = cached_hresult(0x803B0036, 'HCN_INTERFACEPARAMETERS_ALREADY_APPLIED')
    HCN_E_VFP_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x803B0037, 'HCN_E_VFP_NOT_ALLOWED')
    WS_S_ASYNC: Final[ThirtyTwoBits] = cached_hresult(0x003D0000, 'WS_S_ASYNC')
    WS_E_INVALID_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0x803D0000, 'WS_E_INVALID_FORMAT')
    WS_S_END: Final[ThirtyTwoBits] = cached_hresult(0x003D0001, 'WS_S_END')
    WS_E_OBJECT_FAULTED: Final[ThirtyTwoBits] = cached_hresult(0x803D0001, 'WS_E_OBJECT_FAULTED')
    WS_E_NUMERIC_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0x803D0002, 'WS_E_NUMERIC_OVERFLOW')
    WS_E_INVALID_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x803D0003, 'WS_E_INVALID_OPERATION')
    WS_E_OPERATION_ABORTED: Final[ThirtyTwoBits] = cached_hresult(0x803D0004, 'WS_E_OPERATION_ABORTED')
    WS_E_ENDPOINT_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x803D0005, 'WS_E_ENDPOINT_ACCESS_DENIED')
    WS_E_OPERATION_TIMED_OUT: Final[ThirtyTwoBits] = cached_hresult(0x803D0006, 'WS_E_OPERATION_TIMED_OUT')
    WS_E_OPERATION_ABANDONED: Final[ThirtyTwoBits] = cached_hresult(0x803D0007, 'WS_E_OPERATION_ABANDONED')
    WS_E_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x803D0008, 'WS_E_QUOTA_EXCEEDED')
    WS_E_NO_TRANSLATION_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x803D0009, 'WS_E_NO_TRANSLATION_AVAILABLE')
    WS_E_SECURITY_VERIFICATION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x803D000A, 'WS_E_SECURITY_VERIFICATION_FAILURE')
    WS_E_ADDRESS_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x803D000B, 'WS_E_ADDRESS_IN_USE')
    WS_E_ADDRESS_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x803D000C, 'WS_E_ADDRESS_NOT_AVAILABLE')
    WS_E_ENDPOINT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x803D000D, 'WS_E_ENDPOINT_NOT_FOUND')
    WS_E_ENDPOINT_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x803D000E, 'WS_E_ENDPOINT_NOT_AVAILABLE')
    WS_E_ENDPOINT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x803D000F, 'WS_E_ENDPOINT_FAILURE')
    WS_E_ENDPOINT_UNREACHABLE: Final[ThirtyTwoBits] = cached_hresult(0x803D0010, 'WS_E_ENDPOINT_UNREACHABLE')
    WS_E_ENDPOINT_ACTION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x803D0011, 'WS_E_ENDPOINT_ACTION_NOT_SUPPORTED')
    WS_E_ENDPOINT_TOO_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x803D0012, 'WS_E_ENDPOINT_TOO_BUSY')
    WS_E_ENDPOINT_FAULT_RECEIVED: Final[ThirtyTwoBits] = cached_hresult(0x803D0013, 'WS_E_ENDPOINT_FAULT_RECEIVED')
    WS_E_ENDPOINT_DISCONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x803D0014, 'WS_E_ENDPOINT_DISCONNECTED')
    WS_E_PROXY_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x803D0015, 'WS_E_PROXY_FAILURE')
    WS_E_PROXY_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x803D0016, 'WS_E_PROXY_ACCESS_DENIED')
    WS_E_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x803D0017, 'WS_E_NOT_SUPPORTED')
    WS_E_PROXY_REQUIRES_BASIC_AUTH: Final[ThirtyTwoBits] = cached_hresult(0x803D0018, 'WS_E_PROXY_REQUIRES_BASIC_AUTH')
    WS_E_PROXY_REQUIRES_DIGEST_AUTH: Final[ThirtyTwoBits] = cached_hresult(0x803D0019, 'WS_E_PROXY_REQUIRES_DIGEST_AUTH')
    WS_E_PROXY_REQUIRES_NTLM_AUTH: Final[ThirtyTwoBits] = cached_hresult(0x803D001A, 'WS_E_PROXY_REQUIRES_NTLM_AUTH')
    WS_E_PROXY_REQUIRES_NEGOTIATE_AUTH: Final[ThirtyTwoBits] = cached_hresult(0x803D001B, 'WS_E_PROXY_REQUIRES_NEGOTIATE_AUTH')
    WS_E_SERVER_REQUIRES_BASIC_AUTH: Final[ThirtyTwoBits] = cached_hresult(0x803D001C, 'WS_E_SERVER_REQUIRES_BASIC_AUTH')
    WS_E_SERVER_REQUIRES_DIGEST_AUTH: Final[ThirtyTwoBits] = cached_hresult(0x803D001D, 'WS_E_SERVER_REQUIRES_DIGEST_AUTH')
    WS_E_SERVER_REQUIRES_NTLM_AUTH: Final[ThirtyTwoBits] = cached_hresult(0x803D001E, 'WS_E_SERVER_REQUIRES_NTLM_AUTH')
    WS_E_SERVER_REQUIRES_NEGOTIATE_AUTH: Final[ThirtyTwoBits] = cached_hresult(0x803D001F, 'WS_E_SERVER_REQUIRES_NEGOTIATE_AUTH')
    WS_E_INVALID_ENDPOINT_URL: Final[ThirtyTwoBits] = cached_hresult(0x803D0020, 'WS_E_INVALID_ENDPOINT_URL')
    WS_E_OTHER: Final[ThirtyTwoBits] = cached_hresult(0x803D0021, 'WS_E_OTHER')
    WS_E_SECURITY_TOKEN_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x803D0022, 'WS_E_SECURITY_TOKEN_EXPIRED')
    WS_E_SECURITY_SYSTEM_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x803D0023, 'WS_E_SECURITY_SYSTEM_FAILURE')
    WPN_E_CHANNEL_CLOSED: Final[ThirtyTwoBits] = cached_hresult(0x803E0100, 'WPN_E_CHANNEL_CLOSED')
    WPN_E_CHANNEL_REQUEST_NOT_COMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x803E0101, 'WPN_E_CHANNEL_REQUEST_NOT_COMPLETE')
    WPN_E_INVALID_APP: Final[ThirtyTwoBits] = cached_hresult(0x803E0102, 'WPN_E_INVALID_APP')
    WPN_E_OUTSTANDING_CHANNEL_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0x803E0103, 'WPN_E_OUTSTANDING_CHANNEL_REQUEST')
    WPN_E_DUPLICATE_CHANNEL: Final[ThirtyTwoBits] = cached_hresult(0x803E0104, 'WPN_E_DUPLICATE_CHANNEL')
    WPN_E_PLATFORM_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x803E0105, 'WPN_E_PLATFORM_UNAVAILABLE')
    WPN_E_NOTIFICATION_POSTED: Final[ThirtyTwoBits] = cached_hresult(0x803E0106, 'WPN_E_NOTIFICATION_POSTED')
    WPN_E_NOTIFICATION_HIDDEN: Final[ThirtyTwoBits] = cached_hresult(0x803E0107, 'WPN_E_NOTIFICATION_HIDDEN')
    WPN_E_NOTIFICATION_NOT_POSTED: Final[ThirtyTwoBits] = cached_hresult(0x803E0108, 'WPN_E_NOTIFICATION_NOT_POSTED')
    WPN_E_CLOUD_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x803E0109, 'WPN_E_CLOUD_DISABLED')
    WPN_E_CLOUD_INCAPABLE: Final[ThirtyTwoBits] = cached_hresult(0x803E0110, 'WPN_E_CLOUD_INCAPABLE')
    WPN_E_NOTIFICATION_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x803E0111, 'WPN_E_NOTIFICATION_DISABLED')
    WPN_E_NOTIFICATION_INCAPABLE: Final[ThirtyTwoBits] = cached_hresult(0x803E0112, 'WPN_E_NOTIFICATION_INCAPABLE')
    WPN_E_INTERNET_INCAPABLE: Final[ThirtyTwoBits] = cached_hresult(0x803E0113, 'WPN_E_INTERNET_INCAPABLE')
    WPN_E_NOTIFICATION_TYPE_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x803E0114, 'WPN_E_NOTIFICATION_TYPE_DISABLED')
    WPN_E_NOTIFICATION_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x803E0115, 'WPN_E_NOTIFICATION_SIZE')
    WPN_E_TAG_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x803E0116, 'WPN_E_TAG_SIZE')
    WPN_E_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x803E0117, 'WPN_E_ACCESS_DENIED')
    WPN_E_DUPLICATE_REGISTRATION: Final[ThirtyTwoBits] = cached_hresult(0x803E0118, 'WPN_E_DUPLICATE_REGISTRATION')
    WPN_E_PUSH_NOTIFICATION_INCAPABLE: Final[ThirtyTwoBits] = cached_hresult(0x803E0119, 'WPN_E_PUSH_NOTIFICATION_INCAPABLE')
    WPN_E_CLOUD_AUTH_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x803E011A, 'WPN_E_CLOUD_AUTH_UNAVAILABLE')
    WPN_E_CLOUD_SERVICE_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x803E011B, 'WPN_E_CLOUD_SERVICE_UNAVAILABLE')
    WPN_E_FAILED_LOCK_SCREEN_UPDATE_INTIALIZATION: Final[ThirtyTwoBits] = cached_hresult(0x803E011C, 'WPN_E_FAILED_LOCK_SCREEN_UPDATE_INTIALIZATION')
    WPN_E_DEV_ID_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x803E0120, 'WPN_E_DEV_ID_SIZE')
    WPN_E_TAG_ALPHANUMERIC: Final[ThirtyTwoBits] = cached_hresult(0x803E012A, 'WPN_E_TAG_ALPHANUMERIC')
    WPN_E_INVALID_HTTP_STATUS_CODE: Final[ThirtyTwoBits] = cached_hresult(0x803E012B, 'WPN_E_INVALID_HTTP_STATUS_CODE')
    WPN_E_OUT_OF_SESSION: Final[ThirtyTwoBits] = cached_hresult(0x803E0200, 'WPN_E_OUT_OF_SESSION')
    WPN_E_POWER_SAVE: Final[ThirtyTwoBits] = cached_hresult(0x803E0201, 'WPN_E_POWER_SAVE')
    WPN_E_IMAGE_NOT_FOUND_IN_CACHE: Final[ThirtyTwoBits] = cached_hresult(0x803E0202, 'WPN_E_IMAGE_NOT_FOUND_IN_CACHE')
    WPN_E_ALL_URL_NOT_COMPLETED: Final[ThirtyTwoBits] = cached_hresult(0x803E0203, 'WPN_E_ALL_URL_NOT_COMPLETED')
    WPN_E_INVALID_CLOUD_IMAGE: Final[ThirtyTwoBits] = cached_hresult(0x803E0204, 'WPN_E_INVALID_CLOUD_IMAGE')
    WPN_E_NOTIFICATION_ID_MATCHED: Final[ThirtyTwoBits] = cached_hresult(0x803E0205, 'WPN_E_NOTIFICATION_ID_MATCHED')
    WPN_E_CALLBACK_ALREADY_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x803E0206, 'WPN_E_CALLBACK_ALREADY_REGISTERED')
    WPN_E_TOAST_NOTIFICATION_DROPPED: Final[ThirtyTwoBits] = cached_hresult(0x803E0207, 'WPN_E_TOAST_NOTIFICATION_DROPPED')
    WPN_E_STORAGE_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x803E0208, 'WPN_E_STORAGE_LOCKED')
    WPN_E_GROUP_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x803E0209, 'WPN_E_GROUP_SIZE')
    WPN_E_GROUP_ALPHANUMERIC: Final[ThirtyTwoBits] = cached_hresult(0x803E020A, 'WPN_E_GROUP_ALPHANUMERIC')
    WPN_E_CLOUD_DISABLED_FOR_APP: Final[ThirtyTwoBits] = cached_hresult(0x803E020B, 'WPN_E_CLOUD_DISABLED_FOR_APP')
    INPUT_E_OUT_OF_ORDER: Final[ThirtyTwoBits] = cached_hresult(0x80400000, 'INPUT_E_OUT_OF_ORDER')
    INPUT_E_REENTRANCY: Final[ThirtyTwoBits] = cached_hresult(0x80400001, 'INPUT_E_REENTRANCY')
    INPUT_E_MULTIMODAL: Final[ThirtyTwoBits] = cached_hresult(0x80400002, 'INPUT_E_MULTIMODAL')
    INPUT_E_PACKET: Final[ThirtyTwoBits] = cached_hresult(0x80400003, 'INPUT_E_PACKET')
    INPUT_E_FRAME: Final[ThirtyTwoBits] = cached_hresult(0x80400004, 'INPUT_E_FRAME')
    INPUT_E_HISTORY: Final[ThirtyTwoBits] = cached_hresult(0x80400005, 'INPUT_E_HISTORY')
    INPUT_E_DEVICE_INFO: Final[ThirtyTwoBits] = cached_hresult(0x80400006, 'INPUT_E_DEVICE_INFO')
    INPUT_E_TRANSFORM: Final[ThirtyTwoBits] = cached_hresult(0x80400007, 'INPUT_E_TRANSFORM')
    INPUT_E_DEVICE_PROPERTY: Final[ThirtyTwoBits] = cached_hresult(0x80400008, 'INPUT_E_DEVICE_PROPERTY')
    ERROR_QUIC_HANDSHAKE_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80410000, 'ERROR_QUIC_HANDSHAKE_FAILURE')
    ERROR_QUIC_VER_NEG_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80410001, 'ERROR_QUIC_VER_NEG_FAILURE')
    ERROR_QUIC_USER_CANCELED: Final[ThirtyTwoBits] = cached_hresult(0x80410002, 'ERROR_QUIC_USER_CANCELED')
    ERROR_QUIC_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80410003, 'ERROR_QUIC_INTERNAL_ERROR')
    ERROR_QUIC_PROTOCOL_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x80410004, 'ERROR_QUIC_PROTOCOL_VIOLATION')
    ERROR_QUIC_CONNECTION_IDLE: Final[ThirtyTwoBits] = cached_hresult(0x80410005, 'ERROR_QUIC_CONNECTION_IDLE')
    ERROR_QUIC_CONNECTION_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x80410006, 'ERROR_QUIC_CONNECTION_TIMEOUT')
    ERROR_QUIC_ALPN_NEG_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80410007, 'ERROR_QUIC_ALPN_NEG_FAILURE')
    IORING_E_REQUIRED_FLAG_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80460001, 'IORING_E_REQUIRED_FLAG_NOT_SUPPORTED')
    IORING_E_SUBMISSION_QUEUE_FULL: Final[ThirtyTwoBits] = cached_hresult(0x80460002, 'IORING_E_SUBMISSION_QUEUE_FULL')
    IORING_E_VERSION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80460003, 'IORING_E_VERSION_NOT_SUPPORTED')
    IORING_E_SUBMISSION_QUEUE_TOO_BIG: Final[ThirtyTwoBits] = cached_hresult(0x80460004, 'IORING_E_SUBMISSION_QUEUE_TOO_BIG')
    IORING_E_COMPLETION_QUEUE_TOO_BIG: Final[ThirtyTwoBits] = cached_hresult(0x80460005, 'IORING_E_COMPLETION_QUEUE_TOO_BIG')
    IORING_E_SUBMIT_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x80460006, 'IORING_E_SUBMIT_IN_PROGRESS')
    IORING_E_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x80460007, 'IORING_E_CORRUPT')
    IORING_E_COMPLETION_QUEUE_TOO_FULL: Final[ThirtyTwoBits] = cached_hresult(0x80460008, 'IORING_E_COMPLETION_QUEUE_TOO_FULL')
    E_MBN_CONTEXT_NOT_ACTIVATED: Final[ThirtyTwoBits] = cached_hresult(0x80548201, 'E_MBN_CONTEXT_NOT_ACTIVATED')
    E_MBN_BAD_SIM: Final[ThirtyTwoBits] = cached_hresult(0x80548202, 'E_MBN_BAD_SIM')
    E_MBN_DATA_CLASS_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80548203, 'E_MBN_DATA_CLASS_NOT_AVAILABLE')
    E_MBN_INVALID_ACCESS_STRING: Final[ThirtyTwoBits] = cached_hresult(0x80548204, 'E_MBN_INVALID_ACCESS_STRING')
    E_MBN_MAX_ACTIVATED_CONTEXTS: Final[ThirtyTwoBits] = cached_hresult(0x80548205, 'E_MBN_MAX_ACTIVATED_CONTEXTS')
    E_MBN_PACKET_SVC_DETACHED: Final[ThirtyTwoBits] = cached_hresult(0x80548206, 'E_MBN_PACKET_SVC_DETACHED')
    E_MBN_PROVIDER_NOT_VISIBLE: Final[ThirtyTwoBits] = cached_hresult(0x80548207, 'E_MBN_PROVIDER_NOT_VISIBLE')
    E_MBN_RADIO_POWER_OFF: Final[ThirtyTwoBits] = cached_hresult(0x80548208, 'E_MBN_RADIO_POWER_OFF')
    E_MBN_SERVICE_NOT_ACTIVATED: Final[ThirtyTwoBits] = cached_hresult(0x80548209, 'E_MBN_SERVICE_NOT_ACTIVATED')
    E_MBN_SIM_NOT_INSERTED: Final[ThirtyTwoBits] = cached_hresult(0x8054820A, 'E_MBN_SIM_NOT_INSERTED')
    E_MBN_VOICE_CALL_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x8054820B, 'E_MBN_VOICE_CALL_IN_PROGRESS')
    E_MBN_INVALID_CACHE: Final[ThirtyTwoBits] = cached_hresult(0x8054820C, 'E_MBN_INVALID_CACHE')
    E_MBN_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x8054820D, 'E_MBN_NOT_REGISTERED')
    E_MBN_PROVIDERS_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8054820E, 'E_MBN_PROVIDERS_NOT_FOUND')
    E_MBN_PIN_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8054820F, 'E_MBN_PIN_NOT_SUPPORTED')
    E_MBN_PIN_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80548210, 'E_MBN_PIN_REQUIRED')
    E_MBN_PIN_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80548211, 'E_MBN_PIN_DISABLED')
    E_MBN_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80548212, 'E_MBN_FAILURE')
    E_MBN_INVALID_PROFILE: Final[ThirtyTwoBits] = cached_hresult(0x80548218, 'E_MBN_INVALID_PROFILE')
    E_MBN_DEFAULT_PROFILE_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x80548219, 'E_MBN_DEFAULT_PROFILE_EXIST')
    E_MBN_SMS_ENCODING_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80548220, 'E_MBN_SMS_ENCODING_NOT_SUPPORTED')
    E_MBN_SMS_FILTER_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80548221, 'E_MBN_SMS_FILTER_NOT_SUPPORTED')
    E_MBN_SMS_INVALID_MEMORY_INDEX: Final[ThirtyTwoBits] = cached_hresult(0x80548222, 'E_MBN_SMS_INVALID_MEMORY_INDEX')
    E_MBN_SMS_LANG_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80548223, 'E_MBN_SMS_LANG_NOT_SUPPORTED')
    E_MBN_SMS_MEMORY_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80548224, 'E_MBN_SMS_MEMORY_FAILURE')
    E_MBN_SMS_NETWORK_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x80548225, 'E_MBN_SMS_NETWORK_TIMEOUT')
    E_MBN_SMS_UNKNOWN_SMSC_ADDRESS: Final[ThirtyTwoBits] = cached_hresult(0x80548226, 'E_MBN_SMS_UNKNOWN_SMSC_ADDRESS')
    E_MBN_SMS_FORMAT_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80548227, 'E_MBN_SMS_FORMAT_NOT_SUPPORTED')
    E_MBN_SMS_OPERATION_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80548228, 'E_MBN_SMS_OPERATION_NOT_ALLOWED')
    E_MBN_SMS_MEMORY_FULL: Final[ThirtyTwoBits] = cached_hresult(0x80548229, 'E_MBN_SMS_MEMORY_FULL')
    EAS_E_POLICY_NOT_MANAGED_BY_OS: Final[ThirtyTwoBits] = cached_hresult(0x80550001, 'EAS_E_POLICY_NOT_MANAGED_BY_OS')
    EAS_E_POLICY_COMPLIANT_WITH_ACTIONS: Final[ThirtyTwoBits] = cached_hresult(0x80550002, 'EAS_E_POLICY_COMPLIANT_WITH_ACTIONS')
    EAS_E_REQUESTED_POLICY_NOT_ENFORCEABLE: Final[ThirtyTwoBits] = cached_hresult(0x80550003, 'EAS_E_REQUESTED_POLICY_NOT_ENFORCEABLE')
    EAS_E_CURRENT_USER_HAS_BLANK_PASSWORD: Final[ThirtyTwoBits] = cached_hresult(0x80550004, 'EAS_E_CURRENT_USER_HAS_BLANK_PASSWORD')
    EAS_E_REQUESTED_POLICY_PASSWORD_EXPIRATION_INCOMPATIBLE: Final[ThirtyTwoBits] = cached_hresult(0x80550005, 'EAS_E_REQUESTED_POLICY_PASSWORD_EXPIRATION_INCOMPATIBLE')
    EAS_E_USER_CANNOT_CHANGE_PASSWORD: Final[ThirtyTwoBits] = cached_hresult(0x80550006, 'EAS_E_USER_CANNOT_CHANGE_PASSWORD')
    EAS_E_ADMINS_HAVE_BLANK_PASSWORD: Final[ThirtyTwoBits] = cached_hresult(0x80550007, 'EAS_E_ADMINS_HAVE_BLANK_PASSWORD')
    EAS_E_ADMINS_CANNOT_CHANGE_PASSWORD: Final[ThirtyTwoBits] = cached_hresult(0x80550008, 'EAS_E_ADMINS_CANNOT_CHANGE_PASSWORD')
    EAS_E_LOCAL_CONTROLLED_USERS_CANNOT_CHANGE_PASSWORD: Final[ThirtyTwoBits] = cached_hresult(0x80550009, 'EAS_E_LOCAL_CONTROLLED_USERS_CANNOT_CHANGE_PASSWORD')
    EAS_E_PASSWORD_POLICY_NOT_ENFORCEABLE_FOR_CONNECTED_ADMINS: Final[ThirtyTwoBits] = cached_hresult(0x8055000A, 'EAS_E_PASSWORD_POLICY_NOT_ENFORCEABLE_FOR_CONNECTED_ADMINS')
    EAS_E_CONNECTED_ADMINS_NEED_TO_CHANGE_PASSWORD: Final[ThirtyTwoBits] = cached_hresult(0x8055000B, 'EAS_E_CONNECTED_ADMINS_NEED_TO_CHANGE_PASSWORD')
    EAS_E_PASSWORD_POLICY_NOT_ENFORCEABLE_FOR_CURRENT_CONNECTED_USER: Final[ThirtyTwoBits] = cached_hresult(0x8055000C, 'EAS_E_PASSWORD_POLICY_NOT_ENFORCEABLE_FOR_CURRENT_CONNECTED_USER')
    EAS_E_CURRENT_CONNECTED_USER_NEED_TO_CHANGE_PASSWORD: Final[ThirtyTwoBits] = cached_hresult(0x8055000D, 'EAS_E_CURRENT_CONNECTED_USER_NEED_TO_CHANGE_PASSWORD')
    ERROR_SVHDX_ERROR_STORED: Final[ThirtyTwoBits] = cached_hresult(0xC05C0000, 'ERROR_SVHDX_ERROR_STORED')
    ERROR_SVHDX_ERROR_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xC05CFF00, 'ERROR_SVHDX_ERROR_NOT_AVAILABLE')
    ERROR_SVHDX_UNIT_ATTENTION_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xC05CFF01, 'ERROR_SVHDX_UNIT_ATTENTION_AVAILABLE')
    ERROR_SVHDX_UNIT_ATTENTION_CAPACITY_DATA_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0xC05CFF02, 'ERROR_SVHDX_UNIT_ATTENTION_CAPACITY_DATA_CHANGED')
    ERROR_SVHDX_UNIT_ATTENTION_RESERVATIONS_PREEMPTED: Final[ThirtyTwoBits] = cached_hresult(0xC05CFF03, 'ERROR_SVHDX_UNIT_ATTENTION_RESERVATIONS_PREEMPTED')
    ERROR_SVHDX_UNIT_ATTENTION_RESERVATIONS_RELEASED: Final[ThirtyTwoBits] = cached_hresult(0xC05CFF04, 'ERROR_SVHDX_UNIT_ATTENTION_RESERVATIONS_RELEASED')
    ERROR_SVHDX_UNIT_ATTENTION_REGISTRATIONS_PREEMPTED: Final[ThirtyTwoBits] = cached_hresult(0xC05CFF05, 'ERROR_SVHDX_UNIT_ATTENTION_REGISTRATIONS_PREEMPTED')
    ERROR_SVHDX_UNIT_ATTENTION_OPERATING_DEFINITION_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0xC05CFF06, 'ERROR_SVHDX_UNIT_ATTENTION_OPERATING_DEFINITION_CHANGED')
    ERROR_SVHDX_RESERVATION_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0xC05CFF07, 'ERROR_SVHDX_RESERVATION_CONFLICT')
    ERROR_SVHDX_WRONG_FILE_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xC05CFF08, 'ERROR_SVHDX_WRONG_FILE_TYPE')
    ERROR_SVHDX_VERSION_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xC05CFF09, 'ERROR_SVHDX_VERSION_MISMATCH')
    ERROR_VHD_SHARED: Final[ThirtyTwoBits] = cached_hresult(0xC05CFF0A, 'ERROR_VHD_SHARED')
    ERROR_SVHDX_NO_INITIATOR: Final[ThirtyTwoBits] = cached_hresult(0xC05CFF0B, 'ERROR_SVHDX_NO_INITIATOR')
    ERROR_VHDSET_BACKING_STORAGE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xC05CFF0C, 'ERROR_VHDSET_BACKING_STORAGE_NOT_FOUND')
    ERROR_SMB_NO_PREAUTH_INTEGRITY_HASH_OVERLAP: Final[ThirtyTwoBits] = cached_hresult(0xC05D0000, 'ERROR_SMB_NO_PREAUTH_INTEGRITY_HASH_OVERLAP')
    ERROR_SMB_BAD_CLUSTER_DIALECT: Final[ThirtyTwoBits] = cached_hresult(0xC05D0001, 'ERROR_SMB_BAD_CLUSTER_DIALECT')
    ERROR_SMB_NO_SIGNING_ALGORITHM_OVERLAP: Final[ThirtyTwoBits] = cached_hresult(0xC05D0002, 'ERROR_SMB_NO_SIGNING_ALGORITHM_OVERLAP')
    PEER_S_GRAPH_DATA_CREATED: Final[ThirtyTwoBits] = cached_hresult(0x00630001, 'PEER_S_GRAPH_DATA_CREATED')
    PEER_E_IPV6_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x80630001, 'PEER_E_IPV6_NOT_INSTALLED')
    PEER_S_NO_EVENT_DATA: Final[ThirtyTwoBits] = cached_hresult(0x00630002, 'PEER_S_NO_EVENT_DATA')
    PEER_E_NOT_INITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0x80630002, 'PEER_E_NOT_INITIALIZED')
    PEER_E_CANNOT_START_SERVICE: Final[ThirtyTwoBits] = cached_hresult(0x80630003, 'PEER_E_CANNOT_START_SERVICE')
    PEER_E_NOT_LICENSED: Final[ThirtyTwoBits] = cached_hresult(0x80630004, 'PEER_E_NOT_LICENSED')
    PEER_S_NO_CONNECTIVITY: Final[ThirtyTwoBits] = cached_hresult(0x00630005, 'PEER_S_NO_CONNECTIVITY')
    PEER_S_ALREADY_A_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0x00630006, 'PEER_S_ALREADY_A_MEMBER')
    PEER_E_INVALID_GRAPH: Final[ThirtyTwoBits] = cached_hresult(0x80630010, 'PEER_E_INVALID_GRAPH')
    PEER_E_DBNAME_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0x80630011, 'PEER_E_DBNAME_CHANGED')
    PEER_E_DUPLICATE_GRAPH: Final[ThirtyTwoBits] = cached_hresult(0x80630012, 'PEER_E_DUPLICATE_GRAPH')
    PEER_E_GRAPH_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0x80630013, 'PEER_E_GRAPH_NOT_READY')
    PEER_E_GRAPH_SHUTTING_DOWN: Final[ThirtyTwoBits] = cached_hresult(0x80630014, 'PEER_E_GRAPH_SHUTTING_DOWN')
    PEER_E_GRAPH_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x80630015, 'PEER_E_GRAPH_IN_USE')
    PEER_E_INVALID_DATABASE: Final[ThirtyTwoBits] = cached_hresult(0x80630016, 'PEER_E_INVALID_DATABASE')
    PEER_E_TOO_MANY_ATTRIBUTES: Final[ThirtyTwoBits] = cached_hresult(0x80630017, 'PEER_E_TOO_MANY_ATTRIBUTES')
    PEER_E_CONNECTION_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80630103, 'PEER_E_CONNECTION_NOT_FOUND')
    PEER_E_CONNECT_SELF: Final[ThirtyTwoBits] = cached_hresult(0x80630106, 'PEER_E_CONNECT_SELF')
    PEER_E_ALREADY_LISTENING: Final[ThirtyTwoBits] = cached_hresult(0x80630107, 'PEER_E_ALREADY_LISTENING')
    PEER_E_NODE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80630108, 'PEER_E_NODE_NOT_FOUND')
    PEER_E_CONNECTION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80630109, 'PEER_E_CONNECTION_FAILED')
    PEER_E_CONNECTION_NOT_AUTHENTICATED: Final[ThirtyTwoBits] = cached_hresult(0x8063010A, 'PEER_E_CONNECTION_NOT_AUTHENTICATED')
    PEER_E_CONNECTION_REFUSED: Final[ThirtyTwoBits] = cached_hresult(0x8063010B, 'PEER_E_CONNECTION_REFUSED')
    PEER_E_CLASSIFIER_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0x80630201, 'PEER_E_CLASSIFIER_TOO_LONG')
    PEER_E_TOO_MANY_IDENTITIES: Final[ThirtyTwoBits] = cached_hresult(0x80630202, 'PEER_E_TOO_MANY_IDENTITIES')
    PEER_E_NO_KEY_ACCESS: Final[ThirtyTwoBits] = cached_hresult(0x80630203, 'PEER_E_NO_KEY_ACCESS')
    PEER_E_GROUPS_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x80630204, 'PEER_E_GROUPS_EXIST')
    PEER_E_RECORD_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80630301, 'PEER_E_RECORD_NOT_FOUND')
    PEER_E_DATABASE_ACCESSDENIED: Final[ThirtyTwoBits] = cached_hresult(0x80630302, 'PEER_E_DATABASE_ACCESSDENIED')
    PEER_E_DBINITIALIZATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80630303, 'PEER_E_DBINITIALIZATION_FAILED')
    PEER_E_MAX_RECORD_SIZE_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80630304, 'PEER_E_MAX_RECORD_SIZE_EXCEEDED')
    PEER_E_DATABASE_ALREADY_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0x80630305, 'PEER_E_DATABASE_ALREADY_PRESENT')
    PEER_E_DATABASE_NOT_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0x80630306, 'PEER_E_DATABASE_NOT_PRESENT')
    PEER_E_IDENTITY_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80630401, 'PEER_E_IDENTITY_NOT_FOUND')
    PEER_E_EVENT_HANDLE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80630501, 'PEER_E_EVENT_HANDLE_NOT_FOUND')
    PEER_E_INVALID_SEARCH: Final[ThirtyTwoBits] = cached_hresult(0x80630601, 'PEER_E_INVALID_SEARCH')
    PEER_E_INVALID_ATTRIBUTES: Final[ThirtyTwoBits] = cached_hresult(0x80630602, 'PEER_E_INVALID_ATTRIBUTES')
    PEER_E_INVITATION_NOT_TRUSTED: Final[ThirtyTwoBits] = cached_hresult(0x80630701, 'PEER_E_INVITATION_NOT_TRUSTED')
    PEER_E_CHAIN_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0x80630703, 'PEER_E_CHAIN_TOO_LONG')
    PEER_E_INVALID_TIME_PERIOD: Final[ThirtyTwoBits] = cached_hresult(0x80630705, 'PEER_E_INVALID_TIME_PERIOD')
    PEER_E_CIRCULAR_CHAIN_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0x80630706, 'PEER_E_CIRCULAR_CHAIN_DETECTED')
    PEER_E_CERT_STORE_CORRUPTED: Final[ThirtyTwoBits] = cached_hresult(0x80630801, 'PEER_E_CERT_STORE_CORRUPTED')
    PEER_E_NO_CLOUD: Final[ThirtyTwoBits] = cached_hresult(0x80631001, 'PEER_E_NO_CLOUD')
    PEER_E_CLOUD_NAME_AMBIGUOUS: Final[ThirtyTwoBits] = cached_hresult(0x80631005, 'PEER_E_CLOUD_NAME_AMBIGUOUS')
    PEER_S_ALREADY_CONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x00632000, 'PEER_S_ALREADY_CONNECTED')
    PEER_E_INVALID_RECORD: Final[ThirtyTwoBits] = cached_hresult(0x80632010, 'PEER_E_INVALID_RECORD')
    PEER_E_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_hresult(0x80632020, 'PEER_E_NOT_AUTHORIZED')
    PEER_E_PASSWORD_DOES_NOT_MEET_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x80632021, 'PEER_E_PASSWORD_DOES_NOT_MEET_POLICY')
    PEER_E_DEFERRED_VALIDATION: Final[ThirtyTwoBits] = cached_hresult(0x80632030, 'PEER_E_DEFERRED_VALIDATION')
    PEER_E_INVALID_GROUP_PROPERTIES: Final[ThirtyTwoBits] = cached_hresult(0x80632040, 'PEER_E_INVALID_GROUP_PROPERTIES')
    PEER_E_INVALID_PEER_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80632050, 'PEER_E_INVALID_PEER_NAME')
    PEER_E_INVALID_CLASSIFIER: Final[ThirtyTwoBits] = cached_hresult(0x80632060, 'PEER_E_INVALID_CLASSIFIER')
    PEER_E_INVALID_FRIENDLY_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80632070, 'PEER_E_INVALID_FRIENDLY_NAME')
    PEER_E_INVALID_ROLE_PROPERTY: Final[ThirtyTwoBits] = cached_hresult(0x80632071, 'PEER_E_INVALID_ROLE_PROPERTY')
    PEER_E_INVALID_CLASSIFIER_PROPERTY: Final[ThirtyTwoBits] = cached_hresult(0x80632072, 'PEER_E_INVALID_CLASSIFIER_PROPERTY')
    PEER_E_INVALID_RECORD_EXPIRATION: Final[ThirtyTwoBits] = cached_hresult(0x80632080, 'PEER_E_INVALID_RECORD_EXPIRATION')
    PEER_E_INVALID_CREDENTIAL_INFO: Final[ThirtyTwoBits] = cached_hresult(0x80632081, 'PEER_E_INVALID_CREDENTIAL_INFO')
    PEER_E_INVALID_CREDENTIAL: Final[ThirtyTwoBits] = cached_hresult(0x80632082, 'PEER_E_INVALID_CREDENTIAL')
    PEER_E_INVALID_RECORD_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x80632083, 'PEER_E_INVALID_RECORD_SIZE')
    PEER_E_UNSUPPORTED_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80632090, 'PEER_E_UNSUPPORTED_VERSION')
    PEER_E_GROUP_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0x80632091, 'PEER_E_GROUP_NOT_READY')
    PEER_E_GROUP_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x80632092, 'PEER_E_GROUP_IN_USE')
    PEER_E_INVALID_GROUP: Final[ThirtyTwoBits] = cached_hresult(0x80632093, 'PEER_E_INVALID_GROUP')
    PEER_E_NO_MEMBERS_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80632094, 'PEER_E_NO_MEMBERS_FOUND')
    PEER_E_NO_MEMBER_CONNECTIONS: Final[ThirtyTwoBits] = cached_hresult(0x80632095, 'PEER_E_NO_MEMBER_CONNECTIONS')
    PEER_E_UNABLE_TO_LISTEN: Final[ThirtyTwoBits] = cached_hresult(0x80632096, 'PEER_E_UNABLE_TO_LISTEN')
    PEER_E_IDENTITY_DELETED: Final[ThirtyTwoBits] = cached_hresult(0x806320A0, 'PEER_E_IDENTITY_DELETED')
    PEER_E_SERVICE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x806320A1, 'PEER_E_SERVICE_NOT_AVAILABLE')
    PEER_E_CANNOT_CONVERT_PEER_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80634001, 'PEER_E_CANNOT_CONVERT_PEER_NAME')
    PEER_E_INVALID_PEER_HOST_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80634002, 'PEER_E_INVALID_PEER_HOST_NAME')
    PEER_E_NO_MORE: Final[ThirtyTwoBits] = cached_hresult(0x80634003, 'PEER_E_NO_MORE')
    PEER_E_PNRP_DUPLICATE_PEER_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80634005, 'PEER_E_PNRP_DUPLICATE_PEER_NAME')
    PEER_S_SUBSCRIPTION_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x00636000, 'PEER_S_SUBSCRIPTION_EXISTS')
    PEER_E_CONTACT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80636001, 'PEER_E_CONTACT_NOT_FOUND')
    PEER_E_INVITE_CANCELLED: Final[ThirtyTwoBits] = cached_hresult(0x80637000, 'PEER_E_INVITE_CANCELLED')
    PEER_E_INVITE_RESPONSE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80637001, 'PEER_E_INVITE_RESPONSE_NOT_AVAILABLE')
    PEER_E_NOT_SIGNED_IN: Final[ThirtyTwoBits] = cached_hresult(0x80637003, 'PEER_E_NOT_SIGNED_IN')
    PEER_E_PRIVACY_DECLINED: Final[ThirtyTwoBits] = cached_hresult(0x80637004, 'PEER_E_PRIVACY_DECLINED')
    PEER_E_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x80637005, 'PEER_E_TIMEOUT')
    PEER_E_INVALID_ADDRESS: Final[ThirtyTwoBits] = cached_hresult(0x80637007, 'PEER_E_INVALID_ADDRESS')
    PEER_E_FW_EXCEPTION_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80637008, 'PEER_E_FW_EXCEPTION_DISABLED')
    PEER_E_FW_BLOCKED_BY_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x80637009, 'PEER_E_FW_BLOCKED_BY_POLICY')
    PEER_E_FW_BLOCKED_BY_SHIELDS_UP: Final[ThirtyTwoBits] = cached_hresult(0x8063700A, 'PEER_E_FW_BLOCKED_BY_SHIELDS_UP')
    PEER_E_FW_DECLINED: Final[ThirtyTwoBits] = cached_hresult(0x8063700B, 'PEER_E_FW_DECLINED')
    E_BLUETOOTH_ATT_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80650001, 'E_BLUETOOTH_ATT_INVALID_HANDLE')
    E_BLUETOOTH_ATT_READ_NOT_PERMITTED: Final[ThirtyTwoBits] = cached_hresult(0x80650002, 'E_BLUETOOTH_ATT_READ_NOT_PERMITTED')
    E_BLUETOOTH_ATT_WRITE_NOT_PERMITTED: Final[ThirtyTwoBits] = cached_hresult(0x80650003, 'E_BLUETOOTH_ATT_WRITE_NOT_PERMITTED')
    E_BLUETOOTH_ATT_INVALID_PDU: Final[ThirtyTwoBits] = cached_hresult(0x80650004, 'E_BLUETOOTH_ATT_INVALID_PDU')
    E_BLUETOOTH_ATT_INSUFFICIENT_AUTHENTICATION: Final[ThirtyTwoBits] = cached_hresult(0x80650005, 'E_BLUETOOTH_ATT_INSUFFICIENT_AUTHENTICATION')
    E_BLUETOOTH_ATT_REQUEST_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80650006, 'E_BLUETOOTH_ATT_REQUEST_NOT_SUPPORTED')
    E_BLUETOOTH_ATT_INVALID_OFFSET: Final[ThirtyTwoBits] = cached_hresult(0x80650007, 'E_BLUETOOTH_ATT_INVALID_OFFSET')
    E_BLUETOOTH_ATT_INSUFFICIENT_AUTHORIZATION: Final[ThirtyTwoBits] = cached_hresult(0x80650008, 'E_BLUETOOTH_ATT_INSUFFICIENT_AUTHORIZATION')
    E_BLUETOOTH_ATT_PREPARE_QUEUE_FULL: Final[ThirtyTwoBits] = cached_hresult(0x80650009, 'E_BLUETOOTH_ATT_PREPARE_QUEUE_FULL')
    E_BLUETOOTH_ATT_ATTRIBUTE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8065000A, 'E_BLUETOOTH_ATT_ATTRIBUTE_NOT_FOUND')
    E_BLUETOOTH_ATT_ATTRIBUTE_NOT_LONG: Final[ThirtyTwoBits] = cached_hresult(0x8065000B, 'E_BLUETOOTH_ATT_ATTRIBUTE_NOT_LONG')
    E_BLUETOOTH_ATT_INSUFFICIENT_ENCRYPTION_KEY_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x8065000C, 'E_BLUETOOTH_ATT_INSUFFICIENT_ENCRYPTION_KEY_SIZE')
    E_BLUETOOTH_ATT_INVALID_ATTRIBUTE_VALUE_LENGTH: Final[ThirtyTwoBits] = cached_hresult(0x8065000D, 'E_BLUETOOTH_ATT_INVALID_ATTRIBUTE_VALUE_LENGTH')
    E_BLUETOOTH_ATT_UNLIKELY: Final[ThirtyTwoBits] = cached_hresult(0x8065000E, 'E_BLUETOOTH_ATT_UNLIKELY')
    E_BLUETOOTH_ATT_INSUFFICIENT_ENCRYPTION: Final[ThirtyTwoBits] = cached_hresult(0x8065000F, 'E_BLUETOOTH_ATT_INSUFFICIENT_ENCRYPTION')
    E_BLUETOOTH_ATT_UNSUPPORTED_GROUP_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x80650010, 'E_BLUETOOTH_ATT_UNSUPPORTED_GROUP_TYPE')
    E_BLUETOOTH_ATT_INSUFFICIENT_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0x80650011, 'E_BLUETOOTH_ATT_INSUFFICIENT_RESOURCES')
    E_BLUETOOTH_ATT_UNKNOWN_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80651000, 'E_BLUETOOTH_ATT_UNKNOWN_ERROR')
    E_AUDIO_ENGINE_NODE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80660001, 'E_AUDIO_ENGINE_NODE_NOT_FOUND')
    E_HDAUDIO_EMPTY_CONNECTION_LIST: Final[ThirtyTwoBits] = cached_hresult(0x80660002, 'E_HDAUDIO_EMPTY_CONNECTION_LIST')
    E_HDAUDIO_CONNECTION_LIST_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80660003, 'E_HDAUDIO_CONNECTION_LIST_NOT_SUPPORTED')
    E_HDAUDIO_NO_LOGICAL_DEVICES_CREATED: Final[ThirtyTwoBits] = cached_hresult(0x80660004, 'E_HDAUDIO_NO_LOGICAL_DEVICES_CREATED')
    E_HDAUDIO_NULL_LINKED_LIST_ENTRY: Final[ThirtyTwoBits] = cached_hresult(0x80660005, 'E_HDAUDIO_NULL_LINKED_LIST_ENTRY')
    STATEREPOSITORY_E_CONCURRENCY_LOCKING_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80670001, 'STATEREPOSITORY_E_CONCURRENCY_LOCKING_FAILURE')
    STATEREPOSITORY_E_STATEMENT_INPROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x80670002, 'STATEREPOSITORY_E_STATEMENT_INPROGRESS')
    STATEREPOSITORY_E_CONFIGURATION_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80670003, 'STATEREPOSITORY_E_CONFIGURATION_INVALID')
    STATEREPOSITORY_E_UNKNOWN_SCHEMA_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80670004, 'STATEREPOSITORY_E_UNKNOWN_SCHEMA_VERSION')
    STATEREPOSITORY_ERROR_DICTIONARY_CORRUPTED: Final[ThirtyTwoBits] = cached_hresult(0x80670005, 'STATEREPOSITORY_ERROR_DICTIONARY_CORRUPTED')
    STATEREPOSITORY_E_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x80670006, 'STATEREPOSITORY_E_BLOCKED')
    STATEREPOSITORY_E_BUSY_RETRY: Final[ThirtyTwoBits] = cached_hresult(0x80670007, 'STATEREPOSITORY_E_BUSY_RETRY')
    STATEREPOSITORY_E_BUSY_RECOVERY_RETRY: Final[ThirtyTwoBits] = cached_hresult(0x80670008, 'STATEREPOSITORY_E_BUSY_RECOVERY_RETRY')
    STATEREPOSITORY_E_LOCKED_RETRY: Final[ThirtyTwoBits] = cached_hresult(0x80670009, 'STATEREPOSITORY_E_LOCKED_RETRY')
    STATEREPOSITORY_E_LOCKED_SHAREDCACHE_RETRY: Final[ThirtyTwoBits] = cached_hresult(0x8067000A, 'STATEREPOSITORY_E_LOCKED_SHAREDCACHE_RETRY')
    STATEREPOSITORY_E_TRANSACTION_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8067000B, 'STATEREPOSITORY_E_TRANSACTION_REQUIRED')
    STATEREPOSITORY_E_BUSY_TIMEOUT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8067000C, 'STATEREPOSITORY_E_BUSY_TIMEOUT_EXCEEDED')
    STATEREPOSITORY_E_BUSY_RECOVERY_TIMEOUT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8067000D, 'STATEREPOSITORY_E_BUSY_RECOVERY_TIMEOUT_EXCEEDED')
    STATEREPOSITORY_E_LOCKED_TIMEOUT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8067000E, 'STATEREPOSITORY_E_LOCKED_TIMEOUT_EXCEEDED')
    STATEREPOSITORY_E_LOCKED_SHAREDCACHE_TIMEOUT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8067000F, 'STATEREPOSITORY_E_LOCKED_SHAREDCACHE_TIMEOUT_EXCEEDED')
    STATEREPOSITORY_E_SERVICE_STOP_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x80670010, 'STATEREPOSITORY_E_SERVICE_STOP_IN_PROGRESS')
    STATEREPOSTORY_E_NESTED_TRANSACTION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80670011, 'STATEREPOSTORY_E_NESTED_TRANSACTION_NOT_SUPPORTED')
    STATEREPOSITORY_ERROR_CACHE_CORRUPTED: Final[ThirtyTwoBits] = cached_hresult(0x80670012, 'STATEREPOSITORY_ERROR_CACHE_CORRUPTED')
    STATEREPOSITORY_TRANSACTION_CALLER_ID_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0x00670013, 'STATEREPOSITORY_TRANSACTION_CALLER_ID_CHANGED')
    STATEREPOSITORY_TRANSACTION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x80670014, 'STATEREPOSITORY_TRANSACTION_IN_PROGRESS')
    STATEREPOSITORY_E_CACHE_NOT_INIITALIZED: Final[ThirtyTwoBits] = cached_hresult(0x80670015, 'STATEREPOSITORY_E_CACHE_NOT_INIITALIZED')
    STATEREPOSITORY_E_DEPENDENCY_NOT_RESOLVED: Final[ThirtyTwoBits] = cached_hresult(0x80670016, 'STATEREPOSITORY_E_DEPENDENCY_NOT_RESOLVED')
    ERROR_VOLSNAP_BOOTFILE_NOT_VALID: Final[ThirtyTwoBits] = cached_hresult(0x80820001, 'ERROR_VOLSNAP_BOOTFILE_NOT_VALID')
    ERROR_VOLSNAP_ACTIVATION_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x80820002, 'ERROR_VOLSNAP_ACTIVATION_TIMEOUT')
    ERROR_VOLSNAP_NO_BYPASSIO_WITH_SNAPSHOT: Final[ThirtyTwoBits] = cached_hresult(0x80820003, 'ERROR_VOLSNAP_NO_BYPASSIO_WITH_SNAPSHOT')
    ERROR_TIERING_NOT_SUPPORTED_ON_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0x80830001, 'ERROR_TIERING_NOT_SUPPORTED_ON_VOLUME')
    ERROR_TIERING_VOLUME_DISMOUNT_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x80830002, 'ERROR_TIERING_VOLUME_DISMOUNT_IN_PROGRESS')
    ERROR_TIERING_STORAGE_TIER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80830003, 'ERROR_TIERING_STORAGE_TIER_NOT_FOUND')
    ERROR_TIERING_INVALID_FILE_ID: Final[ThirtyTwoBits] = cached_hresult(0x80830004, 'ERROR_TIERING_INVALID_FILE_ID')
    ERROR_TIERING_WRONG_CLUSTER_NODE: Final[ThirtyTwoBits] = cached_hresult(0x80830005, 'ERROR_TIERING_WRONG_CLUSTER_NODE')
    ERROR_TIERING_ALREADY_PROCESSING: Final[ThirtyTwoBits] = cached_hresult(0x80830006, 'ERROR_TIERING_ALREADY_PROCESSING')
    ERROR_TIERING_CANNOT_PIN_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x80830007, 'ERROR_TIERING_CANNOT_PIN_OBJECT')
    ERROR_TIERING_FILE_IS_NOT_PINNED: Final[ThirtyTwoBits] = cached_hresult(0x80830008, 'ERROR_TIERING_FILE_IS_NOT_PINNED')
    ERROR_NOT_A_TIERED_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0x80830009, 'ERROR_NOT_A_TIERED_VOLUME')
    ERROR_ATTRIBUTE_NOT_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0x8083000A, 'ERROR_ATTRIBUTE_NOT_PRESENT')
    ONL_E_INVALID_AUTHENTICATION_TARGET: Final[ThirtyTwoBits] = cached_hresult(0x80860001, 'ONL_E_INVALID_AUTHENTICATION_TARGET')
    ONL_E_ACCESS_DENIED_BY_TOU: Final[ThirtyTwoBits] = cached_hresult(0x80860002, 'ONL_E_ACCESS_DENIED_BY_TOU')
    ONL_E_INVALID_APPLICATION: Final[ThirtyTwoBits] = cached_hresult(0x80860003, 'ONL_E_INVALID_APPLICATION')
    ONL_E_PASSWORD_UPDATE_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80860004, 'ONL_E_PASSWORD_UPDATE_REQUIRED')
    ONL_E_ACCOUNT_UPDATE_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80860005, 'ONL_E_ACCOUNT_UPDATE_REQUIRED')
    ONL_E_FORCESIGNIN: Final[ThirtyTwoBits] = cached_hresult(0x80860006, 'ONL_E_FORCESIGNIN')
    ONL_E_ACCOUNT_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x80860007, 'ONL_E_ACCOUNT_LOCKED')
    ONL_E_PARENTAL_CONSENT_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80860008, 'ONL_E_PARENTAL_CONSENT_REQUIRED')
    ONL_E_EMAIL_VERIFICATION_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80860009, 'ONL_E_EMAIL_VERIFICATION_REQUIRED')
    ONL_E_ACCOUNT_SUSPENDED_COMPROIMISE: Final[ThirtyTwoBits] = cached_hresult(0x8086000A, 'ONL_E_ACCOUNT_SUSPENDED_COMPROIMISE')
    ONL_E_ACCOUNT_SUSPENDED_ABUSE: Final[ThirtyTwoBits] = cached_hresult(0x8086000B, 'ONL_E_ACCOUNT_SUSPENDED_ABUSE')
    ONL_E_ACTION_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8086000C, 'ONL_E_ACTION_REQUIRED')
    ONL_CONNECTION_COUNT_LIMIT: Final[ThirtyTwoBits] = cached_hresult(0x8086000D, 'ONL_CONNECTION_COUNT_LIMIT')
    ONL_E_CONNECTED_ACCOUNT_CAN_NOT_SIGNOUT: Final[ThirtyTwoBits] = cached_hresult(0x8086000E, 'ONL_E_CONNECTED_ACCOUNT_CAN_NOT_SIGNOUT')
    ONL_E_USER_AUTHENTICATION_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8086000F, 'ONL_E_USER_AUTHENTICATION_REQUIRED')
    ONL_E_REQUEST_THROTTLED: Final[ThirtyTwoBits] = cached_hresult(0x80860010, 'ONL_E_REQUEST_THROTTLED')
    ERROR_DBG_CREATE_PROCESS_FAILURE_LOCKDOWN: Final[ThirtyTwoBits] = cached_hresult(0x80B00001, 'ERROR_DBG_CREATE_PROCESS_FAILURE_LOCKDOWN')
    ERROR_DBG_ATTACH_PROCESS_FAILURE_LOCKDOWN: Final[ThirtyTwoBits] = cached_hresult(0x80B00002, 'ERROR_DBG_ATTACH_PROCESS_FAILURE_LOCKDOWN')
    ERROR_DBG_CONNECT_SERVER_FAILURE_LOCKDOWN: Final[ThirtyTwoBits] = cached_hresult(0x80B00003, 'ERROR_DBG_CONNECT_SERVER_FAILURE_LOCKDOWN')
    ERROR_DBG_START_SERVER_FAILURE_LOCKDOWN: Final[ThirtyTwoBits] = cached_hresult(0x80B00004, 'ERROR_DBG_START_SERVER_FAILURE_LOCKDOWN')
    ERROR_SPACES_POOL_WAS_DELETED: Final[ThirtyTwoBits] = cached_hresult(0x00E70001, 'ERROR_SPACES_POOL_WAS_DELETED')
    ERROR_SPACES_FAULT_DOMAIN_TYPE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80E70001, 'ERROR_SPACES_FAULT_DOMAIN_TYPE_INVALID')
    ERROR_SPACES_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80E70002, 'ERROR_SPACES_INTERNAL_ERROR')
    ERROR_SPACES_RESILIENCY_TYPE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80E70003, 'ERROR_SPACES_RESILIENCY_TYPE_INVALID')
    ERROR_SPACES_DRIVE_SECTOR_SIZE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80E70004, 'ERROR_SPACES_DRIVE_SECTOR_SIZE_INVALID')
    ERROR_SPACES_DRIVE_REDUNDANCY_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80E70006, 'ERROR_SPACES_DRIVE_REDUNDANCY_INVALID')
    ERROR_SPACES_NUMBER_OF_DATA_COPIES_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80E70007, 'ERROR_SPACES_NUMBER_OF_DATA_COPIES_INVALID')
    ERROR_SPACES_PARITY_LAYOUT_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80E70008, 'ERROR_SPACES_PARITY_LAYOUT_INVALID')
    ERROR_SPACES_INTERLEAVE_LENGTH_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80E70009, 'ERROR_SPACES_INTERLEAVE_LENGTH_INVALID')
    ERROR_SPACES_NUMBER_OF_COLUMNS_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80E7000A, 'ERROR_SPACES_NUMBER_OF_COLUMNS_INVALID')
    ERROR_SPACES_NOT_ENOUGH_DRIVES: Final[ThirtyTwoBits] = cached_hresult(0x80E7000B, 'ERROR_SPACES_NOT_ENOUGH_DRIVES')
    ERROR_SPACES_EXTENDED_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80E7000C, 'ERROR_SPACES_EXTENDED_ERROR')
    ERROR_SPACES_PROVISIONING_TYPE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80E7000D, 'ERROR_SPACES_PROVISIONING_TYPE_INVALID')
    ERROR_SPACES_ALLOCATION_SIZE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80E7000E, 'ERROR_SPACES_ALLOCATION_SIZE_INVALID')
    ERROR_SPACES_ENCLOSURE_AWARE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80E7000F, 'ERROR_SPACES_ENCLOSURE_AWARE_INVALID')
    ERROR_SPACES_WRITE_CACHE_SIZE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80E70010, 'ERROR_SPACES_WRITE_CACHE_SIZE_INVALID')
    ERROR_SPACES_NUMBER_OF_GROUPS_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80E70011, 'ERROR_SPACES_NUMBER_OF_GROUPS_INVALID')
    ERROR_SPACES_DRIVE_OPERATIONAL_STATE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80E70012, 'ERROR_SPACES_DRIVE_OPERATIONAL_STATE_INVALID')
    ERROR_SPACES_ENTRY_INCOMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x80E70013, 'ERROR_SPACES_ENTRY_INCOMPLETE')
    ERROR_SPACES_ENTRY_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80E70014, 'ERROR_SPACES_ENTRY_INVALID')
    ERROR_SPACES_UPDATE_COLUMN_STATE: Final[ThirtyTwoBits] = cached_hresult(0x80E70015, 'ERROR_SPACES_UPDATE_COLUMN_STATE')
    ERROR_SPACES_MAP_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80E70016, 'ERROR_SPACES_MAP_REQUIRED')
    ERROR_SPACES_UNSUPPORTED_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80E70017, 'ERROR_SPACES_UNSUPPORTED_VERSION')
    ERROR_SPACES_CORRUPT_METADATA: Final[ThirtyTwoBits] = cached_hresult(0x80E70018, 'ERROR_SPACES_CORRUPT_METADATA')
    ERROR_SPACES_DRT_FULL: Final[ThirtyTwoBits] = cached_hresult(0x80E70019, 'ERROR_SPACES_DRT_FULL')
    ERROR_SPACES_INCONSISTENCY: Final[ThirtyTwoBits] = cached_hresult(0x80E7001A, 'ERROR_SPACES_INCONSISTENCY')
    ERROR_SPACES_LOG_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0x80E7001B, 'ERROR_SPACES_LOG_NOT_READY')
    ERROR_SPACES_NO_REDUNDANCY: Final[ThirtyTwoBits] = cached_hresult(0x80E7001C, 'ERROR_SPACES_NO_REDUNDANCY')
    ERROR_SPACES_DRIVE_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0x80E7001D, 'ERROR_SPACES_DRIVE_NOT_READY')
    ERROR_SPACES_DRIVE_SPLIT: Final[ThirtyTwoBits] = cached_hresult(0x80E7001E, 'ERROR_SPACES_DRIVE_SPLIT')
    ERROR_SPACES_DRIVE_LOST_DATA: Final[ThirtyTwoBits] = cached_hresult(0x80E7001F, 'ERROR_SPACES_DRIVE_LOST_DATA')
    ERROR_SPACES_MARK_DIRTY: Final[ThirtyTwoBits] = cached_hresult(0x80E70020, 'ERROR_SPACES_MARK_DIRTY')
    ERROR_SPACES_FLUSH_METADATA: Final[ThirtyTwoBits] = cached_hresult(0x80E70025, 'ERROR_SPACES_FLUSH_METADATA')
    ERROR_SPACES_CACHE_FULL: Final[ThirtyTwoBits] = cached_hresult(0x80E70026, 'ERROR_SPACES_CACHE_FULL')
    ERROR_SPACES_REPAIR_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x80E70027, 'ERROR_SPACES_REPAIR_IN_PROGRESS')
    ERROR_SECCORE_INVALID_COMMAND: Final[ThirtyTwoBits] = cached_hresult(0xC0E80000, 'ERROR_SECCORE_INVALID_COMMAND')
    ERROR_NO_APPLICABLE_APP_LICENSES_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xC0EA0001, 'ERROR_NO_APPLICABLE_APP_LICENSES_FOUND')
    ERROR_CLIP_LICENSE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xC0EA0002, 'ERROR_CLIP_LICENSE_NOT_FOUND')
    ERROR_CLIP_DEVICE_LICENSE_MISSING: Final[ThirtyTwoBits] = cached_hresult(0xC0EA0003, 'ERROR_CLIP_DEVICE_LICENSE_MISSING')
    ERROR_CLIP_LICENSE_INVALID_SIGNATURE: Final[ThirtyTwoBits] = cached_hresult(0xC0EA0004, 'ERROR_CLIP_LICENSE_INVALID_SIGNATURE')
    ERROR_CLIP_KEYHOLDER_LICENSE_MISSING_OR_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xC0EA0005, 'ERROR_CLIP_KEYHOLDER_LICENSE_MISSING_OR_INVALID')
    ERROR_CLIP_LICENSE_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0xC0EA0006, 'ERROR_CLIP_LICENSE_EXPIRED')
    ERROR_CLIP_LICENSE_SIGNED_BY_UNKNOWN_SOURCE: Final[ThirtyTwoBits] = cached_hresult(0xC0EA0007, 'ERROR_CLIP_LICENSE_SIGNED_BY_UNKNOWN_SOURCE')
    ERROR_CLIP_LICENSE_NOT_SIGNED: Final[ThirtyTwoBits] = cached_hresult(0xC0EA0008, 'ERROR_CLIP_LICENSE_NOT_SIGNED')
    ERROR_CLIP_LICENSE_HARDWARE_ID_OUT_OF_TOLERANCE: Final[ThirtyTwoBits] = cached_hresult(0xC0EA0009, 'ERROR_CLIP_LICENSE_HARDWARE_ID_OUT_OF_TOLERANCE')
    ERROR_CLIP_LICENSE_DEVICE_ID_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xC0EA000A, 'ERROR_CLIP_LICENSE_DEVICE_ID_MISMATCH')
    HSP_E_ERROR_MASK: Final[ThirtyTwoBits] = cached_hresult(0x81280000, 'HSP_E_ERROR_MASK')
    HSP_E_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x81280FFF, 'HSP_E_INTERNAL_ERROR')
    HSP_BS_ERROR_MASK: Final[ThirtyTwoBits] = cached_hresult(0x81281000, 'HSP_BS_ERROR_MASK')
    HSP_BS_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x812810FF, 'HSP_BS_INTERNAL_ERROR')
    HSP_DRV_ERROR_MASK: Final[ThirtyTwoBits] = cached_hresult(0x81290000, 'HSP_DRV_ERROR_MASK')
    HSP_DRV_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x812900FF, 'HSP_DRV_INTERNAL_ERROR')
    HSP_BASE_ERROR_MASK: Final[ThirtyTwoBits] = cached_hresult(0x81290100, 'HSP_BASE_ERROR_MASK')
    HSP_BASE_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x812901FF, 'HSP_BASE_INTERNAL_ERROR')
    HSP_KSP_ERROR_MASK: Final[ThirtyTwoBits] = cached_hresult(0x81290200, 'HSP_KSP_ERROR_MASK')
    HSP_KSP_DEVICE_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0x81290201, 'HSP_KSP_DEVICE_NOT_READY')
    HSP_KSP_INVALID_PROVIDER_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x81290202, 'HSP_KSP_INVALID_PROVIDER_HANDLE')
    HSP_KSP_INVALID_KEY_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x81290203, 'HSP_KSP_INVALID_KEY_HANDLE')
    HSP_KSP_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0x81290204, 'HSP_KSP_INVALID_PARAMETER')
    HSP_KSP_BUFFER_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0x81290205, 'HSP_KSP_BUFFER_TOO_SMALL')
    HSP_KSP_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x81290206, 'HSP_KSP_NOT_SUPPORTED')
    HSP_KSP_INVALID_DATA: Final[ThirtyTwoBits] = cached_hresult(0x81290207, 'HSP_KSP_INVALID_DATA')
    HSP_KSP_INVALID_FLAGS: Final[ThirtyTwoBits] = cached_hresult(0x81290208, 'HSP_KSP_INVALID_FLAGS')
    HSP_KSP_ALGORITHM_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x81290209, 'HSP_KSP_ALGORITHM_NOT_SUPPORTED')
    HSP_KSP_KEY_ALREADY_FINALIZED: Final[ThirtyTwoBits] = cached_hresult(0x8129020A, 'HSP_KSP_KEY_ALREADY_FINALIZED')
    HSP_KSP_KEY_NOT_FINALIZED: Final[ThirtyTwoBits] = cached_hresult(0x8129020B, 'HSP_KSP_KEY_NOT_FINALIZED')
    HSP_KSP_INVALID_KEY_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x8129020C, 'HSP_KSP_INVALID_KEY_TYPE')
    HSP_KSP_NO_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0x81290210, 'HSP_KSP_NO_MEMORY')
    HSP_KSP_PARAMETER_NOT_SET: Final[ThirtyTwoBits] = cached_hresult(0x81290211, 'HSP_KSP_PARAMETER_NOT_SET')
    HSP_KSP_KEY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x81290215, 'HSP_KSP_KEY_EXISTS')
    HSP_KSP_KEY_MISSING: Final[ThirtyTwoBits] = cached_hresult(0x81290216, 'HSP_KSP_KEY_MISSING')
    HSP_KSP_KEY_LOAD_FAIL: Final[ThirtyTwoBits] = cached_hresult(0x81290217, 'HSP_KSP_KEY_LOAD_FAIL')
    HSP_KSP_NO_MORE_ITEMS: Final[ThirtyTwoBits] = cached_hresult(0x81290218, 'HSP_KSP_NO_MORE_ITEMS')
    HSP_KSP_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x812902FF, 'HSP_KSP_INTERNAL_ERROR')
    WEB_E_UNSUPPORTED_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0x83750001, 'WEB_E_UNSUPPORTED_FORMAT')
    WEB_E_INVALID_XML: Final[ThirtyTwoBits] = cached_hresult(0x83750002, 'WEB_E_INVALID_XML')
    WEB_E_MISSING_REQUIRED_ELEMENT: Final[ThirtyTwoBits] = cached_hresult(0x83750003, 'WEB_E_MISSING_REQUIRED_ELEMENT')
    WEB_E_MISSING_REQUIRED_ATTRIBUTE: Final[ThirtyTwoBits] = cached_hresult(0x83750004, 'WEB_E_MISSING_REQUIRED_ATTRIBUTE')
    WEB_E_UNEXPECTED_CONTENT: Final[ThirtyTwoBits] = cached_hresult(0x83750005, 'WEB_E_UNEXPECTED_CONTENT')
    WEB_E_RESOURCE_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0x83750006, 'WEB_E_RESOURCE_TOO_LARGE')
    WEB_E_INVALID_JSON_STRING: Final[ThirtyTwoBits] = cached_hresult(0x83750007, 'WEB_E_INVALID_JSON_STRING')
    WEB_E_INVALID_JSON_NUMBER: Final[ThirtyTwoBits] = cached_hresult(0x83750008, 'WEB_E_INVALID_JSON_NUMBER')
    WEB_E_JSON_VALUE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x83750009, 'WEB_E_JSON_VALUE_NOT_FOUND')
    E_INVALID_PROTOCOL_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x83760001, 'E_INVALID_PROTOCOL_OPERATION')
    E_INVALID_PROTOCOL_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0x83760002, 'E_INVALID_PROTOCOL_FORMAT')
    E_PROTOCOL_EXTENSIONS_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x83760003, 'E_PROTOCOL_EXTENSIONS_NOT_SUPPORTED')
    E_SUBPROTOCOL_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x83760004, 'E_SUBPROTOCOL_NOT_SUPPORTED')
    E_PROTOCOL_VERSION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x83760005, 'E_PROTOCOL_VERSION_NOT_SUPPORTED')
    SQLITE_E_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x87AF0001, 'SQLITE_E_ERROR')
    SQLITE_E_INTERNAL: Final[ThirtyTwoBits] = cached_hresult(0x87AF0002, 'SQLITE_E_INTERNAL')
    SQLITE_E_PERM: Final[ThirtyTwoBits] = cached_hresult(0x87AF0003, 'SQLITE_E_PERM')
    SQLITE_E_ABORT: Final[ThirtyTwoBits] = cached_hresult(0x87AF0004, 'SQLITE_E_ABORT')
    SQLITE_E_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x87AF0005, 'SQLITE_E_BUSY')
    SQLITE_E_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x87AF0006, 'SQLITE_E_LOCKED')
    SQLITE_E_NOMEM: Final[ThirtyTwoBits] = cached_hresult(0x87AF0007, 'SQLITE_E_NOMEM')
    SQLITE_E_READONLY: Final[ThirtyTwoBits] = cached_hresult(0x87AF0008, 'SQLITE_E_READONLY')
    SQLITE_E_INTERRUPT: Final[ThirtyTwoBits] = cached_hresult(0x87AF0009, 'SQLITE_E_INTERRUPT')
    SQLITE_E_IOERR: Final[ThirtyTwoBits] = cached_hresult(0x87AF000A, 'SQLITE_E_IOERR')
    SQLITE_E_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x87AF000B, 'SQLITE_E_CORRUPT')
    SQLITE_E_NOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x87AF000C, 'SQLITE_E_NOTFOUND')
    SQLITE_E_FULL: Final[ThirtyTwoBits] = cached_hresult(0x87AF000D, 'SQLITE_E_FULL')
    SQLITE_E_CANTOPEN: Final[ThirtyTwoBits] = cached_hresult(0x87AF000E, 'SQLITE_E_CANTOPEN')
    SQLITE_E_PROTOCOL: Final[ThirtyTwoBits] = cached_hresult(0x87AF000F, 'SQLITE_E_PROTOCOL')
    SQLITE_E_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0x87AF0010, 'SQLITE_E_EMPTY')
    SQLITE_E_SCHEMA: Final[ThirtyTwoBits] = cached_hresult(0x87AF0011, 'SQLITE_E_SCHEMA')
    SQLITE_E_TOOBIG: Final[ThirtyTwoBits] = cached_hresult(0x87AF0012, 'SQLITE_E_TOOBIG')
    SQLITE_E_CONSTRAINT: Final[ThirtyTwoBits] = cached_hresult(0x87AF0013, 'SQLITE_E_CONSTRAINT')
    SQLITE_E_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x87AF0014, 'SQLITE_E_MISMATCH')
    SQLITE_E_MISUSE: Final[ThirtyTwoBits] = cached_hresult(0x87AF0015, 'SQLITE_E_MISUSE')
    SQLITE_E_NOLFS: Final[ThirtyTwoBits] = cached_hresult(0x87AF0016, 'SQLITE_E_NOLFS')
    SQLITE_E_AUTH: Final[ThirtyTwoBits] = cached_hresult(0x87AF0017, 'SQLITE_E_AUTH')
    SQLITE_E_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0x87AF0018, 'SQLITE_E_FORMAT')
    SQLITE_E_RANGE: Final[ThirtyTwoBits] = cached_hresult(0x87AF0019, 'SQLITE_E_RANGE')
    SQLITE_E_NOTADB: Final[ThirtyTwoBits] = cached_hresult(0x87AF001A, 'SQLITE_E_NOTADB')
    SQLITE_E_NOTICE: Final[ThirtyTwoBits] = cached_hresult(0x87AF001B, 'SQLITE_E_NOTICE')
    SQLITE_E_WARNING: Final[ThirtyTwoBits] = cached_hresult(0x87AF001C, 'SQLITE_E_WARNING')
    SQLITE_E_ROW: Final[ThirtyTwoBits] = cached_hresult(0x87AF0064, 'SQLITE_E_ROW')
    SQLITE_E_DONE: Final[ThirtyTwoBits] = cached_hresult(0x87AF0065, 'SQLITE_E_DONE')
    SQLITE_E_BUSY_RECOVERY: Final[ThirtyTwoBits] = cached_hresult(0x87AF0105, 'SQLITE_E_BUSY_RECOVERY')
    SQLITE_E_LOCKED_SHAREDCACHE: Final[ThirtyTwoBits] = cached_hresult(0x87AF0106, 'SQLITE_E_LOCKED_SHAREDCACHE')
    SQLITE_E_READONLY_RECOVERY: Final[ThirtyTwoBits] = cached_hresult(0x87AF0108, 'SQLITE_E_READONLY_RECOVERY')
    SQLITE_E_IOERR_READ: Final[ThirtyTwoBits] = cached_hresult(0x87AF010A, 'SQLITE_E_IOERR_READ')
    SQLITE_E_CORRUPT_VTAB: Final[ThirtyTwoBits] = cached_hresult(0x87AF010B, 'SQLITE_E_CORRUPT_VTAB')
    SQLITE_E_CANTOPEN_NOTEMPDIR: Final[ThirtyTwoBits] = cached_hresult(0x87AF010E, 'SQLITE_E_CANTOPEN_NOTEMPDIR')
    SQLITE_E_CONSTRAINT_CHECK: Final[ThirtyTwoBits] = cached_hresult(0x87AF0113, 'SQLITE_E_CONSTRAINT_CHECK')
    SQLITE_E_NOTICE_RECOVER_WAL: Final[ThirtyTwoBits] = cached_hresult(0x87AF011B, 'SQLITE_E_NOTICE_RECOVER_WAL')
    SQLITE_E_WARNING_AUTOINDEX: Final[ThirtyTwoBits] = cached_hresult(0x87AF011C, 'SQLITE_E_WARNING_AUTOINDEX')
    SQLITE_E_ABORT_ROLLBACK: Final[ThirtyTwoBits] = cached_hresult(0x87AF0204, 'SQLITE_E_ABORT_ROLLBACK')
    SQLITE_E_BUSY_SNAPSHOT: Final[ThirtyTwoBits] = cached_hresult(0x87AF0205, 'SQLITE_E_BUSY_SNAPSHOT')
    SQLITE_E_READONLY_CANTLOCK: Final[ThirtyTwoBits] = cached_hresult(0x87AF0208, 'SQLITE_E_READONLY_CANTLOCK')
    SQLITE_E_IOERR_SHORT_READ: Final[ThirtyTwoBits] = cached_hresult(0x87AF020A, 'SQLITE_E_IOERR_SHORT_READ')
    SQLITE_E_CANTOPEN_ISDIR: Final[ThirtyTwoBits] = cached_hresult(0x87AF020E, 'SQLITE_E_CANTOPEN_ISDIR')
    SQLITE_E_CONSTRAINT_COMMITHOOK: Final[ThirtyTwoBits] = cached_hresult(0x87AF0213, 'SQLITE_E_CONSTRAINT_COMMITHOOK')
    SQLITE_E_NOTICE_RECOVER_ROLLBACK: Final[ThirtyTwoBits] = cached_hresult(0x87AF021B, 'SQLITE_E_NOTICE_RECOVER_ROLLBACK')
    SQLITE_E_READONLY_ROLLBACK: Final[ThirtyTwoBits] = cached_hresult(0x87AF0308, 'SQLITE_E_READONLY_ROLLBACK')
    SQLITE_E_IOERR_WRITE: Final[ThirtyTwoBits] = cached_hresult(0x87AF030A, 'SQLITE_E_IOERR_WRITE')
    SQLITE_E_CANTOPEN_FULLPATH: Final[ThirtyTwoBits] = cached_hresult(0x87AF030E, 'SQLITE_E_CANTOPEN_FULLPATH')
    SQLITE_E_CONSTRAINT_FOREIGNKEY: Final[ThirtyTwoBits] = cached_hresult(0x87AF0313, 'SQLITE_E_CONSTRAINT_FOREIGNKEY')
    SQLITE_E_READONLY_DBMOVED: Final[ThirtyTwoBits] = cached_hresult(0x87AF0408, 'SQLITE_E_READONLY_DBMOVED')
    SQLITE_E_IOERR_FSYNC: Final[ThirtyTwoBits] = cached_hresult(0x87AF040A, 'SQLITE_E_IOERR_FSYNC')
    SQLITE_E_CANTOPEN_CONVPATH: Final[ThirtyTwoBits] = cached_hresult(0x87AF040E, 'SQLITE_E_CANTOPEN_CONVPATH')
    SQLITE_E_CONSTRAINT_FUNCTION: Final[ThirtyTwoBits] = cached_hresult(0x87AF0413, 'SQLITE_E_CONSTRAINT_FUNCTION')
    SQLITE_E_IOERR_DIR_FSYNC: Final[ThirtyTwoBits] = cached_hresult(0x87AF050A, 'SQLITE_E_IOERR_DIR_FSYNC')
    SQLITE_E_CONSTRAINT_NOTNULL: Final[ThirtyTwoBits] = cached_hresult(0x87AF0513, 'SQLITE_E_CONSTRAINT_NOTNULL')
    SQLITE_E_IOERR_TRUNCATE: Final[ThirtyTwoBits] = cached_hresult(0x87AF060A, 'SQLITE_E_IOERR_TRUNCATE')
    SQLITE_E_CONSTRAINT_PRIMARYKEY: Final[ThirtyTwoBits] = cached_hresult(0x87AF0613, 'SQLITE_E_CONSTRAINT_PRIMARYKEY')
    SQLITE_E_IOERR_FSTAT: Final[ThirtyTwoBits] = cached_hresult(0x87AF070A, 'SQLITE_E_IOERR_FSTAT')
    SQLITE_E_CONSTRAINT_TRIGGER: Final[ThirtyTwoBits] = cached_hresult(0x87AF0713, 'SQLITE_E_CONSTRAINT_TRIGGER')
    SQLITE_E_IOERR_UNLOCK: Final[ThirtyTwoBits] = cached_hresult(0x87AF080A, 'SQLITE_E_IOERR_UNLOCK')
    SQLITE_E_CONSTRAINT_UNIQUE: Final[ThirtyTwoBits] = cached_hresult(0x87AF0813, 'SQLITE_E_CONSTRAINT_UNIQUE')
    SQLITE_E_IOERR_RDLOCK: Final[ThirtyTwoBits] = cached_hresult(0x87AF090A, 'SQLITE_E_IOERR_RDLOCK')
    SQLITE_E_CONSTRAINT_VTAB: Final[ThirtyTwoBits] = cached_hresult(0x87AF0913, 'SQLITE_E_CONSTRAINT_VTAB')
    SQLITE_E_IOERR_DELETE: Final[ThirtyTwoBits] = cached_hresult(0x87AF0A0A, 'SQLITE_E_IOERR_DELETE')
    SQLITE_E_CONSTRAINT_ROWID: Final[ThirtyTwoBits] = cached_hresult(0x87AF0A13, 'SQLITE_E_CONSTRAINT_ROWID')
    SQLITE_E_IOERR_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x87AF0B0A, 'SQLITE_E_IOERR_BLOCKED')
    SQLITE_E_IOERR_NOMEM: Final[ThirtyTwoBits] = cached_hresult(0x87AF0C0A, 'SQLITE_E_IOERR_NOMEM')
    SQLITE_E_IOERR_ACCESS: Final[ThirtyTwoBits] = cached_hresult(0x87AF0D0A, 'SQLITE_E_IOERR_ACCESS')
    SQLITE_E_IOERR_CHECKRESERVEDLOCK: Final[ThirtyTwoBits] = cached_hresult(0x87AF0E0A, 'SQLITE_E_IOERR_CHECKRESERVEDLOCK')
    SQLITE_E_IOERR_LOCK: Final[ThirtyTwoBits] = cached_hresult(0x87AF0F0A, 'SQLITE_E_IOERR_LOCK')
    SQLITE_E_IOERR_CLOSE: Final[ThirtyTwoBits] = cached_hresult(0x87AF100A, 'SQLITE_E_IOERR_CLOSE')
    SQLITE_E_IOERR_DIR_CLOSE: Final[ThirtyTwoBits] = cached_hresult(0x87AF110A, 'SQLITE_E_IOERR_DIR_CLOSE')
    SQLITE_E_IOERR_SHMOPEN: Final[ThirtyTwoBits] = cached_hresult(0x87AF120A, 'SQLITE_E_IOERR_SHMOPEN')
    SQLITE_E_IOERR_SHMSIZE: Final[ThirtyTwoBits] = cached_hresult(0x87AF130A, 'SQLITE_E_IOERR_SHMSIZE')
    SQLITE_E_IOERR_SHMLOCK: Final[ThirtyTwoBits] = cached_hresult(0x87AF140A, 'SQLITE_E_IOERR_SHMLOCK')
    SQLITE_E_IOERR_SHMMAP: Final[ThirtyTwoBits] = cached_hresult(0x87AF150A, 'SQLITE_E_IOERR_SHMMAP')
    SQLITE_E_IOERR_SEEK: Final[ThirtyTwoBits] = cached_hresult(0x87AF160A, 'SQLITE_E_IOERR_SEEK')
    SQLITE_E_IOERR_DELETE_NOENT: Final[ThirtyTwoBits] = cached_hresult(0x87AF170A, 'SQLITE_E_IOERR_DELETE_NOENT')
    SQLITE_E_IOERR_MMAP: Final[ThirtyTwoBits] = cached_hresult(0x87AF180A, 'SQLITE_E_IOERR_MMAP')
    SQLITE_E_IOERR_GETTEMPPATH: Final[ThirtyTwoBits] = cached_hresult(0x87AF190A, 'SQLITE_E_IOERR_GETTEMPPATH')
    SQLITE_E_IOERR_VNODE: Final[ThirtyTwoBits] = cached_hresult(0x87AF1A02, 'SQLITE_E_IOERR_VNODE')
    SQLITE_E_IOERR_AUTH: Final[ThirtyTwoBits] = cached_hresult(0x87AF1A03, 'SQLITE_E_IOERR_AUTH')
    SQLITE_E_IOERR_CONVPATH: Final[ThirtyTwoBits] = cached_hresult(0x87AF1A0A, 'SQLITE_E_IOERR_CONVPATH')
    UTC_E_TOGGLE_TRACE_STARTED: Final[ThirtyTwoBits] = cached_hresult(0x87C51001, 'UTC_E_TOGGLE_TRACE_STARTED')
    UTC_E_ALTERNATIVE_TRACE_CANNOT_PREEMPT: Final[ThirtyTwoBits] = cached_hresult(0x87C51002, 'UTC_E_ALTERNATIVE_TRACE_CANNOT_PREEMPT')
    UTC_E_AOT_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x87C51003, 'UTC_E_AOT_NOT_RUNNING')
    UTC_E_SCRIPT_TYPE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x87C51004, 'UTC_E_SCRIPT_TYPE_INVALID')
    UTC_E_SCENARIODEF_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x87C51005, 'UTC_E_SCENARIODEF_NOT_FOUND')
    UTC_E_TRACEPROFILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x87C51006, 'UTC_E_TRACEPROFILE_NOT_FOUND')
    UTC_E_FORWARDER_ALREADY_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0x87C51007, 'UTC_E_FORWARDER_ALREADY_ENABLED')
    UTC_E_FORWARDER_ALREADY_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x87C51008, 'UTC_E_FORWARDER_ALREADY_DISABLED')
    UTC_E_EVENTLOG_ENTRY_MALFORMED: Final[ThirtyTwoBits] = cached_hresult(0x87C51009, 'UTC_E_EVENTLOG_ENTRY_MALFORMED')
    UTC_E_DIAGRULES_SCHEMAVERSION_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x87C5100A, 'UTC_E_DIAGRULES_SCHEMAVERSION_MISMATCH')
    UTC_E_SCRIPT_TERMINATED: Final[ThirtyTwoBits] = cached_hresult(0x87C5100B, 'UTC_E_SCRIPT_TERMINATED')
    UTC_E_INVALID_CUSTOM_FILTER: Final[ThirtyTwoBits] = cached_hresult(0x87C5100C, 'UTC_E_INVALID_CUSTOM_FILTER')
    UTC_E_TRACE_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x87C5100D, 'UTC_E_TRACE_NOT_RUNNING')
    UTC_E_REESCALATED_TOO_QUICKLY: Final[ThirtyTwoBits] = cached_hresult(0x87C5100E, 'UTC_E_REESCALATED_TOO_QUICKLY')
    UTC_E_ESCALATION_ALREADY_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x87C5100F, 'UTC_E_ESCALATION_ALREADY_RUNNING')
    UTC_E_PERFTRACK_ALREADY_TRACING: Final[ThirtyTwoBits] = cached_hresult(0x87C51010, 'UTC_E_PERFTRACK_ALREADY_TRACING')
    UTC_E_REACHED_MAX_ESCALATIONS: Final[ThirtyTwoBits] = cached_hresult(0x87C51011, 'UTC_E_REACHED_MAX_ESCALATIONS')
    UTC_E_FORWARDER_PRODUCER_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x87C51012, 'UTC_E_FORWARDER_PRODUCER_MISMATCH')
    UTC_E_INTENTIONAL_SCRIPT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x87C51013, 'UTC_E_INTENTIONAL_SCRIPT_FAILURE')
    UTC_E_SQM_INIT_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x87C51014, 'UTC_E_SQM_INIT_FAILED')
    UTC_E_NO_WER_LOGGER_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x87C51015, 'UTC_E_NO_WER_LOGGER_SUPPORTED')
    UTC_E_TRACERS_DONT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x87C51016, 'UTC_E_TRACERS_DONT_EXIST')
    UTC_E_WINRT_INIT_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x87C51017, 'UTC_E_WINRT_INIT_FAILED')
    UTC_E_SCENARIODEF_SCHEMAVERSION_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x87C51018, 'UTC_E_SCENARIODEF_SCHEMAVERSION_MISMATCH')
    UTC_E_INVALID_FILTER: Final[ThirtyTwoBits] = cached_hresult(0x87C51019, 'UTC_E_INVALID_FILTER')
    UTC_E_EXE_TERMINATED: Final[ThirtyTwoBits] = cached_hresult(0x87C5101A, 'UTC_E_EXE_TERMINATED')
    UTC_E_ESCALATION_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_hresult(0x87C5101B, 'UTC_E_ESCALATION_NOT_AUTHORIZED')
    UTC_E_SETUP_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_hresult(0x87C5101C, 'UTC_E_SETUP_NOT_AUTHORIZED')
    UTC_E_CHILD_PROCESS_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x87C5101D, 'UTC_E_CHILD_PROCESS_FAILED')
    UTC_E_COMMAND_LINE_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_hresult(0x87C5101E, 'UTC_E_COMMAND_LINE_NOT_AUTHORIZED')
    UTC_E_CANNOT_LOAD_SCENARIO_EDITOR_XML: Final[ThirtyTwoBits] = cached_hresult(0x87C5101F, 'UTC_E_CANNOT_LOAD_SCENARIO_EDITOR_XML')
    UTC_E_ESCALATION_TIMED_OUT: Final[ThirtyTwoBits] = cached_hresult(0x87C51020, 'UTC_E_ESCALATION_TIMED_OUT')
    UTC_E_SETUP_TIMED_OUT: Final[ThirtyTwoBits] = cached_hresult(0x87C51021, 'UTC_E_SETUP_TIMED_OUT')
    UTC_E_TRIGGER_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x87C51022, 'UTC_E_TRIGGER_MISMATCH')
    UTC_E_TRIGGER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x87C51023, 'UTC_E_TRIGGER_NOT_FOUND')
    UTC_E_SIF_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x87C51024, 'UTC_E_SIF_NOT_SUPPORTED')
    UTC_E_DELAY_TERMINATED: Final[ThirtyTwoBits] = cached_hresult(0x87C51025, 'UTC_E_DELAY_TERMINATED')
    UTC_E_DEVICE_TICKET_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x87C51026, 'UTC_E_DEVICE_TICKET_ERROR')
    UTC_E_TRACE_BUFFER_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x87C51027, 'UTC_E_TRACE_BUFFER_LIMIT_EXCEEDED')
    UTC_E_API_RESULT_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x87C51028, 'UTC_E_API_RESULT_UNAVAILABLE')
    UTC_E_RPC_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x87C51029, 'UTC_E_RPC_TIMEOUT')
    UTC_E_RPC_WAIT_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x87C5102A, 'UTC_E_RPC_WAIT_FAILED')
    UTC_E_API_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x87C5102B, 'UTC_E_API_BUSY')
    UTC_E_TRACE_MIN_DURATION_REQUIREMENT_NOT_MET: Final[ThirtyTwoBits] = cached_hresult(0x87C5102C, 'UTC_E_TRACE_MIN_DURATION_REQUIREMENT_NOT_MET')
    UTC_E_EXCLUSIVITY_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x87C5102D, 'UTC_E_EXCLUSIVITY_NOT_AVAILABLE')
    UTC_E_GETFILE_FILE_PATH_NOT_APPROVED: Final[ThirtyTwoBits] = cached_hresult(0x87C5102E, 'UTC_E_GETFILE_FILE_PATH_NOT_APPROVED')
    UTC_E_ESCALATION_DIRECTORY_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x87C5102F, 'UTC_E_ESCALATION_DIRECTORY_ALREADY_EXISTS')
    UTC_E_TIME_TRIGGER_ON_START_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x87C51030, 'UTC_E_TIME_TRIGGER_ON_START_INVALID')
    UTC_E_TIME_TRIGGER_ONLY_VALID_ON_SINGLE_TRANSITION: Final[ThirtyTwoBits] = cached_hresult(0x87C51031, 'UTC_E_TIME_TRIGGER_ONLY_VALID_ON_SINGLE_TRANSITION')
    UTC_E_TIME_TRIGGER_INVALID_TIME_RANGE: Final[ThirtyTwoBits] = cached_hresult(0x87C51032, 'UTC_E_TIME_TRIGGER_INVALID_TIME_RANGE')
    UTC_E_MULTIPLE_TIME_TRIGGER_ON_SINGLE_STATE: Final[ThirtyTwoBits] = cached_hresult(0x87C51033, 'UTC_E_MULTIPLE_TIME_TRIGGER_ON_SINGLE_STATE')
    UTC_E_BINARY_MISSING: Final[ThirtyTwoBits] = cached_hresult(0x87C51034, 'UTC_E_BINARY_MISSING')
    UTC_E_FAILED_TO_RESOLVE_CONTAINER_ID: Final[ThirtyTwoBits] = cached_hresult(0x87C51036, 'UTC_E_FAILED_TO_RESOLVE_CONTAINER_ID')
    UTC_E_UNABLE_TO_RESOLVE_SESSION: Final[ThirtyTwoBits] = cached_hresult(0x87C51037, 'UTC_E_UNABLE_TO_RESOLVE_SESSION')
    UTC_E_THROTTLED: Final[ThirtyTwoBits] = cached_hresult(0x87C51038, 'UTC_E_THROTTLED')
    UTC_E_UNAPPROVED_SCRIPT: Final[ThirtyTwoBits] = cached_hresult(0x87C51039, 'UTC_E_UNAPPROVED_SCRIPT')
    UTC_E_SCRIPT_MISSING: Final[ThirtyTwoBits] = cached_hresult(0x87C5103A, 'UTC_E_SCRIPT_MISSING')
    UTC_E_SCENARIO_THROTTLED: Final[ThirtyTwoBits] = cached_hresult(0x87C5103B, 'UTC_E_SCENARIO_THROTTLED')
    UTC_E_API_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x87C5103C, 'UTC_E_API_NOT_SUPPORTED')
    UTC_E_GETFILE_EXTERNAL_PATH_NOT_APPROVED: Final[ThirtyTwoBits] = cached_hresult(0x87C5103D, 'UTC_E_GETFILE_EXTERNAL_PATH_NOT_APPROVED')
    UTC_E_TRY_GET_SCENARIO_TIMEOUT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x87C5103E, 'UTC_E_TRY_GET_SCENARIO_TIMEOUT_EXCEEDED')
    UTC_E_CERT_REV_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x87C5103F, 'UTC_E_CERT_REV_FAILED')
    UTC_E_FAILED_TO_START_NDISCAP: Final[ThirtyTwoBits] = cached_hresult(0x87C51040, 'UTC_E_FAILED_TO_START_NDISCAP')
    UTC_E_KERNELDUMP_LIMIT_REACHED: Final[ThirtyTwoBits] = cached_hresult(0x87C51041, 'UTC_E_KERNELDUMP_LIMIT_REACHED')
    UTC_E_MISSING_AGGREGATE_EVENT_TAG: Final[ThirtyTwoBits] = cached_hresult(0x87C51042, 'UTC_E_MISSING_AGGREGATE_EVENT_TAG')
    UTC_E_INVALID_AGGREGATION_STRUCT: Final[ThirtyTwoBits] = cached_hresult(0x87C51043, 'UTC_E_INVALID_AGGREGATION_STRUCT')
    UTC_E_ACTION_NOT_SUPPORTED_IN_DESTINATION: Final[ThirtyTwoBits] = cached_hresult(0x87C51044, 'UTC_E_ACTION_NOT_SUPPORTED_IN_DESTINATION')
    UTC_E_FILTER_MISSING_ATTRIBUTE: Final[ThirtyTwoBits] = cached_hresult(0x87C51045, 'UTC_E_FILTER_MISSING_ATTRIBUTE')
    UTC_E_FILTER_INVALID_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x87C51046, 'UTC_E_FILTER_INVALID_TYPE')
    UTC_E_FILTER_VARIABLE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x87C51047, 'UTC_E_FILTER_VARIABLE_NOT_FOUND')
    UTC_E_FILTER_FUNCTION_RESTRICTED: Final[ThirtyTwoBits] = cached_hresult(0x87C51048, 'UTC_E_FILTER_FUNCTION_RESTRICTED')
    UTC_E_FILTER_VERSION_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x87C51049, 'UTC_E_FILTER_VERSION_MISMATCH')
    UTC_E_FILTER_INVALID_FUNCTION: Final[ThirtyTwoBits] = cached_hresult(0x87C51050, 'UTC_E_FILTER_INVALID_FUNCTION')
    UTC_E_FILTER_INVALID_FUNCTION_PARAMS: Final[ThirtyTwoBits] = cached_hresult(0x87C51051, 'UTC_E_FILTER_INVALID_FUNCTION_PARAMS')
    UTC_E_FILTER_INVALID_COMMAND: Final[ThirtyTwoBits] = cached_hresult(0x87C51052, 'UTC_E_FILTER_INVALID_COMMAND')
    UTC_E_FILTER_ILLEGAL_EVAL: Final[ThirtyTwoBits] = cached_hresult(0x87C51053, 'UTC_E_FILTER_ILLEGAL_EVAL')
    UTC_E_TTTRACER_RETURNED_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x87C51054, 'UTC_E_TTTRACER_RETURNED_ERROR')
    UTC_E_AGENT_DIAGNOSTICS_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0x87C51055, 'UTC_E_AGENT_DIAGNOSTICS_TOO_LARGE')
    UTC_E_FAILED_TO_RECEIVE_AGENT_DIAGNOSTICS: Final[ThirtyTwoBits] = cached_hresult(0x87C51056, 'UTC_E_FAILED_TO_RECEIVE_AGENT_DIAGNOSTICS')
    UTC_E_SCENARIO_HAS_NO_ACTIONS: Final[ThirtyTwoBits] = cached_hresult(0x87C51057, 'UTC_E_SCENARIO_HAS_NO_ACTIONS')
    UTC_E_TTTRACER_STORAGE_FULL: Final[ThirtyTwoBits] = cached_hresult(0x87C51058, 'UTC_E_TTTRACER_STORAGE_FULL')
    UTC_E_INSUFFICIENT_SPACE_TO_START_TRACE: Final[ThirtyTwoBits] = cached_hresult(0x87C51059, 'UTC_E_INSUFFICIENT_SPACE_TO_START_TRACE')
    UTC_E_ESCALATION_CANCELLED_AT_SHUTDOWN: Final[ThirtyTwoBits] = cached_hresult(0x87C5105A, 'UTC_E_ESCALATION_CANCELLED_AT_SHUTDOWN')
    UTC_E_GETFILEINFOACTION_FILE_NOT_APPROVED: Final[ThirtyTwoBits] = cached_hresult(0x87C5105B, 'UTC_E_GETFILEINFOACTION_FILE_NOT_APPROVED')
    UTC_E_SETREGKEYACTION_TYPE_NOT_APPROVED: Final[ThirtyTwoBits] = cached_hresult(0x87C5105C, 'UTC_E_SETREGKEYACTION_TYPE_NOT_APPROVED')
    UTC_E_TRACE_THROTTLED: Final[ThirtyTwoBits] = cached_hresult(0x87C5105D, 'UTC_E_TRACE_THROTTLED')
    WEP_E_NOT_PROVISIONED_ON_ALL_VOLUMES: Final[ThirtyTwoBits] = cached_hresult(0x88010001, 'WEP_E_NOT_PROVISIONED_ON_ALL_VOLUMES')
    WEP_E_FIXED_DATA_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x88010002, 'WEP_E_FIXED_DATA_NOT_SUPPORTED')
    WEP_E_HARDWARE_NOT_COMPLIANT: Final[ThirtyTwoBits] = cached_hresult(0x88010003, 'WEP_E_HARDWARE_NOT_COMPLIANT')
    WEP_E_LOCK_NOT_CONFIGURED: Final[ThirtyTwoBits] = cached_hresult(0x88010004, 'WEP_E_LOCK_NOT_CONFIGURED')
    WEP_E_PROTECTION_SUSPENDED: Final[ThirtyTwoBits] = cached_hresult(0x88010005, 'WEP_E_PROTECTION_SUSPENDED')
    WEP_E_NO_LICENSE: Final[ThirtyTwoBits] = cached_hresult(0x88010006, 'WEP_E_NO_LICENSE')
    WEP_E_OS_NOT_PROTECTED: Final[ThirtyTwoBits] = cached_hresult(0x88010007, 'WEP_E_OS_NOT_PROTECTED')
    WEP_E_UNEXPECTED_FAIL: Final[ThirtyTwoBits] = cached_hresult(0x88010008, 'WEP_E_UNEXPECTED_FAIL')
    WEP_E_BUFFER_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0x88010009, 'WEP_E_BUFFER_TOO_LARGE')
    E_SYNCENGINE_FILE_SIZE_OVER_LIMIT: Final[ThirtyTwoBits] = cached_hresult(0x8802B001, 'E_SYNCENGINE_FILE_SIZE_OVER_LIMIT')
    E_SYNCENGINE_FILE_SIZE_EXCEEDS_REMAINING_QUOTA: Final[ThirtyTwoBits] = cached_hresult(0x8802B002, 'E_SYNCENGINE_FILE_SIZE_EXCEEDS_REMAINING_QUOTA')
    E_SYNCENGINE_UNSUPPORTED_FILE_NAME: Final[ThirtyTwoBits] = cached_hresult(0x8802B003, 'E_SYNCENGINE_UNSUPPORTED_FILE_NAME')
    E_SYNCENGINE_FOLDER_ITEM_COUNT_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8802B004, 'E_SYNCENGINE_FOLDER_ITEM_COUNT_LIMIT_EXCEEDED')
    E_SYNCENGINE_FILE_SYNC_PARTNER_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8802B005, 'E_SYNCENGINE_FILE_SYNC_PARTNER_ERROR')
    E_SYNCENGINE_SYNC_PAUSED_BY_SERVICE: Final[ThirtyTwoBits] = cached_hresult(0x8802B006, 'E_SYNCENGINE_SYNC_PAUSED_BY_SERVICE')
    E_SYNCENGINE_FILE_IDENTIFIER_UNKNOWN: Final[ThirtyTwoBits] = cached_hresult(0x8802C002, 'E_SYNCENGINE_FILE_IDENTIFIER_UNKNOWN')
    E_SYNCENGINE_SERVICE_AUTHENTICATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8802C003, 'E_SYNCENGINE_SERVICE_AUTHENTICATION_FAILED')
    E_SYNCENGINE_UNKNOWN_SERVICE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8802C004, 'E_SYNCENGINE_UNKNOWN_SERVICE_ERROR')
    E_SYNCENGINE_SERVICE_RETURNED_UNEXPECTED_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x8802C005, 'E_SYNCENGINE_SERVICE_RETURNED_UNEXPECTED_SIZE')
    E_SYNCENGINE_REQUEST_BLOCKED_BY_SERVICE: Final[ThirtyTwoBits] = cached_hresult(0x8802C006, 'E_SYNCENGINE_REQUEST_BLOCKED_BY_SERVICE')
    E_SYNCENGINE_REQUEST_BLOCKED_DUE_TO_CLIENT_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8802C007, 'E_SYNCENGINE_REQUEST_BLOCKED_DUE_TO_CLIENT_ERROR')
    E_SYNCENGINE_FOLDER_INACCESSIBLE: Final[ThirtyTwoBits] = cached_hresult(0x8802D001, 'E_SYNCENGINE_FOLDER_INACCESSIBLE')
    E_SYNCENGINE_UNSUPPORTED_FOLDER_NAME: Final[ThirtyTwoBits] = cached_hresult(0x8802D002, 'E_SYNCENGINE_UNSUPPORTED_FOLDER_NAME')
    E_SYNCENGINE_UNSUPPORTED_MARKET: Final[ThirtyTwoBits] = cached_hresult(0x8802D003, 'E_SYNCENGINE_UNSUPPORTED_MARKET')
    E_SYNCENGINE_PATH_LENGTH_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8802D004, 'E_SYNCENGINE_PATH_LENGTH_LIMIT_EXCEEDED')
    E_SYNCENGINE_REMOTE_PATH_LENGTH_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8802D005, 'E_SYNCENGINE_REMOTE_PATH_LENGTH_LIMIT_EXCEEDED')
    E_SYNCENGINE_CLIENT_UPDATE_NEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8802D006, 'E_SYNCENGINE_CLIENT_UPDATE_NEEDED')
    E_SYNCENGINE_PROXY_AUTHENTICATION_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8802D007, 'E_SYNCENGINE_PROXY_AUTHENTICATION_REQUIRED')
    E_SYNCENGINE_STORAGE_SERVICE_PROVISIONING_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8802D008, 'E_SYNCENGINE_STORAGE_SERVICE_PROVISIONING_FAILED')
    E_SYNCENGINE_UNSUPPORTED_REPARSE_POINT: Final[ThirtyTwoBits] = cached_hresult(0x8802D009, 'E_SYNCENGINE_UNSUPPORTED_REPARSE_POINT')
    E_SYNCENGINE_STORAGE_SERVICE_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x8802D00A, 'E_SYNCENGINE_STORAGE_SERVICE_BLOCKED')
    E_SYNCENGINE_FOLDER_IN_REDIRECTION: Final[ThirtyTwoBits] = cached_hresult(0x8802D00B, 'E_SYNCENGINE_FOLDER_IN_REDIRECTION')
    D3D10_ERROR_TOO_MANY_UNIQUE_STATE_OBJECTS: Final[ThirtyTwoBits] = cached_hresult(0x88790001, 'D3D10_ERROR_TOO_MANY_UNIQUE_STATE_OBJECTS')
    D3D10_ERROR_FILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x88790002, 'D3D10_ERROR_FILE_NOT_FOUND')
    DXGI_STATUS_OCCLUDED: Final[ThirtyTwoBits] = cached_hresult(0x087A0001, 'DXGI_STATUS_OCCLUDED')
    DXGI_ERROR_INVALID_CALL: Final[ThirtyTwoBits] = cached_hresult(0x887A0001, 'DXGI_ERROR_INVALID_CALL')
    DXGI_STATUS_CLIPPED: Final[ThirtyTwoBits] = cached_hresult(0x087A0002, 'DXGI_STATUS_CLIPPED')
    DXGI_ERROR_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x887A0002, 'DXGI_ERROR_NOT_FOUND')
    DXGI_ERROR_MORE_DATA: Final[ThirtyTwoBits] = cached_hresult(0x887A0003, 'DXGI_ERROR_MORE_DATA')
    DXGI_STATUS_NO_REDIRECTION: Final[ThirtyTwoBits] = cached_hresult(0x087A0004, 'DXGI_STATUS_NO_REDIRECTION')
    DXGI_ERROR_UNSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x887A0004, 'DXGI_ERROR_UNSUPPORTED')
    DXGI_STATUS_NO_DESKTOP_ACCESS: Final[ThirtyTwoBits] = cached_hresult(0x087A0005, 'DXGI_STATUS_NO_DESKTOP_ACCESS')
    DXGI_ERROR_DEVICE_REMOVED: Final[ThirtyTwoBits] = cached_hresult(0x887A0005, 'DXGI_ERROR_DEVICE_REMOVED')
    DXGI_STATUS_GRAPHICS_VIDPN_SOURCE_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x087A0006, 'DXGI_STATUS_GRAPHICS_VIDPN_SOURCE_IN_USE')
    DXGI_ERROR_DEVICE_HUNG: Final[ThirtyTwoBits] = cached_hresult(0x887A0006, 'DXGI_ERROR_DEVICE_HUNG')
    DXGI_STATUS_MODE_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0x087A0007, 'DXGI_STATUS_MODE_CHANGED')
    DXGI_ERROR_DEVICE_RESET: Final[ThirtyTwoBits] = cached_hresult(0x887A0007, 'DXGI_ERROR_DEVICE_RESET')
    DXGI_STATUS_MODE_CHANGE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x087A0008, 'DXGI_STATUS_MODE_CHANGE_IN_PROGRESS')
    DXGI_STATUS_UNOCCLUDED: Final[ThirtyTwoBits] = cached_hresult(0x087A0009, 'DXGI_STATUS_UNOCCLUDED')
    DXGI_STATUS_DDA_WAS_STILL_DRAWING: Final[ThirtyTwoBits] = cached_hresult(0x087A000A, 'DXGI_STATUS_DDA_WAS_STILL_DRAWING')
    DXGI_ERROR_WAS_STILL_DRAWING: Final[ThirtyTwoBits] = cached_hresult(0x887A000A, 'DXGI_ERROR_WAS_STILL_DRAWING')
    DXGI_ERROR_FRAME_STATISTICS_DISJOINT: Final[ThirtyTwoBits] = cached_hresult(0x887A000B, 'DXGI_ERROR_FRAME_STATISTICS_DISJOINT')
    DXGI_ERROR_GRAPHICS_VIDPN_SOURCE_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x887A000C, 'DXGI_ERROR_GRAPHICS_VIDPN_SOURCE_IN_USE')
    DXGI_ERROR_DRIVER_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x887A0020, 'DXGI_ERROR_DRIVER_INTERNAL_ERROR')
    DXGI_ERROR_NONEXCLUSIVE: Final[ThirtyTwoBits] = cached_hresult(0x887A0021, 'DXGI_ERROR_NONEXCLUSIVE')
    DXGI_ERROR_NOT_CURRENTLY_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x887A0022, 'DXGI_ERROR_NOT_CURRENTLY_AVAILABLE')
    DXGI_ERROR_REMOTE_CLIENT_DISCONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x887A0023, 'DXGI_ERROR_REMOTE_CLIENT_DISCONNECTED')
    DXGI_ERROR_REMOTE_OUTOFMEMORY: Final[ThirtyTwoBits] = cached_hresult(0x887A0024, 'DXGI_ERROR_REMOTE_OUTOFMEMORY')
    DXGI_ERROR_MODE_CHANGE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x887A0025, 'DXGI_ERROR_MODE_CHANGE_IN_PROGRESS')
    DXGI_ERROR_ACCESS_LOST: Final[ThirtyTwoBits] = cached_hresult(0x887A0026, 'DXGI_ERROR_ACCESS_LOST')
    DXGI_ERROR_WAIT_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x887A0027, 'DXGI_ERROR_WAIT_TIMEOUT')
    DXGI_ERROR_SESSION_DISCONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x887A0028, 'DXGI_ERROR_SESSION_DISCONNECTED')
    DXGI_ERROR_RESTRICT_TO_OUTPUT_STALE: Final[ThirtyTwoBits] = cached_hresult(0x887A0029, 'DXGI_ERROR_RESTRICT_TO_OUTPUT_STALE')
    DXGI_ERROR_CANNOT_PROTECT_CONTENT: Final[ThirtyTwoBits] = cached_hresult(0x887A002A, 'DXGI_ERROR_CANNOT_PROTECT_CONTENT')
    DXGI_ERROR_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x887A002B, 'DXGI_ERROR_ACCESS_DENIED')
    DXGI_ERROR_NAME_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x887A002C, 'DXGI_ERROR_NAME_ALREADY_EXISTS')
    DXGI_ERROR_SDK_COMPONENT_MISSING: Final[ThirtyTwoBits] = cached_hresult(0x887A002D, 'DXGI_ERROR_SDK_COMPONENT_MISSING')
    DXGI_ERROR_NOT_CURRENT: Final[ThirtyTwoBits] = cached_hresult(0x887A002E, 'DXGI_ERROR_NOT_CURRENT')
    DXGI_STATUS_PRESENT_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x087A002F, 'DXGI_STATUS_PRESENT_REQUIRED')
    DXGI_ERROR_HW_PROTECTION_OUTOFMEMORY: Final[ThirtyTwoBits] = cached_hresult(0x887A0030, 'DXGI_ERROR_HW_PROTECTION_OUTOFMEMORY')
    DXGI_ERROR_DYNAMIC_CODE_POLICY_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x887A0031, 'DXGI_ERROR_DYNAMIC_CODE_POLICY_VIOLATION')
    DXGI_ERROR_NON_COMPOSITED_UI: Final[ThirtyTwoBits] = cached_hresult(0x887A0032, 'DXGI_ERROR_NON_COMPOSITED_UI')
    DXGI_ERROR_CACHE_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x887A0033, 'DXGI_ERROR_CACHE_CORRUPT')
    DXGI_ERROR_CACHE_FULL: Final[ThirtyTwoBits] = cached_hresult(0x887A0034, 'DXGI_ERROR_CACHE_FULL')
    DXGI_ERROR_CACHE_HASH_COLLISION: Final[ThirtyTwoBits] = cached_hresult(0x887A0035, 'DXGI_ERROR_CACHE_HASH_COLLISION')
    DXGI_ERROR_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x887A0036, 'DXGI_ERROR_ALREADY_EXISTS')
    DXGI_ERROR_MPO_UNPINNED: Final[ThirtyTwoBits] = cached_hresult(0x887A0064, 'DXGI_ERROR_MPO_UNPINNED')
    DXGI_DDI_ERR_WASSTILLDRAWING: Final[ThirtyTwoBits] = cached_hresult(0x887B0001, 'DXGI_DDI_ERR_WASSTILLDRAWING')
    DXGI_DDI_ERR_UNSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x887B0002, 'DXGI_DDI_ERR_UNSUPPORTED')
    DXGI_DDI_ERR_NONEXCLUSIVE: Final[ThirtyTwoBits] = cached_hresult(0x887B0003, 'DXGI_DDI_ERR_NONEXCLUSIVE')
    D3D11_ERROR_TOO_MANY_UNIQUE_STATE_OBJECTS: Final[ThirtyTwoBits] = cached_hresult(0x887C0001, 'D3D11_ERROR_TOO_MANY_UNIQUE_STATE_OBJECTS')
    D3D11_ERROR_FILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x887C0002, 'D3D11_ERROR_FILE_NOT_FOUND')
    D3D11_ERROR_TOO_MANY_UNIQUE_VIEW_OBJECTS: Final[ThirtyTwoBits] = cached_hresult(0x887C0003, 'D3D11_ERROR_TOO_MANY_UNIQUE_VIEW_OBJECTS')
    D3D11_ERROR_DEFERRED_CONTEXT_MAP_WITHOUT_INITIAL_DISCARD: Final[ThirtyTwoBits] = cached_hresult(0x887C0004, 'D3D11_ERROR_DEFERRED_CONTEXT_MAP_WITHOUT_INITIAL_DISCARD')
    D3D12_ERROR_ADAPTER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x887E0001, 'D3D12_ERROR_ADAPTER_NOT_FOUND')
    D3D12_ERROR_DRIVER_VERSION_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x887E0002, 'D3D12_ERROR_DRIVER_VERSION_MISMATCH')
    D3D12_ERROR_INVALID_REDIST: Final[ThirtyTwoBits] = cached_hresult(0x887E0003, 'D3D12_ERROR_INVALID_REDIST')
    DXCORE_ERROR_EVENT_NOT_UNREGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x88800001, 'DXCORE_ERROR_EVENT_NOT_UNREGISTERED')
    PRESENTATION_ERROR_LOST: Final[ThirtyTwoBits] = cached_hresult(0x88810001, 'PRESENTATION_ERROR_LOST')
    WINML_ERR_INVALID_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0x88900001, 'WINML_ERR_INVALID_DEVICE')
    WINML_ERR_INVALID_BINDING: Final[ThirtyTwoBits] = cached_hresult(0x88900002, 'WINML_ERR_INVALID_BINDING')
    WINML_ERR_VALUE_NOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x88900003, 'WINML_ERR_VALUE_NOTFOUND')
    WINML_ERR_SIZE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x88900004, 'WINML_ERR_SIZE_MISMATCH')
    MILERR_OBJECTBUSY: Final[ThirtyTwoBits] = cached_hresult(0x88980001, 'MILERR_OBJECTBUSY')
    MILERR_INSUFFICIENTBUFFER: Final[ThirtyTwoBits] = cached_hresult(0x88980002, 'MILERR_INSUFFICIENTBUFFER')
    MILERR_WIN32ERROR: Final[ThirtyTwoBits] = cached_hresult(0x88980003, 'MILERR_WIN32ERROR')
    MILERR_SCANNER_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x88980004, 'MILERR_SCANNER_FAILED')
    MILERR_SCREENACCESSDENIED: Final[ThirtyTwoBits] = cached_hresult(0x88980005, 'MILERR_SCREENACCESSDENIED')
    MILERR_DISPLAYSTATEINVALID: Final[ThirtyTwoBits] = cached_hresult(0x88980006, 'MILERR_DISPLAYSTATEINVALID')
    MILERR_NONINVERTIBLEMATRIX: Final[ThirtyTwoBits] = cached_hresult(0x88980007, 'MILERR_NONINVERTIBLEMATRIX')
    MILERR_ZEROVECTOR: Final[ThirtyTwoBits] = cached_hresult(0x88980008, 'MILERR_ZEROVECTOR')
    MILERR_TERMINATED: Final[ThirtyTwoBits] = cached_hresult(0x88980009, 'MILERR_TERMINATED')
    MILERR_BADNUMBER: Final[ThirtyTwoBits] = cached_hresult(0x8898000A, 'MILERR_BADNUMBER')
    MILERR_INTERNALERROR: Final[ThirtyTwoBits] = cached_hresult(0x88980080, 'MILERR_INTERNALERROR')
    MILERR_DISPLAYFORMATNOTSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x88980084, 'MILERR_DISPLAYFORMATNOTSUPPORTED')
    MILERR_INVALIDCALL: Final[ThirtyTwoBits] = cached_hresult(0x88980085, 'MILERR_INVALIDCALL')
    MILERR_ALREADYLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x88980086, 'MILERR_ALREADYLOCKED')
    MILERR_NOTLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x88980087, 'MILERR_NOTLOCKED')
    MILERR_DEVICECANNOTRENDERTEXT: Final[ThirtyTwoBits] = cached_hresult(0x88980088, 'MILERR_DEVICECANNOTRENDERTEXT')
    MILERR_GLYPHBITMAPMISSED: Final[ThirtyTwoBits] = cached_hresult(0x88980089, 'MILERR_GLYPHBITMAPMISSED')
    MILERR_MALFORMEDGLYPHCACHE: Final[ThirtyTwoBits] = cached_hresult(0x8898008A, 'MILERR_MALFORMEDGLYPHCACHE')
    MILERR_GENERIC_IGNORE: Final[ThirtyTwoBits] = cached_hresult(0x8898008B, 'MILERR_GENERIC_IGNORE')
    MILERR_MALFORMED_GUIDELINE_DATA: Final[ThirtyTwoBits] = cached_hresult(0x8898008C, 'MILERR_MALFORMED_GUIDELINE_DATA')
    MILERR_NO_HARDWARE_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0x8898008D, 'MILERR_NO_HARDWARE_DEVICE')
    MILERR_NEED_RECREATE_AND_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0x8898008E, 'MILERR_NEED_RECREATE_AND_PRESENT')
    MILERR_ALREADY_INITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0x8898008F, 'MILERR_ALREADY_INITIALIZED')
    MILERR_MISMATCHED_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x88980090, 'MILERR_MISMATCHED_SIZE')
    MILERR_NO_REDIRECTION_SURFACE_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x88980091, 'MILERR_NO_REDIRECTION_SURFACE_AVAILABLE')
    MILERR_REMOTING_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x88980092, 'MILERR_REMOTING_NOT_SUPPORTED')
    MILERR_QUEUED_PRESENT_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x88980093, 'MILERR_QUEUED_PRESENT_NOT_SUPPORTED')
    MILERR_NOT_QUEUING_PRESENTS: Final[ThirtyTwoBits] = cached_hresult(0x88980094, 'MILERR_NOT_QUEUING_PRESENTS')
    MILERR_NO_REDIRECTION_SURFACE_RETRY_LATER: Final[ThirtyTwoBits] = cached_hresult(0x88980095, 'MILERR_NO_REDIRECTION_SURFACE_RETRY_LATER')
    MILERR_TOOMANYSHADERELEMNTS: Final[ThirtyTwoBits] = cached_hresult(0x88980096, 'MILERR_TOOMANYSHADERELEMNTS')
    MILERR_MROW_READLOCK_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x88980097, 'MILERR_MROW_READLOCK_FAILED')
    MILERR_MROW_UPDATE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x88980098, 'MILERR_MROW_UPDATE_FAILED')
    MILERR_SHADER_COMPILE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x88980099, 'MILERR_SHADER_COMPILE_FAILED')
    MILERR_MAX_TEXTURE_SIZE_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8898009A, 'MILERR_MAX_TEXTURE_SIZE_EXCEEDED')
    MILERR_QPC_TIME_WENT_BACKWARD: Final[ThirtyTwoBits] = cached_hresult(0x8898009B, 'MILERR_QPC_TIME_WENT_BACKWARD')
    MILERR_DXGI_ENUMERATION_OUT_OF_SYNC: Final[ThirtyTwoBits] = cached_hresult(0x8898009D, 'MILERR_DXGI_ENUMERATION_OUT_OF_SYNC')
    MILERR_ADAPTER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8898009E, 'MILERR_ADAPTER_NOT_FOUND')
    MILERR_COLORSPACE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8898009F, 'MILERR_COLORSPACE_NOT_SUPPORTED')
    MILERR_PREFILTER_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x889800A0, 'MILERR_PREFILTER_NOT_SUPPORTED')
    MILERR_DISPLAYID_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x889800A1, 'MILERR_DISPLAYID_ACCESS_DENIED')
    UCEERR_INVALIDPACKETHEADER: Final[ThirtyTwoBits] = cached_hresult(0x88980400, 'UCEERR_INVALIDPACKETHEADER')
    UCEERR_UNKNOWNPACKET: Final[ThirtyTwoBits] = cached_hresult(0x88980401, 'UCEERR_UNKNOWNPACKET')
    UCEERR_ILLEGALPACKET: Final[ThirtyTwoBits] = cached_hresult(0x88980402, 'UCEERR_ILLEGALPACKET')
    UCEERR_MALFORMEDPACKET: Final[ThirtyTwoBits] = cached_hresult(0x88980403, 'UCEERR_MALFORMEDPACKET')
    UCEERR_ILLEGALHANDLE: Final[ThirtyTwoBits] = cached_hresult(0x88980404, 'UCEERR_ILLEGALHANDLE')
    UCEERR_HANDLELOOKUPFAILED: Final[ThirtyTwoBits] = cached_hresult(0x88980405, 'UCEERR_HANDLELOOKUPFAILED')
    UCEERR_RENDERTHREADFAILURE: Final[ThirtyTwoBits] = cached_hresult(0x88980406, 'UCEERR_RENDERTHREADFAILURE')
    UCEERR_CTXSTACKFRSTTARGETNULL: Final[ThirtyTwoBits] = cached_hresult(0x88980407, 'UCEERR_CTXSTACKFRSTTARGETNULL')
    UCEERR_CONNECTIONIDLOOKUPFAILED: Final[ThirtyTwoBits] = cached_hresult(0x88980408, 'UCEERR_CONNECTIONIDLOOKUPFAILED')
    UCEERR_BLOCKSFULL: Final[ThirtyTwoBits] = cached_hresult(0x88980409, 'UCEERR_BLOCKSFULL')
    UCEERR_MEMORYFAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8898040A, 'UCEERR_MEMORYFAILURE')
    UCEERR_PACKETRECORDOUTOFRANGE: Final[ThirtyTwoBits] = cached_hresult(0x8898040B, 'UCEERR_PACKETRECORDOUTOFRANGE')
    UCEERR_ILLEGALRECORDTYPE: Final[ThirtyTwoBits] = cached_hresult(0x8898040C, 'UCEERR_ILLEGALRECORDTYPE')
    UCEERR_OUTOFHANDLES: Final[ThirtyTwoBits] = cached_hresult(0x8898040D, 'UCEERR_OUTOFHANDLES')
    UCEERR_UNCHANGABLE_UPDATE_ATTEMPTED: Final[ThirtyTwoBits] = cached_hresult(0x8898040E, 'UCEERR_UNCHANGABLE_UPDATE_ATTEMPTED')
    UCEERR_NO_MULTIPLE_WORKER_THREADS: Final[ThirtyTwoBits] = cached_hresult(0x8898040F, 'UCEERR_NO_MULTIPLE_WORKER_THREADS')
    UCEERR_REMOTINGNOTSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x88980410, 'UCEERR_REMOTINGNOTSUPPORTED')
    UCEERR_MISSINGENDCOMMAND: Final[ThirtyTwoBits] = cached_hresult(0x88980411, 'UCEERR_MISSINGENDCOMMAND')
    UCEERR_MISSINGBEGINCOMMAND: Final[ThirtyTwoBits] = cached_hresult(0x88980412, 'UCEERR_MISSINGBEGINCOMMAND')
    UCEERR_CHANNELSYNCTIMEDOUT: Final[ThirtyTwoBits] = cached_hresult(0x88980413, 'UCEERR_CHANNELSYNCTIMEDOUT')
    UCEERR_CHANNELSYNCABANDONED: Final[ThirtyTwoBits] = cached_hresult(0x88980414, 'UCEERR_CHANNELSYNCABANDONED')
    UCEERR_UNSUPPORTEDTRANSPORTVERSION: Final[ThirtyTwoBits] = cached_hresult(0x88980415, 'UCEERR_UNSUPPORTEDTRANSPORTVERSION')
    UCEERR_TRANSPORTUNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x88980416, 'UCEERR_TRANSPORTUNAVAILABLE')
    UCEERR_FEEDBACK_UNSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x88980417, 'UCEERR_FEEDBACK_UNSUPPORTED')
    UCEERR_COMMANDTRANSPORTDENIED: Final[ThirtyTwoBits] = cached_hresult(0x88980418, 'UCEERR_COMMANDTRANSPORTDENIED')
    UCEERR_GRAPHICSSTREAMUNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x88980419, 'UCEERR_GRAPHICSSTREAMUNAVAILABLE')
    UCEERR_GRAPHICSSTREAMALREADYOPEN: Final[ThirtyTwoBits] = cached_hresult(0x88980420, 'UCEERR_GRAPHICSSTREAMALREADYOPEN')
    UCEERR_TRANSPORTDISCONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x88980421, 'UCEERR_TRANSPORTDISCONNECTED')
    UCEERR_TRANSPORTOVERLOADED: Final[ThirtyTwoBits] = cached_hresult(0x88980422, 'UCEERR_TRANSPORTOVERLOADED')
    UCEERR_PARTITION_ZOMBIED: Final[ThirtyTwoBits] = cached_hresult(0x88980423, 'UCEERR_PARTITION_ZOMBIED')
    MILAVERR_NOCLOCK: Final[ThirtyTwoBits] = cached_hresult(0x88980500, 'MILAVERR_NOCLOCK')
    MILAVERR_NOMEDIATYPE: Final[ThirtyTwoBits] = cached_hresult(0x88980501, 'MILAVERR_NOMEDIATYPE')
    MILAVERR_NOVIDEOMIXER: Final[ThirtyTwoBits] = cached_hresult(0x88980502, 'MILAVERR_NOVIDEOMIXER')
    MILAVERR_NOVIDEOPRESENTER: Final[ThirtyTwoBits] = cached_hresult(0x88980503, 'MILAVERR_NOVIDEOPRESENTER')
    MILAVERR_NOREADYFRAMES: Final[ThirtyTwoBits] = cached_hresult(0x88980504, 'MILAVERR_NOREADYFRAMES')
    MILAVERR_MODULENOTLOADED: Final[ThirtyTwoBits] = cached_hresult(0x88980505, 'MILAVERR_MODULENOTLOADED')
    MILAVERR_WMPFACTORYNOTREGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x88980506, 'MILAVERR_WMPFACTORYNOTREGISTERED')
    MILAVERR_INVALIDWMPVERSION: Final[ThirtyTwoBits] = cached_hresult(0x88980507, 'MILAVERR_INVALIDWMPVERSION')
    MILAVERR_INSUFFICIENTVIDEORESOURCES: Final[ThirtyTwoBits] = cached_hresult(0x88980508, 'MILAVERR_INSUFFICIENTVIDEORESOURCES')
    MILAVERR_VIDEOACCELERATIONNOTAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x88980509, 'MILAVERR_VIDEOACCELERATIONNOTAVAILABLE')
    MILAVERR_REQUESTEDTEXTURETOOBIG: Final[ThirtyTwoBits] = cached_hresult(0x8898050A, 'MILAVERR_REQUESTEDTEXTURETOOBIG')
    MILAVERR_SEEKFAILED: Final[ThirtyTwoBits] = cached_hresult(0x8898050B, 'MILAVERR_SEEKFAILED')
    MILAVERR_UNEXPECTEDWMPFAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8898050C, 'MILAVERR_UNEXPECTEDWMPFAILURE')
    MILAVERR_MEDIAPLAYERCLOSED: Final[ThirtyTwoBits] = cached_hresult(0x8898050D, 'MILAVERR_MEDIAPLAYERCLOSED')
    MILAVERR_UNKNOWNHARDWAREERROR: Final[ThirtyTwoBits] = cached_hresult(0x8898050E, 'MILAVERR_UNKNOWNHARDWAREERROR')
    MILEFFECTSERR_UNKNOWNPROPERTY: Final[ThirtyTwoBits] = cached_hresult(0x8898060E, 'MILEFFECTSERR_UNKNOWNPROPERTY')
    MILEFFECTSERR_EFFECTNOTPARTOFGROUP: Final[ThirtyTwoBits] = cached_hresult(0x8898060F, 'MILEFFECTSERR_EFFECTNOTPARTOFGROUP')
    MILEFFECTSERR_NOINPUTSOURCEATTACHED: Final[ThirtyTwoBits] = cached_hresult(0x88980610, 'MILEFFECTSERR_NOINPUTSOURCEATTACHED')
    MILEFFECTSERR_CONNECTORNOTCONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x88980611, 'MILEFFECTSERR_CONNECTORNOTCONNECTED')
    MILEFFECTSERR_CONNECTORNOTASSOCIATEDWITHEFFECT: Final[ThirtyTwoBits] = cached_hresult(0x88980612, 'MILEFFECTSERR_CONNECTORNOTASSOCIATEDWITHEFFECT')
    MILEFFECTSERR_RESERVED: Final[ThirtyTwoBits] = cached_hresult(0x88980613, 'MILEFFECTSERR_RESERVED')
    MILEFFECTSERR_CYCLEDETECTED: Final[ThirtyTwoBits] = cached_hresult(0x88980614, 'MILEFFECTSERR_CYCLEDETECTED')
    MILEFFECTSERR_EFFECTINMORETHANONEGRAPH: Final[ThirtyTwoBits] = cached_hresult(0x88980615, 'MILEFFECTSERR_EFFECTINMORETHANONEGRAPH')
    MILEFFECTSERR_EFFECTALREADYINAGRAPH: Final[ThirtyTwoBits] = cached_hresult(0x88980616, 'MILEFFECTSERR_EFFECTALREADYINAGRAPH')
    MILEFFECTSERR_EFFECTHASNOCHILDREN: Final[ThirtyTwoBits] = cached_hresult(0x88980617, 'MILEFFECTSERR_EFFECTHASNOCHILDREN')
    MILEFFECTSERR_ALREADYATTACHEDTOLISTENER: Final[ThirtyTwoBits] = cached_hresult(0x88980618, 'MILEFFECTSERR_ALREADYATTACHEDTOLISTENER')
    MILEFFECTSERR_NOTAFFINETRANSFORM: Final[ThirtyTwoBits] = cached_hresult(0x88980619, 'MILEFFECTSERR_NOTAFFINETRANSFORM')
    MILEFFECTSERR_EMPTYBOUNDS: Final[ThirtyTwoBits] = cached_hresult(0x8898061A, 'MILEFFECTSERR_EMPTYBOUNDS')
    MILEFFECTSERR_OUTPUTSIZETOOLARGE: Final[ThirtyTwoBits] = cached_hresult(0x8898061B, 'MILEFFECTSERR_OUTPUTSIZETOOLARGE')
    DWMERR_STATE_TRANSITION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x88980700, 'DWMERR_STATE_TRANSITION_FAILED')
    DWMERR_THEME_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x88980701, 'DWMERR_THEME_FAILED')
    DWMERR_CATASTROPHIC_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x88980702, 'DWMERR_CATASTROPHIC_FAILURE')
    DCOMPOSITION_ERROR_WINDOW_ALREADY_COMPOSED: Final[ThirtyTwoBits] = cached_hresult(0x88980800, 'DCOMPOSITION_ERROR_WINDOW_ALREADY_COMPOSED')
    DCOMPOSITION_ERROR_SURFACE_BEING_RENDERED: Final[ThirtyTwoBits] = cached_hresult(0x88980801, 'DCOMPOSITION_ERROR_SURFACE_BEING_RENDERED')
    DCOMPOSITION_ERROR_SURFACE_NOT_BEING_RENDERED: Final[ThirtyTwoBits] = cached_hresult(0x88980802, 'DCOMPOSITION_ERROR_SURFACE_NOT_BEING_RENDERED')
    WINCODEC_ERR_WRONGSTATE: Final[ThirtyTwoBits] = cached_hresult(0x88982F04, 'WINCODEC_ERR_WRONGSTATE')
    WINCODEC_ERR_VALUEOUTOFRANGE: Final[ThirtyTwoBits] = cached_hresult(0x88982F05, 'WINCODEC_ERR_VALUEOUTOFRANGE')
    WINCODEC_ERR_UNKNOWNIMAGEFORMAT: Final[ThirtyTwoBits] = cached_hresult(0x88982F07, 'WINCODEC_ERR_UNKNOWNIMAGEFORMAT')
    WINCODEC_ERR_UNSUPPORTEDVERSION: Final[ThirtyTwoBits] = cached_hresult(0x88982F0B, 'WINCODEC_ERR_UNSUPPORTEDVERSION')
    WINCODEC_ERR_NOTINITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0x88982F0C, 'WINCODEC_ERR_NOTINITIALIZED')
    WINCODEC_ERR_ALREADYLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x88982F0D, 'WINCODEC_ERR_ALREADYLOCKED')
    WINCODEC_ERR_PROPERTYNOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x88982F40, 'WINCODEC_ERR_PROPERTYNOTFOUND')
    WINCODEC_ERR_PROPERTYNOTSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x88982F41, 'WINCODEC_ERR_PROPERTYNOTSUPPORTED')
    WINCODEC_ERR_PROPERTYSIZE: Final[ThirtyTwoBits] = cached_hresult(0x88982F42, 'WINCODEC_ERR_PROPERTYSIZE')
    WINCODEC_ERR_CODECPRESENT: Final[ThirtyTwoBits] = cached_hresult(0x88982F43, 'WINCODEC_ERR_CODECPRESENT')
    WINCODEC_ERR_CODECNOTHUMBNAIL: Final[ThirtyTwoBits] = cached_hresult(0x88982F44, 'WINCODEC_ERR_CODECNOTHUMBNAIL')
    WINCODEC_ERR_PALETTEUNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x88982F45, 'WINCODEC_ERR_PALETTEUNAVAILABLE')
    WINCODEC_ERR_CODECTOOMANYSCANLINES: Final[ThirtyTwoBits] = cached_hresult(0x88982F46, 'WINCODEC_ERR_CODECTOOMANYSCANLINES')
    WINCODEC_ERR_INTERNALERROR: Final[ThirtyTwoBits] = cached_hresult(0x88982F48, 'WINCODEC_ERR_INTERNALERROR')
    WINCODEC_ERR_SOURCERECTDOESNOTMATCHDIMENSIONS: Final[ThirtyTwoBits] = cached_hresult(0x88982F49, 'WINCODEC_ERR_SOURCERECTDOESNOTMATCHDIMENSIONS')
    WINCODEC_ERR_COMPONENTNOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x88982F50, 'WINCODEC_ERR_COMPONENTNOTFOUND')
    WINCODEC_ERR_IMAGESIZEOUTOFRANGE: Final[ThirtyTwoBits] = cached_hresult(0x88982F51, 'WINCODEC_ERR_IMAGESIZEOUTOFRANGE')
    WINCODEC_ERR_TOOMUCHMETADATA: Final[ThirtyTwoBits] = cached_hresult(0x88982F52, 'WINCODEC_ERR_TOOMUCHMETADATA')
    WINCODEC_ERR_BADIMAGE: Final[ThirtyTwoBits] = cached_hresult(0x88982F60, 'WINCODEC_ERR_BADIMAGE')
    WINCODEC_ERR_BADHEADER: Final[ThirtyTwoBits] = cached_hresult(0x88982F61, 'WINCODEC_ERR_BADHEADER')
    WINCODEC_ERR_FRAMEMISSING: Final[ThirtyTwoBits] = cached_hresult(0x88982F62, 'WINCODEC_ERR_FRAMEMISSING')
    WINCODEC_ERR_BADMETADATAHEADER: Final[ThirtyTwoBits] = cached_hresult(0x88982F63, 'WINCODEC_ERR_BADMETADATAHEADER')
    WINCODEC_ERR_BADSTREAMDATA: Final[ThirtyTwoBits] = cached_hresult(0x88982F70, 'WINCODEC_ERR_BADSTREAMDATA')
    WINCODEC_ERR_STREAMWRITE: Final[ThirtyTwoBits] = cached_hresult(0x88982F71, 'WINCODEC_ERR_STREAMWRITE')
    WINCODEC_ERR_STREAMREAD: Final[ThirtyTwoBits] = cached_hresult(0x88982F72, 'WINCODEC_ERR_STREAMREAD')
    WINCODEC_ERR_STREAMNOTAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x88982F73, 'WINCODEC_ERR_STREAMNOTAVAILABLE')
    WINCODEC_ERR_UNSUPPORTEDPIXELFORMAT: Final[ThirtyTwoBits] = cached_hresult(0x88982F80, 'WINCODEC_ERR_UNSUPPORTEDPIXELFORMAT')
    WINCODEC_ERR_UNSUPPORTEDOPERATION: Final[ThirtyTwoBits] = cached_hresult(0x88982F81, 'WINCODEC_ERR_UNSUPPORTEDOPERATION')
    WINCODEC_ERR_INVALIDREGISTRATION: Final[ThirtyTwoBits] = cached_hresult(0x88982F8A, 'WINCODEC_ERR_INVALIDREGISTRATION')
    WINCODEC_ERR_COMPONENTINITIALIZEFAILURE: Final[ThirtyTwoBits] = cached_hresult(0x88982F8B, 'WINCODEC_ERR_COMPONENTINITIALIZEFAILURE')
    WINCODEC_ERR_INSUFFICIENTBUFFER: Final[ThirtyTwoBits] = cached_hresult(0x88982F8C, 'WINCODEC_ERR_INSUFFICIENTBUFFER')
    WINCODEC_ERR_DUPLICATEMETADATAPRESENT: Final[ThirtyTwoBits] = cached_hresult(0x88982F8D, 'WINCODEC_ERR_DUPLICATEMETADATAPRESENT')
    WINCODEC_ERR_PROPERTYUNEXPECTEDTYPE: Final[ThirtyTwoBits] = cached_hresult(0x88982F8E, 'WINCODEC_ERR_PROPERTYUNEXPECTEDTYPE')
    WINCODEC_ERR_UNEXPECTEDSIZE: Final[ThirtyTwoBits] = cached_hresult(0x88982F8F, 'WINCODEC_ERR_UNEXPECTEDSIZE')
    WINCODEC_ERR_INVALIDQUERYREQUEST: Final[ThirtyTwoBits] = cached_hresult(0x88982F90, 'WINCODEC_ERR_INVALIDQUERYREQUEST')
    WINCODEC_ERR_UNEXPECTEDMETADATATYPE: Final[ThirtyTwoBits] = cached_hresult(0x88982F91, 'WINCODEC_ERR_UNEXPECTEDMETADATATYPE')
    WINCODEC_ERR_REQUESTONLYVALIDATMETADATAROOT: Final[ThirtyTwoBits] = cached_hresult(0x88982F92, 'WINCODEC_ERR_REQUESTONLYVALIDATMETADATAROOT')
    WINCODEC_ERR_INVALIDQUERYCHARACTER: Final[ThirtyTwoBits] = cached_hresult(0x88982F93, 'WINCODEC_ERR_INVALIDQUERYCHARACTER')
    WINCODEC_ERR_WIN32ERROR: Final[ThirtyTwoBits] = cached_hresult(0x88982F94, 'WINCODEC_ERR_WIN32ERROR')
    WINCODEC_ERR_INVALIDPROGRESSIVELEVEL: Final[ThirtyTwoBits] = cached_hresult(0x88982F95, 'WINCODEC_ERR_INVALIDPROGRESSIVELEVEL')
    WINCODEC_ERR_INVALIDJPEGSCANINDEX: Final[ThirtyTwoBits] = cached_hresult(0x88982F96, 'WINCODEC_ERR_INVALIDJPEGSCANINDEX')
    DWRITE_E_FILEFORMAT: Final[ThirtyTwoBits] = cached_hresult(0x88985000, 'DWRITE_E_FILEFORMAT')
    DWRITE_E_UNEXPECTED: Final[ThirtyTwoBits] = cached_hresult(0x88985001, 'DWRITE_E_UNEXPECTED')
    DWRITE_E_NOFONT: Final[ThirtyTwoBits] = cached_hresult(0x88985002, 'DWRITE_E_NOFONT')
    DWRITE_E_FILENOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x88985003, 'DWRITE_E_FILENOTFOUND')
    DWRITE_E_FILEACCESS: Final[ThirtyTwoBits] = cached_hresult(0x88985004, 'DWRITE_E_FILEACCESS')
    DWRITE_E_FONTCOLLECTIONOBSOLETE: Final[ThirtyTwoBits] = cached_hresult(0x88985005, 'DWRITE_E_FONTCOLLECTIONOBSOLETE')
    DWRITE_E_ALREADYREGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x88985006, 'DWRITE_E_ALREADYREGISTERED')
    DWRITE_E_CACHEFORMAT: Final[ThirtyTwoBits] = cached_hresult(0x88985007, 'DWRITE_E_CACHEFORMAT')
    DWRITE_E_CACHEVERSION: Final[ThirtyTwoBits] = cached_hresult(0x88985008, 'DWRITE_E_CACHEVERSION')
    DWRITE_E_UNSUPPORTEDOPERATION: Final[ThirtyTwoBits] = cached_hresult(0x88985009, 'DWRITE_E_UNSUPPORTEDOPERATION')
    DWRITE_E_TEXTRENDERERINCOMPATIBLE: Final[ThirtyTwoBits] = cached_hresult(0x8898500A, 'DWRITE_E_TEXTRENDERERINCOMPATIBLE')
    DWRITE_E_FLOWDIRECTIONCONFLICTS: Final[ThirtyTwoBits] = cached_hresult(0x8898500B, 'DWRITE_E_FLOWDIRECTIONCONFLICTS')
    DWRITE_E_NOCOLOR: Final[ThirtyTwoBits] = cached_hresult(0x8898500C, 'DWRITE_E_NOCOLOR')
    DWRITE_E_REMOTEFONT: Final[ThirtyTwoBits] = cached_hresult(0x8898500D, 'DWRITE_E_REMOTEFONT')
    DWRITE_E_DOWNLOADCANCELLED: Final[ThirtyTwoBits] = cached_hresult(0x8898500E, 'DWRITE_E_DOWNLOADCANCELLED')
    DWRITE_E_DOWNLOADFAILED: Final[ThirtyTwoBits] = cached_hresult(0x8898500F, 'DWRITE_E_DOWNLOADFAILED')
    DWRITE_E_TOOMANYDOWNLOADS: Final[ThirtyTwoBits] = cached_hresult(0x88985010, 'DWRITE_E_TOOMANYDOWNLOADS')
    D2DERR_WRONG_STATE: Final[ThirtyTwoBits] = cached_hresult(0x88990001, 'D2DERR_WRONG_STATE')
    D2DERR_NOT_INITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0x88990002, 'D2DERR_NOT_INITIALIZED')
    D2DERR_UNSUPPORTED_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x88990003, 'D2DERR_UNSUPPORTED_OPERATION')
    D2DERR_SCANNER_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x88990004, 'D2DERR_SCANNER_FAILED')
    D2DERR_SCREEN_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x88990005, 'D2DERR_SCREEN_ACCESS_DENIED')
    D2DERR_DISPLAY_STATE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x88990006, 'D2DERR_DISPLAY_STATE_INVALID')
    D2DERR_ZERO_VECTOR: Final[ThirtyTwoBits] = cached_hresult(0x88990007, 'D2DERR_ZERO_VECTOR')
    D2DERR_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x88990008, 'D2DERR_INTERNAL_ERROR')
    D2DERR_DISPLAY_FORMAT_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x88990009, 'D2DERR_DISPLAY_FORMAT_NOT_SUPPORTED')
    D2DERR_INVALID_CALL: Final[ThirtyTwoBits] = cached_hresult(0x8899000A, 'D2DERR_INVALID_CALL')
    D2DERR_NO_HARDWARE_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0x8899000B, 'D2DERR_NO_HARDWARE_DEVICE')
    D2DERR_RECREATE_TARGET: Final[ThirtyTwoBits] = cached_hresult(0x8899000C, 'D2DERR_RECREATE_TARGET')
    D2DERR_TOO_MANY_SHADER_ELEMENTS: Final[ThirtyTwoBits] = cached_hresult(0x8899000D, 'D2DERR_TOO_MANY_SHADER_ELEMENTS')
    D2DERR_SHADER_COMPILE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8899000E, 'D2DERR_SHADER_COMPILE_FAILED')
    D2DERR_MAX_TEXTURE_SIZE_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8899000F, 'D2DERR_MAX_TEXTURE_SIZE_EXCEEDED')
    D2DERR_UNSUPPORTED_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x88990010, 'D2DERR_UNSUPPORTED_VERSION')
    D2DERR_BAD_NUMBER: Final[ThirtyTwoBits] = cached_hresult(0x88990011, 'D2DERR_BAD_NUMBER')
    D2DERR_WRONG_FACTORY: Final[ThirtyTwoBits] = cached_hresult(0x88990012, 'D2DERR_WRONG_FACTORY')
    D2DERR_LAYER_ALREADY_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x88990013, 'D2DERR_LAYER_ALREADY_IN_USE')
    D2DERR_POP_CALL_DID_NOT_MATCH_PUSH: Final[ThirtyTwoBits] = cached_hresult(0x88990014, 'D2DERR_POP_CALL_DID_NOT_MATCH_PUSH')
    D2DERR_WRONG_RESOURCE_DOMAIN: Final[ThirtyTwoBits] = cached_hresult(0x88990015, 'D2DERR_WRONG_RESOURCE_DOMAIN')
    D2DERR_PUSH_POP_UNBALANCED: Final[ThirtyTwoBits] = cached_hresult(0x88990016, 'D2DERR_PUSH_POP_UNBALANCED')
    D2DERR_RENDER_TARGET_HAS_LAYER_OR_CLIPRECT: Final[ThirtyTwoBits] = cached_hresult(0x88990017, 'D2DERR_RENDER_TARGET_HAS_LAYER_OR_CLIPRECT')
    D2DERR_INCOMPATIBLE_BRUSH_TYPES: Final[ThirtyTwoBits] = cached_hresult(0x88990018, 'D2DERR_INCOMPATIBLE_BRUSH_TYPES')
    D2DERR_WIN32_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x88990019, 'D2DERR_WIN32_ERROR')
    D2DERR_TARGET_NOT_GDI_COMPATIBLE: Final[ThirtyTwoBits] = cached_hresult(0x8899001A, 'D2DERR_TARGET_NOT_GDI_COMPATIBLE')
    D2DERR_TEXT_EFFECT_IS_WRONG_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x8899001B, 'D2DERR_TEXT_EFFECT_IS_WRONG_TYPE')
    D2DERR_TEXT_RENDERER_NOT_RELEASED: Final[ThirtyTwoBits] = cached_hresult(0x8899001C, 'D2DERR_TEXT_RENDERER_NOT_RELEASED')
    D2DERR_EXCEEDS_MAX_BITMAP_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x8899001D, 'D2DERR_EXCEEDS_MAX_BITMAP_SIZE')
    D2DERR_INVALID_GRAPH_CONFIGURATION: Final[ThirtyTwoBits] = cached_hresult(0x8899001E, 'D2DERR_INVALID_GRAPH_CONFIGURATION')
    D2DERR_INVALID_INTERNAL_GRAPH_CONFIGURATION: Final[ThirtyTwoBits] = cached_hresult(0x8899001F, 'D2DERR_INVALID_INTERNAL_GRAPH_CONFIGURATION')
    D2DERR_CYCLIC_GRAPH: Final[ThirtyTwoBits] = cached_hresult(0x88990020, 'D2DERR_CYCLIC_GRAPH')
    D2DERR_BITMAP_CANNOT_DRAW: Final[ThirtyTwoBits] = cached_hresult(0x88990021, 'D2DERR_BITMAP_CANNOT_DRAW')
    D2DERR_OUTSTANDING_BITMAP_REFERENCES: Final[ThirtyTwoBits] = cached_hresult(0x88990022, 'D2DERR_OUTSTANDING_BITMAP_REFERENCES')
    D2DERR_ORIGINAL_TARGET_NOT_BOUND: Final[ThirtyTwoBits] = cached_hresult(0x88990023, 'D2DERR_ORIGINAL_TARGET_NOT_BOUND')
    D2DERR_INVALID_TARGET: Final[ThirtyTwoBits] = cached_hresult(0x88990024, 'D2DERR_INVALID_TARGET')
    D2DERR_BITMAP_BOUND_AS_TARGET: Final[ThirtyTwoBits] = cached_hresult(0x88990025, 'D2DERR_BITMAP_BOUND_AS_TARGET')
    D2DERR_INSUFFICIENT_DEVICE_CAPABILITIES: Final[ThirtyTwoBits] = cached_hresult(0x88990026, 'D2DERR_INSUFFICIENT_DEVICE_CAPABILITIES')
    D2DERR_INTERMEDIATE_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0x88990027, 'D2DERR_INTERMEDIATE_TOO_LARGE')
    D2DERR_EFFECT_IS_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x88990028, 'D2DERR_EFFECT_IS_NOT_REGISTERED')
    D2DERR_INVALID_PROPERTY: Final[ThirtyTwoBits] = cached_hresult(0x88990029, 'D2DERR_INVALID_PROPERTY')
    D2DERR_NO_SUBPROPERTIES: Final[ThirtyTwoBits] = cached_hresult(0x8899002A, 'D2DERR_NO_SUBPROPERTIES')
    D2DERR_PRINT_JOB_CLOSED: Final[ThirtyTwoBits] = cached_hresult(0x8899002B, 'D2DERR_PRINT_JOB_CLOSED')
    D2DERR_PRINT_FORMAT_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8899002C, 'D2DERR_PRINT_FORMAT_NOT_SUPPORTED')
    D2DERR_TOO_MANY_TRANSFORM_INPUTS: Final[ThirtyTwoBits] = cached_hresult(0x8899002D, 'D2DERR_TOO_MANY_TRANSFORM_INPUTS')
    D2DERR_INVALID_GLYPH_IMAGE: Final[ThirtyTwoBits] = cached_hresult(0x8899002E, 'D2DERR_INVALID_GLYPH_IMAGE')
    ERROR_IO_PREEMPTED: Final[ThirtyTwoBits] = cached_hresult(0x89010001, 'ERROR_IO_PREEMPTED')
    JSCRIPT_E_CANTEXECUTE: Final[ThirtyTwoBits] = cached_hresult(0x89020001, 'JSCRIPT_E_CANTEXECUTE')

    # WIN32 Error Codes converted to HRESULT
    WIN32__ERROR_SUCCESS: Final[ThirtyTwoBits] = cached_hresult(0x00000000, 'HRESULT_FROM_WIN32(ERROR_SUCCESS)')
    WIN32__ERROR_INVALID_FUNCTION: Final[ThirtyTwoBits] = cached_hresult(0x80070001, 'HRESULT_FROM_WIN32(ERROR_INVALID_FUNCTION)')
    WIN32__ERROR_FILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80070002, 'HRESULT_FROM_WIN32(ERROR_FILE_NOT_FOUND)')
    WIN32__ERROR_PATH_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80070003, 'HRESULT_FROM_WIN32(ERROR_PATH_NOT_FOUND)')
    WIN32__ERROR_TOO_MANY_OPEN_FILES: Final[ThirtyTwoBits] = cached_hresult(0x80070004, 'HRESULT_FROM_WIN32(ERROR_TOO_MANY_OPEN_FILES)')
    WIN32__ERROR_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x80070005, 'HRESULT_FROM_WIN32(ERROR_ACCESS_DENIED)')
    WIN32__ERROR_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80070006, 'HRESULT_FROM_WIN32(ERROR_INVALID_HANDLE)')
    WIN32__ERROR_ARENA_TRASHED: Final[ThirtyTwoBits] = cached_hresult(0x80070007, 'HRESULT_FROM_WIN32(ERROR_ARENA_TRASHED)')
    WIN32__ERROR_NOT_ENOUGH_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0x80070008, 'HRESULT_FROM_WIN32(ERROR_NOT_ENOUGH_MEMORY)')
    WIN32__ERROR_INVALID_BLOCK: Final[ThirtyTwoBits] = cached_hresult(0x80070009, 'HRESULT_FROM_WIN32(ERROR_INVALID_BLOCK)')
    WIN32__ERROR_BAD_ENVIRONMENT: Final[ThirtyTwoBits] = cached_hresult(0x8007000A, 'HRESULT_FROM_WIN32(ERROR_BAD_ENVIRONMENT)')
    WIN32__ERROR_BAD_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0x8007000B, 'HRESULT_FROM_WIN32(ERROR_BAD_FORMAT)')
    WIN32__ERROR_INVALID_ACCESS: Final[ThirtyTwoBits] = cached_hresult(0x8007000C, 'HRESULT_FROM_WIN32(ERROR_INVALID_ACCESS)')
    WIN32__ERROR_INVALID_DATA: Final[ThirtyTwoBits] = cached_hresult(0x8007000D, 'HRESULT_FROM_WIN32(ERROR_INVALID_DATA)')
    WIN32__ERROR_OUTOFMEMORY: Final[ThirtyTwoBits] = cached_hresult(0x8007000E, 'HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY)')
    WIN32__ERROR_INVALID_DRIVE: Final[ThirtyTwoBits] = cached_hresult(0x8007000F, 'HRESULT_FROM_WIN32(ERROR_INVALID_DRIVE)')
    WIN32__ERROR_CURRENT_DIRECTORY: Final[ThirtyTwoBits] = cached_hresult(0x80070010, 'HRESULT_FROM_WIN32(ERROR_CURRENT_DIRECTORY)')
    WIN32__ERROR_NOT_SAME_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0x80070011, 'HRESULT_FROM_WIN32(ERROR_NOT_SAME_DEVICE)')
    WIN32__ERROR_NO_MORE_FILES: Final[ThirtyTwoBits] = cached_hresult(0x80070012, 'HRESULT_FROM_WIN32(ERROR_NO_MORE_FILES)')
    WIN32__ERROR_WRITE_PROTECT: Final[ThirtyTwoBits] = cached_hresult(0x80070013, 'HRESULT_FROM_WIN32(ERROR_WRITE_PROTECT)')
    WIN32__ERROR_BAD_UNIT: Final[ThirtyTwoBits] = cached_hresult(0x80070014, 'HRESULT_FROM_WIN32(ERROR_BAD_UNIT)')
    WIN32__ERROR_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0x80070015, 'HRESULT_FROM_WIN32(ERROR_NOT_READY)')
    WIN32__ERROR_BAD_COMMAND: Final[ThirtyTwoBits] = cached_hresult(0x80070016, 'HRESULT_FROM_WIN32(ERROR_BAD_COMMAND)')
    WIN32__ERROR_CRC: Final[ThirtyTwoBits] = cached_hresult(0x80070017, 'HRESULT_FROM_WIN32(ERROR_CRC)')
    WIN32__ERROR_BAD_LENGTH: Final[ThirtyTwoBits] = cached_hresult(0x80070018, 'HRESULT_FROM_WIN32(ERROR_BAD_LENGTH)')
    WIN32__ERROR_SEEK: Final[ThirtyTwoBits] = cached_hresult(0x80070019, 'HRESULT_FROM_WIN32(ERROR_SEEK)')
    WIN32__ERROR_NOT_DOS_DISK: Final[ThirtyTwoBits] = cached_hresult(0x8007001A, 'HRESULT_FROM_WIN32(ERROR_NOT_DOS_DISK)')
    WIN32__ERROR_SECTOR_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8007001B, 'HRESULT_FROM_WIN32(ERROR_SECTOR_NOT_FOUND)')
    WIN32__ERROR_OUT_OF_PAPER: Final[ThirtyTwoBits] = cached_hresult(0x8007001C, 'HRESULT_FROM_WIN32(ERROR_OUT_OF_PAPER)')
    WIN32__ERROR_WRITE_FAULT: Final[ThirtyTwoBits] = cached_hresult(0x8007001D, 'HRESULT_FROM_WIN32(ERROR_WRITE_FAULT)')
    WIN32__ERROR_READ_FAULT: Final[ThirtyTwoBits] = cached_hresult(0x8007001E, 'HRESULT_FROM_WIN32(ERROR_READ_FAULT)')
    WIN32__ERROR_GEN_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8007001F, 'HRESULT_FROM_WIN32(ERROR_GEN_FAILURE)')
    WIN32__ERROR_SHARING_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x80070020, 'HRESULT_FROM_WIN32(ERROR_SHARING_VIOLATION)')
    WIN32__ERROR_LOCK_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x80070021, 'HRESULT_FROM_WIN32(ERROR_LOCK_VIOLATION)')
    WIN32__ERROR_WRONG_DISK: Final[ThirtyTwoBits] = cached_hresult(0x80070022, 'HRESULT_FROM_WIN32(ERROR_WRONG_DISK)')
    WIN32__ERROR_SHARING_BUFFER_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80070024, 'HRESULT_FROM_WIN32(ERROR_SHARING_BUFFER_EXCEEDED)')
    WIN32__ERROR_HANDLE_EOF: Final[ThirtyTwoBits] = cached_hresult(0x80070026, 'HRESULT_FROM_WIN32(ERROR_HANDLE_EOF)')
    WIN32__ERROR_HANDLE_DISK_FULL: Final[ThirtyTwoBits] = cached_hresult(0x80070027, 'HRESULT_FROM_WIN32(ERROR_HANDLE_DISK_FULL)')
    WIN32__ERROR_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80070032, 'HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED)')
    WIN32__ERROR_REM_NOT_LIST: Final[ThirtyTwoBits] = cached_hresult(0x80070033, 'HRESULT_FROM_WIN32(ERROR_REM_NOT_LIST)')
    WIN32__ERROR_DUP_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80070034, 'HRESULT_FROM_WIN32(ERROR_DUP_NAME)')
    WIN32__ERROR_BAD_NETPATH: Final[ThirtyTwoBits] = cached_hresult(0x80070035, 'HRESULT_FROM_WIN32(ERROR_BAD_NETPATH)')
    WIN32__ERROR_NETWORK_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x80070036, 'HRESULT_FROM_WIN32(ERROR_NETWORK_BUSY)')
    WIN32__ERROR_DEV_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x80070037, 'HRESULT_FROM_WIN32(ERROR_DEV_NOT_EXIST)')
    WIN32__ERROR_TOO_MANY_CMDS: Final[ThirtyTwoBits] = cached_hresult(0x80070038, 'HRESULT_FROM_WIN32(ERROR_TOO_MANY_CMDS)')
    WIN32__ERROR_ADAP_HDW_ERR: Final[ThirtyTwoBits] = cached_hresult(0x80070039, 'HRESULT_FROM_WIN32(ERROR_ADAP_HDW_ERR)')
    WIN32__ERROR_BAD_NET_RESP: Final[ThirtyTwoBits] = cached_hresult(0x8007003A, 'HRESULT_FROM_WIN32(ERROR_BAD_NET_RESP)')
    WIN32__ERROR_UNEXP_NET_ERR: Final[ThirtyTwoBits] = cached_hresult(0x8007003B, 'HRESULT_FROM_WIN32(ERROR_UNEXP_NET_ERR)')
    WIN32__ERROR_BAD_REM_ADAP: Final[ThirtyTwoBits] = cached_hresult(0x8007003C, 'HRESULT_FROM_WIN32(ERROR_BAD_REM_ADAP)')
    WIN32__ERROR_PRINTQ_FULL: Final[ThirtyTwoBits] = cached_hresult(0x8007003D, 'HRESULT_FROM_WIN32(ERROR_PRINTQ_FULL)')
    WIN32__ERROR_NO_SPOOL_SPACE: Final[ThirtyTwoBits] = cached_hresult(0x8007003E, 'HRESULT_FROM_WIN32(ERROR_NO_SPOOL_SPACE)')
    WIN32__ERROR_PRINT_CANCELLED: Final[ThirtyTwoBits] = cached_hresult(0x8007003F, 'HRESULT_FROM_WIN32(ERROR_PRINT_CANCELLED)')
    WIN32__ERROR_NETNAME_DELETED: Final[ThirtyTwoBits] = cached_hresult(0x80070040, 'HRESULT_FROM_WIN32(ERROR_NETNAME_DELETED)')
    WIN32__ERROR_NETWORK_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x80070041, 'HRESULT_FROM_WIN32(ERROR_NETWORK_ACCESS_DENIED)')
    WIN32__ERROR_BAD_DEV_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x80070042, 'HRESULT_FROM_WIN32(ERROR_BAD_DEV_TYPE)')
    WIN32__ERROR_BAD_NET_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80070043, 'HRESULT_FROM_WIN32(ERROR_BAD_NET_NAME)')
    WIN32__ERROR_TOO_MANY_NAMES: Final[ThirtyTwoBits] = cached_hresult(0x80070044, 'HRESULT_FROM_WIN32(ERROR_TOO_MANY_NAMES)')
    WIN32__ERROR_TOO_MANY_SESS: Final[ThirtyTwoBits] = cached_hresult(0x80070045, 'HRESULT_FROM_WIN32(ERROR_TOO_MANY_SESS)')
    WIN32__ERROR_SHARING_PAUSED: Final[ThirtyTwoBits] = cached_hresult(0x80070046, 'HRESULT_FROM_WIN32(ERROR_SHARING_PAUSED)')
    WIN32__ERROR_REQ_NOT_ACCEP: Final[ThirtyTwoBits] = cached_hresult(0x80070047, 'HRESULT_FROM_WIN32(ERROR_REQ_NOT_ACCEP)')
    WIN32__ERROR_REDIR_PAUSED: Final[ThirtyTwoBits] = cached_hresult(0x80070048, 'HRESULT_FROM_WIN32(ERROR_REDIR_PAUSED)')
    WIN32__ERROR_FILE_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80070050, 'HRESULT_FROM_WIN32(ERROR_FILE_EXISTS)')
    WIN32__ERROR_CANNOT_MAKE: Final[ThirtyTwoBits] = cached_hresult(0x80070052, 'HRESULT_FROM_WIN32(ERROR_CANNOT_MAKE)')
    WIN32__ERROR_FAIL_I24: Final[ThirtyTwoBits] = cached_hresult(0x80070053, 'HRESULT_FROM_WIN32(ERROR_FAIL_I24)')
    WIN32__ERROR_OUT_OF_STRUCTURES: Final[ThirtyTwoBits] = cached_hresult(0x80070054, 'HRESULT_FROM_WIN32(ERROR_OUT_OF_STRUCTURES)')
    WIN32__ERROR_ALREADY_ASSIGNED: Final[ThirtyTwoBits] = cached_hresult(0x80070055, 'HRESULT_FROM_WIN32(ERROR_ALREADY_ASSIGNED)')
    WIN32__ERROR_INVALID_PASSWORD: Final[ThirtyTwoBits] = cached_hresult(0x80070056, 'HRESULT_FROM_WIN32(ERROR_INVALID_PASSWORD)')
    WIN32__ERROR_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0x80070057, 'HRESULT_FROM_WIN32(ERROR_INVALID_PARAMETER)')
    WIN32__ERROR_NET_WRITE_FAULT: Final[ThirtyTwoBits] = cached_hresult(0x80070058, 'HRESULT_FROM_WIN32(ERROR_NET_WRITE_FAULT)')
    WIN32__ERROR_NO_PROC_SLOTS: Final[ThirtyTwoBits] = cached_hresult(0x80070059, 'HRESULT_FROM_WIN32(ERROR_NO_PROC_SLOTS)')
    WIN32__ERROR_TOO_MANY_SEMAPHORES: Final[ThirtyTwoBits] = cached_hresult(0x80070064, 'HRESULT_FROM_WIN32(ERROR_TOO_MANY_SEMAPHORES)')
    WIN32__ERROR_EXCL_SEM_ALREADY_OWNED: Final[ThirtyTwoBits] = cached_hresult(0x80070065, 'HRESULT_FROM_WIN32(ERROR_EXCL_SEM_ALREADY_OWNED)')
    WIN32__ERROR_SEM_IS_SET: Final[ThirtyTwoBits] = cached_hresult(0x80070066, 'HRESULT_FROM_WIN32(ERROR_SEM_IS_SET)')
    WIN32__ERROR_TOO_MANY_SEM_REQUESTS: Final[ThirtyTwoBits] = cached_hresult(0x80070067, 'HRESULT_FROM_WIN32(ERROR_TOO_MANY_SEM_REQUESTS)')
    WIN32__ERROR_INVALID_AT_INTERRUPT_TIME: Final[ThirtyTwoBits] = cached_hresult(0x80070068, 'HRESULT_FROM_WIN32(ERROR_INVALID_AT_INTERRUPT_TIME)')
    WIN32__ERROR_SEM_OWNER_DIED: Final[ThirtyTwoBits] = cached_hresult(0x80070069, 'HRESULT_FROM_WIN32(ERROR_SEM_OWNER_DIED)')
    WIN32__ERROR_SEM_USER_LIMIT: Final[ThirtyTwoBits] = cached_hresult(0x8007006A, 'HRESULT_FROM_WIN32(ERROR_SEM_USER_LIMIT)')
    WIN32__ERROR_DISK_CHANGE: Final[ThirtyTwoBits] = cached_hresult(0x8007006B, 'HRESULT_FROM_WIN32(ERROR_DISK_CHANGE)')
    WIN32__ERROR_DRIVE_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x8007006C, 'HRESULT_FROM_WIN32(ERROR_DRIVE_LOCKED)')
    WIN32__ERROR_BROKEN_PIPE: Final[ThirtyTwoBits] = cached_hresult(0x8007006D, 'HRESULT_FROM_WIN32(ERROR_BROKEN_PIPE)')
    WIN32__ERROR_OPEN_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8007006E, 'HRESULT_FROM_WIN32(ERROR_OPEN_FAILED)')
    WIN32__ERROR_BUFFER_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0x8007006F, 'HRESULT_FROM_WIN32(ERROR_BUFFER_OVERFLOW)')
    WIN32__ERROR_DISK_FULL: Final[ThirtyTwoBits] = cached_hresult(0x80070070, 'HRESULT_FROM_WIN32(ERROR_DISK_FULL)')
    WIN32__ERROR_NO_MORE_SEARCH_HANDLES: Final[ThirtyTwoBits] = cached_hresult(0x80070071, 'HRESULT_FROM_WIN32(ERROR_NO_MORE_SEARCH_HANDLES)')
    WIN32__ERROR_INVALID_TARGET_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80070072, 'HRESULT_FROM_WIN32(ERROR_INVALID_TARGET_HANDLE)')
    WIN32__ERROR_INVALID_CATEGORY: Final[ThirtyTwoBits] = cached_hresult(0x80070075, 'HRESULT_FROM_WIN32(ERROR_INVALID_CATEGORY)')
    WIN32__ERROR_INVALID_VERIFY_SWITCH: Final[ThirtyTwoBits] = cached_hresult(0x80070076, 'HRESULT_FROM_WIN32(ERROR_INVALID_VERIFY_SWITCH)')
    WIN32__ERROR_BAD_DRIVER_LEVEL: Final[ThirtyTwoBits] = cached_hresult(0x80070077, 'HRESULT_FROM_WIN32(ERROR_BAD_DRIVER_LEVEL)')
    WIN32__ERROR_CALL_NOT_IMPLEMENTED: Final[ThirtyTwoBits] = cached_hresult(0x80070078, 'HRESULT_FROM_WIN32(ERROR_CALL_NOT_IMPLEMENTED)')
    WIN32__ERROR_SEM_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x80070079, 'HRESULT_FROM_WIN32(ERROR_SEM_TIMEOUT)')
    WIN32__ERROR_INSUFFICIENT_BUFFER: Final[ThirtyTwoBits] = cached_hresult(0x8007007A, 'HRESULT_FROM_WIN32(ERROR_INSUFFICIENT_BUFFER)')
    WIN32__ERROR_INVALID_NAME: Final[ThirtyTwoBits] = cached_hresult(0x8007007B, 'HRESULT_FROM_WIN32(ERROR_INVALID_NAME)')
    WIN32__ERROR_INVALID_LEVEL: Final[ThirtyTwoBits] = cached_hresult(0x8007007C, 'HRESULT_FROM_WIN32(ERROR_INVALID_LEVEL)')
    WIN32__ERROR_NO_VOLUME_LABEL: Final[ThirtyTwoBits] = cached_hresult(0x8007007D, 'HRESULT_FROM_WIN32(ERROR_NO_VOLUME_LABEL)')
    WIN32__ERROR_MOD_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8007007E, 'HRESULT_FROM_WIN32(ERROR_MOD_NOT_FOUND)')
    WIN32__ERROR_PROC_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8007007F, 'HRESULT_FROM_WIN32(ERROR_PROC_NOT_FOUND)')
    WIN32__ERROR_WAIT_NO_CHILDREN: Final[ThirtyTwoBits] = cached_hresult(0x80070080, 'HRESULT_FROM_WIN32(ERROR_WAIT_NO_CHILDREN)')
    WIN32__ERROR_CHILD_NOT_COMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x80070081, 'HRESULT_FROM_WIN32(ERROR_CHILD_NOT_COMPLETE)')
    WIN32__ERROR_DIRECT_ACCESS_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80070082, 'HRESULT_FROM_WIN32(ERROR_DIRECT_ACCESS_HANDLE)')
    WIN32__ERROR_NEGATIVE_SEEK: Final[ThirtyTwoBits] = cached_hresult(0x80070083, 'HRESULT_FROM_WIN32(ERROR_NEGATIVE_SEEK)')
    WIN32__ERROR_SEEK_ON_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0x80070084, 'HRESULT_FROM_WIN32(ERROR_SEEK_ON_DEVICE)')
    WIN32__ERROR_IS_JOIN_TARGET: Final[ThirtyTwoBits] = cached_hresult(0x80070085, 'HRESULT_FROM_WIN32(ERROR_IS_JOIN_TARGET)')
    WIN32__ERROR_IS_JOINED: Final[ThirtyTwoBits] = cached_hresult(0x80070086, 'HRESULT_FROM_WIN32(ERROR_IS_JOINED)')
    WIN32__ERROR_IS_SUBSTED: Final[ThirtyTwoBits] = cached_hresult(0x80070087, 'HRESULT_FROM_WIN32(ERROR_IS_SUBSTED)')
    WIN32__ERROR_NOT_JOINED: Final[ThirtyTwoBits] = cached_hresult(0x80070088, 'HRESULT_FROM_WIN32(ERROR_NOT_JOINED)')
    WIN32__ERROR_NOT_SUBSTED: Final[ThirtyTwoBits] = cached_hresult(0x80070089, 'HRESULT_FROM_WIN32(ERROR_NOT_SUBSTED)')
    WIN32__ERROR_JOIN_TO_JOIN: Final[ThirtyTwoBits] = cached_hresult(0x8007008A, 'HRESULT_FROM_WIN32(ERROR_JOIN_TO_JOIN)')
    WIN32__ERROR_SUBST_TO_SUBST: Final[ThirtyTwoBits] = cached_hresult(0x8007008B, 'HRESULT_FROM_WIN32(ERROR_SUBST_TO_SUBST)')
    WIN32__ERROR_JOIN_TO_SUBST: Final[ThirtyTwoBits] = cached_hresult(0x8007008C, 'HRESULT_FROM_WIN32(ERROR_JOIN_TO_SUBST)')
    WIN32__ERROR_SUBST_TO_JOIN: Final[ThirtyTwoBits] = cached_hresult(0x8007008D, 'HRESULT_FROM_WIN32(ERROR_SUBST_TO_JOIN)')
    WIN32__ERROR_BUSY_DRIVE: Final[ThirtyTwoBits] = cached_hresult(0x8007008E, 'HRESULT_FROM_WIN32(ERROR_BUSY_DRIVE)')
    WIN32__ERROR_SAME_DRIVE: Final[ThirtyTwoBits] = cached_hresult(0x8007008F, 'HRESULT_FROM_WIN32(ERROR_SAME_DRIVE)')
    WIN32__ERROR_DIR_NOT_ROOT: Final[ThirtyTwoBits] = cached_hresult(0x80070090, 'HRESULT_FROM_WIN32(ERROR_DIR_NOT_ROOT)')
    WIN32__ERROR_DIR_NOT_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0x80070091, 'HRESULT_FROM_WIN32(ERROR_DIR_NOT_EMPTY)')
    WIN32__ERROR_IS_SUBST_PATH: Final[ThirtyTwoBits] = cached_hresult(0x80070092, 'HRESULT_FROM_WIN32(ERROR_IS_SUBST_PATH)')
    WIN32__ERROR_IS_JOIN_PATH: Final[ThirtyTwoBits] = cached_hresult(0x80070093, 'HRESULT_FROM_WIN32(ERROR_IS_JOIN_PATH)')
    WIN32__ERROR_PATH_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x80070094, 'HRESULT_FROM_WIN32(ERROR_PATH_BUSY)')
    WIN32__ERROR_IS_SUBST_TARGET: Final[ThirtyTwoBits] = cached_hresult(0x80070095, 'HRESULT_FROM_WIN32(ERROR_IS_SUBST_TARGET)')
    WIN32__ERROR_SYSTEM_TRACE: Final[ThirtyTwoBits] = cached_hresult(0x80070096, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_TRACE)')
    WIN32__ERROR_INVALID_EVENT_COUNT: Final[ThirtyTwoBits] = cached_hresult(0x80070097, 'HRESULT_FROM_WIN32(ERROR_INVALID_EVENT_COUNT)')
    WIN32__ERROR_TOO_MANY_MUXWAITERS: Final[ThirtyTwoBits] = cached_hresult(0x80070098, 'HRESULT_FROM_WIN32(ERROR_TOO_MANY_MUXWAITERS)')
    WIN32__ERROR_INVALID_LIST_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0x80070099, 'HRESULT_FROM_WIN32(ERROR_INVALID_LIST_FORMAT)')
    WIN32__ERROR_LABEL_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0x8007009A, 'HRESULT_FROM_WIN32(ERROR_LABEL_TOO_LONG)')
    WIN32__ERROR_TOO_MANY_TCBS: Final[ThirtyTwoBits] = cached_hresult(0x8007009B, 'HRESULT_FROM_WIN32(ERROR_TOO_MANY_TCBS)')
    WIN32__ERROR_SIGNAL_REFUSED: Final[ThirtyTwoBits] = cached_hresult(0x8007009C, 'HRESULT_FROM_WIN32(ERROR_SIGNAL_REFUSED)')
    WIN32__ERROR_DISCARDED: Final[ThirtyTwoBits] = cached_hresult(0x8007009D, 'HRESULT_FROM_WIN32(ERROR_DISCARDED)')
    WIN32__ERROR_NOT_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x8007009E, 'HRESULT_FROM_WIN32(ERROR_NOT_LOCKED)')
    WIN32__ERROR_BAD_THREADID_ADDR: Final[ThirtyTwoBits] = cached_hresult(0x8007009F, 'HRESULT_FROM_WIN32(ERROR_BAD_THREADID_ADDR)')
    WIN32__ERROR_BAD_ARGUMENTS: Final[ThirtyTwoBits] = cached_hresult(0x800700A0, 'HRESULT_FROM_WIN32(ERROR_BAD_ARGUMENTS)')
    WIN32__ERROR_BAD_PATHNAME: Final[ThirtyTwoBits] = cached_hresult(0x800700A1, 'HRESULT_FROM_WIN32(ERROR_BAD_PATHNAME)')
    WIN32__ERROR_SIGNAL_PENDING: Final[ThirtyTwoBits] = cached_hresult(0x800700A2, 'HRESULT_FROM_WIN32(ERROR_SIGNAL_PENDING)')
    WIN32__ERROR_MAX_THRDS_REACHED: Final[ThirtyTwoBits] = cached_hresult(0x800700A4, 'HRESULT_FROM_WIN32(ERROR_MAX_THRDS_REACHED)')
    WIN32__ERROR_LOCK_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800700A7, 'HRESULT_FROM_WIN32(ERROR_LOCK_FAILED)')
    WIN32__ERROR_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x800700AA, 'HRESULT_FROM_WIN32(ERROR_BUSY)')
    WIN32__ERROR_DEVICE_SUPPORT_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x800700AB, 'HRESULT_FROM_WIN32(ERROR_DEVICE_SUPPORT_IN_PROGRESS)')
    WIN32__ERROR_CANCEL_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x800700AD, 'HRESULT_FROM_WIN32(ERROR_CANCEL_VIOLATION)')
    WIN32__ERROR_ATOMIC_LOCKS_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x800700AE, 'HRESULT_FROM_WIN32(ERROR_ATOMIC_LOCKS_NOT_SUPPORTED)')
    WIN32__ERROR_INVALID_SEGMENT_NUMBER: Final[ThirtyTwoBits] = cached_hresult(0x800700B4, 'HRESULT_FROM_WIN32(ERROR_INVALID_SEGMENT_NUMBER)')
    WIN32__ERROR_INVALID_ORDINAL: Final[ThirtyTwoBits] = cached_hresult(0x800700B6, 'HRESULT_FROM_WIN32(ERROR_INVALID_ORDINAL)')
    WIN32__ERROR_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800700B7, 'HRESULT_FROM_WIN32(ERROR_ALREADY_EXISTS)')
    WIN32__ERROR_INVALID_FLAG_NUMBER: Final[ThirtyTwoBits] = cached_hresult(0x800700BA, 'HRESULT_FROM_WIN32(ERROR_INVALID_FLAG_NUMBER)')
    WIN32__ERROR_SEM_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800700BB, 'HRESULT_FROM_WIN32(ERROR_SEM_NOT_FOUND)')
    WIN32__ERROR_INVALID_STARTING_CODESEG: Final[ThirtyTwoBits] = cached_hresult(0x800700BC, 'HRESULT_FROM_WIN32(ERROR_INVALID_STARTING_CODESEG)')
    WIN32__ERROR_INVALID_STACKSEG: Final[ThirtyTwoBits] = cached_hresult(0x800700BD, 'HRESULT_FROM_WIN32(ERROR_INVALID_STACKSEG)')
    WIN32__ERROR_INVALID_MODULETYPE: Final[ThirtyTwoBits] = cached_hresult(0x800700BE, 'HRESULT_FROM_WIN32(ERROR_INVALID_MODULETYPE)')
    WIN32__ERROR_INVALID_EXE_SIGNATURE: Final[ThirtyTwoBits] = cached_hresult(0x800700BF, 'HRESULT_FROM_WIN32(ERROR_INVALID_EXE_SIGNATURE)')
    WIN32__ERROR_EXE_MARKED_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800700C0, 'HRESULT_FROM_WIN32(ERROR_EXE_MARKED_INVALID)')
    WIN32__ERROR_BAD_EXE_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0x800700C1, 'HRESULT_FROM_WIN32(ERROR_BAD_EXE_FORMAT)')
    WIN32__ERROR_ITERATED_DATA_EXCEEDS_64k: Final[ThirtyTwoBits] = cached_hresult(0x800700C2, 'HRESULT_FROM_WIN32(ERROR_ITERATED_DATA_EXCEEDS_64k)')
    WIN32__ERROR_INVALID_MINALLOCSIZE: Final[ThirtyTwoBits] = cached_hresult(0x800700C3, 'HRESULT_FROM_WIN32(ERROR_INVALID_MINALLOCSIZE)')
    WIN32__ERROR_DYNLINK_FROM_INVALID_RING: Final[ThirtyTwoBits] = cached_hresult(0x800700C4, 'HRESULT_FROM_WIN32(ERROR_DYNLINK_FROM_INVALID_RING)')
    WIN32__ERROR_IOPL_NOT_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0x800700C5, 'HRESULT_FROM_WIN32(ERROR_IOPL_NOT_ENABLED)')
    WIN32__ERROR_INVALID_SEGDPL: Final[ThirtyTwoBits] = cached_hresult(0x800700C6, 'HRESULT_FROM_WIN32(ERROR_INVALID_SEGDPL)')
    WIN32__ERROR_AUTODATASEG_EXCEEDS_64k: Final[ThirtyTwoBits] = cached_hresult(0x800700C7, 'HRESULT_FROM_WIN32(ERROR_AUTODATASEG_EXCEEDS_64k)')
    WIN32__ERROR_RING2SEG_MUST_BE_MOVABLE: Final[ThirtyTwoBits] = cached_hresult(0x800700C8, 'HRESULT_FROM_WIN32(ERROR_RING2SEG_MUST_BE_MOVABLE)')
    WIN32__ERROR_RELOC_CHAIN_XEEDS_SEGLIM: Final[ThirtyTwoBits] = cached_hresult(0x800700C9, 'HRESULT_FROM_WIN32(ERROR_RELOC_CHAIN_XEEDS_SEGLIM)')
    WIN32__ERROR_INFLOOP_IN_RELOC_CHAIN: Final[ThirtyTwoBits] = cached_hresult(0x800700CA, 'HRESULT_FROM_WIN32(ERROR_INFLOOP_IN_RELOC_CHAIN)')
    WIN32__ERROR_ENVVAR_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800700CB, 'HRESULT_FROM_WIN32(ERROR_ENVVAR_NOT_FOUND)')
    WIN32__ERROR_NO_SIGNAL_SENT: Final[ThirtyTwoBits] = cached_hresult(0x800700CD, 'HRESULT_FROM_WIN32(ERROR_NO_SIGNAL_SENT)')
    WIN32__ERROR_FILENAME_EXCED_RANGE: Final[ThirtyTwoBits] = cached_hresult(0x800700CE, 'HRESULT_FROM_WIN32(ERROR_FILENAME_EXCED_RANGE)')
    WIN32__ERROR_RING2_STACK_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x800700CF, 'HRESULT_FROM_WIN32(ERROR_RING2_STACK_IN_USE)')
    WIN32__ERROR_META_EXPANSION_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0x800700D0, 'HRESULT_FROM_WIN32(ERROR_META_EXPANSION_TOO_LONG)')
    WIN32__ERROR_INVALID_SIGNAL_NUMBER: Final[ThirtyTwoBits] = cached_hresult(0x800700D1, 'HRESULT_FROM_WIN32(ERROR_INVALID_SIGNAL_NUMBER)')
    WIN32__ERROR_THREAD_1_INACTIVE: Final[ThirtyTwoBits] = cached_hresult(0x800700D2, 'HRESULT_FROM_WIN32(ERROR_THREAD_1_INACTIVE)')
    WIN32__ERROR_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x800700D4, 'HRESULT_FROM_WIN32(ERROR_LOCKED)')
    WIN32__ERROR_TOO_MANY_MODULES: Final[ThirtyTwoBits] = cached_hresult(0x800700D6, 'HRESULT_FROM_WIN32(ERROR_TOO_MANY_MODULES)')
    WIN32__ERROR_NESTING_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x800700D7, 'HRESULT_FROM_WIN32(ERROR_NESTING_NOT_ALLOWED)')
    WIN32__ERROR_EXE_MACHINE_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x800700D8, 'HRESULT_FROM_WIN32(ERROR_EXE_MACHINE_TYPE_MISMATCH)')
    WIN32__ERROR_EXE_CANNOT_MODIFY_SIGNED_BINARY: Final[ThirtyTwoBits] = cached_hresult(0x800700D9, 'HRESULT_FROM_WIN32(ERROR_EXE_CANNOT_MODIFY_SIGNED_BINARY)')
    WIN32__ERROR_EXE_CANNOT_MODIFY_STRONG_SIGNED_BINARY: Final[ThirtyTwoBits] = cached_hresult(0x800700DA, 'HRESULT_FROM_WIN32(ERROR_EXE_CANNOT_MODIFY_STRONG_SIGNED_BINARY)')
    WIN32__ERROR_FILE_CHECKED_OUT: Final[ThirtyTwoBits] = cached_hresult(0x800700DC, 'HRESULT_FROM_WIN32(ERROR_FILE_CHECKED_OUT)')
    WIN32__ERROR_CHECKOUT_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x800700DD, 'HRESULT_FROM_WIN32(ERROR_CHECKOUT_REQUIRED)')
    WIN32__ERROR_BAD_FILE_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x800700DE, 'HRESULT_FROM_WIN32(ERROR_BAD_FILE_TYPE)')
    WIN32__ERROR_FILE_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0x800700DF, 'HRESULT_FROM_WIN32(ERROR_FILE_TOO_LARGE)')
    WIN32__ERROR_FORMS_AUTH_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x800700E0, 'HRESULT_FROM_WIN32(ERROR_FORMS_AUTH_REQUIRED)')
    WIN32__ERROR_VIRUS_INFECTED: Final[ThirtyTwoBits] = cached_hresult(0x800700E1, 'HRESULT_FROM_WIN32(ERROR_VIRUS_INFECTED)')
    WIN32__ERROR_VIRUS_DELETED: Final[ThirtyTwoBits] = cached_hresult(0x800700E2, 'HRESULT_FROM_WIN32(ERROR_VIRUS_DELETED)')
    WIN32__ERROR_PIPE_LOCAL: Final[ThirtyTwoBits] = cached_hresult(0x800700E5, 'HRESULT_FROM_WIN32(ERROR_PIPE_LOCAL)')
    WIN32__ERROR_BAD_PIPE: Final[ThirtyTwoBits] = cached_hresult(0x800700E6, 'HRESULT_FROM_WIN32(ERROR_BAD_PIPE)')
    WIN32__ERROR_PIPE_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x800700E7, 'HRESULT_FROM_WIN32(ERROR_PIPE_BUSY)')
    WIN32__ERROR_NO_DATA: Final[ThirtyTwoBits] = cached_hresult(0x800700E8, 'HRESULT_FROM_WIN32(ERROR_NO_DATA)')
    WIN32__ERROR_PIPE_NOT_CONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x800700E9, 'HRESULT_FROM_WIN32(ERROR_PIPE_NOT_CONNECTED)')
    WIN32__ERROR_MORE_DATA: Final[ThirtyTwoBits] = cached_hresult(0x800700EA, 'HRESULT_FROM_WIN32(ERROR_MORE_DATA)')
    WIN32__ERROR_NO_WORK_DONE: Final[ThirtyTwoBits] = cached_hresult(0x800700EB, 'HRESULT_FROM_WIN32(ERROR_NO_WORK_DONE)')
    WIN32__ERROR_VC_DISCONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x800700F0, 'HRESULT_FROM_WIN32(ERROR_VC_DISCONNECTED)')
    WIN32__ERROR_INVALID_EA_NAME: Final[ThirtyTwoBits] = cached_hresult(0x800700FE, 'HRESULT_FROM_WIN32(ERROR_INVALID_EA_NAME)')
    WIN32__ERROR_EA_LIST_INCONSISTENT: Final[ThirtyTwoBits] = cached_hresult(0x800700FF, 'HRESULT_FROM_WIN32(ERROR_EA_LIST_INCONSISTENT)')
    WIN32__WAIT_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x80070102, 'HRESULT_FROM_WIN32(WAIT_TIMEOUT)')
    WIN32__ERROR_NO_MORE_ITEMS: Final[ThirtyTwoBits] = cached_hresult(0x80070103, 'HRESULT_FROM_WIN32(ERROR_NO_MORE_ITEMS)')
    WIN32__ERROR_CANNOT_COPY: Final[ThirtyTwoBits] = cached_hresult(0x8007010A, 'HRESULT_FROM_WIN32(ERROR_CANNOT_COPY)')
    WIN32__ERROR_DIRECTORY: Final[ThirtyTwoBits] = cached_hresult(0x8007010B, 'HRESULT_FROM_WIN32(ERROR_DIRECTORY)')
    WIN32__ERROR_EAS_DIDNT_FIT: Final[ThirtyTwoBits] = cached_hresult(0x80070113, 'HRESULT_FROM_WIN32(ERROR_EAS_DIDNT_FIT)')
    WIN32__ERROR_EA_FILE_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x80070114, 'HRESULT_FROM_WIN32(ERROR_EA_FILE_CORRUPT)')
    WIN32__ERROR_EA_TABLE_FULL: Final[ThirtyTwoBits] = cached_hresult(0x80070115, 'HRESULT_FROM_WIN32(ERROR_EA_TABLE_FULL)')
    WIN32__ERROR_INVALID_EA_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80070116, 'HRESULT_FROM_WIN32(ERROR_INVALID_EA_HANDLE)')
    WIN32__ERROR_EAS_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8007011A, 'HRESULT_FROM_WIN32(ERROR_EAS_NOT_SUPPORTED)')
    WIN32__ERROR_NOT_OWNER: Final[ThirtyTwoBits] = cached_hresult(0x80070120, 'HRESULT_FROM_WIN32(ERROR_NOT_OWNER)')
    WIN32__ERROR_TOO_MANY_POSTS: Final[ThirtyTwoBits] = cached_hresult(0x8007012A, 'HRESULT_FROM_WIN32(ERROR_TOO_MANY_POSTS)')
    WIN32__ERROR_PARTIAL_COPY: Final[ThirtyTwoBits] = cached_hresult(0x8007012B, 'HRESULT_FROM_WIN32(ERROR_PARTIAL_COPY)')
    WIN32__ERROR_OPLOCK_NOT_GRANTED: Final[ThirtyTwoBits] = cached_hresult(0x8007012C, 'HRESULT_FROM_WIN32(ERROR_OPLOCK_NOT_GRANTED)')
    WIN32__ERROR_INVALID_OPLOCK_PROTOCOL: Final[ThirtyTwoBits] = cached_hresult(0x8007012D, 'HRESULT_FROM_WIN32(ERROR_INVALID_OPLOCK_PROTOCOL)')
    WIN32__ERROR_DISK_TOO_FRAGMENTED: Final[ThirtyTwoBits] = cached_hresult(0x8007012E, 'HRESULT_FROM_WIN32(ERROR_DISK_TOO_FRAGMENTED)')
    WIN32__ERROR_DELETE_PENDING: Final[ThirtyTwoBits] = cached_hresult(0x8007012F, 'HRESULT_FROM_WIN32(ERROR_DELETE_PENDING)')
    WIN32__ERROR_INCOMPATIBLE_WITH_GLOBAL_SHORT_NAME_REGISTRY_SETTING: Final[ThirtyTwoBits] = cached_hresult(0x80070130, 'HRESULT_FROM_WIN32(ERROR_INCOMPATIBLE_WITH_GLOBAL_SHORT_NAME_REGISTRY_SETTING)')
    WIN32__ERROR_SHORT_NAMES_NOT_ENABLED_ON_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0x80070131, 'HRESULT_FROM_WIN32(ERROR_SHORT_NAMES_NOT_ENABLED_ON_VOLUME)')
    WIN32__ERROR_SECURITY_STREAM_IS_INCONSISTENT: Final[ThirtyTwoBits] = cached_hresult(0x80070132, 'HRESULT_FROM_WIN32(ERROR_SECURITY_STREAM_IS_INCONSISTENT)')
    WIN32__ERROR_INVALID_LOCK_RANGE: Final[ThirtyTwoBits] = cached_hresult(0x80070133, 'HRESULT_FROM_WIN32(ERROR_INVALID_LOCK_RANGE)')
    WIN32__ERROR_IMAGE_SUBSYSTEM_NOT_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0x80070134, 'HRESULT_FROM_WIN32(ERROR_IMAGE_SUBSYSTEM_NOT_PRESENT)')
    WIN32__ERROR_NOTIFICATION_GUID_ALREADY_DEFINED: Final[ThirtyTwoBits] = cached_hresult(0x80070135, 'HRESULT_FROM_WIN32(ERROR_NOTIFICATION_GUID_ALREADY_DEFINED)')
    WIN32__ERROR_INVALID_EXCEPTION_HANDLER: Final[ThirtyTwoBits] = cached_hresult(0x80070136, 'HRESULT_FROM_WIN32(ERROR_INVALID_EXCEPTION_HANDLER)')
    WIN32__ERROR_DUPLICATE_PRIVILEGES: Final[ThirtyTwoBits] = cached_hresult(0x80070137, 'HRESULT_FROM_WIN32(ERROR_DUPLICATE_PRIVILEGES)')
    WIN32__ERROR_NO_RANGES_PROCESSED: Final[ThirtyTwoBits] = cached_hresult(0x80070138, 'HRESULT_FROM_WIN32(ERROR_NO_RANGES_PROCESSED)')
    WIN32__ERROR_NOT_ALLOWED_ON_SYSTEM_FILE: Final[ThirtyTwoBits] = cached_hresult(0x80070139, 'HRESULT_FROM_WIN32(ERROR_NOT_ALLOWED_ON_SYSTEM_FILE)')
    WIN32__ERROR_DISK_RESOURCES_EXHAUSTED: Final[ThirtyTwoBits] = cached_hresult(0x8007013A, 'HRESULT_FROM_WIN32(ERROR_DISK_RESOURCES_EXHAUSTED)')
    WIN32__ERROR_INVALID_TOKEN: Final[ThirtyTwoBits] = cached_hresult(0x8007013B, 'HRESULT_FROM_WIN32(ERROR_INVALID_TOKEN)')
    WIN32__ERROR_DEVICE_FEATURE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8007013C, 'HRESULT_FROM_WIN32(ERROR_DEVICE_FEATURE_NOT_SUPPORTED)')
    WIN32__ERROR_MR_MID_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8007013D, 'HRESULT_FROM_WIN32(ERROR_MR_MID_NOT_FOUND)')
    WIN32__ERROR_SCOPE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8007013E, 'HRESULT_FROM_WIN32(ERROR_SCOPE_NOT_FOUND)')
    WIN32__ERROR_UNDEFINED_SCOPE: Final[ThirtyTwoBits] = cached_hresult(0x8007013F, 'HRESULT_FROM_WIN32(ERROR_UNDEFINED_SCOPE)')
    WIN32__ERROR_INVALID_CAP: Final[ThirtyTwoBits] = cached_hresult(0x80070140, 'HRESULT_FROM_WIN32(ERROR_INVALID_CAP)')
    WIN32__ERROR_DEVICE_UNREACHABLE: Final[ThirtyTwoBits] = cached_hresult(0x80070141, 'HRESULT_FROM_WIN32(ERROR_DEVICE_UNREACHABLE)')
    WIN32__ERROR_DEVICE_NO_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0x80070142, 'HRESULT_FROM_WIN32(ERROR_DEVICE_NO_RESOURCES)')
    WIN32__ERROR_DATA_CHECKSUM_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80070143, 'HRESULT_FROM_WIN32(ERROR_DATA_CHECKSUM_ERROR)')
    WIN32__ERROR_INTERMIXED_KERNEL_EA_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x80070144, 'HRESULT_FROM_WIN32(ERROR_INTERMIXED_KERNEL_EA_OPERATION)')
    WIN32__ERROR_FILE_LEVEL_TRIM_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80070146, 'HRESULT_FROM_WIN32(ERROR_FILE_LEVEL_TRIM_NOT_SUPPORTED)')
    WIN32__ERROR_OFFSET_ALIGNMENT_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x80070147, 'HRESULT_FROM_WIN32(ERROR_OFFSET_ALIGNMENT_VIOLATION)')
    WIN32__ERROR_INVALID_FIELD_IN_PARAMETER_LIST: Final[ThirtyTwoBits] = cached_hresult(0x80070148, 'HRESULT_FROM_WIN32(ERROR_INVALID_FIELD_IN_PARAMETER_LIST)')
    WIN32__ERROR_OPERATION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x80070149, 'HRESULT_FROM_WIN32(ERROR_OPERATION_IN_PROGRESS)')
    WIN32__ERROR_BAD_DEVICE_PATH: Final[ThirtyTwoBits] = cached_hresult(0x8007014A, 'HRESULT_FROM_WIN32(ERROR_BAD_DEVICE_PATH)')
    WIN32__ERROR_TOO_MANY_DESCRIPTORS: Final[ThirtyTwoBits] = cached_hresult(0x8007014B, 'HRESULT_FROM_WIN32(ERROR_TOO_MANY_DESCRIPTORS)')
    WIN32__ERROR_SCRUB_DATA_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x8007014C, 'HRESULT_FROM_WIN32(ERROR_SCRUB_DATA_DISABLED)')
    WIN32__ERROR_NOT_REDUNDANT_STORAGE: Final[ThirtyTwoBits] = cached_hresult(0x8007014D, 'HRESULT_FROM_WIN32(ERROR_NOT_REDUNDANT_STORAGE)')
    WIN32__ERROR_RESIDENT_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8007014E, 'HRESULT_FROM_WIN32(ERROR_RESIDENT_FILE_NOT_SUPPORTED)')
    WIN32__ERROR_COMPRESSED_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8007014F, 'HRESULT_FROM_WIN32(ERROR_COMPRESSED_FILE_NOT_SUPPORTED)')
    WIN32__ERROR_DIRECTORY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80070150, 'HRESULT_FROM_WIN32(ERROR_DIRECTORY_NOT_SUPPORTED)')
    WIN32__ERROR_NOT_READ_FROM_COPY: Final[ThirtyTwoBits] = cached_hresult(0x80070151, 'HRESULT_FROM_WIN32(ERROR_NOT_READ_FROM_COPY)')
    WIN32__ERROR_FT_WRITE_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80070152, 'HRESULT_FROM_WIN32(ERROR_FT_WRITE_FAILURE)')
    WIN32__ERROR_FT_DI_SCAN_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80070153, 'HRESULT_FROM_WIN32(ERROR_FT_DI_SCAN_REQUIRED)')
    WIN32__ERROR_INVALID_KERNEL_INFO_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80070154, 'HRESULT_FROM_WIN32(ERROR_INVALID_KERNEL_INFO_VERSION)')
    WIN32__ERROR_INVALID_PEP_INFO_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80070155, 'HRESULT_FROM_WIN32(ERROR_INVALID_PEP_INFO_VERSION)')
    WIN32__ERROR_OBJECT_NOT_EXTERNALLY_BACKED: Final[ThirtyTwoBits] = cached_hresult(0x80070156, 'HRESULT_FROM_WIN32(ERROR_OBJECT_NOT_EXTERNALLY_BACKED)')
    WIN32__ERROR_EXTERNAL_BACKING_PROVIDER_UNKNOWN: Final[ThirtyTwoBits] = cached_hresult(0x80070157, 'HRESULT_FROM_WIN32(ERROR_EXTERNAL_BACKING_PROVIDER_UNKNOWN)')
    WIN32__ERROR_COMPRESSION_NOT_BENEFICIAL: Final[ThirtyTwoBits] = cached_hresult(0x80070158, 'HRESULT_FROM_WIN32(ERROR_COMPRESSION_NOT_BENEFICIAL)')
    WIN32__ERROR_STORAGE_TOPOLOGY_ID_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80070159, 'HRESULT_FROM_WIN32(ERROR_STORAGE_TOPOLOGY_ID_MISMATCH)')
    WIN32__ERROR_BLOCKED_BY_PARENTAL_CONTROLS: Final[ThirtyTwoBits] = cached_hresult(0x8007015A, 'HRESULT_FROM_WIN32(ERROR_BLOCKED_BY_PARENTAL_CONTROLS)')
    WIN32__ERROR_BLOCK_TOO_MANY_REFERENCES: Final[ThirtyTwoBits] = cached_hresult(0x8007015B, 'HRESULT_FROM_WIN32(ERROR_BLOCK_TOO_MANY_REFERENCES)')
    WIN32__ERROR_MARKED_TO_DISALLOW_WRITES: Final[ThirtyTwoBits] = cached_hresult(0x8007015C, 'HRESULT_FROM_WIN32(ERROR_MARKED_TO_DISALLOW_WRITES)')
    WIN32__ERROR_ENCLAVE_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8007015D, 'HRESULT_FROM_WIN32(ERROR_ENCLAVE_FAILURE)')
    WIN32__ERROR_FAIL_NOACTION_REBOOT: Final[ThirtyTwoBits] = cached_hresult(0x8007015E, 'HRESULT_FROM_WIN32(ERROR_FAIL_NOACTION_REBOOT)')
    WIN32__ERROR_FAIL_SHUTDOWN: Final[ThirtyTwoBits] = cached_hresult(0x8007015F, 'HRESULT_FROM_WIN32(ERROR_FAIL_SHUTDOWN)')
    WIN32__ERROR_FAIL_RESTART: Final[ThirtyTwoBits] = cached_hresult(0x80070160, 'HRESULT_FROM_WIN32(ERROR_FAIL_RESTART)')
    WIN32__ERROR_MAX_SESSIONS_REACHED: Final[ThirtyTwoBits] = cached_hresult(0x80070161, 'HRESULT_FROM_WIN32(ERROR_MAX_SESSIONS_REACHED)')
    WIN32__ERROR_NETWORK_ACCESS_DENIED_EDP: Final[ThirtyTwoBits] = cached_hresult(0x80070162, 'HRESULT_FROM_WIN32(ERROR_NETWORK_ACCESS_DENIED_EDP)')
    WIN32__ERROR_DEVICE_HINT_NAME_BUFFER_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0x80070163, 'HRESULT_FROM_WIN32(ERROR_DEVICE_HINT_NAME_BUFFER_TOO_SMALL)')
    WIN32__ERROR_EDP_POLICY_DENIES_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x80070164, 'HRESULT_FROM_WIN32(ERROR_EDP_POLICY_DENIES_OPERATION)')
    WIN32__ERROR_EDP_DPL_POLICY_CANT_BE_SATISFIED: Final[ThirtyTwoBits] = cached_hresult(0x80070165, 'HRESULT_FROM_WIN32(ERROR_EDP_DPL_POLICY_CANT_BE_SATISFIED)')
    WIN32__ERROR_CLOUD_FILE_SYNC_ROOT_METADATA_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x80070166, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_SYNC_ROOT_METADATA_CORRUPT)')
    WIN32__ERROR_DEVICE_IN_MAINTENANCE: Final[ThirtyTwoBits] = cached_hresult(0x80070167, 'HRESULT_FROM_WIN32(ERROR_DEVICE_IN_MAINTENANCE)')
    WIN32__ERROR_NOT_SUPPORTED_ON_DAX: Final[ThirtyTwoBits] = cached_hresult(0x80070168, 'HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED_ON_DAX)')
    WIN32__ERROR_DAX_MAPPING_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80070169, 'HRESULT_FROM_WIN32(ERROR_DAX_MAPPING_EXISTS)')
    WIN32__ERROR_CLOUD_FILE_PROVIDER_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x8007016A, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_PROVIDER_NOT_RUNNING)')
    WIN32__ERROR_CLOUD_FILE_METADATA_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x8007016B, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_METADATA_CORRUPT)')
    WIN32__ERROR_CLOUD_FILE_METADATA_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0x8007016C, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_METADATA_TOO_LARGE)')
    WIN32__ERROR_CLOUD_FILE_PROPERTY_BLOB_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0x8007016D, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_PROPERTY_BLOB_TOO_LARGE)')
    WIN32__ERROR_CLOUD_FILE_PROPERTY_BLOB_CHECKSUM_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x8007016E, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_PROPERTY_BLOB_CHECKSUM_MISMATCH)')
    WIN32__ERROR_CHILD_PROCESS_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x8007016F, 'HRESULT_FROM_WIN32(ERROR_CHILD_PROCESS_BLOCKED)')
    WIN32__ERROR_STORAGE_LOST_DATA_PERSISTENCE: Final[ThirtyTwoBits] = cached_hresult(0x80070170, 'HRESULT_FROM_WIN32(ERROR_STORAGE_LOST_DATA_PERSISTENCE)')
    WIN32__ERROR_FILE_SYSTEM_VIRTUALIZATION_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80070171, 'HRESULT_FROM_WIN32(ERROR_FILE_SYSTEM_VIRTUALIZATION_UNAVAILABLE)')
    WIN32__ERROR_FILE_SYSTEM_VIRTUALIZATION_METADATA_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x80070172, 'HRESULT_FROM_WIN32(ERROR_FILE_SYSTEM_VIRTUALIZATION_METADATA_CORRUPT)')
    WIN32__ERROR_FILE_SYSTEM_VIRTUALIZATION_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x80070173, 'HRESULT_FROM_WIN32(ERROR_FILE_SYSTEM_VIRTUALIZATION_BUSY)')
    WIN32__ERROR_FILE_SYSTEM_VIRTUALIZATION_PROVIDER_UNKNOWN: Final[ThirtyTwoBits] = cached_hresult(0x80070174, 'HRESULT_FROM_WIN32(ERROR_FILE_SYSTEM_VIRTUALIZATION_PROVIDER_UNKNOWN)')
    WIN32__ERROR_GDI_HANDLE_LEAK: Final[ThirtyTwoBits] = cached_hresult(0x80070175, 'HRESULT_FROM_WIN32(ERROR_GDI_HANDLE_LEAK)')
    WIN32__ERROR_CLOUD_FILE_TOO_MANY_PROPERTY_BLOBS: Final[ThirtyTwoBits] = cached_hresult(0x80070176, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_TOO_MANY_PROPERTY_BLOBS)')
    WIN32__ERROR_CLOUD_FILE_PROPERTY_VERSION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80070177, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_PROPERTY_VERSION_NOT_SUPPORTED)')
    WIN32__ERROR_NOT_A_CLOUD_FILE: Final[ThirtyTwoBits] = cached_hresult(0x80070178, 'HRESULT_FROM_WIN32(ERROR_NOT_A_CLOUD_FILE)')
    WIN32__ERROR_CLOUD_FILE_NOT_IN_SYNC: Final[ThirtyTwoBits] = cached_hresult(0x80070179, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_NOT_IN_SYNC)')
    WIN32__ERROR_CLOUD_FILE_ALREADY_CONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x8007017A, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_ALREADY_CONNECTED)')
    WIN32__ERROR_CLOUD_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8007017B, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_NOT_SUPPORTED)')
    WIN32__ERROR_CLOUD_FILE_INVALID_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0x8007017C, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_INVALID_REQUEST)')
    WIN32__ERROR_CLOUD_FILE_READ_ONLY_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0x8007017D, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_READ_ONLY_VOLUME)')
    WIN32__ERROR_CLOUD_FILE_CONNECTED_PROVIDER_ONLY: Final[ThirtyTwoBits] = cached_hresult(0x8007017E, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_CONNECTED_PROVIDER_ONLY)')
    WIN32__ERROR_CLOUD_FILE_VALIDATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8007017F, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_VALIDATION_FAILED)')
    WIN32__ERROR_SMB1_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80070180, 'HRESULT_FROM_WIN32(ERROR_SMB1_NOT_AVAILABLE)')
    WIN32__ERROR_FILE_SYSTEM_VIRTUALIZATION_INVALID_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x80070181, 'HRESULT_FROM_WIN32(ERROR_FILE_SYSTEM_VIRTUALIZATION_INVALID_OPERATION)')
    WIN32__ERROR_CLOUD_FILE_AUTHENTICATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80070182, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_AUTHENTICATION_FAILED)')
    WIN32__ERROR_CLOUD_FILE_INSUFFICIENT_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0x80070183, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_INSUFFICIENT_RESOURCES)')
    WIN32__ERROR_CLOUD_FILE_NETWORK_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80070184, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_NETWORK_UNAVAILABLE)')
    WIN32__ERROR_CLOUD_FILE_UNSUCCESSFUL: Final[ThirtyTwoBits] = cached_hresult(0x80070185, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_UNSUCCESSFUL)')
    WIN32__ERROR_CLOUD_FILE_NOT_UNDER_SYNC_ROOT: Final[ThirtyTwoBits] = cached_hresult(0x80070186, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_NOT_UNDER_SYNC_ROOT)')
    WIN32__ERROR_CLOUD_FILE_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x80070187, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_IN_USE)')
    WIN32__ERROR_CLOUD_FILE_PINNED: Final[ThirtyTwoBits] = cached_hresult(0x80070188, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_PINNED)')
    WIN32__ERROR_CLOUD_FILE_REQUEST_ABORTED: Final[ThirtyTwoBits] = cached_hresult(0x80070189, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_REQUEST_ABORTED)')
    WIN32__ERROR_CLOUD_FILE_PROPERTY_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x8007018A, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_PROPERTY_CORRUPT)')
    WIN32__ERROR_CLOUD_FILE_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x8007018B, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_ACCESS_DENIED)')
    WIN32__ERROR_CLOUD_FILE_INCOMPATIBLE_HARDLINKS: Final[ThirtyTwoBits] = cached_hresult(0x8007018C, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_INCOMPATIBLE_HARDLINKS)')
    WIN32__ERROR_CLOUD_FILE_PROPERTY_LOCK_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x8007018D, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_PROPERTY_LOCK_CONFLICT)')
    WIN32__ERROR_CLOUD_FILE_REQUEST_CANCELED: Final[ThirtyTwoBits] = cached_hresult(0x8007018E, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_REQUEST_CANCELED)')
    WIN32__ERROR_EXTERNAL_SYSKEY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8007018F, 'HRESULT_FROM_WIN32(ERROR_EXTERNAL_SYSKEY_NOT_SUPPORTED)')
    WIN32__ERROR_THREAD_MODE_ALREADY_BACKGROUND: Final[ThirtyTwoBits] = cached_hresult(0x80070190, 'HRESULT_FROM_WIN32(ERROR_THREAD_MODE_ALREADY_BACKGROUND)')
    WIN32__ERROR_THREAD_MODE_NOT_BACKGROUND: Final[ThirtyTwoBits] = cached_hresult(0x80070191, 'HRESULT_FROM_WIN32(ERROR_THREAD_MODE_NOT_BACKGROUND)')
    WIN32__ERROR_PROCESS_MODE_ALREADY_BACKGROUND: Final[ThirtyTwoBits] = cached_hresult(0x80070192, 'HRESULT_FROM_WIN32(ERROR_PROCESS_MODE_ALREADY_BACKGROUND)')
    WIN32__ERROR_PROCESS_MODE_NOT_BACKGROUND: Final[ThirtyTwoBits] = cached_hresult(0x80070193, 'HRESULT_FROM_WIN32(ERROR_PROCESS_MODE_NOT_BACKGROUND)')
    WIN32__ERROR_CLOUD_FILE_PROVIDER_TERMINATED: Final[ThirtyTwoBits] = cached_hresult(0x80070194, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_PROVIDER_TERMINATED)')
    WIN32__ERROR_NOT_A_CLOUD_SYNC_ROOT: Final[ThirtyTwoBits] = cached_hresult(0x80070195, 'HRESULT_FROM_WIN32(ERROR_NOT_A_CLOUD_SYNC_ROOT)')
    WIN32__ERROR_FILE_PROTECTED_UNDER_DPL: Final[ThirtyTwoBits] = cached_hresult(0x80070196, 'HRESULT_FROM_WIN32(ERROR_FILE_PROTECTED_UNDER_DPL)')
    WIN32__ERROR_VOLUME_NOT_CLUSTER_ALIGNED: Final[ThirtyTwoBits] = cached_hresult(0x80070197, 'HRESULT_FROM_WIN32(ERROR_VOLUME_NOT_CLUSTER_ALIGNED)')
    WIN32__ERROR_NO_PHYSICALLY_ALIGNED_FREE_SPACE_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80070198, 'HRESULT_FROM_WIN32(ERROR_NO_PHYSICALLY_ALIGNED_FREE_SPACE_FOUND)')
    WIN32__ERROR_APPX_FILE_NOT_ENCRYPTED: Final[ThirtyTwoBits] = cached_hresult(0x80070199, 'HRESULT_FROM_WIN32(ERROR_APPX_FILE_NOT_ENCRYPTED)')
    WIN32__ERROR_RWRAW_ENCRYPTED_FILE_NOT_ENCRYPTED: Final[ThirtyTwoBits] = cached_hresult(0x8007019A, 'HRESULT_FROM_WIN32(ERROR_RWRAW_ENCRYPTED_FILE_NOT_ENCRYPTED)')
    WIN32__ERROR_RWRAW_ENCRYPTED_INVALID_EDATAINFO_FILEOFFSET: Final[ThirtyTwoBits] = cached_hresult(0x8007019B, 'HRESULT_FROM_WIN32(ERROR_RWRAW_ENCRYPTED_INVALID_EDATAINFO_FILEOFFSET)')
    WIN32__ERROR_RWRAW_ENCRYPTED_INVALID_EDATAINFO_FILERANGE: Final[ThirtyTwoBits] = cached_hresult(0x8007019C, 'HRESULT_FROM_WIN32(ERROR_RWRAW_ENCRYPTED_INVALID_EDATAINFO_FILERANGE)')
    WIN32__ERROR_RWRAW_ENCRYPTED_INVALID_EDATAINFO_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0x8007019D, 'HRESULT_FROM_WIN32(ERROR_RWRAW_ENCRYPTED_INVALID_EDATAINFO_PARAMETER)')
    WIN32__ERROR_LINUX_SUBSYSTEM_NOT_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0x8007019E, 'HRESULT_FROM_WIN32(ERROR_LINUX_SUBSYSTEM_NOT_PRESENT)')
    WIN32__ERROR_FT_READ_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8007019F, 'HRESULT_FROM_WIN32(ERROR_FT_READ_FAILURE)')
    WIN32__ERROR_STORAGE_RESERVE_ID_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800701A0, 'HRESULT_FROM_WIN32(ERROR_STORAGE_RESERVE_ID_INVALID)')
    WIN32__ERROR_STORAGE_RESERVE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x800701A1, 'HRESULT_FROM_WIN32(ERROR_STORAGE_RESERVE_DOES_NOT_EXIST)')
    WIN32__ERROR_STORAGE_RESERVE_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800701A2, 'HRESULT_FROM_WIN32(ERROR_STORAGE_RESERVE_ALREADY_EXISTS)')
    WIN32__ERROR_STORAGE_RESERVE_NOT_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0x800701A3, 'HRESULT_FROM_WIN32(ERROR_STORAGE_RESERVE_NOT_EMPTY)')
    WIN32__ERROR_NOT_A_DAX_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0x800701A4, 'HRESULT_FROM_WIN32(ERROR_NOT_A_DAX_VOLUME)')
    WIN32__ERROR_NOT_DAX_MAPPABLE: Final[ThirtyTwoBits] = cached_hresult(0x800701A5, 'HRESULT_FROM_WIN32(ERROR_NOT_DAX_MAPPABLE)')
    WIN32__ERROR_TIME_SENSITIVE_THREAD: Final[ThirtyTwoBits] = cached_hresult(0x800701A6, 'HRESULT_FROM_WIN32(ERROR_TIME_SENSITIVE_THREAD)')
    WIN32__ERROR_DPL_NOT_SUPPORTED_FOR_USER: Final[ThirtyTwoBits] = cached_hresult(0x800701A7, 'HRESULT_FROM_WIN32(ERROR_DPL_NOT_SUPPORTED_FOR_USER)')
    WIN32__ERROR_CASE_DIFFERING_NAMES_IN_DIR: Final[ThirtyTwoBits] = cached_hresult(0x800701A8, 'HRESULT_FROM_WIN32(ERROR_CASE_DIFFERING_NAMES_IN_DIR)')
    WIN32__ERROR_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x800701A9, 'HRESULT_FROM_WIN32(ERROR_FILE_NOT_SUPPORTED)')
    WIN32__ERROR_CLOUD_FILE_REQUEST_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x800701AA, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_REQUEST_TIMEOUT)')
    WIN32__ERROR_NO_TASK_QUEUE: Final[ThirtyTwoBits] = cached_hresult(0x800701AB, 'HRESULT_FROM_WIN32(ERROR_NO_TASK_QUEUE)')
    WIN32__ERROR_SRC_SRV_DLL_LOAD_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800701AC, 'HRESULT_FROM_WIN32(ERROR_SRC_SRV_DLL_LOAD_FAILED)')
    WIN32__ERROR_NOT_SUPPORTED_WITH_BTT: Final[ThirtyTwoBits] = cached_hresult(0x800701AD, 'HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED_WITH_BTT)')
    WIN32__ERROR_ENCRYPTION_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x800701AE, 'HRESULT_FROM_WIN32(ERROR_ENCRYPTION_DISABLED)')
    WIN32__ERROR_ENCRYPTING_METADATA_DISALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x800701AF, 'HRESULT_FROM_WIN32(ERROR_ENCRYPTING_METADATA_DISALLOWED)')
    WIN32__ERROR_CANT_CLEAR_ENCRYPTION_FLAG: Final[ThirtyTwoBits] = cached_hresult(0x800701B0, 'HRESULT_FROM_WIN32(ERROR_CANT_CLEAR_ENCRYPTION_FLAG)')
    WIN32__ERROR_NO_SUCH_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0x800701B1, 'HRESULT_FROM_WIN32(ERROR_NO_SUCH_DEVICE)')
    WIN32__ERROR_CLOUD_FILE_DEHYDRATION_DISALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x800701B2, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_DEHYDRATION_DISALLOWED)')
    WIN32__ERROR_FILE_SNAP_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x800701B3, 'HRESULT_FROM_WIN32(ERROR_FILE_SNAP_IN_PROGRESS)')
    WIN32__ERROR_FILE_SNAP_USER_SECTION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x800701B4, 'HRESULT_FROM_WIN32(ERROR_FILE_SNAP_USER_SECTION_NOT_SUPPORTED)')
    WIN32__ERROR_FILE_SNAP_MODIFY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x800701B5, 'HRESULT_FROM_WIN32(ERROR_FILE_SNAP_MODIFY_NOT_SUPPORTED)')
    WIN32__ERROR_FILE_SNAP_IO_NOT_COORDINATED: Final[ThirtyTwoBits] = cached_hresult(0x800701B6, 'HRESULT_FROM_WIN32(ERROR_FILE_SNAP_IO_NOT_COORDINATED)')
    WIN32__ERROR_FILE_SNAP_UNEXPECTED_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800701B7, 'HRESULT_FROM_WIN32(ERROR_FILE_SNAP_UNEXPECTED_ERROR)')
    WIN32__ERROR_FILE_SNAP_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0x800701B8, 'HRESULT_FROM_WIN32(ERROR_FILE_SNAP_INVALID_PARAMETER)')
    WIN32__ERROR_UNSATISFIED_DEPENDENCIES: Final[ThirtyTwoBits] = cached_hresult(0x800701B9, 'HRESULT_FROM_WIN32(ERROR_UNSATISFIED_DEPENDENCIES)')
    WIN32__ERROR_CASE_SENSITIVE_PATH: Final[ThirtyTwoBits] = cached_hresult(0x800701BA, 'HRESULT_FROM_WIN32(ERROR_CASE_SENSITIVE_PATH)')
    WIN32__ERROR_UNEXPECTED_NTCACHEMANAGER_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800701BB, 'HRESULT_FROM_WIN32(ERROR_UNEXPECTED_NTCACHEMANAGER_ERROR)')
    WIN32__ERROR_LINUX_SUBSYSTEM_UPDATE_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x800701BC, 'HRESULT_FROM_WIN32(ERROR_LINUX_SUBSYSTEM_UPDATE_REQUIRED)')
    WIN32__ERROR_DLP_POLICY_WARNS_AGAINST_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x800701BD, 'HRESULT_FROM_WIN32(ERROR_DLP_POLICY_WARNS_AGAINST_OPERATION)')
    WIN32__ERROR_DLP_POLICY_DENIES_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x800701BE, 'HRESULT_FROM_WIN32(ERROR_DLP_POLICY_DENIES_OPERATION)')
    WIN32__ERROR_SECURITY_DENIES_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x800701BF, 'HRESULT_FROM_WIN32(ERROR_SECURITY_DENIES_OPERATION)')
    WIN32__ERROR_UNTRUSTED_MOUNT_POINT: Final[ThirtyTwoBits] = cached_hresult(0x800701C0, 'HRESULT_FROM_WIN32(ERROR_UNTRUSTED_MOUNT_POINT)')
    WIN32__ERROR_DLP_POLICY_SILENTLY_FAIL: Final[ThirtyTwoBits] = cached_hresult(0x800701C1, 'HRESULT_FROM_WIN32(ERROR_DLP_POLICY_SILENTLY_FAIL)')
    WIN32__ERROR_CAPAUTHZ_NOT_DEVUNLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x800701C2, 'HRESULT_FROM_WIN32(ERROR_CAPAUTHZ_NOT_DEVUNLOCKED)')
    WIN32__ERROR_CAPAUTHZ_CHANGE_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x800701C3, 'HRESULT_FROM_WIN32(ERROR_CAPAUTHZ_CHANGE_TYPE)')
    WIN32__ERROR_CAPAUTHZ_NOT_PROVISIONED: Final[ThirtyTwoBits] = cached_hresult(0x800701C4, 'HRESULT_FROM_WIN32(ERROR_CAPAUTHZ_NOT_PROVISIONED)')
    WIN32__ERROR_CAPAUTHZ_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_hresult(0x800701C5, 'HRESULT_FROM_WIN32(ERROR_CAPAUTHZ_NOT_AUTHORIZED)')
    WIN32__ERROR_CAPAUTHZ_NO_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x800701C6, 'HRESULT_FROM_WIN32(ERROR_CAPAUTHZ_NO_POLICY)')
    WIN32__ERROR_CAPAUTHZ_DB_CORRUPTED: Final[ThirtyTwoBits] = cached_hresult(0x800701C7, 'HRESULT_FROM_WIN32(ERROR_CAPAUTHZ_DB_CORRUPTED)')
    WIN32__ERROR_CAPAUTHZ_SCCD_INVALID_CATALOG: Final[ThirtyTwoBits] = cached_hresult(0x800701C8, 'HRESULT_FROM_WIN32(ERROR_CAPAUTHZ_SCCD_INVALID_CATALOG)')
    WIN32__ERROR_CAPAUTHZ_SCCD_NO_AUTH_ENTITY: Final[ThirtyTwoBits] = cached_hresult(0x800701C9, 'HRESULT_FROM_WIN32(ERROR_CAPAUTHZ_SCCD_NO_AUTH_ENTITY)')
    WIN32__ERROR_CAPAUTHZ_SCCD_PARSE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800701CA, 'HRESULT_FROM_WIN32(ERROR_CAPAUTHZ_SCCD_PARSE_ERROR)')
    WIN32__ERROR_CAPAUTHZ_SCCD_DEV_MODE_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x800701CB, 'HRESULT_FROM_WIN32(ERROR_CAPAUTHZ_SCCD_DEV_MODE_REQUIRED)')
    WIN32__ERROR_CAPAUTHZ_SCCD_NO_CAPABILITY_MATCH: Final[ThirtyTwoBits] = cached_hresult(0x800701CC, 'HRESULT_FROM_WIN32(ERROR_CAPAUTHZ_SCCD_NO_CAPABILITY_MATCH)')
    WIN32__ERROR_CIMFS_IMAGE_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x800701D6, 'HRESULT_FROM_WIN32(ERROR_CIMFS_IMAGE_CORRUPT)')
    WIN32__ERROR_CIMFS_IMAGE_VERSION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x800701D7, 'HRESULT_FROM_WIN32(ERROR_CIMFS_IMAGE_VERSION_NOT_SUPPORTED)')
    WIN32__ERROR_STORAGE_STACK_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x800701D8, 'HRESULT_FROM_WIN32(ERROR_STORAGE_STACK_ACCESS_DENIED)')
    WIN32__ERROR_INSUFFICIENT_VIRTUAL_ADDR_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0x800701D9, 'HRESULT_FROM_WIN32(ERROR_INSUFFICIENT_VIRTUAL_ADDR_RESOURCES)')
    WIN32__ERROR_INDEX_OUT_OF_BOUNDS: Final[ThirtyTwoBits] = cached_hresult(0x800701DA, 'HRESULT_FROM_WIN32(ERROR_INDEX_OUT_OF_BOUNDS)')
    WIN32__ERROR_CLOUD_FILE_US_MESSAGE_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x800701DB, 'HRESULT_FROM_WIN32(ERROR_CLOUD_FILE_US_MESSAGE_TIMEOUT)')
    WIN32__ERROR_NOT_A_DEV_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0x800701DC, 'HRESULT_FROM_WIN32(ERROR_NOT_A_DEV_VOLUME)')
    WIN32__ERROR_FS_GUID_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x800701DD, 'HRESULT_FROM_WIN32(ERROR_FS_GUID_MISMATCH)')
    WIN32__ERROR_CANT_ATTACH_TO_DEV_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0x800701DE, 'HRESULT_FROM_WIN32(ERROR_CANT_ATTACH_TO_DEV_VOLUME)')
    WIN32__ERROR_INVALID_CONFIG_VALUE: Final[ThirtyTwoBits] = cached_hresult(0x800701DF, 'HRESULT_FROM_WIN32(ERROR_INVALID_CONFIG_VALUE)')
    WIN32__ERROR_PNP_QUERY_REMOVE_DEVICE_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x800701E0, 'HRESULT_FROM_WIN32(ERROR_PNP_QUERY_REMOVE_DEVICE_TIMEOUT)')
    WIN32__ERROR_PNP_QUERY_REMOVE_RELATED_DEVICE_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x800701E1, 'HRESULT_FROM_WIN32(ERROR_PNP_QUERY_REMOVE_RELATED_DEVICE_TIMEOUT)')
    WIN32__ERROR_PNP_QUERY_REMOVE_UNRELATED_DEVICE_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x800701E2, 'HRESULT_FROM_WIN32(ERROR_PNP_QUERY_REMOVE_UNRELATED_DEVICE_TIMEOUT)')
    WIN32__ERROR_DEVICE_HARDWARE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800701E3, 'HRESULT_FROM_WIN32(ERROR_DEVICE_HARDWARE_ERROR)')
    WIN32__ERROR_INVALID_ADDRESS: Final[ThirtyTwoBits] = cached_hresult(0x800701E7, 'HRESULT_FROM_WIN32(ERROR_INVALID_ADDRESS)')
    WIN32__ERROR_HAS_SYSTEM_CRITICAL_FILES: Final[ThirtyTwoBits] = cached_hresult(0x800701E8, 'HRESULT_FROM_WIN32(ERROR_HAS_SYSTEM_CRITICAL_FILES)')
    WIN32__ERROR_ENCRYPTED_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x800701E9, 'HRESULT_FROM_WIN32(ERROR_ENCRYPTED_FILE_NOT_SUPPORTED)')
    WIN32__ERROR_SPARSE_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x800701EA, 'HRESULT_FROM_WIN32(ERROR_SPARSE_FILE_NOT_SUPPORTED)')
    WIN32__ERROR_PAGEFILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x800701EB, 'HRESULT_FROM_WIN32(ERROR_PAGEFILE_NOT_SUPPORTED)')
    WIN32__ERROR_VOLUME_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x800701EC, 'HRESULT_FROM_WIN32(ERROR_VOLUME_NOT_SUPPORTED)')
    WIN32__ERROR_NOT_SUPPORTED_WITH_BYPASSIO: Final[ThirtyTwoBits] = cached_hresult(0x800701ED, 'HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED_WITH_BYPASSIO)')
    WIN32__ERROR_NO_BYPASSIO_DRIVER_SUPPORT: Final[ThirtyTwoBits] = cached_hresult(0x800701EE, 'HRESULT_FROM_WIN32(ERROR_NO_BYPASSIO_DRIVER_SUPPORT)')
    WIN32__ERROR_NOT_SUPPORTED_WITH_ENCRYPTION: Final[ThirtyTwoBits] = cached_hresult(0x800701EF, 'HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED_WITH_ENCRYPTION)')
    WIN32__ERROR_NOT_SUPPORTED_WITH_COMPRESSION: Final[ThirtyTwoBits] = cached_hresult(0x800701F0, 'HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED_WITH_COMPRESSION)')
    WIN32__ERROR_NOT_SUPPORTED_WITH_REPLICATION: Final[ThirtyTwoBits] = cached_hresult(0x800701F1, 'HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED_WITH_REPLICATION)')
    WIN32__ERROR_NOT_SUPPORTED_WITH_DEDUPLICATION: Final[ThirtyTwoBits] = cached_hresult(0x800701F2, 'HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED_WITH_DEDUPLICATION)')
    WIN32__ERROR_NOT_SUPPORTED_WITH_AUDITING: Final[ThirtyTwoBits] = cached_hresult(0x800701F3, 'HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED_WITH_AUDITING)')
    WIN32__ERROR_USER_PROFILE_LOAD: Final[ThirtyTwoBits] = cached_hresult(0x800701F4, 'HRESULT_FROM_WIN32(ERROR_USER_PROFILE_LOAD)')
    WIN32__ERROR_SESSION_KEY_TOO_SHORT: Final[ThirtyTwoBits] = cached_hresult(0x800701F5, 'HRESULT_FROM_WIN32(ERROR_SESSION_KEY_TOO_SHORT)')
    WIN32__ERROR_ACCESS_DENIED_APPDATA: Final[ThirtyTwoBits] = cached_hresult(0x800701F6, 'HRESULT_FROM_WIN32(ERROR_ACCESS_DENIED_APPDATA)')
    WIN32__ERROR_NOT_SUPPORTED_WITH_MONITORING: Final[ThirtyTwoBits] = cached_hresult(0x800701F7, 'HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED_WITH_MONITORING)')
    WIN32__ERROR_NOT_SUPPORTED_WITH_SNAPSHOT: Final[ThirtyTwoBits] = cached_hresult(0x800701F8, 'HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED_WITH_SNAPSHOT)')
    WIN32__ERROR_NOT_SUPPORTED_WITH_VIRTUALIZATION: Final[ThirtyTwoBits] = cached_hresult(0x800701F9, 'HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED_WITH_VIRTUALIZATION)')
    WIN32__ERROR_BYPASSIO_FLT_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x800701FA, 'HRESULT_FROM_WIN32(ERROR_BYPASSIO_FLT_NOT_SUPPORTED)')
    WIN32__ERROR_DEVICE_RESET_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x800701FB, 'HRESULT_FROM_WIN32(ERROR_DEVICE_RESET_REQUIRED)')
    WIN32__ERROR_VOLUME_WRITE_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x800701FC, 'HRESULT_FROM_WIN32(ERROR_VOLUME_WRITE_ACCESS_DENIED)')
    WIN32__ERROR_NOT_SUPPORTED_WITH_CACHED_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x800701FD, 'HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED_WITH_CACHED_HANDLE)')
    WIN32__ERROR_FS_METADATA_INCONSISTENT: Final[ThirtyTwoBits] = cached_hresult(0x800701FE, 'HRESULT_FROM_WIN32(ERROR_FS_METADATA_INCONSISTENT)')
    WIN32__ERROR_BLOCK_WEAK_REFERENCE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800701FF, 'HRESULT_FROM_WIN32(ERROR_BLOCK_WEAK_REFERENCE_INVALID)')
    WIN32__ERROR_BLOCK_SOURCE_WEAK_REFERENCE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80070200, 'HRESULT_FROM_WIN32(ERROR_BLOCK_SOURCE_WEAK_REFERENCE_INVALID)')
    WIN32__ERROR_BLOCK_TARGET_WEAK_REFERENCE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80070201, 'HRESULT_FROM_WIN32(ERROR_BLOCK_TARGET_WEAK_REFERENCE_INVALID)')
    WIN32__ERROR_BLOCK_SHARED: Final[ThirtyTwoBits] = cached_hresult(0x80070202, 'HRESULT_FROM_WIN32(ERROR_BLOCK_SHARED)')
    WIN32__ERROR_VOLUME_UPGRADE_NOT_NEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80070203, 'HRESULT_FROM_WIN32(ERROR_VOLUME_UPGRADE_NOT_NEEDED)')
    WIN32__ERROR_VOLUME_UPGRADE_PENDING: Final[ThirtyTwoBits] = cached_hresult(0x80070204, 'HRESULT_FROM_WIN32(ERROR_VOLUME_UPGRADE_PENDING)')
    WIN32__ERROR_VOLUME_UPGRADE_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80070205, 'HRESULT_FROM_WIN32(ERROR_VOLUME_UPGRADE_DISABLED)')
    WIN32__ERROR_VOLUME_UPGRADE_DISABLED_TILL_OS_DOWNGRADE_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x80070206, 'HRESULT_FROM_WIN32(ERROR_VOLUME_UPGRADE_DISABLED_TILL_OS_DOWNGRADE_EXPIRED)')
    WIN32__ERROR_ARITHMETIC_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0x80070216, 'HRESULT_FROM_WIN32(ERROR_ARITHMETIC_OVERFLOW)')
    WIN32__ERROR_PIPE_CONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x80070217, 'HRESULT_FROM_WIN32(ERROR_PIPE_CONNECTED)')
    WIN32__ERROR_PIPE_LISTENING: Final[ThirtyTwoBits] = cached_hresult(0x80070218, 'HRESULT_FROM_WIN32(ERROR_PIPE_LISTENING)')
    WIN32__ERROR_VERIFIER_STOP: Final[ThirtyTwoBits] = cached_hresult(0x80070219, 'HRESULT_FROM_WIN32(ERROR_VERIFIER_STOP)')
    WIN32__ERROR_ABIOS_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8007021A, 'HRESULT_FROM_WIN32(ERROR_ABIOS_ERROR)')
    WIN32__ERROR_WX86_WARNING: Final[ThirtyTwoBits] = cached_hresult(0x8007021B, 'HRESULT_FROM_WIN32(ERROR_WX86_WARNING)')
    WIN32__ERROR_WX86_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8007021C, 'HRESULT_FROM_WIN32(ERROR_WX86_ERROR)')
    WIN32__ERROR_TIMER_NOT_CANCELED: Final[ThirtyTwoBits] = cached_hresult(0x8007021D, 'HRESULT_FROM_WIN32(ERROR_TIMER_NOT_CANCELED)')
    WIN32__ERROR_UNWIND: Final[ThirtyTwoBits] = cached_hresult(0x8007021E, 'HRESULT_FROM_WIN32(ERROR_UNWIND)')
    WIN32__ERROR_BAD_STACK: Final[ThirtyTwoBits] = cached_hresult(0x8007021F, 'HRESULT_FROM_WIN32(ERROR_BAD_STACK)')
    WIN32__ERROR_INVALID_UNWIND_TARGET: Final[ThirtyTwoBits] = cached_hresult(0x80070220, 'HRESULT_FROM_WIN32(ERROR_INVALID_UNWIND_TARGET)')
    WIN32__ERROR_INVALID_PORT_ATTRIBUTES: Final[ThirtyTwoBits] = cached_hresult(0x80070221, 'HRESULT_FROM_WIN32(ERROR_INVALID_PORT_ATTRIBUTES)')
    WIN32__ERROR_PORT_MESSAGE_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0x80070222, 'HRESULT_FROM_WIN32(ERROR_PORT_MESSAGE_TOO_LONG)')
    WIN32__ERROR_INVALID_QUOTA_LOWER: Final[ThirtyTwoBits] = cached_hresult(0x80070223, 'HRESULT_FROM_WIN32(ERROR_INVALID_QUOTA_LOWER)')
    WIN32__ERROR_DEVICE_ALREADY_ATTACHED: Final[ThirtyTwoBits] = cached_hresult(0x80070224, 'HRESULT_FROM_WIN32(ERROR_DEVICE_ALREADY_ATTACHED)')
    WIN32__ERROR_INSTRUCTION_MISALIGNMENT: Final[ThirtyTwoBits] = cached_hresult(0x80070225, 'HRESULT_FROM_WIN32(ERROR_INSTRUCTION_MISALIGNMENT)')
    WIN32__ERROR_PROFILING_NOT_STARTED: Final[ThirtyTwoBits] = cached_hresult(0x80070226, 'HRESULT_FROM_WIN32(ERROR_PROFILING_NOT_STARTED)')
    WIN32__ERROR_PROFILING_NOT_STOPPED: Final[ThirtyTwoBits] = cached_hresult(0x80070227, 'HRESULT_FROM_WIN32(ERROR_PROFILING_NOT_STOPPED)')
    WIN32__ERROR_COULD_NOT_INTERPRET: Final[ThirtyTwoBits] = cached_hresult(0x80070228, 'HRESULT_FROM_WIN32(ERROR_COULD_NOT_INTERPRET)')
    WIN32__ERROR_PROFILING_AT_LIMIT: Final[ThirtyTwoBits] = cached_hresult(0x80070229, 'HRESULT_FROM_WIN32(ERROR_PROFILING_AT_LIMIT)')
    WIN32__ERROR_CANT_WAIT: Final[ThirtyTwoBits] = cached_hresult(0x8007022A, 'HRESULT_FROM_WIN32(ERROR_CANT_WAIT)')
    WIN32__ERROR_CANT_TERMINATE_SELF: Final[ThirtyTwoBits] = cached_hresult(0x8007022B, 'HRESULT_FROM_WIN32(ERROR_CANT_TERMINATE_SELF)')
    WIN32__ERROR_UNEXPECTED_MM_CREATE_ERR: Final[ThirtyTwoBits] = cached_hresult(0x8007022C, 'HRESULT_FROM_WIN32(ERROR_UNEXPECTED_MM_CREATE_ERR)')
    WIN32__ERROR_UNEXPECTED_MM_MAP_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8007022D, 'HRESULT_FROM_WIN32(ERROR_UNEXPECTED_MM_MAP_ERROR)')
    WIN32__ERROR_UNEXPECTED_MM_EXTEND_ERR: Final[ThirtyTwoBits] = cached_hresult(0x8007022E, 'HRESULT_FROM_WIN32(ERROR_UNEXPECTED_MM_EXTEND_ERR)')
    WIN32__ERROR_BAD_FUNCTION_TABLE: Final[ThirtyTwoBits] = cached_hresult(0x8007022F, 'HRESULT_FROM_WIN32(ERROR_BAD_FUNCTION_TABLE)')
    WIN32__ERROR_NO_GUID_TRANSLATION: Final[ThirtyTwoBits] = cached_hresult(0x80070230, 'HRESULT_FROM_WIN32(ERROR_NO_GUID_TRANSLATION)')
    WIN32__ERROR_INVALID_LDT_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x80070231, 'HRESULT_FROM_WIN32(ERROR_INVALID_LDT_SIZE)')
    WIN32__ERROR_INVALID_LDT_OFFSET: Final[ThirtyTwoBits] = cached_hresult(0x80070233, 'HRESULT_FROM_WIN32(ERROR_INVALID_LDT_OFFSET)')
    WIN32__ERROR_INVALID_LDT_DESCRIPTOR: Final[ThirtyTwoBits] = cached_hresult(0x80070234, 'HRESULT_FROM_WIN32(ERROR_INVALID_LDT_DESCRIPTOR)')
    WIN32__ERROR_TOO_MANY_THREADS: Final[ThirtyTwoBits] = cached_hresult(0x80070235, 'HRESULT_FROM_WIN32(ERROR_TOO_MANY_THREADS)')
    WIN32__ERROR_THREAD_NOT_IN_PROCESS: Final[ThirtyTwoBits] = cached_hresult(0x80070236, 'HRESULT_FROM_WIN32(ERROR_THREAD_NOT_IN_PROCESS)')
    WIN32__ERROR_PAGEFILE_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80070237, 'HRESULT_FROM_WIN32(ERROR_PAGEFILE_QUOTA_EXCEEDED)')
    WIN32__ERROR_LOGON_SERVER_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x80070238, 'HRESULT_FROM_WIN32(ERROR_LOGON_SERVER_CONFLICT)')
    WIN32__ERROR_SYNCHRONIZATION_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80070239, 'HRESULT_FROM_WIN32(ERROR_SYNCHRONIZATION_REQUIRED)')
    WIN32__ERROR_NET_OPEN_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8007023A, 'HRESULT_FROM_WIN32(ERROR_NET_OPEN_FAILED)')
    WIN32__ERROR_IO_PRIVILEGE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8007023B, 'HRESULT_FROM_WIN32(ERROR_IO_PRIVILEGE_FAILED)')
    WIN32__ERROR_CONTROL_C_EXIT: Final[ThirtyTwoBits] = cached_hresult(0x8007023C, 'HRESULT_FROM_WIN32(ERROR_CONTROL_C_EXIT)')
    WIN32__ERROR_MISSING_SYSTEMFILE: Final[ThirtyTwoBits] = cached_hresult(0x8007023D, 'HRESULT_FROM_WIN32(ERROR_MISSING_SYSTEMFILE)')
    WIN32__ERROR_UNHANDLED_EXCEPTION: Final[ThirtyTwoBits] = cached_hresult(0x8007023E, 'HRESULT_FROM_WIN32(ERROR_UNHANDLED_EXCEPTION)')
    WIN32__ERROR_APP_INIT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8007023F, 'HRESULT_FROM_WIN32(ERROR_APP_INIT_FAILURE)')
    WIN32__ERROR_PAGEFILE_CREATE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80070240, 'HRESULT_FROM_WIN32(ERROR_PAGEFILE_CREATE_FAILED)')
    WIN32__ERROR_INVALID_IMAGE_HASH: Final[ThirtyTwoBits] = cached_hresult(0x80070241, 'HRESULT_FROM_WIN32(ERROR_INVALID_IMAGE_HASH)')
    WIN32__ERROR_NO_PAGEFILE: Final[ThirtyTwoBits] = cached_hresult(0x80070242, 'HRESULT_FROM_WIN32(ERROR_NO_PAGEFILE)')
    WIN32__ERROR_ILLEGAL_FLOAT_CONTEXT: Final[ThirtyTwoBits] = cached_hresult(0x80070243, 'HRESULT_FROM_WIN32(ERROR_ILLEGAL_FLOAT_CONTEXT)')
    WIN32__ERROR_NO_EVENT_PAIR: Final[ThirtyTwoBits] = cached_hresult(0x80070244, 'HRESULT_FROM_WIN32(ERROR_NO_EVENT_PAIR)')
    WIN32__ERROR_DOMAIN_CTRLR_CONFIG_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80070245, 'HRESULT_FROM_WIN32(ERROR_DOMAIN_CTRLR_CONFIG_ERROR)')
    WIN32__ERROR_ILLEGAL_CHARACTER: Final[ThirtyTwoBits] = cached_hresult(0x80070246, 'HRESULT_FROM_WIN32(ERROR_ILLEGAL_CHARACTER)')
    WIN32__ERROR_UNDEFINED_CHARACTER: Final[ThirtyTwoBits] = cached_hresult(0x80070247, 'HRESULT_FROM_WIN32(ERROR_UNDEFINED_CHARACTER)')
    WIN32__ERROR_FLOPPY_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0x80070248, 'HRESULT_FROM_WIN32(ERROR_FLOPPY_VOLUME)')
    WIN32__ERROR_BIOS_FAILED_TO_CONNECT_INTERRUPT: Final[ThirtyTwoBits] = cached_hresult(0x80070249, 'HRESULT_FROM_WIN32(ERROR_BIOS_FAILED_TO_CONNECT_INTERRUPT)')
    WIN32__ERROR_BACKUP_CONTROLLER: Final[ThirtyTwoBits] = cached_hresult(0x8007024A, 'HRESULT_FROM_WIN32(ERROR_BACKUP_CONTROLLER)')
    WIN32__ERROR_MUTANT_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8007024B, 'HRESULT_FROM_WIN32(ERROR_MUTANT_LIMIT_EXCEEDED)')
    WIN32__ERROR_FS_DRIVER_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8007024C, 'HRESULT_FROM_WIN32(ERROR_FS_DRIVER_REQUIRED)')
    WIN32__ERROR_CANNOT_LOAD_REGISTRY_FILE: Final[ThirtyTwoBits] = cached_hresult(0x8007024D, 'HRESULT_FROM_WIN32(ERROR_CANNOT_LOAD_REGISTRY_FILE)')
    WIN32__ERROR_DEBUG_ATTACH_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8007024E, 'HRESULT_FROM_WIN32(ERROR_DEBUG_ATTACH_FAILED)')
    WIN32__ERROR_SYSTEM_PROCESS_TERMINATED: Final[ThirtyTwoBits] = cached_hresult(0x8007024F, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_PROCESS_TERMINATED)')
    WIN32__ERROR_DATA_NOT_ACCEPTED: Final[ThirtyTwoBits] = cached_hresult(0x80070250, 'HRESULT_FROM_WIN32(ERROR_DATA_NOT_ACCEPTED)')
    WIN32__ERROR_VDM_HARD_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80070251, 'HRESULT_FROM_WIN32(ERROR_VDM_HARD_ERROR)')
    WIN32__ERROR_DRIVER_CANCEL_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x80070252, 'HRESULT_FROM_WIN32(ERROR_DRIVER_CANCEL_TIMEOUT)')
    WIN32__ERROR_REPLY_MESSAGE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80070253, 'HRESULT_FROM_WIN32(ERROR_REPLY_MESSAGE_MISMATCH)')
    WIN32__ERROR_LOST_WRITEBEHIND_DATA: Final[ThirtyTwoBits] = cached_hresult(0x80070254, 'HRESULT_FROM_WIN32(ERROR_LOST_WRITEBEHIND_DATA)')
    WIN32__ERROR_CLIENT_SERVER_PARAMETERS_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80070255, 'HRESULT_FROM_WIN32(ERROR_CLIENT_SERVER_PARAMETERS_INVALID)')
    WIN32__ERROR_NOT_TINY_STREAM: Final[ThirtyTwoBits] = cached_hresult(0x80070256, 'HRESULT_FROM_WIN32(ERROR_NOT_TINY_STREAM)')
    WIN32__ERROR_STACK_OVERFLOW_READ: Final[ThirtyTwoBits] = cached_hresult(0x80070257, 'HRESULT_FROM_WIN32(ERROR_STACK_OVERFLOW_READ)')
    WIN32__ERROR_CONVERT_TO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0x80070258, 'HRESULT_FROM_WIN32(ERROR_CONVERT_TO_LARGE)')
    WIN32__ERROR_FOUND_OUT_OF_SCOPE: Final[ThirtyTwoBits] = cached_hresult(0x80070259, 'HRESULT_FROM_WIN32(ERROR_FOUND_OUT_OF_SCOPE)')
    WIN32__ERROR_ALLOCATE_BUCKET: Final[ThirtyTwoBits] = cached_hresult(0x8007025A, 'HRESULT_FROM_WIN32(ERROR_ALLOCATE_BUCKET)')
    WIN32__ERROR_MARSHALL_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0x8007025B, 'HRESULT_FROM_WIN32(ERROR_MARSHALL_OVERFLOW)')
    WIN32__ERROR_INVALID_VARIANT: Final[ThirtyTwoBits] = cached_hresult(0x8007025C, 'HRESULT_FROM_WIN32(ERROR_INVALID_VARIANT)')
    WIN32__ERROR_BAD_COMPRESSION_BUFFER: Final[ThirtyTwoBits] = cached_hresult(0x8007025D, 'HRESULT_FROM_WIN32(ERROR_BAD_COMPRESSION_BUFFER)')
    WIN32__ERROR_AUDIT_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8007025E, 'HRESULT_FROM_WIN32(ERROR_AUDIT_FAILED)')
    WIN32__ERROR_TIMER_RESOLUTION_NOT_SET: Final[ThirtyTwoBits] = cached_hresult(0x8007025F, 'HRESULT_FROM_WIN32(ERROR_TIMER_RESOLUTION_NOT_SET)')
    WIN32__ERROR_INSUFFICIENT_LOGON_INFO: Final[ThirtyTwoBits] = cached_hresult(0x80070260, 'HRESULT_FROM_WIN32(ERROR_INSUFFICIENT_LOGON_INFO)')
    WIN32__ERROR_BAD_DLL_ENTRYPOINT: Final[ThirtyTwoBits] = cached_hresult(0x80070261, 'HRESULT_FROM_WIN32(ERROR_BAD_DLL_ENTRYPOINT)')
    WIN32__ERROR_BAD_SERVICE_ENTRYPOINT: Final[ThirtyTwoBits] = cached_hresult(0x80070262, 'HRESULT_FROM_WIN32(ERROR_BAD_SERVICE_ENTRYPOINT)')
    WIN32__ERROR_IP_ADDRESS_CONFLICT1: Final[ThirtyTwoBits] = cached_hresult(0x80070263, 'HRESULT_FROM_WIN32(ERROR_IP_ADDRESS_CONFLICT1)')
    WIN32__ERROR_IP_ADDRESS_CONFLICT2: Final[ThirtyTwoBits] = cached_hresult(0x80070264, 'HRESULT_FROM_WIN32(ERROR_IP_ADDRESS_CONFLICT2)')
    WIN32__ERROR_REGISTRY_QUOTA_LIMIT: Final[ThirtyTwoBits] = cached_hresult(0x80070265, 'HRESULT_FROM_WIN32(ERROR_REGISTRY_QUOTA_LIMIT)')
    WIN32__ERROR_NO_CALLBACK_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0x80070266, 'HRESULT_FROM_WIN32(ERROR_NO_CALLBACK_ACTIVE)')
    WIN32__ERROR_PWD_TOO_SHORT: Final[ThirtyTwoBits] = cached_hresult(0x80070267, 'HRESULT_FROM_WIN32(ERROR_PWD_TOO_SHORT)')
    WIN32__ERROR_PWD_TOO_RECENT: Final[ThirtyTwoBits] = cached_hresult(0x80070268, 'HRESULT_FROM_WIN32(ERROR_PWD_TOO_RECENT)')
    WIN32__ERROR_PWD_HISTORY_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x80070269, 'HRESULT_FROM_WIN32(ERROR_PWD_HISTORY_CONFLICT)')
    WIN32__ERROR_UNSUPPORTED_COMPRESSION: Final[ThirtyTwoBits] = cached_hresult(0x8007026A, 'HRESULT_FROM_WIN32(ERROR_UNSUPPORTED_COMPRESSION)')
    WIN32__ERROR_INVALID_HW_PROFILE: Final[ThirtyTwoBits] = cached_hresult(0x8007026B, 'HRESULT_FROM_WIN32(ERROR_INVALID_HW_PROFILE)')
    WIN32__ERROR_INVALID_PLUGPLAY_DEVICE_PATH: Final[ThirtyTwoBits] = cached_hresult(0x8007026C, 'HRESULT_FROM_WIN32(ERROR_INVALID_PLUGPLAY_DEVICE_PATH)')
    WIN32__ERROR_QUOTA_LIST_INCONSISTENT: Final[ThirtyTwoBits] = cached_hresult(0x8007026D, 'HRESULT_FROM_WIN32(ERROR_QUOTA_LIST_INCONSISTENT)')
    WIN32__ERROR_EVALUATION_EXPIRATION: Final[ThirtyTwoBits] = cached_hresult(0x8007026E, 'HRESULT_FROM_WIN32(ERROR_EVALUATION_EXPIRATION)')
    WIN32__ERROR_ILLEGAL_DLL_RELOCATION: Final[ThirtyTwoBits] = cached_hresult(0x8007026F, 'HRESULT_FROM_WIN32(ERROR_ILLEGAL_DLL_RELOCATION)')
    WIN32__ERROR_DLL_INIT_FAILED_LOGOFF: Final[ThirtyTwoBits] = cached_hresult(0x80070270, 'HRESULT_FROM_WIN32(ERROR_DLL_INIT_FAILED_LOGOFF)')
    WIN32__ERROR_VALIDATE_CONTINUE: Final[ThirtyTwoBits] = cached_hresult(0x80070271, 'HRESULT_FROM_WIN32(ERROR_VALIDATE_CONTINUE)')
    WIN32__ERROR_NO_MORE_MATCHES: Final[ThirtyTwoBits] = cached_hresult(0x80070272, 'HRESULT_FROM_WIN32(ERROR_NO_MORE_MATCHES)')
    WIN32__ERROR_RANGE_LIST_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x80070273, 'HRESULT_FROM_WIN32(ERROR_RANGE_LIST_CONFLICT)')
    WIN32__ERROR_SERVER_SID_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80070274, 'HRESULT_FROM_WIN32(ERROR_SERVER_SID_MISMATCH)')
    WIN32__ERROR_CANT_ENABLE_DENY_ONLY: Final[ThirtyTwoBits] = cached_hresult(0x80070275, 'HRESULT_FROM_WIN32(ERROR_CANT_ENABLE_DENY_ONLY)')
    WIN32__ERROR_FLOAT_MULTIPLE_FAULTS: Final[ThirtyTwoBits] = cached_hresult(0x80070276, 'HRESULT_FROM_WIN32(ERROR_FLOAT_MULTIPLE_FAULTS)')
    WIN32__ERROR_FLOAT_MULTIPLE_TRAPS: Final[ThirtyTwoBits] = cached_hresult(0x80070277, 'HRESULT_FROM_WIN32(ERROR_FLOAT_MULTIPLE_TRAPS)')
    WIN32__ERROR_NOINTERFACE: Final[ThirtyTwoBits] = cached_hresult(0x80070278, 'HRESULT_FROM_WIN32(ERROR_NOINTERFACE)')
    WIN32__ERROR_DRIVER_FAILED_SLEEP: Final[ThirtyTwoBits] = cached_hresult(0x80070279, 'HRESULT_FROM_WIN32(ERROR_DRIVER_FAILED_SLEEP)')
    WIN32__ERROR_CORRUPT_SYSTEM_FILE: Final[ThirtyTwoBits] = cached_hresult(0x8007027A, 'HRESULT_FROM_WIN32(ERROR_CORRUPT_SYSTEM_FILE)')
    WIN32__ERROR_COMMITMENT_MINIMUM: Final[ThirtyTwoBits] = cached_hresult(0x8007027B, 'HRESULT_FROM_WIN32(ERROR_COMMITMENT_MINIMUM)')
    WIN32__ERROR_PNP_RESTART_ENUMERATION: Final[ThirtyTwoBits] = cached_hresult(0x8007027C, 'HRESULT_FROM_WIN32(ERROR_PNP_RESTART_ENUMERATION)')
    WIN32__ERROR_SYSTEM_IMAGE_BAD_SIGNATURE: Final[ThirtyTwoBits] = cached_hresult(0x8007027D, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_IMAGE_BAD_SIGNATURE)')
    WIN32__ERROR_PNP_REBOOT_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8007027E, 'HRESULT_FROM_WIN32(ERROR_PNP_REBOOT_REQUIRED)')
    WIN32__ERROR_INSUFFICIENT_POWER: Final[ThirtyTwoBits] = cached_hresult(0x8007027F, 'HRESULT_FROM_WIN32(ERROR_INSUFFICIENT_POWER)')
    WIN32__ERROR_MULTIPLE_FAULT_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x80070280, 'HRESULT_FROM_WIN32(ERROR_MULTIPLE_FAULT_VIOLATION)')
    WIN32__ERROR_SYSTEM_SHUTDOWN: Final[ThirtyTwoBits] = cached_hresult(0x80070281, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_SHUTDOWN)')
    WIN32__ERROR_PORT_NOT_SET: Final[ThirtyTwoBits] = cached_hresult(0x80070282, 'HRESULT_FROM_WIN32(ERROR_PORT_NOT_SET)')
    WIN32__ERROR_DS_VERSION_CHECK_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80070283, 'HRESULT_FROM_WIN32(ERROR_DS_VERSION_CHECK_FAILURE)')
    WIN32__ERROR_RANGE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80070284, 'HRESULT_FROM_WIN32(ERROR_RANGE_NOT_FOUND)')
    WIN32__ERROR_NOT_SAFE_MODE_DRIVER: Final[ThirtyTwoBits] = cached_hresult(0x80070286, 'HRESULT_FROM_WIN32(ERROR_NOT_SAFE_MODE_DRIVER)')
    WIN32__ERROR_FAILED_DRIVER_ENTRY: Final[ThirtyTwoBits] = cached_hresult(0x80070287, 'HRESULT_FROM_WIN32(ERROR_FAILED_DRIVER_ENTRY)')
    WIN32__ERROR_DEVICE_ENUMERATION_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80070288, 'HRESULT_FROM_WIN32(ERROR_DEVICE_ENUMERATION_ERROR)')
    WIN32__ERROR_MOUNT_POINT_NOT_RESOLVED: Final[ThirtyTwoBits] = cached_hresult(0x80070289, 'HRESULT_FROM_WIN32(ERROR_MOUNT_POINT_NOT_RESOLVED)')
    WIN32__ERROR_INVALID_DEVICE_OBJECT_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0x8007028A, 'HRESULT_FROM_WIN32(ERROR_INVALID_DEVICE_OBJECT_PARAMETER)')
    WIN32__ERROR_MCA_OCCURED: Final[ThirtyTwoBits] = cached_hresult(0x8007028B, 'HRESULT_FROM_WIN32(ERROR_MCA_OCCURED)')
    WIN32__ERROR_DRIVER_DATABASE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8007028C, 'HRESULT_FROM_WIN32(ERROR_DRIVER_DATABASE_ERROR)')
    WIN32__ERROR_SYSTEM_HIVE_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0x8007028D, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_HIVE_TOO_LARGE)')
    WIN32__ERROR_DRIVER_FAILED_PRIOR_UNLOAD: Final[ThirtyTwoBits] = cached_hresult(0x8007028E, 'HRESULT_FROM_WIN32(ERROR_DRIVER_FAILED_PRIOR_UNLOAD)')
    WIN32__ERROR_VOLSNAP_PREPARE_HIBERNATE: Final[ThirtyTwoBits] = cached_hresult(0x8007028F, 'HRESULT_FROM_WIN32(ERROR_VOLSNAP_PREPARE_HIBERNATE)')
    WIN32__ERROR_HIBERNATION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80070290, 'HRESULT_FROM_WIN32(ERROR_HIBERNATION_FAILURE)')
    WIN32__ERROR_PWD_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0x80070291, 'HRESULT_FROM_WIN32(ERROR_PWD_TOO_LONG)')
    WIN32__ERROR_FILE_SYSTEM_LIMITATION: Final[ThirtyTwoBits] = cached_hresult(0x80070299, 'HRESULT_FROM_WIN32(ERROR_FILE_SYSTEM_LIMITATION)')
    WIN32__ERROR_ASSERTION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8007029C, 'HRESULT_FROM_WIN32(ERROR_ASSERTION_FAILURE)')
    WIN32__ERROR_ACPI_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8007029D, 'HRESULT_FROM_WIN32(ERROR_ACPI_ERROR)')
    WIN32__ERROR_WOW_ASSERTION: Final[ThirtyTwoBits] = cached_hresult(0x8007029E, 'HRESULT_FROM_WIN32(ERROR_WOW_ASSERTION)')
    WIN32__ERROR_PNP_BAD_MPS_TABLE: Final[ThirtyTwoBits] = cached_hresult(0x8007029F, 'HRESULT_FROM_WIN32(ERROR_PNP_BAD_MPS_TABLE)')
    WIN32__ERROR_PNP_TRANSLATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800702A0, 'HRESULT_FROM_WIN32(ERROR_PNP_TRANSLATION_FAILED)')
    WIN32__ERROR_PNP_IRQ_TRANSLATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800702A1, 'HRESULT_FROM_WIN32(ERROR_PNP_IRQ_TRANSLATION_FAILED)')
    WIN32__ERROR_PNP_INVALID_ID: Final[ThirtyTwoBits] = cached_hresult(0x800702A2, 'HRESULT_FROM_WIN32(ERROR_PNP_INVALID_ID)')
    WIN32__ERROR_WAKE_SYSTEM_DEBUGGER: Final[ThirtyTwoBits] = cached_hresult(0x800702A3, 'HRESULT_FROM_WIN32(ERROR_WAKE_SYSTEM_DEBUGGER)')
    WIN32__ERROR_HANDLES_CLOSED: Final[ThirtyTwoBits] = cached_hresult(0x800702A4, 'HRESULT_FROM_WIN32(ERROR_HANDLES_CLOSED)')
    WIN32__ERROR_EXTRANEOUS_INFORMATION: Final[ThirtyTwoBits] = cached_hresult(0x800702A5, 'HRESULT_FROM_WIN32(ERROR_EXTRANEOUS_INFORMATION)')
    WIN32__ERROR_RXACT_COMMIT_NECESSARY: Final[ThirtyTwoBits] = cached_hresult(0x800702A6, 'HRESULT_FROM_WIN32(ERROR_RXACT_COMMIT_NECESSARY)')
    WIN32__ERROR_MEDIA_CHECK: Final[ThirtyTwoBits] = cached_hresult(0x800702A7, 'HRESULT_FROM_WIN32(ERROR_MEDIA_CHECK)')
    WIN32__ERROR_GUID_SUBSTITUTION_MADE: Final[ThirtyTwoBits] = cached_hresult(0x800702A8, 'HRESULT_FROM_WIN32(ERROR_GUID_SUBSTITUTION_MADE)')
    WIN32__ERROR_STOPPED_ON_SYMLINK: Final[ThirtyTwoBits] = cached_hresult(0x800702A9, 'HRESULT_FROM_WIN32(ERROR_STOPPED_ON_SYMLINK)')
    WIN32__ERROR_LONGJUMP: Final[ThirtyTwoBits] = cached_hresult(0x800702AA, 'HRESULT_FROM_WIN32(ERROR_LONGJUMP)')
    WIN32__ERROR_PLUGPLAY_QUERY_VETOED: Final[ThirtyTwoBits] = cached_hresult(0x800702AB, 'HRESULT_FROM_WIN32(ERROR_PLUGPLAY_QUERY_VETOED)')
    WIN32__ERROR_UNWIND_CONSOLIDATE: Final[ThirtyTwoBits] = cached_hresult(0x800702AC, 'HRESULT_FROM_WIN32(ERROR_UNWIND_CONSOLIDATE)')
    WIN32__ERROR_REGISTRY_HIVE_RECOVERED: Final[ThirtyTwoBits] = cached_hresult(0x800702AD, 'HRESULT_FROM_WIN32(ERROR_REGISTRY_HIVE_RECOVERED)')
    WIN32__ERROR_DLL_MIGHT_BE_INSECURE: Final[ThirtyTwoBits] = cached_hresult(0x800702AE, 'HRESULT_FROM_WIN32(ERROR_DLL_MIGHT_BE_INSECURE)')
    WIN32__ERROR_DLL_MIGHT_BE_INCOMPATIBLE: Final[ThirtyTwoBits] = cached_hresult(0x800702AF, 'HRESULT_FROM_WIN32(ERROR_DLL_MIGHT_BE_INCOMPATIBLE)')
    WIN32__ERROR_DBG_EXCEPTION_NOT_HANDLED: Final[ThirtyTwoBits] = cached_hresult(0x800702B0, 'HRESULT_FROM_WIN32(ERROR_DBG_EXCEPTION_NOT_HANDLED)')
    WIN32__ERROR_DBG_REPLY_LATER: Final[ThirtyTwoBits] = cached_hresult(0x800702B1, 'HRESULT_FROM_WIN32(ERROR_DBG_REPLY_LATER)')
    WIN32__ERROR_DBG_UNABLE_TO_PROVIDE_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x800702B2, 'HRESULT_FROM_WIN32(ERROR_DBG_UNABLE_TO_PROVIDE_HANDLE)')
    WIN32__ERROR_DBG_TERMINATE_THREAD: Final[ThirtyTwoBits] = cached_hresult(0x800702B3, 'HRESULT_FROM_WIN32(ERROR_DBG_TERMINATE_THREAD)')
    WIN32__ERROR_DBG_TERMINATE_PROCESS: Final[ThirtyTwoBits] = cached_hresult(0x800702B4, 'HRESULT_FROM_WIN32(ERROR_DBG_TERMINATE_PROCESS)')
    WIN32__ERROR_DBG_CONTROL_C: Final[ThirtyTwoBits] = cached_hresult(0x800702B5, 'HRESULT_FROM_WIN32(ERROR_DBG_CONTROL_C)')
    WIN32__ERROR_DBG_PRINTEXCEPTION_C: Final[ThirtyTwoBits] = cached_hresult(0x800702B6, 'HRESULT_FROM_WIN32(ERROR_DBG_PRINTEXCEPTION_C)')
    WIN32__ERROR_DBG_RIPEXCEPTION: Final[ThirtyTwoBits] = cached_hresult(0x800702B7, 'HRESULT_FROM_WIN32(ERROR_DBG_RIPEXCEPTION)')
    WIN32__ERROR_DBG_CONTROL_BREAK: Final[ThirtyTwoBits] = cached_hresult(0x800702B8, 'HRESULT_FROM_WIN32(ERROR_DBG_CONTROL_BREAK)')
    WIN32__ERROR_DBG_COMMAND_EXCEPTION: Final[ThirtyTwoBits] = cached_hresult(0x800702B9, 'HRESULT_FROM_WIN32(ERROR_DBG_COMMAND_EXCEPTION)')
    WIN32__ERROR_OBJECT_NAME_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800702BA, 'HRESULT_FROM_WIN32(ERROR_OBJECT_NAME_EXISTS)')
    WIN32__ERROR_THREAD_WAS_SUSPENDED: Final[ThirtyTwoBits] = cached_hresult(0x800702BB, 'HRESULT_FROM_WIN32(ERROR_THREAD_WAS_SUSPENDED)')
    WIN32__ERROR_IMAGE_NOT_AT_BASE: Final[ThirtyTwoBits] = cached_hresult(0x800702BC, 'HRESULT_FROM_WIN32(ERROR_IMAGE_NOT_AT_BASE)')
    WIN32__ERROR_RXACT_STATE_CREATED: Final[ThirtyTwoBits] = cached_hresult(0x800702BD, 'HRESULT_FROM_WIN32(ERROR_RXACT_STATE_CREATED)')
    WIN32__ERROR_SEGMENT_NOTIFICATION: Final[ThirtyTwoBits] = cached_hresult(0x800702BE, 'HRESULT_FROM_WIN32(ERROR_SEGMENT_NOTIFICATION)')
    WIN32__ERROR_BAD_CURRENT_DIRECTORY: Final[ThirtyTwoBits] = cached_hresult(0x800702BF, 'HRESULT_FROM_WIN32(ERROR_BAD_CURRENT_DIRECTORY)')
    WIN32__ERROR_FT_READ_RECOVERY_FROM_BACKUP: Final[ThirtyTwoBits] = cached_hresult(0x800702C0, 'HRESULT_FROM_WIN32(ERROR_FT_READ_RECOVERY_FROM_BACKUP)')
    WIN32__ERROR_FT_WRITE_RECOVERY: Final[ThirtyTwoBits] = cached_hresult(0x800702C1, 'HRESULT_FROM_WIN32(ERROR_FT_WRITE_RECOVERY)')
    WIN32__ERROR_IMAGE_MACHINE_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x800702C2, 'HRESULT_FROM_WIN32(ERROR_IMAGE_MACHINE_TYPE_MISMATCH)')
    WIN32__ERROR_RECEIVE_PARTIAL: Final[ThirtyTwoBits] = cached_hresult(0x800702C3, 'HRESULT_FROM_WIN32(ERROR_RECEIVE_PARTIAL)')
    WIN32__ERROR_RECEIVE_EXPEDITED: Final[ThirtyTwoBits] = cached_hresult(0x800702C4, 'HRESULT_FROM_WIN32(ERROR_RECEIVE_EXPEDITED)')
    WIN32__ERROR_RECEIVE_PARTIAL_EXPEDITED: Final[ThirtyTwoBits] = cached_hresult(0x800702C5, 'HRESULT_FROM_WIN32(ERROR_RECEIVE_PARTIAL_EXPEDITED)')
    WIN32__ERROR_EVENT_DONE: Final[ThirtyTwoBits] = cached_hresult(0x800702C6, 'HRESULT_FROM_WIN32(ERROR_EVENT_DONE)')
    WIN32__ERROR_EVENT_PENDING: Final[ThirtyTwoBits] = cached_hresult(0x800702C7, 'HRESULT_FROM_WIN32(ERROR_EVENT_PENDING)')
    WIN32__ERROR_CHECKING_FILE_SYSTEM: Final[ThirtyTwoBits] = cached_hresult(0x800702C8, 'HRESULT_FROM_WIN32(ERROR_CHECKING_FILE_SYSTEM)')
    WIN32__ERROR_FATAL_APP_EXIT: Final[ThirtyTwoBits] = cached_hresult(0x800702C9, 'HRESULT_FROM_WIN32(ERROR_FATAL_APP_EXIT)')
    WIN32__ERROR_PREDEFINED_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x800702CA, 'HRESULT_FROM_WIN32(ERROR_PREDEFINED_HANDLE)')
    WIN32__ERROR_WAS_UNLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x800702CB, 'HRESULT_FROM_WIN32(ERROR_WAS_UNLOCKED)')
    WIN32__ERROR_SERVICE_NOTIFICATION: Final[ThirtyTwoBits] = cached_hresult(0x800702CC, 'HRESULT_FROM_WIN32(ERROR_SERVICE_NOTIFICATION)')
    WIN32__ERROR_WAS_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x800702CD, 'HRESULT_FROM_WIN32(ERROR_WAS_LOCKED)')
    WIN32__ERROR_LOG_HARD_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800702CE, 'HRESULT_FROM_WIN32(ERROR_LOG_HARD_ERROR)')
    WIN32__ERROR_ALREADY_WIN32: Final[ThirtyTwoBits] = cached_hresult(0x800702CF, 'HRESULT_FROM_WIN32(ERROR_ALREADY_WIN32)')
    WIN32__ERROR_IMAGE_MACHINE_TYPE_MISMATCH_EXE: Final[ThirtyTwoBits] = cached_hresult(0x800702D0, 'HRESULT_FROM_WIN32(ERROR_IMAGE_MACHINE_TYPE_MISMATCH_EXE)')
    WIN32__ERROR_NO_YIELD_PERFORMED: Final[ThirtyTwoBits] = cached_hresult(0x800702D1, 'HRESULT_FROM_WIN32(ERROR_NO_YIELD_PERFORMED)')
    WIN32__ERROR_TIMER_RESUME_IGNORED: Final[ThirtyTwoBits] = cached_hresult(0x800702D2, 'HRESULT_FROM_WIN32(ERROR_TIMER_RESUME_IGNORED)')
    WIN32__ERROR_ARBITRATION_UNHANDLED: Final[ThirtyTwoBits] = cached_hresult(0x800702D3, 'HRESULT_FROM_WIN32(ERROR_ARBITRATION_UNHANDLED)')
    WIN32__ERROR_CARDBUS_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x800702D4, 'HRESULT_FROM_WIN32(ERROR_CARDBUS_NOT_SUPPORTED)')
    WIN32__ERROR_MP_PROCESSOR_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x800702D5, 'HRESULT_FROM_WIN32(ERROR_MP_PROCESSOR_MISMATCH)')
    WIN32__ERROR_HIBERNATED: Final[ThirtyTwoBits] = cached_hresult(0x800702D6, 'HRESULT_FROM_WIN32(ERROR_HIBERNATED)')
    WIN32__ERROR_RESUME_HIBERNATION: Final[ThirtyTwoBits] = cached_hresult(0x800702D7, 'HRESULT_FROM_WIN32(ERROR_RESUME_HIBERNATION)')
    WIN32__ERROR_FIRMWARE_UPDATED: Final[ThirtyTwoBits] = cached_hresult(0x800702D8, 'HRESULT_FROM_WIN32(ERROR_FIRMWARE_UPDATED)')
    WIN32__ERROR_DRIVERS_LEAKING_LOCKED_PAGES: Final[ThirtyTwoBits] = cached_hresult(0x800702D9, 'HRESULT_FROM_WIN32(ERROR_DRIVERS_LEAKING_LOCKED_PAGES)')
    WIN32__ERROR_WAKE_SYSTEM: Final[ThirtyTwoBits] = cached_hresult(0x800702DA, 'HRESULT_FROM_WIN32(ERROR_WAKE_SYSTEM)')
    WIN32__ERROR_WAIT_1: Final[ThirtyTwoBits] = cached_hresult(0x800702DB, 'HRESULT_FROM_WIN32(ERROR_WAIT_1)')
    WIN32__ERROR_WAIT_2: Final[ThirtyTwoBits] = cached_hresult(0x800702DC, 'HRESULT_FROM_WIN32(ERROR_WAIT_2)')
    WIN32__ERROR_WAIT_3: Final[ThirtyTwoBits] = cached_hresult(0x800702DD, 'HRESULT_FROM_WIN32(ERROR_WAIT_3)')
    WIN32__ERROR_WAIT_63: Final[ThirtyTwoBits] = cached_hresult(0x800702DE, 'HRESULT_FROM_WIN32(ERROR_WAIT_63)')
    WIN32__ERROR_ABANDONED_WAIT_0: Final[ThirtyTwoBits] = cached_hresult(0x800702DF, 'HRESULT_FROM_WIN32(ERROR_ABANDONED_WAIT_0)')
    WIN32__ERROR_ABANDONED_WAIT_63: Final[ThirtyTwoBits] = cached_hresult(0x800702E0, 'HRESULT_FROM_WIN32(ERROR_ABANDONED_WAIT_63)')
    WIN32__ERROR_USER_APC: Final[ThirtyTwoBits] = cached_hresult(0x800702E1, 'HRESULT_FROM_WIN32(ERROR_USER_APC)')
    WIN32__ERROR_KERNEL_APC: Final[ThirtyTwoBits] = cached_hresult(0x800702E2, 'HRESULT_FROM_WIN32(ERROR_KERNEL_APC)')
    WIN32__ERROR_ALERTED: Final[ThirtyTwoBits] = cached_hresult(0x800702E3, 'HRESULT_FROM_WIN32(ERROR_ALERTED)')
    WIN32__ERROR_ELEVATION_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x800702E4, 'HRESULT_FROM_WIN32(ERROR_ELEVATION_REQUIRED)')
    WIN32__ERROR_REPARSE: Final[ThirtyTwoBits] = cached_hresult(0x800702E5, 'HRESULT_FROM_WIN32(ERROR_REPARSE)')
    WIN32__ERROR_OPLOCK_BREAK_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x800702E6, 'HRESULT_FROM_WIN32(ERROR_OPLOCK_BREAK_IN_PROGRESS)')
    WIN32__ERROR_VOLUME_MOUNTED: Final[ThirtyTwoBits] = cached_hresult(0x800702E7, 'HRESULT_FROM_WIN32(ERROR_VOLUME_MOUNTED)')
    WIN32__ERROR_RXACT_COMMITTED: Final[ThirtyTwoBits] = cached_hresult(0x800702E8, 'HRESULT_FROM_WIN32(ERROR_RXACT_COMMITTED)')
    WIN32__ERROR_NOTIFY_CLEANUP: Final[ThirtyTwoBits] = cached_hresult(0x800702E9, 'HRESULT_FROM_WIN32(ERROR_NOTIFY_CLEANUP)')
    WIN32__ERROR_PRIMARY_TRANSPORT_CONNECT_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800702EA, 'HRESULT_FROM_WIN32(ERROR_PRIMARY_TRANSPORT_CONNECT_FAILED)')
    WIN32__ERROR_PAGE_FAULT_TRANSITION: Final[ThirtyTwoBits] = cached_hresult(0x800702EB, 'HRESULT_FROM_WIN32(ERROR_PAGE_FAULT_TRANSITION)')
    WIN32__ERROR_PAGE_FAULT_DEMAND_ZERO: Final[ThirtyTwoBits] = cached_hresult(0x800702EC, 'HRESULT_FROM_WIN32(ERROR_PAGE_FAULT_DEMAND_ZERO)')
    WIN32__ERROR_PAGE_FAULT_COPY_ON_WRITE: Final[ThirtyTwoBits] = cached_hresult(0x800702ED, 'HRESULT_FROM_WIN32(ERROR_PAGE_FAULT_COPY_ON_WRITE)')
    WIN32__ERROR_PAGE_FAULT_GUARD_PAGE: Final[ThirtyTwoBits] = cached_hresult(0x800702EE, 'HRESULT_FROM_WIN32(ERROR_PAGE_FAULT_GUARD_PAGE)')
    WIN32__ERROR_PAGE_FAULT_PAGING_FILE: Final[ThirtyTwoBits] = cached_hresult(0x800702EF, 'HRESULT_FROM_WIN32(ERROR_PAGE_FAULT_PAGING_FILE)')
    WIN32__ERROR_CACHE_PAGE_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x800702F0, 'HRESULT_FROM_WIN32(ERROR_CACHE_PAGE_LOCKED)')
    WIN32__ERROR_CRASH_DUMP: Final[ThirtyTwoBits] = cached_hresult(0x800702F1, 'HRESULT_FROM_WIN32(ERROR_CRASH_DUMP)')
    WIN32__ERROR_BUFFER_ALL_ZEROS: Final[ThirtyTwoBits] = cached_hresult(0x800702F2, 'HRESULT_FROM_WIN32(ERROR_BUFFER_ALL_ZEROS)')
    WIN32__ERROR_REPARSE_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x800702F3, 'HRESULT_FROM_WIN32(ERROR_REPARSE_OBJECT)')
    WIN32__ERROR_RESOURCE_REQUIREMENTS_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0x800702F4, 'HRESULT_FROM_WIN32(ERROR_RESOURCE_REQUIREMENTS_CHANGED)')
    WIN32__ERROR_TRANSLATION_COMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x800702F5, 'HRESULT_FROM_WIN32(ERROR_TRANSLATION_COMPLETE)')
    WIN32__ERROR_NOTHING_TO_TERMINATE: Final[ThirtyTwoBits] = cached_hresult(0x800702F6, 'HRESULT_FROM_WIN32(ERROR_NOTHING_TO_TERMINATE)')
    WIN32__ERROR_PROCESS_NOT_IN_JOB: Final[ThirtyTwoBits] = cached_hresult(0x800702F7, 'HRESULT_FROM_WIN32(ERROR_PROCESS_NOT_IN_JOB)')
    WIN32__ERROR_PROCESS_IN_JOB: Final[ThirtyTwoBits] = cached_hresult(0x800702F8, 'HRESULT_FROM_WIN32(ERROR_PROCESS_IN_JOB)')
    WIN32__ERROR_VOLSNAP_HIBERNATE_READY: Final[ThirtyTwoBits] = cached_hresult(0x800702F9, 'HRESULT_FROM_WIN32(ERROR_VOLSNAP_HIBERNATE_READY)')
    WIN32__ERROR_FSFILTER_OP_COMPLETED_SUCCESSFULLY: Final[ThirtyTwoBits] = cached_hresult(0x800702FA, 'HRESULT_FROM_WIN32(ERROR_FSFILTER_OP_COMPLETED_SUCCESSFULLY)')
    WIN32__ERROR_INTERRUPT_VECTOR_ALREADY_CONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x800702FB, 'HRESULT_FROM_WIN32(ERROR_INTERRUPT_VECTOR_ALREADY_CONNECTED)')
    WIN32__ERROR_INTERRUPT_STILL_CONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x800702FC, 'HRESULT_FROM_WIN32(ERROR_INTERRUPT_STILL_CONNECTED)')
    WIN32__ERROR_WAIT_FOR_OPLOCK: Final[ThirtyTwoBits] = cached_hresult(0x800702FD, 'HRESULT_FROM_WIN32(ERROR_WAIT_FOR_OPLOCK)')
    WIN32__ERROR_DBG_EXCEPTION_HANDLED: Final[ThirtyTwoBits] = cached_hresult(0x800702FE, 'HRESULT_FROM_WIN32(ERROR_DBG_EXCEPTION_HANDLED)')
    WIN32__ERROR_DBG_CONTINUE: Final[ThirtyTwoBits] = cached_hresult(0x800702FF, 'HRESULT_FROM_WIN32(ERROR_DBG_CONTINUE)')
    WIN32__ERROR_CALLBACK_POP_STACK: Final[ThirtyTwoBits] = cached_hresult(0x80070300, 'HRESULT_FROM_WIN32(ERROR_CALLBACK_POP_STACK)')
    WIN32__ERROR_COMPRESSION_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80070301, 'HRESULT_FROM_WIN32(ERROR_COMPRESSION_DISABLED)')
    WIN32__ERROR_CANTFETCHBACKWARDS: Final[ThirtyTwoBits] = cached_hresult(0x80070302, 'HRESULT_FROM_WIN32(ERROR_CANTFETCHBACKWARDS)')
    WIN32__ERROR_CANTSCROLLBACKWARDS: Final[ThirtyTwoBits] = cached_hresult(0x80070303, 'HRESULT_FROM_WIN32(ERROR_CANTSCROLLBACKWARDS)')
    WIN32__ERROR_ROWSNOTRELEASED: Final[ThirtyTwoBits] = cached_hresult(0x80070304, 'HRESULT_FROM_WIN32(ERROR_ROWSNOTRELEASED)')
    WIN32__ERROR_BAD_ACCESSOR_FLAGS: Final[ThirtyTwoBits] = cached_hresult(0x80070305, 'HRESULT_FROM_WIN32(ERROR_BAD_ACCESSOR_FLAGS)')
    WIN32__ERROR_ERRORS_ENCOUNTERED: Final[ThirtyTwoBits] = cached_hresult(0x80070306, 'HRESULT_FROM_WIN32(ERROR_ERRORS_ENCOUNTERED)')
    WIN32__ERROR_NOT_CAPABLE: Final[ThirtyTwoBits] = cached_hresult(0x80070307, 'HRESULT_FROM_WIN32(ERROR_NOT_CAPABLE)')
    WIN32__ERROR_REQUEST_OUT_OF_SEQUENCE: Final[ThirtyTwoBits] = cached_hresult(0x80070308, 'HRESULT_FROM_WIN32(ERROR_REQUEST_OUT_OF_SEQUENCE)')
    WIN32__ERROR_VERSION_PARSE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80070309, 'HRESULT_FROM_WIN32(ERROR_VERSION_PARSE_ERROR)')
    WIN32__ERROR_BADSTARTPOSITION: Final[ThirtyTwoBits] = cached_hresult(0x8007030A, 'HRESULT_FROM_WIN32(ERROR_BADSTARTPOSITION)')
    WIN32__ERROR_MEMORY_HARDWARE: Final[ThirtyTwoBits] = cached_hresult(0x8007030B, 'HRESULT_FROM_WIN32(ERROR_MEMORY_HARDWARE)')
    WIN32__ERROR_DISK_REPAIR_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x8007030C, 'HRESULT_FROM_WIN32(ERROR_DISK_REPAIR_DISABLED)')
    WIN32__ERROR_INSUFFICIENT_RESOURCE_FOR_SPECIFIED_SHARED_SECTION_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x8007030D, 'HRESULT_FROM_WIN32(ERROR_INSUFFICIENT_RESOURCE_FOR_SPECIFIED_SHARED_SECTION_SIZE)')
    WIN32__ERROR_SYSTEM_POWERSTATE_TRANSITION: Final[ThirtyTwoBits] = cached_hresult(0x8007030E, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_POWERSTATE_TRANSITION)')
    WIN32__ERROR_SYSTEM_POWERSTATE_COMPLEX_TRANSITION: Final[ThirtyTwoBits] = cached_hresult(0x8007030F, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_POWERSTATE_COMPLEX_TRANSITION)')
    WIN32__ERROR_MCA_EXCEPTION: Final[ThirtyTwoBits] = cached_hresult(0x80070310, 'HRESULT_FROM_WIN32(ERROR_MCA_EXCEPTION)')
    WIN32__ERROR_ACCESS_AUDIT_BY_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x80070311, 'HRESULT_FROM_WIN32(ERROR_ACCESS_AUDIT_BY_POLICY)')
    WIN32__ERROR_ACCESS_DISABLED_NO_SAFER_UI_BY_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x80070312, 'HRESULT_FROM_WIN32(ERROR_ACCESS_DISABLED_NO_SAFER_UI_BY_POLICY)')
    WIN32__ERROR_ABANDON_HIBERFILE: Final[ThirtyTwoBits] = cached_hresult(0x80070313, 'HRESULT_FROM_WIN32(ERROR_ABANDON_HIBERFILE)')
    WIN32__ERROR_LOST_WRITEBEHIND_DATA_NETWORK_DISCONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x80070314, 'HRESULT_FROM_WIN32(ERROR_LOST_WRITEBEHIND_DATA_NETWORK_DISCONNECTED)')
    WIN32__ERROR_LOST_WRITEBEHIND_DATA_NETWORK_SERVER_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80070315, 'HRESULT_FROM_WIN32(ERROR_LOST_WRITEBEHIND_DATA_NETWORK_SERVER_ERROR)')
    WIN32__ERROR_LOST_WRITEBEHIND_DATA_LOCAL_DISK_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80070316, 'HRESULT_FROM_WIN32(ERROR_LOST_WRITEBEHIND_DATA_LOCAL_DISK_ERROR)')
    WIN32__ERROR_BAD_MCFG_TABLE: Final[ThirtyTwoBits] = cached_hresult(0x80070317, 'HRESULT_FROM_WIN32(ERROR_BAD_MCFG_TABLE)')
    WIN32__ERROR_DISK_REPAIR_REDIRECTED: Final[ThirtyTwoBits] = cached_hresult(0x80070318, 'HRESULT_FROM_WIN32(ERROR_DISK_REPAIR_REDIRECTED)')
    WIN32__ERROR_DISK_REPAIR_UNSUCCESSFUL: Final[ThirtyTwoBits] = cached_hresult(0x80070319, 'HRESULT_FROM_WIN32(ERROR_DISK_REPAIR_UNSUCCESSFUL)')
    WIN32__ERROR_CORRUPT_LOG_OVERFULL: Final[ThirtyTwoBits] = cached_hresult(0x8007031A, 'HRESULT_FROM_WIN32(ERROR_CORRUPT_LOG_OVERFULL)')
    WIN32__ERROR_CORRUPT_LOG_CORRUPTED: Final[ThirtyTwoBits] = cached_hresult(0x8007031B, 'HRESULT_FROM_WIN32(ERROR_CORRUPT_LOG_CORRUPTED)')
    WIN32__ERROR_CORRUPT_LOG_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x8007031C, 'HRESULT_FROM_WIN32(ERROR_CORRUPT_LOG_UNAVAILABLE)')
    WIN32__ERROR_CORRUPT_LOG_DELETED_FULL: Final[ThirtyTwoBits] = cached_hresult(0x8007031D, 'HRESULT_FROM_WIN32(ERROR_CORRUPT_LOG_DELETED_FULL)')
    WIN32__ERROR_CORRUPT_LOG_CLEARED: Final[ThirtyTwoBits] = cached_hresult(0x8007031E, 'HRESULT_FROM_WIN32(ERROR_CORRUPT_LOG_CLEARED)')
    WIN32__ERROR_ORPHAN_NAME_EXHAUSTED: Final[ThirtyTwoBits] = cached_hresult(0x8007031F, 'HRESULT_FROM_WIN32(ERROR_ORPHAN_NAME_EXHAUSTED)')
    WIN32__ERROR_OPLOCK_SWITCHED_TO_NEW_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80070320, 'HRESULT_FROM_WIN32(ERROR_OPLOCK_SWITCHED_TO_NEW_HANDLE)')
    WIN32__ERROR_CANNOT_GRANT_REQUESTED_OPLOCK: Final[ThirtyTwoBits] = cached_hresult(0x80070321, 'HRESULT_FROM_WIN32(ERROR_CANNOT_GRANT_REQUESTED_OPLOCK)')
    WIN32__ERROR_CANNOT_BREAK_OPLOCK: Final[ThirtyTwoBits] = cached_hresult(0x80070322, 'HRESULT_FROM_WIN32(ERROR_CANNOT_BREAK_OPLOCK)')
    WIN32__ERROR_OPLOCK_HANDLE_CLOSED: Final[ThirtyTwoBits] = cached_hresult(0x80070323, 'HRESULT_FROM_WIN32(ERROR_OPLOCK_HANDLE_CLOSED)')
    WIN32__ERROR_NO_ACE_CONDITION: Final[ThirtyTwoBits] = cached_hresult(0x80070324, 'HRESULT_FROM_WIN32(ERROR_NO_ACE_CONDITION)')
    WIN32__ERROR_INVALID_ACE_CONDITION: Final[ThirtyTwoBits] = cached_hresult(0x80070325, 'HRESULT_FROM_WIN32(ERROR_INVALID_ACE_CONDITION)')
    WIN32__ERROR_FILE_HANDLE_REVOKED: Final[ThirtyTwoBits] = cached_hresult(0x80070326, 'HRESULT_FROM_WIN32(ERROR_FILE_HANDLE_REVOKED)')
    WIN32__ERROR_IMAGE_AT_DIFFERENT_BASE: Final[ThirtyTwoBits] = cached_hresult(0x80070327, 'HRESULT_FROM_WIN32(ERROR_IMAGE_AT_DIFFERENT_BASE)')
    WIN32__ERROR_ENCRYPTED_IO_NOT_POSSIBLE: Final[ThirtyTwoBits] = cached_hresult(0x80070328, 'HRESULT_FROM_WIN32(ERROR_ENCRYPTED_IO_NOT_POSSIBLE)')
    WIN32__ERROR_FILE_METADATA_OPTIMIZATION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x80070329, 'HRESULT_FROM_WIN32(ERROR_FILE_METADATA_OPTIMIZATION_IN_PROGRESS)')
    WIN32__ERROR_QUOTA_ACTIVITY: Final[ThirtyTwoBits] = cached_hresult(0x8007032A, 'HRESULT_FROM_WIN32(ERROR_QUOTA_ACTIVITY)')
    WIN32__ERROR_HANDLE_REVOKED: Final[ThirtyTwoBits] = cached_hresult(0x8007032B, 'HRESULT_FROM_WIN32(ERROR_HANDLE_REVOKED)')
    WIN32__ERROR_CALLBACK_INVOKE_INLINE: Final[ThirtyTwoBits] = cached_hresult(0x8007032C, 'HRESULT_FROM_WIN32(ERROR_CALLBACK_INVOKE_INLINE)')
    WIN32__ERROR_CPU_SET_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x8007032D, 'HRESULT_FROM_WIN32(ERROR_CPU_SET_INVALID)')
    WIN32__ERROR_ENCLAVE_NOT_TERMINATED: Final[ThirtyTwoBits] = cached_hresult(0x8007032E, 'HRESULT_FROM_WIN32(ERROR_ENCLAVE_NOT_TERMINATED)')
    WIN32__ERROR_ENCLAVE_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x8007032F, 'HRESULT_FROM_WIN32(ERROR_ENCLAVE_VIOLATION)')
    WIN32__ERROR_SERVER_TRANSPORT_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x80070330, 'HRESULT_FROM_WIN32(ERROR_SERVER_TRANSPORT_CONFLICT)')
    WIN32__ERROR_CERTIFICATE_VALIDATION_PREFERENCE_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x80070331, 'HRESULT_FROM_WIN32(ERROR_CERTIFICATE_VALIDATION_PREFERENCE_CONFLICT)')
    WIN32__ERROR_FT_READ_FROM_COPY_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80070332, 'HRESULT_FROM_WIN32(ERROR_FT_READ_FROM_COPY_FAILURE)')
    WIN32__ERROR_SECTION_DIRECT_MAP_ONLY: Final[ThirtyTwoBits] = cached_hresult(0x80070333, 'HRESULT_FROM_WIN32(ERROR_SECTION_DIRECT_MAP_ONLY)')
    WIN32__ERROR_EA_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x800703E2, 'HRESULT_FROM_WIN32(ERROR_EA_ACCESS_DENIED)')
    WIN32__ERROR_OPERATION_ABORTED: Final[ThirtyTwoBits] = cached_hresult(0x800703E3, 'HRESULT_FROM_WIN32(ERROR_OPERATION_ABORTED)')
    WIN32__ERROR_IO_INCOMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x800703E4, 'HRESULT_FROM_WIN32(ERROR_IO_INCOMPLETE)')
    WIN32__ERROR_IO_PENDING: Final[ThirtyTwoBits] = cached_hresult(0x800703E5, 'HRESULT_FROM_WIN32(ERROR_IO_PENDING)')
    WIN32__ERROR_NOACCESS: Final[ThirtyTwoBits] = cached_hresult(0x800703E6, 'HRESULT_FROM_WIN32(ERROR_NOACCESS)')
    WIN32__ERROR_SWAPERROR: Final[ThirtyTwoBits] = cached_hresult(0x800703E7, 'HRESULT_FROM_WIN32(ERROR_SWAPERROR)')
    WIN32__ERROR_STACK_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0x800703E9, 'HRESULT_FROM_WIN32(ERROR_STACK_OVERFLOW)')
    WIN32__ERROR_INVALID_MESSAGE: Final[ThirtyTwoBits] = cached_hresult(0x800703EA, 'HRESULT_FROM_WIN32(ERROR_INVALID_MESSAGE)')
    WIN32__ERROR_CAN_NOT_COMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x800703EB, 'HRESULT_FROM_WIN32(ERROR_CAN_NOT_COMPLETE)')
    WIN32__ERROR_INVALID_FLAGS: Final[ThirtyTwoBits] = cached_hresult(0x800703EC, 'HRESULT_FROM_WIN32(ERROR_INVALID_FLAGS)')
    WIN32__ERROR_UNRECOGNIZED_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0x800703ED, 'HRESULT_FROM_WIN32(ERROR_UNRECOGNIZED_VOLUME)')
    WIN32__ERROR_FILE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800703EE, 'HRESULT_FROM_WIN32(ERROR_FILE_INVALID)')
    WIN32__ERROR_FULLSCREEN_MODE: Final[ThirtyTwoBits] = cached_hresult(0x800703EF, 'HRESULT_FROM_WIN32(ERROR_FULLSCREEN_MODE)')
    WIN32__ERROR_NO_TOKEN: Final[ThirtyTwoBits] = cached_hresult(0x800703F0, 'HRESULT_FROM_WIN32(ERROR_NO_TOKEN)')
    WIN32__ERROR_BADDB: Final[ThirtyTwoBits] = cached_hresult(0x800703F1, 'HRESULT_FROM_WIN32(ERROR_BADDB)')
    WIN32__ERROR_BADKEY: Final[ThirtyTwoBits] = cached_hresult(0x800703F2, 'HRESULT_FROM_WIN32(ERROR_BADKEY)')
    WIN32__ERROR_CANTOPEN: Final[ThirtyTwoBits] = cached_hresult(0x800703F3, 'HRESULT_FROM_WIN32(ERROR_CANTOPEN)')
    WIN32__ERROR_CANTREAD: Final[ThirtyTwoBits] = cached_hresult(0x800703F4, 'HRESULT_FROM_WIN32(ERROR_CANTREAD)')
    WIN32__ERROR_CANTWRITE: Final[ThirtyTwoBits] = cached_hresult(0x800703F5, 'HRESULT_FROM_WIN32(ERROR_CANTWRITE)')
    WIN32__ERROR_REGISTRY_RECOVERED: Final[ThirtyTwoBits] = cached_hresult(0x800703F6, 'HRESULT_FROM_WIN32(ERROR_REGISTRY_RECOVERED)')
    WIN32__ERROR_REGISTRY_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x800703F7, 'HRESULT_FROM_WIN32(ERROR_REGISTRY_CORRUPT)')
    WIN32__ERROR_REGISTRY_IO_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800703F8, 'HRESULT_FROM_WIN32(ERROR_REGISTRY_IO_FAILED)')
    WIN32__ERROR_NOT_REGISTRY_FILE: Final[ThirtyTwoBits] = cached_hresult(0x800703F9, 'HRESULT_FROM_WIN32(ERROR_NOT_REGISTRY_FILE)')
    WIN32__ERROR_KEY_DELETED: Final[ThirtyTwoBits] = cached_hresult(0x800703FA, 'HRESULT_FROM_WIN32(ERROR_KEY_DELETED)')
    WIN32__ERROR_NO_LOG_SPACE: Final[ThirtyTwoBits] = cached_hresult(0x800703FB, 'HRESULT_FROM_WIN32(ERROR_NO_LOG_SPACE)')
    WIN32__ERROR_KEY_HAS_CHILDREN: Final[ThirtyTwoBits] = cached_hresult(0x800703FC, 'HRESULT_FROM_WIN32(ERROR_KEY_HAS_CHILDREN)')
    WIN32__ERROR_CHILD_MUST_BE_VOLATILE: Final[ThirtyTwoBits] = cached_hresult(0x800703FD, 'HRESULT_FROM_WIN32(ERROR_CHILD_MUST_BE_VOLATILE)')
    WIN32__ERROR_NOTIFY_ENUM_DIR: Final[ThirtyTwoBits] = cached_hresult(0x800703FE, 'HRESULT_FROM_WIN32(ERROR_NOTIFY_ENUM_DIR)')
    WIN32__ERROR_DEPENDENT_SERVICES_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x8007041B, 'HRESULT_FROM_WIN32(ERROR_DEPENDENT_SERVICES_RUNNING)')
    WIN32__ERROR_INVALID_SERVICE_CONTROL: Final[ThirtyTwoBits] = cached_hresult(0x8007041C, 'HRESULT_FROM_WIN32(ERROR_INVALID_SERVICE_CONTROL)')
    WIN32__ERROR_SERVICE_REQUEST_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x8007041D, 'HRESULT_FROM_WIN32(ERROR_SERVICE_REQUEST_TIMEOUT)')
    WIN32__ERROR_SERVICE_NO_THREAD: Final[ThirtyTwoBits] = cached_hresult(0x8007041E, 'HRESULT_FROM_WIN32(ERROR_SERVICE_NO_THREAD)')
    WIN32__ERROR_SERVICE_DATABASE_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x8007041F, 'HRESULT_FROM_WIN32(ERROR_SERVICE_DATABASE_LOCKED)')
    WIN32__ERROR_SERVICE_ALREADY_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x80070420, 'HRESULT_FROM_WIN32(ERROR_SERVICE_ALREADY_RUNNING)')
    WIN32__ERROR_INVALID_SERVICE_ACCOUNT: Final[ThirtyTwoBits] = cached_hresult(0x80070421, 'HRESULT_FROM_WIN32(ERROR_INVALID_SERVICE_ACCOUNT)')
    WIN32__ERROR_SERVICE_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80070422, 'HRESULT_FROM_WIN32(ERROR_SERVICE_DISABLED)')
    WIN32__ERROR_CIRCULAR_DEPENDENCY: Final[ThirtyTwoBits] = cached_hresult(0x80070423, 'HRESULT_FROM_WIN32(ERROR_CIRCULAR_DEPENDENCY)')
    WIN32__ERROR_SERVICE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x80070424, 'HRESULT_FROM_WIN32(ERROR_SERVICE_DOES_NOT_EXIST)')
    WIN32__ERROR_SERVICE_CANNOT_ACCEPT_CTRL: Final[ThirtyTwoBits] = cached_hresult(0x80070425, 'HRESULT_FROM_WIN32(ERROR_SERVICE_CANNOT_ACCEPT_CTRL)')
    WIN32__ERROR_SERVICE_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0x80070426, 'HRESULT_FROM_WIN32(ERROR_SERVICE_NOT_ACTIVE)')
    WIN32__ERROR_FAILED_SERVICE_CONTROLLER_CONNECT: Final[ThirtyTwoBits] = cached_hresult(0x80070427, 'HRESULT_FROM_WIN32(ERROR_FAILED_SERVICE_CONTROLLER_CONNECT)')
    WIN32__ERROR_EXCEPTION_IN_SERVICE: Final[ThirtyTwoBits] = cached_hresult(0x80070428, 'HRESULT_FROM_WIN32(ERROR_EXCEPTION_IN_SERVICE)')
    WIN32__ERROR_DATABASE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x80070429, 'HRESULT_FROM_WIN32(ERROR_DATABASE_DOES_NOT_EXIST)')
    WIN32__ERROR_SERVICE_SPECIFIC_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8007042A, 'HRESULT_FROM_WIN32(ERROR_SERVICE_SPECIFIC_ERROR)')
    WIN32__ERROR_PROCESS_ABORTED: Final[ThirtyTwoBits] = cached_hresult(0x8007042B, 'HRESULT_FROM_WIN32(ERROR_PROCESS_ABORTED)')
    WIN32__ERROR_SERVICE_DEPENDENCY_FAIL: Final[ThirtyTwoBits] = cached_hresult(0x8007042C, 'HRESULT_FROM_WIN32(ERROR_SERVICE_DEPENDENCY_FAIL)')
    WIN32__ERROR_SERVICE_LOGON_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8007042D, 'HRESULT_FROM_WIN32(ERROR_SERVICE_LOGON_FAILED)')
    WIN32__ERROR_SERVICE_START_HANG: Final[ThirtyTwoBits] = cached_hresult(0x8007042E, 'HRESULT_FROM_WIN32(ERROR_SERVICE_START_HANG)')
    WIN32__ERROR_INVALID_SERVICE_LOCK: Final[ThirtyTwoBits] = cached_hresult(0x8007042F, 'HRESULT_FROM_WIN32(ERROR_INVALID_SERVICE_LOCK)')
    WIN32__ERROR_SERVICE_MARKED_FOR_DELETE: Final[ThirtyTwoBits] = cached_hresult(0x80070430, 'HRESULT_FROM_WIN32(ERROR_SERVICE_MARKED_FOR_DELETE)')
    WIN32__ERROR_SERVICE_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80070431, 'HRESULT_FROM_WIN32(ERROR_SERVICE_EXISTS)')
    WIN32__ERROR_ALREADY_RUNNING_LKG: Final[ThirtyTwoBits] = cached_hresult(0x80070432, 'HRESULT_FROM_WIN32(ERROR_ALREADY_RUNNING_LKG)')
    WIN32__ERROR_SERVICE_DEPENDENCY_DELETED: Final[ThirtyTwoBits] = cached_hresult(0x80070433, 'HRESULT_FROM_WIN32(ERROR_SERVICE_DEPENDENCY_DELETED)')
    WIN32__ERROR_BOOT_ALREADY_ACCEPTED: Final[ThirtyTwoBits] = cached_hresult(0x80070434, 'HRESULT_FROM_WIN32(ERROR_BOOT_ALREADY_ACCEPTED)')
    WIN32__ERROR_SERVICE_NEVER_STARTED: Final[ThirtyTwoBits] = cached_hresult(0x80070435, 'HRESULT_FROM_WIN32(ERROR_SERVICE_NEVER_STARTED)')
    WIN32__ERROR_DUPLICATE_SERVICE_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80070436, 'HRESULT_FROM_WIN32(ERROR_DUPLICATE_SERVICE_NAME)')
    WIN32__ERROR_DIFFERENT_SERVICE_ACCOUNT: Final[ThirtyTwoBits] = cached_hresult(0x80070437, 'HRESULT_FROM_WIN32(ERROR_DIFFERENT_SERVICE_ACCOUNT)')
    WIN32__ERROR_CANNOT_DETECT_DRIVER_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80070438, 'HRESULT_FROM_WIN32(ERROR_CANNOT_DETECT_DRIVER_FAILURE)')
    WIN32__ERROR_CANNOT_DETECT_PROCESS_ABORT: Final[ThirtyTwoBits] = cached_hresult(0x80070439, 'HRESULT_FROM_WIN32(ERROR_CANNOT_DETECT_PROCESS_ABORT)')
    WIN32__ERROR_NO_RECOVERY_PROGRAM: Final[ThirtyTwoBits] = cached_hresult(0x8007043A, 'HRESULT_FROM_WIN32(ERROR_NO_RECOVERY_PROGRAM)')
    WIN32__ERROR_SERVICE_NOT_IN_EXE: Final[ThirtyTwoBits] = cached_hresult(0x8007043B, 'HRESULT_FROM_WIN32(ERROR_SERVICE_NOT_IN_EXE)')
    WIN32__ERROR_NOT_SAFEBOOT_SERVICE: Final[ThirtyTwoBits] = cached_hresult(0x8007043C, 'HRESULT_FROM_WIN32(ERROR_NOT_SAFEBOOT_SERVICE)')
    WIN32__ERROR_END_OF_MEDIA: Final[ThirtyTwoBits] = cached_hresult(0x8007044C, 'HRESULT_FROM_WIN32(ERROR_END_OF_MEDIA)')
    WIN32__ERROR_FILEMARK_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0x8007044D, 'HRESULT_FROM_WIN32(ERROR_FILEMARK_DETECTED)')
    WIN32__ERROR_BEGINNING_OF_MEDIA: Final[ThirtyTwoBits] = cached_hresult(0x8007044E, 'HRESULT_FROM_WIN32(ERROR_BEGINNING_OF_MEDIA)')
    WIN32__ERROR_SETMARK_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0x8007044F, 'HRESULT_FROM_WIN32(ERROR_SETMARK_DETECTED)')
    WIN32__ERROR_NO_DATA_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0x80070450, 'HRESULT_FROM_WIN32(ERROR_NO_DATA_DETECTED)')
    WIN32__ERROR_PARTITION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80070451, 'HRESULT_FROM_WIN32(ERROR_PARTITION_FAILURE)')
    WIN32__ERROR_INVALID_BLOCK_LENGTH: Final[ThirtyTwoBits] = cached_hresult(0x80070452, 'HRESULT_FROM_WIN32(ERROR_INVALID_BLOCK_LENGTH)')
    WIN32__ERROR_DEVICE_NOT_PARTITIONED: Final[ThirtyTwoBits] = cached_hresult(0x80070453, 'HRESULT_FROM_WIN32(ERROR_DEVICE_NOT_PARTITIONED)')
    WIN32__ERROR_UNABLE_TO_LOCK_MEDIA: Final[ThirtyTwoBits] = cached_hresult(0x80070454, 'HRESULT_FROM_WIN32(ERROR_UNABLE_TO_LOCK_MEDIA)')
    WIN32__ERROR_UNABLE_TO_UNLOAD_MEDIA: Final[ThirtyTwoBits] = cached_hresult(0x80070455, 'HRESULT_FROM_WIN32(ERROR_UNABLE_TO_UNLOAD_MEDIA)')
    WIN32__ERROR_MEDIA_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0x80070456, 'HRESULT_FROM_WIN32(ERROR_MEDIA_CHANGED)')
    WIN32__ERROR_BUS_RESET: Final[ThirtyTwoBits] = cached_hresult(0x80070457, 'HRESULT_FROM_WIN32(ERROR_BUS_RESET)')
    WIN32__ERROR_NO_MEDIA_IN_DRIVE: Final[ThirtyTwoBits] = cached_hresult(0x80070458, 'HRESULT_FROM_WIN32(ERROR_NO_MEDIA_IN_DRIVE)')
    WIN32__ERROR_NO_UNICODE_TRANSLATION: Final[ThirtyTwoBits] = cached_hresult(0x80070459, 'HRESULT_FROM_WIN32(ERROR_NO_UNICODE_TRANSLATION)')
    WIN32__ERROR_DLL_INIT_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8007045A, 'HRESULT_FROM_WIN32(ERROR_DLL_INIT_FAILED)')
    WIN32__ERROR_SHUTDOWN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x8007045B, 'HRESULT_FROM_WIN32(ERROR_SHUTDOWN_IN_PROGRESS)')
    WIN32__ERROR_NO_SHUTDOWN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x8007045C, 'HRESULT_FROM_WIN32(ERROR_NO_SHUTDOWN_IN_PROGRESS)')
    WIN32__ERROR_IO_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0x8007045D, 'HRESULT_FROM_WIN32(ERROR_IO_DEVICE)')
    WIN32__ERROR_SERIAL_NO_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0x8007045E, 'HRESULT_FROM_WIN32(ERROR_SERIAL_NO_DEVICE)')
    WIN32__ERROR_IRQ_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x8007045F, 'HRESULT_FROM_WIN32(ERROR_IRQ_BUSY)')
    WIN32__ERROR_MORE_WRITES: Final[ThirtyTwoBits] = cached_hresult(0x80070460, 'HRESULT_FROM_WIN32(ERROR_MORE_WRITES)')
    WIN32__ERROR_COUNTER_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x80070461, 'HRESULT_FROM_WIN32(ERROR_COUNTER_TIMEOUT)')
    WIN32__ERROR_FLOPPY_ID_MARK_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80070462, 'HRESULT_FROM_WIN32(ERROR_FLOPPY_ID_MARK_NOT_FOUND)')
    WIN32__ERROR_FLOPPY_WRONG_CYLINDER: Final[ThirtyTwoBits] = cached_hresult(0x80070463, 'HRESULT_FROM_WIN32(ERROR_FLOPPY_WRONG_CYLINDER)')
    WIN32__ERROR_FLOPPY_UNKNOWN_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80070464, 'HRESULT_FROM_WIN32(ERROR_FLOPPY_UNKNOWN_ERROR)')
    WIN32__ERROR_FLOPPY_BAD_REGISTERS: Final[ThirtyTwoBits] = cached_hresult(0x80070465, 'HRESULT_FROM_WIN32(ERROR_FLOPPY_BAD_REGISTERS)')
    WIN32__ERROR_DISK_RECALIBRATE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80070466, 'HRESULT_FROM_WIN32(ERROR_DISK_RECALIBRATE_FAILED)')
    WIN32__ERROR_DISK_OPERATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80070467, 'HRESULT_FROM_WIN32(ERROR_DISK_OPERATION_FAILED)')
    WIN32__ERROR_DISK_RESET_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80070468, 'HRESULT_FROM_WIN32(ERROR_DISK_RESET_FAILED)')
    WIN32__ERROR_EOM_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0x80070469, 'HRESULT_FROM_WIN32(ERROR_EOM_OVERFLOW)')
    WIN32__ERROR_NOT_ENOUGH_SERVER_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0x8007046A, 'HRESULT_FROM_WIN32(ERROR_NOT_ENOUGH_SERVER_MEMORY)')
    WIN32__ERROR_POSSIBLE_DEADLOCK: Final[ThirtyTwoBits] = cached_hresult(0x8007046B, 'HRESULT_FROM_WIN32(ERROR_POSSIBLE_DEADLOCK)')
    WIN32__ERROR_MAPPED_ALIGNMENT: Final[ThirtyTwoBits] = cached_hresult(0x8007046C, 'HRESULT_FROM_WIN32(ERROR_MAPPED_ALIGNMENT)')
    WIN32__ERROR_SET_POWER_STATE_VETOED: Final[ThirtyTwoBits] = cached_hresult(0x80070474, 'HRESULT_FROM_WIN32(ERROR_SET_POWER_STATE_VETOED)')
    WIN32__ERROR_SET_POWER_STATE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80070475, 'HRESULT_FROM_WIN32(ERROR_SET_POWER_STATE_FAILED)')
    WIN32__ERROR_TOO_MANY_LINKS: Final[ThirtyTwoBits] = cached_hresult(0x80070476, 'HRESULT_FROM_WIN32(ERROR_TOO_MANY_LINKS)')
    WIN32__ERROR_OLD_WIN_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x8007047E, 'HRESULT_FROM_WIN32(ERROR_OLD_WIN_VERSION)')
    WIN32__ERROR_APP_WRONG_OS: Final[ThirtyTwoBits] = cached_hresult(0x8007047F, 'HRESULT_FROM_WIN32(ERROR_APP_WRONG_OS)')
    WIN32__ERROR_SINGLE_INSTANCE_APP: Final[ThirtyTwoBits] = cached_hresult(0x80070480, 'HRESULT_FROM_WIN32(ERROR_SINGLE_INSTANCE_APP)')
    WIN32__ERROR_RMODE_APP: Final[ThirtyTwoBits] = cached_hresult(0x80070481, 'HRESULT_FROM_WIN32(ERROR_RMODE_APP)')
    WIN32__ERROR_INVALID_DLL: Final[ThirtyTwoBits] = cached_hresult(0x80070482, 'HRESULT_FROM_WIN32(ERROR_INVALID_DLL)')
    WIN32__ERROR_NO_ASSOCIATION: Final[ThirtyTwoBits] = cached_hresult(0x80070483, 'HRESULT_FROM_WIN32(ERROR_NO_ASSOCIATION)')
    WIN32__ERROR_DDE_FAIL: Final[ThirtyTwoBits] = cached_hresult(0x80070484, 'HRESULT_FROM_WIN32(ERROR_DDE_FAIL)')
    WIN32__ERROR_DLL_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80070485, 'HRESULT_FROM_WIN32(ERROR_DLL_NOT_FOUND)')
    WIN32__ERROR_NO_MORE_USER_HANDLES: Final[ThirtyTwoBits] = cached_hresult(0x80070486, 'HRESULT_FROM_WIN32(ERROR_NO_MORE_USER_HANDLES)')
    WIN32__ERROR_MESSAGE_SYNC_ONLY: Final[ThirtyTwoBits] = cached_hresult(0x80070487, 'HRESULT_FROM_WIN32(ERROR_MESSAGE_SYNC_ONLY)')
    WIN32__ERROR_SOURCE_ELEMENT_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0x80070488, 'HRESULT_FROM_WIN32(ERROR_SOURCE_ELEMENT_EMPTY)')
    WIN32__ERROR_DESTINATION_ELEMENT_FULL: Final[ThirtyTwoBits] = cached_hresult(0x80070489, 'HRESULT_FROM_WIN32(ERROR_DESTINATION_ELEMENT_FULL)')
    WIN32__ERROR_ILLEGAL_ELEMENT_ADDRESS: Final[ThirtyTwoBits] = cached_hresult(0x8007048A, 'HRESULT_FROM_WIN32(ERROR_ILLEGAL_ELEMENT_ADDRESS)')
    WIN32__ERROR_MAGAZINE_NOT_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0x8007048B, 'HRESULT_FROM_WIN32(ERROR_MAGAZINE_NOT_PRESENT)')
    WIN32__ERROR_DEVICE_REINITIALIZATION_NEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8007048C, 'HRESULT_FROM_WIN32(ERROR_DEVICE_REINITIALIZATION_NEEDED)')
    WIN32__ERROR_DEVICE_REQUIRES_CLEANING: Final[ThirtyTwoBits] = cached_hresult(0x8007048D, 'HRESULT_FROM_WIN32(ERROR_DEVICE_REQUIRES_CLEANING)')
    WIN32__ERROR_DEVICE_DOOR_OPEN: Final[ThirtyTwoBits] = cached_hresult(0x8007048E, 'HRESULT_FROM_WIN32(ERROR_DEVICE_DOOR_OPEN)')
    WIN32__ERROR_DEVICE_NOT_CONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x8007048F, 'HRESULT_FROM_WIN32(ERROR_DEVICE_NOT_CONNECTED)')
    WIN32__ERROR_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80070490, 'HRESULT_FROM_WIN32(ERROR_NOT_FOUND)')
    WIN32__ERROR_NO_MATCH: Final[ThirtyTwoBits] = cached_hresult(0x80070491, 'HRESULT_FROM_WIN32(ERROR_NO_MATCH)')
    WIN32__ERROR_SET_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80070492, 'HRESULT_FROM_WIN32(ERROR_SET_NOT_FOUND)')
    WIN32__ERROR_POINT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80070493, 'HRESULT_FROM_WIN32(ERROR_POINT_NOT_FOUND)')
    WIN32__ERROR_NO_TRACKING_SERVICE: Final[ThirtyTwoBits] = cached_hresult(0x80070494, 'HRESULT_FROM_WIN32(ERROR_NO_TRACKING_SERVICE)')
    WIN32__ERROR_NO_VOLUME_ID: Final[ThirtyTwoBits] = cached_hresult(0x80070495, 'HRESULT_FROM_WIN32(ERROR_NO_VOLUME_ID)')
    WIN32__ERROR_UNABLE_TO_REMOVE_REPLACED: Final[ThirtyTwoBits] = cached_hresult(0x80070497, 'HRESULT_FROM_WIN32(ERROR_UNABLE_TO_REMOVE_REPLACED)')
    WIN32__ERROR_UNABLE_TO_MOVE_REPLACEMENT: Final[ThirtyTwoBits] = cached_hresult(0x80070498, 'HRESULT_FROM_WIN32(ERROR_UNABLE_TO_MOVE_REPLACEMENT)')
    WIN32__ERROR_UNABLE_TO_MOVE_REPLACEMENT_2: Final[ThirtyTwoBits] = cached_hresult(0x80070499, 'HRESULT_FROM_WIN32(ERROR_UNABLE_TO_MOVE_REPLACEMENT_2)')
    WIN32__ERROR_JOURNAL_DELETE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x8007049A, 'HRESULT_FROM_WIN32(ERROR_JOURNAL_DELETE_IN_PROGRESS)')
    WIN32__ERROR_JOURNAL_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0x8007049B, 'HRESULT_FROM_WIN32(ERROR_JOURNAL_NOT_ACTIVE)')
    WIN32__ERROR_POTENTIAL_FILE_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8007049C, 'HRESULT_FROM_WIN32(ERROR_POTENTIAL_FILE_FOUND)')
    WIN32__ERROR_JOURNAL_ENTRY_DELETED: Final[ThirtyTwoBits] = cached_hresult(0x8007049D, 'HRESULT_FROM_WIN32(ERROR_JOURNAL_ENTRY_DELETED)')
    WIN32__ERROR_PARTITION_TERMINATING: Final[ThirtyTwoBits] = cached_hresult(0x800704A0, 'HRESULT_FROM_WIN32(ERROR_PARTITION_TERMINATING)')
    WIN32__ERROR_SHUTDOWN_IS_SCHEDULED: Final[ThirtyTwoBits] = cached_hresult(0x800704A6, 'HRESULT_FROM_WIN32(ERROR_SHUTDOWN_IS_SCHEDULED)')
    WIN32__ERROR_SHUTDOWN_USERS_LOGGED_ON: Final[ThirtyTwoBits] = cached_hresult(0x800704A7, 'HRESULT_FROM_WIN32(ERROR_SHUTDOWN_USERS_LOGGED_ON)')
    WIN32__ERROR_SHUTDOWN_DISKS_NOT_IN_MAINTENANCE_MODE: Final[ThirtyTwoBits] = cached_hresult(0x800704A8, 'HRESULT_FROM_WIN32(ERROR_SHUTDOWN_DISKS_NOT_IN_MAINTENANCE_MODE)')
    WIN32__ERROR_BAD_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0x800704B0, 'HRESULT_FROM_WIN32(ERROR_BAD_DEVICE)')
    WIN32__ERROR_CONNECTION_UNAVAIL: Final[ThirtyTwoBits] = cached_hresult(0x800704B1, 'HRESULT_FROM_WIN32(ERROR_CONNECTION_UNAVAIL)')
    WIN32__ERROR_DEVICE_ALREADY_REMEMBERED: Final[ThirtyTwoBits] = cached_hresult(0x800704B2, 'HRESULT_FROM_WIN32(ERROR_DEVICE_ALREADY_REMEMBERED)')
    WIN32__ERROR_NO_NET_OR_BAD_PATH: Final[ThirtyTwoBits] = cached_hresult(0x800704B3, 'HRESULT_FROM_WIN32(ERROR_NO_NET_OR_BAD_PATH)')
    WIN32__ERROR_BAD_PROVIDER: Final[ThirtyTwoBits] = cached_hresult(0x800704B4, 'HRESULT_FROM_WIN32(ERROR_BAD_PROVIDER)')
    WIN32__ERROR_CANNOT_OPEN_PROFILE: Final[ThirtyTwoBits] = cached_hresult(0x800704B5, 'HRESULT_FROM_WIN32(ERROR_CANNOT_OPEN_PROFILE)')
    WIN32__ERROR_BAD_PROFILE: Final[ThirtyTwoBits] = cached_hresult(0x800704B6, 'HRESULT_FROM_WIN32(ERROR_BAD_PROFILE)')
    WIN32__ERROR_NOT_CONTAINER: Final[ThirtyTwoBits] = cached_hresult(0x800704B7, 'HRESULT_FROM_WIN32(ERROR_NOT_CONTAINER)')
    WIN32__ERROR_EXTENDED_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800704B8, 'HRESULT_FROM_WIN32(ERROR_EXTENDED_ERROR)')
    WIN32__ERROR_INVALID_GROUPNAME: Final[ThirtyTwoBits] = cached_hresult(0x800704B9, 'HRESULT_FROM_WIN32(ERROR_INVALID_GROUPNAME)')
    WIN32__ERROR_INVALID_COMPUTERNAME: Final[ThirtyTwoBits] = cached_hresult(0x800704BA, 'HRESULT_FROM_WIN32(ERROR_INVALID_COMPUTERNAME)')
    WIN32__ERROR_INVALID_EVENTNAME: Final[ThirtyTwoBits] = cached_hresult(0x800704BB, 'HRESULT_FROM_WIN32(ERROR_INVALID_EVENTNAME)')
    WIN32__ERROR_INVALID_DOMAINNAME: Final[ThirtyTwoBits] = cached_hresult(0x800704BC, 'HRESULT_FROM_WIN32(ERROR_INVALID_DOMAINNAME)')
    WIN32__ERROR_INVALID_SERVICENAME: Final[ThirtyTwoBits] = cached_hresult(0x800704BD, 'HRESULT_FROM_WIN32(ERROR_INVALID_SERVICENAME)')
    WIN32__ERROR_INVALID_NETNAME: Final[ThirtyTwoBits] = cached_hresult(0x800704BE, 'HRESULT_FROM_WIN32(ERROR_INVALID_NETNAME)')
    WIN32__ERROR_INVALID_SHARENAME: Final[ThirtyTwoBits] = cached_hresult(0x800704BF, 'HRESULT_FROM_WIN32(ERROR_INVALID_SHARENAME)')
    WIN32__ERROR_INVALID_PASSWORDNAME: Final[ThirtyTwoBits] = cached_hresult(0x800704C0, 'HRESULT_FROM_WIN32(ERROR_INVALID_PASSWORDNAME)')
    WIN32__ERROR_INVALID_MESSAGENAME: Final[ThirtyTwoBits] = cached_hresult(0x800704C1, 'HRESULT_FROM_WIN32(ERROR_INVALID_MESSAGENAME)')
    WIN32__ERROR_INVALID_MESSAGEDEST: Final[ThirtyTwoBits] = cached_hresult(0x800704C2, 'HRESULT_FROM_WIN32(ERROR_INVALID_MESSAGEDEST)')
    WIN32__ERROR_SESSION_CREDENTIAL_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x800704C3, 'HRESULT_FROM_WIN32(ERROR_SESSION_CREDENTIAL_CONFLICT)')
    WIN32__ERROR_REMOTE_SESSION_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x800704C4, 'HRESULT_FROM_WIN32(ERROR_REMOTE_SESSION_LIMIT_EXCEEDED)')
    WIN32__ERROR_DUP_DOMAINNAME: Final[ThirtyTwoBits] = cached_hresult(0x800704C5, 'HRESULT_FROM_WIN32(ERROR_DUP_DOMAINNAME)')
    WIN32__ERROR_NO_NETWORK: Final[ThirtyTwoBits] = cached_hresult(0x800704C6, 'HRESULT_FROM_WIN32(ERROR_NO_NETWORK)')
    WIN32__ERROR_CANCELLED: Final[ThirtyTwoBits] = cached_hresult(0x800704C7, 'HRESULT_FROM_WIN32(ERROR_CANCELLED)')
    WIN32__ERROR_USER_MAPPED_FILE: Final[ThirtyTwoBits] = cached_hresult(0x800704C8, 'HRESULT_FROM_WIN32(ERROR_USER_MAPPED_FILE)')
    WIN32__ERROR_CONNECTION_REFUSED: Final[ThirtyTwoBits] = cached_hresult(0x800704C9, 'HRESULT_FROM_WIN32(ERROR_CONNECTION_REFUSED)')
    WIN32__ERROR_GRACEFUL_DISCONNECT: Final[ThirtyTwoBits] = cached_hresult(0x800704CA, 'HRESULT_FROM_WIN32(ERROR_GRACEFUL_DISCONNECT)')
    WIN32__ERROR_ADDRESS_ALREADY_ASSOCIATED: Final[ThirtyTwoBits] = cached_hresult(0x800704CB, 'HRESULT_FROM_WIN32(ERROR_ADDRESS_ALREADY_ASSOCIATED)')
    WIN32__ERROR_ADDRESS_NOT_ASSOCIATED: Final[ThirtyTwoBits] = cached_hresult(0x800704CC, 'HRESULT_FROM_WIN32(ERROR_ADDRESS_NOT_ASSOCIATED)')
    WIN32__ERROR_CONNECTION_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800704CD, 'HRESULT_FROM_WIN32(ERROR_CONNECTION_INVALID)')
    WIN32__ERROR_CONNECTION_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0x800704CE, 'HRESULT_FROM_WIN32(ERROR_CONNECTION_ACTIVE)')
    WIN32__ERROR_NETWORK_UNREACHABLE: Final[ThirtyTwoBits] = cached_hresult(0x800704CF, 'HRESULT_FROM_WIN32(ERROR_NETWORK_UNREACHABLE)')
    WIN32__ERROR_HOST_UNREACHABLE: Final[ThirtyTwoBits] = cached_hresult(0x800704D0, 'HRESULT_FROM_WIN32(ERROR_HOST_UNREACHABLE)')
    WIN32__ERROR_PROTOCOL_UNREACHABLE: Final[ThirtyTwoBits] = cached_hresult(0x800704D1, 'HRESULT_FROM_WIN32(ERROR_PROTOCOL_UNREACHABLE)')
    WIN32__ERROR_PORT_UNREACHABLE: Final[ThirtyTwoBits] = cached_hresult(0x800704D2, 'HRESULT_FROM_WIN32(ERROR_PORT_UNREACHABLE)')
    WIN32__ERROR_REQUEST_ABORTED: Final[ThirtyTwoBits] = cached_hresult(0x800704D3, 'HRESULT_FROM_WIN32(ERROR_REQUEST_ABORTED)')
    WIN32__ERROR_CONNECTION_ABORTED: Final[ThirtyTwoBits] = cached_hresult(0x800704D4, 'HRESULT_FROM_WIN32(ERROR_CONNECTION_ABORTED)')
    WIN32__ERROR_RETRY: Final[ThirtyTwoBits] = cached_hresult(0x800704D5, 'HRESULT_FROM_WIN32(ERROR_RETRY)')
    WIN32__ERROR_CONNECTION_COUNT_LIMIT: Final[ThirtyTwoBits] = cached_hresult(0x800704D6, 'HRESULT_FROM_WIN32(ERROR_CONNECTION_COUNT_LIMIT)')
    WIN32__ERROR_LOGIN_TIME_RESTRICTION: Final[ThirtyTwoBits] = cached_hresult(0x800704D7, 'HRESULT_FROM_WIN32(ERROR_LOGIN_TIME_RESTRICTION)')
    WIN32__ERROR_LOGIN_WKSTA_RESTRICTION: Final[ThirtyTwoBits] = cached_hresult(0x800704D8, 'HRESULT_FROM_WIN32(ERROR_LOGIN_WKSTA_RESTRICTION)')
    WIN32__ERROR_INCORRECT_ADDRESS: Final[ThirtyTwoBits] = cached_hresult(0x800704D9, 'HRESULT_FROM_WIN32(ERROR_INCORRECT_ADDRESS)')
    WIN32__ERROR_ALREADY_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x800704DA, 'HRESULT_FROM_WIN32(ERROR_ALREADY_REGISTERED)')
    WIN32__ERROR_SERVICE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800704DB, 'HRESULT_FROM_WIN32(ERROR_SERVICE_NOT_FOUND)')
    WIN32__ERROR_NOT_AUTHENTICATED: Final[ThirtyTwoBits] = cached_hresult(0x800704DC, 'HRESULT_FROM_WIN32(ERROR_NOT_AUTHENTICATED)')
    WIN32__ERROR_NOT_LOGGED_ON: Final[ThirtyTwoBits] = cached_hresult(0x800704DD, 'HRESULT_FROM_WIN32(ERROR_NOT_LOGGED_ON)')
    WIN32__ERROR_CONTINUE: Final[ThirtyTwoBits] = cached_hresult(0x800704DE, 'HRESULT_FROM_WIN32(ERROR_CONTINUE)')
    WIN32__ERROR_ALREADY_INITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0x800704DF, 'HRESULT_FROM_WIN32(ERROR_ALREADY_INITIALIZED)')
    WIN32__ERROR_NO_MORE_DEVICES: Final[ThirtyTwoBits] = cached_hresult(0x800704E0, 'HRESULT_FROM_WIN32(ERROR_NO_MORE_DEVICES)')
    WIN32__ERROR_NO_SUCH_SITE: Final[ThirtyTwoBits] = cached_hresult(0x800704E1, 'HRESULT_FROM_WIN32(ERROR_NO_SUCH_SITE)')
    WIN32__ERROR_DOMAIN_CONTROLLER_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800704E2, 'HRESULT_FROM_WIN32(ERROR_DOMAIN_CONTROLLER_EXISTS)')
    WIN32__ERROR_ONLY_IF_CONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x800704E3, 'HRESULT_FROM_WIN32(ERROR_ONLY_IF_CONNECTED)')
    WIN32__ERROR_OVERRIDE_NOCHANGES: Final[ThirtyTwoBits] = cached_hresult(0x800704E4, 'HRESULT_FROM_WIN32(ERROR_OVERRIDE_NOCHANGES)')
    WIN32__ERROR_BAD_USER_PROFILE: Final[ThirtyTwoBits] = cached_hresult(0x800704E5, 'HRESULT_FROM_WIN32(ERROR_BAD_USER_PROFILE)')
    WIN32__ERROR_NOT_SUPPORTED_ON_SBS: Final[ThirtyTwoBits] = cached_hresult(0x800704E6, 'HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED_ON_SBS)')
    WIN32__ERROR_SERVER_SHUTDOWN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x800704E7, 'HRESULT_FROM_WIN32(ERROR_SERVER_SHUTDOWN_IN_PROGRESS)')
    WIN32__ERROR_HOST_DOWN: Final[ThirtyTwoBits] = cached_hresult(0x800704E8, 'HRESULT_FROM_WIN32(ERROR_HOST_DOWN)')
    WIN32__ERROR_NON_ACCOUNT_SID: Final[ThirtyTwoBits] = cached_hresult(0x800704E9, 'HRESULT_FROM_WIN32(ERROR_NON_ACCOUNT_SID)')
    WIN32__ERROR_NON_DOMAIN_SID: Final[ThirtyTwoBits] = cached_hresult(0x800704EA, 'HRESULT_FROM_WIN32(ERROR_NON_DOMAIN_SID)')
    WIN32__ERROR_APPHELP_BLOCK: Final[ThirtyTwoBits] = cached_hresult(0x800704EB, 'HRESULT_FROM_WIN32(ERROR_APPHELP_BLOCK)')
    WIN32__ERROR_ACCESS_DISABLED_BY_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x800704EC, 'HRESULT_FROM_WIN32(ERROR_ACCESS_DISABLED_BY_POLICY)')
    WIN32__ERROR_REG_NAT_CONSUMPTION: Final[ThirtyTwoBits] = cached_hresult(0x800704ED, 'HRESULT_FROM_WIN32(ERROR_REG_NAT_CONSUMPTION)')
    WIN32__ERROR_CSCSHARE_OFFLINE: Final[ThirtyTwoBits] = cached_hresult(0x800704EE, 'HRESULT_FROM_WIN32(ERROR_CSCSHARE_OFFLINE)')
    WIN32__ERROR_PKINIT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x800704EF, 'HRESULT_FROM_WIN32(ERROR_PKINIT_FAILURE)')
    WIN32__ERROR_SMARTCARD_SUBSYSTEM_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x800704F0, 'HRESULT_FROM_WIN32(ERROR_SMARTCARD_SUBSYSTEM_FAILURE)')
    WIN32__ERROR_DOWNGRADE_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0x800704F1, 'HRESULT_FROM_WIN32(ERROR_DOWNGRADE_DETECTED)')
    WIN32__ERROR_MACHINE_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x800704F7, 'HRESULT_FROM_WIN32(ERROR_MACHINE_LOCKED)')
    WIN32__ERROR_SMB_GUEST_LOGON_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x800704F8, 'HRESULT_FROM_WIN32(ERROR_SMB_GUEST_LOGON_BLOCKED)')
    WIN32__ERROR_CALLBACK_SUPPLIED_INVALID_DATA: Final[ThirtyTwoBits] = cached_hresult(0x800704F9, 'HRESULT_FROM_WIN32(ERROR_CALLBACK_SUPPLIED_INVALID_DATA)')
    WIN32__ERROR_SYNC_FOREGROUND_REFRESH_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x800704FA, 'HRESULT_FROM_WIN32(ERROR_SYNC_FOREGROUND_REFRESH_REQUIRED)')
    WIN32__ERROR_DRIVER_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x800704FB, 'HRESULT_FROM_WIN32(ERROR_DRIVER_BLOCKED)')
    WIN32__ERROR_INVALID_IMPORT_OF_NON_DLL: Final[ThirtyTwoBits] = cached_hresult(0x800704FC, 'HRESULT_FROM_WIN32(ERROR_INVALID_IMPORT_OF_NON_DLL)')
    WIN32__ERROR_ACCESS_DISABLED_WEBBLADE: Final[ThirtyTwoBits] = cached_hresult(0x800704FD, 'HRESULT_FROM_WIN32(ERROR_ACCESS_DISABLED_WEBBLADE)')
    WIN32__ERROR_ACCESS_DISABLED_WEBBLADE_TAMPER: Final[ThirtyTwoBits] = cached_hresult(0x800704FE, 'HRESULT_FROM_WIN32(ERROR_ACCESS_DISABLED_WEBBLADE_TAMPER)')
    WIN32__ERROR_RECOVERY_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x800704FF, 'HRESULT_FROM_WIN32(ERROR_RECOVERY_FAILURE)')
    WIN32__ERROR_ALREADY_FIBER: Final[ThirtyTwoBits] = cached_hresult(0x80070500, 'HRESULT_FROM_WIN32(ERROR_ALREADY_FIBER)')
    WIN32__ERROR_ALREADY_THREAD: Final[ThirtyTwoBits] = cached_hresult(0x80070501, 'HRESULT_FROM_WIN32(ERROR_ALREADY_THREAD)')
    WIN32__ERROR_STACK_BUFFER_OVERRUN: Final[ThirtyTwoBits] = cached_hresult(0x80070502, 'HRESULT_FROM_WIN32(ERROR_STACK_BUFFER_OVERRUN)')
    WIN32__ERROR_PARAMETER_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80070503, 'HRESULT_FROM_WIN32(ERROR_PARAMETER_QUOTA_EXCEEDED)')
    WIN32__ERROR_DEBUGGER_INACTIVE: Final[ThirtyTwoBits] = cached_hresult(0x80070504, 'HRESULT_FROM_WIN32(ERROR_DEBUGGER_INACTIVE)')
    WIN32__ERROR_DELAY_LOAD_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80070505, 'HRESULT_FROM_WIN32(ERROR_DELAY_LOAD_FAILED)')
    WIN32__ERROR_VDM_DISALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80070506, 'HRESULT_FROM_WIN32(ERROR_VDM_DISALLOWED)')
    WIN32__ERROR_UNIDENTIFIED_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80070507, 'HRESULT_FROM_WIN32(ERROR_UNIDENTIFIED_ERROR)')
    WIN32__ERROR_INVALID_CRUNTIME_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0x80070508, 'HRESULT_FROM_WIN32(ERROR_INVALID_CRUNTIME_PARAMETER)')
    WIN32__ERROR_BEYOND_VDL: Final[ThirtyTwoBits] = cached_hresult(0x80070509, 'HRESULT_FROM_WIN32(ERROR_BEYOND_VDL)')
    WIN32__ERROR_INCOMPATIBLE_SERVICE_SID_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x8007050A, 'HRESULT_FROM_WIN32(ERROR_INCOMPATIBLE_SERVICE_SID_TYPE)')
    WIN32__ERROR_DRIVER_PROCESS_TERMINATED: Final[ThirtyTwoBits] = cached_hresult(0x8007050B, 'HRESULT_FROM_WIN32(ERROR_DRIVER_PROCESS_TERMINATED)')
    WIN32__ERROR_IMPLEMENTATION_LIMIT: Final[ThirtyTwoBits] = cached_hresult(0x8007050C, 'HRESULT_FROM_WIN32(ERROR_IMPLEMENTATION_LIMIT)')
    WIN32__ERROR_PROCESS_IS_PROTECTED: Final[ThirtyTwoBits] = cached_hresult(0x8007050D, 'HRESULT_FROM_WIN32(ERROR_PROCESS_IS_PROTECTED)')
    WIN32__ERROR_SERVICE_NOTIFY_CLIENT_LAGGING: Final[ThirtyTwoBits] = cached_hresult(0x8007050E, 'HRESULT_FROM_WIN32(ERROR_SERVICE_NOTIFY_CLIENT_LAGGING)')
    WIN32__ERROR_DISK_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8007050F, 'HRESULT_FROM_WIN32(ERROR_DISK_QUOTA_EXCEEDED)')
    WIN32__ERROR_CONTENT_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x80070510, 'HRESULT_FROM_WIN32(ERROR_CONTENT_BLOCKED)')
    WIN32__ERROR_INCOMPATIBLE_SERVICE_PRIVILEGE: Final[ThirtyTwoBits] = cached_hresult(0x80070511, 'HRESULT_FROM_WIN32(ERROR_INCOMPATIBLE_SERVICE_PRIVILEGE)')
    WIN32__ERROR_APP_HANG: Final[ThirtyTwoBits] = cached_hresult(0x80070512, 'HRESULT_FROM_WIN32(ERROR_APP_HANG)')
    WIN32__ERROR_INVALID_LABEL: Final[ThirtyTwoBits] = cached_hresult(0x80070513, 'HRESULT_FROM_WIN32(ERROR_INVALID_LABEL)')
    WIN32__ERROR_NOT_ALL_ASSIGNED: Final[ThirtyTwoBits] = cached_hresult(0x80070514, 'HRESULT_FROM_WIN32(ERROR_NOT_ALL_ASSIGNED)')
    WIN32__ERROR_SOME_NOT_MAPPED: Final[ThirtyTwoBits] = cached_hresult(0x80070515, 'HRESULT_FROM_WIN32(ERROR_SOME_NOT_MAPPED)')
    WIN32__ERROR_NO_QUOTAS_FOR_ACCOUNT: Final[ThirtyTwoBits] = cached_hresult(0x80070516, 'HRESULT_FROM_WIN32(ERROR_NO_QUOTAS_FOR_ACCOUNT)')
    WIN32__ERROR_LOCAL_USER_SESSION_KEY: Final[ThirtyTwoBits] = cached_hresult(0x80070517, 'HRESULT_FROM_WIN32(ERROR_LOCAL_USER_SESSION_KEY)')
    WIN32__ERROR_NULL_LM_PASSWORD: Final[ThirtyTwoBits] = cached_hresult(0x80070518, 'HRESULT_FROM_WIN32(ERROR_NULL_LM_PASSWORD)')
    WIN32__ERROR_UNKNOWN_REVISION: Final[ThirtyTwoBits] = cached_hresult(0x80070519, 'HRESULT_FROM_WIN32(ERROR_UNKNOWN_REVISION)')
    WIN32__ERROR_REVISION_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x8007051A, 'HRESULT_FROM_WIN32(ERROR_REVISION_MISMATCH)')
    WIN32__ERROR_INVALID_OWNER: Final[ThirtyTwoBits] = cached_hresult(0x8007051B, 'HRESULT_FROM_WIN32(ERROR_INVALID_OWNER)')
    WIN32__ERROR_INVALID_PRIMARY_GROUP: Final[ThirtyTwoBits] = cached_hresult(0x8007051C, 'HRESULT_FROM_WIN32(ERROR_INVALID_PRIMARY_GROUP)')
    WIN32__ERROR_NO_IMPERSONATION_TOKEN: Final[ThirtyTwoBits] = cached_hresult(0x8007051D, 'HRESULT_FROM_WIN32(ERROR_NO_IMPERSONATION_TOKEN)')
    WIN32__ERROR_CANT_DISABLE_MANDATORY: Final[ThirtyTwoBits] = cached_hresult(0x8007051E, 'HRESULT_FROM_WIN32(ERROR_CANT_DISABLE_MANDATORY)')
    WIN32__ERROR_NO_LOGON_SERVERS: Final[ThirtyTwoBits] = cached_hresult(0x8007051F, 'HRESULT_FROM_WIN32(ERROR_NO_LOGON_SERVERS)')
    WIN32__ERROR_NO_SUCH_LOGON_SESSION: Final[ThirtyTwoBits] = cached_hresult(0x80070520, 'HRESULT_FROM_WIN32(ERROR_NO_SUCH_LOGON_SESSION)')
    WIN32__ERROR_NO_SUCH_PRIVILEGE: Final[ThirtyTwoBits] = cached_hresult(0x80070521, 'HRESULT_FROM_WIN32(ERROR_NO_SUCH_PRIVILEGE)')
    WIN32__ERROR_PRIVILEGE_NOT_HELD: Final[ThirtyTwoBits] = cached_hresult(0x80070522, 'HRESULT_FROM_WIN32(ERROR_PRIVILEGE_NOT_HELD)')
    WIN32__ERROR_INVALID_ACCOUNT_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80070523, 'HRESULT_FROM_WIN32(ERROR_INVALID_ACCOUNT_NAME)')
    WIN32__ERROR_USER_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80070524, 'HRESULT_FROM_WIN32(ERROR_USER_EXISTS)')
    WIN32__ERROR_NO_SUCH_USER: Final[ThirtyTwoBits] = cached_hresult(0x80070525, 'HRESULT_FROM_WIN32(ERROR_NO_SUCH_USER)')
    WIN32__ERROR_GROUP_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80070526, 'HRESULT_FROM_WIN32(ERROR_GROUP_EXISTS)')
    WIN32__ERROR_NO_SUCH_GROUP: Final[ThirtyTwoBits] = cached_hresult(0x80070527, 'HRESULT_FROM_WIN32(ERROR_NO_SUCH_GROUP)')
    WIN32__ERROR_MEMBER_IN_GROUP: Final[ThirtyTwoBits] = cached_hresult(0x80070528, 'HRESULT_FROM_WIN32(ERROR_MEMBER_IN_GROUP)')
    WIN32__ERROR_MEMBER_NOT_IN_GROUP: Final[ThirtyTwoBits] = cached_hresult(0x80070529, 'HRESULT_FROM_WIN32(ERROR_MEMBER_NOT_IN_GROUP)')
    WIN32__ERROR_LAST_ADMIN: Final[ThirtyTwoBits] = cached_hresult(0x8007052A, 'HRESULT_FROM_WIN32(ERROR_LAST_ADMIN)')
    WIN32__ERROR_WRONG_PASSWORD: Final[ThirtyTwoBits] = cached_hresult(0x8007052B, 'HRESULT_FROM_WIN32(ERROR_WRONG_PASSWORD)')
    WIN32__ERROR_ILL_FORMED_PASSWORD: Final[ThirtyTwoBits] = cached_hresult(0x8007052C, 'HRESULT_FROM_WIN32(ERROR_ILL_FORMED_PASSWORD)')
    WIN32__ERROR_PASSWORD_RESTRICTION: Final[ThirtyTwoBits] = cached_hresult(0x8007052D, 'HRESULT_FROM_WIN32(ERROR_PASSWORD_RESTRICTION)')
    WIN32__ERROR_LOGON_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8007052E, 'HRESULT_FROM_WIN32(ERROR_LOGON_FAILURE)')
    WIN32__ERROR_ACCOUNT_RESTRICTION: Final[ThirtyTwoBits] = cached_hresult(0x8007052F, 'HRESULT_FROM_WIN32(ERROR_ACCOUNT_RESTRICTION)')
    WIN32__ERROR_INVALID_LOGON_HOURS: Final[ThirtyTwoBits] = cached_hresult(0x80070530, 'HRESULT_FROM_WIN32(ERROR_INVALID_LOGON_HOURS)')
    WIN32__ERROR_INVALID_WORKSTATION: Final[ThirtyTwoBits] = cached_hresult(0x80070531, 'HRESULT_FROM_WIN32(ERROR_INVALID_WORKSTATION)')
    WIN32__ERROR_PASSWORD_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x80070532, 'HRESULT_FROM_WIN32(ERROR_PASSWORD_EXPIRED)')
    WIN32__ERROR_ACCOUNT_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80070533, 'HRESULT_FROM_WIN32(ERROR_ACCOUNT_DISABLED)')
    WIN32__ERROR_NONE_MAPPED: Final[ThirtyTwoBits] = cached_hresult(0x80070534, 'HRESULT_FROM_WIN32(ERROR_NONE_MAPPED)')
    WIN32__ERROR_TOO_MANY_LUIDS_REQUESTED: Final[ThirtyTwoBits] = cached_hresult(0x80070535, 'HRESULT_FROM_WIN32(ERROR_TOO_MANY_LUIDS_REQUESTED)')
    WIN32__ERROR_LUIDS_EXHAUSTED: Final[ThirtyTwoBits] = cached_hresult(0x80070536, 'HRESULT_FROM_WIN32(ERROR_LUIDS_EXHAUSTED)')
    WIN32__ERROR_INVALID_SUB_AUTHORITY: Final[ThirtyTwoBits] = cached_hresult(0x80070537, 'HRESULT_FROM_WIN32(ERROR_INVALID_SUB_AUTHORITY)')
    WIN32__ERROR_INVALID_ACL: Final[ThirtyTwoBits] = cached_hresult(0x80070538, 'HRESULT_FROM_WIN32(ERROR_INVALID_ACL)')
    WIN32__ERROR_INVALID_SID: Final[ThirtyTwoBits] = cached_hresult(0x80070539, 'HRESULT_FROM_WIN32(ERROR_INVALID_SID)')
    WIN32__ERROR_INVALID_SECURITY_DESCR: Final[ThirtyTwoBits] = cached_hresult(0x8007053A, 'HRESULT_FROM_WIN32(ERROR_INVALID_SECURITY_DESCR)')
    WIN32__ERROR_BAD_INHERITANCE_ACL: Final[ThirtyTwoBits] = cached_hresult(0x8007053C, 'HRESULT_FROM_WIN32(ERROR_BAD_INHERITANCE_ACL)')
    WIN32__ERROR_SERVER_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x8007053D, 'HRESULT_FROM_WIN32(ERROR_SERVER_DISABLED)')
    WIN32__ERROR_SERVER_NOT_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x8007053E, 'HRESULT_FROM_WIN32(ERROR_SERVER_NOT_DISABLED)')
    WIN32__ERROR_INVALID_ID_AUTHORITY: Final[ThirtyTwoBits] = cached_hresult(0x8007053F, 'HRESULT_FROM_WIN32(ERROR_INVALID_ID_AUTHORITY)')
    WIN32__ERROR_ALLOTTED_SPACE_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80070540, 'HRESULT_FROM_WIN32(ERROR_ALLOTTED_SPACE_EXCEEDED)')
    WIN32__ERROR_INVALID_GROUP_ATTRIBUTES: Final[ThirtyTwoBits] = cached_hresult(0x80070541, 'HRESULT_FROM_WIN32(ERROR_INVALID_GROUP_ATTRIBUTES)')
    WIN32__ERROR_BAD_IMPERSONATION_LEVEL: Final[ThirtyTwoBits] = cached_hresult(0x80070542, 'HRESULT_FROM_WIN32(ERROR_BAD_IMPERSONATION_LEVEL)')
    WIN32__ERROR_CANT_OPEN_ANONYMOUS: Final[ThirtyTwoBits] = cached_hresult(0x80070543, 'HRESULT_FROM_WIN32(ERROR_CANT_OPEN_ANONYMOUS)')
    WIN32__ERROR_BAD_VALIDATION_CLASS: Final[ThirtyTwoBits] = cached_hresult(0x80070544, 'HRESULT_FROM_WIN32(ERROR_BAD_VALIDATION_CLASS)')
    WIN32__ERROR_BAD_TOKEN_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x80070545, 'HRESULT_FROM_WIN32(ERROR_BAD_TOKEN_TYPE)')
    WIN32__ERROR_NO_SECURITY_ON_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x80070546, 'HRESULT_FROM_WIN32(ERROR_NO_SECURITY_ON_OBJECT)')
    WIN32__ERROR_CANT_ACCESS_DOMAIN_INFO: Final[ThirtyTwoBits] = cached_hresult(0x80070547, 'HRESULT_FROM_WIN32(ERROR_CANT_ACCESS_DOMAIN_INFO)')
    WIN32__ERROR_INVALID_SERVER_STATE: Final[ThirtyTwoBits] = cached_hresult(0x80070548, 'HRESULT_FROM_WIN32(ERROR_INVALID_SERVER_STATE)')
    WIN32__ERROR_INVALID_DOMAIN_STATE: Final[ThirtyTwoBits] = cached_hresult(0x80070549, 'HRESULT_FROM_WIN32(ERROR_INVALID_DOMAIN_STATE)')
    WIN32__ERROR_INVALID_DOMAIN_ROLE: Final[ThirtyTwoBits] = cached_hresult(0x8007054A, 'HRESULT_FROM_WIN32(ERROR_INVALID_DOMAIN_ROLE)')
    WIN32__ERROR_NO_SUCH_DOMAIN: Final[ThirtyTwoBits] = cached_hresult(0x8007054B, 'HRESULT_FROM_WIN32(ERROR_NO_SUCH_DOMAIN)')
    WIN32__ERROR_DOMAIN_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x8007054C, 'HRESULT_FROM_WIN32(ERROR_DOMAIN_EXISTS)')
    WIN32__ERROR_DOMAIN_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8007054D, 'HRESULT_FROM_WIN32(ERROR_DOMAIN_LIMIT_EXCEEDED)')
    WIN32__ERROR_INTERNAL_DB_CORRUPTION: Final[ThirtyTwoBits] = cached_hresult(0x8007054E, 'HRESULT_FROM_WIN32(ERROR_INTERNAL_DB_CORRUPTION)')
    WIN32__ERROR_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8007054F, 'HRESULT_FROM_WIN32(ERROR_INTERNAL_ERROR)')
    WIN32__ERROR_GENERIC_NOT_MAPPED: Final[ThirtyTwoBits] = cached_hresult(0x80070550, 'HRESULT_FROM_WIN32(ERROR_GENERIC_NOT_MAPPED)')
    WIN32__ERROR_BAD_DESCRIPTOR_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0x80070551, 'HRESULT_FROM_WIN32(ERROR_BAD_DESCRIPTOR_FORMAT)')
    WIN32__ERROR_NOT_LOGON_PROCESS: Final[ThirtyTwoBits] = cached_hresult(0x80070552, 'HRESULT_FROM_WIN32(ERROR_NOT_LOGON_PROCESS)')
    WIN32__ERROR_LOGON_SESSION_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80070553, 'HRESULT_FROM_WIN32(ERROR_LOGON_SESSION_EXISTS)')
    WIN32__ERROR_NO_SUCH_PACKAGE: Final[ThirtyTwoBits] = cached_hresult(0x80070554, 'HRESULT_FROM_WIN32(ERROR_NO_SUCH_PACKAGE)')
    WIN32__ERROR_BAD_LOGON_SESSION_STATE: Final[ThirtyTwoBits] = cached_hresult(0x80070555, 'HRESULT_FROM_WIN32(ERROR_BAD_LOGON_SESSION_STATE)')
    WIN32__ERROR_LOGON_SESSION_COLLISION: Final[ThirtyTwoBits] = cached_hresult(0x80070556, 'HRESULT_FROM_WIN32(ERROR_LOGON_SESSION_COLLISION)')
    WIN32__ERROR_INVALID_LOGON_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x80070557, 'HRESULT_FROM_WIN32(ERROR_INVALID_LOGON_TYPE)')
    WIN32__ERROR_CANNOT_IMPERSONATE: Final[ThirtyTwoBits] = cached_hresult(0x80070558, 'HRESULT_FROM_WIN32(ERROR_CANNOT_IMPERSONATE)')
    WIN32__ERROR_RXACT_INVALID_STATE: Final[ThirtyTwoBits] = cached_hresult(0x80070559, 'HRESULT_FROM_WIN32(ERROR_RXACT_INVALID_STATE)')
    WIN32__ERROR_RXACT_COMMIT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8007055A, 'HRESULT_FROM_WIN32(ERROR_RXACT_COMMIT_FAILURE)')
    WIN32__ERROR_SPECIAL_ACCOUNT: Final[ThirtyTwoBits] = cached_hresult(0x8007055B, 'HRESULT_FROM_WIN32(ERROR_SPECIAL_ACCOUNT)')
    WIN32__ERROR_SPECIAL_GROUP: Final[ThirtyTwoBits] = cached_hresult(0x8007055C, 'HRESULT_FROM_WIN32(ERROR_SPECIAL_GROUP)')
    WIN32__ERROR_SPECIAL_USER: Final[ThirtyTwoBits] = cached_hresult(0x8007055D, 'HRESULT_FROM_WIN32(ERROR_SPECIAL_USER)')
    WIN32__ERROR_MEMBERS_PRIMARY_GROUP: Final[ThirtyTwoBits] = cached_hresult(0x8007055E, 'HRESULT_FROM_WIN32(ERROR_MEMBERS_PRIMARY_GROUP)')
    WIN32__ERROR_TOKEN_ALREADY_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x8007055F, 'HRESULT_FROM_WIN32(ERROR_TOKEN_ALREADY_IN_USE)')
    WIN32__ERROR_NO_SUCH_ALIAS: Final[ThirtyTwoBits] = cached_hresult(0x80070560, 'HRESULT_FROM_WIN32(ERROR_NO_SUCH_ALIAS)')
    WIN32__ERROR_MEMBER_NOT_IN_ALIAS: Final[ThirtyTwoBits] = cached_hresult(0x80070561, 'HRESULT_FROM_WIN32(ERROR_MEMBER_NOT_IN_ALIAS)')
    WIN32__ERROR_MEMBER_IN_ALIAS: Final[ThirtyTwoBits] = cached_hresult(0x80070562, 'HRESULT_FROM_WIN32(ERROR_MEMBER_IN_ALIAS)')
    WIN32__ERROR_ALIAS_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80070563, 'HRESULT_FROM_WIN32(ERROR_ALIAS_EXISTS)')
    WIN32__ERROR_LOGON_NOT_GRANTED: Final[ThirtyTwoBits] = cached_hresult(0x80070564, 'HRESULT_FROM_WIN32(ERROR_LOGON_NOT_GRANTED)')
    WIN32__ERROR_TOO_MANY_SECRETS: Final[ThirtyTwoBits] = cached_hresult(0x80070565, 'HRESULT_FROM_WIN32(ERROR_TOO_MANY_SECRETS)')
    WIN32__ERROR_SECRET_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0x80070566, 'HRESULT_FROM_WIN32(ERROR_SECRET_TOO_LONG)')
    WIN32__ERROR_INTERNAL_DB_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80070567, 'HRESULT_FROM_WIN32(ERROR_INTERNAL_DB_ERROR)')
    WIN32__ERROR_TOO_MANY_CONTEXT_IDS: Final[ThirtyTwoBits] = cached_hresult(0x80070568, 'HRESULT_FROM_WIN32(ERROR_TOO_MANY_CONTEXT_IDS)')
    WIN32__ERROR_LOGON_TYPE_NOT_GRANTED: Final[ThirtyTwoBits] = cached_hresult(0x80070569, 'HRESULT_FROM_WIN32(ERROR_LOGON_TYPE_NOT_GRANTED)')
    WIN32__ERROR_NT_CROSS_ENCRYPTION_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8007056A, 'HRESULT_FROM_WIN32(ERROR_NT_CROSS_ENCRYPTION_REQUIRED)')
    WIN32__ERROR_NO_SUCH_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0x8007056B, 'HRESULT_FROM_WIN32(ERROR_NO_SUCH_MEMBER)')
    WIN32__ERROR_INVALID_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0x8007056C, 'HRESULT_FROM_WIN32(ERROR_INVALID_MEMBER)')
    WIN32__ERROR_TOO_MANY_SIDS: Final[ThirtyTwoBits] = cached_hresult(0x8007056D, 'HRESULT_FROM_WIN32(ERROR_TOO_MANY_SIDS)')
    WIN32__ERROR_LM_CROSS_ENCRYPTION_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8007056E, 'HRESULT_FROM_WIN32(ERROR_LM_CROSS_ENCRYPTION_REQUIRED)')
    WIN32__ERROR_NO_INHERITANCE: Final[ThirtyTwoBits] = cached_hresult(0x8007056F, 'HRESULT_FROM_WIN32(ERROR_NO_INHERITANCE)')
    WIN32__ERROR_FILE_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x80070570, 'HRESULT_FROM_WIN32(ERROR_FILE_CORRUPT)')
    WIN32__ERROR_DISK_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x80070571, 'HRESULT_FROM_WIN32(ERROR_DISK_CORRUPT)')
    WIN32__ERROR_NO_USER_SESSION_KEY: Final[ThirtyTwoBits] = cached_hresult(0x80070572, 'HRESULT_FROM_WIN32(ERROR_NO_USER_SESSION_KEY)')
    WIN32__ERROR_LICENSE_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80070573, 'HRESULT_FROM_WIN32(ERROR_LICENSE_QUOTA_EXCEEDED)')
    WIN32__ERROR_WRONG_TARGET_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80070574, 'HRESULT_FROM_WIN32(ERROR_WRONG_TARGET_NAME)')
    WIN32__ERROR_MUTUAL_AUTH_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80070575, 'HRESULT_FROM_WIN32(ERROR_MUTUAL_AUTH_FAILED)')
    WIN32__ERROR_TIME_SKEW: Final[ThirtyTwoBits] = cached_hresult(0x80070576, 'HRESULT_FROM_WIN32(ERROR_TIME_SKEW)')
    WIN32__ERROR_CURRENT_DOMAIN_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80070577, 'HRESULT_FROM_WIN32(ERROR_CURRENT_DOMAIN_NOT_ALLOWED)')
    WIN32__ERROR_INVALID_WINDOW_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80070578, 'HRESULT_FROM_WIN32(ERROR_INVALID_WINDOW_HANDLE)')
    WIN32__ERROR_INVALID_MENU_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80070579, 'HRESULT_FROM_WIN32(ERROR_INVALID_MENU_HANDLE)')
    WIN32__ERROR_INVALID_CURSOR_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x8007057A, 'HRESULT_FROM_WIN32(ERROR_INVALID_CURSOR_HANDLE)')
    WIN32__ERROR_INVALID_ACCEL_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x8007057B, 'HRESULT_FROM_WIN32(ERROR_INVALID_ACCEL_HANDLE)')
    WIN32__ERROR_INVALID_HOOK_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x8007057C, 'HRESULT_FROM_WIN32(ERROR_INVALID_HOOK_HANDLE)')
    WIN32__ERROR_INVALID_DWP_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x8007057D, 'HRESULT_FROM_WIN32(ERROR_INVALID_DWP_HANDLE)')
    WIN32__ERROR_TLW_WITH_WSCHILD: Final[ThirtyTwoBits] = cached_hresult(0x8007057E, 'HRESULT_FROM_WIN32(ERROR_TLW_WITH_WSCHILD)')
    WIN32__ERROR_CANNOT_FIND_WND_CLASS: Final[ThirtyTwoBits] = cached_hresult(0x8007057F, 'HRESULT_FROM_WIN32(ERROR_CANNOT_FIND_WND_CLASS)')
    WIN32__ERROR_WINDOW_OF_OTHER_THREAD: Final[ThirtyTwoBits] = cached_hresult(0x80070580, 'HRESULT_FROM_WIN32(ERROR_WINDOW_OF_OTHER_THREAD)')
    WIN32__ERROR_HOTKEY_ALREADY_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x80070581, 'HRESULT_FROM_WIN32(ERROR_HOTKEY_ALREADY_REGISTERED)')
    WIN32__ERROR_CLASS_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80070582, 'HRESULT_FROM_WIN32(ERROR_CLASS_ALREADY_EXISTS)')
    WIN32__ERROR_CLASS_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x80070583, 'HRESULT_FROM_WIN32(ERROR_CLASS_DOES_NOT_EXIST)')
    WIN32__ERROR_CLASS_HAS_WINDOWS: Final[ThirtyTwoBits] = cached_hresult(0x80070584, 'HRESULT_FROM_WIN32(ERROR_CLASS_HAS_WINDOWS)')
    WIN32__ERROR_INVALID_INDEX: Final[ThirtyTwoBits] = cached_hresult(0x80070585, 'HRESULT_FROM_WIN32(ERROR_INVALID_INDEX)')
    WIN32__ERROR_INVALID_ICON_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80070586, 'HRESULT_FROM_WIN32(ERROR_INVALID_ICON_HANDLE)')
    WIN32__ERROR_PRIVATE_DIALOG_INDEX: Final[ThirtyTwoBits] = cached_hresult(0x80070587, 'HRESULT_FROM_WIN32(ERROR_PRIVATE_DIALOG_INDEX)')
    WIN32__ERROR_LISTBOX_ID_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80070588, 'HRESULT_FROM_WIN32(ERROR_LISTBOX_ID_NOT_FOUND)')
    WIN32__ERROR_NO_WILDCARD_CHARACTERS: Final[ThirtyTwoBits] = cached_hresult(0x80070589, 'HRESULT_FROM_WIN32(ERROR_NO_WILDCARD_CHARACTERS)')
    WIN32__ERROR_CLIPBOARD_NOT_OPEN: Final[ThirtyTwoBits] = cached_hresult(0x8007058A, 'HRESULT_FROM_WIN32(ERROR_CLIPBOARD_NOT_OPEN)')
    WIN32__ERROR_HOTKEY_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x8007058B, 'HRESULT_FROM_WIN32(ERROR_HOTKEY_NOT_REGISTERED)')
    WIN32__ERROR_WINDOW_NOT_DIALOG: Final[ThirtyTwoBits] = cached_hresult(0x8007058C, 'HRESULT_FROM_WIN32(ERROR_WINDOW_NOT_DIALOG)')
    WIN32__ERROR_CONTROL_ID_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8007058D, 'HRESULT_FROM_WIN32(ERROR_CONTROL_ID_NOT_FOUND)')
    WIN32__ERROR_INVALID_COMBOBOX_MESSAGE: Final[ThirtyTwoBits] = cached_hresult(0x8007058E, 'HRESULT_FROM_WIN32(ERROR_INVALID_COMBOBOX_MESSAGE)')
    WIN32__ERROR_WINDOW_NOT_COMBOBOX: Final[ThirtyTwoBits] = cached_hresult(0x8007058F, 'HRESULT_FROM_WIN32(ERROR_WINDOW_NOT_COMBOBOX)')
    WIN32__ERROR_INVALID_EDIT_HEIGHT: Final[ThirtyTwoBits] = cached_hresult(0x80070590, 'HRESULT_FROM_WIN32(ERROR_INVALID_EDIT_HEIGHT)')
    WIN32__ERROR_DC_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80070591, 'HRESULT_FROM_WIN32(ERROR_DC_NOT_FOUND)')
    WIN32__ERROR_INVALID_HOOK_FILTER: Final[ThirtyTwoBits] = cached_hresult(0x80070592, 'HRESULT_FROM_WIN32(ERROR_INVALID_HOOK_FILTER)')
    WIN32__ERROR_INVALID_FILTER_PROC: Final[ThirtyTwoBits] = cached_hresult(0x80070593, 'HRESULT_FROM_WIN32(ERROR_INVALID_FILTER_PROC)')
    WIN32__ERROR_HOOK_NEEDS_HMOD: Final[ThirtyTwoBits] = cached_hresult(0x80070594, 'HRESULT_FROM_WIN32(ERROR_HOOK_NEEDS_HMOD)')
    WIN32__ERROR_GLOBAL_ONLY_HOOK: Final[ThirtyTwoBits] = cached_hresult(0x80070595, 'HRESULT_FROM_WIN32(ERROR_GLOBAL_ONLY_HOOK)')
    WIN32__ERROR_JOURNAL_HOOK_SET: Final[ThirtyTwoBits] = cached_hresult(0x80070596, 'HRESULT_FROM_WIN32(ERROR_JOURNAL_HOOK_SET)')
    WIN32__ERROR_HOOK_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x80070597, 'HRESULT_FROM_WIN32(ERROR_HOOK_NOT_INSTALLED)')
    WIN32__ERROR_INVALID_LB_MESSAGE: Final[ThirtyTwoBits] = cached_hresult(0x80070598, 'HRESULT_FROM_WIN32(ERROR_INVALID_LB_MESSAGE)')
    WIN32__ERROR_SETCOUNT_ON_BAD_LB: Final[ThirtyTwoBits] = cached_hresult(0x80070599, 'HRESULT_FROM_WIN32(ERROR_SETCOUNT_ON_BAD_LB)')
    WIN32__ERROR_LB_WITHOUT_TABSTOPS: Final[ThirtyTwoBits] = cached_hresult(0x8007059A, 'HRESULT_FROM_WIN32(ERROR_LB_WITHOUT_TABSTOPS)')
    WIN32__ERROR_DESTROY_OBJECT_OF_OTHER_THREAD: Final[ThirtyTwoBits] = cached_hresult(0x8007059B, 'HRESULT_FROM_WIN32(ERROR_DESTROY_OBJECT_OF_OTHER_THREAD)')
    WIN32__ERROR_CHILD_WINDOW_MENU: Final[ThirtyTwoBits] = cached_hresult(0x8007059C, 'HRESULT_FROM_WIN32(ERROR_CHILD_WINDOW_MENU)')
    WIN32__ERROR_NO_SYSTEM_MENU: Final[ThirtyTwoBits] = cached_hresult(0x8007059D, 'HRESULT_FROM_WIN32(ERROR_NO_SYSTEM_MENU)')
    WIN32__ERROR_INVALID_MSGBOX_STYLE: Final[ThirtyTwoBits] = cached_hresult(0x8007059E, 'HRESULT_FROM_WIN32(ERROR_INVALID_MSGBOX_STYLE)')
    WIN32__ERROR_INVALID_SPI_VALUE: Final[ThirtyTwoBits] = cached_hresult(0x8007059F, 'HRESULT_FROM_WIN32(ERROR_INVALID_SPI_VALUE)')
    WIN32__ERROR_SCREEN_ALREADY_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x800705A0, 'HRESULT_FROM_WIN32(ERROR_SCREEN_ALREADY_LOCKED)')
    WIN32__ERROR_HWNDS_HAVE_DIFF_PARENT: Final[ThirtyTwoBits] = cached_hresult(0x800705A1, 'HRESULT_FROM_WIN32(ERROR_HWNDS_HAVE_DIFF_PARENT)')
    WIN32__ERROR_NOT_CHILD_WINDOW: Final[ThirtyTwoBits] = cached_hresult(0x800705A2, 'HRESULT_FROM_WIN32(ERROR_NOT_CHILD_WINDOW)')
    WIN32__ERROR_INVALID_GW_COMMAND: Final[ThirtyTwoBits] = cached_hresult(0x800705A3, 'HRESULT_FROM_WIN32(ERROR_INVALID_GW_COMMAND)')
    WIN32__ERROR_INVALID_THREAD_ID: Final[ThirtyTwoBits] = cached_hresult(0x800705A4, 'HRESULT_FROM_WIN32(ERROR_INVALID_THREAD_ID)')
    WIN32__ERROR_NON_MDICHILD_WINDOW: Final[ThirtyTwoBits] = cached_hresult(0x800705A5, 'HRESULT_FROM_WIN32(ERROR_NON_MDICHILD_WINDOW)')
    WIN32__ERROR_POPUP_ALREADY_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0x800705A6, 'HRESULT_FROM_WIN32(ERROR_POPUP_ALREADY_ACTIVE)')
    WIN32__ERROR_NO_SCROLLBARS: Final[ThirtyTwoBits] = cached_hresult(0x800705A7, 'HRESULT_FROM_WIN32(ERROR_NO_SCROLLBARS)')
    WIN32__ERROR_INVALID_SCROLLBAR_RANGE: Final[ThirtyTwoBits] = cached_hresult(0x800705A8, 'HRESULT_FROM_WIN32(ERROR_INVALID_SCROLLBAR_RANGE)')
    WIN32__ERROR_INVALID_SHOWWIN_COMMAND: Final[ThirtyTwoBits] = cached_hresult(0x800705A9, 'HRESULT_FROM_WIN32(ERROR_INVALID_SHOWWIN_COMMAND)')
    WIN32__ERROR_NO_SYSTEM_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0x800705AA, 'HRESULT_FROM_WIN32(ERROR_NO_SYSTEM_RESOURCES)')
    WIN32__ERROR_NONPAGED_SYSTEM_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0x800705AB, 'HRESULT_FROM_WIN32(ERROR_NONPAGED_SYSTEM_RESOURCES)')
    WIN32__ERROR_PAGED_SYSTEM_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0x800705AC, 'HRESULT_FROM_WIN32(ERROR_PAGED_SYSTEM_RESOURCES)')
    WIN32__ERROR_WORKING_SET_QUOTA: Final[ThirtyTwoBits] = cached_hresult(0x800705AD, 'HRESULT_FROM_WIN32(ERROR_WORKING_SET_QUOTA)')
    WIN32__ERROR_PAGEFILE_QUOTA: Final[ThirtyTwoBits] = cached_hresult(0x800705AE, 'HRESULT_FROM_WIN32(ERROR_PAGEFILE_QUOTA)')
    WIN32__ERROR_COMMITMENT_LIMIT: Final[ThirtyTwoBits] = cached_hresult(0x800705AF, 'HRESULT_FROM_WIN32(ERROR_COMMITMENT_LIMIT)')
    WIN32__ERROR_MENU_ITEM_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800705B0, 'HRESULT_FROM_WIN32(ERROR_MENU_ITEM_NOT_FOUND)')
    WIN32__ERROR_INVALID_KEYBOARD_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x800705B1, 'HRESULT_FROM_WIN32(ERROR_INVALID_KEYBOARD_HANDLE)')
    WIN32__ERROR_HOOK_TYPE_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x800705B2, 'HRESULT_FROM_WIN32(ERROR_HOOK_TYPE_NOT_ALLOWED)')
    WIN32__ERROR_REQUIRES_INTERACTIVE_WINDOWSTATION: Final[ThirtyTwoBits] = cached_hresult(0x800705B3, 'HRESULT_FROM_WIN32(ERROR_REQUIRES_INTERACTIVE_WINDOWSTATION)')
    WIN32__ERROR_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x800705B4, 'HRESULT_FROM_WIN32(ERROR_TIMEOUT)')
    WIN32__ERROR_INVALID_MONITOR_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x800705B5, 'HRESULT_FROM_WIN32(ERROR_INVALID_MONITOR_HANDLE)')
    WIN32__ERROR_INCORRECT_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x800705B6, 'HRESULT_FROM_WIN32(ERROR_INCORRECT_SIZE)')
    WIN32__ERROR_SYMLINK_CLASS_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x800705B7, 'HRESULT_FROM_WIN32(ERROR_SYMLINK_CLASS_DISABLED)')
    WIN32__ERROR_SYMLINK_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x800705B8, 'HRESULT_FROM_WIN32(ERROR_SYMLINK_NOT_SUPPORTED)')
    WIN32__ERROR_XML_PARSE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800705B9, 'HRESULT_FROM_WIN32(ERROR_XML_PARSE_ERROR)')
    WIN32__ERROR_XMLDSIG_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800705BA, 'HRESULT_FROM_WIN32(ERROR_XMLDSIG_ERROR)')
    WIN32__ERROR_RESTART_APPLICATION: Final[ThirtyTwoBits] = cached_hresult(0x800705BB, 'HRESULT_FROM_WIN32(ERROR_RESTART_APPLICATION)')
    WIN32__ERROR_WRONG_COMPARTMENT: Final[ThirtyTwoBits] = cached_hresult(0x800705BC, 'HRESULT_FROM_WIN32(ERROR_WRONG_COMPARTMENT)')
    WIN32__ERROR_AUTHIP_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x800705BD, 'HRESULT_FROM_WIN32(ERROR_AUTHIP_FAILURE)')
    WIN32__ERROR_NO_NVRAM_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0x800705BE, 'HRESULT_FROM_WIN32(ERROR_NO_NVRAM_RESOURCES)')
    WIN32__ERROR_NOT_GUI_PROCESS: Final[ThirtyTwoBits] = cached_hresult(0x800705BF, 'HRESULT_FROM_WIN32(ERROR_NOT_GUI_PROCESS)')
    WIN32__ERROR_EVENTLOG_FILE_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x800705DC, 'HRESULT_FROM_WIN32(ERROR_EVENTLOG_FILE_CORRUPT)')
    WIN32__ERROR_EVENTLOG_CANT_START: Final[ThirtyTwoBits] = cached_hresult(0x800705DD, 'HRESULT_FROM_WIN32(ERROR_EVENTLOG_CANT_START)')
    WIN32__ERROR_LOG_FILE_FULL: Final[ThirtyTwoBits] = cached_hresult(0x800705DE, 'HRESULT_FROM_WIN32(ERROR_LOG_FILE_FULL)')
    WIN32__ERROR_EVENTLOG_FILE_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0x800705DF, 'HRESULT_FROM_WIN32(ERROR_EVENTLOG_FILE_CHANGED)')
    WIN32__ERROR_CONTAINER_ASSIGNED: Final[ThirtyTwoBits] = cached_hresult(0x800705E0, 'HRESULT_FROM_WIN32(ERROR_CONTAINER_ASSIGNED)')
    WIN32__ERROR_JOB_NO_CONTAINER: Final[ThirtyTwoBits] = cached_hresult(0x800705E1, 'HRESULT_FROM_WIN32(ERROR_JOB_NO_CONTAINER)')
    WIN32__ERROR_INVALID_TASK_NAME: Final[ThirtyTwoBits] = cached_hresult(0x8007060E, 'HRESULT_FROM_WIN32(ERROR_INVALID_TASK_NAME)')
    WIN32__ERROR_INVALID_TASK_INDEX: Final[ThirtyTwoBits] = cached_hresult(0x8007060F, 'HRESULT_FROM_WIN32(ERROR_INVALID_TASK_INDEX)')
    WIN32__ERROR_THREAD_ALREADY_IN_TASK: Final[ThirtyTwoBits] = cached_hresult(0x80070610, 'HRESULT_FROM_WIN32(ERROR_THREAD_ALREADY_IN_TASK)')
    WIN32__ERROR_INSTALL_SERVICE_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80070641, 'HRESULT_FROM_WIN32(ERROR_INSTALL_SERVICE_FAILURE)')
    WIN32__ERROR_INSTALL_USEREXIT: Final[ThirtyTwoBits] = cached_hresult(0x80070642, 'HRESULT_FROM_WIN32(ERROR_INSTALL_USEREXIT)')
    WIN32__ERROR_INSTALL_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80070643, 'HRESULT_FROM_WIN32(ERROR_INSTALL_FAILURE)')
    WIN32__ERROR_INSTALL_SUSPEND: Final[ThirtyTwoBits] = cached_hresult(0x80070644, 'HRESULT_FROM_WIN32(ERROR_INSTALL_SUSPEND)')
    WIN32__ERROR_UNKNOWN_PRODUCT: Final[ThirtyTwoBits] = cached_hresult(0x80070645, 'HRESULT_FROM_WIN32(ERROR_UNKNOWN_PRODUCT)')
    WIN32__ERROR_UNKNOWN_FEATURE: Final[ThirtyTwoBits] = cached_hresult(0x80070646, 'HRESULT_FROM_WIN32(ERROR_UNKNOWN_FEATURE)')
    WIN32__ERROR_UNKNOWN_COMPONENT: Final[ThirtyTwoBits] = cached_hresult(0x80070647, 'HRESULT_FROM_WIN32(ERROR_UNKNOWN_COMPONENT)')
    WIN32__ERROR_UNKNOWN_PROPERTY: Final[ThirtyTwoBits] = cached_hresult(0x80070648, 'HRESULT_FROM_WIN32(ERROR_UNKNOWN_PROPERTY)')
    WIN32__ERROR_INVALID_HANDLE_STATE: Final[ThirtyTwoBits] = cached_hresult(0x80070649, 'HRESULT_FROM_WIN32(ERROR_INVALID_HANDLE_STATE)')
    WIN32__ERROR_BAD_CONFIGURATION: Final[ThirtyTwoBits] = cached_hresult(0x8007064A, 'HRESULT_FROM_WIN32(ERROR_BAD_CONFIGURATION)')
    WIN32__ERROR_INDEX_ABSENT: Final[ThirtyTwoBits] = cached_hresult(0x8007064B, 'HRESULT_FROM_WIN32(ERROR_INDEX_ABSENT)')
    WIN32__ERROR_INSTALL_SOURCE_ABSENT: Final[ThirtyTwoBits] = cached_hresult(0x8007064C, 'HRESULT_FROM_WIN32(ERROR_INSTALL_SOURCE_ABSENT)')
    WIN32__ERROR_INSTALL_PACKAGE_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x8007064D, 'HRESULT_FROM_WIN32(ERROR_INSTALL_PACKAGE_VERSION)')
    WIN32__ERROR_PRODUCT_UNINSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x8007064E, 'HRESULT_FROM_WIN32(ERROR_PRODUCT_UNINSTALLED)')
    WIN32__ERROR_BAD_QUERY_SYNTAX: Final[ThirtyTwoBits] = cached_hresult(0x8007064F, 'HRESULT_FROM_WIN32(ERROR_BAD_QUERY_SYNTAX)')
    WIN32__ERROR_INVALID_FIELD: Final[ThirtyTwoBits] = cached_hresult(0x80070650, 'HRESULT_FROM_WIN32(ERROR_INVALID_FIELD)')
    WIN32__ERROR_DEVICE_REMOVED: Final[ThirtyTwoBits] = cached_hresult(0x80070651, 'HRESULT_FROM_WIN32(ERROR_DEVICE_REMOVED)')
    WIN32__ERROR_INSTALL_ALREADY_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x80070652, 'HRESULT_FROM_WIN32(ERROR_INSTALL_ALREADY_RUNNING)')
    WIN32__ERROR_INSTALL_PACKAGE_OPEN_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80070653, 'HRESULT_FROM_WIN32(ERROR_INSTALL_PACKAGE_OPEN_FAILED)')
    WIN32__ERROR_INSTALL_PACKAGE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80070654, 'HRESULT_FROM_WIN32(ERROR_INSTALL_PACKAGE_INVALID)')
    WIN32__ERROR_INSTALL_UI_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80070655, 'HRESULT_FROM_WIN32(ERROR_INSTALL_UI_FAILURE)')
    WIN32__ERROR_INSTALL_LOG_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80070656, 'HRESULT_FROM_WIN32(ERROR_INSTALL_LOG_FAILURE)')
    WIN32__ERROR_INSTALL_LANGUAGE_UNSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80070657, 'HRESULT_FROM_WIN32(ERROR_INSTALL_LANGUAGE_UNSUPPORTED)')
    WIN32__ERROR_INSTALL_TRANSFORM_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80070658, 'HRESULT_FROM_WIN32(ERROR_INSTALL_TRANSFORM_FAILURE)')
    WIN32__ERROR_INSTALL_PACKAGE_REJECTED: Final[ThirtyTwoBits] = cached_hresult(0x80070659, 'HRESULT_FROM_WIN32(ERROR_INSTALL_PACKAGE_REJECTED)')
    WIN32__ERROR_FUNCTION_NOT_CALLED: Final[ThirtyTwoBits] = cached_hresult(0x8007065A, 'HRESULT_FROM_WIN32(ERROR_FUNCTION_NOT_CALLED)')
    WIN32__ERROR_FUNCTION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8007065B, 'HRESULT_FROM_WIN32(ERROR_FUNCTION_FAILED)')
    WIN32__ERROR_INVALID_TABLE: Final[ThirtyTwoBits] = cached_hresult(0x8007065C, 'HRESULT_FROM_WIN32(ERROR_INVALID_TABLE)')
    WIN32__ERROR_DATATYPE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x8007065D, 'HRESULT_FROM_WIN32(ERROR_DATATYPE_MISMATCH)')
    WIN32__ERROR_UNSUPPORTED_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x8007065E, 'HRESULT_FROM_WIN32(ERROR_UNSUPPORTED_TYPE)')
    WIN32__ERROR_CREATE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8007065F, 'HRESULT_FROM_WIN32(ERROR_CREATE_FAILED)')
    WIN32__ERROR_INSTALL_TEMP_UNWRITABLE: Final[ThirtyTwoBits] = cached_hresult(0x80070660, 'HRESULT_FROM_WIN32(ERROR_INSTALL_TEMP_UNWRITABLE)')
    WIN32__ERROR_INSTALL_PLATFORM_UNSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80070661, 'HRESULT_FROM_WIN32(ERROR_INSTALL_PLATFORM_UNSUPPORTED)')
    WIN32__ERROR_INSTALL_NOTUSED: Final[ThirtyTwoBits] = cached_hresult(0x80070662, 'HRESULT_FROM_WIN32(ERROR_INSTALL_NOTUSED)')
    WIN32__ERROR_PATCH_PACKAGE_OPEN_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80070663, 'HRESULT_FROM_WIN32(ERROR_PATCH_PACKAGE_OPEN_FAILED)')
    WIN32__ERROR_PATCH_PACKAGE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80070664, 'HRESULT_FROM_WIN32(ERROR_PATCH_PACKAGE_INVALID)')
    WIN32__ERROR_PATCH_PACKAGE_UNSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80070665, 'HRESULT_FROM_WIN32(ERROR_PATCH_PACKAGE_UNSUPPORTED)')
    WIN32__ERROR_PRODUCT_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80070666, 'HRESULT_FROM_WIN32(ERROR_PRODUCT_VERSION)')
    WIN32__ERROR_INVALID_COMMAND_LINE: Final[ThirtyTwoBits] = cached_hresult(0x80070667, 'HRESULT_FROM_WIN32(ERROR_INVALID_COMMAND_LINE)')
    WIN32__ERROR_INSTALL_REMOTE_DISALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80070668, 'HRESULT_FROM_WIN32(ERROR_INSTALL_REMOTE_DISALLOWED)')
    WIN32__ERROR_SUCCESS_REBOOT_INITIATED: Final[ThirtyTwoBits] = cached_hresult(0x80070669, 'HRESULT_FROM_WIN32(ERROR_SUCCESS_REBOOT_INITIATED)')
    WIN32__ERROR_PATCH_TARGET_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8007066A, 'HRESULT_FROM_WIN32(ERROR_PATCH_TARGET_NOT_FOUND)')
    WIN32__ERROR_PATCH_PACKAGE_REJECTED: Final[ThirtyTwoBits] = cached_hresult(0x8007066B, 'HRESULT_FROM_WIN32(ERROR_PATCH_PACKAGE_REJECTED)')
    WIN32__ERROR_INSTALL_TRANSFORM_REJECTED: Final[ThirtyTwoBits] = cached_hresult(0x8007066C, 'HRESULT_FROM_WIN32(ERROR_INSTALL_TRANSFORM_REJECTED)')
    WIN32__ERROR_INSTALL_REMOTE_PROHIBITED: Final[ThirtyTwoBits] = cached_hresult(0x8007066D, 'HRESULT_FROM_WIN32(ERROR_INSTALL_REMOTE_PROHIBITED)')
    WIN32__ERROR_PATCH_REMOVAL_UNSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8007066E, 'HRESULT_FROM_WIN32(ERROR_PATCH_REMOVAL_UNSUPPORTED)')
    WIN32__ERROR_UNKNOWN_PATCH: Final[ThirtyTwoBits] = cached_hresult(0x8007066F, 'HRESULT_FROM_WIN32(ERROR_UNKNOWN_PATCH)')
    WIN32__ERROR_PATCH_NO_SEQUENCE: Final[ThirtyTwoBits] = cached_hresult(0x80070670, 'HRESULT_FROM_WIN32(ERROR_PATCH_NO_SEQUENCE)')
    WIN32__ERROR_PATCH_REMOVAL_DISALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80070671, 'HRESULT_FROM_WIN32(ERROR_PATCH_REMOVAL_DISALLOWED)')
    WIN32__ERROR_INVALID_PATCH_XML: Final[ThirtyTwoBits] = cached_hresult(0x80070672, 'HRESULT_FROM_WIN32(ERROR_INVALID_PATCH_XML)')
    WIN32__ERROR_PATCH_MANAGED_ADVERTISED_PRODUCT: Final[ThirtyTwoBits] = cached_hresult(0x80070673, 'HRESULT_FROM_WIN32(ERROR_PATCH_MANAGED_ADVERTISED_PRODUCT)')
    WIN32__ERROR_INSTALL_SERVICE_SAFEBOOT: Final[ThirtyTwoBits] = cached_hresult(0x80070674, 'HRESULT_FROM_WIN32(ERROR_INSTALL_SERVICE_SAFEBOOT)')
    WIN32__ERROR_FAIL_FAST_EXCEPTION: Final[ThirtyTwoBits] = cached_hresult(0x80070675, 'HRESULT_FROM_WIN32(ERROR_FAIL_FAST_EXCEPTION)')
    WIN32__ERROR_INSTALL_REJECTED: Final[ThirtyTwoBits] = cached_hresult(0x80070676, 'HRESULT_FROM_WIN32(ERROR_INSTALL_REJECTED)')
    WIN32__ERROR_DYNAMIC_CODE_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x80070677, 'HRESULT_FROM_WIN32(ERROR_DYNAMIC_CODE_BLOCKED)')
    WIN32__ERROR_NOT_SAME_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x80070678, 'HRESULT_FROM_WIN32(ERROR_NOT_SAME_OBJECT)')
    WIN32__ERROR_STRICT_CFG_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x80070679, 'HRESULT_FROM_WIN32(ERROR_STRICT_CFG_VIOLATION)')
    WIN32__ERROR_SET_CONTEXT_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x8007067C, 'HRESULT_FROM_WIN32(ERROR_SET_CONTEXT_DENIED)')
    WIN32__ERROR_CROSS_PARTITION_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x8007067D, 'HRESULT_FROM_WIN32(ERROR_CROSS_PARTITION_VIOLATION)')
    WIN32__ERROR_RETURN_ADDRESS_HIJACK_ATTEMPT: Final[ThirtyTwoBits] = cached_hresult(0x8007067E, 'HRESULT_FROM_WIN32(ERROR_RETURN_ADDRESS_HIJACK_ATTEMPT)')
    WIN32__RPC_S_INVALID_STRING_BINDING: Final[ThirtyTwoBits] = cached_hresult(0x800706A4, 'HRESULT_FROM_WIN32(RPC_S_INVALID_STRING_BINDING)')
    WIN32__RPC_S_WRONG_KIND_OF_BINDING: Final[ThirtyTwoBits] = cached_hresult(0x800706A5, 'HRESULT_FROM_WIN32(RPC_S_WRONG_KIND_OF_BINDING)')
    WIN32__RPC_S_INVALID_BINDING: Final[ThirtyTwoBits] = cached_hresult(0x800706A6, 'HRESULT_FROM_WIN32(RPC_S_INVALID_BINDING)')
    WIN32__RPC_S_PROTSEQ_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x800706A7, 'HRESULT_FROM_WIN32(RPC_S_PROTSEQ_NOT_SUPPORTED)')
    WIN32__RPC_S_INVALID_RPC_PROTSEQ: Final[ThirtyTwoBits] = cached_hresult(0x800706A8, 'HRESULT_FROM_WIN32(RPC_S_INVALID_RPC_PROTSEQ)')
    WIN32__RPC_S_INVALID_STRING_UUID: Final[ThirtyTwoBits] = cached_hresult(0x800706A9, 'HRESULT_FROM_WIN32(RPC_S_INVALID_STRING_UUID)')
    WIN32__RPC_S_INVALID_ENDPOINT_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0x800706AA, 'HRESULT_FROM_WIN32(RPC_S_INVALID_ENDPOINT_FORMAT)')
    WIN32__RPC_S_INVALID_NET_ADDR: Final[ThirtyTwoBits] = cached_hresult(0x800706AB, 'HRESULT_FROM_WIN32(RPC_S_INVALID_NET_ADDR)')
    WIN32__RPC_S_NO_ENDPOINT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800706AC, 'HRESULT_FROM_WIN32(RPC_S_NO_ENDPOINT_FOUND)')
    WIN32__RPC_S_INVALID_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x800706AD, 'HRESULT_FROM_WIN32(RPC_S_INVALID_TIMEOUT)')
    WIN32__RPC_S_OBJECT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800706AE, 'HRESULT_FROM_WIN32(RPC_S_OBJECT_NOT_FOUND)')
    WIN32__RPC_S_ALREADY_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x800706AF, 'HRESULT_FROM_WIN32(RPC_S_ALREADY_REGISTERED)')
    WIN32__RPC_S_TYPE_ALREADY_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x800706B0, 'HRESULT_FROM_WIN32(RPC_S_TYPE_ALREADY_REGISTERED)')
    WIN32__RPC_S_ALREADY_LISTENING: Final[ThirtyTwoBits] = cached_hresult(0x800706B1, 'HRESULT_FROM_WIN32(RPC_S_ALREADY_LISTENING)')
    WIN32__RPC_S_NO_PROTSEQS_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x800706B2, 'HRESULT_FROM_WIN32(RPC_S_NO_PROTSEQS_REGISTERED)')
    WIN32__RPC_S_NOT_LISTENING: Final[ThirtyTwoBits] = cached_hresult(0x800706B3, 'HRESULT_FROM_WIN32(RPC_S_NOT_LISTENING)')
    WIN32__RPC_S_UNKNOWN_MGR_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x800706B4, 'HRESULT_FROM_WIN32(RPC_S_UNKNOWN_MGR_TYPE)')
    WIN32__RPC_S_UNKNOWN_IF: Final[ThirtyTwoBits] = cached_hresult(0x800706B5, 'HRESULT_FROM_WIN32(RPC_S_UNKNOWN_IF)')
    WIN32__RPC_S_NO_BINDINGS: Final[ThirtyTwoBits] = cached_hresult(0x800706B6, 'HRESULT_FROM_WIN32(RPC_S_NO_BINDINGS)')
    WIN32__RPC_S_NO_PROTSEQS: Final[ThirtyTwoBits] = cached_hresult(0x800706B7, 'HRESULT_FROM_WIN32(RPC_S_NO_PROTSEQS)')
    WIN32__RPC_S_CANT_CREATE_ENDPOINT: Final[ThirtyTwoBits] = cached_hresult(0x800706B8, 'HRESULT_FROM_WIN32(RPC_S_CANT_CREATE_ENDPOINT)')
    WIN32__RPC_S_OUT_OF_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0x800706B9, 'HRESULT_FROM_WIN32(RPC_S_OUT_OF_RESOURCES)')
    WIN32__RPC_S_SERVER_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x800706BA, 'HRESULT_FROM_WIN32(RPC_S_SERVER_UNAVAILABLE)')
    WIN32__RPC_S_SERVER_TOO_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x800706BB, 'HRESULT_FROM_WIN32(RPC_S_SERVER_TOO_BUSY)')
    WIN32__RPC_S_INVALID_NETWORK_OPTIONS: Final[ThirtyTwoBits] = cached_hresult(0x800706BC, 'HRESULT_FROM_WIN32(RPC_S_INVALID_NETWORK_OPTIONS)')
    WIN32__RPC_S_NO_CALL_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0x800706BD, 'HRESULT_FROM_WIN32(RPC_S_NO_CALL_ACTIVE)')
    WIN32__RPC_S_CALL_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800706BE, 'HRESULT_FROM_WIN32(RPC_S_CALL_FAILED)')
    WIN32__RPC_S_CALL_FAILED_DNE: Final[ThirtyTwoBits] = cached_hresult(0x800706BF, 'HRESULT_FROM_WIN32(RPC_S_CALL_FAILED_DNE)')
    WIN32__RPC_S_PROTOCOL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800706C0, 'HRESULT_FROM_WIN32(RPC_S_PROTOCOL_ERROR)')
    WIN32__RPC_S_PROXY_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x800706C1, 'HRESULT_FROM_WIN32(RPC_S_PROXY_ACCESS_DENIED)')
    WIN32__RPC_S_UNSUPPORTED_TRANS_SYN: Final[ThirtyTwoBits] = cached_hresult(0x800706C2, 'HRESULT_FROM_WIN32(RPC_S_UNSUPPORTED_TRANS_SYN)')
    WIN32__RPC_S_UNSUPPORTED_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x800706C4, 'HRESULT_FROM_WIN32(RPC_S_UNSUPPORTED_TYPE)')
    WIN32__RPC_S_INVALID_TAG: Final[ThirtyTwoBits] = cached_hresult(0x800706C5, 'HRESULT_FROM_WIN32(RPC_S_INVALID_TAG)')
    WIN32__RPC_S_INVALID_BOUND: Final[ThirtyTwoBits] = cached_hresult(0x800706C6, 'HRESULT_FROM_WIN32(RPC_S_INVALID_BOUND)')
    WIN32__RPC_S_NO_ENTRY_NAME: Final[ThirtyTwoBits] = cached_hresult(0x800706C7, 'HRESULT_FROM_WIN32(RPC_S_NO_ENTRY_NAME)')
    WIN32__RPC_S_INVALID_NAME_SYNTAX: Final[ThirtyTwoBits] = cached_hresult(0x800706C8, 'HRESULT_FROM_WIN32(RPC_S_INVALID_NAME_SYNTAX)')
    WIN32__RPC_S_UNSUPPORTED_NAME_SYNTAX: Final[ThirtyTwoBits] = cached_hresult(0x800706C9, 'HRESULT_FROM_WIN32(RPC_S_UNSUPPORTED_NAME_SYNTAX)')
    WIN32__RPC_S_UUID_NO_ADDRESS: Final[ThirtyTwoBits] = cached_hresult(0x800706CB, 'HRESULT_FROM_WIN32(RPC_S_UUID_NO_ADDRESS)')
    WIN32__RPC_S_DUPLICATE_ENDPOINT: Final[ThirtyTwoBits] = cached_hresult(0x800706CC, 'HRESULT_FROM_WIN32(RPC_S_DUPLICATE_ENDPOINT)')
    WIN32__RPC_S_UNKNOWN_AUTHN_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x800706CD, 'HRESULT_FROM_WIN32(RPC_S_UNKNOWN_AUTHN_TYPE)')
    WIN32__RPC_S_MAX_CALLS_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0x800706CE, 'HRESULT_FROM_WIN32(RPC_S_MAX_CALLS_TOO_SMALL)')
    WIN32__RPC_S_STRING_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0x800706CF, 'HRESULT_FROM_WIN32(RPC_S_STRING_TOO_LONG)')
    WIN32__RPC_S_PROTSEQ_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800706D0, 'HRESULT_FROM_WIN32(RPC_S_PROTSEQ_NOT_FOUND)')
    WIN32__RPC_S_PROCNUM_OUT_OF_RANGE: Final[ThirtyTwoBits] = cached_hresult(0x800706D1, 'HRESULT_FROM_WIN32(RPC_S_PROCNUM_OUT_OF_RANGE)')
    WIN32__RPC_S_BINDING_HAS_NO_AUTH: Final[ThirtyTwoBits] = cached_hresult(0x800706D2, 'HRESULT_FROM_WIN32(RPC_S_BINDING_HAS_NO_AUTH)')
    WIN32__RPC_S_UNKNOWN_AUTHN_SERVICE: Final[ThirtyTwoBits] = cached_hresult(0x800706D3, 'HRESULT_FROM_WIN32(RPC_S_UNKNOWN_AUTHN_SERVICE)')
    WIN32__RPC_S_UNKNOWN_AUTHN_LEVEL: Final[ThirtyTwoBits] = cached_hresult(0x800706D4, 'HRESULT_FROM_WIN32(RPC_S_UNKNOWN_AUTHN_LEVEL)')
    WIN32__RPC_S_INVALID_AUTH_IDENTITY: Final[ThirtyTwoBits] = cached_hresult(0x800706D5, 'HRESULT_FROM_WIN32(RPC_S_INVALID_AUTH_IDENTITY)')
    WIN32__RPC_S_UNKNOWN_AUTHZ_SERVICE: Final[ThirtyTwoBits] = cached_hresult(0x800706D6, 'HRESULT_FROM_WIN32(RPC_S_UNKNOWN_AUTHZ_SERVICE)')
    WIN32__EPT_S_INVALID_ENTRY: Final[ThirtyTwoBits] = cached_hresult(0x800706D7, 'HRESULT_FROM_WIN32(EPT_S_INVALID_ENTRY)')
    WIN32__EPT_S_CANT_PERFORM_OP: Final[ThirtyTwoBits] = cached_hresult(0x800706D8, 'HRESULT_FROM_WIN32(EPT_S_CANT_PERFORM_OP)')
    WIN32__EPT_S_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x800706D9, 'HRESULT_FROM_WIN32(EPT_S_NOT_REGISTERED)')
    WIN32__RPC_S_NOTHING_TO_EXPORT: Final[ThirtyTwoBits] = cached_hresult(0x800706DA, 'HRESULT_FROM_WIN32(RPC_S_NOTHING_TO_EXPORT)')
    WIN32__RPC_S_INCOMPLETE_NAME: Final[ThirtyTwoBits] = cached_hresult(0x800706DB, 'HRESULT_FROM_WIN32(RPC_S_INCOMPLETE_NAME)')
    WIN32__RPC_S_INVALID_VERS_OPTION: Final[ThirtyTwoBits] = cached_hresult(0x800706DC, 'HRESULT_FROM_WIN32(RPC_S_INVALID_VERS_OPTION)')
    WIN32__RPC_S_NO_MORE_MEMBERS: Final[ThirtyTwoBits] = cached_hresult(0x800706DD, 'HRESULT_FROM_WIN32(RPC_S_NO_MORE_MEMBERS)')
    WIN32__RPC_S_NOT_ALL_OBJS_UNEXPORTED: Final[ThirtyTwoBits] = cached_hresult(0x800706DE, 'HRESULT_FROM_WIN32(RPC_S_NOT_ALL_OBJS_UNEXPORTED)')
    WIN32__RPC_S_INTERFACE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800706DF, 'HRESULT_FROM_WIN32(RPC_S_INTERFACE_NOT_FOUND)')
    WIN32__RPC_S_ENTRY_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800706E0, 'HRESULT_FROM_WIN32(RPC_S_ENTRY_ALREADY_EXISTS)')
    WIN32__RPC_S_ENTRY_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800706E1, 'HRESULT_FROM_WIN32(RPC_S_ENTRY_NOT_FOUND)')
    WIN32__RPC_S_NAME_SERVICE_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x800706E2, 'HRESULT_FROM_WIN32(RPC_S_NAME_SERVICE_UNAVAILABLE)')
    WIN32__RPC_S_INVALID_NAF_ID: Final[ThirtyTwoBits] = cached_hresult(0x800706E3, 'HRESULT_FROM_WIN32(RPC_S_INVALID_NAF_ID)')
    WIN32__RPC_S_CANNOT_SUPPORT: Final[ThirtyTwoBits] = cached_hresult(0x800706E4, 'HRESULT_FROM_WIN32(RPC_S_CANNOT_SUPPORT)')
    WIN32__RPC_S_NO_CONTEXT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x800706E5, 'HRESULT_FROM_WIN32(RPC_S_NO_CONTEXT_AVAILABLE)')
    WIN32__RPC_S_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800706E6, 'HRESULT_FROM_WIN32(RPC_S_INTERNAL_ERROR)')
    WIN32__RPC_S_ZERO_DIVIDE: Final[ThirtyTwoBits] = cached_hresult(0x800706E7, 'HRESULT_FROM_WIN32(RPC_S_ZERO_DIVIDE)')
    WIN32__RPC_S_ADDRESS_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800706E8, 'HRESULT_FROM_WIN32(RPC_S_ADDRESS_ERROR)')
    WIN32__RPC_S_FP_DIV_ZERO: Final[ThirtyTwoBits] = cached_hresult(0x800706E9, 'HRESULT_FROM_WIN32(RPC_S_FP_DIV_ZERO)')
    WIN32__RPC_S_FP_UNDERFLOW: Final[ThirtyTwoBits] = cached_hresult(0x800706EA, 'HRESULT_FROM_WIN32(RPC_S_FP_UNDERFLOW)')
    WIN32__RPC_S_FP_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0x800706EB, 'HRESULT_FROM_WIN32(RPC_S_FP_OVERFLOW)')
    WIN32__RPC_X_NO_MORE_ENTRIES: Final[ThirtyTwoBits] = cached_hresult(0x800706EC, 'HRESULT_FROM_WIN32(RPC_X_NO_MORE_ENTRIES)')
    WIN32__RPC_X_SS_CHAR_TRANS_OPEN_FAIL: Final[ThirtyTwoBits] = cached_hresult(0x800706ED, 'HRESULT_FROM_WIN32(RPC_X_SS_CHAR_TRANS_OPEN_FAIL)')
    WIN32__RPC_X_SS_CHAR_TRANS_SHORT_FILE: Final[ThirtyTwoBits] = cached_hresult(0x800706EE, 'HRESULT_FROM_WIN32(RPC_X_SS_CHAR_TRANS_SHORT_FILE)')
    WIN32__RPC_X_SS_IN_NULL_CONTEXT: Final[ThirtyTwoBits] = cached_hresult(0x800706EF, 'HRESULT_FROM_WIN32(RPC_X_SS_IN_NULL_CONTEXT)')
    WIN32__RPC_X_SS_CONTEXT_DAMAGED: Final[ThirtyTwoBits] = cached_hresult(0x800706F1, 'HRESULT_FROM_WIN32(RPC_X_SS_CONTEXT_DAMAGED)')
    WIN32__RPC_X_SS_HANDLES_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x800706F2, 'HRESULT_FROM_WIN32(RPC_X_SS_HANDLES_MISMATCH)')
    WIN32__RPC_X_SS_CANNOT_GET_CALL_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x800706F3, 'HRESULT_FROM_WIN32(RPC_X_SS_CANNOT_GET_CALL_HANDLE)')
    WIN32__RPC_X_NULL_REF_POINTER: Final[ThirtyTwoBits] = cached_hresult(0x800706F4, 'HRESULT_FROM_WIN32(RPC_X_NULL_REF_POINTER)')
    WIN32__RPC_X_ENUM_VALUE_OUT_OF_RANGE: Final[ThirtyTwoBits] = cached_hresult(0x800706F5, 'HRESULT_FROM_WIN32(RPC_X_ENUM_VALUE_OUT_OF_RANGE)')
    WIN32__RPC_X_BYTE_COUNT_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0x800706F6, 'HRESULT_FROM_WIN32(RPC_X_BYTE_COUNT_TOO_SMALL)')
    WIN32__RPC_X_BAD_STUB_DATA: Final[ThirtyTwoBits] = cached_hresult(0x800706F7, 'HRESULT_FROM_WIN32(RPC_X_BAD_STUB_DATA)')
    WIN32__ERROR_INVALID_USER_BUFFER: Final[ThirtyTwoBits] = cached_hresult(0x800706F8, 'HRESULT_FROM_WIN32(ERROR_INVALID_USER_BUFFER)')
    WIN32__ERROR_UNRECOGNIZED_MEDIA: Final[ThirtyTwoBits] = cached_hresult(0x800706F9, 'HRESULT_FROM_WIN32(ERROR_UNRECOGNIZED_MEDIA)')
    WIN32__ERROR_NO_TRUST_LSA_SECRET: Final[ThirtyTwoBits] = cached_hresult(0x800706FA, 'HRESULT_FROM_WIN32(ERROR_NO_TRUST_LSA_SECRET)')
    WIN32__ERROR_NO_TRUST_SAM_ACCOUNT: Final[ThirtyTwoBits] = cached_hresult(0x800706FB, 'HRESULT_FROM_WIN32(ERROR_NO_TRUST_SAM_ACCOUNT)')
    WIN32__ERROR_TRUSTED_DOMAIN_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x800706FC, 'HRESULT_FROM_WIN32(ERROR_TRUSTED_DOMAIN_FAILURE)')
    WIN32__ERROR_TRUSTED_RELATIONSHIP_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x800706FD, 'HRESULT_FROM_WIN32(ERROR_TRUSTED_RELATIONSHIP_FAILURE)')
    WIN32__ERROR_TRUST_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x800706FE, 'HRESULT_FROM_WIN32(ERROR_TRUST_FAILURE)')
    WIN32__RPC_S_CALL_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x800706FF, 'HRESULT_FROM_WIN32(RPC_S_CALL_IN_PROGRESS)')
    WIN32__ERROR_NETLOGON_NOT_STARTED: Final[ThirtyTwoBits] = cached_hresult(0x80070700, 'HRESULT_FROM_WIN32(ERROR_NETLOGON_NOT_STARTED)')
    WIN32__ERROR_ACCOUNT_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x80070701, 'HRESULT_FROM_WIN32(ERROR_ACCOUNT_EXPIRED)')
    WIN32__ERROR_REDIRECTOR_HAS_OPEN_HANDLES: Final[ThirtyTwoBits] = cached_hresult(0x80070702, 'HRESULT_FROM_WIN32(ERROR_REDIRECTOR_HAS_OPEN_HANDLES)')
    WIN32__ERROR_PRINTER_DRIVER_ALREADY_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x80070703, 'HRESULT_FROM_WIN32(ERROR_PRINTER_DRIVER_ALREADY_INSTALLED)')
    WIN32__ERROR_UNKNOWN_PORT: Final[ThirtyTwoBits] = cached_hresult(0x80070704, 'HRESULT_FROM_WIN32(ERROR_UNKNOWN_PORT)')
    WIN32__ERROR_UNKNOWN_PRINTER_DRIVER: Final[ThirtyTwoBits] = cached_hresult(0x80070705, 'HRESULT_FROM_WIN32(ERROR_UNKNOWN_PRINTER_DRIVER)')
    WIN32__ERROR_UNKNOWN_PRINTPROCESSOR: Final[ThirtyTwoBits] = cached_hresult(0x80070706, 'HRESULT_FROM_WIN32(ERROR_UNKNOWN_PRINTPROCESSOR)')
    WIN32__ERROR_INVALID_SEPARATOR_FILE: Final[ThirtyTwoBits] = cached_hresult(0x80070707, 'HRESULT_FROM_WIN32(ERROR_INVALID_SEPARATOR_FILE)')
    WIN32__ERROR_INVALID_PRIORITY: Final[ThirtyTwoBits] = cached_hresult(0x80070708, 'HRESULT_FROM_WIN32(ERROR_INVALID_PRIORITY)')
    WIN32__ERROR_INVALID_PRINTER_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80070709, 'HRESULT_FROM_WIN32(ERROR_INVALID_PRINTER_NAME)')
    WIN32__ERROR_PRINTER_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x8007070A, 'HRESULT_FROM_WIN32(ERROR_PRINTER_ALREADY_EXISTS)')
    WIN32__ERROR_INVALID_PRINTER_COMMAND: Final[ThirtyTwoBits] = cached_hresult(0x8007070B, 'HRESULT_FROM_WIN32(ERROR_INVALID_PRINTER_COMMAND)')
    WIN32__ERROR_INVALID_DATATYPE: Final[ThirtyTwoBits] = cached_hresult(0x8007070C, 'HRESULT_FROM_WIN32(ERROR_INVALID_DATATYPE)')
    WIN32__ERROR_INVALID_ENVIRONMENT: Final[ThirtyTwoBits] = cached_hresult(0x8007070D, 'HRESULT_FROM_WIN32(ERROR_INVALID_ENVIRONMENT)')
    WIN32__RPC_S_NO_MORE_BINDINGS: Final[ThirtyTwoBits] = cached_hresult(0x8007070E, 'HRESULT_FROM_WIN32(RPC_S_NO_MORE_BINDINGS)')
    WIN32__ERROR_NOLOGON_INTERDOMAIN_TRUST_ACCOUNT: Final[ThirtyTwoBits] = cached_hresult(0x8007070F, 'HRESULT_FROM_WIN32(ERROR_NOLOGON_INTERDOMAIN_TRUST_ACCOUNT)')
    WIN32__ERROR_NOLOGON_WORKSTATION_TRUST_ACCOUNT: Final[ThirtyTwoBits] = cached_hresult(0x80070710, 'HRESULT_FROM_WIN32(ERROR_NOLOGON_WORKSTATION_TRUST_ACCOUNT)')
    WIN32__ERROR_NOLOGON_SERVER_TRUST_ACCOUNT: Final[ThirtyTwoBits] = cached_hresult(0x80070711, 'HRESULT_FROM_WIN32(ERROR_NOLOGON_SERVER_TRUST_ACCOUNT)')
    WIN32__ERROR_DOMAIN_TRUST_INCONSISTENT: Final[ThirtyTwoBits] = cached_hresult(0x80070712, 'HRESULT_FROM_WIN32(ERROR_DOMAIN_TRUST_INCONSISTENT)')
    WIN32__ERROR_SERVER_HAS_OPEN_HANDLES: Final[ThirtyTwoBits] = cached_hresult(0x80070713, 'HRESULT_FROM_WIN32(ERROR_SERVER_HAS_OPEN_HANDLES)')
    WIN32__ERROR_RESOURCE_DATA_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80070714, 'HRESULT_FROM_WIN32(ERROR_RESOURCE_DATA_NOT_FOUND)')
    WIN32__ERROR_RESOURCE_TYPE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80070715, 'HRESULT_FROM_WIN32(ERROR_RESOURCE_TYPE_NOT_FOUND)')
    WIN32__ERROR_RESOURCE_NAME_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80070716, 'HRESULT_FROM_WIN32(ERROR_RESOURCE_NAME_NOT_FOUND)')
    WIN32__ERROR_RESOURCE_LANG_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80070717, 'HRESULT_FROM_WIN32(ERROR_RESOURCE_LANG_NOT_FOUND)')
    WIN32__ERROR_NOT_ENOUGH_QUOTA: Final[ThirtyTwoBits] = cached_hresult(0x80070718, 'HRESULT_FROM_WIN32(ERROR_NOT_ENOUGH_QUOTA)')
    WIN32__RPC_S_NO_INTERFACES: Final[ThirtyTwoBits] = cached_hresult(0x80070719, 'HRESULT_FROM_WIN32(RPC_S_NO_INTERFACES)')
    WIN32__RPC_S_CALL_CANCELLED: Final[ThirtyTwoBits] = cached_hresult(0x8007071A, 'HRESULT_FROM_WIN32(RPC_S_CALL_CANCELLED)')
    WIN32__RPC_S_BINDING_INCOMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x8007071B, 'HRESULT_FROM_WIN32(RPC_S_BINDING_INCOMPLETE)')
    WIN32__RPC_S_COMM_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8007071C, 'HRESULT_FROM_WIN32(RPC_S_COMM_FAILURE)')
    WIN32__RPC_S_UNSUPPORTED_AUTHN_LEVEL: Final[ThirtyTwoBits] = cached_hresult(0x8007071D, 'HRESULT_FROM_WIN32(RPC_S_UNSUPPORTED_AUTHN_LEVEL)')
    WIN32__RPC_S_NO_PRINC_NAME: Final[ThirtyTwoBits] = cached_hresult(0x8007071E, 'HRESULT_FROM_WIN32(RPC_S_NO_PRINC_NAME)')
    WIN32__RPC_S_NOT_RPC_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8007071F, 'HRESULT_FROM_WIN32(RPC_S_NOT_RPC_ERROR)')
    WIN32__RPC_S_UUID_LOCAL_ONLY: Final[ThirtyTwoBits] = cached_hresult(0x80070720, 'HRESULT_FROM_WIN32(RPC_S_UUID_LOCAL_ONLY)')
    WIN32__RPC_S_SEC_PKG_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80070721, 'HRESULT_FROM_WIN32(RPC_S_SEC_PKG_ERROR)')
    WIN32__RPC_S_NOT_CANCELLED: Final[ThirtyTwoBits] = cached_hresult(0x80070722, 'HRESULT_FROM_WIN32(RPC_S_NOT_CANCELLED)')
    WIN32__RPC_X_INVALID_ES_ACTION: Final[ThirtyTwoBits] = cached_hresult(0x80070723, 'HRESULT_FROM_WIN32(RPC_X_INVALID_ES_ACTION)')
    WIN32__RPC_X_WRONG_ES_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80070724, 'HRESULT_FROM_WIN32(RPC_X_WRONG_ES_VERSION)')
    WIN32__RPC_X_WRONG_STUB_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80070725, 'HRESULT_FROM_WIN32(RPC_X_WRONG_STUB_VERSION)')
    WIN32__RPC_X_INVALID_PIPE_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x80070726, 'HRESULT_FROM_WIN32(RPC_X_INVALID_PIPE_OBJECT)')
    WIN32__RPC_X_WRONG_PIPE_ORDER: Final[ThirtyTwoBits] = cached_hresult(0x80070727, 'HRESULT_FROM_WIN32(RPC_X_WRONG_PIPE_ORDER)')
    WIN32__RPC_X_WRONG_PIPE_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80070728, 'HRESULT_FROM_WIN32(RPC_X_WRONG_PIPE_VERSION)')
    WIN32__RPC_S_COOKIE_AUTH_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80070729, 'HRESULT_FROM_WIN32(RPC_S_COOKIE_AUTH_FAILED)')
    WIN32__RPC_S_DO_NOT_DISTURB: Final[ThirtyTwoBits] = cached_hresult(0x8007072A, 'HRESULT_FROM_WIN32(RPC_S_DO_NOT_DISTURB)')
    WIN32__RPC_S_SYSTEM_HANDLE_COUNT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8007072B, 'HRESULT_FROM_WIN32(RPC_S_SYSTEM_HANDLE_COUNT_EXCEEDED)')
    WIN32__RPC_S_SYSTEM_HANDLE_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x8007072C, 'HRESULT_FROM_WIN32(RPC_S_SYSTEM_HANDLE_TYPE_MISMATCH)')
    WIN32__RPC_S_GROUP_MEMBER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8007076A, 'HRESULT_FROM_WIN32(RPC_S_GROUP_MEMBER_NOT_FOUND)')
    WIN32__EPT_S_CANT_CREATE: Final[ThirtyTwoBits] = cached_hresult(0x8007076B, 'HRESULT_FROM_WIN32(EPT_S_CANT_CREATE)')
    WIN32__RPC_S_INVALID_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x8007076C, 'HRESULT_FROM_WIN32(RPC_S_INVALID_OBJECT)')
    WIN32__ERROR_INVALID_TIME: Final[ThirtyTwoBits] = cached_hresult(0x8007076D, 'HRESULT_FROM_WIN32(ERROR_INVALID_TIME)')
    WIN32__ERROR_INVALID_FORM_NAME: Final[ThirtyTwoBits] = cached_hresult(0x8007076E, 'HRESULT_FROM_WIN32(ERROR_INVALID_FORM_NAME)')
    WIN32__ERROR_INVALID_FORM_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x8007076F, 'HRESULT_FROM_WIN32(ERROR_INVALID_FORM_SIZE)')
    WIN32__ERROR_ALREADY_WAITING: Final[ThirtyTwoBits] = cached_hresult(0x80070770, 'HRESULT_FROM_WIN32(ERROR_ALREADY_WAITING)')
    WIN32__ERROR_PRINTER_DELETED: Final[ThirtyTwoBits] = cached_hresult(0x80070771, 'HRESULT_FROM_WIN32(ERROR_PRINTER_DELETED)')
    WIN32__ERROR_INVALID_PRINTER_STATE: Final[ThirtyTwoBits] = cached_hresult(0x80070772, 'HRESULT_FROM_WIN32(ERROR_INVALID_PRINTER_STATE)')
    WIN32__ERROR_PASSWORD_MUST_CHANGE: Final[ThirtyTwoBits] = cached_hresult(0x80070773, 'HRESULT_FROM_WIN32(ERROR_PASSWORD_MUST_CHANGE)')
    WIN32__ERROR_DOMAIN_CONTROLLER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80070774, 'HRESULT_FROM_WIN32(ERROR_DOMAIN_CONTROLLER_NOT_FOUND)')
    WIN32__ERROR_ACCOUNT_LOCKED_OUT: Final[ThirtyTwoBits] = cached_hresult(0x80070775, 'HRESULT_FROM_WIN32(ERROR_ACCOUNT_LOCKED_OUT)')
    WIN32__OR_INVALID_OXID: Final[ThirtyTwoBits] = cached_hresult(0x80070776, 'HRESULT_FROM_WIN32(OR_INVALID_OXID)')
    WIN32__OR_INVALID_OID: Final[ThirtyTwoBits] = cached_hresult(0x80070777, 'HRESULT_FROM_WIN32(OR_INVALID_OID)')
    WIN32__OR_INVALID_SET: Final[ThirtyTwoBits] = cached_hresult(0x80070778, 'HRESULT_FROM_WIN32(OR_INVALID_SET)')
    WIN32__RPC_S_SEND_INCOMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x80070779, 'HRESULT_FROM_WIN32(RPC_S_SEND_INCOMPLETE)')
    WIN32__RPC_S_INVALID_ASYNC_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x8007077A, 'HRESULT_FROM_WIN32(RPC_S_INVALID_ASYNC_HANDLE)')
    WIN32__RPC_S_INVALID_ASYNC_CALL: Final[ThirtyTwoBits] = cached_hresult(0x8007077B, 'HRESULT_FROM_WIN32(RPC_S_INVALID_ASYNC_CALL)')
    WIN32__RPC_X_PIPE_CLOSED: Final[ThirtyTwoBits] = cached_hresult(0x8007077C, 'HRESULT_FROM_WIN32(RPC_X_PIPE_CLOSED)')
    WIN32__RPC_X_PIPE_DISCIPLINE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8007077D, 'HRESULT_FROM_WIN32(RPC_X_PIPE_DISCIPLINE_ERROR)')
    WIN32__RPC_X_PIPE_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0x8007077E, 'HRESULT_FROM_WIN32(RPC_X_PIPE_EMPTY)')
    WIN32__ERROR_NO_SITENAME: Final[ThirtyTwoBits] = cached_hresult(0x8007077F, 'HRESULT_FROM_WIN32(ERROR_NO_SITENAME)')
    WIN32__ERROR_CANT_ACCESS_FILE: Final[ThirtyTwoBits] = cached_hresult(0x80070780, 'HRESULT_FROM_WIN32(ERROR_CANT_ACCESS_FILE)')
    WIN32__ERROR_CANT_RESOLVE_FILENAME: Final[ThirtyTwoBits] = cached_hresult(0x80070781, 'HRESULT_FROM_WIN32(ERROR_CANT_RESOLVE_FILENAME)')
    WIN32__RPC_S_ENTRY_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80070782, 'HRESULT_FROM_WIN32(RPC_S_ENTRY_TYPE_MISMATCH)')
    WIN32__RPC_S_NOT_ALL_OBJS_EXPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80070783, 'HRESULT_FROM_WIN32(RPC_S_NOT_ALL_OBJS_EXPORTED)')
    WIN32__RPC_S_INTERFACE_NOT_EXPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80070784, 'HRESULT_FROM_WIN32(RPC_S_INTERFACE_NOT_EXPORTED)')
    WIN32__RPC_S_PROFILE_NOT_ADDED: Final[ThirtyTwoBits] = cached_hresult(0x80070785, 'HRESULT_FROM_WIN32(RPC_S_PROFILE_NOT_ADDED)')
    WIN32__RPC_S_PRF_ELT_NOT_ADDED: Final[ThirtyTwoBits] = cached_hresult(0x80070786, 'HRESULT_FROM_WIN32(RPC_S_PRF_ELT_NOT_ADDED)')
    WIN32__RPC_S_PRF_ELT_NOT_REMOVED: Final[ThirtyTwoBits] = cached_hresult(0x80070787, 'HRESULT_FROM_WIN32(RPC_S_PRF_ELT_NOT_REMOVED)')
    WIN32__RPC_S_GRP_ELT_NOT_ADDED: Final[ThirtyTwoBits] = cached_hresult(0x80070788, 'HRESULT_FROM_WIN32(RPC_S_GRP_ELT_NOT_ADDED)')
    WIN32__RPC_S_GRP_ELT_NOT_REMOVED: Final[ThirtyTwoBits] = cached_hresult(0x80070789, 'HRESULT_FROM_WIN32(RPC_S_GRP_ELT_NOT_REMOVED)')
    WIN32__ERROR_KM_DRIVER_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x8007078A, 'HRESULT_FROM_WIN32(ERROR_KM_DRIVER_BLOCKED)')
    WIN32__ERROR_CONTEXT_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x8007078B, 'HRESULT_FROM_WIN32(ERROR_CONTEXT_EXPIRED)')
    WIN32__ERROR_PER_USER_TRUST_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8007078C, 'HRESULT_FROM_WIN32(ERROR_PER_USER_TRUST_QUOTA_EXCEEDED)')
    WIN32__ERROR_ALL_USER_TRUST_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8007078D, 'HRESULT_FROM_WIN32(ERROR_ALL_USER_TRUST_QUOTA_EXCEEDED)')
    WIN32__ERROR_USER_DELETE_TRUST_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8007078E, 'HRESULT_FROM_WIN32(ERROR_USER_DELETE_TRUST_QUOTA_EXCEEDED)')
    WIN32__ERROR_AUTHENTICATION_FIREWALL_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8007078F, 'HRESULT_FROM_WIN32(ERROR_AUTHENTICATION_FIREWALL_FAILED)')
    WIN32__ERROR_REMOTE_PRINT_CONNECTIONS_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x80070790, 'HRESULT_FROM_WIN32(ERROR_REMOTE_PRINT_CONNECTIONS_BLOCKED)')
    WIN32__ERROR_NTLM_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x80070791, 'HRESULT_FROM_WIN32(ERROR_NTLM_BLOCKED)')
    WIN32__ERROR_PASSWORD_CHANGE_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80070792, 'HRESULT_FROM_WIN32(ERROR_PASSWORD_CHANGE_REQUIRED)')
    WIN32__ERROR_LOST_MODE_LOGON_RESTRICTION: Final[ThirtyTwoBits] = cached_hresult(0x80070793, 'HRESULT_FROM_WIN32(ERROR_LOST_MODE_LOGON_RESTRICTION)')
    WIN32__ERROR_INVALID_PIXEL_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0x800707D0, 'HRESULT_FROM_WIN32(ERROR_INVALID_PIXEL_FORMAT)')
    WIN32__ERROR_BAD_DRIVER: Final[ThirtyTwoBits] = cached_hresult(0x800707D1, 'HRESULT_FROM_WIN32(ERROR_BAD_DRIVER)')
    WIN32__ERROR_INVALID_WINDOW_STYLE: Final[ThirtyTwoBits] = cached_hresult(0x800707D2, 'HRESULT_FROM_WIN32(ERROR_INVALID_WINDOW_STYLE)')
    WIN32__ERROR_METAFILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x800707D3, 'HRESULT_FROM_WIN32(ERROR_METAFILE_NOT_SUPPORTED)')
    WIN32__ERROR_TRANSFORM_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x800707D4, 'HRESULT_FROM_WIN32(ERROR_TRANSFORM_NOT_SUPPORTED)')
    WIN32__ERROR_CLIPPING_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x800707D5, 'HRESULT_FROM_WIN32(ERROR_CLIPPING_NOT_SUPPORTED)')
    WIN32__ERROR_INVALID_CMM: Final[ThirtyTwoBits] = cached_hresult(0x800707DA, 'HRESULT_FROM_WIN32(ERROR_INVALID_CMM)')
    WIN32__ERROR_INVALID_PROFILE: Final[ThirtyTwoBits] = cached_hresult(0x800707DB, 'HRESULT_FROM_WIN32(ERROR_INVALID_PROFILE)')
    WIN32__ERROR_TAG_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800707DC, 'HRESULT_FROM_WIN32(ERROR_TAG_NOT_FOUND)')
    WIN32__ERROR_TAG_NOT_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0x800707DD, 'HRESULT_FROM_WIN32(ERROR_TAG_NOT_PRESENT)')
    WIN32__ERROR_DUPLICATE_TAG: Final[ThirtyTwoBits] = cached_hresult(0x800707DE, 'HRESULT_FROM_WIN32(ERROR_DUPLICATE_TAG)')
    WIN32__ERROR_PROFILE_NOT_ASSOCIATED_WITH_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0x800707DF, 'HRESULT_FROM_WIN32(ERROR_PROFILE_NOT_ASSOCIATED_WITH_DEVICE)')
    WIN32__ERROR_PROFILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800707E0, 'HRESULT_FROM_WIN32(ERROR_PROFILE_NOT_FOUND)')
    WIN32__ERROR_INVALID_COLORSPACE: Final[ThirtyTwoBits] = cached_hresult(0x800707E1, 'HRESULT_FROM_WIN32(ERROR_INVALID_COLORSPACE)')
    WIN32__ERROR_ICM_NOT_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0x800707E2, 'HRESULT_FROM_WIN32(ERROR_ICM_NOT_ENABLED)')
    WIN32__ERROR_DELETING_ICM_XFORM: Final[ThirtyTwoBits] = cached_hresult(0x800707E3, 'HRESULT_FROM_WIN32(ERROR_DELETING_ICM_XFORM)')
    WIN32__ERROR_INVALID_TRANSFORM: Final[ThirtyTwoBits] = cached_hresult(0x800707E4, 'HRESULT_FROM_WIN32(ERROR_INVALID_TRANSFORM)')
    WIN32__ERROR_COLORSPACE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x800707E5, 'HRESULT_FROM_WIN32(ERROR_COLORSPACE_MISMATCH)')
    WIN32__ERROR_INVALID_COLORINDEX: Final[ThirtyTwoBits] = cached_hresult(0x800707E6, 'HRESULT_FROM_WIN32(ERROR_INVALID_COLORINDEX)')
    WIN32__ERROR_PROFILE_DOES_NOT_MATCH_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0x800707E7, 'HRESULT_FROM_WIN32(ERROR_PROFILE_DOES_NOT_MATCH_DEVICE)')
    WIN32__ERROR_CONNECTED_OTHER_PASSWORD: Final[ThirtyTwoBits] = cached_hresult(0x8007083C, 'HRESULT_FROM_WIN32(ERROR_CONNECTED_OTHER_PASSWORD)')
    WIN32__ERROR_CONNECTED_OTHER_PASSWORD_DEFAULT: Final[ThirtyTwoBits] = cached_hresult(0x8007083D, 'HRESULT_FROM_WIN32(ERROR_CONNECTED_OTHER_PASSWORD_DEFAULT)')
    WIN32__ERROR_BAD_USERNAME: Final[ThirtyTwoBits] = cached_hresult(0x8007089A, 'HRESULT_FROM_WIN32(ERROR_BAD_USERNAME)')
    WIN32__ERROR_NOT_CONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x800708CA, 'HRESULT_FROM_WIN32(ERROR_NOT_CONNECTED)')
    WIN32__ERROR_OPEN_FILES: Final[ThirtyTwoBits] = cached_hresult(0x80070961, 'HRESULT_FROM_WIN32(ERROR_OPEN_FILES)')
    WIN32__ERROR_ACTIVE_CONNECTIONS: Final[ThirtyTwoBits] = cached_hresult(0x80070962, 'HRESULT_FROM_WIN32(ERROR_ACTIVE_CONNECTIONS)')
    WIN32__ERROR_DEVICE_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x80070964, 'HRESULT_FROM_WIN32(ERROR_DEVICE_IN_USE)')
    WIN32__ERROR_UNKNOWN_PRINT_MONITOR: Final[ThirtyTwoBits] = cached_hresult(0x80070BB8, 'HRESULT_FROM_WIN32(ERROR_UNKNOWN_PRINT_MONITOR)')
    WIN32__ERROR_PRINTER_DRIVER_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x80070BB9, 'HRESULT_FROM_WIN32(ERROR_PRINTER_DRIVER_IN_USE)')
    WIN32__ERROR_SPOOL_FILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80070BBA, 'HRESULT_FROM_WIN32(ERROR_SPOOL_FILE_NOT_FOUND)')
    WIN32__ERROR_SPL_NO_STARTDOC: Final[ThirtyTwoBits] = cached_hresult(0x80070BBB, 'HRESULT_FROM_WIN32(ERROR_SPL_NO_STARTDOC)')
    WIN32__ERROR_SPL_NO_ADDJOB: Final[ThirtyTwoBits] = cached_hresult(0x80070BBC, 'HRESULT_FROM_WIN32(ERROR_SPL_NO_ADDJOB)')
    WIN32__ERROR_PRINT_PROCESSOR_ALREADY_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x80070BBD, 'HRESULT_FROM_WIN32(ERROR_PRINT_PROCESSOR_ALREADY_INSTALLED)')
    WIN32__ERROR_PRINT_MONITOR_ALREADY_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x80070BBE, 'HRESULT_FROM_WIN32(ERROR_PRINT_MONITOR_ALREADY_INSTALLED)')
    WIN32__ERROR_INVALID_PRINT_MONITOR: Final[ThirtyTwoBits] = cached_hresult(0x80070BBF, 'HRESULT_FROM_WIN32(ERROR_INVALID_PRINT_MONITOR)')
    WIN32__ERROR_PRINT_MONITOR_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x80070BC0, 'HRESULT_FROM_WIN32(ERROR_PRINT_MONITOR_IN_USE)')
    WIN32__ERROR_PRINTER_HAS_JOBS_QUEUED: Final[ThirtyTwoBits] = cached_hresult(0x80070BC1, 'HRESULT_FROM_WIN32(ERROR_PRINTER_HAS_JOBS_QUEUED)')
    WIN32__ERROR_SUCCESS_REBOOT_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80070BC2, 'HRESULT_FROM_WIN32(ERROR_SUCCESS_REBOOT_REQUIRED)')
    WIN32__ERROR_SUCCESS_RESTART_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80070BC3, 'HRESULT_FROM_WIN32(ERROR_SUCCESS_RESTART_REQUIRED)')
    WIN32__ERROR_PRINTER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80070BC4, 'HRESULT_FROM_WIN32(ERROR_PRINTER_NOT_FOUND)')
    WIN32__ERROR_PRINTER_DRIVER_WARNED: Final[ThirtyTwoBits] = cached_hresult(0x80070BC5, 'HRESULT_FROM_WIN32(ERROR_PRINTER_DRIVER_WARNED)')
    WIN32__ERROR_PRINTER_DRIVER_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x80070BC6, 'HRESULT_FROM_WIN32(ERROR_PRINTER_DRIVER_BLOCKED)')
    WIN32__ERROR_PRINTER_DRIVER_PACKAGE_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x80070BC7, 'HRESULT_FROM_WIN32(ERROR_PRINTER_DRIVER_PACKAGE_IN_USE)')
    WIN32__ERROR_CORE_DRIVER_PACKAGE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80070BC8, 'HRESULT_FROM_WIN32(ERROR_CORE_DRIVER_PACKAGE_NOT_FOUND)')
    WIN32__ERROR_FAIL_REBOOT_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80070BC9, 'HRESULT_FROM_WIN32(ERROR_FAIL_REBOOT_REQUIRED)')
    WIN32__ERROR_FAIL_REBOOT_INITIATED: Final[ThirtyTwoBits] = cached_hresult(0x80070BCA, 'HRESULT_FROM_WIN32(ERROR_FAIL_REBOOT_INITIATED)')
    WIN32__ERROR_PRINTER_DRIVER_DOWNLOAD_NEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80070BCB, 'HRESULT_FROM_WIN32(ERROR_PRINTER_DRIVER_DOWNLOAD_NEEDED)')
    WIN32__ERROR_PRINT_JOB_RESTART_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80070BCC, 'HRESULT_FROM_WIN32(ERROR_PRINT_JOB_RESTART_REQUIRED)')
    WIN32__ERROR_INVALID_PRINTER_DRIVER_MANIFEST: Final[ThirtyTwoBits] = cached_hresult(0x80070BCD, 'HRESULT_FROM_WIN32(ERROR_INVALID_PRINTER_DRIVER_MANIFEST)')
    WIN32__ERROR_PRINTER_NOT_SHAREABLE: Final[ThirtyTwoBits] = cached_hresult(0x80070BCE, 'HRESULT_FROM_WIN32(ERROR_PRINTER_NOT_SHAREABLE)')
    WIN32__ERROR_SERVER_SERVICE_CALL_REQUIRES_SMB1: Final[ThirtyTwoBits] = cached_hresult(0x80070BCF, 'HRESULT_FROM_WIN32(ERROR_SERVER_SERVICE_CALL_REQUIRES_SMB1)')
    WIN32__ERROR_NETWORK_AUTHENTICATION_PROMPT_CANCELED: Final[ThirtyTwoBits] = cached_hresult(0x80070BD0, 'HRESULT_FROM_WIN32(ERROR_NETWORK_AUTHENTICATION_PROMPT_CANCELED)')
    WIN32__ERROR_REQUEST_PAUSED: Final[ThirtyTwoBits] = cached_hresult(0x80070BEA, 'HRESULT_FROM_WIN32(ERROR_REQUEST_PAUSED)')
    WIN32__ERROR_APPEXEC_CONDITION_NOT_SATISFIED: Final[ThirtyTwoBits] = cached_hresult(0x80070BF4, 'HRESULT_FROM_WIN32(ERROR_APPEXEC_CONDITION_NOT_SATISFIED)')
    WIN32__ERROR_APPEXEC_HANDLE_INVALIDATED: Final[ThirtyTwoBits] = cached_hresult(0x80070BF5, 'HRESULT_FROM_WIN32(ERROR_APPEXEC_HANDLE_INVALIDATED)')
    WIN32__ERROR_APPEXEC_INVALID_HOST_GENERATION: Final[ThirtyTwoBits] = cached_hresult(0x80070BF6, 'HRESULT_FROM_WIN32(ERROR_APPEXEC_INVALID_HOST_GENERATION)')
    WIN32__ERROR_APPEXEC_UNEXPECTED_PROCESS_REGISTRATION: Final[ThirtyTwoBits] = cached_hresult(0x80070BF7, 'HRESULT_FROM_WIN32(ERROR_APPEXEC_UNEXPECTED_PROCESS_REGISTRATION)')
    WIN32__ERROR_APPEXEC_INVALID_HOST_STATE: Final[ThirtyTwoBits] = cached_hresult(0x80070BF8, 'HRESULT_FROM_WIN32(ERROR_APPEXEC_INVALID_HOST_STATE)')
    WIN32__ERROR_APPEXEC_NO_DONOR: Final[ThirtyTwoBits] = cached_hresult(0x80070BF9, 'HRESULT_FROM_WIN32(ERROR_APPEXEC_NO_DONOR)')
    WIN32__ERROR_APPEXEC_HOST_ID_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80070BFA, 'HRESULT_FROM_WIN32(ERROR_APPEXEC_HOST_ID_MISMATCH)')
    WIN32__ERROR_APPEXEC_UNKNOWN_USER: Final[ThirtyTwoBits] = cached_hresult(0x80070BFB, 'HRESULT_FROM_WIN32(ERROR_APPEXEC_UNKNOWN_USER)')
    WIN32__ERROR_APPEXEC_APP_COMPAT_BLOCK: Final[ThirtyTwoBits] = cached_hresult(0x80070BFC, 'HRESULT_FROM_WIN32(ERROR_APPEXEC_APP_COMPAT_BLOCK)')
    WIN32__ERROR_APPEXEC_CALLER_WAIT_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x80070BFD, 'HRESULT_FROM_WIN32(ERROR_APPEXEC_CALLER_WAIT_TIMEOUT)')
    WIN32__ERROR_APPEXEC_CALLER_WAIT_TIMEOUT_TERMINATION: Final[ThirtyTwoBits] = cached_hresult(0x80070BFE, 'HRESULT_FROM_WIN32(ERROR_APPEXEC_CALLER_WAIT_TIMEOUT_TERMINATION)')
    WIN32__ERROR_APPEXEC_CALLER_WAIT_TIMEOUT_LICENSING: Final[ThirtyTwoBits] = cached_hresult(0x80070BFF, 'HRESULT_FROM_WIN32(ERROR_APPEXEC_CALLER_WAIT_TIMEOUT_LICENSING)')
    WIN32__ERROR_APPEXEC_CALLER_WAIT_TIMEOUT_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0x80070C00, 'HRESULT_FROM_WIN32(ERROR_APPEXEC_CALLER_WAIT_TIMEOUT_RESOURCES)')
    WIN32__ERROR_VRF_VOLATILE_CFG_AND_IO_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0x80070C08, 'HRESULT_FROM_WIN32(ERROR_VRF_VOLATILE_CFG_AND_IO_ENABLED)')
    WIN32__ERROR_VRF_VOLATILE_NOT_STOPPABLE: Final[ThirtyTwoBits] = cached_hresult(0x80070C09, 'HRESULT_FROM_WIN32(ERROR_VRF_VOLATILE_NOT_STOPPABLE)')
    WIN32__ERROR_VRF_VOLATILE_SAFE_MODE: Final[ThirtyTwoBits] = cached_hresult(0x80070C0A, 'HRESULT_FROM_WIN32(ERROR_VRF_VOLATILE_SAFE_MODE)')
    WIN32__ERROR_VRF_VOLATILE_NOT_RUNNABLE_SYSTEM: Final[ThirtyTwoBits] = cached_hresult(0x80070C0B, 'HRESULT_FROM_WIN32(ERROR_VRF_VOLATILE_NOT_RUNNABLE_SYSTEM)')
    WIN32__ERROR_VRF_VOLATILE_NOT_SUPPORTED_RULECLASS: Final[ThirtyTwoBits] = cached_hresult(0x80070C0C, 'HRESULT_FROM_WIN32(ERROR_VRF_VOLATILE_NOT_SUPPORTED_RULECLASS)')
    WIN32__ERROR_VRF_VOLATILE_PROTECTED_DRIVER: Final[ThirtyTwoBits] = cached_hresult(0x80070C0D, 'HRESULT_FROM_WIN32(ERROR_VRF_VOLATILE_PROTECTED_DRIVER)')
    WIN32__ERROR_VRF_VOLATILE_NMI_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x80070C0E, 'HRESULT_FROM_WIN32(ERROR_VRF_VOLATILE_NMI_REGISTERED)')
    WIN32__ERROR_VRF_VOLATILE_SETTINGS_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x80070C0F, 'HRESULT_FROM_WIN32(ERROR_VRF_VOLATILE_SETTINGS_CONFLICT)')
    WIN32__ERROR_DIF_IOCALLBACK_NOT_REPLACED: Final[ThirtyTwoBits] = cached_hresult(0x80070C76, 'HRESULT_FROM_WIN32(ERROR_DIF_IOCALLBACK_NOT_REPLACED)')
    WIN32__ERROR_DIF_LIVEDUMP_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80070C77, 'HRESULT_FROM_WIN32(ERROR_DIF_LIVEDUMP_LIMIT_EXCEEDED)')
    WIN32__ERROR_DIF_VOLATILE_SECTION_NOT_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x80070C78, 'HRESULT_FROM_WIN32(ERROR_DIF_VOLATILE_SECTION_NOT_LOCKED)')
    WIN32__ERROR_DIF_VOLATILE_DRIVER_HOTPATCHED: Final[ThirtyTwoBits] = cached_hresult(0x80070C79, 'HRESULT_FROM_WIN32(ERROR_DIF_VOLATILE_DRIVER_HOTPATCHED)')
    WIN32__ERROR_DIF_VOLATILE_INVALID_INFO: Final[ThirtyTwoBits] = cached_hresult(0x80070C7A, 'HRESULT_FROM_WIN32(ERROR_DIF_VOLATILE_INVALID_INFO)')
    WIN32__ERROR_DIF_VOLATILE_DRIVER_IS_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x80070C7B, 'HRESULT_FROM_WIN32(ERROR_DIF_VOLATILE_DRIVER_IS_NOT_RUNNING)')
    WIN32__ERROR_DIF_VOLATILE_PLUGIN_IS_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x80070C7C, 'HRESULT_FROM_WIN32(ERROR_DIF_VOLATILE_PLUGIN_IS_NOT_RUNNING)')
    WIN32__ERROR_DIF_VOLATILE_PLUGIN_CHANGE_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80070C7D, 'HRESULT_FROM_WIN32(ERROR_DIF_VOLATILE_PLUGIN_CHANGE_NOT_ALLOWED)')
    WIN32__ERROR_DIF_VOLATILE_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80070C7E, 'HRESULT_FROM_WIN32(ERROR_DIF_VOLATILE_NOT_ALLOWED)')
    WIN32__ERROR_DIF_BINDING_API_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80070C7F, 'HRESULT_FROM_WIN32(ERROR_DIF_BINDING_API_NOT_FOUND)')
    WIN32__ERROR_IO_REISSUE_AS_CACHED: Final[ThirtyTwoBits] = cached_hresult(0x80070F6E, 'HRESULT_FROM_WIN32(ERROR_IO_REISSUE_AS_CACHED)')
    WIN32__ERROR_WINS_INTERNAL: Final[ThirtyTwoBits] = cached_hresult(0x80070FA0, 'HRESULT_FROM_WIN32(ERROR_WINS_INTERNAL)')
    WIN32__ERROR_CAN_NOT_DEL_LOCAL_WINS: Final[ThirtyTwoBits] = cached_hresult(0x80070FA1, 'HRESULT_FROM_WIN32(ERROR_CAN_NOT_DEL_LOCAL_WINS)')
    WIN32__ERROR_STATIC_INIT: Final[ThirtyTwoBits] = cached_hresult(0x80070FA2, 'HRESULT_FROM_WIN32(ERROR_STATIC_INIT)')
    WIN32__ERROR_INC_BACKUP: Final[ThirtyTwoBits] = cached_hresult(0x80070FA3, 'HRESULT_FROM_WIN32(ERROR_INC_BACKUP)')
    WIN32__ERROR_FULL_BACKUP: Final[ThirtyTwoBits] = cached_hresult(0x80070FA4, 'HRESULT_FROM_WIN32(ERROR_FULL_BACKUP)')
    WIN32__ERROR_REC_NON_EXISTENT: Final[ThirtyTwoBits] = cached_hresult(0x80070FA5, 'HRESULT_FROM_WIN32(ERROR_REC_NON_EXISTENT)')
    WIN32__ERROR_RPL_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80070FA6, 'HRESULT_FROM_WIN32(ERROR_RPL_NOT_ALLOWED)')
    WIN32__PEERDIST_ERROR_CONTENTINFO_VERSION_UNSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80070FD2, 'HRESULT_FROM_WIN32(PEERDIST_ERROR_CONTENTINFO_VERSION_UNSUPPORTED)')
    WIN32__PEERDIST_ERROR_CANNOT_PARSE_CONTENTINFO: Final[ThirtyTwoBits] = cached_hresult(0x80070FD3, 'HRESULT_FROM_WIN32(PEERDIST_ERROR_CANNOT_PARSE_CONTENTINFO)')
    WIN32__PEERDIST_ERROR_MISSING_DATA: Final[ThirtyTwoBits] = cached_hresult(0x80070FD4, 'HRESULT_FROM_WIN32(PEERDIST_ERROR_MISSING_DATA)')
    WIN32__PEERDIST_ERROR_NO_MORE: Final[ThirtyTwoBits] = cached_hresult(0x80070FD5, 'HRESULT_FROM_WIN32(PEERDIST_ERROR_NO_MORE)')
    WIN32__PEERDIST_ERROR_NOT_INITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0x80070FD6, 'HRESULT_FROM_WIN32(PEERDIST_ERROR_NOT_INITIALIZED)')
    WIN32__PEERDIST_ERROR_ALREADY_INITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0x80070FD7, 'HRESULT_FROM_WIN32(PEERDIST_ERROR_ALREADY_INITIALIZED)')
    WIN32__PEERDIST_ERROR_SHUTDOWN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x80070FD8, 'HRESULT_FROM_WIN32(PEERDIST_ERROR_SHUTDOWN_IN_PROGRESS)')
    WIN32__PEERDIST_ERROR_INVALIDATED: Final[ThirtyTwoBits] = cached_hresult(0x80070FD9, 'HRESULT_FROM_WIN32(PEERDIST_ERROR_INVALIDATED)')
    WIN32__PEERDIST_ERROR_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80070FDA, 'HRESULT_FROM_WIN32(PEERDIST_ERROR_ALREADY_EXISTS)')
    WIN32__PEERDIST_ERROR_OPERATION_NOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0x80070FDB, 'HRESULT_FROM_WIN32(PEERDIST_ERROR_OPERATION_NOTFOUND)')
    WIN32__PEERDIST_ERROR_ALREADY_COMPLETED: Final[ThirtyTwoBits] = cached_hresult(0x80070FDC, 'HRESULT_FROM_WIN32(PEERDIST_ERROR_ALREADY_COMPLETED)')
    WIN32__PEERDIST_ERROR_OUT_OF_BOUNDS: Final[ThirtyTwoBits] = cached_hresult(0x80070FDD, 'HRESULT_FROM_WIN32(PEERDIST_ERROR_OUT_OF_BOUNDS)')
    WIN32__PEERDIST_ERROR_VERSION_UNSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80070FDE, 'HRESULT_FROM_WIN32(PEERDIST_ERROR_VERSION_UNSUPPORTED)')
    WIN32__PEERDIST_ERROR_INVALID_CONFIGURATION: Final[ThirtyTwoBits] = cached_hresult(0x80070FDF, 'HRESULT_FROM_WIN32(PEERDIST_ERROR_INVALID_CONFIGURATION)')
    WIN32__PEERDIST_ERROR_NOT_LICENSED: Final[ThirtyTwoBits] = cached_hresult(0x80070FE0, 'HRESULT_FROM_WIN32(PEERDIST_ERROR_NOT_LICENSED)')
    WIN32__PEERDIST_ERROR_SERVICE_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80070FE1, 'HRESULT_FROM_WIN32(PEERDIST_ERROR_SERVICE_UNAVAILABLE)')
    WIN32__PEERDIST_ERROR_TRUST_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80070FE2, 'HRESULT_FROM_WIN32(PEERDIST_ERROR_TRUST_FAILURE)')
    WIN32__ERROR_DHCP_ADDRESS_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x80071004, 'HRESULT_FROM_WIN32(ERROR_DHCP_ADDRESS_CONFLICT)')
    WIN32__ERROR_WMI_GUID_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80071068, 'HRESULT_FROM_WIN32(ERROR_WMI_GUID_NOT_FOUND)')
    WIN32__ERROR_WMI_INSTANCE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80071069, 'HRESULT_FROM_WIN32(ERROR_WMI_INSTANCE_NOT_FOUND)')
    WIN32__ERROR_WMI_ITEMID_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8007106A, 'HRESULT_FROM_WIN32(ERROR_WMI_ITEMID_NOT_FOUND)')
    WIN32__ERROR_WMI_TRY_AGAIN: Final[ThirtyTwoBits] = cached_hresult(0x8007106B, 'HRESULT_FROM_WIN32(ERROR_WMI_TRY_AGAIN)')
    WIN32__ERROR_WMI_DP_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8007106C, 'HRESULT_FROM_WIN32(ERROR_WMI_DP_NOT_FOUND)')
    WIN32__ERROR_WMI_UNRESOLVED_INSTANCE_REF: Final[ThirtyTwoBits] = cached_hresult(0x8007106D, 'HRESULT_FROM_WIN32(ERROR_WMI_UNRESOLVED_INSTANCE_REF)')
    WIN32__ERROR_WMI_ALREADY_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0x8007106E, 'HRESULT_FROM_WIN32(ERROR_WMI_ALREADY_ENABLED)')
    WIN32__ERROR_WMI_GUID_DISCONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x8007106F, 'HRESULT_FROM_WIN32(ERROR_WMI_GUID_DISCONNECTED)')
    WIN32__ERROR_WMI_SERVER_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80071070, 'HRESULT_FROM_WIN32(ERROR_WMI_SERVER_UNAVAILABLE)')
    WIN32__ERROR_WMI_DP_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80071071, 'HRESULT_FROM_WIN32(ERROR_WMI_DP_FAILED)')
    WIN32__ERROR_WMI_INVALID_MOF: Final[ThirtyTwoBits] = cached_hresult(0x80071072, 'HRESULT_FROM_WIN32(ERROR_WMI_INVALID_MOF)')
    WIN32__ERROR_WMI_INVALID_REGINFO: Final[ThirtyTwoBits] = cached_hresult(0x80071073, 'HRESULT_FROM_WIN32(ERROR_WMI_INVALID_REGINFO)')
    WIN32__ERROR_WMI_ALREADY_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80071074, 'HRESULT_FROM_WIN32(ERROR_WMI_ALREADY_DISABLED)')
    WIN32__ERROR_WMI_READ_ONLY: Final[ThirtyTwoBits] = cached_hresult(0x80071075, 'HRESULT_FROM_WIN32(ERROR_WMI_READ_ONLY)')
    WIN32__ERROR_WMI_SET_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80071076, 'HRESULT_FROM_WIN32(ERROR_WMI_SET_FAILURE)')
    WIN32__ERROR_NOT_APPCONTAINER: Final[ThirtyTwoBits] = cached_hresult(0x8007109A, 'HRESULT_FROM_WIN32(ERROR_NOT_APPCONTAINER)')
    WIN32__ERROR_APPCONTAINER_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8007109B, 'HRESULT_FROM_WIN32(ERROR_APPCONTAINER_REQUIRED)')
    WIN32__ERROR_NOT_SUPPORTED_IN_APPCONTAINER: Final[ThirtyTwoBits] = cached_hresult(0x8007109C, 'HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED_IN_APPCONTAINER)')
    WIN32__ERROR_INVALID_PACKAGE_SID_LENGTH: Final[ThirtyTwoBits] = cached_hresult(0x8007109D, 'HRESULT_FROM_WIN32(ERROR_INVALID_PACKAGE_SID_LENGTH)')
    WIN32__ERROR_INVALID_MEDIA: Final[ThirtyTwoBits] = cached_hresult(0x800710CC, 'HRESULT_FROM_WIN32(ERROR_INVALID_MEDIA)')
    WIN32__ERROR_INVALID_LIBRARY: Final[ThirtyTwoBits] = cached_hresult(0x800710CD, 'HRESULT_FROM_WIN32(ERROR_INVALID_LIBRARY)')
    WIN32__ERROR_INVALID_MEDIA_POOL: Final[ThirtyTwoBits] = cached_hresult(0x800710CE, 'HRESULT_FROM_WIN32(ERROR_INVALID_MEDIA_POOL)')
    WIN32__ERROR_DRIVE_MEDIA_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x800710CF, 'HRESULT_FROM_WIN32(ERROR_DRIVE_MEDIA_MISMATCH)')
    WIN32__ERROR_MEDIA_OFFLINE: Final[ThirtyTwoBits] = cached_hresult(0x800710D0, 'HRESULT_FROM_WIN32(ERROR_MEDIA_OFFLINE)')
    WIN32__ERROR_LIBRARY_OFFLINE: Final[ThirtyTwoBits] = cached_hresult(0x800710D1, 'HRESULT_FROM_WIN32(ERROR_LIBRARY_OFFLINE)')
    WIN32__ERROR_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0x800710D2, 'HRESULT_FROM_WIN32(ERROR_EMPTY)')
    WIN32__ERROR_NOT_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0x800710D3, 'HRESULT_FROM_WIN32(ERROR_NOT_EMPTY)')
    WIN32__ERROR_MEDIA_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x800710D4, 'HRESULT_FROM_WIN32(ERROR_MEDIA_UNAVAILABLE)')
    WIN32__ERROR_RESOURCE_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x800710D5, 'HRESULT_FROM_WIN32(ERROR_RESOURCE_DISABLED)')
    WIN32__ERROR_INVALID_CLEANER: Final[ThirtyTwoBits] = cached_hresult(0x800710D6, 'HRESULT_FROM_WIN32(ERROR_INVALID_CLEANER)')
    WIN32__ERROR_UNABLE_TO_CLEAN: Final[ThirtyTwoBits] = cached_hresult(0x800710D7, 'HRESULT_FROM_WIN32(ERROR_UNABLE_TO_CLEAN)')
    WIN32__ERROR_OBJECT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800710D8, 'HRESULT_FROM_WIN32(ERROR_OBJECT_NOT_FOUND)')
    WIN32__ERROR_DATABASE_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x800710D9, 'HRESULT_FROM_WIN32(ERROR_DATABASE_FAILURE)')
    WIN32__ERROR_DATABASE_FULL: Final[ThirtyTwoBits] = cached_hresult(0x800710DA, 'HRESULT_FROM_WIN32(ERROR_DATABASE_FULL)')
    WIN32__ERROR_MEDIA_INCOMPATIBLE: Final[ThirtyTwoBits] = cached_hresult(0x800710DB, 'HRESULT_FROM_WIN32(ERROR_MEDIA_INCOMPATIBLE)')
    WIN32__ERROR_RESOURCE_NOT_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0x800710DC, 'HRESULT_FROM_WIN32(ERROR_RESOURCE_NOT_PRESENT)')
    WIN32__ERROR_INVALID_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x800710DD, 'HRESULT_FROM_WIN32(ERROR_INVALID_OPERATION)')
    WIN32__ERROR_MEDIA_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x800710DE, 'HRESULT_FROM_WIN32(ERROR_MEDIA_NOT_AVAILABLE)')
    WIN32__ERROR_DEVICE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x800710DF, 'HRESULT_FROM_WIN32(ERROR_DEVICE_NOT_AVAILABLE)')
    WIN32__ERROR_REQUEST_REFUSED: Final[ThirtyTwoBits] = cached_hresult(0x800710E0, 'HRESULT_FROM_WIN32(ERROR_REQUEST_REFUSED)')
    WIN32__ERROR_INVALID_DRIVE_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x800710E1, 'HRESULT_FROM_WIN32(ERROR_INVALID_DRIVE_OBJECT)')
    WIN32__ERROR_LIBRARY_FULL: Final[ThirtyTwoBits] = cached_hresult(0x800710E2, 'HRESULT_FROM_WIN32(ERROR_LIBRARY_FULL)')
    WIN32__ERROR_MEDIUM_NOT_ACCESSIBLE: Final[ThirtyTwoBits] = cached_hresult(0x800710E3, 'HRESULT_FROM_WIN32(ERROR_MEDIUM_NOT_ACCESSIBLE)')
    WIN32__ERROR_UNABLE_TO_LOAD_MEDIUM: Final[ThirtyTwoBits] = cached_hresult(0x800710E4, 'HRESULT_FROM_WIN32(ERROR_UNABLE_TO_LOAD_MEDIUM)')
    WIN32__ERROR_UNABLE_TO_INVENTORY_DRIVE: Final[ThirtyTwoBits] = cached_hresult(0x800710E5, 'HRESULT_FROM_WIN32(ERROR_UNABLE_TO_INVENTORY_DRIVE)')
    WIN32__ERROR_UNABLE_TO_INVENTORY_SLOT: Final[ThirtyTwoBits] = cached_hresult(0x800710E6, 'HRESULT_FROM_WIN32(ERROR_UNABLE_TO_INVENTORY_SLOT)')
    WIN32__ERROR_UNABLE_TO_INVENTORY_TRANSPORT: Final[ThirtyTwoBits] = cached_hresult(0x800710E7, 'HRESULT_FROM_WIN32(ERROR_UNABLE_TO_INVENTORY_TRANSPORT)')
    WIN32__ERROR_TRANSPORT_FULL: Final[ThirtyTwoBits] = cached_hresult(0x800710E8, 'HRESULT_FROM_WIN32(ERROR_TRANSPORT_FULL)')
    WIN32__ERROR_CONTROLLING_IEPORT: Final[ThirtyTwoBits] = cached_hresult(0x800710E9, 'HRESULT_FROM_WIN32(ERROR_CONTROLLING_IEPORT)')
    WIN32__ERROR_UNABLE_TO_EJECT_MOUNTED_MEDIA: Final[ThirtyTwoBits] = cached_hresult(0x800710EA, 'HRESULT_FROM_WIN32(ERROR_UNABLE_TO_EJECT_MOUNTED_MEDIA)')
    WIN32__ERROR_CLEANER_SLOT_SET: Final[ThirtyTwoBits] = cached_hresult(0x800710EB, 'HRESULT_FROM_WIN32(ERROR_CLEANER_SLOT_SET)')
    WIN32__ERROR_CLEANER_SLOT_NOT_SET: Final[ThirtyTwoBits] = cached_hresult(0x800710EC, 'HRESULT_FROM_WIN32(ERROR_CLEANER_SLOT_NOT_SET)')
    WIN32__ERROR_CLEANER_CARTRIDGE_SPENT: Final[ThirtyTwoBits] = cached_hresult(0x800710ED, 'HRESULT_FROM_WIN32(ERROR_CLEANER_CARTRIDGE_SPENT)')
    WIN32__ERROR_UNEXPECTED_OMID: Final[ThirtyTwoBits] = cached_hresult(0x800710EE, 'HRESULT_FROM_WIN32(ERROR_UNEXPECTED_OMID)')
    WIN32__ERROR_CANT_DELETE_LAST_ITEM: Final[ThirtyTwoBits] = cached_hresult(0x800710EF, 'HRESULT_FROM_WIN32(ERROR_CANT_DELETE_LAST_ITEM)')
    WIN32__ERROR_MESSAGE_EXCEEDS_MAX_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x800710F0, 'HRESULT_FROM_WIN32(ERROR_MESSAGE_EXCEEDS_MAX_SIZE)')
    WIN32__ERROR_VOLUME_CONTAINS_SYS_FILES: Final[ThirtyTwoBits] = cached_hresult(0x800710F1, 'HRESULT_FROM_WIN32(ERROR_VOLUME_CONTAINS_SYS_FILES)')
    WIN32__ERROR_INDIGENOUS_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x800710F2, 'HRESULT_FROM_WIN32(ERROR_INDIGENOUS_TYPE)')
    WIN32__ERROR_NO_SUPPORTING_DRIVES: Final[ThirtyTwoBits] = cached_hresult(0x800710F3, 'HRESULT_FROM_WIN32(ERROR_NO_SUPPORTING_DRIVES)')
    WIN32__ERROR_CLEANER_CARTRIDGE_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x800710F4, 'HRESULT_FROM_WIN32(ERROR_CLEANER_CARTRIDGE_INSTALLED)')
    WIN32__ERROR_IEPORT_FULL: Final[ThirtyTwoBits] = cached_hresult(0x800710F5, 'HRESULT_FROM_WIN32(ERROR_IEPORT_FULL)')
    WIN32__ERROR_FILE_OFFLINE: Final[ThirtyTwoBits] = cached_hresult(0x800710FE, 'HRESULT_FROM_WIN32(ERROR_FILE_OFFLINE)')
    WIN32__ERROR_REMOTE_STORAGE_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0x800710FF, 'HRESULT_FROM_WIN32(ERROR_REMOTE_STORAGE_NOT_ACTIVE)')
    WIN32__ERROR_REMOTE_STORAGE_MEDIA_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80071100, 'HRESULT_FROM_WIN32(ERROR_REMOTE_STORAGE_MEDIA_ERROR)')
    WIN32__ERROR_NOT_A_REPARSE_POINT: Final[ThirtyTwoBits] = cached_hresult(0x80071126, 'HRESULT_FROM_WIN32(ERROR_NOT_A_REPARSE_POINT)')
    WIN32__ERROR_REPARSE_ATTRIBUTE_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x80071127, 'HRESULT_FROM_WIN32(ERROR_REPARSE_ATTRIBUTE_CONFLICT)')
    WIN32__ERROR_INVALID_REPARSE_DATA: Final[ThirtyTwoBits] = cached_hresult(0x80071128, 'HRESULT_FROM_WIN32(ERROR_INVALID_REPARSE_DATA)')
    WIN32__ERROR_REPARSE_TAG_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80071129, 'HRESULT_FROM_WIN32(ERROR_REPARSE_TAG_INVALID)')
    WIN32__ERROR_REPARSE_TAG_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x8007112A, 'HRESULT_FROM_WIN32(ERROR_REPARSE_TAG_MISMATCH)')
    WIN32__ERROR_REPARSE_POINT_ENCOUNTERED: Final[ThirtyTwoBits] = cached_hresult(0x8007112B, 'HRESULT_FROM_WIN32(ERROR_REPARSE_POINT_ENCOUNTERED)')
    WIN32__ERROR_APP_DATA_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80071130, 'HRESULT_FROM_WIN32(ERROR_APP_DATA_NOT_FOUND)')
    WIN32__ERROR_APP_DATA_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x80071131, 'HRESULT_FROM_WIN32(ERROR_APP_DATA_EXPIRED)')
    WIN32__ERROR_APP_DATA_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x80071132, 'HRESULT_FROM_WIN32(ERROR_APP_DATA_CORRUPT)')
    WIN32__ERROR_APP_DATA_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80071133, 'HRESULT_FROM_WIN32(ERROR_APP_DATA_LIMIT_EXCEEDED)')
    WIN32__ERROR_APP_DATA_REBOOT_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80071134, 'HRESULT_FROM_WIN32(ERROR_APP_DATA_REBOOT_REQUIRED)')
    WIN32__ERROR_SECUREBOOT_ROLLBACK_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0x80071144, 'HRESULT_FROM_WIN32(ERROR_SECUREBOOT_ROLLBACK_DETECTED)')
    WIN32__ERROR_SECUREBOOT_POLICY_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x80071145, 'HRESULT_FROM_WIN32(ERROR_SECUREBOOT_POLICY_VIOLATION)')
    WIN32__ERROR_SECUREBOOT_INVALID_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x80071146, 'HRESULT_FROM_WIN32(ERROR_SECUREBOOT_INVALID_POLICY)')
    WIN32__ERROR_SECUREBOOT_POLICY_PUBLISHER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80071147, 'HRESULT_FROM_WIN32(ERROR_SECUREBOOT_POLICY_PUBLISHER_NOT_FOUND)')
    WIN32__ERROR_SECUREBOOT_POLICY_NOT_SIGNED: Final[ThirtyTwoBits] = cached_hresult(0x80071148, 'HRESULT_FROM_WIN32(ERROR_SECUREBOOT_POLICY_NOT_SIGNED)')
    WIN32__ERROR_SECUREBOOT_NOT_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0x80071149, 'HRESULT_FROM_WIN32(ERROR_SECUREBOOT_NOT_ENABLED)')
    WIN32__ERROR_SECUREBOOT_FILE_REPLACED: Final[ThirtyTwoBits] = cached_hresult(0x8007114A, 'HRESULT_FROM_WIN32(ERROR_SECUREBOOT_FILE_REPLACED)')
    WIN32__ERROR_SECUREBOOT_POLICY_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_hresult(0x8007114B, 'HRESULT_FROM_WIN32(ERROR_SECUREBOOT_POLICY_NOT_AUTHORIZED)')
    WIN32__ERROR_SECUREBOOT_POLICY_UNKNOWN: Final[ThirtyTwoBits] = cached_hresult(0x8007114C, 'HRESULT_FROM_WIN32(ERROR_SECUREBOOT_POLICY_UNKNOWN)')
    WIN32__ERROR_SECUREBOOT_POLICY_MISSING_ANTIROLLBACKVERSION: Final[ThirtyTwoBits] = cached_hresult(0x8007114D, 'HRESULT_FROM_WIN32(ERROR_SECUREBOOT_POLICY_MISSING_ANTIROLLBACKVERSION)')
    WIN32__ERROR_SECUREBOOT_PLATFORM_ID_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x8007114E, 'HRESULT_FROM_WIN32(ERROR_SECUREBOOT_PLATFORM_ID_MISMATCH)')
    WIN32__ERROR_SECUREBOOT_POLICY_ROLLBACK_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0x8007114F, 'HRESULT_FROM_WIN32(ERROR_SECUREBOOT_POLICY_ROLLBACK_DETECTED)')
    WIN32__ERROR_SECUREBOOT_POLICY_UPGRADE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80071150, 'HRESULT_FROM_WIN32(ERROR_SECUREBOOT_POLICY_UPGRADE_MISMATCH)')
    WIN32__ERROR_SECUREBOOT_REQUIRED_POLICY_FILE_MISSING: Final[ThirtyTwoBits] = cached_hresult(0x80071151, 'HRESULT_FROM_WIN32(ERROR_SECUREBOOT_REQUIRED_POLICY_FILE_MISSING)')
    WIN32__ERROR_SECUREBOOT_NOT_BASE_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x80071152, 'HRESULT_FROM_WIN32(ERROR_SECUREBOOT_NOT_BASE_POLICY)')
    WIN32__ERROR_SECUREBOOT_NOT_SUPPLEMENTAL_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x80071153, 'HRESULT_FROM_WIN32(ERROR_SECUREBOOT_NOT_SUPPLEMENTAL_POLICY)')
    WIN32__ERROR_OFFLOAD_READ_FLT_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80071158, 'HRESULT_FROM_WIN32(ERROR_OFFLOAD_READ_FLT_NOT_SUPPORTED)')
    WIN32__ERROR_OFFLOAD_WRITE_FLT_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80071159, 'HRESULT_FROM_WIN32(ERROR_OFFLOAD_WRITE_FLT_NOT_SUPPORTED)')
    WIN32__ERROR_OFFLOAD_READ_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8007115A, 'HRESULT_FROM_WIN32(ERROR_OFFLOAD_READ_FILE_NOT_SUPPORTED)')
    WIN32__ERROR_OFFLOAD_WRITE_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8007115B, 'HRESULT_FROM_WIN32(ERROR_OFFLOAD_WRITE_FILE_NOT_SUPPORTED)')
    WIN32__ERROR_ALREADY_HAS_STREAM_ID: Final[ThirtyTwoBits] = cached_hresult(0x8007115C, 'HRESULT_FROM_WIN32(ERROR_ALREADY_HAS_STREAM_ID)')
    WIN32__ERROR_SMR_GARBAGE_COLLECTION_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8007115D, 'HRESULT_FROM_WIN32(ERROR_SMR_GARBAGE_COLLECTION_REQUIRED)')
    WIN32__ERROR_WOF_WIM_HEADER_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x8007115E, 'HRESULT_FROM_WIN32(ERROR_WOF_WIM_HEADER_CORRUPT)')
    WIN32__ERROR_WOF_WIM_RESOURCE_TABLE_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x8007115F, 'HRESULT_FROM_WIN32(ERROR_WOF_WIM_RESOURCE_TABLE_CORRUPT)')
    WIN32__ERROR_WOF_FILE_RESOURCE_TABLE_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x80071160, 'HRESULT_FROM_WIN32(ERROR_WOF_FILE_RESOURCE_TABLE_CORRUPT)')
    WIN32__ERROR_OBJECT_IS_IMMUTABLE: Final[ThirtyTwoBits] = cached_hresult(0x80071161, 'HRESULT_FROM_WIN32(ERROR_OBJECT_IS_IMMUTABLE)')
    WIN32__ERROR_VOLUME_NOT_SIS_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0x80071194, 'HRESULT_FROM_WIN32(ERROR_VOLUME_NOT_SIS_ENABLED)')
    WIN32__ERROR_SYSTEM_INTEGRITY_ROLLBACK_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0x800711C6, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_INTEGRITY_ROLLBACK_DETECTED)')
    WIN32__ERROR_SYSTEM_INTEGRITY_POLICY_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x800711C7, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_INTEGRITY_POLICY_VIOLATION)')
    WIN32__ERROR_SYSTEM_INTEGRITY_INVALID_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x800711C8, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_INTEGRITY_INVALID_POLICY)')
    WIN32__ERROR_SYSTEM_INTEGRITY_POLICY_NOT_SIGNED: Final[ThirtyTwoBits] = cached_hresult(0x800711C9, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_INTEGRITY_POLICY_NOT_SIGNED)')
    WIN32__ERROR_SYSTEM_INTEGRITY_TOO_MANY_POLICIES: Final[ThirtyTwoBits] = cached_hresult(0x800711CA, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_INTEGRITY_TOO_MANY_POLICIES)')
    WIN32__ERROR_SYSTEM_INTEGRITY_SUPPLEMENTAL_POLICY_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_hresult(0x800711CB, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_INTEGRITY_SUPPLEMENTAL_POLICY_NOT_AUTHORIZED)')
    WIN32__ERROR_SYSTEM_INTEGRITY_REPUTATION_MALICIOUS: Final[ThirtyTwoBits] = cached_hresult(0x800711CC, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_INTEGRITY_REPUTATION_MALICIOUS)')
    WIN32__ERROR_SYSTEM_INTEGRITY_REPUTATION_PUA: Final[ThirtyTwoBits] = cached_hresult(0x800711CD, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_INTEGRITY_REPUTATION_PUA)')
    WIN32__ERROR_SYSTEM_INTEGRITY_REPUTATION_DANGEROUS_EXT: Final[ThirtyTwoBits] = cached_hresult(0x800711CE, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_INTEGRITY_REPUTATION_DANGEROUS_EXT)')
    WIN32__ERROR_SYSTEM_INTEGRITY_REPUTATION_OFFLINE: Final[ThirtyTwoBits] = cached_hresult(0x800711CF, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_INTEGRITY_REPUTATION_OFFLINE)')
    WIN32__ERROR_VSM_NOT_INITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0x800711D0, 'HRESULT_FROM_WIN32(ERROR_VSM_NOT_INITIALIZED)')
    WIN32__ERROR_VSM_DMA_PROTECTION_NOT_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x800711D1, 'HRESULT_FROM_WIN32(ERROR_VSM_DMA_PROTECTION_NOT_IN_USE)')
    WIN32__ERROR_PLATFORM_MANIFEST_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_hresult(0x800711DA, 'HRESULT_FROM_WIN32(ERROR_PLATFORM_MANIFEST_NOT_AUTHORIZED)')
    WIN32__ERROR_PLATFORM_MANIFEST_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800711DB, 'HRESULT_FROM_WIN32(ERROR_PLATFORM_MANIFEST_INVALID)')
    WIN32__ERROR_PLATFORM_MANIFEST_FILE_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_hresult(0x800711DC, 'HRESULT_FROM_WIN32(ERROR_PLATFORM_MANIFEST_FILE_NOT_AUTHORIZED)')
    WIN32__ERROR_PLATFORM_MANIFEST_CATALOG_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_hresult(0x800711DD, 'HRESULT_FROM_WIN32(ERROR_PLATFORM_MANIFEST_CATALOG_NOT_AUTHORIZED)')
    WIN32__ERROR_PLATFORM_MANIFEST_BINARY_ID_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800711DE, 'HRESULT_FROM_WIN32(ERROR_PLATFORM_MANIFEST_BINARY_ID_NOT_FOUND)')
    WIN32__ERROR_PLATFORM_MANIFEST_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0x800711DF, 'HRESULT_FROM_WIN32(ERROR_PLATFORM_MANIFEST_NOT_ACTIVE)')
    WIN32__ERROR_PLATFORM_MANIFEST_NOT_SIGNED: Final[ThirtyTwoBits] = cached_hresult(0x800711E0, 'HRESULT_FROM_WIN32(ERROR_PLATFORM_MANIFEST_NOT_SIGNED)')
    WIN32__ERROR_SYSTEM_INTEGRITY_REPUTATION_UNFRIENDLY_FILE: Final[ThirtyTwoBits] = cached_hresult(0x800711E4, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_INTEGRITY_REPUTATION_UNFRIENDLY_FILE)')
    WIN32__ERROR_SYSTEM_INTEGRITY_REPUTATION_UNATTAINABLE: Final[ThirtyTwoBits] = cached_hresult(0x800711E5, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_INTEGRITY_REPUTATION_UNATTAINABLE)')
    WIN32__ERROR_SYSTEM_INTEGRITY_REPUTATION_EXPLICIT_DENY_FILE: Final[ThirtyTwoBits] = cached_hresult(0x800711E6, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_INTEGRITY_REPUTATION_EXPLICIT_DENY_FILE)')
    WIN32__ERROR_DEPENDENT_RESOURCE_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80071389, 'HRESULT_FROM_WIN32(ERROR_DEPENDENT_RESOURCE_EXISTS)')
    WIN32__ERROR_DEPENDENCY_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8007138A, 'HRESULT_FROM_WIN32(ERROR_DEPENDENCY_NOT_FOUND)')
    WIN32__ERROR_DEPENDENCY_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x8007138B, 'HRESULT_FROM_WIN32(ERROR_DEPENDENCY_ALREADY_EXISTS)')
    WIN32__ERROR_RESOURCE_NOT_ONLINE: Final[ThirtyTwoBits] = cached_hresult(0x8007138C, 'HRESULT_FROM_WIN32(ERROR_RESOURCE_NOT_ONLINE)')
    WIN32__ERROR_HOST_NODE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x8007138D, 'HRESULT_FROM_WIN32(ERROR_HOST_NODE_NOT_AVAILABLE)')
    WIN32__ERROR_RESOURCE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x8007138E, 'HRESULT_FROM_WIN32(ERROR_RESOURCE_NOT_AVAILABLE)')
    WIN32__ERROR_RESOURCE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8007138F, 'HRESULT_FROM_WIN32(ERROR_RESOURCE_NOT_FOUND)')
    WIN32__ERROR_SHUTDOWN_CLUSTER: Final[ThirtyTwoBits] = cached_hresult(0x80071390, 'HRESULT_FROM_WIN32(ERROR_SHUTDOWN_CLUSTER)')
    WIN32__ERROR_CANT_EVICT_ACTIVE_NODE: Final[ThirtyTwoBits] = cached_hresult(0x80071391, 'HRESULT_FROM_WIN32(ERROR_CANT_EVICT_ACTIVE_NODE)')
    WIN32__ERROR_OBJECT_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80071392, 'HRESULT_FROM_WIN32(ERROR_OBJECT_ALREADY_EXISTS)')
    WIN32__ERROR_OBJECT_IN_LIST: Final[ThirtyTwoBits] = cached_hresult(0x80071393, 'HRESULT_FROM_WIN32(ERROR_OBJECT_IN_LIST)')
    WIN32__ERROR_GROUP_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80071394, 'HRESULT_FROM_WIN32(ERROR_GROUP_NOT_AVAILABLE)')
    WIN32__ERROR_GROUP_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80071395, 'HRESULT_FROM_WIN32(ERROR_GROUP_NOT_FOUND)')
    WIN32__ERROR_GROUP_NOT_ONLINE: Final[ThirtyTwoBits] = cached_hresult(0x80071396, 'HRESULT_FROM_WIN32(ERROR_GROUP_NOT_ONLINE)')
    WIN32__ERROR_HOST_NODE_NOT_RESOURCE_OWNER: Final[ThirtyTwoBits] = cached_hresult(0x80071397, 'HRESULT_FROM_WIN32(ERROR_HOST_NODE_NOT_RESOURCE_OWNER)')
    WIN32__ERROR_HOST_NODE_NOT_GROUP_OWNER: Final[ThirtyTwoBits] = cached_hresult(0x80071398, 'HRESULT_FROM_WIN32(ERROR_HOST_NODE_NOT_GROUP_OWNER)')
    WIN32__ERROR_RESMON_CREATE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80071399, 'HRESULT_FROM_WIN32(ERROR_RESMON_CREATE_FAILED)')
    WIN32__ERROR_RESMON_ONLINE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8007139A, 'HRESULT_FROM_WIN32(ERROR_RESMON_ONLINE_FAILED)')
    WIN32__ERROR_RESOURCE_ONLINE: Final[ThirtyTwoBits] = cached_hresult(0x8007139B, 'HRESULT_FROM_WIN32(ERROR_RESOURCE_ONLINE)')
    WIN32__ERROR_QUORUM_RESOURCE: Final[ThirtyTwoBits] = cached_hresult(0x8007139C, 'HRESULT_FROM_WIN32(ERROR_QUORUM_RESOURCE)')
    WIN32__ERROR_NOT_QUORUM_CAPABLE: Final[ThirtyTwoBits] = cached_hresult(0x8007139D, 'HRESULT_FROM_WIN32(ERROR_NOT_QUORUM_CAPABLE)')
    WIN32__ERROR_CLUSTER_SHUTTING_DOWN: Final[ThirtyTwoBits] = cached_hresult(0x8007139E, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_SHUTTING_DOWN)')
    WIN32__ERROR_INVALID_STATE: Final[ThirtyTwoBits] = cached_hresult(0x8007139F, 'HRESULT_FROM_WIN32(ERROR_INVALID_STATE)')
    WIN32__ERROR_RESOURCE_PROPERTIES_STORED: Final[ThirtyTwoBits] = cached_hresult(0x800713A0, 'HRESULT_FROM_WIN32(ERROR_RESOURCE_PROPERTIES_STORED)')
    WIN32__ERROR_NOT_QUORUM_CLASS: Final[ThirtyTwoBits] = cached_hresult(0x800713A1, 'HRESULT_FROM_WIN32(ERROR_NOT_QUORUM_CLASS)')
    WIN32__ERROR_CORE_RESOURCE: Final[ThirtyTwoBits] = cached_hresult(0x800713A2, 'HRESULT_FROM_WIN32(ERROR_CORE_RESOURCE)')
    WIN32__ERROR_QUORUM_RESOURCE_ONLINE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800713A3, 'HRESULT_FROM_WIN32(ERROR_QUORUM_RESOURCE_ONLINE_FAILED)')
    WIN32__ERROR_QUORUMLOG_OPEN_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800713A4, 'HRESULT_FROM_WIN32(ERROR_QUORUMLOG_OPEN_FAILED)')
    WIN32__ERROR_CLUSTERLOG_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x800713A5, 'HRESULT_FROM_WIN32(ERROR_CLUSTERLOG_CORRUPT)')
    WIN32__ERROR_CLUSTERLOG_RECORD_EXCEEDS_MAXSIZE: Final[ThirtyTwoBits] = cached_hresult(0x800713A6, 'HRESULT_FROM_WIN32(ERROR_CLUSTERLOG_RECORD_EXCEEDS_MAXSIZE)')
    WIN32__ERROR_CLUSTERLOG_EXCEEDS_MAXSIZE: Final[ThirtyTwoBits] = cached_hresult(0x800713A7, 'HRESULT_FROM_WIN32(ERROR_CLUSTERLOG_EXCEEDS_MAXSIZE)')
    WIN32__ERROR_CLUSTERLOG_CHKPOINT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800713A8, 'HRESULT_FROM_WIN32(ERROR_CLUSTERLOG_CHKPOINT_NOT_FOUND)')
    WIN32__ERROR_CLUSTERLOG_NOT_ENOUGH_SPACE: Final[ThirtyTwoBits] = cached_hresult(0x800713A9, 'HRESULT_FROM_WIN32(ERROR_CLUSTERLOG_NOT_ENOUGH_SPACE)')
    WIN32__ERROR_QUORUM_OWNER_ALIVE: Final[ThirtyTwoBits] = cached_hresult(0x800713AA, 'HRESULT_FROM_WIN32(ERROR_QUORUM_OWNER_ALIVE)')
    WIN32__ERROR_NETWORK_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x800713AB, 'HRESULT_FROM_WIN32(ERROR_NETWORK_NOT_AVAILABLE)')
    WIN32__ERROR_NODE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x800713AC, 'HRESULT_FROM_WIN32(ERROR_NODE_NOT_AVAILABLE)')
    WIN32__ERROR_ALL_NODES_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x800713AD, 'HRESULT_FROM_WIN32(ERROR_ALL_NODES_NOT_AVAILABLE)')
    WIN32__ERROR_RESOURCE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800713AE, 'HRESULT_FROM_WIN32(ERROR_RESOURCE_FAILED)')
    WIN32__ERROR_CLUSTER_INVALID_NODE: Final[ThirtyTwoBits] = cached_hresult(0x800713AF, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_INVALID_NODE)')
    WIN32__ERROR_CLUSTER_NODE_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800713B0, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NODE_EXISTS)')
    WIN32__ERROR_CLUSTER_JOIN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x800713B1, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_JOIN_IN_PROGRESS)')
    WIN32__ERROR_CLUSTER_NODE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800713B2, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NODE_NOT_FOUND)')
    WIN32__ERROR_CLUSTER_LOCAL_NODE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800713B3, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_LOCAL_NODE_NOT_FOUND)')
    WIN32__ERROR_CLUSTER_NETWORK_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800713B4, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NETWORK_EXISTS)')
    WIN32__ERROR_CLUSTER_NETWORK_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800713B5, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NETWORK_NOT_FOUND)')
    WIN32__ERROR_CLUSTER_NETINTERFACE_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800713B6, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NETINTERFACE_EXISTS)')
    WIN32__ERROR_CLUSTER_NETINTERFACE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800713B7, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NETINTERFACE_NOT_FOUND)')
    WIN32__ERROR_CLUSTER_INVALID_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0x800713B8, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_INVALID_REQUEST)')
    WIN32__ERROR_CLUSTER_INVALID_NETWORK_PROVIDER: Final[ThirtyTwoBits] = cached_hresult(0x800713B9, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_INVALID_NETWORK_PROVIDER)')
    WIN32__ERROR_CLUSTER_NODE_DOWN: Final[ThirtyTwoBits] = cached_hresult(0x800713BA, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NODE_DOWN)')
    WIN32__ERROR_CLUSTER_NODE_UNREACHABLE: Final[ThirtyTwoBits] = cached_hresult(0x800713BB, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NODE_UNREACHABLE)')
    WIN32__ERROR_CLUSTER_NODE_NOT_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0x800713BC, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NODE_NOT_MEMBER)')
    WIN32__ERROR_CLUSTER_JOIN_NOT_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x800713BD, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_JOIN_NOT_IN_PROGRESS)')
    WIN32__ERROR_CLUSTER_INVALID_NETWORK: Final[ThirtyTwoBits] = cached_hresult(0x800713BE, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_INVALID_NETWORK)')
    WIN32__ERROR_CLUSTER_NODE_UP: Final[ThirtyTwoBits] = cached_hresult(0x800713C0, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NODE_UP)')
    WIN32__ERROR_CLUSTER_IPADDR_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x800713C1, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_IPADDR_IN_USE)')
    WIN32__ERROR_CLUSTER_NODE_NOT_PAUSED: Final[ThirtyTwoBits] = cached_hresult(0x800713C2, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NODE_NOT_PAUSED)')
    WIN32__ERROR_CLUSTER_NO_SECURITY_CONTEXT: Final[ThirtyTwoBits] = cached_hresult(0x800713C3, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NO_SECURITY_CONTEXT)')
    WIN32__ERROR_CLUSTER_NETWORK_NOT_INTERNAL: Final[ThirtyTwoBits] = cached_hresult(0x800713C4, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NETWORK_NOT_INTERNAL)')
    WIN32__ERROR_CLUSTER_NODE_ALREADY_UP: Final[ThirtyTwoBits] = cached_hresult(0x800713C5, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NODE_ALREADY_UP)')
    WIN32__ERROR_CLUSTER_NODE_ALREADY_DOWN: Final[ThirtyTwoBits] = cached_hresult(0x800713C6, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NODE_ALREADY_DOWN)')
    WIN32__ERROR_CLUSTER_NETWORK_ALREADY_ONLINE: Final[ThirtyTwoBits] = cached_hresult(0x800713C7, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NETWORK_ALREADY_ONLINE)')
    WIN32__ERROR_CLUSTER_NETWORK_ALREADY_OFFLINE: Final[ThirtyTwoBits] = cached_hresult(0x800713C8, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NETWORK_ALREADY_OFFLINE)')
    WIN32__ERROR_CLUSTER_NODE_ALREADY_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0x800713C9, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NODE_ALREADY_MEMBER)')
    WIN32__ERROR_CLUSTER_LAST_INTERNAL_NETWORK: Final[ThirtyTwoBits] = cached_hresult(0x800713CA, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_LAST_INTERNAL_NETWORK)')
    WIN32__ERROR_CLUSTER_NETWORK_HAS_DEPENDENTS: Final[ThirtyTwoBits] = cached_hresult(0x800713CB, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NETWORK_HAS_DEPENDENTS)')
    WIN32__ERROR_INVALID_OPERATION_ON_QUORUM: Final[ThirtyTwoBits] = cached_hresult(0x800713CC, 'HRESULT_FROM_WIN32(ERROR_INVALID_OPERATION_ON_QUORUM)')
    WIN32__ERROR_DEPENDENCY_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x800713CD, 'HRESULT_FROM_WIN32(ERROR_DEPENDENCY_NOT_ALLOWED)')
    WIN32__ERROR_CLUSTER_NODE_PAUSED: Final[ThirtyTwoBits] = cached_hresult(0x800713CE, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NODE_PAUSED)')
    WIN32__ERROR_NODE_CANT_HOST_RESOURCE: Final[ThirtyTwoBits] = cached_hresult(0x800713CF, 'HRESULT_FROM_WIN32(ERROR_NODE_CANT_HOST_RESOURCE)')
    WIN32__ERROR_CLUSTER_NODE_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0x800713D0, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NODE_NOT_READY)')
    WIN32__ERROR_CLUSTER_NODE_SHUTTING_DOWN: Final[ThirtyTwoBits] = cached_hresult(0x800713D1, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NODE_SHUTTING_DOWN)')
    WIN32__ERROR_CLUSTER_JOIN_ABORTED: Final[ThirtyTwoBits] = cached_hresult(0x800713D2, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_JOIN_ABORTED)')
    WIN32__ERROR_CLUSTER_INCOMPATIBLE_VERSIONS: Final[ThirtyTwoBits] = cached_hresult(0x800713D3, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_INCOMPATIBLE_VERSIONS)')
    WIN32__ERROR_CLUSTER_MAXNUM_OF_RESOURCES_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x800713D4, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_MAXNUM_OF_RESOURCES_EXCEEDED)')
    WIN32__ERROR_CLUSTER_SYSTEM_CONFIG_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0x800713D5, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_SYSTEM_CONFIG_CHANGED)')
    WIN32__ERROR_CLUSTER_RESOURCE_TYPE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800713D6, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_RESOURCE_TYPE_NOT_FOUND)')
    WIN32__ERROR_CLUSTER_RESTYPE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x800713D7, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_RESTYPE_NOT_SUPPORTED)')
    WIN32__ERROR_CLUSTER_RESNAME_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800713D8, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_RESNAME_NOT_FOUND)')
    WIN32__ERROR_CLUSTER_NO_RPC_PACKAGES_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x800713D9, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NO_RPC_PACKAGES_REGISTERED)')
    WIN32__ERROR_CLUSTER_OWNER_NOT_IN_PREFLIST: Final[ThirtyTwoBits] = cached_hresult(0x800713DA, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_OWNER_NOT_IN_PREFLIST)')
    WIN32__ERROR_CLUSTER_DATABASE_SEQMISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x800713DB, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_DATABASE_SEQMISMATCH)')
    WIN32__ERROR_RESMON_INVALID_STATE: Final[ThirtyTwoBits] = cached_hresult(0x800713DC, 'HRESULT_FROM_WIN32(ERROR_RESMON_INVALID_STATE)')
    WIN32__ERROR_CLUSTER_GUM_NOT_LOCKER: Final[ThirtyTwoBits] = cached_hresult(0x800713DD, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_GUM_NOT_LOCKER)')
    WIN32__ERROR_QUORUM_DISK_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800713DE, 'HRESULT_FROM_WIN32(ERROR_QUORUM_DISK_NOT_FOUND)')
    WIN32__ERROR_DATABASE_BACKUP_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x800713DF, 'HRESULT_FROM_WIN32(ERROR_DATABASE_BACKUP_CORRUPT)')
    WIN32__ERROR_CLUSTER_NODE_ALREADY_HAS_DFS_ROOT: Final[ThirtyTwoBits] = cached_hresult(0x800713E0, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NODE_ALREADY_HAS_DFS_ROOT)')
    WIN32__ERROR_RESOURCE_PROPERTY_UNCHANGEABLE: Final[ThirtyTwoBits] = cached_hresult(0x800713E1, 'HRESULT_FROM_WIN32(ERROR_RESOURCE_PROPERTY_UNCHANGEABLE)')
    WIN32__ERROR_NO_ADMIN_ACCESS_POINT: Final[ThirtyTwoBits] = cached_hresult(0x800713E2, 'HRESULT_FROM_WIN32(ERROR_NO_ADMIN_ACCESS_POINT)')
    WIN32__ERROR_CLUSTER_MEMBERSHIP_INVALID_STATE: Final[ThirtyTwoBits] = cached_hresult(0x80071702, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_MEMBERSHIP_INVALID_STATE)')
    WIN32__ERROR_CLUSTER_QUORUMLOG_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80071703, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_QUORUMLOG_NOT_FOUND)')
    WIN32__ERROR_CLUSTER_MEMBERSHIP_HALT: Final[ThirtyTwoBits] = cached_hresult(0x80071704, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_MEMBERSHIP_HALT)')
    WIN32__ERROR_CLUSTER_INSTANCE_ID_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80071705, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_INSTANCE_ID_MISMATCH)')
    WIN32__ERROR_CLUSTER_NETWORK_NOT_FOUND_FOR_IP: Final[ThirtyTwoBits] = cached_hresult(0x80071706, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NETWORK_NOT_FOUND_FOR_IP)')
    WIN32__ERROR_CLUSTER_PROPERTY_DATA_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80071707, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_PROPERTY_DATA_TYPE_MISMATCH)')
    WIN32__ERROR_CLUSTER_EVICT_WITHOUT_CLEANUP: Final[ThirtyTwoBits] = cached_hresult(0x80071708, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_EVICT_WITHOUT_CLEANUP)')
    WIN32__ERROR_CLUSTER_PARAMETER_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80071709, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_PARAMETER_MISMATCH)')
    WIN32__ERROR_NODE_CANNOT_BE_CLUSTERED: Final[ThirtyTwoBits] = cached_hresult(0x8007170A, 'HRESULT_FROM_WIN32(ERROR_NODE_CANNOT_BE_CLUSTERED)')
    WIN32__ERROR_CLUSTER_WRONG_OS_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x8007170B, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_WRONG_OS_VERSION)')
    WIN32__ERROR_CLUSTER_CANT_CREATE_DUP_CLUSTER_NAME: Final[ThirtyTwoBits] = cached_hresult(0x8007170C, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_CANT_CREATE_DUP_CLUSTER_NAME)')
    WIN32__ERROR_CLUSCFG_ALREADY_COMMITTED: Final[ThirtyTwoBits] = cached_hresult(0x8007170D, 'HRESULT_FROM_WIN32(ERROR_CLUSCFG_ALREADY_COMMITTED)')
    WIN32__ERROR_CLUSCFG_ROLLBACK_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8007170E, 'HRESULT_FROM_WIN32(ERROR_CLUSCFG_ROLLBACK_FAILED)')
    WIN32__ERROR_CLUSCFG_SYSTEM_DISK_DRIVE_LETTER_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x8007170F, 'HRESULT_FROM_WIN32(ERROR_CLUSCFG_SYSTEM_DISK_DRIVE_LETTER_CONFLICT)')
    WIN32__ERROR_CLUSTER_OLD_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80071710, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_OLD_VERSION)')
    WIN32__ERROR_CLUSTER_MISMATCHED_COMPUTER_ACCT_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80071711, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_MISMATCHED_COMPUTER_ACCT_NAME)')
    WIN32__ERROR_CLUSTER_NO_NET_ADAPTERS: Final[ThirtyTwoBits] = cached_hresult(0x80071712, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NO_NET_ADAPTERS)')
    WIN32__ERROR_CLUSTER_POISONED: Final[ThirtyTwoBits] = cached_hresult(0x80071713, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_POISONED)')
    WIN32__ERROR_CLUSTER_GROUP_MOVING: Final[ThirtyTwoBits] = cached_hresult(0x80071714, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_GROUP_MOVING)')
    WIN32__ERROR_CLUSTER_RESOURCE_TYPE_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x80071715, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_RESOURCE_TYPE_BUSY)')
    WIN32__ERROR_RESOURCE_CALL_TIMED_OUT: Final[ThirtyTwoBits] = cached_hresult(0x80071716, 'HRESULT_FROM_WIN32(ERROR_RESOURCE_CALL_TIMED_OUT)')
    WIN32__ERROR_INVALID_CLUSTER_IPV6_ADDRESS: Final[ThirtyTwoBits] = cached_hresult(0x80071717, 'HRESULT_FROM_WIN32(ERROR_INVALID_CLUSTER_IPV6_ADDRESS)')
    WIN32__ERROR_CLUSTER_INTERNAL_INVALID_FUNCTION: Final[ThirtyTwoBits] = cached_hresult(0x80071718, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_INTERNAL_INVALID_FUNCTION)')
    WIN32__ERROR_CLUSTER_PARAMETER_OUT_OF_BOUNDS: Final[ThirtyTwoBits] = cached_hresult(0x80071719, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_PARAMETER_OUT_OF_BOUNDS)')
    WIN32__ERROR_CLUSTER_PARTIAL_SEND: Final[ThirtyTwoBits] = cached_hresult(0x8007171A, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_PARTIAL_SEND)')
    WIN32__ERROR_CLUSTER_REGISTRY_INVALID_FUNCTION: Final[ThirtyTwoBits] = cached_hresult(0x8007171B, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_REGISTRY_INVALID_FUNCTION)')
    WIN32__ERROR_CLUSTER_INVALID_STRING_TERMINATION: Final[ThirtyTwoBits] = cached_hresult(0x8007171C, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_INVALID_STRING_TERMINATION)')
    WIN32__ERROR_CLUSTER_INVALID_STRING_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0x8007171D, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_INVALID_STRING_FORMAT)')
    WIN32__ERROR_CLUSTER_DATABASE_TRANSACTION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x8007171E, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_DATABASE_TRANSACTION_IN_PROGRESS)')
    WIN32__ERROR_CLUSTER_DATABASE_TRANSACTION_NOT_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x8007171F, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_DATABASE_TRANSACTION_NOT_IN_PROGRESS)')
    WIN32__ERROR_CLUSTER_NULL_DATA: Final[ThirtyTwoBits] = cached_hresult(0x80071720, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NULL_DATA)')
    WIN32__ERROR_CLUSTER_PARTIAL_READ: Final[ThirtyTwoBits] = cached_hresult(0x80071721, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_PARTIAL_READ)')
    WIN32__ERROR_CLUSTER_PARTIAL_WRITE: Final[ThirtyTwoBits] = cached_hresult(0x80071722, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_PARTIAL_WRITE)')
    WIN32__ERROR_CLUSTER_CANT_DESERIALIZE_DATA: Final[ThirtyTwoBits] = cached_hresult(0x80071723, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_CANT_DESERIALIZE_DATA)')
    WIN32__ERROR_DEPENDENT_RESOURCE_PROPERTY_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x80071724, 'HRESULT_FROM_WIN32(ERROR_DEPENDENT_RESOURCE_PROPERTY_CONFLICT)')
    WIN32__ERROR_CLUSTER_NO_QUORUM: Final[ThirtyTwoBits] = cached_hresult(0x80071725, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NO_QUORUM)')
    WIN32__ERROR_CLUSTER_INVALID_IPV6_NETWORK: Final[ThirtyTwoBits] = cached_hresult(0x80071726, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_INVALID_IPV6_NETWORK)')
    WIN32__ERROR_CLUSTER_INVALID_IPV6_TUNNEL_NETWORK: Final[ThirtyTwoBits] = cached_hresult(0x80071727, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_INVALID_IPV6_TUNNEL_NETWORK)')
    WIN32__ERROR_QUORUM_NOT_ALLOWED_IN_THIS_GROUP: Final[ThirtyTwoBits] = cached_hresult(0x80071728, 'HRESULT_FROM_WIN32(ERROR_QUORUM_NOT_ALLOWED_IN_THIS_GROUP)')
    WIN32__ERROR_DEPENDENCY_TREE_TOO_COMPLEX: Final[ThirtyTwoBits] = cached_hresult(0x80071729, 'HRESULT_FROM_WIN32(ERROR_DEPENDENCY_TREE_TOO_COMPLEX)')
    WIN32__ERROR_EXCEPTION_IN_RESOURCE_CALL: Final[ThirtyTwoBits] = cached_hresult(0x8007172A, 'HRESULT_FROM_WIN32(ERROR_EXCEPTION_IN_RESOURCE_CALL)')
    WIN32__ERROR_CLUSTER_RHS_FAILED_INITIALIZATION: Final[ThirtyTwoBits] = cached_hresult(0x8007172B, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_RHS_FAILED_INITIALIZATION)')
    WIN32__ERROR_CLUSTER_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x8007172C, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NOT_INSTALLED)')
    WIN32__ERROR_CLUSTER_RESOURCES_MUST_BE_ONLINE_ON_THE_SAME_NODE: Final[ThirtyTwoBits] = cached_hresult(0x8007172D, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_RESOURCES_MUST_BE_ONLINE_ON_THE_SAME_NODE)')
    WIN32__ERROR_CLUSTER_MAX_NODES_IN_CLUSTER: Final[ThirtyTwoBits] = cached_hresult(0x8007172E, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_MAX_NODES_IN_CLUSTER)')
    WIN32__ERROR_CLUSTER_TOO_MANY_NODES: Final[ThirtyTwoBits] = cached_hresult(0x8007172F, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_TOO_MANY_NODES)')
    WIN32__ERROR_CLUSTER_OBJECT_ALREADY_USED: Final[ThirtyTwoBits] = cached_hresult(0x80071730, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_OBJECT_ALREADY_USED)')
    WIN32__ERROR_NONCORE_GROUPS_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80071731, 'HRESULT_FROM_WIN32(ERROR_NONCORE_GROUPS_FOUND)')
    WIN32__ERROR_FILE_SHARE_RESOURCE_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x80071732, 'HRESULT_FROM_WIN32(ERROR_FILE_SHARE_RESOURCE_CONFLICT)')
    WIN32__ERROR_CLUSTER_EVICT_INVALID_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0x80071733, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_EVICT_INVALID_REQUEST)')
    WIN32__ERROR_CLUSTER_SINGLETON_RESOURCE: Final[ThirtyTwoBits] = cached_hresult(0x80071734, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_SINGLETON_RESOURCE)')
    WIN32__ERROR_CLUSTER_GROUP_SINGLETON_RESOURCE: Final[ThirtyTwoBits] = cached_hresult(0x80071735, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_GROUP_SINGLETON_RESOURCE)')
    WIN32__ERROR_CLUSTER_RESOURCE_PROVIDER_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80071736, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_RESOURCE_PROVIDER_FAILED)')
    WIN32__ERROR_CLUSTER_RESOURCE_CONFIGURATION_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80071737, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_RESOURCE_CONFIGURATION_ERROR)')
    WIN32__ERROR_CLUSTER_GROUP_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x80071738, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_GROUP_BUSY)')
    WIN32__ERROR_CLUSTER_NOT_SHARED_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0x80071739, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NOT_SHARED_VOLUME)')
    WIN32__ERROR_CLUSTER_INVALID_SECURITY_DESCRIPTOR: Final[ThirtyTwoBits] = cached_hresult(0x8007173A, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_INVALID_SECURITY_DESCRIPTOR)')
    WIN32__ERROR_CLUSTER_SHARED_VOLUMES_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x8007173B, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_SHARED_VOLUMES_IN_USE)')
    WIN32__ERROR_CLUSTER_USE_SHARED_VOLUMES_API: Final[ThirtyTwoBits] = cached_hresult(0x8007173C, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_USE_SHARED_VOLUMES_API)')
    WIN32__ERROR_CLUSTER_BACKUP_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x8007173D, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_BACKUP_IN_PROGRESS)')
    WIN32__ERROR_NON_CSV_PATH: Final[ThirtyTwoBits] = cached_hresult(0x8007173E, 'HRESULT_FROM_WIN32(ERROR_NON_CSV_PATH)')
    WIN32__ERROR_CSV_VOLUME_NOT_LOCAL: Final[ThirtyTwoBits] = cached_hresult(0x8007173F, 'HRESULT_FROM_WIN32(ERROR_CSV_VOLUME_NOT_LOCAL)')
    WIN32__ERROR_CLUSTER_WATCHDOG_TERMINATING: Final[ThirtyTwoBits] = cached_hresult(0x80071740, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_WATCHDOG_TERMINATING)')
    WIN32__ERROR_CLUSTER_RESOURCE_VETOED_MOVE_INCOMPATIBLE_NODES: Final[ThirtyTwoBits] = cached_hresult(0x80071741, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_RESOURCE_VETOED_MOVE_INCOMPATIBLE_NODES)')
    WIN32__ERROR_CLUSTER_INVALID_NODE_WEIGHT: Final[ThirtyTwoBits] = cached_hresult(0x80071742, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_INVALID_NODE_WEIGHT)')
    WIN32__ERROR_CLUSTER_RESOURCE_VETOED_CALL: Final[ThirtyTwoBits] = cached_hresult(0x80071743, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_RESOURCE_VETOED_CALL)')
    WIN32__ERROR_RESMON_SYSTEM_RESOURCES_LACKING: Final[ThirtyTwoBits] = cached_hresult(0x80071744, 'HRESULT_FROM_WIN32(ERROR_RESMON_SYSTEM_RESOURCES_LACKING)')
    WIN32__ERROR_CLUSTER_RESOURCE_VETOED_MOVE_NOT_ENOUGH_RESOURCES_ON_DESTINATION: Final[ThirtyTwoBits] = cached_hresult(0x80071745, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_RESOURCE_VETOED_MOVE_NOT_ENOUGH_RESOURCES_ON_DESTINATION)')
    WIN32__ERROR_CLUSTER_RESOURCE_VETOED_MOVE_NOT_ENOUGH_RESOURCES_ON_SOURCE: Final[ThirtyTwoBits] = cached_hresult(0x80071746, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_RESOURCE_VETOED_MOVE_NOT_ENOUGH_RESOURCES_ON_SOURCE)')
    WIN32__ERROR_CLUSTER_GROUP_QUEUED: Final[ThirtyTwoBits] = cached_hresult(0x80071747, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_GROUP_QUEUED)')
    WIN32__ERROR_CLUSTER_RESOURCE_LOCKED_STATUS: Final[ThirtyTwoBits] = cached_hresult(0x80071748, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_RESOURCE_LOCKED_STATUS)')
    WIN32__ERROR_CLUSTER_SHARED_VOLUME_FAILOVER_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80071749, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_SHARED_VOLUME_FAILOVER_NOT_ALLOWED)')
    WIN32__ERROR_CLUSTER_NODE_DRAIN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x8007174A, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NODE_DRAIN_IN_PROGRESS)')
    WIN32__ERROR_CLUSTER_DISK_NOT_CONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x8007174B, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_DISK_NOT_CONNECTED)')
    WIN32__ERROR_DISK_NOT_CSV_CAPABLE: Final[ThirtyTwoBits] = cached_hresult(0x8007174C, 'HRESULT_FROM_WIN32(ERROR_DISK_NOT_CSV_CAPABLE)')
    WIN32__ERROR_RESOURCE_NOT_IN_AVAILABLE_STORAGE: Final[ThirtyTwoBits] = cached_hresult(0x8007174D, 'HRESULT_FROM_WIN32(ERROR_RESOURCE_NOT_IN_AVAILABLE_STORAGE)')
    WIN32__ERROR_CLUSTER_SHARED_VOLUME_REDIRECTED: Final[ThirtyTwoBits] = cached_hresult(0x8007174E, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_SHARED_VOLUME_REDIRECTED)')
    WIN32__ERROR_CLUSTER_SHARED_VOLUME_NOT_REDIRECTED: Final[ThirtyTwoBits] = cached_hresult(0x8007174F, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_SHARED_VOLUME_NOT_REDIRECTED)')
    WIN32__ERROR_CLUSTER_CANNOT_RETURN_PROPERTIES: Final[ThirtyTwoBits] = cached_hresult(0x80071750, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_CANNOT_RETURN_PROPERTIES)')
    WIN32__ERROR_CLUSTER_RESOURCE_CONTAINS_UNSUPPORTED_DIFF_AREA_FOR_SHARED_VOLUMES: Final[ThirtyTwoBits] = cached_hresult(0x80071751, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_RESOURCE_CONTAINS_UNSUPPORTED_DIFF_AREA_FOR_SHARED_VOLUMES)')
    WIN32__ERROR_CLUSTER_RESOURCE_IS_IN_MAINTENANCE_MODE: Final[ThirtyTwoBits] = cached_hresult(0x80071752, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_RESOURCE_IS_IN_MAINTENANCE_MODE)')
    WIN32__ERROR_CLUSTER_AFFINITY_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x80071753, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_AFFINITY_CONFLICT)')
    WIN32__ERROR_CLUSTER_RESOURCE_IS_REPLICA_VIRTUAL_MACHINE: Final[ThirtyTwoBits] = cached_hresult(0x80071754, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_RESOURCE_IS_REPLICA_VIRTUAL_MACHINE)')
    WIN32__ERROR_CLUSTER_UPGRADE_INCOMPATIBLE_VERSIONS: Final[ThirtyTwoBits] = cached_hresult(0x80071755, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_UPGRADE_INCOMPATIBLE_VERSIONS)')
    WIN32__ERROR_CLUSTER_UPGRADE_FIX_QUORUM_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80071756, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_UPGRADE_FIX_QUORUM_NOT_SUPPORTED)')
    WIN32__ERROR_CLUSTER_UPGRADE_RESTART_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80071757, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_UPGRADE_RESTART_REQUIRED)')
    WIN32__ERROR_CLUSTER_UPGRADE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x80071758, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_UPGRADE_IN_PROGRESS)')
    WIN32__ERROR_CLUSTER_UPGRADE_INCOMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x80071759, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_UPGRADE_INCOMPLETE)')
    WIN32__ERROR_CLUSTER_NODE_IN_GRACE_PERIOD: Final[ThirtyTwoBits] = cached_hresult(0x8007175A, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NODE_IN_GRACE_PERIOD)')
    WIN32__ERROR_CLUSTER_CSV_IO_PAUSE_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x8007175B, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_CSV_IO_PAUSE_TIMEOUT)')
    WIN32__ERROR_NODE_NOT_ACTIVE_CLUSTER_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0x8007175C, 'HRESULT_FROM_WIN32(ERROR_NODE_NOT_ACTIVE_CLUSTER_MEMBER)')
    WIN32__ERROR_CLUSTER_RESOURCE_NOT_MONITORED: Final[ThirtyTwoBits] = cached_hresult(0x8007175D, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_RESOURCE_NOT_MONITORED)')
    WIN32__ERROR_CLUSTER_RESOURCE_DOES_NOT_SUPPORT_UNMONITORED: Final[ThirtyTwoBits] = cached_hresult(0x8007175E, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_RESOURCE_DOES_NOT_SUPPORT_UNMONITORED)')
    WIN32__ERROR_CLUSTER_RESOURCE_IS_REPLICATED: Final[ThirtyTwoBits] = cached_hresult(0x8007175F, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_RESOURCE_IS_REPLICATED)')
    WIN32__ERROR_CLUSTER_NODE_ISOLATED: Final[ThirtyTwoBits] = cached_hresult(0x80071760, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NODE_ISOLATED)')
    WIN32__ERROR_CLUSTER_NODE_QUARANTINED: Final[ThirtyTwoBits] = cached_hresult(0x80071761, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_NODE_QUARANTINED)')
    WIN32__ERROR_CLUSTER_DATABASE_UPDATE_CONDITION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80071762, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_DATABASE_UPDATE_CONDITION_FAILED)')
    WIN32__ERROR_CLUSTER_SPACE_DEGRADED: Final[ThirtyTwoBits] = cached_hresult(0x80071763, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_SPACE_DEGRADED)')
    WIN32__ERROR_CLUSTER_TOKEN_DELEGATION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80071764, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_TOKEN_DELEGATION_NOT_SUPPORTED)')
    WIN32__ERROR_CLUSTER_CSV_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80071765, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_CSV_INVALID_HANDLE)')
    WIN32__ERROR_CLUSTER_CSV_SUPPORTED_ONLY_ON_COORDINATOR: Final[ThirtyTwoBits] = cached_hresult(0x80071766, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_CSV_SUPPORTED_ONLY_ON_COORDINATOR)')
    WIN32__ERROR_GROUPSET_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80071767, 'HRESULT_FROM_WIN32(ERROR_GROUPSET_NOT_AVAILABLE)')
    WIN32__ERROR_GROUPSET_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80071768, 'HRESULT_FROM_WIN32(ERROR_GROUPSET_NOT_FOUND)')
    WIN32__ERROR_GROUPSET_CANT_PROVIDE: Final[ThirtyTwoBits] = cached_hresult(0x80071769, 'HRESULT_FROM_WIN32(ERROR_GROUPSET_CANT_PROVIDE)')
    WIN32__ERROR_CLUSTER_FAULT_DOMAIN_PARENT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8007176A, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_FAULT_DOMAIN_PARENT_NOT_FOUND)')
    WIN32__ERROR_CLUSTER_FAULT_DOMAIN_INVALID_HIERARCHY: Final[ThirtyTwoBits] = cached_hresult(0x8007176B, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_FAULT_DOMAIN_INVALID_HIERARCHY)')
    WIN32__ERROR_CLUSTER_FAULT_DOMAIN_FAILED_S2D_VALIDATION: Final[ThirtyTwoBits] = cached_hresult(0x8007176C, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_FAULT_DOMAIN_FAILED_S2D_VALIDATION)')
    WIN32__ERROR_CLUSTER_FAULT_DOMAIN_S2D_CONNECTIVITY_LOSS: Final[ThirtyTwoBits] = cached_hresult(0x8007176D, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_FAULT_DOMAIN_S2D_CONNECTIVITY_LOSS)')
    WIN32__ERROR_CLUSTER_INVALID_INFRASTRUCTURE_FILESERVER_NAME: Final[ThirtyTwoBits] = cached_hresult(0x8007176E, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_INVALID_INFRASTRUCTURE_FILESERVER_NAME)')
    WIN32__ERROR_CLUSTERSET_MANAGEMENT_CLUSTER_UNREACHABLE: Final[ThirtyTwoBits] = cached_hresult(0x8007176F, 'HRESULT_FROM_WIN32(ERROR_CLUSTERSET_MANAGEMENT_CLUSTER_UNREACHABLE)')
    WIN32__ERROR_ENCRYPTION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80071770, 'HRESULT_FROM_WIN32(ERROR_ENCRYPTION_FAILED)')
    WIN32__ERROR_DECRYPTION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80071771, 'HRESULT_FROM_WIN32(ERROR_DECRYPTION_FAILED)')
    WIN32__ERROR_FILE_ENCRYPTED: Final[ThirtyTwoBits] = cached_hresult(0x80071772, 'HRESULT_FROM_WIN32(ERROR_FILE_ENCRYPTED)')
    WIN32__ERROR_NO_RECOVERY_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x80071773, 'HRESULT_FROM_WIN32(ERROR_NO_RECOVERY_POLICY)')
    WIN32__ERROR_NO_EFS: Final[ThirtyTwoBits] = cached_hresult(0x80071774, 'HRESULT_FROM_WIN32(ERROR_NO_EFS)')
    WIN32__ERROR_WRONG_EFS: Final[ThirtyTwoBits] = cached_hresult(0x80071775, 'HRESULT_FROM_WIN32(ERROR_WRONG_EFS)')
    WIN32__ERROR_NO_USER_KEYS: Final[ThirtyTwoBits] = cached_hresult(0x80071776, 'HRESULT_FROM_WIN32(ERROR_NO_USER_KEYS)')
    WIN32__ERROR_FILE_NOT_ENCRYPTED: Final[ThirtyTwoBits] = cached_hresult(0x80071777, 'HRESULT_FROM_WIN32(ERROR_FILE_NOT_ENCRYPTED)')
    WIN32__ERROR_NOT_EXPORT_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0x80071778, 'HRESULT_FROM_WIN32(ERROR_NOT_EXPORT_FORMAT)')
    WIN32__ERROR_FILE_READ_ONLY: Final[ThirtyTwoBits] = cached_hresult(0x80071779, 'HRESULT_FROM_WIN32(ERROR_FILE_READ_ONLY)')
    WIN32__ERROR_DIR_EFS_DISALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x8007177A, 'HRESULT_FROM_WIN32(ERROR_DIR_EFS_DISALLOWED)')
    WIN32__ERROR_EFS_SERVER_NOT_TRUSTED: Final[ThirtyTwoBits] = cached_hresult(0x8007177B, 'HRESULT_FROM_WIN32(ERROR_EFS_SERVER_NOT_TRUSTED)')
    WIN32__ERROR_BAD_RECOVERY_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x8007177C, 'HRESULT_FROM_WIN32(ERROR_BAD_RECOVERY_POLICY)')
    WIN32__ERROR_EFS_ALG_BLOB_TOO_BIG: Final[ThirtyTwoBits] = cached_hresult(0x8007177D, 'HRESULT_FROM_WIN32(ERROR_EFS_ALG_BLOB_TOO_BIG)')
    WIN32__ERROR_VOLUME_NOT_SUPPORT_EFS: Final[ThirtyTwoBits] = cached_hresult(0x8007177E, 'HRESULT_FROM_WIN32(ERROR_VOLUME_NOT_SUPPORT_EFS)')
    WIN32__ERROR_EFS_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x8007177F, 'HRESULT_FROM_WIN32(ERROR_EFS_DISABLED)')
    WIN32__ERROR_EFS_VERSION_NOT_SUPPORT: Final[ThirtyTwoBits] = cached_hresult(0x80071780, 'HRESULT_FROM_WIN32(ERROR_EFS_VERSION_NOT_SUPPORT)')
    WIN32__ERROR_CS_ENCRYPTION_INVALID_SERVER_RESPONSE: Final[ThirtyTwoBits] = cached_hresult(0x80071781, 'HRESULT_FROM_WIN32(ERROR_CS_ENCRYPTION_INVALID_SERVER_RESPONSE)')
    WIN32__ERROR_CS_ENCRYPTION_UNSUPPORTED_SERVER: Final[ThirtyTwoBits] = cached_hresult(0x80071782, 'HRESULT_FROM_WIN32(ERROR_CS_ENCRYPTION_UNSUPPORTED_SERVER)')
    WIN32__ERROR_CS_ENCRYPTION_EXISTING_ENCRYPTED_FILE: Final[ThirtyTwoBits] = cached_hresult(0x80071783, 'HRESULT_FROM_WIN32(ERROR_CS_ENCRYPTION_EXISTING_ENCRYPTED_FILE)')
    WIN32__ERROR_CS_ENCRYPTION_NEW_ENCRYPTED_FILE: Final[ThirtyTwoBits] = cached_hresult(0x80071784, 'HRESULT_FROM_WIN32(ERROR_CS_ENCRYPTION_NEW_ENCRYPTED_FILE)')
    WIN32__ERROR_CS_ENCRYPTION_FILE_NOT_CSE: Final[ThirtyTwoBits] = cached_hresult(0x80071785, 'HRESULT_FROM_WIN32(ERROR_CS_ENCRYPTION_FILE_NOT_CSE)')
    WIN32__ERROR_ENCRYPTION_POLICY_DENIES_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x80071786, 'HRESULT_FROM_WIN32(ERROR_ENCRYPTION_POLICY_DENIES_OPERATION)')
    WIN32__ERROR_WIP_ENCRYPTION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80071787, 'HRESULT_FROM_WIN32(ERROR_WIP_ENCRYPTION_FAILED)')
    WIN32__ERROR_NO_BROWSER_SERVERS_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800717E6, 'HRESULT_FROM_WIN32(ERROR_NO_BROWSER_SERVERS_FOUND)')
    WIN32__SCHED_E_SERVICE_NOT_LOCALSYSTEM: Final[ThirtyTwoBits] = cached_hresult(0x80071838, 'HRESULT_FROM_WIN32(SCHED_E_SERVICE_NOT_LOCALSYSTEM)')
    WIN32__ERROR_CLUSTER_OBJECT_IS_CLUSTER_SET_VM: Final[ThirtyTwoBits] = cached_hresult(0x8007186A, 'HRESULT_FROM_WIN32(ERROR_CLUSTER_OBJECT_IS_CLUSTER_SET_VM)')
    WIN32__ERROR_LOG_SECTOR_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800719C8, 'HRESULT_FROM_WIN32(ERROR_LOG_SECTOR_INVALID)')
    WIN32__ERROR_LOG_SECTOR_PARITY_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800719C9, 'HRESULT_FROM_WIN32(ERROR_LOG_SECTOR_PARITY_INVALID)')
    WIN32__ERROR_LOG_SECTOR_REMAPPED: Final[ThirtyTwoBits] = cached_hresult(0x800719CA, 'HRESULT_FROM_WIN32(ERROR_LOG_SECTOR_REMAPPED)')
    WIN32__ERROR_LOG_BLOCK_INCOMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x800719CB, 'HRESULT_FROM_WIN32(ERROR_LOG_BLOCK_INCOMPLETE)')
    WIN32__ERROR_LOG_INVALID_RANGE: Final[ThirtyTwoBits] = cached_hresult(0x800719CC, 'HRESULT_FROM_WIN32(ERROR_LOG_INVALID_RANGE)')
    WIN32__ERROR_LOG_BLOCKS_EXHAUSTED: Final[ThirtyTwoBits] = cached_hresult(0x800719CD, 'HRESULT_FROM_WIN32(ERROR_LOG_BLOCKS_EXHAUSTED)')
    WIN32__ERROR_LOG_READ_CONTEXT_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800719CE, 'HRESULT_FROM_WIN32(ERROR_LOG_READ_CONTEXT_INVALID)')
    WIN32__ERROR_LOG_RESTART_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800719CF, 'HRESULT_FROM_WIN32(ERROR_LOG_RESTART_INVALID)')
    WIN32__ERROR_LOG_BLOCK_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x800719D0, 'HRESULT_FROM_WIN32(ERROR_LOG_BLOCK_VERSION)')
    WIN32__ERROR_LOG_BLOCK_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800719D1, 'HRESULT_FROM_WIN32(ERROR_LOG_BLOCK_INVALID)')
    WIN32__ERROR_LOG_READ_MODE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800719D2, 'HRESULT_FROM_WIN32(ERROR_LOG_READ_MODE_INVALID)')
    WIN32__ERROR_LOG_NO_RESTART: Final[ThirtyTwoBits] = cached_hresult(0x800719D3, 'HRESULT_FROM_WIN32(ERROR_LOG_NO_RESTART)')
    WIN32__ERROR_LOG_METADATA_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x800719D4, 'HRESULT_FROM_WIN32(ERROR_LOG_METADATA_CORRUPT)')
    WIN32__ERROR_LOG_METADATA_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800719D5, 'HRESULT_FROM_WIN32(ERROR_LOG_METADATA_INVALID)')
    WIN32__ERROR_LOG_METADATA_INCONSISTENT: Final[ThirtyTwoBits] = cached_hresult(0x800719D6, 'HRESULT_FROM_WIN32(ERROR_LOG_METADATA_INCONSISTENT)')
    WIN32__ERROR_LOG_RESERVATION_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800719D7, 'HRESULT_FROM_WIN32(ERROR_LOG_RESERVATION_INVALID)')
    WIN32__ERROR_LOG_CANT_DELETE: Final[ThirtyTwoBits] = cached_hresult(0x800719D8, 'HRESULT_FROM_WIN32(ERROR_LOG_CANT_DELETE)')
    WIN32__ERROR_LOG_CONTAINER_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x800719D9, 'HRESULT_FROM_WIN32(ERROR_LOG_CONTAINER_LIMIT_EXCEEDED)')
    WIN32__ERROR_LOG_START_OF_LOG: Final[ThirtyTwoBits] = cached_hresult(0x800719DA, 'HRESULT_FROM_WIN32(ERROR_LOG_START_OF_LOG)')
    WIN32__ERROR_LOG_POLICY_ALREADY_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x800719DB, 'HRESULT_FROM_WIN32(ERROR_LOG_POLICY_ALREADY_INSTALLED)')
    WIN32__ERROR_LOG_POLICY_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x800719DC, 'HRESULT_FROM_WIN32(ERROR_LOG_POLICY_NOT_INSTALLED)')
    WIN32__ERROR_LOG_POLICY_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800719DD, 'HRESULT_FROM_WIN32(ERROR_LOG_POLICY_INVALID)')
    WIN32__ERROR_LOG_POLICY_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x800719DE, 'HRESULT_FROM_WIN32(ERROR_LOG_POLICY_CONFLICT)')
    WIN32__ERROR_LOG_PINNED_ARCHIVE_TAIL: Final[ThirtyTwoBits] = cached_hresult(0x800719DF, 'HRESULT_FROM_WIN32(ERROR_LOG_PINNED_ARCHIVE_TAIL)')
    WIN32__ERROR_LOG_RECORD_NONEXISTENT: Final[ThirtyTwoBits] = cached_hresult(0x800719E0, 'HRESULT_FROM_WIN32(ERROR_LOG_RECORD_NONEXISTENT)')
    WIN32__ERROR_LOG_RECORDS_RESERVED_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800719E1, 'HRESULT_FROM_WIN32(ERROR_LOG_RECORDS_RESERVED_INVALID)')
    WIN32__ERROR_LOG_SPACE_RESERVED_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800719E2, 'HRESULT_FROM_WIN32(ERROR_LOG_SPACE_RESERVED_INVALID)')
    WIN32__ERROR_LOG_TAIL_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800719E3, 'HRESULT_FROM_WIN32(ERROR_LOG_TAIL_INVALID)')
    WIN32__ERROR_LOG_FULL: Final[ThirtyTwoBits] = cached_hresult(0x800719E4, 'HRESULT_FROM_WIN32(ERROR_LOG_FULL)')
    WIN32__ERROR_COULD_NOT_RESIZE_LOG: Final[ThirtyTwoBits] = cached_hresult(0x800719E5, 'HRESULT_FROM_WIN32(ERROR_COULD_NOT_RESIZE_LOG)')
    WIN32__ERROR_LOG_MULTIPLEXED: Final[ThirtyTwoBits] = cached_hresult(0x800719E6, 'HRESULT_FROM_WIN32(ERROR_LOG_MULTIPLEXED)')
    WIN32__ERROR_LOG_DEDICATED: Final[ThirtyTwoBits] = cached_hresult(0x800719E7, 'HRESULT_FROM_WIN32(ERROR_LOG_DEDICATED)')
    WIN32__ERROR_LOG_ARCHIVE_NOT_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x800719E8, 'HRESULT_FROM_WIN32(ERROR_LOG_ARCHIVE_NOT_IN_PROGRESS)')
    WIN32__ERROR_LOG_ARCHIVE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x800719E9, 'HRESULT_FROM_WIN32(ERROR_LOG_ARCHIVE_IN_PROGRESS)')
    WIN32__ERROR_LOG_EPHEMERAL: Final[ThirtyTwoBits] = cached_hresult(0x800719EA, 'HRESULT_FROM_WIN32(ERROR_LOG_EPHEMERAL)')
    WIN32__ERROR_LOG_NOT_ENOUGH_CONTAINERS: Final[ThirtyTwoBits] = cached_hresult(0x800719EB, 'HRESULT_FROM_WIN32(ERROR_LOG_NOT_ENOUGH_CONTAINERS)')
    WIN32__ERROR_LOG_CLIENT_ALREADY_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x800719EC, 'HRESULT_FROM_WIN32(ERROR_LOG_CLIENT_ALREADY_REGISTERED)')
    WIN32__ERROR_LOG_CLIENT_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x800719ED, 'HRESULT_FROM_WIN32(ERROR_LOG_CLIENT_NOT_REGISTERED)')
    WIN32__ERROR_LOG_FULL_HANDLER_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x800719EE, 'HRESULT_FROM_WIN32(ERROR_LOG_FULL_HANDLER_IN_PROGRESS)')
    WIN32__ERROR_LOG_CONTAINER_READ_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800719EF, 'HRESULT_FROM_WIN32(ERROR_LOG_CONTAINER_READ_FAILED)')
    WIN32__ERROR_LOG_CONTAINER_WRITE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800719F0, 'HRESULT_FROM_WIN32(ERROR_LOG_CONTAINER_WRITE_FAILED)')
    WIN32__ERROR_LOG_CONTAINER_OPEN_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800719F1, 'HRESULT_FROM_WIN32(ERROR_LOG_CONTAINER_OPEN_FAILED)')
    WIN32__ERROR_LOG_CONTAINER_STATE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800719F2, 'HRESULT_FROM_WIN32(ERROR_LOG_CONTAINER_STATE_INVALID)')
    WIN32__ERROR_LOG_STATE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800719F3, 'HRESULT_FROM_WIN32(ERROR_LOG_STATE_INVALID)')
    WIN32__ERROR_LOG_PINNED: Final[ThirtyTwoBits] = cached_hresult(0x800719F4, 'HRESULT_FROM_WIN32(ERROR_LOG_PINNED)')
    WIN32__ERROR_LOG_METADATA_FLUSH_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800719F5, 'HRESULT_FROM_WIN32(ERROR_LOG_METADATA_FLUSH_FAILED)')
    WIN32__ERROR_LOG_INCONSISTENT_SECURITY: Final[ThirtyTwoBits] = cached_hresult(0x800719F6, 'HRESULT_FROM_WIN32(ERROR_LOG_INCONSISTENT_SECURITY)')
    WIN32__ERROR_LOG_APPENDED_FLUSH_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800719F7, 'HRESULT_FROM_WIN32(ERROR_LOG_APPENDED_FLUSH_FAILED)')
    WIN32__ERROR_LOG_PINNED_RESERVATION: Final[ThirtyTwoBits] = cached_hresult(0x800719F8, 'HRESULT_FROM_WIN32(ERROR_LOG_PINNED_RESERVATION)')
    WIN32__ERROR_INVALID_TRANSACTION: Final[ThirtyTwoBits] = cached_hresult(0x80071A2C, 'HRESULT_FROM_WIN32(ERROR_INVALID_TRANSACTION)')
    WIN32__ERROR_TRANSACTION_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0x80071A2D, 'HRESULT_FROM_WIN32(ERROR_TRANSACTION_NOT_ACTIVE)')
    WIN32__ERROR_TRANSACTION_REQUEST_NOT_VALID: Final[ThirtyTwoBits] = cached_hresult(0x80071A2E, 'HRESULT_FROM_WIN32(ERROR_TRANSACTION_REQUEST_NOT_VALID)')
    WIN32__ERROR_TRANSACTION_NOT_REQUESTED: Final[ThirtyTwoBits] = cached_hresult(0x80071A2F, 'HRESULT_FROM_WIN32(ERROR_TRANSACTION_NOT_REQUESTED)')
    WIN32__ERROR_TRANSACTION_ALREADY_ABORTED: Final[ThirtyTwoBits] = cached_hresult(0x80071A30, 'HRESULT_FROM_WIN32(ERROR_TRANSACTION_ALREADY_ABORTED)')
    WIN32__ERROR_TRANSACTION_ALREADY_COMMITTED: Final[ThirtyTwoBits] = cached_hresult(0x80071A31, 'HRESULT_FROM_WIN32(ERROR_TRANSACTION_ALREADY_COMMITTED)')
    WIN32__ERROR_TM_INITIALIZATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80071A32, 'HRESULT_FROM_WIN32(ERROR_TM_INITIALIZATION_FAILED)')
    WIN32__ERROR_RESOURCEMANAGER_READ_ONLY: Final[ThirtyTwoBits] = cached_hresult(0x80071A33, 'HRESULT_FROM_WIN32(ERROR_RESOURCEMANAGER_READ_ONLY)')
    WIN32__ERROR_TRANSACTION_NOT_JOINED: Final[ThirtyTwoBits] = cached_hresult(0x80071A34, 'HRESULT_FROM_WIN32(ERROR_TRANSACTION_NOT_JOINED)')
    WIN32__ERROR_TRANSACTION_SUPERIOR_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80071A35, 'HRESULT_FROM_WIN32(ERROR_TRANSACTION_SUPERIOR_EXISTS)')
    WIN32__ERROR_CRM_PROTOCOL_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80071A36, 'HRESULT_FROM_WIN32(ERROR_CRM_PROTOCOL_ALREADY_EXISTS)')
    WIN32__ERROR_TRANSACTION_PROPAGATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80071A37, 'HRESULT_FROM_WIN32(ERROR_TRANSACTION_PROPAGATION_FAILED)')
    WIN32__ERROR_CRM_PROTOCOL_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80071A38, 'HRESULT_FROM_WIN32(ERROR_CRM_PROTOCOL_NOT_FOUND)')
    WIN32__ERROR_TRANSACTION_INVALID_MARSHALL_BUFFER: Final[ThirtyTwoBits] = cached_hresult(0x80071A39, 'HRESULT_FROM_WIN32(ERROR_TRANSACTION_INVALID_MARSHALL_BUFFER)')
    WIN32__ERROR_CURRENT_TRANSACTION_NOT_VALID: Final[ThirtyTwoBits] = cached_hresult(0x80071A3A, 'HRESULT_FROM_WIN32(ERROR_CURRENT_TRANSACTION_NOT_VALID)')
    WIN32__ERROR_TRANSACTION_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80071A3B, 'HRESULT_FROM_WIN32(ERROR_TRANSACTION_NOT_FOUND)')
    WIN32__ERROR_RESOURCEMANAGER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80071A3C, 'HRESULT_FROM_WIN32(ERROR_RESOURCEMANAGER_NOT_FOUND)')
    WIN32__ERROR_ENLISTMENT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80071A3D, 'HRESULT_FROM_WIN32(ERROR_ENLISTMENT_NOT_FOUND)')
    WIN32__ERROR_TRANSACTIONMANAGER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80071A3E, 'HRESULT_FROM_WIN32(ERROR_TRANSACTIONMANAGER_NOT_FOUND)')
    WIN32__ERROR_TRANSACTIONMANAGER_NOT_ONLINE: Final[ThirtyTwoBits] = cached_hresult(0x80071A3F, 'HRESULT_FROM_WIN32(ERROR_TRANSACTIONMANAGER_NOT_ONLINE)')
    WIN32__ERROR_TRANSACTIONMANAGER_RECOVERY_NAME_COLLISION: Final[ThirtyTwoBits] = cached_hresult(0x80071A40, 'HRESULT_FROM_WIN32(ERROR_TRANSACTIONMANAGER_RECOVERY_NAME_COLLISION)')
    WIN32__ERROR_TRANSACTION_NOT_ROOT: Final[ThirtyTwoBits] = cached_hresult(0x80071A41, 'HRESULT_FROM_WIN32(ERROR_TRANSACTION_NOT_ROOT)')
    WIN32__ERROR_TRANSACTION_OBJECT_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x80071A42, 'HRESULT_FROM_WIN32(ERROR_TRANSACTION_OBJECT_EXPIRED)')
    WIN32__ERROR_TRANSACTION_RESPONSE_NOT_ENLISTED: Final[ThirtyTwoBits] = cached_hresult(0x80071A43, 'HRESULT_FROM_WIN32(ERROR_TRANSACTION_RESPONSE_NOT_ENLISTED)')
    WIN32__ERROR_TRANSACTION_RECORD_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0x80071A44, 'HRESULT_FROM_WIN32(ERROR_TRANSACTION_RECORD_TOO_LONG)')
    WIN32__ERROR_IMPLICIT_TRANSACTION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80071A45, 'HRESULT_FROM_WIN32(ERROR_IMPLICIT_TRANSACTION_NOT_SUPPORTED)')
    WIN32__ERROR_TRANSACTION_INTEGRITY_VIOLATED: Final[ThirtyTwoBits] = cached_hresult(0x80071A46, 'HRESULT_FROM_WIN32(ERROR_TRANSACTION_INTEGRITY_VIOLATED)')
    WIN32__ERROR_TRANSACTIONMANAGER_IDENTITY_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80071A47, 'HRESULT_FROM_WIN32(ERROR_TRANSACTIONMANAGER_IDENTITY_MISMATCH)')
    WIN32__ERROR_RM_CANNOT_BE_FROZEN_FOR_SNAPSHOT: Final[ThirtyTwoBits] = cached_hresult(0x80071A48, 'HRESULT_FROM_WIN32(ERROR_RM_CANNOT_BE_FROZEN_FOR_SNAPSHOT)')
    WIN32__ERROR_TRANSACTION_MUST_WRITETHROUGH: Final[ThirtyTwoBits] = cached_hresult(0x80071A49, 'HRESULT_FROM_WIN32(ERROR_TRANSACTION_MUST_WRITETHROUGH)')
    WIN32__ERROR_TRANSACTION_NO_SUPERIOR: Final[ThirtyTwoBits] = cached_hresult(0x80071A4A, 'HRESULT_FROM_WIN32(ERROR_TRANSACTION_NO_SUPERIOR)')
    WIN32__ERROR_HEURISTIC_DAMAGE_POSSIBLE: Final[ThirtyTwoBits] = cached_hresult(0x80071A4B, 'HRESULT_FROM_WIN32(ERROR_HEURISTIC_DAMAGE_POSSIBLE)')
    WIN32__ERROR_TRANSACTIONAL_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x80071A90, 'HRESULT_FROM_WIN32(ERROR_TRANSACTIONAL_CONFLICT)')
    WIN32__ERROR_RM_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0x80071A91, 'HRESULT_FROM_WIN32(ERROR_RM_NOT_ACTIVE)')
    WIN32__ERROR_RM_METADATA_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x80071A92, 'HRESULT_FROM_WIN32(ERROR_RM_METADATA_CORRUPT)')
    WIN32__ERROR_DIRECTORY_NOT_RM: Final[ThirtyTwoBits] = cached_hresult(0x80071A93, 'HRESULT_FROM_WIN32(ERROR_DIRECTORY_NOT_RM)')
    WIN32__ERROR_TRANSACTIONS_UNSUPPORTED_REMOTE: Final[ThirtyTwoBits] = cached_hresult(0x80071A95, 'HRESULT_FROM_WIN32(ERROR_TRANSACTIONS_UNSUPPORTED_REMOTE)')
    WIN32__ERROR_LOG_RESIZE_INVALID_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x80071A96, 'HRESULT_FROM_WIN32(ERROR_LOG_RESIZE_INVALID_SIZE)')
    WIN32__ERROR_OBJECT_NO_LONGER_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80071A97, 'HRESULT_FROM_WIN32(ERROR_OBJECT_NO_LONGER_EXISTS)')
    WIN32__ERROR_STREAM_MINIVERSION_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80071A98, 'HRESULT_FROM_WIN32(ERROR_STREAM_MINIVERSION_NOT_FOUND)')
    WIN32__ERROR_STREAM_MINIVERSION_NOT_VALID: Final[ThirtyTwoBits] = cached_hresult(0x80071A99, 'HRESULT_FROM_WIN32(ERROR_STREAM_MINIVERSION_NOT_VALID)')
    WIN32__ERROR_MINIVERSION_INACCESSIBLE_FROM_SPECIFIED_TRANSACTION: Final[ThirtyTwoBits] = cached_hresult(0x80071A9A, 'HRESULT_FROM_WIN32(ERROR_MINIVERSION_INACCESSIBLE_FROM_SPECIFIED_TRANSACTION)')
    WIN32__ERROR_CANT_OPEN_MINIVERSION_WITH_MODIFY_INTENT: Final[ThirtyTwoBits] = cached_hresult(0x80071A9B, 'HRESULT_FROM_WIN32(ERROR_CANT_OPEN_MINIVERSION_WITH_MODIFY_INTENT)')
    WIN32__ERROR_CANT_CREATE_MORE_STREAM_MINIVERSIONS: Final[ThirtyTwoBits] = cached_hresult(0x80071A9C, 'HRESULT_FROM_WIN32(ERROR_CANT_CREATE_MORE_STREAM_MINIVERSIONS)')
    WIN32__ERROR_REMOTE_FILE_VERSION_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80071A9E, 'HRESULT_FROM_WIN32(ERROR_REMOTE_FILE_VERSION_MISMATCH)')
    WIN32__ERROR_HANDLE_NO_LONGER_VALID: Final[ThirtyTwoBits] = cached_hresult(0x80071A9F, 'HRESULT_FROM_WIN32(ERROR_HANDLE_NO_LONGER_VALID)')
    WIN32__ERROR_NO_TXF_METADATA: Final[ThirtyTwoBits] = cached_hresult(0x80071AA0, 'HRESULT_FROM_WIN32(ERROR_NO_TXF_METADATA)')
    WIN32__ERROR_LOG_CORRUPTION_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0x80071AA1, 'HRESULT_FROM_WIN32(ERROR_LOG_CORRUPTION_DETECTED)')
    WIN32__ERROR_CANT_RECOVER_WITH_HANDLE_OPEN: Final[ThirtyTwoBits] = cached_hresult(0x80071AA2, 'HRESULT_FROM_WIN32(ERROR_CANT_RECOVER_WITH_HANDLE_OPEN)')
    WIN32__ERROR_RM_DISCONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x80071AA3, 'HRESULT_FROM_WIN32(ERROR_RM_DISCONNECTED)')
    WIN32__ERROR_ENLISTMENT_NOT_SUPERIOR: Final[ThirtyTwoBits] = cached_hresult(0x80071AA4, 'HRESULT_FROM_WIN32(ERROR_ENLISTMENT_NOT_SUPERIOR)')
    WIN32__ERROR_RECOVERY_NOT_NEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80071AA5, 'HRESULT_FROM_WIN32(ERROR_RECOVERY_NOT_NEEDED)')
    WIN32__ERROR_RM_ALREADY_STARTED: Final[ThirtyTwoBits] = cached_hresult(0x80071AA6, 'HRESULT_FROM_WIN32(ERROR_RM_ALREADY_STARTED)')
    WIN32__ERROR_FILE_IDENTITY_NOT_PERSISTENT: Final[ThirtyTwoBits] = cached_hresult(0x80071AA7, 'HRESULT_FROM_WIN32(ERROR_FILE_IDENTITY_NOT_PERSISTENT)')
    WIN32__ERROR_CANT_BREAK_TRANSACTIONAL_DEPENDENCY: Final[ThirtyTwoBits] = cached_hresult(0x80071AA8, 'HRESULT_FROM_WIN32(ERROR_CANT_BREAK_TRANSACTIONAL_DEPENDENCY)')
    WIN32__ERROR_CANT_CROSS_RM_BOUNDARY: Final[ThirtyTwoBits] = cached_hresult(0x80071AA9, 'HRESULT_FROM_WIN32(ERROR_CANT_CROSS_RM_BOUNDARY)')
    WIN32__ERROR_TXF_DIR_NOT_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0x80071AAA, 'HRESULT_FROM_WIN32(ERROR_TXF_DIR_NOT_EMPTY)')
    WIN32__ERROR_INDOUBT_TRANSACTIONS_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x80071AAB, 'HRESULT_FROM_WIN32(ERROR_INDOUBT_TRANSACTIONS_EXIST)')
    WIN32__ERROR_TM_VOLATILE: Final[ThirtyTwoBits] = cached_hresult(0x80071AAC, 'HRESULT_FROM_WIN32(ERROR_TM_VOLATILE)')
    WIN32__ERROR_ROLLBACK_TIMER_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x80071AAD, 'HRESULT_FROM_WIN32(ERROR_ROLLBACK_TIMER_EXPIRED)')
    WIN32__ERROR_TXF_ATTRIBUTE_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x80071AAE, 'HRESULT_FROM_WIN32(ERROR_TXF_ATTRIBUTE_CORRUPT)')
    WIN32__ERROR_EFS_NOT_ALLOWED_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_hresult(0x80071AAF, 'HRESULT_FROM_WIN32(ERROR_EFS_NOT_ALLOWED_IN_TRANSACTION)')
    WIN32__ERROR_TRANSACTIONAL_OPEN_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80071AB0, 'HRESULT_FROM_WIN32(ERROR_TRANSACTIONAL_OPEN_NOT_ALLOWED)')
    WIN32__ERROR_LOG_GROWTH_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80071AB1, 'HRESULT_FROM_WIN32(ERROR_LOG_GROWTH_FAILED)')
    WIN32__ERROR_TRANSACTED_MAPPING_UNSUPPORTED_REMOTE: Final[ThirtyTwoBits] = cached_hresult(0x80071AB2, 'HRESULT_FROM_WIN32(ERROR_TRANSACTED_MAPPING_UNSUPPORTED_REMOTE)')
    WIN32__ERROR_TXF_METADATA_ALREADY_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0x80071AB3, 'HRESULT_FROM_WIN32(ERROR_TXF_METADATA_ALREADY_PRESENT)')
    WIN32__ERROR_TRANSACTION_SCOPE_CALLBACKS_NOT_SET: Final[ThirtyTwoBits] = cached_hresult(0x80071AB4, 'HRESULT_FROM_WIN32(ERROR_TRANSACTION_SCOPE_CALLBACKS_NOT_SET)')
    WIN32__ERROR_TRANSACTION_REQUIRED_PROMOTION: Final[ThirtyTwoBits] = cached_hresult(0x80071AB5, 'HRESULT_FROM_WIN32(ERROR_TRANSACTION_REQUIRED_PROMOTION)')
    WIN32__ERROR_CANNOT_EXECUTE_FILE_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_hresult(0x80071AB6, 'HRESULT_FROM_WIN32(ERROR_CANNOT_EXECUTE_FILE_IN_TRANSACTION)')
    WIN32__ERROR_TRANSACTIONS_NOT_FROZEN: Final[ThirtyTwoBits] = cached_hresult(0x80071AB7, 'HRESULT_FROM_WIN32(ERROR_TRANSACTIONS_NOT_FROZEN)')
    WIN32__ERROR_TRANSACTION_FREEZE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x80071AB8, 'HRESULT_FROM_WIN32(ERROR_TRANSACTION_FREEZE_IN_PROGRESS)')
    WIN32__ERROR_NOT_SNAPSHOT_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0x80071AB9, 'HRESULT_FROM_WIN32(ERROR_NOT_SNAPSHOT_VOLUME)')
    WIN32__ERROR_NO_SAVEPOINT_WITH_OPEN_FILES: Final[ThirtyTwoBits] = cached_hresult(0x80071ABA, 'HRESULT_FROM_WIN32(ERROR_NO_SAVEPOINT_WITH_OPEN_FILES)')
    WIN32__ERROR_DATA_LOST_REPAIR: Final[ThirtyTwoBits] = cached_hresult(0x80071ABB, 'HRESULT_FROM_WIN32(ERROR_DATA_LOST_REPAIR)')
    WIN32__ERROR_SPARSE_NOT_ALLOWED_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_hresult(0x80071ABC, 'HRESULT_FROM_WIN32(ERROR_SPARSE_NOT_ALLOWED_IN_TRANSACTION)')
    WIN32__ERROR_TM_IDENTITY_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80071ABD, 'HRESULT_FROM_WIN32(ERROR_TM_IDENTITY_MISMATCH)')
    WIN32__ERROR_FLOATED_SECTION: Final[ThirtyTwoBits] = cached_hresult(0x80071ABE, 'HRESULT_FROM_WIN32(ERROR_FLOATED_SECTION)')
    WIN32__ERROR_CANNOT_ACCEPT_TRANSACTED_WORK: Final[ThirtyTwoBits] = cached_hresult(0x80071ABF, 'HRESULT_FROM_WIN32(ERROR_CANNOT_ACCEPT_TRANSACTED_WORK)')
    WIN32__ERROR_CANNOT_ABORT_TRANSACTIONS: Final[ThirtyTwoBits] = cached_hresult(0x80071AC0, 'HRESULT_FROM_WIN32(ERROR_CANNOT_ABORT_TRANSACTIONS)')
    WIN32__ERROR_BAD_CLUSTERS: Final[ThirtyTwoBits] = cached_hresult(0x80071AC1, 'HRESULT_FROM_WIN32(ERROR_BAD_CLUSTERS)')
    WIN32__ERROR_COMPRESSION_NOT_ALLOWED_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_hresult(0x80071AC2, 'HRESULT_FROM_WIN32(ERROR_COMPRESSION_NOT_ALLOWED_IN_TRANSACTION)')
    WIN32__ERROR_VOLUME_DIRTY: Final[ThirtyTwoBits] = cached_hresult(0x80071AC3, 'HRESULT_FROM_WIN32(ERROR_VOLUME_DIRTY)')
    WIN32__ERROR_NO_LINK_TRACKING_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_hresult(0x80071AC4, 'HRESULT_FROM_WIN32(ERROR_NO_LINK_TRACKING_IN_TRANSACTION)')
    WIN32__ERROR_OPERATION_NOT_SUPPORTED_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_hresult(0x80071AC5, 'HRESULT_FROM_WIN32(ERROR_OPERATION_NOT_SUPPORTED_IN_TRANSACTION)')
    WIN32__ERROR_EXPIRED_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x80071AC6, 'HRESULT_FROM_WIN32(ERROR_EXPIRED_HANDLE)')
    WIN32__ERROR_TRANSACTION_NOT_ENLISTED: Final[ThirtyTwoBits] = cached_hresult(0x80071AC7, 'HRESULT_FROM_WIN32(ERROR_TRANSACTION_NOT_ENLISTED)')
    WIN32__ERROR_CTX_WINSTATION_NAME_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80071B59, 'HRESULT_FROM_WIN32(ERROR_CTX_WINSTATION_NAME_INVALID)')
    WIN32__ERROR_CTX_INVALID_PD: Final[ThirtyTwoBits] = cached_hresult(0x80071B5A, 'HRESULT_FROM_WIN32(ERROR_CTX_INVALID_PD)')
    WIN32__ERROR_CTX_PD_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80071B5B, 'HRESULT_FROM_WIN32(ERROR_CTX_PD_NOT_FOUND)')
    WIN32__ERROR_CTX_WD_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80071B5C, 'HRESULT_FROM_WIN32(ERROR_CTX_WD_NOT_FOUND)')
    WIN32__ERROR_CTX_CANNOT_MAKE_EVENTLOG_ENTRY: Final[ThirtyTwoBits] = cached_hresult(0x80071B5D, 'HRESULT_FROM_WIN32(ERROR_CTX_CANNOT_MAKE_EVENTLOG_ENTRY)')
    WIN32__ERROR_CTX_SERVICE_NAME_COLLISION: Final[ThirtyTwoBits] = cached_hresult(0x80071B5E, 'HRESULT_FROM_WIN32(ERROR_CTX_SERVICE_NAME_COLLISION)')
    WIN32__ERROR_CTX_CLOSE_PENDING: Final[ThirtyTwoBits] = cached_hresult(0x80071B5F, 'HRESULT_FROM_WIN32(ERROR_CTX_CLOSE_PENDING)')
    WIN32__ERROR_CTX_NO_OUTBUF: Final[ThirtyTwoBits] = cached_hresult(0x80071B60, 'HRESULT_FROM_WIN32(ERROR_CTX_NO_OUTBUF)')
    WIN32__ERROR_CTX_MODEM_INF_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80071B61, 'HRESULT_FROM_WIN32(ERROR_CTX_MODEM_INF_NOT_FOUND)')
    WIN32__ERROR_CTX_INVALID_MODEMNAME: Final[ThirtyTwoBits] = cached_hresult(0x80071B62, 'HRESULT_FROM_WIN32(ERROR_CTX_INVALID_MODEMNAME)')
    WIN32__ERROR_CTX_MODEM_RESPONSE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80071B63, 'HRESULT_FROM_WIN32(ERROR_CTX_MODEM_RESPONSE_ERROR)')
    WIN32__ERROR_CTX_MODEM_RESPONSE_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x80071B64, 'HRESULT_FROM_WIN32(ERROR_CTX_MODEM_RESPONSE_TIMEOUT)')
    WIN32__ERROR_CTX_MODEM_RESPONSE_NO_CARRIER: Final[ThirtyTwoBits] = cached_hresult(0x80071B65, 'HRESULT_FROM_WIN32(ERROR_CTX_MODEM_RESPONSE_NO_CARRIER)')
    WIN32__ERROR_CTX_MODEM_RESPONSE_NO_DIALTONE: Final[ThirtyTwoBits] = cached_hresult(0x80071B66, 'HRESULT_FROM_WIN32(ERROR_CTX_MODEM_RESPONSE_NO_DIALTONE)')
    WIN32__ERROR_CTX_MODEM_RESPONSE_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x80071B67, 'HRESULT_FROM_WIN32(ERROR_CTX_MODEM_RESPONSE_BUSY)')
    WIN32__ERROR_CTX_MODEM_RESPONSE_VOICE: Final[ThirtyTwoBits] = cached_hresult(0x80071B68, 'HRESULT_FROM_WIN32(ERROR_CTX_MODEM_RESPONSE_VOICE)')
    WIN32__ERROR_CTX_TD_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80071B69, 'HRESULT_FROM_WIN32(ERROR_CTX_TD_ERROR)')
    WIN32__ERROR_CTX_WINSTATION_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80071B6E, 'HRESULT_FROM_WIN32(ERROR_CTX_WINSTATION_NOT_FOUND)')
    WIN32__ERROR_CTX_WINSTATION_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80071B6F, 'HRESULT_FROM_WIN32(ERROR_CTX_WINSTATION_ALREADY_EXISTS)')
    WIN32__ERROR_CTX_WINSTATION_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x80071B70, 'HRESULT_FROM_WIN32(ERROR_CTX_WINSTATION_BUSY)')
    WIN32__ERROR_CTX_BAD_VIDEO_MODE: Final[ThirtyTwoBits] = cached_hresult(0x80071B71, 'HRESULT_FROM_WIN32(ERROR_CTX_BAD_VIDEO_MODE)')
    WIN32__ERROR_CTX_GRAPHICS_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80071B7B, 'HRESULT_FROM_WIN32(ERROR_CTX_GRAPHICS_INVALID)')
    WIN32__ERROR_CTX_LOGON_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80071B7D, 'HRESULT_FROM_WIN32(ERROR_CTX_LOGON_DISABLED)')
    WIN32__ERROR_CTX_NOT_CONSOLE: Final[ThirtyTwoBits] = cached_hresult(0x80071B7E, 'HRESULT_FROM_WIN32(ERROR_CTX_NOT_CONSOLE)')
    WIN32__ERROR_CTX_CLIENT_QUERY_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x80071B80, 'HRESULT_FROM_WIN32(ERROR_CTX_CLIENT_QUERY_TIMEOUT)')
    WIN32__ERROR_CTX_CONSOLE_DISCONNECT: Final[ThirtyTwoBits] = cached_hresult(0x80071B81, 'HRESULT_FROM_WIN32(ERROR_CTX_CONSOLE_DISCONNECT)')
    WIN32__ERROR_CTX_CONSOLE_CONNECT: Final[ThirtyTwoBits] = cached_hresult(0x80071B82, 'HRESULT_FROM_WIN32(ERROR_CTX_CONSOLE_CONNECT)')
    WIN32__ERROR_CTX_SHADOW_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x80071B84, 'HRESULT_FROM_WIN32(ERROR_CTX_SHADOW_DENIED)')
    WIN32__ERROR_CTX_WINSTATION_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x80071B85, 'HRESULT_FROM_WIN32(ERROR_CTX_WINSTATION_ACCESS_DENIED)')
    WIN32__ERROR_CTX_INVALID_WD: Final[ThirtyTwoBits] = cached_hresult(0x80071B89, 'HRESULT_FROM_WIN32(ERROR_CTX_INVALID_WD)')
    WIN32__ERROR_CTX_SHADOW_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80071B8A, 'HRESULT_FROM_WIN32(ERROR_CTX_SHADOW_INVALID)')
    WIN32__ERROR_CTX_SHADOW_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80071B8B, 'HRESULT_FROM_WIN32(ERROR_CTX_SHADOW_DISABLED)')
    WIN32__ERROR_CTX_CLIENT_LICENSE_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x80071B8C, 'HRESULT_FROM_WIN32(ERROR_CTX_CLIENT_LICENSE_IN_USE)')
    WIN32__ERROR_CTX_CLIENT_LICENSE_NOT_SET: Final[ThirtyTwoBits] = cached_hresult(0x80071B8D, 'HRESULT_FROM_WIN32(ERROR_CTX_CLIENT_LICENSE_NOT_SET)')
    WIN32__ERROR_CTX_LICENSE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80071B8E, 'HRESULT_FROM_WIN32(ERROR_CTX_LICENSE_NOT_AVAILABLE)')
    WIN32__ERROR_CTX_LICENSE_CLIENT_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80071B8F, 'HRESULT_FROM_WIN32(ERROR_CTX_LICENSE_CLIENT_INVALID)')
    WIN32__ERROR_CTX_LICENSE_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x80071B90, 'HRESULT_FROM_WIN32(ERROR_CTX_LICENSE_EXPIRED)')
    WIN32__ERROR_CTX_SHADOW_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x80071B91, 'HRESULT_FROM_WIN32(ERROR_CTX_SHADOW_NOT_RUNNING)')
    WIN32__ERROR_CTX_SHADOW_ENDED_BY_MODE_CHANGE: Final[ThirtyTwoBits] = cached_hresult(0x80071B92, 'HRESULT_FROM_WIN32(ERROR_CTX_SHADOW_ENDED_BY_MODE_CHANGE)')
    WIN32__ERROR_ACTIVATION_COUNT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80071B93, 'HRESULT_FROM_WIN32(ERROR_ACTIVATION_COUNT_EXCEEDED)')
    WIN32__ERROR_CTX_WINSTATIONS_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80071B94, 'HRESULT_FROM_WIN32(ERROR_CTX_WINSTATIONS_DISABLED)')
    WIN32__ERROR_CTX_ENCRYPTION_LEVEL_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80071B95, 'HRESULT_FROM_WIN32(ERROR_CTX_ENCRYPTION_LEVEL_REQUIRED)')
    WIN32__ERROR_CTX_SESSION_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x80071B96, 'HRESULT_FROM_WIN32(ERROR_CTX_SESSION_IN_USE)')
    WIN32__ERROR_CTX_NO_FORCE_LOGOFF: Final[ThirtyTwoBits] = cached_hresult(0x80071B97, 'HRESULT_FROM_WIN32(ERROR_CTX_NO_FORCE_LOGOFF)')
    WIN32__ERROR_CTX_ACCOUNT_RESTRICTION: Final[ThirtyTwoBits] = cached_hresult(0x80071B98, 'HRESULT_FROM_WIN32(ERROR_CTX_ACCOUNT_RESTRICTION)')
    WIN32__ERROR_RDP_PROTOCOL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80071B99, 'HRESULT_FROM_WIN32(ERROR_RDP_PROTOCOL_ERROR)')
    WIN32__ERROR_CTX_CDM_CONNECT: Final[ThirtyTwoBits] = cached_hresult(0x80071B9A, 'HRESULT_FROM_WIN32(ERROR_CTX_CDM_CONNECT)')
    WIN32__ERROR_CTX_CDM_DISCONNECT: Final[ThirtyTwoBits] = cached_hresult(0x80071B9B, 'HRESULT_FROM_WIN32(ERROR_CTX_CDM_DISCONNECT)')
    WIN32__ERROR_CTX_SECURITY_LAYER_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80071B9C, 'HRESULT_FROM_WIN32(ERROR_CTX_SECURITY_LAYER_ERROR)')
    WIN32__ERROR_TS_INCOMPATIBLE_SESSIONS: Final[ThirtyTwoBits] = cached_hresult(0x80071B9D, 'HRESULT_FROM_WIN32(ERROR_TS_INCOMPATIBLE_SESSIONS)')
    WIN32__ERROR_TS_VIDEO_SUBSYSTEM_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80071B9E, 'HRESULT_FROM_WIN32(ERROR_TS_VIDEO_SUBSYSTEM_ERROR)')
    WIN32__FRS_ERR_INVALID_API_SEQUENCE: Final[ThirtyTwoBits] = cached_hresult(0x80071F41, 'HRESULT_FROM_WIN32(FRS_ERR_INVALID_API_SEQUENCE)')
    WIN32__FRS_ERR_STARTING_SERVICE: Final[ThirtyTwoBits] = cached_hresult(0x80071F42, 'HRESULT_FROM_WIN32(FRS_ERR_STARTING_SERVICE)')
    WIN32__FRS_ERR_STOPPING_SERVICE: Final[ThirtyTwoBits] = cached_hresult(0x80071F43, 'HRESULT_FROM_WIN32(FRS_ERR_STOPPING_SERVICE)')
    WIN32__FRS_ERR_INTERNAL_API: Final[ThirtyTwoBits] = cached_hresult(0x80071F44, 'HRESULT_FROM_WIN32(FRS_ERR_INTERNAL_API)')
    WIN32__FRS_ERR_INTERNAL: Final[ThirtyTwoBits] = cached_hresult(0x80071F45, 'HRESULT_FROM_WIN32(FRS_ERR_INTERNAL)')
    WIN32__FRS_ERR_SERVICE_COMM: Final[ThirtyTwoBits] = cached_hresult(0x80071F46, 'HRESULT_FROM_WIN32(FRS_ERR_SERVICE_COMM)')
    WIN32__FRS_ERR_INSUFFICIENT_PRIV: Final[ThirtyTwoBits] = cached_hresult(0x80071F47, 'HRESULT_FROM_WIN32(FRS_ERR_INSUFFICIENT_PRIV)')
    WIN32__FRS_ERR_AUTHENTICATION: Final[ThirtyTwoBits] = cached_hresult(0x80071F48, 'HRESULT_FROM_WIN32(FRS_ERR_AUTHENTICATION)')
    WIN32__FRS_ERR_PARENT_INSUFFICIENT_PRIV: Final[ThirtyTwoBits] = cached_hresult(0x80071F49, 'HRESULT_FROM_WIN32(FRS_ERR_PARENT_INSUFFICIENT_PRIV)')
    WIN32__FRS_ERR_PARENT_AUTHENTICATION: Final[ThirtyTwoBits] = cached_hresult(0x80071F4A, 'HRESULT_FROM_WIN32(FRS_ERR_PARENT_AUTHENTICATION)')
    WIN32__FRS_ERR_CHILD_TO_PARENT_COMM: Final[ThirtyTwoBits] = cached_hresult(0x80071F4B, 'HRESULT_FROM_WIN32(FRS_ERR_CHILD_TO_PARENT_COMM)')
    WIN32__FRS_ERR_PARENT_TO_CHILD_COMM: Final[ThirtyTwoBits] = cached_hresult(0x80071F4C, 'HRESULT_FROM_WIN32(FRS_ERR_PARENT_TO_CHILD_COMM)')
    WIN32__FRS_ERR_SYSVOL_POPULATE: Final[ThirtyTwoBits] = cached_hresult(0x80071F4D, 'HRESULT_FROM_WIN32(FRS_ERR_SYSVOL_POPULATE)')
    WIN32__FRS_ERR_SYSVOL_POPULATE_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x80071F4E, 'HRESULT_FROM_WIN32(FRS_ERR_SYSVOL_POPULATE_TIMEOUT)')
    WIN32__FRS_ERR_SYSVOL_IS_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x80071F4F, 'HRESULT_FROM_WIN32(FRS_ERR_SYSVOL_IS_BUSY)')
    WIN32__FRS_ERR_SYSVOL_DEMOTE: Final[ThirtyTwoBits] = cached_hresult(0x80071F50, 'HRESULT_FROM_WIN32(FRS_ERR_SYSVOL_DEMOTE)')
    WIN32__FRS_ERR_INVALID_SERVICE_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0x80071F51, 'HRESULT_FROM_WIN32(FRS_ERR_INVALID_SERVICE_PARAMETER)')
    WIN32__ERROR_DS_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x80072008, 'HRESULT_FROM_WIN32(ERROR_DS_NOT_INSTALLED)')
    WIN32__ERROR_DS_MEMBERSHIP_EVALUATED_LOCALLY: Final[ThirtyTwoBits] = cached_hresult(0x80072009, 'HRESULT_FROM_WIN32(ERROR_DS_MEMBERSHIP_EVALUATED_LOCALLY)')
    WIN32__ERROR_DS_NO_ATTRIBUTE_OR_VALUE: Final[ThirtyTwoBits] = cached_hresult(0x8007200A, 'HRESULT_FROM_WIN32(ERROR_DS_NO_ATTRIBUTE_OR_VALUE)')
    WIN32__ERROR_DS_INVALID_ATTRIBUTE_SYNTAX: Final[ThirtyTwoBits] = cached_hresult(0x8007200B, 'HRESULT_FROM_WIN32(ERROR_DS_INVALID_ATTRIBUTE_SYNTAX)')
    WIN32__ERROR_DS_ATTRIBUTE_TYPE_UNDEFINED: Final[ThirtyTwoBits] = cached_hresult(0x8007200C, 'HRESULT_FROM_WIN32(ERROR_DS_ATTRIBUTE_TYPE_UNDEFINED)')
    WIN32__ERROR_DS_ATTRIBUTE_OR_VALUE_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x8007200D, 'HRESULT_FROM_WIN32(ERROR_DS_ATTRIBUTE_OR_VALUE_EXISTS)')
    WIN32__ERROR_DS_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x8007200E, 'HRESULT_FROM_WIN32(ERROR_DS_BUSY)')
    WIN32__ERROR_DS_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x8007200F, 'HRESULT_FROM_WIN32(ERROR_DS_UNAVAILABLE)')
    WIN32__ERROR_DS_NO_RIDS_ALLOCATED: Final[ThirtyTwoBits] = cached_hresult(0x80072010, 'HRESULT_FROM_WIN32(ERROR_DS_NO_RIDS_ALLOCATED)')
    WIN32__ERROR_DS_NO_MORE_RIDS: Final[ThirtyTwoBits] = cached_hresult(0x80072011, 'HRESULT_FROM_WIN32(ERROR_DS_NO_MORE_RIDS)')
    WIN32__ERROR_DS_INCORRECT_ROLE_OWNER: Final[ThirtyTwoBits] = cached_hresult(0x80072012, 'HRESULT_FROM_WIN32(ERROR_DS_INCORRECT_ROLE_OWNER)')
    WIN32__ERROR_DS_RIDMGR_INIT_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80072013, 'HRESULT_FROM_WIN32(ERROR_DS_RIDMGR_INIT_ERROR)')
    WIN32__ERROR_DS_OBJ_CLASS_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x80072014, 'HRESULT_FROM_WIN32(ERROR_DS_OBJ_CLASS_VIOLATION)')
    WIN32__ERROR_DS_CANT_ON_NON_LEAF: Final[ThirtyTwoBits] = cached_hresult(0x80072015, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_ON_NON_LEAF)')
    WIN32__ERROR_DS_CANT_ON_RDN: Final[ThirtyTwoBits] = cached_hresult(0x80072016, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_ON_RDN)')
    WIN32__ERROR_DS_CANT_MOD_OBJ_CLASS: Final[ThirtyTwoBits] = cached_hresult(0x80072017, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_MOD_OBJ_CLASS)')
    WIN32__ERROR_DS_CROSS_DOM_MOVE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80072018, 'HRESULT_FROM_WIN32(ERROR_DS_CROSS_DOM_MOVE_ERROR)')
    WIN32__ERROR_DS_GC_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80072019, 'HRESULT_FROM_WIN32(ERROR_DS_GC_NOT_AVAILABLE)')
    WIN32__ERROR_SHARED_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x8007201A, 'HRESULT_FROM_WIN32(ERROR_SHARED_POLICY)')
    WIN32__ERROR_POLICY_OBJECT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8007201B, 'HRESULT_FROM_WIN32(ERROR_POLICY_OBJECT_NOT_FOUND)')
    WIN32__ERROR_POLICY_ONLY_IN_DS: Final[ThirtyTwoBits] = cached_hresult(0x8007201C, 'HRESULT_FROM_WIN32(ERROR_POLICY_ONLY_IN_DS)')
    WIN32__ERROR_PROMOTION_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0x8007201D, 'HRESULT_FROM_WIN32(ERROR_PROMOTION_ACTIVE)')
    WIN32__ERROR_NO_PROMOTION_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0x8007201E, 'HRESULT_FROM_WIN32(ERROR_NO_PROMOTION_ACTIVE)')
    WIN32__ERROR_DS_OPERATIONS_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80072020, 'HRESULT_FROM_WIN32(ERROR_DS_OPERATIONS_ERROR)')
    WIN32__ERROR_DS_PROTOCOL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80072021, 'HRESULT_FROM_WIN32(ERROR_DS_PROTOCOL_ERROR)')
    WIN32__ERROR_DS_TIMELIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80072022, 'HRESULT_FROM_WIN32(ERROR_DS_TIMELIMIT_EXCEEDED)')
    WIN32__ERROR_DS_SIZELIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80072023, 'HRESULT_FROM_WIN32(ERROR_DS_SIZELIMIT_EXCEEDED)')
    WIN32__ERROR_DS_ADMIN_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80072024, 'HRESULT_FROM_WIN32(ERROR_DS_ADMIN_LIMIT_EXCEEDED)')
    WIN32__ERROR_DS_COMPARE_FALSE: Final[ThirtyTwoBits] = cached_hresult(0x80072025, 'HRESULT_FROM_WIN32(ERROR_DS_COMPARE_FALSE)')
    WIN32__ERROR_DS_COMPARE_TRUE: Final[ThirtyTwoBits] = cached_hresult(0x80072026, 'HRESULT_FROM_WIN32(ERROR_DS_COMPARE_TRUE)')
    WIN32__ERROR_DS_AUTH_METHOD_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80072027, 'HRESULT_FROM_WIN32(ERROR_DS_AUTH_METHOD_NOT_SUPPORTED)')
    WIN32__ERROR_DS_STRONG_AUTH_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80072028, 'HRESULT_FROM_WIN32(ERROR_DS_STRONG_AUTH_REQUIRED)')
    WIN32__ERROR_DS_INAPPROPRIATE_AUTH: Final[ThirtyTwoBits] = cached_hresult(0x80072029, 'HRESULT_FROM_WIN32(ERROR_DS_INAPPROPRIATE_AUTH)')
    WIN32__ERROR_DS_AUTH_UNKNOWN: Final[ThirtyTwoBits] = cached_hresult(0x8007202A, 'HRESULT_FROM_WIN32(ERROR_DS_AUTH_UNKNOWN)')
    WIN32__ERROR_DS_REFERRAL: Final[ThirtyTwoBits] = cached_hresult(0x8007202B, 'HRESULT_FROM_WIN32(ERROR_DS_REFERRAL)')
    WIN32__ERROR_DS_UNAVAILABLE_CRIT_EXTENSION: Final[ThirtyTwoBits] = cached_hresult(0x8007202C, 'HRESULT_FROM_WIN32(ERROR_DS_UNAVAILABLE_CRIT_EXTENSION)')
    WIN32__ERROR_DS_CONFIDENTIALITY_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8007202D, 'HRESULT_FROM_WIN32(ERROR_DS_CONFIDENTIALITY_REQUIRED)')
    WIN32__ERROR_DS_INAPPROPRIATE_MATCHING: Final[ThirtyTwoBits] = cached_hresult(0x8007202E, 'HRESULT_FROM_WIN32(ERROR_DS_INAPPROPRIATE_MATCHING)')
    WIN32__ERROR_DS_CONSTRAINT_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x8007202F, 'HRESULT_FROM_WIN32(ERROR_DS_CONSTRAINT_VIOLATION)')
    WIN32__ERROR_DS_NO_SUCH_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x80072030, 'HRESULT_FROM_WIN32(ERROR_DS_NO_SUCH_OBJECT)')
    WIN32__ERROR_DS_ALIAS_PROBLEM: Final[ThirtyTwoBits] = cached_hresult(0x80072031, 'HRESULT_FROM_WIN32(ERROR_DS_ALIAS_PROBLEM)')
    WIN32__ERROR_DS_INVALID_DN_SYNTAX: Final[ThirtyTwoBits] = cached_hresult(0x80072032, 'HRESULT_FROM_WIN32(ERROR_DS_INVALID_DN_SYNTAX)')
    WIN32__ERROR_DS_IS_LEAF: Final[ThirtyTwoBits] = cached_hresult(0x80072033, 'HRESULT_FROM_WIN32(ERROR_DS_IS_LEAF)')
    WIN32__ERROR_DS_ALIAS_DEREF_PROBLEM: Final[ThirtyTwoBits] = cached_hresult(0x80072034, 'HRESULT_FROM_WIN32(ERROR_DS_ALIAS_DEREF_PROBLEM)')
    WIN32__ERROR_DS_UNWILLING_TO_PERFORM: Final[ThirtyTwoBits] = cached_hresult(0x80072035, 'HRESULT_FROM_WIN32(ERROR_DS_UNWILLING_TO_PERFORM)')
    WIN32__ERROR_DS_LOOP_DETECT: Final[ThirtyTwoBits] = cached_hresult(0x80072036, 'HRESULT_FROM_WIN32(ERROR_DS_LOOP_DETECT)')
    WIN32__ERROR_DS_NAMING_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x80072037, 'HRESULT_FROM_WIN32(ERROR_DS_NAMING_VIOLATION)')
    WIN32__ERROR_DS_OBJECT_RESULTS_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0x80072038, 'HRESULT_FROM_WIN32(ERROR_DS_OBJECT_RESULTS_TOO_LARGE)')
    WIN32__ERROR_DS_AFFECTS_MULTIPLE_DSAS: Final[ThirtyTwoBits] = cached_hresult(0x80072039, 'HRESULT_FROM_WIN32(ERROR_DS_AFFECTS_MULTIPLE_DSAS)')
    WIN32__ERROR_DS_SERVER_DOWN: Final[ThirtyTwoBits] = cached_hresult(0x8007203A, 'HRESULT_FROM_WIN32(ERROR_DS_SERVER_DOWN)')
    WIN32__ERROR_DS_LOCAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8007203B, 'HRESULT_FROM_WIN32(ERROR_DS_LOCAL_ERROR)')
    WIN32__ERROR_DS_ENCODING_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8007203C, 'HRESULT_FROM_WIN32(ERROR_DS_ENCODING_ERROR)')
    WIN32__ERROR_DS_DECODING_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8007203D, 'HRESULT_FROM_WIN32(ERROR_DS_DECODING_ERROR)')
    WIN32__ERROR_DS_FILTER_UNKNOWN: Final[ThirtyTwoBits] = cached_hresult(0x8007203E, 'HRESULT_FROM_WIN32(ERROR_DS_FILTER_UNKNOWN)')
    WIN32__ERROR_DS_PARAM_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8007203F, 'HRESULT_FROM_WIN32(ERROR_DS_PARAM_ERROR)')
    WIN32__ERROR_DS_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80072040, 'HRESULT_FROM_WIN32(ERROR_DS_NOT_SUPPORTED)')
    WIN32__ERROR_DS_NO_RESULTS_RETURNED: Final[ThirtyTwoBits] = cached_hresult(0x80072041, 'HRESULT_FROM_WIN32(ERROR_DS_NO_RESULTS_RETURNED)')
    WIN32__ERROR_DS_CONTROL_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80072042, 'HRESULT_FROM_WIN32(ERROR_DS_CONTROL_NOT_FOUND)')
    WIN32__ERROR_DS_CLIENT_LOOP: Final[ThirtyTwoBits] = cached_hresult(0x80072043, 'HRESULT_FROM_WIN32(ERROR_DS_CLIENT_LOOP)')
    WIN32__ERROR_DS_REFERRAL_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80072044, 'HRESULT_FROM_WIN32(ERROR_DS_REFERRAL_LIMIT_EXCEEDED)')
    WIN32__ERROR_DS_SORT_CONTROL_MISSING: Final[ThirtyTwoBits] = cached_hresult(0x80072045, 'HRESULT_FROM_WIN32(ERROR_DS_SORT_CONTROL_MISSING)')
    WIN32__ERROR_DS_OFFSET_RANGE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80072046, 'HRESULT_FROM_WIN32(ERROR_DS_OFFSET_RANGE_ERROR)')
    WIN32__ERROR_DS_RIDMGR_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80072047, 'HRESULT_FROM_WIN32(ERROR_DS_RIDMGR_DISABLED)')
    WIN32__ERROR_DS_ROOT_MUST_BE_NC: Final[ThirtyTwoBits] = cached_hresult(0x8007206D, 'HRESULT_FROM_WIN32(ERROR_DS_ROOT_MUST_BE_NC)')
    WIN32__ERROR_DS_ADD_REPLICA_INHIBITED: Final[ThirtyTwoBits] = cached_hresult(0x8007206E, 'HRESULT_FROM_WIN32(ERROR_DS_ADD_REPLICA_INHIBITED)')
    WIN32__ERROR_DS_ATT_NOT_DEF_IN_SCHEMA: Final[ThirtyTwoBits] = cached_hresult(0x8007206F, 'HRESULT_FROM_WIN32(ERROR_DS_ATT_NOT_DEF_IN_SCHEMA)')
    WIN32__ERROR_DS_MAX_OBJ_SIZE_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80072070, 'HRESULT_FROM_WIN32(ERROR_DS_MAX_OBJ_SIZE_EXCEEDED)')
    WIN32__ERROR_DS_OBJ_STRING_NAME_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80072071, 'HRESULT_FROM_WIN32(ERROR_DS_OBJ_STRING_NAME_EXISTS)')
    WIN32__ERROR_DS_NO_RDN_DEFINED_IN_SCHEMA: Final[ThirtyTwoBits] = cached_hresult(0x80072072, 'HRESULT_FROM_WIN32(ERROR_DS_NO_RDN_DEFINED_IN_SCHEMA)')
    WIN32__ERROR_DS_RDN_DOESNT_MATCH_SCHEMA: Final[ThirtyTwoBits] = cached_hresult(0x80072073, 'HRESULT_FROM_WIN32(ERROR_DS_RDN_DOESNT_MATCH_SCHEMA)')
    WIN32__ERROR_DS_NO_REQUESTED_ATTS_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80072074, 'HRESULT_FROM_WIN32(ERROR_DS_NO_REQUESTED_ATTS_FOUND)')
    WIN32__ERROR_DS_USER_BUFFER_TO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0x80072075, 'HRESULT_FROM_WIN32(ERROR_DS_USER_BUFFER_TO_SMALL)')
    WIN32__ERROR_DS_ATT_IS_NOT_ON_OBJ: Final[ThirtyTwoBits] = cached_hresult(0x80072076, 'HRESULT_FROM_WIN32(ERROR_DS_ATT_IS_NOT_ON_OBJ)')
    WIN32__ERROR_DS_ILLEGAL_MOD_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x80072077, 'HRESULT_FROM_WIN32(ERROR_DS_ILLEGAL_MOD_OPERATION)')
    WIN32__ERROR_DS_OBJ_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0x80072078, 'HRESULT_FROM_WIN32(ERROR_DS_OBJ_TOO_LARGE)')
    WIN32__ERROR_DS_BAD_INSTANCE_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x80072079, 'HRESULT_FROM_WIN32(ERROR_DS_BAD_INSTANCE_TYPE)')
    WIN32__ERROR_DS_MASTERDSA_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8007207A, 'HRESULT_FROM_WIN32(ERROR_DS_MASTERDSA_REQUIRED)')
    WIN32__ERROR_DS_OBJECT_CLASS_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8007207B, 'HRESULT_FROM_WIN32(ERROR_DS_OBJECT_CLASS_REQUIRED)')
    WIN32__ERROR_DS_MISSING_REQUIRED_ATT: Final[ThirtyTwoBits] = cached_hresult(0x8007207C, 'HRESULT_FROM_WIN32(ERROR_DS_MISSING_REQUIRED_ATT)')
    WIN32__ERROR_DS_ATT_NOT_DEF_FOR_CLASS: Final[ThirtyTwoBits] = cached_hresult(0x8007207D, 'HRESULT_FROM_WIN32(ERROR_DS_ATT_NOT_DEF_FOR_CLASS)')
    WIN32__ERROR_DS_ATT_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x8007207E, 'HRESULT_FROM_WIN32(ERROR_DS_ATT_ALREADY_EXISTS)')
    WIN32__ERROR_DS_CANT_ADD_ATT_VALUES: Final[ThirtyTwoBits] = cached_hresult(0x80072080, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_ADD_ATT_VALUES)')
    WIN32__ERROR_DS_SINGLE_VALUE_CONSTRAINT: Final[ThirtyTwoBits] = cached_hresult(0x80072081, 'HRESULT_FROM_WIN32(ERROR_DS_SINGLE_VALUE_CONSTRAINT)')
    WIN32__ERROR_DS_RANGE_CONSTRAINT: Final[ThirtyTwoBits] = cached_hresult(0x80072082, 'HRESULT_FROM_WIN32(ERROR_DS_RANGE_CONSTRAINT)')
    WIN32__ERROR_DS_ATT_VAL_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80072083, 'HRESULT_FROM_WIN32(ERROR_DS_ATT_VAL_ALREADY_EXISTS)')
    WIN32__ERROR_DS_CANT_REM_MISSING_ATT: Final[ThirtyTwoBits] = cached_hresult(0x80072084, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_REM_MISSING_ATT)')
    WIN32__ERROR_DS_CANT_REM_MISSING_ATT_VAL: Final[ThirtyTwoBits] = cached_hresult(0x80072085, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_REM_MISSING_ATT_VAL)')
    WIN32__ERROR_DS_ROOT_CANT_BE_SUBREF: Final[ThirtyTwoBits] = cached_hresult(0x80072086, 'HRESULT_FROM_WIN32(ERROR_DS_ROOT_CANT_BE_SUBREF)')
    WIN32__ERROR_DS_NO_CHAINING: Final[ThirtyTwoBits] = cached_hresult(0x80072087, 'HRESULT_FROM_WIN32(ERROR_DS_NO_CHAINING)')
    WIN32__ERROR_DS_NO_CHAINED_EVAL: Final[ThirtyTwoBits] = cached_hresult(0x80072088, 'HRESULT_FROM_WIN32(ERROR_DS_NO_CHAINED_EVAL)')
    WIN32__ERROR_DS_NO_PARENT_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x80072089, 'HRESULT_FROM_WIN32(ERROR_DS_NO_PARENT_OBJECT)')
    WIN32__ERROR_DS_PARENT_IS_AN_ALIAS: Final[ThirtyTwoBits] = cached_hresult(0x8007208A, 'HRESULT_FROM_WIN32(ERROR_DS_PARENT_IS_AN_ALIAS)')
    WIN32__ERROR_DS_CANT_MIX_MASTER_AND_REPS: Final[ThirtyTwoBits] = cached_hresult(0x8007208B, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_MIX_MASTER_AND_REPS)')
    WIN32__ERROR_DS_CHILDREN_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x8007208C, 'HRESULT_FROM_WIN32(ERROR_DS_CHILDREN_EXIST)')
    WIN32__ERROR_DS_OBJ_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8007208D, 'HRESULT_FROM_WIN32(ERROR_DS_OBJ_NOT_FOUND)')
    WIN32__ERROR_DS_ALIASED_OBJ_MISSING: Final[ThirtyTwoBits] = cached_hresult(0x8007208E, 'HRESULT_FROM_WIN32(ERROR_DS_ALIASED_OBJ_MISSING)')
    WIN32__ERROR_DS_BAD_NAME_SYNTAX: Final[ThirtyTwoBits] = cached_hresult(0x8007208F, 'HRESULT_FROM_WIN32(ERROR_DS_BAD_NAME_SYNTAX)')
    WIN32__ERROR_DS_ALIAS_POINTS_TO_ALIAS: Final[ThirtyTwoBits] = cached_hresult(0x80072090, 'HRESULT_FROM_WIN32(ERROR_DS_ALIAS_POINTS_TO_ALIAS)')
    WIN32__ERROR_DS_CANT_DEREF_ALIAS: Final[ThirtyTwoBits] = cached_hresult(0x80072091, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_DEREF_ALIAS)')
    WIN32__ERROR_DS_OUT_OF_SCOPE: Final[ThirtyTwoBits] = cached_hresult(0x80072092, 'HRESULT_FROM_WIN32(ERROR_DS_OUT_OF_SCOPE)')
    WIN32__ERROR_DS_OBJECT_BEING_REMOVED: Final[ThirtyTwoBits] = cached_hresult(0x80072093, 'HRESULT_FROM_WIN32(ERROR_DS_OBJECT_BEING_REMOVED)')
    WIN32__ERROR_DS_CANT_DELETE_DSA_OBJ: Final[ThirtyTwoBits] = cached_hresult(0x80072094, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_DELETE_DSA_OBJ)')
    WIN32__ERROR_DS_GENERIC_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80072095, 'HRESULT_FROM_WIN32(ERROR_DS_GENERIC_ERROR)')
    WIN32__ERROR_DS_DSA_MUST_BE_INT_MASTER: Final[ThirtyTwoBits] = cached_hresult(0x80072096, 'HRESULT_FROM_WIN32(ERROR_DS_DSA_MUST_BE_INT_MASTER)')
    WIN32__ERROR_DS_CLASS_NOT_DSA: Final[ThirtyTwoBits] = cached_hresult(0x80072097, 'HRESULT_FROM_WIN32(ERROR_DS_CLASS_NOT_DSA)')
    WIN32__ERROR_DS_INSUFF_ACCESS_RIGHTS: Final[ThirtyTwoBits] = cached_hresult(0x80072098, 'HRESULT_FROM_WIN32(ERROR_DS_INSUFF_ACCESS_RIGHTS)')
    WIN32__ERROR_DS_ILLEGAL_SUPERIOR: Final[ThirtyTwoBits] = cached_hresult(0x80072099, 'HRESULT_FROM_WIN32(ERROR_DS_ILLEGAL_SUPERIOR)')
    WIN32__ERROR_DS_ATTRIBUTE_OWNED_BY_SAM: Final[ThirtyTwoBits] = cached_hresult(0x8007209A, 'HRESULT_FROM_WIN32(ERROR_DS_ATTRIBUTE_OWNED_BY_SAM)')
    WIN32__ERROR_DS_NAME_TOO_MANY_PARTS: Final[ThirtyTwoBits] = cached_hresult(0x8007209B, 'HRESULT_FROM_WIN32(ERROR_DS_NAME_TOO_MANY_PARTS)')
    WIN32__ERROR_DS_NAME_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0x8007209C, 'HRESULT_FROM_WIN32(ERROR_DS_NAME_TOO_LONG)')
    WIN32__ERROR_DS_NAME_VALUE_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0x8007209D, 'HRESULT_FROM_WIN32(ERROR_DS_NAME_VALUE_TOO_LONG)')
    WIN32__ERROR_DS_NAME_UNPARSEABLE: Final[ThirtyTwoBits] = cached_hresult(0x8007209E, 'HRESULT_FROM_WIN32(ERROR_DS_NAME_UNPARSEABLE)')
    WIN32__ERROR_DS_NAME_TYPE_UNKNOWN: Final[ThirtyTwoBits] = cached_hresult(0x8007209F, 'HRESULT_FROM_WIN32(ERROR_DS_NAME_TYPE_UNKNOWN)')
    WIN32__ERROR_DS_NOT_AN_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x800720A0, 'HRESULT_FROM_WIN32(ERROR_DS_NOT_AN_OBJECT)')
    WIN32__ERROR_DS_SEC_DESC_TOO_SHORT: Final[ThirtyTwoBits] = cached_hresult(0x800720A1, 'HRESULT_FROM_WIN32(ERROR_DS_SEC_DESC_TOO_SHORT)')
    WIN32__ERROR_DS_SEC_DESC_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800720A2, 'HRESULT_FROM_WIN32(ERROR_DS_SEC_DESC_INVALID)')
    WIN32__ERROR_DS_NO_DELETED_NAME: Final[ThirtyTwoBits] = cached_hresult(0x800720A3, 'HRESULT_FROM_WIN32(ERROR_DS_NO_DELETED_NAME)')
    WIN32__ERROR_DS_SUBREF_MUST_HAVE_PARENT: Final[ThirtyTwoBits] = cached_hresult(0x800720A4, 'HRESULT_FROM_WIN32(ERROR_DS_SUBREF_MUST_HAVE_PARENT)')
    WIN32__ERROR_DS_NCNAME_MUST_BE_NC: Final[ThirtyTwoBits] = cached_hresult(0x800720A5, 'HRESULT_FROM_WIN32(ERROR_DS_NCNAME_MUST_BE_NC)')
    WIN32__ERROR_DS_CANT_ADD_SYSTEM_ONLY: Final[ThirtyTwoBits] = cached_hresult(0x800720A6, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_ADD_SYSTEM_ONLY)')
    WIN32__ERROR_DS_CLASS_MUST_BE_CONCRETE: Final[ThirtyTwoBits] = cached_hresult(0x800720A7, 'HRESULT_FROM_WIN32(ERROR_DS_CLASS_MUST_BE_CONCRETE)')
    WIN32__ERROR_DS_INVALID_DMD: Final[ThirtyTwoBits] = cached_hresult(0x800720A8, 'HRESULT_FROM_WIN32(ERROR_DS_INVALID_DMD)')
    WIN32__ERROR_DS_OBJ_GUID_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800720A9, 'HRESULT_FROM_WIN32(ERROR_DS_OBJ_GUID_EXISTS)')
    WIN32__ERROR_DS_NOT_ON_BACKLINK: Final[ThirtyTwoBits] = cached_hresult(0x800720AA, 'HRESULT_FROM_WIN32(ERROR_DS_NOT_ON_BACKLINK)')
    WIN32__ERROR_DS_NO_CROSSREF_FOR_NC: Final[ThirtyTwoBits] = cached_hresult(0x800720AB, 'HRESULT_FROM_WIN32(ERROR_DS_NO_CROSSREF_FOR_NC)')
    WIN32__ERROR_DS_SHUTTING_DOWN: Final[ThirtyTwoBits] = cached_hresult(0x800720AC, 'HRESULT_FROM_WIN32(ERROR_DS_SHUTTING_DOWN)')
    WIN32__ERROR_DS_UNKNOWN_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x800720AD, 'HRESULT_FROM_WIN32(ERROR_DS_UNKNOWN_OPERATION)')
    WIN32__ERROR_DS_INVALID_ROLE_OWNER: Final[ThirtyTwoBits] = cached_hresult(0x800720AE, 'HRESULT_FROM_WIN32(ERROR_DS_INVALID_ROLE_OWNER)')
    WIN32__ERROR_DS_COULDNT_CONTACT_FSMO: Final[ThirtyTwoBits] = cached_hresult(0x800720AF, 'HRESULT_FROM_WIN32(ERROR_DS_COULDNT_CONTACT_FSMO)')
    WIN32__ERROR_DS_CROSS_NC_DN_RENAME: Final[ThirtyTwoBits] = cached_hresult(0x800720B0, 'HRESULT_FROM_WIN32(ERROR_DS_CROSS_NC_DN_RENAME)')
    WIN32__ERROR_DS_CANT_MOD_SYSTEM_ONLY: Final[ThirtyTwoBits] = cached_hresult(0x800720B1, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_MOD_SYSTEM_ONLY)')
    WIN32__ERROR_DS_REPLICATOR_ONLY: Final[ThirtyTwoBits] = cached_hresult(0x800720B2, 'HRESULT_FROM_WIN32(ERROR_DS_REPLICATOR_ONLY)')
    WIN32__ERROR_DS_OBJ_CLASS_NOT_DEFINED: Final[ThirtyTwoBits] = cached_hresult(0x800720B3, 'HRESULT_FROM_WIN32(ERROR_DS_OBJ_CLASS_NOT_DEFINED)')
    WIN32__ERROR_DS_OBJ_CLASS_NOT_SUBCLASS: Final[ThirtyTwoBits] = cached_hresult(0x800720B4, 'HRESULT_FROM_WIN32(ERROR_DS_OBJ_CLASS_NOT_SUBCLASS)')
    WIN32__ERROR_DS_NAME_REFERENCE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x800720B5, 'HRESULT_FROM_WIN32(ERROR_DS_NAME_REFERENCE_INVALID)')
    WIN32__ERROR_DS_CROSS_REF_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800720B6, 'HRESULT_FROM_WIN32(ERROR_DS_CROSS_REF_EXISTS)')
    WIN32__ERROR_DS_CANT_DEL_MASTER_CROSSREF: Final[ThirtyTwoBits] = cached_hresult(0x800720B7, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_DEL_MASTER_CROSSREF)')
    WIN32__ERROR_DS_SUBTREE_NOTIFY_NOT_NC_HEAD: Final[ThirtyTwoBits] = cached_hresult(0x800720B8, 'HRESULT_FROM_WIN32(ERROR_DS_SUBTREE_NOTIFY_NOT_NC_HEAD)')
    WIN32__ERROR_DS_NOTIFY_FILTER_TOO_COMPLEX: Final[ThirtyTwoBits] = cached_hresult(0x800720B9, 'HRESULT_FROM_WIN32(ERROR_DS_NOTIFY_FILTER_TOO_COMPLEX)')
    WIN32__ERROR_DS_DUP_RDN: Final[ThirtyTwoBits] = cached_hresult(0x800720BA, 'HRESULT_FROM_WIN32(ERROR_DS_DUP_RDN)')
    WIN32__ERROR_DS_DUP_OID: Final[ThirtyTwoBits] = cached_hresult(0x800720BB, 'HRESULT_FROM_WIN32(ERROR_DS_DUP_OID)')
    WIN32__ERROR_DS_DUP_MAPI_ID: Final[ThirtyTwoBits] = cached_hresult(0x800720BC, 'HRESULT_FROM_WIN32(ERROR_DS_DUP_MAPI_ID)')
    WIN32__ERROR_DS_DUP_SCHEMA_ID_GUID: Final[ThirtyTwoBits] = cached_hresult(0x800720BD, 'HRESULT_FROM_WIN32(ERROR_DS_DUP_SCHEMA_ID_GUID)')
    WIN32__ERROR_DS_DUP_LDAP_DISPLAY_NAME: Final[ThirtyTwoBits] = cached_hresult(0x800720BE, 'HRESULT_FROM_WIN32(ERROR_DS_DUP_LDAP_DISPLAY_NAME)')
    WIN32__ERROR_DS_SEMANTIC_ATT_TEST: Final[ThirtyTwoBits] = cached_hresult(0x800720BF, 'HRESULT_FROM_WIN32(ERROR_DS_SEMANTIC_ATT_TEST)')
    WIN32__ERROR_DS_SYNTAX_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x800720C0, 'HRESULT_FROM_WIN32(ERROR_DS_SYNTAX_MISMATCH)')
    WIN32__ERROR_DS_EXISTS_IN_MUST_HAVE: Final[ThirtyTwoBits] = cached_hresult(0x800720C1, 'HRESULT_FROM_WIN32(ERROR_DS_EXISTS_IN_MUST_HAVE)')
    WIN32__ERROR_DS_EXISTS_IN_MAY_HAVE: Final[ThirtyTwoBits] = cached_hresult(0x800720C2, 'HRESULT_FROM_WIN32(ERROR_DS_EXISTS_IN_MAY_HAVE)')
    WIN32__ERROR_DS_NONEXISTENT_MAY_HAVE: Final[ThirtyTwoBits] = cached_hresult(0x800720C3, 'HRESULT_FROM_WIN32(ERROR_DS_NONEXISTENT_MAY_HAVE)')
    WIN32__ERROR_DS_NONEXISTENT_MUST_HAVE: Final[ThirtyTwoBits] = cached_hresult(0x800720C4, 'HRESULT_FROM_WIN32(ERROR_DS_NONEXISTENT_MUST_HAVE)')
    WIN32__ERROR_DS_AUX_CLS_TEST_FAIL: Final[ThirtyTwoBits] = cached_hresult(0x800720C5, 'HRESULT_FROM_WIN32(ERROR_DS_AUX_CLS_TEST_FAIL)')
    WIN32__ERROR_DS_NONEXISTENT_POSS_SUP: Final[ThirtyTwoBits] = cached_hresult(0x800720C6, 'HRESULT_FROM_WIN32(ERROR_DS_NONEXISTENT_POSS_SUP)')
    WIN32__ERROR_DS_SUB_CLS_TEST_FAIL: Final[ThirtyTwoBits] = cached_hresult(0x800720C7, 'HRESULT_FROM_WIN32(ERROR_DS_SUB_CLS_TEST_FAIL)')
    WIN32__ERROR_DS_BAD_RDN_ATT_ID_SYNTAX: Final[ThirtyTwoBits] = cached_hresult(0x800720C8, 'HRESULT_FROM_WIN32(ERROR_DS_BAD_RDN_ATT_ID_SYNTAX)')
    WIN32__ERROR_DS_EXISTS_IN_AUX_CLS: Final[ThirtyTwoBits] = cached_hresult(0x800720C9, 'HRESULT_FROM_WIN32(ERROR_DS_EXISTS_IN_AUX_CLS)')
    WIN32__ERROR_DS_EXISTS_IN_SUB_CLS: Final[ThirtyTwoBits] = cached_hresult(0x800720CA, 'HRESULT_FROM_WIN32(ERROR_DS_EXISTS_IN_SUB_CLS)')
    WIN32__ERROR_DS_EXISTS_IN_POSS_SUP: Final[ThirtyTwoBits] = cached_hresult(0x800720CB, 'HRESULT_FROM_WIN32(ERROR_DS_EXISTS_IN_POSS_SUP)')
    WIN32__ERROR_DS_RECALCSCHEMA_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800720CC, 'HRESULT_FROM_WIN32(ERROR_DS_RECALCSCHEMA_FAILED)')
    WIN32__ERROR_DS_TREE_DELETE_NOT_FINISHED: Final[ThirtyTwoBits] = cached_hresult(0x800720CD, 'HRESULT_FROM_WIN32(ERROR_DS_TREE_DELETE_NOT_FINISHED)')
    WIN32__ERROR_DS_CANT_DELETE: Final[ThirtyTwoBits] = cached_hresult(0x800720CE, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_DELETE)')
    WIN32__ERROR_DS_ATT_SCHEMA_REQ_ID: Final[ThirtyTwoBits] = cached_hresult(0x800720CF, 'HRESULT_FROM_WIN32(ERROR_DS_ATT_SCHEMA_REQ_ID)')
    WIN32__ERROR_DS_BAD_ATT_SCHEMA_SYNTAX: Final[ThirtyTwoBits] = cached_hresult(0x800720D0, 'HRESULT_FROM_WIN32(ERROR_DS_BAD_ATT_SCHEMA_SYNTAX)')
    WIN32__ERROR_DS_CANT_CACHE_ATT: Final[ThirtyTwoBits] = cached_hresult(0x800720D1, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_CACHE_ATT)')
    WIN32__ERROR_DS_CANT_CACHE_CLASS: Final[ThirtyTwoBits] = cached_hresult(0x800720D2, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_CACHE_CLASS)')
    WIN32__ERROR_DS_CANT_REMOVE_ATT_CACHE: Final[ThirtyTwoBits] = cached_hresult(0x800720D3, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_REMOVE_ATT_CACHE)')
    WIN32__ERROR_DS_CANT_REMOVE_CLASS_CACHE: Final[ThirtyTwoBits] = cached_hresult(0x800720D4, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_REMOVE_CLASS_CACHE)')
    WIN32__ERROR_DS_CANT_RETRIEVE_DN: Final[ThirtyTwoBits] = cached_hresult(0x800720D5, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_RETRIEVE_DN)')
    WIN32__ERROR_DS_MISSING_SUPREF: Final[ThirtyTwoBits] = cached_hresult(0x800720D6, 'HRESULT_FROM_WIN32(ERROR_DS_MISSING_SUPREF)')
    WIN32__ERROR_DS_CANT_RETRIEVE_INSTANCE: Final[ThirtyTwoBits] = cached_hresult(0x800720D7, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_RETRIEVE_INSTANCE)')
    WIN32__ERROR_DS_CODE_INCONSISTENCY: Final[ThirtyTwoBits] = cached_hresult(0x800720D8, 'HRESULT_FROM_WIN32(ERROR_DS_CODE_INCONSISTENCY)')
    WIN32__ERROR_DS_DATABASE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800720D9, 'HRESULT_FROM_WIN32(ERROR_DS_DATABASE_ERROR)')
    WIN32__ERROR_DS_GOVERNSID_MISSING: Final[ThirtyTwoBits] = cached_hresult(0x800720DA, 'HRESULT_FROM_WIN32(ERROR_DS_GOVERNSID_MISSING)')
    WIN32__ERROR_DS_MISSING_EXPECTED_ATT: Final[ThirtyTwoBits] = cached_hresult(0x800720DB, 'HRESULT_FROM_WIN32(ERROR_DS_MISSING_EXPECTED_ATT)')
    WIN32__ERROR_DS_NCNAME_MISSING_CR_REF: Final[ThirtyTwoBits] = cached_hresult(0x800720DC, 'HRESULT_FROM_WIN32(ERROR_DS_NCNAME_MISSING_CR_REF)')
    WIN32__ERROR_DS_SECURITY_CHECKING_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800720DD, 'HRESULT_FROM_WIN32(ERROR_DS_SECURITY_CHECKING_ERROR)')
    WIN32__ERROR_DS_SCHEMA_NOT_LOADED: Final[ThirtyTwoBits] = cached_hresult(0x800720DE, 'HRESULT_FROM_WIN32(ERROR_DS_SCHEMA_NOT_LOADED)')
    WIN32__ERROR_DS_SCHEMA_ALLOC_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800720DF, 'HRESULT_FROM_WIN32(ERROR_DS_SCHEMA_ALLOC_FAILED)')
    WIN32__ERROR_DS_ATT_SCHEMA_REQ_SYNTAX: Final[ThirtyTwoBits] = cached_hresult(0x800720E0, 'HRESULT_FROM_WIN32(ERROR_DS_ATT_SCHEMA_REQ_SYNTAX)')
    WIN32__ERROR_DS_GCVERIFY_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800720E1, 'HRESULT_FROM_WIN32(ERROR_DS_GCVERIFY_ERROR)')
    WIN32__ERROR_DS_DRA_SCHEMA_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x800720E2, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_SCHEMA_MISMATCH)')
    WIN32__ERROR_DS_CANT_FIND_DSA_OBJ: Final[ThirtyTwoBits] = cached_hresult(0x800720E3, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_FIND_DSA_OBJ)')
    WIN32__ERROR_DS_CANT_FIND_EXPECTED_NC: Final[ThirtyTwoBits] = cached_hresult(0x800720E4, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_FIND_EXPECTED_NC)')
    WIN32__ERROR_DS_CANT_FIND_NC_IN_CACHE: Final[ThirtyTwoBits] = cached_hresult(0x800720E5, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_FIND_NC_IN_CACHE)')
    WIN32__ERROR_DS_CANT_RETRIEVE_CHILD: Final[ThirtyTwoBits] = cached_hresult(0x800720E6, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_RETRIEVE_CHILD)')
    WIN32__ERROR_DS_SECURITY_ILLEGAL_MODIFY: Final[ThirtyTwoBits] = cached_hresult(0x800720E7, 'HRESULT_FROM_WIN32(ERROR_DS_SECURITY_ILLEGAL_MODIFY)')
    WIN32__ERROR_DS_CANT_REPLACE_HIDDEN_REC: Final[ThirtyTwoBits] = cached_hresult(0x800720E8, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_REPLACE_HIDDEN_REC)')
    WIN32__ERROR_DS_BAD_HIERARCHY_FILE: Final[ThirtyTwoBits] = cached_hresult(0x800720E9, 'HRESULT_FROM_WIN32(ERROR_DS_BAD_HIERARCHY_FILE)')
    WIN32__ERROR_DS_BUILD_HIERARCHY_TABLE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800720EA, 'HRESULT_FROM_WIN32(ERROR_DS_BUILD_HIERARCHY_TABLE_FAILED)')
    WIN32__ERROR_DS_CONFIG_PARAM_MISSING: Final[ThirtyTwoBits] = cached_hresult(0x800720EB, 'HRESULT_FROM_WIN32(ERROR_DS_CONFIG_PARAM_MISSING)')
    WIN32__ERROR_DS_COUNTING_AB_INDICES_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800720EC, 'HRESULT_FROM_WIN32(ERROR_DS_COUNTING_AB_INDICES_FAILED)')
    WIN32__ERROR_DS_HIERARCHY_TABLE_MALLOC_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800720ED, 'HRESULT_FROM_WIN32(ERROR_DS_HIERARCHY_TABLE_MALLOC_FAILED)')
    WIN32__ERROR_DS_INTERNAL_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x800720EE, 'HRESULT_FROM_WIN32(ERROR_DS_INTERNAL_FAILURE)')
    WIN32__ERROR_DS_UNKNOWN_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800720EF, 'HRESULT_FROM_WIN32(ERROR_DS_UNKNOWN_ERROR)')
    WIN32__ERROR_DS_ROOT_REQUIRES_CLASS_TOP: Final[ThirtyTwoBits] = cached_hresult(0x800720F0, 'HRESULT_FROM_WIN32(ERROR_DS_ROOT_REQUIRES_CLASS_TOP)')
    WIN32__ERROR_DS_REFUSING_FSMO_ROLES: Final[ThirtyTwoBits] = cached_hresult(0x800720F1, 'HRESULT_FROM_WIN32(ERROR_DS_REFUSING_FSMO_ROLES)')
    WIN32__ERROR_DS_MISSING_FSMO_SETTINGS: Final[ThirtyTwoBits] = cached_hresult(0x800720F2, 'HRESULT_FROM_WIN32(ERROR_DS_MISSING_FSMO_SETTINGS)')
    WIN32__ERROR_DS_UNABLE_TO_SURRENDER_ROLES: Final[ThirtyTwoBits] = cached_hresult(0x800720F3, 'HRESULT_FROM_WIN32(ERROR_DS_UNABLE_TO_SURRENDER_ROLES)')
    WIN32__ERROR_DS_DRA_GENERIC: Final[ThirtyTwoBits] = cached_hresult(0x800720F4, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_GENERIC)')
    WIN32__ERROR_DS_DRA_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0x800720F5, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_INVALID_PARAMETER)')
    WIN32__ERROR_DS_DRA_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x800720F6, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_BUSY)')
    WIN32__ERROR_DS_DRA_BAD_DN: Final[ThirtyTwoBits] = cached_hresult(0x800720F7, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_BAD_DN)')
    WIN32__ERROR_DS_DRA_BAD_NC: Final[ThirtyTwoBits] = cached_hresult(0x800720F8, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_BAD_NC)')
    WIN32__ERROR_DS_DRA_DN_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800720F9, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_DN_EXISTS)')
    WIN32__ERROR_DS_DRA_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800720FA, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_INTERNAL_ERROR)')
    WIN32__ERROR_DS_DRA_INCONSISTENT_DIT: Final[ThirtyTwoBits] = cached_hresult(0x800720FB, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_INCONSISTENT_DIT)')
    WIN32__ERROR_DS_DRA_CONNECTION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800720FC, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_CONNECTION_FAILED)')
    WIN32__ERROR_DS_DRA_BAD_INSTANCE_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x800720FD, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_BAD_INSTANCE_TYPE)')
    WIN32__ERROR_DS_DRA_OUT_OF_MEM: Final[ThirtyTwoBits] = cached_hresult(0x800720FE, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_OUT_OF_MEM)')
    WIN32__ERROR_DS_DRA_MAIL_PROBLEM: Final[ThirtyTwoBits] = cached_hresult(0x800720FF, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_MAIL_PROBLEM)')
    WIN32__ERROR_DS_DRA_REF_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80072100, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_REF_ALREADY_EXISTS)')
    WIN32__ERROR_DS_DRA_REF_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80072101, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_REF_NOT_FOUND)')
    WIN32__ERROR_DS_DRA_OBJ_IS_REP_SOURCE: Final[ThirtyTwoBits] = cached_hresult(0x80072102, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_OBJ_IS_REP_SOURCE)')
    WIN32__ERROR_DS_DRA_DB_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80072103, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_DB_ERROR)')
    WIN32__ERROR_DS_DRA_NO_REPLICA: Final[ThirtyTwoBits] = cached_hresult(0x80072104, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_NO_REPLICA)')
    WIN32__ERROR_DS_DRA_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x80072105, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_ACCESS_DENIED)')
    WIN32__ERROR_DS_DRA_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80072106, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_NOT_SUPPORTED)')
    WIN32__ERROR_DS_DRA_RPC_CANCELLED: Final[ThirtyTwoBits] = cached_hresult(0x80072107, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_RPC_CANCELLED)')
    WIN32__ERROR_DS_DRA_SOURCE_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80072108, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_SOURCE_DISABLED)')
    WIN32__ERROR_DS_DRA_SINK_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80072109, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_SINK_DISABLED)')
    WIN32__ERROR_DS_DRA_NAME_COLLISION: Final[ThirtyTwoBits] = cached_hresult(0x8007210A, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_NAME_COLLISION)')
    WIN32__ERROR_DS_DRA_SOURCE_REINSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x8007210B, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_SOURCE_REINSTALLED)')
    WIN32__ERROR_DS_DRA_MISSING_PARENT: Final[ThirtyTwoBits] = cached_hresult(0x8007210C, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_MISSING_PARENT)')
    WIN32__ERROR_DS_DRA_PREEMPTED: Final[ThirtyTwoBits] = cached_hresult(0x8007210D, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_PREEMPTED)')
    WIN32__ERROR_DS_DRA_ABANDON_SYNC: Final[ThirtyTwoBits] = cached_hresult(0x8007210E, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_ABANDON_SYNC)')
    WIN32__ERROR_DS_DRA_SHUTDOWN: Final[ThirtyTwoBits] = cached_hresult(0x8007210F, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_SHUTDOWN)')
    WIN32__ERROR_DS_DRA_INCOMPATIBLE_PARTIAL_SET: Final[ThirtyTwoBits] = cached_hresult(0x80072110, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_INCOMPATIBLE_PARTIAL_SET)')
    WIN32__ERROR_DS_DRA_SOURCE_IS_PARTIAL_REPLICA: Final[ThirtyTwoBits] = cached_hresult(0x80072111, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_SOURCE_IS_PARTIAL_REPLICA)')
    WIN32__ERROR_DS_DRA_EXTN_CONNECTION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80072112, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_EXTN_CONNECTION_FAILED)')
    WIN32__ERROR_DS_INSTALL_SCHEMA_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80072113, 'HRESULT_FROM_WIN32(ERROR_DS_INSTALL_SCHEMA_MISMATCH)')
    WIN32__ERROR_DS_DUP_LINK_ID: Final[ThirtyTwoBits] = cached_hresult(0x80072114, 'HRESULT_FROM_WIN32(ERROR_DS_DUP_LINK_ID)')
    WIN32__ERROR_DS_NAME_ERROR_RESOLVING: Final[ThirtyTwoBits] = cached_hresult(0x80072115, 'HRESULT_FROM_WIN32(ERROR_DS_NAME_ERROR_RESOLVING)')
    WIN32__ERROR_DS_NAME_ERROR_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80072116, 'HRESULT_FROM_WIN32(ERROR_DS_NAME_ERROR_NOT_FOUND)')
    WIN32__ERROR_DS_NAME_ERROR_NOT_UNIQUE: Final[ThirtyTwoBits] = cached_hresult(0x80072117, 'HRESULT_FROM_WIN32(ERROR_DS_NAME_ERROR_NOT_UNIQUE)')
    WIN32__ERROR_DS_NAME_ERROR_NO_MAPPING: Final[ThirtyTwoBits] = cached_hresult(0x80072118, 'HRESULT_FROM_WIN32(ERROR_DS_NAME_ERROR_NO_MAPPING)')
    WIN32__ERROR_DS_NAME_ERROR_DOMAIN_ONLY: Final[ThirtyTwoBits] = cached_hresult(0x80072119, 'HRESULT_FROM_WIN32(ERROR_DS_NAME_ERROR_DOMAIN_ONLY)')
    WIN32__ERROR_DS_NAME_ERROR_NO_SYNTACTICAL_MAPPING: Final[ThirtyTwoBits] = cached_hresult(0x8007211A, 'HRESULT_FROM_WIN32(ERROR_DS_NAME_ERROR_NO_SYNTACTICAL_MAPPING)')
    WIN32__ERROR_DS_CONSTRUCTED_ATT_MOD: Final[ThirtyTwoBits] = cached_hresult(0x8007211B, 'HRESULT_FROM_WIN32(ERROR_DS_CONSTRUCTED_ATT_MOD)')
    WIN32__ERROR_DS_WRONG_OM_OBJ_CLASS: Final[ThirtyTwoBits] = cached_hresult(0x8007211C, 'HRESULT_FROM_WIN32(ERROR_DS_WRONG_OM_OBJ_CLASS)')
    WIN32__ERROR_DS_DRA_REPL_PENDING: Final[ThirtyTwoBits] = cached_hresult(0x8007211D, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_REPL_PENDING)')
    WIN32__ERROR_DS_DS_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8007211E, 'HRESULT_FROM_WIN32(ERROR_DS_DS_REQUIRED)')
    WIN32__ERROR_DS_INVALID_LDAP_DISPLAY_NAME: Final[ThirtyTwoBits] = cached_hresult(0x8007211F, 'HRESULT_FROM_WIN32(ERROR_DS_INVALID_LDAP_DISPLAY_NAME)')
    WIN32__ERROR_DS_NON_BASE_SEARCH: Final[ThirtyTwoBits] = cached_hresult(0x80072120, 'HRESULT_FROM_WIN32(ERROR_DS_NON_BASE_SEARCH)')
    WIN32__ERROR_DS_CANT_RETRIEVE_ATTS: Final[ThirtyTwoBits] = cached_hresult(0x80072121, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_RETRIEVE_ATTS)')
    WIN32__ERROR_DS_BACKLINK_WITHOUT_LINK: Final[ThirtyTwoBits] = cached_hresult(0x80072122, 'HRESULT_FROM_WIN32(ERROR_DS_BACKLINK_WITHOUT_LINK)')
    WIN32__ERROR_DS_EPOCH_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80072123, 'HRESULT_FROM_WIN32(ERROR_DS_EPOCH_MISMATCH)')
    WIN32__ERROR_DS_SRC_NAME_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80072124, 'HRESULT_FROM_WIN32(ERROR_DS_SRC_NAME_MISMATCH)')
    WIN32__ERROR_DS_SRC_AND_DST_NC_IDENTICAL: Final[ThirtyTwoBits] = cached_hresult(0x80072125, 'HRESULT_FROM_WIN32(ERROR_DS_SRC_AND_DST_NC_IDENTICAL)')
    WIN32__ERROR_DS_DST_NC_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80072126, 'HRESULT_FROM_WIN32(ERROR_DS_DST_NC_MISMATCH)')
    WIN32__ERROR_DS_NOT_AUTHORITIVE_FOR_DST_NC: Final[ThirtyTwoBits] = cached_hresult(0x80072127, 'HRESULT_FROM_WIN32(ERROR_DS_NOT_AUTHORITIVE_FOR_DST_NC)')
    WIN32__ERROR_DS_SRC_GUID_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80072128, 'HRESULT_FROM_WIN32(ERROR_DS_SRC_GUID_MISMATCH)')
    WIN32__ERROR_DS_CANT_MOVE_DELETED_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x80072129, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_MOVE_DELETED_OBJECT)')
    WIN32__ERROR_DS_PDC_OPERATION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x8007212A, 'HRESULT_FROM_WIN32(ERROR_DS_PDC_OPERATION_IN_PROGRESS)')
    WIN32__ERROR_DS_CROSS_DOMAIN_CLEANUP_REQD: Final[ThirtyTwoBits] = cached_hresult(0x8007212B, 'HRESULT_FROM_WIN32(ERROR_DS_CROSS_DOMAIN_CLEANUP_REQD)')
    WIN32__ERROR_DS_ILLEGAL_XDOM_MOVE_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x8007212C, 'HRESULT_FROM_WIN32(ERROR_DS_ILLEGAL_XDOM_MOVE_OPERATION)')
    WIN32__ERROR_DS_CANT_WITH_ACCT_GROUP_MEMBERSHPS: Final[ThirtyTwoBits] = cached_hresult(0x8007212D, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_WITH_ACCT_GROUP_MEMBERSHPS)')
    WIN32__ERROR_DS_NC_MUST_HAVE_NC_PARENT: Final[ThirtyTwoBits] = cached_hresult(0x8007212E, 'HRESULT_FROM_WIN32(ERROR_DS_NC_MUST_HAVE_NC_PARENT)')
    WIN32__ERROR_DS_CR_IMPOSSIBLE_TO_VALIDATE: Final[ThirtyTwoBits] = cached_hresult(0x8007212F, 'HRESULT_FROM_WIN32(ERROR_DS_CR_IMPOSSIBLE_TO_VALIDATE)')
    WIN32__ERROR_DS_DST_DOMAIN_NOT_NATIVE: Final[ThirtyTwoBits] = cached_hresult(0x80072130, 'HRESULT_FROM_WIN32(ERROR_DS_DST_DOMAIN_NOT_NATIVE)')
    WIN32__ERROR_DS_MISSING_INFRASTRUCTURE_CONTAINER: Final[ThirtyTwoBits] = cached_hresult(0x80072131, 'HRESULT_FROM_WIN32(ERROR_DS_MISSING_INFRASTRUCTURE_CONTAINER)')
    WIN32__ERROR_DS_CANT_MOVE_ACCOUNT_GROUP: Final[ThirtyTwoBits] = cached_hresult(0x80072132, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_MOVE_ACCOUNT_GROUP)')
    WIN32__ERROR_DS_CANT_MOVE_RESOURCE_GROUP: Final[ThirtyTwoBits] = cached_hresult(0x80072133, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_MOVE_RESOURCE_GROUP)')
    WIN32__ERROR_DS_INVALID_SEARCH_FLAG: Final[ThirtyTwoBits] = cached_hresult(0x80072134, 'HRESULT_FROM_WIN32(ERROR_DS_INVALID_SEARCH_FLAG)')
    WIN32__ERROR_DS_NO_TREE_DELETE_ABOVE_NC: Final[ThirtyTwoBits] = cached_hresult(0x80072135, 'HRESULT_FROM_WIN32(ERROR_DS_NO_TREE_DELETE_ABOVE_NC)')
    WIN32__ERROR_DS_COULDNT_LOCK_TREE_FOR_DELETE: Final[ThirtyTwoBits] = cached_hresult(0x80072136, 'HRESULT_FROM_WIN32(ERROR_DS_COULDNT_LOCK_TREE_FOR_DELETE)')
    WIN32__ERROR_DS_COULDNT_IDENTIFY_OBJECTS_FOR_TREE_DELETE: Final[ThirtyTwoBits] = cached_hresult(0x80072137, 'HRESULT_FROM_WIN32(ERROR_DS_COULDNT_IDENTIFY_OBJECTS_FOR_TREE_DELETE)')
    WIN32__ERROR_DS_SAM_INIT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80072138, 'HRESULT_FROM_WIN32(ERROR_DS_SAM_INIT_FAILURE)')
    WIN32__ERROR_DS_SENSITIVE_GROUP_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x80072139, 'HRESULT_FROM_WIN32(ERROR_DS_SENSITIVE_GROUP_VIOLATION)')
    WIN32__ERROR_DS_CANT_MOD_PRIMARYGROUPID: Final[ThirtyTwoBits] = cached_hresult(0x8007213A, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_MOD_PRIMARYGROUPID)')
    WIN32__ERROR_DS_ILLEGAL_BASE_SCHEMA_MOD: Final[ThirtyTwoBits] = cached_hresult(0x8007213B, 'HRESULT_FROM_WIN32(ERROR_DS_ILLEGAL_BASE_SCHEMA_MOD)')
    WIN32__ERROR_DS_NONSAFE_SCHEMA_CHANGE: Final[ThirtyTwoBits] = cached_hresult(0x8007213C, 'HRESULT_FROM_WIN32(ERROR_DS_NONSAFE_SCHEMA_CHANGE)')
    WIN32__ERROR_DS_SCHEMA_UPDATE_DISALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x8007213D, 'HRESULT_FROM_WIN32(ERROR_DS_SCHEMA_UPDATE_DISALLOWED)')
    WIN32__ERROR_DS_CANT_CREATE_UNDER_SCHEMA: Final[ThirtyTwoBits] = cached_hresult(0x8007213E, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_CREATE_UNDER_SCHEMA)')
    WIN32__ERROR_DS_INSTALL_NO_SRC_SCH_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x8007213F, 'HRESULT_FROM_WIN32(ERROR_DS_INSTALL_NO_SRC_SCH_VERSION)')
    WIN32__ERROR_DS_INSTALL_NO_SCH_VERSION_IN_INIFILE: Final[ThirtyTwoBits] = cached_hresult(0x80072140, 'HRESULT_FROM_WIN32(ERROR_DS_INSTALL_NO_SCH_VERSION_IN_INIFILE)')
    WIN32__ERROR_DS_INVALID_GROUP_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x80072141, 'HRESULT_FROM_WIN32(ERROR_DS_INVALID_GROUP_TYPE)')
    WIN32__ERROR_DS_NO_NEST_GLOBALGROUP_IN_MIXEDDOMAIN: Final[ThirtyTwoBits] = cached_hresult(0x80072142, 'HRESULT_FROM_WIN32(ERROR_DS_NO_NEST_GLOBALGROUP_IN_MIXEDDOMAIN)')
    WIN32__ERROR_DS_NO_NEST_LOCALGROUP_IN_MIXEDDOMAIN: Final[ThirtyTwoBits] = cached_hresult(0x80072143, 'HRESULT_FROM_WIN32(ERROR_DS_NO_NEST_LOCALGROUP_IN_MIXEDDOMAIN)')
    WIN32__ERROR_DS_GLOBAL_CANT_HAVE_LOCAL_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0x80072144, 'HRESULT_FROM_WIN32(ERROR_DS_GLOBAL_CANT_HAVE_LOCAL_MEMBER)')
    WIN32__ERROR_DS_GLOBAL_CANT_HAVE_UNIVERSAL_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0x80072145, 'HRESULT_FROM_WIN32(ERROR_DS_GLOBAL_CANT_HAVE_UNIVERSAL_MEMBER)')
    WIN32__ERROR_DS_UNIVERSAL_CANT_HAVE_LOCAL_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0x80072146, 'HRESULT_FROM_WIN32(ERROR_DS_UNIVERSAL_CANT_HAVE_LOCAL_MEMBER)')
    WIN32__ERROR_DS_GLOBAL_CANT_HAVE_CROSSDOMAIN_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0x80072147, 'HRESULT_FROM_WIN32(ERROR_DS_GLOBAL_CANT_HAVE_CROSSDOMAIN_MEMBER)')
    WIN32__ERROR_DS_LOCAL_CANT_HAVE_CROSSDOMAIN_LOCAL_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0x80072148, 'HRESULT_FROM_WIN32(ERROR_DS_LOCAL_CANT_HAVE_CROSSDOMAIN_LOCAL_MEMBER)')
    WIN32__ERROR_DS_HAVE_PRIMARY_MEMBERS: Final[ThirtyTwoBits] = cached_hresult(0x80072149, 'HRESULT_FROM_WIN32(ERROR_DS_HAVE_PRIMARY_MEMBERS)')
    WIN32__ERROR_DS_STRING_SD_CONVERSION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8007214A, 'HRESULT_FROM_WIN32(ERROR_DS_STRING_SD_CONVERSION_FAILED)')
    WIN32__ERROR_DS_NAMING_MASTER_GC: Final[ThirtyTwoBits] = cached_hresult(0x8007214B, 'HRESULT_FROM_WIN32(ERROR_DS_NAMING_MASTER_GC)')
    WIN32__ERROR_DS_DNS_LOOKUP_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8007214C, 'HRESULT_FROM_WIN32(ERROR_DS_DNS_LOOKUP_FAILURE)')
    WIN32__ERROR_DS_COULDNT_UPDATE_SPNS: Final[ThirtyTwoBits] = cached_hresult(0x8007214D, 'HRESULT_FROM_WIN32(ERROR_DS_COULDNT_UPDATE_SPNS)')
    WIN32__ERROR_DS_CANT_RETRIEVE_SD: Final[ThirtyTwoBits] = cached_hresult(0x8007214E, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_RETRIEVE_SD)')
    WIN32__ERROR_DS_KEY_NOT_UNIQUE: Final[ThirtyTwoBits] = cached_hresult(0x8007214F, 'HRESULT_FROM_WIN32(ERROR_DS_KEY_NOT_UNIQUE)')
    WIN32__ERROR_DS_WRONG_LINKED_ATT_SYNTAX: Final[ThirtyTwoBits] = cached_hresult(0x80072150, 'HRESULT_FROM_WIN32(ERROR_DS_WRONG_LINKED_ATT_SYNTAX)')
    WIN32__ERROR_DS_SAM_NEED_BOOTKEY_PASSWORD: Final[ThirtyTwoBits] = cached_hresult(0x80072151, 'HRESULT_FROM_WIN32(ERROR_DS_SAM_NEED_BOOTKEY_PASSWORD)')
    WIN32__ERROR_DS_SAM_NEED_BOOTKEY_FLOPPY: Final[ThirtyTwoBits] = cached_hresult(0x80072152, 'HRESULT_FROM_WIN32(ERROR_DS_SAM_NEED_BOOTKEY_FLOPPY)')
    WIN32__ERROR_DS_CANT_START: Final[ThirtyTwoBits] = cached_hresult(0x80072153, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_START)')
    WIN32__ERROR_DS_INIT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80072154, 'HRESULT_FROM_WIN32(ERROR_DS_INIT_FAILURE)')
    WIN32__ERROR_DS_NO_PKT_PRIVACY_ON_CONNECTION: Final[ThirtyTwoBits] = cached_hresult(0x80072155, 'HRESULT_FROM_WIN32(ERROR_DS_NO_PKT_PRIVACY_ON_CONNECTION)')
    WIN32__ERROR_DS_SOURCE_DOMAIN_IN_FOREST: Final[ThirtyTwoBits] = cached_hresult(0x80072156, 'HRESULT_FROM_WIN32(ERROR_DS_SOURCE_DOMAIN_IN_FOREST)')
    WIN32__ERROR_DS_DESTINATION_DOMAIN_NOT_IN_FOREST: Final[ThirtyTwoBits] = cached_hresult(0x80072157, 'HRESULT_FROM_WIN32(ERROR_DS_DESTINATION_DOMAIN_NOT_IN_FOREST)')
    WIN32__ERROR_DS_DESTINATION_AUDITING_NOT_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0x80072158, 'HRESULT_FROM_WIN32(ERROR_DS_DESTINATION_AUDITING_NOT_ENABLED)')
    WIN32__ERROR_DS_CANT_FIND_DC_FOR_SRC_DOMAIN: Final[ThirtyTwoBits] = cached_hresult(0x80072159, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_FIND_DC_FOR_SRC_DOMAIN)')
    WIN32__ERROR_DS_SRC_OBJ_NOT_GROUP_OR_USER: Final[ThirtyTwoBits] = cached_hresult(0x8007215A, 'HRESULT_FROM_WIN32(ERROR_DS_SRC_OBJ_NOT_GROUP_OR_USER)')
    WIN32__ERROR_DS_SRC_SID_EXISTS_IN_FOREST: Final[ThirtyTwoBits] = cached_hresult(0x8007215B, 'HRESULT_FROM_WIN32(ERROR_DS_SRC_SID_EXISTS_IN_FOREST)')
    WIN32__ERROR_DS_SRC_AND_DST_OBJECT_CLASS_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x8007215C, 'HRESULT_FROM_WIN32(ERROR_DS_SRC_AND_DST_OBJECT_CLASS_MISMATCH)')
    WIN32__ERROR_SAM_INIT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8007215D, 'HRESULT_FROM_WIN32(ERROR_SAM_INIT_FAILURE)')
    WIN32__ERROR_DS_DRA_SCHEMA_INFO_SHIP: Final[ThirtyTwoBits] = cached_hresult(0x8007215E, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_SCHEMA_INFO_SHIP)')
    WIN32__ERROR_DS_DRA_SCHEMA_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x8007215F, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_SCHEMA_CONFLICT)')
    WIN32__ERROR_DS_DRA_EARLIER_SCHEMA_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x80072160, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_EARLIER_SCHEMA_CONFLICT)')
    WIN32__ERROR_DS_DRA_OBJ_NC_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80072161, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_OBJ_NC_MISMATCH)')
    WIN32__ERROR_DS_NC_STILL_HAS_DSAS: Final[ThirtyTwoBits] = cached_hresult(0x80072162, 'HRESULT_FROM_WIN32(ERROR_DS_NC_STILL_HAS_DSAS)')
    WIN32__ERROR_DS_GC_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80072163, 'HRESULT_FROM_WIN32(ERROR_DS_GC_REQUIRED)')
    WIN32__ERROR_DS_LOCAL_MEMBER_OF_LOCAL_ONLY: Final[ThirtyTwoBits] = cached_hresult(0x80072164, 'HRESULT_FROM_WIN32(ERROR_DS_LOCAL_MEMBER_OF_LOCAL_ONLY)')
    WIN32__ERROR_DS_NO_FPO_IN_UNIVERSAL_GROUPS: Final[ThirtyTwoBits] = cached_hresult(0x80072165, 'HRESULT_FROM_WIN32(ERROR_DS_NO_FPO_IN_UNIVERSAL_GROUPS)')
    WIN32__ERROR_DS_CANT_ADD_TO_GC: Final[ThirtyTwoBits] = cached_hresult(0x80072166, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_ADD_TO_GC)')
    WIN32__ERROR_DS_NO_CHECKPOINT_WITH_PDC: Final[ThirtyTwoBits] = cached_hresult(0x80072167, 'HRESULT_FROM_WIN32(ERROR_DS_NO_CHECKPOINT_WITH_PDC)')
    WIN32__ERROR_DS_SOURCE_AUDITING_NOT_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0x80072168, 'HRESULT_FROM_WIN32(ERROR_DS_SOURCE_AUDITING_NOT_ENABLED)')
    WIN32__ERROR_DS_CANT_CREATE_IN_NONDOMAIN_NC: Final[ThirtyTwoBits] = cached_hresult(0x80072169, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_CREATE_IN_NONDOMAIN_NC)')
    WIN32__ERROR_DS_INVALID_NAME_FOR_SPN: Final[ThirtyTwoBits] = cached_hresult(0x8007216A, 'HRESULT_FROM_WIN32(ERROR_DS_INVALID_NAME_FOR_SPN)')
    WIN32__ERROR_DS_FILTER_USES_CONTRUCTED_ATTRS: Final[ThirtyTwoBits] = cached_hresult(0x8007216B, 'HRESULT_FROM_WIN32(ERROR_DS_FILTER_USES_CONTRUCTED_ATTRS)')
    WIN32__ERROR_DS_UNICODEPWD_NOT_IN_QUOTES: Final[ThirtyTwoBits] = cached_hresult(0x8007216C, 'HRESULT_FROM_WIN32(ERROR_DS_UNICODEPWD_NOT_IN_QUOTES)')
    WIN32__ERROR_DS_MACHINE_ACCOUNT_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8007216D, 'HRESULT_FROM_WIN32(ERROR_DS_MACHINE_ACCOUNT_QUOTA_EXCEEDED)')
    WIN32__ERROR_DS_MUST_BE_RUN_ON_DST_DC: Final[ThirtyTwoBits] = cached_hresult(0x8007216E, 'HRESULT_FROM_WIN32(ERROR_DS_MUST_BE_RUN_ON_DST_DC)')
    WIN32__ERROR_DS_SRC_DC_MUST_BE_SP4_OR_GREATER: Final[ThirtyTwoBits] = cached_hresult(0x8007216F, 'HRESULT_FROM_WIN32(ERROR_DS_SRC_DC_MUST_BE_SP4_OR_GREATER)')
    WIN32__ERROR_DS_CANT_TREE_DELETE_CRITICAL_OBJ: Final[ThirtyTwoBits] = cached_hresult(0x80072170, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_TREE_DELETE_CRITICAL_OBJ)')
    WIN32__ERROR_DS_INIT_FAILURE_CONSOLE: Final[ThirtyTwoBits] = cached_hresult(0x80072171, 'HRESULT_FROM_WIN32(ERROR_DS_INIT_FAILURE_CONSOLE)')
    WIN32__ERROR_DS_SAM_INIT_FAILURE_CONSOLE: Final[ThirtyTwoBits] = cached_hresult(0x80072172, 'HRESULT_FROM_WIN32(ERROR_DS_SAM_INIT_FAILURE_CONSOLE)')
    WIN32__ERROR_DS_FOREST_VERSION_TOO_HIGH: Final[ThirtyTwoBits] = cached_hresult(0x80072173, 'HRESULT_FROM_WIN32(ERROR_DS_FOREST_VERSION_TOO_HIGH)')
    WIN32__ERROR_DS_DOMAIN_VERSION_TOO_HIGH: Final[ThirtyTwoBits] = cached_hresult(0x80072174, 'HRESULT_FROM_WIN32(ERROR_DS_DOMAIN_VERSION_TOO_HIGH)')
    WIN32__ERROR_DS_FOREST_VERSION_TOO_LOW: Final[ThirtyTwoBits] = cached_hresult(0x80072175, 'HRESULT_FROM_WIN32(ERROR_DS_FOREST_VERSION_TOO_LOW)')
    WIN32__ERROR_DS_DOMAIN_VERSION_TOO_LOW: Final[ThirtyTwoBits] = cached_hresult(0x80072176, 'HRESULT_FROM_WIN32(ERROR_DS_DOMAIN_VERSION_TOO_LOW)')
    WIN32__ERROR_DS_INCOMPATIBLE_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80072177, 'HRESULT_FROM_WIN32(ERROR_DS_INCOMPATIBLE_VERSION)')
    WIN32__ERROR_DS_LOW_DSA_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80072178, 'HRESULT_FROM_WIN32(ERROR_DS_LOW_DSA_VERSION)')
    WIN32__ERROR_DS_NO_BEHAVIOR_VERSION_IN_MIXEDDOMAIN: Final[ThirtyTwoBits] = cached_hresult(0x80072179, 'HRESULT_FROM_WIN32(ERROR_DS_NO_BEHAVIOR_VERSION_IN_MIXEDDOMAIN)')
    WIN32__ERROR_DS_NOT_SUPPORTED_SORT_ORDER: Final[ThirtyTwoBits] = cached_hresult(0x8007217A, 'HRESULT_FROM_WIN32(ERROR_DS_NOT_SUPPORTED_SORT_ORDER)')
    WIN32__ERROR_DS_NAME_NOT_UNIQUE: Final[ThirtyTwoBits] = cached_hresult(0x8007217B, 'HRESULT_FROM_WIN32(ERROR_DS_NAME_NOT_UNIQUE)')
    WIN32__ERROR_DS_MACHINE_ACCOUNT_CREATED_PRENT4: Final[ThirtyTwoBits] = cached_hresult(0x8007217C, 'HRESULT_FROM_WIN32(ERROR_DS_MACHINE_ACCOUNT_CREATED_PRENT4)')
    WIN32__ERROR_DS_OUT_OF_VERSION_STORE: Final[ThirtyTwoBits] = cached_hresult(0x8007217D, 'HRESULT_FROM_WIN32(ERROR_DS_OUT_OF_VERSION_STORE)')
    WIN32__ERROR_DS_INCOMPATIBLE_CONTROLS_USED: Final[ThirtyTwoBits] = cached_hresult(0x8007217E, 'HRESULT_FROM_WIN32(ERROR_DS_INCOMPATIBLE_CONTROLS_USED)')
    WIN32__ERROR_DS_NO_REF_DOMAIN: Final[ThirtyTwoBits] = cached_hresult(0x8007217F, 'HRESULT_FROM_WIN32(ERROR_DS_NO_REF_DOMAIN)')
    WIN32__ERROR_DS_RESERVED_LINK_ID: Final[ThirtyTwoBits] = cached_hresult(0x80072180, 'HRESULT_FROM_WIN32(ERROR_DS_RESERVED_LINK_ID)')
    WIN32__ERROR_DS_LINK_ID_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80072181, 'HRESULT_FROM_WIN32(ERROR_DS_LINK_ID_NOT_AVAILABLE)')
    WIN32__ERROR_DS_AG_CANT_HAVE_UNIVERSAL_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0x80072182, 'HRESULT_FROM_WIN32(ERROR_DS_AG_CANT_HAVE_UNIVERSAL_MEMBER)')
    WIN32__ERROR_DS_MODIFYDN_DISALLOWED_BY_INSTANCE_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x80072183, 'HRESULT_FROM_WIN32(ERROR_DS_MODIFYDN_DISALLOWED_BY_INSTANCE_TYPE)')
    WIN32__ERROR_DS_NO_OBJECT_MOVE_IN_SCHEMA_NC: Final[ThirtyTwoBits] = cached_hresult(0x80072184, 'HRESULT_FROM_WIN32(ERROR_DS_NO_OBJECT_MOVE_IN_SCHEMA_NC)')
    WIN32__ERROR_DS_MODIFYDN_DISALLOWED_BY_FLAG: Final[ThirtyTwoBits] = cached_hresult(0x80072185, 'HRESULT_FROM_WIN32(ERROR_DS_MODIFYDN_DISALLOWED_BY_FLAG)')
    WIN32__ERROR_DS_MODIFYDN_WRONG_GRANDPARENT: Final[ThirtyTwoBits] = cached_hresult(0x80072186, 'HRESULT_FROM_WIN32(ERROR_DS_MODIFYDN_WRONG_GRANDPARENT)')
    WIN32__ERROR_DS_NAME_ERROR_TRUST_REFERRAL: Final[ThirtyTwoBits] = cached_hresult(0x80072187, 'HRESULT_FROM_WIN32(ERROR_DS_NAME_ERROR_TRUST_REFERRAL)')
    WIN32__ERROR_NOT_SUPPORTED_ON_STANDARD_SERVER: Final[ThirtyTwoBits] = cached_hresult(0x80072188, 'HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED_ON_STANDARD_SERVER)')
    WIN32__ERROR_DS_CANT_ACCESS_REMOTE_PART_OF_AD: Final[ThirtyTwoBits] = cached_hresult(0x80072189, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_ACCESS_REMOTE_PART_OF_AD)')
    WIN32__ERROR_DS_CR_IMPOSSIBLE_TO_VALIDATE_V2: Final[ThirtyTwoBits] = cached_hresult(0x8007218A, 'HRESULT_FROM_WIN32(ERROR_DS_CR_IMPOSSIBLE_TO_VALIDATE_V2)')
    WIN32__ERROR_DS_THREAD_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x8007218B, 'HRESULT_FROM_WIN32(ERROR_DS_THREAD_LIMIT_EXCEEDED)')
    WIN32__ERROR_DS_NOT_CLOSEST: Final[ThirtyTwoBits] = cached_hresult(0x8007218C, 'HRESULT_FROM_WIN32(ERROR_DS_NOT_CLOSEST)')
    WIN32__ERROR_DS_CANT_DERIVE_SPN_WITHOUT_SERVER_REF: Final[ThirtyTwoBits] = cached_hresult(0x8007218D, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_DERIVE_SPN_WITHOUT_SERVER_REF)')
    WIN32__ERROR_DS_SINGLE_USER_MODE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8007218E, 'HRESULT_FROM_WIN32(ERROR_DS_SINGLE_USER_MODE_FAILED)')
    WIN32__ERROR_DS_NTDSCRIPT_SYNTAX_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8007218F, 'HRESULT_FROM_WIN32(ERROR_DS_NTDSCRIPT_SYNTAX_ERROR)')
    WIN32__ERROR_DS_NTDSCRIPT_PROCESS_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80072190, 'HRESULT_FROM_WIN32(ERROR_DS_NTDSCRIPT_PROCESS_ERROR)')
    WIN32__ERROR_DS_DIFFERENT_REPL_EPOCHS: Final[ThirtyTwoBits] = cached_hresult(0x80072191, 'HRESULT_FROM_WIN32(ERROR_DS_DIFFERENT_REPL_EPOCHS)')
    WIN32__ERROR_DS_DRS_EXTENSIONS_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0x80072192, 'HRESULT_FROM_WIN32(ERROR_DS_DRS_EXTENSIONS_CHANGED)')
    WIN32__ERROR_DS_REPLICA_SET_CHANGE_NOT_ALLOWED_ON_DISABLED_CR: Final[ThirtyTwoBits] = cached_hresult(0x80072193, 'HRESULT_FROM_WIN32(ERROR_DS_REPLICA_SET_CHANGE_NOT_ALLOWED_ON_DISABLED_CR)')
    WIN32__ERROR_DS_NO_MSDS_INTID: Final[ThirtyTwoBits] = cached_hresult(0x80072194, 'HRESULT_FROM_WIN32(ERROR_DS_NO_MSDS_INTID)')
    WIN32__ERROR_DS_DUP_MSDS_INTID: Final[ThirtyTwoBits] = cached_hresult(0x80072195, 'HRESULT_FROM_WIN32(ERROR_DS_DUP_MSDS_INTID)')
    WIN32__ERROR_DS_EXISTS_IN_RDNATTID: Final[ThirtyTwoBits] = cached_hresult(0x80072196, 'HRESULT_FROM_WIN32(ERROR_DS_EXISTS_IN_RDNATTID)')
    WIN32__ERROR_DS_AUTHORIZATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80072197, 'HRESULT_FROM_WIN32(ERROR_DS_AUTHORIZATION_FAILED)')
    WIN32__ERROR_DS_INVALID_SCRIPT: Final[ThirtyTwoBits] = cached_hresult(0x80072198, 'HRESULT_FROM_WIN32(ERROR_DS_INVALID_SCRIPT)')
    WIN32__ERROR_DS_REMOTE_CROSSREF_OP_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80072199, 'HRESULT_FROM_WIN32(ERROR_DS_REMOTE_CROSSREF_OP_FAILED)')
    WIN32__ERROR_DS_CROSS_REF_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x8007219A, 'HRESULT_FROM_WIN32(ERROR_DS_CROSS_REF_BUSY)')
    WIN32__ERROR_DS_CANT_DERIVE_SPN_FOR_DELETED_DOMAIN: Final[ThirtyTwoBits] = cached_hresult(0x8007219B, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_DERIVE_SPN_FOR_DELETED_DOMAIN)')
    WIN32__ERROR_DS_CANT_DEMOTE_WITH_WRITEABLE_NC: Final[ThirtyTwoBits] = cached_hresult(0x8007219C, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_DEMOTE_WITH_WRITEABLE_NC)')
    WIN32__ERROR_DS_DUPLICATE_ID_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8007219D, 'HRESULT_FROM_WIN32(ERROR_DS_DUPLICATE_ID_FOUND)')
    WIN32__ERROR_DS_INSUFFICIENT_ATTR_TO_CREATE_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x8007219E, 'HRESULT_FROM_WIN32(ERROR_DS_INSUFFICIENT_ATTR_TO_CREATE_OBJECT)')
    WIN32__ERROR_DS_GROUP_CONVERSION_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8007219F, 'HRESULT_FROM_WIN32(ERROR_DS_GROUP_CONVERSION_ERROR)')
    WIN32__ERROR_DS_CANT_MOVE_APP_BASIC_GROUP: Final[ThirtyTwoBits] = cached_hresult(0x800721A0, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_MOVE_APP_BASIC_GROUP)')
    WIN32__ERROR_DS_CANT_MOVE_APP_QUERY_GROUP: Final[ThirtyTwoBits] = cached_hresult(0x800721A1, 'HRESULT_FROM_WIN32(ERROR_DS_CANT_MOVE_APP_QUERY_GROUP)')
    WIN32__ERROR_DS_ROLE_NOT_VERIFIED: Final[ThirtyTwoBits] = cached_hresult(0x800721A2, 'HRESULT_FROM_WIN32(ERROR_DS_ROLE_NOT_VERIFIED)')
    WIN32__ERROR_DS_WKO_CONTAINER_CANNOT_BE_SPECIAL: Final[ThirtyTwoBits] = cached_hresult(0x800721A3, 'HRESULT_FROM_WIN32(ERROR_DS_WKO_CONTAINER_CANNOT_BE_SPECIAL)')
    WIN32__ERROR_DS_DOMAIN_RENAME_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x800721A4, 'HRESULT_FROM_WIN32(ERROR_DS_DOMAIN_RENAME_IN_PROGRESS)')
    WIN32__ERROR_DS_EXISTING_AD_CHILD_NC: Final[ThirtyTwoBits] = cached_hresult(0x800721A5, 'HRESULT_FROM_WIN32(ERROR_DS_EXISTING_AD_CHILD_NC)')
    WIN32__ERROR_DS_REPL_LIFETIME_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x800721A6, 'HRESULT_FROM_WIN32(ERROR_DS_REPL_LIFETIME_EXCEEDED)')
    WIN32__ERROR_DS_DISALLOWED_IN_SYSTEM_CONTAINER: Final[ThirtyTwoBits] = cached_hresult(0x800721A7, 'HRESULT_FROM_WIN32(ERROR_DS_DISALLOWED_IN_SYSTEM_CONTAINER)')
    WIN32__ERROR_DS_LDAP_SEND_QUEUE_FULL: Final[ThirtyTwoBits] = cached_hresult(0x800721A8, 'HRESULT_FROM_WIN32(ERROR_DS_LDAP_SEND_QUEUE_FULL)')
    WIN32__ERROR_DS_DRA_OUT_SCHEDULE_WINDOW: Final[ThirtyTwoBits] = cached_hresult(0x800721A9, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_OUT_SCHEDULE_WINDOW)')
    WIN32__ERROR_DS_POLICY_NOT_KNOWN: Final[ThirtyTwoBits] = cached_hresult(0x800721AA, 'HRESULT_FROM_WIN32(ERROR_DS_POLICY_NOT_KNOWN)')
    WIN32__ERROR_NO_SITE_SETTINGS_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x800721AB, 'HRESULT_FROM_WIN32(ERROR_NO_SITE_SETTINGS_OBJECT)')
    WIN32__ERROR_NO_SECRETS: Final[ThirtyTwoBits] = cached_hresult(0x800721AC, 'HRESULT_FROM_WIN32(ERROR_NO_SECRETS)')
    WIN32__ERROR_NO_WRITABLE_DC_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800721AD, 'HRESULT_FROM_WIN32(ERROR_NO_WRITABLE_DC_FOUND)')
    WIN32__ERROR_DS_NO_SERVER_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x800721AE, 'HRESULT_FROM_WIN32(ERROR_DS_NO_SERVER_OBJECT)')
    WIN32__ERROR_DS_NO_NTDSA_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x800721AF, 'HRESULT_FROM_WIN32(ERROR_DS_NO_NTDSA_OBJECT)')
    WIN32__ERROR_DS_NON_ASQ_SEARCH: Final[ThirtyTwoBits] = cached_hresult(0x800721B0, 'HRESULT_FROM_WIN32(ERROR_DS_NON_ASQ_SEARCH)')
    WIN32__ERROR_DS_AUDIT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x800721B1, 'HRESULT_FROM_WIN32(ERROR_DS_AUDIT_FAILURE)')
    WIN32__ERROR_DS_INVALID_SEARCH_FLAG_SUBTREE: Final[ThirtyTwoBits] = cached_hresult(0x800721B2, 'HRESULT_FROM_WIN32(ERROR_DS_INVALID_SEARCH_FLAG_SUBTREE)')
    WIN32__ERROR_DS_INVALID_SEARCH_FLAG_TUPLE: Final[ThirtyTwoBits] = cached_hresult(0x800721B3, 'HRESULT_FROM_WIN32(ERROR_DS_INVALID_SEARCH_FLAG_TUPLE)')
    WIN32__ERROR_DS_HIERARCHY_TABLE_TOO_DEEP: Final[ThirtyTwoBits] = cached_hresult(0x800721B4, 'HRESULT_FROM_WIN32(ERROR_DS_HIERARCHY_TABLE_TOO_DEEP)')
    WIN32__ERROR_DS_DRA_CORRUPT_UTD_VECTOR: Final[ThirtyTwoBits] = cached_hresult(0x800721B5, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_CORRUPT_UTD_VECTOR)')
    WIN32__ERROR_DS_DRA_SECRETS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x800721B6, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_SECRETS_DENIED)')
    WIN32__ERROR_DS_RESERVED_MAPI_ID: Final[ThirtyTwoBits] = cached_hresult(0x800721B7, 'HRESULT_FROM_WIN32(ERROR_DS_RESERVED_MAPI_ID)')
    WIN32__ERROR_DS_MAPI_ID_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x800721B8, 'HRESULT_FROM_WIN32(ERROR_DS_MAPI_ID_NOT_AVAILABLE)')
    WIN32__ERROR_DS_DRA_MISSING_KRBTGT_SECRET: Final[ThirtyTwoBits] = cached_hresult(0x800721B9, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_MISSING_KRBTGT_SECRET)')
    WIN32__ERROR_DS_DOMAIN_NAME_EXISTS_IN_FOREST: Final[ThirtyTwoBits] = cached_hresult(0x800721BA, 'HRESULT_FROM_WIN32(ERROR_DS_DOMAIN_NAME_EXISTS_IN_FOREST)')
    WIN32__ERROR_DS_FLAT_NAME_EXISTS_IN_FOREST: Final[ThirtyTwoBits] = cached_hresult(0x800721BB, 'HRESULT_FROM_WIN32(ERROR_DS_FLAT_NAME_EXISTS_IN_FOREST)')
    WIN32__ERROR_INVALID_USER_PRINCIPAL_NAME: Final[ThirtyTwoBits] = cached_hresult(0x800721BC, 'HRESULT_FROM_WIN32(ERROR_INVALID_USER_PRINCIPAL_NAME)')
    WIN32__ERROR_DS_OID_MAPPED_GROUP_CANT_HAVE_MEMBERS: Final[ThirtyTwoBits] = cached_hresult(0x800721BD, 'HRESULT_FROM_WIN32(ERROR_DS_OID_MAPPED_GROUP_CANT_HAVE_MEMBERS)')
    WIN32__ERROR_DS_OID_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800721BE, 'HRESULT_FROM_WIN32(ERROR_DS_OID_NOT_FOUND)')
    WIN32__ERROR_DS_DRA_RECYCLED_TARGET: Final[ThirtyTwoBits] = cached_hresult(0x800721BF, 'HRESULT_FROM_WIN32(ERROR_DS_DRA_RECYCLED_TARGET)')
    WIN32__ERROR_DS_DISALLOWED_NC_REDIRECT: Final[ThirtyTwoBits] = cached_hresult(0x800721C0, 'HRESULT_FROM_WIN32(ERROR_DS_DISALLOWED_NC_REDIRECT)')
    WIN32__ERROR_DS_HIGH_ADLDS_FFL: Final[ThirtyTwoBits] = cached_hresult(0x800721C1, 'HRESULT_FROM_WIN32(ERROR_DS_HIGH_ADLDS_FFL)')
    WIN32__ERROR_DS_HIGH_DSA_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x800721C2, 'HRESULT_FROM_WIN32(ERROR_DS_HIGH_DSA_VERSION)')
    WIN32__ERROR_DS_LOW_ADLDS_FFL: Final[ThirtyTwoBits] = cached_hresult(0x800721C3, 'HRESULT_FROM_WIN32(ERROR_DS_LOW_ADLDS_FFL)')
    WIN32__ERROR_DOMAIN_SID_SAME_AS_LOCAL_WORKSTATION: Final[ThirtyTwoBits] = cached_hresult(0x800721C4, 'HRESULT_FROM_WIN32(ERROR_DOMAIN_SID_SAME_AS_LOCAL_WORKSTATION)')
    WIN32__ERROR_DS_UNDELETE_SAM_VALIDATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800721C5, 'HRESULT_FROM_WIN32(ERROR_DS_UNDELETE_SAM_VALIDATION_FAILED)')
    WIN32__ERROR_INCORRECT_ACCOUNT_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x800721C6, 'HRESULT_FROM_WIN32(ERROR_INCORRECT_ACCOUNT_TYPE)')
    WIN32__ERROR_DS_SPN_VALUE_NOT_UNIQUE_IN_FOREST: Final[ThirtyTwoBits] = cached_hresult(0x800721C7, 'HRESULT_FROM_WIN32(ERROR_DS_SPN_VALUE_NOT_UNIQUE_IN_FOREST)')
    WIN32__ERROR_DS_UPN_VALUE_NOT_UNIQUE_IN_FOREST: Final[ThirtyTwoBits] = cached_hresult(0x800721C8, 'HRESULT_FROM_WIN32(ERROR_DS_UPN_VALUE_NOT_UNIQUE_IN_FOREST)')
    WIN32__ERROR_DS_MISSING_FOREST_TRUST: Final[ThirtyTwoBits] = cached_hresult(0x800721C9, 'HRESULT_FROM_WIN32(ERROR_DS_MISSING_FOREST_TRUST)')
    WIN32__ERROR_DS_VALUE_KEY_NOT_UNIQUE: Final[ThirtyTwoBits] = cached_hresult(0x800721CA, 'HRESULT_FROM_WIN32(ERROR_DS_VALUE_KEY_NOT_UNIQUE)')
    WIN32__ERROR_WEAK_WHFBKEY_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x800721CB, 'HRESULT_FROM_WIN32(ERROR_WEAK_WHFBKEY_BLOCKED)')
    WIN32__ERROR_DS_PER_ATTRIBUTE_AUTHZ_FAILED_DURING_ADD: Final[ThirtyTwoBits] = cached_hresult(0x800721CC, 'HRESULT_FROM_WIN32(ERROR_DS_PER_ATTRIBUTE_AUTHZ_FAILED_DURING_ADD)')
    WIN32__ERROR_LOCAL_POLICY_MODIFICATION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x800721CD, 'HRESULT_FROM_WIN32(ERROR_LOCAL_POLICY_MODIFICATION_NOT_SUPPORTED)')
    WIN32__ERROR_POLICY_CONTROLLED_ACCOUNT: Final[ThirtyTwoBits] = cached_hresult(0x800721CE, 'HRESULT_FROM_WIN32(ERROR_POLICY_CONTROLLED_ACCOUNT)')
    WIN32__ERROR_LAPS_LEGACY_SCHEMA_MISSING: Final[ThirtyTwoBits] = cached_hresult(0x800721CF, 'HRESULT_FROM_WIN32(ERROR_LAPS_LEGACY_SCHEMA_MISSING)')
    WIN32__ERROR_LAPS_SCHEMA_MISSING: Final[ThirtyTwoBits] = cached_hresult(0x800721D0, 'HRESULT_FROM_WIN32(ERROR_LAPS_SCHEMA_MISSING)')
    WIN32__ERROR_LAPS_ENCRYPTION_REQUIRES_2016_DFL: Final[ThirtyTwoBits] = cached_hresult(0x800721D1, 'HRESULT_FROM_WIN32(ERROR_LAPS_ENCRYPTION_REQUIRES_2016_DFL)')
    WIN32__DNS_ERROR_RCODE_FORMAT_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80072329, 'HRESULT_FROM_WIN32(DNS_ERROR_RCODE_FORMAT_ERROR)')
    WIN32__DNS_ERROR_RCODE_SERVER_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8007232A, 'HRESULT_FROM_WIN32(DNS_ERROR_RCODE_SERVER_FAILURE)')
    WIN32__DNS_ERROR_RCODE_NAME_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8007232B, 'HRESULT_FROM_WIN32(DNS_ERROR_RCODE_NAME_ERROR)')
    WIN32__DNS_ERROR_RCODE_NOT_IMPLEMENTED: Final[ThirtyTwoBits] = cached_hresult(0x8007232C, 'HRESULT_FROM_WIN32(DNS_ERROR_RCODE_NOT_IMPLEMENTED)')
    WIN32__DNS_ERROR_RCODE_REFUSED: Final[ThirtyTwoBits] = cached_hresult(0x8007232D, 'HRESULT_FROM_WIN32(DNS_ERROR_RCODE_REFUSED)')
    WIN32__DNS_ERROR_RCODE_YXDOMAIN: Final[ThirtyTwoBits] = cached_hresult(0x8007232E, 'HRESULT_FROM_WIN32(DNS_ERROR_RCODE_YXDOMAIN)')
    WIN32__DNS_ERROR_RCODE_YXRRSET: Final[ThirtyTwoBits] = cached_hresult(0x8007232F, 'HRESULT_FROM_WIN32(DNS_ERROR_RCODE_YXRRSET)')
    WIN32__DNS_ERROR_RCODE_NXRRSET: Final[ThirtyTwoBits] = cached_hresult(0x80072330, 'HRESULT_FROM_WIN32(DNS_ERROR_RCODE_NXRRSET)')
    WIN32__DNS_ERROR_RCODE_NOTAUTH: Final[ThirtyTwoBits] = cached_hresult(0x80072331, 'HRESULT_FROM_WIN32(DNS_ERROR_RCODE_NOTAUTH)')
    WIN32__DNS_ERROR_RCODE_NOTZONE: Final[ThirtyTwoBits] = cached_hresult(0x80072332, 'HRESULT_FROM_WIN32(DNS_ERROR_RCODE_NOTZONE)')
    WIN32__DNS_ERROR_RCODE_BADSIG: Final[ThirtyTwoBits] = cached_hresult(0x80072338, 'HRESULT_FROM_WIN32(DNS_ERROR_RCODE_BADSIG)')
    WIN32__DNS_ERROR_RCODE_BADKEY: Final[ThirtyTwoBits] = cached_hresult(0x80072339, 'HRESULT_FROM_WIN32(DNS_ERROR_RCODE_BADKEY)')
    WIN32__DNS_ERROR_RCODE_BADTIME: Final[ThirtyTwoBits] = cached_hresult(0x8007233A, 'HRESULT_FROM_WIN32(DNS_ERROR_RCODE_BADTIME)')
    WIN32__DNS_ERROR_KEYMASTER_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x8007238D, 'HRESULT_FROM_WIN32(DNS_ERROR_KEYMASTER_REQUIRED)')
    WIN32__DNS_ERROR_NOT_ALLOWED_ON_SIGNED_ZONE: Final[ThirtyTwoBits] = cached_hresult(0x8007238E, 'HRESULT_FROM_WIN32(DNS_ERROR_NOT_ALLOWED_ON_SIGNED_ZONE)')
    WIN32__DNS_ERROR_NSEC3_INCOMPATIBLE_WITH_RSA_SHA1: Final[ThirtyTwoBits] = cached_hresult(0x8007238F, 'HRESULT_FROM_WIN32(DNS_ERROR_NSEC3_INCOMPATIBLE_WITH_RSA_SHA1)')
    WIN32__DNS_ERROR_NOT_ENOUGH_SIGNING_KEY_DESCRIPTORS: Final[ThirtyTwoBits] = cached_hresult(0x80072390, 'HRESULT_FROM_WIN32(DNS_ERROR_NOT_ENOUGH_SIGNING_KEY_DESCRIPTORS)')
    WIN32__DNS_ERROR_UNSUPPORTED_ALGORITHM: Final[ThirtyTwoBits] = cached_hresult(0x80072391, 'HRESULT_FROM_WIN32(DNS_ERROR_UNSUPPORTED_ALGORITHM)')
    WIN32__DNS_ERROR_INVALID_KEY_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x80072392, 'HRESULT_FROM_WIN32(DNS_ERROR_INVALID_KEY_SIZE)')
    WIN32__DNS_ERROR_SIGNING_KEY_NOT_ACCESSIBLE: Final[ThirtyTwoBits] = cached_hresult(0x80072393, 'HRESULT_FROM_WIN32(DNS_ERROR_SIGNING_KEY_NOT_ACCESSIBLE)')
    WIN32__DNS_ERROR_KSP_DOES_NOT_SUPPORT_PROTECTION: Final[ThirtyTwoBits] = cached_hresult(0x80072394, 'HRESULT_FROM_WIN32(DNS_ERROR_KSP_DOES_NOT_SUPPORT_PROTECTION)')
    WIN32__DNS_ERROR_UNEXPECTED_DATA_PROTECTION_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80072395, 'HRESULT_FROM_WIN32(DNS_ERROR_UNEXPECTED_DATA_PROTECTION_ERROR)')
    WIN32__DNS_ERROR_UNEXPECTED_CNG_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80072396, 'HRESULT_FROM_WIN32(DNS_ERROR_UNEXPECTED_CNG_ERROR)')
    WIN32__DNS_ERROR_UNKNOWN_SIGNING_PARAMETER_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80072397, 'HRESULT_FROM_WIN32(DNS_ERROR_UNKNOWN_SIGNING_PARAMETER_VERSION)')
    WIN32__DNS_ERROR_KSP_NOT_ACCESSIBLE: Final[ThirtyTwoBits] = cached_hresult(0x80072398, 'HRESULT_FROM_WIN32(DNS_ERROR_KSP_NOT_ACCESSIBLE)')
    WIN32__DNS_ERROR_TOO_MANY_SKDS: Final[ThirtyTwoBits] = cached_hresult(0x80072399, 'HRESULT_FROM_WIN32(DNS_ERROR_TOO_MANY_SKDS)')
    WIN32__DNS_ERROR_INVALID_ROLLOVER_PERIOD: Final[ThirtyTwoBits] = cached_hresult(0x8007239A, 'HRESULT_FROM_WIN32(DNS_ERROR_INVALID_ROLLOVER_PERIOD)')
    WIN32__DNS_ERROR_INVALID_INITIAL_ROLLOVER_OFFSET: Final[ThirtyTwoBits] = cached_hresult(0x8007239B, 'HRESULT_FROM_WIN32(DNS_ERROR_INVALID_INITIAL_ROLLOVER_OFFSET)')
    WIN32__DNS_ERROR_ROLLOVER_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x8007239C, 'HRESULT_FROM_WIN32(DNS_ERROR_ROLLOVER_IN_PROGRESS)')
    WIN32__DNS_ERROR_STANDBY_KEY_NOT_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0x8007239D, 'HRESULT_FROM_WIN32(DNS_ERROR_STANDBY_KEY_NOT_PRESENT)')
    WIN32__DNS_ERROR_NOT_ALLOWED_ON_ZSK: Final[ThirtyTwoBits] = cached_hresult(0x8007239E, 'HRESULT_FROM_WIN32(DNS_ERROR_NOT_ALLOWED_ON_ZSK)')
    WIN32__DNS_ERROR_NOT_ALLOWED_ON_ACTIVE_SKD: Final[ThirtyTwoBits] = cached_hresult(0x8007239F, 'HRESULT_FROM_WIN32(DNS_ERROR_NOT_ALLOWED_ON_ACTIVE_SKD)')
    WIN32__DNS_ERROR_ROLLOVER_ALREADY_QUEUED: Final[ThirtyTwoBits] = cached_hresult(0x800723A0, 'HRESULT_FROM_WIN32(DNS_ERROR_ROLLOVER_ALREADY_QUEUED)')
    WIN32__DNS_ERROR_NOT_ALLOWED_ON_UNSIGNED_ZONE: Final[ThirtyTwoBits] = cached_hresult(0x800723A1, 'HRESULT_FROM_WIN32(DNS_ERROR_NOT_ALLOWED_ON_UNSIGNED_ZONE)')
    WIN32__DNS_ERROR_BAD_KEYMASTER: Final[ThirtyTwoBits] = cached_hresult(0x800723A2, 'HRESULT_FROM_WIN32(DNS_ERROR_BAD_KEYMASTER)')
    WIN32__DNS_ERROR_INVALID_SIGNATURE_VALIDITY_PERIOD: Final[ThirtyTwoBits] = cached_hresult(0x800723A3, 'HRESULT_FROM_WIN32(DNS_ERROR_INVALID_SIGNATURE_VALIDITY_PERIOD)')
    WIN32__DNS_ERROR_INVALID_NSEC3_ITERATION_COUNT: Final[ThirtyTwoBits] = cached_hresult(0x800723A4, 'HRESULT_FROM_WIN32(DNS_ERROR_INVALID_NSEC3_ITERATION_COUNT)')
    WIN32__DNS_ERROR_DNSSEC_IS_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x800723A5, 'HRESULT_FROM_WIN32(DNS_ERROR_DNSSEC_IS_DISABLED)')
    WIN32__DNS_ERROR_INVALID_XML: Final[ThirtyTwoBits] = cached_hresult(0x800723A6, 'HRESULT_FROM_WIN32(DNS_ERROR_INVALID_XML)')
    WIN32__DNS_ERROR_NO_VALID_TRUST_ANCHORS: Final[ThirtyTwoBits] = cached_hresult(0x800723A7, 'HRESULT_FROM_WIN32(DNS_ERROR_NO_VALID_TRUST_ANCHORS)')
    WIN32__DNS_ERROR_ROLLOVER_NOT_POKEABLE: Final[ThirtyTwoBits] = cached_hresult(0x800723A8, 'HRESULT_FROM_WIN32(DNS_ERROR_ROLLOVER_NOT_POKEABLE)')
    WIN32__DNS_ERROR_NSEC3_NAME_COLLISION: Final[ThirtyTwoBits] = cached_hresult(0x800723A9, 'HRESULT_FROM_WIN32(DNS_ERROR_NSEC3_NAME_COLLISION)')
    WIN32__DNS_ERROR_NSEC_INCOMPATIBLE_WITH_NSEC3_RSA_SHA1: Final[ThirtyTwoBits] = cached_hresult(0x800723AA, 'HRESULT_FROM_WIN32(DNS_ERROR_NSEC_INCOMPATIBLE_WITH_NSEC3_RSA_SHA1)')
    WIN32__DNS_INFO_NO_RECORDS: Final[ThirtyTwoBits] = cached_hresult(0x8007251D, 'HRESULT_FROM_WIN32(DNS_INFO_NO_RECORDS)')
    WIN32__DNS_ERROR_BAD_PACKET: Final[ThirtyTwoBits] = cached_hresult(0x8007251E, 'HRESULT_FROM_WIN32(DNS_ERROR_BAD_PACKET)')
    WIN32__DNS_ERROR_NO_PACKET: Final[ThirtyTwoBits] = cached_hresult(0x8007251F, 'HRESULT_FROM_WIN32(DNS_ERROR_NO_PACKET)')
    WIN32__DNS_ERROR_RCODE: Final[ThirtyTwoBits] = cached_hresult(0x80072520, 'HRESULT_FROM_WIN32(DNS_ERROR_RCODE)')
    WIN32__DNS_ERROR_UNSECURE_PACKET: Final[ThirtyTwoBits] = cached_hresult(0x80072521, 'HRESULT_FROM_WIN32(DNS_ERROR_UNSECURE_PACKET)')
    WIN32__DNS_REQUEST_PENDING: Final[ThirtyTwoBits] = cached_hresult(0x80072522, 'HRESULT_FROM_WIN32(DNS_REQUEST_PENDING)')
    WIN32__DNS_ERROR_INVALID_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x8007254F, 'HRESULT_FROM_WIN32(DNS_ERROR_INVALID_TYPE)')
    WIN32__DNS_ERROR_INVALID_IP_ADDRESS: Final[ThirtyTwoBits] = cached_hresult(0x80072550, 'HRESULT_FROM_WIN32(DNS_ERROR_INVALID_IP_ADDRESS)')
    WIN32__DNS_ERROR_INVALID_PROPERTY: Final[ThirtyTwoBits] = cached_hresult(0x80072551, 'HRESULT_FROM_WIN32(DNS_ERROR_INVALID_PROPERTY)')
    WIN32__DNS_ERROR_TRY_AGAIN_LATER: Final[ThirtyTwoBits] = cached_hresult(0x80072552, 'HRESULT_FROM_WIN32(DNS_ERROR_TRY_AGAIN_LATER)')
    WIN32__DNS_ERROR_NOT_UNIQUE: Final[ThirtyTwoBits] = cached_hresult(0x80072553, 'HRESULT_FROM_WIN32(DNS_ERROR_NOT_UNIQUE)')
    WIN32__DNS_ERROR_NON_RFC_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80072554, 'HRESULT_FROM_WIN32(DNS_ERROR_NON_RFC_NAME)')
    WIN32__DNS_STATUS_FQDN: Final[ThirtyTwoBits] = cached_hresult(0x80072555, 'HRESULT_FROM_WIN32(DNS_STATUS_FQDN)')
    WIN32__DNS_STATUS_DOTTED_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80072556, 'HRESULT_FROM_WIN32(DNS_STATUS_DOTTED_NAME)')
    WIN32__DNS_STATUS_SINGLE_PART_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80072557, 'HRESULT_FROM_WIN32(DNS_STATUS_SINGLE_PART_NAME)')
    WIN32__DNS_ERROR_INVALID_NAME_CHAR: Final[ThirtyTwoBits] = cached_hresult(0x80072558, 'HRESULT_FROM_WIN32(DNS_ERROR_INVALID_NAME_CHAR)')
    WIN32__DNS_ERROR_NUMERIC_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80072559, 'HRESULT_FROM_WIN32(DNS_ERROR_NUMERIC_NAME)')
    WIN32__DNS_ERROR_NOT_ALLOWED_ON_ROOT_SERVER: Final[ThirtyTwoBits] = cached_hresult(0x8007255A, 'HRESULT_FROM_WIN32(DNS_ERROR_NOT_ALLOWED_ON_ROOT_SERVER)')
    WIN32__DNS_ERROR_NOT_ALLOWED_UNDER_DELEGATION: Final[ThirtyTwoBits] = cached_hresult(0x8007255B, 'HRESULT_FROM_WIN32(DNS_ERROR_NOT_ALLOWED_UNDER_DELEGATION)')
    WIN32__DNS_ERROR_CANNOT_FIND_ROOT_HINTS: Final[ThirtyTwoBits] = cached_hresult(0x8007255C, 'HRESULT_FROM_WIN32(DNS_ERROR_CANNOT_FIND_ROOT_HINTS)')
    WIN32__DNS_ERROR_INCONSISTENT_ROOT_HINTS: Final[ThirtyTwoBits] = cached_hresult(0x8007255D, 'HRESULT_FROM_WIN32(DNS_ERROR_INCONSISTENT_ROOT_HINTS)')
    WIN32__DNS_ERROR_DWORD_VALUE_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0x8007255E, 'HRESULT_FROM_WIN32(DNS_ERROR_DWORD_VALUE_TOO_SMALL)')
    WIN32__DNS_ERROR_DWORD_VALUE_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0x8007255F, 'HRESULT_FROM_WIN32(DNS_ERROR_DWORD_VALUE_TOO_LARGE)')
    WIN32__DNS_ERROR_BACKGROUND_LOADING: Final[ThirtyTwoBits] = cached_hresult(0x80072560, 'HRESULT_FROM_WIN32(DNS_ERROR_BACKGROUND_LOADING)')
    WIN32__DNS_ERROR_NOT_ALLOWED_ON_RODC: Final[ThirtyTwoBits] = cached_hresult(0x80072561, 'HRESULT_FROM_WIN32(DNS_ERROR_NOT_ALLOWED_ON_RODC)')
    WIN32__DNS_ERROR_NOT_ALLOWED_UNDER_DNAME: Final[ThirtyTwoBits] = cached_hresult(0x80072562, 'HRESULT_FROM_WIN32(DNS_ERROR_NOT_ALLOWED_UNDER_DNAME)')
    WIN32__DNS_ERROR_DELEGATION_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80072563, 'HRESULT_FROM_WIN32(DNS_ERROR_DELEGATION_REQUIRED)')
    WIN32__DNS_ERROR_INVALID_POLICY_TABLE: Final[ThirtyTwoBits] = cached_hresult(0x80072564, 'HRESULT_FROM_WIN32(DNS_ERROR_INVALID_POLICY_TABLE)')
    WIN32__DNS_ERROR_ADDRESS_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80072565, 'HRESULT_FROM_WIN32(DNS_ERROR_ADDRESS_REQUIRED)')
    WIN32__DNS_ERROR_ZONE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x80072581, 'HRESULT_FROM_WIN32(DNS_ERROR_ZONE_DOES_NOT_EXIST)')
    WIN32__DNS_ERROR_NO_ZONE_INFO: Final[ThirtyTwoBits] = cached_hresult(0x80072582, 'HRESULT_FROM_WIN32(DNS_ERROR_NO_ZONE_INFO)')
    WIN32__DNS_ERROR_INVALID_ZONE_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x80072583, 'HRESULT_FROM_WIN32(DNS_ERROR_INVALID_ZONE_OPERATION)')
    WIN32__DNS_ERROR_ZONE_CONFIGURATION_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80072584, 'HRESULT_FROM_WIN32(DNS_ERROR_ZONE_CONFIGURATION_ERROR)')
    WIN32__DNS_ERROR_ZONE_HAS_NO_SOA_RECORD: Final[ThirtyTwoBits] = cached_hresult(0x80072585, 'HRESULT_FROM_WIN32(DNS_ERROR_ZONE_HAS_NO_SOA_RECORD)')
    WIN32__DNS_ERROR_ZONE_HAS_NO_NS_RECORDS: Final[ThirtyTwoBits] = cached_hresult(0x80072586, 'HRESULT_FROM_WIN32(DNS_ERROR_ZONE_HAS_NO_NS_RECORDS)')
    WIN32__DNS_ERROR_ZONE_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x80072587, 'HRESULT_FROM_WIN32(DNS_ERROR_ZONE_LOCKED)')
    WIN32__DNS_ERROR_ZONE_CREATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80072588, 'HRESULT_FROM_WIN32(DNS_ERROR_ZONE_CREATION_FAILED)')
    WIN32__DNS_ERROR_ZONE_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80072589, 'HRESULT_FROM_WIN32(DNS_ERROR_ZONE_ALREADY_EXISTS)')
    WIN32__DNS_ERROR_AUTOZONE_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x8007258A, 'HRESULT_FROM_WIN32(DNS_ERROR_AUTOZONE_ALREADY_EXISTS)')
    WIN32__DNS_ERROR_INVALID_ZONE_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x8007258B, 'HRESULT_FROM_WIN32(DNS_ERROR_INVALID_ZONE_TYPE)')
    WIN32__DNS_ERROR_SECONDARY_REQUIRES_MASTER_IP: Final[ThirtyTwoBits] = cached_hresult(0x8007258C, 'HRESULT_FROM_WIN32(DNS_ERROR_SECONDARY_REQUIRES_MASTER_IP)')
    WIN32__DNS_ERROR_ZONE_NOT_SECONDARY: Final[ThirtyTwoBits] = cached_hresult(0x8007258D, 'HRESULT_FROM_WIN32(DNS_ERROR_ZONE_NOT_SECONDARY)')
    WIN32__DNS_ERROR_NEED_SECONDARY_ADDRESSES: Final[ThirtyTwoBits] = cached_hresult(0x8007258E, 'HRESULT_FROM_WIN32(DNS_ERROR_NEED_SECONDARY_ADDRESSES)')
    WIN32__DNS_ERROR_WINS_INIT_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8007258F, 'HRESULT_FROM_WIN32(DNS_ERROR_WINS_INIT_FAILED)')
    WIN32__DNS_ERROR_NEED_WINS_SERVERS: Final[ThirtyTwoBits] = cached_hresult(0x80072590, 'HRESULT_FROM_WIN32(DNS_ERROR_NEED_WINS_SERVERS)')
    WIN32__DNS_ERROR_NBSTAT_INIT_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80072591, 'HRESULT_FROM_WIN32(DNS_ERROR_NBSTAT_INIT_FAILED)')
    WIN32__DNS_ERROR_SOA_DELETE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80072592, 'HRESULT_FROM_WIN32(DNS_ERROR_SOA_DELETE_INVALID)')
    WIN32__DNS_ERROR_FORWARDER_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80072593, 'HRESULT_FROM_WIN32(DNS_ERROR_FORWARDER_ALREADY_EXISTS)')
    WIN32__DNS_ERROR_ZONE_REQUIRES_MASTER_IP: Final[ThirtyTwoBits] = cached_hresult(0x80072594, 'HRESULT_FROM_WIN32(DNS_ERROR_ZONE_REQUIRES_MASTER_IP)')
    WIN32__DNS_ERROR_ZONE_IS_SHUTDOWN: Final[ThirtyTwoBits] = cached_hresult(0x80072595, 'HRESULT_FROM_WIN32(DNS_ERROR_ZONE_IS_SHUTDOWN)')
    WIN32__DNS_ERROR_ZONE_LOCKED_FOR_SIGNING: Final[ThirtyTwoBits] = cached_hresult(0x80072596, 'HRESULT_FROM_WIN32(DNS_ERROR_ZONE_LOCKED_FOR_SIGNING)')
    WIN32__DNS_ERROR_PRIMARY_REQUIRES_DATAFILE: Final[ThirtyTwoBits] = cached_hresult(0x800725B3, 'HRESULT_FROM_WIN32(DNS_ERROR_PRIMARY_REQUIRES_DATAFILE)')
    WIN32__DNS_ERROR_INVALID_DATAFILE_NAME: Final[ThirtyTwoBits] = cached_hresult(0x800725B4, 'HRESULT_FROM_WIN32(DNS_ERROR_INVALID_DATAFILE_NAME)')
    WIN32__DNS_ERROR_DATAFILE_OPEN_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x800725B5, 'HRESULT_FROM_WIN32(DNS_ERROR_DATAFILE_OPEN_FAILURE)')
    WIN32__DNS_ERROR_FILE_WRITEBACK_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800725B6, 'HRESULT_FROM_WIN32(DNS_ERROR_FILE_WRITEBACK_FAILED)')
    WIN32__DNS_ERROR_DATAFILE_PARSING: Final[ThirtyTwoBits] = cached_hresult(0x800725B7, 'HRESULT_FROM_WIN32(DNS_ERROR_DATAFILE_PARSING)')
    WIN32__DNS_ERROR_RECORD_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x800725E5, 'HRESULT_FROM_WIN32(DNS_ERROR_RECORD_DOES_NOT_EXIST)')
    WIN32__DNS_ERROR_RECORD_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0x800725E6, 'HRESULT_FROM_WIN32(DNS_ERROR_RECORD_FORMAT)')
    WIN32__DNS_ERROR_NODE_CREATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800725E7, 'HRESULT_FROM_WIN32(DNS_ERROR_NODE_CREATION_FAILED)')
    WIN32__DNS_ERROR_UNKNOWN_RECORD_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x800725E8, 'HRESULT_FROM_WIN32(DNS_ERROR_UNKNOWN_RECORD_TYPE)')
    WIN32__DNS_ERROR_RECORD_TIMED_OUT: Final[ThirtyTwoBits] = cached_hresult(0x800725E9, 'HRESULT_FROM_WIN32(DNS_ERROR_RECORD_TIMED_OUT)')
    WIN32__DNS_ERROR_NAME_NOT_IN_ZONE: Final[ThirtyTwoBits] = cached_hresult(0x800725EA, 'HRESULT_FROM_WIN32(DNS_ERROR_NAME_NOT_IN_ZONE)')
    WIN32__DNS_ERROR_CNAME_LOOP: Final[ThirtyTwoBits] = cached_hresult(0x800725EB, 'HRESULT_FROM_WIN32(DNS_ERROR_CNAME_LOOP)')
    WIN32__DNS_ERROR_NODE_IS_CNAME: Final[ThirtyTwoBits] = cached_hresult(0x800725EC, 'HRESULT_FROM_WIN32(DNS_ERROR_NODE_IS_CNAME)')
    WIN32__DNS_ERROR_CNAME_COLLISION: Final[ThirtyTwoBits] = cached_hresult(0x800725ED, 'HRESULT_FROM_WIN32(DNS_ERROR_CNAME_COLLISION)')
    WIN32__DNS_ERROR_RECORD_ONLY_AT_ZONE_ROOT: Final[ThirtyTwoBits] = cached_hresult(0x800725EE, 'HRESULT_FROM_WIN32(DNS_ERROR_RECORD_ONLY_AT_ZONE_ROOT)')
    WIN32__DNS_ERROR_RECORD_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800725EF, 'HRESULT_FROM_WIN32(DNS_ERROR_RECORD_ALREADY_EXISTS)')
    WIN32__DNS_ERROR_SECONDARY_DATA: Final[ThirtyTwoBits] = cached_hresult(0x800725F0, 'HRESULT_FROM_WIN32(DNS_ERROR_SECONDARY_DATA)')
    WIN32__DNS_ERROR_NO_CREATE_CACHE_DATA: Final[ThirtyTwoBits] = cached_hresult(0x800725F1, 'HRESULT_FROM_WIN32(DNS_ERROR_NO_CREATE_CACHE_DATA)')
    WIN32__DNS_ERROR_NAME_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x800725F2, 'HRESULT_FROM_WIN32(DNS_ERROR_NAME_DOES_NOT_EXIST)')
    WIN32__DNS_WARNING_PTR_CREATE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800725F3, 'HRESULT_FROM_WIN32(DNS_WARNING_PTR_CREATE_FAILED)')
    WIN32__DNS_WARNING_DOMAIN_UNDELETED: Final[ThirtyTwoBits] = cached_hresult(0x800725F4, 'HRESULT_FROM_WIN32(DNS_WARNING_DOMAIN_UNDELETED)')
    WIN32__DNS_ERROR_DS_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x800725F5, 'HRESULT_FROM_WIN32(DNS_ERROR_DS_UNAVAILABLE)')
    WIN32__DNS_ERROR_DS_ZONE_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800725F6, 'HRESULT_FROM_WIN32(DNS_ERROR_DS_ZONE_ALREADY_EXISTS)')
    WIN32__DNS_ERROR_NO_BOOTFILE_IF_DS_ZONE: Final[ThirtyTwoBits] = cached_hresult(0x800725F7, 'HRESULT_FROM_WIN32(DNS_ERROR_NO_BOOTFILE_IF_DS_ZONE)')
    WIN32__DNS_ERROR_NODE_IS_DNAME: Final[ThirtyTwoBits] = cached_hresult(0x800725F8, 'HRESULT_FROM_WIN32(DNS_ERROR_NODE_IS_DNAME)')
    WIN32__DNS_ERROR_DNAME_COLLISION: Final[ThirtyTwoBits] = cached_hresult(0x800725F9, 'HRESULT_FROM_WIN32(DNS_ERROR_DNAME_COLLISION)')
    WIN32__DNS_ERROR_ALIAS_LOOP: Final[ThirtyTwoBits] = cached_hresult(0x800725FA, 'HRESULT_FROM_WIN32(DNS_ERROR_ALIAS_LOOP)')
    WIN32__DNS_INFO_AXFR_COMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x80072617, 'HRESULT_FROM_WIN32(DNS_INFO_AXFR_COMPLETE)')
    WIN32__DNS_ERROR_AXFR: Final[ThirtyTwoBits] = cached_hresult(0x80072618, 'HRESULT_FROM_WIN32(DNS_ERROR_AXFR)')
    WIN32__DNS_INFO_ADDED_LOCAL_WINS: Final[ThirtyTwoBits] = cached_hresult(0x80072619, 'HRESULT_FROM_WIN32(DNS_INFO_ADDED_LOCAL_WINS)')
    WIN32__DNS_STATUS_CONTINUE_NEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80072649, 'HRESULT_FROM_WIN32(DNS_STATUS_CONTINUE_NEEDED)')
    WIN32__DNS_ERROR_NO_TCPIP: Final[ThirtyTwoBits] = cached_hresult(0x8007267B, 'HRESULT_FROM_WIN32(DNS_ERROR_NO_TCPIP)')
    WIN32__DNS_ERROR_NO_DNS_SERVERS: Final[ThirtyTwoBits] = cached_hresult(0x8007267C, 'HRESULT_FROM_WIN32(DNS_ERROR_NO_DNS_SERVERS)')
    WIN32__DNS_ERROR_DP_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x800726AD, 'HRESULT_FROM_WIN32(DNS_ERROR_DP_DOES_NOT_EXIST)')
    WIN32__DNS_ERROR_DP_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800726AE, 'HRESULT_FROM_WIN32(DNS_ERROR_DP_ALREADY_EXISTS)')
    WIN32__DNS_ERROR_DP_NOT_ENLISTED: Final[ThirtyTwoBits] = cached_hresult(0x800726AF, 'HRESULT_FROM_WIN32(DNS_ERROR_DP_NOT_ENLISTED)')
    WIN32__DNS_ERROR_DP_ALREADY_ENLISTED: Final[ThirtyTwoBits] = cached_hresult(0x800726B0, 'HRESULT_FROM_WIN32(DNS_ERROR_DP_ALREADY_ENLISTED)')
    WIN32__DNS_ERROR_DP_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x800726B1, 'HRESULT_FROM_WIN32(DNS_ERROR_DP_NOT_AVAILABLE)')
    WIN32__DNS_ERROR_DP_FSMO_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800726B2, 'HRESULT_FROM_WIN32(DNS_ERROR_DP_FSMO_ERROR)')
    WIN32__DNS_ERROR_RRL_NOT_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0x800726B7, 'HRESULT_FROM_WIN32(DNS_ERROR_RRL_NOT_ENABLED)')
    WIN32__DNS_ERROR_RRL_INVALID_WINDOW_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x800726B8, 'HRESULT_FROM_WIN32(DNS_ERROR_RRL_INVALID_WINDOW_SIZE)')
    WIN32__DNS_ERROR_RRL_INVALID_IPV4_PREFIX: Final[ThirtyTwoBits] = cached_hresult(0x800726B9, 'HRESULT_FROM_WIN32(DNS_ERROR_RRL_INVALID_IPV4_PREFIX)')
    WIN32__DNS_ERROR_RRL_INVALID_IPV6_PREFIX: Final[ThirtyTwoBits] = cached_hresult(0x800726BA, 'HRESULT_FROM_WIN32(DNS_ERROR_RRL_INVALID_IPV6_PREFIX)')
    WIN32__DNS_ERROR_RRL_INVALID_TC_RATE: Final[ThirtyTwoBits] = cached_hresult(0x800726BB, 'HRESULT_FROM_WIN32(DNS_ERROR_RRL_INVALID_TC_RATE)')
    WIN32__DNS_ERROR_RRL_INVALID_LEAK_RATE: Final[ThirtyTwoBits] = cached_hresult(0x800726BC, 'HRESULT_FROM_WIN32(DNS_ERROR_RRL_INVALID_LEAK_RATE)')
    WIN32__DNS_ERROR_RRL_LEAK_RATE_LESSTHAN_TC_RATE: Final[ThirtyTwoBits] = cached_hresult(0x800726BD, 'HRESULT_FROM_WIN32(DNS_ERROR_RRL_LEAK_RATE_LESSTHAN_TC_RATE)')
    WIN32__DNS_ERROR_VIRTUALIZATION_INSTANCE_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800726C1, 'HRESULT_FROM_WIN32(DNS_ERROR_VIRTUALIZATION_INSTANCE_ALREADY_EXISTS)')
    WIN32__DNS_ERROR_VIRTUALIZATION_INSTANCE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x800726C2, 'HRESULT_FROM_WIN32(DNS_ERROR_VIRTUALIZATION_INSTANCE_DOES_NOT_EXIST)')
    WIN32__DNS_ERROR_VIRTUALIZATION_TREE_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x800726C3, 'HRESULT_FROM_WIN32(DNS_ERROR_VIRTUALIZATION_TREE_LOCKED)')
    WIN32__DNS_ERROR_INVAILD_VIRTUALIZATION_INSTANCE_NAME: Final[ThirtyTwoBits] = cached_hresult(0x800726C4, 'HRESULT_FROM_WIN32(DNS_ERROR_INVAILD_VIRTUALIZATION_INSTANCE_NAME)')
    WIN32__DNS_ERROR_DEFAULT_VIRTUALIZATION_INSTANCE: Final[ThirtyTwoBits] = cached_hresult(0x800726C5, 'HRESULT_FROM_WIN32(DNS_ERROR_DEFAULT_VIRTUALIZATION_INSTANCE)')
    WIN32__DNS_ERROR_ZONESCOPE_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800726DF, 'HRESULT_FROM_WIN32(DNS_ERROR_ZONESCOPE_ALREADY_EXISTS)')
    WIN32__DNS_ERROR_ZONESCOPE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x800726E0, 'HRESULT_FROM_WIN32(DNS_ERROR_ZONESCOPE_DOES_NOT_EXIST)')
    WIN32__DNS_ERROR_DEFAULT_ZONESCOPE: Final[ThirtyTwoBits] = cached_hresult(0x800726E1, 'HRESULT_FROM_WIN32(DNS_ERROR_DEFAULT_ZONESCOPE)')
    WIN32__DNS_ERROR_INVALID_ZONESCOPE_NAME: Final[ThirtyTwoBits] = cached_hresult(0x800726E2, 'HRESULT_FROM_WIN32(DNS_ERROR_INVALID_ZONESCOPE_NAME)')
    WIN32__DNS_ERROR_NOT_ALLOWED_WITH_ZONESCOPES: Final[ThirtyTwoBits] = cached_hresult(0x800726E3, 'HRESULT_FROM_WIN32(DNS_ERROR_NOT_ALLOWED_WITH_ZONESCOPES)')
    WIN32__DNS_ERROR_LOAD_ZONESCOPE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800726E4, 'HRESULT_FROM_WIN32(DNS_ERROR_LOAD_ZONESCOPE_FAILED)')
    WIN32__DNS_ERROR_ZONESCOPE_FILE_WRITEBACK_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800726E5, 'HRESULT_FROM_WIN32(DNS_ERROR_ZONESCOPE_FILE_WRITEBACK_FAILED)')
    WIN32__DNS_ERROR_INVALID_SCOPE_NAME: Final[ThirtyTwoBits] = cached_hresult(0x800726E6, 'HRESULT_FROM_WIN32(DNS_ERROR_INVALID_SCOPE_NAME)')
    WIN32__DNS_ERROR_SCOPE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x800726E7, 'HRESULT_FROM_WIN32(DNS_ERROR_SCOPE_DOES_NOT_EXIST)')
    WIN32__DNS_ERROR_DEFAULT_SCOPE: Final[ThirtyTwoBits] = cached_hresult(0x800726E8, 'HRESULT_FROM_WIN32(DNS_ERROR_DEFAULT_SCOPE)')
    WIN32__DNS_ERROR_INVALID_SCOPE_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0x800726E9, 'HRESULT_FROM_WIN32(DNS_ERROR_INVALID_SCOPE_OPERATION)')
    WIN32__DNS_ERROR_SCOPE_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x800726EA, 'HRESULT_FROM_WIN32(DNS_ERROR_SCOPE_LOCKED)')
    WIN32__DNS_ERROR_SCOPE_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800726EB, 'HRESULT_FROM_WIN32(DNS_ERROR_SCOPE_ALREADY_EXISTS)')
    WIN32__DNS_ERROR_POLICY_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800726F3, 'HRESULT_FROM_WIN32(DNS_ERROR_POLICY_ALREADY_EXISTS)')
    WIN32__DNS_ERROR_POLICY_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x800726F4, 'HRESULT_FROM_WIN32(DNS_ERROR_POLICY_DOES_NOT_EXIST)')
    WIN32__DNS_ERROR_POLICY_INVALID_CRITERIA: Final[ThirtyTwoBits] = cached_hresult(0x800726F5, 'HRESULT_FROM_WIN32(DNS_ERROR_POLICY_INVALID_CRITERIA)')
    WIN32__DNS_ERROR_POLICY_INVALID_SETTINGS: Final[ThirtyTwoBits] = cached_hresult(0x800726F6, 'HRESULT_FROM_WIN32(DNS_ERROR_POLICY_INVALID_SETTINGS)')
    WIN32__DNS_ERROR_CLIENT_SUBNET_IS_ACCESSED: Final[ThirtyTwoBits] = cached_hresult(0x800726F7, 'HRESULT_FROM_WIN32(DNS_ERROR_CLIENT_SUBNET_IS_ACCESSED)')
    WIN32__DNS_ERROR_CLIENT_SUBNET_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x800726F8, 'HRESULT_FROM_WIN32(DNS_ERROR_CLIENT_SUBNET_DOES_NOT_EXIST)')
    WIN32__DNS_ERROR_CLIENT_SUBNET_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800726F9, 'HRESULT_FROM_WIN32(DNS_ERROR_CLIENT_SUBNET_ALREADY_EXISTS)')
    WIN32__DNS_ERROR_SUBNET_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x800726FA, 'HRESULT_FROM_WIN32(DNS_ERROR_SUBNET_DOES_NOT_EXIST)')
    WIN32__DNS_ERROR_SUBNET_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800726FB, 'HRESULT_FROM_WIN32(DNS_ERROR_SUBNET_ALREADY_EXISTS)')
    WIN32__DNS_ERROR_POLICY_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x800726FC, 'HRESULT_FROM_WIN32(DNS_ERROR_POLICY_LOCKED)')
    WIN32__DNS_ERROR_POLICY_INVALID_WEIGHT: Final[ThirtyTwoBits] = cached_hresult(0x800726FD, 'HRESULT_FROM_WIN32(DNS_ERROR_POLICY_INVALID_WEIGHT)')
    WIN32__DNS_ERROR_POLICY_INVALID_NAME: Final[ThirtyTwoBits] = cached_hresult(0x800726FE, 'HRESULT_FROM_WIN32(DNS_ERROR_POLICY_INVALID_NAME)')
    WIN32__DNS_ERROR_POLICY_MISSING_CRITERIA: Final[ThirtyTwoBits] = cached_hresult(0x800726FF, 'HRESULT_FROM_WIN32(DNS_ERROR_POLICY_MISSING_CRITERIA)')
    WIN32__DNS_ERROR_INVALID_CLIENT_SUBNET_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80072700, 'HRESULT_FROM_WIN32(DNS_ERROR_INVALID_CLIENT_SUBNET_NAME)')
    WIN32__DNS_ERROR_POLICY_PROCESSING_ORDER_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80072701, 'HRESULT_FROM_WIN32(DNS_ERROR_POLICY_PROCESSING_ORDER_INVALID)')
    WIN32__DNS_ERROR_POLICY_SCOPE_MISSING: Final[ThirtyTwoBits] = cached_hresult(0x80072702, 'HRESULT_FROM_WIN32(DNS_ERROR_POLICY_SCOPE_MISSING)')
    WIN32__DNS_ERROR_POLICY_SCOPE_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80072703, 'HRESULT_FROM_WIN32(DNS_ERROR_POLICY_SCOPE_NOT_ALLOWED)')
    WIN32__DNS_ERROR_SERVERSCOPE_IS_REFERENCED: Final[ThirtyTwoBits] = cached_hresult(0x80072704, 'HRESULT_FROM_WIN32(DNS_ERROR_SERVERSCOPE_IS_REFERENCED)')
    WIN32__DNS_ERROR_ZONESCOPE_IS_REFERENCED: Final[ThirtyTwoBits] = cached_hresult(0x80072705, 'HRESULT_FROM_WIN32(DNS_ERROR_ZONESCOPE_IS_REFERENCED)')
    WIN32__DNS_ERROR_POLICY_INVALID_CRITERIA_CLIENT_SUBNET: Final[ThirtyTwoBits] = cached_hresult(0x80072706, 'HRESULT_FROM_WIN32(DNS_ERROR_POLICY_INVALID_CRITERIA_CLIENT_SUBNET)')
    WIN32__DNS_ERROR_POLICY_INVALID_CRITERIA_TRANSPORT_PROTOCOL: Final[ThirtyTwoBits] = cached_hresult(0x80072707, 'HRESULT_FROM_WIN32(DNS_ERROR_POLICY_INVALID_CRITERIA_TRANSPORT_PROTOCOL)')
    WIN32__DNS_ERROR_POLICY_INVALID_CRITERIA_NETWORK_PROTOCOL: Final[ThirtyTwoBits] = cached_hresult(0x80072708, 'HRESULT_FROM_WIN32(DNS_ERROR_POLICY_INVALID_CRITERIA_NETWORK_PROTOCOL)')
    WIN32__DNS_ERROR_POLICY_INVALID_CRITERIA_INTERFACE: Final[ThirtyTwoBits] = cached_hresult(0x80072709, 'HRESULT_FROM_WIN32(DNS_ERROR_POLICY_INVALID_CRITERIA_INTERFACE)')
    WIN32__DNS_ERROR_POLICY_INVALID_CRITERIA_FQDN: Final[ThirtyTwoBits] = cached_hresult(0x8007270A, 'HRESULT_FROM_WIN32(DNS_ERROR_POLICY_INVALID_CRITERIA_FQDN)')
    WIN32__DNS_ERROR_POLICY_INVALID_CRITERIA_QUERY_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x8007270B, 'HRESULT_FROM_WIN32(DNS_ERROR_POLICY_INVALID_CRITERIA_QUERY_TYPE)')
    WIN32__DNS_ERROR_POLICY_INVALID_CRITERIA_TIME_OF_DAY: Final[ThirtyTwoBits] = cached_hresult(0x8007270C, 'HRESULT_FROM_WIN32(DNS_ERROR_POLICY_INVALID_CRITERIA_TIME_OF_DAY)')
    WIN32__WSAEINTR: Final[ThirtyTwoBits] = cached_hresult(0x80072714, 'HRESULT_FROM_WIN32(WSAEINTR)')
    WIN32__WSAEBADF: Final[ThirtyTwoBits] = cached_hresult(0x80072719, 'HRESULT_FROM_WIN32(WSAEBADF)')
    WIN32__WSAEACCES: Final[ThirtyTwoBits] = cached_hresult(0x8007271D, 'HRESULT_FROM_WIN32(WSAEACCES)')
    WIN32__WSAEFAULT: Final[ThirtyTwoBits] = cached_hresult(0x8007271E, 'HRESULT_FROM_WIN32(WSAEFAULT)')
    WIN32__WSAEINVAL: Final[ThirtyTwoBits] = cached_hresult(0x80072726, 'HRESULT_FROM_WIN32(WSAEINVAL)')
    WIN32__WSAEMFILE: Final[ThirtyTwoBits] = cached_hresult(0x80072728, 'HRESULT_FROM_WIN32(WSAEMFILE)')
    WIN32__WSAEWOULDBLOCK: Final[ThirtyTwoBits] = cached_hresult(0x80072733, 'HRESULT_FROM_WIN32(WSAEWOULDBLOCK)')
    WIN32__WSAEINPROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x80072734, 'HRESULT_FROM_WIN32(WSAEINPROGRESS)')
    WIN32__WSAEALREADY: Final[ThirtyTwoBits] = cached_hresult(0x80072735, 'HRESULT_FROM_WIN32(WSAEALREADY)')
    WIN32__WSAENOTSOCK: Final[ThirtyTwoBits] = cached_hresult(0x80072736, 'HRESULT_FROM_WIN32(WSAENOTSOCK)')
    WIN32__WSAEDESTADDRREQ: Final[ThirtyTwoBits] = cached_hresult(0x80072737, 'HRESULT_FROM_WIN32(WSAEDESTADDRREQ)')
    WIN32__WSAEMSGSIZE: Final[ThirtyTwoBits] = cached_hresult(0x80072738, 'HRESULT_FROM_WIN32(WSAEMSGSIZE)')
    WIN32__WSAEPROTOTYPE: Final[ThirtyTwoBits] = cached_hresult(0x80072739, 'HRESULT_FROM_WIN32(WSAEPROTOTYPE)')
    WIN32__WSAENOPROTOOPT: Final[ThirtyTwoBits] = cached_hresult(0x8007273A, 'HRESULT_FROM_WIN32(WSAENOPROTOOPT)')
    WIN32__WSAEPROTONOSUPPORT: Final[ThirtyTwoBits] = cached_hresult(0x8007273B, 'HRESULT_FROM_WIN32(WSAEPROTONOSUPPORT)')
    WIN32__WSAESOCKTNOSUPPORT: Final[ThirtyTwoBits] = cached_hresult(0x8007273C, 'HRESULT_FROM_WIN32(WSAESOCKTNOSUPPORT)')
    WIN32__WSAEOPNOTSUPP: Final[ThirtyTwoBits] = cached_hresult(0x8007273D, 'HRESULT_FROM_WIN32(WSAEOPNOTSUPP)')
    WIN32__WSAEPFNOSUPPORT: Final[ThirtyTwoBits] = cached_hresult(0x8007273E, 'HRESULT_FROM_WIN32(WSAEPFNOSUPPORT)')
    WIN32__WSAEAFNOSUPPORT: Final[ThirtyTwoBits] = cached_hresult(0x8007273F, 'HRESULT_FROM_WIN32(WSAEAFNOSUPPORT)')
    WIN32__WSAEADDRINUSE: Final[ThirtyTwoBits] = cached_hresult(0x80072740, 'HRESULT_FROM_WIN32(WSAEADDRINUSE)')
    WIN32__WSAEADDRNOTAVAIL: Final[ThirtyTwoBits] = cached_hresult(0x80072741, 'HRESULT_FROM_WIN32(WSAEADDRNOTAVAIL)')
    WIN32__WSAENETDOWN: Final[ThirtyTwoBits] = cached_hresult(0x80072742, 'HRESULT_FROM_WIN32(WSAENETDOWN)')
    WIN32__WSAENETUNREACH: Final[ThirtyTwoBits] = cached_hresult(0x80072743, 'HRESULT_FROM_WIN32(WSAENETUNREACH)')
    WIN32__WSAENETRESET: Final[ThirtyTwoBits] = cached_hresult(0x80072744, 'HRESULT_FROM_WIN32(WSAENETRESET)')
    WIN32__WSAECONNABORTED: Final[ThirtyTwoBits] = cached_hresult(0x80072745, 'HRESULT_FROM_WIN32(WSAECONNABORTED)')
    WIN32__WSAECONNRESET: Final[ThirtyTwoBits] = cached_hresult(0x80072746, 'HRESULT_FROM_WIN32(WSAECONNRESET)')
    WIN32__WSAENOBUFS: Final[ThirtyTwoBits] = cached_hresult(0x80072747, 'HRESULT_FROM_WIN32(WSAENOBUFS)')
    WIN32__WSAEISCONN: Final[ThirtyTwoBits] = cached_hresult(0x80072748, 'HRESULT_FROM_WIN32(WSAEISCONN)')
    WIN32__WSAENOTCONN: Final[ThirtyTwoBits] = cached_hresult(0x80072749, 'HRESULT_FROM_WIN32(WSAENOTCONN)')
    WIN32__WSAESHUTDOWN: Final[ThirtyTwoBits] = cached_hresult(0x8007274A, 'HRESULT_FROM_WIN32(WSAESHUTDOWN)')
    WIN32__WSAETOOMANYREFS: Final[ThirtyTwoBits] = cached_hresult(0x8007274B, 'HRESULT_FROM_WIN32(WSAETOOMANYREFS)')
    WIN32__WSAETIMEDOUT: Final[ThirtyTwoBits] = cached_hresult(0x8007274C, 'HRESULT_FROM_WIN32(WSAETIMEDOUT)')
    WIN32__WSAECONNREFUSED: Final[ThirtyTwoBits] = cached_hresult(0x8007274D, 'HRESULT_FROM_WIN32(WSAECONNREFUSED)')
    WIN32__WSAELOOP: Final[ThirtyTwoBits] = cached_hresult(0x8007274E, 'HRESULT_FROM_WIN32(WSAELOOP)')
    WIN32__WSAENAMETOOLONG: Final[ThirtyTwoBits] = cached_hresult(0x8007274F, 'HRESULT_FROM_WIN32(WSAENAMETOOLONG)')
    WIN32__WSAEHOSTDOWN: Final[ThirtyTwoBits] = cached_hresult(0x80072750, 'HRESULT_FROM_WIN32(WSAEHOSTDOWN)')
    WIN32__WSAEHOSTUNREACH: Final[ThirtyTwoBits] = cached_hresult(0x80072751, 'HRESULT_FROM_WIN32(WSAEHOSTUNREACH)')
    WIN32__WSAENOTEMPTY: Final[ThirtyTwoBits] = cached_hresult(0x80072752, 'HRESULT_FROM_WIN32(WSAENOTEMPTY)')
    WIN32__WSAEPROCLIM: Final[ThirtyTwoBits] = cached_hresult(0x80072753, 'HRESULT_FROM_WIN32(WSAEPROCLIM)')
    WIN32__WSAEUSERS: Final[ThirtyTwoBits] = cached_hresult(0x80072754, 'HRESULT_FROM_WIN32(WSAEUSERS)')
    WIN32__WSAEDQUOT: Final[ThirtyTwoBits] = cached_hresult(0x80072755, 'HRESULT_FROM_WIN32(WSAEDQUOT)')
    WIN32__WSAESTALE: Final[ThirtyTwoBits] = cached_hresult(0x80072756, 'HRESULT_FROM_WIN32(WSAESTALE)')
    WIN32__WSAEREMOTE: Final[ThirtyTwoBits] = cached_hresult(0x80072757, 'HRESULT_FROM_WIN32(WSAEREMOTE)')
    WIN32__WSASYSNOTREADY: Final[ThirtyTwoBits] = cached_hresult(0x8007276B, 'HRESULT_FROM_WIN32(WSASYSNOTREADY)')
    WIN32__WSAVERNOTSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x8007276C, 'HRESULT_FROM_WIN32(WSAVERNOTSUPPORTED)')
    WIN32__WSANOTINITIALISED: Final[ThirtyTwoBits] = cached_hresult(0x8007276D, 'HRESULT_FROM_WIN32(WSANOTINITIALISED)')
    WIN32__WSAEDISCON: Final[ThirtyTwoBits] = cached_hresult(0x80072775, 'HRESULT_FROM_WIN32(WSAEDISCON)')
    WIN32__WSAENOMORE: Final[ThirtyTwoBits] = cached_hresult(0x80072776, 'HRESULT_FROM_WIN32(WSAENOMORE)')
    WIN32__WSAECANCELLED: Final[ThirtyTwoBits] = cached_hresult(0x80072777, 'HRESULT_FROM_WIN32(WSAECANCELLED)')
    WIN32__WSAEINVALIDPROCTABLE: Final[ThirtyTwoBits] = cached_hresult(0x80072778, 'HRESULT_FROM_WIN32(WSAEINVALIDPROCTABLE)')
    WIN32__WSAEINVALIDPROVIDER: Final[ThirtyTwoBits] = cached_hresult(0x80072779, 'HRESULT_FROM_WIN32(WSAEINVALIDPROVIDER)')
    WIN32__WSAEPROVIDERFAILEDINIT: Final[ThirtyTwoBits] = cached_hresult(0x8007277A, 'HRESULT_FROM_WIN32(WSAEPROVIDERFAILEDINIT)')
    WIN32__WSASYSCALLFAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8007277B, 'HRESULT_FROM_WIN32(WSASYSCALLFAILURE)')
    WIN32__WSASERVICE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8007277C, 'HRESULT_FROM_WIN32(WSASERVICE_NOT_FOUND)')
    WIN32__WSATYPE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x8007277D, 'HRESULT_FROM_WIN32(WSATYPE_NOT_FOUND)')
    WIN32__WSA_E_NO_MORE: Final[ThirtyTwoBits] = cached_hresult(0x8007277E, 'HRESULT_FROM_WIN32(WSA_E_NO_MORE)')
    WIN32__WSA_E_CANCELLED: Final[ThirtyTwoBits] = cached_hresult(0x8007277F, 'HRESULT_FROM_WIN32(WSA_E_CANCELLED)')
    WIN32__WSAEREFUSED: Final[ThirtyTwoBits] = cached_hresult(0x80072780, 'HRESULT_FROM_WIN32(WSAEREFUSED)')
    WIN32__WSAHOST_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80072AF9, 'HRESULT_FROM_WIN32(WSAHOST_NOT_FOUND)')
    WIN32__WSATRY_AGAIN: Final[ThirtyTwoBits] = cached_hresult(0x80072AFA, 'HRESULT_FROM_WIN32(WSATRY_AGAIN)')
    WIN32__WSANO_RECOVERY: Final[ThirtyTwoBits] = cached_hresult(0x80072AFB, 'HRESULT_FROM_WIN32(WSANO_RECOVERY)')
    WIN32__WSANO_DATA: Final[ThirtyTwoBits] = cached_hresult(0x80072AFC, 'HRESULT_FROM_WIN32(WSANO_DATA)')
    WIN32__WSA_QOS_RECEIVERS: Final[ThirtyTwoBits] = cached_hresult(0x80072AFD, 'HRESULT_FROM_WIN32(WSA_QOS_RECEIVERS)')
    WIN32__WSA_QOS_SENDERS: Final[ThirtyTwoBits] = cached_hresult(0x80072AFE, 'HRESULT_FROM_WIN32(WSA_QOS_SENDERS)')
    WIN32__WSA_QOS_NO_SENDERS: Final[ThirtyTwoBits] = cached_hresult(0x80072AFF, 'HRESULT_FROM_WIN32(WSA_QOS_NO_SENDERS)')
    WIN32__WSA_QOS_NO_RECEIVERS: Final[ThirtyTwoBits] = cached_hresult(0x80072B00, 'HRESULT_FROM_WIN32(WSA_QOS_NO_RECEIVERS)')
    WIN32__WSA_QOS_REQUEST_CONFIRMED: Final[ThirtyTwoBits] = cached_hresult(0x80072B01, 'HRESULT_FROM_WIN32(WSA_QOS_REQUEST_CONFIRMED)')
    WIN32__WSA_QOS_ADMISSION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80072B02, 'HRESULT_FROM_WIN32(WSA_QOS_ADMISSION_FAILURE)')
    WIN32__WSA_QOS_POLICY_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80072B03, 'HRESULT_FROM_WIN32(WSA_QOS_POLICY_FAILURE)')
    WIN32__WSA_QOS_BAD_STYLE: Final[ThirtyTwoBits] = cached_hresult(0x80072B04, 'HRESULT_FROM_WIN32(WSA_QOS_BAD_STYLE)')
    WIN32__WSA_QOS_BAD_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x80072B05, 'HRESULT_FROM_WIN32(WSA_QOS_BAD_OBJECT)')
    WIN32__WSA_QOS_TRAFFIC_CTRL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80072B06, 'HRESULT_FROM_WIN32(WSA_QOS_TRAFFIC_CTRL_ERROR)')
    WIN32__WSA_QOS_GENERIC_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80072B07, 'HRESULT_FROM_WIN32(WSA_QOS_GENERIC_ERROR)')
    WIN32__WSA_QOS_ESERVICETYPE: Final[ThirtyTwoBits] = cached_hresult(0x80072B08, 'HRESULT_FROM_WIN32(WSA_QOS_ESERVICETYPE)')
    WIN32__WSA_QOS_EFLOWSPEC: Final[ThirtyTwoBits] = cached_hresult(0x80072B09, 'HRESULT_FROM_WIN32(WSA_QOS_EFLOWSPEC)')
    WIN32__WSA_QOS_EPROVSPECBUF: Final[ThirtyTwoBits] = cached_hresult(0x80072B0A, 'HRESULT_FROM_WIN32(WSA_QOS_EPROVSPECBUF)')
    WIN32__WSA_QOS_EFILTERSTYLE: Final[ThirtyTwoBits] = cached_hresult(0x80072B0B, 'HRESULT_FROM_WIN32(WSA_QOS_EFILTERSTYLE)')
    WIN32__WSA_QOS_EFILTERTYPE: Final[ThirtyTwoBits] = cached_hresult(0x80072B0C, 'HRESULT_FROM_WIN32(WSA_QOS_EFILTERTYPE)')
    WIN32__WSA_QOS_EFILTERCOUNT: Final[ThirtyTwoBits] = cached_hresult(0x80072B0D, 'HRESULT_FROM_WIN32(WSA_QOS_EFILTERCOUNT)')
    WIN32__WSA_QOS_EOBJLENGTH: Final[ThirtyTwoBits] = cached_hresult(0x80072B0E, 'HRESULT_FROM_WIN32(WSA_QOS_EOBJLENGTH)')
    WIN32__WSA_QOS_EFLOWCOUNT: Final[ThirtyTwoBits] = cached_hresult(0x80072B0F, 'HRESULT_FROM_WIN32(WSA_QOS_EFLOWCOUNT)')
    WIN32__WSA_QOS_EUNKOWNPSOBJ: Final[ThirtyTwoBits] = cached_hresult(0x80072B10, 'HRESULT_FROM_WIN32(WSA_QOS_EUNKOWNPSOBJ)')
    WIN32__WSA_QOS_EPOLICYOBJ: Final[ThirtyTwoBits] = cached_hresult(0x80072B11, 'HRESULT_FROM_WIN32(WSA_QOS_EPOLICYOBJ)')
    WIN32__WSA_QOS_EFLOWDESC: Final[ThirtyTwoBits] = cached_hresult(0x80072B12, 'HRESULT_FROM_WIN32(WSA_QOS_EFLOWDESC)')
    WIN32__WSA_QOS_EPSFLOWSPEC: Final[ThirtyTwoBits] = cached_hresult(0x80072B13, 'HRESULT_FROM_WIN32(WSA_QOS_EPSFLOWSPEC)')
    WIN32__WSA_QOS_EPSFILTERSPEC: Final[ThirtyTwoBits] = cached_hresult(0x80072B14, 'HRESULT_FROM_WIN32(WSA_QOS_EPSFILTERSPEC)')
    WIN32__WSA_QOS_ESDMODEOBJ: Final[ThirtyTwoBits] = cached_hresult(0x80072B15, 'HRESULT_FROM_WIN32(WSA_QOS_ESDMODEOBJ)')
    WIN32__WSA_QOS_ESHAPERATEOBJ: Final[ThirtyTwoBits] = cached_hresult(0x80072B16, 'HRESULT_FROM_WIN32(WSA_QOS_ESHAPERATEOBJ)')
    WIN32__WSA_QOS_RESERVED_PETYPE: Final[ThirtyTwoBits] = cached_hresult(0x80072B17, 'HRESULT_FROM_WIN32(WSA_QOS_RESERVED_PETYPE)')
    WIN32__WSA_SECURE_HOST_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80072B18, 'HRESULT_FROM_WIN32(WSA_SECURE_HOST_NOT_FOUND)')
    WIN32__WSA_IPSEC_NAME_POLICY_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80072B19, 'HRESULT_FROM_WIN32(WSA_IPSEC_NAME_POLICY_ERROR)')
    WIN32__ERROR_IPSEC_QM_POLICY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800732C8, 'HRESULT_FROM_WIN32(ERROR_IPSEC_QM_POLICY_EXISTS)')
    WIN32__ERROR_IPSEC_QM_POLICY_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800732C9, 'HRESULT_FROM_WIN32(ERROR_IPSEC_QM_POLICY_NOT_FOUND)')
    WIN32__ERROR_IPSEC_QM_POLICY_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x800732CA, 'HRESULT_FROM_WIN32(ERROR_IPSEC_QM_POLICY_IN_USE)')
    WIN32__ERROR_IPSEC_MM_POLICY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800732CB, 'HRESULT_FROM_WIN32(ERROR_IPSEC_MM_POLICY_EXISTS)')
    WIN32__ERROR_IPSEC_MM_POLICY_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800732CC, 'HRESULT_FROM_WIN32(ERROR_IPSEC_MM_POLICY_NOT_FOUND)')
    WIN32__ERROR_IPSEC_MM_POLICY_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x800732CD, 'HRESULT_FROM_WIN32(ERROR_IPSEC_MM_POLICY_IN_USE)')
    WIN32__ERROR_IPSEC_MM_FILTER_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800732CE, 'HRESULT_FROM_WIN32(ERROR_IPSEC_MM_FILTER_EXISTS)')
    WIN32__ERROR_IPSEC_MM_FILTER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800732CF, 'HRESULT_FROM_WIN32(ERROR_IPSEC_MM_FILTER_NOT_FOUND)')
    WIN32__ERROR_IPSEC_TRANSPORT_FILTER_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800732D0, 'HRESULT_FROM_WIN32(ERROR_IPSEC_TRANSPORT_FILTER_EXISTS)')
    WIN32__ERROR_IPSEC_TRANSPORT_FILTER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800732D1, 'HRESULT_FROM_WIN32(ERROR_IPSEC_TRANSPORT_FILTER_NOT_FOUND)')
    WIN32__ERROR_IPSEC_MM_AUTH_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800732D2, 'HRESULT_FROM_WIN32(ERROR_IPSEC_MM_AUTH_EXISTS)')
    WIN32__ERROR_IPSEC_MM_AUTH_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800732D3, 'HRESULT_FROM_WIN32(ERROR_IPSEC_MM_AUTH_NOT_FOUND)')
    WIN32__ERROR_IPSEC_MM_AUTH_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x800732D4, 'HRESULT_FROM_WIN32(ERROR_IPSEC_MM_AUTH_IN_USE)')
    WIN32__ERROR_IPSEC_DEFAULT_MM_POLICY_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800732D5, 'HRESULT_FROM_WIN32(ERROR_IPSEC_DEFAULT_MM_POLICY_NOT_FOUND)')
    WIN32__ERROR_IPSEC_DEFAULT_MM_AUTH_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800732D6, 'HRESULT_FROM_WIN32(ERROR_IPSEC_DEFAULT_MM_AUTH_NOT_FOUND)')
    WIN32__ERROR_IPSEC_DEFAULT_QM_POLICY_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800732D7, 'HRESULT_FROM_WIN32(ERROR_IPSEC_DEFAULT_QM_POLICY_NOT_FOUND)')
    WIN32__ERROR_IPSEC_TUNNEL_FILTER_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x800732D8, 'HRESULT_FROM_WIN32(ERROR_IPSEC_TUNNEL_FILTER_EXISTS)')
    WIN32__ERROR_IPSEC_TUNNEL_FILTER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800732D9, 'HRESULT_FROM_WIN32(ERROR_IPSEC_TUNNEL_FILTER_NOT_FOUND)')
    WIN32__ERROR_IPSEC_MM_FILTER_PENDING_DELETION: Final[ThirtyTwoBits] = cached_hresult(0x800732DA, 'HRESULT_FROM_WIN32(ERROR_IPSEC_MM_FILTER_PENDING_DELETION)')
    WIN32__ERROR_IPSEC_TRANSPORT_FILTER_PENDING_DELETION: Final[ThirtyTwoBits] = cached_hresult(0x800732DB, 'HRESULT_FROM_WIN32(ERROR_IPSEC_TRANSPORT_FILTER_PENDING_DELETION)')
    WIN32__ERROR_IPSEC_TUNNEL_FILTER_PENDING_DELETION: Final[ThirtyTwoBits] = cached_hresult(0x800732DC, 'HRESULT_FROM_WIN32(ERROR_IPSEC_TUNNEL_FILTER_PENDING_DELETION)')
    WIN32__ERROR_IPSEC_MM_POLICY_PENDING_DELETION: Final[ThirtyTwoBits] = cached_hresult(0x800732DD, 'HRESULT_FROM_WIN32(ERROR_IPSEC_MM_POLICY_PENDING_DELETION)')
    WIN32__ERROR_IPSEC_MM_AUTH_PENDING_DELETION: Final[ThirtyTwoBits] = cached_hresult(0x800732DE, 'HRESULT_FROM_WIN32(ERROR_IPSEC_MM_AUTH_PENDING_DELETION)')
    WIN32__ERROR_IPSEC_QM_POLICY_PENDING_DELETION: Final[ThirtyTwoBits] = cached_hresult(0x800732DF, 'HRESULT_FROM_WIN32(ERROR_IPSEC_QM_POLICY_PENDING_DELETION)')
    WIN32__WARNING_IPSEC_MM_POLICY_PRUNED: Final[ThirtyTwoBits] = cached_hresult(0x800732E0, 'HRESULT_FROM_WIN32(WARNING_IPSEC_MM_POLICY_PRUNED)')
    WIN32__WARNING_IPSEC_QM_POLICY_PRUNED: Final[ThirtyTwoBits] = cached_hresult(0x800732E1, 'HRESULT_FROM_WIN32(WARNING_IPSEC_QM_POLICY_PRUNED)')
    WIN32__ERROR_IPSEC_IKE_NEG_STATUS_BEGIN: Final[ThirtyTwoBits] = cached_hresult(0x800735E8, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_NEG_STATUS_BEGIN)')
    WIN32__ERROR_IPSEC_IKE_AUTH_FAIL: Final[ThirtyTwoBits] = cached_hresult(0x800735E9, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_AUTH_FAIL)')
    WIN32__ERROR_IPSEC_IKE_ATTRIB_FAIL: Final[ThirtyTwoBits] = cached_hresult(0x800735EA, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_ATTRIB_FAIL)')
    WIN32__ERROR_IPSEC_IKE_NEGOTIATION_PENDING: Final[ThirtyTwoBits] = cached_hresult(0x800735EB, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_NEGOTIATION_PENDING)')
    WIN32__ERROR_IPSEC_IKE_GENERAL_PROCESSING_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800735EC, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_GENERAL_PROCESSING_ERROR)')
    WIN32__ERROR_IPSEC_IKE_TIMED_OUT: Final[ThirtyTwoBits] = cached_hresult(0x800735ED, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_TIMED_OUT)')
    WIN32__ERROR_IPSEC_IKE_NO_CERT: Final[ThirtyTwoBits] = cached_hresult(0x800735EE, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_NO_CERT)')
    WIN32__ERROR_IPSEC_IKE_SA_DELETED: Final[ThirtyTwoBits] = cached_hresult(0x800735EF, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_SA_DELETED)')
    WIN32__ERROR_IPSEC_IKE_SA_REAPED: Final[ThirtyTwoBits] = cached_hresult(0x800735F0, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_SA_REAPED)')
    WIN32__ERROR_IPSEC_IKE_MM_ACQUIRE_DROP: Final[ThirtyTwoBits] = cached_hresult(0x800735F1, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_MM_ACQUIRE_DROP)')
    WIN32__ERROR_IPSEC_IKE_QM_ACQUIRE_DROP: Final[ThirtyTwoBits] = cached_hresult(0x800735F2, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_QM_ACQUIRE_DROP)')
    WIN32__ERROR_IPSEC_IKE_QUEUE_DROP_MM: Final[ThirtyTwoBits] = cached_hresult(0x800735F3, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_QUEUE_DROP_MM)')
    WIN32__ERROR_IPSEC_IKE_QUEUE_DROP_NO_MM: Final[ThirtyTwoBits] = cached_hresult(0x800735F4, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_QUEUE_DROP_NO_MM)')
    WIN32__ERROR_IPSEC_IKE_DROP_NO_RESPONSE: Final[ThirtyTwoBits] = cached_hresult(0x800735F5, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_DROP_NO_RESPONSE)')
    WIN32__ERROR_IPSEC_IKE_MM_DELAY_DROP: Final[ThirtyTwoBits] = cached_hresult(0x800735F6, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_MM_DELAY_DROP)')
    WIN32__ERROR_IPSEC_IKE_QM_DELAY_DROP: Final[ThirtyTwoBits] = cached_hresult(0x800735F7, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_QM_DELAY_DROP)')
    WIN32__ERROR_IPSEC_IKE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800735F8, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_ERROR)')
    WIN32__ERROR_IPSEC_IKE_CRL_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800735F9, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_CRL_FAILED)')
    WIN32__ERROR_IPSEC_IKE_INVALID_KEY_USAGE: Final[ThirtyTwoBits] = cached_hresult(0x800735FA, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INVALID_KEY_USAGE)')
    WIN32__ERROR_IPSEC_IKE_INVALID_CERT_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x800735FB, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INVALID_CERT_TYPE)')
    WIN32__ERROR_IPSEC_IKE_NO_PRIVATE_KEY: Final[ThirtyTwoBits] = cached_hresult(0x800735FC, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_NO_PRIVATE_KEY)')
    WIN32__ERROR_IPSEC_IKE_SIMULTANEOUS_REKEY: Final[ThirtyTwoBits] = cached_hresult(0x800735FD, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_SIMULTANEOUS_REKEY)')
    WIN32__ERROR_IPSEC_IKE_DH_FAIL: Final[ThirtyTwoBits] = cached_hresult(0x800735FE, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_DH_FAIL)')
    WIN32__ERROR_IPSEC_IKE_CRITICAL_PAYLOAD_NOT_RECOGNIZED: Final[ThirtyTwoBits] = cached_hresult(0x800735FF, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_CRITICAL_PAYLOAD_NOT_RECOGNIZED)')
    WIN32__ERROR_IPSEC_IKE_INVALID_HEADER: Final[ThirtyTwoBits] = cached_hresult(0x80073600, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INVALID_HEADER)')
    WIN32__ERROR_IPSEC_IKE_NO_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x80073601, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_NO_POLICY)')
    WIN32__ERROR_IPSEC_IKE_INVALID_SIGNATURE: Final[ThirtyTwoBits] = cached_hresult(0x80073602, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INVALID_SIGNATURE)')
    WIN32__ERROR_IPSEC_IKE_KERBEROS_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80073603, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_KERBEROS_ERROR)')
    WIN32__ERROR_IPSEC_IKE_NO_PUBLIC_KEY: Final[ThirtyTwoBits] = cached_hresult(0x80073604, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_NO_PUBLIC_KEY)')
    WIN32__ERROR_IPSEC_IKE_PROCESS_ERR: Final[ThirtyTwoBits] = cached_hresult(0x80073605, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_PROCESS_ERR)')
    WIN32__ERROR_IPSEC_IKE_PROCESS_ERR_SA: Final[ThirtyTwoBits] = cached_hresult(0x80073606, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_PROCESS_ERR_SA)')
    WIN32__ERROR_IPSEC_IKE_PROCESS_ERR_PROP: Final[ThirtyTwoBits] = cached_hresult(0x80073607, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_PROCESS_ERR_PROP)')
    WIN32__ERROR_IPSEC_IKE_PROCESS_ERR_TRANS: Final[ThirtyTwoBits] = cached_hresult(0x80073608, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_PROCESS_ERR_TRANS)')
    WIN32__ERROR_IPSEC_IKE_PROCESS_ERR_KE: Final[ThirtyTwoBits] = cached_hresult(0x80073609, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_PROCESS_ERR_KE)')
    WIN32__ERROR_IPSEC_IKE_PROCESS_ERR_ID: Final[ThirtyTwoBits] = cached_hresult(0x8007360A, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_PROCESS_ERR_ID)')
    WIN32__ERROR_IPSEC_IKE_PROCESS_ERR_CERT: Final[ThirtyTwoBits] = cached_hresult(0x8007360B, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_PROCESS_ERR_CERT)')
    WIN32__ERROR_IPSEC_IKE_PROCESS_ERR_CERT_REQ: Final[ThirtyTwoBits] = cached_hresult(0x8007360C, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_PROCESS_ERR_CERT_REQ)')
    WIN32__ERROR_IPSEC_IKE_PROCESS_ERR_HASH: Final[ThirtyTwoBits] = cached_hresult(0x8007360D, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_PROCESS_ERR_HASH)')
    WIN32__ERROR_IPSEC_IKE_PROCESS_ERR_SIG: Final[ThirtyTwoBits] = cached_hresult(0x8007360E, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_PROCESS_ERR_SIG)')
    WIN32__ERROR_IPSEC_IKE_PROCESS_ERR_NONCE: Final[ThirtyTwoBits] = cached_hresult(0x8007360F, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_PROCESS_ERR_NONCE)')
    WIN32__ERROR_IPSEC_IKE_PROCESS_ERR_NOTIFY: Final[ThirtyTwoBits] = cached_hresult(0x80073610, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_PROCESS_ERR_NOTIFY)')
    WIN32__ERROR_IPSEC_IKE_PROCESS_ERR_DELETE: Final[ThirtyTwoBits] = cached_hresult(0x80073611, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_PROCESS_ERR_DELETE)')
    WIN32__ERROR_IPSEC_IKE_PROCESS_ERR_VENDOR: Final[ThirtyTwoBits] = cached_hresult(0x80073612, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_PROCESS_ERR_VENDOR)')
    WIN32__ERROR_IPSEC_IKE_INVALID_PAYLOAD: Final[ThirtyTwoBits] = cached_hresult(0x80073613, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INVALID_PAYLOAD)')
    WIN32__ERROR_IPSEC_IKE_LOAD_SOFT_SA: Final[ThirtyTwoBits] = cached_hresult(0x80073614, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_LOAD_SOFT_SA)')
    WIN32__ERROR_IPSEC_IKE_SOFT_SA_TORN_DOWN: Final[ThirtyTwoBits] = cached_hresult(0x80073615, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_SOFT_SA_TORN_DOWN)')
    WIN32__ERROR_IPSEC_IKE_INVALID_COOKIE: Final[ThirtyTwoBits] = cached_hresult(0x80073616, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INVALID_COOKIE)')
    WIN32__ERROR_IPSEC_IKE_NO_PEER_CERT: Final[ThirtyTwoBits] = cached_hresult(0x80073617, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_NO_PEER_CERT)')
    WIN32__ERROR_IPSEC_IKE_PEER_CRL_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073618, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_PEER_CRL_FAILED)')
    WIN32__ERROR_IPSEC_IKE_POLICY_CHANGE: Final[ThirtyTwoBits] = cached_hresult(0x80073619, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_POLICY_CHANGE)')
    WIN32__ERROR_IPSEC_IKE_NO_MM_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x8007361A, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_NO_MM_POLICY)')
    WIN32__ERROR_IPSEC_IKE_NOTCBPRIV: Final[ThirtyTwoBits] = cached_hresult(0x8007361B, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_NOTCBPRIV)')
    WIN32__ERROR_IPSEC_IKE_SECLOADFAIL: Final[ThirtyTwoBits] = cached_hresult(0x8007361C, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_SECLOADFAIL)')
    WIN32__ERROR_IPSEC_IKE_FAILSSPINIT: Final[ThirtyTwoBits] = cached_hresult(0x8007361D, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_FAILSSPINIT)')
    WIN32__ERROR_IPSEC_IKE_FAILQUERYSSP: Final[ThirtyTwoBits] = cached_hresult(0x8007361E, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_FAILQUERYSSP)')
    WIN32__ERROR_IPSEC_IKE_SRVACQFAIL: Final[ThirtyTwoBits] = cached_hresult(0x8007361F, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_SRVACQFAIL)')
    WIN32__ERROR_IPSEC_IKE_SRVQUERYCRED: Final[ThirtyTwoBits] = cached_hresult(0x80073620, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_SRVQUERYCRED)')
    WIN32__ERROR_IPSEC_IKE_GETSPIFAIL: Final[ThirtyTwoBits] = cached_hresult(0x80073621, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_GETSPIFAIL)')
    WIN32__ERROR_IPSEC_IKE_INVALID_FILTER: Final[ThirtyTwoBits] = cached_hresult(0x80073622, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INVALID_FILTER)')
    WIN32__ERROR_IPSEC_IKE_OUT_OF_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0x80073623, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_OUT_OF_MEMORY)')
    WIN32__ERROR_IPSEC_IKE_ADD_UPDATE_KEY_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073624, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_ADD_UPDATE_KEY_FAILED)')
    WIN32__ERROR_IPSEC_IKE_INVALID_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x80073625, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INVALID_POLICY)')
    WIN32__ERROR_IPSEC_IKE_UNKNOWN_DOI: Final[ThirtyTwoBits] = cached_hresult(0x80073626, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_UNKNOWN_DOI)')
    WIN32__ERROR_IPSEC_IKE_INVALID_SITUATION: Final[ThirtyTwoBits] = cached_hresult(0x80073627, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INVALID_SITUATION)')
    WIN32__ERROR_IPSEC_IKE_DH_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80073628, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_DH_FAILURE)')
    WIN32__ERROR_IPSEC_IKE_INVALID_GROUP: Final[ThirtyTwoBits] = cached_hresult(0x80073629, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INVALID_GROUP)')
    WIN32__ERROR_IPSEC_IKE_ENCRYPT: Final[ThirtyTwoBits] = cached_hresult(0x8007362A, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_ENCRYPT)')
    WIN32__ERROR_IPSEC_IKE_DECRYPT: Final[ThirtyTwoBits] = cached_hresult(0x8007362B, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_DECRYPT)')
    WIN32__ERROR_IPSEC_IKE_POLICY_MATCH: Final[ThirtyTwoBits] = cached_hresult(0x8007362C, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_POLICY_MATCH)')
    WIN32__ERROR_IPSEC_IKE_UNSUPPORTED_ID: Final[ThirtyTwoBits] = cached_hresult(0x8007362D, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_UNSUPPORTED_ID)')
    WIN32__ERROR_IPSEC_IKE_INVALID_HASH: Final[ThirtyTwoBits] = cached_hresult(0x8007362E, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INVALID_HASH)')
    WIN32__ERROR_IPSEC_IKE_INVALID_HASH_ALG: Final[ThirtyTwoBits] = cached_hresult(0x8007362F, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INVALID_HASH_ALG)')
    WIN32__ERROR_IPSEC_IKE_INVALID_HASH_SIZE: Final[ThirtyTwoBits] = cached_hresult(0x80073630, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INVALID_HASH_SIZE)')
    WIN32__ERROR_IPSEC_IKE_INVALID_ENCRYPT_ALG: Final[ThirtyTwoBits] = cached_hresult(0x80073631, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INVALID_ENCRYPT_ALG)')
    WIN32__ERROR_IPSEC_IKE_INVALID_AUTH_ALG: Final[ThirtyTwoBits] = cached_hresult(0x80073632, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INVALID_AUTH_ALG)')
    WIN32__ERROR_IPSEC_IKE_INVALID_SIG: Final[ThirtyTwoBits] = cached_hresult(0x80073633, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INVALID_SIG)')
    WIN32__ERROR_IPSEC_IKE_LOAD_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073634, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_LOAD_FAILED)')
    WIN32__ERROR_IPSEC_IKE_RPC_DELETE: Final[ThirtyTwoBits] = cached_hresult(0x80073635, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_RPC_DELETE)')
    WIN32__ERROR_IPSEC_IKE_BENIGN_REINIT: Final[ThirtyTwoBits] = cached_hresult(0x80073636, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_BENIGN_REINIT)')
    WIN32__ERROR_IPSEC_IKE_INVALID_RESPONDER_LIFETIME_NOTIFY: Final[ThirtyTwoBits] = cached_hresult(0x80073637, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INVALID_RESPONDER_LIFETIME_NOTIFY)')
    WIN32__ERROR_IPSEC_IKE_INVALID_MAJOR_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80073638, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INVALID_MAJOR_VERSION)')
    WIN32__ERROR_IPSEC_IKE_INVALID_CERT_KEYLEN: Final[ThirtyTwoBits] = cached_hresult(0x80073639, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INVALID_CERT_KEYLEN)')
    WIN32__ERROR_IPSEC_IKE_MM_LIMIT: Final[ThirtyTwoBits] = cached_hresult(0x8007363A, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_MM_LIMIT)')
    WIN32__ERROR_IPSEC_IKE_NEGOTIATION_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x8007363B, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_NEGOTIATION_DISABLED)')
    WIN32__ERROR_IPSEC_IKE_QM_LIMIT: Final[ThirtyTwoBits] = cached_hresult(0x8007363C, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_QM_LIMIT)')
    WIN32__ERROR_IPSEC_IKE_MM_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x8007363D, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_MM_EXPIRED)')
    WIN32__ERROR_IPSEC_IKE_PEER_MM_ASSUMED_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x8007363E, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_PEER_MM_ASSUMED_INVALID)')
    WIN32__ERROR_IPSEC_IKE_CERT_CHAIN_POLICY_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x8007363F, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_CERT_CHAIN_POLICY_MISMATCH)')
    WIN32__ERROR_IPSEC_IKE_UNEXPECTED_MESSAGE_ID: Final[ThirtyTwoBits] = cached_hresult(0x80073640, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_UNEXPECTED_MESSAGE_ID)')
    WIN32__ERROR_IPSEC_IKE_INVALID_AUTH_PAYLOAD: Final[ThirtyTwoBits] = cached_hresult(0x80073641, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INVALID_AUTH_PAYLOAD)')
    WIN32__ERROR_IPSEC_IKE_DOS_COOKIE_SENT: Final[ThirtyTwoBits] = cached_hresult(0x80073642, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_DOS_COOKIE_SENT)')
    WIN32__ERROR_IPSEC_IKE_SHUTTING_DOWN: Final[ThirtyTwoBits] = cached_hresult(0x80073643, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_SHUTTING_DOWN)')
    WIN32__ERROR_IPSEC_IKE_CGA_AUTH_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073644, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_CGA_AUTH_FAILED)')
    WIN32__ERROR_IPSEC_IKE_PROCESS_ERR_NATOA: Final[ThirtyTwoBits] = cached_hresult(0x80073645, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_PROCESS_ERR_NATOA)')
    WIN32__ERROR_IPSEC_IKE_INVALID_MM_FOR_QM: Final[ThirtyTwoBits] = cached_hresult(0x80073646, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INVALID_MM_FOR_QM)')
    WIN32__ERROR_IPSEC_IKE_QM_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x80073647, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_QM_EXPIRED)')
    WIN32__ERROR_IPSEC_IKE_TOO_MANY_FILTERS: Final[ThirtyTwoBits] = cached_hresult(0x80073648, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_TOO_MANY_FILTERS)')
    WIN32__ERROR_IPSEC_IKE_NEG_STATUS_END: Final[ThirtyTwoBits] = cached_hresult(0x80073649, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_NEG_STATUS_END)')
    WIN32__ERROR_IPSEC_IKE_KILL_DUMMY_NAP_TUNNEL: Final[ThirtyTwoBits] = cached_hresult(0x8007364A, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_KILL_DUMMY_NAP_TUNNEL)')
    WIN32__ERROR_IPSEC_IKE_INNER_IP_ASSIGNMENT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x8007364B, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_INNER_IP_ASSIGNMENT_FAILURE)')
    WIN32__ERROR_IPSEC_IKE_REQUIRE_CP_PAYLOAD_MISSING: Final[ThirtyTwoBits] = cached_hresult(0x8007364C, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_REQUIRE_CP_PAYLOAD_MISSING)')
    WIN32__ERROR_IPSEC_KEY_MODULE_IMPERSONATION_NEGOTIATION_PENDING: Final[ThirtyTwoBits] = cached_hresult(0x8007364D, 'HRESULT_FROM_WIN32(ERROR_IPSEC_KEY_MODULE_IMPERSONATION_NEGOTIATION_PENDING)')
    WIN32__ERROR_IPSEC_IKE_COEXISTENCE_SUPPRESS: Final[ThirtyTwoBits] = cached_hresult(0x8007364E, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_COEXISTENCE_SUPPRESS)')
    WIN32__ERROR_IPSEC_IKE_RATELIMIT_DROP: Final[ThirtyTwoBits] = cached_hresult(0x8007364F, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_RATELIMIT_DROP)')
    WIN32__ERROR_IPSEC_IKE_PEER_DOESNT_SUPPORT_MOBIKE: Final[ThirtyTwoBits] = cached_hresult(0x80073650, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_PEER_DOESNT_SUPPORT_MOBIKE)')
    WIN32__ERROR_IPSEC_IKE_AUTHORIZATION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80073651, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_AUTHORIZATION_FAILURE)')
    WIN32__ERROR_IPSEC_IKE_STRONG_CRED_AUTHORIZATION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80073652, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_STRONG_CRED_AUTHORIZATION_FAILURE)')
    WIN32__ERROR_IPSEC_IKE_AUTHORIZATION_FAILURE_WITH_OPTIONAL_RETRY: Final[ThirtyTwoBits] = cached_hresult(0x80073653, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_AUTHORIZATION_FAILURE_WITH_OPTIONAL_RETRY)')
    WIN32__ERROR_IPSEC_IKE_STRONG_CRED_AUTHORIZATION_AND_CERTMAP_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80073654, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_STRONG_CRED_AUTHORIZATION_AND_CERTMAP_FAILURE)')
    WIN32__ERROR_IPSEC_IKE_NEG_STATUS_EXTENDED_END: Final[ThirtyTwoBits] = cached_hresult(0x80073655, 'HRESULT_FROM_WIN32(ERROR_IPSEC_IKE_NEG_STATUS_EXTENDED_END)')
    WIN32__ERROR_IPSEC_BAD_SPI: Final[ThirtyTwoBits] = cached_hresult(0x80073656, 'HRESULT_FROM_WIN32(ERROR_IPSEC_BAD_SPI)')
    WIN32__ERROR_IPSEC_SA_LIFETIME_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0x80073657, 'HRESULT_FROM_WIN32(ERROR_IPSEC_SA_LIFETIME_EXPIRED)')
    WIN32__ERROR_IPSEC_WRONG_SA: Final[ThirtyTwoBits] = cached_hresult(0x80073658, 'HRESULT_FROM_WIN32(ERROR_IPSEC_WRONG_SA)')
    WIN32__ERROR_IPSEC_REPLAY_CHECK_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073659, 'HRESULT_FROM_WIN32(ERROR_IPSEC_REPLAY_CHECK_FAILED)')
    WIN32__ERROR_IPSEC_INVALID_PACKET: Final[ThirtyTwoBits] = cached_hresult(0x8007365A, 'HRESULT_FROM_WIN32(ERROR_IPSEC_INVALID_PACKET)')
    WIN32__ERROR_IPSEC_INTEGRITY_CHECK_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8007365B, 'HRESULT_FROM_WIN32(ERROR_IPSEC_INTEGRITY_CHECK_FAILED)')
    WIN32__ERROR_IPSEC_CLEAR_TEXT_DROP: Final[ThirtyTwoBits] = cached_hresult(0x8007365C, 'HRESULT_FROM_WIN32(ERROR_IPSEC_CLEAR_TEXT_DROP)')
    WIN32__ERROR_IPSEC_AUTH_FIREWALL_DROP: Final[ThirtyTwoBits] = cached_hresult(0x8007365D, 'HRESULT_FROM_WIN32(ERROR_IPSEC_AUTH_FIREWALL_DROP)')
    WIN32__ERROR_IPSEC_THROTTLE_DROP: Final[ThirtyTwoBits] = cached_hresult(0x8007365E, 'HRESULT_FROM_WIN32(ERROR_IPSEC_THROTTLE_DROP)')
    WIN32__ERROR_IPSEC_DOSP_BLOCK: Final[ThirtyTwoBits] = cached_hresult(0x80073665, 'HRESULT_FROM_WIN32(ERROR_IPSEC_DOSP_BLOCK)')
    WIN32__ERROR_IPSEC_DOSP_RECEIVED_MULTICAST: Final[ThirtyTwoBits] = cached_hresult(0x80073666, 'HRESULT_FROM_WIN32(ERROR_IPSEC_DOSP_RECEIVED_MULTICAST)')
    WIN32__ERROR_IPSEC_DOSP_INVALID_PACKET: Final[ThirtyTwoBits] = cached_hresult(0x80073667, 'HRESULT_FROM_WIN32(ERROR_IPSEC_DOSP_INVALID_PACKET)')
    WIN32__ERROR_IPSEC_DOSP_STATE_LOOKUP_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073668, 'HRESULT_FROM_WIN32(ERROR_IPSEC_DOSP_STATE_LOOKUP_FAILED)')
    WIN32__ERROR_IPSEC_DOSP_MAX_ENTRIES: Final[ThirtyTwoBits] = cached_hresult(0x80073669, 'HRESULT_FROM_WIN32(ERROR_IPSEC_DOSP_MAX_ENTRIES)')
    WIN32__ERROR_IPSEC_DOSP_KEYMOD_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x8007366A, 'HRESULT_FROM_WIN32(ERROR_IPSEC_DOSP_KEYMOD_NOT_ALLOWED)')
    WIN32__ERROR_IPSEC_DOSP_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x8007366B, 'HRESULT_FROM_WIN32(ERROR_IPSEC_DOSP_NOT_INSTALLED)')
    WIN32__ERROR_IPSEC_DOSP_MAX_PER_IP_RATELIMIT_QUEUES: Final[ThirtyTwoBits] = cached_hresult(0x8007366C, 'HRESULT_FROM_WIN32(ERROR_IPSEC_DOSP_MAX_PER_IP_RATELIMIT_QUEUES)')
    WIN32__ERROR_SXS_SECTION_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800736B0, 'HRESULT_FROM_WIN32(ERROR_SXS_SECTION_NOT_FOUND)')
    WIN32__ERROR_SXS_CANT_GEN_ACTCTX: Final[ThirtyTwoBits] = cached_hresult(0x800736B1, 'HRESULT_FROM_WIN32(ERROR_SXS_CANT_GEN_ACTCTX)')
    WIN32__ERROR_SXS_INVALID_ACTCTXDATA_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0x800736B2, 'HRESULT_FROM_WIN32(ERROR_SXS_INVALID_ACTCTXDATA_FORMAT)')
    WIN32__ERROR_SXS_ASSEMBLY_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800736B3, 'HRESULT_FROM_WIN32(ERROR_SXS_ASSEMBLY_NOT_FOUND)')
    WIN32__ERROR_SXS_MANIFEST_FORMAT_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800736B4, 'HRESULT_FROM_WIN32(ERROR_SXS_MANIFEST_FORMAT_ERROR)')
    WIN32__ERROR_SXS_MANIFEST_PARSE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800736B5, 'HRESULT_FROM_WIN32(ERROR_SXS_MANIFEST_PARSE_ERROR)')
    WIN32__ERROR_SXS_ACTIVATION_CONTEXT_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x800736B6, 'HRESULT_FROM_WIN32(ERROR_SXS_ACTIVATION_CONTEXT_DISABLED)')
    WIN32__ERROR_SXS_KEY_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x800736B7, 'HRESULT_FROM_WIN32(ERROR_SXS_KEY_NOT_FOUND)')
    WIN32__ERROR_SXS_VERSION_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x800736B8, 'HRESULT_FROM_WIN32(ERROR_SXS_VERSION_CONFLICT)')
    WIN32__ERROR_SXS_WRONG_SECTION_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x800736B9, 'HRESULT_FROM_WIN32(ERROR_SXS_WRONG_SECTION_TYPE)')
    WIN32__ERROR_SXS_THREAD_QUERIES_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x800736BA, 'HRESULT_FROM_WIN32(ERROR_SXS_THREAD_QUERIES_DISABLED)')
    WIN32__ERROR_SXS_PROCESS_DEFAULT_ALREADY_SET: Final[ThirtyTwoBits] = cached_hresult(0x800736BB, 'HRESULT_FROM_WIN32(ERROR_SXS_PROCESS_DEFAULT_ALREADY_SET)')
    WIN32__ERROR_SXS_UNKNOWN_ENCODING_GROUP: Final[ThirtyTwoBits] = cached_hresult(0x800736BC, 'HRESULT_FROM_WIN32(ERROR_SXS_UNKNOWN_ENCODING_GROUP)')
    WIN32__ERROR_SXS_UNKNOWN_ENCODING: Final[ThirtyTwoBits] = cached_hresult(0x800736BD, 'HRESULT_FROM_WIN32(ERROR_SXS_UNKNOWN_ENCODING)')
    WIN32__ERROR_SXS_INVALID_XML_NAMESPACE_URI: Final[ThirtyTwoBits] = cached_hresult(0x800736BE, 'HRESULT_FROM_WIN32(ERROR_SXS_INVALID_XML_NAMESPACE_URI)')
    WIN32__ERROR_SXS_ROOT_MANIFEST_DEPENDENCY_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x800736BF, 'HRESULT_FROM_WIN32(ERROR_SXS_ROOT_MANIFEST_DEPENDENCY_NOT_INSTALLED)')
    WIN32__ERROR_SXS_LEAF_MANIFEST_DEPENDENCY_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x800736C0, 'HRESULT_FROM_WIN32(ERROR_SXS_LEAF_MANIFEST_DEPENDENCY_NOT_INSTALLED)')
    WIN32__ERROR_SXS_INVALID_ASSEMBLY_IDENTITY_ATTRIBUTE: Final[ThirtyTwoBits] = cached_hresult(0x800736C1, 'HRESULT_FROM_WIN32(ERROR_SXS_INVALID_ASSEMBLY_IDENTITY_ATTRIBUTE)')
    WIN32__ERROR_SXS_MANIFEST_MISSING_REQUIRED_DEFAULT_NAMESPACE: Final[ThirtyTwoBits] = cached_hresult(0x800736C2, 'HRESULT_FROM_WIN32(ERROR_SXS_MANIFEST_MISSING_REQUIRED_DEFAULT_NAMESPACE)')
    WIN32__ERROR_SXS_MANIFEST_INVALID_REQUIRED_DEFAULT_NAMESPACE: Final[ThirtyTwoBits] = cached_hresult(0x800736C3, 'HRESULT_FROM_WIN32(ERROR_SXS_MANIFEST_INVALID_REQUIRED_DEFAULT_NAMESPACE)')
    WIN32__ERROR_SXS_PRIVATE_MANIFEST_CROSS_PATH_WITH_REPARSE_POINT: Final[ThirtyTwoBits] = cached_hresult(0x800736C4, 'HRESULT_FROM_WIN32(ERROR_SXS_PRIVATE_MANIFEST_CROSS_PATH_WITH_REPARSE_POINT)')
    WIN32__ERROR_SXS_DUPLICATE_DLL_NAME: Final[ThirtyTwoBits] = cached_hresult(0x800736C5, 'HRESULT_FROM_WIN32(ERROR_SXS_DUPLICATE_DLL_NAME)')
    WIN32__ERROR_SXS_DUPLICATE_WINDOWCLASS_NAME: Final[ThirtyTwoBits] = cached_hresult(0x800736C6, 'HRESULT_FROM_WIN32(ERROR_SXS_DUPLICATE_WINDOWCLASS_NAME)')
    WIN32__ERROR_SXS_DUPLICATE_CLSID: Final[ThirtyTwoBits] = cached_hresult(0x800736C7, 'HRESULT_FROM_WIN32(ERROR_SXS_DUPLICATE_CLSID)')
    WIN32__ERROR_SXS_DUPLICATE_IID: Final[ThirtyTwoBits] = cached_hresult(0x800736C8, 'HRESULT_FROM_WIN32(ERROR_SXS_DUPLICATE_IID)')
    WIN32__ERROR_SXS_DUPLICATE_TLBID: Final[ThirtyTwoBits] = cached_hresult(0x800736C9, 'HRESULT_FROM_WIN32(ERROR_SXS_DUPLICATE_TLBID)')
    WIN32__ERROR_SXS_DUPLICATE_PROGID: Final[ThirtyTwoBits] = cached_hresult(0x800736CA, 'HRESULT_FROM_WIN32(ERROR_SXS_DUPLICATE_PROGID)')
    WIN32__ERROR_SXS_DUPLICATE_ASSEMBLY_NAME: Final[ThirtyTwoBits] = cached_hresult(0x800736CB, 'HRESULT_FROM_WIN32(ERROR_SXS_DUPLICATE_ASSEMBLY_NAME)')
    WIN32__ERROR_SXS_FILE_HASH_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x800736CC, 'HRESULT_FROM_WIN32(ERROR_SXS_FILE_HASH_MISMATCH)')
    WIN32__ERROR_SXS_POLICY_PARSE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x800736CD, 'HRESULT_FROM_WIN32(ERROR_SXS_POLICY_PARSE_ERROR)')
    WIN32__ERROR_SXS_XML_E_MISSINGQUOTE: Final[ThirtyTwoBits] = cached_hresult(0x800736CE, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_MISSINGQUOTE)')
    WIN32__ERROR_SXS_XML_E_COMMENTSYNTAX: Final[ThirtyTwoBits] = cached_hresult(0x800736CF, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_COMMENTSYNTAX)')
    WIN32__ERROR_SXS_XML_E_BADSTARTNAMECHAR: Final[ThirtyTwoBits] = cached_hresult(0x800736D0, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_BADSTARTNAMECHAR)')
    WIN32__ERROR_SXS_XML_E_BADNAMECHAR: Final[ThirtyTwoBits] = cached_hresult(0x800736D1, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_BADNAMECHAR)')
    WIN32__ERROR_SXS_XML_E_BADCHARINSTRING: Final[ThirtyTwoBits] = cached_hresult(0x800736D2, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_BADCHARINSTRING)')
    WIN32__ERROR_SXS_XML_E_XMLDECLSYNTAX: Final[ThirtyTwoBits] = cached_hresult(0x800736D3, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_XMLDECLSYNTAX)')
    WIN32__ERROR_SXS_XML_E_BADCHARDATA: Final[ThirtyTwoBits] = cached_hresult(0x800736D4, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_BADCHARDATA)')
    WIN32__ERROR_SXS_XML_E_MISSINGWHITESPACE: Final[ThirtyTwoBits] = cached_hresult(0x800736D5, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_MISSINGWHITESPACE)')
    WIN32__ERROR_SXS_XML_E_EXPECTINGTAGEND: Final[ThirtyTwoBits] = cached_hresult(0x800736D6, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_EXPECTINGTAGEND)')
    WIN32__ERROR_SXS_XML_E_MISSINGSEMICOLON: Final[ThirtyTwoBits] = cached_hresult(0x800736D7, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_MISSINGSEMICOLON)')
    WIN32__ERROR_SXS_XML_E_UNBALANCEDPAREN: Final[ThirtyTwoBits] = cached_hresult(0x800736D8, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_UNBALANCEDPAREN)')
    WIN32__ERROR_SXS_XML_E_INTERNALERROR: Final[ThirtyTwoBits] = cached_hresult(0x800736D9, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_INTERNALERROR)')
    WIN32__ERROR_SXS_XML_E_UNEXPECTED_WHITESPACE: Final[ThirtyTwoBits] = cached_hresult(0x800736DA, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_UNEXPECTED_WHITESPACE)')
    WIN32__ERROR_SXS_XML_E_INCOMPLETE_ENCODING: Final[ThirtyTwoBits] = cached_hresult(0x800736DB, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_INCOMPLETE_ENCODING)')
    WIN32__ERROR_SXS_XML_E_MISSING_PAREN: Final[ThirtyTwoBits] = cached_hresult(0x800736DC, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_MISSING_PAREN)')
    WIN32__ERROR_SXS_XML_E_EXPECTINGCLOSEQUOTE: Final[ThirtyTwoBits] = cached_hresult(0x800736DD, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_EXPECTINGCLOSEQUOTE)')
    WIN32__ERROR_SXS_XML_E_MULTIPLE_COLONS: Final[ThirtyTwoBits] = cached_hresult(0x800736DE, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_MULTIPLE_COLONS)')
    WIN32__ERROR_SXS_XML_E_INVALID_DECIMAL: Final[ThirtyTwoBits] = cached_hresult(0x800736DF, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_INVALID_DECIMAL)')
    WIN32__ERROR_SXS_XML_E_INVALID_HEXIDECIMAL: Final[ThirtyTwoBits] = cached_hresult(0x800736E0, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_INVALID_HEXIDECIMAL)')
    WIN32__ERROR_SXS_XML_E_INVALID_UNICODE: Final[ThirtyTwoBits] = cached_hresult(0x800736E1, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_INVALID_UNICODE)')
    WIN32__ERROR_SXS_XML_E_WHITESPACEORQUESTIONMARK: Final[ThirtyTwoBits] = cached_hresult(0x800736E2, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_WHITESPACEORQUESTIONMARK)')
    WIN32__ERROR_SXS_XML_E_UNEXPECTEDENDTAG: Final[ThirtyTwoBits] = cached_hresult(0x800736E3, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_UNEXPECTEDENDTAG)')
    WIN32__ERROR_SXS_XML_E_UNCLOSEDTAG: Final[ThirtyTwoBits] = cached_hresult(0x800736E4, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_UNCLOSEDTAG)')
    WIN32__ERROR_SXS_XML_E_DUPLICATEATTRIBUTE: Final[ThirtyTwoBits] = cached_hresult(0x800736E5, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_DUPLICATEATTRIBUTE)')
    WIN32__ERROR_SXS_XML_E_MULTIPLEROOTS: Final[ThirtyTwoBits] = cached_hresult(0x800736E6, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_MULTIPLEROOTS)')
    WIN32__ERROR_SXS_XML_E_INVALIDATROOTLEVEL: Final[ThirtyTwoBits] = cached_hresult(0x800736E7, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_INVALIDATROOTLEVEL)')
    WIN32__ERROR_SXS_XML_E_BADXMLDECL: Final[ThirtyTwoBits] = cached_hresult(0x800736E8, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_BADXMLDECL)')
    WIN32__ERROR_SXS_XML_E_MISSINGROOT: Final[ThirtyTwoBits] = cached_hresult(0x800736E9, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_MISSINGROOT)')
    WIN32__ERROR_SXS_XML_E_UNEXPECTEDEOF: Final[ThirtyTwoBits] = cached_hresult(0x800736EA, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_UNEXPECTEDEOF)')
    WIN32__ERROR_SXS_XML_E_BADPEREFINSUBSET: Final[ThirtyTwoBits] = cached_hresult(0x800736EB, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_BADPEREFINSUBSET)')
    WIN32__ERROR_SXS_XML_E_UNCLOSEDSTARTTAG: Final[ThirtyTwoBits] = cached_hresult(0x800736EC, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_UNCLOSEDSTARTTAG)')
    WIN32__ERROR_SXS_XML_E_UNCLOSEDENDTAG: Final[ThirtyTwoBits] = cached_hresult(0x800736ED, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_UNCLOSEDENDTAG)')
    WIN32__ERROR_SXS_XML_E_UNCLOSEDSTRING: Final[ThirtyTwoBits] = cached_hresult(0x800736EE, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_UNCLOSEDSTRING)')
    WIN32__ERROR_SXS_XML_E_UNCLOSEDCOMMENT: Final[ThirtyTwoBits] = cached_hresult(0x800736EF, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_UNCLOSEDCOMMENT)')
    WIN32__ERROR_SXS_XML_E_UNCLOSEDDECL: Final[ThirtyTwoBits] = cached_hresult(0x800736F0, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_UNCLOSEDDECL)')
    WIN32__ERROR_SXS_XML_E_UNCLOSEDCDATA: Final[ThirtyTwoBits] = cached_hresult(0x800736F1, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_UNCLOSEDCDATA)')
    WIN32__ERROR_SXS_XML_E_RESERVEDNAMESPACE: Final[ThirtyTwoBits] = cached_hresult(0x800736F2, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_RESERVEDNAMESPACE)')
    WIN32__ERROR_SXS_XML_E_INVALIDENCODING: Final[ThirtyTwoBits] = cached_hresult(0x800736F3, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_INVALIDENCODING)')
    WIN32__ERROR_SXS_XML_E_INVALIDSWITCH: Final[ThirtyTwoBits] = cached_hresult(0x800736F4, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_INVALIDSWITCH)')
    WIN32__ERROR_SXS_XML_E_BADXMLCASE: Final[ThirtyTwoBits] = cached_hresult(0x800736F5, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_BADXMLCASE)')
    WIN32__ERROR_SXS_XML_E_INVALID_STANDALONE: Final[ThirtyTwoBits] = cached_hresult(0x800736F6, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_INVALID_STANDALONE)')
    WIN32__ERROR_SXS_XML_E_UNEXPECTED_STANDALONE: Final[ThirtyTwoBits] = cached_hresult(0x800736F7, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_UNEXPECTED_STANDALONE)')
    WIN32__ERROR_SXS_XML_E_INVALID_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x800736F8, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_INVALID_VERSION)')
    WIN32__ERROR_SXS_XML_E_MISSINGEQUALS: Final[ThirtyTwoBits] = cached_hresult(0x800736F9, 'HRESULT_FROM_WIN32(ERROR_SXS_XML_E_MISSINGEQUALS)')
    WIN32__ERROR_SXS_PROTECTION_RECOVERY_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x800736FA, 'HRESULT_FROM_WIN32(ERROR_SXS_PROTECTION_RECOVERY_FAILED)')
    WIN32__ERROR_SXS_PROTECTION_PUBLIC_KEY_TOO_SHORT: Final[ThirtyTwoBits] = cached_hresult(0x800736FB, 'HRESULT_FROM_WIN32(ERROR_SXS_PROTECTION_PUBLIC_KEY_TOO_SHORT)')
    WIN32__ERROR_SXS_PROTECTION_CATALOG_NOT_VALID: Final[ThirtyTwoBits] = cached_hresult(0x800736FC, 'HRESULT_FROM_WIN32(ERROR_SXS_PROTECTION_CATALOG_NOT_VALID)')
    WIN32__ERROR_SXS_UNTRANSLATABLE_HRESULT: Final[ThirtyTwoBits] = cached_hresult(0x800736FD, 'HRESULT_FROM_WIN32(ERROR_SXS_UNTRANSLATABLE_HRESULT)')
    WIN32__ERROR_SXS_PROTECTION_CATALOG_FILE_MISSING: Final[ThirtyTwoBits] = cached_hresult(0x800736FE, 'HRESULT_FROM_WIN32(ERROR_SXS_PROTECTION_CATALOG_FILE_MISSING)')
    WIN32__ERROR_SXS_MISSING_ASSEMBLY_IDENTITY_ATTRIBUTE: Final[ThirtyTwoBits] = cached_hresult(0x800736FF, 'HRESULT_FROM_WIN32(ERROR_SXS_MISSING_ASSEMBLY_IDENTITY_ATTRIBUTE)')
    WIN32__ERROR_SXS_INVALID_ASSEMBLY_IDENTITY_ATTRIBUTE_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80073700, 'HRESULT_FROM_WIN32(ERROR_SXS_INVALID_ASSEMBLY_IDENTITY_ATTRIBUTE_NAME)')
    WIN32__ERROR_SXS_ASSEMBLY_MISSING: Final[ThirtyTwoBits] = cached_hresult(0x80073701, 'HRESULT_FROM_WIN32(ERROR_SXS_ASSEMBLY_MISSING)')
    WIN32__ERROR_SXS_CORRUPT_ACTIVATION_STACK: Final[ThirtyTwoBits] = cached_hresult(0x80073702, 'HRESULT_FROM_WIN32(ERROR_SXS_CORRUPT_ACTIVATION_STACK)')
    WIN32__ERROR_SXS_CORRUPTION: Final[ThirtyTwoBits] = cached_hresult(0x80073703, 'HRESULT_FROM_WIN32(ERROR_SXS_CORRUPTION)')
    WIN32__ERROR_SXS_EARLY_DEACTIVATION: Final[ThirtyTwoBits] = cached_hresult(0x80073704, 'HRESULT_FROM_WIN32(ERROR_SXS_EARLY_DEACTIVATION)')
    WIN32__ERROR_SXS_INVALID_DEACTIVATION: Final[ThirtyTwoBits] = cached_hresult(0x80073705, 'HRESULT_FROM_WIN32(ERROR_SXS_INVALID_DEACTIVATION)')
    WIN32__ERROR_SXS_MULTIPLE_DEACTIVATION: Final[ThirtyTwoBits] = cached_hresult(0x80073706, 'HRESULT_FROM_WIN32(ERROR_SXS_MULTIPLE_DEACTIVATION)')
    WIN32__ERROR_SXS_PROCESS_TERMINATION_REQUESTED: Final[ThirtyTwoBits] = cached_hresult(0x80073707, 'HRESULT_FROM_WIN32(ERROR_SXS_PROCESS_TERMINATION_REQUESTED)')
    WIN32__ERROR_SXS_RELEASE_ACTIVATION_CONTEXT: Final[ThirtyTwoBits] = cached_hresult(0x80073708, 'HRESULT_FROM_WIN32(ERROR_SXS_RELEASE_ACTIVATION_CONTEXT)')
    WIN32__ERROR_SXS_SYSTEM_DEFAULT_ACTIVATION_CONTEXT_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0x80073709, 'HRESULT_FROM_WIN32(ERROR_SXS_SYSTEM_DEFAULT_ACTIVATION_CONTEXT_EMPTY)')
    WIN32__ERROR_SXS_INVALID_IDENTITY_ATTRIBUTE_VALUE: Final[ThirtyTwoBits] = cached_hresult(0x8007370A, 'HRESULT_FROM_WIN32(ERROR_SXS_INVALID_IDENTITY_ATTRIBUTE_VALUE)')
    WIN32__ERROR_SXS_INVALID_IDENTITY_ATTRIBUTE_NAME: Final[ThirtyTwoBits] = cached_hresult(0x8007370B, 'HRESULT_FROM_WIN32(ERROR_SXS_INVALID_IDENTITY_ATTRIBUTE_NAME)')
    WIN32__ERROR_SXS_IDENTITY_DUPLICATE_ATTRIBUTE: Final[ThirtyTwoBits] = cached_hresult(0x8007370C, 'HRESULT_FROM_WIN32(ERROR_SXS_IDENTITY_DUPLICATE_ATTRIBUTE)')
    WIN32__ERROR_SXS_IDENTITY_PARSE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x8007370D, 'HRESULT_FROM_WIN32(ERROR_SXS_IDENTITY_PARSE_ERROR)')
    WIN32__ERROR_MALFORMED_SUBSTITUTION_STRING: Final[ThirtyTwoBits] = cached_hresult(0x8007370E, 'HRESULT_FROM_WIN32(ERROR_MALFORMED_SUBSTITUTION_STRING)')
    WIN32__ERROR_SXS_INCORRECT_PUBLIC_KEY_TOKEN: Final[ThirtyTwoBits] = cached_hresult(0x8007370F, 'HRESULT_FROM_WIN32(ERROR_SXS_INCORRECT_PUBLIC_KEY_TOKEN)')
    WIN32__ERROR_UNMAPPED_SUBSTITUTION_STRING: Final[ThirtyTwoBits] = cached_hresult(0x80073710, 'HRESULT_FROM_WIN32(ERROR_UNMAPPED_SUBSTITUTION_STRING)')
    WIN32__ERROR_SXS_ASSEMBLY_NOT_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x80073711, 'HRESULT_FROM_WIN32(ERROR_SXS_ASSEMBLY_NOT_LOCKED)')
    WIN32__ERROR_SXS_COMPONENT_STORE_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x80073712, 'HRESULT_FROM_WIN32(ERROR_SXS_COMPONENT_STORE_CORRUPT)')
    WIN32__ERROR_ADVANCED_INSTALLER_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073713, 'HRESULT_FROM_WIN32(ERROR_ADVANCED_INSTALLER_FAILED)')
    WIN32__ERROR_XML_ENCODING_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80073714, 'HRESULT_FROM_WIN32(ERROR_XML_ENCODING_MISMATCH)')
    WIN32__ERROR_SXS_MANIFEST_IDENTITY_SAME_BUT_CONTENTS_DIFFERENT: Final[ThirtyTwoBits] = cached_hresult(0x80073715, 'HRESULT_FROM_WIN32(ERROR_SXS_MANIFEST_IDENTITY_SAME_BUT_CONTENTS_DIFFERENT)')
    WIN32__ERROR_SXS_IDENTITIES_DIFFERENT: Final[ThirtyTwoBits] = cached_hresult(0x80073716, 'HRESULT_FROM_WIN32(ERROR_SXS_IDENTITIES_DIFFERENT)')
    WIN32__ERROR_SXS_ASSEMBLY_IS_NOT_A_DEPLOYMENT: Final[ThirtyTwoBits] = cached_hresult(0x80073717, 'HRESULT_FROM_WIN32(ERROR_SXS_ASSEMBLY_IS_NOT_A_DEPLOYMENT)')
    WIN32__ERROR_SXS_FILE_NOT_PART_OF_ASSEMBLY: Final[ThirtyTwoBits] = cached_hresult(0x80073718, 'HRESULT_FROM_WIN32(ERROR_SXS_FILE_NOT_PART_OF_ASSEMBLY)')
    WIN32__ERROR_SXS_MANIFEST_TOO_BIG: Final[ThirtyTwoBits] = cached_hresult(0x80073719, 'HRESULT_FROM_WIN32(ERROR_SXS_MANIFEST_TOO_BIG)')
    WIN32__ERROR_SXS_SETTING_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x8007371A, 'HRESULT_FROM_WIN32(ERROR_SXS_SETTING_NOT_REGISTERED)')
    WIN32__ERROR_SXS_TRANSACTION_CLOSURE_INCOMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x8007371B, 'HRESULT_FROM_WIN32(ERROR_SXS_TRANSACTION_CLOSURE_INCOMPLETE)')
    WIN32__ERROR_SMI_PRIMITIVE_INSTALLER_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8007371C, 'HRESULT_FROM_WIN32(ERROR_SMI_PRIMITIVE_INSTALLER_FAILED)')
    WIN32__ERROR_GENERIC_COMMAND_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x8007371D, 'HRESULT_FROM_WIN32(ERROR_GENERIC_COMMAND_FAILED)')
    WIN32__ERROR_SXS_FILE_HASH_MISSING: Final[ThirtyTwoBits] = cached_hresult(0x8007371E, 'HRESULT_FROM_WIN32(ERROR_SXS_FILE_HASH_MISSING)')
    WIN32__ERROR_SXS_DUPLICATE_ACTIVATABLE_CLASS: Final[ThirtyTwoBits] = cached_hresult(0x8007371F, 'HRESULT_FROM_WIN32(ERROR_SXS_DUPLICATE_ACTIVATABLE_CLASS)')
    WIN32__ERROR_EVT_INVALID_CHANNEL_PATH: Final[ThirtyTwoBits] = cached_hresult(0x80073A98, 'HRESULT_FROM_WIN32(ERROR_EVT_INVALID_CHANNEL_PATH)')
    WIN32__ERROR_EVT_INVALID_QUERY: Final[ThirtyTwoBits] = cached_hresult(0x80073A99, 'HRESULT_FROM_WIN32(ERROR_EVT_INVALID_QUERY)')
    WIN32__ERROR_EVT_PUBLISHER_METADATA_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80073A9A, 'HRESULT_FROM_WIN32(ERROR_EVT_PUBLISHER_METADATA_NOT_FOUND)')
    WIN32__ERROR_EVT_EVENT_TEMPLATE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80073A9B, 'HRESULT_FROM_WIN32(ERROR_EVT_EVENT_TEMPLATE_NOT_FOUND)')
    WIN32__ERROR_EVT_INVALID_PUBLISHER_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80073A9C, 'HRESULT_FROM_WIN32(ERROR_EVT_INVALID_PUBLISHER_NAME)')
    WIN32__ERROR_EVT_INVALID_EVENT_DATA: Final[ThirtyTwoBits] = cached_hresult(0x80073A9D, 'HRESULT_FROM_WIN32(ERROR_EVT_INVALID_EVENT_DATA)')
    WIN32__ERROR_EVT_CHANNEL_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80073A9F, 'HRESULT_FROM_WIN32(ERROR_EVT_CHANNEL_NOT_FOUND)')
    WIN32__ERROR_EVT_MALFORMED_XML_TEXT: Final[ThirtyTwoBits] = cached_hresult(0x80073AA0, 'HRESULT_FROM_WIN32(ERROR_EVT_MALFORMED_XML_TEXT)')
    WIN32__ERROR_EVT_SUBSCRIPTION_TO_DIRECT_CHANNEL: Final[ThirtyTwoBits] = cached_hresult(0x80073AA1, 'HRESULT_FROM_WIN32(ERROR_EVT_SUBSCRIPTION_TO_DIRECT_CHANNEL)')
    WIN32__ERROR_EVT_CONFIGURATION_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80073AA2, 'HRESULT_FROM_WIN32(ERROR_EVT_CONFIGURATION_ERROR)')
    WIN32__ERROR_EVT_QUERY_RESULT_STALE: Final[ThirtyTwoBits] = cached_hresult(0x80073AA3, 'HRESULT_FROM_WIN32(ERROR_EVT_QUERY_RESULT_STALE)')
    WIN32__ERROR_EVT_QUERY_RESULT_INVALID_POSITION: Final[ThirtyTwoBits] = cached_hresult(0x80073AA4, 'HRESULT_FROM_WIN32(ERROR_EVT_QUERY_RESULT_INVALID_POSITION)')
    WIN32__ERROR_EVT_NON_VALIDATING_MSXML: Final[ThirtyTwoBits] = cached_hresult(0x80073AA5, 'HRESULT_FROM_WIN32(ERROR_EVT_NON_VALIDATING_MSXML)')
    WIN32__ERROR_EVT_FILTER_ALREADYSCOPED: Final[ThirtyTwoBits] = cached_hresult(0x80073AA6, 'HRESULT_FROM_WIN32(ERROR_EVT_FILTER_ALREADYSCOPED)')
    WIN32__ERROR_EVT_FILTER_NOTELTSET: Final[ThirtyTwoBits] = cached_hresult(0x80073AA7, 'HRESULT_FROM_WIN32(ERROR_EVT_FILTER_NOTELTSET)')
    WIN32__ERROR_EVT_FILTER_INVARG: Final[ThirtyTwoBits] = cached_hresult(0x80073AA8, 'HRESULT_FROM_WIN32(ERROR_EVT_FILTER_INVARG)')
    WIN32__ERROR_EVT_FILTER_INVTEST: Final[ThirtyTwoBits] = cached_hresult(0x80073AA9, 'HRESULT_FROM_WIN32(ERROR_EVT_FILTER_INVTEST)')
    WIN32__ERROR_EVT_FILTER_INVTYPE: Final[ThirtyTwoBits] = cached_hresult(0x80073AAA, 'HRESULT_FROM_WIN32(ERROR_EVT_FILTER_INVTYPE)')
    WIN32__ERROR_EVT_FILTER_PARSEERR: Final[ThirtyTwoBits] = cached_hresult(0x80073AAB, 'HRESULT_FROM_WIN32(ERROR_EVT_FILTER_PARSEERR)')
    WIN32__ERROR_EVT_FILTER_UNSUPPORTEDOP: Final[ThirtyTwoBits] = cached_hresult(0x80073AAC, 'HRESULT_FROM_WIN32(ERROR_EVT_FILTER_UNSUPPORTEDOP)')
    WIN32__ERROR_EVT_FILTER_UNEXPECTEDTOKEN: Final[ThirtyTwoBits] = cached_hresult(0x80073AAD, 'HRESULT_FROM_WIN32(ERROR_EVT_FILTER_UNEXPECTEDTOKEN)')
    WIN32__ERROR_EVT_INVALID_OPERATION_OVER_ENABLED_DIRECT_CHANNEL: Final[ThirtyTwoBits] = cached_hresult(0x80073AAE, 'HRESULT_FROM_WIN32(ERROR_EVT_INVALID_OPERATION_OVER_ENABLED_DIRECT_CHANNEL)')
    WIN32__ERROR_EVT_INVALID_CHANNEL_PROPERTY_VALUE: Final[ThirtyTwoBits] = cached_hresult(0x80073AAF, 'HRESULT_FROM_WIN32(ERROR_EVT_INVALID_CHANNEL_PROPERTY_VALUE)')
    WIN32__ERROR_EVT_INVALID_PUBLISHER_PROPERTY_VALUE: Final[ThirtyTwoBits] = cached_hresult(0x80073AB0, 'HRESULT_FROM_WIN32(ERROR_EVT_INVALID_PUBLISHER_PROPERTY_VALUE)')
    WIN32__ERROR_EVT_CHANNEL_CANNOT_ACTIVATE: Final[ThirtyTwoBits] = cached_hresult(0x80073AB1, 'HRESULT_FROM_WIN32(ERROR_EVT_CHANNEL_CANNOT_ACTIVATE)')
    WIN32__ERROR_EVT_FILTER_TOO_COMPLEX: Final[ThirtyTwoBits] = cached_hresult(0x80073AB2, 'HRESULT_FROM_WIN32(ERROR_EVT_FILTER_TOO_COMPLEX)')
    WIN32__ERROR_EVT_MESSAGE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80073AB3, 'HRESULT_FROM_WIN32(ERROR_EVT_MESSAGE_NOT_FOUND)')
    WIN32__ERROR_EVT_MESSAGE_ID_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80073AB4, 'HRESULT_FROM_WIN32(ERROR_EVT_MESSAGE_ID_NOT_FOUND)')
    WIN32__ERROR_EVT_UNRESOLVED_VALUE_INSERT: Final[ThirtyTwoBits] = cached_hresult(0x80073AB5, 'HRESULT_FROM_WIN32(ERROR_EVT_UNRESOLVED_VALUE_INSERT)')
    WIN32__ERROR_EVT_UNRESOLVED_PARAMETER_INSERT: Final[ThirtyTwoBits] = cached_hresult(0x80073AB6, 'HRESULT_FROM_WIN32(ERROR_EVT_UNRESOLVED_PARAMETER_INSERT)')
    WIN32__ERROR_EVT_MAX_INSERTS_REACHED: Final[ThirtyTwoBits] = cached_hresult(0x80073AB7, 'HRESULT_FROM_WIN32(ERROR_EVT_MAX_INSERTS_REACHED)')
    WIN32__ERROR_EVT_EVENT_DEFINITION_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80073AB8, 'HRESULT_FROM_WIN32(ERROR_EVT_EVENT_DEFINITION_NOT_FOUND)')
    WIN32__ERROR_EVT_MESSAGE_LOCALE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80073AB9, 'HRESULT_FROM_WIN32(ERROR_EVT_MESSAGE_LOCALE_NOT_FOUND)')
    WIN32__ERROR_EVT_VERSION_TOO_OLD: Final[ThirtyTwoBits] = cached_hresult(0x80073ABA, 'HRESULT_FROM_WIN32(ERROR_EVT_VERSION_TOO_OLD)')
    WIN32__ERROR_EVT_VERSION_TOO_NEW: Final[ThirtyTwoBits] = cached_hresult(0x80073ABB, 'HRESULT_FROM_WIN32(ERROR_EVT_VERSION_TOO_NEW)')
    WIN32__ERROR_EVT_CANNOT_OPEN_CHANNEL_OF_QUERY: Final[ThirtyTwoBits] = cached_hresult(0x80073ABC, 'HRESULT_FROM_WIN32(ERROR_EVT_CANNOT_OPEN_CHANNEL_OF_QUERY)')
    WIN32__ERROR_EVT_PUBLISHER_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80073ABD, 'HRESULT_FROM_WIN32(ERROR_EVT_PUBLISHER_DISABLED)')
    WIN32__ERROR_EVT_FILTER_OUT_OF_RANGE: Final[ThirtyTwoBits] = cached_hresult(0x80073ABE, 'HRESULT_FROM_WIN32(ERROR_EVT_FILTER_OUT_OF_RANGE)')
    WIN32__ERROR_EC_SUBSCRIPTION_CANNOT_ACTIVATE: Final[ThirtyTwoBits] = cached_hresult(0x80073AE8, 'HRESULT_FROM_WIN32(ERROR_EC_SUBSCRIPTION_CANNOT_ACTIVATE)')
    WIN32__ERROR_EC_LOG_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0x80073AE9, 'HRESULT_FROM_WIN32(ERROR_EC_LOG_DISABLED)')
    WIN32__ERROR_EC_CIRCULAR_FORWARDING: Final[ThirtyTwoBits] = cached_hresult(0x80073AEA, 'HRESULT_FROM_WIN32(ERROR_EC_CIRCULAR_FORWARDING)')
    WIN32__ERROR_EC_CREDSTORE_FULL: Final[ThirtyTwoBits] = cached_hresult(0x80073AEB, 'HRESULT_FROM_WIN32(ERROR_EC_CREDSTORE_FULL)')
    WIN32__ERROR_EC_CRED_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80073AEC, 'HRESULT_FROM_WIN32(ERROR_EC_CRED_NOT_FOUND)')
    WIN32__ERROR_EC_NO_ACTIVE_CHANNEL: Final[ThirtyTwoBits] = cached_hresult(0x80073AED, 'HRESULT_FROM_WIN32(ERROR_EC_NO_ACTIVE_CHANNEL)')
    WIN32__ERROR_MUI_FILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80073AFC, 'HRESULT_FROM_WIN32(ERROR_MUI_FILE_NOT_FOUND)')
    WIN32__ERROR_MUI_INVALID_FILE: Final[ThirtyTwoBits] = cached_hresult(0x80073AFD, 'HRESULT_FROM_WIN32(ERROR_MUI_INVALID_FILE)')
    WIN32__ERROR_MUI_INVALID_RC_CONFIG: Final[ThirtyTwoBits] = cached_hresult(0x80073AFE, 'HRESULT_FROM_WIN32(ERROR_MUI_INVALID_RC_CONFIG)')
    WIN32__ERROR_MUI_INVALID_LOCALE_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80073AFF, 'HRESULT_FROM_WIN32(ERROR_MUI_INVALID_LOCALE_NAME)')
    WIN32__ERROR_MUI_INVALID_ULTIMATEFALLBACK_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80073B00, 'HRESULT_FROM_WIN32(ERROR_MUI_INVALID_ULTIMATEFALLBACK_NAME)')
    WIN32__ERROR_MUI_FILE_NOT_LOADED: Final[ThirtyTwoBits] = cached_hresult(0x80073B01, 'HRESULT_FROM_WIN32(ERROR_MUI_FILE_NOT_LOADED)')
    WIN32__ERROR_RESOURCE_ENUM_USER_STOP: Final[ThirtyTwoBits] = cached_hresult(0x80073B02, 'HRESULT_FROM_WIN32(ERROR_RESOURCE_ENUM_USER_STOP)')
    WIN32__ERROR_MUI_INTLSETTINGS_UILANG_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x80073B03, 'HRESULT_FROM_WIN32(ERROR_MUI_INTLSETTINGS_UILANG_NOT_INSTALLED)')
    WIN32__ERROR_MUI_INTLSETTINGS_INVALID_LOCALE_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80073B04, 'HRESULT_FROM_WIN32(ERROR_MUI_INTLSETTINGS_INVALID_LOCALE_NAME)')
    WIN32__ERROR_MRM_RUNTIME_NO_DEFAULT_OR_NEUTRAL_RESOURCE: Final[ThirtyTwoBits] = cached_hresult(0x80073B06, 'HRESULT_FROM_WIN32(ERROR_MRM_RUNTIME_NO_DEFAULT_OR_NEUTRAL_RESOURCE)')
    WIN32__ERROR_MRM_INVALID_PRICONFIG: Final[ThirtyTwoBits] = cached_hresult(0x80073B07, 'HRESULT_FROM_WIN32(ERROR_MRM_INVALID_PRICONFIG)')
    WIN32__ERROR_MRM_INVALID_FILE_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x80073B08, 'HRESULT_FROM_WIN32(ERROR_MRM_INVALID_FILE_TYPE)')
    WIN32__ERROR_MRM_UNKNOWN_QUALIFIER: Final[ThirtyTwoBits] = cached_hresult(0x80073B09, 'HRESULT_FROM_WIN32(ERROR_MRM_UNKNOWN_QUALIFIER)')
    WIN32__ERROR_MRM_INVALID_QUALIFIER_VALUE: Final[ThirtyTwoBits] = cached_hresult(0x80073B0A, 'HRESULT_FROM_WIN32(ERROR_MRM_INVALID_QUALIFIER_VALUE)')
    WIN32__ERROR_MRM_NO_CANDIDATE: Final[ThirtyTwoBits] = cached_hresult(0x80073B0B, 'HRESULT_FROM_WIN32(ERROR_MRM_NO_CANDIDATE)')
    WIN32__ERROR_MRM_NO_MATCH_OR_DEFAULT_CANDIDATE: Final[ThirtyTwoBits] = cached_hresult(0x80073B0C, 'HRESULT_FROM_WIN32(ERROR_MRM_NO_MATCH_OR_DEFAULT_CANDIDATE)')
    WIN32__ERROR_MRM_RESOURCE_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80073B0D, 'HRESULT_FROM_WIN32(ERROR_MRM_RESOURCE_TYPE_MISMATCH)')
    WIN32__ERROR_MRM_DUPLICATE_MAP_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80073B0E, 'HRESULT_FROM_WIN32(ERROR_MRM_DUPLICATE_MAP_NAME)')
    WIN32__ERROR_MRM_DUPLICATE_ENTRY: Final[ThirtyTwoBits] = cached_hresult(0x80073B0F, 'HRESULT_FROM_WIN32(ERROR_MRM_DUPLICATE_ENTRY)')
    WIN32__ERROR_MRM_INVALID_RESOURCE_IDENTIFIER: Final[ThirtyTwoBits] = cached_hresult(0x80073B10, 'HRESULT_FROM_WIN32(ERROR_MRM_INVALID_RESOURCE_IDENTIFIER)')
    WIN32__ERROR_MRM_FILEPATH_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0x80073B11, 'HRESULT_FROM_WIN32(ERROR_MRM_FILEPATH_TOO_LONG)')
    WIN32__ERROR_MRM_UNSUPPORTED_DIRECTORY_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x80073B12, 'HRESULT_FROM_WIN32(ERROR_MRM_UNSUPPORTED_DIRECTORY_TYPE)')
    WIN32__ERROR_MRM_INVALID_PRI_FILE: Final[ThirtyTwoBits] = cached_hresult(0x80073B16, 'HRESULT_FROM_WIN32(ERROR_MRM_INVALID_PRI_FILE)')
    WIN32__ERROR_MRM_NAMED_RESOURCE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80073B17, 'HRESULT_FROM_WIN32(ERROR_MRM_NAMED_RESOURCE_NOT_FOUND)')
    WIN32__ERROR_MRM_MAP_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80073B1F, 'HRESULT_FROM_WIN32(ERROR_MRM_MAP_NOT_FOUND)')
    WIN32__ERROR_MRM_UNSUPPORTED_PROFILE_TYPE: Final[ThirtyTwoBits] = cached_hresult(0x80073B20, 'HRESULT_FROM_WIN32(ERROR_MRM_UNSUPPORTED_PROFILE_TYPE)')
    WIN32__ERROR_MRM_INVALID_QUALIFIER_OPERATOR: Final[ThirtyTwoBits] = cached_hresult(0x80073B21, 'HRESULT_FROM_WIN32(ERROR_MRM_INVALID_QUALIFIER_OPERATOR)')
    WIN32__ERROR_MRM_INDETERMINATE_QUALIFIER_VALUE: Final[ThirtyTwoBits] = cached_hresult(0x80073B22, 'HRESULT_FROM_WIN32(ERROR_MRM_INDETERMINATE_QUALIFIER_VALUE)')
    WIN32__ERROR_MRM_AUTOMERGE_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0x80073B23, 'HRESULT_FROM_WIN32(ERROR_MRM_AUTOMERGE_ENABLED)')
    WIN32__ERROR_MRM_TOO_MANY_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0x80073B24, 'HRESULT_FROM_WIN32(ERROR_MRM_TOO_MANY_RESOURCES)')
    WIN32__ERROR_MRM_UNSUPPORTED_FILE_TYPE_FOR_MERGE: Final[ThirtyTwoBits] = cached_hresult(0x80073B25, 'HRESULT_FROM_WIN32(ERROR_MRM_UNSUPPORTED_FILE_TYPE_FOR_MERGE)')
    WIN32__ERROR_MRM_UNSUPPORTED_FILE_TYPE_FOR_LOAD_UNLOAD_PRI_FILE: Final[ThirtyTwoBits] = cached_hresult(0x80073B26, 'HRESULT_FROM_WIN32(ERROR_MRM_UNSUPPORTED_FILE_TYPE_FOR_LOAD_UNLOAD_PRI_FILE)')
    WIN32__ERROR_MRM_NO_CURRENT_VIEW_ON_THREAD: Final[ThirtyTwoBits] = cached_hresult(0x80073B27, 'HRESULT_FROM_WIN32(ERROR_MRM_NO_CURRENT_VIEW_ON_THREAD)')
    WIN32__ERROR_DIFFERENT_PROFILE_RESOURCE_MANAGER_EXIST: Final[ThirtyTwoBits] = cached_hresult(0x80073B28, 'HRESULT_FROM_WIN32(ERROR_DIFFERENT_PROFILE_RESOURCE_MANAGER_EXIST)')
    WIN32__ERROR_OPERATION_NOT_ALLOWED_FROM_SYSTEM_COMPONENT: Final[ThirtyTwoBits] = cached_hresult(0x80073B29, 'HRESULT_FROM_WIN32(ERROR_OPERATION_NOT_ALLOWED_FROM_SYSTEM_COMPONENT)')
    WIN32__ERROR_MRM_DIRECT_REF_TO_NON_DEFAULT_RESOURCE: Final[ThirtyTwoBits] = cached_hresult(0x80073B2A, 'HRESULT_FROM_WIN32(ERROR_MRM_DIRECT_REF_TO_NON_DEFAULT_RESOURCE)')
    WIN32__ERROR_MRM_GENERATION_COUNT_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80073B2B, 'HRESULT_FROM_WIN32(ERROR_MRM_GENERATION_COUNT_MISMATCH)')
    WIN32__ERROR_PRI_MERGE_VERSION_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80073B2C, 'HRESULT_FROM_WIN32(ERROR_PRI_MERGE_VERSION_MISMATCH)')
    WIN32__ERROR_PRI_MERGE_MISSING_SCHEMA: Final[ThirtyTwoBits] = cached_hresult(0x80073B2D, 'HRESULT_FROM_WIN32(ERROR_PRI_MERGE_MISSING_SCHEMA)')
    WIN32__ERROR_PRI_MERGE_LOAD_FILE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073B2E, 'HRESULT_FROM_WIN32(ERROR_PRI_MERGE_LOAD_FILE_FAILED)')
    WIN32__ERROR_PRI_MERGE_ADD_FILE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073B2F, 'HRESULT_FROM_WIN32(ERROR_PRI_MERGE_ADD_FILE_FAILED)')
    WIN32__ERROR_PRI_MERGE_WRITE_FILE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073B30, 'HRESULT_FROM_WIN32(ERROR_PRI_MERGE_WRITE_FILE_FAILED)')
    WIN32__ERROR_PRI_MERGE_MULTIPLE_PACKAGE_FAMILIES_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80073B31, 'HRESULT_FROM_WIN32(ERROR_PRI_MERGE_MULTIPLE_PACKAGE_FAMILIES_NOT_ALLOWED)')
    WIN32__ERROR_PRI_MERGE_MULTIPLE_MAIN_PACKAGES_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80073B32, 'HRESULT_FROM_WIN32(ERROR_PRI_MERGE_MULTIPLE_MAIN_PACKAGES_NOT_ALLOWED)')
    WIN32__ERROR_PRI_MERGE_BUNDLE_PACKAGES_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80073B33, 'HRESULT_FROM_WIN32(ERROR_PRI_MERGE_BUNDLE_PACKAGES_NOT_ALLOWED)')
    WIN32__ERROR_PRI_MERGE_MAIN_PACKAGE_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80073B34, 'HRESULT_FROM_WIN32(ERROR_PRI_MERGE_MAIN_PACKAGE_REQUIRED)')
    WIN32__ERROR_PRI_MERGE_RESOURCE_PACKAGE_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x80073B35, 'HRESULT_FROM_WIN32(ERROR_PRI_MERGE_RESOURCE_PACKAGE_REQUIRED)')
    WIN32__ERROR_PRI_MERGE_INVALID_FILE_NAME: Final[ThirtyTwoBits] = cached_hresult(0x80073B36, 'HRESULT_FROM_WIN32(ERROR_PRI_MERGE_INVALID_FILE_NAME)')
    WIN32__ERROR_MRM_PACKAGE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80073B37, 'HRESULT_FROM_WIN32(ERROR_MRM_PACKAGE_NOT_FOUND)')
    WIN32__ERROR_MRM_MISSING_DEFAULT_LANGUAGE: Final[ThirtyTwoBits] = cached_hresult(0x80073B38, 'HRESULT_FROM_WIN32(ERROR_MRM_MISSING_DEFAULT_LANGUAGE)')
    WIN32__ERROR_MRM_SCOPE_ITEM_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0x80073B39, 'HRESULT_FROM_WIN32(ERROR_MRM_SCOPE_ITEM_CONFLICT)')
    WIN32__ERROR_MCA_INVALID_CAPABILITIES_STRING: Final[ThirtyTwoBits] = cached_hresult(0x80073B60, 'HRESULT_FROM_WIN32(ERROR_MCA_INVALID_CAPABILITIES_STRING)')
    WIN32__ERROR_MCA_INVALID_VCP_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80073B61, 'HRESULT_FROM_WIN32(ERROR_MCA_INVALID_VCP_VERSION)')
    WIN32__ERROR_MCA_MONITOR_VIOLATES_MCCS_SPECIFICATION: Final[ThirtyTwoBits] = cached_hresult(0x80073B62, 'HRESULT_FROM_WIN32(ERROR_MCA_MONITOR_VIOLATES_MCCS_SPECIFICATION)')
    WIN32__ERROR_MCA_MCCS_VERSION_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80073B63, 'HRESULT_FROM_WIN32(ERROR_MCA_MCCS_VERSION_MISMATCH)')
    WIN32__ERROR_MCA_UNSUPPORTED_MCCS_VERSION: Final[ThirtyTwoBits] = cached_hresult(0x80073B64, 'HRESULT_FROM_WIN32(ERROR_MCA_UNSUPPORTED_MCCS_VERSION)')
    WIN32__ERROR_MCA_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x80073B65, 'HRESULT_FROM_WIN32(ERROR_MCA_INTERNAL_ERROR)')
    WIN32__ERROR_MCA_INVALID_TECHNOLOGY_TYPE_RETURNED: Final[ThirtyTwoBits] = cached_hresult(0x80073B66, 'HRESULT_FROM_WIN32(ERROR_MCA_INVALID_TECHNOLOGY_TYPE_RETURNED)')
    WIN32__ERROR_MCA_UNSUPPORTED_COLOR_TEMPERATURE: Final[ThirtyTwoBits] = cached_hresult(0x80073B67, 'HRESULT_FROM_WIN32(ERROR_MCA_UNSUPPORTED_COLOR_TEMPERATURE)')
    WIN32__ERROR_AMBIGUOUS_SYSTEM_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0x80073B92, 'HRESULT_FROM_WIN32(ERROR_AMBIGUOUS_SYSTEM_DEVICE)')
    WIN32__ERROR_SYSTEM_DEVICE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80073BC3, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_DEVICE_NOT_FOUND)')
    WIN32__ERROR_HASH_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80073BC4, 'HRESULT_FROM_WIN32(ERROR_HASH_NOT_SUPPORTED)')
    WIN32__ERROR_HASH_NOT_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0x80073BC5, 'HRESULT_FROM_WIN32(ERROR_HASH_NOT_PRESENT)')
    WIN32__ERROR_SECONDARY_IC_PROVIDER_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0x80073BD9, 'HRESULT_FROM_WIN32(ERROR_SECONDARY_IC_PROVIDER_NOT_REGISTERED)')
    WIN32__ERROR_GPIO_CLIENT_INFORMATION_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80073BDA, 'HRESULT_FROM_WIN32(ERROR_GPIO_CLIENT_INFORMATION_INVALID)')
    WIN32__ERROR_GPIO_VERSION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80073BDB, 'HRESULT_FROM_WIN32(ERROR_GPIO_VERSION_NOT_SUPPORTED)')
    WIN32__ERROR_GPIO_INVALID_REGISTRATION_PACKET: Final[ThirtyTwoBits] = cached_hresult(0x80073BDC, 'HRESULT_FROM_WIN32(ERROR_GPIO_INVALID_REGISTRATION_PACKET)')
    WIN32__ERROR_GPIO_OPERATION_DENIED: Final[ThirtyTwoBits] = cached_hresult(0x80073BDD, 'HRESULT_FROM_WIN32(ERROR_GPIO_OPERATION_DENIED)')
    WIN32__ERROR_GPIO_INCOMPATIBLE_CONNECT_MODE: Final[ThirtyTwoBits] = cached_hresult(0x80073BDE, 'HRESULT_FROM_WIN32(ERROR_GPIO_INCOMPATIBLE_CONNECT_MODE)')
    WIN32__ERROR_GPIO_INTERRUPT_ALREADY_UNMASKED: Final[ThirtyTwoBits] = cached_hresult(0x80073BDF, 'HRESULT_FROM_WIN32(ERROR_GPIO_INTERRUPT_ALREADY_UNMASKED)')
    WIN32__ERROR_CANNOT_SWITCH_RUNLEVEL: Final[ThirtyTwoBits] = cached_hresult(0x80073C28, 'HRESULT_FROM_WIN32(ERROR_CANNOT_SWITCH_RUNLEVEL)')
    WIN32__ERROR_INVALID_RUNLEVEL_SETTING: Final[ThirtyTwoBits] = cached_hresult(0x80073C29, 'HRESULT_FROM_WIN32(ERROR_INVALID_RUNLEVEL_SETTING)')
    WIN32__ERROR_RUNLEVEL_SWITCH_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x80073C2A, 'HRESULT_FROM_WIN32(ERROR_RUNLEVEL_SWITCH_TIMEOUT)')
    WIN32__ERROR_RUNLEVEL_SWITCH_AGENT_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x80073C2B, 'HRESULT_FROM_WIN32(ERROR_RUNLEVEL_SWITCH_AGENT_TIMEOUT)')
    WIN32__ERROR_RUNLEVEL_SWITCH_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x80073C2C, 'HRESULT_FROM_WIN32(ERROR_RUNLEVEL_SWITCH_IN_PROGRESS)')
    WIN32__ERROR_SERVICES_FAILED_AUTOSTART: Final[ThirtyTwoBits] = cached_hresult(0x80073C2D, 'HRESULT_FROM_WIN32(ERROR_SERVICES_FAILED_AUTOSTART)')
    WIN32__ERROR_COM_TASK_STOP_PENDING: Final[ThirtyTwoBits] = cached_hresult(0x80073C8D, 'HRESULT_FROM_WIN32(ERROR_COM_TASK_STOP_PENDING)')
    WIN32__ERROR_INSTALL_OPEN_PACKAGE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073CF0, 'HRESULT_FROM_WIN32(ERROR_INSTALL_OPEN_PACKAGE_FAILED)')
    WIN32__ERROR_INSTALL_PACKAGE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0x80073CF1, 'HRESULT_FROM_WIN32(ERROR_INSTALL_PACKAGE_NOT_FOUND)')
    WIN32__ERROR_INSTALL_INVALID_PACKAGE: Final[ThirtyTwoBits] = cached_hresult(0x80073CF2, 'HRESULT_FROM_WIN32(ERROR_INSTALL_INVALID_PACKAGE)')
    WIN32__ERROR_INSTALL_RESOLVE_DEPENDENCY_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073CF3, 'HRESULT_FROM_WIN32(ERROR_INSTALL_RESOLVE_DEPENDENCY_FAILED)')
    WIN32__ERROR_INSTALL_OUT_OF_DISK_SPACE: Final[ThirtyTwoBits] = cached_hresult(0x80073CF4, 'HRESULT_FROM_WIN32(ERROR_INSTALL_OUT_OF_DISK_SPACE)')
    WIN32__ERROR_INSTALL_NETWORK_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80073CF5, 'HRESULT_FROM_WIN32(ERROR_INSTALL_NETWORK_FAILURE)')
    WIN32__ERROR_INSTALL_REGISTRATION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80073CF6, 'HRESULT_FROM_WIN32(ERROR_INSTALL_REGISTRATION_FAILURE)')
    WIN32__ERROR_INSTALL_DEREGISTRATION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80073CF7, 'HRESULT_FROM_WIN32(ERROR_INSTALL_DEREGISTRATION_FAILURE)')
    WIN32__ERROR_INSTALL_CANCEL: Final[ThirtyTwoBits] = cached_hresult(0x80073CF8, 'HRESULT_FROM_WIN32(ERROR_INSTALL_CANCEL)')
    WIN32__ERROR_INSTALL_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073CF9, 'HRESULT_FROM_WIN32(ERROR_INSTALL_FAILED)')
    WIN32__ERROR_REMOVE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073CFA, 'HRESULT_FROM_WIN32(ERROR_REMOVE_FAILED)')
    WIN32__ERROR_PACKAGE_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x80073CFB, 'HRESULT_FROM_WIN32(ERROR_PACKAGE_ALREADY_EXISTS)')
    WIN32__ERROR_NEEDS_REMEDIATION: Final[ThirtyTwoBits] = cached_hresult(0x80073CFC, 'HRESULT_FROM_WIN32(ERROR_NEEDS_REMEDIATION)')
    WIN32__ERROR_INSTALL_PREREQUISITE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073CFD, 'HRESULT_FROM_WIN32(ERROR_INSTALL_PREREQUISITE_FAILED)')
    WIN32__ERROR_PACKAGE_REPOSITORY_CORRUPTED: Final[ThirtyTwoBits] = cached_hresult(0x80073CFE, 'HRESULT_FROM_WIN32(ERROR_PACKAGE_REPOSITORY_CORRUPTED)')
    WIN32__ERROR_INSTALL_POLICY_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0x80073CFF, 'HRESULT_FROM_WIN32(ERROR_INSTALL_POLICY_FAILURE)')
    WIN32__ERROR_PACKAGE_UPDATING: Final[ThirtyTwoBits] = cached_hresult(0x80073D00, 'HRESULT_FROM_WIN32(ERROR_PACKAGE_UPDATING)')
    WIN32__ERROR_DEPLOYMENT_BLOCKED_BY_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x80073D01, 'HRESULT_FROM_WIN32(ERROR_DEPLOYMENT_BLOCKED_BY_POLICY)')
    WIN32__ERROR_PACKAGES_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x80073D02, 'HRESULT_FROM_WIN32(ERROR_PACKAGES_IN_USE)')
    WIN32__ERROR_RECOVERY_FILE_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x80073D03, 'HRESULT_FROM_WIN32(ERROR_RECOVERY_FILE_CORRUPT)')
    WIN32__ERROR_INVALID_STAGED_SIGNATURE: Final[ThirtyTwoBits] = cached_hresult(0x80073D04, 'HRESULT_FROM_WIN32(ERROR_INVALID_STAGED_SIGNATURE)')
    WIN32__ERROR_DELETING_EXISTING_APPLICATIONDATA_STORE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073D05, 'HRESULT_FROM_WIN32(ERROR_DELETING_EXISTING_APPLICATIONDATA_STORE_FAILED)')
    WIN32__ERROR_INSTALL_PACKAGE_DOWNGRADE: Final[ThirtyTwoBits] = cached_hresult(0x80073D06, 'HRESULT_FROM_WIN32(ERROR_INSTALL_PACKAGE_DOWNGRADE)')
    WIN32__ERROR_SYSTEM_NEEDS_REMEDIATION: Final[ThirtyTwoBits] = cached_hresult(0x80073D07, 'HRESULT_FROM_WIN32(ERROR_SYSTEM_NEEDS_REMEDIATION)')
    WIN32__ERROR_APPX_INTEGRITY_FAILURE_CLR_NGEN: Final[ThirtyTwoBits] = cached_hresult(0x80073D08, 'HRESULT_FROM_WIN32(ERROR_APPX_INTEGRITY_FAILURE_CLR_NGEN)')
    WIN32__ERROR_RESILIENCY_FILE_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x80073D09, 'HRESULT_FROM_WIN32(ERROR_RESILIENCY_FILE_CORRUPT)')
    WIN32__ERROR_INSTALL_FIREWALL_SERVICE_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0x80073D0A, 'HRESULT_FROM_WIN32(ERROR_INSTALL_FIREWALL_SERVICE_NOT_RUNNING)')
    WIN32__ERROR_PACKAGE_MOVE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073D0B, 'HRESULT_FROM_WIN32(ERROR_PACKAGE_MOVE_FAILED)')
    WIN32__ERROR_INSTALL_VOLUME_NOT_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0x80073D0C, 'HRESULT_FROM_WIN32(ERROR_INSTALL_VOLUME_NOT_EMPTY)')
    WIN32__ERROR_INSTALL_VOLUME_OFFLINE: Final[ThirtyTwoBits] = cached_hresult(0x80073D0D, 'HRESULT_FROM_WIN32(ERROR_INSTALL_VOLUME_OFFLINE)')
    WIN32__ERROR_INSTALL_VOLUME_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x80073D0E, 'HRESULT_FROM_WIN32(ERROR_INSTALL_VOLUME_CORRUPT)')
    WIN32__ERROR_NEEDS_REGISTRATION: Final[ThirtyTwoBits] = cached_hresult(0x80073D0F, 'HRESULT_FROM_WIN32(ERROR_NEEDS_REGISTRATION)')
    WIN32__ERROR_INSTALL_WRONG_PROCESSOR_ARCHITECTURE: Final[ThirtyTwoBits] = cached_hresult(0x80073D10, 'HRESULT_FROM_WIN32(ERROR_INSTALL_WRONG_PROCESSOR_ARCHITECTURE)')
    WIN32__ERROR_DEV_SIDELOAD_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80073D11, 'HRESULT_FROM_WIN32(ERROR_DEV_SIDELOAD_LIMIT_EXCEEDED)')
    WIN32__ERROR_INSTALL_OPTIONAL_PACKAGE_REQUIRES_MAIN_PACKAGE: Final[ThirtyTwoBits] = cached_hresult(0x80073D12, 'HRESULT_FROM_WIN32(ERROR_INSTALL_OPTIONAL_PACKAGE_REQUIRES_MAIN_PACKAGE)')
    WIN32__ERROR_PACKAGE_NOT_SUPPORTED_ON_FILESYSTEM: Final[ThirtyTwoBits] = cached_hresult(0x80073D13, 'HRESULT_FROM_WIN32(ERROR_PACKAGE_NOT_SUPPORTED_ON_FILESYSTEM)')
    WIN32__ERROR_PACKAGE_MOVE_BLOCKED_BY_STREAMING: Final[ThirtyTwoBits] = cached_hresult(0x80073D14, 'HRESULT_FROM_WIN32(ERROR_PACKAGE_MOVE_BLOCKED_BY_STREAMING)')
    WIN32__ERROR_INSTALL_OPTIONAL_PACKAGE_APPLICATIONID_NOT_UNIQUE: Final[ThirtyTwoBits] = cached_hresult(0x80073D15, 'HRESULT_FROM_WIN32(ERROR_INSTALL_OPTIONAL_PACKAGE_APPLICATIONID_NOT_UNIQUE)')
    WIN32__ERROR_PACKAGE_STAGING_ONHOLD: Final[ThirtyTwoBits] = cached_hresult(0x80073D16, 'HRESULT_FROM_WIN32(ERROR_PACKAGE_STAGING_ONHOLD)')
    WIN32__ERROR_INSTALL_INVALID_RELATED_SET_UPDATE: Final[ThirtyTwoBits] = cached_hresult(0x80073D17, 'HRESULT_FROM_WIN32(ERROR_INSTALL_INVALID_RELATED_SET_UPDATE)')
    WIN32__ERROR_INSTALL_OPTIONAL_PACKAGE_REQUIRES_MAIN_PACKAGE_FULLTRUST_CAPABILITY: Final[ThirtyTwoBits] = cached_hresult(0x80073D18, 'HRESULT_FROM_WIN32(ERROR_INSTALL_OPTIONAL_PACKAGE_REQUIRES_MAIN_PACKAGE_FULLTRUST_CAPABILITY)')
    WIN32__ERROR_DEPLOYMENT_BLOCKED_BY_USER_LOG_OFF: Final[ThirtyTwoBits] = cached_hresult(0x80073D19, 'HRESULT_FROM_WIN32(ERROR_DEPLOYMENT_BLOCKED_BY_USER_LOG_OFF)')
    WIN32__ERROR_PROVISION_OPTIONAL_PACKAGE_REQUIRES_MAIN_PACKAGE_PROVISIONED: Final[ThirtyTwoBits] = cached_hresult(0x80073D1A, 'HRESULT_FROM_WIN32(ERROR_PROVISION_OPTIONAL_PACKAGE_REQUIRES_MAIN_PACKAGE_PROVISIONED)')
    WIN32__ERROR_PACKAGES_REPUTATION_CHECK_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073D1B, 'HRESULT_FROM_WIN32(ERROR_PACKAGES_REPUTATION_CHECK_FAILED)')
    WIN32__ERROR_PACKAGES_REPUTATION_CHECK_TIMEDOUT: Final[ThirtyTwoBits] = cached_hresult(0x80073D1C, 'HRESULT_FROM_WIN32(ERROR_PACKAGES_REPUTATION_CHECK_TIMEDOUT)')
    WIN32__ERROR_DEPLOYMENT_OPTION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80073D1D, 'HRESULT_FROM_WIN32(ERROR_DEPLOYMENT_OPTION_NOT_SUPPORTED)')
    WIN32__ERROR_APPINSTALLER_ACTIVATION_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x80073D1E, 'HRESULT_FROM_WIN32(ERROR_APPINSTALLER_ACTIVATION_BLOCKED)')
    WIN32__ERROR_REGISTRATION_FROM_REMOTE_DRIVE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x80073D1F, 'HRESULT_FROM_WIN32(ERROR_REGISTRATION_FROM_REMOTE_DRIVE_NOT_SUPPORTED)')
    WIN32__ERROR_APPX_RAW_DATA_WRITE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073D20, 'HRESULT_FROM_WIN32(ERROR_APPX_RAW_DATA_WRITE_FAILED)')
    WIN32__ERROR_DEPLOYMENT_BLOCKED_BY_VOLUME_POLICY_PACKAGE: Final[ThirtyTwoBits] = cached_hresult(0x80073D21, 'HRESULT_FROM_WIN32(ERROR_DEPLOYMENT_BLOCKED_BY_VOLUME_POLICY_PACKAGE)')
    WIN32__ERROR_DEPLOYMENT_BLOCKED_BY_VOLUME_POLICY_MACHINE: Final[ThirtyTwoBits] = cached_hresult(0x80073D22, 'HRESULT_FROM_WIN32(ERROR_DEPLOYMENT_BLOCKED_BY_VOLUME_POLICY_MACHINE)')
    WIN32__ERROR_DEPLOYMENT_BLOCKED_BY_PROFILE_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x80073D23, 'HRESULT_FROM_WIN32(ERROR_DEPLOYMENT_BLOCKED_BY_PROFILE_POLICY)')
    WIN32__ERROR_DEPLOYMENT_FAILED_CONFLICTING_MUTABLE_PACKAGE_DIRECTORY: Final[ThirtyTwoBits] = cached_hresult(0x80073D24, 'HRESULT_FROM_WIN32(ERROR_DEPLOYMENT_FAILED_CONFLICTING_MUTABLE_PACKAGE_DIRECTORY)')
    WIN32__ERROR_SINGLETON_RESOURCE_INSTALLED_IN_ACTIVE_USER: Final[ThirtyTwoBits] = cached_hresult(0x80073D25, 'HRESULT_FROM_WIN32(ERROR_SINGLETON_RESOURCE_INSTALLED_IN_ACTIVE_USER)')
    WIN32__ERROR_DIFFERENT_VERSION_OF_PACKAGED_SERVICE_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x80073D26, 'HRESULT_FROM_WIN32(ERROR_DIFFERENT_VERSION_OF_PACKAGED_SERVICE_INSTALLED)')
    WIN32__ERROR_SERVICE_EXISTS_AS_NON_PACKAGED_SERVICE: Final[ThirtyTwoBits] = cached_hresult(0x80073D27, 'HRESULT_FROM_WIN32(ERROR_SERVICE_EXISTS_AS_NON_PACKAGED_SERVICE)')
    WIN32__ERROR_PACKAGED_SERVICE_REQUIRES_ADMIN_PRIVILEGES: Final[ThirtyTwoBits] = cached_hresult(0x80073D28, 'HRESULT_FROM_WIN32(ERROR_PACKAGED_SERVICE_REQUIRES_ADMIN_PRIVILEGES)')
    WIN32__ERROR_REDIRECTION_TO_DEFAULT_ACCOUNT_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80073D29, 'HRESULT_FROM_WIN32(ERROR_REDIRECTION_TO_DEFAULT_ACCOUNT_NOT_ALLOWED)')
    WIN32__ERROR_PACKAGE_LACKS_CAPABILITY_TO_DEPLOY_ON_HOST: Final[ThirtyTwoBits] = cached_hresult(0x80073D2A, 'HRESULT_FROM_WIN32(ERROR_PACKAGE_LACKS_CAPABILITY_TO_DEPLOY_ON_HOST)')
    WIN32__ERROR_UNSIGNED_PACKAGE_INVALID_CONTENT: Final[ThirtyTwoBits] = cached_hresult(0x80073D2B, 'HRESULT_FROM_WIN32(ERROR_UNSIGNED_PACKAGE_INVALID_CONTENT)')
    WIN32__ERROR_UNSIGNED_PACKAGE_INVALID_PUBLISHER_NAMESPACE: Final[ThirtyTwoBits] = cached_hresult(0x80073D2C, 'HRESULT_FROM_WIN32(ERROR_UNSIGNED_PACKAGE_INVALID_PUBLISHER_NAMESPACE)')
    WIN32__ERROR_SIGNED_PACKAGE_INVALID_PUBLISHER_NAMESPACE: Final[ThirtyTwoBits] = cached_hresult(0x80073D2D, 'HRESULT_FROM_WIN32(ERROR_SIGNED_PACKAGE_INVALID_PUBLISHER_NAMESPACE)')
    WIN32__ERROR_PACKAGE_EXTERNAL_LOCATION_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80073D2E, 'HRESULT_FROM_WIN32(ERROR_PACKAGE_EXTERNAL_LOCATION_NOT_ALLOWED)')
    WIN32__ERROR_INSTALL_FULLTRUST_HOSTRUNTIME_REQUIRES_MAIN_PACKAGE_FULLTRUST_CAPABILITY: Final[ThirtyTwoBits] = cached_hresult(0x80073D2F, 'HRESULT_FROM_WIN32(ERROR_INSTALL_FULLTRUST_HOSTRUNTIME_REQUIRES_MAIN_PACKAGE_FULLTRUST_CAPABILITY)')
    WIN32__ERROR_PACKAGE_LACKS_CAPABILITY_FOR_MANDATORY_STARTUPTASKS: Final[ThirtyTwoBits] = cached_hresult(0x80073D30, 'HRESULT_FROM_WIN32(ERROR_PACKAGE_LACKS_CAPABILITY_FOR_MANDATORY_STARTUPTASKS)')
    WIN32__ERROR_INSTALL_RESOLVE_HOSTRUNTIME_DEPENDENCY_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073D31, 'HRESULT_FROM_WIN32(ERROR_INSTALL_RESOLVE_HOSTRUNTIME_DEPENDENCY_FAILED)')
    WIN32__ERROR_MACHINE_SCOPE_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80073D32, 'HRESULT_FROM_WIN32(ERROR_MACHINE_SCOPE_NOT_ALLOWED)')
    WIN32__ERROR_CLASSIC_COMPAT_MODE_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0x80073D33, 'HRESULT_FROM_WIN32(ERROR_CLASSIC_COMPAT_MODE_NOT_ALLOWED)')
    WIN32__ERROR_STAGEFROMUPDATEAGENT_PACKAGE_NOT_APPLICABLE: Final[ThirtyTwoBits] = cached_hresult(0x80073D34, 'HRESULT_FROM_WIN32(ERROR_STAGEFROMUPDATEAGENT_PACKAGE_NOT_APPLICABLE)')
    WIN32__ERROR_PACKAGE_NOT_REGISTERED_FOR_USER: Final[ThirtyTwoBits] = cached_hresult(0x80073D35, 'HRESULT_FROM_WIN32(ERROR_PACKAGE_NOT_REGISTERED_FOR_USER)')
    WIN32__ERROR_PACKAGE_NAME_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x80073D36, 'HRESULT_FROM_WIN32(ERROR_PACKAGE_NAME_MISMATCH)')
    WIN32__ERROR_APPINSTALLER_URI_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0x80073D37, 'HRESULT_FROM_WIN32(ERROR_APPINSTALLER_URI_IN_USE)')
    WIN32__ERROR_APPINSTALLER_IS_MANAGED_BY_SYSTEM: Final[ThirtyTwoBits] = cached_hresult(0x80073D38, 'HRESULT_FROM_WIN32(ERROR_APPINSTALLER_IS_MANAGED_BY_SYSTEM)')
    WIN32__APPMODEL_ERROR_NO_PACKAGE: Final[ThirtyTwoBits] = cached_hresult(0x80073D54, 'HRESULT_FROM_WIN32(APPMODEL_ERROR_NO_PACKAGE)')
    WIN32__APPMODEL_ERROR_PACKAGE_RUNTIME_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x80073D55, 'HRESULT_FROM_WIN32(APPMODEL_ERROR_PACKAGE_RUNTIME_CORRUPT)')
    WIN32__APPMODEL_ERROR_PACKAGE_IDENTITY_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0x80073D56, 'HRESULT_FROM_WIN32(APPMODEL_ERROR_PACKAGE_IDENTITY_CORRUPT)')
    WIN32__APPMODEL_ERROR_NO_APPLICATION: Final[ThirtyTwoBits] = cached_hresult(0x80073D57, 'HRESULT_FROM_WIN32(APPMODEL_ERROR_NO_APPLICATION)')
    WIN32__APPMODEL_ERROR_DYNAMIC_PROPERTY_READ_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073D58, 'HRESULT_FROM_WIN32(APPMODEL_ERROR_DYNAMIC_PROPERTY_READ_FAILED)')
    WIN32__APPMODEL_ERROR_DYNAMIC_PROPERTY_INVALID: Final[ThirtyTwoBits] = cached_hresult(0x80073D59, 'HRESULT_FROM_WIN32(APPMODEL_ERROR_DYNAMIC_PROPERTY_INVALID)')
    WIN32__APPMODEL_ERROR_PACKAGE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80073D5A, 'HRESULT_FROM_WIN32(APPMODEL_ERROR_PACKAGE_NOT_AVAILABLE)')
    WIN32__APPMODEL_ERROR_NO_MUTABLE_DIRECTORY: Final[ThirtyTwoBits] = cached_hresult(0x80073D5B, 'HRESULT_FROM_WIN32(APPMODEL_ERROR_NO_MUTABLE_DIRECTORY)')
    WIN32__ERROR_STATE_LOAD_STORE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073DB8, 'HRESULT_FROM_WIN32(ERROR_STATE_LOAD_STORE_FAILED)')
    WIN32__ERROR_STATE_GET_VERSION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073DB9, 'HRESULT_FROM_WIN32(ERROR_STATE_GET_VERSION_FAILED)')
    WIN32__ERROR_STATE_SET_VERSION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073DBA, 'HRESULT_FROM_WIN32(ERROR_STATE_SET_VERSION_FAILED)')
    WIN32__ERROR_STATE_STRUCTURED_RESET_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073DBB, 'HRESULT_FROM_WIN32(ERROR_STATE_STRUCTURED_RESET_FAILED)')
    WIN32__ERROR_STATE_OPEN_CONTAINER_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073DBC, 'HRESULT_FROM_WIN32(ERROR_STATE_OPEN_CONTAINER_FAILED)')
    WIN32__ERROR_STATE_CREATE_CONTAINER_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073DBD, 'HRESULT_FROM_WIN32(ERROR_STATE_CREATE_CONTAINER_FAILED)')
    WIN32__ERROR_STATE_DELETE_CONTAINER_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073DBE, 'HRESULT_FROM_WIN32(ERROR_STATE_DELETE_CONTAINER_FAILED)')
    WIN32__ERROR_STATE_READ_SETTING_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073DBF, 'HRESULT_FROM_WIN32(ERROR_STATE_READ_SETTING_FAILED)')
    WIN32__ERROR_STATE_WRITE_SETTING_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073DC0, 'HRESULT_FROM_WIN32(ERROR_STATE_WRITE_SETTING_FAILED)')
    WIN32__ERROR_STATE_DELETE_SETTING_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073DC1, 'HRESULT_FROM_WIN32(ERROR_STATE_DELETE_SETTING_FAILED)')
    WIN32__ERROR_STATE_QUERY_SETTING_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073DC2, 'HRESULT_FROM_WIN32(ERROR_STATE_QUERY_SETTING_FAILED)')
    WIN32__ERROR_STATE_READ_COMPOSITE_SETTING_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073DC3, 'HRESULT_FROM_WIN32(ERROR_STATE_READ_COMPOSITE_SETTING_FAILED)')
    WIN32__ERROR_STATE_WRITE_COMPOSITE_SETTING_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073DC4, 'HRESULT_FROM_WIN32(ERROR_STATE_WRITE_COMPOSITE_SETTING_FAILED)')
    WIN32__ERROR_STATE_ENUMERATE_CONTAINER_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073DC5, 'HRESULT_FROM_WIN32(ERROR_STATE_ENUMERATE_CONTAINER_FAILED)')
    WIN32__ERROR_STATE_ENUMERATE_SETTINGS_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x80073DC6, 'HRESULT_FROM_WIN32(ERROR_STATE_ENUMERATE_SETTINGS_FAILED)')
    WIN32__ERROR_STATE_COMPOSITE_SETTING_VALUE_SIZE_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80073DC7, 'HRESULT_FROM_WIN32(ERROR_STATE_COMPOSITE_SETTING_VALUE_SIZE_LIMIT_EXCEEDED)')
    WIN32__ERROR_STATE_SETTING_VALUE_SIZE_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80073DC8, 'HRESULT_FROM_WIN32(ERROR_STATE_SETTING_VALUE_SIZE_LIMIT_EXCEEDED)')
    WIN32__ERROR_STATE_SETTING_NAME_SIZE_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80073DC9, 'HRESULT_FROM_WIN32(ERROR_STATE_SETTING_NAME_SIZE_LIMIT_EXCEEDED)')
    WIN32__ERROR_STATE_CONTAINER_NAME_SIZE_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0x80073DCA, 'HRESULT_FROM_WIN32(ERROR_STATE_CONTAINER_NAME_SIZE_LIMIT_EXCEEDED)')
    WIN32__ERROR_API_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0x80073DE1, 'HRESULT_FROM_WIN32(ERROR_API_UNAVAILABLE)')
    WIN32__STORE_ERROR_UNLICENSED: Final[ThirtyTwoBits] = cached_hresult(0x80073DF5, 'HRESULT_FROM_WIN32(STORE_ERROR_UNLICENSED)')
    WIN32__STORE_ERROR_UNLICENSED_USER: Final[ThirtyTwoBits] = cached_hresult(0x80073DF6, 'HRESULT_FROM_WIN32(STORE_ERROR_UNLICENSED_USER)')
    WIN32__STORE_ERROR_PENDING_COM_TRANSACTION: Final[ThirtyTwoBits] = cached_hresult(0x80073DF7, 'HRESULT_FROM_WIN32(STORE_ERROR_PENDING_COM_TRANSACTION)')
    WIN32__STORE_ERROR_LICENSE_REVOKED: Final[ThirtyTwoBits] = cached_hresult(0x80073DF8, 'HRESULT_FROM_WIN32(STORE_ERROR_LICENSE_REVOKED)')

    # NTSTATUS values converted to HRESULT
    NTSTATUS__STATUS_SUCCESS: Final[ThirtyTwoBits] = cached_hresult(0x10000000, 'HRESULT_FROM_NT(STATUS_SUCCESS)')
    NTSTATUS__STATUS_ABANDONED: Final[ThirtyTwoBits] = cached_hresult(0x10000080, 'HRESULT_FROM_NT(STATUS_ABANDONED)')
    NTSTATUS__STATUS_USER_APC: Final[ThirtyTwoBits] = cached_hresult(0x100000C0, 'HRESULT_FROM_NT(STATUS_USER_APC)')
    NTSTATUS__STATUS_ALREADY_COMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x100000FF, 'HRESULT_FROM_NT(STATUS_ALREADY_COMPLETE)')
    NTSTATUS__STATUS_KERNEL_APC: Final[ThirtyTwoBits] = cached_hresult(0x10000100, 'HRESULT_FROM_NT(STATUS_KERNEL_APC)')
    NTSTATUS__STATUS_ALERTED: Final[ThirtyTwoBits] = cached_hresult(0x10000101, 'HRESULT_FROM_NT(STATUS_ALERTED)')
    NTSTATUS__STATUS_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x10000102, 'HRESULT_FROM_NT(STATUS_TIMEOUT)')
    NTSTATUS__STATUS_PENDING: Final[ThirtyTwoBits] = cached_hresult(0x10000103, 'HRESULT_FROM_NT(STATUS_PENDING)')
    NTSTATUS__STATUS_REPARSE: Final[ThirtyTwoBits] = cached_hresult(0x10000104, 'HRESULT_FROM_NT(STATUS_REPARSE)')
    NTSTATUS__STATUS_MORE_ENTRIES: Final[ThirtyTwoBits] = cached_hresult(0x10000105, 'HRESULT_FROM_NT(STATUS_MORE_ENTRIES)')
    NTSTATUS__STATUS_NOT_ALL_ASSIGNED: Final[ThirtyTwoBits] = cached_hresult(0x10000106, 'HRESULT_FROM_NT(STATUS_NOT_ALL_ASSIGNED)')
    NTSTATUS__STATUS_SOME_NOT_MAPPED: Final[ThirtyTwoBits] = cached_hresult(0x10000107, 'HRESULT_FROM_NT(STATUS_SOME_NOT_MAPPED)')
    NTSTATUS__STATUS_OPLOCK_BREAK_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x10000108, 'HRESULT_FROM_NT(STATUS_OPLOCK_BREAK_IN_PROGRESS)')
    NTSTATUS__STATUS_VOLUME_MOUNTED: Final[ThirtyTwoBits] = cached_hresult(0x10000109, 'HRESULT_FROM_NT(STATUS_VOLUME_MOUNTED)')
    NTSTATUS__STATUS_RXACT_COMMITTED: Final[ThirtyTwoBits] = cached_hresult(0x1000010A, 'HRESULT_FROM_NT(STATUS_RXACT_COMMITTED)')
    NTSTATUS__STATUS_NOTIFY_CLEANUP: Final[ThirtyTwoBits] = cached_hresult(0x1000010B, 'HRESULT_FROM_NT(STATUS_NOTIFY_CLEANUP)')
    NTSTATUS__STATUS_NOTIFY_ENUM_DIR: Final[ThirtyTwoBits] = cached_hresult(0x1000010C, 'HRESULT_FROM_NT(STATUS_NOTIFY_ENUM_DIR)')
    NTSTATUS__STATUS_NO_QUOTAS_FOR_ACCOUNT: Final[ThirtyTwoBits] = cached_hresult(0x1000010D, 'HRESULT_FROM_NT(STATUS_NO_QUOTAS_FOR_ACCOUNT)')
    NTSTATUS__STATUS_PRIMARY_TRANSPORT_CONNECT_FAILED: Final[ThirtyTwoBits] = cached_hresult(0x1000010E, 'HRESULT_FROM_NT(STATUS_PRIMARY_TRANSPORT_CONNECT_FAILED)')
    NTSTATUS__STATUS_PAGE_FAULT_TRANSITION: Final[ThirtyTwoBits] = cached_hresult(0x10000110, 'HRESULT_FROM_NT(STATUS_PAGE_FAULT_TRANSITION)')
    NTSTATUS__STATUS_PAGE_FAULT_DEMAND_ZERO: Final[ThirtyTwoBits] = cached_hresult(0x10000111, 'HRESULT_FROM_NT(STATUS_PAGE_FAULT_DEMAND_ZERO)')
    NTSTATUS__STATUS_PAGE_FAULT_COPY_ON_WRITE: Final[ThirtyTwoBits] = cached_hresult(0x10000112, 'HRESULT_FROM_NT(STATUS_PAGE_FAULT_COPY_ON_WRITE)')
    NTSTATUS__STATUS_PAGE_FAULT_GUARD_PAGE: Final[ThirtyTwoBits] = cached_hresult(0x10000113, 'HRESULT_FROM_NT(STATUS_PAGE_FAULT_GUARD_PAGE)')
    NTSTATUS__STATUS_PAGE_FAULT_PAGING_FILE: Final[ThirtyTwoBits] = cached_hresult(0x10000114, 'HRESULT_FROM_NT(STATUS_PAGE_FAULT_PAGING_FILE)')
    NTSTATUS__STATUS_CACHE_PAGE_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x10000115, 'HRESULT_FROM_NT(STATUS_CACHE_PAGE_LOCKED)')
    NTSTATUS__STATUS_CRASH_DUMP: Final[ThirtyTwoBits] = cached_hresult(0x10000116, 'HRESULT_FROM_NT(STATUS_CRASH_DUMP)')
    NTSTATUS__STATUS_BUFFER_ALL_ZEROS: Final[ThirtyTwoBits] = cached_hresult(0x10000117, 'HRESULT_FROM_NT(STATUS_BUFFER_ALL_ZEROS)')
    NTSTATUS__STATUS_REPARSE_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0x10000118, 'HRESULT_FROM_NT(STATUS_REPARSE_OBJECT)')
    NTSTATUS__STATUS_RESOURCE_REQUIREMENTS_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0x10000119, 'HRESULT_FROM_NT(STATUS_RESOURCE_REQUIREMENTS_CHANGED)')
    NTSTATUS__STATUS_TRANSLATION_COMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x10000120, 'HRESULT_FROM_NT(STATUS_TRANSLATION_COMPLETE)')
    NTSTATUS__STATUS_DS_MEMBERSHIP_EVALUATED_LOCALLY: Final[ThirtyTwoBits] = cached_hresult(0x10000121, 'HRESULT_FROM_NT(STATUS_DS_MEMBERSHIP_EVALUATED_LOCALLY)')
    NTSTATUS__STATUS_NOTHING_TO_TERMINATE: Final[ThirtyTwoBits] = cached_hresult(0x10000122, 'HRESULT_FROM_NT(STATUS_NOTHING_TO_TERMINATE)')
    NTSTATUS__STATUS_PROCESS_NOT_IN_JOB: Final[ThirtyTwoBits] = cached_hresult(0x10000123, 'HRESULT_FROM_NT(STATUS_PROCESS_NOT_IN_JOB)')
    NTSTATUS__STATUS_PROCESS_IN_JOB: Final[ThirtyTwoBits] = cached_hresult(0x10000124, 'HRESULT_FROM_NT(STATUS_PROCESS_IN_JOB)')
    NTSTATUS__STATUS_VOLSNAP_HIBERNATE_READY: Final[ThirtyTwoBits] = cached_hresult(0x10000125, 'HRESULT_FROM_NT(STATUS_VOLSNAP_HIBERNATE_READY)')
    NTSTATUS__STATUS_FSFILTER_OP_COMPLETED_SUCCESSFULLY: Final[ThirtyTwoBits] = cached_hresult(0x10000126, 'HRESULT_FROM_NT(STATUS_FSFILTER_OP_COMPLETED_SUCCESSFULLY)')
    NTSTATUS__STATUS_INTERRUPT_VECTOR_ALREADY_CONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x10000127, 'HRESULT_FROM_NT(STATUS_INTERRUPT_VECTOR_ALREADY_CONNECTED)')
    NTSTATUS__STATUS_INTERRUPT_STILL_CONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x10000128, 'HRESULT_FROM_NT(STATUS_INTERRUPT_STILL_CONNECTED)')
    NTSTATUS__STATUS_PROCESS_CLONED: Final[ThirtyTwoBits] = cached_hresult(0x10000129, 'HRESULT_FROM_NT(STATUS_PROCESS_CLONED)')
    NTSTATUS__STATUS_FILE_LOCKED_WITH_ONLY_READERS: Final[ThirtyTwoBits] = cached_hresult(0x1000012A, 'HRESULT_FROM_NT(STATUS_FILE_LOCKED_WITH_ONLY_READERS)')
    NTSTATUS__STATUS_FILE_LOCKED_WITH_WRITERS: Final[ThirtyTwoBits] = cached_hresult(0x1000012B, 'HRESULT_FROM_NT(STATUS_FILE_LOCKED_WITH_WRITERS)')
    NTSTATUS__STATUS_VALID_IMAGE_HASH: Final[ThirtyTwoBits] = cached_hresult(0x1000012C, 'HRESULT_FROM_NT(STATUS_VALID_IMAGE_HASH)')
    NTSTATUS__STATUS_VALID_CATALOG_HASH: Final[ThirtyTwoBits] = cached_hresult(0x1000012D, 'HRESULT_FROM_NT(STATUS_VALID_CATALOG_HASH)')
    NTSTATUS__STATUS_VALID_STRONG_CODE_HASH: Final[ThirtyTwoBits] = cached_hresult(0x1000012E, 'HRESULT_FROM_NT(STATUS_VALID_STRONG_CODE_HASH)')
    NTSTATUS__STATUS_GHOSTED: Final[ThirtyTwoBits] = cached_hresult(0x1000012F, 'HRESULT_FROM_NT(STATUS_GHOSTED)')
    NTSTATUS__STATUS_DATA_OVERWRITTEN: Final[ThirtyTwoBits] = cached_hresult(0x10000130, 'HRESULT_FROM_NT(STATUS_DATA_OVERWRITTEN)')
    NTSTATUS__STATUS_RESOURCEMANAGER_READ_ONLY: Final[ThirtyTwoBits] = cached_hresult(0x10000202, 'HRESULT_FROM_NT(STATUS_RESOURCEMANAGER_READ_ONLY)')
    NTSTATUS__STATUS_RING_PREVIOUSLY_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0x10000210, 'HRESULT_FROM_NT(STATUS_RING_PREVIOUSLY_EMPTY)')
    NTSTATUS__STATUS_RING_PREVIOUSLY_FULL: Final[ThirtyTwoBits] = cached_hresult(0x10000211, 'HRESULT_FROM_NT(STATUS_RING_PREVIOUSLY_FULL)')
    NTSTATUS__STATUS_RING_PREVIOUSLY_ABOVE_QUOTA: Final[ThirtyTwoBits] = cached_hresult(0x10000212, 'HRESULT_FROM_NT(STATUS_RING_PREVIOUSLY_ABOVE_QUOTA)')
    NTSTATUS__STATUS_RING_NEWLY_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0x10000213, 'HRESULT_FROM_NT(STATUS_RING_NEWLY_EMPTY)')
    NTSTATUS__STATUS_RING_SIGNAL_OPPOSITE_ENDPOINT: Final[ThirtyTwoBits] = cached_hresult(0x10000214, 'HRESULT_FROM_NT(STATUS_RING_SIGNAL_OPPOSITE_ENDPOINT)')
    NTSTATUS__STATUS_OPLOCK_SWITCHED_TO_NEW_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x10000215, 'HRESULT_FROM_NT(STATUS_OPLOCK_SWITCHED_TO_NEW_HANDLE)')
    NTSTATUS__STATUS_OPLOCK_HANDLE_CLOSED: Final[ThirtyTwoBits] = cached_hresult(0x10000216, 'HRESULT_FROM_NT(STATUS_OPLOCK_HANDLE_CLOSED)')
    NTSTATUS__STATUS_WAIT_FOR_OPLOCK: Final[ThirtyTwoBits] = cached_hresult(0x10000367, 'HRESULT_FROM_NT(STATUS_WAIT_FOR_OPLOCK)')
    NTSTATUS__STATUS_REPARSE_GLOBAL: Final[ThirtyTwoBits] = cached_hresult(0x10000368, 'HRESULT_FROM_NT(STATUS_REPARSE_GLOBAL)')
    NTSTATUS__STATUS_PAGE_FAULT_RETRY: Final[ThirtyTwoBits] = cached_hresult(0x10000369, 'HRESULT_FROM_NT(STATUS_PAGE_FAULT_RETRY)')
    NTSTATUS__STATUS_FLT_IO_COMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x101C0001, 'HRESULT_FROM_NT(STATUS_FLT_IO_COMPLETE)')
    NTSTATUS__STATUS_RTPM_CONTEXT_CONTINUE: Final[ThirtyTwoBits] = cached_hresult(0x10293000, 'HRESULT_FROM_NT(STATUS_RTPM_CONTEXT_CONTINUE)')
    NTSTATUS__STATUS_RTPM_CONTEXT_COMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x10293001, 'HRESULT_FROM_NT(STATUS_RTPM_CONTEXT_COMPLETE)')
    NTSTATUS__STATUS_HV_PENDING_PAGE_REQUESTS: Final[ThirtyTwoBits] = cached_hresult(0x10350059, 'HRESULT_FROM_NT(STATUS_HV_PENDING_PAGE_REQUESTS)')
    NTSTATUS__STATUS_SPACES_REPAIRED: Final[ThirtyTwoBits] = cached_hresult(0x10E70000, 'HRESULT_FROM_NT(STATUS_SPACES_REPAIRED)')
    NTSTATUS__STATUS_SPACES_PAUSE: Final[ThirtyTwoBits] = cached_hresult(0x10E70001, 'HRESULT_FROM_NT(STATUS_SPACES_PAUSE)')
    NTSTATUS__STATUS_SPACES_COMPLETE: Final[ThirtyTwoBits] = cached_hresult(0x10E70002, 'HRESULT_FROM_NT(STATUS_SPACES_COMPLETE)')
    NTSTATUS__STATUS_SPACES_REDIRECT: Final[ThirtyTwoBits] = cached_hresult(0x10E70003, 'HRESULT_FROM_NT(STATUS_SPACES_REDIRECT)')
    NTSTATUS__STATUS_OBJECT_NAME_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0x50000000, 'HRESULT_FROM_NT(STATUS_OBJECT_NAME_EXISTS)')
    NTSTATUS__STATUS_THREAD_WAS_SUSPENDED: Final[ThirtyTwoBits] = cached_hresult(0x50000001, 'HRESULT_FROM_NT(STATUS_THREAD_WAS_SUSPENDED)')
    NTSTATUS__STATUS_WORKING_SET_LIMIT_RANGE: Final[ThirtyTwoBits] = cached_hresult(0x50000002, 'HRESULT_FROM_NT(STATUS_WORKING_SET_LIMIT_RANGE)')
    NTSTATUS__STATUS_IMAGE_NOT_AT_BASE: Final[ThirtyTwoBits] = cached_hresult(0x50000003, 'HRESULT_FROM_NT(STATUS_IMAGE_NOT_AT_BASE)')
    NTSTATUS__STATUS_RXACT_STATE_CREATED: Final[ThirtyTwoBits] = cached_hresult(0x50000004, 'HRESULT_FROM_NT(STATUS_RXACT_STATE_CREATED)')
    NTSTATUS__STATUS_SEGMENT_NOTIFICATION: Final[ThirtyTwoBits] = cached_hresult(0x50000005, 'HRESULT_FROM_NT(STATUS_SEGMENT_NOTIFICATION)')
    NTSTATUS__STATUS_LOCAL_USER_SESSION_KEY: Final[ThirtyTwoBits] = cached_hresult(0x50000006, 'HRESULT_FROM_NT(STATUS_LOCAL_USER_SESSION_KEY)')
    NTSTATUS__STATUS_BAD_CURRENT_DIRECTORY: Final[ThirtyTwoBits] = cached_hresult(0x50000007, 'HRESULT_FROM_NT(STATUS_BAD_CURRENT_DIRECTORY)')
    NTSTATUS__STATUS_SERIAL_MORE_WRITES: Final[ThirtyTwoBits] = cached_hresult(0x50000008, 'HRESULT_FROM_NT(STATUS_SERIAL_MORE_WRITES)')
    NTSTATUS__STATUS_REGISTRY_RECOVERED: Final[ThirtyTwoBits] = cached_hresult(0x50000009, 'HRESULT_FROM_NT(STATUS_REGISTRY_RECOVERED)')
    NTSTATUS__STATUS_FT_READ_RECOVERY_FROM_BACKUP: Final[ThirtyTwoBits] = cached_hresult(0x5000000A, 'HRESULT_FROM_NT(STATUS_FT_READ_RECOVERY_FROM_BACKUP)')
    NTSTATUS__STATUS_FT_WRITE_RECOVERY: Final[ThirtyTwoBits] = cached_hresult(0x5000000B, 'HRESULT_FROM_NT(STATUS_FT_WRITE_RECOVERY)')
    NTSTATUS__STATUS_SERIAL_COUNTER_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0x5000000C, 'HRESULT_FROM_NT(STATUS_SERIAL_COUNTER_TIMEOUT)')
    NTSTATUS__STATUS_NULL_LM_PASSWORD: Final[ThirtyTwoBits] = cached_hresult(0x5000000D, 'HRESULT_FROM_NT(STATUS_NULL_LM_PASSWORD)')
    NTSTATUS__STATUS_IMAGE_MACHINE_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x5000000E, 'HRESULT_FROM_NT(STATUS_IMAGE_MACHINE_TYPE_MISMATCH)')
    NTSTATUS__STATUS_RECEIVE_PARTIAL: Final[ThirtyTwoBits] = cached_hresult(0x5000000F, 'HRESULT_FROM_NT(STATUS_RECEIVE_PARTIAL)')
    NTSTATUS__STATUS_RECEIVE_EXPEDITED: Final[ThirtyTwoBits] = cached_hresult(0x50000010, 'HRESULT_FROM_NT(STATUS_RECEIVE_EXPEDITED)')
    NTSTATUS__STATUS_RECEIVE_PARTIAL_EXPEDITED: Final[ThirtyTwoBits] = cached_hresult(0x50000011, 'HRESULT_FROM_NT(STATUS_RECEIVE_PARTIAL_EXPEDITED)')
    NTSTATUS__STATUS_EVENT_DONE: Final[ThirtyTwoBits] = cached_hresult(0x50000012, 'HRESULT_FROM_NT(STATUS_EVENT_DONE)')
    NTSTATUS__STATUS_EVENT_PENDING: Final[ThirtyTwoBits] = cached_hresult(0x50000013, 'HRESULT_FROM_NT(STATUS_EVENT_PENDING)')
    NTSTATUS__STATUS_CHECKING_FILE_SYSTEM: Final[ThirtyTwoBits] = cached_hresult(0x50000014, 'HRESULT_FROM_NT(STATUS_CHECKING_FILE_SYSTEM)')
    NTSTATUS__STATUS_FATAL_APP_EXIT: Final[ThirtyTwoBits] = cached_hresult(0x50000015, 'HRESULT_FROM_NT(STATUS_FATAL_APP_EXIT)')
    NTSTATUS__STATUS_PREDEFINED_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0x50000016, 'HRESULT_FROM_NT(STATUS_PREDEFINED_HANDLE)')
    NTSTATUS__STATUS_WAS_UNLOCKED: Final[ThirtyTwoBits] = cached_hresult(0x50000017, 'HRESULT_FROM_NT(STATUS_WAS_UNLOCKED)')
    NTSTATUS__STATUS_SERVICE_NOTIFICATION: Final[ThirtyTwoBits] = cached_hresult(0x50000018, 'HRESULT_FROM_NT(STATUS_SERVICE_NOTIFICATION)')
    NTSTATUS__STATUS_WAS_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0x50000019, 'HRESULT_FROM_NT(STATUS_WAS_LOCKED)')
    NTSTATUS__STATUS_LOG_HARD_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x5000001A, 'HRESULT_FROM_NT(STATUS_LOG_HARD_ERROR)')
    NTSTATUS__STATUS_ALREADY_WIN32: Final[ThirtyTwoBits] = cached_hresult(0x5000001B, 'HRESULT_FROM_NT(STATUS_ALREADY_WIN32)')
    NTSTATUS__STATUS_WX86_UNSIMULATE: Final[ThirtyTwoBits] = cached_hresult(0x5000001C, 'HRESULT_FROM_NT(STATUS_WX86_UNSIMULATE)')
    NTSTATUS__STATUS_WX86_CONTINUE: Final[ThirtyTwoBits] = cached_hresult(0x5000001D, 'HRESULT_FROM_NT(STATUS_WX86_CONTINUE)')
    NTSTATUS__STATUS_WX86_SINGLE_STEP: Final[ThirtyTwoBits] = cached_hresult(0x5000001E, 'HRESULT_FROM_NT(STATUS_WX86_SINGLE_STEP)')
    NTSTATUS__STATUS_WX86_BREAKPOINT: Final[ThirtyTwoBits] = cached_hresult(0x5000001F, 'HRESULT_FROM_NT(STATUS_WX86_BREAKPOINT)')
    NTSTATUS__STATUS_WX86_EXCEPTION_CONTINUE: Final[ThirtyTwoBits] = cached_hresult(0x50000020, 'HRESULT_FROM_NT(STATUS_WX86_EXCEPTION_CONTINUE)')
    NTSTATUS__STATUS_WX86_EXCEPTION_LASTCHANCE: Final[ThirtyTwoBits] = cached_hresult(0x50000021, 'HRESULT_FROM_NT(STATUS_WX86_EXCEPTION_LASTCHANCE)')
    NTSTATUS__STATUS_WX86_EXCEPTION_CHAIN: Final[ThirtyTwoBits] = cached_hresult(0x50000022, 'HRESULT_FROM_NT(STATUS_WX86_EXCEPTION_CHAIN)')
    NTSTATUS__STATUS_IMAGE_MACHINE_TYPE_MISMATCH_EXE: Final[ThirtyTwoBits] = cached_hresult(0x50000023, 'HRESULT_FROM_NT(STATUS_IMAGE_MACHINE_TYPE_MISMATCH_EXE)')
    NTSTATUS__STATUS_NO_YIELD_PERFORMED: Final[ThirtyTwoBits] = cached_hresult(0x50000024, 'HRESULT_FROM_NT(STATUS_NO_YIELD_PERFORMED)')
    NTSTATUS__STATUS_TIMER_RESUME_IGNORED: Final[ThirtyTwoBits] = cached_hresult(0x50000025, 'HRESULT_FROM_NT(STATUS_TIMER_RESUME_IGNORED)')
    NTSTATUS__STATUS_ARBITRATION_UNHANDLED: Final[ThirtyTwoBits] = cached_hresult(0x50000026, 'HRESULT_FROM_NT(STATUS_ARBITRATION_UNHANDLED)')
    NTSTATUS__STATUS_CARDBUS_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0x50000027, 'HRESULT_FROM_NT(STATUS_CARDBUS_NOT_SUPPORTED)')
    NTSTATUS__STATUS_WX86_CREATEWX86TIB: Final[ThirtyTwoBits] = cached_hresult(0x50000028, 'HRESULT_FROM_NT(STATUS_WX86_CREATEWX86TIB)')
    NTSTATUS__STATUS_MP_PROCESSOR_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x50000029, 'HRESULT_FROM_NT(STATUS_MP_PROCESSOR_MISMATCH)')
    NTSTATUS__STATUS_HIBERNATED: Final[ThirtyTwoBits] = cached_hresult(0x5000002A, 'HRESULT_FROM_NT(STATUS_HIBERNATED)')
    NTSTATUS__STATUS_RESUME_HIBERNATION: Final[ThirtyTwoBits] = cached_hresult(0x5000002B, 'HRESULT_FROM_NT(STATUS_RESUME_HIBERNATION)')
    NTSTATUS__STATUS_FIRMWARE_UPDATED: Final[ThirtyTwoBits] = cached_hresult(0x5000002C, 'HRESULT_FROM_NT(STATUS_FIRMWARE_UPDATED)')
    NTSTATUS__STATUS_DRIVERS_LEAKING_LOCKED_PAGES: Final[ThirtyTwoBits] = cached_hresult(0x5000002D, 'HRESULT_FROM_NT(STATUS_DRIVERS_LEAKING_LOCKED_PAGES)')
    NTSTATUS__STATUS_MESSAGE_RETRIEVED: Final[ThirtyTwoBits] = cached_hresult(0x5000002E, 'HRESULT_FROM_NT(STATUS_MESSAGE_RETRIEVED)')
    NTSTATUS__STATUS_SYSTEM_POWERSTATE_TRANSITION: Final[ThirtyTwoBits] = cached_hresult(0x5000002F, 'HRESULT_FROM_NT(STATUS_SYSTEM_POWERSTATE_TRANSITION)')
    NTSTATUS__STATUS_ALPC_CHECK_COMPLETION_LIST: Final[ThirtyTwoBits] = cached_hresult(0x50000030, 'HRESULT_FROM_NT(STATUS_ALPC_CHECK_COMPLETION_LIST)')
    NTSTATUS__STATUS_SYSTEM_POWERSTATE_COMPLEX_TRANSITION: Final[ThirtyTwoBits] = cached_hresult(0x50000031, 'HRESULT_FROM_NT(STATUS_SYSTEM_POWERSTATE_COMPLEX_TRANSITION)')
    NTSTATUS__STATUS_ACCESS_AUDIT_BY_POLICY: Final[ThirtyTwoBits] = cached_hresult(0x50000032, 'HRESULT_FROM_NT(STATUS_ACCESS_AUDIT_BY_POLICY)')
    NTSTATUS__STATUS_ABANDON_HIBERFILE: Final[ThirtyTwoBits] = cached_hresult(0x50000033, 'HRESULT_FROM_NT(STATUS_ABANDON_HIBERFILE)')
    NTSTATUS__STATUS_BIZRULES_NOT_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0x50000034, 'HRESULT_FROM_NT(STATUS_BIZRULES_NOT_ENABLED)')
    NTSTATUS__STATUS_FT_READ_FROM_COPY: Final[ThirtyTwoBits] = cached_hresult(0x50000035, 'HRESULT_FROM_NT(STATUS_FT_READ_FROM_COPY)')
    NTSTATUS__STATUS_IMAGE_AT_DIFFERENT_BASE: Final[ThirtyTwoBits] = cached_hresult(0x50000036, 'HRESULT_FROM_NT(STATUS_IMAGE_AT_DIFFERENT_BASE)')
    NTSTATUS__STATUS_PATCH_DEFERRED: Final[ThirtyTwoBits] = cached_hresult(0x50000037, 'HRESULT_FROM_NT(STATUS_PATCH_DEFERRED)')
    NTSTATUS__STATUS_EMULATION_BREAKPOINT: Final[ThirtyTwoBits] = cached_hresult(0x50000038, 'HRESULT_FROM_NT(STATUS_EMULATION_BREAKPOINT)')
    NTSTATUS__STATUS_EMULATION_SYSCALL: Final[ThirtyTwoBits] = cached_hresult(0x50000039, 'HRESULT_FROM_NT(STATUS_EMULATION_SYSCALL)')
    NTSTATUS__STATUS_WAKE_SYSTEM: Final[ThirtyTwoBits] = cached_hresult(0x50000294, 'HRESULT_FROM_NT(STATUS_WAKE_SYSTEM)')
    NTSTATUS__STATUS_DS_SHUTTING_DOWN: Final[ThirtyTwoBits] = cached_hresult(0x50000370, 'HRESULT_FROM_NT(STATUS_DS_SHUTTING_DOWN)')
    NTSTATUS__STATUS_DISK_REPAIR_REDIRECTED: Final[ThirtyTwoBits] = cached_hresult(0x50000807, 'HRESULT_FROM_NT(STATUS_DISK_REPAIR_REDIRECTED)')
    NTSTATUS__STATUS_SERVICES_FAILED_AUTOSTART: Final[ThirtyTwoBits] = cached_hresult(0x5000A144, 'HRESULT_FROM_NT(STATUS_SERVICES_FAILED_AUTOSTART)')
    NTSTATUS__STATUS_CTX_CDM_CONNECT: Final[ThirtyTwoBits] = cached_hresult(0x500A0004, 'HRESULT_FROM_NT(STATUS_CTX_CDM_CONNECT)')
    NTSTATUS__STATUS_CTX_CDM_DISCONNECT: Final[ThirtyTwoBits] = cached_hresult(0x500A0005, 'HRESULT_FROM_NT(STATUS_CTX_CDM_DISCONNECT)')
    NTSTATUS__STATUS_SXS_RELEASE_ACTIVATION_CONTEXT: Final[ThirtyTwoBits] = cached_hresult(0x5015000D, 'HRESULT_FROM_NT(STATUS_SXS_RELEASE_ACTIVATION_CONTEXT)')
    NTSTATUS__STATUS_HEURISTIC_DAMAGE_POSSIBLE: Final[ThirtyTwoBits] = cached_hresult(0x50190001, 'HRESULT_FROM_NT(STATUS_HEURISTIC_DAMAGE_POSSIBLE)')
    NTSTATUS__STATUS_RECOVERY_NOT_NEEDED: Final[ThirtyTwoBits] = cached_hresult(0x50190034, 'HRESULT_FROM_NT(STATUS_RECOVERY_NOT_NEEDED)')
    NTSTATUS__STATUS_RM_ALREADY_STARTED: Final[ThirtyTwoBits] = cached_hresult(0x50190035, 'HRESULT_FROM_NT(STATUS_RM_ALREADY_STARTED)')
    NTSTATUS__STATUS_LOG_NO_RESTART: Final[ThirtyTwoBits] = cached_hresult(0x501A000C, 'HRESULT_FROM_NT(STATUS_LOG_NO_RESTART)')
    NTSTATUS__STATUS_VIDEO_DRIVER_DEBUG_REPORT_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0x501B00EC, 'HRESULT_FROM_NT(STATUS_VIDEO_DRIVER_DEBUG_REPORT_REQUEST)')
    NTSTATUS__STATUS_GRAPHICS_PARTIAL_DATA_POPULATED: Final[ThirtyTwoBits] = cached_hresult(0x501E000A, 'HRESULT_FROM_NT(STATUS_GRAPHICS_PARTIAL_DATA_POPULATED)')
    NTSTATUS__STATUS_GRAPHICS_SKIP_ALLOCATION_PREPARATION: Final[ThirtyTwoBits] = cached_hresult(0x501E0201, 'HRESULT_FROM_NT(STATUS_GRAPHICS_SKIP_ALLOCATION_PREPARATION)')
    NTSTATUS__STATUS_GRAPHICS_MODE_NOT_PINNED: Final[ThirtyTwoBits] = cached_hresult(0x501E0307, 'HRESULT_FROM_NT(STATUS_GRAPHICS_MODE_NOT_PINNED)')
    NTSTATUS__STATUS_GRAPHICS_NO_PREFERRED_MODE: Final[ThirtyTwoBits] = cached_hresult(0x501E031E, 'HRESULT_FROM_NT(STATUS_GRAPHICS_NO_PREFERRED_MODE)')
    NTSTATUS__STATUS_GRAPHICS_DATASET_IS_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0x501E034B, 'HRESULT_FROM_NT(STATUS_GRAPHICS_DATASET_IS_EMPTY)')
    NTSTATUS__STATUS_GRAPHICS_NO_MORE_ELEMENTS_IN_DATASET: Final[ThirtyTwoBits] = cached_hresult(0x501E034C, 'HRESULT_FROM_NT(STATUS_GRAPHICS_NO_MORE_ELEMENTS_IN_DATASET)')
    NTSTATUS__STATUS_GRAPHICS_PATH_CONTENT_GEOMETRY_TRANSFORMATION_NOT_PINNED: Final[ThirtyTwoBits] = cached_hresult(0x501E0351, 'HRESULT_FROM_NT(STATUS_GRAPHICS_PATH_CONTENT_GEOMETRY_TRANSFORMATION_NOT_PINNED)')
    NTSTATUS__STATUS_GRAPHICS_UNKNOWN_CHILD_STATUS: Final[ThirtyTwoBits] = cached_hresult(0x501E042F, 'HRESULT_FROM_NT(STATUS_GRAPHICS_UNKNOWN_CHILD_STATUS)')
    NTSTATUS__STATUS_GRAPHICS_LEADLINK_START_DEFERRED: Final[ThirtyTwoBits] = cached_hresult(0x501E0437, 'HRESULT_FROM_NT(STATUS_GRAPHICS_LEADLINK_START_DEFERRED)')
    NTSTATUS__STATUS_GRAPHICS_POLLING_TOO_FREQUENTLY: Final[ThirtyTwoBits] = cached_hresult(0x501E0439, 'HRESULT_FROM_NT(STATUS_GRAPHICS_POLLING_TOO_FREQUENTLY)')
    NTSTATUS__STATUS_GRAPHICS_START_DEFERRED: Final[ThirtyTwoBits] = cached_hresult(0x501E043A, 'HRESULT_FROM_NT(STATUS_GRAPHICS_START_DEFERRED)')
    NTSTATUS__STATUS_GRAPHICS_DEPENDABLE_CHILD_STATUS: Final[ThirtyTwoBits] = cached_hresult(0x501E043C, 'HRESULT_FROM_NT(STATUS_GRAPHICS_DEPENDABLE_CHILD_STATUS)')
    NTSTATUS__STATUS_NDIS_INDICATION_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x50230001, 'HRESULT_FROM_NT(STATUS_NDIS_INDICATION_REQUIRED)')
    NTSTATUS__STATUS_PCP_UNSUPPORTED_PSS_SALT: Final[ThirtyTwoBits] = cached_hresult(0x50292023, 'HRESULT_FROM_NT(STATUS_PCP_UNSUPPORTED_PSS_SALT)')
    NTSTATUS__STATUS_GUARD_PAGE_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0x90000001, 'HRESULT_FROM_NT(STATUS_GUARD_PAGE_VIOLATION)')
    NTSTATUS__STATUS_DATATYPE_MISALIGNMENT: Final[ThirtyTwoBits] = cached_hresult(0x90000002, 'HRESULT_FROM_NT(STATUS_DATATYPE_MISALIGNMENT)')
    NTSTATUS__STATUS_BREAKPOINT: Final[ThirtyTwoBits] = cached_hresult(0x90000003, 'HRESULT_FROM_NT(STATUS_BREAKPOINT)')
    NTSTATUS__STATUS_SINGLE_STEP: Final[ThirtyTwoBits] = cached_hresult(0x90000004, 'HRESULT_FROM_NT(STATUS_SINGLE_STEP)')
    NTSTATUS__STATUS_BUFFER_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0x90000005, 'HRESULT_FROM_NT(STATUS_BUFFER_OVERFLOW)')
    NTSTATUS__STATUS_NO_MORE_FILES: Final[ThirtyTwoBits] = cached_hresult(0x90000006, 'HRESULT_FROM_NT(STATUS_NO_MORE_FILES)')
    NTSTATUS__STATUS_WAKE_SYSTEM_DEBUGGER: Final[ThirtyTwoBits] = cached_hresult(0x90000007, 'HRESULT_FROM_NT(STATUS_WAKE_SYSTEM_DEBUGGER)')
    NTSTATUS__STATUS_HANDLES_CLOSED: Final[ThirtyTwoBits] = cached_hresult(0x9000000A, 'HRESULT_FROM_NT(STATUS_HANDLES_CLOSED)')
    NTSTATUS__STATUS_NO_INHERITANCE: Final[ThirtyTwoBits] = cached_hresult(0x9000000B, 'HRESULT_FROM_NT(STATUS_NO_INHERITANCE)')
    NTSTATUS__STATUS_GUID_SUBSTITUTION_MADE: Final[ThirtyTwoBits] = cached_hresult(0x9000000C, 'HRESULT_FROM_NT(STATUS_GUID_SUBSTITUTION_MADE)')
    NTSTATUS__STATUS_PARTIAL_COPY: Final[ThirtyTwoBits] = cached_hresult(0x9000000D, 'HRESULT_FROM_NT(STATUS_PARTIAL_COPY)')
    NTSTATUS__STATUS_DEVICE_PAPER_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0x9000000E, 'HRESULT_FROM_NT(STATUS_DEVICE_PAPER_EMPTY)')
    NTSTATUS__STATUS_DEVICE_POWERED_OFF: Final[ThirtyTwoBits] = cached_hresult(0x9000000F, 'HRESULT_FROM_NT(STATUS_DEVICE_POWERED_OFF)')
    NTSTATUS__STATUS_DEVICE_OFF_LINE: Final[ThirtyTwoBits] = cached_hresult(0x90000010, 'HRESULT_FROM_NT(STATUS_DEVICE_OFF_LINE)')
    NTSTATUS__STATUS_DEVICE_BUSY: Final[ThirtyTwoBits] = cached_hresult(0x90000011, 'HRESULT_FROM_NT(STATUS_DEVICE_BUSY)')
    NTSTATUS__STATUS_NO_MORE_EAS: Final[ThirtyTwoBits] = cached_hresult(0x90000012, 'HRESULT_FROM_NT(STATUS_NO_MORE_EAS)')
    NTSTATUS__STATUS_INVALID_EA_NAME: Final[ThirtyTwoBits] = cached_hresult(0x90000013, 'HRESULT_FROM_NT(STATUS_INVALID_EA_NAME)')
    NTSTATUS__STATUS_EA_LIST_INCONSISTENT: Final[ThirtyTwoBits] = cached_hresult(0x90000014, 'HRESULT_FROM_NT(STATUS_EA_LIST_INCONSISTENT)')
    NTSTATUS__STATUS_INVALID_EA_FLAG: Final[ThirtyTwoBits] = cached_hresult(0x90000015, 'HRESULT_FROM_NT(STATUS_INVALID_EA_FLAG)')
    NTSTATUS__STATUS_VERIFY_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x90000016, 'HRESULT_FROM_NT(STATUS_VERIFY_REQUIRED)')
    NTSTATUS__STATUS_EXTRANEOUS_INFORMATION: Final[ThirtyTwoBits] = cached_hresult(0x90000017, 'HRESULT_FROM_NT(STATUS_EXTRANEOUS_INFORMATION)')
    NTSTATUS__STATUS_RXACT_COMMIT_NECESSARY: Final[ThirtyTwoBits] = cached_hresult(0x90000018, 'HRESULT_FROM_NT(STATUS_RXACT_COMMIT_NECESSARY)')
    NTSTATUS__STATUS_NO_MORE_ENTRIES: Final[ThirtyTwoBits] = cached_hresult(0x9000001A, 'HRESULT_FROM_NT(STATUS_NO_MORE_ENTRIES)')
    NTSTATUS__STATUS_FILEMARK_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0x9000001B, 'HRESULT_FROM_NT(STATUS_FILEMARK_DETECTED)')
    NTSTATUS__STATUS_MEDIA_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0x9000001C, 'HRESULT_FROM_NT(STATUS_MEDIA_CHANGED)')
    NTSTATUS__STATUS_BUS_RESET: Final[ThirtyTwoBits] = cached_hresult(0x9000001D, 'HRESULT_FROM_NT(STATUS_BUS_RESET)')
    NTSTATUS__STATUS_END_OF_MEDIA: Final[ThirtyTwoBits] = cached_hresult(0x9000001E, 'HRESULT_FROM_NT(STATUS_END_OF_MEDIA)')
    NTSTATUS__STATUS_BEGINNING_OF_MEDIA: Final[ThirtyTwoBits] = cached_hresult(0x9000001F, 'HRESULT_FROM_NT(STATUS_BEGINNING_OF_MEDIA)')
    NTSTATUS__STATUS_MEDIA_CHECK: Final[ThirtyTwoBits] = cached_hresult(0x90000020, 'HRESULT_FROM_NT(STATUS_MEDIA_CHECK)')
    NTSTATUS__STATUS_SETMARK_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0x90000021, 'HRESULT_FROM_NT(STATUS_SETMARK_DETECTED)')
    NTSTATUS__STATUS_NO_DATA_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0x90000022, 'HRESULT_FROM_NT(STATUS_NO_DATA_DETECTED)')
    NTSTATUS__STATUS_REDIRECTOR_HAS_OPEN_HANDLES: Final[ThirtyTwoBits] = cached_hresult(0x90000023, 'HRESULT_FROM_NT(STATUS_REDIRECTOR_HAS_OPEN_HANDLES)')
    NTSTATUS__STATUS_SERVER_HAS_OPEN_HANDLES: Final[ThirtyTwoBits] = cached_hresult(0x90000024, 'HRESULT_FROM_NT(STATUS_SERVER_HAS_OPEN_HANDLES)')
    NTSTATUS__STATUS_ALREADY_DISCONNECTED: Final[ThirtyTwoBits] = cached_hresult(0x90000025, 'HRESULT_FROM_NT(STATUS_ALREADY_DISCONNECTED)')
    NTSTATUS__STATUS_LONGJUMP: Final[ThirtyTwoBits] = cached_hresult(0x90000026, 'HRESULT_FROM_NT(STATUS_LONGJUMP)')
    NTSTATUS__STATUS_CLEANER_CARTRIDGE_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0x90000027, 'HRESULT_FROM_NT(STATUS_CLEANER_CARTRIDGE_INSTALLED)')
    NTSTATUS__STATUS_PLUGPLAY_QUERY_VETOED: Final[ThirtyTwoBits] = cached_hresult(0x90000028, 'HRESULT_FROM_NT(STATUS_PLUGPLAY_QUERY_VETOED)')
    NTSTATUS__STATUS_UNWIND_CONSOLIDATE: Final[ThirtyTwoBits] = cached_hresult(0x90000029, 'HRESULT_FROM_NT(STATUS_UNWIND_CONSOLIDATE)')
    NTSTATUS__STATUS_REGISTRY_HIVE_RECOVERED: Final[ThirtyTwoBits] = cached_hresult(0x9000002A, 'HRESULT_FROM_NT(STATUS_REGISTRY_HIVE_RECOVERED)')
    NTSTATUS__STATUS_DLL_MIGHT_BE_INSECURE: Final[ThirtyTwoBits] = cached_hresult(0x9000002B, 'HRESULT_FROM_NT(STATUS_DLL_MIGHT_BE_INSECURE)')
    NTSTATUS__STATUS_DLL_MIGHT_BE_INCOMPATIBLE: Final[ThirtyTwoBits] = cached_hresult(0x9000002C, 'HRESULT_FROM_NT(STATUS_DLL_MIGHT_BE_INCOMPATIBLE)')
    NTSTATUS__STATUS_STOPPED_ON_SYMLINK: Final[ThirtyTwoBits] = cached_hresult(0x9000002D, 'HRESULT_FROM_NT(STATUS_STOPPED_ON_SYMLINK)')
    NTSTATUS__STATUS_CANNOT_GRANT_REQUESTED_OPLOCK: Final[ThirtyTwoBits] = cached_hresult(0x9000002E, 'HRESULT_FROM_NT(STATUS_CANNOT_GRANT_REQUESTED_OPLOCK)')
    NTSTATUS__STATUS_NO_ACE_CONDITION: Final[ThirtyTwoBits] = cached_hresult(0x9000002F, 'HRESULT_FROM_NT(STATUS_NO_ACE_CONDITION)')
    NTSTATUS__STATUS_DEVICE_SUPPORT_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x90000030, 'HRESULT_FROM_NT(STATUS_DEVICE_SUPPORT_IN_PROGRESS)')
    NTSTATUS__STATUS_DEVICE_POWER_CYCLE_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x90000031, 'HRESULT_FROM_NT(STATUS_DEVICE_POWER_CYCLE_REQUIRED)')
    NTSTATUS__STATUS_NO_WORK_DONE: Final[ThirtyTwoBits] = cached_hresult(0x90000032, 'HRESULT_FROM_NT(STATUS_NO_WORK_DONE)')
    NTSTATUS__STATUS_RETURN_ADDRESS_HIJACK_ATTEMPT: Final[ThirtyTwoBits] = cached_hresult(0x90000033, 'HRESULT_FROM_NT(STATUS_RETURN_ADDRESS_HIJACK_ATTEMPT)')
    NTSTATUS__STATUS_RECOVERABLE_BUGCHECK: Final[ThirtyTwoBits] = cached_hresult(0x90000034, 'HRESULT_FROM_NT(STATUS_RECOVERABLE_BUGCHECK)')
    NTSTATUS__STATUS_DEVICE_RESET_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0x900001B6, 'HRESULT_FROM_NT(STATUS_DEVICE_RESET_REQUIRED)')
    NTSTATUS__STATUS_DEVICE_REQUIRES_CLEANING: Final[ThirtyTwoBits] = cached_hresult(0x90000288, 'HRESULT_FROM_NT(STATUS_DEVICE_REQUIRES_CLEANING)')
    NTSTATUS__STATUS_DEVICE_DOOR_OPEN: Final[ThirtyTwoBits] = cached_hresult(0x90000289, 'HRESULT_FROM_NT(STATUS_DEVICE_DOOR_OPEN)')
    NTSTATUS__STATUS_DATA_LOST_REPAIR: Final[ThirtyTwoBits] = cached_hresult(0x90000803, 'HRESULT_FROM_NT(STATUS_DATA_LOST_REPAIR)')
    NTSTATUS__STATUS_GPIO_INTERRUPT_ALREADY_UNMASKED: Final[ThirtyTwoBits] = cached_hresult(0x9000A127, 'HRESULT_FROM_NT(STATUS_GPIO_INTERRUPT_ALREADY_UNMASKED)')
    NTSTATUS__STATUS_CLOUD_FILE_PROPERTY_BLOB_CHECKSUM_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0x9000CF00, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_PROPERTY_BLOB_CHECKSUM_MISMATCH)')
    NTSTATUS__STATUS_CLOUD_FILE_PROPERTY_BLOB_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0x9000CF04, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_PROPERTY_BLOB_TOO_LARGE)')
    NTSTATUS__STATUS_CLOUD_FILE_TOO_MANY_PROPERTY_BLOBS: Final[ThirtyTwoBits] = cached_hresult(0x9000CF05, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_TOO_MANY_PROPERTY_BLOBS)')
    NTSTATUS__STATUS_CLUSTER_NODE_ALREADY_UP: Final[ThirtyTwoBits] = cached_hresult(0x90130001, 'HRESULT_FROM_NT(STATUS_CLUSTER_NODE_ALREADY_UP)')
    NTSTATUS__STATUS_CLUSTER_NODE_ALREADY_DOWN: Final[ThirtyTwoBits] = cached_hresult(0x90130002, 'HRESULT_FROM_NT(STATUS_CLUSTER_NODE_ALREADY_DOWN)')
    NTSTATUS__STATUS_CLUSTER_NETWORK_ALREADY_ONLINE: Final[ThirtyTwoBits] = cached_hresult(0x90130003, 'HRESULT_FROM_NT(STATUS_CLUSTER_NETWORK_ALREADY_ONLINE)')
    NTSTATUS__STATUS_CLUSTER_NETWORK_ALREADY_OFFLINE: Final[ThirtyTwoBits] = cached_hresult(0x90130004, 'HRESULT_FROM_NT(STATUS_CLUSTER_NETWORK_ALREADY_OFFLINE)')
    NTSTATUS__STATUS_CLUSTER_NODE_ALREADY_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0x90130005, 'HRESULT_FROM_NT(STATUS_CLUSTER_NODE_ALREADY_MEMBER)')
    NTSTATUS__STATUS_COULD_NOT_RESIZE_LOG: Final[ThirtyTwoBits] = cached_hresult(0x90190009, 'HRESULT_FROM_NT(STATUS_COULD_NOT_RESIZE_LOG)')
    NTSTATUS__STATUS_NO_TXF_METADATA: Final[ThirtyTwoBits] = cached_hresult(0x90190029, 'HRESULT_FROM_NT(STATUS_NO_TXF_METADATA)')
    NTSTATUS__STATUS_CANT_RECOVER_WITH_HANDLE_OPEN: Final[ThirtyTwoBits] = cached_hresult(0x90190031, 'HRESULT_FROM_NT(STATUS_CANT_RECOVER_WITH_HANDLE_OPEN)')
    NTSTATUS__STATUS_TXF_METADATA_ALREADY_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0x90190041, 'HRESULT_FROM_NT(STATUS_TXF_METADATA_ALREADY_PRESENT)')
    NTSTATUS__STATUS_TRANSACTION_SCOPE_CALLBACKS_NOT_SET: Final[ThirtyTwoBits] = cached_hresult(0x90190042, 'HRESULT_FROM_NT(STATUS_TRANSACTION_SCOPE_CALLBACKS_NOT_SET)')
    NTSTATUS__STATUS_VIDEO_HUNG_DISPLAY_DRIVER_THREAD_RECOVERED: Final[ThirtyTwoBits] = cached_hresult(0x901B00EB, 'HRESULT_FROM_NT(STATUS_VIDEO_HUNG_DISPLAY_DRIVER_THREAD_RECOVERED)')
    NTSTATUS__STATUS_FLT_BUFFER_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0x901C0001, 'HRESULT_FROM_NT(STATUS_FLT_BUFFER_TOO_SMALL)')
    NTSTATUS__STATUS_GRAPHICS_LINK_CONFIGURATION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0x901E0000, 'HRESULT_FROM_NT(STATUS_GRAPHICS_LINK_CONFIGURATION_IN_PROGRESS)')
    NTSTATUS__STATUS_FVE_PARTIAL_METADATA: Final[ThirtyTwoBits] = cached_hresult(0x90210001, 'HRESULT_FROM_NT(STATUS_FVE_PARTIAL_METADATA)')
    NTSTATUS__STATUS_FVE_TRANSIENT_STATE: Final[ThirtyTwoBits] = cached_hresult(0x90210002, 'HRESULT_FROM_NT(STATUS_FVE_TRANSIENT_STATE)')
    NTSTATUS__STATUS_VID_REMOTE_NODE_PARENT_GPA_PAGES_USED: Final[ThirtyTwoBits] = cached_hresult(0x90370001, 'HRESULT_FROM_NT(STATUS_VID_REMOTE_NODE_PARENT_GPA_PAGES_USED)')
    NTSTATUS__STATUS_VOLMGR_INCOMPLETE_REGENERATION: Final[ThirtyTwoBits] = cached_hresult(0x90380001, 'HRESULT_FROM_NT(STATUS_VOLMGR_INCOMPLETE_REGENERATION)')
    NTSTATUS__STATUS_VOLMGR_INCOMPLETE_DISK_MIGRATION: Final[ThirtyTwoBits] = cached_hresult(0x90380002, 'HRESULT_FROM_NT(STATUS_VOLMGR_INCOMPLETE_DISK_MIGRATION)')
    NTSTATUS__STATUS_BCD_NOT_ALL_ENTRIES_IMPORTED: Final[ThirtyTwoBits] = cached_hresult(0x90390001, 'HRESULT_FROM_NT(STATUS_BCD_NOT_ALL_ENTRIES_IMPORTED)')
    NTSTATUS__STATUS_BCD_NOT_ALL_ENTRIES_SYNCHRONIZED: Final[ThirtyTwoBits] = cached_hresult(0x90390003, 'HRESULT_FROM_NT(STATUS_BCD_NOT_ALL_ENTRIES_SYNCHRONIZED)')
    NTSTATUS__STATUS_QUERY_STORAGE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0x903A0001, 'HRESULT_FROM_NT(STATUS_QUERY_STORAGE_ERROR)')
    NTSTATUS__STATUS_GDI_HANDLE_LEAK: Final[ThirtyTwoBits] = cached_hresult(0x903F0001, 'HRESULT_FROM_NT(STATUS_GDI_HANDLE_LEAK)')
    NTSTATUS__STATUS_SECUREBOOT_NOT_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0x90430006, 'HRESULT_FROM_NT(STATUS_SECUREBOOT_NOT_ENABLED)')
    NTSTATUS__STATUS_UNSUCCESSFUL: Final[ThirtyTwoBits] = cached_hresult(0xD0000001, 'HRESULT_FROM_NT(STATUS_UNSUCCESSFUL)')
    NTSTATUS__STATUS_NOT_IMPLEMENTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000002, 'HRESULT_FROM_NT(STATUS_NOT_IMPLEMENTED)')
    NTSTATUS__STATUS_INVALID_INFO_CLASS: Final[ThirtyTwoBits] = cached_hresult(0xD0000003, 'HRESULT_FROM_NT(STATUS_INVALID_INFO_CLASS)')
    NTSTATUS__STATUS_INFO_LENGTH_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD0000004, 'HRESULT_FROM_NT(STATUS_INFO_LENGTH_MISMATCH)')
    NTSTATUS__STATUS_ACCESS_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0xD0000005, 'HRESULT_FROM_NT(STATUS_ACCESS_VIOLATION)')
    NTSTATUS__STATUS_IN_PAGE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0000006, 'HRESULT_FROM_NT(STATUS_IN_PAGE_ERROR)')
    NTSTATUS__STATUS_PAGEFILE_QUOTA: Final[ThirtyTwoBits] = cached_hresult(0xD0000007, 'HRESULT_FROM_NT(STATUS_PAGEFILE_QUOTA)')
    NTSTATUS__STATUS_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0xD0000008, 'HRESULT_FROM_NT(STATUS_INVALID_HANDLE)')
    NTSTATUS__STATUS_BAD_INITIAL_STACK: Final[ThirtyTwoBits] = cached_hresult(0xD0000009, 'HRESULT_FROM_NT(STATUS_BAD_INITIAL_STACK)')
    NTSTATUS__STATUS_BAD_INITIAL_PC: Final[ThirtyTwoBits] = cached_hresult(0xD000000A, 'HRESULT_FROM_NT(STATUS_BAD_INITIAL_PC)')
    NTSTATUS__STATUS_INVALID_CID: Final[ThirtyTwoBits] = cached_hresult(0xD000000B, 'HRESULT_FROM_NT(STATUS_INVALID_CID)')
    NTSTATUS__STATUS_TIMER_NOT_CANCELED: Final[ThirtyTwoBits] = cached_hresult(0xD000000C, 'HRESULT_FROM_NT(STATUS_TIMER_NOT_CANCELED)')
    NTSTATUS__STATUS_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0xD000000D, 'HRESULT_FROM_NT(STATUS_INVALID_PARAMETER)')
    NTSTATUS__STATUS_NO_SUCH_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0xD000000E, 'HRESULT_FROM_NT(STATUS_NO_SUCH_DEVICE)')
    NTSTATUS__STATUS_NO_SUCH_FILE: Final[ThirtyTwoBits] = cached_hresult(0xD000000F, 'HRESULT_FROM_NT(STATUS_NO_SUCH_FILE)')
    NTSTATUS__STATUS_INVALID_DEVICE_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0xD0000010, 'HRESULT_FROM_NT(STATUS_INVALID_DEVICE_REQUEST)')
    NTSTATUS__STATUS_END_OF_FILE: Final[ThirtyTwoBits] = cached_hresult(0xD0000011, 'HRESULT_FROM_NT(STATUS_END_OF_FILE)')
    NTSTATUS__STATUS_WRONG_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0xD0000012, 'HRESULT_FROM_NT(STATUS_WRONG_VOLUME)')
    NTSTATUS__STATUS_NO_MEDIA_IN_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0xD0000013, 'HRESULT_FROM_NT(STATUS_NO_MEDIA_IN_DEVICE)')
    NTSTATUS__STATUS_UNRECOGNIZED_MEDIA: Final[ThirtyTwoBits] = cached_hresult(0xD0000014, 'HRESULT_FROM_NT(STATUS_UNRECOGNIZED_MEDIA)')
    NTSTATUS__STATUS_NONEXISTENT_SECTOR: Final[ThirtyTwoBits] = cached_hresult(0xD0000015, 'HRESULT_FROM_NT(STATUS_NONEXISTENT_SECTOR)')
    NTSTATUS__STATUS_MORE_PROCESSING_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0xD0000016, 'HRESULT_FROM_NT(STATUS_MORE_PROCESSING_REQUIRED)')
    NTSTATUS__STATUS_NO_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0xD0000017, 'HRESULT_FROM_NT(STATUS_NO_MEMORY)')
    NTSTATUS__STATUS_CONFLICTING_ADDRESSES: Final[ThirtyTwoBits] = cached_hresult(0xD0000018, 'HRESULT_FROM_NT(STATUS_CONFLICTING_ADDRESSES)')
    NTSTATUS__STATUS_NOT_MAPPED_VIEW: Final[ThirtyTwoBits] = cached_hresult(0xD0000019, 'HRESULT_FROM_NT(STATUS_NOT_MAPPED_VIEW)')
    NTSTATUS__STATUS_UNABLE_TO_FREE_VM: Final[ThirtyTwoBits] = cached_hresult(0xD000001A, 'HRESULT_FROM_NT(STATUS_UNABLE_TO_FREE_VM)')
    NTSTATUS__STATUS_UNABLE_TO_DELETE_SECTION: Final[ThirtyTwoBits] = cached_hresult(0xD000001B, 'HRESULT_FROM_NT(STATUS_UNABLE_TO_DELETE_SECTION)')
    NTSTATUS__STATUS_INVALID_SYSTEM_SERVICE: Final[ThirtyTwoBits] = cached_hresult(0xD000001C, 'HRESULT_FROM_NT(STATUS_INVALID_SYSTEM_SERVICE)')
    NTSTATUS__STATUS_ILLEGAL_INSTRUCTION: Final[ThirtyTwoBits] = cached_hresult(0xD000001D, 'HRESULT_FROM_NT(STATUS_ILLEGAL_INSTRUCTION)')
    NTSTATUS__STATUS_INVALID_LOCK_SEQUENCE: Final[ThirtyTwoBits] = cached_hresult(0xD000001E, 'HRESULT_FROM_NT(STATUS_INVALID_LOCK_SEQUENCE)')
    NTSTATUS__STATUS_INVALID_VIEW_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD000001F, 'HRESULT_FROM_NT(STATUS_INVALID_VIEW_SIZE)')
    NTSTATUS__STATUS_INVALID_FILE_FOR_SECTION: Final[ThirtyTwoBits] = cached_hresult(0xD0000020, 'HRESULT_FROM_NT(STATUS_INVALID_FILE_FOR_SECTION)')
    NTSTATUS__STATUS_ALREADY_COMMITTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000021, 'HRESULT_FROM_NT(STATUS_ALREADY_COMMITTED)')
    NTSTATUS__STATUS_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0xD0000022, 'HRESULT_FROM_NT(STATUS_ACCESS_DENIED)')
    NTSTATUS__STATUS_BUFFER_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0xD0000023, 'HRESULT_FROM_NT(STATUS_BUFFER_TOO_SMALL)')
    NTSTATUS__STATUS_OBJECT_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD0000024, 'HRESULT_FROM_NT(STATUS_OBJECT_TYPE_MISMATCH)')
    NTSTATUS__STATUS_NONCONTINUABLE_EXCEPTION: Final[ThirtyTwoBits] = cached_hresult(0xD0000025, 'HRESULT_FROM_NT(STATUS_NONCONTINUABLE_EXCEPTION)')
    NTSTATUS__STATUS_INVALID_DISPOSITION: Final[ThirtyTwoBits] = cached_hresult(0xD0000026, 'HRESULT_FROM_NT(STATUS_INVALID_DISPOSITION)')
    NTSTATUS__STATUS_UNWIND: Final[ThirtyTwoBits] = cached_hresult(0xD0000027, 'HRESULT_FROM_NT(STATUS_UNWIND)')
    NTSTATUS__STATUS_BAD_STACK: Final[ThirtyTwoBits] = cached_hresult(0xD0000028, 'HRESULT_FROM_NT(STATUS_BAD_STACK)')
    NTSTATUS__STATUS_INVALID_UNWIND_TARGET: Final[ThirtyTwoBits] = cached_hresult(0xD0000029, 'HRESULT_FROM_NT(STATUS_INVALID_UNWIND_TARGET)')
    NTSTATUS__STATUS_NOT_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0xD000002A, 'HRESULT_FROM_NT(STATUS_NOT_LOCKED)')
    NTSTATUS__STATUS_PARITY_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD000002B, 'HRESULT_FROM_NT(STATUS_PARITY_ERROR)')
    NTSTATUS__STATUS_UNABLE_TO_DECOMMIT_VM: Final[ThirtyTwoBits] = cached_hresult(0xD000002C, 'HRESULT_FROM_NT(STATUS_UNABLE_TO_DECOMMIT_VM)')
    NTSTATUS__STATUS_NOT_COMMITTED: Final[ThirtyTwoBits] = cached_hresult(0xD000002D, 'HRESULT_FROM_NT(STATUS_NOT_COMMITTED)')
    NTSTATUS__STATUS_INVALID_PORT_ATTRIBUTES: Final[ThirtyTwoBits] = cached_hresult(0xD000002E, 'HRESULT_FROM_NT(STATUS_INVALID_PORT_ATTRIBUTES)')
    NTSTATUS__STATUS_PORT_MESSAGE_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0xD000002F, 'HRESULT_FROM_NT(STATUS_PORT_MESSAGE_TOO_LONG)')
    NTSTATUS__STATUS_INVALID_PARAMETER_MIX: Final[ThirtyTwoBits] = cached_hresult(0xD0000030, 'HRESULT_FROM_NT(STATUS_INVALID_PARAMETER_MIX)')
    NTSTATUS__STATUS_INVALID_QUOTA_LOWER: Final[ThirtyTwoBits] = cached_hresult(0xD0000031, 'HRESULT_FROM_NT(STATUS_INVALID_QUOTA_LOWER)')
    NTSTATUS__STATUS_DISK_CORRUPT_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0000032, 'HRESULT_FROM_NT(STATUS_DISK_CORRUPT_ERROR)')
    NTSTATUS__STATUS_OBJECT_NAME_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0000033, 'HRESULT_FROM_NT(STATUS_OBJECT_NAME_INVALID)')
    NTSTATUS__STATUS_OBJECT_NAME_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000034, 'HRESULT_FROM_NT(STATUS_OBJECT_NAME_NOT_FOUND)')
    NTSTATUS__STATUS_OBJECT_NAME_COLLISION: Final[ThirtyTwoBits] = cached_hresult(0xD0000035, 'HRESULT_FROM_NT(STATUS_OBJECT_NAME_COLLISION)')
    NTSTATUS__STATUS_PORT_DO_NOT_DISTURB: Final[ThirtyTwoBits] = cached_hresult(0xD0000036, 'HRESULT_FROM_NT(STATUS_PORT_DO_NOT_DISTURB)')
    NTSTATUS__STATUS_PORT_DISCONNECTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000037, 'HRESULT_FROM_NT(STATUS_PORT_DISCONNECTED)')
    NTSTATUS__STATUS_DEVICE_ALREADY_ATTACHED: Final[ThirtyTwoBits] = cached_hresult(0xD0000038, 'HRESULT_FROM_NT(STATUS_DEVICE_ALREADY_ATTACHED)')
    NTSTATUS__STATUS_OBJECT_PATH_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0000039, 'HRESULT_FROM_NT(STATUS_OBJECT_PATH_INVALID)')
    NTSTATUS__STATUS_OBJECT_PATH_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD000003A, 'HRESULT_FROM_NT(STATUS_OBJECT_PATH_NOT_FOUND)')
    NTSTATUS__STATUS_OBJECT_PATH_SYNTAX_BAD: Final[ThirtyTwoBits] = cached_hresult(0xD000003B, 'HRESULT_FROM_NT(STATUS_OBJECT_PATH_SYNTAX_BAD)')
    NTSTATUS__STATUS_DATA_OVERRUN: Final[ThirtyTwoBits] = cached_hresult(0xD000003C, 'HRESULT_FROM_NT(STATUS_DATA_OVERRUN)')
    NTSTATUS__STATUS_DATA_LATE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD000003D, 'HRESULT_FROM_NT(STATUS_DATA_LATE_ERROR)')
    NTSTATUS__STATUS_DATA_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD000003E, 'HRESULT_FROM_NT(STATUS_DATA_ERROR)')
    NTSTATUS__STATUS_CRC_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD000003F, 'HRESULT_FROM_NT(STATUS_CRC_ERROR)')
    NTSTATUS__STATUS_SECTION_TOO_BIG: Final[ThirtyTwoBits] = cached_hresult(0xD0000040, 'HRESULT_FROM_NT(STATUS_SECTION_TOO_BIG)')
    NTSTATUS__STATUS_PORT_CONNECTION_REFUSED: Final[ThirtyTwoBits] = cached_hresult(0xD0000041, 'HRESULT_FROM_NT(STATUS_PORT_CONNECTION_REFUSED)')
    NTSTATUS__STATUS_INVALID_PORT_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0xD0000042, 'HRESULT_FROM_NT(STATUS_INVALID_PORT_HANDLE)')
    NTSTATUS__STATUS_SHARING_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0xD0000043, 'HRESULT_FROM_NT(STATUS_SHARING_VIOLATION)')
    NTSTATUS__STATUS_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD0000044, 'HRESULT_FROM_NT(STATUS_QUOTA_EXCEEDED)')
    NTSTATUS__STATUS_INVALID_PAGE_PROTECTION: Final[ThirtyTwoBits] = cached_hresult(0xD0000045, 'HRESULT_FROM_NT(STATUS_INVALID_PAGE_PROTECTION)')
    NTSTATUS__STATUS_MUTANT_NOT_OWNED: Final[ThirtyTwoBits] = cached_hresult(0xD0000046, 'HRESULT_FROM_NT(STATUS_MUTANT_NOT_OWNED)')
    NTSTATUS__STATUS_SEMAPHORE_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD0000047, 'HRESULT_FROM_NT(STATUS_SEMAPHORE_LIMIT_EXCEEDED)')
    NTSTATUS__STATUS_PORT_ALREADY_SET: Final[ThirtyTwoBits] = cached_hresult(0xD0000048, 'HRESULT_FROM_NT(STATUS_PORT_ALREADY_SET)')
    NTSTATUS__STATUS_SECTION_NOT_IMAGE: Final[ThirtyTwoBits] = cached_hresult(0xD0000049, 'HRESULT_FROM_NT(STATUS_SECTION_NOT_IMAGE)')
    NTSTATUS__STATUS_SUSPEND_COUNT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD000004A, 'HRESULT_FROM_NT(STATUS_SUSPEND_COUNT_EXCEEDED)')
    NTSTATUS__STATUS_THREAD_IS_TERMINATING: Final[ThirtyTwoBits] = cached_hresult(0xD000004B, 'HRESULT_FROM_NT(STATUS_THREAD_IS_TERMINATING)')
    NTSTATUS__STATUS_BAD_WORKING_SET_LIMIT: Final[ThirtyTwoBits] = cached_hresult(0xD000004C, 'HRESULT_FROM_NT(STATUS_BAD_WORKING_SET_LIMIT)')
    NTSTATUS__STATUS_INCOMPATIBLE_FILE_MAP: Final[ThirtyTwoBits] = cached_hresult(0xD000004D, 'HRESULT_FROM_NT(STATUS_INCOMPATIBLE_FILE_MAP)')
    NTSTATUS__STATUS_SECTION_PROTECTION: Final[ThirtyTwoBits] = cached_hresult(0xD000004E, 'HRESULT_FROM_NT(STATUS_SECTION_PROTECTION)')
    NTSTATUS__STATUS_EAS_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD000004F, 'HRESULT_FROM_NT(STATUS_EAS_NOT_SUPPORTED)')
    NTSTATUS__STATUS_EA_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0xD0000050, 'HRESULT_FROM_NT(STATUS_EA_TOO_LARGE)')
    NTSTATUS__STATUS_NONEXISTENT_EA_ENTRY: Final[ThirtyTwoBits] = cached_hresult(0xD0000051, 'HRESULT_FROM_NT(STATUS_NONEXISTENT_EA_ENTRY)')
    NTSTATUS__STATUS_NO_EAS_ON_FILE: Final[ThirtyTwoBits] = cached_hresult(0xD0000052, 'HRESULT_FROM_NT(STATUS_NO_EAS_ON_FILE)')
    NTSTATUS__STATUS_EA_CORRUPT_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0000053, 'HRESULT_FROM_NT(STATUS_EA_CORRUPT_ERROR)')
    NTSTATUS__STATUS_FILE_LOCK_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0xD0000054, 'HRESULT_FROM_NT(STATUS_FILE_LOCK_CONFLICT)')
    NTSTATUS__STATUS_LOCK_NOT_GRANTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000055, 'HRESULT_FROM_NT(STATUS_LOCK_NOT_GRANTED)')
    NTSTATUS__STATUS_DELETE_PENDING: Final[ThirtyTwoBits] = cached_hresult(0xD0000056, 'HRESULT_FROM_NT(STATUS_DELETE_PENDING)')
    NTSTATUS__STATUS_CTL_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000057, 'HRESULT_FROM_NT(STATUS_CTL_FILE_NOT_SUPPORTED)')
    NTSTATUS__STATUS_UNKNOWN_REVISION: Final[ThirtyTwoBits] = cached_hresult(0xD0000058, 'HRESULT_FROM_NT(STATUS_UNKNOWN_REVISION)')
    NTSTATUS__STATUS_REVISION_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD0000059, 'HRESULT_FROM_NT(STATUS_REVISION_MISMATCH)')
    NTSTATUS__STATUS_INVALID_OWNER: Final[ThirtyTwoBits] = cached_hresult(0xD000005A, 'HRESULT_FROM_NT(STATUS_INVALID_OWNER)')
    NTSTATUS__STATUS_INVALID_PRIMARY_GROUP: Final[ThirtyTwoBits] = cached_hresult(0xD000005B, 'HRESULT_FROM_NT(STATUS_INVALID_PRIMARY_GROUP)')
    NTSTATUS__STATUS_NO_IMPERSONATION_TOKEN: Final[ThirtyTwoBits] = cached_hresult(0xD000005C, 'HRESULT_FROM_NT(STATUS_NO_IMPERSONATION_TOKEN)')
    NTSTATUS__STATUS_CANT_DISABLE_MANDATORY: Final[ThirtyTwoBits] = cached_hresult(0xD000005D, 'HRESULT_FROM_NT(STATUS_CANT_DISABLE_MANDATORY)')
    NTSTATUS__STATUS_NO_LOGON_SERVERS: Final[ThirtyTwoBits] = cached_hresult(0xD000005E, 'HRESULT_FROM_NT(STATUS_NO_LOGON_SERVERS)')
    NTSTATUS__STATUS_NO_SUCH_LOGON_SESSION: Final[ThirtyTwoBits] = cached_hresult(0xD000005F, 'HRESULT_FROM_NT(STATUS_NO_SUCH_LOGON_SESSION)')
    NTSTATUS__STATUS_NO_SUCH_PRIVILEGE: Final[ThirtyTwoBits] = cached_hresult(0xD0000060, 'HRESULT_FROM_NT(STATUS_NO_SUCH_PRIVILEGE)')
    NTSTATUS__STATUS_PRIVILEGE_NOT_HELD: Final[ThirtyTwoBits] = cached_hresult(0xD0000061, 'HRESULT_FROM_NT(STATUS_PRIVILEGE_NOT_HELD)')
    NTSTATUS__STATUS_INVALID_ACCOUNT_NAME: Final[ThirtyTwoBits] = cached_hresult(0xD0000062, 'HRESULT_FROM_NT(STATUS_INVALID_ACCOUNT_NAME)')
    NTSTATUS__STATUS_USER_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD0000063, 'HRESULT_FROM_NT(STATUS_USER_EXISTS)')
    NTSTATUS__STATUS_NO_SUCH_USER: Final[ThirtyTwoBits] = cached_hresult(0xD0000064, 'HRESULT_FROM_NT(STATUS_NO_SUCH_USER)')
    NTSTATUS__STATUS_GROUP_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD0000065, 'HRESULT_FROM_NT(STATUS_GROUP_EXISTS)')
    NTSTATUS__STATUS_NO_SUCH_GROUP: Final[ThirtyTwoBits] = cached_hresult(0xD0000066, 'HRESULT_FROM_NT(STATUS_NO_SUCH_GROUP)')
    NTSTATUS__STATUS_MEMBER_IN_GROUP: Final[ThirtyTwoBits] = cached_hresult(0xD0000067, 'HRESULT_FROM_NT(STATUS_MEMBER_IN_GROUP)')
    NTSTATUS__STATUS_MEMBER_NOT_IN_GROUP: Final[ThirtyTwoBits] = cached_hresult(0xD0000068, 'HRESULT_FROM_NT(STATUS_MEMBER_NOT_IN_GROUP)')
    NTSTATUS__STATUS_LAST_ADMIN: Final[ThirtyTwoBits] = cached_hresult(0xD0000069, 'HRESULT_FROM_NT(STATUS_LAST_ADMIN)')
    NTSTATUS__STATUS_WRONG_PASSWORD: Final[ThirtyTwoBits] = cached_hresult(0xD000006A, 'HRESULT_FROM_NT(STATUS_WRONG_PASSWORD)')
    NTSTATUS__STATUS_ILL_FORMED_PASSWORD: Final[ThirtyTwoBits] = cached_hresult(0xD000006B, 'HRESULT_FROM_NT(STATUS_ILL_FORMED_PASSWORD)')
    NTSTATUS__STATUS_PASSWORD_RESTRICTION: Final[ThirtyTwoBits] = cached_hresult(0xD000006C, 'HRESULT_FROM_NT(STATUS_PASSWORD_RESTRICTION)')
    NTSTATUS__STATUS_LOGON_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD000006D, 'HRESULT_FROM_NT(STATUS_LOGON_FAILURE)')
    NTSTATUS__STATUS_ACCOUNT_RESTRICTION: Final[ThirtyTwoBits] = cached_hresult(0xD000006E, 'HRESULT_FROM_NT(STATUS_ACCOUNT_RESTRICTION)')
    NTSTATUS__STATUS_INVALID_LOGON_HOURS: Final[ThirtyTwoBits] = cached_hresult(0xD000006F, 'HRESULT_FROM_NT(STATUS_INVALID_LOGON_HOURS)')
    NTSTATUS__STATUS_INVALID_WORKSTATION: Final[ThirtyTwoBits] = cached_hresult(0xD0000070, 'HRESULT_FROM_NT(STATUS_INVALID_WORKSTATION)')
    NTSTATUS__STATUS_PASSWORD_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0xD0000071, 'HRESULT_FROM_NT(STATUS_PASSWORD_EXPIRED)')
    NTSTATUS__STATUS_ACCOUNT_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD0000072, 'HRESULT_FROM_NT(STATUS_ACCOUNT_DISABLED)')
    NTSTATUS__STATUS_NONE_MAPPED: Final[ThirtyTwoBits] = cached_hresult(0xD0000073, 'HRESULT_FROM_NT(STATUS_NONE_MAPPED)')
    NTSTATUS__STATUS_TOO_MANY_LUIDS_REQUESTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000074, 'HRESULT_FROM_NT(STATUS_TOO_MANY_LUIDS_REQUESTED)')
    NTSTATUS__STATUS_LUIDS_EXHAUSTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000075, 'HRESULT_FROM_NT(STATUS_LUIDS_EXHAUSTED)')
    NTSTATUS__STATUS_INVALID_SUB_AUTHORITY: Final[ThirtyTwoBits] = cached_hresult(0xD0000076, 'HRESULT_FROM_NT(STATUS_INVALID_SUB_AUTHORITY)')
    NTSTATUS__STATUS_INVALID_ACL: Final[ThirtyTwoBits] = cached_hresult(0xD0000077, 'HRESULT_FROM_NT(STATUS_INVALID_ACL)')
    NTSTATUS__STATUS_INVALID_SID: Final[ThirtyTwoBits] = cached_hresult(0xD0000078, 'HRESULT_FROM_NT(STATUS_INVALID_SID)')
    NTSTATUS__STATUS_INVALID_SECURITY_DESCR: Final[ThirtyTwoBits] = cached_hresult(0xD0000079, 'HRESULT_FROM_NT(STATUS_INVALID_SECURITY_DESCR)')
    NTSTATUS__STATUS_PROCEDURE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD000007A, 'HRESULT_FROM_NT(STATUS_PROCEDURE_NOT_FOUND)')
    NTSTATUS__STATUS_INVALID_IMAGE_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0xD000007B, 'HRESULT_FROM_NT(STATUS_INVALID_IMAGE_FORMAT)')
    NTSTATUS__STATUS_NO_TOKEN: Final[ThirtyTwoBits] = cached_hresult(0xD000007C, 'HRESULT_FROM_NT(STATUS_NO_TOKEN)')
    NTSTATUS__STATUS_BAD_INHERITANCE_ACL: Final[ThirtyTwoBits] = cached_hresult(0xD000007D, 'HRESULT_FROM_NT(STATUS_BAD_INHERITANCE_ACL)')
    NTSTATUS__STATUS_RANGE_NOT_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0xD000007E, 'HRESULT_FROM_NT(STATUS_RANGE_NOT_LOCKED)')
    NTSTATUS__STATUS_DISK_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD000007F, 'HRESULT_FROM_NT(STATUS_DISK_FULL)')
    NTSTATUS__STATUS_SERVER_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD0000080, 'HRESULT_FROM_NT(STATUS_SERVER_DISABLED)')
    NTSTATUS__STATUS_SERVER_NOT_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD0000081, 'HRESULT_FROM_NT(STATUS_SERVER_NOT_DISABLED)')
    NTSTATUS__STATUS_TOO_MANY_GUIDS_REQUESTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000082, 'HRESULT_FROM_NT(STATUS_TOO_MANY_GUIDS_REQUESTED)')
    NTSTATUS__STATUS_GUIDS_EXHAUSTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000083, 'HRESULT_FROM_NT(STATUS_GUIDS_EXHAUSTED)')
    NTSTATUS__STATUS_INVALID_ID_AUTHORITY: Final[ThirtyTwoBits] = cached_hresult(0xD0000084, 'HRESULT_FROM_NT(STATUS_INVALID_ID_AUTHORITY)')
    NTSTATUS__STATUS_AGENTS_EXHAUSTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000085, 'HRESULT_FROM_NT(STATUS_AGENTS_EXHAUSTED)')
    NTSTATUS__STATUS_INVALID_VOLUME_LABEL: Final[ThirtyTwoBits] = cached_hresult(0xD0000086, 'HRESULT_FROM_NT(STATUS_INVALID_VOLUME_LABEL)')
    NTSTATUS__STATUS_SECTION_NOT_EXTENDED: Final[ThirtyTwoBits] = cached_hresult(0xD0000087, 'HRESULT_FROM_NT(STATUS_SECTION_NOT_EXTENDED)')
    NTSTATUS__STATUS_NOT_MAPPED_DATA: Final[ThirtyTwoBits] = cached_hresult(0xD0000088, 'HRESULT_FROM_NT(STATUS_NOT_MAPPED_DATA)')
    NTSTATUS__STATUS_RESOURCE_DATA_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000089, 'HRESULT_FROM_NT(STATUS_RESOURCE_DATA_NOT_FOUND)')
    NTSTATUS__STATUS_RESOURCE_TYPE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD000008A, 'HRESULT_FROM_NT(STATUS_RESOURCE_TYPE_NOT_FOUND)')
    NTSTATUS__STATUS_RESOURCE_NAME_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD000008B, 'HRESULT_FROM_NT(STATUS_RESOURCE_NAME_NOT_FOUND)')
    NTSTATUS__STATUS_ARRAY_BOUNDS_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD000008C, 'HRESULT_FROM_NT(STATUS_ARRAY_BOUNDS_EXCEEDED)')
    NTSTATUS__STATUS_FLOAT_DENORMAL_OPERAND: Final[ThirtyTwoBits] = cached_hresult(0xD000008D, 'HRESULT_FROM_NT(STATUS_FLOAT_DENORMAL_OPERAND)')
    NTSTATUS__STATUS_FLOAT_DIVIDE_BY_ZERO: Final[ThirtyTwoBits] = cached_hresult(0xD000008E, 'HRESULT_FROM_NT(STATUS_FLOAT_DIVIDE_BY_ZERO)')
    NTSTATUS__STATUS_FLOAT_INEXACT_RESULT: Final[ThirtyTwoBits] = cached_hresult(0xD000008F, 'HRESULT_FROM_NT(STATUS_FLOAT_INEXACT_RESULT)')
    NTSTATUS__STATUS_FLOAT_INVALID_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0xD0000090, 'HRESULT_FROM_NT(STATUS_FLOAT_INVALID_OPERATION)')
    NTSTATUS__STATUS_FLOAT_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0xD0000091, 'HRESULT_FROM_NT(STATUS_FLOAT_OVERFLOW)')
    NTSTATUS__STATUS_FLOAT_STACK_CHECK: Final[ThirtyTwoBits] = cached_hresult(0xD0000092, 'HRESULT_FROM_NT(STATUS_FLOAT_STACK_CHECK)')
    NTSTATUS__STATUS_FLOAT_UNDERFLOW: Final[ThirtyTwoBits] = cached_hresult(0xD0000093, 'HRESULT_FROM_NT(STATUS_FLOAT_UNDERFLOW)')
    NTSTATUS__STATUS_INTEGER_DIVIDE_BY_ZERO: Final[ThirtyTwoBits] = cached_hresult(0xD0000094, 'HRESULT_FROM_NT(STATUS_INTEGER_DIVIDE_BY_ZERO)')
    NTSTATUS__STATUS_INTEGER_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0xD0000095, 'HRESULT_FROM_NT(STATUS_INTEGER_OVERFLOW)')
    NTSTATUS__STATUS_PRIVILEGED_INSTRUCTION: Final[ThirtyTwoBits] = cached_hresult(0xD0000096, 'HRESULT_FROM_NT(STATUS_PRIVILEGED_INSTRUCTION)')
    NTSTATUS__STATUS_TOO_MANY_PAGING_FILES: Final[ThirtyTwoBits] = cached_hresult(0xD0000097, 'HRESULT_FROM_NT(STATUS_TOO_MANY_PAGING_FILES)')
    NTSTATUS__STATUS_FILE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0000098, 'HRESULT_FROM_NT(STATUS_FILE_INVALID)')
    NTSTATUS__STATUS_ALLOTTED_SPACE_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD0000099, 'HRESULT_FROM_NT(STATUS_ALLOTTED_SPACE_EXCEEDED)')
    NTSTATUS__STATUS_INSUFFICIENT_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0xD000009A, 'HRESULT_FROM_NT(STATUS_INSUFFICIENT_RESOURCES)')
    NTSTATUS__STATUS_DFS_EXIT_PATH_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD000009B, 'HRESULT_FROM_NT(STATUS_DFS_EXIT_PATH_FOUND)')
    NTSTATUS__STATUS_DEVICE_DATA_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD000009C, 'HRESULT_FROM_NT(STATUS_DEVICE_DATA_ERROR)')
    NTSTATUS__STATUS_DEVICE_NOT_CONNECTED: Final[ThirtyTwoBits] = cached_hresult(0xD000009D, 'HRESULT_FROM_NT(STATUS_DEVICE_NOT_CONNECTED)')
    NTSTATUS__STATUS_DEVICE_POWER_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD000009E, 'HRESULT_FROM_NT(STATUS_DEVICE_POWER_FAILURE)')
    NTSTATUS__STATUS_FREE_VM_NOT_AT_BASE: Final[ThirtyTwoBits] = cached_hresult(0xD000009F, 'HRESULT_FROM_NT(STATUS_FREE_VM_NOT_AT_BASE)')
    NTSTATUS__STATUS_MEMORY_NOT_ALLOCATED: Final[ThirtyTwoBits] = cached_hresult(0xD00000A0, 'HRESULT_FROM_NT(STATUS_MEMORY_NOT_ALLOCATED)')
    NTSTATUS__STATUS_WORKING_SET_QUOTA: Final[ThirtyTwoBits] = cached_hresult(0xD00000A1, 'HRESULT_FROM_NT(STATUS_WORKING_SET_QUOTA)')
    NTSTATUS__STATUS_MEDIA_WRITE_PROTECTED: Final[ThirtyTwoBits] = cached_hresult(0xD00000A2, 'HRESULT_FROM_NT(STATUS_MEDIA_WRITE_PROTECTED)')
    NTSTATUS__STATUS_DEVICE_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0xD00000A3, 'HRESULT_FROM_NT(STATUS_DEVICE_NOT_READY)')
    NTSTATUS__STATUS_INVALID_GROUP_ATTRIBUTES: Final[ThirtyTwoBits] = cached_hresult(0xD00000A4, 'HRESULT_FROM_NT(STATUS_INVALID_GROUP_ATTRIBUTES)')
    NTSTATUS__STATUS_BAD_IMPERSONATION_LEVEL: Final[ThirtyTwoBits] = cached_hresult(0xD00000A5, 'HRESULT_FROM_NT(STATUS_BAD_IMPERSONATION_LEVEL)')
    NTSTATUS__STATUS_CANT_OPEN_ANONYMOUS: Final[ThirtyTwoBits] = cached_hresult(0xD00000A6, 'HRESULT_FROM_NT(STATUS_CANT_OPEN_ANONYMOUS)')
    NTSTATUS__STATUS_BAD_VALIDATION_CLASS: Final[ThirtyTwoBits] = cached_hresult(0xD00000A7, 'HRESULT_FROM_NT(STATUS_BAD_VALIDATION_CLASS)')
    NTSTATUS__STATUS_BAD_TOKEN_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xD00000A8, 'HRESULT_FROM_NT(STATUS_BAD_TOKEN_TYPE)')
    NTSTATUS__STATUS_BAD_MASTER_BOOT_RECORD: Final[ThirtyTwoBits] = cached_hresult(0xD00000A9, 'HRESULT_FROM_NT(STATUS_BAD_MASTER_BOOT_RECORD)')
    NTSTATUS__STATUS_INSTRUCTION_MISALIGNMENT: Final[ThirtyTwoBits] = cached_hresult(0xD00000AA, 'HRESULT_FROM_NT(STATUS_INSTRUCTION_MISALIGNMENT)')
    NTSTATUS__STATUS_INSTANCE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD00000AB, 'HRESULT_FROM_NT(STATUS_INSTANCE_NOT_AVAILABLE)')
    NTSTATUS__STATUS_PIPE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD00000AC, 'HRESULT_FROM_NT(STATUS_PIPE_NOT_AVAILABLE)')
    NTSTATUS__STATUS_INVALID_PIPE_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD00000AD, 'HRESULT_FROM_NT(STATUS_INVALID_PIPE_STATE)')
    NTSTATUS__STATUS_PIPE_BUSY: Final[ThirtyTwoBits] = cached_hresult(0xD00000AE, 'HRESULT_FROM_NT(STATUS_PIPE_BUSY)')
    NTSTATUS__STATUS_ILLEGAL_FUNCTION: Final[ThirtyTwoBits] = cached_hresult(0xD00000AF, 'HRESULT_FROM_NT(STATUS_ILLEGAL_FUNCTION)')
    NTSTATUS__STATUS_PIPE_DISCONNECTED: Final[ThirtyTwoBits] = cached_hresult(0xD00000B0, 'HRESULT_FROM_NT(STATUS_PIPE_DISCONNECTED)')
    NTSTATUS__STATUS_PIPE_CLOSING: Final[ThirtyTwoBits] = cached_hresult(0xD00000B1, 'HRESULT_FROM_NT(STATUS_PIPE_CLOSING)')
    NTSTATUS__STATUS_PIPE_CONNECTED: Final[ThirtyTwoBits] = cached_hresult(0xD00000B2, 'HRESULT_FROM_NT(STATUS_PIPE_CONNECTED)')
    NTSTATUS__STATUS_PIPE_LISTENING: Final[ThirtyTwoBits] = cached_hresult(0xD00000B3, 'HRESULT_FROM_NT(STATUS_PIPE_LISTENING)')
    NTSTATUS__STATUS_INVALID_READ_MODE: Final[ThirtyTwoBits] = cached_hresult(0xD00000B4, 'HRESULT_FROM_NT(STATUS_INVALID_READ_MODE)')
    NTSTATUS__STATUS_IO_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0xD00000B5, 'HRESULT_FROM_NT(STATUS_IO_TIMEOUT)')
    NTSTATUS__STATUS_FILE_FORCED_CLOSED: Final[ThirtyTwoBits] = cached_hresult(0xD00000B6, 'HRESULT_FROM_NT(STATUS_FILE_FORCED_CLOSED)')
    NTSTATUS__STATUS_PROFILING_NOT_STARTED: Final[ThirtyTwoBits] = cached_hresult(0xD00000B7, 'HRESULT_FROM_NT(STATUS_PROFILING_NOT_STARTED)')
    NTSTATUS__STATUS_PROFILING_NOT_STOPPED: Final[ThirtyTwoBits] = cached_hresult(0xD00000B8, 'HRESULT_FROM_NT(STATUS_PROFILING_NOT_STOPPED)')
    NTSTATUS__STATUS_COULD_NOT_INTERPRET: Final[ThirtyTwoBits] = cached_hresult(0xD00000B9, 'HRESULT_FROM_NT(STATUS_COULD_NOT_INTERPRET)')
    NTSTATUS__STATUS_FILE_IS_A_DIRECTORY: Final[ThirtyTwoBits] = cached_hresult(0xD00000BA, 'HRESULT_FROM_NT(STATUS_FILE_IS_A_DIRECTORY)')
    NTSTATUS__STATUS_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD00000BB, 'HRESULT_FROM_NT(STATUS_NOT_SUPPORTED)')
    NTSTATUS__STATUS_REMOTE_NOT_LISTENING: Final[ThirtyTwoBits] = cached_hresult(0xD00000BC, 'HRESULT_FROM_NT(STATUS_REMOTE_NOT_LISTENING)')
    NTSTATUS__STATUS_DUPLICATE_NAME: Final[ThirtyTwoBits] = cached_hresult(0xD00000BD, 'HRESULT_FROM_NT(STATUS_DUPLICATE_NAME)')
    NTSTATUS__STATUS_BAD_NETWORK_PATH: Final[ThirtyTwoBits] = cached_hresult(0xD00000BE, 'HRESULT_FROM_NT(STATUS_BAD_NETWORK_PATH)')
    NTSTATUS__STATUS_NETWORK_BUSY: Final[ThirtyTwoBits] = cached_hresult(0xD00000BF, 'HRESULT_FROM_NT(STATUS_NETWORK_BUSY)')
    NTSTATUS__STATUS_DEVICE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0xD00000C0, 'HRESULT_FROM_NT(STATUS_DEVICE_DOES_NOT_EXIST)')
    NTSTATUS__STATUS_TOO_MANY_COMMANDS: Final[ThirtyTwoBits] = cached_hresult(0xD00000C1, 'HRESULT_FROM_NT(STATUS_TOO_MANY_COMMANDS)')
    NTSTATUS__STATUS_ADAPTER_HARDWARE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD00000C2, 'HRESULT_FROM_NT(STATUS_ADAPTER_HARDWARE_ERROR)')
    NTSTATUS__STATUS_INVALID_NETWORK_RESPONSE: Final[ThirtyTwoBits] = cached_hresult(0xD00000C3, 'HRESULT_FROM_NT(STATUS_INVALID_NETWORK_RESPONSE)')
    NTSTATUS__STATUS_UNEXPECTED_NETWORK_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD00000C4, 'HRESULT_FROM_NT(STATUS_UNEXPECTED_NETWORK_ERROR)')
    NTSTATUS__STATUS_BAD_REMOTE_ADAPTER: Final[ThirtyTwoBits] = cached_hresult(0xD00000C5, 'HRESULT_FROM_NT(STATUS_BAD_REMOTE_ADAPTER)')
    NTSTATUS__STATUS_PRINT_QUEUE_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD00000C6, 'HRESULT_FROM_NT(STATUS_PRINT_QUEUE_FULL)')
    NTSTATUS__STATUS_NO_SPOOL_SPACE: Final[ThirtyTwoBits] = cached_hresult(0xD00000C7, 'HRESULT_FROM_NT(STATUS_NO_SPOOL_SPACE)')
    NTSTATUS__STATUS_PRINT_CANCELLED: Final[ThirtyTwoBits] = cached_hresult(0xD00000C8, 'HRESULT_FROM_NT(STATUS_PRINT_CANCELLED)')
    NTSTATUS__STATUS_NETWORK_NAME_DELETED: Final[ThirtyTwoBits] = cached_hresult(0xD00000C9, 'HRESULT_FROM_NT(STATUS_NETWORK_NAME_DELETED)')
    NTSTATUS__STATUS_NETWORK_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0xD00000CA, 'HRESULT_FROM_NT(STATUS_NETWORK_ACCESS_DENIED)')
    NTSTATUS__STATUS_BAD_DEVICE_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xD00000CB, 'HRESULT_FROM_NT(STATUS_BAD_DEVICE_TYPE)')
    NTSTATUS__STATUS_BAD_NETWORK_NAME: Final[ThirtyTwoBits] = cached_hresult(0xD00000CC, 'HRESULT_FROM_NT(STATUS_BAD_NETWORK_NAME)')
    NTSTATUS__STATUS_TOO_MANY_NAMES: Final[ThirtyTwoBits] = cached_hresult(0xD00000CD, 'HRESULT_FROM_NT(STATUS_TOO_MANY_NAMES)')
    NTSTATUS__STATUS_TOO_MANY_SESSIONS: Final[ThirtyTwoBits] = cached_hresult(0xD00000CE, 'HRESULT_FROM_NT(STATUS_TOO_MANY_SESSIONS)')
    NTSTATUS__STATUS_SHARING_PAUSED: Final[ThirtyTwoBits] = cached_hresult(0xD00000CF, 'HRESULT_FROM_NT(STATUS_SHARING_PAUSED)')
    NTSTATUS__STATUS_REQUEST_NOT_ACCEPTED: Final[ThirtyTwoBits] = cached_hresult(0xD00000D0, 'HRESULT_FROM_NT(STATUS_REQUEST_NOT_ACCEPTED)')
    NTSTATUS__STATUS_REDIRECTOR_PAUSED: Final[ThirtyTwoBits] = cached_hresult(0xD00000D1, 'HRESULT_FROM_NT(STATUS_REDIRECTOR_PAUSED)')
    NTSTATUS__STATUS_NET_WRITE_FAULT: Final[ThirtyTwoBits] = cached_hresult(0xD00000D2, 'HRESULT_FROM_NT(STATUS_NET_WRITE_FAULT)')
    NTSTATUS__STATUS_PROFILING_AT_LIMIT: Final[ThirtyTwoBits] = cached_hresult(0xD00000D3, 'HRESULT_FROM_NT(STATUS_PROFILING_AT_LIMIT)')
    NTSTATUS__STATUS_NOT_SAME_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0xD00000D4, 'HRESULT_FROM_NT(STATUS_NOT_SAME_DEVICE)')
    NTSTATUS__STATUS_FILE_RENAMED: Final[ThirtyTwoBits] = cached_hresult(0xD00000D5, 'HRESULT_FROM_NT(STATUS_FILE_RENAMED)')
    NTSTATUS__STATUS_VIRTUAL_CIRCUIT_CLOSED: Final[ThirtyTwoBits] = cached_hresult(0xD00000D6, 'HRESULT_FROM_NT(STATUS_VIRTUAL_CIRCUIT_CLOSED)')
    NTSTATUS__STATUS_NO_SECURITY_ON_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0xD00000D7, 'HRESULT_FROM_NT(STATUS_NO_SECURITY_ON_OBJECT)')
    NTSTATUS__STATUS_CANT_WAIT: Final[ThirtyTwoBits] = cached_hresult(0xD00000D8, 'HRESULT_FROM_NT(STATUS_CANT_WAIT)')
    NTSTATUS__STATUS_PIPE_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0xD00000D9, 'HRESULT_FROM_NT(STATUS_PIPE_EMPTY)')
    NTSTATUS__STATUS_CANT_ACCESS_DOMAIN_INFO: Final[ThirtyTwoBits] = cached_hresult(0xD00000DA, 'HRESULT_FROM_NT(STATUS_CANT_ACCESS_DOMAIN_INFO)')
    NTSTATUS__STATUS_CANT_TERMINATE_SELF: Final[ThirtyTwoBits] = cached_hresult(0xD00000DB, 'HRESULT_FROM_NT(STATUS_CANT_TERMINATE_SELF)')
    NTSTATUS__STATUS_INVALID_SERVER_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD00000DC, 'HRESULT_FROM_NT(STATUS_INVALID_SERVER_STATE)')
    NTSTATUS__STATUS_INVALID_DOMAIN_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD00000DD, 'HRESULT_FROM_NT(STATUS_INVALID_DOMAIN_STATE)')
    NTSTATUS__STATUS_INVALID_DOMAIN_ROLE: Final[ThirtyTwoBits] = cached_hresult(0xD00000DE, 'HRESULT_FROM_NT(STATUS_INVALID_DOMAIN_ROLE)')
    NTSTATUS__STATUS_NO_SUCH_DOMAIN: Final[ThirtyTwoBits] = cached_hresult(0xD00000DF, 'HRESULT_FROM_NT(STATUS_NO_SUCH_DOMAIN)')
    NTSTATUS__STATUS_DOMAIN_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD00000E0, 'HRESULT_FROM_NT(STATUS_DOMAIN_EXISTS)')
    NTSTATUS__STATUS_DOMAIN_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD00000E1, 'HRESULT_FROM_NT(STATUS_DOMAIN_LIMIT_EXCEEDED)')
    NTSTATUS__STATUS_OPLOCK_NOT_GRANTED: Final[ThirtyTwoBits] = cached_hresult(0xD00000E2, 'HRESULT_FROM_NT(STATUS_OPLOCK_NOT_GRANTED)')
    NTSTATUS__STATUS_INVALID_OPLOCK_PROTOCOL: Final[ThirtyTwoBits] = cached_hresult(0xD00000E3, 'HRESULT_FROM_NT(STATUS_INVALID_OPLOCK_PROTOCOL)')
    NTSTATUS__STATUS_INTERNAL_DB_CORRUPTION: Final[ThirtyTwoBits] = cached_hresult(0xD00000E4, 'HRESULT_FROM_NT(STATUS_INTERNAL_DB_CORRUPTION)')
    NTSTATUS__STATUS_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD00000E5, 'HRESULT_FROM_NT(STATUS_INTERNAL_ERROR)')
    NTSTATUS__STATUS_GENERIC_NOT_MAPPED: Final[ThirtyTwoBits] = cached_hresult(0xD00000E6, 'HRESULT_FROM_NT(STATUS_GENERIC_NOT_MAPPED)')
    NTSTATUS__STATUS_BAD_DESCRIPTOR_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0xD00000E7, 'HRESULT_FROM_NT(STATUS_BAD_DESCRIPTOR_FORMAT)')
    NTSTATUS__STATUS_INVALID_USER_BUFFER: Final[ThirtyTwoBits] = cached_hresult(0xD00000E8, 'HRESULT_FROM_NT(STATUS_INVALID_USER_BUFFER)')
    NTSTATUS__STATUS_UNEXPECTED_IO_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD00000E9, 'HRESULT_FROM_NT(STATUS_UNEXPECTED_IO_ERROR)')
    NTSTATUS__STATUS_UNEXPECTED_MM_CREATE_ERR: Final[ThirtyTwoBits] = cached_hresult(0xD00000EA, 'HRESULT_FROM_NT(STATUS_UNEXPECTED_MM_CREATE_ERR)')
    NTSTATUS__STATUS_UNEXPECTED_MM_MAP_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD00000EB, 'HRESULT_FROM_NT(STATUS_UNEXPECTED_MM_MAP_ERROR)')
    NTSTATUS__STATUS_UNEXPECTED_MM_EXTEND_ERR: Final[ThirtyTwoBits] = cached_hresult(0xD00000EC, 'HRESULT_FROM_NT(STATUS_UNEXPECTED_MM_EXTEND_ERR)')
    NTSTATUS__STATUS_NOT_LOGON_PROCESS: Final[ThirtyTwoBits] = cached_hresult(0xD00000ED, 'HRESULT_FROM_NT(STATUS_NOT_LOGON_PROCESS)')
    NTSTATUS__STATUS_LOGON_SESSION_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD00000EE, 'HRESULT_FROM_NT(STATUS_LOGON_SESSION_EXISTS)')
    NTSTATUS__STATUS_INVALID_PARAMETER_1: Final[ThirtyTwoBits] = cached_hresult(0xD00000EF, 'HRESULT_FROM_NT(STATUS_INVALID_PARAMETER_1)')
    NTSTATUS__STATUS_INVALID_PARAMETER_2: Final[ThirtyTwoBits] = cached_hresult(0xD00000F0, 'HRESULT_FROM_NT(STATUS_INVALID_PARAMETER_2)')
    NTSTATUS__STATUS_INVALID_PARAMETER_3: Final[ThirtyTwoBits] = cached_hresult(0xD00000F1, 'HRESULT_FROM_NT(STATUS_INVALID_PARAMETER_3)')
    NTSTATUS__STATUS_INVALID_PARAMETER_4: Final[ThirtyTwoBits] = cached_hresult(0xD00000F2, 'HRESULT_FROM_NT(STATUS_INVALID_PARAMETER_4)')
    NTSTATUS__STATUS_INVALID_PARAMETER_5: Final[ThirtyTwoBits] = cached_hresult(0xD00000F3, 'HRESULT_FROM_NT(STATUS_INVALID_PARAMETER_5)')
    NTSTATUS__STATUS_INVALID_PARAMETER_6: Final[ThirtyTwoBits] = cached_hresult(0xD00000F4, 'HRESULT_FROM_NT(STATUS_INVALID_PARAMETER_6)')
    NTSTATUS__STATUS_INVALID_PARAMETER_7: Final[ThirtyTwoBits] = cached_hresult(0xD00000F5, 'HRESULT_FROM_NT(STATUS_INVALID_PARAMETER_7)')
    NTSTATUS__STATUS_INVALID_PARAMETER_8: Final[ThirtyTwoBits] = cached_hresult(0xD00000F6, 'HRESULT_FROM_NT(STATUS_INVALID_PARAMETER_8)')
    NTSTATUS__STATUS_INVALID_PARAMETER_9: Final[ThirtyTwoBits] = cached_hresult(0xD00000F7, 'HRESULT_FROM_NT(STATUS_INVALID_PARAMETER_9)')
    NTSTATUS__STATUS_INVALID_PARAMETER_10: Final[ThirtyTwoBits] = cached_hresult(0xD00000F8, 'HRESULT_FROM_NT(STATUS_INVALID_PARAMETER_10)')
    NTSTATUS__STATUS_INVALID_PARAMETER_11: Final[ThirtyTwoBits] = cached_hresult(0xD00000F9, 'HRESULT_FROM_NT(STATUS_INVALID_PARAMETER_11)')
    NTSTATUS__STATUS_INVALID_PARAMETER_12: Final[ThirtyTwoBits] = cached_hresult(0xD00000FA, 'HRESULT_FROM_NT(STATUS_INVALID_PARAMETER_12)')
    NTSTATUS__STATUS_REDIRECTOR_NOT_STARTED: Final[ThirtyTwoBits] = cached_hresult(0xD00000FB, 'HRESULT_FROM_NT(STATUS_REDIRECTOR_NOT_STARTED)')
    NTSTATUS__STATUS_REDIRECTOR_STARTED: Final[ThirtyTwoBits] = cached_hresult(0xD00000FC, 'HRESULT_FROM_NT(STATUS_REDIRECTOR_STARTED)')
    NTSTATUS__STATUS_STACK_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0xD00000FD, 'HRESULT_FROM_NT(STATUS_STACK_OVERFLOW)')
    NTSTATUS__STATUS_NO_SUCH_PACKAGE: Final[ThirtyTwoBits] = cached_hresult(0xD00000FE, 'HRESULT_FROM_NT(STATUS_NO_SUCH_PACKAGE)')
    NTSTATUS__STATUS_BAD_FUNCTION_TABLE: Final[ThirtyTwoBits] = cached_hresult(0xD00000FF, 'HRESULT_FROM_NT(STATUS_BAD_FUNCTION_TABLE)')
    NTSTATUS__STATUS_VARIABLE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000100, 'HRESULT_FROM_NT(STATUS_VARIABLE_NOT_FOUND)')
    NTSTATUS__STATUS_DIRECTORY_NOT_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0xD0000101, 'HRESULT_FROM_NT(STATUS_DIRECTORY_NOT_EMPTY)')
    NTSTATUS__STATUS_FILE_CORRUPT_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0000102, 'HRESULT_FROM_NT(STATUS_FILE_CORRUPT_ERROR)')
    NTSTATUS__STATUS_NOT_A_DIRECTORY: Final[ThirtyTwoBits] = cached_hresult(0xD0000103, 'HRESULT_FROM_NT(STATUS_NOT_A_DIRECTORY)')
    NTSTATUS__STATUS_BAD_LOGON_SESSION_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD0000104, 'HRESULT_FROM_NT(STATUS_BAD_LOGON_SESSION_STATE)')
    NTSTATUS__STATUS_LOGON_SESSION_COLLISION: Final[ThirtyTwoBits] = cached_hresult(0xD0000105, 'HRESULT_FROM_NT(STATUS_LOGON_SESSION_COLLISION)')
    NTSTATUS__STATUS_NAME_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0xD0000106, 'HRESULT_FROM_NT(STATUS_NAME_TOO_LONG)')
    NTSTATUS__STATUS_FILES_OPEN: Final[ThirtyTwoBits] = cached_hresult(0xD0000107, 'HRESULT_FROM_NT(STATUS_FILES_OPEN)')
    NTSTATUS__STATUS_CONNECTION_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0xD0000108, 'HRESULT_FROM_NT(STATUS_CONNECTION_IN_USE)')
    NTSTATUS__STATUS_MESSAGE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000109, 'HRESULT_FROM_NT(STATUS_MESSAGE_NOT_FOUND)')
    NTSTATUS__STATUS_PROCESS_IS_TERMINATING: Final[ThirtyTwoBits] = cached_hresult(0xD000010A, 'HRESULT_FROM_NT(STATUS_PROCESS_IS_TERMINATING)')
    NTSTATUS__STATUS_INVALID_LOGON_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xD000010B, 'HRESULT_FROM_NT(STATUS_INVALID_LOGON_TYPE)')
    NTSTATUS__STATUS_NO_GUID_TRANSLATION: Final[ThirtyTwoBits] = cached_hresult(0xD000010C, 'HRESULT_FROM_NT(STATUS_NO_GUID_TRANSLATION)')
    NTSTATUS__STATUS_CANNOT_IMPERSONATE: Final[ThirtyTwoBits] = cached_hresult(0xD000010D, 'HRESULT_FROM_NT(STATUS_CANNOT_IMPERSONATE)')
    NTSTATUS__STATUS_IMAGE_ALREADY_LOADED: Final[ThirtyTwoBits] = cached_hresult(0xD000010E, 'HRESULT_FROM_NT(STATUS_IMAGE_ALREADY_LOADED)')
    NTSTATUS__STATUS_ABIOS_NOT_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0xD000010F, 'HRESULT_FROM_NT(STATUS_ABIOS_NOT_PRESENT)')
    NTSTATUS__STATUS_ABIOS_LID_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0xD0000110, 'HRESULT_FROM_NT(STATUS_ABIOS_LID_NOT_EXIST)')
    NTSTATUS__STATUS_ABIOS_LID_ALREADY_OWNED: Final[ThirtyTwoBits] = cached_hresult(0xD0000111, 'HRESULT_FROM_NT(STATUS_ABIOS_LID_ALREADY_OWNED)')
    NTSTATUS__STATUS_ABIOS_NOT_LID_OWNER: Final[ThirtyTwoBits] = cached_hresult(0xD0000112, 'HRESULT_FROM_NT(STATUS_ABIOS_NOT_LID_OWNER)')
    NTSTATUS__STATUS_ABIOS_INVALID_COMMAND: Final[ThirtyTwoBits] = cached_hresult(0xD0000113, 'HRESULT_FROM_NT(STATUS_ABIOS_INVALID_COMMAND)')
    NTSTATUS__STATUS_ABIOS_INVALID_LID: Final[ThirtyTwoBits] = cached_hresult(0xD0000114, 'HRESULT_FROM_NT(STATUS_ABIOS_INVALID_LID)')
    NTSTATUS__STATUS_ABIOS_SELECTOR_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD0000115, 'HRESULT_FROM_NT(STATUS_ABIOS_SELECTOR_NOT_AVAILABLE)')
    NTSTATUS__STATUS_ABIOS_INVALID_SELECTOR: Final[ThirtyTwoBits] = cached_hresult(0xD0000116, 'HRESULT_FROM_NT(STATUS_ABIOS_INVALID_SELECTOR)')
    NTSTATUS__STATUS_NO_LDT: Final[ThirtyTwoBits] = cached_hresult(0xD0000117, 'HRESULT_FROM_NT(STATUS_NO_LDT)')
    NTSTATUS__STATUS_INVALID_LDT_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD0000118, 'HRESULT_FROM_NT(STATUS_INVALID_LDT_SIZE)')
    NTSTATUS__STATUS_INVALID_LDT_OFFSET: Final[ThirtyTwoBits] = cached_hresult(0xD0000119, 'HRESULT_FROM_NT(STATUS_INVALID_LDT_OFFSET)')
    NTSTATUS__STATUS_INVALID_LDT_DESCRIPTOR: Final[ThirtyTwoBits] = cached_hresult(0xD000011A, 'HRESULT_FROM_NT(STATUS_INVALID_LDT_DESCRIPTOR)')
    NTSTATUS__STATUS_INVALID_IMAGE_NE_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0xD000011B, 'HRESULT_FROM_NT(STATUS_INVALID_IMAGE_NE_FORMAT)')
    NTSTATUS__STATUS_RXACT_INVALID_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD000011C, 'HRESULT_FROM_NT(STATUS_RXACT_INVALID_STATE)')
    NTSTATUS__STATUS_RXACT_COMMIT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD000011D, 'HRESULT_FROM_NT(STATUS_RXACT_COMMIT_FAILURE)')
    NTSTATUS__STATUS_MAPPED_FILE_SIZE_ZERO: Final[ThirtyTwoBits] = cached_hresult(0xD000011E, 'HRESULT_FROM_NT(STATUS_MAPPED_FILE_SIZE_ZERO)')
    NTSTATUS__STATUS_TOO_MANY_OPENED_FILES: Final[ThirtyTwoBits] = cached_hresult(0xD000011F, 'HRESULT_FROM_NT(STATUS_TOO_MANY_OPENED_FILES)')
    NTSTATUS__STATUS_CANCELLED: Final[ThirtyTwoBits] = cached_hresult(0xD0000120, 'HRESULT_FROM_NT(STATUS_CANCELLED)')
    NTSTATUS__STATUS_CANNOT_DELETE: Final[ThirtyTwoBits] = cached_hresult(0xD0000121, 'HRESULT_FROM_NT(STATUS_CANNOT_DELETE)')
    NTSTATUS__STATUS_INVALID_COMPUTER_NAME: Final[ThirtyTwoBits] = cached_hresult(0xD0000122, 'HRESULT_FROM_NT(STATUS_INVALID_COMPUTER_NAME)')
    NTSTATUS__STATUS_FILE_DELETED: Final[ThirtyTwoBits] = cached_hresult(0xD0000123, 'HRESULT_FROM_NT(STATUS_FILE_DELETED)')
    NTSTATUS__STATUS_SPECIAL_ACCOUNT: Final[ThirtyTwoBits] = cached_hresult(0xD0000124, 'HRESULT_FROM_NT(STATUS_SPECIAL_ACCOUNT)')
    NTSTATUS__STATUS_SPECIAL_GROUP: Final[ThirtyTwoBits] = cached_hresult(0xD0000125, 'HRESULT_FROM_NT(STATUS_SPECIAL_GROUP)')
    NTSTATUS__STATUS_SPECIAL_USER: Final[ThirtyTwoBits] = cached_hresult(0xD0000126, 'HRESULT_FROM_NT(STATUS_SPECIAL_USER)')
    NTSTATUS__STATUS_MEMBERS_PRIMARY_GROUP: Final[ThirtyTwoBits] = cached_hresult(0xD0000127, 'HRESULT_FROM_NT(STATUS_MEMBERS_PRIMARY_GROUP)')
    NTSTATUS__STATUS_FILE_CLOSED: Final[ThirtyTwoBits] = cached_hresult(0xD0000128, 'HRESULT_FROM_NT(STATUS_FILE_CLOSED)')
    NTSTATUS__STATUS_TOO_MANY_THREADS: Final[ThirtyTwoBits] = cached_hresult(0xD0000129, 'HRESULT_FROM_NT(STATUS_TOO_MANY_THREADS)')
    NTSTATUS__STATUS_THREAD_NOT_IN_PROCESS: Final[ThirtyTwoBits] = cached_hresult(0xD000012A, 'HRESULT_FROM_NT(STATUS_THREAD_NOT_IN_PROCESS)')
    NTSTATUS__STATUS_TOKEN_ALREADY_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0xD000012B, 'HRESULT_FROM_NT(STATUS_TOKEN_ALREADY_IN_USE)')
    NTSTATUS__STATUS_PAGEFILE_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD000012C, 'HRESULT_FROM_NT(STATUS_PAGEFILE_QUOTA_EXCEEDED)')
    NTSTATUS__STATUS_COMMITMENT_LIMIT: Final[ThirtyTwoBits] = cached_hresult(0xD000012D, 'HRESULT_FROM_NT(STATUS_COMMITMENT_LIMIT)')
    NTSTATUS__STATUS_INVALID_IMAGE_LE_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0xD000012E, 'HRESULT_FROM_NT(STATUS_INVALID_IMAGE_LE_FORMAT)')
    NTSTATUS__STATUS_INVALID_IMAGE_NOT_MZ: Final[ThirtyTwoBits] = cached_hresult(0xD000012F, 'HRESULT_FROM_NT(STATUS_INVALID_IMAGE_NOT_MZ)')
    NTSTATUS__STATUS_INVALID_IMAGE_PROTECT: Final[ThirtyTwoBits] = cached_hresult(0xD0000130, 'HRESULT_FROM_NT(STATUS_INVALID_IMAGE_PROTECT)')
    NTSTATUS__STATUS_INVALID_IMAGE_WIN_16: Final[ThirtyTwoBits] = cached_hresult(0xD0000131, 'HRESULT_FROM_NT(STATUS_INVALID_IMAGE_WIN_16)')
    NTSTATUS__STATUS_LOGON_SERVER_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0xD0000132, 'HRESULT_FROM_NT(STATUS_LOGON_SERVER_CONFLICT)')
    NTSTATUS__STATUS_TIME_DIFFERENCE_AT_DC: Final[ThirtyTwoBits] = cached_hresult(0xD0000133, 'HRESULT_FROM_NT(STATUS_TIME_DIFFERENCE_AT_DC)')
    NTSTATUS__STATUS_SYNCHRONIZATION_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0xD0000134, 'HRESULT_FROM_NT(STATUS_SYNCHRONIZATION_REQUIRED)')
    NTSTATUS__STATUS_DLL_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000135, 'HRESULT_FROM_NT(STATUS_DLL_NOT_FOUND)')
    NTSTATUS__STATUS_OPEN_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0000136, 'HRESULT_FROM_NT(STATUS_OPEN_FAILED)')
    NTSTATUS__STATUS_IO_PRIVILEGE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0000137, 'HRESULT_FROM_NT(STATUS_IO_PRIVILEGE_FAILED)')
    NTSTATUS__STATUS_ORDINAL_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000138, 'HRESULT_FROM_NT(STATUS_ORDINAL_NOT_FOUND)')
    NTSTATUS__STATUS_ENTRYPOINT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000139, 'HRESULT_FROM_NT(STATUS_ENTRYPOINT_NOT_FOUND)')
    NTSTATUS__STATUS_CONTROL_C_EXIT: Final[ThirtyTwoBits] = cached_hresult(0xD000013A, 'HRESULT_FROM_NT(STATUS_CONTROL_C_EXIT)')
    NTSTATUS__STATUS_LOCAL_DISCONNECT: Final[ThirtyTwoBits] = cached_hresult(0xD000013B, 'HRESULT_FROM_NT(STATUS_LOCAL_DISCONNECT)')
    NTSTATUS__STATUS_REMOTE_DISCONNECT: Final[ThirtyTwoBits] = cached_hresult(0xD000013C, 'HRESULT_FROM_NT(STATUS_REMOTE_DISCONNECT)')
    NTSTATUS__STATUS_REMOTE_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0xD000013D, 'HRESULT_FROM_NT(STATUS_REMOTE_RESOURCES)')
    NTSTATUS__STATUS_LINK_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD000013E, 'HRESULT_FROM_NT(STATUS_LINK_FAILED)')
    NTSTATUS__STATUS_LINK_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0xD000013F, 'HRESULT_FROM_NT(STATUS_LINK_TIMEOUT)')
    NTSTATUS__STATUS_INVALID_CONNECTION: Final[ThirtyTwoBits] = cached_hresult(0xD0000140, 'HRESULT_FROM_NT(STATUS_INVALID_CONNECTION)')
    NTSTATUS__STATUS_INVALID_ADDRESS: Final[ThirtyTwoBits] = cached_hresult(0xD0000141, 'HRESULT_FROM_NT(STATUS_INVALID_ADDRESS)')
    NTSTATUS__STATUS_DLL_INIT_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0000142, 'HRESULT_FROM_NT(STATUS_DLL_INIT_FAILED)')
    NTSTATUS__STATUS_MISSING_SYSTEMFILE: Final[ThirtyTwoBits] = cached_hresult(0xD0000143, 'HRESULT_FROM_NT(STATUS_MISSING_SYSTEMFILE)')
    NTSTATUS__STATUS_UNHANDLED_EXCEPTION: Final[ThirtyTwoBits] = cached_hresult(0xD0000144, 'HRESULT_FROM_NT(STATUS_UNHANDLED_EXCEPTION)')
    NTSTATUS__STATUS_APP_INIT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD0000145, 'HRESULT_FROM_NT(STATUS_APP_INIT_FAILURE)')
    NTSTATUS__STATUS_PAGEFILE_CREATE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0000146, 'HRESULT_FROM_NT(STATUS_PAGEFILE_CREATE_FAILED)')
    NTSTATUS__STATUS_NO_PAGEFILE: Final[ThirtyTwoBits] = cached_hresult(0xD0000147, 'HRESULT_FROM_NT(STATUS_NO_PAGEFILE)')
    NTSTATUS__STATUS_INVALID_LEVEL: Final[ThirtyTwoBits] = cached_hresult(0xD0000148, 'HRESULT_FROM_NT(STATUS_INVALID_LEVEL)')
    NTSTATUS__STATUS_WRONG_PASSWORD_CORE: Final[ThirtyTwoBits] = cached_hresult(0xD0000149, 'HRESULT_FROM_NT(STATUS_WRONG_PASSWORD_CORE)')
    NTSTATUS__STATUS_ILLEGAL_FLOAT_CONTEXT: Final[ThirtyTwoBits] = cached_hresult(0xD000014A, 'HRESULT_FROM_NT(STATUS_ILLEGAL_FLOAT_CONTEXT)')
    NTSTATUS__STATUS_PIPE_BROKEN: Final[ThirtyTwoBits] = cached_hresult(0xD000014B, 'HRESULT_FROM_NT(STATUS_PIPE_BROKEN)')
    NTSTATUS__STATUS_REGISTRY_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0xD000014C, 'HRESULT_FROM_NT(STATUS_REGISTRY_CORRUPT)')
    NTSTATUS__STATUS_REGISTRY_IO_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD000014D, 'HRESULT_FROM_NT(STATUS_REGISTRY_IO_FAILED)')
    NTSTATUS__STATUS_NO_EVENT_PAIR: Final[ThirtyTwoBits] = cached_hresult(0xD000014E, 'HRESULT_FROM_NT(STATUS_NO_EVENT_PAIR)')
    NTSTATUS__STATUS_UNRECOGNIZED_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0xD000014F, 'HRESULT_FROM_NT(STATUS_UNRECOGNIZED_VOLUME)')
    NTSTATUS__STATUS_SERIAL_NO_DEVICE_INITED: Final[ThirtyTwoBits] = cached_hresult(0xD0000150, 'HRESULT_FROM_NT(STATUS_SERIAL_NO_DEVICE_INITED)')
    NTSTATUS__STATUS_NO_SUCH_ALIAS: Final[ThirtyTwoBits] = cached_hresult(0xD0000151, 'HRESULT_FROM_NT(STATUS_NO_SUCH_ALIAS)')
    NTSTATUS__STATUS_MEMBER_NOT_IN_ALIAS: Final[ThirtyTwoBits] = cached_hresult(0xD0000152, 'HRESULT_FROM_NT(STATUS_MEMBER_NOT_IN_ALIAS)')
    NTSTATUS__STATUS_MEMBER_IN_ALIAS: Final[ThirtyTwoBits] = cached_hresult(0xD0000153, 'HRESULT_FROM_NT(STATUS_MEMBER_IN_ALIAS)')
    NTSTATUS__STATUS_ALIAS_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD0000154, 'HRESULT_FROM_NT(STATUS_ALIAS_EXISTS)')
    NTSTATUS__STATUS_LOGON_NOT_GRANTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000155, 'HRESULT_FROM_NT(STATUS_LOGON_NOT_GRANTED)')
    NTSTATUS__STATUS_TOO_MANY_SECRETS: Final[ThirtyTwoBits] = cached_hresult(0xD0000156, 'HRESULT_FROM_NT(STATUS_TOO_MANY_SECRETS)')
    NTSTATUS__STATUS_SECRET_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0xD0000157, 'HRESULT_FROM_NT(STATUS_SECRET_TOO_LONG)')
    NTSTATUS__STATUS_INTERNAL_DB_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0000158, 'HRESULT_FROM_NT(STATUS_INTERNAL_DB_ERROR)')
    NTSTATUS__STATUS_FULLSCREEN_MODE: Final[ThirtyTwoBits] = cached_hresult(0xD0000159, 'HRESULT_FROM_NT(STATUS_FULLSCREEN_MODE)')
    NTSTATUS__STATUS_TOO_MANY_CONTEXT_IDS: Final[ThirtyTwoBits] = cached_hresult(0xD000015A, 'HRESULT_FROM_NT(STATUS_TOO_MANY_CONTEXT_IDS)')
    NTSTATUS__STATUS_LOGON_TYPE_NOT_GRANTED: Final[ThirtyTwoBits] = cached_hresult(0xD000015B, 'HRESULT_FROM_NT(STATUS_LOGON_TYPE_NOT_GRANTED)')
    NTSTATUS__STATUS_NOT_REGISTRY_FILE: Final[ThirtyTwoBits] = cached_hresult(0xD000015C, 'HRESULT_FROM_NT(STATUS_NOT_REGISTRY_FILE)')
    NTSTATUS__STATUS_NT_CROSS_ENCRYPTION_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0xD000015D, 'HRESULT_FROM_NT(STATUS_NT_CROSS_ENCRYPTION_REQUIRED)')
    NTSTATUS__STATUS_DOMAIN_CTRLR_CONFIG_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD000015E, 'HRESULT_FROM_NT(STATUS_DOMAIN_CTRLR_CONFIG_ERROR)')
    NTSTATUS__STATUS_FT_MISSING_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0xD000015F, 'HRESULT_FROM_NT(STATUS_FT_MISSING_MEMBER)')
    NTSTATUS__STATUS_ILL_FORMED_SERVICE_ENTRY: Final[ThirtyTwoBits] = cached_hresult(0xD0000160, 'HRESULT_FROM_NT(STATUS_ILL_FORMED_SERVICE_ENTRY)')
    NTSTATUS__STATUS_ILLEGAL_CHARACTER: Final[ThirtyTwoBits] = cached_hresult(0xD0000161, 'HRESULT_FROM_NT(STATUS_ILLEGAL_CHARACTER)')
    NTSTATUS__STATUS_UNMAPPABLE_CHARACTER: Final[ThirtyTwoBits] = cached_hresult(0xD0000162, 'HRESULT_FROM_NT(STATUS_UNMAPPABLE_CHARACTER)')
    NTSTATUS__STATUS_UNDEFINED_CHARACTER: Final[ThirtyTwoBits] = cached_hresult(0xD0000163, 'HRESULT_FROM_NT(STATUS_UNDEFINED_CHARACTER)')
    NTSTATUS__STATUS_FLOPPY_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0xD0000164, 'HRESULT_FROM_NT(STATUS_FLOPPY_VOLUME)')
    NTSTATUS__STATUS_FLOPPY_ID_MARK_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000165, 'HRESULT_FROM_NT(STATUS_FLOPPY_ID_MARK_NOT_FOUND)')
    NTSTATUS__STATUS_FLOPPY_WRONG_CYLINDER: Final[ThirtyTwoBits] = cached_hresult(0xD0000166, 'HRESULT_FROM_NT(STATUS_FLOPPY_WRONG_CYLINDER)')
    NTSTATUS__STATUS_FLOPPY_UNKNOWN_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0000167, 'HRESULT_FROM_NT(STATUS_FLOPPY_UNKNOWN_ERROR)')
    NTSTATUS__STATUS_FLOPPY_BAD_REGISTERS: Final[ThirtyTwoBits] = cached_hresult(0xD0000168, 'HRESULT_FROM_NT(STATUS_FLOPPY_BAD_REGISTERS)')
    NTSTATUS__STATUS_DISK_RECALIBRATE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0000169, 'HRESULT_FROM_NT(STATUS_DISK_RECALIBRATE_FAILED)')
    NTSTATUS__STATUS_DISK_OPERATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD000016A, 'HRESULT_FROM_NT(STATUS_DISK_OPERATION_FAILED)')
    NTSTATUS__STATUS_DISK_RESET_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD000016B, 'HRESULT_FROM_NT(STATUS_DISK_RESET_FAILED)')
    NTSTATUS__STATUS_SHARED_IRQ_BUSY: Final[ThirtyTwoBits] = cached_hresult(0xD000016C, 'HRESULT_FROM_NT(STATUS_SHARED_IRQ_BUSY)')
    NTSTATUS__STATUS_FT_ORPHANING: Final[ThirtyTwoBits] = cached_hresult(0xD000016D, 'HRESULT_FROM_NT(STATUS_FT_ORPHANING)')
    NTSTATUS__STATUS_BIOS_FAILED_TO_CONNECT_INTERRUPT: Final[ThirtyTwoBits] = cached_hresult(0xD000016E, 'HRESULT_FROM_NT(STATUS_BIOS_FAILED_TO_CONNECT_INTERRUPT)')
    NTSTATUS__STATUS_PARTITION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD0000172, 'HRESULT_FROM_NT(STATUS_PARTITION_FAILURE)')
    NTSTATUS__STATUS_INVALID_BLOCK_LENGTH: Final[ThirtyTwoBits] = cached_hresult(0xD0000173, 'HRESULT_FROM_NT(STATUS_INVALID_BLOCK_LENGTH)')
    NTSTATUS__STATUS_DEVICE_NOT_PARTITIONED: Final[ThirtyTwoBits] = cached_hresult(0xD0000174, 'HRESULT_FROM_NT(STATUS_DEVICE_NOT_PARTITIONED)')
    NTSTATUS__STATUS_UNABLE_TO_LOCK_MEDIA: Final[ThirtyTwoBits] = cached_hresult(0xD0000175, 'HRESULT_FROM_NT(STATUS_UNABLE_TO_LOCK_MEDIA)')
    NTSTATUS__STATUS_UNABLE_TO_UNLOAD_MEDIA: Final[ThirtyTwoBits] = cached_hresult(0xD0000176, 'HRESULT_FROM_NT(STATUS_UNABLE_TO_UNLOAD_MEDIA)')
    NTSTATUS__STATUS_EOM_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0xD0000177, 'HRESULT_FROM_NT(STATUS_EOM_OVERFLOW)')
    NTSTATUS__STATUS_NO_MEDIA: Final[ThirtyTwoBits] = cached_hresult(0xD0000178, 'HRESULT_FROM_NT(STATUS_NO_MEDIA)')
    NTSTATUS__STATUS_NO_SUCH_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0xD000017A, 'HRESULT_FROM_NT(STATUS_NO_SUCH_MEMBER)')
    NTSTATUS__STATUS_INVALID_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0xD000017B, 'HRESULT_FROM_NT(STATUS_INVALID_MEMBER)')
    NTSTATUS__STATUS_KEY_DELETED: Final[ThirtyTwoBits] = cached_hresult(0xD000017C, 'HRESULT_FROM_NT(STATUS_KEY_DELETED)')
    NTSTATUS__STATUS_NO_LOG_SPACE: Final[ThirtyTwoBits] = cached_hresult(0xD000017D, 'HRESULT_FROM_NT(STATUS_NO_LOG_SPACE)')
    NTSTATUS__STATUS_TOO_MANY_SIDS: Final[ThirtyTwoBits] = cached_hresult(0xD000017E, 'HRESULT_FROM_NT(STATUS_TOO_MANY_SIDS)')
    NTSTATUS__STATUS_LM_CROSS_ENCRYPTION_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0xD000017F, 'HRESULT_FROM_NT(STATUS_LM_CROSS_ENCRYPTION_REQUIRED)')
    NTSTATUS__STATUS_KEY_HAS_CHILDREN: Final[ThirtyTwoBits] = cached_hresult(0xD0000180, 'HRESULT_FROM_NT(STATUS_KEY_HAS_CHILDREN)')
    NTSTATUS__STATUS_CHILD_MUST_BE_VOLATILE: Final[ThirtyTwoBits] = cached_hresult(0xD0000181, 'HRESULT_FROM_NT(STATUS_CHILD_MUST_BE_VOLATILE)')
    NTSTATUS__STATUS_DEVICE_CONFIGURATION_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0000182, 'HRESULT_FROM_NT(STATUS_DEVICE_CONFIGURATION_ERROR)')
    NTSTATUS__STATUS_DRIVER_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0000183, 'HRESULT_FROM_NT(STATUS_DRIVER_INTERNAL_ERROR)')
    NTSTATUS__STATUS_INVALID_DEVICE_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD0000184, 'HRESULT_FROM_NT(STATUS_INVALID_DEVICE_STATE)')
    NTSTATUS__STATUS_IO_DEVICE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0000185, 'HRESULT_FROM_NT(STATUS_IO_DEVICE_ERROR)')
    NTSTATUS__STATUS_DEVICE_PROTOCOL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0000186, 'HRESULT_FROM_NT(STATUS_DEVICE_PROTOCOL_ERROR)')
    NTSTATUS__STATUS_BACKUP_CONTROLLER: Final[ThirtyTwoBits] = cached_hresult(0xD0000187, 'HRESULT_FROM_NT(STATUS_BACKUP_CONTROLLER)')
    NTSTATUS__STATUS_LOG_FILE_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD0000188, 'HRESULT_FROM_NT(STATUS_LOG_FILE_FULL)')
    NTSTATUS__STATUS_TOO_LATE: Final[ThirtyTwoBits] = cached_hresult(0xD0000189, 'HRESULT_FROM_NT(STATUS_TOO_LATE)')
    NTSTATUS__STATUS_NO_TRUST_LSA_SECRET: Final[ThirtyTwoBits] = cached_hresult(0xD000018A, 'HRESULT_FROM_NT(STATUS_NO_TRUST_LSA_SECRET)')
    NTSTATUS__STATUS_NO_TRUST_SAM_ACCOUNT: Final[ThirtyTwoBits] = cached_hresult(0xD000018B, 'HRESULT_FROM_NT(STATUS_NO_TRUST_SAM_ACCOUNT)')
    NTSTATUS__STATUS_TRUSTED_DOMAIN_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD000018C, 'HRESULT_FROM_NT(STATUS_TRUSTED_DOMAIN_FAILURE)')
    NTSTATUS__STATUS_TRUSTED_RELATIONSHIP_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD000018D, 'HRESULT_FROM_NT(STATUS_TRUSTED_RELATIONSHIP_FAILURE)')
    NTSTATUS__STATUS_EVENTLOG_FILE_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0xD000018E, 'HRESULT_FROM_NT(STATUS_EVENTLOG_FILE_CORRUPT)')
    NTSTATUS__STATUS_EVENTLOG_CANT_START: Final[ThirtyTwoBits] = cached_hresult(0xD000018F, 'HRESULT_FROM_NT(STATUS_EVENTLOG_CANT_START)')
    NTSTATUS__STATUS_TRUST_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD0000190, 'HRESULT_FROM_NT(STATUS_TRUST_FAILURE)')
    NTSTATUS__STATUS_MUTANT_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD0000191, 'HRESULT_FROM_NT(STATUS_MUTANT_LIMIT_EXCEEDED)')
    NTSTATUS__STATUS_NETLOGON_NOT_STARTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000192, 'HRESULT_FROM_NT(STATUS_NETLOGON_NOT_STARTED)')
    NTSTATUS__STATUS_ACCOUNT_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0xD0000193, 'HRESULT_FROM_NT(STATUS_ACCOUNT_EXPIRED)')
    NTSTATUS__STATUS_POSSIBLE_DEADLOCK: Final[ThirtyTwoBits] = cached_hresult(0xD0000194, 'HRESULT_FROM_NT(STATUS_POSSIBLE_DEADLOCK)')
    NTSTATUS__STATUS_NETWORK_CREDENTIAL_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0xD0000195, 'HRESULT_FROM_NT(STATUS_NETWORK_CREDENTIAL_CONFLICT)')
    NTSTATUS__STATUS_REMOTE_SESSION_LIMIT: Final[ThirtyTwoBits] = cached_hresult(0xD0000196, 'HRESULT_FROM_NT(STATUS_REMOTE_SESSION_LIMIT)')
    NTSTATUS__STATUS_EVENTLOG_FILE_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0xD0000197, 'HRESULT_FROM_NT(STATUS_EVENTLOG_FILE_CHANGED)')
    NTSTATUS__STATUS_NOLOGON_INTERDOMAIN_TRUST_ACCOUNT: Final[ThirtyTwoBits] = cached_hresult(0xD0000198, 'HRESULT_FROM_NT(STATUS_NOLOGON_INTERDOMAIN_TRUST_ACCOUNT)')
    NTSTATUS__STATUS_NOLOGON_WORKSTATION_TRUST_ACCOUNT: Final[ThirtyTwoBits] = cached_hresult(0xD0000199, 'HRESULT_FROM_NT(STATUS_NOLOGON_WORKSTATION_TRUST_ACCOUNT)')
    NTSTATUS__STATUS_NOLOGON_SERVER_TRUST_ACCOUNT: Final[ThirtyTwoBits] = cached_hresult(0xD000019A, 'HRESULT_FROM_NT(STATUS_NOLOGON_SERVER_TRUST_ACCOUNT)')
    NTSTATUS__STATUS_DOMAIN_TRUST_INCONSISTENT: Final[ThirtyTwoBits] = cached_hresult(0xD000019B, 'HRESULT_FROM_NT(STATUS_DOMAIN_TRUST_INCONSISTENT)')
    NTSTATUS__STATUS_FS_DRIVER_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0xD000019C, 'HRESULT_FROM_NT(STATUS_FS_DRIVER_REQUIRED)')
    NTSTATUS__STATUS_IMAGE_ALREADY_LOADED_AS_DLL: Final[ThirtyTwoBits] = cached_hresult(0xD000019D, 'HRESULT_FROM_NT(STATUS_IMAGE_ALREADY_LOADED_AS_DLL)')
    NTSTATUS__STATUS_INCOMPATIBLE_WITH_GLOBAL_SHORT_NAME_REGISTRY_SETTING: Final[ThirtyTwoBits] = cached_hresult(0xD000019E, 'HRESULT_FROM_NT(STATUS_INCOMPATIBLE_WITH_GLOBAL_SHORT_NAME_REGISTRY_SETTING)')
    NTSTATUS__STATUS_SHORT_NAMES_NOT_ENABLED_ON_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0xD000019F, 'HRESULT_FROM_NT(STATUS_SHORT_NAMES_NOT_ENABLED_ON_VOLUME)')
    NTSTATUS__STATUS_SECURITY_STREAM_IS_INCONSISTENT: Final[ThirtyTwoBits] = cached_hresult(0xD00001A0, 'HRESULT_FROM_NT(STATUS_SECURITY_STREAM_IS_INCONSISTENT)')
    NTSTATUS__STATUS_INVALID_LOCK_RANGE: Final[ThirtyTwoBits] = cached_hresult(0xD00001A1, 'HRESULT_FROM_NT(STATUS_INVALID_LOCK_RANGE)')
    NTSTATUS__STATUS_INVALID_ACE_CONDITION: Final[ThirtyTwoBits] = cached_hresult(0xD00001A2, 'HRESULT_FROM_NT(STATUS_INVALID_ACE_CONDITION)')
    NTSTATUS__STATUS_IMAGE_SUBSYSTEM_NOT_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0xD00001A3, 'HRESULT_FROM_NT(STATUS_IMAGE_SUBSYSTEM_NOT_PRESENT)')
    NTSTATUS__STATUS_NOTIFICATION_GUID_ALREADY_DEFINED: Final[ThirtyTwoBits] = cached_hresult(0xD00001A4, 'HRESULT_FROM_NT(STATUS_NOTIFICATION_GUID_ALREADY_DEFINED)')
    NTSTATUS__STATUS_INVALID_EXCEPTION_HANDLER: Final[ThirtyTwoBits] = cached_hresult(0xD00001A5, 'HRESULT_FROM_NT(STATUS_INVALID_EXCEPTION_HANDLER)')
    NTSTATUS__STATUS_DUPLICATE_PRIVILEGES: Final[ThirtyTwoBits] = cached_hresult(0xD00001A6, 'HRESULT_FROM_NT(STATUS_DUPLICATE_PRIVILEGES)')
    NTSTATUS__STATUS_NOT_ALLOWED_ON_SYSTEM_FILE: Final[ThirtyTwoBits] = cached_hresult(0xD00001A7, 'HRESULT_FROM_NT(STATUS_NOT_ALLOWED_ON_SYSTEM_FILE)')
    NTSTATUS__STATUS_REPAIR_NEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD00001A8, 'HRESULT_FROM_NT(STATUS_REPAIR_NEEDED)')
    NTSTATUS__STATUS_QUOTA_NOT_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0xD00001A9, 'HRESULT_FROM_NT(STATUS_QUOTA_NOT_ENABLED)')
    NTSTATUS__STATUS_NO_APPLICATION_PACKAGE: Final[ThirtyTwoBits] = cached_hresult(0xD00001AA, 'HRESULT_FROM_NT(STATUS_NO_APPLICATION_PACKAGE)')
    NTSTATUS__STATUS_FILE_METADATA_OPTIMIZATION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD00001AB, 'HRESULT_FROM_NT(STATUS_FILE_METADATA_OPTIMIZATION_IN_PROGRESS)')
    NTSTATUS__STATUS_NOT_SAME_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0xD00001AC, 'HRESULT_FROM_NT(STATUS_NOT_SAME_OBJECT)')
    NTSTATUS__STATUS_FATAL_MEMORY_EXHAUSTION: Final[ThirtyTwoBits] = cached_hresult(0xD00001AD, 'HRESULT_FROM_NT(STATUS_FATAL_MEMORY_EXHAUSTION)')
    NTSTATUS__STATUS_ERROR_PROCESS_NOT_IN_JOB: Final[ThirtyTwoBits] = cached_hresult(0xD00001AE, 'HRESULT_FROM_NT(STATUS_ERROR_PROCESS_NOT_IN_JOB)')
    NTSTATUS__STATUS_CPU_SET_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD00001AF, 'HRESULT_FROM_NT(STATUS_CPU_SET_INVALID)')
    NTSTATUS__STATUS_IO_DEVICE_INVALID_DATA: Final[ThirtyTwoBits] = cached_hresult(0xD00001B0, 'HRESULT_FROM_NT(STATUS_IO_DEVICE_INVALID_DATA)')
    NTSTATUS__STATUS_IO_UNALIGNED_WRITE: Final[ThirtyTwoBits] = cached_hresult(0xD00001B1, 'HRESULT_FROM_NT(STATUS_IO_UNALIGNED_WRITE)')
    NTSTATUS__STATUS_CONTROL_STACK_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0xD00001B2, 'HRESULT_FROM_NT(STATUS_CONTROL_STACK_VIOLATION)')
    NTSTATUS__STATUS_WEAK_WHFBKEY_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0xD00001B3, 'HRESULT_FROM_NT(STATUS_WEAK_WHFBKEY_BLOCKED)')
    NTSTATUS__STATUS_SERVER_TRANSPORT_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0xD00001B4, 'HRESULT_FROM_NT(STATUS_SERVER_TRANSPORT_CONFLICT)')
    NTSTATUS__STATUS_CERTIFICATE_VALIDATION_PREFERENCE_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0xD00001B5, 'HRESULT_FROM_NT(STATUS_CERTIFICATE_VALIDATION_PREFERENCE_CONFLICT)')
    NTSTATUS__STATUS_NETWORK_OPEN_RESTRICTION: Final[ThirtyTwoBits] = cached_hresult(0xD0000201, 'HRESULT_FROM_NT(STATUS_NETWORK_OPEN_RESTRICTION)')
    NTSTATUS__STATUS_NO_USER_SESSION_KEY: Final[ThirtyTwoBits] = cached_hresult(0xD0000202, 'HRESULT_FROM_NT(STATUS_NO_USER_SESSION_KEY)')
    NTSTATUS__STATUS_USER_SESSION_DELETED: Final[ThirtyTwoBits] = cached_hresult(0xD0000203, 'HRESULT_FROM_NT(STATUS_USER_SESSION_DELETED)')
    NTSTATUS__STATUS_RESOURCE_LANG_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000204, 'HRESULT_FROM_NT(STATUS_RESOURCE_LANG_NOT_FOUND)')
    NTSTATUS__STATUS_INSUFF_SERVER_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0xD0000205, 'HRESULT_FROM_NT(STATUS_INSUFF_SERVER_RESOURCES)')
    NTSTATUS__STATUS_INVALID_BUFFER_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD0000206, 'HRESULT_FROM_NT(STATUS_INVALID_BUFFER_SIZE)')
    NTSTATUS__STATUS_INVALID_ADDRESS_COMPONENT: Final[ThirtyTwoBits] = cached_hresult(0xD0000207, 'HRESULT_FROM_NT(STATUS_INVALID_ADDRESS_COMPONENT)')
    NTSTATUS__STATUS_INVALID_ADDRESS_WILDCARD: Final[ThirtyTwoBits] = cached_hresult(0xD0000208, 'HRESULT_FROM_NT(STATUS_INVALID_ADDRESS_WILDCARD)')
    NTSTATUS__STATUS_TOO_MANY_ADDRESSES: Final[ThirtyTwoBits] = cached_hresult(0xD0000209, 'HRESULT_FROM_NT(STATUS_TOO_MANY_ADDRESSES)')
    NTSTATUS__STATUS_ADDRESS_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD000020A, 'HRESULT_FROM_NT(STATUS_ADDRESS_ALREADY_EXISTS)')
    NTSTATUS__STATUS_ADDRESS_CLOSED: Final[ThirtyTwoBits] = cached_hresult(0xD000020B, 'HRESULT_FROM_NT(STATUS_ADDRESS_CLOSED)')
    NTSTATUS__STATUS_CONNECTION_DISCONNECTED: Final[ThirtyTwoBits] = cached_hresult(0xD000020C, 'HRESULT_FROM_NT(STATUS_CONNECTION_DISCONNECTED)')
    NTSTATUS__STATUS_CONNECTION_RESET: Final[ThirtyTwoBits] = cached_hresult(0xD000020D, 'HRESULT_FROM_NT(STATUS_CONNECTION_RESET)')
    NTSTATUS__STATUS_TOO_MANY_NODES: Final[ThirtyTwoBits] = cached_hresult(0xD000020E, 'HRESULT_FROM_NT(STATUS_TOO_MANY_NODES)')
    NTSTATUS__STATUS_TRANSACTION_ABORTED: Final[ThirtyTwoBits] = cached_hresult(0xD000020F, 'HRESULT_FROM_NT(STATUS_TRANSACTION_ABORTED)')
    NTSTATUS__STATUS_TRANSACTION_TIMED_OUT: Final[ThirtyTwoBits] = cached_hresult(0xD0000210, 'HRESULT_FROM_NT(STATUS_TRANSACTION_TIMED_OUT)')
    NTSTATUS__STATUS_TRANSACTION_NO_RELEASE: Final[ThirtyTwoBits] = cached_hresult(0xD0000211, 'HRESULT_FROM_NT(STATUS_TRANSACTION_NO_RELEASE)')
    NTSTATUS__STATUS_TRANSACTION_NO_MATCH: Final[ThirtyTwoBits] = cached_hresult(0xD0000212, 'HRESULT_FROM_NT(STATUS_TRANSACTION_NO_MATCH)')
    NTSTATUS__STATUS_TRANSACTION_RESPONDED: Final[ThirtyTwoBits] = cached_hresult(0xD0000213, 'HRESULT_FROM_NT(STATUS_TRANSACTION_RESPONDED)')
    NTSTATUS__STATUS_TRANSACTION_INVALID_ID: Final[ThirtyTwoBits] = cached_hresult(0xD0000214, 'HRESULT_FROM_NT(STATUS_TRANSACTION_INVALID_ID)')
    NTSTATUS__STATUS_TRANSACTION_INVALID_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xD0000215, 'HRESULT_FROM_NT(STATUS_TRANSACTION_INVALID_TYPE)')
    NTSTATUS__STATUS_NOT_SERVER_SESSION: Final[ThirtyTwoBits] = cached_hresult(0xD0000216, 'HRESULT_FROM_NT(STATUS_NOT_SERVER_SESSION)')
    NTSTATUS__STATUS_NOT_CLIENT_SESSION: Final[ThirtyTwoBits] = cached_hresult(0xD0000217, 'HRESULT_FROM_NT(STATUS_NOT_CLIENT_SESSION)')
    NTSTATUS__STATUS_CANNOT_LOAD_REGISTRY_FILE: Final[ThirtyTwoBits] = cached_hresult(0xD0000218, 'HRESULT_FROM_NT(STATUS_CANNOT_LOAD_REGISTRY_FILE)')
    NTSTATUS__STATUS_DEBUG_ATTACH_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0000219, 'HRESULT_FROM_NT(STATUS_DEBUG_ATTACH_FAILED)')
    NTSTATUS__STATUS_SYSTEM_PROCESS_TERMINATED: Final[ThirtyTwoBits] = cached_hresult(0xD000021A, 'HRESULT_FROM_NT(STATUS_SYSTEM_PROCESS_TERMINATED)')
    NTSTATUS__STATUS_DATA_NOT_ACCEPTED: Final[ThirtyTwoBits] = cached_hresult(0xD000021B, 'HRESULT_FROM_NT(STATUS_DATA_NOT_ACCEPTED)')
    NTSTATUS__STATUS_NO_BROWSER_SERVERS_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD000021C, 'HRESULT_FROM_NT(STATUS_NO_BROWSER_SERVERS_FOUND)')
    NTSTATUS__STATUS_VDM_HARD_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD000021D, 'HRESULT_FROM_NT(STATUS_VDM_HARD_ERROR)')
    NTSTATUS__STATUS_DRIVER_CANCEL_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0xD000021E, 'HRESULT_FROM_NT(STATUS_DRIVER_CANCEL_TIMEOUT)')
    NTSTATUS__STATUS_REPLY_MESSAGE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD000021F, 'HRESULT_FROM_NT(STATUS_REPLY_MESSAGE_MISMATCH)')
    NTSTATUS__STATUS_MAPPED_ALIGNMENT: Final[ThirtyTwoBits] = cached_hresult(0xD0000220, 'HRESULT_FROM_NT(STATUS_MAPPED_ALIGNMENT)')
    NTSTATUS__STATUS_IMAGE_CHECKSUM_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD0000221, 'HRESULT_FROM_NT(STATUS_IMAGE_CHECKSUM_MISMATCH)')
    NTSTATUS__STATUS_LOST_WRITEBEHIND_DATA: Final[ThirtyTwoBits] = cached_hresult(0xD0000222, 'HRESULT_FROM_NT(STATUS_LOST_WRITEBEHIND_DATA)')
    NTSTATUS__STATUS_CLIENT_SERVER_PARAMETERS_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0000223, 'HRESULT_FROM_NT(STATUS_CLIENT_SERVER_PARAMETERS_INVALID)')
    NTSTATUS__STATUS_PASSWORD_MUST_CHANGE: Final[ThirtyTwoBits] = cached_hresult(0xD0000224, 'HRESULT_FROM_NT(STATUS_PASSWORD_MUST_CHANGE)')
    NTSTATUS__STATUS_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000225, 'HRESULT_FROM_NT(STATUS_NOT_FOUND)')
    NTSTATUS__STATUS_NOT_TINY_STREAM: Final[ThirtyTwoBits] = cached_hresult(0xD0000226, 'HRESULT_FROM_NT(STATUS_NOT_TINY_STREAM)')
    NTSTATUS__STATUS_RECOVERY_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD0000227, 'HRESULT_FROM_NT(STATUS_RECOVERY_FAILURE)')
    NTSTATUS__STATUS_STACK_OVERFLOW_READ: Final[ThirtyTwoBits] = cached_hresult(0xD0000228, 'HRESULT_FROM_NT(STATUS_STACK_OVERFLOW_READ)')
    NTSTATUS__STATUS_FAIL_CHECK: Final[ThirtyTwoBits] = cached_hresult(0xD0000229, 'HRESULT_FROM_NT(STATUS_FAIL_CHECK)')
    NTSTATUS__STATUS_DUPLICATE_OBJECTID: Final[ThirtyTwoBits] = cached_hresult(0xD000022A, 'HRESULT_FROM_NT(STATUS_DUPLICATE_OBJECTID)')
    NTSTATUS__STATUS_OBJECTID_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD000022B, 'HRESULT_FROM_NT(STATUS_OBJECTID_EXISTS)')
    NTSTATUS__STATUS_CONVERT_TO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0xD000022C, 'HRESULT_FROM_NT(STATUS_CONVERT_TO_LARGE)')
    NTSTATUS__STATUS_RETRY: Final[ThirtyTwoBits] = cached_hresult(0xD000022D, 'HRESULT_FROM_NT(STATUS_RETRY)')
    NTSTATUS__STATUS_FOUND_OUT_OF_SCOPE: Final[ThirtyTwoBits] = cached_hresult(0xD000022E, 'HRESULT_FROM_NT(STATUS_FOUND_OUT_OF_SCOPE)')
    NTSTATUS__STATUS_ALLOCATE_BUCKET: Final[ThirtyTwoBits] = cached_hresult(0xD000022F, 'HRESULT_FROM_NT(STATUS_ALLOCATE_BUCKET)')
    NTSTATUS__STATUS_PROPSET_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000230, 'HRESULT_FROM_NT(STATUS_PROPSET_NOT_FOUND)')
    NTSTATUS__STATUS_MARSHALL_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0xD0000231, 'HRESULT_FROM_NT(STATUS_MARSHALL_OVERFLOW)')
    NTSTATUS__STATUS_INVALID_VARIANT: Final[ThirtyTwoBits] = cached_hresult(0xD0000232, 'HRESULT_FROM_NT(STATUS_INVALID_VARIANT)')
    NTSTATUS__STATUS_DOMAIN_CONTROLLER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000233, 'HRESULT_FROM_NT(STATUS_DOMAIN_CONTROLLER_NOT_FOUND)')
    NTSTATUS__STATUS_ACCOUNT_LOCKED_OUT: Final[ThirtyTwoBits] = cached_hresult(0xD0000234, 'HRESULT_FROM_NT(STATUS_ACCOUNT_LOCKED_OUT)')
    NTSTATUS__STATUS_HANDLE_NOT_CLOSABLE: Final[ThirtyTwoBits] = cached_hresult(0xD0000235, 'HRESULT_FROM_NT(STATUS_HANDLE_NOT_CLOSABLE)')
    NTSTATUS__STATUS_CONNECTION_REFUSED: Final[ThirtyTwoBits] = cached_hresult(0xD0000236, 'HRESULT_FROM_NT(STATUS_CONNECTION_REFUSED)')
    NTSTATUS__STATUS_GRACEFUL_DISCONNECT: Final[ThirtyTwoBits] = cached_hresult(0xD0000237, 'HRESULT_FROM_NT(STATUS_GRACEFUL_DISCONNECT)')
    NTSTATUS__STATUS_ADDRESS_ALREADY_ASSOCIATED: Final[ThirtyTwoBits] = cached_hresult(0xD0000238, 'HRESULT_FROM_NT(STATUS_ADDRESS_ALREADY_ASSOCIATED)')
    NTSTATUS__STATUS_ADDRESS_NOT_ASSOCIATED: Final[ThirtyTwoBits] = cached_hresult(0xD0000239, 'HRESULT_FROM_NT(STATUS_ADDRESS_NOT_ASSOCIATED)')
    NTSTATUS__STATUS_CONNECTION_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD000023A, 'HRESULT_FROM_NT(STATUS_CONNECTION_INVALID)')
    NTSTATUS__STATUS_CONNECTION_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0xD000023B, 'HRESULT_FROM_NT(STATUS_CONNECTION_ACTIVE)')
    NTSTATUS__STATUS_NETWORK_UNREACHABLE: Final[ThirtyTwoBits] = cached_hresult(0xD000023C, 'HRESULT_FROM_NT(STATUS_NETWORK_UNREACHABLE)')
    NTSTATUS__STATUS_HOST_UNREACHABLE: Final[ThirtyTwoBits] = cached_hresult(0xD000023D, 'HRESULT_FROM_NT(STATUS_HOST_UNREACHABLE)')
    NTSTATUS__STATUS_PROTOCOL_UNREACHABLE: Final[ThirtyTwoBits] = cached_hresult(0xD000023E, 'HRESULT_FROM_NT(STATUS_PROTOCOL_UNREACHABLE)')
    NTSTATUS__STATUS_PORT_UNREACHABLE: Final[ThirtyTwoBits] = cached_hresult(0xD000023F, 'HRESULT_FROM_NT(STATUS_PORT_UNREACHABLE)')
    NTSTATUS__STATUS_REQUEST_ABORTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000240, 'HRESULT_FROM_NT(STATUS_REQUEST_ABORTED)')
    NTSTATUS__STATUS_CONNECTION_ABORTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000241, 'HRESULT_FROM_NT(STATUS_CONNECTION_ABORTED)')
    NTSTATUS__STATUS_BAD_COMPRESSION_BUFFER: Final[ThirtyTwoBits] = cached_hresult(0xD0000242, 'HRESULT_FROM_NT(STATUS_BAD_COMPRESSION_BUFFER)')
    NTSTATUS__STATUS_USER_MAPPED_FILE: Final[ThirtyTwoBits] = cached_hresult(0xD0000243, 'HRESULT_FROM_NT(STATUS_USER_MAPPED_FILE)')
    NTSTATUS__STATUS_AUDIT_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0000244, 'HRESULT_FROM_NT(STATUS_AUDIT_FAILED)')
    NTSTATUS__STATUS_TIMER_RESOLUTION_NOT_SET: Final[ThirtyTwoBits] = cached_hresult(0xD0000245, 'HRESULT_FROM_NT(STATUS_TIMER_RESOLUTION_NOT_SET)')
    NTSTATUS__STATUS_CONNECTION_COUNT_LIMIT: Final[ThirtyTwoBits] = cached_hresult(0xD0000246, 'HRESULT_FROM_NT(STATUS_CONNECTION_COUNT_LIMIT)')
    NTSTATUS__STATUS_LOGIN_TIME_RESTRICTION: Final[ThirtyTwoBits] = cached_hresult(0xD0000247, 'HRESULT_FROM_NT(STATUS_LOGIN_TIME_RESTRICTION)')
    NTSTATUS__STATUS_LOGIN_WKSTA_RESTRICTION: Final[ThirtyTwoBits] = cached_hresult(0xD0000248, 'HRESULT_FROM_NT(STATUS_LOGIN_WKSTA_RESTRICTION)')
    NTSTATUS__STATUS_IMAGE_MP_UP_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD0000249, 'HRESULT_FROM_NT(STATUS_IMAGE_MP_UP_MISMATCH)')
    NTSTATUS__STATUS_INSUFFICIENT_LOGON_INFO: Final[ThirtyTwoBits] = cached_hresult(0xD0000250, 'HRESULT_FROM_NT(STATUS_INSUFFICIENT_LOGON_INFO)')
    NTSTATUS__STATUS_BAD_DLL_ENTRYPOINT: Final[ThirtyTwoBits] = cached_hresult(0xD0000251, 'HRESULT_FROM_NT(STATUS_BAD_DLL_ENTRYPOINT)')
    NTSTATUS__STATUS_BAD_SERVICE_ENTRYPOINT: Final[ThirtyTwoBits] = cached_hresult(0xD0000252, 'HRESULT_FROM_NT(STATUS_BAD_SERVICE_ENTRYPOINT)')
    NTSTATUS__STATUS_LPC_REPLY_LOST: Final[ThirtyTwoBits] = cached_hresult(0xD0000253, 'HRESULT_FROM_NT(STATUS_LPC_REPLY_LOST)')
    NTSTATUS__STATUS_IP_ADDRESS_CONFLICT1: Final[ThirtyTwoBits] = cached_hresult(0xD0000254, 'HRESULT_FROM_NT(STATUS_IP_ADDRESS_CONFLICT1)')
    NTSTATUS__STATUS_IP_ADDRESS_CONFLICT2: Final[ThirtyTwoBits] = cached_hresult(0xD0000255, 'HRESULT_FROM_NT(STATUS_IP_ADDRESS_CONFLICT2)')
    NTSTATUS__STATUS_REGISTRY_QUOTA_LIMIT: Final[ThirtyTwoBits] = cached_hresult(0xD0000256, 'HRESULT_FROM_NT(STATUS_REGISTRY_QUOTA_LIMIT)')
    NTSTATUS__STATUS_PATH_NOT_COVERED: Final[ThirtyTwoBits] = cached_hresult(0xD0000257, 'HRESULT_FROM_NT(STATUS_PATH_NOT_COVERED)')
    NTSTATUS__STATUS_NO_CALLBACK_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0xD0000258, 'HRESULT_FROM_NT(STATUS_NO_CALLBACK_ACTIVE)')
    NTSTATUS__STATUS_LICENSE_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD0000259, 'HRESULT_FROM_NT(STATUS_LICENSE_QUOTA_EXCEEDED)')
    NTSTATUS__STATUS_PWD_TOO_SHORT: Final[ThirtyTwoBits] = cached_hresult(0xD000025A, 'HRESULT_FROM_NT(STATUS_PWD_TOO_SHORT)')
    NTSTATUS__STATUS_PWD_TOO_RECENT: Final[ThirtyTwoBits] = cached_hresult(0xD000025B, 'HRESULT_FROM_NT(STATUS_PWD_TOO_RECENT)')
    NTSTATUS__STATUS_PWD_HISTORY_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0xD000025C, 'HRESULT_FROM_NT(STATUS_PWD_HISTORY_CONFLICT)')
    NTSTATUS__STATUS_PLUGPLAY_NO_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0xD000025E, 'HRESULT_FROM_NT(STATUS_PLUGPLAY_NO_DEVICE)')
    NTSTATUS__STATUS_UNSUPPORTED_COMPRESSION: Final[ThirtyTwoBits] = cached_hresult(0xD000025F, 'HRESULT_FROM_NT(STATUS_UNSUPPORTED_COMPRESSION)')
    NTSTATUS__STATUS_INVALID_HW_PROFILE: Final[ThirtyTwoBits] = cached_hresult(0xD0000260, 'HRESULT_FROM_NT(STATUS_INVALID_HW_PROFILE)')
    NTSTATUS__STATUS_INVALID_PLUGPLAY_DEVICE_PATH: Final[ThirtyTwoBits] = cached_hresult(0xD0000261, 'HRESULT_FROM_NT(STATUS_INVALID_PLUGPLAY_DEVICE_PATH)')
    NTSTATUS__STATUS_DRIVER_ORDINAL_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000262, 'HRESULT_FROM_NT(STATUS_DRIVER_ORDINAL_NOT_FOUND)')
    NTSTATUS__STATUS_DRIVER_ENTRYPOINT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000263, 'HRESULT_FROM_NT(STATUS_DRIVER_ENTRYPOINT_NOT_FOUND)')
    NTSTATUS__STATUS_RESOURCE_NOT_OWNED: Final[ThirtyTwoBits] = cached_hresult(0xD0000264, 'HRESULT_FROM_NT(STATUS_RESOURCE_NOT_OWNED)')
    NTSTATUS__STATUS_TOO_MANY_LINKS: Final[ThirtyTwoBits] = cached_hresult(0xD0000265, 'HRESULT_FROM_NT(STATUS_TOO_MANY_LINKS)')
    NTSTATUS__STATUS_QUOTA_LIST_INCONSISTENT: Final[ThirtyTwoBits] = cached_hresult(0xD0000266, 'HRESULT_FROM_NT(STATUS_QUOTA_LIST_INCONSISTENT)')
    NTSTATUS__STATUS_FILE_IS_OFFLINE: Final[ThirtyTwoBits] = cached_hresult(0xD0000267, 'HRESULT_FROM_NT(STATUS_FILE_IS_OFFLINE)')
    NTSTATUS__STATUS_EVALUATION_EXPIRATION: Final[ThirtyTwoBits] = cached_hresult(0xD0000268, 'HRESULT_FROM_NT(STATUS_EVALUATION_EXPIRATION)')
    NTSTATUS__STATUS_ILLEGAL_DLL_RELOCATION: Final[ThirtyTwoBits] = cached_hresult(0xD0000269, 'HRESULT_FROM_NT(STATUS_ILLEGAL_DLL_RELOCATION)')
    NTSTATUS__STATUS_LICENSE_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0xD000026A, 'HRESULT_FROM_NT(STATUS_LICENSE_VIOLATION)')
    NTSTATUS__STATUS_DLL_INIT_FAILED_LOGOFF: Final[ThirtyTwoBits] = cached_hresult(0xD000026B, 'HRESULT_FROM_NT(STATUS_DLL_INIT_FAILED_LOGOFF)')
    NTSTATUS__STATUS_DRIVER_UNABLE_TO_LOAD: Final[ThirtyTwoBits] = cached_hresult(0xD000026C, 'HRESULT_FROM_NT(STATUS_DRIVER_UNABLE_TO_LOAD)')
    NTSTATUS__STATUS_DFS_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD000026D, 'HRESULT_FROM_NT(STATUS_DFS_UNAVAILABLE)')
    NTSTATUS__STATUS_VOLUME_DISMOUNTED: Final[ThirtyTwoBits] = cached_hresult(0xD000026E, 'HRESULT_FROM_NT(STATUS_VOLUME_DISMOUNTED)')
    NTSTATUS__STATUS_WX86_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD000026F, 'HRESULT_FROM_NT(STATUS_WX86_INTERNAL_ERROR)')
    NTSTATUS__STATUS_WX86_FLOAT_STACK_CHECK: Final[ThirtyTwoBits] = cached_hresult(0xD0000270, 'HRESULT_FROM_NT(STATUS_WX86_FLOAT_STACK_CHECK)')
    NTSTATUS__STATUS_VALIDATE_CONTINUE: Final[ThirtyTwoBits] = cached_hresult(0xD0000271, 'HRESULT_FROM_NT(STATUS_VALIDATE_CONTINUE)')
    NTSTATUS__STATUS_NO_MATCH: Final[ThirtyTwoBits] = cached_hresult(0xD0000272, 'HRESULT_FROM_NT(STATUS_NO_MATCH)')
    NTSTATUS__STATUS_NO_MORE_MATCHES: Final[ThirtyTwoBits] = cached_hresult(0xD0000273, 'HRESULT_FROM_NT(STATUS_NO_MORE_MATCHES)')
    NTSTATUS__STATUS_NOT_A_REPARSE_POINT: Final[ThirtyTwoBits] = cached_hresult(0xD0000275, 'HRESULT_FROM_NT(STATUS_NOT_A_REPARSE_POINT)')
    NTSTATUS__STATUS_IO_REPARSE_TAG_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0000276, 'HRESULT_FROM_NT(STATUS_IO_REPARSE_TAG_INVALID)')
    NTSTATUS__STATUS_IO_REPARSE_TAG_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD0000277, 'HRESULT_FROM_NT(STATUS_IO_REPARSE_TAG_MISMATCH)')
    NTSTATUS__STATUS_IO_REPARSE_DATA_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0000278, 'HRESULT_FROM_NT(STATUS_IO_REPARSE_DATA_INVALID)')
    NTSTATUS__STATUS_IO_REPARSE_TAG_NOT_HANDLED: Final[ThirtyTwoBits] = cached_hresult(0xD0000279, 'HRESULT_FROM_NT(STATUS_IO_REPARSE_TAG_NOT_HANDLED)')
    NTSTATUS__STATUS_PWD_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0xD000027A, 'HRESULT_FROM_NT(STATUS_PWD_TOO_LONG)')
    NTSTATUS__STATUS_STOWED_EXCEPTION: Final[ThirtyTwoBits] = cached_hresult(0xD000027B, 'HRESULT_FROM_NT(STATUS_STOWED_EXCEPTION)')
    NTSTATUS__STATUS_CONTEXT_STOWED_EXCEPTION: Final[ThirtyTwoBits] = cached_hresult(0xD000027C, 'HRESULT_FROM_NT(STATUS_CONTEXT_STOWED_EXCEPTION)')
    NTSTATUS__STATUS_REPARSE_POINT_NOT_RESOLVED: Final[ThirtyTwoBits] = cached_hresult(0xD0000280, 'HRESULT_FROM_NT(STATUS_REPARSE_POINT_NOT_RESOLVED)')
    NTSTATUS__STATUS_DIRECTORY_IS_A_REPARSE_POINT: Final[ThirtyTwoBits] = cached_hresult(0xD0000281, 'HRESULT_FROM_NT(STATUS_DIRECTORY_IS_A_REPARSE_POINT)')
    NTSTATUS__STATUS_RANGE_LIST_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0xD0000282, 'HRESULT_FROM_NT(STATUS_RANGE_LIST_CONFLICT)')
    NTSTATUS__STATUS_SOURCE_ELEMENT_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0xD0000283, 'HRESULT_FROM_NT(STATUS_SOURCE_ELEMENT_EMPTY)')
    NTSTATUS__STATUS_DESTINATION_ELEMENT_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD0000284, 'HRESULT_FROM_NT(STATUS_DESTINATION_ELEMENT_FULL)')
    NTSTATUS__STATUS_ILLEGAL_ELEMENT_ADDRESS: Final[ThirtyTwoBits] = cached_hresult(0xD0000285, 'HRESULT_FROM_NT(STATUS_ILLEGAL_ELEMENT_ADDRESS)')
    NTSTATUS__STATUS_MAGAZINE_NOT_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0xD0000286, 'HRESULT_FROM_NT(STATUS_MAGAZINE_NOT_PRESENT)')
    NTSTATUS__STATUS_REINITIALIZATION_NEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD0000287, 'HRESULT_FROM_NT(STATUS_REINITIALIZATION_NEEDED)')
    NTSTATUS__STATUS_ENCRYPTION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD000028A, 'HRESULT_FROM_NT(STATUS_ENCRYPTION_FAILED)')
    NTSTATUS__STATUS_DECRYPTION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD000028B, 'HRESULT_FROM_NT(STATUS_DECRYPTION_FAILED)')
    NTSTATUS__STATUS_RANGE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD000028C, 'HRESULT_FROM_NT(STATUS_RANGE_NOT_FOUND)')
    NTSTATUS__STATUS_NO_RECOVERY_POLICY: Final[ThirtyTwoBits] = cached_hresult(0xD000028D, 'HRESULT_FROM_NT(STATUS_NO_RECOVERY_POLICY)')
    NTSTATUS__STATUS_NO_EFS: Final[ThirtyTwoBits] = cached_hresult(0xD000028E, 'HRESULT_FROM_NT(STATUS_NO_EFS)')
    NTSTATUS__STATUS_WRONG_EFS: Final[ThirtyTwoBits] = cached_hresult(0xD000028F, 'HRESULT_FROM_NT(STATUS_WRONG_EFS)')
    NTSTATUS__STATUS_NO_USER_KEYS: Final[ThirtyTwoBits] = cached_hresult(0xD0000290, 'HRESULT_FROM_NT(STATUS_NO_USER_KEYS)')
    NTSTATUS__STATUS_FILE_NOT_ENCRYPTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000291, 'HRESULT_FROM_NT(STATUS_FILE_NOT_ENCRYPTED)')
    NTSTATUS__STATUS_NOT_EXPORT_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0xD0000292, 'HRESULT_FROM_NT(STATUS_NOT_EXPORT_FORMAT)')
    NTSTATUS__STATUS_FILE_ENCRYPTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000293, 'HRESULT_FROM_NT(STATUS_FILE_ENCRYPTED)')
    NTSTATUS__STATUS_WMI_GUID_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000295, 'HRESULT_FROM_NT(STATUS_WMI_GUID_NOT_FOUND)')
    NTSTATUS__STATUS_WMI_INSTANCE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000296, 'HRESULT_FROM_NT(STATUS_WMI_INSTANCE_NOT_FOUND)')
    NTSTATUS__STATUS_WMI_ITEMID_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000297, 'HRESULT_FROM_NT(STATUS_WMI_ITEMID_NOT_FOUND)')
    NTSTATUS__STATUS_WMI_TRY_AGAIN: Final[ThirtyTwoBits] = cached_hresult(0xD0000298, 'HRESULT_FROM_NT(STATUS_WMI_TRY_AGAIN)')
    NTSTATUS__STATUS_SHARED_POLICY: Final[ThirtyTwoBits] = cached_hresult(0xD0000299, 'HRESULT_FROM_NT(STATUS_SHARED_POLICY)')
    NTSTATUS__STATUS_POLICY_OBJECT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD000029A, 'HRESULT_FROM_NT(STATUS_POLICY_OBJECT_NOT_FOUND)')
    NTSTATUS__STATUS_POLICY_ONLY_IN_DS: Final[ThirtyTwoBits] = cached_hresult(0xD000029B, 'HRESULT_FROM_NT(STATUS_POLICY_ONLY_IN_DS)')
    NTSTATUS__STATUS_VOLUME_NOT_UPGRADED: Final[ThirtyTwoBits] = cached_hresult(0xD000029C, 'HRESULT_FROM_NT(STATUS_VOLUME_NOT_UPGRADED)')
    NTSTATUS__STATUS_REMOTE_STORAGE_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0xD000029D, 'HRESULT_FROM_NT(STATUS_REMOTE_STORAGE_NOT_ACTIVE)')
    NTSTATUS__STATUS_REMOTE_STORAGE_MEDIA_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD000029E, 'HRESULT_FROM_NT(STATUS_REMOTE_STORAGE_MEDIA_ERROR)')
    NTSTATUS__STATUS_NO_TRACKING_SERVICE: Final[ThirtyTwoBits] = cached_hresult(0xD000029F, 'HRESULT_FROM_NT(STATUS_NO_TRACKING_SERVICE)')
    NTSTATUS__STATUS_SERVER_SID_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD00002A0, 'HRESULT_FROM_NT(STATUS_SERVER_SID_MISMATCH)')
    NTSTATUS__STATUS_DS_NO_ATTRIBUTE_OR_VALUE: Final[ThirtyTwoBits] = cached_hresult(0xD00002A1, 'HRESULT_FROM_NT(STATUS_DS_NO_ATTRIBUTE_OR_VALUE)')
    NTSTATUS__STATUS_DS_INVALID_ATTRIBUTE_SYNTAX: Final[ThirtyTwoBits] = cached_hresult(0xD00002A2, 'HRESULT_FROM_NT(STATUS_DS_INVALID_ATTRIBUTE_SYNTAX)')
    NTSTATUS__STATUS_DS_ATTRIBUTE_TYPE_UNDEFINED: Final[ThirtyTwoBits] = cached_hresult(0xD00002A3, 'HRESULT_FROM_NT(STATUS_DS_ATTRIBUTE_TYPE_UNDEFINED)')
    NTSTATUS__STATUS_DS_ATTRIBUTE_OR_VALUE_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD00002A4, 'HRESULT_FROM_NT(STATUS_DS_ATTRIBUTE_OR_VALUE_EXISTS)')
    NTSTATUS__STATUS_DS_BUSY: Final[ThirtyTwoBits] = cached_hresult(0xD00002A5, 'HRESULT_FROM_NT(STATUS_DS_BUSY)')
    NTSTATUS__STATUS_DS_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD00002A6, 'HRESULT_FROM_NT(STATUS_DS_UNAVAILABLE)')
    NTSTATUS__STATUS_DS_NO_RIDS_ALLOCATED: Final[ThirtyTwoBits] = cached_hresult(0xD00002A7, 'HRESULT_FROM_NT(STATUS_DS_NO_RIDS_ALLOCATED)')
    NTSTATUS__STATUS_DS_NO_MORE_RIDS: Final[ThirtyTwoBits] = cached_hresult(0xD00002A8, 'HRESULT_FROM_NT(STATUS_DS_NO_MORE_RIDS)')
    NTSTATUS__STATUS_DS_INCORRECT_ROLE_OWNER: Final[ThirtyTwoBits] = cached_hresult(0xD00002A9, 'HRESULT_FROM_NT(STATUS_DS_INCORRECT_ROLE_OWNER)')
    NTSTATUS__STATUS_DS_RIDMGR_INIT_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD00002AA, 'HRESULT_FROM_NT(STATUS_DS_RIDMGR_INIT_ERROR)')
    NTSTATUS__STATUS_DS_OBJ_CLASS_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0xD00002AB, 'HRESULT_FROM_NT(STATUS_DS_OBJ_CLASS_VIOLATION)')
    NTSTATUS__STATUS_DS_CANT_ON_NON_LEAF: Final[ThirtyTwoBits] = cached_hresult(0xD00002AC, 'HRESULT_FROM_NT(STATUS_DS_CANT_ON_NON_LEAF)')
    NTSTATUS__STATUS_DS_CANT_ON_RDN: Final[ThirtyTwoBits] = cached_hresult(0xD00002AD, 'HRESULT_FROM_NT(STATUS_DS_CANT_ON_RDN)')
    NTSTATUS__STATUS_DS_CANT_MOD_OBJ_CLASS: Final[ThirtyTwoBits] = cached_hresult(0xD00002AE, 'HRESULT_FROM_NT(STATUS_DS_CANT_MOD_OBJ_CLASS)')
    NTSTATUS__STATUS_DS_CROSS_DOM_MOVE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD00002AF, 'HRESULT_FROM_NT(STATUS_DS_CROSS_DOM_MOVE_FAILED)')
    NTSTATUS__STATUS_DS_GC_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD00002B0, 'HRESULT_FROM_NT(STATUS_DS_GC_NOT_AVAILABLE)')
    NTSTATUS__STATUS_DIRECTORY_SERVICE_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0xD00002B1, 'HRESULT_FROM_NT(STATUS_DIRECTORY_SERVICE_REQUIRED)')
    NTSTATUS__STATUS_REPARSE_ATTRIBUTE_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0xD00002B2, 'HRESULT_FROM_NT(STATUS_REPARSE_ATTRIBUTE_CONFLICT)')
    NTSTATUS__STATUS_CANT_ENABLE_DENY_ONLY: Final[ThirtyTwoBits] = cached_hresult(0xD00002B3, 'HRESULT_FROM_NT(STATUS_CANT_ENABLE_DENY_ONLY)')
    NTSTATUS__STATUS_FLOAT_MULTIPLE_FAULTS: Final[ThirtyTwoBits] = cached_hresult(0xD00002B4, 'HRESULT_FROM_NT(STATUS_FLOAT_MULTIPLE_FAULTS)')
    NTSTATUS__STATUS_FLOAT_MULTIPLE_TRAPS: Final[ThirtyTwoBits] = cached_hresult(0xD00002B5, 'HRESULT_FROM_NT(STATUS_FLOAT_MULTIPLE_TRAPS)')
    NTSTATUS__STATUS_DEVICE_REMOVED: Final[ThirtyTwoBits] = cached_hresult(0xD00002B6, 'HRESULT_FROM_NT(STATUS_DEVICE_REMOVED)')
    NTSTATUS__STATUS_JOURNAL_DELETE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD00002B7, 'HRESULT_FROM_NT(STATUS_JOURNAL_DELETE_IN_PROGRESS)')
    NTSTATUS__STATUS_JOURNAL_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0xD00002B8, 'HRESULT_FROM_NT(STATUS_JOURNAL_NOT_ACTIVE)')
    NTSTATUS__STATUS_NOINTERFACE: Final[ThirtyTwoBits] = cached_hresult(0xD00002B9, 'HRESULT_FROM_NT(STATUS_NOINTERFACE)')
    NTSTATUS__STATUS_DS_RIDMGR_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD00002BA, 'HRESULT_FROM_NT(STATUS_DS_RIDMGR_DISABLED)')
    NTSTATUS__STATUS_DS_ADMIN_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD00002C1, 'HRESULT_FROM_NT(STATUS_DS_ADMIN_LIMIT_EXCEEDED)')
    NTSTATUS__STATUS_DRIVER_FAILED_SLEEP: Final[ThirtyTwoBits] = cached_hresult(0xD00002C2, 'HRESULT_FROM_NT(STATUS_DRIVER_FAILED_SLEEP)')
    NTSTATUS__STATUS_MUTUAL_AUTHENTICATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD00002C3, 'HRESULT_FROM_NT(STATUS_MUTUAL_AUTHENTICATION_FAILED)')
    NTSTATUS__STATUS_CORRUPT_SYSTEM_FILE: Final[ThirtyTwoBits] = cached_hresult(0xD00002C4, 'HRESULT_FROM_NT(STATUS_CORRUPT_SYSTEM_FILE)')
    NTSTATUS__STATUS_DATATYPE_MISALIGNMENT_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD00002C5, 'HRESULT_FROM_NT(STATUS_DATATYPE_MISALIGNMENT_ERROR)')
    NTSTATUS__STATUS_WMI_READ_ONLY: Final[ThirtyTwoBits] = cached_hresult(0xD00002C6, 'HRESULT_FROM_NT(STATUS_WMI_READ_ONLY)')
    NTSTATUS__STATUS_WMI_SET_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD00002C7, 'HRESULT_FROM_NT(STATUS_WMI_SET_FAILURE)')
    NTSTATUS__STATUS_COMMITMENT_MINIMUM: Final[ThirtyTwoBits] = cached_hresult(0xD00002C8, 'HRESULT_FROM_NT(STATUS_COMMITMENT_MINIMUM)')
    NTSTATUS__STATUS_REG_NAT_CONSUMPTION: Final[ThirtyTwoBits] = cached_hresult(0xD00002C9, 'HRESULT_FROM_NT(STATUS_REG_NAT_CONSUMPTION)')
    NTSTATUS__STATUS_TRANSPORT_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD00002CA, 'HRESULT_FROM_NT(STATUS_TRANSPORT_FULL)')
    NTSTATUS__STATUS_DS_SAM_INIT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD00002CB, 'HRESULT_FROM_NT(STATUS_DS_SAM_INIT_FAILURE)')
    NTSTATUS__STATUS_ONLY_IF_CONNECTED: Final[ThirtyTwoBits] = cached_hresult(0xD00002CC, 'HRESULT_FROM_NT(STATUS_ONLY_IF_CONNECTED)')
    NTSTATUS__STATUS_DS_SENSITIVE_GROUP_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0xD00002CD, 'HRESULT_FROM_NT(STATUS_DS_SENSITIVE_GROUP_VIOLATION)')
    NTSTATUS__STATUS_PNP_RESTART_ENUMERATION: Final[ThirtyTwoBits] = cached_hresult(0xD00002CE, 'HRESULT_FROM_NT(STATUS_PNP_RESTART_ENUMERATION)')
    NTSTATUS__STATUS_JOURNAL_ENTRY_DELETED: Final[ThirtyTwoBits] = cached_hresult(0xD00002CF, 'HRESULT_FROM_NT(STATUS_JOURNAL_ENTRY_DELETED)')
    NTSTATUS__STATUS_DS_CANT_MOD_PRIMARYGROUPID: Final[ThirtyTwoBits] = cached_hresult(0xD00002D0, 'HRESULT_FROM_NT(STATUS_DS_CANT_MOD_PRIMARYGROUPID)')
    NTSTATUS__STATUS_SYSTEM_IMAGE_BAD_SIGNATURE: Final[ThirtyTwoBits] = cached_hresult(0xD00002D1, 'HRESULT_FROM_NT(STATUS_SYSTEM_IMAGE_BAD_SIGNATURE)')
    NTSTATUS__STATUS_PNP_REBOOT_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0xD00002D2, 'HRESULT_FROM_NT(STATUS_PNP_REBOOT_REQUIRED)')
    NTSTATUS__STATUS_POWER_STATE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD00002D3, 'HRESULT_FROM_NT(STATUS_POWER_STATE_INVALID)')
    NTSTATUS__STATUS_DS_INVALID_GROUP_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xD00002D4, 'HRESULT_FROM_NT(STATUS_DS_INVALID_GROUP_TYPE)')
    NTSTATUS__STATUS_DS_NO_NEST_GLOBALGROUP_IN_MIXEDDOMAIN: Final[ThirtyTwoBits] = cached_hresult(0xD00002D5, 'HRESULT_FROM_NT(STATUS_DS_NO_NEST_GLOBALGROUP_IN_MIXEDDOMAIN)')
    NTSTATUS__STATUS_DS_NO_NEST_LOCALGROUP_IN_MIXEDDOMAIN: Final[ThirtyTwoBits] = cached_hresult(0xD00002D6, 'HRESULT_FROM_NT(STATUS_DS_NO_NEST_LOCALGROUP_IN_MIXEDDOMAIN)')
    NTSTATUS__STATUS_DS_GLOBAL_CANT_HAVE_LOCAL_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0xD00002D7, 'HRESULT_FROM_NT(STATUS_DS_GLOBAL_CANT_HAVE_LOCAL_MEMBER)')
    NTSTATUS__STATUS_DS_GLOBAL_CANT_HAVE_UNIVERSAL_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0xD00002D8, 'HRESULT_FROM_NT(STATUS_DS_GLOBAL_CANT_HAVE_UNIVERSAL_MEMBER)')
    NTSTATUS__STATUS_DS_UNIVERSAL_CANT_HAVE_LOCAL_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0xD00002D9, 'HRESULT_FROM_NT(STATUS_DS_UNIVERSAL_CANT_HAVE_LOCAL_MEMBER)')
    NTSTATUS__STATUS_DS_GLOBAL_CANT_HAVE_CROSSDOMAIN_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0xD00002DA, 'HRESULT_FROM_NT(STATUS_DS_GLOBAL_CANT_HAVE_CROSSDOMAIN_MEMBER)')
    NTSTATUS__STATUS_DS_LOCAL_CANT_HAVE_CROSSDOMAIN_LOCAL_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0xD00002DB, 'HRESULT_FROM_NT(STATUS_DS_LOCAL_CANT_HAVE_CROSSDOMAIN_LOCAL_MEMBER)')
    NTSTATUS__STATUS_DS_HAVE_PRIMARY_MEMBERS: Final[ThirtyTwoBits] = cached_hresult(0xD00002DC, 'HRESULT_FROM_NT(STATUS_DS_HAVE_PRIMARY_MEMBERS)')
    NTSTATUS__STATUS_WMI_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD00002DD, 'HRESULT_FROM_NT(STATUS_WMI_NOT_SUPPORTED)')
    NTSTATUS__STATUS_INSUFFICIENT_POWER: Final[ThirtyTwoBits] = cached_hresult(0xD00002DE, 'HRESULT_FROM_NT(STATUS_INSUFFICIENT_POWER)')
    NTSTATUS__STATUS_SAM_NEED_BOOTKEY_PASSWORD: Final[ThirtyTwoBits] = cached_hresult(0xD00002DF, 'HRESULT_FROM_NT(STATUS_SAM_NEED_BOOTKEY_PASSWORD)')
    NTSTATUS__STATUS_SAM_NEED_BOOTKEY_FLOPPY: Final[ThirtyTwoBits] = cached_hresult(0xD00002E0, 'HRESULT_FROM_NT(STATUS_SAM_NEED_BOOTKEY_FLOPPY)')
    NTSTATUS__STATUS_DS_CANT_START: Final[ThirtyTwoBits] = cached_hresult(0xD00002E1, 'HRESULT_FROM_NT(STATUS_DS_CANT_START)')
    NTSTATUS__STATUS_DS_INIT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD00002E2, 'HRESULT_FROM_NT(STATUS_DS_INIT_FAILURE)')
    NTSTATUS__STATUS_SAM_INIT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD00002E3, 'HRESULT_FROM_NT(STATUS_SAM_INIT_FAILURE)')
    NTSTATUS__STATUS_DS_GC_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0xD00002E4, 'HRESULT_FROM_NT(STATUS_DS_GC_REQUIRED)')
    NTSTATUS__STATUS_DS_LOCAL_MEMBER_OF_LOCAL_ONLY: Final[ThirtyTwoBits] = cached_hresult(0xD00002E5, 'HRESULT_FROM_NT(STATUS_DS_LOCAL_MEMBER_OF_LOCAL_ONLY)')
    NTSTATUS__STATUS_DS_NO_FPO_IN_UNIVERSAL_GROUPS: Final[ThirtyTwoBits] = cached_hresult(0xD00002E6, 'HRESULT_FROM_NT(STATUS_DS_NO_FPO_IN_UNIVERSAL_GROUPS)')
    NTSTATUS__STATUS_DS_MACHINE_ACCOUNT_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD00002E7, 'HRESULT_FROM_NT(STATUS_DS_MACHINE_ACCOUNT_QUOTA_EXCEEDED)')
    NTSTATUS__STATUS_MULTIPLE_FAULT_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0xD00002E8, 'HRESULT_FROM_NT(STATUS_MULTIPLE_FAULT_VIOLATION)')
    NTSTATUS__STATUS_CURRENT_DOMAIN_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0xD00002E9, 'HRESULT_FROM_NT(STATUS_CURRENT_DOMAIN_NOT_ALLOWED)')
    NTSTATUS__STATUS_CANNOT_MAKE: Final[ThirtyTwoBits] = cached_hresult(0xD00002EA, 'HRESULT_FROM_NT(STATUS_CANNOT_MAKE)')
    NTSTATUS__STATUS_SYSTEM_SHUTDOWN: Final[ThirtyTwoBits] = cached_hresult(0xD00002EB, 'HRESULT_FROM_NT(STATUS_SYSTEM_SHUTDOWN)')
    NTSTATUS__STATUS_DS_INIT_FAILURE_CONSOLE: Final[ThirtyTwoBits] = cached_hresult(0xD00002EC, 'HRESULT_FROM_NT(STATUS_DS_INIT_FAILURE_CONSOLE)')
    NTSTATUS__STATUS_DS_SAM_INIT_FAILURE_CONSOLE: Final[ThirtyTwoBits] = cached_hresult(0xD00002ED, 'HRESULT_FROM_NT(STATUS_DS_SAM_INIT_FAILURE_CONSOLE)')
    NTSTATUS__STATUS_UNFINISHED_CONTEXT_DELETED: Final[ThirtyTwoBits] = cached_hresult(0xD00002EE, 'HRESULT_FROM_NT(STATUS_UNFINISHED_CONTEXT_DELETED)')
    NTSTATUS__STATUS_NO_TGT_REPLY: Final[ThirtyTwoBits] = cached_hresult(0xD00002EF, 'HRESULT_FROM_NT(STATUS_NO_TGT_REPLY)')
    NTSTATUS__STATUS_OBJECTID_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD00002F0, 'HRESULT_FROM_NT(STATUS_OBJECTID_NOT_FOUND)')
    NTSTATUS__STATUS_NO_IP_ADDRESSES: Final[ThirtyTwoBits] = cached_hresult(0xD00002F1, 'HRESULT_FROM_NT(STATUS_NO_IP_ADDRESSES)')
    NTSTATUS__STATUS_WRONG_CREDENTIAL_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0xD00002F2, 'HRESULT_FROM_NT(STATUS_WRONG_CREDENTIAL_HANDLE)')
    NTSTATUS__STATUS_CRYPTO_SYSTEM_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD00002F3, 'HRESULT_FROM_NT(STATUS_CRYPTO_SYSTEM_INVALID)')
    NTSTATUS__STATUS_MAX_REFERRALS_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD00002F4, 'HRESULT_FROM_NT(STATUS_MAX_REFERRALS_EXCEEDED)')
    NTSTATUS__STATUS_MUST_BE_KDC: Final[ThirtyTwoBits] = cached_hresult(0xD00002F5, 'HRESULT_FROM_NT(STATUS_MUST_BE_KDC)')
    NTSTATUS__STATUS_STRONG_CRYPTO_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD00002F6, 'HRESULT_FROM_NT(STATUS_STRONG_CRYPTO_NOT_SUPPORTED)')
    NTSTATUS__STATUS_TOO_MANY_PRINCIPALS: Final[ThirtyTwoBits] = cached_hresult(0xD00002F7, 'HRESULT_FROM_NT(STATUS_TOO_MANY_PRINCIPALS)')
    NTSTATUS__STATUS_NO_PA_DATA: Final[ThirtyTwoBits] = cached_hresult(0xD00002F8, 'HRESULT_FROM_NT(STATUS_NO_PA_DATA)')
    NTSTATUS__STATUS_PKINIT_NAME_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD00002F9, 'HRESULT_FROM_NT(STATUS_PKINIT_NAME_MISMATCH)')
    NTSTATUS__STATUS_SMARTCARD_LOGON_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0xD00002FA, 'HRESULT_FROM_NT(STATUS_SMARTCARD_LOGON_REQUIRED)')
    NTSTATUS__STATUS_KDC_INVALID_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0xD00002FB, 'HRESULT_FROM_NT(STATUS_KDC_INVALID_REQUEST)')
    NTSTATUS__STATUS_KDC_UNABLE_TO_REFER: Final[ThirtyTwoBits] = cached_hresult(0xD00002FC, 'HRESULT_FROM_NT(STATUS_KDC_UNABLE_TO_REFER)')
    NTSTATUS__STATUS_KDC_UNKNOWN_ETYPE: Final[ThirtyTwoBits] = cached_hresult(0xD00002FD, 'HRESULT_FROM_NT(STATUS_KDC_UNKNOWN_ETYPE)')
    NTSTATUS__STATUS_SHUTDOWN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD00002FE, 'HRESULT_FROM_NT(STATUS_SHUTDOWN_IN_PROGRESS)')
    NTSTATUS__STATUS_SERVER_SHUTDOWN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD00002FF, 'HRESULT_FROM_NT(STATUS_SERVER_SHUTDOWN_IN_PROGRESS)')
    NTSTATUS__STATUS_NOT_SUPPORTED_ON_SBS: Final[ThirtyTwoBits] = cached_hresult(0xD0000300, 'HRESULT_FROM_NT(STATUS_NOT_SUPPORTED_ON_SBS)')
    NTSTATUS__STATUS_WMI_GUID_DISCONNECTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000301, 'HRESULT_FROM_NT(STATUS_WMI_GUID_DISCONNECTED)')
    NTSTATUS__STATUS_WMI_ALREADY_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD0000302, 'HRESULT_FROM_NT(STATUS_WMI_ALREADY_DISABLED)')
    NTSTATUS__STATUS_WMI_ALREADY_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0xD0000303, 'HRESULT_FROM_NT(STATUS_WMI_ALREADY_ENABLED)')
    NTSTATUS__STATUS_MFT_TOO_FRAGMENTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000304, 'HRESULT_FROM_NT(STATUS_MFT_TOO_FRAGMENTED)')
    NTSTATUS__STATUS_COPY_PROTECTION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD0000305, 'HRESULT_FROM_NT(STATUS_COPY_PROTECTION_FAILURE)')
    NTSTATUS__STATUS_CSS_AUTHENTICATION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD0000306, 'HRESULT_FROM_NT(STATUS_CSS_AUTHENTICATION_FAILURE)')
    NTSTATUS__STATUS_CSS_KEY_NOT_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0xD0000307, 'HRESULT_FROM_NT(STATUS_CSS_KEY_NOT_PRESENT)')
    NTSTATUS__STATUS_CSS_KEY_NOT_ESTABLISHED: Final[ThirtyTwoBits] = cached_hresult(0xD0000308, 'HRESULT_FROM_NT(STATUS_CSS_KEY_NOT_ESTABLISHED)')
    NTSTATUS__STATUS_CSS_SCRAMBLED_SECTOR: Final[ThirtyTwoBits] = cached_hresult(0xD0000309, 'HRESULT_FROM_NT(STATUS_CSS_SCRAMBLED_SECTOR)')
    NTSTATUS__STATUS_CSS_REGION_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD000030A, 'HRESULT_FROM_NT(STATUS_CSS_REGION_MISMATCH)')
    NTSTATUS__STATUS_CSS_RESETS_EXHAUSTED: Final[ThirtyTwoBits] = cached_hresult(0xD000030B, 'HRESULT_FROM_NT(STATUS_CSS_RESETS_EXHAUSTED)')
    NTSTATUS__STATUS_PASSWORD_CHANGE_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0xD000030C, 'HRESULT_FROM_NT(STATUS_PASSWORD_CHANGE_REQUIRED)')
    NTSTATUS__STATUS_LOST_MODE_LOGON_RESTRICTION: Final[ThirtyTwoBits] = cached_hresult(0xD000030D, 'HRESULT_FROM_NT(STATUS_LOST_MODE_LOGON_RESTRICTION)')
    NTSTATUS__STATUS_PKINIT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD0000320, 'HRESULT_FROM_NT(STATUS_PKINIT_FAILURE)')
    NTSTATUS__STATUS_SMARTCARD_SUBSYSTEM_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD0000321, 'HRESULT_FROM_NT(STATUS_SMARTCARD_SUBSYSTEM_FAILURE)')
    NTSTATUS__STATUS_NO_KERB_KEY: Final[ThirtyTwoBits] = cached_hresult(0xD0000322, 'HRESULT_FROM_NT(STATUS_NO_KERB_KEY)')
    NTSTATUS__STATUS_HOST_DOWN: Final[ThirtyTwoBits] = cached_hresult(0xD0000350, 'HRESULT_FROM_NT(STATUS_HOST_DOWN)')
    NTSTATUS__STATUS_UNSUPPORTED_PREAUTH: Final[ThirtyTwoBits] = cached_hresult(0xD0000351, 'HRESULT_FROM_NT(STATUS_UNSUPPORTED_PREAUTH)')
    NTSTATUS__STATUS_EFS_ALG_BLOB_TOO_BIG: Final[ThirtyTwoBits] = cached_hresult(0xD0000352, 'HRESULT_FROM_NT(STATUS_EFS_ALG_BLOB_TOO_BIG)')
    NTSTATUS__STATUS_PORT_NOT_SET: Final[ThirtyTwoBits] = cached_hresult(0xD0000353, 'HRESULT_FROM_NT(STATUS_PORT_NOT_SET)')
    NTSTATUS__STATUS_DEBUGGER_INACTIVE: Final[ThirtyTwoBits] = cached_hresult(0xD0000354, 'HRESULT_FROM_NT(STATUS_DEBUGGER_INACTIVE)')
    NTSTATUS__STATUS_DS_VERSION_CHECK_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD0000355, 'HRESULT_FROM_NT(STATUS_DS_VERSION_CHECK_FAILURE)')
    NTSTATUS__STATUS_AUDITING_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD0000356, 'HRESULT_FROM_NT(STATUS_AUDITING_DISABLED)')
    NTSTATUS__STATUS_PRENT4_MACHINE_ACCOUNT: Final[ThirtyTwoBits] = cached_hresult(0xD0000357, 'HRESULT_FROM_NT(STATUS_PRENT4_MACHINE_ACCOUNT)')
    NTSTATUS__STATUS_DS_AG_CANT_HAVE_UNIVERSAL_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0xD0000358, 'HRESULT_FROM_NT(STATUS_DS_AG_CANT_HAVE_UNIVERSAL_MEMBER)')
    NTSTATUS__STATUS_INVALID_IMAGE_WIN_32: Final[ThirtyTwoBits] = cached_hresult(0xD0000359, 'HRESULT_FROM_NT(STATUS_INVALID_IMAGE_WIN_32)')
    NTSTATUS__STATUS_INVALID_IMAGE_WIN_64: Final[ThirtyTwoBits] = cached_hresult(0xD000035A, 'HRESULT_FROM_NT(STATUS_INVALID_IMAGE_WIN_64)')
    NTSTATUS__STATUS_BAD_BINDINGS: Final[ThirtyTwoBits] = cached_hresult(0xD000035B, 'HRESULT_FROM_NT(STATUS_BAD_BINDINGS)')
    NTSTATUS__STATUS_NETWORK_SESSION_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0xD000035C, 'HRESULT_FROM_NT(STATUS_NETWORK_SESSION_EXPIRED)')
    NTSTATUS__STATUS_APPHELP_BLOCK: Final[ThirtyTwoBits] = cached_hresult(0xD000035D, 'HRESULT_FROM_NT(STATUS_APPHELP_BLOCK)')
    NTSTATUS__STATUS_ALL_SIDS_FILTERED: Final[ThirtyTwoBits] = cached_hresult(0xD000035E, 'HRESULT_FROM_NT(STATUS_ALL_SIDS_FILTERED)')
    NTSTATUS__STATUS_NOT_SAFE_MODE_DRIVER: Final[ThirtyTwoBits] = cached_hresult(0xD000035F, 'HRESULT_FROM_NT(STATUS_NOT_SAFE_MODE_DRIVER)')
    NTSTATUS__STATUS_ACCESS_DISABLED_BY_POLICY_DEFAULT: Final[ThirtyTwoBits] = cached_hresult(0xD0000361, 'HRESULT_FROM_NT(STATUS_ACCESS_DISABLED_BY_POLICY_DEFAULT)')
    NTSTATUS__STATUS_ACCESS_DISABLED_BY_POLICY_PATH: Final[ThirtyTwoBits] = cached_hresult(0xD0000362, 'HRESULT_FROM_NT(STATUS_ACCESS_DISABLED_BY_POLICY_PATH)')
    NTSTATUS__STATUS_ACCESS_DISABLED_BY_POLICY_PUBLISHER: Final[ThirtyTwoBits] = cached_hresult(0xD0000363, 'HRESULT_FROM_NT(STATUS_ACCESS_DISABLED_BY_POLICY_PUBLISHER)')
    NTSTATUS__STATUS_ACCESS_DISABLED_BY_POLICY_OTHER: Final[ThirtyTwoBits] = cached_hresult(0xD0000364, 'HRESULT_FROM_NT(STATUS_ACCESS_DISABLED_BY_POLICY_OTHER)')
    NTSTATUS__STATUS_FAILED_DRIVER_ENTRY: Final[ThirtyTwoBits] = cached_hresult(0xD0000365, 'HRESULT_FROM_NT(STATUS_FAILED_DRIVER_ENTRY)')
    NTSTATUS__STATUS_DEVICE_ENUMERATION_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0000366, 'HRESULT_FROM_NT(STATUS_DEVICE_ENUMERATION_ERROR)')
    NTSTATUS__STATUS_MOUNT_POINT_NOT_RESOLVED: Final[ThirtyTwoBits] = cached_hresult(0xD0000368, 'HRESULT_FROM_NT(STATUS_MOUNT_POINT_NOT_RESOLVED)')
    NTSTATUS__STATUS_INVALID_DEVICE_OBJECT_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0xD0000369, 'HRESULT_FROM_NT(STATUS_INVALID_DEVICE_OBJECT_PARAMETER)')
    NTSTATUS__STATUS_MCA_OCCURED: Final[ThirtyTwoBits] = cached_hresult(0xD000036A, 'HRESULT_FROM_NT(STATUS_MCA_OCCURED)')
    NTSTATUS__STATUS_DRIVER_BLOCKED_CRITICAL: Final[ThirtyTwoBits] = cached_hresult(0xD000036B, 'HRESULT_FROM_NT(STATUS_DRIVER_BLOCKED_CRITICAL)')
    NTSTATUS__STATUS_DRIVER_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0xD000036C, 'HRESULT_FROM_NT(STATUS_DRIVER_BLOCKED)')
    NTSTATUS__STATUS_DRIVER_DATABASE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD000036D, 'HRESULT_FROM_NT(STATUS_DRIVER_DATABASE_ERROR)')
    NTSTATUS__STATUS_SYSTEM_HIVE_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0xD000036E, 'HRESULT_FROM_NT(STATUS_SYSTEM_HIVE_TOO_LARGE)')
    NTSTATUS__STATUS_INVALID_IMPORT_OF_NON_DLL: Final[ThirtyTwoBits] = cached_hresult(0xD000036F, 'HRESULT_FROM_NT(STATUS_INVALID_IMPORT_OF_NON_DLL)')
    NTSTATUS__STATUS_NO_SECRETS: Final[ThirtyTwoBits] = cached_hresult(0xD0000371, 'HRESULT_FROM_NT(STATUS_NO_SECRETS)')
    NTSTATUS__STATUS_ACCESS_DISABLED_NO_SAFER_UI_BY_POLICY: Final[ThirtyTwoBits] = cached_hresult(0xD0000372, 'HRESULT_FROM_NT(STATUS_ACCESS_DISABLED_NO_SAFER_UI_BY_POLICY)')
    NTSTATUS__STATUS_FAILED_STACK_SWITCH: Final[ThirtyTwoBits] = cached_hresult(0xD0000373, 'HRESULT_FROM_NT(STATUS_FAILED_STACK_SWITCH)')
    NTSTATUS__STATUS_HEAP_CORRUPTION: Final[ThirtyTwoBits] = cached_hresult(0xD0000374, 'HRESULT_FROM_NT(STATUS_HEAP_CORRUPTION)')
    NTSTATUS__STATUS_SMARTCARD_WRONG_PIN: Final[ThirtyTwoBits] = cached_hresult(0xD0000380, 'HRESULT_FROM_NT(STATUS_SMARTCARD_WRONG_PIN)')
    NTSTATUS__STATUS_SMARTCARD_CARD_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0xD0000381, 'HRESULT_FROM_NT(STATUS_SMARTCARD_CARD_BLOCKED)')
    NTSTATUS__STATUS_SMARTCARD_CARD_NOT_AUTHENTICATED: Final[ThirtyTwoBits] = cached_hresult(0xD0000382, 'HRESULT_FROM_NT(STATUS_SMARTCARD_CARD_NOT_AUTHENTICATED)')
    NTSTATUS__STATUS_SMARTCARD_NO_CARD: Final[ThirtyTwoBits] = cached_hresult(0xD0000383, 'HRESULT_FROM_NT(STATUS_SMARTCARD_NO_CARD)')
    NTSTATUS__STATUS_SMARTCARD_NO_KEY_CONTAINER: Final[ThirtyTwoBits] = cached_hresult(0xD0000384, 'HRESULT_FROM_NT(STATUS_SMARTCARD_NO_KEY_CONTAINER)')
    NTSTATUS__STATUS_SMARTCARD_NO_CERTIFICATE: Final[ThirtyTwoBits] = cached_hresult(0xD0000385, 'HRESULT_FROM_NT(STATUS_SMARTCARD_NO_CERTIFICATE)')
    NTSTATUS__STATUS_SMARTCARD_NO_KEYSET: Final[ThirtyTwoBits] = cached_hresult(0xD0000386, 'HRESULT_FROM_NT(STATUS_SMARTCARD_NO_KEYSET)')
    NTSTATUS__STATUS_SMARTCARD_IO_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0000387, 'HRESULT_FROM_NT(STATUS_SMARTCARD_IO_ERROR)')
    NTSTATUS__STATUS_DOWNGRADE_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000388, 'HRESULT_FROM_NT(STATUS_DOWNGRADE_DETECTED)')
    NTSTATUS__STATUS_SMARTCARD_CERT_REVOKED: Final[ThirtyTwoBits] = cached_hresult(0xD0000389, 'HRESULT_FROM_NT(STATUS_SMARTCARD_CERT_REVOKED)')
    NTSTATUS__STATUS_ISSUING_CA_UNTRUSTED: Final[ThirtyTwoBits] = cached_hresult(0xD000038A, 'HRESULT_FROM_NT(STATUS_ISSUING_CA_UNTRUSTED)')
    NTSTATUS__STATUS_REVOCATION_OFFLINE_C: Final[ThirtyTwoBits] = cached_hresult(0xD000038B, 'HRESULT_FROM_NT(STATUS_REVOCATION_OFFLINE_C)')
    NTSTATUS__STATUS_PKINIT_CLIENT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD000038C, 'HRESULT_FROM_NT(STATUS_PKINIT_CLIENT_FAILURE)')
    NTSTATUS__STATUS_SMARTCARD_CERT_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0xD000038D, 'HRESULT_FROM_NT(STATUS_SMARTCARD_CERT_EXPIRED)')
    NTSTATUS__STATUS_DRIVER_FAILED_PRIOR_UNLOAD: Final[ThirtyTwoBits] = cached_hresult(0xD000038E, 'HRESULT_FROM_NT(STATUS_DRIVER_FAILED_PRIOR_UNLOAD)')
    NTSTATUS__STATUS_SMARTCARD_SILENT_CONTEXT: Final[ThirtyTwoBits] = cached_hresult(0xD000038F, 'HRESULT_FROM_NT(STATUS_SMARTCARD_SILENT_CONTEXT)')
    NTSTATUS__STATUS_PER_USER_TRUST_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD0000401, 'HRESULT_FROM_NT(STATUS_PER_USER_TRUST_QUOTA_EXCEEDED)')
    NTSTATUS__STATUS_ALL_USER_TRUST_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD0000402, 'HRESULT_FROM_NT(STATUS_ALL_USER_TRUST_QUOTA_EXCEEDED)')
    NTSTATUS__STATUS_USER_DELETE_TRUST_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD0000403, 'HRESULT_FROM_NT(STATUS_USER_DELETE_TRUST_QUOTA_EXCEEDED)')
    NTSTATUS__STATUS_DS_NAME_NOT_UNIQUE: Final[ThirtyTwoBits] = cached_hresult(0xD0000404, 'HRESULT_FROM_NT(STATUS_DS_NAME_NOT_UNIQUE)')
    NTSTATUS__STATUS_DS_DUPLICATE_ID_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000405, 'HRESULT_FROM_NT(STATUS_DS_DUPLICATE_ID_FOUND)')
    NTSTATUS__STATUS_DS_GROUP_CONVERSION_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0000406, 'HRESULT_FROM_NT(STATUS_DS_GROUP_CONVERSION_ERROR)')
    NTSTATUS__STATUS_VOLSNAP_PREPARE_HIBERNATE: Final[ThirtyTwoBits] = cached_hresult(0xD0000407, 'HRESULT_FROM_NT(STATUS_VOLSNAP_PREPARE_HIBERNATE)')
    NTSTATUS__STATUS_USER2USER_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0xD0000408, 'HRESULT_FROM_NT(STATUS_USER2USER_REQUIRED)')
    NTSTATUS__STATUS_STACK_BUFFER_OVERRUN: Final[ThirtyTwoBits] = cached_hresult(0xD0000409, 'HRESULT_FROM_NT(STATUS_STACK_BUFFER_OVERRUN)')
    NTSTATUS__STATUS_NO_S4U_PROT_SUPPORT: Final[ThirtyTwoBits] = cached_hresult(0xD000040A, 'HRESULT_FROM_NT(STATUS_NO_S4U_PROT_SUPPORT)')
    NTSTATUS__STATUS_CROSSREALM_DELEGATION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD000040B, 'HRESULT_FROM_NT(STATUS_CROSSREALM_DELEGATION_FAILURE)')
    NTSTATUS__STATUS_REVOCATION_OFFLINE_KDC: Final[ThirtyTwoBits] = cached_hresult(0xD000040C, 'HRESULT_FROM_NT(STATUS_REVOCATION_OFFLINE_KDC)')
    NTSTATUS__STATUS_ISSUING_CA_UNTRUSTED_KDC: Final[ThirtyTwoBits] = cached_hresult(0xD000040D, 'HRESULT_FROM_NT(STATUS_ISSUING_CA_UNTRUSTED_KDC)')
    NTSTATUS__STATUS_KDC_CERT_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0xD000040E, 'HRESULT_FROM_NT(STATUS_KDC_CERT_EXPIRED)')
    NTSTATUS__STATUS_KDC_CERT_REVOKED: Final[ThirtyTwoBits] = cached_hresult(0xD000040F, 'HRESULT_FROM_NT(STATUS_KDC_CERT_REVOKED)')
    NTSTATUS__STATUS_PARAMETER_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD0000410, 'HRESULT_FROM_NT(STATUS_PARAMETER_QUOTA_EXCEEDED)')
    NTSTATUS__STATUS_HIBERNATION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD0000411, 'HRESULT_FROM_NT(STATUS_HIBERNATION_FAILURE)')
    NTSTATUS__STATUS_DELAY_LOAD_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0000412, 'HRESULT_FROM_NT(STATUS_DELAY_LOAD_FAILED)')
    NTSTATUS__STATUS_AUTHENTICATION_FIREWALL_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0000413, 'HRESULT_FROM_NT(STATUS_AUTHENTICATION_FIREWALL_FAILED)')
    NTSTATUS__STATUS_VDM_DISALLOWED: Final[ThirtyTwoBits] = cached_hresult(0xD0000414, 'HRESULT_FROM_NT(STATUS_VDM_DISALLOWED)')
    NTSTATUS__STATUS_HUNG_DISPLAY_DRIVER_THREAD: Final[ThirtyTwoBits] = cached_hresult(0xD0000415, 'HRESULT_FROM_NT(STATUS_HUNG_DISPLAY_DRIVER_THREAD)')
    NTSTATUS__STATUS_INSUFFICIENT_RESOURCE_FOR_SPECIFIED_SHARED_SECTION_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD0000416, 'HRESULT_FROM_NT(STATUS_INSUFFICIENT_RESOURCE_FOR_SPECIFIED_SHARED_SECTION_SIZE)')
    NTSTATUS__STATUS_INVALID_CRUNTIME_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0xD0000417, 'HRESULT_FROM_NT(STATUS_INVALID_CRUNTIME_PARAMETER)')
    NTSTATUS__STATUS_NTLM_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0xD0000418, 'HRESULT_FROM_NT(STATUS_NTLM_BLOCKED)')
    NTSTATUS__STATUS_DS_SRC_SID_EXISTS_IN_FOREST: Final[ThirtyTwoBits] = cached_hresult(0xD0000419, 'HRESULT_FROM_NT(STATUS_DS_SRC_SID_EXISTS_IN_FOREST)')
    NTSTATUS__STATUS_DS_DOMAIN_NAME_EXISTS_IN_FOREST: Final[ThirtyTwoBits] = cached_hresult(0xD000041A, 'HRESULT_FROM_NT(STATUS_DS_DOMAIN_NAME_EXISTS_IN_FOREST)')
    NTSTATUS__STATUS_DS_FLAT_NAME_EXISTS_IN_FOREST: Final[ThirtyTwoBits] = cached_hresult(0xD000041B, 'HRESULT_FROM_NT(STATUS_DS_FLAT_NAME_EXISTS_IN_FOREST)')
    NTSTATUS__STATUS_INVALID_USER_PRINCIPAL_NAME: Final[ThirtyTwoBits] = cached_hresult(0xD000041C, 'HRESULT_FROM_NT(STATUS_INVALID_USER_PRINCIPAL_NAME)')
    NTSTATUS__STATUS_FATAL_USER_CALLBACK_EXCEPTION: Final[ThirtyTwoBits] = cached_hresult(0xD000041D, 'HRESULT_FROM_NT(STATUS_FATAL_USER_CALLBACK_EXCEPTION)')
    NTSTATUS__STATUS_ASSERTION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD0000420, 'HRESULT_FROM_NT(STATUS_ASSERTION_FAILURE)')
    NTSTATUS__STATUS_VERIFIER_STOP: Final[ThirtyTwoBits] = cached_hresult(0xD0000421, 'HRESULT_FROM_NT(STATUS_VERIFIER_STOP)')
    NTSTATUS__STATUS_CALLBACK_POP_STACK: Final[ThirtyTwoBits] = cached_hresult(0xD0000423, 'HRESULT_FROM_NT(STATUS_CALLBACK_POP_STACK)')
    NTSTATUS__STATUS_INCOMPATIBLE_DRIVER_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0xD0000424, 'HRESULT_FROM_NT(STATUS_INCOMPATIBLE_DRIVER_BLOCKED)')
    NTSTATUS__STATUS_HIVE_UNLOADED: Final[ThirtyTwoBits] = cached_hresult(0xD0000425, 'HRESULT_FROM_NT(STATUS_HIVE_UNLOADED)')
    NTSTATUS__STATUS_COMPRESSION_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD0000426, 'HRESULT_FROM_NT(STATUS_COMPRESSION_DISABLED)')
    NTSTATUS__STATUS_FILE_SYSTEM_LIMITATION: Final[ThirtyTwoBits] = cached_hresult(0xD0000427, 'HRESULT_FROM_NT(STATUS_FILE_SYSTEM_LIMITATION)')
    NTSTATUS__STATUS_INVALID_IMAGE_HASH: Final[ThirtyTwoBits] = cached_hresult(0xD0000428, 'HRESULT_FROM_NT(STATUS_INVALID_IMAGE_HASH)')
    NTSTATUS__STATUS_NOT_CAPABLE: Final[ThirtyTwoBits] = cached_hresult(0xD0000429, 'HRESULT_FROM_NT(STATUS_NOT_CAPABLE)')
    NTSTATUS__STATUS_REQUEST_OUT_OF_SEQUENCE: Final[ThirtyTwoBits] = cached_hresult(0xD000042A, 'HRESULT_FROM_NT(STATUS_REQUEST_OUT_OF_SEQUENCE)')
    NTSTATUS__STATUS_IMPLEMENTATION_LIMIT: Final[ThirtyTwoBits] = cached_hresult(0xD000042B, 'HRESULT_FROM_NT(STATUS_IMPLEMENTATION_LIMIT)')
    NTSTATUS__STATUS_ELEVATION_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0xD000042C, 'HRESULT_FROM_NT(STATUS_ELEVATION_REQUIRED)')
    NTSTATUS__STATUS_NO_SECURITY_CONTEXT: Final[ThirtyTwoBits] = cached_hresult(0xD000042D, 'HRESULT_FROM_NT(STATUS_NO_SECURITY_CONTEXT)')
    NTSTATUS__STATUS_PKU2U_CERT_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD000042F, 'HRESULT_FROM_NT(STATUS_PKU2U_CERT_FAILURE)')
    NTSTATUS__STATUS_BEYOND_VDL: Final[ThirtyTwoBits] = cached_hresult(0xD0000432, 'HRESULT_FROM_NT(STATUS_BEYOND_VDL)')
    NTSTATUS__STATUS_ENCOUNTERED_WRITE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD0000433, 'HRESULT_FROM_NT(STATUS_ENCOUNTERED_WRITE_IN_PROGRESS)')
    NTSTATUS__STATUS_PTE_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0xD0000434, 'HRESULT_FROM_NT(STATUS_PTE_CHANGED)')
    NTSTATUS__STATUS_PURGE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0000435, 'HRESULT_FROM_NT(STATUS_PURGE_FAILED)')
    NTSTATUS__STATUS_CRED_REQUIRES_CONFIRMATION: Final[ThirtyTwoBits] = cached_hresult(0xD0000440, 'HRESULT_FROM_NT(STATUS_CRED_REQUIRES_CONFIRMATION)')
    NTSTATUS__STATUS_CS_ENCRYPTION_INVALID_SERVER_RESPONSE: Final[ThirtyTwoBits] = cached_hresult(0xD0000441, 'HRESULT_FROM_NT(STATUS_CS_ENCRYPTION_INVALID_SERVER_RESPONSE)')
    NTSTATUS__STATUS_CS_ENCRYPTION_UNSUPPORTED_SERVER: Final[ThirtyTwoBits] = cached_hresult(0xD0000442, 'HRESULT_FROM_NT(STATUS_CS_ENCRYPTION_UNSUPPORTED_SERVER)')
    NTSTATUS__STATUS_CS_ENCRYPTION_EXISTING_ENCRYPTED_FILE: Final[ThirtyTwoBits] = cached_hresult(0xD0000443, 'HRESULT_FROM_NT(STATUS_CS_ENCRYPTION_EXISTING_ENCRYPTED_FILE)')
    NTSTATUS__STATUS_CS_ENCRYPTION_NEW_ENCRYPTED_FILE: Final[ThirtyTwoBits] = cached_hresult(0xD0000444, 'HRESULT_FROM_NT(STATUS_CS_ENCRYPTION_NEW_ENCRYPTED_FILE)')
    NTSTATUS__STATUS_CS_ENCRYPTION_FILE_NOT_CSE: Final[ThirtyTwoBits] = cached_hresult(0xD0000445, 'HRESULT_FROM_NT(STATUS_CS_ENCRYPTION_FILE_NOT_CSE)')
    NTSTATUS__STATUS_INVALID_LABEL: Final[ThirtyTwoBits] = cached_hresult(0xD0000446, 'HRESULT_FROM_NT(STATUS_INVALID_LABEL)')
    NTSTATUS__STATUS_DRIVER_PROCESS_TERMINATED: Final[ThirtyTwoBits] = cached_hresult(0xD0000450, 'HRESULT_FROM_NT(STATUS_DRIVER_PROCESS_TERMINATED)')
    NTSTATUS__STATUS_AMBIGUOUS_SYSTEM_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0xD0000451, 'HRESULT_FROM_NT(STATUS_AMBIGUOUS_SYSTEM_DEVICE)')
    NTSTATUS__STATUS_SYSTEM_DEVICE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000452, 'HRESULT_FROM_NT(STATUS_SYSTEM_DEVICE_NOT_FOUND)')
    NTSTATUS__STATUS_RESTART_BOOT_APPLICATION: Final[ThirtyTwoBits] = cached_hresult(0xD0000453, 'HRESULT_FROM_NT(STATUS_RESTART_BOOT_APPLICATION)')
    NTSTATUS__STATUS_INSUFFICIENT_NVRAM_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0xD0000454, 'HRESULT_FROM_NT(STATUS_INSUFFICIENT_NVRAM_RESOURCES)')
    NTSTATUS__STATUS_INVALID_SESSION: Final[ThirtyTwoBits] = cached_hresult(0xD0000455, 'HRESULT_FROM_NT(STATUS_INVALID_SESSION)')
    NTSTATUS__STATUS_THREAD_ALREADY_IN_SESSION: Final[ThirtyTwoBits] = cached_hresult(0xD0000456, 'HRESULT_FROM_NT(STATUS_THREAD_ALREADY_IN_SESSION)')
    NTSTATUS__STATUS_THREAD_NOT_IN_SESSION: Final[ThirtyTwoBits] = cached_hresult(0xD0000457, 'HRESULT_FROM_NT(STATUS_THREAD_NOT_IN_SESSION)')
    NTSTATUS__STATUS_INVALID_WEIGHT: Final[ThirtyTwoBits] = cached_hresult(0xD0000458, 'HRESULT_FROM_NT(STATUS_INVALID_WEIGHT)')
    NTSTATUS__STATUS_REQUEST_PAUSED: Final[ThirtyTwoBits] = cached_hresult(0xD0000459, 'HRESULT_FROM_NT(STATUS_REQUEST_PAUSED)')
    NTSTATUS__STATUS_NO_RANGES_PROCESSED: Final[ThirtyTwoBits] = cached_hresult(0xD0000460, 'HRESULT_FROM_NT(STATUS_NO_RANGES_PROCESSED)')
    NTSTATUS__STATUS_DISK_RESOURCES_EXHAUSTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000461, 'HRESULT_FROM_NT(STATUS_DISK_RESOURCES_EXHAUSTED)')
    NTSTATUS__STATUS_NEEDS_REMEDIATION: Final[ThirtyTwoBits] = cached_hresult(0xD0000462, 'HRESULT_FROM_NT(STATUS_NEEDS_REMEDIATION)')
    NTSTATUS__STATUS_DEVICE_FEATURE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000463, 'HRESULT_FROM_NT(STATUS_DEVICE_FEATURE_NOT_SUPPORTED)')
    NTSTATUS__STATUS_DEVICE_UNREACHABLE: Final[ThirtyTwoBits] = cached_hresult(0xD0000464, 'HRESULT_FROM_NT(STATUS_DEVICE_UNREACHABLE)')
    NTSTATUS__STATUS_INVALID_TOKEN: Final[ThirtyTwoBits] = cached_hresult(0xD0000465, 'HRESULT_FROM_NT(STATUS_INVALID_TOKEN)')
    NTSTATUS__STATUS_SERVER_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD0000466, 'HRESULT_FROM_NT(STATUS_SERVER_UNAVAILABLE)')
    NTSTATUS__STATUS_FILE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD0000467, 'HRESULT_FROM_NT(STATUS_FILE_NOT_AVAILABLE)')
    NTSTATUS__STATUS_DEVICE_INSUFFICIENT_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0xD0000468, 'HRESULT_FROM_NT(STATUS_DEVICE_INSUFFICIENT_RESOURCES)')
    NTSTATUS__STATUS_PACKAGE_UPDATING: Final[ThirtyTwoBits] = cached_hresult(0xD0000469, 'HRESULT_FROM_NT(STATUS_PACKAGE_UPDATING)')
    NTSTATUS__STATUS_NOT_READ_FROM_COPY: Final[ThirtyTwoBits] = cached_hresult(0xD000046A, 'HRESULT_FROM_NT(STATUS_NOT_READ_FROM_COPY)')
    NTSTATUS__STATUS_FT_WRITE_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD000046B, 'HRESULT_FROM_NT(STATUS_FT_WRITE_FAILURE)')
    NTSTATUS__STATUS_FT_DI_SCAN_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0xD000046C, 'HRESULT_FROM_NT(STATUS_FT_DI_SCAN_REQUIRED)')
    NTSTATUS__STATUS_OBJECT_NOT_EXTERNALLY_BACKED: Final[ThirtyTwoBits] = cached_hresult(0xD000046D, 'HRESULT_FROM_NT(STATUS_OBJECT_NOT_EXTERNALLY_BACKED)')
    NTSTATUS__STATUS_EXTERNAL_BACKING_PROVIDER_UNKNOWN: Final[ThirtyTwoBits] = cached_hresult(0xD000046E, 'HRESULT_FROM_NT(STATUS_EXTERNAL_BACKING_PROVIDER_UNKNOWN)')
    NTSTATUS__STATUS_COMPRESSION_NOT_BENEFICIAL: Final[ThirtyTwoBits] = cached_hresult(0xD000046F, 'HRESULT_FROM_NT(STATUS_COMPRESSION_NOT_BENEFICIAL)')
    NTSTATUS__STATUS_DATA_CHECKSUM_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0000470, 'HRESULT_FROM_NT(STATUS_DATA_CHECKSUM_ERROR)')
    NTSTATUS__STATUS_INTERMIXED_KERNEL_EA_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0xD0000471, 'HRESULT_FROM_NT(STATUS_INTERMIXED_KERNEL_EA_OPERATION)')
    NTSTATUS__STATUS_TRIM_READ_ZERO_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000472, 'HRESULT_FROM_NT(STATUS_TRIM_READ_ZERO_NOT_SUPPORTED)')
    NTSTATUS__STATUS_TOO_MANY_SEGMENT_DESCRIPTORS: Final[ThirtyTwoBits] = cached_hresult(0xD0000473, 'HRESULT_FROM_NT(STATUS_TOO_MANY_SEGMENT_DESCRIPTORS)')
    NTSTATUS__STATUS_INVALID_OFFSET_ALIGNMENT: Final[ThirtyTwoBits] = cached_hresult(0xD0000474, 'HRESULT_FROM_NT(STATUS_INVALID_OFFSET_ALIGNMENT)')
    NTSTATUS__STATUS_INVALID_FIELD_IN_PARAMETER_LIST: Final[ThirtyTwoBits] = cached_hresult(0xD0000475, 'HRESULT_FROM_NT(STATUS_INVALID_FIELD_IN_PARAMETER_LIST)')
    NTSTATUS__STATUS_OPERATION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD0000476, 'HRESULT_FROM_NT(STATUS_OPERATION_IN_PROGRESS)')
    NTSTATUS__STATUS_INVALID_INITIATOR_TARGET_PATH: Final[ThirtyTwoBits] = cached_hresult(0xD0000477, 'HRESULT_FROM_NT(STATUS_INVALID_INITIATOR_TARGET_PATH)')
    NTSTATUS__STATUS_SCRUB_DATA_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD0000478, 'HRESULT_FROM_NT(STATUS_SCRUB_DATA_DISABLED)')
    NTSTATUS__STATUS_NOT_REDUNDANT_STORAGE: Final[ThirtyTwoBits] = cached_hresult(0xD0000479, 'HRESULT_FROM_NT(STATUS_NOT_REDUNDANT_STORAGE)')
    NTSTATUS__STATUS_RESIDENT_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD000047A, 'HRESULT_FROM_NT(STATUS_RESIDENT_FILE_NOT_SUPPORTED)')
    NTSTATUS__STATUS_COMPRESSED_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD000047B, 'HRESULT_FROM_NT(STATUS_COMPRESSED_FILE_NOT_SUPPORTED)')
    NTSTATUS__STATUS_DIRECTORY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD000047C, 'HRESULT_FROM_NT(STATUS_DIRECTORY_NOT_SUPPORTED)')
    NTSTATUS__STATUS_IO_OPERATION_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0xD000047D, 'HRESULT_FROM_NT(STATUS_IO_OPERATION_TIMEOUT)')
    NTSTATUS__STATUS_SYSTEM_NEEDS_REMEDIATION: Final[ThirtyTwoBits] = cached_hresult(0xD000047E, 'HRESULT_FROM_NT(STATUS_SYSTEM_NEEDS_REMEDIATION)')
    NTSTATUS__STATUS_APPX_INTEGRITY_FAILURE_CLR_NGEN: Final[ThirtyTwoBits] = cached_hresult(0xD000047F, 'HRESULT_FROM_NT(STATUS_APPX_INTEGRITY_FAILURE_CLR_NGEN)')
    NTSTATUS__STATUS_SHARE_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD0000480, 'HRESULT_FROM_NT(STATUS_SHARE_UNAVAILABLE)')
    NTSTATUS__STATUS_APISET_NOT_HOSTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000481, 'HRESULT_FROM_NT(STATUS_APISET_NOT_HOSTED)')
    NTSTATUS__STATUS_APISET_NOT_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0xD0000482, 'HRESULT_FROM_NT(STATUS_APISET_NOT_PRESENT)')
    NTSTATUS__STATUS_DEVICE_HARDWARE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0000483, 'HRESULT_FROM_NT(STATUS_DEVICE_HARDWARE_ERROR)')
    NTSTATUS__STATUS_FIRMWARE_SLOT_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0000484, 'HRESULT_FROM_NT(STATUS_FIRMWARE_SLOT_INVALID)')
    NTSTATUS__STATUS_FIRMWARE_IMAGE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0000485, 'HRESULT_FROM_NT(STATUS_FIRMWARE_IMAGE_INVALID)')
    NTSTATUS__STATUS_STORAGE_TOPOLOGY_ID_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD0000486, 'HRESULT_FROM_NT(STATUS_STORAGE_TOPOLOGY_ID_MISMATCH)')
    NTSTATUS__STATUS_WIM_NOT_BOOTABLE: Final[ThirtyTwoBits] = cached_hresult(0xD0000487, 'HRESULT_FROM_NT(STATUS_WIM_NOT_BOOTABLE)')
    NTSTATUS__STATUS_BLOCKED_BY_PARENTAL_CONTROLS: Final[ThirtyTwoBits] = cached_hresult(0xD0000488, 'HRESULT_FROM_NT(STATUS_BLOCKED_BY_PARENTAL_CONTROLS)')
    NTSTATUS__STATUS_NEEDS_REGISTRATION: Final[ThirtyTwoBits] = cached_hresult(0xD0000489, 'HRESULT_FROM_NT(STATUS_NEEDS_REGISTRATION)')
    NTSTATUS__STATUS_QUOTA_ACTIVITY: Final[ThirtyTwoBits] = cached_hresult(0xD000048A, 'HRESULT_FROM_NT(STATUS_QUOTA_ACTIVITY)')
    NTSTATUS__STATUS_CALLBACK_INVOKE_INLINE: Final[ThirtyTwoBits] = cached_hresult(0xD000048B, 'HRESULT_FROM_NT(STATUS_CALLBACK_INVOKE_INLINE)')
    NTSTATUS__STATUS_BLOCK_TOO_MANY_REFERENCES: Final[ThirtyTwoBits] = cached_hresult(0xD000048C, 'HRESULT_FROM_NT(STATUS_BLOCK_TOO_MANY_REFERENCES)')
    NTSTATUS__STATUS_MARKED_TO_DISALLOW_WRITES: Final[ThirtyTwoBits] = cached_hresult(0xD000048D, 'HRESULT_FROM_NT(STATUS_MARKED_TO_DISALLOW_WRITES)')
    NTSTATUS__STATUS_NETWORK_ACCESS_DENIED_EDP: Final[ThirtyTwoBits] = cached_hresult(0xD000048E, 'HRESULT_FROM_NT(STATUS_NETWORK_ACCESS_DENIED_EDP)')
    NTSTATUS__STATUS_ENCLAVE_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD000048F, 'HRESULT_FROM_NT(STATUS_ENCLAVE_FAILURE)')
    NTSTATUS__STATUS_PNP_NO_COMPAT_DRIVERS: Final[ThirtyTwoBits] = cached_hresult(0xD0000490, 'HRESULT_FROM_NT(STATUS_PNP_NO_COMPAT_DRIVERS)')
    NTSTATUS__STATUS_PNP_DRIVER_PACKAGE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000491, 'HRESULT_FROM_NT(STATUS_PNP_DRIVER_PACKAGE_NOT_FOUND)')
    NTSTATUS__STATUS_PNP_DRIVER_CONFIGURATION_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000492, 'HRESULT_FROM_NT(STATUS_PNP_DRIVER_CONFIGURATION_NOT_FOUND)')
    NTSTATUS__STATUS_PNP_DRIVER_CONFIGURATION_INCOMPLETE: Final[ThirtyTwoBits] = cached_hresult(0xD0000493, 'HRESULT_FROM_NT(STATUS_PNP_DRIVER_CONFIGURATION_INCOMPLETE)')
    NTSTATUS__STATUS_PNP_FUNCTION_DRIVER_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0xD0000494, 'HRESULT_FROM_NT(STATUS_PNP_FUNCTION_DRIVER_REQUIRED)')
    NTSTATUS__STATUS_PNP_DEVICE_CONFIGURATION_PENDING: Final[ThirtyTwoBits] = cached_hresult(0xD0000495, 'HRESULT_FROM_NT(STATUS_PNP_DEVICE_CONFIGURATION_PENDING)')
    NTSTATUS__STATUS_DEVICE_HINT_NAME_BUFFER_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0xD0000496, 'HRESULT_FROM_NT(STATUS_DEVICE_HINT_NAME_BUFFER_TOO_SMALL)')
    NTSTATUS__STATUS_PACKAGE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD0000497, 'HRESULT_FROM_NT(STATUS_PACKAGE_NOT_AVAILABLE)')
    NTSTATUS__STATUS_DEVICE_IN_MAINTENANCE: Final[ThirtyTwoBits] = cached_hresult(0xD0000499, 'HRESULT_FROM_NT(STATUS_DEVICE_IN_MAINTENANCE)')
    NTSTATUS__STATUS_NOT_SUPPORTED_ON_DAX: Final[ThirtyTwoBits] = cached_hresult(0xD000049A, 'HRESULT_FROM_NT(STATUS_NOT_SUPPORTED_ON_DAX)')
    NTSTATUS__STATUS_FREE_SPACE_TOO_FRAGMENTED: Final[ThirtyTwoBits] = cached_hresult(0xD000049B, 'HRESULT_FROM_NT(STATUS_FREE_SPACE_TOO_FRAGMENTED)')
    NTSTATUS__STATUS_DAX_MAPPING_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD000049C, 'HRESULT_FROM_NT(STATUS_DAX_MAPPING_EXISTS)')
    NTSTATUS__STATUS_CHILD_PROCESS_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0xD000049D, 'HRESULT_FROM_NT(STATUS_CHILD_PROCESS_BLOCKED)')
    NTSTATUS__STATUS_STORAGE_LOST_DATA_PERSISTENCE: Final[ThirtyTwoBits] = cached_hresult(0xD000049E, 'HRESULT_FROM_NT(STATUS_STORAGE_LOST_DATA_PERSISTENCE)')
    NTSTATUS__STATUS_PARTITION_TERMINATING: Final[ThirtyTwoBits] = cached_hresult(0xD00004A0, 'HRESULT_FROM_NT(STATUS_PARTITION_TERMINATING)')
    NTSTATUS__STATUS_EXTERNAL_SYSKEY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD00004A1, 'HRESULT_FROM_NT(STATUS_EXTERNAL_SYSKEY_NOT_SUPPORTED)')
    NTSTATUS__STATUS_ENCLAVE_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0xD00004A2, 'HRESULT_FROM_NT(STATUS_ENCLAVE_VIOLATION)')
    NTSTATUS__STATUS_FILE_PROTECTED_UNDER_DPL: Final[ThirtyTwoBits] = cached_hresult(0xD00004A3, 'HRESULT_FROM_NT(STATUS_FILE_PROTECTED_UNDER_DPL)')
    NTSTATUS__STATUS_VOLUME_NOT_CLUSTER_ALIGNED: Final[ThirtyTwoBits] = cached_hresult(0xD00004A4, 'HRESULT_FROM_NT(STATUS_VOLUME_NOT_CLUSTER_ALIGNED)')
    NTSTATUS__STATUS_NO_PHYSICALLY_ALIGNED_FREE_SPACE_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD00004A5, 'HRESULT_FROM_NT(STATUS_NO_PHYSICALLY_ALIGNED_FREE_SPACE_FOUND)')
    NTSTATUS__STATUS_APPX_FILE_NOT_ENCRYPTED: Final[ThirtyTwoBits] = cached_hresult(0xD00004A6, 'HRESULT_FROM_NT(STATUS_APPX_FILE_NOT_ENCRYPTED)')
    NTSTATUS__STATUS_RWRAW_ENCRYPTED_FILE_NOT_ENCRYPTED: Final[ThirtyTwoBits] = cached_hresult(0xD00004A7, 'HRESULT_FROM_NT(STATUS_RWRAW_ENCRYPTED_FILE_NOT_ENCRYPTED)')
    NTSTATUS__STATUS_RWRAW_ENCRYPTED_INVALID_EDATAINFO_FILEOFFSET: Final[ThirtyTwoBits] = cached_hresult(0xD00004A8, 'HRESULT_FROM_NT(STATUS_RWRAW_ENCRYPTED_INVALID_EDATAINFO_FILEOFFSET)')
    NTSTATUS__STATUS_RWRAW_ENCRYPTED_INVALID_EDATAINFO_FILERANGE: Final[ThirtyTwoBits] = cached_hresult(0xD00004A9, 'HRESULT_FROM_NT(STATUS_RWRAW_ENCRYPTED_INVALID_EDATAINFO_FILERANGE)')
    NTSTATUS__STATUS_RWRAW_ENCRYPTED_INVALID_EDATAINFO_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0xD00004AA, 'HRESULT_FROM_NT(STATUS_RWRAW_ENCRYPTED_INVALID_EDATAINFO_PARAMETER)')
    NTSTATUS__STATUS_FT_READ_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD00004AB, 'HRESULT_FROM_NT(STATUS_FT_READ_FAILURE)')
    NTSTATUS__STATUS_PATCH_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0xD00004AC, 'HRESULT_FROM_NT(STATUS_PATCH_CONFLICT)')
    NTSTATUS__STATUS_STORAGE_RESERVE_ID_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD00004AD, 'HRESULT_FROM_NT(STATUS_STORAGE_RESERVE_ID_INVALID)')
    NTSTATUS__STATUS_STORAGE_RESERVE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0xD00004AE, 'HRESULT_FROM_NT(STATUS_STORAGE_RESERVE_DOES_NOT_EXIST)')
    NTSTATUS__STATUS_STORAGE_RESERVE_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD00004AF, 'HRESULT_FROM_NT(STATUS_STORAGE_RESERVE_ALREADY_EXISTS)')
    NTSTATUS__STATUS_STORAGE_RESERVE_NOT_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0xD00004B0, 'HRESULT_FROM_NT(STATUS_STORAGE_RESERVE_NOT_EMPTY)')
    NTSTATUS__STATUS_NOT_A_DAX_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0xD00004B1, 'HRESULT_FROM_NT(STATUS_NOT_A_DAX_VOLUME)')
    NTSTATUS__STATUS_NOT_DAX_MAPPABLE: Final[ThirtyTwoBits] = cached_hresult(0xD00004B2, 'HRESULT_FROM_NT(STATUS_NOT_DAX_MAPPABLE)')
    NTSTATUS__STATUS_CASE_DIFFERING_NAMES_IN_DIR: Final[ThirtyTwoBits] = cached_hresult(0xD00004B3, 'HRESULT_FROM_NT(STATUS_CASE_DIFFERING_NAMES_IN_DIR)')
    NTSTATUS__STATUS_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD00004B4, 'HRESULT_FROM_NT(STATUS_FILE_NOT_SUPPORTED)')
    NTSTATUS__STATUS_NOT_SUPPORTED_WITH_BTT: Final[ThirtyTwoBits] = cached_hresult(0xD00004B5, 'HRESULT_FROM_NT(STATUS_NOT_SUPPORTED_WITH_BTT)')
    NTSTATUS__STATUS_ENCRYPTION_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD00004B6, 'HRESULT_FROM_NT(STATUS_ENCRYPTION_DISABLED)')
    NTSTATUS__STATUS_ENCRYPTING_METADATA_DISALLOWED: Final[ThirtyTwoBits] = cached_hresult(0xD00004B7, 'HRESULT_FROM_NT(STATUS_ENCRYPTING_METADATA_DISALLOWED)')
    NTSTATUS__STATUS_CANT_CLEAR_ENCRYPTION_FLAG: Final[ThirtyTwoBits] = cached_hresult(0xD00004B8, 'HRESULT_FROM_NT(STATUS_CANT_CLEAR_ENCRYPTION_FLAG)')
    NTSTATUS__STATUS_UNSATISFIED_DEPENDENCIES: Final[ThirtyTwoBits] = cached_hresult(0xD00004B9, 'HRESULT_FROM_NT(STATUS_UNSATISFIED_DEPENDENCIES)')
    NTSTATUS__STATUS_CASE_SENSITIVE_PATH: Final[ThirtyTwoBits] = cached_hresult(0xD00004BA, 'HRESULT_FROM_NT(STATUS_CASE_SENSITIVE_PATH)')
    NTSTATUS__STATUS_UNSUPPORTED_PAGING_MODE: Final[ThirtyTwoBits] = cached_hresult(0xD00004BB, 'HRESULT_FROM_NT(STATUS_UNSUPPORTED_PAGING_MODE)')
    NTSTATUS__STATUS_UNTRUSTED_MOUNT_POINT: Final[ThirtyTwoBits] = cached_hresult(0xD00004BC, 'HRESULT_FROM_NT(STATUS_UNTRUSTED_MOUNT_POINT)')
    NTSTATUS__STATUS_HAS_SYSTEM_CRITICAL_FILES: Final[ThirtyTwoBits] = cached_hresult(0xD00004BD, 'HRESULT_FROM_NT(STATUS_HAS_SYSTEM_CRITICAL_FILES)')
    NTSTATUS__STATUS_OBJECT_IS_IMMUTABLE: Final[ThirtyTwoBits] = cached_hresult(0xD00004BE, 'HRESULT_FROM_NT(STATUS_OBJECT_IS_IMMUTABLE)')
    NTSTATUS__STATUS_FT_READ_FROM_COPY_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD00004BF, 'HRESULT_FROM_NT(STATUS_FT_READ_FROM_COPY_FAILURE)')
    NTSTATUS__STATUS_IMAGE_LOADED_AS_PATCH_IMAGE: Final[ThirtyTwoBits] = cached_hresult(0xD00004C0, 'HRESULT_FROM_NT(STATUS_IMAGE_LOADED_AS_PATCH_IMAGE)')
    NTSTATUS__STATUS_STORAGE_STACK_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0xD00004C1, 'HRESULT_FROM_NT(STATUS_STORAGE_STACK_ACCESS_DENIED)')
    NTSTATUS__STATUS_INSUFFICIENT_VIRTUAL_ADDR_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0xD00004C2, 'HRESULT_FROM_NT(STATUS_INSUFFICIENT_VIRTUAL_ADDR_RESOURCES)')
    NTSTATUS__STATUS_ENCRYPTED_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD00004C3, 'HRESULT_FROM_NT(STATUS_ENCRYPTED_FILE_NOT_SUPPORTED)')
    NTSTATUS__STATUS_SPARSE_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD00004C4, 'HRESULT_FROM_NT(STATUS_SPARSE_FILE_NOT_SUPPORTED)')
    NTSTATUS__STATUS_PAGEFILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD00004C5, 'HRESULT_FROM_NT(STATUS_PAGEFILE_NOT_SUPPORTED)')
    NTSTATUS__STATUS_VOLUME_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD00004C6, 'HRESULT_FROM_NT(STATUS_VOLUME_NOT_SUPPORTED)')
    NTSTATUS__STATUS_NOT_SUPPORTED_WITH_BYPASSIO: Final[ThirtyTwoBits] = cached_hresult(0xD00004C7, 'HRESULT_FROM_NT(STATUS_NOT_SUPPORTED_WITH_BYPASSIO)')
    NTSTATUS__STATUS_NO_BYPASSIO_DRIVER_SUPPORT: Final[ThirtyTwoBits] = cached_hresult(0xD00004C8, 'HRESULT_FROM_NT(STATUS_NO_BYPASSIO_DRIVER_SUPPORT)')
    NTSTATUS__STATUS_NOT_SUPPORTED_WITH_ENCRYPTION: Final[ThirtyTwoBits] = cached_hresult(0xD00004C9, 'HRESULT_FROM_NT(STATUS_NOT_SUPPORTED_WITH_ENCRYPTION)')
    NTSTATUS__STATUS_NOT_SUPPORTED_WITH_COMPRESSION: Final[ThirtyTwoBits] = cached_hresult(0xD00004CA, 'HRESULT_FROM_NT(STATUS_NOT_SUPPORTED_WITH_COMPRESSION)')
    NTSTATUS__STATUS_NOT_SUPPORTED_WITH_REPLICATION: Final[ThirtyTwoBits] = cached_hresult(0xD00004CB, 'HRESULT_FROM_NT(STATUS_NOT_SUPPORTED_WITH_REPLICATION)')
    NTSTATUS__STATUS_NOT_SUPPORTED_WITH_DEDUPLICATION: Final[ThirtyTwoBits] = cached_hresult(0xD00004CC, 'HRESULT_FROM_NT(STATUS_NOT_SUPPORTED_WITH_DEDUPLICATION)')
    NTSTATUS__STATUS_NOT_SUPPORTED_WITH_AUDITING: Final[ThirtyTwoBits] = cached_hresult(0xD00004CD, 'HRESULT_FROM_NT(STATUS_NOT_SUPPORTED_WITH_AUDITING)')
    NTSTATUS__STATUS_NOT_SUPPORTED_WITH_MONITORING: Final[ThirtyTwoBits] = cached_hresult(0xD00004CE, 'HRESULT_FROM_NT(STATUS_NOT_SUPPORTED_WITH_MONITORING)')
    NTSTATUS__STATUS_NOT_SUPPORTED_WITH_SNAPSHOT: Final[ThirtyTwoBits] = cached_hresult(0xD00004CF, 'HRESULT_FROM_NT(STATUS_NOT_SUPPORTED_WITH_SNAPSHOT)')
    NTSTATUS__STATUS_NOT_SUPPORTED_WITH_VIRTUALIZATION: Final[ThirtyTwoBits] = cached_hresult(0xD00004D0, 'HRESULT_FROM_NT(STATUS_NOT_SUPPORTED_WITH_VIRTUALIZATION)')
    NTSTATUS__STATUS_INDEX_OUT_OF_BOUNDS: Final[ThirtyTwoBits] = cached_hresult(0xD00004D1, 'HRESULT_FROM_NT(STATUS_INDEX_OUT_OF_BOUNDS)')
    NTSTATUS__STATUS_BYPASSIO_FLT_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD00004D2, 'HRESULT_FROM_NT(STATUS_BYPASSIO_FLT_NOT_SUPPORTED)')
    NTSTATUS__STATUS_VOLUME_WRITE_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0xD00004D3, 'HRESULT_FROM_NT(STATUS_VOLUME_WRITE_ACCESS_DENIED)')
    NTSTATUS__STATUS_PATCH_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0xD00004D4, 'HRESULT_FROM_NT(STATUS_PATCH_NOT_REGISTERED)')
    NTSTATUS__STATUS_NOT_SUPPORTED_WITH_CACHED_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0xD00004D5, 'HRESULT_FROM_NT(STATUS_NOT_SUPPORTED_WITH_CACHED_HANDLE)')
    NTSTATUS__STATUS_VOLUME_UPGRADE_NOT_NEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD00004D9, 'HRESULT_FROM_NT(STATUS_VOLUME_UPGRADE_NOT_NEEDED)')
    NTSTATUS__STATUS_VOLUME_UPGRADE_PENDING: Final[ThirtyTwoBits] = cached_hresult(0xD00004DA, 'HRESULT_FROM_NT(STATUS_VOLUME_UPGRADE_PENDING)')
    NTSTATUS__STATUS_VOLUME_UPGRADE_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD00004DB, 'HRESULT_FROM_NT(STATUS_VOLUME_UPGRADE_DISABLED)')
    NTSTATUS__STATUS_VOLUME_UPGRADE_DISABLED_TILL_OS_DOWNGRADE_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0xD00004DC, 'HRESULT_FROM_NT(STATUS_VOLUME_UPGRADE_DISABLED_TILL_OS_DOWNGRADE_EXPIRED)')
    NTSTATUS__STATUS_NOT_A_DEV_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0xD00004DD, 'HRESULT_FROM_NT(STATUS_NOT_A_DEV_VOLUME)')
    NTSTATUS__STATUS_FS_GUID_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD00004DE, 'HRESULT_FROM_NT(STATUS_FS_GUID_MISMATCH)')
    NTSTATUS__STATUS_CANT_ATTACH_TO_DEV_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0xD00004DF, 'HRESULT_FROM_NT(STATUS_CANT_ATTACH_TO_DEV_VOLUME)')
    NTSTATUS__STATUS_INVALID_CONFIG_VALUE: Final[ThirtyTwoBits] = cached_hresult(0xD00004E0, 'HRESULT_FROM_NT(STATUS_INVALID_CONFIG_VALUE)')
    NTSTATUS__STATUS_INVALID_TASK_NAME: Final[ThirtyTwoBits] = cached_hresult(0xD0000500, 'HRESULT_FROM_NT(STATUS_INVALID_TASK_NAME)')
    NTSTATUS__STATUS_INVALID_TASK_INDEX: Final[ThirtyTwoBits] = cached_hresult(0xD0000501, 'HRESULT_FROM_NT(STATUS_INVALID_TASK_INDEX)')
    NTSTATUS__STATUS_THREAD_ALREADY_IN_TASK: Final[ThirtyTwoBits] = cached_hresult(0xD0000502, 'HRESULT_FROM_NT(STATUS_THREAD_ALREADY_IN_TASK)')
    NTSTATUS__STATUS_CALLBACK_BYPASS: Final[ThirtyTwoBits] = cached_hresult(0xD0000503, 'HRESULT_FROM_NT(STATUS_CALLBACK_BYPASS)')
    NTSTATUS__STATUS_UNDEFINED_SCOPE: Final[ThirtyTwoBits] = cached_hresult(0xD0000504, 'HRESULT_FROM_NT(STATUS_UNDEFINED_SCOPE)')
    NTSTATUS__STATUS_INVALID_CAP: Final[ThirtyTwoBits] = cached_hresult(0xD0000505, 'HRESULT_FROM_NT(STATUS_INVALID_CAP)')
    NTSTATUS__STATUS_NOT_GUI_PROCESS: Final[ThirtyTwoBits] = cached_hresult(0xD0000506, 'HRESULT_FROM_NT(STATUS_NOT_GUI_PROCESS)')
    NTSTATUS__STATUS_DEVICE_HUNG: Final[ThirtyTwoBits] = cached_hresult(0xD0000507, 'HRESULT_FROM_NT(STATUS_DEVICE_HUNG)')
    NTSTATUS__STATUS_CONTAINER_ASSIGNED: Final[ThirtyTwoBits] = cached_hresult(0xD0000508, 'HRESULT_FROM_NT(STATUS_CONTAINER_ASSIGNED)')
    NTSTATUS__STATUS_JOB_NO_CONTAINER: Final[ThirtyTwoBits] = cached_hresult(0xD0000509, 'HRESULT_FROM_NT(STATUS_JOB_NO_CONTAINER)')
    NTSTATUS__STATUS_DEVICE_UNRESPONSIVE: Final[ThirtyTwoBits] = cached_hresult(0xD000050A, 'HRESULT_FROM_NT(STATUS_DEVICE_UNRESPONSIVE)')
    NTSTATUS__STATUS_REPARSE_POINT_ENCOUNTERED: Final[ThirtyTwoBits] = cached_hresult(0xD000050B, 'HRESULT_FROM_NT(STATUS_REPARSE_POINT_ENCOUNTERED)')
    NTSTATUS__STATUS_ATTRIBUTE_NOT_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0xD000050C, 'HRESULT_FROM_NT(STATUS_ATTRIBUTE_NOT_PRESENT)')
    NTSTATUS__STATUS_NOT_A_TIERED_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0xD000050D, 'HRESULT_FROM_NT(STATUS_NOT_A_TIERED_VOLUME)')
    NTSTATUS__STATUS_ALREADY_HAS_STREAM_ID: Final[ThirtyTwoBits] = cached_hresult(0xD000050E, 'HRESULT_FROM_NT(STATUS_ALREADY_HAS_STREAM_ID)')
    NTSTATUS__STATUS_JOB_NOT_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0xD000050F, 'HRESULT_FROM_NT(STATUS_JOB_NOT_EMPTY)')
    NTSTATUS__STATUS_ALREADY_INITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0xD0000510, 'HRESULT_FROM_NT(STATUS_ALREADY_INITIALIZED)')
    NTSTATUS__STATUS_ENCLAVE_NOT_TERMINATED: Final[ThirtyTwoBits] = cached_hresult(0xD0000511, 'HRESULT_FROM_NT(STATUS_ENCLAVE_NOT_TERMINATED)')
    NTSTATUS__STATUS_ENCLAVE_IS_TERMINATING: Final[ThirtyTwoBits] = cached_hresult(0xD0000512, 'HRESULT_FROM_NT(STATUS_ENCLAVE_IS_TERMINATING)')
    NTSTATUS__STATUS_SMB1_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD0000513, 'HRESULT_FROM_NT(STATUS_SMB1_NOT_AVAILABLE)')
    NTSTATUS__STATUS_SMR_GARBAGE_COLLECTION_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0xD0000514, 'HRESULT_FROM_NT(STATUS_SMR_GARBAGE_COLLECTION_REQUIRED)')
    NTSTATUS__STATUS_INTERRUPTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000515, 'HRESULT_FROM_NT(STATUS_INTERRUPTED)')
    NTSTATUS__STATUS_THREAD_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0xD0000516, 'HRESULT_FROM_NT(STATUS_THREAD_NOT_RUNNING)')
    NTSTATUS__STATUS_SESSION_KEY_TOO_SHORT: Final[ThirtyTwoBits] = cached_hresult(0xD0000517, 'HRESULT_FROM_NT(STATUS_SESSION_KEY_TOO_SHORT)')
    NTSTATUS__STATUS_FS_METADATA_INCONSISTENT: Final[ThirtyTwoBits] = cached_hresult(0xD0000518, 'HRESULT_FROM_NT(STATUS_FS_METADATA_INCONSISTENT)')
    NTSTATUS__STATUS_FAIL_FAST_EXCEPTION: Final[ThirtyTwoBits] = cached_hresult(0xD0000602, 'HRESULT_FROM_NT(STATUS_FAIL_FAST_EXCEPTION)')
    NTSTATUS__STATUS_IMAGE_CERT_REVOKED: Final[ThirtyTwoBits] = cached_hresult(0xD0000603, 'HRESULT_FROM_NT(STATUS_IMAGE_CERT_REVOKED)')
    NTSTATUS__STATUS_DYNAMIC_CODE_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0xD0000604, 'HRESULT_FROM_NT(STATUS_DYNAMIC_CODE_BLOCKED)')
    NTSTATUS__STATUS_IMAGE_CERT_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0xD0000605, 'HRESULT_FROM_NT(STATUS_IMAGE_CERT_EXPIRED)')
    NTSTATUS__STATUS_STRICT_CFG_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0xD0000606, 'HRESULT_FROM_NT(STATUS_STRICT_CFG_VIOLATION)')
    NTSTATUS__STATUS_SET_CONTEXT_DENIED: Final[ThirtyTwoBits] = cached_hresult(0xD000060A, 'HRESULT_FROM_NT(STATUS_SET_CONTEXT_DENIED)')
    NTSTATUS__STATUS_CROSS_PARTITION_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0xD000060B, 'HRESULT_FROM_NT(STATUS_CROSS_PARTITION_VIOLATION)')
    NTSTATUS__STATUS_PORT_CLOSED: Final[ThirtyTwoBits] = cached_hresult(0xD0000700, 'HRESULT_FROM_NT(STATUS_PORT_CLOSED)')
    NTSTATUS__STATUS_MESSAGE_LOST: Final[ThirtyTwoBits] = cached_hresult(0xD0000701, 'HRESULT_FROM_NT(STATUS_MESSAGE_LOST)')
    NTSTATUS__STATUS_INVALID_MESSAGE: Final[ThirtyTwoBits] = cached_hresult(0xD0000702, 'HRESULT_FROM_NT(STATUS_INVALID_MESSAGE)')
    NTSTATUS__STATUS_REQUEST_CANCELED: Final[ThirtyTwoBits] = cached_hresult(0xD0000703, 'HRESULT_FROM_NT(STATUS_REQUEST_CANCELED)')
    NTSTATUS__STATUS_RECURSIVE_DISPATCH: Final[ThirtyTwoBits] = cached_hresult(0xD0000704, 'HRESULT_FROM_NT(STATUS_RECURSIVE_DISPATCH)')
    NTSTATUS__STATUS_LPC_RECEIVE_BUFFER_EXPECTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000705, 'HRESULT_FROM_NT(STATUS_LPC_RECEIVE_BUFFER_EXPECTED)')
    NTSTATUS__STATUS_LPC_INVALID_CONNECTION_USAGE: Final[ThirtyTwoBits] = cached_hresult(0xD0000706, 'HRESULT_FROM_NT(STATUS_LPC_INVALID_CONNECTION_USAGE)')
    NTSTATUS__STATUS_LPC_REQUESTS_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0xD0000707, 'HRESULT_FROM_NT(STATUS_LPC_REQUESTS_NOT_ALLOWED)')
    NTSTATUS__STATUS_RESOURCE_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0xD0000708, 'HRESULT_FROM_NT(STATUS_RESOURCE_IN_USE)')
    NTSTATUS__STATUS_HARDWARE_MEMORY_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0000709, 'HRESULT_FROM_NT(STATUS_HARDWARE_MEMORY_ERROR)')
    NTSTATUS__STATUS_THREADPOOL_HANDLE_EXCEPTION: Final[ThirtyTwoBits] = cached_hresult(0xD000070A, 'HRESULT_FROM_NT(STATUS_THREADPOOL_HANDLE_EXCEPTION)')
    NTSTATUS__STATUS_THREADPOOL_SET_EVENT_ON_COMPLETION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD000070B, 'HRESULT_FROM_NT(STATUS_THREADPOOL_SET_EVENT_ON_COMPLETION_FAILED)')
    NTSTATUS__STATUS_THREADPOOL_RELEASE_SEMAPHORE_ON_COMPLETION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD000070C, 'HRESULT_FROM_NT(STATUS_THREADPOOL_RELEASE_SEMAPHORE_ON_COMPLETION_FAILED)')
    NTSTATUS__STATUS_THREADPOOL_RELEASE_MUTEX_ON_COMPLETION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD000070D, 'HRESULT_FROM_NT(STATUS_THREADPOOL_RELEASE_MUTEX_ON_COMPLETION_FAILED)')
    NTSTATUS__STATUS_THREADPOOL_FREE_LIBRARY_ON_COMPLETION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD000070E, 'HRESULT_FROM_NT(STATUS_THREADPOOL_FREE_LIBRARY_ON_COMPLETION_FAILED)')
    NTSTATUS__STATUS_THREADPOOL_RELEASED_DURING_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0xD000070F, 'HRESULT_FROM_NT(STATUS_THREADPOOL_RELEASED_DURING_OPERATION)')
    NTSTATUS__STATUS_CALLBACK_RETURNED_WHILE_IMPERSONATING: Final[ThirtyTwoBits] = cached_hresult(0xD0000710, 'HRESULT_FROM_NT(STATUS_CALLBACK_RETURNED_WHILE_IMPERSONATING)')
    NTSTATUS__STATUS_APC_RETURNED_WHILE_IMPERSONATING: Final[ThirtyTwoBits] = cached_hresult(0xD0000711, 'HRESULT_FROM_NT(STATUS_APC_RETURNED_WHILE_IMPERSONATING)')
    NTSTATUS__STATUS_PROCESS_IS_PROTECTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000712, 'HRESULT_FROM_NT(STATUS_PROCESS_IS_PROTECTED)')
    NTSTATUS__STATUS_MCA_EXCEPTION: Final[ThirtyTwoBits] = cached_hresult(0xD0000713, 'HRESULT_FROM_NT(STATUS_MCA_EXCEPTION)')
    NTSTATUS__STATUS_CERTIFICATE_MAPPING_NOT_UNIQUE: Final[ThirtyTwoBits] = cached_hresult(0xD0000714, 'HRESULT_FROM_NT(STATUS_CERTIFICATE_MAPPING_NOT_UNIQUE)')
    NTSTATUS__STATUS_SYMLINK_CLASS_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD0000715, 'HRESULT_FROM_NT(STATUS_SYMLINK_CLASS_DISABLED)')
    NTSTATUS__STATUS_INVALID_IDN_NORMALIZATION: Final[ThirtyTwoBits] = cached_hresult(0xD0000716, 'HRESULT_FROM_NT(STATUS_INVALID_IDN_NORMALIZATION)')
    NTSTATUS__STATUS_NO_UNICODE_TRANSLATION: Final[ThirtyTwoBits] = cached_hresult(0xD0000717, 'HRESULT_FROM_NT(STATUS_NO_UNICODE_TRANSLATION)')
    NTSTATUS__STATUS_ALREADY_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0xD0000718, 'HRESULT_FROM_NT(STATUS_ALREADY_REGISTERED)')
    NTSTATUS__STATUS_CONTEXT_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD0000719, 'HRESULT_FROM_NT(STATUS_CONTEXT_MISMATCH)')
    NTSTATUS__STATUS_PORT_ALREADY_HAS_COMPLETION_LIST: Final[ThirtyTwoBits] = cached_hresult(0xD000071A, 'HRESULT_FROM_NT(STATUS_PORT_ALREADY_HAS_COMPLETION_LIST)')
    NTSTATUS__STATUS_CALLBACK_RETURNED_THREAD_PRIORITY: Final[ThirtyTwoBits] = cached_hresult(0xD000071B, 'HRESULT_FROM_NT(STATUS_CALLBACK_RETURNED_THREAD_PRIORITY)')
    NTSTATUS__STATUS_INVALID_THREAD: Final[ThirtyTwoBits] = cached_hresult(0xD000071C, 'HRESULT_FROM_NT(STATUS_INVALID_THREAD)')
    NTSTATUS__STATUS_CALLBACK_RETURNED_TRANSACTION: Final[ThirtyTwoBits] = cached_hresult(0xD000071D, 'HRESULT_FROM_NT(STATUS_CALLBACK_RETURNED_TRANSACTION)')
    NTSTATUS__STATUS_CALLBACK_RETURNED_LDR_LOCK: Final[ThirtyTwoBits] = cached_hresult(0xD000071E, 'HRESULT_FROM_NT(STATUS_CALLBACK_RETURNED_LDR_LOCK)')
    NTSTATUS__STATUS_CALLBACK_RETURNED_LANG: Final[ThirtyTwoBits] = cached_hresult(0xD000071F, 'HRESULT_FROM_NT(STATUS_CALLBACK_RETURNED_LANG)')
    NTSTATUS__STATUS_CALLBACK_RETURNED_PRI_BACK: Final[ThirtyTwoBits] = cached_hresult(0xD0000720, 'HRESULT_FROM_NT(STATUS_CALLBACK_RETURNED_PRI_BACK)')
    NTSTATUS__STATUS_CALLBACK_RETURNED_THREAD_AFFINITY: Final[ThirtyTwoBits] = cached_hresult(0xD0000721, 'HRESULT_FROM_NT(STATUS_CALLBACK_RETURNED_THREAD_AFFINITY)')
    NTSTATUS__STATUS_LPC_HANDLE_COUNT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD0000722, 'HRESULT_FROM_NT(STATUS_LPC_HANDLE_COUNT_EXCEEDED)')
    NTSTATUS__STATUS_EXECUTABLE_MEMORY_WRITE: Final[ThirtyTwoBits] = cached_hresult(0xD0000723, 'HRESULT_FROM_NT(STATUS_EXECUTABLE_MEMORY_WRITE)')
    NTSTATUS__STATUS_KERNEL_EXECUTABLE_MEMORY_WRITE: Final[ThirtyTwoBits] = cached_hresult(0xD0000724, 'HRESULT_FROM_NT(STATUS_KERNEL_EXECUTABLE_MEMORY_WRITE)')
    NTSTATUS__STATUS_ATTACHED_EXECUTABLE_MEMORY_WRITE: Final[ThirtyTwoBits] = cached_hresult(0xD0000725, 'HRESULT_FROM_NT(STATUS_ATTACHED_EXECUTABLE_MEMORY_WRITE)')
    NTSTATUS__STATUS_TRIGGERED_EXECUTABLE_MEMORY_WRITE: Final[ThirtyTwoBits] = cached_hresult(0xD0000726, 'HRESULT_FROM_NT(STATUS_TRIGGERED_EXECUTABLE_MEMORY_WRITE)')
    NTSTATUS__STATUS_DISK_REPAIR_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD0000800, 'HRESULT_FROM_NT(STATUS_DISK_REPAIR_DISABLED)')
    NTSTATUS__STATUS_DS_DOMAIN_RENAME_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD0000801, 'HRESULT_FROM_NT(STATUS_DS_DOMAIN_RENAME_IN_PROGRESS)')
    NTSTATUS__STATUS_DISK_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD0000802, 'HRESULT_FROM_NT(STATUS_DISK_QUOTA_EXCEEDED)')
    NTSTATUS__STATUS_CONTENT_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0xD0000804, 'HRESULT_FROM_NT(STATUS_CONTENT_BLOCKED)')
    NTSTATUS__STATUS_BAD_CLUSTERS: Final[ThirtyTwoBits] = cached_hresult(0xD0000805, 'HRESULT_FROM_NT(STATUS_BAD_CLUSTERS)')
    NTSTATUS__STATUS_VOLUME_DIRTY: Final[ThirtyTwoBits] = cached_hresult(0xD0000806, 'HRESULT_FROM_NT(STATUS_VOLUME_DIRTY)')
    NTSTATUS__STATUS_DISK_REPAIR_UNSUCCESSFUL: Final[ThirtyTwoBits] = cached_hresult(0xD0000808, 'HRESULT_FROM_NT(STATUS_DISK_REPAIR_UNSUCCESSFUL)')
    NTSTATUS__STATUS_CORRUPT_LOG_OVERFULL: Final[ThirtyTwoBits] = cached_hresult(0xD0000809, 'HRESULT_FROM_NT(STATUS_CORRUPT_LOG_OVERFULL)')
    NTSTATUS__STATUS_CORRUPT_LOG_CORRUPTED: Final[ThirtyTwoBits] = cached_hresult(0xD000080A, 'HRESULT_FROM_NT(STATUS_CORRUPT_LOG_CORRUPTED)')
    NTSTATUS__STATUS_CORRUPT_LOG_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD000080B, 'HRESULT_FROM_NT(STATUS_CORRUPT_LOG_UNAVAILABLE)')
    NTSTATUS__STATUS_CORRUPT_LOG_DELETED_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD000080C, 'HRESULT_FROM_NT(STATUS_CORRUPT_LOG_DELETED_FULL)')
    NTSTATUS__STATUS_CORRUPT_LOG_CLEARED: Final[ThirtyTwoBits] = cached_hresult(0xD000080D, 'HRESULT_FROM_NT(STATUS_CORRUPT_LOG_CLEARED)')
    NTSTATUS__STATUS_ORPHAN_NAME_EXHAUSTED: Final[ThirtyTwoBits] = cached_hresult(0xD000080E, 'HRESULT_FROM_NT(STATUS_ORPHAN_NAME_EXHAUSTED)')
    NTSTATUS__STATUS_PROACTIVE_SCAN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD000080F, 'HRESULT_FROM_NT(STATUS_PROACTIVE_SCAN_IN_PROGRESS)')
    NTSTATUS__STATUS_ENCRYPTED_IO_NOT_POSSIBLE: Final[ThirtyTwoBits] = cached_hresult(0xD0000810, 'HRESULT_FROM_NT(STATUS_ENCRYPTED_IO_NOT_POSSIBLE)')
    NTSTATUS__STATUS_CORRUPT_LOG_UPLEVEL_RECORDS: Final[ThirtyTwoBits] = cached_hresult(0xD0000811, 'HRESULT_FROM_NT(STATUS_CORRUPT_LOG_UPLEVEL_RECORDS)')
    NTSTATUS__STATUS_FILE_CHECKED_OUT: Final[ThirtyTwoBits] = cached_hresult(0xD0000901, 'HRESULT_FROM_NT(STATUS_FILE_CHECKED_OUT)')
    NTSTATUS__STATUS_CHECKOUT_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0xD0000902, 'HRESULT_FROM_NT(STATUS_CHECKOUT_REQUIRED)')
    NTSTATUS__STATUS_BAD_FILE_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xD0000903, 'HRESULT_FROM_NT(STATUS_BAD_FILE_TYPE)')
    NTSTATUS__STATUS_FILE_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0xD0000904, 'HRESULT_FROM_NT(STATUS_FILE_TOO_LARGE)')
    NTSTATUS__STATUS_FORMS_AUTH_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0xD0000905, 'HRESULT_FROM_NT(STATUS_FORMS_AUTH_REQUIRED)')
    NTSTATUS__STATUS_VIRUS_INFECTED: Final[ThirtyTwoBits] = cached_hresult(0xD0000906, 'HRESULT_FROM_NT(STATUS_VIRUS_INFECTED)')
    NTSTATUS__STATUS_VIRUS_DELETED: Final[ThirtyTwoBits] = cached_hresult(0xD0000907, 'HRESULT_FROM_NT(STATUS_VIRUS_DELETED)')
    NTSTATUS__STATUS_BAD_MCFG_TABLE: Final[ThirtyTwoBits] = cached_hresult(0xD0000908, 'HRESULT_FROM_NT(STATUS_BAD_MCFG_TABLE)')
    NTSTATUS__STATUS_CANNOT_BREAK_OPLOCK: Final[ThirtyTwoBits] = cached_hresult(0xD0000909, 'HRESULT_FROM_NT(STATUS_CANNOT_BREAK_OPLOCK)')
    NTSTATUS__STATUS_BAD_KEY: Final[ThirtyTwoBits] = cached_hresult(0xD000090A, 'HRESULT_FROM_NT(STATUS_BAD_KEY)')
    NTSTATUS__STATUS_BAD_DATA: Final[ThirtyTwoBits] = cached_hresult(0xD000090B, 'HRESULT_FROM_NT(STATUS_BAD_DATA)')
    NTSTATUS__STATUS_NO_KEY: Final[ThirtyTwoBits] = cached_hresult(0xD000090C, 'HRESULT_FROM_NT(STATUS_NO_KEY)')
    NTSTATUS__STATUS_FILE_HANDLE_REVOKED: Final[ThirtyTwoBits] = cached_hresult(0xD0000910, 'HRESULT_FROM_NT(STATUS_FILE_HANDLE_REVOKED)')
    NTSTATUS__STATUS_SECTION_DIRECT_MAP_ONLY: Final[ThirtyTwoBits] = cached_hresult(0xD0000911, 'HRESULT_FROM_NT(STATUS_SECTION_DIRECT_MAP_ONLY)')
    NTSTATUS__STATUS_BLOCK_WEAK_REFERENCE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0000912, 'HRESULT_FROM_NT(STATUS_BLOCK_WEAK_REFERENCE_INVALID)')
    NTSTATUS__STATUS_BLOCK_SOURCE_WEAK_REFERENCE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0000913, 'HRESULT_FROM_NT(STATUS_BLOCK_SOURCE_WEAK_REFERENCE_INVALID)')
    NTSTATUS__STATUS_BLOCK_TARGET_WEAK_REFERENCE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0000914, 'HRESULT_FROM_NT(STATUS_BLOCK_TARGET_WEAK_REFERENCE_INVALID)')
    NTSTATUS__STATUS_BLOCK_SHARED: Final[ThirtyTwoBits] = cached_hresult(0xD0000915, 'HRESULT_FROM_NT(STATUS_BLOCK_SHARED)')
    NTSTATUS__STATUS_VRF_VOLATILE_CFG_AND_IO_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0xD0000C08, 'HRESULT_FROM_NT(STATUS_VRF_VOLATILE_CFG_AND_IO_ENABLED)')
    NTSTATUS__STATUS_VRF_VOLATILE_NOT_STOPPABLE: Final[ThirtyTwoBits] = cached_hresult(0xD0000C09, 'HRESULT_FROM_NT(STATUS_VRF_VOLATILE_NOT_STOPPABLE)')
    NTSTATUS__STATUS_VRF_VOLATILE_SAFE_MODE: Final[ThirtyTwoBits] = cached_hresult(0xD0000C0A, 'HRESULT_FROM_NT(STATUS_VRF_VOLATILE_SAFE_MODE)')
    NTSTATUS__STATUS_VRF_VOLATILE_NOT_RUNNABLE_SYSTEM: Final[ThirtyTwoBits] = cached_hresult(0xD0000C0B, 'HRESULT_FROM_NT(STATUS_VRF_VOLATILE_NOT_RUNNABLE_SYSTEM)')
    NTSTATUS__STATUS_VRF_VOLATILE_NOT_SUPPORTED_RULECLASS: Final[ThirtyTwoBits] = cached_hresult(0xD0000C0C, 'HRESULT_FROM_NT(STATUS_VRF_VOLATILE_NOT_SUPPORTED_RULECLASS)')
    NTSTATUS__STATUS_VRF_VOLATILE_PROTECTED_DRIVER: Final[ThirtyTwoBits] = cached_hresult(0xD0000C0D, 'HRESULT_FROM_NT(STATUS_VRF_VOLATILE_PROTECTED_DRIVER)')
    NTSTATUS__STATUS_VRF_VOLATILE_NMI_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0xD0000C0E, 'HRESULT_FROM_NT(STATUS_VRF_VOLATILE_NMI_REGISTERED)')
    NTSTATUS__STATUS_VRF_VOLATILE_SETTINGS_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0xD0000C0F, 'HRESULT_FROM_NT(STATUS_VRF_VOLATILE_SETTINGS_CONFLICT)')
    NTSTATUS__STATUS_DIF_IOCALLBACK_NOT_REPLACED: Final[ThirtyTwoBits] = cached_hresult(0xD0000C76, 'HRESULT_FROM_NT(STATUS_DIF_IOCALLBACK_NOT_REPLACED)')
    NTSTATUS__STATUS_DIF_LIVEDUMP_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD0000C77, 'HRESULT_FROM_NT(STATUS_DIF_LIVEDUMP_LIMIT_EXCEEDED)')
    NTSTATUS__STATUS_DIF_VOLATILE_SECTION_NOT_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0xD0000C78, 'HRESULT_FROM_NT(STATUS_DIF_VOLATILE_SECTION_NOT_LOCKED)')
    NTSTATUS__STATUS_DIF_VOLATILE_DRIVER_HOTPATCHED: Final[ThirtyTwoBits] = cached_hresult(0xD0000C79, 'HRESULT_FROM_NT(STATUS_DIF_VOLATILE_DRIVER_HOTPATCHED)')
    NTSTATUS__STATUS_DIF_VOLATILE_INVALID_INFO: Final[ThirtyTwoBits] = cached_hresult(0xD0000C7A, 'HRESULT_FROM_NT(STATUS_DIF_VOLATILE_INVALID_INFO)')
    NTSTATUS__STATUS_DIF_VOLATILE_DRIVER_IS_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0xD0000C7B, 'HRESULT_FROM_NT(STATUS_DIF_VOLATILE_DRIVER_IS_NOT_RUNNING)')
    NTSTATUS__STATUS_DIF_VOLATILE_PLUGIN_IS_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0xD0000C7C, 'HRESULT_FROM_NT(STATUS_DIF_VOLATILE_PLUGIN_IS_NOT_RUNNING)')
    NTSTATUS__STATUS_DIF_VOLATILE_PLUGIN_CHANGE_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0xD0000C7D, 'HRESULT_FROM_NT(STATUS_DIF_VOLATILE_PLUGIN_CHANGE_NOT_ALLOWED)')
    NTSTATUS__STATUS_DIF_VOLATILE_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0xD0000C7E, 'HRESULT_FROM_NT(STATUS_DIF_VOLATILE_NOT_ALLOWED)')
    NTSTATUS__STATUS_DIF_BINDING_API_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0000C7F, 'HRESULT_FROM_NT(STATUS_DIF_BINDING_API_NOT_FOUND)')
    NTSTATUS__STATUS_WOW_ASSERTION: Final[ThirtyTwoBits] = cached_hresult(0xD0009898, 'HRESULT_FROM_NT(STATUS_WOW_ASSERTION)')
    NTSTATUS__STATUS_INVALID_SIGNATURE: Final[ThirtyTwoBits] = cached_hresult(0xD000A000, 'HRESULT_FROM_NT(STATUS_INVALID_SIGNATURE)')
    NTSTATUS__STATUS_HMAC_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD000A001, 'HRESULT_FROM_NT(STATUS_HMAC_NOT_SUPPORTED)')
    NTSTATUS__STATUS_AUTH_TAG_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD000A002, 'HRESULT_FROM_NT(STATUS_AUTH_TAG_MISMATCH)')
    NTSTATUS__STATUS_INVALID_STATE_TRANSITION: Final[ThirtyTwoBits] = cached_hresult(0xD000A003, 'HRESULT_FROM_NT(STATUS_INVALID_STATE_TRANSITION)')
    NTSTATUS__STATUS_INVALID_KERNEL_INFO_VERSION: Final[ThirtyTwoBits] = cached_hresult(0xD000A004, 'HRESULT_FROM_NT(STATUS_INVALID_KERNEL_INFO_VERSION)')
    NTSTATUS__STATUS_INVALID_PEP_INFO_VERSION: Final[ThirtyTwoBits] = cached_hresult(0xD000A005, 'HRESULT_FROM_NT(STATUS_INVALID_PEP_INFO_VERSION)')
    NTSTATUS__STATUS_HANDLE_REVOKED: Final[ThirtyTwoBits] = cached_hresult(0xD000A006, 'HRESULT_FROM_NT(STATUS_HANDLE_REVOKED)')
    NTSTATUS__STATUS_EOF_ON_GHOSTED_RANGE: Final[ThirtyTwoBits] = cached_hresult(0xD000A007, 'HRESULT_FROM_NT(STATUS_EOF_ON_GHOSTED_RANGE)')
    NTSTATUS__STATUS_CC_NEEDS_CALLBACK_SECTION_DRAIN: Final[ThirtyTwoBits] = cached_hresult(0xD000A008, 'HRESULT_FROM_NT(STATUS_CC_NEEDS_CALLBACK_SECTION_DRAIN)')
    NTSTATUS__STATUS_IPSEC_QUEUE_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0xD000A010, 'HRESULT_FROM_NT(STATUS_IPSEC_QUEUE_OVERFLOW)')
    NTSTATUS__STATUS_ND_QUEUE_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0xD000A011, 'HRESULT_FROM_NT(STATUS_ND_QUEUE_OVERFLOW)')
    NTSTATUS__STATUS_HOPLIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD000A012, 'HRESULT_FROM_NT(STATUS_HOPLIMIT_EXCEEDED)')
    NTSTATUS__STATUS_PROTOCOL_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD000A013, 'HRESULT_FROM_NT(STATUS_PROTOCOL_NOT_SUPPORTED)')
    NTSTATUS__STATUS_FASTPATH_REJECTED: Final[ThirtyTwoBits] = cached_hresult(0xD000A014, 'HRESULT_FROM_NT(STATUS_FASTPATH_REJECTED)')
    NTSTATUS__STATUS_LOST_WRITEBEHIND_DATA_NETWORK_DISCONNECTED: Final[ThirtyTwoBits] = cached_hresult(0xD000A080, 'HRESULT_FROM_NT(STATUS_LOST_WRITEBEHIND_DATA_NETWORK_DISCONNECTED)')
    NTSTATUS__STATUS_LOST_WRITEBEHIND_DATA_NETWORK_SERVER_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD000A081, 'HRESULT_FROM_NT(STATUS_LOST_WRITEBEHIND_DATA_NETWORK_SERVER_ERROR)')
    NTSTATUS__STATUS_LOST_WRITEBEHIND_DATA_LOCAL_DISK_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD000A082, 'HRESULT_FROM_NT(STATUS_LOST_WRITEBEHIND_DATA_LOCAL_DISK_ERROR)')
    NTSTATUS__STATUS_XML_PARSE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD000A083, 'HRESULT_FROM_NT(STATUS_XML_PARSE_ERROR)')
    NTSTATUS__STATUS_XMLDSIG_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD000A084, 'HRESULT_FROM_NT(STATUS_XMLDSIG_ERROR)')
    NTSTATUS__STATUS_WRONG_COMPARTMENT: Final[ThirtyTwoBits] = cached_hresult(0xD000A085, 'HRESULT_FROM_NT(STATUS_WRONG_COMPARTMENT)')
    NTSTATUS__STATUS_AUTHIP_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD000A086, 'HRESULT_FROM_NT(STATUS_AUTHIP_FAILURE)')
    NTSTATUS__STATUS_DS_OID_MAPPED_GROUP_CANT_HAVE_MEMBERS: Final[ThirtyTwoBits] = cached_hresult(0xD000A087, 'HRESULT_FROM_NT(STATUS_DS_OID_MAPPED_GROUP_CANT_HAVE_MEMBERS)')
    NTSTATUS__STATUS_DS_OID_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD000A088, 'HRESULT_FROM_NT(STATUS_DS_OID_NOT_FOUND)')
    NTSTATUS__STATUS_INCORRECT_ACCOUNT_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xD000A089, 'HRESULT_FROM_NT(STATUS_INCORRECT_ACCOUNT_TYPE)')
    NTSTATUS__STATUS_LOCAL_POLICY_MODIFICATION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD000A08A, 'HRESULT_FROM_NT(STATUS_LOCAL_POLICY_MODIFICATION_NOT_SUPPORTED)')
    NTSTATUS__STATUS_POLICY_CONTROLLED_ACCOUNT: Final[ThirtyTwoBits] = cached_hresult(0xD000A08B, 'HRESULT_FROM_NT(STATUS_POLICY_CONTROLLED_ACCOUNT)')
    NTSTATUS__STATUS_LAPS_LEGACY_SCHEMA_MISSING: Final[ThirtyTwoBits] = cached_hresult(0xD000A08C, 'HRESULT_FROM_NT(STATUS_LAPS_LEGACY_SCHEMA_MISSING)')
    NTSTATUS__STATUS_LAPS_SCHEMA_MISSING: Final[ThirtyTwoBits] = cached_hresult(0xD000A08D, 'HRESULT_FROM_NT(STATUS_LAPS_SCHEMA_MISSING)')
    NTSTATUS__STATUS_LAPS_ENCRYPTION_REQUIRES_2016_DFL: Final[ThirtyTwoBits] = cached_hresult(0xD000A08E, 'HRESULT_FROM_NT(STATUS_LAPS_ENCRYPTION_REQUIRES_2016_DFL)')
    NTSTATUS__STATUS_HASH_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD000A100, 'HRESULT_FROM_NT(STATUS_HASH_NOT_SUPPORTED)')
    NTSTATUS__STATUS_HASH_NOT_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0xD000A101, 'HRESULT_FROM_NT(STATUS_HASH_NOT_PRESENT)')
    NTSTATUS__STATUS_SECONDARY_IC_PROVIDER_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0xD000A121, 'HRESULT_FROM_NT(STATUS_SECONDARY_IC_PROVIDER_NOT_REGISTERED)')
    NTSTATUS__STATUS_GPIO_CLIENT_INFORMATION_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD000A122, 'HRESULT_FROM_NT(STATUS_GPIO_CLIENT_INFORMATION_INVALID)')
    NTSTATUS__STATUS_GPIO_VERSION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD000A123, 'HRESULT_FROM_NT(STATUS_GPIO_VERSION_NOT_SUPPORTED)')
    NTSTATUS__STATUS_GPIO_INVALID_REGISTRATION_PACKET: Final[ThirtyTwoBits] = cached_hresult(0xD000A124, 'HRESULT_FROM_NT(STATUS_GPIO_INVALID_REGISTRATION_PACKET)')
    NTSTATUS__STATUS_GPIO_OPERATION_DENIED: Final[ThirtyTwoBits] = cached_hresult(0xD000A125, 'HRESULT_FROM_NT(STATUS_GPIO_OPERATION_DENIED)')
    NTSTATUS__STATUS_GPIO_INCOMPATIBLE_CONNECT_MODE: Final[ThirtyTwoBits] = cached_hresult(0xD000A126, 'HRESULT_FROM_NT(STATUS_GPIO_INCOMPATIBLE_CONNECT_MODE)')
    NTSTATUS__STATUS_CANNOT_SWITCH_RUNLEVEL: Final[ThirtyTwoBits] = cached_hresult(0xD000A141, 'HRESULT_FROM_NT(STATUS_CANNOT_SWITCH_RUNLEVEL)')
    NTSTATUS__STATUS_INVALID_RUNLEVEL_SETTING: Final[ThirtyTwoBits] = cached_hresult(0xD000A142, 'HRESULT_FROM_NT(STATUS_INVALID_RUNLEVEL_SETTING)')
    NTSTATUS__STATUS_RUNLEVEL_SWITCH_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0xD000A143, 'HRESULT_FROM_NT(STATUS_RUNLEVEL_SWITCH_TIMEOUT)')
    NTSTATUS__STATUS_RUNLEVEL_SWITCH_AGENT_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0xD000A145, 'HRESULT_FROM_NT(STATUS_RUNLEVEL_SWITCH_AGENT_TIMEOUT)')
    NTSTATUS__STATUS_RUNLEVEL_SWITCH_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD000A146, 'HRESULT_FROM_NT(STATUS_RUNLEVEL_SWITCH_IN_PROGRESS)')
    NTSTATUS__STATUS_NOT_APPCONTAINER: Final[ThirtyTwoBits] = cached_hresult(0xD000A200, 'HRESULT_FROM_NT(STATUS_NOT_APPCONTAINER)')
    NTSTATUS__STATUS_NOT_SUPPORTED_IN_APPCONTAINER: Final[ThirtyTwoBits] = cached_hresult(0xD000A201, 'HRESULT_FROM_NT(STATUS_NOT_SUPPORTED_IN_APPCONTAINER)')
    NTSTATUS__STATUS_INVALID_PACKAGE_SID_LENGTH: Final[ThirtyTwoBits] = cached_hresult(0xD000A202, 'HRESULT_FROM_NT(STATUS_INVALID_PACKAGE_SID_LENGTH)')
    NTSTATUS__STATUS_LPAC_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0xD000A203, 'HRESULT_FROM_NT(STATUS_LPAC_ACCESS_DENIED)')
    NTSTATUS__STATUS_ADMINLESS_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0xD000A204, 'HRESULT_FROM_NT(STATUS_ADMINLESS_ACCESS_DENIED)')
    NTSTATUS__STATUS_APP_DATA_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD000A281, 'HRESULT_FROM_NT(STATUS_APP_DATA_NOT_FOUND)')
    NTSTATUS__STATUS_APP_DATA_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0xD000A282, 'HRESULT_FROM_NT(STATUS_APP_DATA_EXPIRED)')
    NTSTATUS__STATUS_APP_DATA_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0xD000A283, 'HRESULT_FROM_NT(STATUS_APP_DATA_CORRUPT)')
    NTSTATUS__STATUS_APP_DATA_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD000A284, 'HRESULT_FROM_NT(STATUS_APP_DATA_LIMIT_EXCEEDED)')
    NTSTATUS__STATUS_APP_DATA_REBOOT_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0xD000A285, 'HRESULT_FROM_NT(STATUS_APP_DATA_REBOOT_REQUIRED)')
    NTSTATUS__STATUS_OFFLOAD_READ_FLT_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD000A2A1, 'HRESULT_FROM_NT(STATUS_OFFLOAD_READ_FLT_NOT_SUPPORTED)')
    NTSTATUS__STATUS_OFFLOAD_WRITE_FLT_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD000A2A2, 'HRESULT_FROM_NT(STATUS_OFFLOAD_WRITE_FLT_NOT_SUPPORTED)')
    NTSTATUS__STATUS_OFFLOAD_READ_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD000A2A3, 'HRESULT_FROM_NT(STATUS_OFFLOAD_READ_FILE_NOT_SUPPORTED)')
    NTSTATUS__STATUS_OFFLOAD_WRITE_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD000A2A4, 'HRESULT_FROM_NT(STATUS_OFFLOAD_WRITE_FILE_NOT_SUPPORTED)')
    NTSTATUS__STATUS_WOF_WIM_HEADER_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0xD000A2A5, 'HRESULT_FROM_NT(STATUS_WOF_WIM_HEADER_CORRUPT)')
    NTSTATUS__STATUS_WOF_WIM_RESOURCE_TABLE_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0xD000A2A6, 'HRESULT_FROM_NT(STATUS_WOF_WIM_RESOURCE_TABLE_CORRUPT)')
    NTSTATUS__STATUS_WOF_FILE_RESOURCE_TABLE_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0xD000A2A7, 'HRESULT_FROM_NT(STATUS_WOF_FILE_RESOURCE_TABLE_CORRUPT)')
    NTSTATUS__STATUS_CIMFS_IMAGE_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0xD000C001, 'HRESULT_FROM_NT(STATUS_CIMFS_IMAGE_CORRUPT)')
    NTSTATUS__STATUS_CIMFS_IMAGE_VERSION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD000C002, 'HRESULT_FROM_NT(STATUS_CIMFS_IMAGE_VERSION_NOT_SUPPORTED)')
    NTSTATUS__STATUS_FILE_SYSTEM_VIRTUALIZATION_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD000CE01, 'HRESULT_FROM_NT(STATUS_FILE_SYSTEM_VIRTUALIZATION_UNAVAILABLE)')
    NTSTATUS__STATUS_FILE_SYSTEM_VIRTUALIZATION_METADATA_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0xD000CE02, 'HRESULT_FROM_NT(STATUS_FILE_SYSTEM_VIRTUALIZATION_METADATA_CORRUPT)')
    NTSTATUS__STATUS_FILE_SYSTEM_VIRTUALIZATION_BUSY: Final[ThirtyTwoBits] = cached_hresult(0xD000CE03, 'HRESULT_FROM_NT(STATUS_FILE_SYSTEM_VIRTUALIZATION_BUSY)')
    NTSTATUS__STATUS_FILE_SYSTEM_VIRTUALIZATION_PROVIDER_UNKNOWN: Final[ThirtyTwoBits] = cached_hresult(0xD000CE04, 'HRESULT_FROM_NT(STATUS_FILE_SYSTEM_VIRTUALIZATION_PROVIDER_UNKNOWN)')
    NTSTATUS__STATUS_FILE_SYSTEM_VIRTUALIZATION_INVALID_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0xD000CE05, 'HRESULT_FROM_NT(STATUS_FILE_SYSTEM_VIRTUALIZATION_INVALID_OPERATION)')
    NTSTATUS__STATUS_CLOUD_FILE_SYNC_ROOT_METADATA_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0xD000CF00, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_SYNC_ROOT_METADATA_CORRUPT)')
    NTSTATUS__STATUS_CLOUD_FILE_PROVIDER_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0xD000CF01, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_PROVIDER_NOT_RUNNING)')
    NTSTATUS__STATUS_CLOUD_FILE_METADATA_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0xD000CF02, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_METADATA_CORRUPT)')
    NTSTATUS__STATUS_CLOUD_FILE_METADATA_TOO_LARGE: Final[ThirtyTwoBits] = cached_hresult(0xD000CF03, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_METADATA_TOO_LARGE)')
    NTSTATUS__STATUS_CLOUD_FILE_PROPERTY_VERSION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD000CF06, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_PROPERTY_VERSION_NOT_SUPPORTED)')
    NTSTATUS__STATUS_NOT_A_CLOUD_FILE: Final[ThirtyTwoBits] = cached_hresult(0xD000CF07, 'HRESULT_FROM_NT(STATUS_NOT_A_CLOUD_FILE)')
    NTSTATUS__STATUS_CLOUD_FILE_NOT_IN_SYNC: Final[ThirtyTwoBits] = cached_hresult(0xD000CF08, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_NOT_IN_SYNC)')
    NTSTATUS__STATUS_CLOUD_FILE_ALREADY_CONNECTED: Final[ThirtyTwoBits] = cached_hresult(0xD000CF09, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_ALREADY_CONNECTED)')
    NTSTATUS__STATUS_CLOUD_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD000CF0A, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_NOT_SUPPORTED)')
    NTSTATUS__STATUS_CLOUD_FILE_INVALID_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0xD000CF0B, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_INVALID_REQUEST)')
    NTSTATUS__STATUS_CLOUD_FILE_READ_ONLY_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0xD000CF0C, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_READ_ONLY_VOLUME)')
    NTSTATUS__STATUS_CLOUD_FILE_CONNECTED_PROVIDER_ONLY: Final[ThirtyTwoBits] = cached_hresult(0xD000CF0D, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_CONNECTED_PROVIDER_ONLY)')
    NTSTATUS__STATUS_CLOUD_FILE_VALIDATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD000CF0E, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_VALIDATION_FAILED)')
    NTSTATUS__STATUS_CLOUD_FILE_AUTHENTICATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD000CF0F, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_AUTHENTICATION_FAILED)')
    NTSTATUS__STATUS_CLOUD_FILE_INSUFFICIENT_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0xD000CF10, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_INSUFFICIENT_RESOURCES)')
    NTSTATUS__STATUS_CLOUD_FILE_NETWORK_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD000CF11, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_NETWORK_UNAVAILABLE)')
    NTSTATUS__STATUS_CLOUD_FILE_UNSUCCESSFUL: Final[ThirtyTwoBits] = cached_hresult(0xD000CF12, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_UNSUCCESSFUL)')
    NTSTATUS__STATUS_CLOUD_FILE_NOT_UNDER_SYNC_ROOT: Final[ThirtyTwoBits] = cached_hresult(0xD000CF13, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_NOT_UNDER_SYNC_ROOT)')
    NTSTATUS__STATUS_CLOUD_FILE_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0xD000CF14, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_IN_USE)')
    NTSTATUS__STATUS_CLOUD_FILE_PINNED: Final[ThirtyTwoBits] = cached_hresult(0xD000CF15, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_PINNED)')
    NTSTATUS__STATUS_CLOUD_FILE_REQUEST_ABORTED: Final[ThirtyTwoBits] = cached_hresult(0xD000CF16, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_REQUEST_ABORTED)')
    NTSTATUS__STATUS_CLOUD_FILE_PROPERTY_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0xD000CF17, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_PROPERTY_CORRUPT)')
    NTSTATUS__STATUS_CLOUD_FILE_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0xD000CF18, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_ACCESS_DENIED)')
    NTSTATUS__STATUS_CLOUD_FILE_INCOMPATIBLE_HARDLINKS: Final[ThirtyTwoBits] = cached_hresult(0xD000CF19, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_INCOMPATIBLE_HARDLINKS)')
    NTSTATUS__STATUS_CLOUD_FILE_PROPERTY_LOCK_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0xD000CF1A, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_PROPERTY_LOCK_CONFLICT)')
    NTSTATUS__STATUS_CLOUD_FILE_REQUEST_CANCELED: Final[ThirtyTwoBits] = cached_hresult(0xD000CF1B, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_REQUEST_CANCELED)')
    NTSTATUS__STATUS_CLOUD_FILE_PROVIDER_TERMINATED: Final[ThirtyTwoBits] = cached_hresult(0xD000CF1D, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_PROVIDER_TERMINATED)')
    NTSTATUS__STATUS_NOT_A_CLOUD_SYNC_ROOT: Final[ThirtyTwoBits] = cached_hresult(0xD000CF1E, 'HRESULT_FROM_NT(STATUS_NOT_A_CLOUD_SYNC_ROOT)')
    NTSTATUS__STATUS_CLOUD_FILE_REQUEST_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0xD000CF1F, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_REQUEST_TIMEOUT)')
    NTSTATUS__STATUS_CLOUD_FILE_DEHYDRATION_DISALLOWED: Final[ThirtyTwoBits] = cached_hresult(0xD000CF20, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_DEHYDRATION_DISALLOWED)')
    NTSTATUS__STATUS_CLOUD_FILE_US_MESSAGE_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0xD000CF21, 'HRESULT_FROM_NT(STATUS_CLOUD_FILE_US_MESSAGE_TIMEOUT)')
    NTSTATUS__STATUS_FILE_SNAP_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD000F500, 'HRESULT_FROM_NT(STATUS_FILE_SNAP_IN_PROGRESS)')
    NTSTATUS__STATUS_FILE_SNAP_USER_SECTION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD000F501, 'HRESULT_FROM_NT(STATUS_FILE_SNAP_USER_SECTION_NOT_SUPPORTED)')
    NTSTATUS__STATUS_FILE_SNAP_MODIFY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD000F502, 'HRESULT_FROM_NT(STATUS_FILE_SNAP_MODIFY_NOT_SUPPORTED)')
    NTSTATUS__STATUS_FILE_SNAP_IO_NOT_COORDINATED: Final[ThirtyTwoBits] = cached_hresult(0xD000F503, 'HRESULT_FROM_NT(STATUS_FILE_SNAP_IO_NOT_COORDINATED)')
    NTSTATUS__STATUS_FILE_SNAP_UNEXPECTED_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD000F504, 'HRESULT_FROM_NT(STATUS_FILE_SNAP_UNEXPECTED_ERROR)')
    NTSTATUS__STATUS_FILE_SNAP_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0xD000F505, 'HRESULT_FROM_NT(STATUS_FILE_SNAP_INVALID_PARAMETER)')
    NTSTATUS__STATUS_PNP_BAD_MPS_TABLE: Final[ThirtyTwoBits] = cached_hresult(0xD0040035, 'HRESULT_FROM_NT(STATUS_PNP_BAD_MPS_TABLE)')
    NTSTATUS__STATUS_PNP_TRANSLATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0040036, 'HRESULT_FROM_NT(STATUS_PNP_TRANSLATION_FAILED)')
    NTSTATUS__STATUS_PNP_IRQ_TRANSLATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0040037, 'HRESULT_FROM_NT(STATUS_PNP_IRQ_TRANSLATION_FAILED)')
    NTSTATUS__STATUS_PNP_INVALID_ID: Final[ThirtyTwoBits] = cached_hresult(0xD0040038, 'HRESULT_FROM_NT(STATUS_PNP_INVALID_ID)')
    NTSTATUS__STATUS_IO_REISSUE_AS_CACHED: Final[ThirtyTwoBits] = cached_hresult(0xD0040039, 'HRESULT_FROM_NT(STATUS_IO_REISSUE_AS_CACHED)')
    NTSTATUS__STATUS_CTX_WINSTATION_NAME_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD00A0001, 'HRESULT_FROM_NT(STATUS_CTX_WINSTATION_NAME_INVALID)')
    NTSTATUS__STATUS_CTX_INVALID_PD: Final[ThirtyTwoBits] = cached_hresult(0xD00A0002, 'HRESULT_FROM_NT(STATUS_CTX_INVALID_PD)')
    NTSTATUS__STATUS_CTX_PD_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD00A0003, 'HRESULT_FROM_NT(STATUS_CTX_PD_NOT_FOUND)')
    NTSTATUS__STATUS_CTX_CLOSE_PENDING: Final[ThirtyTwoBits] = cached_hresult(0xD00A0006, 'HRESULT_FROM_NT(STATUS_CTX_CLOSE_PENDING)')
    NTSTATUS__STATUS_CTX_NO_OUTBUF: Final[ThirtyTwoBits] = cached_hresult(0xD00A0007, 'HRESULT_FROM_NT(STATUS_CTX_NO_OUTBUF)')
    NTSTATUS__STATUS_CTX_MODEM_INF_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD00A0008, 'HRESULT_FROM_NT(STATUS_CTX_MODEM_INF_NOT_FOUND)')
    NTSTATUS__STATUS_CTX_INVALID_MODEMNAME: Final[ThirtyTwoBits] = cached_hresult(0xD00A0009, 'HRESULT_FROM_NT(STATUS_CTX_INVALID_MODEMNAME)')
    NTSTATUS__STATUS_CTX_RESPONSE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD00A000A, 'HRESULT_FROM_NT(STATUS_CTX_RESPONSE_ERROR)')
    NTSTATUS__STATUS_CTX_MODEM_RESPONSE_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0xD00A000B, 'HRESULT_FROM_NT(STATUS_CTX_MODEM_RESPONSE_TIMEOUT)')
    NTSTATUS__STATUS_CTX_MODEM_RESPONSE_NO_CARRIER: Final[ThirtyTwoBits] = cached_hresult(0xD00A000C, 'HRESULT_FROM_NT(STATUS_CTX_MODEM_RESPONSE_NO_CARRIER)')
    NTSTATUS__STATUS_CTX_MODEM_RESPONSE_NO_DIALTONE: Final[ThirtyTwoBits] = cached_hresult(0xD00A000D, 'HRESULT_FROM_NT(STATUS_CTX_MODEM_RESPONSE_NO_DIALTONE)')
    NTSTATUS__STATUS_CTX_MODEM_RESPONSE_BUSY: Final[ThirtyTwoBits] = cached_hresult(0xD00A000E, 'HRESULT_FROM_NT(STATUS_CTX_MODEM_RESPONSE_BUSY)')
    NTSTATUS__STATUS_CTX_MODEM_RESPONSE_VOICE: Final[ThirtyTwoBits] = cached_hresult(0xD00A000F, 'HRESULT_FROM_NT(STATUS_CTX_MODEM_RESPONSE_VOICE)')
    NTSTATUS__STATUS_CTX_TD_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD00A0010, 'HRESULT_FROM_NT(STATUS_CTX_TD_ERROR)')
    NTSTATUS__STATUS_CTX_LICENSE_CLIENT_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD00A0012, 'HRESULT_FROM_NT(STATUS_CTX_LICENSE_CLIENT_INVALID)')
    NTSTATUS__STATUS_CTX_LICENSE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD00A0013, 'HRESULT_FROM_NT(STATUS_CTX_LICENSE_NOT_AVAILABLE)')
    NTSTATUS__STATUS_CTX_LICENSE_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0xD00A0014, 'HRESULT_FROM_NT(STATUS_CTX_LICENSE_EXPIRED)')
    NTSTATUS__STATUS_CTX_WINSTATION_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD00A0015, 'HRESULT_FROM_NT(STATUS_CTX_WINSTATION_NOT_FOUND)')
    NTSTATUS__STATUS_CTX_WINSTATION_NAME_COLLISION: Final[ThirtyTwoBits] = cached_hresult(0xD00A0016, 'HRESULT_FROM_NT(STATUS_CTX_WINSTATION_NAME_COLLISION)')
    NTSTATUS__STATUS_CTX_WINSTATION_BUSY: Final[ThirtyTwoBits] = cached_hresult(0xD00A0017, 'HRESULT_FROM_NT(STATUS_CTX_WINSTATION_BUSY)')
    NTSTATUS__STATUS_CTX_BAD_VIDEO_MODE: Final[ThirtyTwoBits] = cached_hresult(0xD00A0018, 'HRESULT_FROM_NT(STATUS_CTX_BAD_VIDEO_MODE)')
    NTSTATUS__STATUS_CTX_GRAPHICS_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD00A0022, 'HRESULT_FROM_NT(STATUS_CTX_GRAPHICS_INVALID)')
    NTSTATUS__STATUS_CTX_NOT_CONSOLE: Final[ThirtyTwoBits] = cached_hresult(0xD00A0024, 'HRESULT_FROM_NT(STATUS_CTX_NOT_CONSOLE)')
    NTSTATUS__STATUS_CTX_CLIENT_QUERY_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0xD00A0026, 'HRESULT_FROM_NT(STATUS_CTX_CLIENT_QUERY_TIMEOUT)')
    NTSTATUS__STATUS_CTX_CONSOLE_DISCONNECT: Final[ThirtyTwoBits] = cached_hresult(0xD00A0027, 'HRESULT_FROM_NT(STATUS_CTX_CONSOLE_DISCONNECT)')
    NTSTATUS__STATUS_CTX_CONSOLE_CONNECT: Final[ThirtyTwoBits] = cached_hresult(0xD00A0028, 'HRESULT_FROM_NT(STATUS_CTX_CONSOLE_CONNECT)')
    NTSTATUS__STATUS_CTX_SHADOW_DENIED: Final[ThirtyTwoBits] = cached_hresult(0xD00A002A, 'HRESULT_FROM_NT(STATUS_CTX_SHADOW_DENIED)')
    NTSTATUS__STATUS_CTX_WINSTATION_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0xD00A002B, 'HRESULT_FROM_NT(STATUS_CTX_WINSTATION_ACCESS_DENIED)')
    NTSTATUS__STATUS_CTX_INVALID_WD: Final[ThirtyTwoBits] = cached_hresult(0xD00A002E, 'HRESULT_FROM_NT(STATUS_CTX_INVALID_WD)')
    NTSTATUS__STATUS_CTX_WD_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD00A002F, 'HRESULT_FROM_NT(STATUS_CTX_WD_NOT_FOUND)')
    NTSTATUS__STATUS_CTX_SHADOW_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD00A0030, 'HRESULT_FROM_NT(STATUS_CTX_SHADOW_INVALID)')
    NTSTATUS__STATUS_CTX_SHADOW_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD00A0031, 'HRESULT_FROM_NT(STATUS_CTX_SHADOW_DISABLED)')
    NTSTATUS__STATUS_RDP_PROTOCOL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD00A0032, 'HRESULT_FROM_NT(STATUS_RDP_PROTOCOL_ERROR)')
    NTSTATUS__STATUS_CTX_CLIENT_LICENSE_NOT_SET: Final[ThirtyTwoBits] = cached_hresult(0xD00A0033, 'HRESULT_FROM_NT(STATUS_CTX_CLIENT_LICENSE_NOT_SET)')
    NTSTATUS__STATUS_CTX_CLIENT_LICENSE_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0xD00A0034, 'HRESULT_FROM_NT(STATUS_CTX_CLIENT_LICENSE_IN_USE)')
    NTSTATUS__STATUS_CTX_SHADOW_ENDED_BY_MODE_CHANGE: Final[ThirtyTwoBits] = cached_hresult(0xD00A0035, 'HRESULT_FROM_NT(STATUS_CTX_SHADOW_ENDED_BY_MODE_CHANGE)')
    NTSTATUS__STATUS_CTX_SHADOW_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0xD00A0036, 'HRESULT_FROM_NT(STATUS_CTX_SHADOW_NOT_RUNNING)')
    NTSTATUS__STATUS_CTX_LOGON_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD00A0037, 'HRESULT_FROM_NT(STATUS_CTX_LOGON_DISABLED)')
    NTSTATUS__STATUS_CTX_SECURITY_LAYER_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD00A0038, 'HRESULT_FROM_NT(STATUS_CTX_SECURITY_LAYER_ERROR)')
    NTSTATUS__STATUS_TS_INCOMPATIBLE_SESSIONS: Final[ThirtyTwoBits] = cached_hresult(0xD00A0039, 'HRESULT_FROM_NT(STATUS_TS_INCOMPATIBLE_SESSIONS)')
    NTSTATUS__STATUS_TS_VIDEO_SUBSYSTEM_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD00A003A, 'HRESULT_FROM_NT(STATUS_TS_VIDEO_SUBSYSTEM_ERROR)')
    NTSTATUS__STATUS_MUI_FILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD00B0001, 'HRESULT_FROM_NT(STATUS_MUI_FILE_NOT_FOUND)')
    NTSTATUS__STATUS_MUI_INVALID_FILE: Final[ThirtyTwoBits] = cached_hresult(0xD00B0002, 'HRESULT_FROM_NT(STATUS_MUI_INVALID_FILE)')
    NTSTATUS__STATUS_MUI_INVALID_RC_CONFIG: Final[ThirtyTwoBits] = cached_hresult(0xD00B0003, 'HRESULT_FROM_NT(STATUS_MUI_INVALID_RC_CONFIG)')
    NTSTATUS__STATUS_MUI_INVALID_LOCALE_NAME: Final[ThirtyTwoBits] = cached_hresult(0xD00B0004, 'HRESULT_FROM_NT(STATUS_MUI_INVALID_LOCALE_NAME)')
    NTSTATUS__STATUS_MUI_INVALID_ULTIMATEFALLBACK_NAME: Final[ThirtyTwoBits] = cached_hresult(0xD00B0005, 'HRESULT_FROM_NT(STATUS_MUI_INVALID_ULTIMATEFALLBACK_NAME)')
    NTSTATUS__STATUS_MUI_FILE_NOT_LOADED: Final[ThirtyTwoBits] = cached_hresult(0xD00B0006, 'HRESULT_FROM_NT(STATUS_MUI_FILE_NOT_LOADED)')
    NTSTATUS__STATUS_RESOURCE_ENUM_USER_STOP: Final[ThirtyTwoBits] = cached_hresult(0xD00B0007, 'HRESULT_FROM_NT(STATUS_RESOURCE_ENUM_USER_STOP)')
    NTSTATUS__STATUS_CLUSTER_INVALID_NODE: Final[ThirtyTwoBits] = cached_hresult(0xD0130001, 'HRESULT_FROM_NT(STATUS_CLUSTER_INVALID_NODE)')
    NTSTATUS__STATUS_CLUSTER_NODE_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD0130002, 'HRESULT_FROM_NT(STATUS_CLUSTER_NODE_EXISTS)')
    NTSTATUS__STATUS_CLUSTER_JOIN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD0130003, 'HRESULT_FROM_NT(STATUS_CLUSTER_JOIN_IN_PROGRESS)')
    NTSTATUS__STATUS_CLUSTER_NODE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0130004, 'HRESULT_FROM_NT(STATUS_CLUSTER_NODE_NOT_FOUND)')
    NTSTATUS__STATUS_CLUSTER_LOCAL_NODE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0130005, 'HRESULT_FROM_NT(STATUS_CLUSTER_LOCAL_NODE_NOT_FOUND)')
    NTSTATUS__STATUS_CLUSTER_NETWORK_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD0130006, 'HRESULT_FROM_NT(STATUS_CLUSTER_NETWORK_EXISTS)')
    NTSTATUS__STATUS_CLUSTER_NETWORK_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0130007, 'HRESULT_FROM_NT(STATUS_CLUSTER_NETWORK_NOT_FOUND)')
    NTSTATUS__STATUS_CLUSTER_NETINTERFACE_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD0130008, 'HRESULT_FROM_NT(STATUS_CLUSTER_NETINTERFACE_EXISTS)')
    NTSTATUS__STATUS_CLUSTER_NETINTERFACE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0130009, 'HRESULT_FROM_NT(STATUS_CLUSTER_NETINTERFACE_NOT_FOUND)')
    NTSTATUS__STATUS_CLUSTER_INVALID_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0xD013000A, 'HRESULT_FROM_NT(STATUS_CLUSTER_INVALID_REQUEST)')
    NTSTATUS__STATUS_CLUSTER_INVALID_NETWORK_PROVIDER: Final[ThirtyTwoBits] = cached_hresult(0xD013000B, 'HRESULT_FROM_NT(STATUS_CLUSTER_INVALID_NETWORK_PROVIDER)')
    NTSTATUS__STATUS_CLUSTER_NODE_DOWN: Final[ThirtyTwoBits] = cached_hresult(0xD013000C, 'HRESULT_FROM_NT(STATUS_CLUSTER_NODE_DOWN)')
    NTSTATUS__STATUS_CLUSTER_NODE_UNREACHABLE: Final[ThirtyTwoBits] = cached_hresult(0xD013000D, 'HRESULT_FROM_NT(STATUS_CLUSTER_NODE_UNREACHABLE)')
    NTSTATUS__STATUS_CLUSTER_NODE_NOT_MEMBER: Final[ThirtyTwoBits] = cached_hresult(0xD013000E, 'HRESULT_FROM_NT(STATUS_CLUSTER_NODE_NOT_MEMBER)')
    NTSTATUS__STATUS_CLUSTER_JOIN_NOT_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD013000F, 'HRESULT_FROM_NT(STATUS_CLUSTER_JOIN_NOT_IN_PROGRESS)')
    NTSTATUS__STATUS_CLUSTER_INVALID_NETWORK: Final[ThirtyTwoBits] = cached_hresult(0xD0130010, 'HRESULT_FROM_NT(STATUS_CLUSTER_INVALID_NETWORK)')
    NTSTATUS__STATUS_CLUSTER_NO_NET_ADAPTERS: Final[ThirtyTwoBits] = cached_hresult(0xD0130011, 'HRESULT_FROM_NT(STATUS_CLUSTER_NO_NET_ADAPTERS)')
    NTSTATUS__STATUS_CLUSTER_NODE_UP: Final[ThirtyTwoBits] = cached_hresult(0xD0130012, 'HRESULT_FROM_NT(STATUS_CLUSTER_NODE_UP)')
    NTSTATUS__STATUS_CLUSTER_NODE_PAUSED: Final[ThirtyTwoBits] = cached_hresult(0xD0130013, 'HRESULT_FROM_NT(STATUS_CLUSTER_NODE_PAUSED)')
    NTSTATUS__STATUS_CLUSTER_NODE_NOT_PAUSED: Final[ThirtyTwoBits] = cached_hresult(0xD0130014, 'HRESULT_FROM_NT(STATUS_CLUSTER_NODE_NOT_PAUSED)')
    NTSTATUS__STATUS_CLUSTER_NO_SECURITY_CONTEXT: Final[ThirtyTwoBits] = cached_hresult(0xD0130015, 'HRESULT_FROM_NT(STATUS_CLUSTER_NO_SECURITY_CONTEXT)')
    NTSTATUS__STATUS_CLUSTER_NETWORK_NOT_INTERNAL: Final[ThirtyTwoBits] = cached_hresult(0xD0130016, 'HRESULT_FROM_NT(STATUS_CLUSTER_NETWORK_NOT_INTERNAL)')
    NTSTATUS__STATUS_CLUSTER_POISONED: Final[ThirtyTwoBits] = cached_hresult(0xD0130017, 'HRESULT_FROM_NT(STATUS_CLUSTER_POISONED)')
    NTSTATUS__STATUS_CLUSTER_NON_CSV_PATH: Final[ThirtyTwoBits] = cached_hresult(0xD0130018, 'HRESULT_FROM_NT(STATUS_CLUSTER_NON_CSV_PATH)')
    NTSTATUS__STATUS_CLUSTER_CSV_VOLUME_NOT_LOCAL: Final[ThirtyTwoBits] = cached_hresult(0xD0130019, 'HRESULT_FROM_NT(STATUS_CLUSTER_CSV_VOLUME_NOT_LOCAL)')
    NTSTATUS__STATUS_CLUSTER_CSV_READ_OPLOCK_BREAK_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD0130020, 'HRESULT_FROM_NT(STATUS_CLUSTER_CSV_READ_OPLOCK_BREAK_IN_PROGRESS)')
    NTSTATUS__STATUS_CLUSTER_CSV_AUTO_PAUSE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0130021, 'HRESULT_FROM_NT(STATUS_CLUSTER_CSV_AUTO_PAUSE_ERROR)')
    NTSTATUS__STATUS_CLUSTER_CSV_REDIRECTED: Final[ThirtyTwoBits] = cached_hresult(0xD0130022, 'HRESULT_FROM_NT(STATUS_CLUSTER_CSV_REDIRECTED)')
    NTSTATUS__STATUS_CLUSTER_CSV_NOT_REDIRECTED: Final[ThirtyTwoBits] = cached_hresult(0xD0130023, 'HRESULT_FROM_NT(STATUS_CLUSTER_CSV_NOT_REDIRECTED)')
    NTSTATUS__STATUS_CLUSTER_CSV_VOLUME_DRAINING: Final[ThirtyTwoBits] = cached_hresult(0xD0130024, 'HRESULT_FROM_NT(STATUS_CLUSTER_CSV_VOLUME_DRAINING)')
    NTSTATUS__STATUS_CLUSTER_CSV_SNAPSHOT_CREATION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD0130025, 'HRESULT_FROM_NT(STATUS_CLUSTER_CSV_SNAPSHOT_CREATION_IN_PROGRESS)')
    NTSTATUS__STATUS_CLUSTER_CSV_VOLUME_DRAINING_SUCCEEDED_DOWNLEVEL: Final[ThirtyTwoBits] = cached_hresult(0xD0130026, 'HRESULT_FROM_NT(STATUS_CLUSTER_CSV_VOLUME_DRAINING_SUCCEEDED_DOWNLEVEL)')
    NTSTATUS__STATUS_CLUSTER_CSV_NO_SNAPSHOTS: Final[ThirtyTwoBits] = cached_hresult(0xD0130027, 'HRESULT_FROM_NT(STATUS_CLUSTER_CSV_NO_SNAPSHOTS)')
    NTSTATUS__STATUS_CSV_IO_PAUSE_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0xD0130028, 'HRESULT_FROM_NT(STATUS_CSV_IO_PAUSE_TIMEOUT)')
    NTSTATUS__STATUS_CLUSTER_CSV_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0xD0130029, 'HRESULT_FROM_NT(STATUS_CLUSTER_CSV_INVALID_HANDLE)')
    NTSTATUS__STATUS_CLUSTER_CSV_SUPPORTED_ONLY_ON_COORDINATOR: Final[ThirtyTwoBits] = cached_hresult(0xD0130030, 'HRESULT_FROM_NT(STATUS_CLUSTER_CSV_SUPPORTED_ONLY_ON_COORDINATOR)')
    NTSTATUS__STATUS_CLUSTER_CAM_TICKET_REPLAY_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0xD0130031, 'HRESULT_FROM_NT(STATUS_CLUSTER_CAM_TICKET_REPLAY_DETECTED)')
    NTSTATUS__STATUS_ACPI_INVALID_OPCODE: Final[ThirtyTwoBits] = cached_hresult(0xD0140001, 'HRESULT_FROM_NT(STATUS_ACPI_INVALID_OPCODE)')
    NTSTATUS__STATUS_ACPI_STACK_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0xD0140002, 'HRESULT_FROM_NT(STATUS_ACPI_STACK_OVERFLOW)')
    NTSTATUS__STATUS_ACPI_ASSERT_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0140003, 'HRESULT_FROM_NT(STATUS_ACPI_ASSERT_FAILED)')
    NTSTATUS__STATUS_ACPI_INVALID_INDEX: Final[ThirtyTwoBits] = cached_hresult(0xD0140004, 'HRESULT_FROM_NT(STATUS_ACPI_INVALID_INDEX)')
    NTSTATUS__STATUS_ACPI_INVALID_ARGUMENT: Final[ThirtyTwoBits] = cached_hresult(0xD0140005, 'HRESULT_FROM_NT(STATUS_ACPI_INVALID_ARGUMENT)')
    NTSTATUS__STATUS_ACPI_FATAL: Final[ThirtyTwoBits] = cached_hresult(0xD0140006, 'HRESULT_FROM_NT(STATUS_ACPI_FATAL)')
    NTSTATUS__STATUS_ACPI_INVALID_SUPERNAME: Final[ThirtyTwoBits] = cached_hresult(0xD0140007, 'HRESULT_FROM_NT(STATUS_ACPI_INVALID_SUPERNAME)')
    NTSTATUS__STATUS_ACPI_INVALID_ARGTYPE: Final[ThirtyTwoBits] = cached_hresult(0xD0140008, 'HRESULT_FROM_NT(STATUS_ACPI_INVALID_ARGTYPE)')
    NTSTATUS__STATUS_ACPI_INVALID_OBJTYPE: Final[ThirtyTwoBits] = cached_hresult(0xD0140009, 'HRESULT_FROM_NT(STATUS_ACPI_INVALID_OBJTYPE)')
    NTSTATUS__STATUS_ACPI_INVALID_TARGETTYPE: Final[ThirtyTwoBits] = cached_hresult(0xD014000A, 'HRESULT_FROM_NT(STATUS_ACPI_INVALID_TARGETTYPE)')
    NTSTATUS__STATUS_ACPI_INCORRECT_ARGUMENT_COUNT: Final[ThirtyTwoBits] = cached_hresult(0xD014000B, 'HRESULT_FROM_NT(STATUS_ACPI_INCORRECT_ARGUMENT_COUNT)')
    NTSTATUS__STATUS_ACPI_ADDRESS_NOT_MAPPED: Final[ThirtyTwoBits] = cached_hresult(0xD014000C, 'HRESULT_FROM_NT(STATUS_ACPI_ADDRESS_NOT_MAPPED)')
    NTSTATUS__STATUS_ACPI_INVALID_EVENTTYPE: Final[ThirtyTwoBits] = cached_hresult(0xD014000D, 'HRESULT_FROM_NT(STATUS_ACPI_INVALID_EVENTTYPE)')
    NTSTATUS__STATUS_ACPI_HANDLER_COLLISION: Final[ThirtyTwoBits] = cached_hresult(0xD014000E, 'HRESULT_FROM_NT(STATUS_ACPI_HANDLER_COLLISION)')
    NTSTATUS__STATUS_ACPI_INVALID_DATA: Final[ThirtyTwoBits] = cached_hresult(0xD014000F, 'HRESULT_FROM_NT(STATUS_ACPI_INVALID_DATA)')
    NTSTATUS__STATUS_ACPI_INVALID_REGION: Final[ThirtyTwoBits] = cached_hresult(0xD0140010, 'HRESULT_FROM_NT(STATUS_ACPI_INVALID_REGION)')
    NTSTATUS__STATUS_ACPI_INVALID_ACCESS_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD0140011, 'HRESULT_FROM_NT(STATUS_ACPI_INVALID_ACCESS_SIZE)')
    NTSTATUS__STATUS_ACPI_ACQUIRE_GLOBAL_LOCK: Final[ThirtyTwoBits] = cached_hresult(0xD0140012, 'HRESULT_FROM_NT(STATUS_ACPI_ACQUIRE_GLOBAL_LOCK)')
    NTSTATUS__STATUS_ACPI_ALREADY_INITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0xD0140013, 'HRESULT_FROM_NT(STATUS_ACPI_ALREADY_INITIALIZED)')
    NTSTATUS__STATUS_ACPI_NOT_INITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0xD0140014, 'HRESULT_FROM_NT(STATUS_ACPI_NOT_INITIALIZED)')
    NTSTATUS__STATUS_ACPI_INVALID_MUTEX_LEVEL: Final[ThirtyTwoBits] = cached_hresult(0xD0140015, 'HRESULT_FROM_NT(STATUS_ACPI_INVALID_MUTEX_LEVEL)')
    NTSTATUS__STATUS_ACPI_MUTEX_NOT_OWNED: Final[ThirtyTwoBits] = cached_hresult(0xD0140016, 'HRESULT_FROM_NT(STATUS_ACPI_MUTEX_NOT_OWNED)')
    NTSTATUS__STATUS_ACPI_MUTEX_NOT_OWNER: Final[ThirtyTwoBits] = cached_hresult(0xD0140017, 'HRESULT_FROM_NT(STATUS_ACPI_MUTEX_NOT_OWNER)')
    NTSTATUS__STATUS_ACPI_RS_ACCESS: Final[ThirtyTwoBits] = cached_hresult(0xD0140018, 'HRESULT_FROM_NT(STATUS_ACPI_RS_ACCESS)')
    NTSTATUS__STATUS_ACPI_INVALID_TABLE: Final[ThirtyTwoBits] = cached_hresult(0xD0140019, 'HRESULT_FROM_NT(STATUS_ACPI_INVALID_TABLE)')
    NTSTATUS__STATUS_ACPI_REG_HANDLER_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0140020, 'HRESULT_FROM_NT(STATUS_ACPI_REG_HANDLER_FAILED)')
    NTSTATUS__STATUS_ACPI_POWER_REQUEST_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0140021, 'HRESULT_FROM_NT(STATUS_ACPI_POWER_REQUEST_FAILED)')
    NTSTATUS__STATUS_SXS_SECTION_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0150001, 'HRESULT_FROM_NT(STATUS_SXS_SECTION_NOT_FOUND)')
    NTSTATUS__STATUS_SXS_CANT_GEN_ACTCTX: Final[ThirtyTwoBits] = cached_hresult(0xD0150002, 'HRESULT_FROM_NT(STATUS_SXS_CANT_GEN_ACTCTX)')
    NTSTATUS__STATUS_SXS_INVALID_ACTCTXDATA_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0xD0150003, 'HRESULT_FROM_NT(STATUS_SXS_INVALID_ACTCTXDATA_FORMAT)')
    NTSTATUS__STATUS_SXS_ASSEMBLY_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0150004, 'HRESULT_FROM_NT(STATUS_SXS_ASSEMBLY_NOT_FOUND)')
    NTSTATUS__STATUS_SXS_MANIFEST_FORMAT_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0150005, 'HRESULT_FROM_NT(STATUS_SXS_MANIFEST_FORMAT_ERROR)')
    NTSTATUS__STATUS_SXS_MANIFEST_PARSE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0150006, 'HRESULT_FROM_NT(STATUS_SXS_MANIFEST_PARSE_ERROR)')
    NTSTATUS__STATUS_SXS_ACTIVATION_CONTEXT_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD0150007, 'HRESULT_FROM_NT(STATUS_SXS_ACTIVATION_CONTEXT_DISABLED)')
    NTSTATUS__STATUS_SXS_KEY_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0150008, 'HRESULT_FROM_NT(STATUS_SXS_KEY_NOT_FOUND)')
    NTSTATUS__STATUS_SXS_VERSION_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0xD0150009, 'HRESULT_FROM_NT(STATUS_SXS_VERSION_CONFLICT)')
    NTSTATUS__STATUS_SXS_WRONG_SECTION_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xD015000A, 'HRESULT_FROM_NT(STATUS_SXS_WRONG_SECTION_TYPE)')
    NTSTATUS__STATUS_SXS_THREAD_QUERIES_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD015000B, 'HRESULT_FROM_NT(STATUS_SXS_THREAD_QUERIES_DISABLED)')
    NTSTATUS__STATUS_SXS_ASSEMBLY_MISSING: Final[ThirtyTwoBits] = cached_hresult(0xD015000C, 'HRESULT_FROM_NT(STATUS_SXS_ASSEMBLY_MISSING)')
    NTSTATUS__STATUS_SXS_PROCESS_DEFAULT_ALREADY_SET: Final[ThirtyTwoBits] = cached_hresult(0xD015000E, 'HRESULT_FROM_NT(STATUS_SXS_PROCESS_DEFAULT_ALREADY_SET)')
    NTSTATUS__STATUS_SXS_EARLY_DEACTIVATION: Final[ThirtyTwoBits] = cached_hresult(0xD015000F, 'HRESULT_FROM_NT(STATUS_SXS_EARLY_DEACTIVATION)')
    NTSTATUS__STATUS_SXS_INVALID_DEACTIVATION: Final[ThirtyTwoBits] = cached_hresult(0xD0150010, 'HRESULT_FROM_NT(STATUS_SXS_INVALID_DEACTIVATION)')
    NTSTATUS__STATUS_SXS_MULTIPLE_DEACTIVATION: Final[ThirtyTwoBits] = cached_hresult(0xD0150011, 'HRESULT_FROM_NT(STATUS_SXS_MULTIPLE_DEACTIVATION)')
    NTSTATUS__STATUS_SXS_SYSTEM_DEFAULT_ACTIVATION_CONTEXT_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0xD0150012, 'HRESULT_FROM_NT(STATUS_SXS_SYSTEM_DEFAULT_ACTIVATION_CONTEXT_EMPTY)')
    NTSTATUS__STATUS_SXS_PROCESS_TERMINATION_REQUESTED: Final[ThirtyTwoBits] = cached_hresult(0xD0150013, 'HRESULT_FROM_NT(STATUS_SXS_PROCESS_TERMINATION_REQUESTED)')
    NTSTATUS__STATUS_SXS_CORRUPT_ACTIVATION_STACK: Final[ThirtyTwoBits] = cached_hresult(0xD0150014, 'HRESULT_FROM_NT(STATUS_SXS_CORRUPT_ACTIVATION_STACK)')
    NTSTATUS__STATUS_SXS_CORRUPTION: Final[ThirtyTwoBits] = cached_hresult(0xD0150015, 'HRESULT_FROM_NT(STATUS_SXS_CORRUPTION)')
    NTSTATUS__STATUS_SXS_INVALID_IDENTITY_ATTRIBUTE_VALUE: Final[ThirtyTwoBits] = cached_hresult(0xD0150016, 'HRESULT_FROM_NT(STATUS_SXS_INVALID_IDENTITY_ATTRIBUTE_VALUE)')
    NTSTATUS__STATUS_SXS_INVALID_IDENTITY_ATTRIBUTE_NAME: Final[ThirtyTwoBits] = cached_hresult(0xD0150017, 'HRESULT_FROM_NT(STATUS_SXS_INVALID_IDENTITY_ATTRIBUTE_NAME)')
    NTSTATUS__STATUS_SXS_IDENTITY_DUPLICATE_ATTRIBUTE: Final[ThirtyTwoBits] = cached_hresult(0xD0150018, 'HRESULT_FROM_NT(STATUS_SXS_IDENTITY_DUPLICATE_ATTRIBUTE)')
    NTSTATUS__STATUS_SXS_IDENTITY_PARSE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0150019, 'HRESULT_FROM_NT(STATUS_SXS_IDENTITY_PARSE_ERROR)')
    NTSTATUS__STATUS_SXS_COMPONENT_STORE_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0xD015001A, 'HRESULT_FROM_NT(STATUS_SXS_COMPONENT_STORE_CORRUPT)')
    NTSTATUS__STATUS_SXS_FILE_HASH_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD015001B, 'HRESULT_FROM_NT(STATUS_SXS_FILE_HASH_MISMATCH)')
    NTSTATUS__STATUS_SXS_MANIFEST_IDENTITY_SAME_BUT_CONTENTS_DIFFERENT: Final[ThirtyTwoBits] = cached_hresult(0xD015001C, 'HRESULT_FROM_NT(STATUS_SXS_MANIFEST_IDENTITY_SAME_BUT_CONTENTS_DIFFERENT)')
    NTSTATUS__STATUS_SXS_IDENTITIES_DIFFERENT: Final[ThirtyTwoBits] = cached_hresult(0xD015001D, 'HRESULT_FROM_NT(STATUS_SXS_IDENTITIES_DIFFERENT)')
    NTSTATUS__STATUS_SXS_ASSEMBLY_IS_NOT_A_DEPLOYMENT: Final[ThirtyTwoBits] = cached_hresult(0xD015001E, 'HRESULT_FROM_NT(STATUS_SXS_ASSEMBLY_IS_NOT_A_DEPLOYMENT)')
    NTSTATUS__STATUS_SXS_FILE_NOT_PART_OF_ASSEMBLY: Final[ThirtyTwoBits] = cached_hresult(0xD015001F, 'HRESULT_FROM_NT(STATUS_SXS_FILE_NOT_PART_OF_ASSEMBLY)')
    NTSTATUS__STATUS_ADVANCED_INSTALLER_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0150020, 'HRESULT_FROM_NT(STATUS_ADVANCED_INSTALLER_FAILED)')
    NTSTATUS__STATUS_XML_ENCODING_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD0150021, 'HRESULT_FROM_NT(STATUS_XML_ENCODING_MISMATCH)')
    NTSTATUS__STATUS_SXS_MANIFEST_TOO_BIG: Final[ThirtyTwoBits] = cached_hresult(0xD0150022, 'HRESULT_FROM_NT(STATUS_SXS_MANIFEST_TOO_BIG)')
    NTSTATUS__STATUS_SXS_SETTING_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0xD0150023, 'HRESULT_FROM_NT(STATUS_SXS_SETTING_NOT_REGISTERED)')
    NTSTATUS__STATUS_SXS_TRANSACTION_CLOSURE_INCOMPLETE: Final[ThirtyTwoBits] = cached_hresult(0xD0150024, 'HRESULT_FROM_NT(STATUS_SXS_TRANSACTION_CLOSURE_INCOMPLETE)')
    NTSTATUS__STATUS_SMI_PRIMITIVE_INSTALLER_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0150025, 'HRESULT_FROM_NT(STATUS_SMI_PRIMITIVE_INSTALLER_FAILED)')
    NTSTATUS__STATUS_GENERIC_COMMAND_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0150026, 'HRESULT_FROM_NT(STATUS_GENERIC_COMMAND_FAILED)')
    NTSTATUS__STATUS_SXS_FILE_HASH_MISSING: Final[ThirtyTwoBits] = cached_hresult(0xD0150027, 'HRESULT_FROM_NT(STATUS_SXS_FILE_HASH_MISSING)')
    NTSTATUS__STATUS_TRANSACTIONAL_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0xD0190001, 'HRESULT_FROM_NT(STATUS_TRANSACTIONAL_CONFLICT)')
    NTSTATUS__STATUS_INVALID_TRANSACTION: Final[ThirtyTwoBits] = cached_hresult(0xD0190002, 'HRESULT_FROM_NT(STATUS_INVALID_TRANSACTION)')
    NTSTATUS__STATUS_TRANSACTION_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0xD0190003, 'HRESULT_FROM_NT(STATUS_TRANSACTION_NOT_ACTIVE)')
    NTSTATUS__STATUS_TM_INITIALIZATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0190004, 'HRESULT_FROM_NT(STATUS_TM_INITIALIZATION_FAILED)')
    NTSTATUS__STATUS_RM_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0xD0190005, 'HRESULT_FROM_NT(STATUS_RM_NOT_ACTIVE)')
    NTSTATUS__STATUS_RM_METADATA_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0xD0190006, 'HRESULT_FROM_NT(STATUS_RM_METADATA_CORRUPT)')
    NTSTATUS__STATUS_TRANSACTION_NOT_JOINED: Final[ThirtyTwoBits] = cached_hresult(0xD0190007, 'HRESULT_FROM_NT(STATUS_TRANSACTION_NOT_JOINED)')
    NTSTATUS__STATUS_DIRECTORY_NOT_RM: Final[ThirtyTwoBits] = cached_hresult(0xD0190008, 'HRESULT_FROM_NT(STATUS_DIRECTORY_NOT_RM)')
    NTSTATUS__STATUS_TRANSACTIONS_UNSUPPORTED_REMOTE: Final[ThirtyTwoBits] = cached_hresult(0xD019000A, 'HRESULT_FROM_NT(STATUS_TRANSACTIONS_UNSUPPORTED_REMOTE)')
    NTSTATUS__STATUS_LOG_RESIZE_INVALID_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD019000B, 'HRESULT_FROM_NT(STATUS_LOG_RESIZE_INVALID_SIZE)')
    NTSTATUS__STATUS_REMOTE_FILE_VERSION_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD019000C, 'HRESULT_FROM_NT(STATUS_REMOTE_FILE_VERSION_MISMATCH)')
    NTSTATUS__STATUS_CRM_PROTOCOL_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD019000F, 'HRESULT_FROM_NT(STATUS_CRM_PROTOCOL_ALREADY_EXISTS)')
    NTSTATUS__STATUS_TRANSACTION_PROPAGATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0190010, 'HRESULT_FROM_NT(STATUS_TRANSACTION_PROPAGATION_FAILED)')
    NTSTATUS__STATUS_CRM_PROTOCOL_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0190011, 'HRESULT_FROM_NT(STATUS_CRM_PROTOCOL_NOT_FOUND)')
    NTSTATUS__STATUS_TRANSACTION_SUPERIOR_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD0190012, 'HRESULT_FROM_NT(STATUS_TRANSACTION_SUPERIOR_EXISTS)')
    NTSTATUS__STATUS_TRANSACTION_REQUEST_NOT_VALID: Final[ThirtyTwoBits] = cached_hresult(0xD0190013, 'HRESULT_FROM_NT(STATUS_TRANSACTION_REQUEST_NOT_VALID)')
    NTSTATUS__STATUS_TRANSACTION_NOT_REQUESTED: Final[ThirtyTwoBits] = cached_hresult(0xD0190014, 'HRESULT_FROM_NT(STATUS_TRANSACTION_NOT_REQUESTED)')
    NTSTATUS__STATUS_TRANSACTION_ALREADY_ABORTED: Final[ThirtyTwoBits] = cached_hresult(0xD0190015, 'HRESULT_FROM_NT(STATUS_TRANSACTION_ALREADY_ABORTED)')
    NTSTATUS__STATUS_TRANSACTION_ALREADY_COMMITTED: Final[ThirtyTwoBits] = cached_hresult(0xD0190016, 'HRESULT_FROM_NT(STATUS_TRANSACTION_ALREADY_COMMITTED)')
    NTSTATUS__STATUS_TRANSACTION_INVALID_MARSHALL_BUFFER: Final[ThirtyTwoBits] = cached_hresult(0xD0190017, 'HRESULT_FROM_NT(STATUS_TRANSACTION_INVALID_MARSHALL_BUFFER)')
    NTSTATUS__STATUS_CURRENT_TRANSACTION_NOT_VALID: Final[ThirtyTwoBits] = cached_hresult(0xD0190018, 'HRESULT_FROM_NT(STATUS_CURRENT_TRANSACTION_NOT_VALID)')
    NTSTATUS__STATUS_LOG_GROWTH_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0190019, 'HRESULT_FROM_NT(STATUS_LOG_GROWTH_FAILED)')
    NTSTATUS__STATUS_OBJECT_NO_LONGER_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD0190021, 'HRESULT_FROM_NT(STATUS_OBJECT_NO_LONGER_EXISTS)')
    NTSTATUS__STATUS_STREAM_MINIVERSION_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0190022, 'HRESULT_FROM_NT(STATUS_STREAM_MINIVERSION_NOT_FOUND)')
    NTSTATUS__STATUS_STREAM_MINIVERSION_NOT_VALID: Final[ThirtyTwoBits] = cached_hresult(0xD0190023, 'HRESULT_FROM_NT(STATUS_STREAM_MINIVERSION_NOT_VALID)')
    NTSTATUS__STATUS_MINIVERSION_INACCESSIBLE_FROM_SPECIFIED_TRANSACTION: Final[ThirtyTwoBits] = cached_hresult(0xD0190024, 'HRESULT_FROM_NT(STATUS_MINIVERSION_INACCESSIBLE_FROM_SPECIFIED_TRANSACTION)')
    NTSTATUS__STATUS_CANT_OPEN_MINIVERSION_WITH_MODIFY_INTENT: Final[ThirtyTwoBits] = cached_hresult(0xD0190025, 'HRESULT_FROM_NT(STATUS_CANT_OPEN_MINIVERSION_WITH_MODIFY_INTENT)')
    NTSTATUS__STATUS_CANT_CREATE_MORE_STREAM_MINIVERSIONS: Final[ThirtyTwoBits] = cached_hresult(0xD0190026, 'HRESULT_FROM_NT(STATUS_CANT_CREATE_MORE_STREAM_MINIVERSIONS)')
    NTSTATUS__STATUS_HANDLE_NO_LONGER_VALID: Final[ThirtyTwoBits] = cached_hresult(0xD0190028, 'HRESULT_FROM_NT(STATUS_HANDLE_NO_LONGER_VALID)')
    NTSTATUS__STATUS_LOG_CORRUPTION_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0xD0190030, 'HRESULT_FROM_NT(STATUS_LOG_CORRUPTION_DETECTED)')
    NTSTATUS__STATUS_RM_DISCONNECTED: Final[ThirtyTwoBits] = cached_hresult(0xD0190032, 'HRESULT_FROM_NT(STATUS_RM_DISCONNECTED)')
    NTSTATUS__STATUS_ENLISTMENT_NOT_SUPERIOR: Final[ThirtyTwoBits] = cached_hresult(0xD0190033, 'HRESULT_FROM_NT(STATUS_ENLISTMENT_NOT_SUPERIOR)')
    NTSTATUS__STATUS_FILE_IDENTITY_NOT_PERSISTENT: Final[ThirtyTwoBits] = cached_hresult(0xD0190036, 'HRESULT_FROM_NT(STATUS_FILE_IDENTITY_NOT_PERSISTENT)')
    NTSTATUS__STATUS_CANT_BREAK_TRANSACTIONAL_DEPENDENCY: Final[ThirtyTwoBits] = cached_hresult(0xD0190037, 'HRESULT_FROM_NT(STATUS_CANT_BREAK_TRANSACTIONAL_DEPENDENCY)')
    NTSTATUS__STATUS_CANT_CROSS_RM_BOUNDARY: Final[ThirtyTwoBits] = cached_hresult(0xD0190038, 'HRESULT_FROM_NT(STATUS_CANT_CROSS_RM_BOUNDARY)')
    NTSTATUS__STATUS_TXF_DIR_NOT_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0xD0190039, 'HRESULT_FROM_NT(STATUS_TXF_DIR_NOT_EMPTY)')
    NTSTATUS__STATUS_INDOUBT_TRANSACTIONS_EXIST: Final[ThirtyTwoBits] = cached_hresult(0xD019003A, 'HRESULT_FROM_NT(STATUS_INDOUBT_TRANSACTIONS_EXIST)')
    NTSTATUS__STATUS_TM_VOLATILE: Final[ThirtyTwoBits] = cached_hresult(0xD019003B, 'HRESULT_FROM_NT(STATUS_TM_VOLATILE)')
    NTSTATUS__STATUS_ROLLBACK_TIMER_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0xD019003C, 'HRESULT_FROM_NT(STATUS_ROLLBACK_TIMER_EXPIRED)')
    NTSTATUS__STATUS_TXF_ATTRIBUTE_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0xD019003D, 'HRESULT_FROM_NT(STATUS_TXF_ATTRIBUTE_CORRUPT)')
    NTSTATUS__STATUS_EFS_NOT_ALLOWED_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_hresult(0xD019003E, 'HRESULT_FROM_NT(STATUS_EFS_NOT_ALLOWED_IN_TRANSACTION)')
    NTSTATUS__STATUS_TRANSACTIONAL_OPEN_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0xD019003F, 'HRESULT_FROM_NT(STATUS_TRANSACTIONAL_OPEN_NOT_ALLOWED)')
    NTSTATUS__STATUS_TRANSACTED_MAPPING_UNSUPPORTED_REMOTE: Final[ThirtyTwoBits] = cached_hresult(0xD0190040, 'HRESULT_FROM_NT(STATUS_TRANSACTED_MAPPING_UNSUPPORTED_REMOTE)')
    NTSTATUS__STATUS_TRANSACTION_REQUIRED_PROMOTION: Final[ThirtyTwoBits] = cached_hresult(0xD0190043, 'HRESULT_FROM_NT(STATUS_TRANSACTION_REQUIRED_PROMOTION)')
    NTSTATUS__STATUS_CANNOT_EXECUTE_FILE_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_hresult(0xD0190044, 'HRESULT_FROM_NT(STATUS_CANNOT_EXECUTE_FILE_IN_TRANSACTION)')
    NTSTATUS__STATUS_TRANSACTIONS_NOT_FROZEN: Final[ThirtyTwoBits] = cached_hresult(0xD0190045, 'HRESULT_FROM_NT(STATUS_TRANSACTIONS_NOT_FROZEN)')
    NTSTATUS__STATUS_TRANSACTION_FREEZE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD0190046, 'HRESULT_FROM_NT(STATUS_TRANSACTION_FREEZE_IN_PROGRESS)')
    NTSTATUS__STATUS_NOT_SNAPSHOT_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0xD0190047, 'HRESULT_FROM_NT(STATUS_NOT_SNAPSHOT_VOLUME)')
    NTSTATUS__STATUS_NO_SAVEPOINT_WITH_OPEN_FILES: Final[ThirtyTwoBits] = cached_hresult(0xD0190048, 'HRESULT_FROM_NT(STATUS_NO_SAVEPOINT_WITH_OPEN_FILES)')
    NTSTATUS__STATUS_SPARSE_NOT_ALLOWED_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_hresult(0xD0190049, 'HRESULT_FROM_NT(STATUS_SPARSE_NOT_ALLOWED_IN_TRANSACTION)')
    NTSTATUS__STATUS_TM_IDENTITY_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD019004A, 'HRESULT_FROM_NT(STATUS_TM_IDENTITY_MISMATCH)')
    NTSTATUS__STATUS_FLOATED_SECTION: Final[ThirtyTwoBits] = cached_hresult(0xD019004B, 'HRESULT_FROM_NT(STATUS_FLOATED_SECTION)')
    NTSTATUS__STATUS_CANNOT_ACCEPT_TRANSACTED_WORK: Final[ThirtyTwoBits] = cached_hresult(0xD019004C, 'HRESULT_FROM_NT(STATUS_CANNOT_ACCEPT_TRANSACTED_WORK)')
    NTSTATUS__STATUS_CANNOT_ABORT_TRANSACTIONS: Final[ThirtyTwoBits] = cached_hresult(0xD019004D, 'HRESULT_FROM_NT(STATUS_CANNOT_ABORT_TRANSACTIONS)')
    NTSTATUS__STATUS_TRANSACTION_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD019004E, 'HRESULT_FROM_NT(STATUS_TRANSACTION_NOT_FOUND)')
    NTSTATUS__STATUS_RESOURCEMANAGER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD019004F, 'HRESULT_FROM_NT(STATUS_RESOURCEMANAGER_NOT_FOUND)')
    NTSTATUS__STATUS_ENLISTMENT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0190050, 'HRESULT_FROM_NT(STATUS_ENLISTMENT_NOT_FOUND)')
    NTSTATUS__STATUS_TRANSACTIONMANAGER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0190051, 'HRESULT_FROM_NT(STATUS_TRANSACTIONMANAGER_NOT_FOUND)')
    NTSTATUS__STATUS_TRANSACTIONMANAGER_NOT_ONLINE: Final[ThirtyTwoBits] = cached_hresult(0xD0190052, 'HRESULT_FROM_NT(STATUS_TRANSACTIONMANAGER_NOT_ONLINE)')
    NTSTATUS__STATUS_TRANSACTIONMANAGER_RECOVERY_NAME_COLLISION: Final[ThirtyTwoBits] = cached_hresult(0xD0190053, 'HRESULT_FROM_NT(STATUS_TRANSACTIONMANAGER_RECOVERY_NAME_COLLISION)')
    NTSTATUS__STATUS_TRANSACTION_NOT_ROOT: Final[ThirtyTwoBits] = cached_hresult(0xD0190054, 'HRESULT_FROM_NT(STATUS_TRANSACTION_NOT_ROOT)')
    NTSTATUS__STATUS_TRANSACTION_OBJECT_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0xD0190055, 'HRESULT_FROM_NT(STATUS_TRANSACTION_OBJECT_EXPIRED)')
    NTSTATUS__STATUS_COMPRESSION_NOT_ALLOWED_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_hresult(0xD0190056, 'HRESULT_FROM_NT(STATUS_COMPRESSION_NOT_ALLOWED_IN_TRANSACTION)')
    NTSTATUS__STATUS_TRANSACTION_RESPONSE_NOT_ENLISTED: Final[ThirtyTwoBits] = cached_hresult(0xD0190057, 'HRESULT_FROM_NT(STATUS_TRANSACTION_RESPONSE_NOT_ENLISTED)')
    NTSTATUS__STATUS_TRANSACTION_RECORD_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0xD0190058, 'HRESULT_FROM_NT(STATUS_TRANSACTION_RECORD_TOO_LONG)')
    NTSTATUS__STATUS_NO_LINK_TRACKING_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_hresult(0xD0190059, 'HRESULT_FROM_NT(STATUS_NO_LINK_TRACKING_IN_TRANSACTION)')
    NTSTATUS__STATUS_OPERATION_NOT_SUPPORTED_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_hresult(0xD019005A, 'HRESULT_FROM_NT(STATUS_OPERATION_NOT_SUPPORTED_IN_TRANSACTION)')
    NTSTATUS__STATUS_TRANSACTION_INTEGRITY_VIOLATED: Final[ThirtyTwoBits] = cached_hresult(0xD019005B, 'HRESULT_FROM_NT(STATUS_TRANSACTION_INTEGRITY_VIOLATED)')
    NTSTATUS__STATUS_TRANSACTIONMANAGER_IDENTITY_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD019005C, 'HRESULT_FROM_NT(STATUS_TRANSACTIONMANAGER_IDENTITY_MISMATCH)')
    NTSTATUS__STATUS_RM_CANNOT_BE_FROZEN_FOR_SNAPSHOT: Final[ThirtyTwoBits] = cached_hresult(0xD019005D, 'HRESULT_FROM_NT(STATUS_RM_CANNOT_BE_FROZEN_FOR_SNAPSHOT)')
    NTSTATUS__STATUS_TRANSACTION_MUST_WRITETHROUGH: Final[ThirtyTwoBits] = cached_hresult(0xD019005E, 'HRESULT_FROM_NT(STATUS_TRANSACTION_MUST_WRITETHROUGH)')
    NTSTATUS__STATUS_TRANSACTION_NO_SUPERIOR: Final[ThirtyTwoBits] = cached_hresult(0xD019005F, 'HRESULT_FROM_NT(STATUS_TRANSACTION_NO_SUPERIOR)')
    NTSTATUS__STATUS_EXPIRED_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0xD0190060, 'HRESULT_FROM_NT(STATUS_EXPIRED_HANDLE)')
    NTSTATUS__STATUS_TRANSACTION_NOT_ENLISTED: Final[ThirtyTwoBits] = cached_hresult(0xD0190061, 'HRESULT_FROM_NT(STATUS_TRANSACTION_NOT_ENLISTED)')
    NTSTATUS__STATUS_LOG_SECTOR_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD01A0001, 'HRESULT_FROM_NT(STATUS_LOG_SECTOR_INVALID)')
    NTSTATUS__STATUS_LOG_SECTOR_PARITY_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD01A0002, 'HRESULT_FROM_NT(STATUS_LOG_SECTOR_PARITY_INVALID)')
    NTSTATUS__STATUS_LOG_SECTOR_REMAPPED: Final[ThirtyTwoBits] = cached_hresult(0xD01A0003, 'HRESULT_FROM_NT(STATUS_LOG_SECTOR_REMAPPED)')
    NTSTATUS__STATUS_LOG_BLOCK_INCOMPLETE: Final[ThirtyTwoBits] = cached_hresult(0xD01A0004, 'HRESULT_FROM_NT(STATUS_LOG_BLOCK_INCOMPLETE)')
    NTSTATUS__STATUS_LOG_INVALID_RANGE: Final[ThirtyTwoBits] = cached_hresult(0xD01A0005, 'HRESULT_FROM_NT(STATUS_LOG_INVALID_RANGE)')
    NTSTATUS__STATUS_LOG_BLOCKS_EXHAUSTED: Final[ThirtyTwoBits] = cached_hresult(0xD01A0006, 'HRESULT_FROM_NT(STATUS_LOG_BLOCKS_EXHAUSTED)')
    NTSTATUS__STATUS_LOG_READ_CONTEXT_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD01A0007, 'HRESULT_FROM_NT(STATUS_LOG_READ_CONTEXT_INVALID)')
    NTSTATUS__STATUS_LOG_RESTART_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD01A0008, 'HRESULT_FROM_NT(STATUS_LOG_RESTART_INVALID)')
    NTSTATUS__STATUS_LOG_BLOCK_VERSION: Final[ThirtyTwoBits] = cached_hresult(0xD01A0009, 'HRESULT_FROM_NT(STATUS_LOG_BLOCK_VERSION)')
    NTSTATUS__STATUS_LOG_BLOCK_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD01A000A, 'HRESULT_FROM_NT(STATUS_LOG_BLOCK_INVALID)')
    NTSTATUS__STATUS_LOG_READ_MODE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD01A000B, 'HRESULT_FROM_NT(STATUS_LOG_READ_MODE_INVALID)')
    NTSTATUS__STATUS_LOG_METADATA_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0xD01A000D, 'HRESULT_FROM_NT(STATUS_LOG_METADATA_CORRUPT)')
    NTSTATUS__STATUS_LOG_METADATA_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD01A000E, 'HRESULT_FROM_NT(STATUS_LOG_METADATA_INVALID)')
    NTSTATUS__STATUS_LOG_METADATA_INCONSISTENT: Final[ThirtyTwoBits] = cached_hresult(0xD01A000F, 'HRESULT_FROM_NT(STATUS_LOG_METADATA_INCONSISTENT)')
    NTSTATUS__STATUS_LOG_RESERVATION_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD01A0010, 'HRESULT_FROM_NT(STATUS_LOG_RESERVATION_INVALID)')
    NTSTATUS__STATUS_LOG_CANT_DELETE: Final[ThirtyTwoBits] = cached_hresult(0xD01A0011, 'HRESULT_FROM_NT(STATUS_LOG_CANT_DELETE)')
    NTSTATUS__STATUS_LOG_CONTAINER_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD01A0012, 'HRESULT_FROM_NT(STATUS_LOG_CONTAINER_LIMIT_EXCEEDED)')
    NTSTATUS__STATUS_LOG_START_OF_LOG: Final[ThirtyTwoBits] = cached_hresult(0xD01A0013, 'HRESULT_FROM_NT(STATUS_LOG_START_OF_LOG)')
    NTSTATUS__STATUS_LOG_POLICY_ALREADY_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0xD01A0014, 'HRESULT_FROM_NT(STATUS_LOG_POLICY_ALREADY_INSTALLED)')
    NTSTATUS__STATUS_LOG_POLICY_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_hresult(0xD01A0015, 'HRESULT_FROM_NT(STATUS_LOG_POLICY_NOT_INSTALLED)')
    NTSTATUS__STATUS_LOG_POLICY_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD01A0016, 'HRESULT_FROM_NT(STATUS_LOG_POLICY_INVALID)')
    NTSTATUS__STATUS_LOG_POLICY_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0xD01A0017, 'HRESULT_FROM_NT(STATUS_LOG_POLICY_CONFLICT)')
    NTSTATUS__STATUS_LOG_PINNED_ARCHIVE_TAIL: Final[ThirtyTwoBits] = cached_hresult(0xD01A0018, 'HRESULT_FROM_NT(STATUS_LOG_PINNED_ARCHIVE_TAIL)')
    NTSTATUS__STATUS_LOG_RECORD_NONEXISTENT: Final[ThirtyTwoBits] = cached_hresult(0xD01A0019, 'HRESULT_FROM_NT(STATUS_LOG_RECORD_NONEXISTENT)')
    NTSTATUS__STATUS_LOG_RECORDS_RESERVED_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD01A001A, 'HRESULT_FROM_NT(STATUS_LOG_RECORDS_RESERVED_INVALID)')
    NTSTATUS__STATUS_LOG_SPACE_RESERVED_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD01A001B, 'HRESULT_FROM_NT(STATUS_LOG_SPACE_RESERVED_INVALID)')
    NTSTATUS__STATUS_LOG_TAIL_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD01A001C, 'HRESULT_FROM_NT(STATUS_LOG_TAIL_INVALID)')
    NTSTATUS__STATUS_LOG_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD01A001D, 'HRESULT_FROM_NT(STATUS_LOG_FULL)')
    NTSTATUS__STATUS_LOG_MULTIPLEXED: Final[ThirtyTwoBits] = cached_hresult(0xD01A001E, 'HRESULT_FROM_NT(STATUS_LOG_MULTIPLEXED)')
    NTSTATUS__STATUS_LOG_DEDICATED: Final[ThirtyTwoBits] = cached_hresult(0xD01A001F, 'HRESULT_FROM_NT(STATUS_LOG_DEDICATED)')
    NTSTATUS__STATUS_LOG_ARCHIVE_NOT_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD01A0020, 'HRESULT_FROM_NT(STATUS_LOG_ARCHIVE_NOT_IN_PROGRESS)')
    NTSTATUS__STATUS_LOG_ARCHIVE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD01A0021, 'HRESULT_FROM_NT(STATUS_LOG_ARCHIVE_IN_PROGRESS)')
    NTSTATUS__STATUS_LOG_EPHEMERAL: Final[ThirtyTwoBits] = cached_hresult(0xD01A0022, 'HRESULT_FROM_NT(STATUS_LOG_EPHEMERAL)')
    NTSTATUS__STATUS_LOG_NOT_ENOUGH_CONTAINERS: Final[ThirtyTwoBits] = cached_hresult(0xD01A0023, 'HRESULT_FROM_NT(STATUS_LOG_NOT_ENOUGH_CONTAINERS)')
    NTSTATUS__STATUS_LOG_CLIENT_ALREADY_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0xD01A0024, 'HRESULT_FROM_NT(STATUS_LOG_CLIENT_ALREADY_REGISTERED)')
    NTSTATUS__STATUS_LOG_CLIENT_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0xD01A0025, 'HRESULT_FROM_NT(STATUS_LOG_CLIENT_NOT_REGISTERED)')
    NTSTATUS__STATUS_LOG_FULL_HANDLER_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD01A0026, 'HRESULT_FROM_NT(STATUS_LOG_FULL_HANDLER_IN_PROGRESS)')
    NTSTATUS__STATUS_LOG_CONTAINER_READ_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD01A0027, 'HRESULT_FROM_NT(STATUS_LOG_CONTAINER_READ_FAILED)')
    NTSTATUS__STATUS_LOG_CONTAINER_WRITE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD01A0028, 'HRESULT_FROM_NT(STATUS_LOG_CONTAINER_WRITE_FAILED)')
    NTSTATUS__STATUS_LOG_CONTAINER_OPEN_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD01A0029, 'HRESULT_FROM_NT(STATUS_LOG_CONTAINER_OPEN_FAILED)')
    NTSTATUS__STATUS_LOG_CONTAINER_STATE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD01A002A, 'HRESULT_FROM_NT(STATUS_LOG_CONTAINER_STATE_INVALID)')
    NTSTATUS__STATUS_LOG_STATE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD01A002B, 'HRESULT_FROM_NT(STATUS_LOG_STATE_INVALID)')
    NTSTATUS__STATUS_LOG_PINNED: Final[ThirtyTwoBits] = cached_hresult(0xD01A002C, 'HRESULT_FROM_NT(STATUS_LOG_PINNED)')
    NTSTATUS__STATUS_LOG_METADATA_FLUSH_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD01A002D, 'HRESULT_FROM_NT(STATUS_LOG_METADATA_FLUSH_FAILED)')
    NTSTATUS__STATUS_LOG_INCONSISTENT_SECURITY: Final[ThirtyTwoBits] = cached_hresult(0xD01A002E, 'HRESULT_FROM_NT(STATUS_LOG_INCONSISTENT_SECURITY)')
    NTSTATUS__STATUS_LOG_APPENDED_FLUSH_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD01A002F, 'HRESULT_FROM_NT(STATUS_LOG_APPENDED_FLUSH_FAILED)')
    NTSTATUS__STATUS_LOG_PINNED_RESERVATION: Final[ThirtyTwoBits] = cached_hresult(0xD01A0030, 'HRESULT_FROM_NT(STATUS_LOG_PINNED_RESERVATION)')
    NTSTATUS__STATUS_VIDEO_HUNG_DISPLAY_DRIVER_THREAD: Final[ThirtyTwoBits] = cached_hresult(0xD01B00EA, 'HRESULT_FROM_NT(STATUS_VIDEO_HUNG_DISPLAY_DRIVER_THREAD)')
    NTSTATUS__STATUS_FLT_NO_HANDLER_DEFINED: Final[ThirtyTwoBits] = cached_hresult(0xD01C0001, 'HRESULT_FROM_NT(STATUS_FLT_NO_HANDLER_DEFINED)')
    NTSTATUS__STATUS_FLT_CONTEXT_ALREADY_DEFINED: Final[ThirtyTwoBits] = cached_hresult(0xD01C0002, 'HRESULT_FROM_NT(STATUS_FLT_CONTEXT_ALREADY_DEFINED)')
    NTSTATUS__STATUS_FLT_INVALID_ASYNCHRONOUS_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0xD01C0003, 'HRESULT_FROM_NT(STATUS_FLT_INVALID_ASYNCHRONOUS_REQUEST)')
    NTSTATUS__STATUS_FLT_DISALLOW_FAST_IO: Final[ThirtyTwoBits] = cached_hresult(0xD01C0004, 'HRESULT_FROM_NT(STATUS_FLT_DISALLOW_FAST_IO)')
    NTSTATUS__STATUS_FLT_INVALID_NAME_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0xD01C0005, 'HRESULT_FROM_NT(STATUS_FLT_INVALID_NAME_REQUEST)')
    NTSTATUS__STATUS_FLT_NOT_SAFE_TO_POST_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0xD01C0006, 'HRESULT_FROM_NT(STATUS_FLT_NOT_SAFE_TO_POST_OPERATION)')
    NTSTATUS__STATUS_FLT_NOT_INITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0xD01C0007, 'HRESULT_FROM_NT(STATUS_FLT_NOT_INITIALIZED)')
    NTSTATUS__STATUS_FLT_FILTER_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0xD01C0008, 'HRESULT_FROM_NT(STATUS_FLT_FILTER_NOT_READY)')
    NTSTATUS__STATUS_FLT_POST_OPERATION_CLEANUP: Final[ThirtyTwoBits] = cached_hresult(0xD01C0009, 'HRESULT_FROM_NT(STATUS_FLT_POST_OPERATION_CLEANUP)')
    NTSTATUS__STATUS_FLT_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD01C000A, 'HRESULT_FROM_NT(STATUS_FLT_INTERNAL_ERROR)')
    NTSTATUS__STATUS_FLT_DELETING_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0xD01C000B, 'HRESULT_FROM_NT(STATUS_FLT_DELETING_OBJECT)')
    NTSTATUS__STATUS_FLT_MUST_BE_NONPAGED_POOL: Final[ThirtyTwoBits] = cached_hresult(0xD01C000C, 'HRESULT_FROM_NT(STATUS_FLT_MUST_BE_NONPAGED_POOL)')
    NTSTATUS__STATUS_FLT_DUPLICATE_ENTRY: Final[ThirtyTwoBits] = cached_hresult(0xD01C000D, 'HRESULT_FROM_NT(STATUS_FLT_DUPLICATE_ENTRY)')
    NTSTATUS__STATUS_FLT_CBDQ_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD01C000E, 'HRESULT_FROM_NT(STATUS_FLT_CBDQ_DISABLED)')
    NTSTATUS__STATUS_FLT_DO_NOT_ATTACH: Final[ThirtyTwoBits] = cached_hresult(0xD01C000F, 'HRESULT_FROM_NT(STATUS_FLT_DO_NOT_ATTACH)')
    NTSTATUS__STATUS_FLT_DO_NOT_DETACH: Final[ThirtyTwoBits] = cached_hresult(0xD01C0010, 'HRESULT_FROM_NT(STATUS_FLT_DO_NOT_DETACH)')
    NTSTATUS__STATUS_FLT_INSTANCE_ALTITUDE_COLLISION: Final[ThirtyTwoBits] = cached_hresult(0xD01C0011, 'HRESULT_FROM_NT(STATUS_FLT_INSTANCE_ALTITUDE_COLLISION)')
    NTSTATUS__STATUS_FLT_INSTANCE_NAME_COLLISION: Final[ThirtyTwoBits] = cached_hresult(0xD01C0012, 'HRESULT_FROM_NT(STATUS_FLT_INSTANCE_NAME_COLLISION)')
    NTSTATUS__STATUS_FLT_FILTER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD01C0013, 'HRESULT_FROM_NT(STATUS_FLT_FILTER_NOT_FOUND)')
    NTSTATUS__STATUS_FLT_VOLUME_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD01C0014, 'HRESULT_FROM_NT(STATUS_FLT_VOLUME_NOT_FOUND)')
    NTSTATUS__STATUS_FLT_INSTANCE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD01C0015, 'HRESULT_FROM_NT(STATUS_FLT_INSTANCE_NOT_FOUND)')
    NTSTATUS__STATUS_FLT_CONTEXT_ALLOCATION_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD01C0016, 'HRESULT_FROM_NT(STATUS_FLT_CONTEXT_ALLOCATION_NOT_FOUND)')
    NTSTATUS__STATUS_FLT_INVALID_CONTEXT_REGISTRATION: Final[ThirtyTwoBits] = cached_hresult(0xD01C0017, 'HRESULT_FROM_NT(STATUS_FLT_INVALID_CONTEXT_REGISTRATION)')
    NTSTATUS__STATUS_FLT_NAME_CACHE_MISS: Final[ThirtyTwoBits] = cached_hresult(0xD01C0018, 'HRESULT_FROM_NT(STATUS_FLT_NAME_CACHE_MISS)')
    NTSTATUS__STATUS_FLT_NO_DEVICE_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0xD01C0019, 'HRESULT_FROM_NT(STATUS_FLT_NO_DEVICE_OBJECT)')
    NTSTATUS__STATUS_FLT_VOLUME_ALREADY_MOUNTED: Final[ThirtyTwoBits] = cached_hresult(0xD01C001A, 'HRESULT_FROM_NT(STATUS_FLT_VOLUME_ALREADY_MOUNTED)')
    NTSTATUS__STATUS_FLT_ALREADY_ENLISTED: Final[ThirtyTwoBits] = cached_hresult(0xD01C001B, 'HRESULT_FROM_NT(STATUS_FLT_ALREADY_ENLISTED)')
    NTSTATUS__STATUS_FLT_CONTEXT_ALREADY_LINKED: Final[ThirtyTwoBits] = cached_hresult(0xD01C001C, 'HRESULT_FROM_NT(STATUS_FLT_CONTEXT_ALREADY_LINKED)')
    NTSTATUS__STATUS_FLT_NO_WAITER_FOR_REPLY: Final[ThirtyTwoBits] = cached_hresult(0xD01C0020, 'HRESULT_FROM_NT(STATUS_FLT_NO_WAITER_FOR_REPLY)')
    NTSTATUS__STATUS_FLT_REGISTRATION_BUSY: Final[ThirtyTwoBits] = cached_hresult(0xD01C0023, 'HRESULT_FROM_NT(STATUS_FLT_REGISTRATION_BUSY)')
    NTSTATUS__STATUS_FLT_WCOS_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD01C0024, 'HRESULT_FROM_NT(STATUS_FLT_WCOS_NOT_SUPPORTED)')
    NTSTATUS__STATUS_MONITOR_NO_DESCRIPTOR: Final[ThirtyTwoBits] = cached_hresult(0xD01D0001, 'HRESULT_FROM_NT(STATUS_MONITOR_NO_DESCRIPTOR)')
    NTSTATUS__STATUS_MONITOR_UNKNOWN_DESCRIPTOR_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0xD01D0002, 'HRESULT_FROM_NT(STATUS_MONITOR_UNKNOWN_DESCRIPTOR_FORMAT)')
    NTSTATUS__STATUS_MONITOR_INVALID_DESCRIPTOR_CHECKSUM: Final[ThirtyTwoBits] = cached_hresult(0xD01D0003, 'HRESULT_FROM_NT(STATUS_MONITOR_INVALID_DESCRIPTOR_CHECKSUM)')
    NTSTATUS__STATUS_MONITOR_INVALID_STANDARD_TIMING_BLOCK: Final[ThirtyTwoBits] = cached_hresult(0xD01D0004, 'HRESULT_FROM_NT(STATUS_MONITOR_INVALID_STANDARD_TIMING_BLOCK)')
    NTSTATUS__STATUS_MONITOR_WMI_DATABLOCK_REGISTRATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD01D0005, 'HRESULT_FROM_NT(STATUS_MONITOR_WMI_DATABLOCK_REGISTRATION_FAILED)')
    NTSTATUS__STATUS_MONITOR_INVALID_SERIAL_NUMBER_MONDSC_BLOCK: Final[ThirtyTwoBits] = cached_hresult(0xD01D0006, 'HRESULT_FROM_NT(STATUS_MONITOR_INVALID_SERIAL_NUMBER_MONDSC_BLOCK)')
    NTSTATUS__STATUS_MONITOR_INVALID_USER_FRIENDLY_MONDSC_BLOCK: Final[ThirtyTwoBits] = cached_hresult(0xD01D0007, 'HRESULT_FROM_NT(STATUS_MONITOR_INVALID_USER_FRIENDLY_MONDSC_BLOCK)')
    NTSTATUS__STATUS_MONITOR_NO_MORE_DESCRIPTOR_DATA: Final[ThirtyTwoBits] = cached_hresult(0xD01D0008, 'HRESULT_FROM_NT(STATUS_MONITOR_NO_MORE_DESCRIPTOR_DATA)')
    NTSTATUS__STATUS_MONITOR_INVALID_DETAILED_TIMING_BLOCK: Final[ThirtyTwoBits] = cached_hresult(0xD01D0009, 'HRESULT_FROM_NT(STATUS_MONITOR_INVALID_DETAILED_TIMING_BLOCK)')
    NTSTATUS__STATUS_MONITOR_INVALID_MANUFACTURE_DATE: Final[ThirtyTwoBits] = cached_hresult(0xD01D000A, 'HRESULT_FROM_NT(STATUS_MONITOR_INVALID_MANUFACTURE_DATE)')
    NTSTATUS__STATUS_GRAPHICS_NOT_EXCLUSIVE_MODE_OWNER: Final[ThirtyTwoBits] = cached_hresult(0xD01E0000, 'HRESULT_FROM_NT(STATUS_GRAPHICS_NOT_EXCLUSIVE_MODE_OWNER)')
    NTSTATUS__STATUS_GRAPHICS_INSUFFICIENT_DMA_BUFFER: Final[ThirtyTwoBits] = cached_hresult(0xD01E0001, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INSUFFICIENT_DMA_BUFFER)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_DISPLAY_ADAPTER: Final[ThirtyTwoBits] = cached_hresult(0xD01E0002, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_DISPLAY_ADAPTER)')
    NTSTATUS__STATUS_GRAPHICS_ADAPTER_WAS_RESET: Final[ThirtyTwoBits] = cached_hresult(0xD01E0003, 'HRESULT_FROM_NT(STATUS_GRAPHICS_ADAPTER_WAS_RESET)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_DRIVER_MODEL: Final[ThirtyTwoBits] = cached_hresult(0xD01E0004, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_DRIVER_MODEL)')
    NTSTATUS__STATUS_GRAPHICS_PRESENT_MODE_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0005, 'HRESULT_FROM_NT(STATUS_GRAPHICS_PRESENT_MODE_CHANGED)')
    NTSTATUS__STATUS_GRAPHICS_PRESENT_OCCLUDED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0006, 'HRESULT_FROM_NT(STATUS_GRAPHICS_PRESENT_OCCLUDED)')
    NTSTATUS__STATUS_GRAPHICS_PRESENT_DENIED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0007, 'HRESULT_FROM_NT(STATUS_GRAPHICS_PRESENT_DENIED)')
    NTSTATUS__STATUS_GRAPHICS_CANNOTCOLORCONVERT: Final[ThirtyTwoBits] = cached_hresult(0xD01E0008, 'HRESULT_FROM_NT(STATUS_GRAPHICS_CANNOTCOLORCONVERT)')
    NTSTATUS__STATUS_GRAPHICS_DRIVER_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD01E0009, 'HRESULT_FROM_NT(STATUS_GRAPHICS_DRIVER_MISMATCH)')
    NTSTATUS__STATUS_GRAPHICS_PRESENT_REDIRECTION_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD01E000B, 'HRESULT_FROM_NT(STATUS_GRAPHICS_PRESENT_REDIRECTION_DISABLED)')
    NTSTATUS__STATUS_GRAPHICS_PRESENT_UNOCCLUDED: Final[ThirtyTwoBits] = cached_hresult(0xD01E000C, 'HRESULT_FROM_NT(STATUS_GRAPHICS_PRESENT_UNOCCLUDED)')
    NTSTATUS__STATUS_GRAPHICS_WINDOWDC_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD01E000D, 'HRESULT_FROM_NT(STATUS_GRAPHICS_WINDOWDC_NOT_AVAILABLE)')
    NTSTATUS__STATUS_GRAPHICS_WINDOWLESS_PRESENT_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD01E000E, 'HRESULT_FROM_NT(STATUS_GRAPHICS_WINDOWLESS_PRESENT_DISABLED)')
    NTSTATUS__STATUS_GRAPHICS_PRESENT_INVALID_WINDOW: Final[ThirtyTwoBits] = cached_hresult(0xD01E000F, 'HRESULT_FROM_NT(STATUS_GRAPHICS_PRESENT_INVALID_WINDOW)')
    NTSTATUS__STATUS_GRAPHICS_PRESENT_BUFFER_NOT_BOUND: Final[ThirtyTwoBits] = cached_hresult(0xD01E0010, 'HRESULT_FROM_NT(STATUS_GRAPHICS_PRESENT_BUFFER_NOT_BOUND)')
    NTSTATUS__STATUS_GRAPHICS_VAIL_STATE_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0011, 'HRESULT_FROM_NT(STATUS_GRAPHICS_VAIL_STATE_CHANGED)')
    NTSTATUS__STATUS_GRAPHICS_INDIRECT_DISPLAY_ABANDON_SWAPCHAIN: Final[ThirtyTwoBits] = cached_hresult(0xD01E0012, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INDIRECT_DISPLAY_ABANDON_SWAPCHAIN)')
    NTSTATUS__STATUS_GRAPHICS_INDIRECT_DISPLAY_DEVICE_STOPPED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0013, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INDIRECT_DISPLAY_DEVICE_STOPPED)')
    NTSTATUS__STATUS_GRAPHICS_MPO_ALLOCATION_UNPINNED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0018, 'HRESULT_FROM_NT(STATUS_GRAPHICS_MPO_ALLOCATION_UNPINNED)')
    NTSTATUS__STATUS_GRAPHICS_NO_VIDEO_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0xD01E0100, 'HRESULT_FROM_NT(STATUS_GRAPHICS_NO_VIDEO_MEMORY)')
    NTSTATUS__STATUS_GRAPHICS_CANT_LOCK_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0xD01E0101, 'HRESULT_FROM_NT(STATUS_GRAPHICS_CANT_LOCK_MEMORY)')
    NTSTATUS__STATUS_GRAPHICS_ALLOCATION_BUSY: Final[ThirtyTwoBits] = cached_hresult(0xD01E0102, 'HRESULT_FROM_NT(STATUS_GRAPHICS_ALLOCATION_BUSY)')
    NTSTATUS__STATUS_GRAPHICS_TOO_MANY_REFERENCES: Final[ThirtyTwoBits] = cached_hresult(0xD01E0103, 'HRESULT_FROM_NT(STATUS_GRAPHICS_TOO_MANY_REFERENCES)')
    NTSTATUS__STATUS_GRAPHICS_TRY_AGAIN_LATER: Final[ThirtyTwoBits] = cached_hresult(0xD01E0104, 'HRESULT_FROM_NT(STATUS_GRAPHICS_TRY_AGAIN_LATER)')
    NTSTATUS__STATUS_GRAPHICS_TRY_AGAIN_NOW: Final[ThirtyTwoBits] = cached_hresult(0xD01E0105, 'HRESULT_FROM_NT(STATUS_GRAPHICS_TRY_AGAIN_NOW)')
    NTSTATUS__STATUS_GRAPHICS_ALLOCATION_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD01E0106, 'HRESULT_FROM_NT(STATUS_GRAPHICS_ALLOCATION_INVALID)')
    NTSTATUS__STATUS_GRAPHICS_UNSWIZZLING_APERTURE_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD01E0107, 'HRESULT_FROM_NT(STATUS_GRAPHICS_UNSWIZZLING_APERTURE_UNAVAILABLE)')
    NTSTATUS__STATUS_GRAPHICS_UNSWIZZLING_APERTURE_UNSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0108, 'HRESULT_FROM_NT(STATUS_GRAPHICS_UNSWIZZLING_APERTURE_UNSUPPORTED)')
    NTSTATUS__STATUS_GRAPHICS_CANT_EVICT_PINNED_ALLOCATION: Final[ThirtyTwoBits] = cached_hresult(0xD01E0109, 'HRESULT_FROM_NT(STATUS_GRAPHICS_CANT_EVICT_PINNED_ALLOCATION)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_ALLOCATION_USAGE: Final[ThirtyTwoBits] = cached_hresult(0xD01E0110, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_ALLOCATION_USAGE)')
    NTSTATUS__STATUS_GRAPHICS_CANT_RENDER_LOCKED_ALLOCATION: Final[ThirtyTwoBits] = cached_hresult(0xD01E0111, 'HRESULT_FROM_NT(STATUS_GRAPHICS_CANT_RENDER_LOCKED_ALLOCATION)')
    NTSTATUS__STATUS_GRAPHICS_ALLOCATION_CLOSED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0112, 'HRESULT_FROM_NT(STATUS_GRAPHICS_ALLOCATION_CLOSED)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_ALLOCATION_INSTANCE: Final[ThirtyTwoBits] = cached_hresult(0xD01E0113, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_ALLOCATION_INSTANCE)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_ALLOCATION_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0xD01E0114, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_ALLOCATION_HANDLE)')
    NTSTATUS__STATUS_GRAPHICS_WRONG_ALLOCATION_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0xD01E0115, 'HRESULT_FROM_NT(STATUS_GRAPHICS_WRONG_ALLOCATION_DEVICE)')
    NTSTATUS__STATUS_GRAPHICS_ALLOCATION_CONTENT_LOST: Final[ThirtyTwoBits] = cached_hresult(0xD01E0116, 'HRESULT_FROM_NT(STATUS_GRAPHICS_ALLOCATION_CONTENT_LOST)')
    NTSTATUS__STATUS_GRAPHICS_GPU_EXCEPTION_ON_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0xD01E0200, 'HRESULT_FROM_NT(STATUS_GRAPHICS_GPU_EXCEPTION_ON_DEVICE)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_VIDPN_TOPOLOGY: Final[ThirtyTwoBits] = cached_hresult(0xD01E0300, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_VIDPN_TOPOLOGY)')
    NTSTATUS__STATUS_GRAPHICS_VIDPN_TOPOLOGY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0301, 'HRESULT_FROM_NT(STATUS_GRAPHICS_VIDPN_TOPOLOGY_NOT_SUPPORTED)')
    NTSTATUS__STATUS_GRAPHICS_VIDPN_TOPOLOGY_CURRENTLY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0302, 'HRESULT_FROM_NT(STATUS_GRAPHICS_VIDPN_TOPOLOGY_CURRENTLY_NOT_SUPPORTED)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_VIDPN: Final[ThirtyTwoBits] = cached_hresult(0xD01E0303, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_VIDPN)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_VIDEO_PRESENT_SOURCE: Final[ThirtyTwoBits] = cached_hresult(0xD01E0304, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_VIDEO_PRESENT_SOURCE)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_VIDEO_PRESENT_TARGET: Final[ThirtyTwoBits] = cached_hresult(0xD01E0305, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_VIDEO_PRESENT_TARGET)')
    NTSTATUS__STATUS_GRAPHICS_VIDPN_MODALITY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0306, 'HRESULT_FROM_NT(STATUS_GRAPHICS_VIDPN_MODALITY_NOT_SUPPORTED)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_VIDPN_SOURCEMODESET: Final[ThirtyTwoBits] = cached_hresult(0xD01E0308, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_VIDPN_SOURCEMODESET)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_VIDPN_TARGETMODESET: Final[ThirtyTwoBits] = cached_hresult(0xD01E0309, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_VIDPN_TARGETMODESET)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_FREQUENCY: Final[ThirtyTwoBits] = cached_hresult(0xD01E030A, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_FREQUENCY)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_ACTIVE_REGION: Final[ThirtyTwoBits] = cached_hresult(0xD01E030B, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_ACTIVE_REGION)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_TOTAL_REGION: Final[ThirtyTwoBits] = cached_hresult(0xD01E030C, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_TOTAL_REGION)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_VIDEO_PRESENT_SOURCE_MODE: Final[ThirtyTwoBits] = cached_hresult(0xD01E0310, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_VIDEO_PRESENT_SOURCE_MODE)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_VIDEO_PRESENT_TARGET_MODE: Final[ThirtyTwoBits] = cached_hresult(0xD01E0311, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_VIDEO_PRESENT_TARGET_MODE)')
    NTSTATUS__STATUS_GRAPHICS_PINNED_MODE_MUST_REMAIN_IN_SET: Final[ThirtyTwoBits] = cached_hresult(0xD01E0312, 'HRESULT_FROM_NT(STATUS_GRAPHICS_PINNED_MODE_MUST_REMAIN_IN_SET)')
    NTSTATUS__STATUS_GRAPHICS_PATH_ALREADY_IN_TOPOLOGY: Final[ThirtyTwoBits] = cached_hresult(0xD01E0313, 'HRESULT_FROM_NT(STATUS_GRAPHICS_PATH_ALREADY_IN_TOPOLOGY)')
    NTSTATUS__STATUS_GRAPHICS_MODE_ALREADY_IN_MODESET: Final[ThirtyTwoBits] = cached_hresult(0xD01E0314, 'HRESULT_FROM_NT(STATUS_GRAPHICS_MODE_ALREADY_IN_MODESET)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_VIDEOPRESENTSOURCESET: Final[ThirtyTwoBits] = cached_hresult(0xD01E0315, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_VIDEOPRESENTSOURCESET)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_VIDEOPRESENTTARGETSET: Final[ThirtyTwoBits] = cached_hresult(0xD01E0316, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_VIDEOPRESENTTARGETSET)')
    NTSTATUS__STATUS_GRAPHICS_SOURCE_ALREADY_IN_SET: Final[ThirtyTwoBits] = cached_hresult(0xD01E0317, 'HRESULT_FROM_NT(STATUS_GRAPHICS_SOURCE_ALREADY_IN_SET)')
    NTSTATUS__STATUS_GRAPHICS_TARGET_ALREADY_IN_SET: Final[ThirtyTwoBits] = cached_hresult(0xD01E0318, 'HRESULT_FROM_NT(STATUS_GRAPHICS_TARGET_ALREADY_IN_SET)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_VIDPN_PRESENT_PATH: Final[ThirtyTwoBits] = cached_hresult(0xD01E0319, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_VIDPN_PRESENT_PATH)')
    NTSTATUS__STATUS_GRAPHICS_NO_RECOMMENDED_VIDPN_TOPOLOGY: Final[ThirtyTwoBits] = cached_hresult(0xD01E031A, 'HRESULT_FROM_NT(STATUS_GRAPHICS_NO_RECOMMENDED_VIDPN_TOPOLOGY)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGESET: Final[ThirtyTwoBits] = cached_hresult(0xD01E031B, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGESET)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGE: Final[ThirtyTwoBits] = cached_hresult(0xD01E031C, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGE)')
    NTSTATUS__STATUS_GRAPHICS_FREQUENCYRANGE_NOT_IN_SET: Final[ThirtyTwoBits] = cached_hresult(0xD01E031D, 'HRESULT_FROM_NT(STATUS_GRAPHICS_FREQUENCYRANGE_NOT_IN_SET)')
    NTSTATUS__STATUS_GRAPHICS_FREQUENCYRANGE_ALREADY_IN_SET: Final[ThirtyTwoBits] = cached_hresult(0xD01E031F, 'HRESULT_FROM_NT(STATUS_GRAPHICS_FREQUENCYRANGE_ALREADY_IN_SET)')
    NTSTATUS__STATUS_GRAPHICS_STALE_MODESET: Final[ThirtyTwoBits] = cached_hresult(0xD01E0320, 'HRESULT_FROM_NT(STATUS_GRAPHICS_STALE_MODESET)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_MONITOR_SOURCEMODESET: Final[ThirtyTwoBits] = cached_hresult(0xD01E0321, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_MONITOR_SOURCEMODESET)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_MONITOR_SOURCE_MODE: Final[ThirtyTwoBits] = cached_hresult(0xD01E0322, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_MONITOR_SOURCE_MODE)')
    NTSTATUS__STATUS_GRAPHICS_NO_RECOMMENDED_FUNCTIONAL_VIDPN: Final[ThirtyTwoBits] = cached_hresult(0xD01E0323, 'HRESULT_FROM_NT(STATUS_GRAPHICS_NO_RECOMMENDED_FUNCTIONAL_VIDPN)')
    NTSTATUS__STATUS_GRAPHICS_MODE_ID_MUST_BE_UNIQUE: Final[ThirtyTwoBits] = cached_hresult(0xD01E0324, 'HRESULT_FROM_NT(STATUS_GRAPHICS_MODE_ID_MUST_BE_UNIQUE)')
    NTSTATUS__STATUS_GRAPHICS_EMPTY_ADAPTER_MONITOR_MODE_SUPPORT_INTERSECTION: Final[ThirtyTwoBits] = cached_hresult(0xD01E0325, 'HRESULT_FROM_NT(STATUS_GRAPHICS_EMPTY_ADAPTER_MONITOR_MODE_SUPPORT_INTERSECTION)')
    NTSTATUS__STATUS_GRAPHICS_VIDEO_PRESENT_TARGETS_LESS_THAN_SOURCES: Final[ThirtyTwoBits] = cached_hresult(0xD01E0326, 'HRESULT_FROM_NT(STATUS_GRAPHICS_VIDEO_PRESENT_TARGETS_LESS_THAN_SOURCES)')
    NTSTATUS__STATUS_GRAPHICS_PATH_NOT_IN_TOPOLOGY: Final[ThirtyTwoBits] = cached_hresult(0xD01E0327, 'HRESULT_FROM_NT(STATUS_GRAPHICS_PATH_NOT_IN_TOPOLOGY)')
    NTSTATUS__STATUS_GRAPHICS_ADAPTER_MUST_HAVE_AT_LEAST_ONE_SOURCE: Final[ThirtyTwoBits] = cached_hresult(0xD01E0328, 'HRESULT_FROM_NT(STATUS_GRAPHICS_ADAPTER_MUST_HAVE_AT_LEAST_ONE_SOURCE)')
    NTSTATUS__STATUS_GRAPHICS_ADAPTER_MUST_HAVE_AT_LEAST_ONE_TARGET: Final[ThirtyTwoBits] = cached_hresult(0xD01E0329, 'HRESULT_FROM_NT(STATUS_GRAPHICS_ADAPTER_MUST_HAVE_AT_LEAST_ONE_TARGET)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_MONITORDESCRIPTORSET: Final[ThirtyTwoBits] = cached_hresult(0xD01E032A, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_MONITORDESCRIPTORSET)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_MONITORDESCRIPTOR: Final[ThirtyTwoBits] = cached_hresult(0xD01E032B, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_MONITORDESCRIPTOR)')
    NTSTATUS__STATUS_GRAPHICS_MONITORDESCRIPTOR_NOT_IN_SET: Final[ThirtyTwoBits] = cached_hresult(0xD01E032C, 'HRESULT_FROM_NT(STATUS_GRAPHICS_MONITORDESCRIPTOR_NOT_IN_SET)')
    NTSTATUS__STATUS_GRAPHICS_MONITORDESCRIPTOR_ALREADY_IN_SET: Final[ThirtyTwoBits] = cached_hresult(0xD01E032D, 'HRESULT_FROM_NT(STATUS_GRAPHICS_MONITORDESCRIPTOR_ALREADY_IN_SET)')
    NTSTATUS__STATUS_GRAPHICS_MONITORDESCRIPTOR_ID_MUST_BE_UNIQUE: Final[ThirtyTwoBits] = cached_hresult(0xD01E032E, 'HRESULT_FROM_NT(STATUS_GRAPHICS_MONITORDESCRIPTOR_ID_MUST_BE_UNIQUE)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_VIDPN_TARGET_SUBSET_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xD01E032F, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_VIDPN_TARGET_SUBSET_TYPE)')
    NTSTATUS__STATUS_GRAPHICS_RESOURCES_NOT_RELATED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0330, 'HRESULT_FROM_NT(STATUS_GRAPHICS_RESOURCES_NOT_RELATED)')
    NTSTATUS__STATUS_GRAPHICS_SOURCE_ID_MUST_BE_UNIQUE: Final[ThirtyTwoBits] = cached_hresult(0xD01E0331, 'HRESULT_FROM_NT(STATUS_GRAPHICS_SOURCE_ID_MUST_BE_UNIQUE)')
    NTSTATUS__STATUS_GRAPHICS_TARGET_ID_MUST_BE_UNIQUE: Final[ThirtyTwoBits] = cached_hresult(0xD01E0332, 'HRESULT_FROM_NT(STATUS_GRAPHICS_TARGET_ID_MUST_BE_UNIQUE)')
    NTSTATUS__STATUS_GRAPHICS_NO_AVAILABLE_VIDPN_TARGET: Final[ThirtyTwoBits] = cached_hresult(0xD01E0333, 'HRESULT_FROM_NT(STATUS_GRAPHICS_NO_AVAILABLE_VIDPN_TARGET)')
    NTSTATUS__STATUS_GRAPHICS_MONITOR_COULD_NOT_BE_ASSOCIATED_WITH_ADAPTER: Final[ThirtyTwoBits] = cached_hresult(0xD01E0334, 'HRESULT_FROM_NT(STATUS_GRAPHICS_MONITOR_COULD_NOT_BE_ASSOCIATED_WITH_ADAPTER)')
    NTSTATUS__STATUS_GRAPHICS_NO_VIDPNMGR: Final[ThirtyTwoBits] = cached_hresult(0xD01E0335, 'HRESULT_FROM_NT(STATUS_GRAPHICS_NO_VIDPNMGR)')
    NTSTATUS__STATUS_GRAPHICS_NO_ACTIVE_VIDPN: Final[ThirtyTwoBits] = cached_hresult(0xD01E0336, 'HRESULT_FROM_NT(STATUS_GRAPHICS_NO_ACTIVE_VIDPN)')
    NTSTATUS__STATUS_GRAPHICS_STALE_VIDPN_TOPOLOGY: Final[ThirtyTwoBits] = cached_hresult(0xD01E0337, 'HRESULT_FROM_NT(STATUS_GRAPHICS_STALE_VIDPN_TOPOLOGY)')
    NTSTATUS__STATUS_GRAPHICS_MONITOR_NOT_CONNECTED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0338, 'HRESULT_FROM_NT(STATUS_GRAPHICS_MONITOR_NOT_CONNECTED)')
    NTSTATUS__STATUS_GRAPHICS_SOURCE_NOT_IN_TOPOLOGY: Final[ThirtyTwoBits] = cached_hresult(0xD01E0339, 'HRESULT_FROM_NT(STATUS_GRAPHICS_SOURCE_NOT_IN_TOPOLOGY)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_PRIMARYSURFACE_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD01E033A, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_PRIMARYSURFACE_SIZE)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_VISIBLEREGION_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD01E033B, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_VISIBLEREGION_SIZE)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_STRIDE: Final[ThirtyTwoBits] = cached_hresult(0xD01E033C, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_STRIDE)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_PIXELFORMAT: Final[ThirtyTwoBits] = cached_hresult(0xD01E033D, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_PIXELFORMAT)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_COLORBASIS: Final[ThirtyTwoBits] = cached_hresult(0xD01E033E, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_COLORBASIS)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_PIXELVALUEACCESSMODE: Final[ThirtyTwoBits] = cached_hresult(0xD01E033F, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_PIXELVALUEACCESSMODE)')
    NTSTATUS__STATUS_GRAPHICS_TARGET_NOT_IN_TOPOLOGY: Final[ThirtyTwoBits] = cached_hresult(0xD01E0340, 'HRESULT_FROM_NT(STATUS_GRAPHICS_TARGET_NOT_IN_TOPOLOGY)')
    NTSTATUS__STATUS_GRAPHICS_NO_DISPLAY_MODE_MANAGEMENT_SUPPORT: Final[ThirtyTwoBits] = cached_hresult(0xD01E0341, 'HRESULT_FROM_NT(STATUS_GRAPHICS_NO_DISPLAY_MODE_MANAGEMENT_SUPPORT)')
    NTSTATUS__STATUS_GRAPHICS_VIDPN_SOURCE_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0xD01E0342, 'HRESULT_FROM_NT(STATUS_GRAPHICS_VIDPN_SOURCE_IN_USE)')
    NTSTATUS__STATUS_GRAPHICS_CANT_ACCESS_ACTIVE_VIDPN: Final[ThirtyTwoBits] = cached_hresult(0xD01E0343, 'HRESULT_FROM_NT(STATUS_GRAPHICS_CANT_ACCESS_ACTIVE_VIDPN)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_PATH_IMPORTANCE_ORDINAL: Final[ThirtyTwoBits] = cached_hresult(0xD01E0344, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_PATH_IMPORTANCE_ORDINAL)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_PATH_CONTENT_GEOMETRY_TRANSFORMATION: Final[ThirtyTwoBits] = cached_hresult(0xD01E0345, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_PATH_CONTENT_GEOMETRY_TRANSFORMATION)')
    NTSTATUS__STATUS_GRAPHICS_PATH_CONTENT_GEOMETRY_TRANSFORMATION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0346, 'HRESULT_FROM_NT(STATUS_GRAPHICS_PATH_CONTENT_GEOMETRY_TRANSFORMATION_NOT_SUPPORTED)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_GAMMA_RAMP: Final[ThirtyTwoBits] = cached_hresult(0xD01E0347, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_GAMMA_RAMP)')
    NTSTATUS__STATUS_GRAPHICS_GAMMA_RAMP_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0348, 'HRESULT_FROM_NT(STATUS_GRAPHICS_GAMMA_RAMP_NOT_SUPPORTED)')
    NTSTATUS__STATUS_GRAPHICS_MULTISAMPLING_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0349, 'HRESULT_FROM_NT(STATUS_GRAPHICS_MULTISAMPLING_NOT_SUPPORTED)')
    NTSTATUS__STATUS_GRAPHICS_MODE_NOT_IN_MODESET: Final[ThirtyTwoBits] = cached_hresult(0xD01E034A, 'HRESULT_FROM_NT(STATUS_GRAPHICS_MODE_NOT_IN_MODESET)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_VIDPN_TOPOLOGY_RECOMMENDATION_REASON: Final[ThirtyTwoBits] = cached_hresult(0xD01E034D, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_VIDPN_TOPOLOGY_RECOMMENDATION_REASON)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_PATH_CONTENT_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xD01E034E, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_PATH_CONTENT_TYPE)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_COPYPROTECTION_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xD01E034F, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_COPYPROTECTION_TYPE)')
    NTSTATUS__STATUS_GRAPHICS_UNASSIGNED_MODESET_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD01E0350, 'HRESULT_FROM_NT(STATUS_GRAPHICS_UNASSIGNED_MODESET_ALREADY_EXISTS)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_SCANLINE_ORDERING: Final[ThirtyTwoBits] = cached_hresult(0xD01E0352, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_SCANLINE_ORDERING)')
    NTSTATUS__STATUS_GRAPHICS_TOPOLOGY_CHANGES_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0353, 'HRESULT_FROM_NT(STATUS_GRAPHICS_TOPOLOGY_CHANGES_NOT_ALLOWED)')
    NTSTATUS__STATUS_GRAPHICS_NO_AVAILABLE_IMPORTANCE_ORDINALS: Final[ThirtyTwoBits] = cached_hresult(0xD01E0354, 'HRESULT_FROM_NT(STATUS_GRAPHICS_NO_AVAILABLE_IMPORTANCE_ORDINALS)')
    NTSTATUS__STATUS_GRAPHICS_INCOMPATIBLE_PRIVATE_FORMAT: Final[ThirtyTwoBits] = cached_hresult(0xD01E0355, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INCOMPATIBLE_PRIVATE_FORMAT)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_MODE_PRUNING_ALGORITHM: Final[ThirtyTwoBits] = cached_hresult(0xD01E0356, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_MODE_PRUNING_ALGORITHM)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_MONITOR_CAPABILITY_ORIGIN: Final[ThirtyTwoBits] = cached_hresult(0xD01E0357, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_MONITOR_CAPABILITY_ORIGIN)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGE_CONSTRAINT: Final[ThirtyTwoBits] = cached_hresult(0xD01E0358, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGE_CONSTRAINT)')
    NTSTATUS__STATUS_GRAPHICS_MAX_NUM_PATHS_REACHED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0359, 'HRESULT_FROM_NT(STATUS_GRAPHICS_MAX_NUM_PATHS_REACHED)')
    NTSTATUS__STATUS_GRAPHICS_CANCEL_VIDPN_TOPOLOGY_AUGMENTATION: Final[ThirtyTwoBits] = cached_hresult(0xD01E035A, 'HRESULT_FROM_NT(STATUS_GRAPHICS_CANCEL_VIDPN_TOPOLOGY_AUGMENTATION)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_CLIENT_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xD01E035B, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_CLIENT_TYPE)')
    NTSTATUS__STATUS_GRAPHICS_CLIENTVIDPN_NOT_SET: Final[ThirtyTwoBits] = cached_hresult(0xD01E035C, 'HRESULT_FROM_NT(STATUS_GRAPHICS_CLIENTVIDPN_NOT_SET)')
    NTSTATUS__STATUS_GRAPHICS_SPECIFIED_CHILD_ALREADY_CONNECTED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0400, 'HRESULT_FROM_NT(STATUS_GRAPHICS_SPECIFIED_CHILD_ALREADY_CONNECTED)')
    NTSTATUS__STATUS_GRAPHICS_CHILD_DESCRIPTOR_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0401, 'HRESULT_FROM_NT(STATUS_GRAPHICS_CHILD_DESCRIPTOR_NOT_SUPPORTED)')
    NTSTATUS__STATUS_GRAPHICS_NOT_A_LINKED_ADAPTER: Final[ThirtyTwoBits] = cached_hresult(0xD01E0430, 'HRESULT_FROM_NT(STATUS_GRAPHICS_NOT_A_LINKED_ADAPTER)')
    NTSTATUS__STATUS_GRAPHICS_LEADLINK_NOT_ENUMERATED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0431, 'HRESULT_FROM_NT(STATUS_GRAPHICS_LEADLINK_NOT_ENUMERATED)')
    NTSTATUS__STATUS_GRAPHICS_CHAINLINKS_NOT_ENUMERATED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0432, 'HRESULT_FROM_NT(STATUS_GRAPHICS_CHAINLINKS_NOT_ENUMERATED)')
    NTSTATUS__STATUS_GRAPHICS_ADAPTER_CHAIN_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0xD01E0433, 'HRESULT_FROM_NT(STATUS_GRAPHICS_ADAPTER_CHAIN_NOT_READY)')
    NTSTATUS__STATUS_GRAPHICS_CHAINLINKS_NOT_STARTED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0434, 'HRESULT_FROM_NT(STATUS_GRAPHICS_CHAINLINKS_NOT_STARTED)')
    NTSTATUS__STATUS_GRAPHICS_CHAINLINKS_NOT_POWERED_ON: Final[ThirtyTwoBits] = cached_hresult(0xD01E0435, 'HRESULT_FROM_NT(STATUS_GRAPHICS_CHAINLINKS_NOT_POWERED_ON)')
    NTSTATUS__STATUS_GRAPHICS_INCONSISTENT_DEVICE_LINK_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD01E0436, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INCONSISTENT_DEVICE_LINK_STATE)')
    NTSTATUS__STATUS_GRAPHICS_NOT_POST_DEVICE_DRIVER: Final[ThirtyTwoBits] = cached_hresult(0xD01E0438, 'HRESULT_FROM_NT(STATUS_GRAPHICS_NOT_POST_DEVICE_DRIVER)')
    NTSTATUS__STATUS_GRAPHICS_ADAPTER_ACCESS_NOT_EXCLUDED: Final[ThirtyTwoBits] = cached_hresult(0xD01E043B, 'HRESULT_FROM_NT(STATUS_GRAPHICS_ADAPTER_ACCESS_NOT_EXCLUDED)')
    NTSTATUS__STATUS_GRAPHICS_OPM_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0500, 'HRESULT_FROM_NT(STATUS_GRAPHICS_OPM_NOT_SUPPORTED)')
    NTSTATUS__STATUS_GRAPHICS_COPP_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0501, 'HRESULT_FROM_NT(STATUS_GRAPHICS_COPP_NOT_SUPPORTED)')
    NTSTATUS__STATUS_GRAPHICS_UAB_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0502, 'HRESULT_FROM_NT(STATUS_GRAPHICS_UAB_NOT_SUPPORTED)')
    NTSTATUS__STATUS_GRAPHICS_OPM_INVALID_ENCRYPTED_PARAMETERS: Final[ThirtyTwoBits] = cached_hresult(0xD01E0503, 'HRESULT_FROM_NT(STATUS_GRAPHICS_OPM_INVALID_ENCRYPTED_PARAMETERS)')
    NTSTATUS__STATUS_GRAPHICS_OPM_NO_PROTECTED_OUTPUTS_EXIST: Final[ThirtyTwoBits] = cached_hresult(0xD01E0505, 'HRESULT_FROM_NT(STATUS_GRAPHICS_OPM_NO_PROTECTED_OUTPUTS_EXIST)')
    NTSTATUS__STATUS_GRAPHICS_OPM_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD01E050B, 'HRESULT_FROM_NT(STATUS_GRAPHICS_OPM_INTERNAL_ERROR)')
    NTSTATUS__STATUS_GRAPHICS_OPM_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0xD01E050C, 'HRESULT_FROM_NT(STATUS_GRAPHICS_OPM_INVALID_HANDLE)')
    NTSTATUS__STATUS_GRAPHICS_PVP_INVALID_CERTIFICATE_LENGTH: Final[ThirtyTwoBits] = cached_hresult(0xD01E050E, 'HRESULT_FROM_NT(STATUS_GRAPHICS_PVP_INVALID_CERTIFICATE_LENGTH)')
    NTSTATUS__STATUS_GRAPHICS_OPM_SPANNING_MODE_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0xD01E050F, 'HRESULT_FROM_NT(STATUS_GRAPHICS_OPM_SPANNING_MODE_ENABLED)')
    NTSTATUS__STATUS_GRAPHICS_OPM_THEATER_MODE_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0510, 'HRESULT_FROM_NT(STATUS_GRAPHICS_OPM_THEATER_MODE_ENABLED)')
    NTSTATUS__STATUS_GRAPHICS_PVP_HFS_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0511, 'HRESULT_FROM_NT(STATUS_GRAPHICS_PVP_HFS_FAILED)')
    NTSTATUS__STATUS_GRAPHICS_OPM_INVALID_SRM: Final[ThirtyTwoBits] = cached_hresult(0xD01E0512, 'HRESULT_FROM_NT(STATUS_GRAPHICS_OPM_INVALID_SRM)')
    NTSTATUS__STATUS_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_HDCP: Final[ThirtyTwoBits] = cached_hresult(0xD01E0513, 'HRESULT_FROM_NT(STATUS_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_HDCP)')
    NTSTATUS__STATUS_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_ACP: Final[ThirtyTwoBits] = cached_hresult(0xD01E0514, 'HRESULT_FROM_NT(STATUS_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_ACP)')
    NTSTATUS__STATUS_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_CGMSA: Final[ThirtyTwoBits] = cached_hresult(0xD01E0515, 'HRESULT_FROM_NT(STATUS_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_CGMSA)')
    NTSTATUS__STATUS_GRAPHICS_OPM_HDCP_SRM_NEVER_SET: Final[ThirtyTwoBits] = cached_hresult(0xD01E0516, 'HRESULT_FROM_NT(STATUS_GRAPHICS_OPM_HDCP_SRM_NEVER_SET)')
    NTSTATUS__STATUS_GRAPHICS_OPM_RESOLUTION_TOO_HIGH: Final[ThirtyTwoBits] = cached_hresult(0xD01E0517, 'HRESULT_FROM_NT(STATUS_GRAPHICS_OPM_RESOLUTION_TOO_HIGH)')
    NTSTATUS__STATUS_GRAPHICS_OPM_ALL_HDCP_HARDWARE_ALREADY_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0xD01E0518, 'HRESULT_FROM_NT(STATUS_GRAPHICS_OPM_ALL_HDCP_HARDWARE_ALREADY_IN_USE)')
    NTSTATUS__STATUS_GRAPHICS_OPM_PROTECTED_OUTPUT_NO_LONGER_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD01E051A, 'HRESULT_FROM_NT(STATUS_GRAPHICS_OPM_PROTECTED_OUTPUT_NO_LONGER_EXISTS)')
    NTSTATUS__STATUS_GRAPHICS_OPM_PROTECTED_OUTPUT_DOES_NOT_HAVE_COPP_SEMANTICS: Final[ThirtyTwoBits] = cached_hresult(0xD01E051C, 'HRESULT_FROM_NT(STATUS_GRAPHICS_OPM_PROTECTED_OUTPUT_DOES_NOT_HAVE_COPP_SEMANTICS)')
    NTSTATUS__STATUS_GRAPHICS_OPM_INVALID_INFORMATION_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0xD01E051D, 'HRESULT_FROM_NT(STATUS_GRAPHICS_OPM_INVALID_INFORMATION_REQUEST)')
    NTSTATUS__STATUS_GRAPHICS_OPM_DRIVER_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD01E051E, 'HRESULT_FROM_NT(STATUS_GRAPHICS_OPM_DRIVER_INTERNAL_ERROR)')
    NTSTATUS__STATUS_GRAPHICS_OPM_PROTECTED_OUTPUT_DOES_NOT_HAVE_OPM_SEMANTICS: Final[ThirtyTwoBits] = cached_hresult(0xD01E051F, 'HRESULT_FROM_NT(STATUS_GRAPHICS_OPM_PROTECTED_OUTPUT_DOES_NOT_HAVE_OPM_SEMANTICS)')
    NTSTATUS__STATUS_GRAPHICS_OPM_SIGNALING_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0520, 'HRESULT_FROM_NT(STATUS_GRAPHICS_OPM_SIGNALING_NOT_SUPPORTED)')
    NTSTATUS__STATUS_GRAPHICS_OPM_INVALID_CONFIGURATION_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0xD01E0521, 'HRESULT_FROM_NT(STATUS_GRAPHICS_OPM_INVALID_CONFIGURATION_REQUEST)')
    NTSTATUS__STATUS_GRAPHICS_I2C_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0580, 'HRESULT_FROM_NT(STATUS_GRAPHICS_I2C_NOT_SUPPORTED)')
    NTSTATUS__STATUS_GRAPHICS_I2C_DEVICE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0xD01E0581, 'HRESULT_FROM_NT(STATUS_GRAPHICS_I2C_DEVICE_DOES_NOT_EXIST)')
    NTSTATUS__STATUS_GRAPHICS_I2C_ERROR_TRANSMITTING_DATA: Final[ThirtyTwoBits] = cached_hresult(0xD01E0582, 'HRESULT_FROM_NT(STATUS_GRAPHICS_I2C_ERROR_TRANSMITTING_DATA)')
    NTSTATUS__STATUS_GRAPHICS_I2C_ERROR_RECEIVING_DATA: Final[ThirtyTwoBits] = cached_hresult(0xD01E0583, 'HRESULT_FROM_NT(STATUS_GRAPHICS_I2C_ERROR_RECEIVING_DATA)')
    NTSTATUS__STATUS_GRAPHICS_DDCCI_VCP_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD01E0584, 'HRESULT_FROM_NT(STATUS_GRAPHICS_DDCCI_VCP_NOT_SUPPORTED)')
    NTSTATUS__STATUS_GRAPHICS_DDCCI_INVALID_DATA: Final[ThirtyTwoBits] = cached_hresult(0xD01E0585, 'HRESULT_FROM_NT(STATUS_GRAPHICS_DDCCI_INVALID_DATA)')
    NTSTATUS__STATUS_GRAPHICS_DDCCI_MONITOR_RETURNED_INVALID_TIMING_STATUS_BYTE: Final[ThirtyTwoBits] = cached_hresult(0xD01E0586, 'HRESULT_FROM_NT(STATUS_GRAPHICS_DDCCI_MONITOR_RETURNED_INVALID_TIMING_STATUS_BYTE)')
    NTSTATUS__STATUS_GRAPHICS_DDCCI_INVALID_CAPABILITIES_STRING: Final[ThirtyTwoBits] = cached_hresult(0xD01E0587, 'HRESULT_FROM_NT(STATUS_GRAPHICS_DDCCI_INVALID_CAPABILITIES_STRING)')
    NTSTATUS__STATUS_GRAPHICS_MCA_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD01E0588, 'HRESULT_FROM_NT(STATUS_GRAPHICS_MCA_INTERNAL_ERROR)')
    NTSTATUS__STATUS_GRAPHICS_DDCCI_INVALID_MESSAGE_COMMAND: Final[ThirtyTwoBits] = cached_hresult(0xD01E0589, 'HRESULT_FROM_NT(STATUS_GRAPHICS_DDCCI_INVALID_MESSAGE_COMMAND)')
    NTSTATUS__STATUS_GRAPHICS_DDCCI_INVALID_MESSAGE_LENGTH: Final[ThirtyTwoBits] = cached_hresult(0xD01E058A, 'HRESULT_FROM_NT(STATUS_GRAPHICS_DDCCI_INVALID_MESSAGE_LENGTH)')
    NTSTATUS__STATUS_GRAPHICS_DDCCI_INVALID_MESSAGE_CHECKSUM: Final[ThirtyTwoBits] = cached_hresult(0xD01E058B, 'HRESULT_FROM_NT(STATUS_GRAPHICS_DDCCI_INVALID_MESSAGE_CHECKSUM)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_PHYSICAL_MONITOR_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0xD01E058C, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_PHYSICAL_MONITOR_HANDLE)')
    NTSTATUS__STATUS_GRAPHICS_MONITOR_NO_LONGER_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD01E058D, 'HRESULT_FROM_NT(STATUS_GRAPHICS_MONITOR_NO_LONGER_EXISTS)')
    NTSTATUS__STATUS_GRAPHICS_ONLY_CONSOLE_SESSION_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD01E05E0, 'HRESULT_FROM_NT(STATUS_GRAPHICS_ONLY_CONSOLE_SESSION_SUPPORTED)')
    NTSTATUS__STATUS_GRAPHICS_NO_DISPLAY_DEVICE_CORRESPONDS_TO_NAME: Final[ThirtyTwoBits] = cached_hresult(0xD01E05E1, 'HRESULT_FROM_NT(STATUS_GRAPHICS_NO_DISPLAY_DEVICE_CORRESPONDS_TO_NAME)')
    NTSTATUS__STATUS_GRAPHICS_DISPLAY_DEVICE_NOT_ATTACHED_TO_DESKTOP: Final[ThirtyTwoBits] = cached_hresult(0xD01E05E2, 'HRESULT_FROM_NT(STATUS_GRAPHICS_DISPLAY_DEVICE_NOT_ATTACHED_TO_DESKTOP)')
    NTSTATUS__STATUS_GRAPHICS_MIRRORING_DEVICES_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD01E05E3, 'HRESULT_FROM_NT(STATUS_GRAPHICS_MIRRORING_DEVICES_NOT_SUPPORTED)')
    NTSTATUS__STATUS_GRAPHICS_INVALID_POINTER: Final[ThirtyTwoBits] = cached_hresult(0xD01E05E4, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INVALID_POINTER)')
    NTSTATUS__STATUS_GRAPHICS_NO_MONITORS_CORRESPOND_TO_DISPLAY_DEVICE: Final[ThirtyTwoBits] = cached_hresult(0xD01E05E5, 'HRESULT_FROM_NT(STATUS_GRAPHICS_NO_MONITORS_CORRESPOND_TO_DISPLAY_DEVICE)')
    NTSTATUS__STATUS_GRAPHICS_PARAMETER_ARRAY_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0xD01E05E6, 'HRESULT_FROM_NT(STATUS_GRAPHICS_PARAMETER_ARRAY_TOO_SMALL)')
    NTSTATUS__STATUS_GRAPHICS_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD01E05E7, 'HRESULT_FROM_NT(STATUS_GRAPHICS_INTERNAL_ERROR)')
    NTSTATUS__STATUS_GRAPHICS_SESSION_TYPE_CHANGE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD01E05E8, 'HRESULT_FROM_NT(STATUS_GRAPHICS_SESSION_TYPE_CHANGE_IN_PROGRESS)')
    NTSTATUS__STATUS_FVE_LOCKED_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0xD0210000, 'HRESULT_FROM_NT(STATUS_FVE_LOCKED_VOLUME)')
    NTSTATUS__STATUS_FVE_NOT_ENCRYPTED: Final[ThirtyTwoBits] = cached_hresult(0xD0210001, 'HRESULT_FROM_NT(STATUS_FVE_NOT_ENCRYPTED)')
    NTSTATUS__STATUS_FVE_BAD_INFORMATION: Final[ThirtyTwoBits] = cached_hresult(0xD0210002, 'HRESULT_FROM_NT(STATUS_FVE_BAD_INFORMATION)')
    NTSTATUS__STATUS_FVE_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0xD0210003, 'HRESULT_FROM_NT(STATUS_FVE_TOO_SMALL)')
    NTSTATUS__STATUS_FVE_FAILED_WRONG_FS: Final[ThirtyTwoBits] = cached_hresult(0xD0210004, 'HRESULT_FROM_NT(STATUS_FVE_FAILED_WRONG_FS)')
    NTSTATUS__STATUS_FVE_BAD_PARTITION_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD0210005, 'HRESULT_FROM_NT(STATUS_FVE_BAD_PARTITION_SIZE)')
    NTSTATUS__STATUS_FVE_FS_NOT_EXTENDED: Final[ThirtyTwoBits] = cached_hresult(0xD0210006, 'HRESULT_FROM_NT(STATUS_FVE_FS_NOT_EXTENDED)')
    NTSTATUS__STATUS_FVE_FS_MOUNTED: Final[ThirtyTwoBits] = cached_hresult(0xD0210007, 'HRESULT_FROM_NT(STATUS_FVE_FS_MOUNTED)')
    NTSTATUS__STATUS_FVE_NO_LICENSE: Final[ThirtyTwoBits] = cached_hresult(0xD0210008, 'HRESULT_FROM_NT(STATUS_FVE_NO_LICENSE)')
    NTSTATUS__STATUS_FVE_ACTION_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0xD0210009, 'HRESULT_FROM_NT(STATUS_FVE_ACTION_NOT_ALLOWED)')
    NTSTATUS__STATUS_FVE_BAD_DATA: Final[ThirtyTwoBits] = cached_hresult(0xD021000A, 'HRESULT_FROM_NT(STATUS_FVE_BAD_DATA)')
    NTSTATUS__STATUS_FVE_VOLUME_NOT_BOUND: Final[ThirtyTwoBits] = cached_hresult(0xD021000B, 'HRESULT_FROM_NT(STATUS_FVE_VOLUME_NOT_BOUND)')
    NTSTATUS__STATUS_FVE_NOT_DATA_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0xD021000C, 'HRESULT_FROM_NT(STATUS_FVE_NOT_DATA_VOLUME)')
    NTSTATUS__STATUS_FVE_CONV_READ_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD021000D, 'HRESULT_FROM_NT(STATUS_FVE_CONV_READ_ERROR)')
    NTSTATUS__STATUS_FVE_CONV_WRITE_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD021000E, 'HRESULT_FROM_NT(STATUS_FVE_CONV_WRITE_ERROR)')
    NTSTATUS__STATUS_FVE_OVERLAPPED_UPDATE: Final[ThirtyTwoBits] = cached_hresult(0xD021000F, 'HRESULT_FROM_NT(STATUS_FVE_OVERLAPPED_UPDATE)')
    NTSTATUS__STATUS_FVE_FAILED_SECTOR_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD0210010, 'HRESULT_FROM_NT(STATUS_FVE_FAILED_SECTOR_SIZE)')
    NTSTATUS__STATUS_FVE_FAILED_AUTHENTICATION: Final[ThirtyTwoBits] = cached_hresult(0xD0210011, 'HRESULT_FROM_NT(STATUS_FVE_FAILED_AUTHENTICATION)')
    NTSTATUS__STATUS_FVE_NOT_OS_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0xD0210012, 'HRESULT_FROM_NT(STATUS_FVE_NOT_OS_VOLUME)')
    NTSTATUS__STATUS_FVE_KEYFILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0210013, 'HRESULT_FROM_NT(STATUS_FVE_KEYFILE_NOT_FOUND)')
    NTSTATUS__STATUS_FVE_KEYFILE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0210014, 'HRESULT_FROM_NT(STATUS_FVE_KEYFILE_INVALID)')
    NTSTATUS__STATUS_FVE_KEYFILE_NO_VMK: Final[ThirtyTwoBits] = cached_hresult(0xD0210015, 'HRESULT_FROM_NT(STATUS_FVE_KEYFILE_NO_VMK)')
    NTSTATUS__STATUS_FVE_TPM_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD0210016, 'HRESULT_FROM_NT(STATUS_FVE_TPM_DISABLED)')
    NTSTATUS__STATUS_FVE_TPM_SRK_AUTH_NOT_ZERO: Final[ThirtyTwoBits] = cached_hresult(0xD0210017, 'HRESULT_FROM_NT(STATUS_FVE_TPM_SRK_AUTH_NOT_ZERO)')
    NTSTATUS__STATUS_FVE_TPM_INVALID_PCR: Final[ThirtyTwoBits] = cached_hresult(0xD0210018, 'HRESULT_FROM_NT(STATUS_FVE_TPM_INVALID_PCR)')
    NTSTATUS__STATUS_FVE_TPM_NO_VMK: Final[ThirtyTwoBits] = cached_hresult(0xD0210019, 'HRESULT_FROM_NT(STATUS_FVE_TPM_NO_VMK)')
    NTSTATUS__STATUS_FVE_PIN_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD021001A, 'HRESULT_FROM_NT(STATUS_FVE_PIN_INVALID)')
    NTSTATUS__STATUS_FVE_AUTH_INVALID_APPLICATION: Final[ThirtyTwoBits] = cached_hresult(0xD021001B, 'HRESULT_FROM_NT(STATUS_FVE_AUTH_INVALID_APPLICATION)')
    NTSTATUS__STATUS_FVE_AUTH_INVALID_CONFIG: Final[ThirtyTwoBits] = cached_hresult(0xD021001C, 'HRESULT_FROM_NT(STATUS_FVE_AUTH_INVALID_CONFIG)')
    NTSTATUS__STATUS_FVE_DEBUGGER_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0xD021001D, 'HRESULT_FROM_NT(STATUS_FVE_DEBUGGER_ENABLED)')
    NTSTATUS__STATUS_FVE_DRY_RUN_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD021001E, 'HRESULT_FROM_NT(STATUS_FVE_DRY_RUN_FAILED)')
    NTSTATUS__STATUS_FVE_BAD_METADATA_POINTER: Final[ThirtyTwoBits] = cached_hresult(0xD021001F, 'HRESULT_FROM_NT(STATUS_FVE_BAD_METADATA_POINTER)')
    NTSTATUS__STATUS_FVE_OLD_METADATA_COPY: Final[ThirtyTwoBits] = cached_hresult(0xD0210020, 'HRESULT_FROM_NT(STATUS_FVE_OLD_METADATA_COPY)')
    NTSTATUS__STATUS_FVE_REBOOT_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0xD0210021, 'HRESULT_FROM_NT(STATUS_FVE_REBOOT_REQUIRED)')
    NTSTATUS__STATUS_FVE_RAW_ACCESS: Final[ThirtyTwoBits] = cached_hresult(0xD0210022, 'HRESULT_FROM_NT(STATUS_FVE_RAW_ACCESS)')
    NTSTATUS__STATUS_FVE_RAW_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0xD0210023, 'HRESULT_FROM_NT(STATUS_FVE_RAW_BLOCKED)')
    NTSTATUS__STATUS_FVE_NO_AUTOUNLOCK_MASTER_KEY: Final[ThirtyTwoBits] = cached_hresult(0xD0210024, 'HRESULT_FROM_NT(STATUS_FVE_NO_AUTOUNLOCK_MASTER_KEY)')
    NTSTATUS__STATUS_FVE_MOR_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0210025, 'HRESULT_FROM_NT(STATUS_FVE_MOR_FAILED)')
    NTSTATUS__STATUS_FVE_NO_FEATURE_LICENSE: Final[ThirtyTwoBits] = cached_hresult(0xD0210026, 'HRESULT_FROM_NT(STATUS_FVE_NO_FEATURE_LICENSE)')
    NTSTATUS__STATUS_FVE_POLICY_USER_DISABLE_RDV_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0xD0210027, 'HRESULT_FROM_NT(STATUS_FVE_POLICY_USER_DISABLE_RDV_NOT_ALLOWED)')
    NTSTATUS__STATUS_FVE_CONV_RECOVERY_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0210028, 'HRESULT_FROM_NT(STATUS_FVE_CONV_RECOVERY_FAILED)')
    NTSTATUS__STATUS_FVE_VIRTUALIZED_SPACE_TOO_BIG: Final[ThirtyTwoBits] = cached_hresult(0xD0210029, 'HRESULT_FROM_NT(STATUS_FVE_VIRTUALIZED_SPACE_TOO_BIG)')
    NTSTATUS__STATUS_FVE_INVALID_DATUM_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xD021002A, 'HRESULT_FROM_NT(STATUS_FVE_INVALID_DATUM_TYPE)')
    NTSTATUS__STATUS_FVE_VOLUME_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0xD0210030, 'HRESULT_FROM_NT(STATUS_FVE_VOLUME_TOO_SMALL)')
    NTSTATUS__STATUS_FVE_ENH_PIN_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0210031, 'HRESULT_FROM_NT(STATUS_FVE_ENH_PIN_INVALID)')
    NTSTATUS__STATUS_FVE_FULL_ENCRYPTION_NOT_ALLOWED_ON_TP_STORAGE: Final[ThirtyTwoBits] = cached_hresult(0xD0210032, 'HRESULT_FROM_NT(STATUS_FVE_FULL_ENCRYPTION_NOT_ALLOWED_ON_TP_STORAGE)')
    NTSTATUS__STATUS_FVE_WIPE_NOT_ALLOWED_ON_TP_STORAGE: Final[ThirtyTwoBits] = cached_hresult(0xD0210033, 'HRESULT_FROM_NT(STATUS_FVE_WIPE_NOT_ALLOWED_ON_TP_STORAGE)')
    NTSTATUS__STATUS_FVE_NOT_ALLOWED_ON_CSV_STACK: Final[ThirtyTwoBits] = cached_hresult(0xD0210034, 'HRESULT_FROM_NT(STATUS_FVE_NOT_ALLOWED_ON_CSV_STACK)')
    NTSTATUS__STATUS_FVE_NOT_ALLOWED_ON_CLUSTER: Final[ThirtyTwoBits] = cached_hresult(0xD0210035, 'HRESULT_FROM_NT(STATUS_FVE_NOT_ALLOWED_ON_CLUSTER)')
    NTSTATUS__STATUS_FVE_NOT_ALLOWED_TO_UPGRADE_WHILE_CONVERTING: Final[ThirtyTwoBits] = cached_hresult(0xD0210036, 'HRESULT_FROM_NT(STATUS_FVE_NOT_ALLOWED_TO_UPGRADE_WHILE_CONVERTING)')
    NTSTATUS__STATUS_FVE_WIPE_CANCEL_NOT_APPLICABLE: Final[ThirtyTwoBits] = cached_hresult(0xD0210037, 'HRESULT_FROM_NT(STATUS_FVE_WIPE_CANCEL_NOT_APPLICABLE)')
    NTSTATUS__STATUS_FVE_EDRIVE_DRY_RUN_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0210038, 'HRESULT_FROM_NT(STATUS_FVE_EDRIVE_DRY_RUN_FAILED)')
    NTSTATUS__STATUS_FVE_SECUREBOOT_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD0210039, 'HRESULT_FROM_NT(STATUS_FVE_SECUREBOOT_DISABLED)')
    NTSTATUS__STATUS_FVE_SECUREBOOT_CONFIG_CHANGE: Final[ThirtyTwoBits] = cached_hresult(0xD021003A, 'HRESULT_FROM_NT(STATUS_FVE_SECUREBOOT_CONFIG_CHANGE)')
    NTSTATUS__STATUS_FVE_DEVICE_LOCKEDOUT: Final[ThirtyTwoBits] = cached_hresult(0xD021003B, 'HRESULT_FROM_NT(STATUS_FVE_DEVICE_LOCKEDOUT)')
    NTSTATUS__STATUS_FVE_VOLUME_EXTEND_PREVENTS_EOW_DECRYPT: Final[ThirtyTwoBits] = cached_hresult(0xD021003C, 'HRESULT_FROM_NT(STATUS_FVE_VOLUME_EXTEND_PREVENTS_EOW_DECRYPT)')
    NTSTATUS__STATUS_FVE_NOT_DE_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0xD021003D, 'HRESULT_FROM_NT(STATUS_FVE_NOT_DE_VOLUME)')
    NTSTATUS__STATUS_FVE_PROTECTION_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD021003E, 'HRESULT_FROM_NT(STATUS_FVE_PROTECTION_DISABLED)')
    NTSTATUS__STATUS_FVE_PROTECTION_CANNOT_BE_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD021003F, 'HRESULT_FROM_NT(STATUS_FVE_PROTECTION_CANNOT_BE_DISABLED)')
    NTSTATUS__STATUS_FVE_OSV_KSR_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0xD0210040, 'HRESULT_FROM_NT(STATUS_FVE_OSV_KSR_NOT_ALLOWED)')
    NTSTATUS__STATUS_FVE_EDRIVE_BAND_ENUMERATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0210041, 'HRESULT_FROM_NT(STATUS_FVE_EDRIVE_BAND_ENUMERATION_FAILED)')
    NTSTATUS__STATUS_FVE_POLICY_ON_RDV_EXCLUSION_LIST: Final[ThirtyTwoBits] = cached_hresult(0xD0210042, 'HRESULT_FROM_NT(STATUS_FVE_POLICY_ON_RDV_EXCLUSION_LIST)')
    NTSTATUS__STATUS_FVE_DATASET_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD0210043, 'HRESULT_FROM_NT(STATUS_FVE_DATASET_FULL)')
    NTSTATUS__STATUS_FVE_METADATA_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD0210044, 'HRESULT_FROM_NT(STATUS_FVE_METADATA_FULL)')
    NTSTATUS__STATUS_FWP_CALLOUT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0220001, 'HRESULT_FROM_NT(STATUS_FWP_CALLOUT_NOT_FOUND)')
    NTSTATUS__STATUS_FWP_CONDITION_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0220002, 'HRESULT_FROM_NT(STATUS_FWP_CONDITION_NOT_FOUND)')
    NTSTATUS__STATUS_FWP_FILTER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0220003, 'HRESULT_FROM_NT(STATUS_FWP_FILTER_NOT_FOUND)')
    NTSTATUS__STATUS_FWP_LAYER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0220004, 'HRESULT_FROM_NT(STATUS_FWP_LAYER_NOT_FOUND)')
    NTSTATUS__STATUS_FWP_PROVIDER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0220005, 'HRESULT_FROM_NT(STATUS_FWP_PROVIDER_NOT_FOUND)')
    NTSTATUS__STATUS_FWP_PROVIDER_CONTEXT_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0220006, 'HRESULT_FROM_NT(STATUS_FWP_PROVIDER_CONTEXT_NOT_FOUND)')
    NTSTATUS__STATUS_FWP_SUBLAYER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0220007, 'HRESULT_FROM_NT(STATUS_FWP_SUBLAYER_NOT_FOUND)')
    NTSTATUS__STATUS_FWP_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0220008, 'HRESULT_FROM_NT(STATUS_FWP_NOT_FOUND)')
    NTSTATUS__STATUS_FWP_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD0220009, 'HRESULT_FROM_NT(STATUS_FWP_ALREADY_EXISTS)')
    NTSTATUS__STATUS_FWP_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0xD022000A, 'HRESULT_FROM_NT(STATUS_FWP_IN_USE)')
    NTSTATUS__STATUS_FWP_DYNAMIC_SESSION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD022000B, 'HRESULT_FROM_NT(STATUS_FWP_DYNAMIC_SESSION_IN_PROGRESS)')
    NTSTATUS__STATUS_FWP_WRONG_SESSION: Final[ThirtyTwoBits] = cached_hresult(0xD022000C, 'HRESULT_FROM_NT(STATUS_FWP_WRONG_SESSION)')
    NTSTATUS__STATUS_FWP_NO_TXN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD022000D, 'HRESULT_FROM_NT(STATUS_FWP_NO_TXN_IN_PROGRESS)')
    NTSTATUS__STATUS_FWP_TXN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD022000E, 'HRESULT_FROM_NT(STATUS_FWP_TXN_IN_PROGRESS)')
    NTSTATUS__STATUS_FWP_TXN_ABORTED: Final[ThirtyTwoBits] = cached_hresult(0xD022000F, 'HRESULT_FROM_NT(STATUS_FWP_TXN_ABORTED)')
    NTSTATUS__STATUS_FWP_SESSION_ABORTED: Final[ThirtyTwoBits] = cached_hresult(0xD0220010, 'HRESULT_FROM_NT(STATUS_FWP_SESSION_ABORTED)')
    NTSTATUS__STATUS_FWP_INCOMPATIBLE_TXN: Final[ThirtyTwoBits] = cached_hresult(0xD0220011, 'HRESULT_FROM_NT(STATUS_FWP_INCOMPATIBLE_TXN)')
    NTSTATUS__STATUS_FWP_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0xD0220012, 'HRESULT_FROM_NT(STATUS_FWP_TIMEOUT)')
    NTSTATUS__STATUS_FWP_NET_EVENTS_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD0220013, 'HRESULT_FROM_NT(STATUS_FWP_NET_EVENTS_DISABLED)')
    NTSTATUS__STATUS_FWP_INCOMPATIBLE_LAYER: Final[ThirtyTwoBits] = cached_hresult(0xD0220014, 'HRESULT_FROM_NT(STATUS_FWP_INCOMPATIBLE_LAYER)')
    NTSTATUS__STATUS_FWP_KM_CLIENTS_ONLY: Final[ThirtyTwoBits] = cached_hresult(0xD0220015, 'HRESULT_FROM_NT(STATUS_FWP_KM_CLIENTS_ONLY)')
    NTSTATUS__STATUS_FWP_LIFETIME_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD0220016, 'HRESULT_FROM_NT(STATUS_FWP_LIFETIME_MISMATCH)')
    NTSTATUS__STATUS_FWP_BUILTIN_OBJECT: Final[ThirtyTwoBits] = cached_hresult(0xD0220017, 'HRESULT_FROM_NT(STATUS_FWP_BUILTIN_OBJECT)')
    NTSTATUS__STATUS_FWP_TOO_MANY_CALLOUTS: Final[ThirtyTwoBits] = cached_hresult(0xD0220018, 'HRESULT_FROM_NT(STATUS_FWP_TOO_MANY_CALLOUTS)')
    NTSTATUS__STATUS_FWP_NOTIFICATION_DROPPED: Final[ThirtyTwoBits] = cached_hresult(0xD0220019, 'HRESULT_FROM_NT(STATUS_FWP_NOTIFICATION_DROPPED)')
    NTSTATUS__STATUS_FWP_TRAFFIC_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD022001A, 'HRESULT_FROM_NT(STATUS_FWP_TRAFFIC_MISMATCH)')
    NTSTATUS__STATUS_FWP_INCOMPATIBLE_SA_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD022001B, 'HRESULT_FROM_NT(STATUS_FWP_INCOMPATIBLE_SA_STATE)')
    NTSTATUS__STATUS_FWP_NULL_POINTER: Final[ThirtyTwoBits] = cached_hresult(0xD022001C, 'HRESULT_FROM_NT(STATUS_FWP_NULL_POINTER)')
    NTSTATUS__STATUS_FWP_INVALID_ENUMERATOR: Final[ThirtyTwoBits] = cached_hresult(0xD022001D, 'HRESULT_FROM_NT(STATUS_FWP_INVALID_ENUMERATOR)')
    NTSTATUS__STATUS_FWP_INVALID_FLAGS: Final[ThirtyTwoBits] = cached_hresult(0xD022001E, 'HRESULT_FROM_NT(STATUS_FWP_INVALID_FLAGS)')
    NTSTATUS__STATUS_FWP_INVALID_NET_MASK: Final[ThirtyTwoBits] = cached_hresult(0xD022001F, 'HRESULT_FROM_NT(STATUS_FWP_INVALID_NET_MASK)')
    NTSTATUS__STATUS_FWP_INVALID_RANGE: Final[ThirtyTwoBits] = cached_hresult(0xD0220020, 'HRESULT_FROM_NT(STATUS_FWP_INVALID_RANGE)')
    NTSTATUS__STATUS_FWP_INVALID_INTERVAL: Final[ThirtyTwoBits] = cached_hresult(0xD0220021, 'HRESULT_FROM_NT(STATUS_FWP_INVALID_INTERVAL)')
    NTSTATUS__STATUS_FWP_ZERO_LENGTH_ARRAY: Final[ThirtyTwoBits] = cached_hresult(0xD0220022, 'HRESULT_FROM_NT(STATUS_FWP_ZERO_LENGTH_ARRAY)')
    NTSTATUS__STATUS_FWP_NULL_DISPLAY_NAME: Final[ThirtyTwoBits] = cached_hresult(0xD0220023, 'HRESULT_FROM_NT(STATUS_FWP_NULL_DISPLAY_NAME)')
    NTSTATUS__STATUS_FWP_INVALID_ACTION_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xD0220024, 'HRESULT_FROM_NT(STATUS_FWP_INVALID_ACTION_TYPE)')
    NTSTATUS__STATUS_FWP_INVALID_WEIGHT: Final[ThirtyTwoBits] = cached_hresult(0xD0220025, 'HRESULT_FROM_NT(STATUS_FWP_INVALID_WEIGHT)')
    NTSTATUS__STATUS_FWP_MATCH_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD0220026, 'HRESULT_FROM_NT(STATUS_FWP_MATCH_TYPE_MISMATCH)')
    NTSTATUS__STATUS_FWP_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD0220027, 'HRESULT_FROM_NT(STATUS_FWP_TYPE_MISMATCH)')
    NTSTATUS__STATUS_FWP_OUT_OF_BOUNDS: Final[ThirtyTwoBits] = cached_hresult(0xD0220028, 'HRESULT_FROM_NT(STATUS_FWP_OUT_OF_BOUNDS)')
    NTSTATUS__STATUS_FWP_RESERVED: Final[ThirtyTwoBits] = cached_hresult(0xD0220029, 'HRESULT_FROM_NT(STATUS_FWP_RESERVED)')
    NTSTATUS__STATUS_FWP_DUPLICATE_CONDITION: Final[ThirtyTwoBits] = cached_hresult(0xD022002A, 'HRESULT_FROM_NT(STATUS_FWP_DUPLICATE_CONDITION)')
    NTSTATUS__STATUS_FWP_DUPLICATE_KEYMOD: Final[ThirtyTwoBits] = cached_hresult(0xD022002B, 'HRESULT_FROM_NT(STATUS_FWP_DUPLICATE_KEYMOD)')
    NTSTATUS__STATUS_FWP_ACTION_INCOMPATIBLE_WITH_LAYER: Final[ThirtyTwoBits] = cached_hresult(0xD022002C, 'HRESULT_FROM_NT(STATUS_FWP_ACTION_INCOMPATIBLE_WITH_LAYER)')
    NTSTATUS__STATUS_FWP_ACTION_INCOMPATIBLE_WITH_SUBLAYER: Final[ThirtyTwoBits] = cached_hresult(0xD022002D, 'HRESULT_FROM_NT(STATUS_FWP_ACTION_INCOMPATIBLE_WITH_SUBLAYER)')
    NTSTATUS__STATUS_FWP_CONTEXT_INCOMPATIBLE_WITH_LAYER: Final[ThirtyTwoBits] = cached_hresult(0xD022002E, 'HRESULT_FROM_NT(STATUS_FWP_CONTEXT_INCOMPATIBLE_WITH_LAYER)')
    NTSTATUS__STATUS_FWP_CONTEXT_INCOMPATIBLE_WITH_CALLOUT: Final[ThirtyTwoBits] = cached_hresult(0xD022002F, 'HRESULT_FROM_NT(STATUS_FWP_CONTEXT_INCOMPATIBLE_WITH_CALLOUT)')
    NTSTATUS__STATUS_FWP_INCOMPATIBLE_AUTH_METHOD: Final[ThirtyTwoBits] = cached_hresult(0xD0220030, 'HRESULT_FROM_NT(STATUS_FWP_INCOMPATIBLE_AUTH_METHOD)')
    NTSTATUS__STATUS_FWP_INCOMPATIBLE_DH_GROUP: Final[ThirtyTwoBits] = cached_hresult(0xD0220031, 'HRESULT_FROM_NT(STATUS_FWP_INCOMPATIBLE_DH_GROUP)')
    NTSTATUS__STATUS_FWP_EM_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD0220032, 'HRESULT_FROM_NT(STATUS_FWP_EM_NOT_SUPPORTED)')
    NTSTATUS__STATUS_FWP_NEVER_MATCH: Final[ThirtyTwoBits] = cached_hresult(0xD0220033, 'HRESULT_FROM_NT(STATUS_FWP_NEVER_MATCH)')
    NTSTATUS__STATUS_FWP_PROVIDER_CONTEXT_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD0220034, 'HRESULT_FROM_NT(STATUS_FWP_PROVIDER_CONTEXT_MISMATCH)')
    NTSTATUS__STATUS_FWP_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0xD0220035, 'HRESULT_FROM_NT(STATUS_FWP_INVALID_PARAMETER)')
    NTSTATUS__STATUS_FWP_TOO_MANY_SUBLAYERS: Final[ThirtyTwoBits] = cached_hresult(0xD0220036, 'HRESULT_FROM_NT(STATUS_FWP_TOO_MANY_SUBLAYERS)')
    NTSTATUS__STATUS_FWP_CALLOUT_NOTIFICATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0220037, 'HRESULT_FROM_NT(STATUS_FWP_CALLOUT_NOTIFICATION_FAILED)')
    NTSTATUS__STATUS_FWP_INVALID_AUTH_TRANSFORM: Final[ThirtyTwoBits] = cached_hresult(0xD0220038, 'HRESULT_FROM_NT(STATUS_FWP_INVALID_AUTH_TRANSFORM)')
    NTSTATUS__STATUS_FWP_INVALID_CIPHER_TRANSFORM: Final[ThirtyTwoBits] = cached_hresult(0xD0220039, 'HRESULT_FROM_NT(STATUS_FWP_INVALID_CIPHER_TRANSFORM)')
    NTSTATUS__STATUS_FWP_INCOMPATIBLE_CIPHER_TRANSFORM: Final[ThirtyTwoBits] = cached_hresult(0xD022003A, 'HRESULT_FROM_NT(STATUS_FWP_INCOMPATIBLE_CIPHER_TRANSFORM)')
    NTSTATUS__STATUS_FWP_INVALID_TRANSFORM_COMBINATION: Final[ThirtyTwoBits] = cached_hresult(0xD022003B, 'HRESULT_FROM_NT(STATUS_FWP_INVALID_TRANSFORM_COMBINATION)')
    NTSTATUS__STATUS_FWP_DUPLICATE_AUTH_METHOD: Final[ThirtyTwoBits] = cached_hresult(0xD022003C, 'HRESULT_FROM_NT(STATUS_FWP_DUPLICATE_AUTH_METHOD)')
    NTSTATUS__STATUS_FWP_INVALID_TUNNEL_ENDPOINT: Final[ThirtyTwoBits] = cached_hresult(0xD022003D, 'HRESULT_FROM_NT(STATUS_FWP_INVALID_TUNNEL_ENDPOINT)')
    NTSTATUS__STATUS_FWP_L2_DRIVER_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0xD022003E, 'HRESULT_FROM_NT(STATUS_FWP_L2_DRIVER_NOT_READY)')
    NTSTATUS__STATUS_FWP_KEY_DICTATOR_ALREADY_REGISTERED: Final[ThirtyTwoBits] = cached_hresult(0xD022003F, 'HRESULT_FROM_NT(STATUS_FWP_KEY_DICTATOR_ALREADY_REGISTERED)')
    NTSTATUS__STATUS_FWP_KEY_DICTATION_INVALID_KEYING_MATERIAL: Final[ThirtyTwoBits] = cached_hresult(0xD0220040, 'HRESULT_FROM_NT(STATUS_FWP_KEY_DICTATION_INVALID_KEYING_MATERIAL)')
    NTSTATUS__STATUS_FWP_CONNECTIONS_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD0220041, 'HRESULT_FROM_NT(STATUS_FWP_CONNECTIONS_DISABLED)')
    NTSTATUS__STATUS_FWP_INVALID_DNS_NAME: Final[ThirtyTwoBits] = cached_hresult(0xD0220042, 'HRESULT_FROM_NT(STATUS_FWP_INVALID_DNS_NAME)')
    NTSTATUS__STATUS_FWP_STILL_ON: Final[ThirtyTwoBits] = cached_hresult(0xD0220043, 'HRESULT_FROM_NT(STATUS_FWP_STILL_ON)')
    NTSTATUS__STATUS_FWP_IKEEXT_NOT_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0xD0220044, 'HRESULT_FROM_NT(STATUS_FWP_IKEEXT_NOT_RUNNING)')
    NTSTATUS__STATUS_FWP_TCPIP_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0xD0220100, 'HRESULT_FROM_NT(STATUS_FWP_TCPIP_NOT_READY)')
    NTSTATUS__STATUS_FWP_INJECT_HANDLE_CLOSING: Final[ThirtyTwoBits] = cached_hresult(0xD0220101, 'HRESULT_FROM_NT(STATUS_FWP_INJECT_HANDLE_CLOSING)')
    NTSTATUS__STATUS_FWP_INJECT_HANDLE_STALE: Final[ThirtyTwoBits] = cached_hresult(0xD0220102, 'HRESULT_FROM_NT(STATUS_FWP_INJECT_HANDLE_STALE)')
    NTSTATUS__STATUS_FWP_CANNOT_PEND: Final[ThirtyTwoBits] = cached_hresult(0xD0220103, 'HRESULT_FROM_NT(STATUS_FWP_CANNOT_PEND)')
    NTSTATUS__STATUS_FWP_DROP_NOICMP: Final[ThirtyTwoBits] = cached_hresult(0xD0220104, 'HRESULT_FROM_NT(STATUS_FWP_DROP_NOICMP)')
    NTSTATUS__STATUS_NDIS_CLOSING: Final[ThirtyTwoBits] = cached_hresult(0xD0230002, 'HRESULT_FROM_NT(STATUS_NDIS_CLOSING)')
    NTSTATUS__STATUS_NDIS_BAD_VERSION: Final[ThirtyTwoBits] = cached_hresult(0xD0230004, 'HRESULT_FROM_NT(STATUS_NDIS_BAD_VERSION)')
    NTSTATUS__STATUS_NDIS_BAD_CHARACTERISTICS: Final[ThirtyTwoBits] = cached_hresult(0xD0230005, 'HRESULT_FROM_NT(STATUS_NDIS_BAD_CHARACTERISTICS)')
    NTSTATUS__STATUS_NDIS_ADAPTER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0230006, 'HRESULT_FROM_NT(STATUS_NDIS_ADAPTER_NOT_FOUND)')
    NTSTATUS__STATUS_NDIS_OPEN_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0230007, 'HRESULT_FROM_NT(STATUS_NDIS_OPEN_FAILED)')
    NTSTATUS__STATUS_NDIS_DEVICE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0230008, 'HRESULT_FROM_NT(STATUS_NDIS_DEVICE_FAILED)')
    NTSTATUS__STATUS_NDIS_MULTICAST_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD0230009, 'HRESULT_FROM_NT(STATUS_NDIS_MULTICAST_FULL)')
    NTSTATUS__STATUS_NDIS_MULTICAST_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD023000A, 'HRESULT_FROM_NT(STATUS_NDIS_MULTICAST_EXISTS)')
    NTSTATUS__STATUS_NDIS_MULTICAST_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD023000B, 'HRESULT_FROM_NT(STATUS_NDIS_MULTICAST_NOT_FOUND)')
    NTSTATUS__STATUS_NDIS_REQUEST_ABORTED: Final[ThirtyTwoBits] = cached_hresult(0xD023000C, 'HRESULT_FROM_NT(STATUS_NDIS_REQUEST_ABORTED)')
    NTSTATUS__STATUS_NDIS_RESET_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD023000D, 'HRESULT_FROM_NT(STATUS_NDIS_RESET_IN_PROGRESS)')
    NTSTATUS__STATUS_NDIS_INVALID_PACKET: Final[ThirtyTwoBits] = cached_hresult(0xD023000F, 'HRESULT_FROM_NT(STATUS_NDIS_INVALID_PACKET)')
    NTSTATUS__STATUS_NDIS_INVALID_DEVICE_REQUEST: Final[ThirtyTwoBits] = cached_hresult(0xD0230010, 'HRESULT_FROM_NT(STATUS_NDIS_INVALID_DEVICE_REQUEST)')
    NTSTATUS__STATUS_NDIS_ADAPTER_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0xD0230011, 'HRESULT_FROM_NT(STATUS_NDIS_ADAPTER_NOT_READY)')
    NTSTATUS__STATUS_NDIS_INVALID_LENGTH: Final[ThirtyTwoBits] = cached_hresult(0xD0230014, 'HRESULT_FROM_NT(STATUS_NDIS_INVALID_LENGTH)')
    NTSTATUS__STATUS_NDIS_INVALID_DATA: Final[ThirtyTwoBits] = cached_hresult(0xD0230015, 'HRESULT_FROM_NT(STATUS_NDIS_INVALID_DATA)')
    NTSTATUS__STATUS_NDIS_BUFFER_TOO_SHORT: Final[ThirtyTwoBits] = cached_hresult(0xD0230016, 'HRESULT_FROM_NT(STATUS_NDIS_BUFFER_TOO_SHORT)')
    NTSTATUS__STATUS_NDIS_INVALID_OID: Final[ThirtyTwoBits] = cached_hresult(0xD0230017, 'HRESULT_FROM_NT(STATUS_NDIS_INVALID_OID)')
    NTSTATUS__STATUS_NDIS_ADAPTER_REMOVED: Final[ThirtyTwoBits] = cached_hresult(0xD0230018, 'HRESULT_FROM_NT(STATUS_NDIS_ADAPTER_REMOVED)')
    NTSTATUS__STATUS_NDIS_UNSUPPORTED_MEDIA: Final[ThirtyTwoBits] = cached_hresult(0xD0230019, 'HRESULT_FROM_NT(STATUS_NDIS_UNSUPPORTED_MEDIA)')
    NTSTATUS__STATUS_NDIS_GROUP_ADDRESS_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0xD023001A, 'HRESULT_FROM_NT(STATUS_NDIS_GROUP_ADDRESS_IN_USE)')
    NTSTATUS__STATUS_NDIS_FILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD023001B, 'HRESULT_FROM_NT(STATUS_NDIS_FILE_NOT_FOUND)')
    NTSTATUS__STATUS_NDIS_ERROR_READING_FILE: Final[ThirtyTwoBits] = cached_hresult(0xD023001C, 'HRESULT_FROM_NT(STATUS_NDIS_ERROR_READING_FILE)')
    NTSTATUS__STATUS_NDIS_ALREADY_MAPPED: Final[ThirtyTwoBits] = cached_hresult(0xD023001D, 'HRESULT_FROM_NT(STATUS_NDIS_ALREADY_MAPPED)')
    NTSTATUS__STATUS_NDIS_RESOURCE_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0xD023001E, 'HRESULT_FROM_NT(STATUS_NDIS_RESOURCE_CONFLICT)')
    NTSTATUS__STATUS_NDIS_MEDIA_DISCONNECTED: Final[ThirtyTwoBits] = cached_hresult(0xD023001F, 'HRESULT_FROM_NT(STATUS_NDIS_MEDIA_DISCONNECTED)')
    NTSTATUS__STATUS_NDIS_INVALID_ADDRESS: Final[ThirtyTwoBits] = cached_hresult(0xD0230022, 'HRESULT_FROM_NT(STATUS_NDIS_INVALID_ADDRESS)')
    NTSTATUS__STATUS_NDIS_PAUSED: Final[ThirtyTwoBits] = cached_hresult(0xD023002A, 'HRESULT_FROM_NT(STATUS_NDIS_PAUSED)')
    NTSTATUS__STATUS_NDIS_INTERFACE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD023002B, 'HRESULT_FROM_NT(STATUS_NDIS_INTERFACE_NOT_FOUND)')
    NTSTATUS__STATUS_NDIS_UNSUPPORTED_REVISION: Final[ThirtyTwoBits] = cached_hresult(0xD023002C, 'HRESULT_FROM_NT(STATUS_NDIS_UNSUPPORTED_REVISION)')
    NTSTATUS__STATUS_NDIS_INVALID_PORT: Final[ThirtyTwoBits] = cached_hresult(0xD023002D, 'HRESULT_FROM_NT(STATUS_NDIS_INVALID_PORT)')
    NTSTATUS__STATUS_NDIS_INVALID_PORT_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD023002E, 'HRESULT_FROM_NT(STATUS_NDIS_INVALID_PORT_STATE)')
    NTSTATUS__STATUS_NDIS_LOW_POWER_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD023002F, 'HRESULT_FROM_NT(STATUS_NDIS_LOW_POWER_STATE)')
    NTSTATUS__STATUS_NDIS_REINIT_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0xD0230030, 'HRESULT_FROM_NT(STATUS_NDIS_REINIT_REQUIRED)')
    NTSTATUS__STATUS_NDIS_NO_QUEUES: Final[ThirtyTwoBits] = cached_hresult(0xD0230031, 'HRESULT_FROM_NT(STATUS_NDIS_NO_QUEUES)')
    NTSTATUS__STATUS_NDIS_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD02300BB, 'HRESULT_FROM_NT(STATUS_NDIS_NOT_SUPPORTED)')
    NTSTATUS__STATUS_NDIS_OFFLOAD_POLICY: Final[ThirtyTwoBits] = cached_hresult(0xD023100F, 'HRESULT_FROM_NT(STATUS_NDIS_OFFLOAD_POLICY)')
    NTSTATUS__STATUS_NDIS_OFFLOAD_CONNECTION_REJECTED: Final[ThirtyTwoBits] = cached_hresult(0xD0231012, 'HRESULT_FROM_NT(STATUS_NDIS_OFFLOAD_CONNECTION_REJECTED)')
    NTSTATUS__STATUS_NDIS_OFFLOAD_PATH_REJECTED: Final[ThirtyTwoBits] = cached_hresult(0xD0231013, 'HRESULT_FROM_NT(STATUS_NDIS_OFFLOAD_PATH_REJECTED)')
    NTSTATUS__STATUS_NDIS_DOT11_AUTO_CONFIG_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0xD0232000, 'HRESULT_FROM_NT(STATUS_NDIS_DOT11_AUTO_CONFIG_ENABLED)')
    NTSTATUS__STATUS_NDIS_DOT11_MEDIA_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0xD0232001, 'HRESULT_FROM_NT(STATUS_NDIS_DOT11_MEDIA_IN_USE)')
    NTSTATUS__STATUS_NDIS_DOT11_POWER_STATE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0232002, 'HRESULT_FROM_NT(STATUS_NDIS_DOT11_POWER_STATE_INVALID)')
    NTSTATUS__STATUS_NDIS_PM_WOL_PATTERN_LIST_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD0232003, 'HRESULT_FROM_NT(STATUS_NDIS_PM_WOL_PATTERN_LIST_FULL)')
    NTSTATUS__STATUS_NDIS_PM_PROTOCOL_OFFLOAD_LIST_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD0232004, 'HRESULT_FROM_NT(STATUS_NDIS_PM_PROTOCOL_OFFLOAD_LIST_FULL)')
    NTSTATUS__STATUS_NDIS_DOT11_AP_CHANNEL_CURRENTLY_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD0232005, 'HRESULT_FROM_NT(STATUS_NDIS_DOT11_AP_CHANNEL_CURRENTLY_NOT_AVAILABLE)')
    NTSTATUS__STATUS_NDIS_DOT11_AP_BAND_CURRENTLY_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD0232006, 'HRESULT_FROM_NT(STATUS_NDIS_DOT11_AP_BAND_CURRENTLY_NOT_AVAILABLE)')
    NTSTATUS__STATUS_NDIS_DOT11_AP_CHANNEL_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0xD0232007, 'HRESULT_FROM_NT(STATUS_NDIS_DOT11_AP_CHANNEL_NOT_ALLOWED)')
    NTSTATUS__STATUS_NDIS_DOT11_AP_BAND_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0xD0232008, 'HRESULT_FROM_NT(STATUS_NDIS_DOT11_AP_BAND_NOT_ALLOWED)')
    NTSTATUS__STATUS_QUIC_HANDSHAKE_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD0240000, 'HRESULT_FROM_NT(STATUS_QUIC_HANDSHAKE_FAILURE)')
    NTSTATUS__STATUS_QUIC_VER_NEG_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD0240001, 'HRESULT_FROM_NT(STATUS_QUIC_VER_NEG_FAILURE)')
    NTSTATUS__STATUS_QUIC_USER_CANCELED: Final[ThirtyTwoBits] = cached_hresult(0xD0240002, 'HRESULT_FROM_NT(STATUS_QUIC_USER_CANCELED)')
    NTSTATUS__STATUS_QUIC_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0240003, 'HRESULT_FROM_NT(STATUS_QUIC_INTERNAL_ERROR)')
    NTSTATUS__STATUS_QUIC_PROTOCOL_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0xD0240004, 'HRESULT_FROM_NT(STATUS_QUIC_PROTOCOL_VIOLATION)')
    NTSTATUS__STATUS_QUIC_CONNECTION_IDLE: Final[ThirtyTwoBits] = cached_hresult(0xD0240005, 'HRESULT_FROM_NT(STATUS_QUIC_CONNECTION_IDLE)')
    NTSTATUS__STATUS_QUIC_CONNECTION_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0xD0240006, 'HRESULT_FROM_NT(STATUS_QUIC_CONNECTION_TIMEOUT)')
    NTSTATUS__STATUS_QUIC_ALPN_NEG_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD0240007, 'HRESULT_FROM_NT(STATUS_QUIC_ALPN_NEG_FAILURE)')
    NTSTATUS__STATUS_TPM_ERROR_MASK: Final[ThirtyTwoBits] = cached_hresult(0xD0290000, 'HRESULT_FROM_NT(STATUS_TPM_ERROR_MASK)')
    NTSTATUS__STATUS_TPM_AUTHFAIL: Final[ThirtyTwoBits] = cached_hresult(0xD0290001, 'HRESULT_FROM_NT(STATUS_TPM_AUTHFAIL)')
    NTSTATUS__STATUS_TPM_BADINDEX: Final[ThirtyTwoBits] = cached_hresult(0xD0290002, 'HRESULT_FROM_NT(STATUS_TPM_BADINDEX)')
    NTSTATUS__STATUS_TPM_BAD_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0xD0290003, 'HRESULT_FROM_NT(STATUS_TPM_BAD_PARAMETER)')
    NTSTATUS__STATUS_TPM_AUDITFAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD0290004, 'HRESULT_FROM_NT(STATUS_TPM_AUDITFAILURE)')
    NTSTATUS__STATUS_TPM_CLEAR_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD0290005, 'HRESULT_FROM_NT(STATUS_TPM_CLEAR_DISABLED)')
    NTSTATUS__STATUS_TPM_DEACTIVATED: Final[ThirtyTwoBits] = cached_hresult(0xD0290006, 'HRESULT_FROM_NT(STATUS_TPM_DEACTIVATED)')
    NTSTATUS__STATUS_TPM_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD0290007, 'HRESULT_FROM_NT(STATUS_TPM_DISABLED)')
    NTSTATUS__STATUS_TPM_DISABLED_CMD: Final[ThirtyTwoBits] = cached_hresult(0xD0290008, 'HRESULT_FROM_NT(STATUS_TPM_DISABLED_CMD)')
    NTSTATUS__STATUS_TPM_FAIL: Final[ThirtyTwoBits] = cached_hresult(0xD0290009, 'HRESULT_FROM_NT(STATUS_TPM_FAIL)')
    NTSTATUS__STATUS_TPM_BAD_ORDINAL: Final[ThirtyTwoBits] = cached_hresult(0xD029000A, 'HRESULT_FROM_NT(STATUS_TPM_BAD_ORDINAL)')
    NTSTATUS__STATUS_TPM_INSTALL_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD029000B, 'HRESULT_FROM_NT(STATUS_TPM_INSTALL_DISABLED)')
    NTSTATUS__STATUS_TPM_INVALID_KEYHANDLE: Final[ThirtyTwoBits] = cached_hresult(0xD029000C, 'HRESULT_FROM_NT(STATUS_TPM_INVALID_KEYHANDLE)')
    NTSTATUS__STATUS_TPM_KEYNOTFOUND: Final[ThirtyTwoBits] = cached_hresult(0xD029000D, 'HRESULT_FROM_NT(STATUS_TPM_KEYNOTFOUND)')
    NTSTATUS__STATUS_TPM_INAPPROPRIATE_ENC: Final[ThirtyTwoBits] = cached_hresult(0xD029000E, 'HRESULT_FROM_NT(STATUS_TPM_INAPPROPRIATE_ENC)')
    NTSTATUS__STATUS_TPM_MIGRATEFAIL: Final[ThirtyTwoBits] = cached_hresult(0xD029000F, 'HRESULT_FROM_NT(STATUS_TPM_MIGRATEFAIL)')
    NTSTATUS__STATUS_TPM_INVALID_PCR_INFO: Final[ThirtyTwoBits] = cached_hresult(0xD0290010, 'HRESULT_FROM_NT(STATUS_TPM_INVALID_PCR_INFO)')
    NTSTATUS__STATUS_TPM_NOSPACE: Final[ThirtyTwoBits] = cached_hresult(0xD0290011, 'HRESULT_FROM_NT(STATUS_TPM_NOSPACE)')
    NTSTATUS__STATUS_TPM_NOSRK: Final[ThirtyTwoBits] = cached_hresult(0xD0290012, 'HRESULT_FROM_NT(STATUS_TPM_NOSRK)')
    NTSTATUS__STATUS_TPM_NOTSEALED_BLOB: Final[ThirtyTwoBits] = cached_hresult(0xD0290013, 'HRESULT_FROM_NT(STATUS_TPM_NOTSEALED_BLOB)')
    NTSTATUS__STATUS_TPM_OWNER_SET: Final[ThirtyTwoBits] = cached_hresult(0xD0290014, 'HRESULT_FROM_NT(STATUS_TPM_OWNER_SET)')
    NTSTATUS__STATUS_TPM_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0xD0290015, 'HRESULT_FROM_NT(STATUS_TPM_RESOURCES)')
    NTSTATUS__STATUS_TPM_SHORTRANDOM: Final[ThirtyTwoBits] = cached_hresult(0xD0290016, 'HRESULT_FROM_NT(STATUS_TPM_SHORTRANDOM)')
    NTSTATUS__STATUS_TPM_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD0290017, 'HRESULT_FROM_NT(STATUS_TPM_SIZE)')
    NTSTATUS__STATUS_TPM_WRONGPCRVAL: Final[ThirtyTwoBits] = cached_hresult(0xD0290018, 'HRESULT_FROM_NT(STATUS_TPM_WRONGPCRVAL)')
    NTSTATUS__STATUS_TPM_BAD_PARAM_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD0290019, 'HRESULT_FROM_NT(STATUS_TPM_BAD_PARAM_SIZE)')
    NTSTATUS__STATUS_TPM_SHA_THREAD: Final[ThirtyTwoBits] = cached_hresult(0xD029001A, 'HRESULT_FROM_NT(STATUS_TPM_SHA_THREAD)')
    NTSTATUS__STATUS_TPM_SHA_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD029001B, 'HRESULT_FROM_NT(STATUS_TPM_SHA_ERROR)')
    NTSTATUS__STATUS_TPM_FAILEDSELFTEST: Final[ThirtyTwoBits] = cached_hresult(0xD029001C, 'HRESULT_FROM_NT(STATUS_TPM_FAILEDSELFTEST)')
    NTSTATUS__STATUS_TPM_AUTH2FAIL: Final[ThirtyTwoBits] = cached_hresult(0xD029001D, 'HRESULT_FROM_NT(STATUS_TPM_AUTH2FAIL)')
    NTSTATUS__STATUS_TPM_BADTAG: Final[ThirtyTwoBits] = cached_hresult(0xD029001E, 'HRESULT_FROM_NT(STATUS_TPM_BADTAG)')
    NTSTATUS__STATUS_TPM_IOERROR: Final[ThirtyTwoBits] = cached_hresult(0xD029001F, 'HRESULT_FROM_NT(STATUS_TPM_IOERROR)')
    NTSTATUS__STATUS_TPM_ENCRYPT_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0290020, 'HRESULT_FROM_NT(STATUS_TPM_ENCRYPT_ERROR)')
    NTSTATUS__STATUS_TPM_DECRYPT_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0290021, 'HRESULT_FROM_NT(STATUS_TPM_DECRYPT_ERROR)')
    NTSTATUS__STATUS_TPM_INVALID_AUTHHANDLE: Final[ThirtyTwoBits] = cached_hresult(0xD0290022, 'HRESULT_FROM_NT(STATUS_TPM_INVALID_AUTHHANDLE)')
    NTSTATUS__STATUS_TPM_NO_ENDORSEMENT: Final[ThirtyTwoBits] = cached_hresult(0xD0290023, 'HRESULT_FROM_NT(STATUS_TPM_NO_ENDORSEMENT)')
    NTSTATUS__STATUS_TPM_INVALID_KEYUSAGE: Final[ThirtyTwoBits] = cached_hresult(0xD0290024, 'HRESULT_FROM_NT(STATUS_TPM_INVALID_KEYUSAGE)')
    NTSTATUS__STATUS_TPM_WRONG_ENTITYTYPE: Final[ThirtyTwoBits] = cached_hresult(0xD0290025, 'HRESULT_FROM_NT(STATUS_TPM_WRONG_ENTITYTYPE)')
    NTSTATUS__STATUS_TPM_INVALID_POSTINIT: Final[ThirtyTwoBits] = cached_hresult(0xD0290026, 'HRESULT_FROM_NT(STATUS_TPM_INVALID_POSTINIT)')
    NTSTATUS__STATUS_TPM_INAPPROPRIATE_SIG: Final[ThirtyTwoBits] = cached_hresult(0xD0290027, 'HRESULT_FROM_NT(STATUS_TPM_INAPPROPRIATE_SIG)')
    NTSTATUS__STATUS_TPM_BAD_KEY_PROPERTY: Final[ThirtyTwoBits] = cached_hresult(0xD0290028, 'HRESULT_FROM_NT(STATUS_TPM_BAD_KEY_PROPERTY)')
    NTSTATUS__STATUS_TPM_BAD_MIGRATION: Final[ThirtyTwoBits] = cached_hresult(0xD0290029, 'HRESULT_FROM_NT(STATUS_TPM_BAD_MIGRATION)')
    NTSTATUS__STATUS_TPM_BAD_SCHEME: Final[ThirtyTwoBits] = cached_hresult(0xD029002A, 'HRESULT_FROM_NT(STATUS_TPM_BAD_SCHEME)')
    NTSTATUS__STATUS_TPM_BAD_DATASIZE: Final[ThirtyTwoBits] = cached_hresult(0xD029002B, 'HRESULT_FROM_NT(STATUS_TPM_BAD_DATASIZE)')
    NTSTATUS__STATUS_TPM_BAD_MODE: Final[ThirtyTwoBits] = cached_hresult(0xD029002C, 'HRESULT_FROM_NT(STATUS_TPM_BAD_MODE)')
    NTSTATUS__STATUS_TPM_BAD_PRESENCE: Final[ThirtyTwoBits] = cached_hresult(0xD029002D, 'HRESULT_FROM_NT(STATUS_TPM_BAD_PRESENCE)')
    NTSTATUS__STATUS_TPM_BAD_VERSION: Final[ThirtyTwoBits] = cached_hresult(0xD029002E, 'HRESULT_FROM_NT(STATUS_TPM_BAD_VERSION)')
    NTSTATUS__STATUS_TPM_NO_WRAP_TRANSPORT: Final[ThirtyTwoBits] = cached_hresult(0xD029002F, 'HRESULT_FROM_NT(STATUS_TPM_NO_WRAP_TRANSPORT)')
    NTSTATUS__STATUS_TPM_AUDITFAIL_UNSUCCESSFUL: Final[ThirtyTwoBits] = cached_hresult(0xD0290030, 'HRESULT_FROM_NT(STATUS_TPM_AUDITFAIL_UNSUCCESSFUL)')
    NTSTATUS__STATUS_TPM_AUDITFAIL_SUCCESSFUL: Final[ThirtyTwoBits] = cached_hresult(0xD0290031, 'HRESULT_FROM_NT(STATUS_TPM_AUDITFAIL_SUCCESSFUL)')
    NTSTATUS__STATUS_TPM_NOTRESETABLE: Final[ThirtyTwoBits] = cached_hresult(0xD0290032, 'HRESULT_FROM_NT(STATUS_TPM_NOTRESETABLE)')
    NTSTATUS__STATUS_TPM_NOTLOCAL: Final[ThirtyTwoBits] = cached_hresult(0xD0290033, 'HRESULT_FROM_NT(STATUS_TPM_NOTLOCAL)')
    NTSTATUS__STATUS_TPM_BAD_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xD0290034, 'HRESULT_FROM_NT(STATUS_TPM_BAD_TYPE)')
    NTSTATUS__STATUS_TPM_INVALID_RESOURCE: Final[ThirtyTwoBits] = cached_hresult(0xD0290035, 'HRESULT_FROM_NT(STATUS_TPM_INVALID_RESOURCE)')
    NTSTATUS__STATUS_TPM_NOTFIPS: Final[ThirtyTwoBits] = cached_hresult(0xD0290036, 'HRESULT_FROM_NT(STATUS_TPM_NOTFIPS)')
    NTSTATUS__STATUS_TPM_INVALID_FAMILY: Final[ThirtyTwoBits] = cached_hresult(0xD0290037, 'HRESULT_FROM_NT(STATUS_TPM_INVALID_FAMILY)')
    NTSTATUS__STATUS_TPM_NO_NV_PERMISSION: Final[ThirtyTwoBits] = cached_hresult(0xD0290038, 'HRESULT_FROM_NT(STATUS_TPM_NO_NV_PERMISSION)')
    NTSTATUS__STATUS_TPM_REQUIRES_SIGN: Final[ThirtyTwoBits] = cached_hresult(0xD0290039, 'HRESULT_FROM_NT(STATUS_TPM_REQUIRES_SIGN)')
    NTSTATUS__STATUS_TPM_KEY_NOTSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD029003A, 'HRESULT_FROM_NT(STATUS_TPM_KEY_NOTSUPPORTED)')
    NTSTATUS__STATUS_TPM_AUTH_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0xD029003B, 'HRESULT_FROM_NT(STATUS_TPM_AUTH_CONFLICT)')
    NTSTATUS__STATUS_TPM_AREA_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0xD029003C, 'HRESULT_FROM_NT(STATUS_TPM_AREA_LOCKED)')
    NTSTATUS__STATUS_TPM_BAD_LOCALITY: Final[ThirtyTwoBits] = cached_hresult(0xD029003D, 'HRESULT_FROM_NT(STATUS_TPM_BAD_LOCALITY)')
    NTSTATUS__STATUS_TPM_READ_ONLY: Final[ThirtyTwoBits] = cached_hresult(0xD029003E, 'HRESULT_FROM_NT(STATUS_TPM_READ_ONLY)')
    NTSTATUS__STATUS_TPM_PER_NOWRITE: Final[ThirtyTwoBits] = cached_hresult(0xD029003F, 'HRESULT_FROM_NT(STATUS_TPM_PER_NOWRITE)')
    NTSTATUS__STATUS_TPM_FAMILYCOUNT: Final[ThirtyTwoBits] = cached_hresult(0xD0290040, 'HRESULT_FROM_NT(STATUS_TPM_FAMILYCOUNT)')
    NTSTATUS__STATUS_TPM_WRITE_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0xD0290041, 'HRESULT_FROM_NT(STATUS_TPM_WRITE_LOCKED)')
    NTSTATUS__STATUS_TPM_BAD_ATTRIBUTES: Final[ThirtyTwoBits] = cached_hresult(0xD0290042, 'HRESULT_FROM_NT(STATUS_TPM_BAD_ATTRIBUTES)')
    NTSTATUS__STATUS_TPM_INVALID_STRUCTURE: Final[ThirtyTwoBits] = cached_hresult(0xD0290043, 'HRESULT_FROM_NT(STATUS_TPM_INVALID_STRUCTURE)')
    NTSTATUS__STATUS_TPM_KEY_OWNER_CONTROL: Final[ThirtyTwoBits] = cached_hresult(0xD0290044, 'HRESULT_FROM_NT(STATUS_TPM_KEY_OWNER_CONTROL)')
    NTSTATUS__STATUS_TPM_BAD_COUNTER: Final[ThirtyTwoBits] = cached_hresult(0xD0290045, 'HRESULT_FROM_NT(STATUS_TPM_BAD_COUNTER)')
    NTSTATUS__STATUS_TPM_NOT_FULLWRITE: Final[ThirtyTwoBits] = cached_hresult(0xD0290046, 'HRESULT_FROM_NT(STATUS_TPM_NOT_FULLWRITE)')
    NTSTATUS__STATUS_TPM_CONTEXT_GAP: Final[ThirtyTwoBits] = cached_hresult(0xD0290047, 'HRESULT_FROM_NT(STATUS_TPM_CONTEXT_GAP)')
    NTSTATUS__STATUS_TPM_MAXNVWRITES: Final[ThirtyTwoBits] = cached_hresult(0xD0290048, 'HRESULT_FROM_NT(STATUS_TPM_MAXNVWRITES)')
    NTSTATUS__STATUS_TPM_NOOPERATOR: Final[ThirtyTwoBits] = cached_hresult(0xD0290049, 'HRESULT_FROM_NT(STATUS_TPM_NOOPERATOR)')
    NTSTATUS__STATUS_TPM_RESOURCEMISSING: Final[ThirtyTwoBits] = cached_hresult(0xD029004A, 'HRESULT_FROM_NT(STATUS_TPM_RESOURCEMISSING)')
    NTSTATUS__STATUS_TPM_DELEGATE_LOCK: Final[ThirtyTwoBits] = cached_hresult(0xD029004B, 'HRESULT_FROM_NT(STATUS_TPM_DELEGATE_LOCK)')
    NTSTATUS__STATUS_TPM_DELEGATE_FAMILY: Final[ThirtyTwoBits] = cached_hresult(0xD029004C, 'HRESULT_FROM_NT(STATUS_TPM_DELEGATE_FAMILY)')
    NTSTATUS__STATUS_TPM_DELEGATE_ADMIN: Final[ThirtyTwoBits] = cached_hresult(0xD029004D, 'HRESULT_FROM_NT(STATUS_TPM_DELEGATE_ADMIN)')
    NTSTATUS__STATUS_TPM_TRANSPORT_NOTEXCLUSIVE: Final[ThirtyTwoBits] = cached_hresult(0xD029004E, 'HRESULT_FROM_NT(STATUS_TPM_TRANSPORT_NOTEXCLUSIVE)')
    NTSTATUS__STATUS_TPM_OWNER_CONTROL: Final[ThirtyTwoBits] = cached_hresult(0xD029004F, 'HRESULT_FROM_NT(STATUS_TPM_OWNER_CONTROL)')
    NTSTATUS__STATUS_TPM_DAA_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0xD0290050, 'HRESULT_FROM_NT(STATUS_TPM_DAA_RESOURCES)')
    NTSTATUS__STATUS_TPM_DAA_INPUT_DATA0: Final[ThirtyTwoBits] = cached_hresult(0xD0290051, 'HRESULT_FROM_NT(STATUS_TPM_DAA_INPUT_DATA0)')
    NTSTATUS__STATUS_TPM_DAA_INPUT_DATA1: Final[ThirtyTwoBits] = cached_hresult(0xD0290052, 'HRESULT_FROM_NT(STATUS_TPM_DAA_INPUT_DATA1)')
    NTSTATUS__STATUS_TPM_DAA_ISSUER_SETTINGS: Final[ThirtyTwoBits] = cached_hresult(0xD0290053, 'HRESULT_FROM_NT(STATUS_TPM_DAA_ISSUER_SETTINGS)')
    NTSTATUS__STATUS_TPM_DAA_TPM_SETTINGS: Final[ThirtyTwoBits] = cached_hresult(0xD0290054, 'HRESULT_FROM_NT(STATUS_TPM_DAA_TPM_SETTINGS)')
    NTSTATUS__STATUS_TPM_DAA_STAGE: Final[ThirtyTwoBits] = cached_hresult(0xD0290055, 'HRESULT_FROM_NT(STATUS_TPM_DAA_STAGE)')
    NTSTATUS__STATUS_TPM_DAA_ISSUER_VALIDITY: Final[ThirtyTwoBits] = cached_hresult(0xD0290056, 'HRESULT_FROM_NT(STATUS_TPM_DAA_ISSUER_VALIDITY)')
    NTSTATUS__STATUS_TPM_DAA_WRONG_W: Final[ThirtyTwoBits] = cached_hresult(0xD0290057, 'HRESULT_FROM_NT(STATUS_TPM_DAA_WRONG_W)')
    NTSTATUS__STATUS_TPM_BAD_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0xD0290058, 'HRESULT_FROM_NT(STATUS_TPM_BAD_HANDLE)')
    NTSTATUS__STATUS_TPM_BAD_DELEGATE: Final[ThirtyTwoBits] = cached_hresult(0xD0290059, 'HRESULT_FROM_NT(STATUS_TPM_BAD_DELEGATE)')
    NTSTATUS__STATUS_TPM_BADCONTEXT: Final[ThirtyTwoBits] = cached_hresult(0xD029005A, 'HRESULT_FROM_NT(STATUS_TPM_BADCONTEXT)')
    NTSTATUS__STATUS_TPM_TOOMANYCONTEXTS: Final[ThirtyTwoBits] = cached_hresult(0xD029005B, 'HRESULT_FROM_NT(STATUS_TPM_TOOMANYCONTEXTS)')
    NTSTATUS__STATUS_TPM_MA_TICKET_SIGNATURE: Final[ThirtyTwoBits] = cached_hresult(0xD029005C, 'HRESULT_FROM_NT(STATUS_TPM_MA_TICKET_SIGNATURE)')
    NTSTATUS__STATUS_TPM_MA_DESTINATION: Final[ThirtyTwoBits] = cached_hresult(0xD029005D, 'HRESULT_FROM_NT(STATUS_TPM_MA_DESTINATION)')
    NTSTATUS__STATUS_TPM_MA_SOURCE: Final[ThirtyTwoBits] = cached_hresult(0xD029005E, 'HRESULT_FROM_NT(STATUS_TPM_MA_SOURCE)')
    NTSTATUS__STATUS_TPM_MA_AUTHORITY: Final[ThirtyTwoBits] = cached_hresult(0xD029005F, 'HRESULT_FROM_NT(STATUS_TPM_MA_AUTHORITY)')
    NTSTATUS__STATUS_TPM_PERMANENTEK: Final[ThirtyTwoBits] = cached_hresult(0xD0290061, 'HRESULT_FROM_NT(STATUS_TPM_PERMANENTEK)')
    NTSTATUS__STATUS_TPM_BAD_SIGNATURE: Final[ThirtyTwoBits] = cached_hresult(0xD0290062, 'HRESULT_FROM_NT(STATUS_TPM_BAD_SIGNATURE)')
    NTSTATUS__STATUS_TPM_NOCONTEXTSPACE: Final[ThirtyTwoBits] = cached_hresult(0xD0290063, 'HRESULT_FROM_NT(STATUS_TPM_NOCONTEXTSPACE)')
    NTSTATUS__STATUS_TPM_20_E_ASYMMETRIC: Final[ThirtyTwoBits] = cached_hresult(0xD0290081, 'HRESULT_FROM_NT(STATUS_TPM_20_E_ASYMMETRIC)')
    NTSTATUS__STATUS_TPM_20_E_ATTRIBUTES: Final[ThirtyTwoBits] = cached_hresult(0xD0290082, 'HRESULT_FROM_NT(STATUS_TPM_20_E_ATTRIBUTES)')
    NTSTATUS__STATUS_TPM_20_E_HASH: Final[ThirtyTwoBits] = cached_hresult(0xD0290083, 'HRESULT_FROM_NT(STATUS_TPM_20_E_HASH)')
    NTSTATUS__STATUS_TPM_20_E_VALUE: Final[ThirtyTwoBits] = cached_hresult(0xD0290084, 'HRESULT_FROM_NT(STATUS_TPM_20_E_VALUE)')
    NTSTATUS__STATUS_TPM_20_E_HIERARCHY: Final[ThirtyTwoBits] = cached_hresult(0xD0290085, 'HRESULT_FROM_NT(STATUS_TPM_20_E_HIERARCHY)')
    NTSTATUS__STATUS_TPM_20_E_KEY_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD0290087, 'HRESULT_FROM_NT(STATUS_TPM_20_E_KEY_SIZE)')
    NTSTATUS__STATUS_TPM_20_E_MGF: Final[ThirtyTwoBits] = cached_hresult(0xD0290088, 'HRESULT_FROM_NT(STATUS_TPM_20_E_MGF)')
    NTSTATUS__STATUS_TPM_20_E_MODE: Final[ThirtyTwoBits] = cached_hresult(0xD0290089, 'HRESULT_FROM_NT(STATUS_TPM_20_E_MODE)')
    NTSTATUS__STATUS_TPM_20_E_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xD029008A, 'HRESULT_FROM_NT(STATUS_TPM_20_E_TYPE)')
    NTSTATUS__STATUS_TPM_20_E_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0xD029008B, 'HRESULT_FROM_NT(STATUS_TPM_20_E_HANDLE)')
    NTSTATUS__STATUS_TPM_20_E_KDF: Final[ThirtyTwoBits] = cached_hresult(0xD029008C, 'HRESULT_FROM_NT(STATUS_TPM_20_E_KDF)')
    NTSTATUS__STATUS_TPM_20_E_RANGE: Final[ThirtyTwoBits] = cached_hresult(0xD029008D, 'HRESULT_FROM_NT(STATUS_TPM_20_E_RANGE)')
    NTSTATUS__STATUS_TPM_20_E_AUTH_FAIL: Final[ThirtyTwoBits] = cached_hresult(0xD029008E, 'HRESULT_FROM_NT(STATUS_TPM_20_E_AUTH_FAIL)')
    NTSTATUS__STATUS_TPM_20_E_NONCE: Final[ThirtyTwoBits] = cached_hresult(0xD029008F, 'HRESULT_FROM_NT(STATUS_TPM_20_E_NONCE)')
    NTSTATUS__STATUS_TPM_20_E_PP: Final[ThirtyTwoBits] = cached_hresult(0xD0290090, 'HRESULT_FROM_NT(STATUS_TPM_20_E_PP)')
    NTSTATUS__STATUS_TPM_20_E_SCHEME: Final[ThirtyTwoBits] = cached_hresult(0xD0290092, 'HRESULT_FROM_NT(STATUS_TPM_20_E_SCHEME)')
    NTSTATUS__STATUS_TPM_20_E_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD0290095, 'HRESULT_FROM_NT(STATUS_TPM_20_E_SIZE)')
    NTSTATUS__STATUS_TPM_20_E_SYMMETRIC: Final[ThirtyTwoBits] = cached_hresult(0xD0290096, 'HRESULT_FROM_NT(STATUS_TPM_20_E_SYMMETRIC)')
    NTSTATUS__STATUS_TPM_20_E_TAG: Final[ThirtyTwoBits] = cached_hresult(0xD0290097, 'HRESULT_FROM_NT(STATUS_TPM_20_E_TAG)')
    NTSTATUS__STATUS_TPM_20_E_SELECTOR: Final[ThirtyTwoBits] = cached_hresult(0xD0290098, 'HRESULT_FROM_NT(STATUS_TPM_20_E_SELECTOR)')
    NTSTATUS__STATUS_TPM_20_E_INSUFFICIENT: Final[ThirtyTwoBits] = cached_hresult(0xD029009A, 'HRESULT_FROM_NT(STATUS_TPM_20_E_INSUFFICIENT)')
    NTSTATUS__STATUS_TPM_20_E_SIGNATURE: Final[ThirtyTwoBits] = cached_hresult(0xD029009B, 'HRESULT_FROM_NT(STATUS_TPM_20_E_SIGNATURE)')
    NTSTATUS__STATUS_TPM_20_E_KEY: Final[ThirtyTwoBits] = cached_hresult(0xD029009C, 'HRESULT_FROM_NT(STATUS_TPM_20_E_KEY)')
    NTSTATUS__STATUS_TPM_20_E_POLICY_FAIL: Final[ThirtyTwoBits] = cached_hresult(0xD029009D, 'HRESULT_FROM_NT(STATUS_TPM_20_E_POLICY_FAIL)')
    NTSTATUS__STATUS_TPM_20_E_INTEGRITY: Final[ThirtyTwoBits] = cached_hresult(0xD029009F, 'HRESULT_FROM_NT(STATUS_TPM_20_E_INTEGRITY)')
    NTSTATUS__STATUS_TPM_20_E_TICKET: Final[ThirtyTwoBits] = cached_hresult(0xD02900A0, 'HRESULT_FROM_NT(STATUS_TPM_20_E_TICKET)')
    NTSTATUS__STATUS_TPM_20_E_RESERVED_BITS: Final[ThirtyTwoBits] = cached_hresult(0xD02900A1, 'HRESULT_FROM_NT(STATUS_TPM_20_E_RESERVED_BITS)')
    NTSTATUS__STATUS_TPM_20_E_BAD_AUTH: Final[ThirtyTwoBits] = cached_hresult(0xD02900A2, 'HRESULT_FROM_NT(STATUS_TPM_20_E_BAD_AUTH)')
    NTSTATUS__STATUS_TPM_20_E_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0xD02900A3, 'HRESULT_FROM_NT(STATUS_TPM_20_E_EXPIRED)')
    NTSTATUS__STATUS_TPM_20_E_POLICY_CC: Final[ThirtyTwoBits] = cached_hresult(0xD02900A4, 'HRESULT_FROM_NT(STATUS_TPM_20_E_POLICY_CC)')
    NTSTATUS__STATUS_TPM_20_E_BINDING: Final[ThirtyTwoBits] = cached_hresult(0xD02900A5, 'HRESULT_FROM_NT(STATUS_TPM_20_E_BINDING)')
    NTSTATUS__STATUS_TPM_20_E_CURVE: Final[ThirtyTwoBits] = cached_hresult(0xD02900A6, 'HRESULT_FROM_NT(STATUS_TPM_20_E_CURVE)')
    NTSTATUS__STATUS_TPM_20_E_ECC_POINT: Final[ThirtyTwoBits] = cached_hresult(0xD02900A7, 'HRESULT_FROM_NT(STATUS_TPM_20_E_ECC_POINT)')
    NTSTATUS__STATUS_TPM_20_E_INITIALIZE: Final[ThirtyTwoBits] = cached_hresult(0xD0290100, 'HRESULT_FROM_NT(STATUS_TPM_20_E_INITIALIZE)')
    NTSTATUS__STATUS_TPM_20_E_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD0290101, 'HRESULT_FROM_NT(STATUS_TPM_20_E_FAILURE)')
    NTSTATUS__STATUS_TPM_20_E_SEQUENCE: Final[ThirtyTwoBits] = cached_hresult(0xD0290103, 'HRESULT_FROM_NT(STATUS_TPM_20_E_SEQUENCE)')
    NTSTATUS__STATUS_TPM_20_E_PRIVATE: Final[ThirtyTwoBits] = cached_hresult(0xD029010B, 'HRESULT_FROM_NT(STATUS_TPM_20_E_PRIVATE)')
    NTSTATUS__STATUS_TPM_20_E_HMAC: Final[ThirtyTwoBits] = cached_hresult(0xD0290119, 'HRESULT_FROM_NT(STATUS_TPM_20_E_HMAC)')
    NTSTATUS__STATUS_TPM_20_E_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD0290120, 'HRESULT_FROM_NT(STATUS_TPM_20_E_DISABLED)')
    NTSTATUS__STATUS_TPM_20_E_EXCLUSIVE: Final[ThirtyTwoBits] = cached_hresult(0xD0290121, 'HRESULT_FROM_NT(STATUS_TPM_20_E_EXCLUSIVE)')
    NTSTATUS__STATUS_TPM_20_E_ECC_CURVE: Final[ThirtyTwoBits] = cached_hresult(0xD0290123, 'HRESULT_FROM_NT(STATUS_TPM_20_E_ECC_CURVE)')
    NTSTATUS__STATUS_TPM_20_E_AUTH_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xD0290124, 'HRESULT_FROM_NT(STATUS_TPM_20_E_AUTH_TYPE)')
    NTSTATUS__STATUS_TPM_20_E_AUTH_MISSING: Final[ThirtyTwoBits] = cached_hresult(0xD0290125, 'HRESULT_FROM_NT(STATUS_TPM_20_E_AUTH_MISSING)')
    NTSTATUS__STATUS_TPM_20_E_POLICY: Final[ThirtyTwoBits] = cached_hresult(0xD0290126, 'HRESULT_FROM_NT(STATUS_TPM_20_E_POLICY)')
    NTSTATUS__STATUS_TPM_20_E_PCR: Final[ThirtyTwoBits] = cached_hresult(0xD0290127, 'HRESULT_FROM_NT(STATUS_TPM_20_E_PCR)')
    NTSTATUS__STATUS_TPM_20_E_PCR_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0xD0290128, 'HRESULT_FROM_NT(STATUS_TPM_20_E_PCR_CHANGED)')
    NTSTATUS__STATUS_TPM_20_E_UPGRADE: Final[ThirtyTwoBits] = cached_hresult(0xD029012D, 'HRESULT_FROM_NT(STATUS_TPM_20_E_UPGRADE)')
    NTSTATUS__STATUS_TPM_20_E_TOO_MANY_CONTEXTS: Final[ThirtyTwoBits] = cached_hresult(0xD029012E, 'HRESULT_FROM_NT(STATUS_TPM_20_E_TOO_MANY_CONTEXTS)')
    NTSTATUS__STATUS_TPM_20_E_AUTH_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD029012F, 'HRESULT_FROM_NT(STATUS_TPM_20_E_AUTH_UNAVAILABLE)')
    NTSTATUS__STATUS_TPM_20_E_REBOOT: Final[ThirtyTwoBits] = cached_hresult(0xD0290130, 'HRESULT_FROM_NT(STATUS_TPM_20_E_REBOOT)')
    NTSTATUS__STATUS_TPM_20_E_UNBALANCED: Final[ThirtyTwoBits] = cached_hresult(0xD0290131, 'HRESULT_FROM_NT(STATUS_TPM_20_E_UNBALANCED)')
    NTSTATUS__STATUS_TPM_20_E_COMMAND_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD0290142, 'HRESULT_FROM_NT(STATUS_TPM_20_E_COMMAND_SIZE)')
    NTSTATUS__STATUS_TPM_20_E_COMMAND_CODE: Final[ThirtyTwoBits] = cached_hresult(0xD0290143, 'HRESULT_FROM_NT(STATUS_TPM_20_E_COMMAND_CODE)')
    NTSTATUS__STATUS_TPM_20_E_AUTHSIZE: Final[ThirtyTwoBits] = cached_hresult(0xD0290144, 'HRESULT_FROM_NT(STATUS_TPM_20_E_AUTHSIZE)')
    NTSTATUS__STATUS_TPM_20_E_AUTH_CONTEXT: Final[ThirtyTwoBits] = cached_hresult(0xD0290145, 'HRESULT_FROM_NT(STATUS_TPM_20_E_AUTH_CONTEXT)')
    NTSTATUS__STATUS_TPM_20_E_NV_RANGE: Final[ThirtyTwoBits] = cached_hresult(0xD0290146, 'HRESULT_FROM_NT(STATUS_TPM_20_E_NV_RANGE)')
    NTSTATUS__STATUS_TPM_20_E_NV_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD0290147, 'HRESULT_FROM_NT(STATUS_TPM_20_E_NV_SIZE)')
    NTSTATUS__STATUS_TPM_20_E_NV_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0xD0290148, 'HRESULT_FROM_NT(STATUS_TPM_20_E_NV_LOCKED)')
    NTSTATUS__STATUS_TPM_20_E_NV_AUTHORIZATION: Final[ThirtyTwoBits] = cached_hresult(0xD0290149, 'HRESULT_FROM_NT(STATUS_TPM_20_E_NV_AUTHORIZATION)')
    NTSTATUS__STATUS_TPM_20_E_NV_UNINITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0xD029014A, 'HRESULT_FROM_NT(STATUS_TPM_20_E_NV_UNINITIALIZED)')
    NTSTATUS__STATUS_TPM_20_E_NV_SPACE: Final[ThirtyTwoBits] = cached_hresult(0xD029014B, 'HRESULT_FROM_NT(STATUS_TPM_20_E_NV_SPACE)')
    NTSTATUS__STATUS_TPM_20_E_NV_DEFINED: Final[ThirtyTwoBits] = cached_hresult(0xD029014C, 'HRESULT_FROM_NT(STATUS_TPM_20_E_NV_DEFINED)')
    NTSTATUS__STATUS_TPM_20_E_BAD_CONTEXT: Final[ThirtyTwoBits] = cached_hresult(0xD0290150, 'HRESULT_FROM_NT(STATUS_TPM_20_E_BAD_CONTEXT)')
    NTSTATUS__STATUS_TPM_20_E_CPHASH: Final[ThirtyTwoBits] = cached_hresult(0xD0290151, 'HRESULT_FROM_NT(STATUS_TPM_20_E_CPHASH)')
    NTSTATUS__STATUS_TPM_20_E_PARENT: Final[ThirtyTwoBits] = cached_hresult(0xD0290152, 'HRESULT_FROM_NT(STATUS_TPM_20_E_PARENT)')
    NTSTATUS__STATUS_TPM_20_E_NEEDS_TEST: Final[ThirtyTwoBits] = cached_hresult(0xD0290153, 'HRESULT_FROM_NT(STATUS_TPM_20_E_NEEDS_TEST)')
    NTSTATUS__STATUS_TPM_20_E_NO_RESULT: Final[ThirtyTwoBits] = cached_hresult(0xD0290154, 'HRESULT_FROM_NT(STATUS_TPM_20_E_NO_RESULT)')
    NTSTATUS__STATUS_TPM_20_E_SENSITIVE: Final[ThirtyTwoBits] = cached_hresult(0xD0290155, 'HRESULT_FROM_NT(STATUS_TPM_20_E_SENSITIVE)')
    NTSTATUS__STATUS_TPM_COMMAND_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0xD0290400, 'HRESULT_FROM_NT(STATUS_TPM_COMMAND_BLOCKED)')
    NTSTATUS__STATUS_TPM_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0xD0290401, 'HRESULT_FROM_NT(STATUS_TPM_INVALID_HANDLE)')
    NTSTATUS__STATUS_TPM_DUPLICATE_VHANDLE: Final[ThirtyTwoBits] = cached_hresult(0xD0290402, 'HRESULT_FROM_NT(STATUS_TPM_DUPLICATE_VHANDLE)')
    NTSTATUS__STATUS_TPM_EMBEDDED_COMMAND_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0xD0290403, 'HRESULT_FROM_NT(STATUS_TPM_EMBEDDED_COMMAND_BLOCKED)')
    NTSTATUS__STATUS_TPM_EMBEDDED_COMMAND_UNSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD0290404, 'HRESULT_FROM_NT(STATUS_TPM_EMBEDDED_COMMAND_UNSUPPORTED)')
    NTSTATUS__STATUS_TPM_RETRY: Final[ThirtyTwoBits] = cached_hresult(0xD0290800, 'HRESULT_FROM_NT(STATUS_TPM_RETRY)')
    NTSTATUS__STATUS_TPM_NEEDS_SELFTEST: Final[ThirtyTwoBits] = cached_hresult(0xD0290801, 'HRESULT_FROM_NT(STATUS_TPM_NEEDS_SELFTEST)')
    NTSTATUS__STATUS_TPM_DOING_SELFTEST: Final[ThirtyTwoBits] = cached_hresult(0xD0290802, 'HRESULT_FROM_NT(STATUS_TPM_DOING_SELFTEST)')
    NTSTATUS__STATUS_TPM_DEFEND_LOCK_RUNNING: Final[ThirtyTwoBits] = cached_hresult(0xD0290803, 'HRESULT_FROM_NT(STATUS_TPM_DEFEND_LOCK_RUNNING)')
    NTSTATUS__STATUS_TPM_COMMAND_CANCELED: Final[ThirtyTwoBits] = cached_hresult(0xD0291001, 'HRESULT_FROM_NT(STATUS_TPM_COMMAND_CANCELED)')
    NTSTATUS__STATUS_TPM_TOO_MANY_CONTEXTS: Final[ThirtyTwoBits] = cached_hresult(0xD0291002, 'HRESULT_FROM_NT(STATUS_TPM_TOO_MANY_CONTEXTS)')
    NTSTATUS__STATUS_TPM_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0291003, 'HRESULT_FROM_NT(STATUS_TPM_NOT_FOUND)')
    NTSTATUS__STATUS_TPM_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0xD0291004, 'HRESULT_FROM_NT(STATUS_TPM_ACCESS_DENIED)')
    NTSTATUS__STATUS_TPM_INSUFFICIENT_BUFFER: Final[ThirtyTwoBits] = cached_hresult(0xD0291005, 'HRESULT_FROM_NT(STATUS_TPM_INSUFFICIENT_BUFFER)')
    NTSTATUS__STATUS_TPM_PPI_FUNCTION_UNSUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD0291006, 'HRESULT_FROM_NT(STATUS_TPM_PPI_FUNCTION_UNSUPPORTED)')
    NTSTATUS__STATUS_PCP_ERROR_MASK: Final[ThirtyTwoBits] = cached_hresult(0xD0292000, 'HRESULT_FROM_NT(STATUS_PCP_ERROR_MASK)')
    NTSTATUS__STATUS_PCP_DEVICE_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0xD0292001, 'HRESULT_FROM_NT(STATUS_PCP_DEVICE_NOT_READY)')
    NTSTATUS__STATUS_PCP_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0xD0292002, 'HRESULT_FROM_NT(STATUS_PCP_INVALID_HANDLE)')
    NTSTATUS__STATUS_PCP_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0xD0292003, 'HRESULT_FROM_NT(STATUS_PCP_INVALID_PARAMETER)')
    NTSTATUS__STATUS_PCP_FLAG_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD0292004, 'HRESULT_FROM_NT(STATUS_PCP_FLAG_NOT_SUPPORTED)')
    NTSTATUS__STATUS_PCP_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD0292005, 'HRESULT_FROM_NT(STATUS_PCP_NOT_SUPPORTED)')
    NTSTATUS__STATUS_PCP_BUFFER_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0xD0292006, 'HRESULT_FROM_NT(STATUS_PCP_BUFFER_TOO_SMALL)')
    NTSTATUS__STATUS_PCP_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0292007, 'HRESULT_FROM_NT(STATUS_PCP_INTERNAL_ERROR)')
    NTSTATUS__STATUS_PCP_AUTHENTICATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0292008, 'HRESULT_FROM_NT(STATUS_PCP_AUTHENTICATION_FAILED)')
    NTSTATUS__STATUS_PCP_AUTHENTICATION_IGNORED: Final[ThirtyTwoBits] = cached_hresult(0xD0292009, 'HRESULT_FROM_NT(STATUS_PCP_AUTHENTICATION_IGNORED)')
    NTSTATUS__STATUS_PCP_POLICY_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD029200A, 'HRESULT_FROM_NT(STATUS_PCP_POLICY_NOT_FOUND)')
    NTSTATUS__STATUS_PCP_PROFILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD029200B, 'HRESULT_FROM_NT(STATUS_PCP_PROFILE_NOT_FOUND)')
    NTSTATUS__STATUS_PCP_VALIDATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD029200C, 'HRESULT_FROM_NT(STATUS_PCP_VALIDATION_FAILED)')
    NTSTATUS__STATUS_PCP_DEVICE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD029200D, 'HRESULT_FROM_NT(STATUS_PCP_DEVICE_NOT_FOUND)')
    NTSTATUS__STATUS_PCP_WRONG_PARENT: Final[ThirtyTwoBits] = cached_hresult(0xD029200E, 'HRESULT_FROM_NT(STATUS_PCP_WRONG_PARENT)')
    NTSTATUS__STATUS_PCP_KEY_NOT_LOADED: Final[ThirtyTwoBits] = cached_hresult(0xD029200F, 'HRESULT_FROM_NT(STATUS_PCP_KEY_NOT_LOADED)')
    NTSTATUS__STATUS_PCP_NO_KEY_CERTIFICATION: Final[ThirtyTwoBits] = cached_hresult(0xD0292010, 'HRESULT_FROM_NT(STATUS_PCP_NO_KEY_CERTIFICATION)')
    NTSTATUS__STATUS_PCP_KEY_NOT_FINALIZED: Final[ThirtyTwoBits] = cached_hresult(0xD0292011, 'HRESULT_FROM_NT(STATUS_PCP_KEY_NOT_FINALIZED)')
    NTSTATUS__STATUS_PCP_ATTESTATION_CHALLENGE_NOT_SET: Final[ThirtyTwoBits] = cached_hresult(0xD0292012, 'HRESULT_FROM_NT(STATUS_PCP_ATTESTATION_CHALLENGE_NOT_SET)')
    NTSTATUS__STATUS_PCP_NOT_PCR_BOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0292013, 'HRESULT_FROM_NT(STATUS_PCP_NOT_PCR_BOUND)')
    NTSTATUS__STATUS_PCP_KEY_ALREADY_FINALIZED: Final[ThirtyTwoBits] = cached_hresult(0xD0292014, 'HRESULT_FROM_NT(STATUS_PCP_KEY_ALREADY_FINALIZED)')
    NTSTATUS__STATUS_PCP_KEY_USAGE_POLICY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD0292015, 'HRESULT_FROM_NT(STATUS_PCP_KEY_USAGE_POLICY_NOT_SUPPORTED)')
    NTSTATUS__STATUS_PCP_KEY_USAGE_POLICY_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0292016, 'HRESULT_FROM_NT(STATUS_PCP_KEY_USAGE_POLICY_INVALID)')
    NTSTATUS__STATUS_PCP_SOFT_KEY_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0292017, 'HRESULT_FROM_NT(STATUS_PCP_SOFT_KEY_ERROR)')
    NTSTATUS__STATUS_PCP_KEY_NOT_AUTHENTICATED: Final[ThirtyTwoBits] = cached_hresult(0xD0292018, 'HRESULT_FROM_NT(STATUS_PCP_KEY_NOT_AUTHENTICATED)')
    NTSTATUS__STATUS_PCP_KEY_NOT_AIK: Final[ThirtyTwoBits] = cached_hresult(0xD0292019, 'HRESULT_FROM_NT(STATUS_PCP_KEY_NOT_AIK)')
    NTSTATUS__STATUS_PCP_KEY_NOT_SIGNING_KEY: Final[ThirtyTwoBits] = cached_hresult(0xD029201A, 'HRESULT_FROM_NT(STATUS_PCP_KEY_NOT_SIGNING_KEY)')
    NTSTATUS__STATUS_PCP_LOCKED_OUT: Final[ThirtyTwoBits] = cached_hresult(0xD029201B, 'HRESULT_FROM_NT(STATUS_PCP_LOCKED_OUT)')
    NTSTATUS__STATUS_PCP_CLAIM_TYPE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD029201C, 'HRESULT_FROM_NT(STATUS_PCP_CLAIM_TYPE_NOT_SUPPORTED)')
    NTSTATUS__STATUS_PCP_TPM_VERSION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD029201D, 'HRESULT_FROM_NT(STATUS_PCP_TPM_VERSION_NOT_SUPPORTED)')
    NTSTATUS__STATUS_PCP_BUFFER_LENGTH_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD029201E, 'HRESULT_FROM_NT(STATUS_PCP_BUFFER_LENGTH_MISMATCH)')
    NTSTATUS__STATUS_PCP_IFX_RSA_KEY_CREATION_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0xD029201F, 'HRESULT_FROM_NT(STATUS_PCP_IFX_RSA_KEY_CREATION_BLOCKED)')
    NTSTATUS__STATUS_PCP_TICKET_MISSING: Final[ThirtyTwoBits] = cached_hresult(0xD0292020, 'HRESULT_FROM_NT(STATUS_PCP_TICKET_MISSING)')
    NTSTATUS__STATUS_PCP_RAW_POLICY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD0292021, 'HRESULT_FROM_NT(STATUS_PCP_RAW_POLICY_NOT_SUPPORTED)')
    NTSTATUS__STATUS_PCP_KEY_HANDLE_INVALIDATED: Final[ThirtyTwoBits] = cached_hresult(0xD0292022, 'HRESULT_FROM_NT(STATUS_PCP_KEY_HANDLE_INVALIDATED)')
    NTSTATUS__STATUS_RTPM_NO_RESULT: Final[ThirtyTwoBits] = cached_hresult(0xD0293002, 'HRESULT_FROM_NT(STATUS_RTPM_NO_RESULT)')
    NTSTATUS__STATUS_RTPM_PCR_READ_INCOMPLETE: Final[ThirtyTwoBits] = cached_hresult(0xD0293003, 'HRESULT_FROM_NT(STATUS_RTPM_PCR_READ_INCOMPLETE)')
    NTSTATUS__STATUS_RTPM_INVALID_CONTEXT: Final[ThirtyTwoBits] = cached_hresult(0xD0293004, 'HRESULT_FROM_NT(STATUS_RTPM_INVALID_CONTEXT)')
    NTSTATUS__STATUS_RTPM_UNSUPPORTED_CMD: Final[ThirtyTwoBits] = cached_hresult(0xD0293005, 'HRESULT_FROM_NT(STATUS_RTPM_UNSUPPORTED_CMD)')
    NTSTATUS__STATUS_TPM_ZERO_EXHAUST_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0xD0294000, 'HRESULT_FROM_NT(STATUS_TPM_ZERO_EXHAUST_ENABLED)')
    NTSTATUS__STATUS_HV_INVALID_HYPERCALL_CODE: Final[ThirtyTwoBits] = cached_hresult(0xD0350002, 'HRESULT_FROM_NT(STATUS_HV_INVALID_HYPERCALL_CODE)')
    NTSTATUS__STATUS_HV_INVALID_HYPERCALL_INPUT: Final[ThirtyTwoBits] = cached_hresult(0xD0350003, 'HRESULT_FROM_NT(STATUS_HV_INVALID_HYPERCALL_INPUT)')
    NTSTATUS__STATUS_HV_INVALID_ALIGNMENT: Final[ThirtyTwoBits] = cached_hresult(0xD0350004, 'HRESULT_FROM_NT(STATUS_HV_INVALID_ALIGNMENT)')
    NTSTATUS__STATUS_HV_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_hresult(0xD0350005, 'HRESULT_FROM_NT(STATUS_HV_INVALID_PARAMETER)')
    NTSTATUS__STATUS_HV_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0xD0350006, 'HRESULT_FROM_NT(STATUS_HV_ACCESS_DENIED)')
    NTSTATUS__STATUS_HV_INVALID_PARTITION_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD0350007, 'HRESULT_FROM_NT(STATUS_HV_INVALID_PARTITION_STATE)')
    NTSTATUS__STATUS_HV_OPERATION_DENIED: Final[ThirtyTwoBits] = cached_hresult(0xD0350008, 'HRESULT_FROM_NT(STATUS_HV_OPERATION_DENIED)')
    NTSTATUS__STATUS_HV_UNKNOWN_PROPERTY: Final[ThirtyTwoBits] = cached_hresult(0xD0350009, 'HRESULT_FROM_NT(STATUS_HV_UNKNOWN_PROPERTY)')
    NTSTATUS__STATUS_HV_PROPERTY_VALUE_OUT_OF_RANGE: Final[ThirtyTwoBits] = cached_hresult(0xD035000A, 'HRESULT_FROM_NT(STATUS_HV_PROPERTY_VALUE_OUT_OF_RANGE)')
    NTSTATUS__STATUS_HV_INSUFFICIENT_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0xD035000B, 'HRESULT_FROM_NT(STATUS_HV_INSUFFICIENT_MEMORY)')
    NTSTATUS__STATUS_HV_PARTITION_TOO_DEEP: Final[ThirtyTwoBits] = cached_hresult(0xD035000C, 'HRESULT_FROM_NT(STATUS_HV_PARTITION_TOO_DEEP)')
    NTSTATUS__STATUS_HV_INVALID_PARTITION_ID: Final[ThirtyTwoBits] = cached_hresult(0xD035000D, 'HRESULT_FROM_NT(STATUS_HV_INVALID_PARTITION_ID)')
    NTSTATUS__STATUS_HV_INVALID_VP_INDEX: Final[ThirtyTwoBits] = cached_hresult(0xD035000E, 'HRESULT_FROM_NT(STATUS_HV_INVALID_VP_INDEX)')
    NTSTATUS__STATUS_HV_INVALID_PORT_ID: Final[ThirtyTwoBits] = cached_hresult(0xD0350011, 'HRESULT_FROM_NT(STATUS_HV_INVALID_PORT_ID)')
    NTSTATUS__STATUS_HV_INVALID_CONNECTION_ID: Final[ThirtyTwoBits] = cached_hresult(0xD0350012, 'HRESULT_FROM_NT(STATUS_HV_INVALID_CONNECTION_ID)')
    NTSTATUS__STATUS_HV_INSUFFICIENT_BUFFERS: Final[ThirtyTwoBits] = cached_hresult(0xD0350013, 'HRESULT_FROM_NT(STATUS_HV_INSUFFICIENT_BUFFERS)')
    NTSTATUS__STATUS_HV_NOT_ACKNOWLEDGED: Final[ThirtyTwoBits] = cached_hresult(0xD0350014, 'HRESULT_FROM_NT(STATUS_HV_NOT_ACKNOWLEDGED)')
    NTSTATUS__STATUS_HV_INVALID_VP_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD0350015, 'HRESULT_FROM_NT(STATUS_HV_INVALID_VP_STATE)')
    NTSTATUS__STATUS_HV_ACKNOWLEDGED: Final[ThirtyTwoBits] = cached_hresult(0xD0350016, 'HRESULT_FROM_NT(STATUS_HV_ACKNOWLEDGED)')
    NTSTATUS__STATUS_HV_INVALID_SAVE_RESTORE_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD0350017, 'HRESULT_FROM_NT(STATUS_HV_INVALID_SAVE_RESTORE_STATE)')
    NTSTATUS__STATUS_HV_INVALID_SYNIC_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD0350018, 'HRESULT_FROM_NT(STATUS_HV_INVALID_SYNIC_STATE)')
    NTSTATUS__STATUS_HV_OBJECT_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0xD0350019, 'HRESULT_FROM_NT(STATUS_HV_OBJECT_IN_USE)')
    NTSTATUS__STATUS_HV_INVALID_PROXIMITY_DOMAIN_INFO: Final[ThirtyTwoBits] = cached_hresult(0xD035001A, 'HRESULT_FROM_NT(STATUS_HV_INVALID_PROXIMITY_DOMAIN_INFO)')
    NTSTATUS__STATUS_HV_NO_DATA: Final[ThirtyTwoBits] = cached_hresult(0xD035001B, 'HRESULT_FROM_NT(STATUS_HV_NO_DATA)')
    NTSTATUS__STATUS_HV_INACTIVE: Final[ThirtyTwoBits] = cached_hresult(0xD035001C, 'HRESULT_FROM_NT(STATUS_HV_INACTIVE)')
    NTSTATUS__STATUS_HV_NO_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0xD035001D, 'HRESULT_FROM_NT(STATUS_HV_NO_RESOURCES)')
    NTSTATUS__STATUS_HV_FEATURE_UNAVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD035001E, 'HRESULT_FROM_NT(STATUS_HV_FEATURE_UNAVAILABLE)')
    NTSTATUS__STATUS_HV_INSUFFICIENT_BUFFER: Final[ThirtyTwoBits] = cached_hresult(0xD0350033, 'HRESULT_FROM_NT(STATUS_HV_INSUFFICIENT_BUFFER)')
    NTSTATUS__STATUS_HV_INSUFFICIENT_DEVICE_DOMAINS: Final[ThirtyTwoBits] = cached_hresult(0xD0350038, 'HRESULT_FROM_NT(STATUS_HV_INSUFFICIENT_DEVICE_DOMAINS)')
    NTSTATUS__STATUS_HV_CPUID_FEATURE_VALIDATION_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD035003C, 'HRESULT_FROM_NT(STATUS_HV_CPUID_FEATURE_VALIDATION_ERROR)')
    NTSTATUS__STATUS_HV_CPUID_XSAVE_FEATURE_VALIDATION_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD035003D, 'HRESULT_FROM_NT(STATUS_HV_CPUID_XSAVE_FEATURE_VALIDATION_ERROR)')
    NTSTATUS__STATUS_HV_PROCESSOR_STARTUP_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0xD035003E, 'HRESULT_FROM_NT(STATUS_HV_PROCESSOR_STARTUP_TIMEOUT)')
    NTSTATUS__STATUS_HV_SMX_ENABLED: Final[ThirtyTwoBits] = cached_hresult(0xD035003F, 'HRESULT_FROM_NT(STATUS_HV_SMX_ENABLED)')
    NTSTATUS__STATUS_HV_INVALID_LP_INDEX: Final[ThirtyTwoBits] = cached_hresult(0xD0350041, 'HRESULT_FROM_NT(STATUS_HV_INVALID_LP_INDEX)')
    NTSTATUS__STATUS_HV_INVALID_REGISTER_VALUE: Final[ThirtyTwoBits] = cached_hresult(0xD0350050, 'HRESULT_FROM_NT(STATUS_HV_INVALID_REGISTER_VALUE)')
    NTSTATUS__STATUS_HV_INVALID_VTL_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD0350051, 'HRESULT_FROM_NT(STATUS_HV_INVALID_VTL_STATE)')
    NTSTATUS__STATUS_HV_NX_NOT_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0xD0350055, 'HRESULT_FROM_NT(STATUS_HV_NX_NOT_DETECTED)')
    NTSTATUS__STATUS_HV_INVALID_DEVICE_ID: Final[ThirtyTwoBits] = cached_hresult(0xD0350057, 'HRESULT_FROM_NT(STATUS_HV_INVALID_DEVICE_ID)')
    NTSTATUS__STATUS_HV_INVALID_DEVICE_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD0350058, 'HRESULT_FROM_NT(STATUS_HV_INVALID_DEVICE_STATE)')
    NTSTATUS__STATUS_HV_PAGE_REQUEST_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0350060, 'HRESULT_FROM_NT(STATUS_HV_PAGE_REQUEST_INVALID)')
    NTSTATUS__STATUS_HV_INVALID_CPU_GROUP_ID: Final[ThirtyTwoBits] = cached_hresult(0xD035006F, 'HRESULT_FROM_NT(STATUS_HV_INVALID_CPU_GROUP_ID)')
    NTSTATUS__STATUS_HV_INVALID_CPU_GROUP_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD0350070, 'HRESULT_FROM_NT(STATUS_HV_INVALID_CPU_GROUP_STATE)')
    NTSTATUS__STATUS_HV_OPERATION_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0350071, 'HRESULT_FROM_NT(STATUS_HV_OPERATION_FAILED)')
    NTSTATUS__STATUS_HV_NOT_ALLOWED_WITH_NESTED_VIRT_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0xD0350072, 'HRESULT_FROM_NT(STATUS_HV_NOT_ALLOWED_WITH_NESTED_VIRT_ACTIVE)')
    NTSTATUS__STATUS_HV_INSUFFICIENT_ROOT_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0xD0350073, 'HRESULT_FROM_NT(STATUS_HV_INSUFFICIENT_ROOT_MEMORY)')
    NTSTATUS__STATUS_HV_EVENT_BUFFER_ALREADY_FREED: Final[ThirtyTwoBits] = cached_hresult(0xD0350074, 'HRESULT_FROM_NT(STATUS_HV_EVENT_BUFFER_ALREADY_FREED)')
    NTSTATUS__STATUS_HV_INSUFFICIENT_CONTIGUOUS_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0xD0350075, 'HRESULT_FROM_NT(STATUS_HV_INSUFFICIENT_CONTIGUOUS_MEMORY)')
    NTSTATUS__STATUS_HV_DEVICE_NOT_IN_DOMAIN: Final[ThirtyTwoBits] = cached_hresult(0xD0350076, 'HRESULT_FROM_NT(STATUS_HV_DEVICE_NOT_IN_DOMAIN)')
    NTSTATUS__STATUS_HV_NESTED_VM_EXIT: Final[ThirtyTwoBits] = cached_hresult(0xD0350077, 'HRESULT_FROM_NT(STATUS_HV_NESTED_VM_EXIT)')
    NTSTATUS__STATUS_HV_CALL_PENDING: Final[ThirtyTwoBits] = cached_hresult(0xD0350079, 'HRESULT_FROM_NT(STATUS_HV_CALL_PENDING)')
    NTSTATUS__STATUS_HV_MSR_ACCESS_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0350080, 'HRESULT_FROM_NT(STATUS_HV_MSR_ACCESS_FAILED)')
    NTSTATUS__STATUS_HV_INSUFFICIENT_MEMORY_MIRRORING: Final[ThirtyTwoBits] = cached_hresult(0xD0350081, 'HRESULT_FROM_NT(STATUS_HV_INSUFFICIENT_MEMORY_MIRRORING)')
    NTSTATUS__STATUS_HV_INSUFFICIENT_CONTIGUOUS_MEMORY_MIRRORING: Final[ThirtyTwoBits] = cached_hresult(0xD0350082, 'HRESULT_FROM_NT(STATUS_HV_INSUFFICIENT_CONTIGUOUS_MEMORY_MIRRORING)')
    NTSTATUS__STATUS_HV_INSUFFICIENT_CONTIGUOUS_ROOT_MEMORY: Final[ThirtyTwoBits] = cached_hresult(0xD0350083, 'HRESULT_FROM_NT(STATUS_HV_INSUFFICIENT_CONTIGUOUS_ROOT_MEMORY)')
    NTSTATUS__STATUS_HV_INSUFFICIENT_ROOT_MEMORY_MIRRORING: Final[ThirtyTwoBits] = cached_hresult(0xD0350084, 'HRESULT_FROM_NT(STATUS_HV_INSUFFICIENT_ROOT_MEMORY_MIRRORING)')
    NTSTATUS__STATUS_HV_INSUFFICIENT_CONTIGUOUS_ROOT_MEMORY_MIRRORING: Final[ThirtyTwoBits] = cached_hresult(0xD0350085, 'HRESULT_FROM_NT(STATUS_HV_INSUFFICIENT_CONTIGUOUS_ROOT_MEMORY_MIRRORING)')
    NTSTATUS__STATUS_HV_NOT_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0xD0351000, 'HRESULT_FROM_NT(STATUS_HV_NOT_PRESENT)')
    NTSTATUS__STATUS_IPSEC_BAD_SPI: Final[ThirtyTwoBits] = cached_hresult(0xD0360001, 'HRESULT_FROM_NT(STATUS_IPSEC_BAD_SPI)')
    NTSTATUS__STATUS_IPSEC_SA_LIFETIME_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0xD0360002, 'HRESULT_FROM_NT(STATUS_IPSEC_SA_LIFETIME_EXPIRED)')
    NTSTATUS__STATUS_IPSEC_WRONG_SA: Final[ThirtyTwoBits] = cached_hresult(0xD0360003, 'HRESULT_FROM_NT(STATUS_IPSEC_WRONG_SA)')
    NTSTATUS__STATUS_IPSEC_REPLAY_CHECK_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0360004, 'HRESULT_FROM_NT(STATUS_IPSEC_REPLAY_CHECK_FAILED)')
    NTSTATUS__STATUS_IPSEC_INVALID_PACKET: Final[ThirtyTwoBits] = cached_hresult(0xD0360005, 'HRESULT_FROM_NT(STATUS_IPSEC_INVALID_PACKET)')
    NTSTATUS__STATUS_IPSEC_INTEGRITY_CHECK_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0360006, 'HRESULT_FROM_NT(STATUS_IPSEC_INTEGRITY_CHECK_FAILED)')
    NTSTATUS__STATUS_IPSEC_CLEAR_TEXT_DROP: Final[ThirtyTwoBits] = cached_hresult(0xD0360007, 'HRESULT_FROM_NT(STATUS_IPSEC_CLEAR_TEXT_DROP)')
    NTSTATUS__STATUS_IPSEC_AUTH_FIREWALL_DROP: Final[ThirtyTwoBits] = cached_hresult(0xD0360008, 'HRESULT_FROM_NT(STATUS_IPSEC_AUTH_FIREWALL_DROP)')
    NTSTATUS__STATUS_IPSEC_THROTTLE_DROP: Final[ThirtyTwoBits] = cached_hresult(0xD0360009, 'HRESULT_FROM_NT(STATUS_IPSEC_THROTTLE_DROP)')
    NTSTATUS__STATUS_IPSEC_DOSP_BLOCK: Final[ThirtyTwoBits] = cached_hresult(0xD0368000, 'HRESULT_FROM_NT(STATUS_IPSEC_DOSP_BLOCK)')
    NTSTATUS__STATUS_IPSEC_DOSP_RECEIVED_MULTICAST: Final[ThirtyTwoBits] = cached_hresult(0xD0368001, 'HRESULT_FROM_NT(STATUS_IPSEC_DOSP_RECEIVED_MULTICAST)')
    NTSTATUS__STATUS_IPSEC_DOSP_INVALID_PACKET: Final[ThirtyTwoBits] = cached_hresult(0xD0368002, 'HRESULT_FROM_NT(STATUS_IPSEC_DOSP_INVALID_PACKET)')
    NTSTATUS__STATUS_IPSEC_DOSP_STATE_LOOKUP_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0368003, 'HRESULT_FROM_NT(STATUS_IPSEC_DOSP_STATE_LOOKUP_FAILED)')
    NTSTATUS__STATUS_IPSEC_DOSP_MAX_ENTRIES: Final[ThirtyTwoBits] = cached_hresult(0xD0368004, 'HRESULT_FROM_NT(STATUS_IPSEC_DOSP_MAX_ENTRIES)')
    NTSTATUS__STATUS_IPSEC_DOSP_KEYMOD_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_hresult(0xD0368005, 'HRESULT_FROM_NT(STATUS_IPSEC_DOSP_KEYMOD_NOT_ALLOWED)')
    NTSTATUS__STATUS_IPSEC_DOSP_MAX_PER_IP_RATELIMIT_QUEUES: Final[ThirtyTwoBits] = cached_hresult(0xD0368006, 'HRESULT_FROM_NT(STATUS_IPSEC_DOSP_MAX_PER_IP_RATELIMIT_QUEUES)')
    NTSTATUS__STATUS_VID_DUPLICATE_HANDLER: Final[ThirtyTwoBits] = cached_hresult(0xD0370001, 'HRESULT_FROM_NT(STATUS_VID_DUPLICATE_HANDLER)')
    NTSTATUS__STATUS_VID_TOO_MANY_HANDLERS: Final[ThirtyTwoBits] = cached_hresult(0xD0370002, 'HRESULT_FROM_NT(STATUS_VID_TOO_MANY_HANDLERS)')
    NTSTATUS__STATUS_VID_QUEUE_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD0370003, 'HRESULT_FROM_NT(STATUS_VID_QUEUE_FULL)')
    NTSTATUS__STATUS_VID_HANDLER_NOT_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0xD0370004, 'HRESULT_FROM_NT(STATUS_VID_HANDLER_NOT_PRESENT)')
    NTSTATUS__STATUS_VID_INVALID_OBJECT_NAME: Final[ThirtyTwoBits] = cached_hresult(0xD0370005, 'HRESULT_FROM_NT(STATUS_VID_INVALID_OBJECT_NAME)')
    NTSTATUS__STATUS_VID_PARTITION_NAME_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0xD0370006, 'HRESULT_FROM_NT(STATUS_VID_PARTITION_NAME_TOO_LONG)')
    NTSTATUS__STATUS_VID_MESSAGE_QUEUE_NAME_TOO_LONG: Final[ThirtyTwoBits] = cached_hresult(0xD0370007, 'HRESULT_FROM_NT(STATUS_VID_MESSAGE_QUEUE_NAME_TOO_LONG)')
    NTSTATUS__STATUS_VID_PARTITION_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD0370008, 'HRESULT_FROM_NT(STATUS_VID_PARTITION_ALREADY_EXISTS)')
    NTSTATUS__STATUS_VID_PARTITION_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_hresult(0xD0370009, 'HRESULT_FROM_NT(STATUS_VID_PARTITION_DOES_NOT_EXIST)')
    NTSTATUS__STATUS_VID_PARTITION_NAME_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD037000A, 'HRESULT_FROM_NT(STATUS_VID_PARTITION_NAME_NOT_FOUND)')
    NTSTATUS__STATUS_VID_MESSAGE_QUEUE_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_hresult(0xD037000B, 'HRESULT_FROM_NT(STATUS_VID_MESSAGE_QUEUE_ALREADY_EXISTS)')
    NTSTATUS__STATUS_VID_EXCEEDED_MBP_ENTRY_MAP_LIMIT: Final[ThirtyTwoBits] = cached_hresult(0xD037000C, 'HRESULT_FROM_NT(STATUS_VID_EXCEEDED_MBP_ENTRY_MAP_LIMIT)')
    NTSTATUS__STATUS_VID_MB_STILL_REFERENCED: Final[ThirtyTwoBits] = cached_hresult(0xD037000D, 'HRESULT_FROM_NT(STATUS_VID_MB_STILL_REFERENCED)')
    NTSTATUS__STATUS_VID_CHILD_GPA_PAGE_SET_CORRUPTED: Final[ThirtyTwoBits] = cached_hresult(0xD037000E, 'HRESULT_FROM_NT(STATUS_VID_CHILD_GPA_PAGE_SET_CORRUPTED)')
    NTSTATUS__STATUS_VID_INVALID_NUMA_SETTINGS: Final[ThirtyTwoBits] = cached_hresult(0xD037000F, 'HRESULT_FROM_NT(STATUS_VID_INVALID_NUMA_SETTINGS)')
    NTSTATUS__STATUS_VID_INVALID_NUMA_NODE_INDEX: Final[ThirtyTwoBits] = cached_hresult(0xD0370010, 'HRESULT_FROM_NT(STATUS_VID_INVALID_NUMA_NODE_INDEX)')
    NTSTATUS__STATUS_VID_NOTIFICATION_QUEUE_ALREADY_ASSOCIATED: Final[ThirtyTwoBits] = cached_hresult(0xD0370011, 'HRESULT_FROM_NT(STATUS_VID_NOTIFICATION_QUEUE_ALREADY_ASSOCIATED)')
    NTSTATUS__STATUS_VID_INVALID_MEMORY_BLOCK_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0xD0370012, 'HRESULT_FROM_NT(STATUS_VID_INVALID_MEMORY_BLOCK_HANDLE)')
    NTSTATUS__STATUS_VID_PAGE_RANGE_OVERFLOW: Final[ThirtyTwoBits] = cached_hresult(0xD0370013, 'HRESULT_FROM_NT(STATUS_VID_PAGE_RANGE_OVERFLOW)')
    NTSTATUS__STATUS_VID_INVALID_MESSAGE_QUEUE_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0xD0370014, 'HRESULT_FROM_NT(STATUS_VID_INVALID_MESSAGE_QUEUE_HANDLE)')
    NTSTATUS__STATUS_VID_INVALID_GPA_RANGE_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0xD0370015, 'HRESULT_FROM_NT(STATUS_VID_INVALID_GPA_RANGE_HANDLE)')
    NTSTATUS__STATUS_VID_NO_MEMORY_BLOCK_NOTIFICATION_QUEUE: Final[ThirtyTwoBits] = cached_hresult(0xD0370016, 'HRESULT_FROM_NT(STATUS_VID_NO_MEMORY_BLOCK_NOTIFICATION_QUEUE)')
    NTSTATUS__STATUS_VID_MEMORY_BLOCK_LOCK_COUNT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD0370017, 'HRESULT_FROM_NT(STATUS_VID_MEMORY_BLOCK_LOCK_COUNT_EXCEEDED)')
    NTSTATUS__STATUS_VID_INVALID_PPM_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0xD0370018, 'HRESULT_FROM_NT(STATUS_VID_INVALID_PPM_HANDLE)')
    NTSTATUS__STATUS_VID_MBPS_ARE_LOCKED: Final[ThirtyTwoBits] = cached_hresult(0xD0370019, 'HRESULT_FROM_NT(STATUS_VID_MBPS_ARE_LOCKED)')
    NTSTATUS__STATUS_VID_MESSAGE_QUEUE_CLOSED: Final[ThirtyTwoBits] = cached_hresult(0xD037001A, 'HRESULT_FROM_NT(STATUS_VID_MESSAGE_QUEUE_CLOSED)')
    NTSTATUS__STATUS_VID_VIRTUAL_PROCESSOR_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD037001B, 'HRESULT_FROM_NT(STATUS_VID_VIRTUAL_PROCESSOR_LIMIT_EXCEEDED)')
    NTSTATUS__STATUS_VID_STOP_PENDING: Final[ThirtyTwoBits] = cached_hresult(0xD037001C, 'HRESULT_FROM_NT(STATUS_VID_STOP_PENDING)')
    NTSTATUS__STATUS_VID_INVALID_PROCESSOR_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD037001D, 'HRESULT_FROM_NT(STATUS_VID_INVALID_PROCESSOR_STATE)')
    NTSTATUS__STATUS_VID_EXCEEDED_KM_CONTEXT_COUNT_LIMIT: Final[ThirtyTwoBits] = cached_hresult(0xD037001E, 'HRESULT_FROM_NT(STATUS_VID_EXCEEDED_KM_CONTEXT_COUNT_LIMIT)')
    NTSTATUS__STATUS_VID_KM_INTERFACE_ALREADY_INITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0xD037001F, 'HRESULT_FROM_NT(STATUS_VID_KM_INTERFACE_ALREADY_INITIALIZED)')
    NTSTATUS__STATUS_VID_MB_PROPERTY_ALREADY_SET_RESET: Final[ThirtyTwoBits] = cached_hresult(0xD0370020, 'HRESULT_FROM_NT(STATUS_VID_MB_PROPERTY_ALREADY_SET_RESET)')
    NTSTATUS__STATUS_VID_MMIO_RANGE_DESTROYED: Final[ThirtyTwoBits] = cached_hresult(0xD0370021, 'HRESULT_FROM_NT(STATUS_VID_MMIO_RANGE_DESTROYED)')
    NTSTATUS__STATUS_VID_INVALID_CHILD_GPA_PAGE_SET: Final[ThirtyTwoBits] = cached_hresult(0xD0370022, 'HRESULT_FROM_NT(STATUS_VID_INVALID_CHILD_GPA_PAGE_SET)')
    NTSTATUS__STATUS_VID_RESERVE_PAGE_SET_IS_BEING_USED: Final[ThirtyTwoBits] = cached_hresult(0xD0370023, 'HRESULT_FROM_NT(STATUS_VID_RESERVE_PAGE_SET_IS_BEING_USED)')
    NTSTATUS__STATUS_VID_RESERVE_PAGE_SET_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0xD0370024, 'HRESULT_FROM_NT(STATUS_VID_RESERVE_PAGE_SET_TOO_SMALL)')
    NTSTATUS__STATUS_VID_MBP_ALREADY_LOCKED_USING_RESERVED_PAGE: Final[ThirtyTwoBits] = cached_hresult(0xD0370025, 'HRESULT_FROM_NT(STATUS_VID_MBP_ALREADY_LOCKED_USING_RESERVED_PAGE)')
    NTSTATUS__STATUS_VID_MBP_COUNT_EXCEEDED_LIMIT: Final[ThirtyTwoBits] = cached_hresult(0xD0370026, 'HRESULT_FROM_NT(STATUS_VID_MBP_COUNT_EXCEEDED_LIMIT)')
    NTSTATUS__STATUS_VID_SAVED_STATE_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0xD0370027, 'HRESULT_FROM_NT(STATUS_VID_SAVED_STATE_CORRUPT)')
    NTSTATUS__STATUS_VID_SAVED_STATE_UNRECOGNIZED_ITEM: Final[ThirtyTwoBits] = cached_hresult(0xD0370028, 'HRESULT_FROM_NT(STATUS_VID_SAVED_STATE_UNRECOGNIZED_ITEM)')
    NTSTATUS__STATUS_VID_SAVED_STATE_INCOMPATIBLE: Final[ThirtyTwoBits] = cached_hresult(0xD0370029, 'HRESULT_FROM_NT(STATUS_VID_SAVED_STATE_INCOMPATIBLE)')
    NTSTATUS__STATUS_VID_VTL_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0xD037002A, 'HRESULT_FROM_NT(STATUS_VID_VTL_ACCESS_DENIED)')
    NTSTATUS__STATUS_VID_INSUFFICIENT_RESOURCES_RESERVE: Final[ThirtyTwoBits] = cached_hresult(0xD037002B, 'HRESULT_FROM_NT(STATUS_VID_INSUFFICIENT_RESOURCES_RESERVE)')
    NTSTATUS__STATUS_VID_INSUFFICIENT_RESOURCES_PHYSICAL_BUFFER: Final[ThirtyTwoBits] = cached_hresult(0xD037002C, 'HRESULT_FROM_NT(STATUS_VID_INSUFFICIENT_RESOURCES_PHYSICAL_BUFFER)')
    NTSTATUS__STATUS_VID_INSUFFICIENT_RESOURCES_HV_DEPOSIT: Final[ThirtyTwoBits] = cached_hresult(0xD037002D, 'HRESULT_FROM_NT(STATUS_VID_INSUFFICIENT_RESOURCES_HV_DEPOSIT)')
    NTSTATUS__STATUS_VID_MEMORY_TYPE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD037002E, 'HRESULT_FROM_NT(STATUS_VID_MEMORY_TYPE_NOT_SUPPORTED)')
    NTSTATUS__STATUS_VID_INSUFFICIENT_RESOURCES_WITHDRAW: Final[ThirtyTwoBits] = cached_hresult(0xD037002F, 'HRESULT_FROM_NT(STATUS_VID_INSUFFICIENT_RESOURCES_WITHDRAW)')
    NTSTATUS__STATUS_VID_PROCESS_ALREADY_SET: Final[ThirtyTwoBits] = cached_hresult(0xD0370030, 'HRESULT_FROM_NT(STATUS_VID_PROCESS_ALREADY_SET)')
    NTSTATUS__STATUS_DM_OPERATION_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_hresult(0xD0370600, 'HRESULT_FROM_NT(STATUS_DM_OPERATION_LIMIT_EXCEEDED)')
    NTSTATUS__STATUS_VOLMGR_DATABASE_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD0380001, 'HRESULT_FROM_NT(STATUS_VOLMGR_DATABASE_FULL)')
    NTSTATUS__STATUS_VOLMGR_DISK_CONFIGURATION_CORRUPTED: Final[ThirtyTwoBits] = cached_hresult(0xD0380002, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_CONFIGURATION_CORRUPTED)')
    NTSTATUS__STATUS_VOLMGR_DISK_CONFIGURATION_NOT_IN_SYNC: Final[ThirtyTwoBits] = cached_hresult(0xD0380003, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_CONFIGURATION_NOT_IN_SYNC)')
    NTSTATUS__STATUS_VOLMGR_PACK_CONFIG_UPDATE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0380004, 'HRESULT_FROM_NT(STATUS_VOLMGR_PACK_CONFIG_UPDATE_FAILED)')
    NTSTATUS__STATUS_VOLMGR_DISK_CONTAINS_NON_SIMPLE_VOLUME: Final[ThirtyTwoBits] = cached_hresult(0xD0380005, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_CONTAINS_NON_SIMPLE_VOLUME)')
    NTSTATUS__STATUS_VOLMGR_DISK_DUPLICATE: Final[ThirtyTwoBits] = cached_hresult(0xD0380006, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_DUPLICATE)')
    NTSTATUS__STATUS_VOLMGR_DISK_DYNAMIC: Final[ThirtyTwoBits] = cached_hresult(0xD0380007, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_DYNAMIC)')
    NTSTATUS__STATUS_VOLMGR_DISK_ID_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0380008, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_ID_INVALID)')
    NTSTATUS__STATUS_VOLMGR_DISK_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0380009, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_INVALID)')
    NTSTATUS__STATUS_VOLMGR_DISK_LAST_VOTER: Final[ThirtyTwoBits] = cached_hresult(0xD038000A, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_LAST_VOTER)')
    NTSTATUS__STATUS_VOLMGR_DISK_LAYOUT_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD038000B, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_LAYOUT_INVALID)')
    NTSTATUS__STATUS_VOLMGR_DISK_LAYOUT_NON_BASIC_BETWEEN_BASIC_PARTITIONS: Final[ThirtyTwoBits] = cached_hresult(0xD038000C, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_LAYOUT_NON_BASIC_BETWEEN_BASIC_PARTITIONS)')
    NTSTATUS__STATUS_VOLMGR_DISK_LAYOUT_NOT_CYLINDER_ALIGNED: Final[ThirtyTwoBits] = cached_hresult(0xD038000D, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_LAYOUT_NOT_CYLINDER_ALIGNED)')
    NTSTATUS__STATUS_VOLMGR_DISK_LAYOUT_PARTITIONS_TOO_SMALL: Final[ThirtyTwoBits] = cached_hresult(0xD038000E, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_LAYOUT_PARTITIONS_TOO_SMALL)')
    NTSTATUS__STATUS_VOLMGR_DISK_LAYOUT_PRIMARY_BETWEEN_LOGICAL_PARTITIONS: Final[ThirtyTwoBits] = cached_hresult(0xD038000F, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_LAYOUT_PRIMARY_BETWEEN_LOGICAL_PARTITIONS)')
    NTSTATUS__STATUS_VOLMGR_DISK_LAYOUT_TOO_MANY_PARTITIONS: Final[ThirtyTwoBits] = cached_hresult(0xD0380010, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_LAYOUT_TOO_MANY_PARTITIONS)')
    NTSTATUS__STATUS_VOLMGR_DISK_MISSING: Final[ThirtyTwoBits] = cached_hresult(0xD0380011, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_MISSING)')
    NTSTATUS__STATUS_VOLMGR_DISK_NOT_EMPTY: Final[ThirtyTwoBits] = cached_hresult(0xD0380012, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_NOT_EMPTY)')
    NTSTATUS__STATUS_VOLMGR_DISK_NOT_ENOUGH_SPACE: Final[ThirtyTwoBits] = cached_hresult(0xD0380013, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_NOT_ENOUGH_SPACE)')
    NTSTATUS__STATUS_VOLMGR_DISK_REVECTORING_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0380014, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_REVECTORING_FAILED)')
    NTSTATUS__STATUS_VOLMGR_DISK_SECTOR_SIZE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0380015, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_SECTOR_SIZE_INVALID)')
    NTSTATUS__STATUS_VOLMGR_DISK_SET_NOT_CONTAINED: Final[ThirtyTwoBits] = cached_hresult(0xD0380016, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_SET_NOT_CONTAINED)')
    NTSTATUS__STATUS_VOLMGR_DISK_USED_BY_MULTIPLE_MEMBERS: Final[ThirtyTwoBits] = cached_hresult(0xD0380017, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_USED_BY_MULTIPLE_MEMBERS)')
    NTSTATUS__STATUS_VOLMGR_DISK_USED_BY_MULTIPLE_PLEXES: Final[ThirtyTwoBits] = cached_hresult(0xD0380018, 'HRESULT_FROM_NT(STATUS_VOLMGR_DISK_USED_BY_MULTIPLE_PLEXES)')
    NTSTATUS__STATUS_VOLMGR_DYNAMIC_DISK_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD0380019, 'HRESULT_FROM_NT(STATUS_VOLMGR_DYNAMIC_DISK_NOT_SUPPORTED)')
    NTSTATUS__STATUS_VOLMGR_EXTENT_ALREADY_USED: Final[ThirtyTwoBits] = cached_hresult(0xD038001A, 'HRESULT_FROM_NT(STATUS_VOLMGR_EXTENT_ALREADY_USED)')
    NTSTATUS__STATUS_VOLMGR_EXTENT_NOT_CONTIGUOUS: Final[ThirtyTwoBits] = cached_hresult(0xD038001B, 'HRESULT_FROM_NT(STATUS_VOLMGR_EXTENT_NOT_CONTIGUOUS)')
    NTSTATUS__STATUS_VOLMGR_EXTENT_NOT_IN_PUBLIC_REGION: Final[ThirtyTwoBits] = cached_hresult(0xD038001C, 'HRESULT_FROM_NT(STATUS_VOLMGR_EXTENT_NOT_IN_PUBLIC_REGION)')
    NTSTATUS__STATUS_VOLMGR_EXTENT_NOT_SECTOR_ALIGNED: Final[ThirtyTwoBits] = cached_hresult(0xD038001D, 'HRESULT_FROM_NT(STATUS_VOLMGR_EXTENT_NOT_SECTOR_ALIGNED)')
    NTSTATUS__STATUS_VOLMGR_EXTENT_OVERLAPS_EBR_PARTITION: Final[ThirtyTwoBits] = cached_hresult(0xD038001E, 'HRESULT_FROM_NT(STATUS_VOLMGR_EXTENT_OVERLAPS_EBR_PARTITION)')
    NTSTATUS__STATUS_VOLMGR_EXTENT_VOLUME_LENGTHS_DO_NOT_MATCH: Final[ThirtyTwoBits] = cached_hresult(0xD038001F, 'HRESULT_FROM_NT(STATUS_VOLMGR_EXTENT_VOLUME_LENGTHS_DO_NOT_MATCH)')
    NTSTATUS__STATUS_VOLMGR_FAULT_TOLERANT_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD0380020, 'HRESULT_FROM_NT(STATUS_VOLMGR_FAULT_TOLERANT_NOT_SUPPORTED)')
    NTSTATUS__STATUS_VOLMGR_INTERLEAVE_LENGTH_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0380021, 'HRESULT_FROM_NT(STATUS_VOLMGR_INTERLEAVE_LENGTH_INVALID)')
    NTSTATUS__STATUS_VOLMGR_MAXIMUM_REGISTERED_USERS: Final[ThirtyTwoBits] = cached_hresult(0xD0380022, 'HRESULT_FROM_NT(STATUS_VOLMGR_MAXIMUM_REGISTERED_USERS)')
    NTSTATUS__STATUS_VOLMGR_MEMBER_IN_SYNC: Final[ThirtyTwoBits] = cached_hresult(0xD0380023, 'HRESULT_FROM_NT(STATUS_VOLMGR_MEMBER_IN_SYNC)')
    NTSTATUS__STATUS_VOLMGR_MEMBER_INDEX_DUPLICATE: Final[ThirtyTwoBits] = cached_hresult(0xD0380024, 'HRESULT_FROM_NT(STATUS_VOLMGR_MEMBER_INDEX_DUPLICATE)')
    NTSTATUS__STATUS_VOLMGR_MEMBER_INDEX_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0380025, 'HRESULT_FROM_NT(STATUS_VOLMGR_MEMBER_INDEX_INVALID)')
    NTSTATUS__STATUS_VOLMGR_MEMBER_MISSING: Final[ThirtyTwoBits] = cached_hresult(0xD0380026, 'HRESULT_FROM_NT(STATUS_VOLMGR_MEMBER_MISSING)')
    NTSTATUS__STATUS_VOLMGR_MEMBER_NOT_DETACHED: Final[ThirtyTwoBits] = cached_hresult(0xD0380027, 'HRESULT_FROM_NT(STATUS_VOLMGR_MEMBER_NOT_DETACHED)')
    NTSTATUS__STATUS_VOLMGR_MEMBER_REGENERATING: Final[ThirtyTwoBits] = cached_hresult(0xD0380028, 'HRESULT_FROM_NT(STATUS_VOLMGR_MEMBER_REGENERATING)')
    NTSTATUS__STATUS_VOLMGR_ALL_DISKS_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0380029, 'HRESULT_FROM_NT(STATUS_VOLMGR_ALL_DISKS_FAILED)')
    NTSTATUS__STATUS_VOLMGR_NO_REGISTERED_USERS: Final[ThirtyTwoBits] = cached_hresult(0xD038002A, 'HRESULT_FROM_NT(STATUS_VOLMGR_NO_REGISTERED_USERS)')
    NTSTATUS__STATUS_VOLMGR_NO_SUCH_USER: Final[ThirtyTwoBits] = cached_hresult(0xD038002B, 'HRESULT_FROM_NT(STATUS_VOLMGR_NO_SUCH_USER)')
    NTSTATUS__STATUS_VOLMGR_NOTIFICATION_RESET: Final[ThirtyTwoBits] = cached_hresult(0xD038002C, 'HRESULT_FROM_NT(STATUS_VOLMGR_NOTIFICATION_RESET)')
    NTSTATUS__STATUS_VOLMGR_NUMBER_OF_MEMBERS_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD038002D, 'HRESULT_FROM_NT(STATUS_VOLMGR_NUMBER_OF_MEMBERS_INVALID)')
    NTSTATUS__STATUS_VOLMGR_NUMBER_OF_PLEXES_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD038002E, 'HRESULT_FROM_NT(STATUS_VOLMGR_NUMBER_OF_PLEXES_INVALID)')
    NTSTATUS__STATUS_VOLMGR_PACK_DUPLICATE: Final[ThirtyTwoBits] = cached_hresult(0xD038002F, 'HRESULT_FROM_NT(STATUS_VOLMGR_PACK_DUPLICATE)')
    NTSTATUS__STATUS_VOLMGR_PACK_ID_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0380030, 'HRESULT_FROM_NT(STATUS_VOLMGR_PACK_ID_INVALID)')
    NTSTATUS__STATUS_VOLMGR_PACK_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0380031, 'HRESULT_FROM_NT(STATUS_VOLMGR_PACK_INVALID)')
    NTSTATUS__STATUS_VOLMGR_PACK_NAME_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0380032, 'HRESULT_FROM_NT(STATUS_VOLMGR_PACK_NAME_INVALID)')
    NTSTATUS__STATUS_VOLMGR_PACK_OFFLINE: Final[ThirtyTwoBits] = cached_hresult(0xD0380033, 'HRESULT_FROM_NT(STATUS_VOLMGR_PACK_OFFLINE)')
    NTSTATUS__STATUS_VOLMGR_PACK_HAS_QUORUM: Final[ThirtyTwoBits] = cached_hresult(0xD0380034, 'HRESULT_FROM_NT(STATUS_VOLMGR_PACK_HAS_QUORUM)')
    NTSTATUS__STATUS_VOLMGR_PACK_WITHOUT_QUORUM: Final[ThirtyTwoBits] = cached_hresult(0xD0380035, 'HRESULT_FROM_NT(STATUS_VOLMGR_PACK_WITHOUT_QUORUM)')
    NTSTATUS__STATUS_VOLMGR_PARTITION_STYLE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0380036, 'HRESULT_FROM_NT(STATUS_VOLMGR_PARTITION_STYLE_INVALID)')
    NTSTATUS__STATUS_VOLMGR_PARTITION_UPDATE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0380037, 'HRESULT_FROM_NT(STATUS_VOLMGR_PARTITION_UPDATE_FAILED)')
    NTSTATUS__STATUS_VOLMGR_PLEX_IN_SYNC: Final[ThirtyTwoBits] = cached_hresult(0xD0380038, 'HRESULT_FROM_NT(STATUS_VOLMGR_PLEX_IN_SYNC)')
    NTSTATUS__STATUS_VOLMGR_PLEX_INDEX_DUPLICATE: Final[ThirtyTwoBits] = cached_hresult(0xD0380039, 'HRESULT_FROM_NT(STATUS_VOLMGR_PLEX_INDEX_DUPLICATE)')
    NTSTATUS__STATUS_VOLMGR_PLEX_INDEX_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD038003A, 'HRESULT_FROM_NT(STATUS_VOLMGR_PLEX_INDEX_INVALID)')
    NTSTATUS__STATUS_VOLMGR_PLEX_LAST_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0xD038003B, 'HRESULT_FROM_NT(STATUS_VOLMGR_PLEX_LAST_ACTIVE)')
    NTSTATUS__STATUS_VOLMGR_PLEX_MISSING: Final[ThirtyTwoBits] = cached_hresult(0xD038003C, 'HRESULT_FROM_NT(STATUS_VOLMGR_PLEX_MISSING)')
    NTSTATUS__STATUS_VOLMGR_PLEX_REGENERATING: Final[ThirtyTwoBits] = cached_hresult(0xD038003D, 'HRESULT_FROM_NT(STATUS_VOLMGR_PLEX_REGENERATING)')
    NTSTATUS__STATUS_VOLMGR_PLEX_TYPE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD038003E, 'HRESULT_FROM_NT(STATUS_VOLMGR_PLEX_TYPE_INVALID)')
    NTSTATUS__STATUS_VOLMGR_PLEX_NOT_RAID5: Final[ThirtyTwoBits] = cached_hresult(0xD038003F, 'HRESULT_FROM_NT(STATUS_VOLMGR_PLEX_NOT_RAID5)')
    NTSTATUS__STATUS_VOLMGR_PLEX_NOT_SIMPLE: Final[ThirtyTwoBits] = cached_hresult(0xD0380040, 'HRESULT_FROM_NT(STATUS_VOLMGR_PLEX_NOT_SIMPLE)')
    NTSTATUS__STATUS_VOLMGR_STRUCTURE_SIZE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0380041, 'HRESULT_FROM_NT(STATUS_VOLMGR_STRUCTURE_SIZE_INVALID)')
    NTSTATUS__STATUS_VOLMGR_TOO_MANY_NOTIFICATION_REQUESTS: Final[ThirtyTwoBits] = cached_hresult(0xD0380042, 'HRESULT_FROM_NT(STATUS_VOLMGR_TOO_MANY_NOTIFICATION_REQUESTS)')
    NTSTATUS__STATUS_VOLMGR_TRANSACTION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD0380043, 'HRESULT_FROM_NT(STATUS_VOLMGR_TRANSACTION_IN_PROGRESS)')
    NTSTATUS__STATUS_VOLMGR_UNEXPECTED_DISK_LAYOUT_CHANGE: Final[ThirtyTwoBits] = cached_hresult(0xD0380044, 'HRESULT_FROM_NT(STATUS_VOLMGR_UNEXPECTED_DISK_LAYOUT_CHANGE)')
    NTSTATUS__STATUS_VOLMGR_VOLUME_CONTAINS_MISSING_DISK: Final[ThirtyTwoBits] = cached_hresult(0xD0380045, 'HRESULT_FROM_NT(STATUS_VOLMGR_VOLUME_CONTAINS_MISSING_DISK)')
    NTSTATUS__STATUS_VOLMGR_VOLUME_ID_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0380046, 'HRESULT_FROM_NT(STATUS_VOLMGR_VOLUME_ID_INVALID)')
    NTSTATUS__STATUS_VOLMGR_VOLUME_LENGTH_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0380047, 'HRESULT_FROM_NT(STATUS_VOLMGR_VOLUME_LENGTH_INVALID)')
    NTSTATUS__STATUS_VOLMGR_VOLUME_LENGTH_NOT_SECTOR_SIZE_MULTIPLE: Final[ThirtyTwoBits] = cached_hresult(0xD0380048, 'HRESULT_FROM_NT(STATUS_VOLMGR_VOLUME_LENGTH_NOT_SECTOR_SIZE_MULTIPLE)')
    NTSTATUS__STATUS_VOLMGR_VOLUME_NOT_MIRRORED: Final[ThirtyTwoBits] = cached_hresult(0xD0380049, 'HRESULT_FROM_NT(STATUS_VOLMGR_VOLUME_NOT_MIRRORED)')
    NTSTATUS__STATUS_VOLMGR_VOLUME_NOT_RETAINED: Final[ThirtyTwoBits] = cached_hresult(0xD038004A, 'HRESULT_FROM_NT(STATUS_VOLMGR_VOLUME_NOT_RETAINED)')
    NTSTATUS__STATUS_VOLMGR_VOLUME_OFFLINE: Final[ThirtyTwoBits] = cached_hresult(0xD038004B, 'HRESULT_FROM_NT(STATUS_VOLMGR_VOLUME_OFFLINE)')
    NTSTATUS__STATUS_VOLMGR_VOLUME_RETAINED: Final[ThirtyTwoBits] = cached_hresult(0xD038004C, 'HRESULT_FROM_NT(STATUS_VOLMGR_VOLUME_RETAINED)')
    NTSTATUS__STATUS_VOLMGR_NUMBER_OF_EXTENTS_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD038004D, 'HRESULT_FROM_NT(STATUS_VOLMGR_NUMBER_OF_EXTENTS_INVALID)')
    NTSTATUS__STATUS_VOLMGR_DIFFERENT_SECTOR_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD038004E, 'HRESULT_FROM_NT(STATUS_VOLMGR_DIFFERENT_SECTOR_SIZE)')
    NTSTATUS__STATUS_VOLMGR_BAD_BOOT_DISK: Final[ThirtyTwoBits] = cached_hresult(0xD038004F, 'HRESULT_FROM_NT(STATUS_VOLMGR_BAD_BOOT_DISK)')
    NTSTATUS__STATUS_VOLMGR_PACK_CONFIG_OFFLINE: Final[ThirtyTwoBits] = cached_hresult(0xD0380050, 'HRESULT_FROM_NT(STATUS_VOLMGR_PACK_CONFIG_OFFLINE)')
    NTSTATUS__STATUS_VOLMGR_PACK_CONFIG_ONLINE: Final[ThirtyTwoBits] = cached_hresult(0xD0380051, 'HRESULT_FROM_NT(STATUS_VOLMGR_PACK_CONFIG_ONLINE)')
    NTSTATUS__STATUS_VOLMGR_NOT_PRIMARY_PACK: Final[ThirtyTwoBits] = cached_hresult(0xD0380052, 'HRESULT_FROM_NT(STATUS_VOLMGR_NOT_PRIMARY_PACK)')
    NTSTATUS__STATUS_VOLMGR_PACK_LOG_UPDATE_FAILED: Final[ThirtyTwoBits] = cached_hresult(0xD0380053, 'HRESULT_FROM_NT(STATUS_VOLMGR_PACK_LOG_UPDATE_FAILED)')
    NTSTATUS__STATUS_VOLMGR_NUMBER_OF_DISKS_IN_PLEX_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0380054, 'HRESULT_FROM_NT(STATUS_VOLMGR_NUMBER_OF_DISKS_IN_PLEX_INVALID)')
    NTSTATUS__STATUS_VOLMGR_NUMBER_OF_DISKS_IN_MEMBER_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0380055, 'HRESULT_FROM_NT(STATUS_VOLMGR_NUMBER_OF_DISKS_IN_MEMBER_INVALID)')
    NTSTATUS__STATUS_VOLMGR_VOLUME_MIRRORED: Final[ThirtyTwoBits] = cached_hresult(0xD0380056, 'HRESULT_FROM_NT(STATUS_VOLMGR_VOLUME_MIRRORED)')
    NTSTATUS__STATUS_VOLMGR_PLEX_NOT_SIMPLE_SPANNED: Final[ThirtyTwoBits] = cached_hresult(0xD0380057, 'HRESULT_FROM_NT(STATUS_VOLMGR_PLEX_NOT_SIMPLE_SPANNED)')
    NTSTATUS__STATUS_VOLMGR_NO_VALID_LOG_COPIES: Final[ThirtyTwoBits] = cached_hresult(0xD0380058, 'HRESULT_FROM_NT(STATUS_VOLMGR_NO_VALID_LOG_COPIES)')
    NTSTATUS__STATUS_VOLMGR_PRIMARY_PACK_PRESENT: Final[ThirtyTwoBits] = cached_hresult(0xD0380059, 'HRESULT_FROM_NT(STATUS_VOLMGR_PRIMARY_PACK_PRESENT)')
    NTSTATUS__STATUS_VOLMGR_NUMBER_OF_DISKS_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD038005A, 'HRESULT_FROM_NT(STATUS_VOLMGR_NUMBER_OF_DISKS_INVALID)')
    NTSTATUS__STATUS_VOLMGR_MIRROR_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD038005B, 'HRESULT_FROM_NT(STATUS_VOLMGR_MIRROR_NOT_SUPPORTED)')
    NTSTATUS__STATUS_VOLMGR_RAID5_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD038005C, 'HRESULT_FROM_NT(STATUS_VOLMGR_RAID5_NOT_SUPPORTED)')
    NTSTATUS__STATUS_BCD_TOO_MANY_ELEMENTS: Final[ThirtyTwoBits] = cached_hresult(0xD0390002, 'HRESULT_FROM_NT(STATUS_BCD_TOO_MANY_ELEMENTS)')
    NTSTATUS__STATUS_VHD_DRIVE_FOOTER_MISSING: Final[ThirtyTwoBits] = cached_hresult(0xD03A0001, 'HRESULT_FROM_NT(STATUS_VHD_DRIVE_FOOTER_MISSING)')
    NTSTATUS__STATUS_VHD_DRIVE_FOOTER_CHECKSUM_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD03A0002, 'HRESULT_FROM_NT(STATUS_VHD_DRIVE_FOOTER_CHECKSUM_MISMATCH)')
    NTSTATUS__STATUS_VHD_DRIVE_FOOTER_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0xD03A0003, 'HRESULT_FROM_NT(STATUS_VHD_DRIVE_FOOTER_CORRUPT)')
    NTSTATUS__STATUS_VHD_FORMAT_UNKNOWN: Final[ThirtyTwoBits] = cached_hresult(0xD03A0004, 'HRESULT_FROM_NT(STATUS_VHD_FORMAT_UNKNOWN)')
    NTSTATUS__STATUS_VHD_FORMAT_UNSUPPORTED_VERSION: Final[ThirtyTwoBits] = cached_hresult(0xD03A0005, 'HRESULT_FROM_NT(STATUS_VHD_FORMAT_UNSUPPORTED_VERSION)')
    NTSTATUS__STATUS_VHD_SPARSE_HEADER_CHECKSUM_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD03A0006, 'HRESULT_FROM_NT(STATUS_VHD_SPARSE_HEADER_CHECKSUM_MISMATCH)')
    NTSTATUS__STATUS_VHD_SPARSE_HEADER_UNSUPPORTED_VERSION: Final[ThirtyTwoBits] = cached_hresult(0xD03A0007, 'HRESULT_FROM_NT(STATUS_VHD_SPARSE_HEADER_UNSUPPORTED_VERSION)')
    NTSTATUS__STATUS_VHD_SPARSE_HEADER_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0xD03A0008, 'HRESULT_FROM_NT(STATUS_VHD_SPARSE_HEADER_CORRUPT)')
    NTSTATUS__STATUS_VHD_BLOCK_ALLOCATION_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD03A0009, 'HRESULT_FROM_NT(STATUS_VHD_BLOCK_ALLOCATION_FAILURE)')
    NTSTATUS__STATUS_VHD_BLOCK_ALLOCATION_TABLE_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0xD03A000A, 'HRESULT_FROM_NT(STATUS_VHD_BLOCK_ALLOCATION_TABLE_CORRUPT)')
    NTSTATUS__STATUS_VHD_INVALID_BLOCK_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD03A000B, 'HRESULT_FROM_NT(STATUS_VHD_INVALID_BLOCK_SIZE)')
    NTSTATUS__STATUS_VHD_BITMAP_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD03A000C, 'HRESULT_FROM_NT(STATUS_VHD_BITMAP_MISMATCH)')
    NTSTATUS__STATUS_VHD_PARENT_VHD_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD03A000D, 'HRESULT_FROM_NT(STATUS_VHD_PARENT_VHD_NOT_FOUND)')
    NTSTATUS__STATUS_VHD_CHILD_PARENT_ID_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD03A000E, 'HRESULT_FROM_NT(STATUS_VHD_CHILD_PARENT_ID_MISMATCH)')
    NTSTATUS__STATUS_VHD_CHILD_PARENT_TIMESTAMP_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD03A000F, 'HRESULT_FROM_NT(STATUS_VHD_CHILD_PARENT_TIMESTAMP_MISMATCH)')
    NTSTATUS__STATUS_VHD_METADATA_READ_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD03A0010, 'HRESULT_FROM_NT(STATUS_VHD_METADATA_READ_FAILURE)')
    NTSTATUS__STATUS_VHD_METADATA_WRITE_FAILURE: Final[ThirtyTwoBits] = cached_hresult(0xD03A0011, 'HRESULT_FROM_NT(STATUS_VHD_METADATA_WRITE_FAILURE)')
    NTSTATUS__STATUS_VHD_INVALID_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD03A0012, 'HRESULT_FROM_NT(STATUS_VHD_INVALID_SIZE)')
    NTSTATUS__STATUS_VHD_INVALID_FILE_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD03A0013, 'HRESULT_FROM_NT(STATUS_VHD_INVALID_FILE_SIZE)')
    NTSTATUS__STATUS_VIRTDISK_PROVIDER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD03A0014, 'HRESULT_FROM_NT(STATUS_VIRTDISK_PROVIDER_NOT_FOUND)')
    NTSTATUS__STATUS_VIRTDISK_NOT_VIRTUAL_DISK: Final[ThirtyTwoBits] = cached_hresult(0xD03A0015, 'HRESULT_FROM_NT(STATUS_VIRTDISK_NOT_VIRTUAL_DISK)')
    NTSTATUS__STATUS_VHD_PARENT_VHD_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_hresult(0xD03A0016, 'HRESULT_FROM_NT(STATUS_VHD_PARENT_VHD_ACCESS_DENIED)')
    NTSTATUS__STATUS_VHD_CHILD_PARENT_SIZE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD03A0017, 'HRESULT_FROM_NT(STATUS_VHD_CHILD_PARENT_SIZE_MISMATCH)')
    NTSTATUS__STATUS_VHD_DIFFERENCING_CHAIN_CYCLE_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0xD03A0018, 'HRESULT_FROM_NT(STATUS_VHD_DIFFERENCING_CHAIN_CYCLE_DETECTED)')
    NTSTATUS__STATUS_VHD_DIFFERENCING_CHAIN_ERROR_IN_PARENT: Final[ThirtyTwoBits] = cached_hresult(0xD03A0019, 'HRESULT_FROM_NT(STATUS_VHD_DIFFERENCING_CHAIN_ERROR_IN_PARENT)')
    NTSTATUS__STATUS_VIRTUAL_DISK_LIMITATION: Final[ThirtyTwoBits] = cached_hresult(0xD03A001A, 'HRESULT_FROM_NT(STATUS_VIRTUAL_DISK_LIMITATION)')
    NTSTATUS__STATUS_VHD_INVALID_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xD03A001B, 'HRESULT_FROM_NT(STATUS_VHD_INVALID_TYPE)')
    NTSTATUS__STATUS_VHD_INVALID_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD03A001C, 'HRESULT_FROM_NT(STATUS_VHD_INVALID_STATE)')
    NTSTATUS__STATUS_VIRTDISK_UNSUPPORTED_DISK_SECTOR_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD03A001D, 'HRESULT_FROM_NT(STATUS_VIRTDISK_UNSUPPORTED_DISK_SECTOR_SIZE)')
    NTSTATUS__STATUS_VIRTDISK_DISK_ALREADY_OWNED: Final[ThirtyTwoBits] = cached_hresult(0xD03A001E, 'HRESULT_FROM_NT(STATUS_VIRTDISK_DISK_ALREADY_OWNED)')
    NTSTATUS__STATUS_VIRTDISK_DISK_ONLINE_AND_WRITABLE: Final[ThirtyTwoBits] = cached_hresult(0xD03A001F, 'HRESULT_FROM_NT(STATUS_VIRTDISK_DISK_ONLINE_AND_WRITABLE)')
    NTSTATUS__STATUS_CTLOG_TRACKING_NOT_INITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0xD03A0020, 'HRESULT_FROM_NT(STATUS_CTLOG_TRACKING_NOT_INITIALIZED)')
    NTSTATUS__STATUS_CTLOG_LOGFILE_SIZE_EXCEEDED_MAXSIZE: Final[ThirtyTwoBits] = cached_hresult(0xD03A0021, 'HRESULT_FROM_NT(STATUS_CTLOG_LOGFILE_SIZE_EXCEEDED_MAXSIZE)')
    NTSTATUS__STATUS_CTLOG_VHD_CHANGED_OFFLINE: Final[ThirtyTwoBits] = cached_hresult(0xD03A0022, 'HRESULT_FROM_NT(STATUS_CTLOG_VHD_CHANGED_OFFLINE)')
    NTSTATUS__STATUS_CTLOG_INVALID_TRACKING_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD03A0023, 'HRESULT_FROM_NT(STATUS_CTLOG_INVALID_TRACKING_STATE)')
    NTSTATUS__STATUS_CTLOG_INCONSISTENT_TRACKING_FILE: Final[ThirtyTwoBits] = cached_hresult(0xD03A0024, 'HRESULT_FROM_NT(STATUS_CTLOG_INCONSISTENT_TRACKING_FILE)')
    NTSTATUS__STATUS_VHD_METADATA_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD03A0028, 'HRESULT_FROM_NT(STATUS_VHD_METADATA_FULL)')
    NTSTATUS__STATUS_VHD_INVALID_CHANGE_TRACKING_ID: Final[ThirtyTwoBits] = cached_hresult(0xD03A0029, 'HRESULT_FROM_NT(STATUS_VHD_INVALID_CHANGE_TRACKING_ID)')
    NTSTATUS__STATUS_VHD_CHANGE_TRACKING_DISABLED: Final[ThirtyTwoBits] = cached_hresult(0xD03A002A, 'HRESULT_FROM_NT(STATUS_VHD_CHANGE_TRACKING_DISABLED)')
    NTSTATUS__STATUS_VHD_MISSING_CHANGE_TRACKING_INFORMATION: Final[ThirtyTwoBits] = cached_hresult(0xD03A0030, 'HRESULT_FROM_NT(STATUS_VHD_MISSING_CHANGE_TRACKING_INFORMATION)')
    NTSTATUS__STATUS_VHD_RESIZE_WOULD_TRUNCATE_DATA: Final[ThirtyTwoBits] = cached_hresult(0xD03A0031, 'HRESULT_FROM_NT(STATUS_VHD_RESIZE_WOULD_TRUNCATE_DATA)')
    NTSTATUS__STATUS_VHD_COULD_NOT_COMPUTE_MINIMUM_VIRTUAL_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD03A0032, 'HRESULT_FROM_NT(STATUS_VHD_COULD_NOT_COMPUTE_MINIMUM_VIRTUAL_SIZE)')
    NTSTATUS__STATUS_VHD_ALREADY_AT_OR_BELOW_MINIMUM_VIRTUAL_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD03A0033, 'HRESULT_FROM_NT(STATUS_VHD_ALREADY_AT_OR_BELOW_MINIMUM_VIRTUAL_SIZE)')
    NTSTATUS__STATUS_VHD_UNEXPECTED_ID: Final[ThirtyTwoBits] = cached_hresult(0xD03A0034, 'HRESULT_FROM_NT(STATUS_VHD_UNEXPECTED_ID)')
    NTSTATUS__STATUS_RKF_KEY_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0400001, 'HRESULT_FROM_NT(STATUS_RKF_KEY_NOT_FOUND)')
    NTSTATUS__STATUS_RKF_DUPLICATE_KEY: Final[ThirtyTwoBits] = cached_hresult(0xD0400002, 'HRESULT_FROM_NT(STATUS_RKF_DUPLICATE_KEY)')
    NTSTATUS__STATUS_RKF_BLOB_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD0400003, 'HRESULT_FROM_NT(STATUS_RKF_BLOB_FULL)')
    NTSTATUS__STATUS_RKF_STORE_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD0400004, 'HRESULT_FROM_NT(STATUS_RKF_STORE_FULL)')
    NTSTATUS__STATUS_RKF_FILE_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0xD0400005, 'HRESULT_FROM_NT(STATUS_RKF_FILE_BLOCKED)')
    NTSTATUS__STATUS_RKF_ACTIVE_KEY: Final[ThirtyTwoBits] = cached_hresult(0xD0400006, 'HRESULT_FROM_NT(STATUS_RKF_ACTIVE_KEY)')
    NTSTATUS__STATUS_RDBSS_RESTART_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0xD0410001, 'HRESULT_FROM_NT(STATUS_RDBSS_RESTART_OPERATION)')
    NTSTATUS__STATUS_RDBSS_CONTINUE_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0xD0410002, 'HRESULT_FROM_NT(STATUS_RDBSS_CONTINUE_OPERATION)')
    NTSTATUS__STATUS_RDBSS_POST_OPERATION: Final[ThirtyTwoBits] = cached_hresult(0xD0410003, 'HRESULT_FROM_NT(STATUS_RDBSS_POST_OPERATION)')
    NTSTATUS__STATUS_RDBSS_RETRY_LOOKUP: Final[ThirtyTwoBits] = cached_hresult(0xD0410004, 'HRESULT_FROM_NT(STATUS_RDBSS_RETRY_LOOKUP)')
    NTSTATUS__STATUS_BTH_ATT_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_hresult(0xD0420001, 'HRESULT_FROM_NT(STATUS_BTH_ATT_INVALID_HANDLE)')
    NTSTATUS__STATUS_BTH_ATT_READ_NOT_PERMITTED: Final[ThirtyTwoBits] = cached_hresult(0xD0420002, 'HRESULT_FROM_NT(STATUS_BTH_ATT_READ_NOT_PERMITTED)')
    NTSTATUS__STATUS_BTH_ATT_WRITE_NOT_PERMITTED: Final[ThirtyTwoBits] = cached_hresult(0xD0420003, 'HRESULT_FROM_NT(STATUS_BTH_ATT_WRITE_NOT_PERMITTED)')
    NTSTATUS__STATUS_BTH_ATT_INVALID_PDU: Final[ThirtyTwoBits] = cached_hresult(0xD0420004, 'HRESULT_FROM_NT(STATUS_BTH_ATT_INVALID_PDU)')
    NTSTATUS__STATUS_BTH_ATT_INSUFFICIENT_AUTHENTICATION: Final[ThirtyTwoBits] = cached_hresult(0xD0420005, 'HRESULT_FROM_NT(STATUS_BTH_ATT_INSUFFICIENT_AUTHENTICATION)')
    NTSTATUS__STATUS_BTH_ATT_REQUEST_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD0420006, 'HRESULT_FROM_NT(STATUS_BTH_ATT_REQUEST_NOT_SUPPORTED)')
    NTSTATUS__STATUS_BTH_ATT_INVALID_OFFSET: Final[ThirtyTwoBits] = cached_hresult(0xD0420007, 'HRESULT_FROM_NT(STATUS_BTH_ATT_INVALID_OFFSET)')
    NTSTATUS__STATUS_BTH_ATT_INSUFFICIENT_AUTHORIZATION: Final[ThirtyTwoBits] = cached_hresult(0xD0420008, 'HRESULT_FROM_NT(STATUS_BTH_ATT_INSUFFICIENT_AUTHORIZATION)')
    NTSTATUS__STATUS_BTH_ATT_PREPARE_QUEUE_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD0420009, 'HRESULT_FROM_NT(STATUS_BTH_ATT_PREPARE_QUEUE_FULL)')
    NTSTATUS__STATUS_BTH_ATT_ATTRIBUTE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD042000A, 'HRESULT_FROM_NT(STATUS_BTH_ATT_ATTRIBUTE_NOT_FOUND)')
    NTSTATUS__STATUS_BTH_ATT_ATTRIBUTE_NOT_LONG: Final[ThirtyTwoBits] = cached_hresult(0xD042000B, 'HRESULT_FROM_NT(STATUS_BTH_ATT_ATTRIBUTE_NOT_LONG)')
    NTSTATUS__STATUS_BTH_ATT_INSUFFICIENT_ENCRYPTION_KEY_SIZE: Final[ThirtyTwoBits] = cached_hresult(0xD042000C, 'HRESULT_FROM_NT(STATUS_BTH_ATT_INSUFFICIENT_ENCRYPTION_KEY_SIZE)')
    NTSTATUS__STATUS_BTH_ATT_INVALID_ATTRIBUTE_VALUE_LENGTH: Final[ThirtyTwoBits] = cached_hresult(0xD042000D, 'HRESULT_FROM_NT(STATUS_BTH_ATT_INVALID_ATTRIBUTE_VALUE_LENGTH)')
    NTSTATUS__STATUS_BTH_ATT_UNLIKELY: Final[ThirtyTwoBits] = cached_hresult(0xD042000E, 'HRESULT_FROM_NT(STATUS_BTH_ATT_UNLIKELY)')
    NTSTATUS__STATUS_BTH_ATT_INSUFFICIENT_ENCRYPTION: Final[ThirtyTwoBits] = cached_hresult(0xD042000F, 'HRESULT_FROM_NT(STATUS_BTH_ATT_INSUFFICIENT_ENCRYPTION)')
    NTSTATUS__STATUS_BTH_ATT_UNSUPPORTED_GROUP_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xD0420010, 'HRESULT_FROM_NT(STATUS_BTH_ATT_UNSUPPORTED_GROUP_TYPE)')
    NTSTATUS__STATUS_BTH_ATT_INSUFFICIENT_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0xD0420011, 'HRESULT_FROM_NT(STATUS_BTH_ATT_INSUFFICIENT_RESOURCES)')
    NTSTATUS__STATUS_BTH_ATT_UNKNOWN_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0421000, 'HRESULT_FROM_NT(STATUS_BTH_ATT_UNKNOWN_ERROR)')
    NTSTATUS__STATUS_SECUREBOOT_ROLLBACK_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0xD0430001, 'HRESULT_FROM_NT(STATUS_SECUREBOOT_ROLLBACK_DETECTED)')
    NTSTATUS__STATUS_SECUREBOOT_POLICY_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0xD0430002, 'HRESULT_FROM_NT(STATUS_SECUREBOOT_POLICY_VIOLATION)')
    NTSTATUS__STATUS_SECUREBOOT_INVALID_POLICY: Final[ThirtyTwoBits] = cached_hresult(0xD0430003, 'HRESULT_FROM_NT(STATUS_SECUREBOOT_INVALID_POLICY)')
    NTSTATUS__STATUS_SECUREBOOT_POLICY_PUBLISHER_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0430004, 'HRESULT_FROM_NT(STATUS_SECUREBOOT_POLICY_PUBLISHER_NOT_FOUND)')
    NTSTATUS__STATUS_SECUREBOOT_POLICY_NOT_SIGNED: Final[ThirtyTwoBits] = cached_hresult(0xD0430005, 'HRESULT_FROM_NT(STATUS_SECUREBOOT_POLICY_NOT_SIGNED)')
    NTSTATUS__STATUS_SECUREBOOT_FILE_REPLACED: Final[ThirtyTwoBits] = cached_hresult(0xD0430007, 'HRESULT_FROM_NT(STATUS_SECUREBOOT_FILE_REPLACED)')
    NTSTATUS__STATUS_SECUREBOOT_POLICY_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_hresult(0xD0430008, 'HRESULT_FROM_NT(STATUS_SECUREBOOT_POLICY_NOT_AUTHORIZED)')
    NTSTATUS__STATUS_SECUREBOOT_POLICY_UNKNOWN: Final[ThirtyTwoBits] = cached_hresult(0xD0430009, 'HRESULT_FROM_NT(STATUS_SECUREBOOT_POLICY_UNKNOWN)')
    NTSTATUS__STATUS_SECUREBOOT_POLICY_MISSING_ANTIROLLBACKVERSION: Final[ThirtyTwoBits] = cached_hresult(0xD043000A, 'HRESULT_FROM_NT(STATUS_SECUREBOOT_POLICY_MISSING_ANTIROLLBACKVERSION)')
    NTSTATUS__STATUS_SECUREBOOT_PLATFORM_ID_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD043000B, 'HRESULT_FROM_NT(STATUS_SECUREBOOT_PLATFORM_ID_MISMATCH)')
    NTSTATUS__STATUS_SECUREBOOT_POLICY_ROLLBACK_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0xD043000C, 'HRESULT_FROM_NT(STATUS_SECUREBOOT_POLICY_ROLLBACK_DETECTED)')
    NTSTATUS__STATUS_SECUREBOOT_POLICY_UPGRADE_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD043000D, 'HRESULT_FROM_NT(STATUS_SECUREBOOT_POLICY_UPGRADE_MISMATCH)')
    NTSTATUS__STATUS_SECUREBOOT_REQUIRED_POLICY_FILE_MISSING: Final[ThirtyTwoBits] = cached_hresult(0xD043000E, 'HRESULT_FROM_NT(STATUS_SECUREBOOT_REQUIRED_POLICY_FILE_MISSING)')
    NTSTATUS__STATUS_SECUREBOOT_NOT_BASE_POLICY: Final[ThirtyTwoBits] = cached_hresult(0xD043000F, 'HRESULT_FROM_NT(STATUS_SECUREBOOT_NOT_BASE_POLICY)')
    NTSTATUS__STATUS_SECUREBOOT_NOT_SUPPLEMENTAL_POLICY: Final[ThirtyTwoBits] = cached_hresult(0xD0430010, 'HRESULT_FROM_NT(STATUS_SECUREBOOT_NOT_SUPPLEMENTAL_POLICY)')
    NTSTATUS__STATUS_AUDIO_ENGINE_NODE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0440001, 'HRESULT_FROM_NT(STATUS_AUDIO_ENGINE_NODE_NOT_FOUND)')
    NTSTATUS__STATUS_HDAUDIO_EMPTY_CONNECTION_LIST: Final[ThirtyTwoBits] = cached_hresult(0xD0440002, 'HRESULT_FROM_NT(STATUS_HDAUDIO_EMPTY_CONNECTION_LIST)')
    NTSTATUS__STATUS_HDAUDIO_CONNECTION_LIST_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD0440003, 'HRESULT_FROM_NT(STATUS_HDAUDIO_CONNECTION_LIST_NOT_SUPPORTED)')
    NTSTATUS__STATUS_HDAUDIO_NO_LOGICAL_DEVICES_CREATED: Final[ThirtyTwoBits] = cached_hresult(0xD0440004, 'HRESULT_FROM_NT(STATUS_HDAUDIO_NO_LOGICAL_DEVICES_CREATED)')
    NTSTATUS__STATUS_HDAUDIO_NULL_LINKED_LIST_ENTRY: Final[ThirtyTwoBits] = cached_hresult(0xD0440005, 'HRESULT_FROM_NT(STATUS_HDAUDIO_NULL_LINKED_LIST_ENTRY)')
    NTSTATUS__STATUS_VSM_NOT_INITIALIZED: Final[ThirtyTwoBits] = cached_hresult(0xD0450000, 'HRESULT_FROM_NT(STATUS_VSM_NOT_INITIALIZED)')
    NTSTATUS__STATUS_VSM_DMA_PROTECTION_NOT_IN_USE: Final[ThirtyTwoBits] = cached_hresult(0xD0450001, 'HRESULT_FROM_NT(STATUS_VSM_DMA_PROTECTION_NOT_IN_USE)')
    NTSTATUS__STATUS_IORING_REQUIRED_FLAG_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD0460001, 'HRESULT_FROM_NT(STATUS_IORING_REQUIRED_FLAG_NOT_SUPPORTED)')
    NTSTATUS__STATUS_IORING_SUBMISSION_QUEUE_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD0460002, 'HRESULT_FROM_NT(STATUS_IORING_SUBMISSION_QUEUE_FULL)')
    NTSTATUS__STATUS_IORING_VERSION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_hresult(0xD0460003, 'HRESULT_FROM_NT(STATUS_IORING_VERSION_NOT_SUPPORTED)')
    NTSTATUS__STATUS_IORING_SUBMISSION_QUEUE_TOO_BIG: Final[ThirtyTwoBits] = cached_hresult(0xD0460004, 'HRESULT_FROM_NT(STATUS_IORING_SUBMISSION_QUEUE_TOO_BIG)')
    NTSTATUS__STATUS_IORING_COMPLETION_QUEUE_TOO_BIG: Final[ThirtyTwoBits] = cached_hresult(0xD0460005, 'HRESULT_FROM_NT(STATUS_IORING_COMPLETION_QUEUE_TOO_BIG)')
    NTSTATUS__STATUS_IORING_SUBMIT_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD0460006, 'HRESULT_FROM_NT(STATUS_IORING_SUBMIT_IN_PROGRESS)')
    NTSTATUS__STATUS_IORING_CORRUPT: Final[ThirtyTwoBits] = cached_hresult(0xD0460007, 'HRESULT_FROM_NT(STATUS_IORING_CORRUPT)')
    NTSTATUS__STATUS_IORING_COMPLETION_QUEUE_TOO_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD0460008, 'HRESULT_FROM_NT(STATUS_IORING_COMPLETION_QUEUE_TOO_FULL)')
    NTSTATUS__STATUS_VOLSNAP_BOOTFILE_NOT_VALID: Final[ThirtyTwoBits] = cached_hresult(0xD0500003, 'HRESULT_FROM_NT(STATUS_VOLSNAP_BOOTFILE_NOT_VALID)')
    NTSTATUS__STATUS_VOLSNAP_ACTIVATION_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0xD0500004, 'HRESULT_FROM_NT(STATUS_VOLSNAP_ACTIVATION_TIMEOUT)')
    NTSTATUS__STATUS_VOLSNAP_NO_BYPASSIO_WITH_SNAPSHOT: Final[ThirtyTwoBits] = cached_hresult(0xD0500005, 'HRESULT_FROM_NT(STATUS_VOLSNAP_NO_BYPASSIO_WITH_SNAPSHOT)')
    NTSTATUS__STATUS_IO_PREEMPTED: Final[ThirtyTwoBits] = cached_hresult(0xD0510001, 'HRESULT_FROM_NT(STATUS_IO_PREEMPTED)')
    NTSTATUS__STATUS_SVHDX_ERROR_STORED: Final[ThirtyTwoBits] = cached_hresult(0xD05C0000, 'HRESULT_FROM_NT(STATUS_SVHDX_ERROR_STORED)')
    NTSTATUS__STATUS_SVHDX_ERROR_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD05CFF00, 'HRESULT_FROM_NT(STATUS_SVHDX_ERROR_NOT_AVAILABLE)')
    NTSTATUS__STATUS_SVHDX_UNIT_ATTENTION_AVAILABLE: Final[ThirtyTwoBits] = cached_hresult(0xD05CFF01, 'HRESULT_FROM_NT(STATUS_SVHDX_UNIT_ATTENTION_AVAILABLE)')
    NTSTATUS__STATUS_SVHDX_UNIT_ATTENTION_CAPACITY_DATA_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0xD05CFF02, 'HRESULT_FROM_NT(STATUS_SVHDX_UNIT_ATTENTION_CAPACITY_DATA_CHANGED)')
    NTSTATUS__STATUS_SVHDX_UNIT_ATTENTION_RESERVATIONS_PREEMPTED: Final[ThirtyTwoBits] = cached_hresult(0xD05CFF03, 'HRESULT_FROM_NT(STATUS_SVHDX_UNIT_ATTENTION_RESERVATIONS_PREEMPTED)')
    NTSTATUS__STATUS_SVHDX_UNIT_ATTENTION_RESERVATIONS_RELEASED: Final[ThirtyTwoBits] = cached_hresult(0xD05CFF04, 'HRESULT_FROM_NT(STATUS_SVHDX_UNIT_ATTENTION_RESERVATIONS_RELEASED)')
    NTSTATUS__STATUS_SVHDX_UNIT_ATTENTION_REGISTRATIONS_PREEMPTED: Final[ThirtyTwoBits] = cached_hresult(0xD05CFF05, 'HRESULT_FROM_NT(STATUS_SVHDX_UNIT_ATTENTION_REGISTRATIONS_PREEMPTED)')
    NTSTATUS__STATUS_SVHDX_UNIT_ATTENTION_OPERATING_DEFINITION_CHANGED: Final[ThirtyTwoBits] = cached_hresult(0xD05CFF06, 'HRESULT_FROM_NT(STATUS_SVHDX_UNIT_ATTENTION_OPERATING_DEFINITION_CHANGED)')
    NTSTATUS__STATUS_SVHDX_RESERVATION_CONFLICT: Final[ThirtyTwoBits] = cached_hresult(0xD05CFF07, 'HRESULT_FROM_NT(STATUS_SVHDX_RESERVATION_CONFLICT)')
    NTSTATUS__STATUS_SVHDX_WRONG_FILE_TYPE: Final[ThirtyTwoBits] = cached_hresult(0xD05CFF08, 'HRESULT_FROM_NT(STATUS_SVHDX_WRONG_FILE_TYPE)')
    NTSTATUS__STATUS_SVHDX_VERSION_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD05CFF09, 'HRESULT_FROM_NT(STATUS_SVHDX_VERSION_MISMATCH)')
    NTSTATUS__STATUS_VHD_SHARED: Final[ThirtyTwoBits] = cached_hresult(0xD05CFF0A, 'HRESULT_FROM_NT(STATUS_VHD_SHARED)')
    NTSTATUS__STATUS_SVHDX_NO_INITIATOR: Final[ThirtyTwoBits] = cached_hresult(0xD05CFF0B, 'HRESULT_FROM_NT(STATUS_SVHDX_NO_INITIATOR)')
    NTSTATUS__STATUS_VHDSET_BACKING_STORAGE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD05CFF0C, 'HRESULT_FROM_NT(STATUS_VHDSET_BACKING_STORAGE_NOT_FOUND)')
    NTSTATUS__STATUS_SMB_NO_PREAUTH_INTEGRITY_HASH_OVERLAP: Final[ThirtyTwoBits] = cached_hresult(0xD05D0000, 'HRESULT_FROM_NT(STATUS_SMB_NO_PREAUTH_INTEGRITY_HASH_OVERLAP)')
    NTSTATUS__STATUS_SMB_BAD_CLUSTER_DIALECT: Final[ThirtyTwoBits] = cached_hresult(0xD05D0001, 'HRESULT_FROM_NT(STATUS_SMB_BAD_CLUSTER_DIALECT)')
    NTSTATUS__STATUS_SMB_GUEST_LOGON_BLOCKED: Final[ThirtyTwoBits] = cached_hresult(0xD05D0002, 'HRESULT_FROM_NT(STATUS_SMB_GUEST_LOGON_BLOCKED)')
    NTSTATUS__STATUS_SMB_NO_SIGNING_ALGORITHM_OVERLAP: Final[ThirtyTwoBits] = cached_hresult(0xD05D0003, 'HRESULT_FROM_NT(STATUS_SMB_NO_SIGNING_ALGORITHM_OVERLAP)')
    NTSTATUS__STATUS_NETWORK_AUTHENTICATION_PROMPT_CANCELED: Final[ThirtyTwoBits] = cached_hresult(0xD05D0004, 'HRESULT_FROM_NT(STATUS_NETWORK_AUTHENTICATION_PROMPT_CANCELED)')
    NTSTATUS__STATUS_SPACES_FAULT_DOMAIN_TYPE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0E70001, 'HRESULT_FROM_NT(STATUS_SPACES_FAULT_DOMAIN_TYPE_INVALID)')
    NTSTATUS__STATUS_SPACES_RESILIENCY_TYPE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0E70003, 'HRESULT_FROM_NT(STATUS_SPACES_RESILIENCY_TYPE_INVALID)')
    NTSTATUS__STATUS_SPACES_DRIVE_SECTOR_SIZE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0E70004, 'HRESULT_FROM_NT(STATUS_SPACES_DRIVE_SECTOR_SIZE_INVALID)')
    NTSTATUS__STATUS_SPACES_DRIVE_REDUNDANCY_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0E70006, 'HRESULT_FROM_NT(STATUS_SPACES_DRIVE_REDUNDANCY_INVALID)')
    NTSTATUS__STATUS_SPACES_NUMBER_OF_DATA_COPIES_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0E70007, 'HRESULT_FROM_NT(STATUS_SPACES_NUMBER_OF_DATA_COPIES_INVALID)')
    NTSTATUS__STATUS_SPACES_INTERLEAVE_LENGTH_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0E70009, 'HRESULT_FROM_NT(STATUS_SPACES_INTERLEAVE_LENGTH_INVALID)')
    NTSTATUS__STATUS_SPACES_NUMBER_OF_COLUMNS_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0E7000A, 'HRESULT_FROM_NT(STATUS_SPACES_NUMBER_OF_COLUMNS_INVALID)')
    NTSTATUS__STATUS_SPACES_NOT_ENOUGH_DRIVES: Final[ThirtyTwoBits] = cached_hresult(0xD0E7000B, 'HRESULT_FROM_NT(STATUS_SPACES_NOT_ENOUGH_DRIVES)')
    NTSTATUS__STATUS_SPACES_EXTENDED_ERROR: Final[ThirtyTwoBits] = cached_hresult(0xD0E7000C, 'HRESULT_FROM_NT(STATUS_SPACES_EXTENDED_ERROR)')
    NTSTATUS__STATUS_SPACES_PROVISIONING_TYPE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0E7000D, 'HRESULT_FROM_NT(STATUS_SPACES_PROVISIONING_TYPE_INVALID)')
    NTSTATUS__STATUS_SPACES_ALLOCATION_SIZE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0E7000E, 'HRESULT_FROM_NT(STATUS_SPACES_ALLOCATION_SIZE_INVALID)')
    NTSTATUS__STATUS_SPACES_ENCLOSURE_AWARE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0E7000F, 'HRESULT_FROM_NT(STATUS_SPACES_ENCLOSURE_AWARE_INVALID)')
    NTSTATUS__STATUS_SPACES_WRITE_CACHE_SIZE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0E70010, 'HRESULT_FROM_NT(STATUS_SPACES_WRITE_CACHE_SIZE_INVALID)')
    NTSTATUS__STATUS_SPACES_NUMBER_OF_GROUPS_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0E70011, 'HRESULT_FROM_NT(STATUS_SPACES_NUMBER_OF_GROUPS_INVALID)')
    NTSTATUS__STATUS_SPACES_DRIVE_OPERATIONAL_STATE_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0E70012, 'HRESULT_FROM_NT(STATUS_SPACES_DRIVE_OPERATIONAL_STATE_INVALID)')
    NTSTATUS__STATUS_SPACES_UPDATE_COLUMN_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD0E70013, 'HRESULT_FROM_NT(STATUS_SPACES_UPDATE_COLUMN_STATE)')
    NTSTATUS__STATUS_SPACES_MAP_REQUIRED: Final[ThirtyTwoBits] = cached_hresult(0xD0E70014, 'HRESULT_FROM_NT(STATUS_SPACES_MAP_REQUIRED)')
    NTSTATUS__STATUS_SPACES_UNSUPPORTED_VERSION: Final[ThirtyTwoBits] = cached_hresult(0xD0E70015, 'HRESULT_FROM_NT(STATUS_SPACES_UNSUPPORTED_VERSION)')
    NTSTATUS__STATUS_SPACES_CORRUPT_METADATA: Final[ThirtyTwoBits] = cached_hresult(0xD0E70016, 'HRESULT_FROM_NT(STATUS_SPACES_CORRUPT_METADATA)')
    NTSTATUS__STATUS_SPACES_DRT_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD0E70017, 'HRESULT_FROM_NT(STATUS_SPACES_DRT_FULL)')
    NTSTATUS__STATUS_SPACES_INCONSISTENCY: Final[ThirtyTwoBits] = cached_hresult(0xD0E70018, 'HRESULT_FROM_NT(STATUS_SPACES_INCONSISTENCY)')
    NTSTATUS__STATUS_SPACES_LOG_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0xD0E70019, 'HRESULT_FROM_NT(STATUS_SPACES_LOG_NOT_READY)')
    NTSTATUS__STATUS_SPACES_NO_REDUNDANCY: Final[ThirtyTwoBits] = cached_hresult(0xD0E7001A, 'HRESULT_FROM_NT(STATUS_SPACES_NO_REDUNDANCY)')
    NTSTATUS__STATUS_SPACES_DRIVE_NOT_READY: Final[ThirtyTwoBits] = cached_hresult(0xD0E7001B, 'HRESULT_FROM_NT(STATUS_SPACES_DRIVE_NOT_READY)')
    NTSTATUS__STATUS_SPACES_DRIVE_SPLIT: Final[ThirtyTwoBits] = cached_hresult(0xD0E7001C, 'HRESULT_FROM_NT(STATUS_SPACES_DRIVE_SPLIT)')
    NTSTATUS__STATUS_SPACES_DRIVE_LOST_DATA: Final[ThirtyTwoBits] = cached_hresult(0xD0E7001D, 'HRESULT_FROM_NT(STATUS_SPACES_DRIVE_LOST_DATA)')
    NTSTATUS__STATUS_SPACES_ENTRY_INCOMPLETE: Final[ThirtyTwoBits] = cached_hresult(0xD0E7001E, 'HRESULT_FROM_NT(STATUS_SPACES_ENTRY_INCOMPLETE)')
    NTSTATUS__STATUS_SPACES_ENTRY_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0E7001F, 'HRESULT_FROM_NT(STATUS_SPACES_ENTRY_INVALID)')
    NTSTATUS__STATUS_SPACES_MARK_DIRTY: Final[ThirtyTwoBits] = cached_hresult(0xD0E70020, 'HRESULT_FROM_NT(STATUS_SPACES_MARK_DIRTY)')
    NTSTATUS__STATUS_SPACES_PD_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0E70021, 'HRESULT_FROM_NT(STATUS_SPACES_PD_NOT_FOUND)')
    NTSTATUS__STATUS_SPACES_PD_LENGTH_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD0E70022, 'HRESULT_FROM_NT(STATUS_SPACES_PD_LENGTH_MISMATCH)')
    NTSTATUS__STATUS_SPACES_PD_UNSUPPORTED_VERSION: Final[ThirtyTwoBits] = cached_hresult(0xD0E70023, 'HRESULT_FROM_NT(STATUS_SPACES_PD_UNSUPPORTED_VERSION)')
    NTSTATUS__STATUS_SPACES_PD_INVALID_DATA: Final[ThirtyTwoBits] = cached_hresult(0xD0E70024, 'HRESULT_FROM_NT(STATUS_SPACES_PD_INVALID_DATA)')
    NTSTATUS__STATUS_SPACES_FLUSH_METADATA: Final[ThirtyTwoBits] = cached_hresult(0xD0E70025, 'HRESULT_FROM_NT(STATUS_SPACES_FLUSH_METADATA)')
    NTSTATUS__STATUS_SPACES_CACHE_FULL: Final[ThirtyTwoBits] = cached_hresult(0xD0E70026, 'HRESULT_FROM_NT(STATUS_SPACES_CACHE_FULL)')
    NTSTATUS__STATUS_SPACES_REPAIR_IN_PROGRESS: Final[ThirtyTwoBits] = cached_hresult(0xD0E70027, 'HRESULT_FROM_NT(STATUS_SPACES_REPAIR_IN_PROGRESS)')
    NTSTATUS__STATUS_SECCORE_INVALID_COMMAND: Final[ThirtyTwoBits] = cached_hresult(0xD0E80000, 'HRESULT_FROM_NT(STATUS_SECCORE_INVALID_COMMAND)')
    NTSTATUS__STATUS_SYSTEM_INTEGRITY_ROLLBACK_DETECTED: Final[ThirtyTwoBits] = cached_hresult(0xD0E90001, 'HRESULT_FROM_NT(STATUS_SYSTEM_INTEGRITY_ROLLBACK_DETECTED)')
    NTSTATUS__STATUS_SYSTEM_INTEGRITY_POLICY_VIOLATION: Final[ThirtyTwoBits] = cached_hresult(0xD0E90002, 'HRESULT_FROM_NT(STATUS_SYSTEM_INTEGRITY_POLICY_VIOLATION)')
    NTSTATUS__STATUS_SYSTEM_INTEGRITY_INVALID_POLICY: Final[ThirtyTwoBits] = cached_hresult(0xD0E90003, 'HRESULT_FROM_NT(STATUS_SYSTEM_INTEGRITY_INVALID_POLICY)')
    NTSTATUS__STATUS_SYSTEM_INTEGRITY_POLICY_NOT_SIGNED: Final[ThirtyTwoBits] = cached_hresult(0xD0E90004, 'HRESULT_FROM_NT(STATUS_SYSTEM_INTEGRITY_POLICY_NOT_SIGNED)')
    NTSTATUS__STATUS_SYSTEM_INTEGRITY_TOO_MANY_POLICIES: Final[ThirtyTwoBits] = cached_hresult(0xD0E90005, 'HRESULT_FROM_NT(STATUS_SYSTEM_INTEGRITY_TOO_MANY_POLICIES)')
    NTSTATUS__STATUS_SYSTEM_INTEGRITY_SUPPLEMENTAL_POLICY_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_hresult(0xD0E90006, 'HRESULT_FROM_NT(STATUS_SYSTEM_INTEGRITY_SUPPLEMENTAL_POLICY_NOT_AUTHORIZED)')
    NTSTATUS__STATUS_SYSTEM_INTEGRITY_REPUTATION_MALICIOUS: Final[ThirtyTwoBits] = cached_hresult(0xD0E90007, 'HRESULT_FROM_NT(STATUS_SYSTEM_INTEGRITY_REPUTATION_MALICIOUS)')
    NTSTATUS__STATUS_SYSTEM_INTEGRITY_REPUTATION_PUA: Final[ThirtyTwoBits] = cached_hresult(0xD0E90008, 'HRESULT_FROM_NT(STATUS_SYSTEM_INTEGRITY_REPUTATION_PUA)')
    NTSTATUS__STATUS_SYSTEM_INTEGRITY_REPUTATION_DANGEROUS_EXT: Final[ThirtyTwoBits] = cached_hresult(0xD0E90009, 'HRESULT_FROM_NT(STATUS_SYSTEM_INTEGRITY_REPUTATION_DANGEROUS_EXT)')
    NTSTATUS__STATUS_SYSTEM_INTEGRITY_REPUTATION_OFFLINE: Final[ThirtyTwoBits] = cached_hresult(0xD0E9000A, 'HRESULT_FROM_NT(STATUS_SYSTEM_INTEGRITY_REPUTATION_OFFLINE)')
    NTSTATUS__STATUS_SYSTEM_INTEGRITY_REPUTATION_UNFRIENDLY_FILE: Final[ThirtyTwoBits] = cached_hresult(0xD0E9000B, 'HRESULT_FROM_NT(STATUS_SYSTEM_INTEGRITY_REPUTATION_UNFRIENDLY_FILE)')
    NTSTATUS__STATUS_SYSTEM_INTEGRITY_REPUTATION_UNATTAINABLE: Final[ThirtyTwoBits] = cached_hresult(0xD0E9000C, 'HRESULT_FROM_NT(STATUS_SYSTEM_INTEGRITY_REPUTATION_UNATTAINABLE)')
    NTSTATUS__STATUS_SYSTEM_INTEGRITY_REPUTATION_EXPLICIT_DENY_FILE: Final[ThirtyTwoBits] = cached_hresult(0xD0E9000D, 'HRESULT_FROM_NT(STATUS_SYSTEM_INTEGRITY_REPUTATION_EXPLICIT_DENY_FILE)')
    NTSTATUS__STATUS_NO_APPLICABLE_APP_LICENSES_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0EA0001, 'HRESULT_FROM_NT(STATUS_NO_APPLICABLE_APP_LICENSES_FOUND)')
    NTSTATUS__STATUS_CLIP_LICENSE_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0EA0002, 'HRESULT_FROM_NT(STATUS_CLIP_LICENSE_NOT_FOUND)')
    NTSTATUS__STATUS_CLIP_DEVICE_LICENSE_MISSING: Final[ThirtyTwoBits] = cached_hresult(0xD0EA0003, 'HRESULT_FROM_NT(STATUS_CLIP_DEVICE_LICENSE_MISSING)')
    NTSTATUS__STATUS_CLIP_LICENSE_INVALID_SIGNATURE: Final[ThirtyTwoBits] = cached_hresult(0xD0EA0004, 'HRESULT_FROM_NT(STATUS_CLIP_LICENSE_INVALID_SIGNATURE)')
    NTSTATUS__STATUS_CLIP_KEYHOLDER_LICENSE_MISSING_OR_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0EA0005, 'HRESULT_FROM_NT(STATUS_CLIP_KEYHOLDER_LICENSE_MISSING_OR_INVALID)')
    NTSTATUS__STATUS_CLIP_LICENSE_EXPIRED: Final[ThirtyTwoBits] = cached_hresult(0xD0EA0006, 'HRESULT_FROM_NT(STATUS_CLIP_LICENSE_EXPIRED)')
    NTSTATUS__STATUS_CLIP_LICENSE_SIGNED_BY_UNKNOWN_SOURCE: Final[ThirtyTwoBits] = cached_hresult(0xD0EA0007, 'HRESULT_FROM_NT(STATUS_CLIP_LICENSE_SIGNED_BY_UNKNOWN_SOURCE)')
    NTSTATUS__STATUS_CLIP_LICENSE_NOT_SIGNED: Final[ThirtyTwoBits] = cached_hresult(0xD0EA0008, 'HRESULT_FROM_NT(STATUS_CLIP_LICENSE_NOT_SIGNED)')
    NTSTATUS__STATUS_CLIP_LICENSE_HARDWARE_ID_OUT_OF_TOLERANCE: Final[ThirtyTwoBits] = cached_hresult(0xD0EA0009, 'HRESULT_FROM_NT(STATUS_CLIP_LICENSE_HARDWARE_ID_OUT_OF_TOLERANCE)')
    NTSTATUS__STATUS_CLIP_LICENSE_DEVICE_ID_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD0EA000A, 'HRESULT_FROM_NT(STATUS_CLIP_LICENSE_DEVICE_ID_MISMATCH)')
    NTSTATUS__STATUS_PLATFORM_MANIFEST_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_hresult(0xD0EB0001, 'HRESULT_FROM_NT(STATUS_PLATFORM_MANIFEST_NOT_AUTHORIZED)')
    NTSTATUS__STATUS_PLATFORM_MANIFEST_INVALID: Final[ThirtyTwoBits] = cached_hresult(0xD0EB0002, 'HRESULT_FROM_NT(STATUS_PLATFORM_MANIFEST_INVALID)')
    NTSTATUS__STATUS_PLATFORM_MANIFEST_FILE_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_hresult(0xD0EB0003, 'HRESULT_FROM_NT(STATUS_PLATFORM_MANIFEST_FILE_NOT_AUTHORIZED)')
    NTSTATUS__STATUS_PLATFORM_MANIFEST_CATALOG_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_hresult(0xD0EB0004, 'HRESULT_FROM_NT(STATUS_PLATFORM_MANIFEST_CATALOG_NOT_AUTHORIZED)')
    NTSTATUS__STATUS_PLATFORM_MANIFEST_BINARY_ID_NOT_FOUND: Final[ThirtyTwoBits] = cached_hresult(0xD0EB0005, 'HRESULT_FROM_NT(STATUS_PLATFORM_MANIFEST_BINARY_ID_NOT_FOUND)')
    NTSTATUS__STATUS_PLATFORM_MANIFEST_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_hresult(0xD0EB0006, 'HRESULT_FROM_NT(STATUS_PLATFORM_MANIFEST_NOT_ACTIVE)')
    NTSTATUS__STATUS_PLATFORM_MANIFEST_NOT_SIGNED: Final[ThirtyTwoBits] = cached_hresult(0xD0EB0007, 'HRESULT_FROM_NT(STATUS_PLATFORM_MANIFEST_NOT_SIGNED)')
    NTSTATUS__STATUS_APPEXEC_CONDITION_NOT_SATISFIED: Final[ThirtyTwoBits] = cached_hresult(0xD0EC0000, 'HRESULT_FROM_NT(STATUS_APPEXEC_CONDITION_NOT_SATISFIED)')
    NTSTATUS__STATUS_APPEXEC_HANDLE_INVALIDATED: Final[ThirtyTwoBits] = cached_hresult(0xD0EC0001, 'HRESULT_FROM_NT(STATUS_APPEXEC_HANDLE_INVALIDATED)')
    NTSTATUS__STATUS_APPEXEC_INVALID_HOST_GENERATION: Final[ThirtyTwoBits] = cached_hresult(0xD0EC0002, 'HRESULT_FROM_NT(STATUS_APPEXEC_INVALID_HOST_GENERATION)')
    NTSTATUS__STATUS_APPEXEC_UNEXPECTED_PROCESS_REGISTRATION: Final[ThirtyTwoBits] = cached_hresult(0xD0EC0003, 'HRESULT_FROM_NT(STATUS_APPEXEC_UNEXPECTED_PROCESS_REGISTRATION)')
    NTSTATUS__STATUS_APPEXEC_INVALID_HOST_STATE: Final[ThirtyTwoBits] = cached_hresult(0xD0EC0004, 'HRESULT_FROM_NT(STATUS_APPEXEC_INVALID_HOST_STATE)')
    NTSTATUS__STATUS_APPEXEC_NO_DONOR: Final[ThirtyTwoBits] = cached_hresult(0xD0EC0005, 'HRESULT_FROM_NT(STATUS_APPEXEC_NO_DONOR)')
    NTSTATUS__STATUS_APPEXEC_HOST_ID_MISMATCH: Final[ThirtyTwoBits] = cached_hresult(0xD0EC0006, 'HRESULT_FROM_NT(STATUS_APPEXEC_HOST_ID_MISMATCH)')
    NTSTATUS__STATUS_APPEXEC_UNKNOWN_USER: Final[ThirtyTwoBits] = cached_hresult(0xD0EC0007, 'HRESULT_FROM_NT(STATUS_APPEXEC_UNKNOWN_USER)')
    NTSTATUS__STATUS_APPEXEC_APP_COMPAT_BLOCK: Final[ThirtyTwoBits] = cached_hresult(0xD0EC0008, 'HRESULT_FROM_NT(STATUS_APPEXEC_APP_COMPAT_BLOCK)')
    NTSTATUS__STATUS_APPEXEC_CALLER_WAIT_TIMEOUT: Final[ThirtyTwoBits] = cached_hresult(0xD0EC0009, 'HRESULT_FROM_NT(STATUS_APPEXEC_CALLER_WAIT_TIMEOUT)')
    NTSTATUS__STATUS_APPEXEC_CALLER_WAIT_TIMEOUT_TERMINATION: Final[ThirtyTwoBits] = cached_hresult(0xD0EC000A, 'HRESULT_FROM_NT(STATUS_APPEXEC_CALLER_WAIT_TIMEOUT_TERMINATION)')
    NTSTATUS__STATUS_APPEXEC_CALLER_WAIT_TIMEOUT_LICENSING: Final[ThirtyTwoBits] = cached_hresult(0xD0EC000B, 'HRESULT_FROM_NT(STATUS_APPEXEC_CALLER_WAIT_TIMEOUT_LICENSING)')
    NTSTATUS__STATUS_APPEXEC_CALLER_WAIT_TIMEOUT_RESOURCES: Final[ThirtyTwoBits] = cached_hresult(0xD0EC000C, 'HRESULT_FROM_NT(STATUS_APPEXEC_CALLER_WAIT_TIMEOUT_RESOURCES)')
