__all__ = [
    'NtStatus',
    'NtStatusSeverity',
]

from enum import Enum
from typing import Final, Optional

from ntstatus._thirtytwobits import ThirtyTwoBits

NTSTATUS_LOOKUP: dict[int, ThirtyTwoBits] = dict()

def cached_ntstatus(code: int, name: str) -> ThirtyTwoBits:
    if code in NTSTATUS_LOOKUP:
        return NTSTATUS_LOOKUP[code]

    bits = ThirtyTwoBits(code, name)
    NTSTATUS_LOOKUP[bits.signed_value] = bits
    NTSTATUS_LOOKUP[bits.unsigned_value] = bits
    return bits

class NtStatusSeverity(Enum):
    """
    Enumeration representing WinAPI NTSTATUS severity values. Values can be obtained using the 'severity' property of NtStatus values.

    See explanation of severity values here: https://learn.microsoft.com/en-us/windows-hardware/drivers/kernel/using-ntstatus-values
    See how to calculate the severity from a numeric NTSTATUS value here: https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-erref/87fba13e-bf06-450e-83b1-9241dc81e781
    """

    STATUS_SEVERITY_SUCCESS = 0x0
    STATUS_SEVERITY_INFORMATIONAL = 0x1
    STATUS_SEVERITY_WARNING = 0x2
    STATUS_SEVERITY_ERROR = 0x3

class NtStatus:
    """
    This class defines all named [WinAPI NTSTATUS values](https://learn.microsoft.com/en-us/windows-hardware/drivers/kernel/using-ntstatus-values) as attributes.

    Constants are represented using the `ntstatus.ThirtyTwoBits` class which can be equality-compared to integers; the comparision will return True for both the signed and unsigned interpretation of the underlying 32 bits.

    Static methods:
    - `NtStatus.decode(code)`: Looks up an NTSTATUS constant by its numeric value and returns it as an `ntstatus.ThirtyTwoBits` object, with its name set to the the name of the constant. Returns `None` if argument is not a known NTSTATUS constant.
    - `NtStatus.decode_name(code, default='')`: Returns the name of the NTSTATUS value associated with the first argument if there is such a value. Otherwise, returns the second argument.
    - `NtStatus.make(code)`: Looks up an NTSTATUS constant by its numeric value and returns it as an `ntstatus.ThirtyTwoBits` object, with its name set to the the name of the constant. Raises `ValueError` if argument is not a known NTSTATUS value.
    - `NtStatus.severity(status)`: Returns the severity of an NTSTATUS value represented by an `ntstatus.ThirtyTwoBits` object as an `ntstatus.NtStatusSeverity`.

    The complete list of NTSTATUS values can be found [here](https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-erref/596a1078-e883-4972-9bbc-49e60bebca55). `NtStatus` encodes most of them, with the exception of:

    - `STATUS_WAIT_n`, where `n` is 0, 1, 2, 3, and 63
    - `STATUS_ABANDONED_WAIT_n`, where `n` is 0 and 63
    - `STATUS_FLT_DISALLOW_FSFILTER_IO`
    """

    @staticmethod
    def decode(code: int) -> Optional[ThirtyTwoBits]:
        """
        Looks up an NTSTATUS constant by its numeric value and returns it as an `ntstatus.ThirtyTwoBits` object, with its name set to the the name of the constant. Returns `None` if argument is not a known NTSTATUS constant.
        """

        ThirtyTwoBits.check(code)

        return NTSTATUS_LOOKUP.get(code, None)

    @staticmethod
    def decode_name(code: int, default: str = '') -> str:
        """
        Returns the name of the NTSTATUS value associated with the first argument if there is such a value. Otherwise, returns the second argument.
        """

        status = NtStatus.decode(code)
        if status is not None:
            return status.name
        else:
            return default

    @staticmethod
    def make(code: int) -> ThirtyTwoBits:
        """
        Looks up an NTSTATUS constant by its numeric value and returns it as an `ntstatus.ThirtyTwoBits` object, with its name set to the the name of the constant. Raises ValueError if argument is not a known NTSTATUS value.
        """

        status = NtStatus.decode(code)
        if status is not None:
            return status
        else:
            raise ValueError(f'Unknown NTSTATUS value: {code}')

    @staticmethod
    def severity(status: ThirtyTwoBits) -> NtStatusSeverity:
        """
        Returns the severity of an NTSTATUS value represented by an `ntstatus.ThirtyTwoBits` object as an `ntstatus.NtStatusSeverity`.
        """

        return NtStatusSeverity((status.unsigned_value & 0xC0000000) >> 30)

    # AUTOGENERATED FROM HERE, DO NOT TOUCH

    STATUS_SUCCESS: Final[ThirtyTwoBits] = cached_ntstatus(0x00000000, 'STATUS_SUCCESS')
    STATUS_ABANDONED: Final[ThirtyTwoBits] = cached_ntstatus(0x00000080, 'STATUS_ABANDONED')
    STATUS_USER_APC: Final[ThirtyTwoBits] = cached_ntstatus(0x000000C0, 'STATUS_USER_APC')
    STATUS_ALREADY_COMPLETE: Final[ThirtyTwoBits] = cached_ntstatus(0x000000FF, 'STATUS_ALREADY_COMPLETE')
    STATUS_KERNEL_APC: Final[ThirtyTwoBits] = cached_ntstatus(0x00000100, 'STATUS_KERNEL_APC')
    STATUS_ALERTED: Final[ThirtyTwoBits] = cached_ntstatus(0x00000101, 'STATUS_ALERTED')
    STATUS_TIMEOUT: Final[ThirtyTwoBits] = cached_ntstatus(0x00000102, 'STATUS_TIMEOUT')
    STATUS_PENDING: Final[ThirtyTwoBits] = cached_ntstatus(0x00000103, 'STATUS_PENDING')
    STATUS_REPARSE: Final[ThirtyTwoBits] = cached_ntstatus(0x00000104, 'STATUS_REPARSE')
    STATUS_MORE_ENTRIES: Final[ThirtyTwoBits] = cached_ntstatus(0x00000105, 'STATUS_MORE_ENTRIES')
    STATUS_NOT_ALL_ASSIGNED: Final[ThirtyTwoBits] = cached_ntstatus(0x00000106, 'STATUS_NOT_ALL_ASSIGNED')
    STATUS_SOME_NOT_MAPPED: Final[ThirtyTwoBits] = cached_ntstatus(0x00000107, 'STATUS_SOME_NOT_MAPPED')
    STATUS_OPLOCK_BREAK_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0x00000108, 'STATUS_OPLOCK_BREAK_IN_PROGRESS')
    STATUS_VOLUME_MOUNTED: Final[ThirtyTwoBits] = cached_ntstatus(0x00000109, 'STATUS_VOLUME_MOUNTED')
    STATUS_RXACT_COMMITTED: Final[ThirtyTwoBits] = cached_ntstatus(0x0000010A, 'STATUS_RXACT_COMMITTED')
    STATUS_NOTIFY_CLEANUP: Final[ThirtyTwoBits] = cached_ntstatus(0x0000010B, 'STATUS_NOTIFY_CLEANUP')
    STATUS_NOTIFY_ENUM_DIR: Final[ThirtyTwoBits] = cached_ntstatus(0x0000010C, 'STATUS_NOTIFY_ENUM_DIR')
    STATUS_NO_QUOTAS_FOR_ACCOUNT: Final[ThirtyTwoBits] = cached_ntstatus(0x0000010D, 'STATUS_NO_QUOTAS_FOR_ACCOUNT')
    STATUS_PRIMARY_TRANSPORT_CONNECT_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0x0000010E, 'STATUS_PRIMARY_TRANSPORT_CONNECT_FAILED')
    STATUS_PAGE_FAULT_TRANSITION: Final[ThirtyTwoBits] = cached_ntstatus(0x00000110, 'STATUS_PAGE_FAULT_TRANSITION')
    STATUS_PAGE_FAULT_DEMAND_ZERO: Final[ThirtyTwoBits] = cached_ntstatus(0x00000111, 'STATUS_PAGE_FAULT_DEMAND_ZERO')
    STATUS_PAGE_FAULT_COPY_ON_WRITE: Final[ThirtyTwoBits] = cached_ntstatus(0x00000112, 'STATUS_PAGE_FAULT_COPY_ON_WRITE')
    STATUS_PAGE_FAULT_GUARD_PAGE: Final[ThirtyTwoBits] = cached_ntstatus(0x00000113, 'STATUS_PAGE_FAULT_GUARD_PAGE')
    STATUS_PAGE_FAULT_PAGING_FILE: Final[ThirtyTwoBits] = cached_ntstatus(0x00000114, 'STATUS_PAGE_FAULT_PAGING_FILE')
    STATUS_CACHE_PAGE_LOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0x00000115, 'STATUS_CACHE_PAGE_LOCKED')
    STATUS_CRASH_DUMP: Final[ThirtyTwoBits] = cached_ntstatus(0x00000116, 'STATUS_CRASH_DUMP')
    STATUS_BUFFER_ALL_ZEROS: Final[ThirtyTwoBits] = cached_ntstatus(0x00000117, 'STATUS_BUFFER_ALL_ZEROS')
    STATUS_REPARSE_OBJECT: Final[ThirtyTwoBits] = cached_ntstatus(0x00000118, 'STATUS_REPARSE_OBJECT')
    STATUS_RESOURCE_REQUIREMENTS_CHANGED: Final[ThirtyTwoBits] = cached_ntstatus(0x00000119, 'STATUS_RESOURCE_REQUIREMENTS_CHANGED')
    STATUS_TRANSLATION_COMPLETE: Final[ThirtyTwoBits] = cached_ntstatus(0x00000120, 'STATUS_TRANSLATION_COMPLETE')
    STATUS_DS_MEMBERSHIP_EVALUATED_LOCALLY: Final[ThirtyTwoBits] = cached_ntstatus(0x00000121, 'STATUS_DS_MEMBERSHIP_EVALUATED_LOCALLY')
    STATUS_NOTHING_TO_TERMINATE: Final[ThirtyTwoBits] = cached_ntstatus(0x00000122, 'STATUS_NOTHING_TO_TERMINATE')
    STATUS_PROCESS_NOT_IN_JOB: Final[ThirtyTwoBits] = cached_ntstatus(0x00000123, 'STATUS_PROCESS_NOT_IN_JOB')
    STATUS_PROCESS_IN_JOB: Final[ThirtyTwoBits] = cached_ntstatus(0x00000124, 'STATUS_PROCESS_IN_JOB')
    STATUS_VOLSNAP_HIBERNATE_READY: Final[ThirtyTwoBits] = cached_ntstatus(0x00000125, 'STATUS_VOLSNAP_HIBERNATE_READY')
    STATUS_FSFILTER_OP_COMPLETED_SUCCESSFULLY: Final[ThirtyTwoBits] = cached_ntstatus(0x00000126, 'STATUS_FSFILTER_OP_COMPLETED_SUCCESSFULLY')
    STATUS_INTERRUPT_VECTOR_ALREADY_CONNECTED: Final[ThirtyTwoBits] = cached_ntstatus(0x00000127, 'STATUS_INTERRUPT_VECTOR_ALREADY_CONNECTED')
    STATUS_INTERRUPT_STILL_CONNECTED: Final[ThirtyTwoBits] = cached_ntstatus(0x00000128, 'STATUS_INTERRUPT_STILL_CONNECTED')
    STATUS_PROCESS_CLONED: Final[ThirtyTwoBits] = cached_ntstatus(0x00000129, 'STATUS_PROCESS_CLONED')
    STATUS_FILE_LOCKED_WITH_ONLY_READERS: Final[ThirtyTwoBits] = cached_ntstatus(0x0000012A, 'STATUS_FILE_LOCKED_WITH_ONLY_READERS')
    STATUS_FILE_LOCKED_WITH_WRITERS: Final[ThirtyTwoBits] = cached_ntstatus(0x0000012B, 'STATUS_FILE_LOCKED_WITH_WRITERS')
    STATUS_VALID_IMAGE_HASH: Final[ThirtyTwoBits] = cached_ntstatus(0x0000012C, 'STATUS_VALID_IMAGE_HASH')
    STATUS_VALID_CATALOG_HASH: Final[ThirtyTwoBits] = cached_ntstatus(0x0000012D, 'STATUS_VALID_CATALOG_HASH')
    STATUS_VALID_STRONG_CODE_HASH: Final[ThirtyTwoBits] = cached_ntstatus(0x0000012E, 'STATUS_VALID_STRONG_CODE_HASH')
    STATUS_GHOSTED: Final[ThirtyTwoBits] = cached_ntstatus(0x0000012F, 'STATUS_GHOSTED')
    STATUS_DATA_OVERWRITTEN: Final[ThirtyTwoBits] = cached_ntstatus(0x00000130, 'STATUS_DATA_OVERWRITTEN')
    STATUS_RESOURCEMANAGER_READ_ONLY: Final[ThirtyTwoBits] = cached_ntstatus(0x00000202, 'STATUS_RESOURCEMANAGER_READ_ONLY')
    STATUS_RING_PREVIOUSLY_EMPTY: Final[ThirtyTwoBits] = cached_ntstatus(0x00000210, 'STATUS_RING_PREVIOUSLY_EMPTY')
    STATUS_RING_PREVIOUSLY_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0x00000211, 'STATUS_RING_PREVIOUSLY_FULL')
    STATUS_RING_PREVIOUSLY_ABOVE_QUOTA: Final[ThirtyTwoBits] = cached_ntstatus(0x00000212, 'STATUS_RING_PREVIOUSLY_ABOVE_QUOTA')
    STATUS_RING_NEWLY_EMPTY: Final[ThirtyTwoBits] = cached_ntstatus(0x00000213, 'STATUS_RING_NEWLY_EMPTY')
    STATUS_RING_SIGNAL_OPPOSITE_ENDPOINT: Final[ThirtyTwoBits] = cached_ntstatus(0x00000214, 'STATUS_RING_SIGNAL_OPPOSITE_ENDPOINT')
    STATUS_OPLOCK_SWITCHED_TO_NEW_HANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0x00000215, 'STATUS_OPLOCK_SWITCHED_TO_NEW_HANDLE')
    STATUS_OPLOCK_HANDLE_CLOSED: Final[ThirtyTwoBits] = cached_ntstatus(0x00000216, 'STATUS_OPLOCK_HANDLE_CLOSED')
    STATUS_WAIT_FOR_OPLOCK: Final[ThirtyTwoBits] = cached_ntstatus(0x00000367, 'STATUS_WAIT_FOR_OPLOCK')
    STATUS_REPARSE_GLOBAL: Final[ThirtyTwoBits] = cached_ntstatus(0x00000368, 'STATUS_REPARSE_GLOBAL')
    STATUS_PAGE_FAULT_RETRY: Final[ThirtyTwoBits] = cached_ntstatus(0x00000369, 'STATUS_PAGE_FAULT_RETRY')
    STATUS_FLT_IO_COMPLETE: Final[ThirtyTwoBits] = cached_ntstatus(0x001C0001, 'STATUS_FLT_IO_COMPLETE')
    STATUS_RTPM_CONTEXT_CONTINUE: Final[ThirtyTwoBits] = cached_ntstatus(0x00293000, 'STATUS_RTPM_CONTEXT_CONTINUE')
    STATUS_RTPM_CONTEXT_COMPLETE: Final[ThirtyTwoBits] = cached_ntstatus(0x00293001, 'STATUS_RTPM_CONTEXT_COMPLETE')
    STATUS_HV_PENDING_PAGE_REQUESTS: Final[ThirtyTwoBits] = cached_ntstatus(0x00350059, 'STATUS_HV_PENDING_PAGE_REQUESTS')
    STATUS_SPACES_REPAIRED: Final[ThirtyTwoBits] = cached_ntstatus(0x00E70000, 'STATUS_SPACES_REPAIRED')
    STATUS_SPACES_PAUSE: Final[ThirtyTwoBits] = cached_ntstatus(0x00E70001, 'STATUS_SPACES_PAUSE')
    STATUS_SPACES_COMPLETE: Final[ThirtyTwoBits] = cached_ntstatus(0x00E70002, 'STATUS_SPACES_COMPLETE')
    STATUS_SPACES_REDIRECT: Final[ThirtyTwoBits] = cached_ntstatus(0x00E70003, 'STATUS_SPACES_REDIRECT')
    STATUS_OBJECT_NAME_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0x40000000, 'STATUS_OBJECT_NAME_EXISTS')
    STATUS_THREAD_WAS_SUSPENDED: Final[ThirtyTwoBits] = cached_ntstatus(0x40000001, 'STATUS_THREAD_WAS_SUSPENDED')
    STATUS_WORKING_SET_LIMIT_RANGE: Final[ThirtyTwoBits] = cached_ntstatus(0x40000002, 'STATUS_WORKING_SET_LIMIT_RANGE')
    STATUS_IMAGE_NOT_AT_BASE: Final[ThirtyTwoBits] = cached_ntstatus(0x40000003, 'STATUS_IMAGE_NOT_AT_BASE')
    STATUS_RXACT_STATE_CREATED: Final[ThirtyTwoBits] = cached_ntstatus(0x40000004, 'STATUS_RXACT_STATE_CREATED')
    STATUS_SEGMENT_NOTIFICATION: Final[ThirtyTwoBits] = cached_ntstatus(0x40000005, 'STATUS_SEGMENT_NOTIFICATION')
    STATUS_LOCAL_USER_SESSION_KEY: Final[ThirtyTwoBits] = cached_ntstatus(0x40000006, 'STATUS_LOCAL_USER_SESSION_KEY')
    STATUS_BAD_CURRENT_DIRECTORY: Final[ThirtyTwoBits] = cached_ntstatus(0x40000007, 'STATUS_BAD_CURRENT_DIRECTORY')
    STATUS_SERIAL_MORE_WRITES: Final[ThirtyTwoBits] = cached_ntstatus(0x40000008, 'STATUS_SERIAL_MORE_WRITES')
    STATUS_REGISTRY_RECOVERED: Final[ThirtyTwoBits] = cached_ntstatus(0x40000009, 'STATUS_REGISTRY_RECOVERED')
    STATUS_FT_READ_RECOVERY_FROM_BACKUP: Final[ThirtyTwoBits] = cached_ntstatus(0x4000000A, 'STATUS_FT_READ_RECOVERY_FROM_BACKUP')
    STATUS_FT_WRITE_RECOVERY: Final[ThirtyTwoBits] = cached_ntstatus(0x4000000B, 'STATUS_FT_WRITE_RECOVERY')
    STATUS_SERIAL_COUNTER_TIMEOUT: Final[ThirtyTwoBits] = cached_ntstatus(0x4000000C, 'STATUS_SERIAL_COUNTER_TIMEOUT')
    STATUS_NULL_LM_PASSWORD: Final[ThirtyTwoBits] = cached_ntstatus(0x4000000D, 'STATUS_NULL_LM_PASSWORD')
    STATUS_IMAGE_MACHINE_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0x4000000E, 'STATUS_IMAGE_MACHINE_TYPE_MISMATCH')
    STATUS_RECEIVE_PARTIAL: Final[ThirtyTwoBits] = cached_ntstatus(0x4000000F, 'STATUS_RECEIVE_PARTIAL')
    STATUS_RECEIVE_EXPEDITED: Final[ThirtyTwoBits] = cached_ntstatus(0x40000010, 'STATUS_RECEIVE_EXPEDITED')
    STATUS_RECEIVE_PARTIAL_EXPEDITED: Final[ThirtyTwoBits] = cached_ntstatus(0x40000011, 'STATUS_RECEIVE_PARTIAL_EXPEDITED')
    STATUS_EVENT_DONE: Final[ThirtyTwoBits] = cached_ntstatus(0x40000012, 'STATUS_EVENT_DONE')
    STATUS_EVENT_PENDING: Final[ThirtyTwoBits] = cached_ntstatus(0x40000013, 'STATUS_EVENT_PENDING')
    STATUS_CHECKING_FILE_SYSTEM: Final[ThirtyTwoBits] = cached_ntstatus(0x40000014, 'STATUS_CHECKING_FILE_SYSTEM')
    STATUS_FATAL_APP_EXIT: Final[ThirtyTwoBits] = cached_ntstatus(0x40000015, 'STATUS_FATAL_APP_EXIT')
    STATUS_PREDEFINED_HANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0x40000016, 'STATUS_PREDEFINED_HANDLE')
    STATUS_WAS_UNLOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0x40000017, 'STATUS_WAS_UNLOCKED')
    STATUS_SERVICE_NOTIFICATION: Final[ThirtyTwoBits] = cached_ntstatus(0x40000018, 'STATUS_SERVICE_NOTIFICATION')
    STATUS_WAS_LOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0x40000019, 'STATUS_WAS_LOCKED')
    STATUS_LOG_HARD_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0x4000001A, 'STATUS_LOG_HARD_ERROR')
    STATUS_ALREADY_WIN32: Final[ThirtyTwoBits] = cached_ntstatus(0x4000001B, 'STATUS_ALREADY_WIN32')
    STATUS_WX86_UNSIMULATE: Final[ThirtyTwoBits] = cached_ntstatus(0x4000001C, 'STATUS_WX86_UNSIMULATE')
    STATUS_WX86_CONTINUE: Final[ThirtyTwoBits] = cached_ntstatus(0x4000001D, 'STATUS_WX86_CONTINUE')
    STATUS_WX86_SINGLE_STEP: Final[ThirtyTwoBits] = cached_ntstatus(0x4000001E, 'STATUS_WX86_SINGLE_STEP')
    STATUS_WX86_BREAKPOINT: Final[ThirtyTwoBits] = cached_ntstatus(0x4000001F, 'STATUS_WX86_BREAKPOINT')
    STATUS_WX86_EXCEPTION_CONTINUE: Final[ThirtyTwoBits] = cached_ntstatus(0x40000020, 'STATUS_WX86_EXCEPTION_CONTINUE')
    STATUS_WX86_EXCEPTION_LASTCHANCE: Final[ThirtyTwoBits] = cached_ntstatus(0x40000021, 'STATUS_WX86_EXCEPTION_LASTCHANCE')
    STATUS_WX86_EXCEPTION_CHAIN: Final[ThirtyTwoBits] = cached_ntstatus(0x40000022, 'STATUS_WX86_EXCEPTION_CHAIN')
    STATUS_IMAGE_MACHINE_TYPE_MISMATCH_EXE: Final[ThirtyTwoBits] = cached_ntstatus(0x40000023, 'STATUS_IMAGE_MACHINE_TYPE_MISMATCH_EXE')
    STATUS_NO_YIELD_PERFORMED: Final[ThirtyTwoBits] = cached_ntstatus(0x40000024, 'STATUS_NO_YIELD_PERFORMED')
    STATUS_TIMER_RESUME_IGNORED: Final[ThirtyTwoBits] = cached_ntstatus(0x40000025, 'STATUS_TIMER_RESUME_IGNORED')
    STATUS_ARBITRATION_UNHANDLED: Final[ThirtyTwoBits] = cached_ntstatus(0x40000026, 'STATUS_ARBITRATION_UNHANDLED')
    STATUS_CARDBUS_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0x40000027, 'STATUS_CARDBUS_NOT_SUPPORTED')
    STATUS_WX86_CREATEWX86TIB: Final[ThirtyTwoBits] = cached_ntstatus(0x40000028, 'STATUS_WX86_CREATEWX86TIB')
    STATUS_MP_PROCESSOR_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0x40000029, 'STATUS_MP_PROCESSOR_MISMATCH')
    STATUS_HIBERNATED: Final[ThirtyTwoBits] = cached_ntstatus(0x4000002A, 'STATUS_HIBERNATED')
    STATUS_RESUME_HIBERNATION: Final[ThirtyTwoBits] = cached_ntstatus(0x4000002B, 'STATUS_RESUME_HIBERNATION')
    STATUS_FIRMWARE_UPDATED: Final[ThirtyTwoBits] = cached_ntstatus(0x4000002C, 'STATUS_FIRMWARE_UPDATED')
    STATUS_DRIVERS_LEAKING_LOCKED_PAGES: Final[ThirtyTwoBits] = cached_ntstatus(0x4000002D, 'STATUS_DRIVERS_LEAKING_LOCKED_PAGES')
    STATUS_MESSAGE_RETRIEVED: Final[ThirtyTwoBits] = cached_ntstatus(0x4000002E, 'STATUS_MESSAGE_RETRIEVED')
    STATUS_SYSTEM_POWERSTATE_TRANSITION: Final[ThirtyTwoBits] = cached_ntstatus(0x4000002F, 'STATUS_SYSTEM_POWERSTATE_TRANSITION')
    STATUS_ALPC_CHECK_COMPLETION_LIST: Final[ThirtyTwoBits] = cached_ntstatus(0x40000030, 'STATUS_ALPC_CHECK_COMPLETION_LIST')
    STATUS_SYSTEM_POWERSTATE_COMPLEX_TRANSITION: Final[ThirtyTwoBits] = cached_ntstatus(0x40000031, 'STATUS_SYSTEM_POWERSTATE_COMPLEX_TRANSITION')
    STATUS_ACCESS_AUDIT_BY_POLICY: Final[ThirtyTwoBits] = cached_ntstatus(0x40000032, 'STATUS_ACCESS_AUDIT_BY_POLICY')
    STATUS_ABANDON_HIBERFILE: Final[ThirtyTwoBits] = cached_ntstatus(0x40000033, 'STATUS_ABANDON_HIBERFILE')
    STATUS_BIZRULES_NOT_ENABLED: Final[ThirtyTwoBits] = cached_ntstatus(0x40000034, 'STATUS_BIZRULES_NOT_ENABLED')
    STATUS_FT_READ_FROM_COPY: Final[ThirtyTwoBits] = cached_ntstatus(0x40000035, 'STATUS_FT_READ_FROM_COPY')
    STATUS_IMAGE_AT_DIFFERENT_BASE: Final[ThirtyTwoBits] = cached_ntstatus(0x40000036, 'STATUS_IMAGE_AT_DIFFERENT_BASE')
    STATUS_PATCH_DEFERRED: Final[ThirtyTwoBits] = cached_ntstatus(0x40000037, 'STATUS_PATCH_DEFERRED')
    STATUS_EMULATION_BREAKPOINT: Final[ThirtyTwoBits] = cached_ntstatus(0x40000038, 'STATUS_EMULATION_BREAKPOINT')
    STATUS_EMULATION_SYSCALL: Final[ThirtyTwoBits] = cached_ntstatus(0x40000039, 'STATUS_EMULATION_SYSCALL')
    STATUS_WAKE_SYSTEM: Final[ThirtyTwoBits] = cached_ntstatus(0x40000294, 'STATUS_WAKE_SYSTEM')
    STATUS_DS_SHUTTING_DOWN: Final[ThirtyTwoBits] = cached_ntstatus(0x40000370, 'STATUS_DS_SHUTTING_DOWN')
    STATUS_DISK_REPAIR_REDIRECTED: Final[ThirtyTwoBits] = cached_ntstatus(0x40000807, 'STATUS_DISK_REPAIR_REDIRECTED')
    STATUS_SERVICES_FAILED_AUTOSTART: Final[ThirtyTwoBits] = cached_ntstatus(0x4000A144, 'STATUS_SERVICES_FAILED_AUTOSTART')
    STATUS_CTX_CDM_CONNECT: Final[ThirtyTwoBits] = cached_ntstatus(0x400A0004, 'STATUS_CTX_CDM_CONNECT')
    STATUS_CTX_CDM_DISCONNECT: Final[ThirtyTwoBits] = cached_ntstatus(0x400A0005, 'STATUS_CTX_CDM_DISCONNECT')
    STATUS_SXS_RELEASE_ACTIVATION_CONTEXT: Final[ThirtyTwoBits] = cached_ntstatus(0x4015000D, 'STATUS_SXS_RELEASE_ACTIVATION_CONTEXT')
    STATUS_HEURISTIC_DAMAGE_POSSIBLE: Final[ThirtyTwoBits] = cached_ntstatus(0x40190001, 'STATUS_HEURISTIC_DAMAGE_POSSIBLE')
    STATUS_RECOVERY_NOT_NEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0x40190034, 'STATUS_RECOVERY_NOT_NEEDED')
    STATUS_RM_ALREADY_STARTED: Final[ThirtyTwoBits] = cached_ntstatus(0x40190035, 'STATUS_RM_ALREADY_STARTED')
    STATUS_LOG_NO_RESTART: Final[ThirtyTwoBits] = cached_ntstatus(0x401A000C, 'STATUS_LOG_NO_RESTART')
    STATUS_VIDEO_DRIVER_DEBUG_REPORT_REQUEST: Final[ThirtyTwoBits] = cached_ntstatus(0x401B00EC, 'STATUS_VIDEO_DRIVER_DEBUG_REPORT_REQUEST')
    STATUS_GRAPHICS_PARTIAL_DATA_POPULATED: Final[ThirtyTwoBits] = cached_ntstatus(0x401E000A, 'STATUS_GRAPHICS_PARTIAL_DATA_POPULATED')
    STATUS_GRAPHICS_SKIP_ALLOCATION_PREPARATION: Final[ThirtyTwoBits] = cached_ntstatus(0x401E0201, 'STATUS_GRAPHICS_SKIP_ALLOCATION_PREPARATION')
    STATUS_GRAPHICS_MODE_NOT_PINNED: Final[ThirtyTwoBits] = cached_ntstatus(0x401E0307, 'STATUS_GRAPHICS_MODE_NOT_PINNED')
    STATUS_GRAPHICS_NO_PREFERRED_MODE: Final[ThirtyTwoBits] = cached_ntstatus(0x401E031E, 'STATUS_GRAPHICS_NO_PREFERRED_MODE')
    STATUS_GRAPHICS_DATASET_IS_EMPTY: Final[ThirtyTwoBits] = cached_ntstatus(0x401E034B, 'STATUS_GRAPHICS_DATASET_IS_EMPTY')
    STATUS_GRAPHICS_NO_MORE_ELEMENTS_IN_DATASET: Final[ThirtyTwoBits] = cached_ntstatus(0x401E034C, 'STATUS_GRAPHICS_NO_MORE_ELEMENTS_IN_DATASET')
    STATUS_GRAPHICS_PATH_CONTENT_GEOMETRY_TRANSFORMATION_NOT_PINNED: Final[ThirtyTwoBits] = cached_ntstatus(0x401E0351, 'STATUS_GRAPHICS_PATH_CONTENT_GEOMETRY_TRANSFORMATION_NOT_PINNED')
    STATUS_GRAPHICS_UNKNOWN_CHILD_STATUS: Final[ThirtyTwoBits] = cached_ntstatus(0x401E042F, 'STATUS_GRAPHICS_UNKNOWN_CHILD_STATUS')
    STATUS_GRAPHICS_LEADLINK_START_DEFERRED: Final[ThirtyTwoBits] = cached_ntstatus(0x401E0437, 'STATUS_GRAPHICS_LEADLINK_START_DEFERRED')
    STATUS_GRAPHICS_POLLING_TOO_FREQUENTLY: Final[ThirtyTwoBits] = cached_ntstatus(0x401E0439, 'STATUS_GRAPHICS_POLLING_TOO_FREQUENTLY')
    STATUS_GRAPHICS_START_DEFERRED: Final[ThirtyTwoBits] = cached_ntstatus(0x401E043A, 'STATUS_GRAPHICS_START_DEFERRED')
    STATUS_GRAPHICS_DEPENDABLE_CHILD_STATUS: Final[ThirtyTwoBits] = cached_ntstatus(0x401E043C, 'STATUS_GRAPHICS_DEPENDABLE_CHILD_STATUS')
    STATUS_NDIS_INDICATION_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0x40230001, 'STATUS_NDIS_INDICATION_REQUIRED')
    STATUS_PCP_UNSUPPORTED_PSS_SALT: Final[ThirtyTwoBits] = cached_ntstatus(0x40292023, 'STATUS_PCP_UNSUPPORTED_PSS_SALT')
    STATUS_GUARD_PAGE_VIOLATION: Final[ThirtyTwoBits] = cached_ntstatus(0x80000001, 'STATUS_GUARD_PAGE_VIOLATION')
    STATUS_DATATYPE_MISALIGNMENT: Final[ThirtyTwoBits] = cached_ntstatus(0x80000002, 'STATUS_DATATYPE_MISALIGNMENT')
    STATUS_BREAKPOINT: Final[ThirtyTwoBits] = cached_ntstatus(0x80000003, 'STATUS_BREAKPOINT')
    STATUS_SINGLE_STEP: Final[ThirtyTwoBits] = cached_ntstatus(0x80000004, 'STATUS_SINGLE_STEP')
    STATUS_BUFFER_OVERFLOW: Final[ThirtyTwoBits] = cached_ntstatus(0x80000005, 'STATUS_BUFFER_OVERFLOW')
    STATUS_NO_MORE_FILES: Final[ThirtyTwoBits] = cached_ntstatus(0x80000006, 'STATUS_NO_MORE_FILES')
    STATUS_WAKE_SYSTEM_DEBUGGER: Final[ThirtyTwoBits] = cached_ntstatus(0x80000007, 'STATUS_WAKE_SYSTEM_DEBUGGER')
    STATUS_HANDLES_CLOSED: Final[ThirtyTwoBits] = cached_ntstatus(0x8000000A, 'STATUS_HANDLES_CLOSED')
    STATUS_NO_INHERITANCE: Final[ThirtyTwoBits] = cached_ntstatus(0x8000000B, 'STATUS_NO_INHERITANCE')
    STATUS_GUID_SUBSTITUTION_MADE: Final[ThirtyTwoBits] = cached_ntstatus(0x8000000C, 'STATUS_GUID_SUBSTITUTION_MADE')
    STATUS_PARTIAL_COPY: Final[ThirtyTwoBits] = cached_ntstatus(0x8000000D, 'STATUS_PARTIAL_COPY')
    STATUS_DEVICE_PAPER_EMPTY: Final[ThirtyTwoBits] = cached_ntstatus(0x8000000E, 'STATUS_DEVICE_PAPER_EMPTY')
    STATUS_DEVICE_POWERED_OFF: Final[ThirtyTwoBits] = cached_ntstatus(0x8000000F, 'STATUS_DEVICE_POWERED_OFF')
    STATUS_DEVICE_OFF_LINE: Final[ThirtyTwoBits] = cached_ntstatus(0x80000010, 'STATUS_DEVICE_OFF_LINE')
    STATUS_DEVICE_BUSY: Final[ThirtyTwoBits] = cached_ntstatus(0x80000011, 'STATUS_DEVICE_BUSY')
    STATUS_NO_MORE_EAS: Final[ThirtyTwoBits] = cached_ntstatus(0x80000012, 'STATUS_NO_MORE_EAS')
    STATUS_INVALID_EA_NAME: Final[ThirtyTwoBits] = cached_ntstatus(0x80000013, 'STATUS_INVALID_EA_NAME')
    STATUS_EA_LIST_INCONSISTENT: Final[ThirtyTwoBits] = cached_ntstatus(0x80000014, 'STATUS_EA_LIST_INCONSISTENT')
    STATUS_INVALID_EA_FLAG: Final[ThirtyTwoBits] = cached_ntstatus(0x80000015, 'STATUS_INVALID_EA_FLAG')
    STATUS_VERIFY_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0x80000016, 'STATUS_VERIFY_REQUIRED')
    STATUS_EXTRANEOUS_INFORMATION: Final[ThirtyTwoBits] = cached_ntstatus(0x80000017, 'STATUS_EXTRANEOUS_INFORMATION')
    STATUS_RXACT_COMMIT_NECESSARY: Final[ThirtyTwoBits] = cached_ntstatus(0x80000018, 'STATUS_RXACT_COMMIT_NECESSARY')
    STATUS_NO_MORE_ENTRIES: Final[ThirtyTwoBits] = cached_ntstatus(0x8000001A, 'STATUS_NO_MORE_ENTRIES')
    STATUS_FILEMARK_DETECTED: Final[ThirtyTwoBits] = cached_ntstatus(0x8000001B, 'STATUS_FILEMARK_DETECTED')
    STATUS_MEDIA_CHANGED: Final[ThirtyTwoBits] = cached_ntstatus(0x8000001C, 'STATUS_MEDIA_CHANGED')
    STATUS_BUS_RESET: Final[ThirtyTwoBits] = cached_ntstatus(0x8000001D, 'STATUS_BUS_RESET')
    STATUS_END_OF_MEDIA: Final[ThirtyTwoBits] = cached_ntstatus(0x8000001E, 'STATUS_END_OF_MEDIA')
    STATUS_BEGINNING_OF_MEDIA: Final[ThirtyTwoBits] = cached_ntstatus(0x8000001F, 'STATUS_BEGINNING_OF_MEDIA')
    STATUS_MEDIA_CHECK: Final[ThirtyTwoBits] = cached_ntstatus(0x80000020, 'STATUS_MEDIA_CHECK')
    STATUS_SETMARK_DETECTED: Final[ThirtyTwoBits] = cached_ntstatus(0x80000021, 'STATUS_SETMARK_DETECTED')
    STATUS_NO_DATA_DETECTED: Final[ThirtyTwoBits] = cached_ntstatus(0x80000022, 'STATUS_NO_DATA_DETECTED')
    STATUS_REDIRECTOR_HAS_OPEN_HANDLES: Final[ThirtyTwoBits] = cached_ntstatus(0x80000023, 'STATUS_REDIRECTOR_HAS_OPEN_HANDLES')
    STATUS_SERVER_HAS_OPEN_HANDLES: Final[ThirtyTwoBits] = cached_ntstatus(0x80000024, 'STATUS_SERVER_HAS_OPEN_HANDLES')
    STATUS_ALREADY_DISCONNECTED: Final[ThirtyTwoBits] = cached_ntstatus(0x80000025, 'STATUS_ALREADY_DISCONNECTED')
    STATUS_LONGJUMP: Final[ThirtyTwoBits] = cached_ntstatus(0x80000026, 'STATUS_LONGJUMP')
    STATUS_CLEANER_CARTRIDGE_INSTALLED: Final[ThirtyTwoBits] = cached_ntstatus(0x80000027, 'STATUS_CLEANER_CARTRIDGE_INSTALLED')
    STATUS_PLUGPLAY_QUERY_VETOED: Final[ThirtyTwoBits] = cached_ntstatus(0x80000028, 'STATUS_PLUGPLAY_QUERY_VETOED')
    STATUS_UNWIND_CONSOLIDATE: Final[ThirtyTwoBits] = cached_ntstatus(0x80000029, 'STATUS_UNWIND_CONSOLIDATE')
    STATUS_REGISTRY_HIVE_RECOVERED: Final[ThirtyTwoBits] = cached_ntstatus(0x8000002A, 'STATUS_REGISTRY_HIVE_RECOVERED')
    STATUS_DLL_MIGHT_BE_INSECURE: Final[ThirtyTwoBits] = cached_ntstatus(0x8000002B, 'STATUS_DLL_MIGHT_BE_INSECURE')
    STATUS_DLL_MIGHT_BE_INCOMPATIBLE: Final[ThirtyTwoBits] = cached_ntstatus(0x8000002C, 'STATUS_DLL_MIGHT_BE_INCOMPATIBLE')
    STATUS_STOPPED_ON_SYMLINK: Final[ThirtyTwoBits] = cached_ntstatus(0x8000002D, 'STATUS_STOPPED_ON_SYMLINK')
    STATUS_CANNOT_GRANT_REQUESTED_OPLOCK: Final[ThirtyTwoBits] = cached_ntstatus(0x8000002E, 'STATUS_CANNOT_GRANT_REQUESTED_OPLOCK')
    STATUS_NO_ACE_CONDITION: Final[ThirtyTwoBits] = cached_ntstatus(0x8000002F, 'STATUS_NO_ACE_CONDITION')
    STATUS_DEVICE_SUPPORT_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0x80000030, 'STATUS_DEVICE_SUPPORT_IN_PROGRESS')
    STATUS_DEVICE_POWER_CYCLE_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0x80000031, 'STATUS_DEVICE_POWER_CYCLE_REQUIRED')
    STATUS_NO_WORK_DONE: Final[ThirtyTwoBits] = cached_ntstatus(0x80000032, 'STATUS_NO_WORK_DONE')
    STATUS_RETURN_ADDRESS_HIJACK_ATTEMPT: Final[ThirtyTwoBits] = cached_ntstatus(0x80000033, 'STATUS_RETURN_ADDRESS_HIJACK_ATTEMPT')
    STATUS_RECOVERABLE_BUGCHECK: Final[ThirtyTwoBits] = cached_ntstatus(0x80000034, 'STATUS_RECOVERABLE_BUGCHECK')
    STATUS_DEVICE_RESET_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0x800001B6, 'STATUS_DEVICE_RESET_REQUIRED')
    STATUS_DEVICE_REQUIRES_CLEANING: Final[ThirtyTwoBits] = cached_ntstatus(0x80000288, 'STATUS_DEVICE_REQUIRES_CLEANING')
    STATUS_DEVICE_DOOR_OPEN: Final[ThirtyTwoBits] = cached_ntstatus(0x80000289, 'STATUS_DEVICE_DOOR_OPEN')
    STATUS_DATA_LOST_REPAIR: Final[ThirtyTwoBits] = cached_ntstatus(0x80000803, 'STATUS_DATA_LOST_REPAIR')
    STATUS_GPIO_INTERRUPT_ALREADY_UNMASKED: Final[ThirtyTwoBits] = cached_ntstatus(0x8000A127, 'STATUS_GPIO_INTERRUPT_ALREADY_UNMASKED')
    STATUS_CLOUD_FILE_PROPERTY_BLOB_CHECKSUM_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0x8000CF00, 'STATUS_CLOUD_FILE_PROPERTY_BLOB_CHECKSUM_MISMATCH')
    STATUS_CLOUD_FILE_PROPERTY_BLOB_TOO_LARGE: Final[ThirtyTwoBits] = cached_ntstatus(0x8000CF04, 'STATUS_CLOUD_FILE_PROPERTY_BLOB_TOO_LARGE')
    STATUS_CLOUD_FILE_TOO_MANY_PROPERTY_BLOBS: Final[ThirtyTwoBits] = cached_ntstatus(0x8000CF05, 'STATUS_CLOUD_FILE_TOO_MANY_PROPERTY_BLOBS')
    STATUS_CLUSTER_NODE_ALREADY_UP: Final[ThirtyTwoBits] = cached_ntstatus(0x80130001, 'STATUS_CLUSTER_NODE_ALREADY_UP')
    STATUS_CLUSTER_NODE_ALREADY_DOWN: Final[ThirtyTwoBits] = cached_ntstatus(0x80130002, 'STATUS_CLUSTER_NODE_ALREADY_DOWN')
    STATUS_CLUSTER_NETWORK_ALREADY_ONLINE: Final[ThirtyTwoBits] = cached_ntstatus(0x80130003, 'STATUS_CLUSTER_NETWORK_ALREADY_ONLINE')
    STATUS_CLUSTER_NETWORK_ALREADY_OFFLINE: Final[ThirtyTwoBits] = cached_ntstatus(0x80130004, 'STATUS_CLUSTER_NETWORK_ALREADY_OFFLINE')
    STATUS_CLUSTER_NODE_ALREADY_MEMBER: Final[ThirtyTwoBits] = cached_ntstatus(0x80130005, 'STATUS_CLUSTER_NODE_ALREADY_MEMBER')
    STATUS_COULD_NOT_RESIZE_LOG: Final[ThirtyTwoBits] = cached_ntstatus(0x80190009, 'STATUS_COULD_NOT_RESIZE_LOG')
    STATUS_NO_TXF_METADATA: Final[ThirtyTwoBits] = cached_ntstatus(0x80190029, 'STATUS_NO_TXF_METADATA')
    STATUS_CANT_RECOVER_WITH_HANDLE_OPEN: Final[ThirtyTwoBits] = cached_ntstatus(0x80190031, 'STATUS_CANT_RECOVER_WITH_HANDLE_OPEN')
    STATUS_TXF_METADATA_ALREADY_PRESENT: Final[ThirtyTwoBits] = cached_ntstatus(0x80190041, 'STATUS_TXF_METADATA_ALREADY_PRESENT')
    STATUS_TRANSACTION_SCOPE_CALLBACKS_NOT_SET: Final[ThirtyTwoBits] = cached_ntstatus(0x80190042, 'STATUS_TRANSACTION_SCOPE_CALLBACKS_NOT_SET')
    STATUS_VIDEO_HUNG_DISPLAY_DRIVER_THREAD_RECOVERED: Final[ThirtyTwoBits] = cached_ntstatus(0x801B00EB, 'STATUS_VIDEO_HUNG_DISPLAY_DRIVER_THREAD_RECOVERED')
    STATUS_FLT_BUFFER_TOO_SMALL: Final[ThirtyTwoBits] = cached_ntstatus(0x801C0001, 'STATUS_FLT_BUFFER_TOO_SMALL')
    STATUS_GRAPHICS_LINK_CONFIGURATION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0x801E0000, 'STATUS_GRAPHICS_LINK_CONFIGURATION_IN_PROGRESS')
    STATUS_FVE_PARTIAL_METADATA: Final[ThirtyTwoBits] = cached_ntstatus(0x80210001, 'STATUS_FVE_PARTIAL_METADATA')
    STATUS_FVE_TRANSIENT_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0x80210002, 'STATUS_FVE_TRANSIENT_STATE')
    STATUS_VID_REMOTE_NODE_PARENT_GPA_PAGES_USED: Final[ThirtyTwoBits] = cached_ntstatus(0x80370001, 'STATUS_VID_REMOTE_NODE_PARENT_GPA_PAGES_USED')
    STATUS_VOLMGR_INCOMPLETE_REGENERATION: Final[ThirtyTwoBits] = cached_ntstatus(0x80380001, 'STATUS_VOLMGR_INCOMPLETE_REGENERATION')
    STATUS_VOLMGR_INCOMPLETE_DISK_MIGRATION: Final[ThirtyTwoBits] = cached_ntstatus(0x80380002, 'STATUS_VOLMGR_INCOMPLETE_DISK_MIGRATION')
    STATUS_BCD_NOT_ALL_ENTRIES_IMPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0x80390001, 'STATUS_BCD_NOT_ALL_ENTRIES_IMPORTED')
    STATUS_BCD_NOT_ALL_ENTRIES_SYNCHRONIZED: Final[ThirtyTwoBits] = cached_ntstatus(0x80390003, 'STATUS_BCD_NOT_ALL_ENTRIES_SYNCHRONIZED')
    STATUS_QUERY_STORAGE_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0x803A0001, 'STATUS_QUERY_STORAGE_ERROR')
    STATUS_GDI_HANDLE_LEAK: Final[ThirtyTwoBits] = cached_ntstatus(0x803F0001, 'STATUS_GDI_HANDLE_LEAK')
    STATUS_SECUREBOOT_NOT_ENABLED: Final[ThirtyTwoBits] = cached_ntstatus(0x80430006, 'STATUS_SECUREBOOT_NOT_ENABLED')
    STATUS_UNSUCCESSFUL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000001, 'STATUS_UNSUCCESSFUL')
    STATUS_NOT_IMPLEMENTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000002, 'STATUS_NOT_IMPLEMENTED')
    STATUS_INVALID_INFO_CLASS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000003, 'STATUS_INVALID_INFO_CLASS')
    STATUS_INFO_LENGTH_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000004, 'STATUS_INFO_LENGTH_MISMATCH')
    STATUS_ACCESS_VIOLATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000005, 'STATUS_ACCESS_VIOLATION')
    STATUS_IN_PAGE_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000006, 'STATUS_IN_PAGE_ERROR')
    STATUS_PAGEFILE_QUOTA: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000007, 'STATUS_PAGEFILE_QUOTA')
    STATUS_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000008, 'STATUS_INVALID_HANDLE')
    STATUS_BAD_INITIAL_STACK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000009, 'STATUS_BAD_INITIAL_STACK')
    STATUS_BAD_INITIAL_PC: Final[ThirtyTwoBits] = cached_ntstatus(0xC000000A, 'STATUS_BAD_INITIAL_PC')
    STATUS_INVALID_CID: Final[ThirtyTwoBits] = cached_ntstatus(0xC000000B, 'STATUS_INVALID_CID')
    STATUS_TIMER_NOT_CANCELED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000000C, 'STATUS_TIMER_NOT_CANCELED')
    STATUS_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_ntstatus(0xC000000D, 'STATUS_INVALID_PARAMETER')
    STATUS_NO_SUCH_DEVICE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000000E, 'STATUS_NO_SUCH_DEVICE')
    STATUS_NO_SUCH_FILE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000000F, 'STATUS_NO_SUCH_FILE')
    STATUS_INVALID_DEVICE_REQUEST: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000010, 'STATUS_INVALID_DEVICE_REQUEST')
    STATUS_END_OF_FILE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000011, 'STATUS_END_OF_FILE')
    STATUS_WRONG_VOLUME: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000012, 'STATUS_WRONG_VOLUME')
    STATUS_NO_MEDIA_IN_DEVICE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000013, 'STATUS_NO_MEDIA_IN_DEVICE')
    STATUS_UNRECOGNIZED_MEDIA: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000014, 'STATUS_UNRECOGNIZED_MEDIA')
    STATUS_NONEXISTENT_SECTOR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000015, 'STATUS_NONEXISTENT_SECTOR')
    STATUS_MORE_PROCESSING_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000016, 'STATUS_MORE_PROCESSING_REQUIRED')
    STATUS_NO_MEMORY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000017, 'STATUS_NO_MEMORY')
    STATUS_CONFLICTING_ADDRESSES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000018, 'STATUS_CONFLICTING_ADDRESSES')
    STATUS_NOT_MAPPED_VIEW: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000019, 'STATUS_NOT_MAPPED_VIEW')
    STATUS_UNABLE_TO_FREE_VM: Final[ThirtyTwoBits] = cached_ntstatus(0xC000001A, 'STATUS_UNABLE_TO_FREE_VM')
    STATUS_UNABLE_TO_DELETE_SECTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000001B, 'STATUS_UNABLE_TO_DELETE_SECTION')
    STATUS_INVALID_SYSTEM_SERVICE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000001C, 'STATUS_INVALID_SYSTEM_SERVICE')
    STATUS_ILLEGAL_INSTRUCTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000001D, 'STATUS_ILLEGAL_INSTRUCTION')
    STATUS_INVALID_LOCK_SEQUENCE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000001E, 'STATUS_INVALID_LOCK_SEQUENCE')
    STATUS_INVALID_VIEW_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000001F, 'STATUS_INVALID_VIEW_SIZE')
    STATUS_INVALID_FILE_FOR_SECTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000020, 'STATUS_INVALID_FILE_FOR_SECTION')
    STATUS_ALREADY_COMMITTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000021, 'STATUS_ALREADY_COMMITTED')
    STATUS_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000022, 'STATUS_ACCESS_DENIED')
    STATUS_BUFFER_TOO_SMALL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000023, 'STATUS_BUFFER_TOO_SMALL')
    STATUS_OBJECT_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000024, 'STATUS_OBJECT_TYPE_MISMATCH')
    STATUS_NONCONTINUABLE_EXCEPTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000025, 'STATUS_NONCONTINUABLE_EXCEPTION')
    STATUS_INVALID_DISPOSITION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000026, 'STATUS_INVALID_DISPOSITION')
    STATUS_UNWIND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000027, 'STATUS_UNWIND')
    STATUS_BAD_STACK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000028, 'STATUS_BAD_STACK')
    STATUS_INVALID_UNWIND_TARGET: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000029, 'STATUS_INVALID_UNWIND_TARGET')
    STATUS_NOT_LOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000002A, 'STATUS_NOT_LOCKED')
    STATUS_PARITY_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC000002B, 'STATUS_PARITY_ERROR')
    STATUS_UNABLE_TO_DECOMMIT_VM: Final[ThirtyTwoBits] = cached_ntstatus(0xC000002C, 'STATUS_UNABLE_TO_DECOMMIT_VM')
    STATUS_NOT_COMMITTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000002D, 'STATUS_NOT_COMMITTED')
    STATUS_INVALID_PORT_ATTRIBUTES: Final[ThirtyTwoBits] = cached_ntstatus(0xC000002E, 'STATUS_INVALID_PORT_ATTRIBUTES')
    STATUS_PORT_MESSAGE_TOO_LONG: Final[ThirtyTwoBits] = cached_ntstatus(0xC000002F, 'STATUS_PORT_MESSAGE_TOO_LONG')
    STATUS_INVALID_PARAMETER_MIX: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000030, 'STATUS_INVALID_PARAMETER_MIX')
    STATUS_INVALID_QUOTA_LOWER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000031, 'STATUS_INVALID_QUOTA_LOWER')
    STATUS_DISK_CORRUPT_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000032, 'STATUS_DISK_CORRUPT_ERROR')
    STATUS_OBJECT_NAME_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000033, 'STATUS_OBJECT_NAME_INVALID')
    STATUS_OBJECT_NAME_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000034, 'STATUS_OBJECT_NAME_NOT_FOUND')
    STATUS_OBJECT_NAME_COLLISION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000035, 'STATUS_OBJECT_NAME_COLLISION')
    STATUS_PORT_DO_NOT_DISTURB: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000036, 'STATUS_PORT_DO_NOT_DISTURB')
    STATUS_PORT_DISCONNECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000037, 'STATUS_PORT_DISCONNECTED')
    STATUS_DEVICE_ALREADY_ATTACHED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000038, 'STATUS_DEVICE_ALREADY_ATTACHED')
    STATUS_OBJECT_PATH_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000039, 'STATUS_OBJECT_PATH_INVALID')
    STATUS_OBJECT_PATH_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC000003A, 'STATUS_OBJECT_PATH_NOT_FOUND')
    STATUS_OBJECT_PATH_SYNTAX_BAD: Final[ThirtyTwoBits] = cached_ntstatus(0xC000003B, 'STATUS_OBJECT_PATH_SYNTAX_BAD')
    STATUS_DATA_OVERRUN: Final[ThirtyTwoBits] = cached_ntstatus(0xC000003C, 'STATUS_DATA_OVERRUN')
    STATUS_DATA_LATE_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC000003D, 'STATUS_DATA_LATE_ERROR')
    STATUS_DATA_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC000003E, 'STATUS_DATA_ERROR')
    STATUS_CRC_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC000003F, 'STATUS_CRC_ERROR')
    STATUS_SECTION_TOO_BIG: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000040, 'STATUS_SECTION_TOO_BIG')
    STATUS_PORT_CONNECTION_REFUSED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000041, 'STATUS_PORT_CONNECTION_REFUSED')
    STATUS_INVALID_PORT_HANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000042, 'STATUS_INVALID_PORT_HANDLE')
    STATUS_SHARING_VIOLATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000043, 'STATUS_SHARING_VIOLATION')
    STATUS_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000044, 'STATUS_QUOTA_EXCEEDED')
    STATUS_INVALID_PAGE_PROTECTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000045, 'STATUS_INVALID_PAGE_PROTECTION')
    STATUS_MUTANT_NOT_OWNED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000046, 'STATUS_MUTANT_NOT_OWNED')
    STATUS_SEMAPHORE_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000047, 'STATUS_SEMAPHORE_LIMIT_EXCEEDED')
    STATUS_PORT_ALREADY_SET: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000048, 'STATUS_PORT_ALREADY_SET')
    STATUS_SECTION_NOT_IMAGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000049, 'STATUS_SECTION_NOT_IMAGE')
    STATUS_SUSPEND_COUNT_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000004A, 'STATUS_SUSPEND_COUNT_EXCEEDED')
    STATUS_THREAD_IS_TERMINATING: Final[ThirtyTwoBits] = cached_ntstatus(0xC000004B, 'STATUS_THREAD_IS_TERMINATING')
    STATUS_BAD_WORKING_SET_LIMIT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000004C, 'STATUS_BAD_WORKING_SET_LIMIT')
    STATUS_INCOMPATIBLE_FILE_MAP: Final[ThirtyTwoBits] = cached_ntstatus(0xC000004D, 'STATUS_INCOMPATIBLE_FILE_MAP')
    STATUS_SECTION_PROTECTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000004E, 'STATUS_SECTION_PROTECTION')
    STATUS_EAS_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000004F, 'STATUS_EAS_NOT_SUPPORTED')
    STATUS_EA_TOO_LARGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000050, 'STATUS_EA_TOO_LARGE')
    STATUS_NONEXISTENT_EA_ENTRY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000051, 'STATUS_NONEXISTENT_EA_ENTRY')
    STATUS_NO_EAS_ON_FILE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000052, 'STATUS_NO_EAS_ON_FILE')
    STATUS_EA_CORRUPT_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000053, 'STATUS_EA_CORRUPT_ERROR')
    STATUS_FILE_LOCK_CONFLICT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000054, 'STATUS_FILE_LOCK_CONFLICT')
    STATUS_LOCK_NOT_GRANTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000055, 'STATUS_LOCK_NOT_GRANTED')
    STATUS_DELETE_PENDING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000056, 'STATUS_DELETE_PENDING')
    STATUS_CTL_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000057, 'STATUS_CTL_FILE_NOT_SUPPORTED')
    STATUS_UNKNOWN_REVISION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000058, 'STATUS_UNKNOWN_REVISION')
    STATUS_REVISION_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000059, 'STATUS_REVISION_MISMATCH')
    STATUS_INVALID_OWNER: Final[ThirtyTwoBits] = cached_ntstatus(0xC000005A, 'STATUS_INVALID_OWNER')
    STATUS_INVALID_PRIMARY_GROUP: Final[ThirtyTwoBits] = cached_ntstatus(0xC000005B, 'STATUS_INVALID_PRIMARY_GROUP')
    STATUS_NO_IMPERSONATION_TOKEN: Final[ThirtyTwoBits] = cached_ntstatus(0xC000005C, 'STATUS_NO_IMPERSONATION_TOKEN')
    STATUS_CANT_DISABLE_MANDATORY: Final[ThirtyTwoBits] = cached_ntstatus(0xC000005D, 'STATUS_CANT_DISABLE_MANDATORY')
    STATUS_NO_LOGON_SERVERS: Final[ThirtyTwoBits] = cached_ntstatus(0xC000005E, 'STATUS_NO_LOGON_SERVERS')
    STATUS_NO_SUCH_LOGON_SESSION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000005F, 'STATUS_NO_SUCH_LOGON_SESSION')
    STATUS_NO_SUCH_PRIVILEGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000060, 'STATUS_NO_SUCH_PRIVILEGE')
    STATUS_PRIVILEGE_NOT_HELD: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000061, 'STATUS_PRIVILEGE_NOT_HELD')
    STATUS_INVALID_ACCOUNT_NAME: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000062, 'STATUS_INVALID_ACCOUNT_NAME')
    STATUS_USER_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000063, 'STATUS_USER_EXISTS')
    STATUS_NO_SUCH_USER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000064, 'STATUS_NO_SUCH_USER')
    STATUS_GROUP_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000065, 'STATUS_GROUP_EXISTS')
    STATUS_NO_SUCH_GROUP: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000066, 'STATUS_NO_SUCH_GROUP')
    STATUS_MEMBER_IN_GROUP: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000067, 'STATUS_MEMBER_IN_GROUP')
    STATUS_MEMBER_NOT_IN_GROUP: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000068, 'STATUS_MEMBER_NOT_IN_GROUP')
    STATUS_LAST_ADMIN: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000069, 'STATUS_LAST_ADMIN')
    STATUS_WRONG_PASSWORD: Final[ThirtyTwoBits] = cached_ntstatus(0xC000006A, 'STATUS_WRONG_PASSWORD')
    STATUS_ILL_FORMED_PASSWORD: Final[ThirtyTwoBits] = cached_ntstatus(0xC000006B, 'STATUS_ILL_FORMED_PASSWORD')
    STATUS_PASSWORD_RESTRICTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000006C, 'STATUS_PASSWORD_RESTRICTION')
    STATUS_LOGON_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000006D, 'STATUS_LOGON_FAILURE')
    STATUS_ACCOUNT_RESTRICTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000006E, 'STATUS_ACCOUNT_RESTRICTION')
    STATUS_INVALID_LOGON_HOURS: Final[ThirtyTwoBits] = cached_ntstatus(0xC000006F, 'STATUS_INVALID_LOGON_HOURS')
    STATUS_INVALID_WORKSTATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000070, 'STATUS_INVALID_WORKSTATION')
    STATUS_PASSWORD_EXPIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000071, 'STATUS_PASSWORD_EXPIRED')
    STATUS_ACCOUNT_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000072, 'STATUS_ACCOUNT_DISABLED')
    STATUS_NONE_MAPPED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000073, 'STATUS_NONE_MAPPED')
    STATUS_TOO_MANY_LUIDS_REQUESTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000074, 'STATUS_TOO_MANY_LUIDS_REQUESTED')
    STATUS_LUIDS_EXHAUSTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000075, 'STATUS_LUIDS_EXHAUSTED')
    STATUS_INVALID_SUB_AUTHORITY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000076, 'STATUS_INVALID_SUB_AUTHORITY')
    STATUS_INVALID_ACL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000077, 'STATUS_INVALID_ACL')
    STATUS_INVALID_SID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000078, 'STATUS_INVALID_SID')
    STATUS_INVALID_SECURITY_DESCR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000079, 'STATUS_INVALID_SECURITY_DESCR')
    STATUS_PROCEDURE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC000007A, 'STATUS_PROCEDURE_NOT_FOUND')
    STATUS_INVALID_IMAGE_FORMAT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000007B, 'STATUS_INVALID_IMAGE_FORMAT')
    STATUS_NO_TOKEN: Final[ThirtyTwoBits] = cached_ntstatus(0xC000007C, 'STATUS_NO_TOKEN')
    STATUS_BAD_INHERITANCE_ACL: Final[ThirtyTwoBits] = cached_ntstatus(0xC000007D, 'STATUS_BAD_INHERITANCE_ACL')
    STATUS_RANGE_NOT_LOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000007E, 'STATUS_RANGE_NOT_LOCKED')
    STATUS_DISK_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC000007F, 'STATUS_DISK_FULL')
    STATUS_SERVER_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000080, 'STATUS_SERVER_DISABLED')
    STATUS_SERVER_NOT_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000081, 'STATUS_SERVER_NOT_DISABLED')
    STATUS_TOO_MANY_GUIDS_REQUESTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000082, 'STATUS_TOO_MANY_GUIDS_REQUESTED')
    STATUS_GUIDS_EXHAUSTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000083, 'STATUS_GUIDS_EXHAUSTED')
    STATUS_INVALID_ID_AUTHORITY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000084, 'STATUS_INVALID_ID_AUTHORITY')
    STATUS_AGENTS_EXHAUSTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000085, 'STATUS_AGENTS_EXHAUSTED')
    STATUS_INVALID_VOLUME_LABEL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000086, 'STATUS_INVALID_VOLUME_LABEL')
    STATUS_SECTION_NOT_EXTENDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000087, 'STATUS_SECTION_NOT_EXTENDED')
    STATUS_NOT_MAPPED_DATA: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000088, 'STATUS_NOT_MAPPED_DATA')
    STATUS_RESOURCE_DATA_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000089, 'STATUS_RESOURCE_DATA_NOT_FOUND')
    STATUS_RESOURCE_TYPE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC000008A, 'STATUS_RESOURCE_TYPE_NOT_FOUND')
    STATUS_RESOURCE_NAME_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC000008B, 'STATUS_RESOURCE_NAME_NOT_FOUND')
    STATUS_ARRAY_BOUNDS_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000008C, 'STATUS_ARRAY_BOUNDS_EXCEEDED')
    STATUS_FLOAT_DENORMAL_OPERAND: Final[ThirtyTwoBits] = cached_ntstatus(0xC000008D, 'STATUS_FLOAT_DENORMAL_OPERAND')
    STATUS_FLOAT_DIVIDE_BY_ZERO: Final[ThirtyTwoBits] = cached_ntstatus(0xC000008E, 'STATUS_FLOAT_DIVIDE_BY_ZERO')
    STATUS_FLOAT_INEXACT_RESULT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000008F, 'STATUS_FLOAT_INEXACT_RESULT')
    STATUS_FLOAT_INVALID_OPERATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000090, 'STATUS_FLOAT_INVALID_OPERATION')
    STATUS_FLOAT_OVERFLOW: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000091, 'STATUS_FLOAT_OVERFLOW')
    STATUS_FLOAT_STACK_CHECK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000092, 'STATUS_FLOAT_STACK_CHECK')
    STATUS_FLOAT_UNDERFLOW: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000093, 'STATUS_FLOAT_UNDERFLOW')
    STATUS_INTEGER_DIVIDE_BY_ZERO: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000094, 'STATUS_INTEGER_DIVIDE_BY_ZERO')
    STATUS_INTEGER_OVERFLOW: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000095, 'STATUS_INTEGER_OVERFLOW')
    STATUS_PRIVILEGED_INSTRUCTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000096, 'STATUS_PRIVILEGED_INSTRUCTION')
    STATUS_TOO_MANY_PAGING_FILES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000097, 'STATUS_TOO_MANY_PAGING_FILES')
    STATUS_FILE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000098, 'STATUS_FILE_INVALID')
    STATUS_ALLOTTED_SPACE_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000099, 'STATUS_ALLOTTED_SPACE_EXCEEDED')
    STATUS_INSUFFICIENT_RESOURCES: Final[ThirtyTwoBits] = cached_ntstatus(0xC000009A, 'STATUS_INSUFFICIENT_RESOURCES')
    STATUS_DFS_EXIT_PATH_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC000009B, 'STATUS_DFS_EXIT_PATH_FOUND')
    STATUS_DEVICE_DATA_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC000009C, 'STATUS_DEVICE_DATA_ERROR')
    STATUS_DEVICE_NOT_CONNECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000009D, 'STATUS_DEVICE_NOT_CONNECTED')
    STATUS_DEVICE_POWER_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000009E, 'STATUS_DEVICE_POWER_FAILURE')
    STATUS_FREE_VM_NOT_AT_BASE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000009F, 'STATUS_FREE_VM_NOT_AT_BASE')
    STATUS_MEMORY_NOT_ALLOCATED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000A0, 'STATUS_MEMORY_NOT_ALLOCATED')
    STATUS_WORKING_SET_QUOTA: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000A1, 'STATUS_WORKING_SET_QUOTA')
    STATUS_MEDIA_WRITE_PROTECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000A2, 'STATUS_MEDIA_WRITE_PROTECTED')
    STATUS_DEVICE_NOT_READY: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000A3, 'STATUS_DEVICE_NOT_READY')
    STATUS_INVALID_GROUP_ATTRIBUTES: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000A4, 'STATUS_INVALID_GROUP_ATTRIBUTES')
    STATUS_BAD_IMPERSONATION_LEVEL: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000A5, 'STATUS_BAD_IMPERSONATION_LEVEL')
    STATUS_CANT_OPEN_ANONYMOUS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000A6, 'STATUS_CANT_OPEN_ANONYMOUS')
    STATUS_BAD_VALIDATION_CLASS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000A7, 'STATUS_BAD_VALIDATION_CLASS')
    STATUS_BAD_TOKEN_TYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000A8, 'STATUS_BAD_TOKEN_TYPE')
    STATUS_BAD_MASTER_BOOT_RECORD: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000A9, 'STATUS_BAD_MASTER_BOOT_RECORD')
    STATUS_INSTRUCTION_MISALIGNMENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000AA, 'STATUS_INSTRUCTION_MISALIGNMENT')
    STATUS_INSTANCE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000AB, 'STATUS_INSTANCE_NOT_AVAILABLE')
    STATUS_PIPE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000AC, 'STATUS_PIPE_NOT_AVAILABLE')
    STATUS_INVALID_PIPE_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000AD, 'STATUS_INVALID_PIPE_STATE')
    STATUS_PIPE_BUSY: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000AE, 'STATUS_PIPE_BUSY')
    STATUS_ILLEGAL_FUNCTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000AF, 'STATUS_ILLEGAL_FUNCTION')
    STATUS_PIPE_DISCONNECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000B0, 'STATUS_PIPE_DISCONNECTED')
    STATUS_PIPE_CLOSING: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000B1, 'STATUS_PIPE_CLOSING')
    STATUS_PIPE_CONNECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000B2, 'STATUS_PIPE_CONNECTED')
    STATUS_PIPE_LISTENING: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000B3, 'STATUS_PIPE_LISTENING')
    STATUS_INVALID_READ_MODE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000B4, 'STATUS_INVALID_READ_MODE')
    STATUS_IO_TIMEOUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000B5, 'STATUS_IO_TIMEOUT')
    STATUS_FILE_FORCED_CLOSED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000B6, 'STATUS_FILE_FORCED_CLOSED')
    STATUS_PROFILING_NOT_STARTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000B7, 'STATUS_PROFILING_NOT_STARTED')
    STATUS_PROFILING_NOT_STOPPED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000B8, 'STATUS_PROFILING_NOT_STOPPED')
    STATUS_COULD_NOT_INTERPRET: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000B9, 'STATUS_COULD_NOT_INTERPRET')
    STATUS_FILE_IS_A_DIRECTORY: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000BA, 'STATUS_FILE_IS_A_DIRECTORY')
    STATUS_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000BB, 'STATUS_NOT_SUPPORTED')
    STATUS_REMOTE_NOT_LISTENING: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000BC, 'STATUS_REMOTE_NOT_LISTENING')
    STATUS_DUPLICATE_NAME: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000BD, 'STATUS_DUPLICATE_NAME')
    STATUS_BAD_NETWORK_PATH: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000BE, 'STATUS_BAD_NETWORK_PATH')
    STATUS_NETWORK_BUSY: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000BF, 'STATUS_NETWORK_BUSY')
    STATUS_DEVICE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000C0, 'STATUS_DEVICE_DOES_NOT_EXIST')
    STATUS_TOO_MANY_COMMANDS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000C1, 'STATUS_TOO_MANY_COMMANDS')
    STATUS_ADAPTER_HARDWARE_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000C2, 'STATUS_ADAPTER_HARDWARE_ERROR')
    STATUS_INVALID_NETWORK_RESPONSE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000C3, 'STATUS_INVALID_NETWORK_RESPONSE')
    STATUS_UNEXPECTED_NETWORK_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000C4, 'STATUS_UNEXPECTED_NETWORK_ERROR')
    STATUS_BAD_REMOTE_ADAPTER: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000C5, 'STATUS_BAD_REMOTE_ADAPTER')
    STATUS_PRINT_QUEUE_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000C6, 'STATUS_PRINT_QUEUE_FULL')
    STATUS_NO_SPOOL_SPACE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000C7, 'STATUS_NO_SPOOL_SPACE')
    STATUS_PRINT_CANCELLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000C8, 'STATUS_PRINT_CANCELLED')
    STATUS_NETWORK_NAME_DELETED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000C9, 'STATUS_NETWORK_NAME_DELETED')
    STATUS_NETWORK_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000CA, 'STATUS_NETWORK_ACCESS_DENIED')
    STATUS_BAD_DEVICE_TYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000CB, 'STATUS_BAD_DEVICE_TYPE')
    STATUS_BAD_NETWORK_NAME: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000CC, 'STATUS_BAD_NETWORK_NAME')
    STATUS_TOO_MANY_NAMES: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000CD, 'STATUS_TOO_MANY_NAMES')
    STATUS_TOO_MANY_SESSIONS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000CE, 'STATUS_TOO_MANY_SESSIONS')
    STATUS_SHARING_PAUSED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000CF, 'STATUS_SHARING_PAUSED')
    STATUS_REQUEST_NOT_ACCEPTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000D0, 'STATUS_REQUEST_NOT_ACCEPTED')
    STATUS_REDIRECTOR_PAUSED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000D1, 'STATUS_REDIRECTOR_PAUSED')
    STATUS_NET_WRITE_FAULT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000D2, 'STATUS_NET_WRITE_FAULT')
    STATUS_PROFILING_AT_LIMIT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000D3, 'STATUS_PROFILING_AT_LIMIT')
    STATUS_NOT_SAME_DEVICE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000D4, 'STATUS_NOT_SAME_DEVICE')
    STATUS_FILE_RENAMED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000D5, 'STATUS_FILE_RENAMED')
    STATUS_VIRTUAL_CIRCUIT_CLOSED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000D6, 'STATUS_VIRTUAL_CIRCUIT_CLOSED')
    STATUS_NO_SECURITY_ON_OBJECT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000D7, 'STATUS_NO_SECURITY_ON_OBJECT')
    STATUS_CANT_WAIT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000D8, 'STATUS_CANT_WAIT')
    STATUS_PIPE_EMPTY: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000D9, 'STATUS_PIPE_EMPTY')
    STATUS_CANT_ACCESS_DOMAIN_INFO: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000DA, 'STATUS_CANT_ACCESS_DOMAIN_INFO')
    STATUS_CANT_TERMINATE_SELF: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000DB, 'STATUS_CANT_TERMINATE_SELF')
    STATUS_INVALID_SERVER_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000DC, 'STATUS_INVALID_SERVER_STATE')
    STATUS_INVALID_DOMAIN_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000DD, 'STATUS_INVALID_DOMAIN_STATE')
    STATUS_INVALID_DOMAIN_ROLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000DE, 'STATUS_INVALID_DOMAIN_ROLE')
    STATUS_NO_SUCH_DOMAIN: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000DF, 'STATUS_NO_SUCH_DOMAIN')
    STATUS_DOMAIN_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000E0, 'STATUS_DOMAIN_EXISTS')
    STATUS_DOMAIN_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000E1, 'STATUS_DOMAIN_LIMIT_EXCEEDED')
    STATUS_OPLOCK_NOT_GRANTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000E2, 'STATUS_OPLOCK_NOT_GRANTED')
    STATUS_INVALID_OPLOCK_PROTOCOL: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000E3, 'STATUS_INVALID_OPLOCK_PROTOCOL')
    STATUS_INTERNAL_DB_CORRUPTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000E4, 'STATUS_INTERNAL_DB_CORRUPTION')
    STATUS_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000E5, 'STATUS_INTERNAL_ERROR')
    STATUS_GENERIC_NOT_MAPPED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000E6, 'STATUS_GENERIC_NOT_MAPPED')
    STATUS_BAD_DESCRIPTOR_FORMAT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000E7, 'STATUS_BAD_DESCRIPTOR_FORMAT')
    STATUS_INVALID_USER_BUFFER: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000E8, 'STATUS_INVALID_USER_BUFFER')
    STATUS_UNEXPECTED_IO_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000E9, 'STATUS_UNEXPECTED_IO_ERROR')
    STATUS_UNEXPECTED_MM_CREATE_ERR: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000EA, 'STATUS_UNEXPECTED_MM_CREATE_ERR')
    STATUS_UNEXPECTED_MM_MAP_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000EB, 'STATUS_UNEXPECTED_MM_MAP_ERROR')
    STATUS_UNEXPECTED_MM_EXTEND_ERR: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000EC, 'STATUS_UNEXPECTED_MM_EXTEND_ERR')
    STATUS_NOT_LOGON_PROCESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000ED, 'STATUS_NOT_LOGON_PROCESS')
    STATUS_LOGON_SESSION_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000EE, 'STATUS_LOGON_SESSION_EXISTS')
    STATUS_INVALID_PARAMETER_1: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000EF, 'STATUS_INVALID_PARAMETER_1')
    STATUS_INVALID_PARAMETER_2: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000F0, 'STATUS_INVALID_PARAMETER_2')
    STATUS_INVALID_PARAMETER_3: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000F1, 'STATUS_INVALID_PARAMETER_3')
    STATUS_INVALID_PARAMETER_4: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000F2, 'STATUS_INVALID_PARAMETER_4')
    STATUS_INVALID_PARAMETER_5: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000F3, 'STATUS_INVALID_PARAMETER_5')
    STATUS_INVALID_PARAMETER_6: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000F4, 'STATUS_INVALID_PARAMETER_6')
    STATUS_INVALID_PARAMETER_7: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000F5, 'STATUS_INVALID_PARAMETER_7')
    STATUS_INVALID_PARAMETER_8: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000F6, 'STATUS_INVALID_PARAMETER_8')
    STATUS_INVALID_PARAMETER_9: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000F7, 'STATUS_INVALID_PARAMETER_9')
    STATUS_INVALID_PARAMETER_10: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000F8, 'STATUS_INVALID_PARAMETER_10')
    STATUS_INVALID_PARAMETER_11: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000F9, 'STATUS_INVALID_PARAMETER_11')
    STATUS_INVALID_PARAMETER_12: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000FA, 'STATUS_INVALID_PARAMETER_12')
    STATUS_REDIRECTOR_NOT_STARTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000FB, 'STATUS_REDIRECTOR_NOT_STARTED')
    STATUS_REDIRECTOR_STARTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000FC, 'STATUS_REDIRECTOR_STARTED')
    STATUS_STACK_OVERFLOW: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000FD, 'STATUS_STACK_OVERFLOW')
    STATUS_NO_SUCH_PACKAGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000FE, 'STATUS_NO_SUCH_PACKAGE')
    STATUS_BAD_FUNCTION_TABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00000FF, 'STATUS_BAD_FUNCTION_TABLE')
    STATUS_VARIABLE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000100, 'STATUS_VARIABLE_NOT_FOUND')
    STATUS_DIRECTORY_NOT_EMPTY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000101, 'STATUS_DIRECTORY_NOT_EMPTY')
    STATUS_FILE_CORRUPT_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000102, 'STATUS_FILE_CORRUPT_ERROR')
    STATUS_NOT_A_DIRECTORY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000103, 'STATUS_NOT_A_DIRECTORY')
    STATUS_BAD_LOGON_SESSION_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000104, 'STATUS_BAD_LOGON_SESSION_STATE')
    STATUS_LOGON_SESSION_COLLISION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000105, 'STATUS_LOGON_SESSION_COLLISION')
    STATUS_NAME_TOO_LONG: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000106, 'STATUS_NAME_TOO_LONG')
    STATUS_FILES_OPEN: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000107, 'STATUS_FILES_OPEN')
    STATUS_CONNECTION_IN_USE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000108, 'STATUS_CONNECTION_IN_USE')
    STATUS_MESSAGE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000109, 'STATUS_MESSAGE_NOT_FOUND')
    STATUS_PROCESS_IS_TERMINATING: Final[ThirtyTwoBits] = cached_ntstatus(0xC000010A, 'STATUS_PROCESS_IS_TERMINATING')
    STATUS_INVALID_LOGON_TYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000010B, 'STATUS_INVALID_LOGON_TYPE')
    STATUS_NO_GUID_TRANSLATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000010C, 'STATUS_NO_GUID_TRANSLATION')
    STATUS_CANNOT_IMPERSONATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000010D, 'STATUS_CANNOT_IMPERSONATE')
    STATUS_IMAGE_ALREADY_LOADED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000010E, 'STATUS_IMAGE_ALREADY_LOADED')
    STATUS_ABIOS_NOT_PRESENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000010F, 'STATUS_ABIOS_NOT_PRESENT')
    STATUS_ABIOS_LID_NOT_EXIST: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000110, 'STATUS_ABIOS_LID_NOT_EXIST')
    STATUS_ABIOS_LID_ALREADY_OWNED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000111, 'STATUS_ABIOS_LID_ALREADY_OWNED')
    STATUS_ABIOS_NOT_LID_OWNER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000112, 'STATUS_ABIOS_NOT_LID_OWNER')
    STATUS_ABIOS_INVALID_COMMAND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000113, 'STATUS_ABIOS_INVALID_COMMAND')
    STATUS_ABIOS_INVALID_LID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000114, 'STATUS_ABIOS_INVALID_LID')
    STATUS_ABIOS_SELECTOR_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000115, 'STATUS_ABIOS_SELECTOR_NOT_AVAILABLE')
    STATUS_ABIOS_INVALID_SELECTOR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000116, 'STATUS_ABIOS_INVALID_SELECTOR')
    STATUS_NO_LDT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000117, 'STATUS_NO_LDT')
    STATUS_INVALID_LDT_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000118, 'STATUS_INVALID_LDT_SIZE')
    STATUS_INVALID_LDT_OFFSET: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000119, 'STATUS_INVALID_LDT_OFFSET')
    STATUS_INVALID_LDT_DESCRIPTOR: Final[ThirtyTwoBits] = cached_ntstatus(0xC000011A, 'STATUS_INVALID_LDT_DESCRIPTOR')
    STATUS_INVALID_IMAGE_NE_FORMAT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000011B, 'STATUS_INVALID_IMAGE_NE_FORMAT')
    STATUS_RXACT_INVALID_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000011C, 'STATUS_RXACT_INVALID_STATE')
    STATUS_RXACT_COMMIT_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000011D, 'STATUS_RXACT_COMMIT_FAILURE')
    STATUS_MAPPED_FILE_SIZE_ZERO: Final[ThirtyTwoBits] = cached_ntstatus(0xC000011E, 'STATUS_MAPPED_FILE_SIZE_ZERO')
    STATUS_TOO_MANY_OPENED_FILES: Final[ThirtyTwoBits] = cached_ntstatus(0xC000011F, 'STATUS_TOO_MANY_OPENED_FILES')
    STATUS_CANCELLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000120, 'STATUS_CANCELLED')
    STATUS_CANNOT_DELETE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000121, 'STATUS_CANNOT_DELETE')
    STATUS_INVALID_COMPUTER_NAME: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000122, 'STATUS_INVALID_COMPUTER_NAME')
    STATUS_FILE_DELETED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000123, 'STATUS_FILE_DELETED')
    STATUS_SPECIAL_ACCOUNT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000124, 'STATUS_SPECIAL_ACCOUNT')
    STATUS_SPECIAL_GROUP: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000125, 'STATUS_SPECIAL_GROUP')
    STATUS_SPECIAL_USER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000126, 'STATUS_SPECIAL_USER')
    STATUS_MEMBERS_PRIMARY_GROUP: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000127, 'STATUS_MEMBERS_PRIMARY_GROUP')
    STATUS_FILE_CLOSED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000128, 'STATUS_FILE_CLOSED')
    STATUS_TOO_MANY_THREADS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000129, 'STATUS_TOO_MANY_THREADS')
    STATUS_THREAD_NOT_IN_PROCESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC000012A, 'STATUS_THREAD_NOT_IN_PROCESS')
    STATUS_TOKEN_ALREADY_IN_USE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000012B, 'STATUS_TOKEN_ALREADY_IN_USE')
    STATUS_PAGEFILE_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000012C, 'STATUS_PAGEFILE_QUOTA_EXCEEDED')
    STATUS_COMMITMENT_LIMIT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000012D, 'STATUS_COMMITMENT_LIMIT')
    STATUS_INVALID_IMAGE_LE_FORMAT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000012E, 'STATUS_INVALID_IMAGE_LE_FORMAT')
    STATUS_INVALID_IMAGE_NOT_MZ: Final[ThirtyTwoBits] = cached_ntstatus(0xC000012F, 'STATUS_INVALID_IMAGE_NOT_MZ')
    STATUS_INVALID_IMAGE_PROTECT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000130, 'STATUS_INVALID_IMAGE_PROTECT')
    STATUS_INVALID_IMAGE_WIN_16: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000131, 'STATUS_INVALID_IMAGE_WIN_16')
    STATUS_LOGON_SERVER_CONFLICT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000132, 'STATUS_LOGON_SERVER_CONFLICT')
    STATUS_TIME_DIFFERENCE_AT_DC: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000133, 'STATUS_TIME_DIFFERENCE_AT_DC')
    STATUS_SYNCHRONIZATION_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000134, 'STATUS_SYNCHRONIZATION_REQUIRED')
    STATUS_DLL_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000135, 'STATUS_DLL_NOT_FOUND')
    STATUS_OPEN_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000136, 'STATUS_OPEN_FAILED')
    STATUS_IO_PRIVILEGE_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000137, 'STATUS_IO_PRIVILEGE_FAILED')
    STATUS_ORDINAL_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000138, 'STATUS_ORDINAL_NOT_FOUND')
    STATUS_ENTRYPOINT_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000139, 'STATUS_ENTRYPOINT_NOT_FOUND')
    STATUS_CONTROL_C_EXIT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000013A, 'STATUS_CONTROL_C_EXIT')
    STATUS_LOCAL_DISCONNECT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000013B, 'STATUS_LOCAL_DISCONNECT')
    STATUS_REMOTE_DISCONNECT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000013C, 'STATUS_REMOTE_DISCONNECT')
    STATUS_REMOTE_RESOURCES: Final[ThirtyTwoBits] = cached_ntstatus(0xC000013D, 'STATUS_REMOTE_RESOURCES')
    STATUS_LINK_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000013E, 'STATUS_LINK_FAILED')
    STATUS_LINK_TIMEOUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000013F, 'STATUS_LINK_TIMEOUT')
    STATUS_INVALID_CONNECTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000140, 'STATUS_INVALID_CONNECTION')
    STATUS_INVALID_ADDRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000141, 'STATUS_INVALID_ADDRESS')
    STATUS_DLL_INIT_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000142, 'STATUS_DLL_INIT_FAILED')
    STATUS_MISSING_SYSTEMFILE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000143, 'STATUS_MISSING_SYSTEMFILE')
    STATUS_UNHANDLED_EXCEPTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000144, 'STATUS_UNHANDLED_EXCEPTION')
    STATUS_APP_INIT_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000145, 'STATUS_APP_INIT_FAILURE')
    STATUS_PAGEFILE_CREATE_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000146, 'STATUS_PAGEFILE_CREATE_FAILED')
    STATUS_NO_PAGEFILE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000147, 'STATUS_NO_PAGEFILE')
    STATUS_INVALID_LEVEL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000148, 'STATUS_INVALID_LEVEL')
    STATUS_WRONG_PASSWORD_CORE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000149, 'STATUS_WRONG_PASSWORD_CORE')
    STATUS_ILLEGAL_FLOAT_CONTEXT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000014A, 'STATUS_ILLEGAL_FLOAT_CONTEXT')
    STATUS_PIPE_BROKEN: Final[ThirtyTwoBits] = cached_ntstatus(0xC000014B, 'STATUS_PIPE_BROKEN')
    STATUS_REGISTRY_CORRUPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000014C, 'STATUS_REGISTRY_CORRUPT')
    STATUS_REGISTRY_IO_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000014D, 'STATUS_REGISTRY_IO_FAILED')
    STATUS_NO_EVENT_PAIR: Final[ThirtyTwoBits] = cached_ntstatus(0xC000014E, 'STATUS_NO_EVENT_PAIR')
    STATUS_UNRECOGNIZED_VOLUME: Final[ThirtyTwoBits] = cached_ntstatus(0xC000014F, 'STATUS_UNRECOGNIZED_VOLUME')
    STATUS_SERIAL_NO_DEVICE_INITED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000150, 'STATUS_SERIAL_NO_DEVICE_INITED')
    STATUS_NO_SUCH_ALIAS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000151, 'STATUS_NO_SUCH_ALIAS')
    STATUS_MEMBER_NOT_IN_ALIAS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000152, 'STATUS_MEMBER_NOT_IN_ALIAS')
    STATUS_MEMBER_IN_ALIAS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000153, 'STATUS_MEMBER_IN_ALIAS')
    STATUS_ALIAS_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000154, 'STATUS_ALIAS_EXISTS')
    STATUS_LOGON_NOT_GRANTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000155, 'STATUS_LOGON_NOT_GRANTED')
    STATUS_TOO_MANY_SECRETS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000156, 'STATUS_TOO_MANY_SECRETS')
    STATUS_SECRET_TOO_LONG: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000157, 'STATUS_SECRET_TOO_LONG')
    STATUS_INTERNAL_DB_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000158, 'STATUS_INTERNAL_DB_ERROR')
    STATUS_FULLSCREEN_MODE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000159, 'STATUS_FULLSCREEN_MODE')
    STATUS_TOO_MANY_CONTEXT_IDS: Final[ThirtyTwoBits] = cached_ntstatus(0xC000015A, 'STATUS_TOO_MANY_CONTEXT_IDS')
    STATUS_LOGON_TYPE_NOT_GRANTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000015B, 'STATUS_LOGON_TYPE_NOT_GRANTED')
    STATUS_NOT_REGISTRY_FILE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000015C, 'STATUS_NOT_REGISTRY_FILE')
    STATUS_NT_CROSS_ENCRYPTION_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000015D, 'STATUS_NT_CROSS_ENCRYPTION_REQUIRED')
    STATUS_DOMAIN_CTRLR_CONFIG_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC000015E, 'STATUS_DOMAIN_CTRLR_CONFIG_ERROR')
    STATUS_FT_MISSING_MEMBER: Final[ThirtyTwoBits] = cached_ntstatus(0xC000015F, 'STATUS_FT_MISSING_MEMBER')
    STATUS_ILL_FORMED_SERVICE_ENTRY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000160, 'STATUS_ILL_FORMED_SERVICE_ENTRY')
    STATUS_ILLEGAL_CHARACTER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000161, 'STATUS_ILLEGAL_CHARACTER')
    STATUS_UNMAPPABLE_CHARACTER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000162, 'STATUS_UNMAPPABLE_CHARACTER')
    STATUS_UNDEFINED_CHARACTER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000163, 'STATUS_UNDEFINED_CHARACTER')
    STATUS_FLOPPY_VOLUME: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000164, 'STATUS_FLOPPY_VOLUME')
    STATUS_FLOPPY_ID_MARK_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000165, 'STATUS_FLOPPY_ID_MARK_NOT_FOUND')
    STATUS_FLOPPY_WRONG_CYLINDER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000166, 'STATUS_FLOPPY_WRONG_CYLINDER')
    STATUS_FLOPPY_UNKNOWN_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000167, 'STATUS_FLOPPY_UNKNOWN_ERROR')
    STATUS_FLOPPY_BAD_REGISTERS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000168, 'STATUS_FLOPPY_BAD_REGISTERS')
    STATUS_DISK_RECALIBRATE_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000169, 'STATUS_DISK_RECALIBRATE_FAILED')
    STATUS_DISK_OPERATION_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000016A, 'STATUS_DISK_OPERATION_FAILED')
    STATUS_DISK_RESET_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000016B, 'STATUS_DISK_RESET_FAILED')
    STATUS_SHARED_IRQ_BUSY: Final[ThirtyTwoBits] = cached_ntstatus(0xC000016C, 'STATUS_SHARED_IRQ_BUSY')
    STATUS_FT_ORPHANING: Final[ThirtyTwoBits] = cached_ntstatus(0xC000016D, 'STATUS_FT_ORPHANING')
    STATUS_BIOS_FAILED_TO_CONNECT_INTERRUPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000016E, 'STATUS_BIOS_FAILED_TO_CONNECT_INTERRUPT')
    STATUS_PARTITION_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000172, 'STATUS_PARTITION_FAILURE')
    STATUS_INVALID_BLOCK_LENGTH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000173, 'STATUS_INVALID_BLOCK_LENGTH')
    STATUS_DEVICE_NOT_PARTITIONED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000174, 'STATUS_DEVICE_NOT_PARTITIONED')
    STATUS_UNABLE_TO_LOCK_MEDIA: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000175, 'STATUS_UNABLE_TO_LOCK_MEDIA')
    STATUS_UNABLE_TO_UNLOAD_MEDIA: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000176, 'STATUS_UNABLE_TO_UNLOAD_MEDIA')
    STATUS_EOM_OVERFLOW: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000177, 'STATUS_EOM_OVERFLOW')
    STATUS_NO_MEDIA: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000178, 'STATUS_NO_MEDIA')
    STATUS_NO_SUCH_MEMBER: Final[ThirtyTwoBits] = cached_ntstatus(0xC000017A, 'STATUS_NO_SUCH_MEMBER')
    STATUS_INVALID_MEMBER: Final[ThirtyTwoBits] = cached_ntstatus(0xC000017B, 'STATUS_INVALID_MEMBER')
    STATUS_KEY_DELETED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000017C, 'STATUS_KEY_DELETED')
    STATUS_NO_LOG_SPACE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000017D, 'STATUS_NO_LOG_SPACE')
    STATUS_TOO_MANY_SIDS: Final[ThirtyTwoBits] = cached_ntstatus(0xC000017E, 'STATUS_TOO_MANY_SIDS')
    STATUS_LM_CROSS_ENCRYPTION_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000017F, 'STATUS_LM_CROSS_ENCRYPTION_REQUIRED')
    STATUS_KEY_HAS_CHILDREN: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000180, 'STATUS_KEY_HAS_CHILDREN')
    STATUS_CHILD_MUST_BE_VOLATILE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000181, 'STATUS_CHILD_MUST_BE_VOLATILE')
    STATUS_DEVICE_CONFIGURATION_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000182, 'STATUS_DEVICE_CONFIGURATION_ERROR')
    STATUS_DRIVER_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000183, 'STATUS_DRIVER_INTERNAL_ERROR')
    STATUS_INVALID_DEVICE_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000184, 'STATUS_INVALID_DEVICE_STATE')
    STATUS_IO_DEVICE_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000185, 'STATUS_IO_DEVICE_ERROR')
    STATUS_DEVICE_PROTOCOL_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000186, 'STATUS_DEVICE_PROTOCOL_ERROR')
    STATUS_BACKUP_CONTROLLER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000187, 'STATUS_BACKUP_CONTROLLER')
    STATUS_LOG_FILE_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000188, 'STATUS_LOG_FILE_FULL')
    STATUS_TOO_LATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000189, 'STATUS_TOO_LATE')
    STATUS_NO_TRUST_LSA_SECRET: Final[ThirtyTwoBits] = cached_ntstatus(0xC000018A, 'STATUS_NO_TRUST_LSA_SECRET')
    STATUS_NO_TRUST_SAM_ACCOUNT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000018B, 'STATUS_NO_TRUST_SAM_ACCOUNT')
    STATUS_TRUSTED_DOMAIN_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000018C, 'STATUS_TRUSTED_DOMAIN_FAILURE')
    STATUS_TRUSTED_RELATIONSHIP_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000018D, 'STATUS_TRUSTED_RELATIONSHIP_FAILURE')
    STATUS_EVENTLOG_FILE_CORRUPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000018E, 'STATUS_EVENTLOG_FILE_CORRUPT')
    STATUS_EVENTLOG_CANT_START: Final[ThirtyTwoBits] = cached_ntstatus(0xC000018F, 'STATUS_EVENTLOG_CANT_START')
    STATUS_TRUST_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000190, 'STATUS_TRUST_FAILURE')
    STATUS_MUTANT_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000191, 'STATUS_MUTANT_LIMIT_EXCEEDED')
    STATUS_NETLOGON_NOT_STARTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000192, 'STATUS_NETLOGON_NOT_STARTED')
    STATUS_ACCOUNT_EXPIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000193, 'STATUS_ACCOUNT_EXPIRED')
    STATUS_POSSIBLE_DEADLOCK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000194, 'STATUS_POSSIBLE_DEADLOCK')
    STATUS_NETWORK_CREDENTIAL_CONFLICT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000195, 'STATUS_NETWORK_CREDENTIAL_CONFLICT')
    STATUS_REMOTE_SESSION_LIMIT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000196, 'STATUS_REMOTE_SESSION_LIMIT')
    STATUS_EVENTLOG_FILE_CHANGED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000197, 'STATUS_EVENTLOG_FILE_CHANGED')
    STATUS_NOLOGON_INTERDOMAIN_TRUST_ACCOUNT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000198, 'STATUS_NOLOGON_INTERDOMAIN_TRUST_ACCOUNT')
    STATUS_NOLOGON_WORKSTATION_TRUST_ACCOUNT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000199, 'STATUS_NOLOGON_WORKSTATION_TRUST_ACCOUNT')
    STATUS_NOLOGON_SERVER_TRUST_ACCOUNT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000019A, 'STATUS_NOLOGON_SERVER_TRUST_ACCOUNT')
    STATUS_DOMAIN_TRUST_INCONSISTENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000019B, 'STATUS_DOMAIN_TRUST_INCONSISTENT')
    STATUS_FS_DRIVER_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000019C, 'STATUS_FS_DRIVER_REQUIRED')
    STATUS_IMAGE_ALREADY_LOADED_AS_DLL: Final[ThirtyTwoBits] = cached_ntstatus(0xC000019D, 'STATUS_IMAGE_ALREADY_LOADED_AS_DLL')
    STATUS_INCOMPATIBLE_WITH_GLOBAL_SHORT_NAME_REGISTRY_SETTING: Final[ThirtyTwoBits] = cached_ntstatus(0xC000019E, 'STATUS_INCOMPATIBLE_WITH_GLOBAL_SHORT_NAME_REGISTRY_SETTING')
    STATUS_SHORT_NAMES_NOT_ENABLED_ON_VOLUME: Final[ThirtyTwoBits] = cached_ntstatus(0xC000019F, 'STATUS_SHORT_NAMES_NOT_ENABLED_ON_VOLUME')
    STATUS_SECURITY_STREAM_IS_INCONSISTENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001A0, 'STATUS_SECURITY_STREAM_IS_INCONSISTENT')
    STATUS_INVALID_LOCK_RANGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001A1, 'STATUS_INVALID_LOCK_RANGE')
    STATUS_INVALID_ACE_CONDITION: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001A2, 'STATUS_INVALID_ACE_CONDITION')
    STATUS_IMAGE_SUBSYSTEM_NOT_PRESENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001A3, 'STATUS_IMAGE_SUBSYSTEM_NOT_PRESENT')
    STATUS_NOTIFICATION_GUID_ALREADY_DEFINED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001A4, 'STATUS_NOTIFICATION_GUID_ALREADY_DEFINED')
    STATUS_INVALID_EXCEPTION_HANDLER: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001A5, 'STATUS_INVALID_EXCEPTION_HANDLER')
    STATUS_DUPLICATE_PRIVILEGES: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001A6, 'STATUS_DUPLICATE_PRIVILEGES')
    STATUS_NOT_ALLOWED_ON_SYSTEM_FILE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001A7, 'STATUS_NOT_ALLOWED_ON_SYSTEM_FILE')
    STATUS_REPAIR_NEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001A8, 'STATUS_REPAIR_NEEDED')
    STATUS_QUOTA_NOT_ENABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001A9, 'STATUS_QUOTA_NOT_ENABLED')
    STATUS_NO_APPLICATION_PACKAGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001AA, 'STATUS_NO_APPLICATION_PACKAGE')
    STATUS_FILE_METADATA_OPTIMIZATION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001AB, 'STATUS_FILE_METADATA_OPTIMIZATION_IN_PROGRESS')
    STATUS_NOT_SAME_OBJECT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001AC, 'STATUS_NOT_SAME_OBJECT')
    STATUS_FATAL_MEMORY_EXHAUSTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001AD, 'STATUS_FATAL_MEMORY_EXHAUSTION')
    STATUS_ERROR_PROCESS_NOT_IN_JOB: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001AE, 'STATUS_ERROR_PROCESS_NOT_IN_JOB')
    STATUS_CPU_SET_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001AF, 'STATUS_CPU_SET_INVALID')
    STATUS_IO_DEVICE_INVALID_DATA: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001B0, 'STATUS_IO_DEVICE_INVALID_DATA')
    STATUS_IO_UNALIGNED_WRITE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001B1, 'STATUS_IO_UNALIGNED_WRITE')
    STATUS_CONTROL_STACK_VIOLATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001B2, 'STATUS_CONTROL_STACK_VIOLATION')
    STATUS_WEAK_WHFBKEY_BLOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001B3, 'STATUS_WEAK_WHFBKEY_BLOCKED')
    STATUS_SERVER_TRANSPORT_CONFLICT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001B4, 'STATUS_SERVER_TRANSPORT_CONFLICT')
    STATUS_CERTIFICATE_VALIDATION_PREFERENCE_CONFLICT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00001B5, 'STATUS_CERTIFICATE_VALIDATION_PREFERENCE_CONFLICT')
    STATUS_NETWORK_OPEN_RESTRICTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000201, 'STATUS_NETWORK_OPEN_RESTRICTION')
    STATUS_NO_USER_SESSION_KEY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000202, 'STATUS_NO_USER_SESSION_KEY')
    STATUS_USER_SESSION_DELETED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000203, 'STATUS_USER_SESSION_DELETED')
    STATUS_RESOURCE_LANG_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000204, 'STATUS_RESOURCE_LANG_NOT_FOUND')
    STATUS_INSUFF_SERVER_RESOURCES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000205, 'STATUS_INSUFF_SERVER_RESOURCES')
    STATUS_INVALID_BUFFER_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000206, 'STATUS_INVALID_BUFFER_SIZE')
    STATUS_INVALID_ADDRESS_COMPONENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000207, 'STATUS_INVALID_ADDRESS_COMPONENT')
    STATUS_INVALID_ADDRESS_WILDCARD: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000208, 'STATUS_INVALID_ADDRESS_WILDCARD')
    STATUS_TOO_MANY_ADDRESSES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000209, 'STATUS_TOO_MANY_ADDRESSES')
    STATUS_ADDRESS_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC000020A, 'STATUS_ADDRESS_ALREADY_EXISTS')
    STATUS_ADDRESS_CLOSED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000020B, 'STATUS_ADDRESS_CLOSED')
    STATUS_CONNECTION_DISCONNECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000020C, 'STATUS_CONNECTION_DISCONNECTED')
    STATUS_CONNECTION_RESET: Final[ThirtyTwoBits] = cached_ntstatus(0xC000020D, 'STATUS_CONNECTION_RESET')
    STATUS_TOO_MANY_NODES: Final[ThirtyTwoBits] = cached_ntstatus(0xC000020E, 'STATUS_TOO_MANY_NODES')
    STATUS_TRANSACTION_ABORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000020F, 'STATUS_TRANSACTION_ABORTED')
    STATUS_TRANSACTION_TIMED_OUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000210, 'STATUS_TRANSACTION_TIMED_OUT')
    STATUS_TRANSACTION_NO_RELEASE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000211, 'STATUS_TRANSACTION_NO_RELEASE')
    STATUS_TRANSACTION_NO_MATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000212, 'STATUS_TRANSACTION_NO_MATCH')
    STATUS_TRANSACTION_RESPONDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000213, 'STATUS_TRANSACTION_RESPONDED')
    STATUS_TRANSACTION_INVALID_ID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000214, 'STATUS_TRANSACTION_INVALID_ID')
    STATUS_TRANSACTION_INVALID_TYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000215, 'STATUS_TRANSACTION_INVALID_TYPE')
    STATUS_NOT_SERVER_SESSION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000216, 'STATUS_NOT_SERVER_SESSION')
    STATUS_NOT_CLIENT_SESSION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000217, 'STATUS_NOT_CLIENT_SESSION')
    STATUS_CANNOT_LOAD_REGISTRY_FILE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000218, 'STATUS_CANNOT_LOAD_REGISTRY_FILE')
    STATUS_DEBUG_ATTACH_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000219, 'STATUS_DEBUG_ATTACH_FAILED')
    STATUS_SYSTEM_PROCESS_TERMINATED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000021A, 'STATUS_SYSTEM_PROCESS_TERMINATED')
    STATUS_DATA_NOT_ACCEPTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000021B, 'STATUS_DATA_NOT_ACCEPTED')
    STATUS_NO_BROWSER_SERVERS_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC000021C, 'STATUS_NO_BROWSER_SERVERS_FOUND')
    STATUS_VDM_HARD_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC000021D, 'STATUS_VDM_HARD_ERROR')
    STATUS_DRIVER_CANCEL_TIMEOUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000021E, 'STATUS_DRIVER_CANCEL_TIMEOUT')
    STATUS_REPLY_MESSAGE_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC000021F, 'STATUS_REPLY_MESSAGE_MISMATCH')
    STATUS_MAPPED_ALIGNMENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000220, 'STATUS_MAPPED_ALIGNMENT')
    STATUS_IMAGE_CHECKSUM_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000221, 'STATUS_IMAGE_CHECKSUM_MISMATCH')
    STATUS_LOST_WRITEBEHIND_DATA: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000222, 'STATUS_LOST_WRITEBEHIND_DATA')
    STATUS_CLIENT_SERVER_PARAMETERS_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000223, 'STATUS_CLIENT_SERVER_PARAMETERS_INVALID')
    STATUS_PASSWORD_MUST_CHANGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000224, 'STATUS_PASSWORD_MUST_CHANGE')
    STATUS_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000225, 'STATUS_NOT_FOUND')
    STATUS_NOT_TINY_STREAM: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000226, 'STATUS_NOT_TINY_STREAM')
    STATUS_RECOVERY_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000227, 'STATUS_RECOVERY_FAILURE')
    STATUS_STACK_OVERFLOW_READ: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000228, 'STATUS_STACK_OVERFLOW_READ')
    STATUS_FAIL_CHECK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000229, 'STATUS_FAIL_CHECK')
    STATUS_DUPLICATE_OBJECTID: Final[ThirtyTwoBits] = cached_ntstatus(0xC000022A, 'STATUS_DUPLICATE_OBJECTID')
    STATUS_OBJECTID_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC000022B, 'STATUS_OBJECTID_EXISTS')
    STATUS_CONVERT_TO_LARGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000022C, 'STATUS_CONVERT_TO_LARGE')
    STATUS_RETRY: Final[ThirtyTwoBits] = cached_ntstatus(0xC000022D, 'STATUS_RETRY')
    STATUS_FOUND_OUT_OF_SCOPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000022E, 'STATUS_FOUND_OUT_OF_SCOPE')
    STATUS_ALLOCATE_BUCKET: Final[ThirtyTwoBits] = cached_ntstatus(0xC000022F, 'STATUS_ALLOCATE_BUCKET')
    STATUS_PROPSET_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000230, 'STATUS_PROPSET_NOT_FOUND')
    STATUS_MARSHALL_OVERFLOW: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000231, 'STATUS_MARSHALL_OVERFLOW')
    STATUS_INVALID_VARIANT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000232, 'STATUS_INVALID_VARIANT')
    STATUS_DOMAIN_CONTROLLER_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000233, 'STATUS_DOMAIN_CONTROLLER_NOT_FOUND')
    STATUS_ACCOUNT_LOCKED_OUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000234, 'STATUS_ACCOUNT_LOCKED_OUT')
    STATUS_HANDLE_NOT_CLOSABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000235, 'STATUS_HANDLE_NOT_CLOSABLE')
    STATUS_CONNECTION_REFUSED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000236, 'STATUS_CONNECTION_REFUSED')
    STATUS_GRACEFUL_DISCONNECT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000237, 'STATUS_GRACEFUL_DISCONNECT')
    STATUS_ADDRESS_ALREADY_ASSOCIATED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000238, 'STATUS_ADDRESS_ALREADY_ASSOCIATED')
    STATUS_ADDRESS_NOT_ASSOCIATED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000239, 'STATUS_ADDRESS_NOT_ASSOCIATED')
    STATUS_CONNECTION_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC000023A, 'STATUS_CONNECTION_INVALID')
    STATUS_CONNECTION_ACTIVE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000023B, 'STATUS_CONNECTION_ACTIVE')
    STATUS_NETWORK_UNREACHABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000023C, 'STATUS_NETWORK_UNREACHABLE')
    STATUS_HOST_UNREACHABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000023D, 'STATUS_HOST_UNREACHABLE')
    STATUS_PROTOCOL_UNREACHABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000023E, 'STATUS_PROTOCOL_UNREACHABLE')
    STATUS_PORT_UNREACHABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000023F, 'STATUS_PORT_UNREACHABLE')
    STATUS_REQUEST_ABORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000240, 'STATUS_REQUEST_ABORTED')
    STATUS_CONNECTION_ABORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000241, 'STATUS_CONNECTION_ABORTED')
    STATUS_BAD_COMPRESSION_BUFFER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000242, 'STATUS_BAD_COMPRESSION_BUFFER')
    STATUS_USER_MAPPED_FILE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000243, 'STATUS_USER_MAPPED_FILE')
    STATUS_AUDIT_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000244, 'STATUS_AUDIT_FAILED')
    STATUS_TIMER_RESOLUTION_NOT_SET: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000245, 'STATUS_TIMER_RESOLUTION_NOT_SET')
    STATUS_CONNECTION_COUNT_LIMIT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000246, 'STATUS_CONNECTION_COUNT_LIMIT')
    STATUS_LOGIN_TIME_RESTRICTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000247, 'STATUS_LOGIN_TIME_RESTRICTION')
    STATUS_LOGIN_WKSTA_RESTRICTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000248, 'STATUS_LOGIN_WKSTA_RESTRICTION')
    STATUS_IMAGE_MP_UP_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000249, 'STATUS_IMAGE_MP_UP_MISMATCH')
    STATUS_INSUFFICIENT_LOGON_INFO: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000250, 'STATUS_INSUFFICIENT_LOGON_INFO')
    STATUS_BAD_DLL_ENTRYPOINT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000251, 'STATUS_BAD_DLL_ENTRYPOINT')
    STATUS_BAD_SERVICE_ENTRYPOINT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000252, 'STATUS_BAD_SERVICE_ENTRYPOINT')
    STATUS_LPC_REPLY_LOST: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000253, 'STATUS_LPC_REPLY_LOST')
    STATUS_IP_ADDRESS_CONFLICT1: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000254, 'STATUS_IP_ADDRESS_CONFLICT1')
    STATUS_IP_ADDRESS_CONFLICT2: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000255, 'STATUS_IP_ADDRESS_CONFLICT2')
    STATUS_REGISTRY_QUOTA_LIMIT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000256, 'STATUS_REGISTRY_QUOTA_LIMIT')
    STATUS_PATH_NOT_COVERED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000257, 'STATUS_PATH_NOT_COVERED')
    STATUS_NO_CALLBACK_ACTIVE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000258, 'STATUS_NO_CALLBACK_ACTIVE')
    STATUS_LICENSE_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000259, 'STATUS_LICENSE_QUOTA_EXCEEDED')
    STATUS_PWD_TOO_SHORT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000025A, 'STATUS_PWD_TOO_SHORT')
    STATUS_PWD_TOO_RECENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000025B, 'STATUS_PWD_TOO_RECENT')
    STATUS_PWD_HISTORY_CONFLICT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000025C, 'STATUS_PWD_HISTORY_CONFLICT')
    STATUS_PLUGPLAY_NO_DEVICE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000025E, 'STATUS_PLUGPLAY_NO_DEVICE')
    STATUS_UNSUPPORTED_COMPRESSION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000025F, 'STATUS_UNSUPPORTED_COMPRESSION')
    STATUS_INVALID_HW_PROFILE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000260, 'STATUS_INVALID_HW_PROFILE')
    STATUS_INVALID_PLUGPLAY_DEVICE_PATH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000261, 'STATUS_INVALID_PLUGPLAY_DEVICE_PATH')
    STATUS_DRIVER_ORDINAL_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000262, 'STATUS_DRIVER_ORDINAL_NOT_FOUND')
    STATUS_DRIVER_ENTRYPOINT_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000263, 'STATUS_DRIVER_ENTRYPOINT_NOT_FOUND')
    STATUS_RESOURCE_NOT_OWNED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000264, 'STATUS_RESOURCE_NOT_OWNED')
    STATUS_TOO_MANY_LINKS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000265, 'STATUS_TOO_MANY_LINKS')
    STATUS_QUOTA_LIST_INCONSISTENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000266, 'STATUS_QUOTA_LIST_INCONSISTENT')
    STATUS_FILE_IS_OFFLINE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000267, 'STATUS_FILE_IS_OFFLINE')
    STATUS_EVALUATION_EXPIRATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000268, 'STATUS_EVALUATION_EXPIRATION')
    STATUS_ILLEGAL_DLL_RELOCATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000269, 'STATUS_ILLEGAL_DLL_RELOCATION')
    STATUS_LICENSE_VIOLATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000026A, 'STATUS_LICENSE_VIOLATION')
    STATUS_DLL_INIT_FAILED_LOGOFF: Final[ThirtyTwoBits] = cached_ntstatus(0xC000026B, 'STATUS_DLL_INIT_FAILED_LOGOFF')
    STATUS_DRIVER_UNABLE_TO_LOAD: Final[ThirtyTwoBits] = cached_ntstatus(0xC000026C, 'STATUS_DRIVER_UNABLE_TO_LOAD')
    STATUS_DFS_UNAVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000026D, 'STATUS_DFS_UNAVAILABLE')
    STATUS_VOLUME_DISMOUNTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000026E, 'STATUS_VOLUME_DISMOUNTED')
    STATUS_WX86_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC000026F, 'STATUS_WX86_INTERNAL_ERROR')
    STATUS_WX86_FLOAT_STACK_CHECK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000270, 'STATUS_WX86_FLOAT_STACK_CHECK')
    STATUS_VALIDATE_CONTINUE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000271, 'STATUS_VALIDATE_CONTINUE')
    STATUS_NO_MATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000272, 'STATUS_NO_MATCH')
    STATUS_NO_MORE_MATCHES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000273, 'STATUS_NO_MORE_MATCHES')
    STATUS_NOT_A_REPARSE_POINT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000275, 'STATUS_NOT_A_REPARSE_POINT')
    STATUS_IO_REPARSE_TAG_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000276, 'STATUS_IO_REPARSE_TAG_INVALID')
    STATUS_IO_REPARSE_TAG_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000277, 'STATUS_IO_REPARSE_TAG_MISMATCH')
    STATUS_IO_REPARSE_DATA_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000278, 'STATUS_IO_REPARSE_DATA_INVALID')
    STATUS_IO_REPARSE_TAG_NOT_HANDLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000279, 'STATUS_IO_REPARSE_TAG_NOT_HANDLED')
    STATUS_PWD_TOO_LONG: Final[ThirtyTwoBits] = cached_ntstatus(0xC000027A, 'STATUS_PWD_TOO_LONG')
    STATUS_STOWED_EXCEPTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000027B, 'STATUS_STOWED_EXCEPTION')
    STATUS_CONTEXT_STOWED_EXCEPTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000027C, 'STATUS_CONTEXT_STOWED_EXCEPTION')
    STATUS_REPARSE_POINT_NOT_RESOLVED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000280, 'STATUS_REPARSE_POINT_NOT_RESOLVED')
    STATUS_DIRECTORY_IS_A_REPARSE_POINT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000281, 'STATUS_DIRECTORY_IS_A_REPARSE_POINT')
    STATUS_RANGE_LIST_CONFLICT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000282, 'STATUS_RANGE_LIST_CONFLICT')
    STATUS_SOURCE_ELEMENT_EMPTY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000283, 'STATUS_SOURCE_ELEMENT_EMPTY')
    STATUS_DESTINATION_ELEMENT_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000284, 'STATUS_DESTINATION_ELEMENT_FULL')
    STATUS_ILLEGAL_ELEMENT_ADDRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000285, 'STATUS_ILLEGAL_ELEMENT_ADDRESS')
    STATUS_MAGAZINE_NOT_PRESENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000286, 'STATUS_MAGAZINE_NOT_PRESENT')
    STATUS_REINITIALIZATION_NEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000287, 'STATUS_REINITIALIZATION_NEEDED')
    STATUS_ENCRYPTION_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000028A, 'STATUS_ENCRYPTION_FAILED')
    STATUS_DECRYPTION_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000028B, 'STATUS_DECRYPTION_FAILED')
    STATUS_RANGE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC000028C, 'STATUS_RANGE_NOT_FOUND')
    STATUS_NO_RECOVERY_POLICY: Final[ThirtyTwoBits] = cached_ntstatus(0xC000028D, 'STATUS_NO_RECOVERY_POLICY')
    STATUS_NO_EFS: Final[ThirtyTwoBits] = cached_ntstatus(0xC000028E, 'STATUS_NO_EFS')
    STATUS_WRONG_EFS: Final[ThirtyTwoBits] = cached_ntstatus(0xC000028F, 'STATUS_WRONG_EFS')
    STATUS_NO_USER_KEYS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000290, 'STATUS_NO_USER_KEYS')
    STATUS_FILE_NOT_ENCRYPTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000291, 'STATUS_FILE_NOT_ENCRYPTED')
    STATUS_NOT_EXPORT_FORMAT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000292, 'STATUS_NOT_EXPORT_FORMAT')
    STATUS_FILE_ENCRYPTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000293, 'STATUS_FILE_ENCRYPTED')
    STATUS_WMI_GUID_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000295, 'STATUS_WMI_GUID_NOT_FOUND')
    STATUS_WMI_INSTANCE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000296, 'STATUS_WMI_INSTANCE_NOT_FOUND')
    STATUS_WMI_ITEMID_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000297, 'STATUS_WMI_ITEMID_NOT_FOUND')
    STATUS_WMI_TRY_AGAIN: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000298, 'STATUS_WMI_TRY_AGAIN')
    STATUS_SHARED_POLICY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000299, 'STATUS_SHARED_POLICY')
    STATUS_POLICY_OBJECT_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC000029A, 'STATUS_POLICY_OBJECT_NOT_FOUND')
    STATUS_POLICY_ONLY_IN_DS: Final[ThirtyTwoBits] = cached_ntstatus(0xC000029B, 'STATUS_POLICY_ONLY_IN_DS')
    STATUS_VOLUME_NOT_UPGRADED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000029C, 'STATUS_VOLUME_NOT_UPGRADED')
    STATUS_REMOTE_STORAGE_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000029D, 'STATUS_REMOTE_STORAGE_NOT_ACTIVE')
    STATUS_REMOTE_STORAGE_MEDIA_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC000029E, 'STATUS_REMOTE_STORAGE_MEDIA_ERROR')
    STATUS_NO_TRACKING_SERVICE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000029F, 'STATUS_NO_TRACKING_SERVICE')
    STATUS_SERVER_SID_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002A0, 'STATUS_SERVER_SID_MISMATCH')
    STATUS_DS_NO_ATTRIBUTE_OR_VALUE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002A1, 'STATUS_DS_NO_ATTRIBUTE_OR_VALUE')
    STATUS_DS_INVALID_ATTRIBUTE_SYNTAX: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002A2, 'STATUS_DS_INVALID_ATTRIBUTE_SYNTAX')
    STATUS_DS_ATTRIBUTE_TYPE_UNDEFINED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002A3, 'STATUS_DS_ATTRIBUTE_TYPE_UNDEFINED')
    STATUS_DS_ATTRIBUTE_OR_VALUE_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002A4, 'STATUS_DS_ATTRIBUTE_OR_VALUE_EXISTS')
    STATUS_DS_BUSY: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002A5, 'STATUS_DS_BUSY')
    STATUS_DS_UNAVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002A6, 'STATUS_DS_UNAVAILABLE')
    STATUS_DS_NO_RIDS_ALLOCATED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002A7, 'STATUS_DS_NO_RIDS_ALLOCATED')
    STATUS_DS_NO_MORE_RIDS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002A8, 'STATUS_DS_NO_MORE_RIDS')
    STATUS_DS_INCORRECT_ROLE_OWNER: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002A9, 'STATUS_DS_INCORRECT_ROLE_OWNER')
    STATUS_DS_RIDMGR_INIT_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002AA, 'STATUS_DS_RIDMGR_INIT_ERROR')
    STATUS_DS_OBJ_CLASS_VIOLATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002AB, 'STATUS_DS_OBJ_CLASS_VIOLATION')
    STATUS_DS_CANT_ON_NON_LEAF: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002AC, 'STATUS_DS_CANT_ON_NON_LEAF')
    STATUS_DS_CANT_ON_RDN: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002AD, 'STATUS_DS_CANT_ON_RDN')
    STATUS_DS_CANT_MOD_OBJ_CLASS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002AE, 'STATUS_DS_CANT_MOD_OBJ_CLASS')
    STATUS_DS_CROSS_DOM_MOVE_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002AF, 'STATUS_DS_CROSS_DOM_MOVE_FAILED')
    STATUS_DS_GC_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002B0, 'STATUS_DS_GC_NOT_AVAILABLE')
    STATUS_DIRECTORY_SERVICE_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002B1, 'STATUS_DIRECTORY_SERVICE_REQUIRED')
    STATUS_REPARSE_ATTRIBUTE_CONFLICT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002B2, 'STATUS_REPARSE_ATTRIBUTE_CONFLICT')
    STATUS_CANT_ENABLE_DENY_ONLY: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002B3, 'STATUS_CANT_ENABLE_DENY_ONLY')
    STATUS_FLOAT_MULTIPLE_FAULTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002B4, 'STATUS_FLOAT_MULTIPLE_FAULTS')
    STATUS_FLOAT_MULTIPLE_TRAPS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002B5, 'STATUS_FLOAT_MULTIPLE_TRAPS')
    STATUS_DEVICE_REMOVED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002B6, 'STATUS_DEVICE_REMOVED')
    STATUS_JOURNAL_DELETE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002B7, 'STATUS_JOURNAL_DELETE_IN_PROGRESS')
    STATUS_JOURNAL_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002B8, 'STATUS_JOURNAL_NOT_ACTIVE')
    STATUS_NOINTERFACE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002B9, 'STATUS_NOINTERFACE')
    STATUS_DS_RIDMGR_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002BA, 'STATUS_DS_RIDMGR_DISABLED')
    STATUS_DS_ADMIN_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002C1, 'STATUS_DS_ADMIN_LIMIT_EXCEEDED')
    STATUS_DRIVER_FAILED_SLEEP: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002C2, 'STATUS_DRIVER_FAILED_SLEEP')
    STATUS_MUTUAL_AUTHENTICATION_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002C3, 'STATUS_MUTUAL_AUTHENTICATION_FAILED')
    STATUS_CORRUPT_SYSTEM_FILE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002C4, 'STATUS_CORRUPT_SYSTEM_FILE')
    STATUS_DATATYPE_MISALIGNMENT_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002C5, 'STATUS_DATATYPE_MISALIGNMENT_ERROR')
    STATUS_WMI_READ_ONLY: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002C6, 'STATUS_WMI_READ_ONLY')
    STATUS_WMI_SET_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002C7, 'STATUS_WMI_SET_FAILURE')
    STATUS_COMMITMENT_MINIMUM: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002C8, 'STATUS_COMMITMENT_MINIMUM')
    STATUS_REG_NAT_CONSUMPTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002C9, 'STATUS_REG_NAT_CONSUMPTION')
    STATUS_TRANSPORT_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002CA, 'STATUS_TRANSPORT_FULL')
    STATUS_DS_SAM_INIT_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002CB, 'STATUS_DS_SAM_INIT_FAILURE')
    STATUS_ONLY_IF_CONNECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002CC, 'STATUS_ONLY_IF_CONNECTED')
    STATUS_DS_SENSITIVE_GROUP_VIOLATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002CD, 'STATUS_DS_SENSITIVE_GROUP_VIOLATION')
    STATUS_PNP_RESTART_ENUMERATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002CE, 'STATUS_PNP_RESTART_ENUMERATION')
    STATUS_JOURNAL_ENTRY_DELETED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002CF, 'STATUS_JOURNAL_ENTRY_DELETED')
    STATUS_DS_CANT_MOD_PRIMARYGROUPID: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002D0, 'STATUS_DS_CANT_MOD_PRIMARYGROUPID')
    STATUS_SYSTEM_IMAGE_BAD_SIGNATURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002D1, 'STATUS_SYSTEM_IMAGE_BAD_SIGNATURE')
    STATUS_PNP_REBOOT_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002D2, 'STATUS_PNP_REBOOT_REQUIRED')
    STATUS_POWER_STATE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002D3, 'STATUS_POWER_STATE_INVALID')
    STATUS_DS_INVALID_GROUP_TYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002D4, 'STATUS_DS_INVALID_GROUP_TYPE')
    STATUS_DS_NO_NEST_GLOBALGROUP_IN_MIXEDDOMAIN: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002D5, 'STATUS_DS_NO_NEST_GLOBALGROUP_IN_MIXEDDOMAIN')
    STATUS_DS_NO_NEST_LOCALGROUP_IN_MIXEDDOMAIN: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002D6, 'STATUS_DS_NO_NEST_LOCALGROUP_IN_MIXEDDOMAIN')
    STATUS_DS_GLOBAL_CANT_HAVE_LOCAL_MEMBER: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002D7, 'STATUS_DS_GLOBAL_CANT_HAVE_LOCAL_MEMBER')
    STATUS_DS_GLOBAL_CANT_HAVE_UNIVERSAL_MEMBER: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002D8, 'STATUS_DS_GLOBAL_CANT_HAVE_UNIVERSAL_MEMBER')
    STATUS_DS_UNIVERSAL_CANT_HAVE_LOCAL_MEMBER: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002D9, 'STATUS_DS_UNIVERSAL_CANT_HAVE_LOCAL_MEMBER')
    STATUS_DS_GLOBAL_CANT_HAVE_CROSSDOMAIN_MEMBER: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002DA, 'STATUS_DS_GLOBAL_CANT_HAVE_CROSSDOMAIN_MEMBER')
    STATUS_DS_LOCAL_CANT_HAVE_CROSSDOMAIN_LOCAL_MEMBER: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002DB, 'STATUS_DS_LOCAL_CANT_HAVE_CROSSDOMAIN_LOCAL_MEMBER')
    STATUS_DS_HAVE_PRIMARY_MEMBERS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002DC, 'STATUS_DS_HAVE_PRIMARY_MEMBERS')
    STATUS_WMI_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002DD, 'STATUS_WMI_NOT_SUPPORTED')
    STATUS_INSUFFICIENT_POWER: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002DE, 'STATUS_INSUFFICIENT_POWER')
    STATUS_SAM_NEED_BOOTKEY_PASSWORD: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002DF, 'STATUS_SAM_NEED_BOOTKEY_PASSWORD')
    STATUS_SAM_NEED_BOOTKEY_FLOPPY: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002E0, 'STATUS_SAM_NEED_BOOTKEY_FLOPPY')
    STATUS_DS_CANT_START: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002E1, 'STATUS_DS_CANT_START')
    STATUS_DS_INIT_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002E2, 'STATUS_DS_INIT_FAILURE')
    STATUS_SAM_INIT_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002E3, 'STATUS_SAM_INIT_FAILURE')
    STATUS_DS_GC_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002E4, 'STATUS_DS_GC_REQUIRED')
    STATUS_DS_LOCAL_MEMBER_OF_LOCAL_ONLY: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002E5, 'STATUS_DS_LOCAL_MEMBER_OF_LOCAL_ONLY')
    STATUS_DS_NO_FPO_IN_UNIVERSAL_GROUPS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002E6, 'STATUS_DS_NO_FPO_IN_UNIVERSAL_GROUPS')
    STATUS_DS_MACHINE_ACCOUNT_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002E7, 'STATUS_DS_MACHINE_ACCOUNT_QUOTA_EXCEEDED')
    STATUS_MULTIPLE_FAULT_VIOLATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002E8, 'STATUS_MULTIPLE_FAULT_VIOLATION')
    STATUS_CURRENT_DOMAIN_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002E9, 'STATUS_CURRENT_DOMAIN_NOT_ALLOWED')
    STATUS_CANNOT_MAKE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002EA, 'STATUS_CANNOT_MAKE')
    STATUS_SYSTEM_SHUTDOWN: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002EB, 'STATUS_SYSTEM_SHUTDOWN')
    STATUS_DS_INIT_FAILURE_CONSOLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002EC, 'STATUS_DS_INIT_FAILURE_CONSOLE')
    STATUS_DS_SAM_INIT_FAILURE_CONSOLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002ED, 'STATUS_DS_SAM_INIT_FAILURE_CONSOLE')
    STATUS_UNFINISHED_CONTEXT_DELETED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002EE, 'STATUS_UNFINISHED_CONTEXT_DELETED')
    STATUS_NO_TGT_REPLY: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002EF, 'STATUS_NO_TGT_REPLY')
    STATUS_OBJECTID_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002F0, 'STATUS_OBJECTID_NOT_FOUND')
    STATUS_NO_IP_ADDRESSES: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002F1, 'STATUS_NO_IP_ADDRESSES')
    STATUS_WRONG_CREDENTIAL_HANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002F2, 'STATUS_WRONG_CREDENTIAL_HANDLE')
    STATUS_CRYPTO_SYSTEM_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002F3, 'STATUS_CRYPTO_SYSTEM_INVALID')
    STATUS_MAX_REFERRALS_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002F4, 'STATUS_MAX_REFERRALS_EXCEEDED')
    STATUS_MUST_BE_KDC: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002F5, 'STATUS_MUST_BE_KDC')
    STATUS_STRONG_CRYPTO_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002F6, 'STATUS_STRONG_CRYPTO_NOT_SUPPORTED')
    STATUS_TOO_MANY_PRINCIPALS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002F7, 'STATUS_TOO_MANY_PRINCIPALS')
    STATUS_NO_PA_DATA: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002F8, 'STATUS_NO_PA_DATA')
    STATUS_PKINIT_NAME_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002F9, 'STATUS_PKINIT_NAME_MISMATCH')
    STATUS_SMARTCARD_LOGON_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002FA, 'STATUS_SMARTCARD_LOGON_REQUIRED')
    STATUS_KDC_INVALID_REQUEST: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002FB, 'STATUS_KDC_INVALID_REQUEST')
    STATUS_KDC_UNABLE_TO_REFER: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002FC, 'STATUS_KDC_UNABLE_TO_REFER')
    STATUS_KDC_UNKNOWN_ETYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002FD, 'STATUS_KDC_UNKNOWN_ETYPE')
    STATUS_SHUTDOWN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002FE, 'STATUS_SHUTDOWN_IN_PROGRESS')
    STATUS_SERVER_SHUTDOWN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00002FF, 'STATUS_SERVER_SHUTDOWN_IN_PROGRESS')
    STATUS_NOT_SUPPORTED_ON_SBS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000300, 'STATUS_NOT_SUPPORTED_ON_SBS')
    STATUS_WMI_GUID_DISCONNECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000301, 'STATUS_WMI_GUID_DISCONNECTED')
    STATUS_WMI_ALREADY_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000302, 'STATUS_WMI_ALREADY_DISABLED')
    STATUS_WMI_ALREADY_ENABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000303, 'STATUS_WMI_ALREADY_ENABLED')
    STATUS_MFT_TOO_FRAGMENTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000304, 'STATUS_MFT_TOO_FRAGMENTED')
    STATUS_COPY_PROTECTION_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000305, 'STATUS_COPY_PROTECTION_FAILURE')
    STATUS_CSS_AUTHENTICATION_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000306, 'STATUS_CSS_AUTHENTICATION_FAILURE')
    STATUS_CSS_KEY_NOT_PRESENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000307, 'STATUS_CSS_KEY_NOT_PRESENT')
    STATUS_CSS_KEY_NOT_ESTABLISHED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000308, 'STATUS_CSS_KEY_NOT_ESTABLISHED')
    STATUS_CSS_SCRAMBLED_SECTOR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000309, 'STATUS_CSS_SCRAMBLED_SECTOR')
    STATUS_CSS_REGION_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC000030A, 'STATUS_CSS_REGION_MISMATCH')
    STATUS_CSS_RESETS_EXHAUSTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000030B, 'STATUS_CSS_RESETS_EXHAUSTED')
    STATUS_PASSWORD_CHANGE_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000030C, 'STATUS_PASSWORD_CHANGE_REQUIRED')
    STATUS_LOST_MODE_LOGON_RESTRICTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000030D, 'STATUS_LOST_MODE_LOGON_RESTRICTION')
    STATUS_PKINIT_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000320, 'STATUS_PKINIT_FAILURE')
    STATUS_SMARTCARD_SUBSYSTEM_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000321, 'STATUS_SMARTCARD_SUBSYSTEM_FAILURE')
    STATUS_NO_KERB_KEY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000322, 'STATUS_NO_KERB_KEY')
    STATUS_HOST_DOWN: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000350, 'STATUS_HOST_DOWN')
    STATUS_UNSUPPORTED_PREAUTH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000351, 'STATUS_UNSUPPORTED_PREAUTH')
    STATUS_EFS_ALG_BLOB_TOO_BIG: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000352, 'STATUS_EFS_ALG_BLOB_TOO_BIG')
    STATUS_PORT_NOT_SET: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000353, 'STATUS_PORT_NOT_SET')
    STATUS_DEBUGGER_INACTIVE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000354, 'STATUS_DEBUGGER_INACTIVE')
    STATUS_DS_VERSION_CHECK_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000355, 'STATUS_DS_VERSION_CHECK_FAILURE')
    STATUS_AUDITING_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000356, 'STATUS_AUDITING_DISABLED')
    STATUS_PRENT4_MACHINE_ACCOUNT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000357, 'STATUS_PRENT4_MACHINE_ACCOUNT')
    STATUS_DS_AG_CANT_HAVE_UNIVERSAL_MEMBER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000358, 'STATUS_DS_AG_CANT_HAVE_UNIVERSAL_MEMBER')
    STATUS_INVALID_IMAGE_WIN_32: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000359, 'STATUS_INVALID_IMAGE_WIN_32')
    STATUS_INVALID_IMAGE_WIN_64: Final[ThirtyTwoBits] = cached_ntstatus(0xC000035A, 'STATUS_INVALID_IMAGE_WIN_64')
    STATUS_BAD_BINDINGS: Final[ThirtyTwoBits] = cached_ntstatus(0xC000035B, 'STATUS_BAD_BINDINGS')
    STATUS_NETWORK_SESSION_EXPIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000035C, 'STATUS_NETWORK_SESSION_EXPIRED')
    STATUS_APPHELP_BLOCK: Final[ThirtyTwoBits] = cached_ntstatus(0xC000035D, 'STATUS_APPHELP_BLOCK')
    STATUS_ALL_SIDS_FILTERED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000035E, 'STATUS_ALL_SIDS_FILTERED')
    STATUS_NOT_SAFE_MODE_DRIVER: Final[ThirtyTwoBits] = cached_ntstatus(0xC000035F, 'STATUS_NOT_SAFE_MODE_DRIVER')
    STATUS_ACCESS_DISABLED_BY_POLICY_DEFAULT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000361, 'STATUS_ACCESS_DISABLED_BY_POLICY_DEFAULT')
    STATUS_ACCESS_DISABLED_BY_POLICY_PATH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000362, 'STATUS_ACCESS_DISABLED_BY_POLICY_PATH')
    STATUS_ACCESS_DISABLED_BY_POLICY_PUBLISHER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000363, 'STATUS_ACCESS_DISABLED_BY_POLICY_PUBLISHER')
    STATUS_ACCESS_DISABLED_BY_POLICY_OTHER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000364, 'STATUS_ACCESS_DISABLED_BY_POLICY_OTHER')
    STATUS_FAILED_DRIVER_ENTRY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000365, 'STATUS_FAILED_DRIVER_ENTRY')
    STATUS_DEVICE_ENUMERATION_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000366, 'STATUS_DEVICE_ENUMERATION_ERROR')
    STATUS_MOUNT_POINT_NOT_RESOLVED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000368, 'STATUS_MOUNT_POINT_NOT_RESOLVED')
    STATUS_INVALID_DEVICE_OBJECT_PARAMETER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000369, 'STATUS_INVALID_DEVICE_OBJECT_PARAMETER')
    STATUS_MCA_OCCURED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000036A, 'STATUS_MCA_OCCURED')
    STATUS_DRIVER_BLOCKED_CRITICAL: Final[ThirtyTwoBits] = cached_ntstatus(0xC000036B, 'STATUS_DRIVER_BLOCKED_CRITICAL')
    STATUS_DRIVER_BLOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000036C, 'STATUS_DRIVER_BLOCKED')
    STATUS_DRIVER_DATABASE_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC000036D, 'STATUS_DRIVER_DATABASE_ERROR')
    STATUS_SYSTEM_HIVE_TOO_LARGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000036E, 'STATUS_SYSTEM_HIVE_TOO_LARGE')
    STATUS_INVALID_IMPORT_OF_NON_DLL: Final[ThirtyTwoBits] = cached_ntstatus(0xC000036F, 'STATUS_INVALID_IMPORT_OF_NON_DLL')
    STATUS_NO_SECRETS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000371, 'STATUS_NO_SECRETS')
    STATUS_ACCESS_DISABLED_NO_SAFER_UI_BY_POLICY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000372, 'STATUS_ACCESS_DISABLED_NO_SAFER_UI_BY_POLICY')
    STATUS_FAILED_STACK_SWITCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000373, 'STATUS_FAILED_STACK_SWITCH')
    STATUS_HEAP_CORRUPTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000374, 'STATUS_HEAP_CORRUPTION')
    STATUS_SMARTCARD_WRONG_PIN: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000380, 'STATUS_SMARTCARD_WRONG_PIN')
    STATUS_SMARTCARD_CARD_BLOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000381, 'STATUS_SMARTCARD_CARD_BLOCKED')
    STATUS_SMARTCARD_CARD_NOT_AUTHENTICATED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000382, 'STATUS_SMARTCARD_CARD_NOT_AUTHENTICATED')
    STATUS_SMARTCARD_NO_CARD: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000383, 'STATUS_SMARTCARD_NO_CARD')
    STATUS_SMARTCARD_NO_KEY_CONTAINER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000384, 'STATUS_SMARTCARD_NO_KEY_CONTAINER')
    STATUS_SMARTCARD_NO_CERTIFICATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000385, 'STATUS_SMARTCARD_NO_CERTIFICATE')
    STATUS_SMARTCARD_NO_KEYSET: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000386, 'STATUS_SMARTCARD_NO_KEYSET')
    STATUS_SMARTCARD_IO_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000387, 'STATUS_SMARTCARD_IO_ERROR')
    STATUS_DOWNGRADE_DETECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000388, 'STATUS_DOWNGRADE_DETECTED')
    STATUS_SMARTCARD_CERT_REVOKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000389, 'STATUS_SMARTCARD_CERT_REVOKED')
    STATUS_ISSUING_CA_UNTRUSTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000038A, 'STATUS_ISSUING_CA_UNTRUSTED')
    STATUS_REVOCATION_OFFLINE_C: Final[ThirtyTwoBits] = cached_ntstatus(0xC000038B, 'STATUS_REVOCATION_OFFLINE_C')
    STATUS_PKINIT_CLIENT_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000038C, 'STATUS_PKINIT_CLIENT_FAILURE')
    STATUS_SMARTCARD_CERT_EXPIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000038D, 'STATUS_SMARTCARD_CERT_EXPIRED')
    STATUS_DRIVER_FAILED_PRIOR_UNLOAD: Final[ThirtyTwoBits] = cached_ntstatus(0xC000038E, 'STATUS_DRIVER_FAILED_PRIOR_UNLOAD')
    STATUS_SMARTCARD_SILENT_CONTEXT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000038F, 'STATUS_SMARTCARD_SILENT_CONTEXT')
    STATUS_PER_USER_TRUST_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000401, 'STATUS_PER_USER_TRUST_QUOTA_EXCEEDED')
    STATUS_ALL_USER_TRUST_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000402, 'STATUS_ALL_USER_TRUST_QUOTA_EXCEEDED')
    STATUS_USER_DELETE_TRUST_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000403, 'STATUS_USER_DELETE_TRUST_QUOTA_EXCEEDED')
    STATUS_DS_NAME_NOT_UNIQUE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000404, 'STATUS_DS_NAME_NOT_UNIQUE')
    STATUS_DS_DUPLICATE_ID_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000405, 'STATUS_DS_DUPLICATE_ID_FOUND')
    STATUS_DS_GROUP_CONVERSION_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000406, 'STATUS_DS_GROUP_CONVERSION_ERROR')
    STATUS_VOLSNAP_PREPARE_HIBERNATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000407, 'STATUS_VOLSNAP_PREPARE_HIBERNATE')
    STATUS_USER2USER_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000408, 'STATUS_USER2USER_REQUIRED')
    STATUS_STACK_BUFFER_OVERRUN: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000409, 'STATUS_STACK_BUFFER_OVERRUN')
    STATUS_NO_S4U_PROT_SUPPORT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000040A, 'STATUS_NO_S4U_PROT_SUPPORT')
    STATUS_CROSSREALM_DELEGATION_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000040B, 'STATUS_CROSSREALM_DELEGATION_FAILURE')
    STATUS_REVOCATION_OFFLINE_KDC: Final[ThirtyTwoBits] = cached_ntstatus(0xC000040C, 'STATUS_REVOCATION_OFFLINE_KDC')
    STATUS_ISSUING_CA_UNTRUSTED_KDC: Final[ThirtyTwoBits] = cached_ntstatus(0xC000040D, 'STATUS_ISSUING_CA_UNTRUSTED_KDC')
    STATUS_KDC_CERT_EXPIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000040E, 'STATUS_KDC_CERT_EXPIRED')
    STATUS_KDC_CERT_REVOKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000040F, 'STATUS_KDC_CERT_REVOKED')
    STATUS_PARAMETER_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000410, 'STATUS_PARAMETER_QUOTA_EXCEEDED')
    STATUS_HIBERNATION_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000411, 'STATUS_HIBERNATION_FAILURE')
    STATUS_DELAY_LOAD_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000412, 'STATUS_DELAY_LOAD_FAILED')
    STATUS_AUTHENTICATION_FIREWALL_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000413, 'STATUS_AUTHENTICATION_FIREWALL_FAILED')
    STATUS_VDM_DISALLOWED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000414, 'STATUS_VDM_DISALLOWED')
    STATUS_HUNG_DISPLAY_DRIVER_THREAD: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000415, 'STATUS_HUNG_DISPLAY_DRIVER_THREAD')
    STATUS_INSUFFICIENT_RESOURCE_FOR_SPECIFIED_SHARED_SECTION_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000416, 'STATUS_INSUFFICIENT_RESOURCE_FOR_SPECIFIED_SHARED_SECTION_SIZE')
    STATUS_INVALID_CRUNTIME_PARAMETER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000417, 'STATUS_INVALID_CRUNTIME_PARAMETER')
    STATUS_NTLM_BLOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000418, 'STATUS_NTLM_BLOCKED')
    STATUS_DS_SRC_SID_EXISTS_IN_FOREST: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000419, 'STATUS_DS_SRC_SID_EXISTS_IN_FOREST')
    STATUS_DS_DOMAIN_NAME_EXISTS_IN_FOREST: Final[ThirtyTwoBits] = cached_ntstatus(0xC000041A, 'STATUS_DS_DOMAIN_NAME_EXISTS_IN_FOREST')
    STATUS_DS_FLAT_NAME_EXISTS_IN_FOREST: Final[ThirtyTwoBits] = cached_ntstatus(0xC000041B, 'STATUS_DS_FLAT_NAME_EXISTS_IN_FOREST')
    STATUS_INVALID_USER_PRINCIPAL_NAME: Final[ThirtyTwoBits] = cached_ntstatus(0xC000041C, 'STATUS_INVALID_USER_PRINCIPAL_NAME')
    STATUS_FATAL_USER_CALLBACK_EXCEPTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000041D, 'STATUS_FATAL_USER_CALLBACK_EXCEPTION')
    STATUS_ASSERTION_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000420, 'STATUS_ASSERTION_FAILURE')
    STATUS_VERIFIER_STOP: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000421, 'STATUS_VERIFIER_STOP')
    STATUS_CALLBACK_POP_STACK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000423, 'STATUS_CALLBACK_POP_STACK')
    STATUS_INCOMPATIBLE_DRIVER_BLOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000424, 'STATUS_INCOMPATIBLE_DRIVER_BLOCKED')
    STATUS_HIVE_UNLOADED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000425, 'STATUS_HIVE_UNLOADED')
    STATUS_COMPRESSION_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000426, 'STATUS_COMPRESSION_DISABLED')
    STATUS_FILE_SYSTEM_LIMITATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000427, 'STATUS_FILE_SYSTEM_LIMITATION')
    STATUS_INVALID_IMAGE_HASH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000428, 'STATUS_INVALID_IMAGE_HASH')
    STATUS_NOT_CAPABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000429, 'STATUS_NOT_CAPABLE')
    STATUS_REQUEST_OUT_OF_SEQUENCE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000042A, 'STATUS_REQUEST_OUT_OF_SEQUENCE')
    STATUS_IMPLEMENTATION_LIMIT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000042B, 'STATUS_IMPLEMENTATION_LIMIT')
    STATUS_ELEVATION_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000042C, 'STATUS_ELEVATION_REQUIRED')
    STATUS_NO_SECURITY_CONTEXT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000042D, 'STATUS_NO_SECURITY_CONTEXT')
    STATUS_PKU2U_CERT_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000042F, 'STATUS_PKU2U_CERT_FAILURE')
    STATUS_BEYOND_VDL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000432, 'STATUS_BEYOND_VDL')
    STATUS_ENCOUNTERED_WRITE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000433, 'STATUS_ENCOUNTERED_WRITE_IN_PROGRESS')
    STATUS_PTE_CHANGED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000434, 'STATUS_PTE_CHANGED')
    STATUS_PURGE_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000435, 'STATUS_PURGE_FAILED')
    STATUS_CRED_REQUIRES_CONFIRMATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000440, 'STATUS_CRED_REQUIRES_CONFIRMATION')
    STATUS_CS_ENCRYPTION_INVALID_SERVER_RESPONSE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000441, 'STATUS_CS_ENCRYPTION_INVALID_SERVER_RESPONSE')
    STATUS_CS_ENCRYPTION_UNSUPPORTED_SERVER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000442, 'STATUS_CS_ENCRYPTION_UNSUPPORTED_SERVER')
    STATUS_CS_ENCRYPTION_EXISTING_ENCRYPTED_FILE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000443, 'STATUS_CS_ENCRYPTION_EXISTING_ENCRYPTED_FILE')
    STATUS_CS_ENCRYPTION_NEW_ENCRYPTED_FILE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000444, 'STATUS_CS_ENCRYPTION_NEW_ENCRYPTED_FILE')
    STATUS_CS_ENCRYPTION_FILE_NOT_CSE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000445, 'STATUS_CS_ENCRYPTION_FILE_NOT_CSE')
    STATUS_INVALID_LABEL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000446, 'STATUS_INVALID_LABEL')
    STATUS_DRIVER_PROCESS_TERMINATED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000450, 'STATUS_DRIVER_PROCESS_TERMINATED')
    STATUS_AMBIGUOUS_SYSTEM_DEVICE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000451, 'STATUS_AMBIGUOUS_SYSTEM_DEVICE')
    STATUS_SYSTEM_DEVICE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000452, 'STATUS_SYSTEM_DEVICE_NOT_FOUND')
    STATUS_RESTART_BOOT_APPLICATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000453, 'STATUS_RESTART_BOOT_APPLICATION')
    STATUS_INSUFFICIENT_NVRAM_RESOURCES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000454, 'STATUS_INSUFFICIENT_NVRAM_RESOURCES')
    STATUS_INVALID_SESSION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000455, 'STATUS_INVALID_SESSION')
    STATUS_THREAD_ALREADY_IN_SESSION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000456, 'STATUS_THREAD_ALREADY_IN_SESSION')
    STATUS_THREAD_NOT_IN_SESSION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000457, 'STATUS_THREAD_NOT_IN_SESSION')
    STATUS_INVALID_WEIGHT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000458, 'STATUS_INVALID_WEIGHT')
    STATUS_REQUEST_PAUSED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000459, 'STATUS_REQUEST_PAUSED')
    STATUS_NO_RANGES_PROCESSED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000460, 'STATUS_NO_RANGES_PROCESSED')
    STATUS_DISK_RESOURCES_EXHAUSTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000461, 'STATUS_DISK_RESOURCES_EXHAUSTED')
    STATUS_NEEDS_REMEDIATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000462, 'STATUS_NEEDS_REMEDIATION')
    STATUS_DEVICE_FEATURE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000463, 'STATUS_DEVICE_FEATURE_NOT_SUPPORTED')
    STATUS_DEVICE_UNREACHABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000464, 'STATUS_DEVICE_UNREACHABLE')
    STATUS_INVALID_TOKEN: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000465, 'STATUS_INVALID_TOKEN')
    STATUS_SERVER_UNAVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000466, 'STATUS_SERVER_UNAVAILABLE')
    STATUS_FILE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000467, 'STATUS_FILE_NOT_AVAILABLE')
    STATUS_DEVICE_INSUFFICIENT_RESOURCES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000468, 'STATUS_DEVICE_INSUFFICIENT_RESOURCES')
    STATUS_PACKAGE_UPDATING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000469, 'STATUS_PACKAGE_UPDATING')
    STATUS_NOT_READ_FROM_COPY: Final[ThirtyTwoBits] = cached_ntstatus(0xC000046A, 'STATUS_NOT_READ_FROM_COPY')
    STATUS_FT_WRITE_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000046B, 'STATUS_FT_WRITE_FAILURE')
    STATUS_FT_DI_SCAN_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000046C, 'STATUS_FT_DI_SCAN_REQUIRED')
    STATUS_OBJECT_NOT_EXTERNALLY_BACKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000046D, 'STATUS_OBJECT_NOT_EXTERNALLY_BACKED')
    STATUS_EXTERNAL_BACKING_PROVIDER_UNKNOWN: Final[ThirtyTwoBits] = cached_ntstatus(0xC000046E, 'STATUS_EXTERNAL_BACKING_PROVIDER_UNKNOWN')
    STATUS_COMPRESSION_NOT_BENEFICIAL: Final[ThirtyTwoBits] = cached_ntstatus(0xC000046F, 'STATUS_COMPRESSION_NOT_BENEFICIAL')
    STATUS_DATA_CHECKSUM_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000470, 'STATUS_DATA_CHECKSUM_ERROR')
    STATUS_INTERMIXED_KERNEL_EA_OPERATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000471, 'STATUS_INTERMIXED_KERNEL_EA_OPERATION')
    STATUS_TRIM_READ_ZERO_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000472, 'STATUS_TRIM_READ_ZERO_NOT_SUPPORTED')
    STATUS_TOO_MANY_SEGMENT_DESCRIPTORS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000473, 'STATUS_TOO_MANY_SEGMENT_DESCRIPTORS')
    STATUS_INVALID_OFFSET_ALIGNMENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000474, 'STATUS_INVALID_OFFSET_ALIGNMENT')
    STATUS_INVALID_FIELD_IN_PARAMETER_LIST: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000475, 'STATUS_INVALID_FIELD_IN_PARAMETER_LIST')
    STATUS_OPERATION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000476, 'STATUS_OPERATION_IN_PROGRESS')
    STATUS_INVALID_INITIATOR_TARGET_PATH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000477, 'STATUS_INVALID_INITIATOR_TARGET_PATH')
    STATUS_SCRUB_DATA_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000478, 'STATUS_SCRUB_DATA_DISABLED')
    STATUS_NOT_REDUNDANT_STORAGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000479, 'STATUS_NOT_REDUNDANT_STORAGE')
    STATUS_RESIDENT_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000047A, 'STATUS_RESIDENT_FILE_NOT_SUPPORTED')
    STATUS_COMPRESSED_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000047B, 'STATUS_COMPRESSED_FILE_NOT_SUPPORTED')
    STATUS_DIRECTORY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000047C, 'STATUS_DIRECTORY_NOT_SUPPORTED')
    STATUS_IO_OPERATION_TIMEOUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000047D, 'STATUS_IO_OPERATION_TIMEOUT')
    STATUS_SYSTEM_NEEDS_REMEDIATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000047E, 'STATUS_SYSTEM_NEEDS_REMEDIATION')
    STATUS_APPX_INTEGRITY_FAILURE_CLR_NGEN: Final[ThirtyTwoBits] = cached_ntstatus(0xC000047F, 'STATUS_APPX_INTEGRITY_FAILURE_CLR_NGEN')
    STATUS_SHARE_UNAVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000480, 'STATUS_SHARE_UNAVAILABLE')
    STATUS_APISET_NOT_HOSTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000481, 'STATUS_APISET_NOT_HOSTED')
    STATUS_APISET_NOT_PRESENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000482, 'STATUS_APISET_NOT_PRESENT')
    STATUS_DEVICE_HARDWARE_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000483, 'STATUS_DEVICE_HARDWARE_ERROR')
    STATUS_FIRMWARE_SLOT_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000484, 'STATUS_FIRMWARE_SLOT_INVALID')
    STATUS_FIRMWARE_IMAGE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000485, 'STATUS_FIRMWARE_IMAGE_INVALID')
    STATUS_STORAGE_TOPOLOGY_ID_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000486, 'STATUS_STORAGE_TOPOLOGY_ID_MISMATCH')
    STATUS_WIM_NOT_BOOTABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000487, 'STATUS_WIM_NOT_BOOTABLE')
    STATUS_BLOCKED_BY_PARENTAL_CONTROLS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000488, 'STATUS_BLOCKED_BY_PARENTAL_CONTROLS')
    STATUS_NEEDS_REGISTRATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000489, 'STATUS_NEEDS_REGISTRATION')
    STATUS_QUOTA_ACTIVITY: Final[ThirtyTwoBits] = cached_ntstatus(0xC000048A, 'STATUS_QUOTA_ACTIVITY')
    STATUS_CALLBACK_INVOKE_INLINE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000048B, 'STATUS_CALLBACK_INVOKE_INLINE')
    STATUS_BLOCK_TOO_MANY_REFERENCES: Final[ThirtyTwoBits] = cached_ntstatus(0xC000048C, 'STATUS_BLOCK_TOO_MANY_REFERENCES')
    STATUS_MARKED_TO_DISALLOW_WRITES: Final[ThirtyTwoBits] = cached_ntstatus(0xC000048D, 'STATUS_MARKED_TO_DISALLOW_WRITES')
    STATUS_NETWORK_ACCESS_DENIED_EDP: Final[ThirtyTwoBits] = cached_ntstatus(0xC000048E, 'STATUS_NETWORK_ACCESS_DENIED_EDP')
    STATUS_ENCLAVE_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000048F, 'STATUS_ENCLAVE_FAILURE')
    STATUS_PNP_NO_COMPAT_DRIVERS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000490, 'STATUS_PNP_NO_COMPAT_DRIVERS')
    STATUS_PNP_DRIVER_PACKAGE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000491, 'STATUS_PNP_DRIVER_PACKAGE_NOT_FOUND')
    STATUS_PNP_DRIVER_CONFIGURATION_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000492, 'STATUS_PNP_DRIVER_CONFIGURATION_NOT_FOUND')
    STATUS_PNP_DRIVER_CONFIGURATION_INCOMPLETE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000493, 'STATUS_PNP_DRIVER_CONFIGURATION_INCOMPLETE')
    STATUS_PNP_FUNCTION_DRIVER_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000494, 'STATUS_PNP_FUNCTION_DRIVER_REQUIRED')
    STATUS_PNP_DEVICE_CONFIGURATION_PENDING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000495, 'STATUS_PNP_DEVICE_CONFIGURATION_PENDING')
    STATUS_DEVICE_HINT_NAME_BUFFER_TOO_SMALL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000496, 'STATUS_DEVICE_HINT_NAME_BUFFER_TOO_SMALL')
    STATUS_PACKAGE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000497, 'STATUS_PACKAGE_NOT_AVAILABLE')
    STATUS_DEVICE_IN_MAINTENANCE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000499, 'STATUS_DEVICE_IN_MAINTENANCE')
    STATUS_NOT_SUPPORTED_ON_DAX: Final[ThirtyTwoBits] = cached_ntstatus(0xC000049A, 'STATUS_NOT_SUPPORTED_ON_DAX')
    STATUS_FREE_SPACE_TOO_FRAGMENTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000049B, 'STATUS_FREE_SPACE_TOO_FRAGMENTED')
    STATUS_DAX_MAPPING_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC000049C, 'STATUS_DAX_MAPPING_EXISTS')
    STATUS_CHILD_PROCESS_BLOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000049D, 'STATUS_CHILD_PROCESS_BLOCKED')
    STATUS_STORAGE_LOST_DATA_PERSISTENCE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000049E, 'STATUS_STORAGE_LOST_DATA_PERSISTENCE')
    STATUS_PARTITION_TERMINATING: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004A0, 'STATUS_PARTITION_TERMINATING')
    STATUS_EXTERNAL_SYSKEY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004A1, 'STATUS_EXTERNAL_SYSKEY_NOT_SUPPORTED')
    STATUS_ENCLAVE_VIOLATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004A2, 'STATUS_ENCLAVE_VIOLATION')
    STATUS_FILE_PROTECTED_UNDER_DPL: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004A3, 'STATUS_FILE_PROTECTED_UNDER_DPL')
    STATUS_VOLUME_NOT_CLUSTER_ALIGNED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004A4, 'STATUS_VOLUME_NOT_CLUSTER_ALIGNED')
    STATUS_NO_PHYSICALLY_ALIGNED_FREE_SPACE_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004A5, 'STATUS_NO_PHYSICALLY_ALIGNED_FREE_SPACE_FOUND')
    STATUS_APPX_FILE_NOT_ENCRYPTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004A6, 'STATUS_APPX_FILE_NOT_ENCRYPTED')
    STATUS_RWRAW_ENCRYPTED_FILE_NOT_ENCRYPTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004A7, 'STATUS_RWRAW_ENCRYPTED_FILE_NOT_ENCRYPTED')
    STATUS_RWRAW_ENCRYPTED_INVALID_EDATAINFO_FILEOFFSET: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004A8, 'STATUS_RWRAW_ENCRYPTED_INVALID_EDATAINFO_FILEOFFSET')
    STATUS_RWRAW_ENCRYPTED_INVALID_EDATAINFO_FILERANGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004A9, 'STATUS_RWRAW_ENCRYPTED_INVALID_EDATAINFO_FILERANGE')
    STATUS_RWRAW_ENCRYPTED_INVALID_EDATAINFO_PARAMETER: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004AA, 'STATUS_RWRAW_ENCRYPTED_INVALID_EDATAINFO_PARAMETER')
    STATUS_FT_READ_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004AB, 'STATUS_FT_READ_FAILURE')
    STATUS_PATCH_CONFLICT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004AC, 'STATUS_PATCH_CONFLICT')
    STATUS_STORAGE_RESERVE_ID_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004AD, 'STATUS_STORAGE_RESERVE_ID_INVALID')
    STATUS_STORAGE_RESERVE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004AE, 'STATUS_STORAGE_RESERVE_DOES_NOT_EXIST')
    STATUS_STORAGE_RESERVE_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004AF, 'STATUS_STORAGE_RESERVE_ALREADY_EXISTS')
    STATUS_STORAGE_RESERVE_NOT_EMPTY: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004B0, 'STATUS_STORAGE_RESERVE_NOT_EMPTY')
    STATUS_NOT_A_DAX_VOLUME: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004B1, 'STATUS_NOT_A_DAX_VOLUME')
    STATUS_NOT_DAX_MAPPABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004B2, 'STATUS_NOT_DAX_MAPPABLE')
    STATUS_CASE_DIFFERING_NAMES_IN_DIR: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004B3, 'STATUS_CASE_DIFFERING_NAMES_IN_DIR')
    STATUS_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004B4, 'STATUS_FILE_NOT_SUPPORTED')
    STATUS_NOT_SUPPORTED_WITH_BTT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004B5, 'STATUS_NOT_SUPPORTED_WITH_BTT')
    STATUS_ENCRYPTION_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004B6, 'STATUS_ENCRYPTION_DISABLED')
    STATUS_ENCRYPTING_METADATA_DISALLOWED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004B7, 'STATUS_ENCRYPTING_METADATA_DISALLOWED')
    STATUS_CANT_CLEAR_ENCRYPTION_FLAG: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004B8, 'STATUS_CANT_CLEAR_ENCRYPTION_FLAG')
    STATUS_UNSATISFIED_DEPENDENCIES: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004B9, 'STATUS_UNSATISFIED_DEPENDENCIES')
    STATUS_CASE_SENSITIVE_PATH: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004BA, 'STATUS_CASE_SENSITIVE_PATH')
    STATUS_UNSUPPORTED_PAGING_MODE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004BB, 'STATUS_UNSUPPORTED_PAGING_MODE')
    STATUS_UNTRUSTED_MOUNT_POINT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004BC, 'STATUS_UNTRUSTED_MOUNT_POINT')
    STATUS_HAS_SYSTEM_CRITICAL_FILES: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004BD, 'STATUS_HAS_SYSTEM_CRITICAL_FILES')
    STATUS_OBJECT_IS_IMMUTABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004BE, 'STATUS_OBJECT_IS_IMMUTABLE')
    STATUS_FT_READ_FROM_COPY_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004BF, 'STATUS_FT_READ_FROM_COPY_FAILURE')
    STATUS_IMAGE_LOADED_AS_PATCH_IMAGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004C0, 'STATUS_IMAGE_LOADED_AS_PATCH_IMAGE')
    STATUS_STORAGE_STACK_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004C1, 'STATUS_STORAGE_STACK_ACCESS_DENIED')
    STATUS_INSUFFICIENT_VIRTUAL_ADDR_RESOURCES: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004C2, 'STATUS_INSUFFICIENT_VIRTUAL_ADDR_RESOURCES')
    STATUS_ENCRYPTED_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004C3, 'STATUS_ENCRYPTED_FILE_NOT_SUPPORTED')
    STATUS_SPARSE_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004C4, 'STATUS_SPARSE_FILE_NOT_SUPPORTED')
    STATUS_PAGEFILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004C5, 'STATUS_PAGEFILE_NOT_SUPPORTED')
    STATUS_VOLUME_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004C6, 'STATUS_VOLUME_NOT_SUPPORTED')
    STATUS_NOT_SUPPORTED_WITH_BYPASSIO: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004C7, 'STATUS_NOT_SUPPORTED_WITH_BYPASSIO')
    STATUS_NO_BYPASSIO_DRIVER_SUPPORT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004C8, 'STATUS_NO_BYPASSIO_DRIVER_SUPPORT')
    STATUS_NOT_SUPPORTED_WITH_ENCRYPTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004C9, 'STATUS_NOT_SUPPORTED_WITH_ENCRYPTION')
    STATUS_NOT_SUPPORTED_WITH_COMPRESSION: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004CA, 'STATUS_NOT_SUPPORTED_WITH_COMPRESSION')
    STATUS_NOT_SUPPORTED_WITH_REPLICATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004CB, 'STATUS_NOT_SUPPORTED_WITH_REPLICATION')
    STATUS_NOT_SUPPORTED_WITH_DEDUPLICATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004CC, 'STATUS_NOT_SUPPORTED_WITH_DEDUPLICATION')
    STATUS_NOT_SUPPORTED_WITH_AUDITING: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004CD, 'STATUS_NOT_SUPPORTED_WITH_AUDITING')
    STATUS_NOT_SUPPORTED_WITH_MONITORING: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004CE, 'STATUS_NOT_SUPPORTED_WITH_MONITORING')
    STATUS_NOT_SUPPORTED_WITH_SNAPSHOT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004CF, 'STATUS_NOT_SUPPORTED_WITH_SNAPSHOT')
    STATUS_NOT_SUPPORTED_WITH_VIRTUALIZATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004D0, 'STATUS_NOT_SUPPORTED_WITH_VIRTUALIZATION')
    STATUS_INDEX_OUT_OF_BOUNDS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004D1, 'STATUS_INDEX_OUT_OF_BOUNDS')
    STATUS_BYPASSIO_FLT_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004D2, 'STATUS_BYPASSIO_FLT_NOT_SUPPORTED')
    STATUS_VOLUME_WRITE_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004D3, 'STATUS_VOLUME_WRITE_ACCESS_DENIED')
    STATUS_PATCH_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004D4, 'STATUS_PATCH_NOT_REGISTERED')
    STATUS_NOT_SUPPORTED_WITH_CACHED_HANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004D5, 'STATUS_NOT_SUPPORTED_WITH_CACHED_HANDLE')
    STATUS_VOLUME_UPGRADE_NOT_NEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004D9, 'STATUS_VOLUME_UPGRADE_NOT_NEEDED')
    STATUS_VOLUME_UPGRADE_PENDING: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004DA, 'STATUS_VOLUME_UPGRADE_PENDING')
    STATUS_VOLUME_UPGRADE_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004DB, 'STATUS_VOLUME_UPGRADE_DISABLED')
    STATUS_VOLUME_UPGRADE_DISABLED_TILL_OS_DOWNGRADE_EXPIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004DC, 'STATUS_VOLUME_UPGRADE_DISABLED_TILL_OS_DOWNGRADE_EXPIRED')
    STATUS_NOT_A_DEV_VOLUME: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004DD, 'STATUS_NOT_A_DEV_VOLUME')
    STATUS_FS_GUID_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004DE, 'STATUS_FS_GUID_MISMATCH')
    STATUS_CANT_ATTACH_TO_DEV_VOLUME: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004DF, 'STATUS_CANT_ATTACH_TO_DEV_VOLUME')
    STATUS_INVALID_CONFIG_VALUE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00004E0, 'STATUS_INVALID_CONFIG_VALUE')
    STATUS_INVALID_TASK_NAME: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000500, 'STATUS_INVALID_TASK_NAME')
    STATUS_INVALID_TASK_INDEX: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000501, 'STATUS_INVALID_TASK_INDEX')
    STATUS_THREAD_ALREADY_IN_TASK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000502, 'STATUS_THREAD_ALREADY_IN_TASK')
    STATUS_CALLBACK_BYPASS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000503, 'STATUS_CALLBACK_BYPASS')
    STATUS_UNDEFINED_SCOPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000504, 'STATUS_UNDEFINED_SCOPE')
    STATUS_INVALID_CAP: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000505, 'STATUS_INVALID_CAP')
    STATUS_NOT_GUI_PROCESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000506, 'STATUS_NOT_GUI_PROCESS')
    STATUS_DEVICE_HUNG: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000507, 'STATUS_DEVICE_HUNG')
    STATUS_CONTAINER_ASSIGNED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000508, 'STATUS_CONTAINER_ASSIGNED')
    STATUS_JOB_NO_CONTAINER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000509, 'STATUS_JOB_NO_CONTAINER')
    STATUS_DEVICE_UNRESPONSIVE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000050A, 'STATUS_DEVICE_UNRESPONSIVE')
    STATUS_REPARSE_POINT_ENCOUNTERED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000050B, 'STATUS_REPARSE_POINT_ENCOUNTERED')
    STATUS_ATTRIBUTE_NOT_PRESENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000050C, 'STATUS_ATTRIBUTE_NOT_PRESENT')
    STATUS_NOT_A_TIERED_VOLUME: Final[ThirtyTwoBits] = cached_ntstatus(0xC000050D, 'STATUS_NOT_A_TIERED_VOLUME')
    STATUS_ALREADY_HAS_STREAM_ID: Final[ThirtyTwoBits] = cached_ntstatus(0xC000050E, 'STATUS_ALREADY_HAS_STREAM_ID')
    STATUS_JOB_NOT_EMPTY: Final[ThirtyTwoBits] = cached_ntstatus(0xC000050F, 'STATUS_JOB_NOT_EMPTY')
    STATUS_ALREADY_INITIALIZED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000510, 'STATUS_ALREADY_INITIALIZED')
    STATUS_ENCLAVE_NOT_TERMINATED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000511, 'STATUS_ENCLAVE_NOT_TERMINATED')
    STATUS_ENCLAVE_IS_TERMINATING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000512, 'STATUS_ENCLAVE_IS_TERMINATING')
    STATUS_SMB1_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000513, 'STATUS_SMB1_NOT_AVAILABLE')
    STATUS_SMR_GARBAGE_COLLECTION_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000514, 'STATUS_SMR_GARBAGE_COLLECTION_REQUIRED')
    STATUS_INTERRUPTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000515, 'STATUS_INTERRUPTED')
    STATUS_THREAD_NOT_RUNNING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000516, 'STATUS_THREAD_NOT_RUNNING')
    STATUS_SESSION_KEY_TOO_SHORT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000517, 'STATUS_SESSION_KEY_TOO_SHORT')
    STATUS_FS_METADATA_INCONSISTENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000518, 'STATUS_FS_METADATA_INCONSISTENT')
    STATUS_FAIL_FAST_EXCEPTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000602, 'STATUS_FAIL_FAST_EXCEPTION')
    STATUS_IMAGE_CERT_REVOKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000603, 'STATUS_IMAGE_CERT_REVOKED')
    STATUS_DYNAMIC_CODE_BLOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000604, 'STATUS_DYNAMIC_CODE_BLOCKED')
    STATUS_IMAGE_CERT_EXPIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000605, 'STATUS_IMAGE_CERT_EXPIRED')
    STATUS_STRICT_CFG_VIOLATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000606, 'STATUS_STRICT_CFG_VIOLATION')
    STATUS_SET_CONTEXT_DENIED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000060A, 'STATUS_SET_CONTEXT_DENIED')
    STATUS_CROSS_PARTITION_VIOLATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000060B, 'STATUS_CROSS_PARTITION_VIOLATION')
    STATUS_PORT_CLOSED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000700, 'STATUS_PORT_CLOSED')
    STATUS_MESSAGE_LOST: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000701, 'STATUS_MESSAGE_LOST')
    STATUS_INVALID_MESSAGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000702, 'STATUS_INVALID_MESSAGE')
    STATUS_REQUEST_CANCELED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000703, 'STATUS_REQUEST_CANCELED')
    STATUS_RECURSIVE_DISPATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000704, 'STATUS_RECURSIVE_DISPATCH')
    STATUS_LPC_RECEIVE_BUFFER_EXPECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000705, 'STATUS_LPC_RECEIVE_BUFFER_EXPECTED')
    STATUS_LPC_INVALID_CONNECTION_USAGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000706, 'STATUS_LPC_INVALID_CONNECTION_USAGE')
    STATUS_LPC_REQUESTS_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000707, 'STATUS_LPC_REQUESTS_NOT_ALLOWED')
    STATUS_RESOURCE_IN_USE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000708, 'STATUS_RESOURCE_IN_USE')
    STATUS_HARDWARE_MEMORY_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000709, 'STATUS_HARDWARE_MEMORY_ERROR')
    STATUS_THREADPOOL_HANDLE_EXCEPTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000070A, 'STATUS_THREADPOOL_HANDLE_EXCEPTION')
    STATUS_THREADPOOL_SET_EVENT_ON_COMPLETION_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000070B, 'STATUS_THREADPOOL_SET_EVENT_ON_COMPLETION_FAILED')
    STATUS_THREADPOOL_RELEASE_SEMAPHORE_ON_COMPLETION_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000070C, 'STATUS_THREADPOOL_RELEASE_SEMAPHORE_ON_COMPLETION_FAILED')
    STATUS_THREADPOOL_RELEASE_MUTEX_ON_COMPLETION_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000070D, 'STATUS_THREADPOOL_RELEASE_MUTEX_ON_COMPLETION_FAILED')
    STATUS_THREADPOOL_FREE_LIBRARY_ON_COMPLETION_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000070E, 'STATUS_THREADPOOL_FREE_LIBRARY_ON_COMPLETION_FAILED')
    STATUS_THREADPOOL_RELEASED_DURING_OPERATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000070F, 'STATUS_THREADPOOL_RELEASED_DURING_OPERATION')
    STATUS_CALLBACK_RETURNED_WHILE_IMPERSONATING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000710, 'STATUS_CALLBACK_RETURNED_WHILE_IMPERSONATING')
    STATUS_APC_RETURNED_WHILE_IMPERSONATING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000711, 'STATUS_APC_RETURNED_WHILE_IMPERSONATING')
    STATUS_PROCESS_IS_PROTECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000712, 'STATUS_PROCESS_IS_PROTECTED')
    STATUS_MCA_EXCEPTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000713, 'STATUS_MCA_EXCEPTION')
    STATUS_CERTIFICATE_MAPPING_NOT_UNIQUE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000714, 'STATUS_CERTIFICATE_MAPPING_NOT_UNIQUE')
    STATUS_SYMLINK_CLASS_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000715, 'STATUS_SYMLINK_CLASS_DISABLED')
    STATUS_INVALID_IDN_NORMALIZATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000716, 'STATUS_INVALID_IDN_NORMALIZATION')
    STATUS_NO_UNICODE_TRANSLATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000717, 'STATUS_NO_UNICODE_TRANSLATION')
    STATUS_ALREADY_REGISTERED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000718, 'STATUS_ALREADY_REGISTERED')
    STATUS_CONTEXT_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000719, 'STATUS_CONTEXT_MISMATCH')
    STATUS_PORT_ALREADY_HAS_COMPLETION_LIST: Final[ThirtyTwoBits] = cached_ntstatus(0xC000071A, 'STATUS_PORT_ALREADY_HAS_COMPLETION_LIST')
    STATUS_CALLBACK_RETURNED_THREAD_PRIORITY: Final[ThirtyTwoBits] = cached_ntstatus(0xC000071B, 'STATUS_CALLBACK_RETURNED_THREAD_PRIORITY')
    STATUS_INVALID_THREAD: Final[ThirtyTwoBits] = cached_ntstatus(0xC000071C, 'STATUS_INVALID_THREAD')
    STATUS_CALLBACK_RETURNED_TRANSACTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000071D, 'STATUS_CALLBACK_RETURNED_TRANSACTION')
    STATUS_CALLBACK_RETURNED_LDR_LOCK: Final[ThirtyTwoBits] = cached_ntstatus(0xC000071E, 'STATUS_CALLBACK_RETURNED_LDR_LOCK')
    STATUS_CALLBACK_RETURNED_LANG: Final[ThirtyTwoBits] = cached_ntstatus(0xC000071F, 'STATUS_CALLBACK_RETURNED_LANG')
    STATUS_CALLBACK_RETURNED_PRI_BACK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000720, 'STATUS_CALLBACK_RETURNED_PRI_BACK')
    STATUS_CALLBACK_RETURNED_THREAD_AFFINITY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000721, 'STATUS_CALLBACK_RETURNED_THREAD_AFFINITY')
    STATUS_LPC_HANDLE_COUNT_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000722, 'STATUS_LPC_HANDLE_COUNT_EXCEEDED')
    STATUS_EXECUTABLE_MEMORY_WRITE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000723, 'STATUS_EXECUTABLE_MEMORY_WRITE')
    STATUS_KERNEL_EXECUTABLE_MEMORY_WRITE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000724, 'STATUS_KERNEL_EXECUTABLE_MEMORY_WRITE')
    STATUS_ATTACHED_EXECUTABLE_MEMORY_WRITE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000725, 'STATUS_ATTACHED_EXECUTABLE_MEMORY_WRITE')
    STATUS_TRIGGERED_EXECUTABLE_MEMORY_WRITE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000726, 'STATUS_TRIGGERED_EXECUTABLE_MEMORY_WRITE')
    STATUS_DISK_REPAIR_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000800, 'STATUS_DISK_REPAIR_DISABLED')
    STATUS_DS_DOMAIN_RENAME_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000801, 'STATUS_DS_DOMAIN_RENAME_IN_PROGRESS')
    STATUS_DISK_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000802, 'STATUS_DISK_QUOTA_EXCEEDED')
    STATUS_CONTENT_BLOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000804, 'STATUS_CONTENT_BLOCKED')
    STATUS_BAD_CLUSTERS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000805, 'STATUS_BAD_CLUSTERS')
    STATUS_VOLUME_DIRTY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000806, 'STATUS_VOLUME_DIRTY')
    STATUS_DISK_REPAIR_UNSUCCESSFUL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000808, 'STATUS_DISK_REPAIR_UNSUCCESSFUL')
    STATUS_CORRUPT_LOG_OVERFULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000809, 'STATUS_CORRUPT_LOG_OVERFULL')
    STATUS_CORRUPT_LOG_CORRUPTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000080A, 'STATUS_CORRUPT_LOG_CORRUPTED')
    STATUS_CORRUPT_LOG_UNAVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000080B, 'STATUS_CORRUPT_LOG_UNAVAILABLE')
    STATUS_CORRUPT_LOG_DELETED_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC000080C, 'STATUS_CORRUPT_LOG_DELETED_FULL')
    STATUS_CORRUPT_LOG_CLEARED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000080D, 'STATUS_CORRUPT_LOG_CLEARED')
    STATUS_ORPHAN_NAME_EXHAUSTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000080E, 'STATUS_ORPHAN_NAME_EXHAUSTED')
    STATUS_PROACTIVE_SCAN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC000080F, 'STATUS_PROACTIVE_SCAN_IN_PROGRESS')
    STATUS_ENCRYPTED_IO_NOT_POSSIBLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000810, 'STATUS_ENCRYPTED_IO_NOT_POSSIBLE')
    STATUS_CORRUPT_LOG_UPLEVEL_RECORDS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000811, 'STATUS_CORRUPT_LOG_UPLEVEL_RECORDS')
    STATUS_FILE_CHECKED_OUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000901, 'STATUS_FILE_CHECKED_OUT')
    STATUS_CHECKOUT_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000902, 'STATUS_CHECKOUT_REQUIRED')
    STATUS_BAD_FILE_TYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000903, 'STATUS_BAD_FILE_TYPE')
    STATUS_FILE_TOO_LARGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000904, 'STATUS_FILE_TOO_LARGE')
    STATUS_FORMS_AUTH_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000905, 'STATUS_FORMS_AUTH_REQUIRED')
    STATUS_VIRUS_INFECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000906, 'STATUS_VIRUS_INFECTED')
    STATUS_VIRUS_DELETED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000907, 'STATUS_VIRUS_DELETED')
    STATUS_BAD_MCFG_TABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000908, 'STATUS_BAD_MCFG_TABLE')
    STATUS_CANNOT_BREAK_OPLOCK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000909, 'STATUS_CANNOT_BREAK_OPLOCK')
    STATUS_BAD_KEY: Final[ThirtyTwoBits] = cached_ntstatus(0xC000090A, 'STATUS_BAD_KEY')
    STATUS_BAD_DATA: Final[ThirtyTwoBits] = cached_ntstatus(0xC000090B, 'STATUS_BAD_DATA')
    STATUS_NO_KEY: Final[ThirtyTwoBits] = cached_ntstatus(0xC000090C, 'STATUS_NO_KEY')
    STATUS_FILE_HANDLE_REVOKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000910, 'STATUS_FILE_HANDLE_REVOKED')
    STATUS_SECTION_DIRECT_MAP_ONLY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000911, 'STATUS_SECTION_DIRECT_MAP_ONLY')
    STATUS_BLOCK_WEAK_REFERENCE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000912, 'STATUS_BLOCK_WEAK_REFERENCE_INVALID')
    STATUS_BLOCK_SOURCE_WEAK_REFERENCE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000913, 'STATUS_BLOCK_SOURCE_WEAK_REFERENCE_INVALID')
    STATUS_BLOCK_TARGET_WEAK_REFERENCE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000914, 'STATUS_BLOCK_TARGET_WEAK_REFERENCE_INVALID')
    STATUS_BLOCK_SHARED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000915, 'STATUS_BLOCK_SHARED')
    STATUS_VRF_VOLATILE_CFG_AND_IO_ENABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000C08, 'STATUS_VRF_VOLATILE_CFG_AND_IO_ENABLED')
    STATUS_VRF_VOLATILE_NOT_STOPPABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000C09, 'STATUS_VRF_VOLATILE_NOT_STOPPABLE')
    STATUS_VRF_VOLATILE_SAFE_MODE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000C0A, 'STATUS_VRF_VOLATILE_SAFE_MODE')
    STATUS_VRF_VOLATILE_NOT_RUNNABLE_SYSTEM: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000C0B, 'STATUS_VRF_VOLATILE_NOT_RUNNABLE_SYSTEM')
    STATUS_VRF_VOLATILE_NOT_SUPPORTED_RULECLASS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000C0C, 'STATUS_VRF_VOLATILE_NOT_SUPPORTED_RULECLASS')
    STATUS_VRF_VOLATILE_PROTECTED_DRIVER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000C0D, 'STATUS_VRF_VOLATILE_PROTECTED_DRIVER')
    STATUS_VRF_VOLATILE_NMI_REGISTERED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000C0E, 'STATUS_VRF_VOLATILE_NMI_REGISTERED')
    STATUS_VRF_VOLATILE_SETTINGS_CONFLICT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000C0F, 'STATUS_VRF_VOLATILE_SETTINGS_CONFLICT')
    STATUS_DIF_IOCALLBACK_NOT_REPLACED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000C76, 'STATUS_DIF_IOCALLBACK_NOT_REPLACED')
    STATUS_DIF_LIVEDUMP_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000C77, 'STATUS_DIF_LIVEDUMP_LIMIT_EXCEEDED')
    STATUS_DIF_VOLATILE_SECTION_NOT_LOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000C78, 'STATUS_DIF_VOLATILE_SECTION_NOT_LOCKED')
    STATUS_DIF_VOLATILE_DRIVER_HOTPATCHED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000C79, 'STATUS_DIF_VOLATILE_DRIVER_HOTPATCHED')
    STATUS_DIF_VOLATILE_INVALID_INFO: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000C7A, 'STATUS_DIF_VOLATILE_INVALID_INFO')
    STATUS_DIF_VOLATILE_DRIVER_IS_NOT_RUNNING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000C7B, 'STATUS_DIF_VOLATILE_DRIVER_IS_NOT_RUNNING')
    STATUS_DIF_VOLATILE_PLUGIN_IS_NOT_RUNNING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000C7C, 'STATUS_DIF_VOLATILE_PLUGIN_IS_NOT_RUNNING')
    STATUS_DIF_VOLATILE_PLUGIN_CHANGE_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000C7D, 'STATUS_DIF_VOLATILE_PLUGIN_CHANGE_NOT_ALLOWED')
    STATUS_DIF_VOLATILE_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000C7E, 'STATUS_DIF_VOLATILE_NOT_ALLOWED')
    STATUS_DIF_BINDING_API_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0000C7F, 'STATUS_DIF_BINDING_API_NOT_FOUND')
    STATUS_WOW_ASSERTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0009898, 'STATUS_WOW_ASSERTION')
    STATUS_INVALID_SIGNATURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A000, 'STATUS_INVALID_SIGNATURE')
    STATUS_HMAC_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A001, 'STATUS_HMAC_NOT_SUPPORTED')
    STATUS_AUTH_TAG_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A002, 'STATUS_AUTH_TAG_MISMATCH')
    STATUS_INVALID_STATE_TRANSITION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A003, 'STATUS_INVALID_STATE_TRANSITION')
    STATUS_INVALID_KERNEL_INFO_VERSION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A004, 'STATUS_INVALID_KERNEL_INFO_VERSION')
    STATUS_INVALID_PEP_INFO_VERSION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A005, 'STATUS_INVALID_PEP_INFO_VERSION')
    STATUS_HANDLE_REVOKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A006, 'STATUS_HANDLE_REVOKED')
    STATUS_EOF_ON_GHOSTED_RANGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A007, 'STATUS_EOF_ON_GHOSTED_RANGE')
    STATUS_CC_NEEDS_CALLBACK_SECTION_DRAIN: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A008, 'STATUS_CC_NEEDS_CALLBACK_SECTION_DRAIN')
    STATUS_IPSEC_QUEUE_OVERFLOW: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A010, 'STATUS_IPSEC_QUEUE_OVERFLOW')
    STATUS_ND_QUEUE_OVERFLOW: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A011, 'STATUS_ND_QUEUE_OVERFLOW')
    STATUS_HOPLIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A012, 'STATUS_HOPLIMIT_EXCEEDED')
    STATUS_PROTOCOL_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A013, 'STATUS_PROTOCOL_NOT_SUPPORTED')
    STATUS_FASTPATH_REJECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A014, 'STATUS_FASTPATH_REJECTED')
    STATUS_LOST_WRITEBEHIND_DATA_NETWORK_DISCONNECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A080, 'STATUS_LOST_WRITEBEHIND_DATA_NETWORK_DISCONNECTED')
    STATUS_LOST_WRITEBEHIND_DATA_NETWORK_SERVER_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A081, 'STATUS_LOST_WRITEBEHIND_DATA_NETWORK_SERVER_ERROR')
    STATUS_LOST_WRITEBEHIND_DATA_LOCAL_DISK_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A082, 'STATUS_LOST_WRITEBEHIND_DATA_LOCAL_DISK_ERROR')
    STATUS_XML_PARSE_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A083, 'STATUS_XML_PARSE_ERROR')
    STATUS_XMLDSIG_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A084, 'STATUS_XMLDSIG_ERROR')
    STATUS_WRONG_COMPARTMENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A085, 'STATUS_WRONG_COMPARTMENT')
    STATUS_AUTHIP_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A086, 'STATUS_AUTHIP_FAILURE')
    STATUS_DS_OID_MAPPED_GROUP_CANT_HAVE_MEMBERS: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A087, 'STATUS_DS_OID_MAPPED_GROUP_CANT_HAVE_MEMBERS')
    STATUS_DS_OID_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A088, 'STATUS_DS_OID_NOT_FOUND')
    STATUS_INCORRECT_ACCOUNT_TYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A089, 'STATUS_INCORRECT_ACCOUNT_TYPE')
    STATUS_LOCAL_POLICY_MODIFICATION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A08A, 'STATUS_LOCAL_POLICY_MODIFICATION_NOT_SUPPORTED')
    STATUS_POLICY_CONTROLLED_ACCOUNT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A08B, 'STATUS_POLICY_CONTROLLED_ACCOUNT')
    STATUS_LAPS_LEGACY_SCHEMA_MISSING: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A08C, 'STATUS_LAPS_LEGACY_SCHEMA_MISSING')
    STATUS_LAPS_SCHEMA_MISSING: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A08D, 'STATUS_LAPS_SCHEMA_MISSING')
    STATUS_LAPS_ENCRYPTION_REQUIRES_2016_DFL: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A08E, 'STATUS_LAPS_ENCRYPTION_REQUIRES_2016_DFL')
    STATUS_HASH_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A100, 'STATUS_HASH_NOT_SUPPORTED')
    STATUS_HASH_NOT_PRESENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A101, 'STATUS_HASH_NOT_PRESENT')
    STATUS_SECONDARY_IC_PROVIDER_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A121, 'STATUS_SECONDARY_IC_PROVIDER_NOT_REGISTERED')
    STATUS_GPIO_CLIENT_INFORMATION_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A122, 'STATUS_GPIO_CLIENT_INFORMATION_INVALID')
    STATUS_GPIO_VERSION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A123, 'STATUS_GPIO_VERSION_NOT_SUPPORTED')
    STATUS_GPIO_INVALID_REGISTRATION_PACKET: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A124, 'STATUS_GPIO_INVALID_REGISTRATION_PACKET')
    STATUS_GPIO_OPERATION_DENIED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A125, 'STATUS_GPIO_OPERATION_DENIED')
    STATUS_GPIO_INCOMPATIBLE_CONNECT_MODE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A126, 'STATUS_GPIO_INCOMPATIBLE_CONNECT_MODE')
    STATUS_CANNOT_SWITCH_RUNLEVEL: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A141, 'STATUS_CANNOT_SWITCH_RUNLEVEL')
    STATUS_INVALID_RUNLEVEL_SETTING: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A142, 'STATUS_INVALID_RUNLEVEL_SETTING')
    STATUS_RUNLEVEL_SWITCH_TIMEOUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A143, 'STATUS_RUNLEVEL_SWITCH_TIMEOUT')
    STATUS_RUNLEVEL_SWITCH_AGENT_TIMEOUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A145, 'STATUS_RUNLEVEL_SWITCH_AGENT_TIMEOUT')
    STATUS_RUNLEVEL_SWITCH_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A146, 'STATUS_RUNLEVEL_SWITCH_IN_PROGRESS')
    STATUS_NOT_APPCONTAINER: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A200, 'STATUS_NOT_APPCONTAINER')
    STATUS_NOT_SUPPORTED_IN_APPCONTAINER: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A201, 'STATUS_NOT_SUPPORTED_IN_APPCONTAINER')
    STATUS_INVALID_PACKAGE_SID_LENGTH: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A202, 'STATUS_INVALID_PACKAGE_SID_LENGTH')
    STATUS_LPAC_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A203, 'STATUS_LPAC_ACCESS_DENIED')
    STATUS_ADMINLESS_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A204, 'STATUS_ADMINLESS_ACCESS_DENIED')
    STATUS_APP_DATA_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A281, 'STATUS_APP_DATA_NOT_FOUND')
    STATUS_APP_DATA_EXPIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A282, 'STATUS_APP_DATA_EXPIRED')
    STATUS_APP_DATA_CORRUPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A283, 'STATUS_APP_DATA_CORRUPT')
    STATUS_APP_DATA_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A284, 'STATUS_APP_DATA_LIMIT_EXCEEDED')
    STATUS_APP_DATA_REBOOT_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A285, 'STATUS_APP_DATA_REBOOT_REQUIRED')
    STATUS_OFFLOAD_READ_FLT_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A2A1, 'STATUS_OFFLOAD_READ_FLT_NOT_SUPPORTED')
    STATUS_OFFLOAD_WRITE_FLT_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A2A2, 'STATUS_OFFLOAD_WRITE_FLT_NOT_SUPPORTED')
    STATUS_OFFLOAD_READ_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A2A3, 'STATUS_OFFLOAD_READ_FILE_NOT_SUPPORTED')
    STATUS_OFFLOAD_WRITE_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A2A4, 'STATUS_OFFLOAD_WRITE_FILE_NOT_SUPPORTED')
    STATUS_WOF_WIM_HEADER_CORRUPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A2A5, 'STATUS_WOF_WIM_HEADER_CORRUPT')
    STATUS_WOF_WIM_RESOURCE_TABLE_CORRUPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A2A6, 'STATUS_WOF_WIM_RESOURCE_TABLE_CORRUPT')
    STATUS_WOF_FILE_RESOURCE_TABLE_CORRUPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000A2A7, 'STATUS_WOF_FILE_RESOURCE_TABLE_CORRUPT')
    STATUS_CIMFS_IMAGE_CORRUPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000C001, 'STATUS_CIMFS_IMAGE_CORRUPT')
    STATUS_CIMFS_IMAGE_VERSION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000C002, 'STATUS_CIMFS_IMAGE_VERSION_NOT_SUPPORTED')
    STATUS_FILE_SYSTEM_VIRTUALIZATION_UNAVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CE01, 'STATUS_FILE_SYSTEM_VIRTUALIZATION_UNAVAILABLE')
    STATUS_FILE_SYSTEM_VIRTUALIZATION_METADATA_CORRUPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CE02, 'STATUS_FILE_SYSTEM_VIRTUALIZATION_METADATA_CORRUPT')
    STATUS_FILE_SYSTEM_VIRTUALIZATION_BUSY: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CE03, 'STATUS_FILE_SYSTEM_VIRTUALIZATION_BUSY')
    STATUS_FILE_SYSTEM_VIRTUALIZATION_PROVIDER_UNKNOWN: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CE04, 'STATUS_FILE_SYSTEM_VIRTUALIZATION_PROVIDER_UNKNOWN')
    STATUS_FILE_SYSTEM_VIRTUALIZATION_INVALID_OPERATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CE05, 'STATUS_FILE_SYSTEM_VIRTUALIZATION_INVALID_OPERATION')
    STATUS_CLOUD_FILE_SYNC_ROOT_METADATA_CORRUPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF00, 'STATUS_CLOUD_FILE_SYNC_ROOT_METADATA_CORRUPT')
    STATUS_CLOUD_FILE_PROVIDER_NOT_RUNNING: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF01, 'STATUS_CLOUD_FILE_PROVIDER_NOT_RUNNING')
    STATUS_CLOUD_FILE_METADATA_CORRUPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF02, 'STATUS_CLOUD_FILE_METADATA_CORRUPT')
    STATUS_CLOUD_FILE_METADATA_TOO_LARGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF03, 'STATUS_CLOUD_FILE_METADATA_TOO_LARGE')
    STATUS_CLOUD_FILE_PROPERTY_VERSION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF06, 'STATUS_CLOUD_FILE_PROPERTY_VERSION_NOT_SUPPORTED')
    STATUS_NOT_A_CLOUD_FILE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF07, 'STATUS_NOT_A_CLOUD_FILE')
    STATUS_CLOUD_FILE_NOT_IN_SYNC: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF08, 'STATUS_CLOUD_FILE_NOT_IN_SYNC')
    STATUS_CLOUD_FILE_ALREADY_CONNECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF09, 'STATUS_CLOUD_FILE_ALREADY_CONNECTED')
    STATUS_CLOUD_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF0A, 'STATUS_CLOUD_FILE_NOT_SUPPORTED')
    STATUS_CLOUD_FILE_INVALID_REQUEST: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF0B, 'STATUS_CLOUD_FILE_INVALID_REQUEST')
    STATUS_CLOUD_FILE_READ_ONLY_VOLUME: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF0C, 'STATUS_CLOUD_FILE_READ_ONLY_VOLUME')
    STATUS_CLOUD_FILE_CONNECTED_PROVIDER_ONLY: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF0D, 'STATUS_CLOUD_FILE_CONNECTED_PROVIDER_ONLY')
    STATUS_CLOUD_FILE_VALIDATION_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF0E, 'STATUS_CLOUD_FILE_VALIDATION_FAILED')
    STATUS_CLOUD_FILE_AUTHENTICATION_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF0F, 'STATUS_CLOUD_FILE_AUTHENTICATION_FAILED')
    STATUS_CLOUD_FILE_INSUFFICIENT_RESOURCES: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF10, 'STATUS_CLOUD_FILE_INSUFFICIENT_RESOURCES')
    STATUS_CLOUD_FILE_NETWORK_UNAVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF11, 'STATUS_CLOUD_FILE_NETWORK_UNAVAILABLE')
    STATUS_CLOUD_FILE_UNSUCCESSFUL: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF12, 'STATUS_CLOUD_FILE_UNSUCCESSFUL')
    STATUS_CLOUD_FILE_NOT_UNDER_SYNC_ROOT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF13, 'STATUS_CLOUD_FILE_NOT_UNDER_SYNC_ROOT')
    STATUS_CLOUD_FILE_IN_USE: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF14, 'STATUS_CLOUD_FILE_IN_USE')
    STATUS_CLOUD_FILE_PINNED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF15, 'STATUS_CLOUD_FILE_PINNED')
    STATUS_CLOUD_FILE_REQUEST_ABORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF16, 'STATUS_CLOUD_FILE_REQUEST_ABORTED')
    STATUS_CLOUD_FILE_PROPERTY_CORRUPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF17, 'STATUS_CLOUD_FILE_PROPERTY_CORRUPT')
    STATUS_CLOUD_FILE_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF18, 'STATUS_CLOUD_FILE_ACCESS_DENIED')
    STATUS_CLOUD_FILE_INCOMPATIBLE_HARDLINKS: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF19, 'STATUS_CLOUD_FILE_INCOMPATIBLE_HARDLINKS')
    STATUS_CLOUD_FILE_PROPERTY_LOCK_CONFLICT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF1A, 'STATUS_CLOUD_FILE_PROPERTY_LOCK_CONFLICT')
    STATUS_CLOUD_FILE_REQUEST_CANCELED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF1B, 'STATUS_CLOUD_FILE_REQUEST_CANCELED')
    STATUS_CLOUD_FILE_PROVIDER_TERMINATED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF1D, 'STATUS_CLOUD_FILE_PROVIDER_TERMINATED')
    STATUS_NOT_A_CLOUD_SYNC_ROOT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF1E, 'STATUS_NOT_A_CLOUD_SYNC_ROOT')
    STATUS_CLOUD_FILE_REQUEST_TIMEOUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF1F, 'STATUS_CLOUD_FILE_REQUEST_TIMEOUT')
    STATUS_CLOUD_FILE_DEHYDRATION_DISALLOWED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF20, 'STATUS_CLOUD_FILE_DEHYDRATION_DISALLOWED')
    STATUS_CLOUD_FILE_US_MESSAGE_TIMEOUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC000CF21, 'STATUS_CLOUD_FILE_US_MESSAGE_TIMEOUT')
    STATUS_FILE_SNAP_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC000F500, 'STATUS_FILE_SNAP_IN_PROGRESS')
    STATUS_FILE_SNAP_USER_SECTION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000F501, 'STATUS_FILE_SNAP_USER_SECTION_NOT_SUPPORTED')
    STATUS_FILE_SNAP_MODIFY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000F502, 'STATUS_FILE_SNAP_MODIFY_NOT_SUPPORTED')
    STATUS_FILE_SNAP_IO_NOT_COORDINATED: Final[ThirtyTwoBits] = cached_ntstatus(0xC000F503, 'STATUS_FILE_SNAP_IO_NOT_COORDINATED')
    STATUS_FILE_SNAP_UNEXPECTED_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC000F504, 'STATUS_FILE_SNAP_UNEXPECTED_ERROR')
    STATUS_FILE_SNAP_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_ntstatus(0xC000F505, 'STATUS_FILE_SNAP_INVALID_PARAMETER')
    STATUS_PNP_BAD_MPS_TABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0040035, 'STATUS_PNP_BAD_MPS_TABLE')
    STATUS_PNP_TRANSLATION_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0040036, 'STATUS_PNP_TRANSLATION_FAILED')
    STATUS_PNP_IRQ_TRANSLATION_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0040037, 'STATUS_PNP_IRQ_TRANSLATION_FAILED')
    STATUS_PNP_INVALID_ID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0040038, 'STATUS_PNP_INVALID_ID')
    STATUS_IO_REISSUE_AS_CACHED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0040039, 'STATUS_IO_REISSUE_AS_CACHED')
    STATUS_CTX_WINSTATION_NAME_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0001, 'STATUS_CTX_WINSTATION_NAME_INVALID')
    STATUS_CTX_INVALID_PD: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0002, 'STATUS_CTX_INVALID_PD')
    STATUS_CTX_PD_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0003, 'STATUS_CTX_PD_NOT_FOUND')
    STATUS_CTX_CLOSE_PENDING: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0006, 'STATUS_CTX_CLOSE_PENDING')
    STATUS_CTX_NO_OUTBUF: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0007, 'STATUS_CTX_NO_OUTBUF')
    STATUS_CTX_MODEM_INF_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0008, 'STATUS_CTX_MODEM_INF_NOT_FOUND')
    STATUS_CTX_INVALID_MODEMNAME: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0009, 'STATUS_CTX_INVALID_MODEMNAME')
    STATUS_CTX_RESPONSE_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A000A, 'STATUS_CTX_RESPONSE_ERROR')
    STATUS_CTX_MODEM_RESPONSE_TIMEOUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A000B, 'STATUS_CTX_MODEM_RESPONSE_TIMEOUT')
    STATUS_CTX_MODEM_RESPONSE_NO_CARRIER: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A000C, 'STATUS_CTX_MODEM_RESPONSE_NO_CARRIER')
    STATUS_CTX_MODEM_RESPONSE_NO_DIALTONE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A000D, 'STATUS_CTX_MODEM_RESPONSE_NO_DIALTONE')
    STATUS_CTX_MODEM_RESPONSE_BUSY: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A000E, 'STATUS_CTX_MODEM_RESPONSE_BUSY')
    STATUS_CTX_MODEM_RESPONSE_VOICE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A000F, 'STATUS_CTX_MODEM_RESPONSE_VOICE')
    STATUS_CTX_TD_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0010, 'STATUS_CTX_TD_ERROR')
    STATUS_CTX_LICENSE_CLIENT_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0012, 'STATUS_CTX_LICENSE_CLIENT_INVALID')
    STATUS_CTX_LICENSE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0013, 'STATUS_CTX_LICENSE_NOT_AVAILABLE')
    STATUS_CTX_LICENSE_EXPIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0014, 'STATUS_CTX_LICENSE_EXPIRED')
    STATUS_CTX_WINSTATION_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0015, 'STATUS_CTX_WINSTATION_NOT_FOUND')
    STATUS_CTX_WINSTATION_NAME_COLLISION: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0016, 'STATUS_CTX_WINSTATION_NAME_COLLISION')
    STATUS_CTX_WINSTATION_BUSY: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0017, 'STATUS_CTX_WINSTATION_BUSY')
    STATUS_CTX_BAD_VIDEO_MODE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0018, 'STATUS_CTX_BAD_VIDEO_MODE')
    STATUS_CTX_GRAPHICS_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0022, 'STATUS_CTX_GRAPHICS_INVALID')
    STATUS_CTX_NOT_CONSOLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0024, 'STATUS_CTX_NOT_CONSOLE')
    STATUS_CTX_CLIENT_QUERY_TIMEOUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0026, 'STATUS_CTX_CLIENT_QUERY_TIMEOUT')
    STATUS_CTX_CONSOLE_DISCONNECT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0027, 'STATUS_CTX_CONSOLE_DISCONNECT')
    STATUS_CTX_CONSOLE_CONNECT: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0028, 'STATUS_CTX_CONSOLE_CONNECT')
    STATUS_CTX_SHADOW_DENIED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A002A, 'STATUS_CTX_SHADOW_DENIED')
    STATUS_CTX_WINSTATION_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A002B, 'STATUS_CTX_WINSTATION_ACCESS_DENIED')
    STATUS_CTX_INVALID_WD: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A002E, 'STATUS_CTX_INVALID_WD')
    STATUS_CTX_WD_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A002F, 'STATUS_CTX_WD_NOT_FOUND')
    STATUS_CTX_SHADOW_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0030, 'STATUS_CTX_SHADOW_INVALID')
    STATUS_CTX_SHADOW_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0031, 'STATUS_CTX_SHADOW_DISABLED')
    STATUS_RDP_PROTOCOL_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0032, 'STATUS_RDP_PROTOCOL_ERROR')
    STATUS_CTX_CLIENT_LICENSE_NOT_SET: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0033, 'STATUS_CTX_CLIENT_LICENSE_NOT_SET')
    STATUS_CTX_CLIENT_LICENSE_IN_USE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0034, 'STATUS_CTX_CLIENT_LICENSE_IN_USE')
    STATUS_CTX_SHADOW_ENDED_BY_MODE_CHANGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0035, 'STATUS_CTX_SHADOW_ENDED_BY_MODE_CHANGE')
    STATUS_CTX_SHADOW_NOT_RUNNING: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0036, 'STATUS_CTX_SHADOW_NOT_RUNNING')
    STATUS_CTX_LOGON_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0037, 'STATUS_CTX_LOGON_DISABLED')
    STATUS_CTX_SECURITY_LAYER_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0038, 'STATUS_CTX_SECURITY_LAYER_ERROR')
    STATUS_TS_INCOMPATIBLE_SESSIONS: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A0039, 'STATUS_TS_INCOMPATIBLE_SESSIONS')
    STATUS_TS_VIDEO_SUBSYSTEM_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC00A003A, 'STATUS_TS_VIDEO_SUBSYSTEM_ERROR')
    STATUS_MUI_FILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC00B0001, 'STATUS_MUI_FILE_NOT_FOUND')
    STATUS_MUI_INVALID_FILE: Final[ThirtyTwoBits] = cached_ntstatus(0xC00B0002, 'STATUS_MUI_INVALID_FILE')
    STATUS_MUI_INVALID_RC_CONFIG: Final[ThirtyTwoBits] = cached_ntstatus(0xC00B0003, 'STATUS_MUI_INVALID_RC_CONFIG')
    STATUS_MUI_INVALID_LOCALE_NAME: Final[ThirtyTwoBits] = cached_ntstatus(0xC00B0004, 'STATUS_MUI_INVALID_LOCALE_NAME')
    STATUS_MUI_INVALID_ULTIMATEFALLBACK_NAME: Final[ThirtyTwoBits] = cached_ntstatus(0xC00B0005, 'STATUS_MUI_INVALID_ULTIMATEFALLBACK_NAME')
    STATUS_MUI_FILE_NOT_LOADED: Final[ThirtyTwoBits] = cached_ntstatus(0xC00B0006, 'STATUS_MUI_FILE_NOT_LOADED')
    STATUS_RESOURCE_ENUM_USER_STOP: Final[ThirtyTwoBits] = cached_ntstatus(0xC00B0007, 'STATUS_RESOURCE_ENUM_USER_STOP')
    STATUS_CLUSTER_INVALID_NODE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130001, 'STATUS_CLUSTER_INVALID_NODE')
    STATUS_CLUSTER_NODE_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130002, 'STATUS_CLUSTER_NODE_EXISTS')
    STATUS_CLUSTER_JOIN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130003, 'STATUS_CLUSTER_JOIN_IN_PROGRESS')
    STATUS_CLUSTER_NODE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130004, 'STATUS_CLUSTER_NODE_NOT_FOUND')
    STATUS_CLUSTER_LOCAL_NODE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130005, 'STATUS_CLUSTER_LOCAL_NODE_NOT_FOUND')
    STATUS_CLUSTER_NETWORK_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130006, 'STATUS_CLUSTER_NETWORK_EXISTS')
    STATUS_CLUSTER_NETWORK_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130007, 'STATUS_CLUSTER_NETWORK_NOT_FOUND')
    STATUS_CLUSTER_NETINTERFACE_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130008, 'STATUS_CLUSTER_NETINTERFACE_EXISTS')
    STATUS_CLUSTER_NETINTERFACE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130009, 'STATUS_CLUSTER_NETINTERFACE_NOT_FOUND')
    STATUS_CLUSTER_INVALID_REQUEST: Final[ThirtyTwoBits] = cached_ntstatus(0xC013000A, 'STATUS_CLUSTER_INVALID_REQUEST')
    STATUS_CLUSTER_INVALID_NETWORK_PROVIDER: Final[ThirtyTwoBits] = cached_ntstatus(0xC013000B, 'STATUS_CLUSTER_INVALID_NETWORK_PROVIDER')
    STATUS_CLUSTER_NODE_DOWN: Final[ThirtyTwoBits] = cached_ntstatus(0xC013000C, 'STATUS_CLUSTER_NODE_DOWN')
    STATUS_CLUSTER_NODE_UNREACHABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC013000D, 'STATUS_CLUSTER_NODE_UNREACHABLE')
    STATUS_CLUSTER_NODE_NOT_MEMBER: Final[ThirtyTwoBits] = cached_ntstatus(0xC013000E, 'STATUS_CLUSTER_NODE_NOT_MEMBER')
    STATUS_CLUSTER_JOIN_NOT_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC013000F, 'STATUS_CLUSTER_JOIN_NOT_IN_PROGRESS')
    STATUS_CLUSTER_INVALID_NETWORK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130010, 'STATUS_CLUSTER_INVALID_NETWORK')
    STATUS_CLUSTER_NO_NET_ADAPTERS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130011, 'STATUS_CLUSTER_NO_NET_ADAPTERS')
    STATUS_CLUSTER_NODE_UP: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130012, 'STATUS_CLUSTER_NODE_UP')
    STATUS_CLUSTER_NODE_PAUSED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130013, 'STATUS_CLUSTER_NODE_PAUSED')
    STATUS_CLUSTER_NODE_NOT_PAUSED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130014, 'STATUS_CLUSTER_NODE_NOT_PAUSED')
    STATUS_CLUSTER_NO_SECURITY_CONTEXT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130015, 'STATUS_CLUSTER_NO_SECURITY_CONTEXT')
    STATUS_CLUSTER_NETWORK_NOT_INTERNAL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130016, 'STATUS_CLUSTER_NETWORK_NOT_INTERNAL')
    STATUS_CLUSTER_POISONED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130017, 'STATUS_CLUSTER_POISONED')
    STATUS_CLUSTER_NON_CSV_PATH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130018, 'STATUS_CLUSTER_NON_CSV_PATH')
    STATUS_CLUSTER_CSV_VOLUME_NOT_LOCAL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130019, 'STATUS_CLUSTER_CSV_VOLUME_NOT_LOCAL')
    STATUS_CLUSTER_CSV_READ_OPLOCK_BREAK_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130020, 'STATUS_CLUSTER_CSV_READ_OPLOCK_BREAK_IN_PROGRESS')
    STATUS_CLUSTER_CSV_AUTO_PAUSE_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130021, 'STATUS_CLUSTER_CSV_AUTO_PAUSE_ERROR')
    STATUS_CLUSTER_CSV_REDIRECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130022, 'STATUS_CLUSTER_CSV_REDIRECTED')
    STATUS_CLUSTER_CSV_NOT_REDIRECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130023, 'STATUS_CLUSTER_CSV_NOT_REDIRECTED')
    STATUS_CLUSTER_CSV_VOLUME_DRAINING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130024, 'STATUS_CLUSTER_CSV_VOLUME_DRAINING')
    STATUS_CLUSTER_CSV_SNAPSHOT_CREATION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130025, 'STATUS_CLUSTER_CSV_SNAPSHOT_CREATION_IN_PROGRESS')
    STATUS_CLUSTER_CSV_VOLUME_DRAINING_SUCCEEDED_DOWNLEVEL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130026, 'STATUS_CLUSTER_CSV_VOLUME_DRAINING_SUCCEEDED_DOWNLEVEL')
    STATUS_CLUSTER_CSV_NO_SNAPSHOTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130027, 'STATUS_CLUSTER_CSV_NO_SNAPSHOTS')
    STATUS_CSV_IO_PAUSE_TIMEOUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130028, 'STATUS_CSV_IO_PAUSE_TIMEOUT')
    STATUS_CLUSTER_CSV_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130029, 'STATUS_CLUSTER_CSV_INVALID_HANDLE')
    STATUS_CLUSTER_CSV_SUPPORTED_ONLY_ON_COORDINATOR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130030, 'STATUS_CLUSTER_CSV_SUPPORTED_ONLY_ON_COORDINATOR')
    STATUS_CLUSTER_CAM_TICKET_REPLAY_DETECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0130031, 'STATUS_CLUSTER_CAM_TICKET_REPLAY_DETECTED')
    STATUS_ACPI_INVALID_OPCODE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0140001, 'STATUS_ACPI_INVALID_OPCODE')
    STATUS_ACPI_STACK_OVERFLOW: Final[ThirtyTwoBits] = cached_ntstatus(0xC0140002, 'STATUS_ACPI_STACK_OVERFLOW')
    STATUS_ACPI_ASSERT_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0140003, 'STATUS_ACPI_ASSERT_FAILED')
    STATUS_ACPI_INVALID_INDEX: Final[ThirtyTwoBits] = cached_ntstatus(0xC0140004, 'STATUS_ACPI_INVALID_INDEX')
    STATUS_ACPI_INVALID_ARGUMENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0140005, 'STATUS_ACPI_INVALID_ARGUMENT')
    STATUS_ACPI_FATAL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0140006, 'STATUS_ACPI_FATAL')
    STATUS_ACPI_INVALID_SUPERNAME: Final[ThirtyTwoBits] = cached_ntstatus(0xC0140007, 'STATUS_ACPI_INVALID_SUPERNAME')
    STATUS_ACPI_INVALID_ARGTYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0140008, 'STATUS_ACPI_INVALID_ARGTYPE')
    STATUS_ACPI_INVALID_OBJTYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0140009, 'STATUS_ACPI_INVALID_OBJTYPE')
    STATUS_ACPI_INVALID_TARGETTYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC014000A, 'STATUS_ACPI_INVALID_TARGETTYPE')
    STATUS_ACPI_INCORRECT_ARGUMENT_COUNT: Final[ThirtyTwoBits] = cached_ntstatus(0xC014000B, 'STATUS_ACPI_INCORRECT_ARGUMENT_COUNT')
    STATUS_ACPI_ADDRESS_NOT_MAPPED: Final[ThirtyTwoBits] = cached_ntstatus(0xC014000C, 'STATUS_ACPI_ADDRESS_NOT_MAPPED')
    STATUS_ACPI_INVALID_EVENTTYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC014000D, 'STATUS_ACPI_INVALID_EVENTTYPE')
    STATUS_ACPI_HANDLER_COLLISION: Final[ThirtyTwoBits] = cached_ntstatus(0xC014000E, 'STATUS_ACPI_HANDLER_COLLISION')
    STATUS_ACPI_INVALID_DATA: Final[ThirtyTwoBits] = cached_ntstatus(0xC014000F, 'STATUS_ACPI_INVALID_DATA')
    STATUS_ACPI_INVALID_REGION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0140010, 'STATUS_ACPI_INVALID_REGION')
    STATUS_ACPI_INVALID_ACCESS_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0140011, 'STATUS_ACPI_INVALID_ACCESS_SIZE')
    STATUS_ACPI_ACQUIRE_GLOBAL_LOCK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0140012, 'STATUS_ACPI_ACQUIRE_GLOBAL_LOCK')
    STATUS_ACPI_ALREADY_INITIALIZED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0140013, 'STATUS_ACPI_ALREADY_INITIALIZED')
    STATUS_ACPI_NOT_INITIALIZED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0140014, 'STATUS_ACPI_NOT_INITIALIZED')
    STATUS_ACPI_INVALID_MUTEX_LEVEL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0140015, 'STATUS_ACPI_INVALID_MUTEX_LEVEL')
    STATUS_ACPI_MUTEX_NOT_OWNED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0140016, 'STATUS_ACPI_MUTEX_NOT_OWNED')
    STATUS_ACPI_MUTEX_NOT_OWNER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0140017, 'STATUS_ACPI_MUTEX_NOT_OWNER')
    STATUS_ACPI_RS_ACCESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0140018, 'STATUS_ACPI_RS_ACCESS')
    STATUS_ACPI_INVALID_TABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0140019, 'STATUS_ACPI_INVALID_TABLE')
    STATUS_ACPI_REG_HANDLER_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0140020, 'STATUS_ACPI_REG_HANDLER_FAILED')
    STATUS_ACPI_POWER_REQUEST_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0140021, 'STATUS_ACPI_POWER_REQUEST_FAILED')
    STATUS_SXS_SECTION_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150001, 'STATUS_SXS_SECTION_NOT_FOUND')
    STATUS_SXS_CANT_GEN_ACTCTX: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150002, 'STATUS_SXS_CANT_GEN_ACTCTX')
    STATUS_SXS_INVALID_ACTCTXDATA_FORMAT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150003, 'STATUS_SXS_INVALID_ACTCTXDATA_FORMAT')
    STATUS_SXS_ASSEMBLY_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150004, 'STATUS_SXS_ASSEMBLY_NOT_FOUND')
    STATUS_SXS_MANIFEST_FORMAT_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150005, 'STATUS_SXS_MANIFEST_FORMAT_ERROR')
    STATUS_SXS_MANIFEST_PARSE_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150006, 'STATUS_SXS_MANIFEST_PARSE_ERROR')
    STATUS_SXS_ACTIVATION_CONTEXT_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150007, 'STATUS_SXS_ACTIVATION_CONTEXT_DISABLED')
    STATUS_SXS_KEY_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150008, 'STATUS_SXS_KEY_NOT_FOUND')
    STATUS_SXS_VERSION_CONFLICT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150009, 'STATUS_SXS_VERSION_CONFLICT')
    STATUS_SXS_WRONG_SECTION_TYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC015000A, 'STATUS_SXS_WRONG_SECTION_TYPE')
    STATUS_SXS_THREAD_QUERIES_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC015000B, 'STATUS_SXS_THREAD_QUERIES_DISABLED')
    STATUS_SXS_ASSEMBLY_MISSING: Final[ThirtyTwoBits] = cached_ntstatus(0xC015000C, 'STATUS_SXS_ASSEMBLY_MISSING')
    STATUS_SXS_PROCESS_DEFAULT_ALREADY_SET: Final[ThirtyTwoBits] = cached_ntstatus(0xC015000E, 'STATUS_SXS_PROCESS_DEFAULT_ALREADY_SET')
    STATUS_SXS_EARLY_DEACTIVATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC015000F, 'STATUS_SXS_EARLY_DEACTIVATION')
    STATUS_SXS_INVALID_DEACTIVATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150010, 'STATUS_SXS_INVALID_DEACTIVATION')
    STATUS_SXS_MULTIPLE_DEACTIVATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150011, 'STATUS_SXS_MULTIPLE_DEACTIVATION')
    STATUS_SXS_SYSTEM_DEFAULT_ACTIVATION_CONTEXT_EMPTY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150012, 'STATUS_SXS_SYSTEM_DEFAULT_ACTIVATION_CONTEXT_EMPTY')
    STATUS_SXS_PROCESS_TERMINATION_REQUESTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150013, 'STATUS_SXS_PROCESS_TERMINATION_REQUESTED')
    STATUS_SXS_CORRUPT_ACTIVATION_STACK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150014, 'STATUS_SXS_CORRUPT_ACTIVATION_STACK')
    STATUS_SXS_CORRUPTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150015, 'STATUS_SXS_CORRUPTION')
    STATUS_SXS_INVALID_IDENTITY_ATTRIBUTE_VALUE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150016, 'STATUS_SXS_INVALID_IDENTITY_ATTRIBUTE_VALUE')
    STATUS_SXS_INVALID_IDENTITY_ATTRIBUTE_NAME: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150017, 'STATUS_SXS_INVALID_IDENTITY_ATTRIBUTE_NAME')
    STATUS_SXS_IDENTITY_DUPLICATE_ATTRIBUTE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150018, 'STATUS_SXS_IDENTITY_DUPLICATE_ATTRIBUTE')
    STATUS_SXS_IDENTITY_PARSE_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150019, 'STATUS_SXS_IDENTITY_PARSE_ERROR')
    STATUS_SXS_COMPONENT_STORE_CORRUPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC015001A, 'STATUS_SXS_COMPONENT_STORE_CORRUPT')
    STATUS_SXS_FILE_HASH_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC015001B, 'STATUS_SXS_FILE_HASH_MISMATCH')
    STATUS_SXS_MANIFEST_IDENTITY_SAME_BUT_CONTENTS_DIFFERENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC015001C, 'STATUS_SXS_MANIFEST_IDENTITY_SAME_BUT_CONTENTS_DIFFERENT')
    STATUS_SXS_IDENTITIES_DIFFERENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC015001D, 'STATUS_SXS_IDENTITIES_DIFFERENT')
    STATUS_SXS_ASSEMBLY_IS_NOT_A_DEPLOYMENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC015001E, 'STATUS_SXS_ASSEMBLY_IS_NOT_A_DEPLOYMENT')
    STATUS_SXS_FILE_NOT_PART_OF_ASSEMBLY: Final[ThirtyTwoBits] = cached_ntstatus(0xC015001F, 'STATUS_SXS_FILE_NOT_PART_OF_ASSEMBLY')
    STATUS_ADVANCED_INSTALLER_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150020, 'STATUS_ADVANCED_INSTALLER_FAILED')
    STATUS_XML_ENCODING_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150021, 'STATUS_XML_ENCODING_MISMATCH')
    STATUS_SXS_MANIFEST_TOO_BIG: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150022, 'STATUS_SXS_MANIFEST_TOO_BIG')
    STATUS_SXS_SETTING_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150023, 'STATUS_SXS_SETTING_NOT_REGISTERED')
    STATUS_SXS_TRANSACTION_CLOSURE_INCOMPLETE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150024, 'STATUS_SXS_TRANSACTION_CLOSURE_INCOMPLETE')
    STATUS_SMI_PRIMITIVE_INSTALLER_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150025, 'STATUS_SMI_PRIMITIVE_INSTALLER_FAILED')
    STATUS_GENERIC_COMMAND_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150026, 'STATUS_GENERIC_COMMAND_FAILED')
    STATUS_SXS_FILE_HASH_MISSING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0150027, 'STATUS_SXS_FILE_HASH_MISSING')
    STATUS_TRANSACTIONAL_CONFLICT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190001, 'STATUS_TRANSACTIONAL_CONFLICT')
    STATUS_INVALID_TRANSACTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190002, 'STATUS_INVALID_TRANSACTION')
    STATUS_TRANSACTION_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190003, 'STATUS_TRANSACTION_NOT_ACTIVE')
    STATUS_TM_INITIALIZATION_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190004, 'STATUS_TM_INITIALIZATION_FAILED')
    STATUS_RM_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190005, 'STATUS_RM_NOT_ACTIVE')
    STATUS_RM_METADATA_CORRUPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190006, 'STATUS_RM_METADATA_CORRUPT')
    STATUS_TRANSACTION_NOT_JOINED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190007, 'STATUS_TRANSACTION_NOT_JOINED')
    STATUS_DIRECTORY_NOT_RM: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190008, 'STATUS_DIRECTORY_NOT_RM')
    STATUS_TRANSACTIONS_UNSUPPORTED_REMOTE: Final[ThirtyTwoBits] = cached_ntstatus(0xC019000A, 'STATUS_TRANSACTIONS_UNSUPPORTED_REMOTE')
    STATUS_LOG_RESIZE_INVALID_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC019000B, 'STATUS_LOG_RESIZE_INVALID_SIZE')
    STATUS_REMOTE_FILE_VERSION_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC019000C, 'STATUS_REMOTE_FILE_VERSION_MISMATCH')
    STATUS_CRM_PROTOCOL_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC019000F, 'STATUS_CRM_PROTOCOL_ALREADY_EXISTS')
    STATUS_TRANSACTION_PROPAGATION_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190010, 'STATUS_TRANSACTION_PROPAGATION_FAILED')
    STATUS_CRM_PROTOCOL_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190011, 'STATUS_CRM_PROTOCOL_NOT_FOUND')
    STATUS_TRANSACTION_SUPERIOR_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190012, 'STATUS_TRANSACTION_SUPERIOR_EXISTS')
    STATUS_TRANSACTION_REQUEST_NOT_VALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190013, 'STATUS_TRANSACTION_REQUEST_NOT_VALID')
    STATUS_TRANSACTION_NOT_REQUESTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190014, 'STATUS_TRANSACTION_NOT_REQUESTED')
    STATUS_TRANSACTION_ALREADY_ABORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190015, 'STATUS_TRANSACTION_ALREADY_ABORTED')
    STATUS_TRANSACTION_ALREADY_COMMITTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190016, 'STATUS_TRANSACTION_ALREADY_COMMITTED')
    STATUS_TRANSACTION_INVALID_MARSHALL_BUFFER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190017, 'STATUS_TRANSACTION_INVALID_MARSHALL_BUFFER')
    STATUS_CURRENT_TRANSACTION_NOT_VALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190018, 'STATUS_CURRENT_TRANSACTION_NOT_VALID')
    STATUS_LOG_GROWTH_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190019, 'STATUS_LOG_GROWTH_FAILED')
    STATUS_OBJECT_NO_LONGER_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190021, 'STATUS_OBJECT_NO_LONGER_EXISTS')
    STATUS_STREAM_MINIVERSION_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190022, 'STATUS_STREAM_MINIVERSION_NOT_FOUND')
    STATUS_STREAM_MINIVERSION_NOT_VALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190023, 'STATUS_STREAM_MINIVERSION_NOT_VALID')
    STATUS_MINIVERSION_INACCESSIBLE_FROM_SPECIFIED_TRANSACTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190024, 'STATUS_MINIVERSION_INACCESSIBLE_FROM_SPECIFIED_TRANSACTION')
    STATUS_CANT_OPEN_MINIVERSION_WITH_MODIFY_INTENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190025, 'STATUS_CANT_OPEN_MINIVERSION_WITH_MODIFY_INTENT')
    STATUS_CANT_CREATE_MORE_STREAM_MINIVERSIONS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190026, 'STATUS_CANT_CREATE_MORE_STREAM_MINIVERSIONS')
    STATUS_HANDLE_NO_LONGER_VALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190028, 'STATUS_HANDLE_NO_LONGER_VALID')
    STATUS_LOG_CORRUPTION_DETECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190030, 'STATUS_LOG_CORRUPTION_DETECTED')
    STATUS_RM_DISCONNECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190032, 'STATUS_RM_DISCONNECTED')
    STATUS_ENLISTMENT_NOT_SUPERIOR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190033, 'STATUS_ENLISTMENT_NOT_SUPERIOR')
    STATUS_FILE_IDENTITY_NOT_PERSISTENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190036, 'STATUS_FILE_IDENTITY_NOT_PERSISTENT')
    STATUS_CANT_BREAK_TRANSACTIONAL_DEPENDENCY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190037, 'STATUS_CANT_BREAK_TRANSACTIONAL_DEPENDENCY')
    STATUS_CANT_CROSS_RM_BOUNDARY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190038, 'STATUS_CANT_CROSS_RM_BOUNDARY')
    STATUS_TXF_DIR_NOT_EMPTY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190039, 'STATUS_TXF_DIR_NOT_EMPTY')
    STATUS_INDOUBT_TRANSACTIONS_EXIST: Final[ThirtyTwoBits] = cached_ntstatus(0xC019003A, 'STATUS_INDOUBT_TRANSACTIONS_EXIST')
    STATUS_TM_VOLATILE: Final[ThirtyTwoBits] = cached_ntstatus(0xC019003B, 'STATUS_TM_VOLATILE')
    STATUS_ROLLBACK_TIMER_EXPIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC019003C, 'STATUS_ROLLBACK_TIMER_EXPIRED')
    STATUS_TXF_ATTRIBUTE_CORRUPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC019003D, 'STATUS_TXF_ATTRIBUTE_CORRUPT')
    STATUS_EFS_NOT_ALLOWED_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC019003E, 'STATUS_EFS_NOT_ALLOWED_IN_TRANSACTION')
    STATUS_TRANSACTIONAL_OPEN_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_ntstatus(0xC019003F, 'STATUS_TRANSACTIONAL_OPEN_NOT_ALLOWED')
    STATUS_TRANSACTED_MAPPING_UNSUPPORTED_REMOTE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190040, 'STATUS_TRANSACTED_MAPPING_UNSUPPORTED_REMOTE')
    STATUS_TRANSACTION_REQUIRED_PROMOTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190043, 'STATUS_TRANSACTION_REQUIRED_PROMOTION')
    STATUS_CANNOT_EXECUTE_FILE_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190044, 'STATUS_CANNOT_EXECUTE_FILE_IN_TRANSACTION')
    STATUS_TRANSACTIONS_NOT_FROZEN: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190045, 'STATUS_TRANSACTIONS_NOT_FROZEN')
    STATUS_TRANSACTION_FREEZE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190046, 'STATUS_TRANSACTION_FREEZE_IN_PROGRESS')
    STATUS_NOT_SNAPSHOT_VOLUME: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190047, 'STATUS_NOT_SNAPSHOT_VOLUME')
    STATUS_NO_SAVEPOINT_WITH_OPEN_FILES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190048, 'STATUS_NO_SAVEPOINT_WITH_OPEN_FILES')
    STATUS_SPARSE_NOT_ALLOWED_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190049, 'STATUS_SPARSE_NOT_ALLOWED_IN_TRANSACTION')
    STATUS_TM_IDENTITY_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC019004A, 'STATUS_TM_IDENTITY_MISMATCH')
    STATUS_FLOATED_SECTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC019004B, 'STATUS_FLOATED_SECTION')
    STATUS_CANNOT_ACCEPT_TRANSACTED_WORK: Final[ThirtyTwoBits] = cached_ntstatus(0xC019004C, 'STATUS_CANNOT_ACCEPT_TRANSACTED_WORK')
    STATUS_CANNOT_ABORT_TRANSACTIONS: Final[ThirtyTwoBits] = cached_ntstatus(0xC019004D, 'STATUS_CANNOT_ABORT_TRANSACTIONS')
    STATUS_TRANSACTION_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC019004E, 'STATUS_TRANSACTION_NOT_FOUND')
    STATUS_RESOURCEMANAGER_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC019004F, 'STATUS_RESOURCEMANAGER_NOT_FOUND')
    STATUS_ENLISTMENT_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190050, 'STATUS_ENLISTMENT_NOT_FOUND')
    STATUS_TRANSACTIONMANAGER_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190051, 'STATUS_TRANSACTIONMANAGER_NOT_FOUND')
    STATUS_TRANSACTIONMANAGER_NOT_ONLINE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190052, 'STATUS_TRANSACTIONMANAGER_NOT_ONLINE')
    STATUS_TRANSACTIONMANAGER_RECOVERY_NAME_COLLISION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190053, 'STATUS_TRANSACTIONMANAGER_RECOVERY_NAME_COLLISION')
    STATUS_TRANSACTION_NOT_ROOT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190054, 'STATUS_TRANSACTION_NOT_ROOT')
    STATUS_TRANSACTION_OBJECT_EXPIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190055, 'STATUS_TRANSACTION_OBJECT_EXPIRED')
    STATUS_COMPRESSION_NOT_ALLOWED_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190056, 'STATUS_COMPRESSION_NOT_ALLOWED_IN_TRANSACTION')
    STATUS_TRANSACTION_RESPONSE_NOT_ENLISTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190057, 'STATUS_TRANSACTION_RESPONSE_NOT_ENLISTED')
    STATUS_TRANSACTION_RECORD_TOO_LONG: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190058, 'STATUS_TRANSACTION_RECORD_TOO_LONG')
    STATUS_NO_LINK_TRACKING_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190059, 'STATUS_NO_LINK_TRACKING_IN_TRANSACTION')
    STATUS_OPERATION_NOT_SUPPORTED_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC019005A, 'STATUS_OPERATION_NOT_SUPPORTED_IN_TRANSACTION')
    STATUS_TRANSACTION_INTEGRITY_VIOLATED: Final[ThirtyTwoBits] = cached_ntstatus(0xC019005B, 'STATUS_TRANSACTION_INTEGRITY_VIOLATED')
    STATUS_TRANSACTIONMANAGER_IDENTITY_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC019005C, 'STATUS_TRANSACTIONMANAGER_IDENTITY_MISMATCH')
    STATUS_RM_CANNOT_BE_FROZEN_FOR_SNAPSHOT: Final[ThirtyTwoBits] = cached_ntstatus(0xC019005D, 'STATUS_RM_CANNOT_BE_FROZEN_FOR_SNAPSHOT')
    STATUS_TRANSACTION_MUST_WRITETHROUGH: Final[ThirtyTwoBits] = cached_ntstatus(0xC019005E, 'STATUS_TRANSACTION_MUST_WRITETHROUGH')
    STATUS_TRANSACTION_NO_SUPERIOR: Final[ThirtyTwoBits] = cached_ntstatus(0xC019005F, 'STATUS_TRANSACTION_NO_SUPERIOR')
    STATUS_EXPIRED_HANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190060, 'STATUS_EXPIRED_HANDLE')
    STATUS_TRANSACTION_NOT_ENLISTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0190061, 'STATUS_TRANSACTION_NOT_ENLISTED')
    STATUS_LOG_SECTOR_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0001, 'STATUS_LOG_SECTOR_INVALID')
    STATUS_LOG_SECTOR_PARITY_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0002, 'STATUS_LOG_SECTOR_PARITY_INVALID')
    STATUS_LOG_SECTOR_REMAPPED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0003, 'STATUS_LOG_SECTOR_REMAPPED')
    STATUS_LOG_BLOCK_INCOMPLETE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0004, 'STATUS_LOG_BLOCK_INCOMPLETE')
    STATUS_LOG_INVALID_RANGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0005, 'STATUS_LOG_INVALID_RANGE')
    STATUS_LOG_BLOCKS_EXHAUSTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0006, 'STATUS_LOG_BLOCKS_EXHAUSTED')
    STATUS_LOG_READ_CONTEXT_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0007, 'STATUS_LOG_READ_CONTEXT_INVALID')
    STATUS_LOG_RESTART_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0008, 'STATUS_LOG_RESTART_INVALID')
    STATUS_LOG_BLOCK_VERSION: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0009, 'STATUS_LOG_BLOCK_VERSION')
    STATUS_LOG_BLOCK_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A000A, 'STATUS_LOG_BLOCK_INVALID')
    STATUS_LOG_READ_MODE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A000B, 'STATUS_LOG_READ_MODE_INVALID')
    STATUS_LOG_METADATA_CORRUPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A000D, 'STATUS_LOG_METADATA_CORRUPT')
    STATUS_LOG_METADATA_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A000E, 'STATUS_LOG_METADATA_INVALID')
    STATUS_LOG_METADATA_INCONSISTENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A000F, 'STATUS_LOG_METADATA_INCONSISTENT')
    STATUS_LOG_RESERVATION_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0010, 'STATUS_LOG_RESERVATION_INVALID')
    STATUS_LOG_CANT_DELETE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0011, 'STATUS_LOG_CANT_DELETE')
    STATUS_LOG_CONTAINER_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0012, 'STATUS_LOG_CONTAINER_LIMIT_EXCEEDED')
    STATUS_LOG_START_OF_LOG: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0013, 'STATUS_LOG_START_OF_LOG')
    STATUS_LOG_POLICY_ALREADY_INSTALLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0014, 'STATUS_LOG_POLICY_ALREADY_INSTALLED')
    STATUS_LOG_POLICY_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0015, 'STATUS_LOG_POLICY_NOT_INSTALLED')
    STATUS_LOG_POLICY_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0016, 'STATUS_LOG_POLICY_INVALID')
    STATUS_LOG_POLICY_CONFLICT: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0017, 'STATUS_LOG_POLICY_CONFLICT')
    STATUS_LOG_PINNED_ARCHIVE_TAIL: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0018, 'STATUS_LOG_PINNED_ARCHIVE_TAIL')
    STATUS_LOG_RECORD_NONEXISTENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0019, 'STATUS_LOG_RECORD_NONEXISTENT')
    STATUS_LOG_RECORDS_RESERVED_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A001A, 'STATUS_LOG_RECORDS_RESERVED_INVALID')
    STATUS_LOG_SPACE_RESERVED_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A001B, 'STATUS_LOG_SPACE_RESERVED_INVALID')
    STATUS_LOG_TAIL_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A001C, 'STATUS_LOG_TAIL_INVALID')
    STATUS_LOG_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A001D, 'STATUS_LOG_FULL')
    STATUS_LOG_MULTIPLEXED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A001E, 'STATUS_LOG_MULTIPLEXED')
    STATUS_LOG_DEDICATED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A001F, 'STATUS_LOG_DEDICATED')
    STATUS_LOG_ARCHIVE_NOT_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0020, 'STATUS_LOG_ARCHIVE_NOT_IN_PROGRESS')
    STATUS_LOG_ARCHIVE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0021, 'STATUS_LOG_ARCHIVE_IN_PROGRESS')
    STATUS_LOG_EPHEMERAL: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0022, 'STATUS_LOG_EPHEMERAL')
    STATUS_LOG_NOT_ENOUGH_CONTAINERS: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0023, 'STATUS_LOG_NOT_ENOUGH_CONTAINERS')
    STATUS_LOG_CLIENT_ALREADY_REGISTERED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0024, 'STATUS_LOG_CLIENT_ALREADY_REGISTERED')
    STATUS_LOG_CLIENT_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0025, 'STATUS_LOG_CLIENT_NOT_REGISTERED')
    STATUS_LOG_FULL_HANDLER_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0026, 'STATUS_LOG_FULL_HANDLER_IN_PROGRESS')
    STATUS_LOG_CONTAINER_READ_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0027, 'STATUS_LOG_CONTAINER_READ_FAILED')
    STATUS_LOG_CONTAINER_WRITE_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0028, 'STATUS_LOG_CONTAINER_WRITE_FAILED')
    STATUS_LOG_CONTAINER_OPEN_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0029, 'STATUS_LOG_CONTAINER_OPEN_FAILED')
    STATUS_LOG_CONTAINER_STATE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A002A, 'STATUS_LOG_CONTAINER_STATE_INVALID')
    STATUS_LOG_STATE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A002B, 'STATUS_LOG_STATE_INVALID')
    STATUS_LOG_PINNED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A002C, 'STATUS_LOG_PINNED')
    STATUS_LOG_METADATA_FLUSH_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A002D, 'STATUS_LOG_METADATA_FLUSH_FAILED')
    STATUS_LOG_INCONSISTENT_SECURITY: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A002E, 'STATUS_LOG_INCONSISTENT_SECURITY')
    STATUS_LOG_APPENDED_FLUSH_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A002F, 'STATUS_LOG_APPENDED_FLUSH_FAILED')
    STATUS_LOG_PINNED_RESERVATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC01A0030, 'STATUS_LOG_PINNED_RESERVATION')
    STATUS_VIDEO_HUNG_DISPLAY_DRIVER_THREAD: Final[ThirtyTwoBits] = cached_ntstatus(0xC01B00EA, 'STATUS_VIDEO_HUNG_DISPLAY_DRIVER_THREAD')
    STATUS_FLT_NO_HANDLER_DEFINED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0001, 'STATUS_FLT_NO_HANDLER_DEFINED')
    STATUS_FLT_CONTEXT_ALREADY_DEFINED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0002, 'STATUS_FLT_CONTEXT_ALREADY_DEFINED')
    STATUS_FLT_INVALID_ASYNCHRONOUS_REQUEST: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0003, 'STATUS_FLT_INVALID_ASYNCHRONOUS_REQUEST')
    STATUS_FLT_DISALLOW_FAST_IO: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0004, 'STATUS_FLT_DISALLOW_FAST_IO')
    STATUS_FLT_INVALID_NAME_REQUEST: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0005, 'STATUS_FLT_INVALID_NAME_REQUEST')
    STATUS_FLT_NOT_SAFE_TO_POST_OPERATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0006, 'STATUS_FLT_NOT_SAFE_TO_POST_OPERATION')
    STATUS_FLT_NOT_INITIALIZED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0007, 'STATUS_FLT_NOT_INITIALIZED')
    STATUS_FLT_FILTER_NOT_READY: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0008, 'STATUS_FLT_FILTER_NOT_READY')
    STATUS_FLT_POST_OPERATION_CLEANUP: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0009, 'STATUS_FLT_POST_OPERATION_CLEANUP')
    STATUS_FLT_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C000A, 'STATUS_FLT_INTERNAL_ERROR')
    STATUS_FLT_DELETING_OBJECT: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C000B, 'STATUS_FLT_DELETING_OBJECT')
    STATUS_FLT_MUST_BE_NONPAGED_POOL: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C000C, 'STATUS_FLT_MUST_BE_NONPAGED_POOL')
    STATUS_FLT_DUPLICATE_ENTRY: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C000D, 'STATUS_FLT_DUPLICATE_ENTRY')
    STATUS_FLT_CBDQ_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C000E, 'STATUS_FLT_CBDQ_DISABLED')
    STATUS_FLT_DO_NOT_ATTACH: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C000F, 'STATUS_FLT_DO_NOT_ATTACH')
    STATUS_FLT_DO_NOT_DETACH: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0010, 'STATUS_FLT_DO_NOT_DETACH')
    STATUS_FLT_INSTANCE_ALTITUDE_COLLISION: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0011, 'STATUS_FLT_INSTANCE_ALTITUDE_COLLISION')
    STATUS_FLT_INSTANCE_NAME_COLLISION: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0012, 'STATUS_FLT_INSTANCE_NAME_COLLISION')
    STATUS_FLT_FILTER_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0013, 'STATUS_FLT_FILTER_NOT_FOUND')
    STATUS_FLT_VOLUME_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0014, 'STATUS_FLT_VOLUME_NOT_FOUND')
    STATUS_FLT_INSTANCE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0015, 'STATUS_FLT_INSTANCE_NOT_FOUND')
    STATUS_FLT_CONTEXT_ALLOCATION_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0016, 'STATUS_FLT_CONTEXT_ALLOCATION_NOT_FOUND')
    STATUS_FLT_INVALID_CONTEXT_REGISTRATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0017, 'STATUS_FLT_INVALID_CONTEXT_REGISTRATION')
    STATUS_FLT_NAME_CACHE_MISS: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0018, 'STATUS_FLT_NAME_CACHE_MISS')
    STATUS_FLT_NO_DEVICE_OBJECT: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0019, 'STATUS_FLT_NO_DEVICE_OBJECT')
    STATUS_FLT_VOLUME_ALREADY_MOUNTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C001A, 'STATUS_FLT_VOLUME_ALREADY_MOUNTED')
    STATUS_FLT_ALREADY_ENLISTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C001B, 'STATUS_FLT_ALREADY_ENLISTED')
    STATUS_FLT_CONTEXT_ALREADY_LINKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C001C, 'STATUS_FLT_CONTEXT_ALREADY_LINKED')
    STATUS_FLT_NO_WAITER_FOR_REPLY: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0020, 'STATUS_FLT_NO_WAITER_FOR_REPLY')
    STATUS_FLT_REGISTRATION_BUSY: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0023, 'STATUS_FLT_REGISTRATION_BUSY')
    STATUS_FLT_WCOS_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01C0024, 'STATUS_FLT_WCOS_NOT_SUPPORTED')
    STATUS_MONITOR_NO_DESCRIPTOR: Final[ThirtyTwoBits] = cached_ntstatus(0xC01D0001, 'STATUS_MONITOR_NO_DESCRIPTOR')
    STATUS_MONITOR_UNKNOWN_DESCRIPTOR_FORMAT: Final[ThirtyTwoBits] = cached_ntstatus(0xC01D0002, 'STATUS_MONITOR_UNKNOWN_DESCRIPTOR_FORMAT')
    STATUS_MONITOR_INVALID_DESCRIPTOR_CHECKSUM: Final[ThirtyTwoBits] = cached_ntstatus(0xC01D0003, 'STATUS_MONITOR_INVALID_DESCRIPTOR_CHECKSUM')
    STATUS_MONITOR_INVALID_STANDARD_TIMING_BLOCK: Final[ThirtyTwoBits] = cached_ntstatus(0xC01D0004, 'STATUS_MONITOR_INVALID_STANDARD_TIMING_BLOCK')
    STATUS_MONITOR_WMI_DATABLOCK_REGISTRATION_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01D0005, 'STATUS_MONITOR_WMI_DATABLOCK_REGISTRATION_FAILED')
    STATUS_MONITOR_INVALID_SERIAL_NUMBER_MONDSC_BLOCK: Final[ThirtyTwoBits] = cached_ntstatus(0xC01D0006, 'STATUS_MONITOR_INVALID_SERIAL_NUMBER_MONDSC_BLOCK')
    STATUS_MONITOR_INVALID_USER_FRIENDLY_MONDSC_BLOCK: Final[ThirtyTwoBits] = cached_ntstatus(0xC01D0007, 'STATUS_MONITOR_INVALID_USER_FRIENDLY_MONDSC_BLOCK')
    STATUS_MONITOR_NO_MORE_DESCRIPTOR_DATA: Final[ThirtyTwoBits] = cached_ntstatus(0xC01D0008, 'STATUS_MONITOR_NO_MORE_DESCRIPTOR_DATA')
    STATUS_MONITOR_INVALID_DETAILED_TIMING_BLOCK: Final[ThirtyTwoBits] = cached_ntstatus(0xC01D0009, 'STATUS_MONITOR_INVALID_DETAILED_TIMING_BLOCK')
    STATUS_MONITOR_INVALID_MANUFACTURE_DATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01D000A, 'STATUS_MONITOR_INVALID_MANUFACTURE_DATE')
    STATUS_GRAPHICS_NOT_EXCLUSIVE_MODE_OWNER: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0000, 'STATUS_GRAPHICS_NOT_EXCLUSIVE_MODE_OWNER')
    STATUS_GRAPHICS_INSUFFICIENT_DMA_BUFFER: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0001, 'STATUS_GRAPHICS_INSUFFICIENT_DMA_BUFFER')
    STATUS_GRAPHICS_INVALID_DISPLAY_ADAPTER: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0002, 'STATUS_GRAPHICS_INVALID_DISPLAY_ADAPTER')
    STATUS_GRAPHICS_ADAPTER_WAS_RESET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0003, 'STATUS_GRAPHICS_ADAPTER_WAS_RESET')
    STATUS_GRAPHICS_INVALID_DRIVER_MODEL: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0004, 'STATUS_GRAPHICS_INVALID_DRIVER_MODEL')
    STATUS_GRAPHICS_PRESENT_MODE_CHANGED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0005, 'STATUS_GRAPHICS_PRESENT_MODE_CHANGED')
    STATUS_GRAPHICS_PRESENT_OCCLUDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0006, 'STATUS_GRAPHICS_PRESENT_OCCLUDED')
    STATUS_GRAPHICS_PRESENT_DENIED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0007, 'STATUS_GRAPHICS_PRESENT_DENIED')
    STATUS_GRAPHICS_CANNOTCOLORCONVERT: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0008, 'STATUS_GRAPHICS_CANNOTCOLORCONVERT')
    STATUS_GRAPHICS_DRIVER_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0009, 'STATUS_GRAPHICS_DRIVER_MISMATCH')
    STATUS_GRAPHICS_PRESENT_REDIRECTION_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E000B, 'STATUS_GRAPHICS_PRESENT_REDIRECTION_DISABLED')
    STATUS_GRAPHICS_PRESENT_UNOCCLUDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E000C, 'STATUS_GRAPHICS_PRESENT_UNOCCLUDED')
    STATUS_GRAPHICS_WINDOWDC_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E000D, 'STATUS_GRAPHICS_WINDOWDC_NOT_AVAILABLE')
    STATUS_GRAPHICS_WINDOWLESS_PRESENT_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E000E, 'STATUS_GRAPHICS_WINDOWLESS_PRESENT_DISABLED')
    STATUS_GRAPHICS_PRESENT_INVALID_WINDOW: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E000F, 'STATUS_GRAPHICS_PRESENT_INVALID_WINDOW')
    STATUS_GRAPHICS_PRESENT_BUFFER_NOT_BOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0010, 'STATUS_GRAPHICS_PRESENT_BUFFER_NOT_BOUND')
    STATUS_GRAPHICS_VAIL_STATE_CHANGED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0011, 'STATUS_GRAPHICS_VAIL_STATE_CHANGED')
    STATUS_GRAPHICS_INDIRECT_DISPLAY_ABANDON_SWAPCHAIN: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0012, 'STATUS_GRAPHICS_INDIRECT_DISPLAY_ABANDON_SWAPCHAIN')
    STATUS_GRAPHICS_INDIRECT_DISPLAY_DEVICE_STOPPED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0013, 'STATUS_GRAPHICS_INDIRECT_DISPLAY_DEVICE_STOPPED')
    STATUS_GRAPHICS_MPO_ALLOCATION_UNPINNED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0018, 'STATUS_GRAPHICS_MPO_ALLOCATION_UNPINNED')
    STATUS_GRAPHICS_NO_VIDEO_MEMORY: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0100, 'STATUS_GRAPHICS_NO_VIDEO_MEMORY')
    STATUS_GRAPHICS_CANT_LOCK_MEMORY: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0101, 'STATUS_GRAPHICS_CANT_LOCK_MEMORY')
    STATUS_GRAPHICS_ALLOCATION_BUSY: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0102, 'STATUS_GRAPHICS_ALLOCATION_BUSY')
    STATUS_GRAPHICS_TOO_MANY_REFERENCES: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0103, 'STATUS_GRAPHICS_TOO_MANY_REFERENCES')
    STATUS_GRAPHICS_TRY_AGAIN_LATER: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0104, 'STATUS_GRAPHICS_TRY_AGAIN_LATER')
    STATUS_GRAPHICS_TRY_AGAIN_NOW: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0105, 'STATUS_GRAPHICS_TRY_AGAIN_NOW')
    STATUS_GRAPHICS_ALLOCATION_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0106, 'STATUS_GRAPHICS_ALLOCATION_INVALID')
    STATUS_GRAPHICS_UNSWIZZLING_APERTURE_UNAVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0107, 'STATUS_GRAPHICS_UNSWIZZLING_APERTURE_UNAVAILABLE')
    STATUS_GRAPHICS_UNSWIZZLING_APERTURE_UNSUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0108, 'STATUS_GRAPHICS_UNSWIZZLING_APERTURE_UNSUPPORTED')
    STATUS_GRAPHICS_CANT_EVICT_PINNED_ALLOCATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0109, 'STATUS_GRAPHICS_CANT_EVICT_PINNED_ALLOCATION')
    STATUS_GRAPHICS_INVALID_ALLOCATION_USAGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0110, 'STATUS_GRAPHICS_INVALID_ALLOCATION_USAGE')
    STATUS_GRAPHICS_CANT_RENDER_LOCKED_ALLOCATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0111, 'STATUS_GRAPHICS_CANT_RENDER_LOCKED_ALLOCATION')
    STATUS_GRAPHICS_ALLOCATION_CLOSED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0112, 'STATUS_GRAPHICS_ALLOCATION_CLOSED')
    STATUS_GRAPHICS_INVALID_ALLOCATION_INSTANCE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0113, 'STATUS_GRAPHICS_INVALID_ALLOCATION_INSTANCE')
    STATUS_GRAPHICS_INVALID_ALLOCATION_HANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0114, 'STATUS_GRAPHICS_INVALID_ALLOCATION_HANDLE')
    STATUS_GRAPHICS_WRONG_ALLOCATION_DEVICE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0115, 'STATUS_GRAPHICS_WRONG_ALLOCATION_DEVICE')
    STATUS_GRAPHICS_ALLOCATION_CONTENT_LOST: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0116, 'STATUS_GRAPHICS_ALLOCATION_CONTENT_LOST')
    STATUS_GRAPHICS_GPU_EXCEPTION_ON_DEVICE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0200, 'STATUS_GRAPHICS_GPU_EXCEPTION_ON_DEVICE')
    STATUS_GRAPHICS_INVALID_VIDPN_TOPOLOGY: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0300, 'STATUS_GRAPHICS_INVALID_VIDPN_TOPOLOGY')
    STATUS_GRAPHICS_VIDPN_TOPOLOGY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0301, 'STATUS_GRAPHICS_VIDPN_TOPOLOGY_NOT_SUPPORTED')
    STATUS_GRAPHICS_VIDPN_TOPOLOGY_CURRENTLY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0302, 'STATUS_GRAPHICS_VIDPN_TOPOLOGY_CURRENTLY_NOT_SUPPORTED')
    STATUS_GRAPHICS_INVALID_VIDPN: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0303, 'STATUS_GRAPHICS_INVALID_VIDPN')
    STATUS_GRAPHICS_INVALID_VIDEO_PRESENT_SOURCE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0304, 'STATUS_GRAPHICS_INVALID_VIDEO_PRESENT_SOURCE')
    STATUS_GRAPHICS_INVALID_VIDEO_PRESENT_TARGET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0305, 'STATUS_GRAPHICS_INVALID_VIDEO_PRESENT_TARGET')
    STATUS_GRAPHICS_VIDPN_MODALITY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0306, 'STATUS_GRAPHICS_VIDPN_MODALITY_NOT_SUPPORTED')
    STATUS_GRAPHICS_INVALID_VIDPN_SOURCEMODESET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0308, 'STATUS_GRAPHICS_INVALID_VIDPN_SOURCEMODESET')
    STATUS_GRAPHICS_INVALID_VIDPN_TARGETMODESET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0309, 'STATUS_GRAPHICS_INVALID_VIDPN_TARGETMODESET')
    STATUS_GRAPHICS_INVALID_FREQUENCY: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E030A, 'STATUS_GRAPHICS_INVALID_FREQUENCY')
    STATUS_GRAPHICS_INVALID_ACTIVE_REGION: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E030B, 'STATUS_GRAPHICS_INVALID_ACTIVE_REGION')
    STATUS_GRAPHICS_INVALID_TOTAL_REGION: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E030C, 'STATUS_GRAPHICS_INVALID_TOTAL_REGION')
    STATUS_GRAPHICS_INVALID_VIDEO_PRESENT_SOURCE_MODE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0310, 'STATUS_GRAPHICS_INVALID_VIDEO_PRESENT_SOURCE_MODE')
    STATUS_GRAPHICS_INVALID_VIDEO_PRESENT_TARGET_MODE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0311, 'STATUS_GRAPHICS_INVALID_VIDEO_PRESENT_TARGET_MODE')
    STATUS_GRAPHICS_PINNED_MODE_MUST_REMAIN_IN_SET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0312, 'STATUS_GRAPHICS_PINNED_MODE_MUST_REMAIN_IN_SET')
    STATUS_GRAPHICS_PATH_ALREADY_IN_TOPOLOGY: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0313, 'STATUS_GRAPHICS_PATH_ALREADY_IN_TOPOLOGY')
    STATUS_GRAPHICS_MODE_ALREADY_IN_MODESET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0314, 'STATUS_GRAPHICS_MODE_ALREADY_IN_MODESET')
    STATUS_GRAPHICS_INVALID_VIDEOPRESENTSOURCESET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0315, 'STATUS_GRAPHICS_INVALID_VIDEOPRESENTSOURCESET')
    STATUS_GRAPHICS_INVALID_VIDEOPRESENTTARGETSET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0316, 'STATUS_GRAPHICS_INVALID_VIDEOPRESENTTARGETSET')
    STATUS_GRAPHICS_SOURCE_ALREADY_IN_SET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0317, 'STATUS_GRAPHICS_SOURCE_ALREADY_IN_SET')
    STATUS_GRAPHICS_TARGET_ALREADY_IN_SET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0318, 'STATUS_GRAPHICS_TARGET_ALREADY_IN_SET')
    STATUS_GRAPHICS_INVALID_VIDPN_PRESENT_PATH: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0319, 'STATUS_GRAPHICS_INVALID_VIDPN_PRESENT_PATH')
    STATUS_GRAPHICS_NO_RECOMMENDED_VIDPN_TOPOLOGY: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E031A, 'STATUS_GRAPHICS_NO_RECOMMENDED_VIDPN_TOPOLOGY')
    STATUS_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGESET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E031B, 'STATUS_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGESET')
    STATUS_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E031C, 'STATUS_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGE')
    STATUS_GRAPHICS_FREQUENCYRANGE_NOT_IN_SET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E031D, 'STATUS_GRAPHICS_FREQUENCYRANGE_NOT_IN_SET')
    STATUS_GRAPHICS_FREQUENCYRANGE_ALREADY_IN_SET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E031F, 'STATUS_GRAPHICS_FREQUENCYRANGE_ALREADY_IN_SET')
    STATUS_GRAPHICS_STALE_MODESET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0320, 'STATUS_GRAPHICS_STALE_MODESET')
    STATUS_GRAPHICS_INVALID_MONITOR_SOURCEMODESET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0321, 'STATUS_GRAPHICS_INVALID_MONITOR_SOURCEMODESET')
    STATUS_GRAPHICS_INVALID_MONITOR_SOURCE_MODE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0322, 'STATUS_GRAPHICS_INVALID_MONITOR_SOURCE_MODE')
    STATUS_GRAPHICS_NO_RECOMMENDED_FUNCTIONAL_VIDPN: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0323, 'STATUS_GRAPHICS_NO_RECOMMENDED_FUNCTIONAL_VIDPN')
    STATUS_GRAPHICS_MODE_ID_MUST_BE_UNIQUE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0324, 'STATUS_GRAPHICS_MODE_ID_MUST_BE_UNIQUE')
    STATUS_GRAPHICS_EMPTY_ADAPTER_MONITOR_MODE_SUPPORT_INTERSECTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0325, 'STATUS_GRAPHICS_EMPTY_ADAPTER_MONITOR_MODE_SUPPORT_INTERSECTION')
    STATUS_GRAPHICS_VIDEO_PRESENT_TARGETS_LESS_THAN_SOURCES: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0326, 'STATUS_GRAPHICS_VIDEO_PRESENT_TARGETS_LESS_THAN_SOURCES')
    STATUS_GRAPHICS_PATH_NOT_IN_TOPOLOGY: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0327, 'STATUS_GRAPHICS_PATH_NOT_IN_TOPOLOGY')
    STATUS_GRAPHICS_ADAPTER_MUST_HAVE_AT_LEAST_ONE_SOURCE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0328, 'STATUS_GRAPHICS_ADAPTER_MUST_HAVE_AT_LEAST_ONE_SOURCE')
    STATUS_GRAPHICS_ADAPTER_MUST_HAVE_AT_LEAST_ONE_TARGET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0329, 'STATUS_GRAPHICS_ADAPTER_MUST_HAVE_AT_LEAST_ONE_TARGET')
    STATUS_GRAPHICS_INVALID_MONITORDESCRIPTORSET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E032A, 'STATUS_GRAPHICS_INVALID_MONITORDESCRIPTORSET')
    STATUS_GRAPHICS_INVALID_MONITORDESCRIPTOR: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E032B, 'STATUS_GRAPHICS_INVALID_MONITORDESCRIPTOR')
    STATUS_GRAPHICS_MONITORDESCRIPTOR_NOT_IN_SET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E032C, 'STATUS_GRAPHICS_MONITORDESCRIPTOR_NOT_IN_SET')
    STATUS_GRAPHICS_MONITORDESCRIPTOR_ALREADY_IN_SET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E032D, 'STATUS_GRAPHICS_MONITORDESCRIPTOR_ALREADY_IN_SET')
    STATUS_GRAPHICS_MONITORDESCRIPTOR_ID_MUST_BE_UNIQUE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E032E, 'STATUS_GRAPHICS_MONITORDESCRIPTOR_ID_MUST_BE_UNIQUE')
    STATUS_GRAPHICS_INVALID_VIDPN_TARGET_SUBSET_TYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E032F, 'STATUS_GRAPHICS_INVALID_VIDPN_TARGET_SUBSET_TYPE')
    STATUS_GRAPHICS_RESOURCES_NOT_RELATED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0330, 'STATUS_GRAPHICS_RESOURCES_NOT_RELATED')
    STATUS_GRAPHICS_SOURCE_ID_MUST_BE_UNIQUE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0331, 'STATUS_GRAPHICS_SOURCE_ID_MUST_BE_UNIQUE')
    STATUS_GRAPHICS_TARGET_ID_MUST_BE_UNIQUE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0332, 'STATUS_GRAPHICS_TARGET_ID_MUST_BE_UNIQUE')
    STATUS_GRAPHICS_NO_AVAILABLE_VIDPN_TARGET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0333, 'STATUS_GRAPHICS_NO_AVAILABLE_VIDPN_TARGET')
    STATUS_GRAPHICS_MONITOR_COULD_NOT_BE_ASSOCIATED_WITH_ADAPTER: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0334, 'STATUS_GRAPHICS_MONITOR_COULD_NOT_BE_ASSOCIATED_WITH_ADAPTER')
    STATUS_GRAPHICS_NO_VIDPNMGR: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0335, 'STATUS_GRAPHICS_NO_VIDPNMGR')
    STATUS_GRAPHICS_NO_ACTIVE_VIDPN: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0336, 'STATUS_GRAPHICS_NO_ACTIVE_VIDPN')
    STATUS_GRAPHICS_STALE_VIDPN_TOPOLOGY: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0337, 'STATUS_GRAPHICS_STALE_VIDPN_TOPOLOGY')
    STATUS_GRAPHICS_MONITOR_NOT_CONNECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0338, 'STATUS_GRAPHICS_MONITOR_NOT_CONNECTED')
    STATUS_GRAPHICS_SOURCE_NOT_IN_TOPOLOGY: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0339, 'STATUS_GRAPHICS_SOURCE_NOT_IN_TOPOLOGY')
    STATUS_GRAPHICS_INVALID_PRIMARYSURFACE_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E033A, 'STATUS_GRAPHICS_INVALID_PRIMARYSURFACE_SIZE')
    STATUS_GRAPHICS_INVALID_VISIBLEREGION_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E033B, 'STATUS_GRAPHICS_INVALID_VISIBLEREGION_SIZE')
    STATUS_GRAPHICS_INVALID_STRIDE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E033C, 'STATUS_GRAPHICS_INVALID_STRIDE')
    STATUS_GRAPHICS_INVALID_PIXELFORMAT: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E033D, 'STATUS_GRAPHICS_INVALID_PIXELFORMAT')
    STATUS_GRAPHICS_INVALID_COLORBASIS: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E033E, 'STATUS_GRAPHICS_INVALID_COLORBASIS')
    STATUS_GRAPHICS_INVALID_PIXELVALUEACCESSMODE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E033F, 'STATUS_GRAPHICS_INVALID_PIXELVALUEACCESSMODE')
    STATUS_GRAPHICS_TARGET_NOT_IN_TOPOLOGY: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0340, 'STATUS_GRAPHICS_TARGET_NOT_IN_TOPOLOGY')
    STATUS_GRAPHICS_NO_DISPLAY_MODE_MANAGEMENT_SUPPORT: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0341, 'STATUS_GRAPHICS_NO_DISPLAY_MODE_MANAGEMENT_SUPPORT')
    STATUS_GRAPHICS_VIDPN_SOURCE_IN_USE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0342, 'STATUS_GRAPHICS_VIDPN_SOURCE_IN_USE')
    STATUS_GRAPHICS_CANT_ACCESS_ACTIVE_VIDPN: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0343, 'STATUS_GRAPHICS_CANT_ACCESS_ACTIVE_VIDPN')
    STATUS_GRAPHICS_INVALID_PATH_IMPORTANCE_ORDINAL: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0344, 'STATUS_GRAPHICS_INVALID_PATH_IMPORTANCE_ORDINAL')
    STATUS_GRAPHICS_INVALID_PATH_CONTENT_GEOMETRY_TRANSFORMATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0345, 'STATUS_GRAPHICS_INVALID_PATH_CONTENT_GEOMETRY_TRANSFORMATION')
    STATUS_GRAPHICS_PATH_CONTENT_GEOMETRY_TRANSFORMATION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0346, 'STATUS_GRAPHICS_PATH_CONTENT_GEOMETRY_TRANSFORMATION_NOT_SUPPORTED')
    STATUS_GRAPHICS_INVALID_GAMMA_RAMP: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0347, 'STATUS_GRAPHICS_INVALID_GAMMA_RAMP')
    STATUS_GRAPHICS_GAMMA_RAMP_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0348, 'STATUS_GRAPHICS_GAMMA_RAMP_NOT_SUPPORTED')
    STATUS_GRAPHICS_MULTISAMPLING_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0349, 'STATUS_GRAPHICS_MULTISAMPLING_NOT_SUPPORTED')
    STATUS_GRAPHICS_MODE_NOT_IN_MODESET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E034A, 'STATUS_GRAPHICS_MODE_NOT_IN_MODESET')
    STATUS_GRAPHICS_INVALID_VIDPN_TOPOLOGY_RECOMMENDATION_REASON: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E034D, 'STATUS_GRAPHICS_INVALID_VIDPN_TOPOLOGY_RECOMMENDATION_REASON')
    STATUS_GRAPHICS_INVALID_PATH_CONTENT_TYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E034E, 'STATUS_GRAPHICS_INVALID_PATH_CONTENT_TYPE')
    STATUS_GRAPHICS_INVALID_COPYPROTECTION_TYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E034F, 'STATUS_GRAPHICS_INVALID_COPYPROTECTION_TYPE')
    STATUS_GRAPHICS_UNASSIGNED_MODESET_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0350, 'STATUS_GRAPHICS_UNASSIGNED_MODESET_ALREADY_EXISTS')
    STATUS_GRAPHICS_INVALID_SCANLINE_ORDERING: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0352, 'STATUS_GRAPHICS_INVALID_SCANLINE_ORDERING')
    STATUS_GRAPHICS_TOPOLOGY_CHANGES_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0353, 'STATUS_GRAPHICS_TOPOLOGY_CHANGES_NOT_ALLOWED')
    STATUS_GRAPHICS_NO_AVAILABLE_IMPORTANCE_ORDINALS: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0354, 'STATUS_GRAPHICS_NO_AVAILABLE_IMPORTANCE_ORDINALS')
    STATUS_GRAPHICS_INCOMPATIBLE_PRIVATE_FORMAT: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0355, 'STATUS_GRAPHICS_INCOMPATIBLE_PRIVATE_FORMAT')
    STATUS_GRAPHICS_INVALID_MODE_PRUNING_ALGORITHM: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0356, 'STATUS_GRAPHICS_INVALID_MODE_PRUNING_ALGORITHM')
    STATUS_GRAPHICS_INVALID_MONITOR_CAPABILITY_ORIGIN: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0357, 'STATUS_GRAPHICS_INVALID_MONITOR_CAPABILITY_ORIGIN')
    STATUS_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGE_CONSTRAINT: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0358, 'STATUS_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGE_CONSTRAINT')
    STATUS_GRAPHICS_MAX_NUM_PATHS_REACHED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0359, 'STATUS_GRAPHICS_MAX_NUM_PATHS_REACHED')
    STATUS_GRAPHICS_CANCEL_VIDPN_TOPOLOGY_AUGMENTATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E035A, 'STATUS_GRAPHICS_CANCEL_VIDPN_TOPOLOGY_AUGMENTATION')
    STATUS_GRAPHICS_INVALID_CLIENT_TYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E035B, 'STATUS_GRAPHICS_INVALID_CLIENT_TYPE')
    STATUS_GRAPHICS_CLIENTVIDPN_NOT_SET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E035C, 'STATUS_GRAPHICS_CLIENTVIDPN_NOT_SET')
    STATUS_GRAPHICS_SPECIFIED_CHILD_ALREADY_CONNECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0400, 'STATUS_GRAPHICS_SPECIFIED_CHILD_ALREADY_CONNECTED')
    STATUS_GRAPHICS_CHILD_DESCRIPTOR_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0401, 'STATUS_GRAPHICS_CHILD_DESCRIPTOR_NOT_SUPPORTED')
    STATUS_GRAPHICS_NOT_A_LINKED_ADAPTER: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0430, 'STATUS_GRAPHICS_NOT_A_LINKED_ADAPTER')
    STATUS_GRAPHICS_LEADLINK_NOT_ENUMERATED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0431, 'STATUS_GRAPHICS_LEADLINK_NOT_ENUMERATED')
    STATUS_GRAPHICS_CHAINLINKS_NOT_ENUMERATED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0432, 'STATUS_GRAPHICS_CHAINLINKS_NOT_ENUMERATED')
    STATUS_GRAPHICS_ADAPTER_CHAIN_NOT_READY: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0433, 'STATUS_GRAPHICS_ADAPTER_CHAIN_NOT_READY')
    STATUS_GRAPHICS_CHAINLINKS_NOT_STARTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0434, 'STATUS_GRAPHICS_CHAINLINKS_NOT_STARTED')
    STATUS_GRAPHICS_CHAINLINKS_NOT_POWERED_ON: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0435, 'STATUS_GRAPHICS_CHAINLINKS_NOT_POWERED_ON')
    STATUS_GRAPHICS_INCONSISTENT_DEVICE_LINK_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0436, 'STATUS_GRAPHICS_INCONSISTENT_DEVICE_LINK_STATE')
    STATUS_GRAPHICS_NOT_POST_DEVICE_DRIVER: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0438, 'STATUS_GRAPHICS_NOT_POST_DEVICE_DRIVER')
    STATUS_GRAPHICS_ADAPTER_ACCESS_NOT_EXCLUDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E043B, 'STATUS_GRAPHICS_ADAPTER_ACCESS_NOT_EXCLUDED')
    STATUS_GRAPHICS_OPM_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0500, 'STATUS_GRAPHICS_OPM_NOT_SUPPORTED')
    STATUS_GRAPHICS_COPP_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0501, 'STATUS_GRAPHICS_COPP_NOT_SUPPORTED')
    STATUS_GRAPHICS_UAB_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0502, 'STATUS_GRAPHICS_UAB_NOT_SUPPORTED')
    STATUS_GRAPHICS_OPM_INVALID_ENCRYPTED_PARAMETERS: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0503, 'STATUS_GRAPHICS_OPM_INVALID_ENCRYPTED_PARAMETERS')
    STATUS_GRAPHICS_OPM_NO_PROTECTED_OUTPUTS_EXIST: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0505, 'STATUS_GRAPHICS_OPM_NO_PROTECTED_OUTPUTS_EXIST')
    STATUS_GRAPHICS_OPM_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E050B, 'STATUS_GRAPHICS_OPM_INTERNAL_ERROR')
    STATUS_GRAPHICS_OPM_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E050C, 'STATUS_GRAPHICS_OPM_INVALID_HANDLE')
    STATUS_GRAPHICS_PVP_INVALID_CERTIFICATE_LENGTH: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E050E, 'STATUS_GRAPHICS_PVP_INVALID_CERTIFICATE_LENGTH')
    STATUS_GRAPHICS_OPM_SPANNING_MODE_ENABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E050F, 'STATUS_GRAPHICS_OPM_SPANNING_MODE_ENABLED')
    STATUS_GRAPHICS_OPM_THEATER_MODE_ENABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0510, 'STATUS_GRAPHICS_OPM_THEATER_MODE_ENABLED')
    STATUS_GRAPHICS_PVP_HFS_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0511, 'STATUS_GRAPHICS_PVP_HFS_FAILED')
    STATUS_GRAPHICS_OPM_INVALID_SRM: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0512, 'STATUS_GRAPHICS_OPM_INVALID_SRM')
    STATUS_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_HDCP: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0513, 'STATUS_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_HDCP')
    STATUS_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_ACP: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0514, 'STATUS_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_ACP')
    STATUS_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_CGMSA: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0515, 'STATUS_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_CGMSA')
    STATUS_GRAPHICS_OPM_HDCP_SRM_NEVER_SET: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0516, 'STATUS_GRAPHICS_OPM_HDCP_SRM_NEVER_SET')
    STATUS_GRAPHICS_OPM_RESOLUTION_TOO_HIGH: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0517, 'STATUS_GRAPHICS_OPM_RESOLUTION_TOO_HIGH')
    STATUS_GRAPHICS_OPM_ALL_HDCP_HARDWARE_ALREADY_IN_USE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0518, 'STATUS_GRAPHICS_OPM_ALL_HDCP_HARDWARE_ALREADY_IN_USE')
    STATUS_GRAPHICS_OPM_PROTECTED_OUTPUT_NO_LONGER_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E051A, 'STATUS_GRAPHICS_OPM_PROTECTED_OUTPUT_NO_LONGER_EXISTS')
    STATUS_GRAPHICS_OPM_PROTECTED_OUTPUT_DOES_NOT_HAVE_COPP_SEMANTICS: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E051C, 'STATUS_GRAPHICS_OPM_PROTECTED_OUTPUT_DOES_NOT_HAVE_COPP_SEMANTICS')
    STATUS_GRAPHICS_OPM_INVALID_INFORMATION_REQUEST: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E051D, 'STATUS_GRAPHICS_OPM_INVALID_INFORMATION_REQUEST')
    STATUS_GRAPHICS_OPM_DRIVER_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E051E, 'STATUS_GRAPHICS_OPM_DRIVER_INTERNAL_ERROR')
    STATUS_GRAPHICS_OPM_PROTECTED_OUTPUT_DOES_NOT_HAVE_OPM_SEMANTICS: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E051F, 'STATUS_GRAPHICS_OPM_PROTECTED_OUTPUT_DOES_NOT_HAVE_OPM_SEMANTICS')
    STATUS_GRAPHICS_OPM_SIGNALING_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0520, 'STATUS_GRAPHICS_OPM_SIGNALING_NOT_SUPPORTED')
    STATUS_GRAPHICS_OPM_INVALID_CONFIGURATION_REQUEST: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0521, 'STATUS_GRAPHICS_OPM_INVALID_CONFIGURATION_REQUEST')
    STATUS_GRAPHICS_I2C_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0580, 'STATUS_GRAPHICS_I2C_NOT_SUPPORTED')
    STATUS_GRAPHICS_I2C_DEVICE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0581, 'STATUS_GRAPHICS_I2C_DEVICE_DOES_NOT_EXIST')
    STATUS_GRAPHICS_I2C_ERROR_TRANSMITTING_DATA: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0582, 'STATUS_GRAPHICS_I2C_ERROR_TRANSMITTING_DATA')
    STATUS_GRAPHICS_I2C_ERROR_RECEIVING_DATA: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0583, 'STATUS_GRAPHICS_I2C_ERROR_RECEIVING_DATA')
    STATUS_GRAPHICS_DDCCI_VCP_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0584, 'STATUS_GRAPHICS_DDCCI_VCP_NOT_SUPPORTED')
    STATUS_GRAPHICS_DDCCI_INVALID_DATA: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0585, 'STATUS_GRAPHICS_DDCCI_INVALID_DATA')
    STATUS_GRAPHICS_DDCCI_MONITOR_RETURNED_INVALID_TIMING_STATUS_BYTE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0586, 'STATUS_GRAPHICS_DDCCI_MONITOR_RETURNED_INVALID_TIMING_STATUS_BYTE')
    STATUS_GRAPHICS_DDCCI_INVALID_CAPABILITIES_STRING: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0587, 'STATUS_GRAPHICS_DDCCI_INVALID_CAPABILITIES_STRING')
    STATUS_GRAPHICS_MCA_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0588, 'STATUS_GRAPHICS_MCA_INTERNAL_ERROR')
    STATUS_GRAPHICS_DDCCI_INVALID_MESSAGE_COMMAND: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E0589, 'STATUS_GRAPHICS_DDCCI_INVALID_MESSAGE_COMMAND')
    STATUS_GRAPHICS_DDCCI_INVALID_MESSAGE_LENGTH: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E058A, 'STATUS_GRAPHICS_DDCCI_INVALID_MESSAGE_LENGTH')
    STATUS_GRAPHICS_DDCCI_INVALID_MESSAGE_CHECKSUM: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E058B, 'STATUS_GRAPHICS_DDCCI_INVALID_MESSAGE_CHECKSUM')
    STATUS_GRAPHICS_INVALID_PHYSICAL_MONITOR_HANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E058C, 'STATUS_GRAPHICS_INVALID_PHYSICAL_MONITOR_HANDLE')
    STATUS_GRAPHICS_MONITOR_NO_LONGER_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E058D, 'STATUS_GRAPHICS_MONITOR_NO_LONGER_EXISTS')
    STATUS_GRAPHICS_ONLY_CONSOLE_SESSION_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E05E0, 'STATUS_GRAPHICS_ONLY_CONSOLE_SESSION_SUPPORTED')
    STATUS_GRAPHICS_NO_DISPLAY_DEVICE_CORRESPONDS_TO_NAME: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E05E1, 'STATUS_GRAPHICS_NO_DISPLAY_DEVICE_CORRESPONDS_TO_NAME')
    STATUS_GRAPHICS_DISPLAY_DEVICE_NOT_ATTACHED_TO_DESKTOP: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E05E2, 'STATUS_GRAPHICS_DISPLAY_DEVICE_NOT_ATTACHED_TO_DESKTOP')
    STATUS_GRAPHICS_MIRRORING_DEVICES_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E05E3, 'STATUS_GRAPHICS_MIRRORING_DEVICES_NOT_SUPPORTED')
    STATUS_GRAPHICS_INVALID_POINTER: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E05E4, 'STATUS_GRAPHICS_INVALID_POINTER')
    STATUS_GRAPHICS_NO_MONITORS_CORRESPOND_TO_DISPLAY_DEVICE: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E05E5, 'STATUS_GRAPHICS_NO_MONITORS_CORRESPOND_TO_DISPLAY_DEVICE')
    STATUS_GRAPHICS_PARAMETER_ARRAY_TOO_SMALL: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E05E6, 'STATUS_GRAPHICS_PARAMETER_ARRAY_TOO_SMALL')
    STATUS_GRAPHICS_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E05E7, 'STATUS_GRAPHICS_INTERNAL_ERROR')
    STATUS_GRAPHICS_SESSION_TYPE_CHANGE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC01E05E8, 'STATUS_GRAPHICS_SESSION_TYPE_CHANGE_IN_PROGRESS')
    STATUS_FVE_LOCKED_VOLUME: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210000, 'STATUS_FVE_LOCKED_VOLUME')
    STATUS_FVE_NOT_ENCRYPTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210001, 'STATUS_FVE_NOT_ENCRYPTED')
    STATUS_FVE_BAD_INFORMATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210002, 'STATUS_FVE_BAD_INFORMATION')
    STATUS_FVE_TOO_SMALL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210003, 'STATUS_FVE_TOO_SMALL')
    STATUS_FVE_FAILED_WRONG_FS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210004, 'STATUS_FVE_FAILED_WRONG_FS')
    STATUS_FVE_BAD_PARTITION_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210005, 'STATUS_FVE_BAD_PARTITION_SIZE')
    STATUS_FVE_FS_NOT_EXTENDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210006, 'STATUS_FVE_FS_NOT_EXTENDED')
    STATUS_FVE_FS_MOUNTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210007, 'STATUS_FVE_FS_MOUNTED')
    STATUS_FVE_NO_LICENSE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210008, 'STATUS_FVE_NO_LICENSE')
    STATUS_FVE_ACTION_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210009, 'STATUS_FVE_ACTION_NOT_ALLOWED')
    STATUS_FVE_BAD_DATA: Final[ThirtyTwoBits] = cached_ntstatus(0xC021000A, 'STATUS_FVE_BAD_DATA')
    STATUS_FVE_VOLUME_NOT_BOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC021000B, 'STATUS_FVE_VOLUME_NOT_BOUND')
    STATUS_FVE_NOT_DATA_VOLUME: Final[ThirtyTwoBits] = cached_ntstatus(0xC021000C, 'STATUS_FVE_NOT_DATA_VOLUME')
    STATUS_FVE_CONV_READ_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC021000D, 'STATUS_FVE_CONV_READ_ERROR')
    STATUS_FVE_CONV_WRITE_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC021000E, 'STATUS_FVE_CONV_WRITE_ERROR')
    STATUS_FVE_OVERLAPPED_UPDATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC021000F, 'STATUS_FVE_OVERLAPPED_UPDATE')
    STATUS_FVE_FAILED_SECTOR_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210010, 'STATUS_FVE_FAILED_SECTOR_SIZE')
    STATUS_FVE_FAILED_AUTHENTICATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210011, 'STATUS_FVE_FAILED_AUTHENTICATION')
    STATUS_FVE_NOT_OS_VOLUME: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210012, 'STATUS_FVE_NOT_OS_VOLUME')
    STATUS_FVE_KEYFILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210013, 'STATUS_FVE_KEYFILE_NOT_FOUND')
    STATUS_FVE_KEYFILE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210014, 'STATUS_FVE_KEYFILE_INVALID')
    STATUS_FVE_KEYFILE_NO_VMK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210015, 'STATUS_FVE_KEYFILE_NO_VMK')
    STATUS_FVE_TPM_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210016, 'STATUS_FVE_TPM_DISABLED')
    STATUS_FVE_TPM_SRK_AUTH_NOT_ZERO: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210017, 'STATUS_FVE_TPM_SRK_AUTH_NOT_ZERO')
    STATUS_FVE_TPM_INVALID_PCR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210018, 'STATUS_FVE_TPM_INVALID_PCR')
    STATUS_FVE_TPM_NO_VMK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210019, 'STATUS_FVE_TPM_NO_VMK')
    STATUS_FVE_PIN_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC021001A, 'STATUS_FVE_PIN_INVALID')
    STATUS_FVE_AUTH_INVALID_APPLICATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC021001B, 'STATUS_FVE_AUTH_INVALID_APPLICATION')
    STATUS_FVE_AUTH_INVALID_CONFIG: Final[ThirtyTwoBits] = cached_ntstatus(0xC021001C, 'STATUS_FVE_AUTH_INVALID_CONFIG')
    STATUS_FVE_DEBUGGER_ENABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC021001D, 'STATUS_FVE_DEBUGGER_ENABLED')
    STATUS_FVE_DRY_RUN_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC021001E, 'STATUS_FVE_DRY_RUN_FAILED')
    STATUS_FVE_BAD_METADATA_POINTER: Final[ThirtyTwoBits] = cached_ntstatus(0xC021001F, 'STATUS_FVE_BAD_METADATA_POINTER')
    STATUS_FVE_OLD_METADATA_COPY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210020, 'STATUS_FVE_OLD_METADATA_COPY')
    STATUS_FVE_REBOOT_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210021, 'STATUS_FVE_REBOOT_REQUIRED')
    STATUS_FVE_RAW_ACCESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210022, 'STATUS_FVE_RAW_ACCESS')
    STATUS_FVE_RAW_BLOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210023, 'STATUS_FVE_RAW_BLOCKED')
    STATUS_FVE_NO_AUTOUNLOCK_MASTER_KEY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210024, 'STATUS_FVE_NO_AUTOUNLOCK_MASTER_KEY')
    STATUS_FVE_MOR_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210025, 'STATUS_FVE_MOR_FAILED')
    STATUS_FVE_NO_FEATURE_LICENSE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210026, 'STATUS_FVE_NO_FEATURE_LICENSE')
    STATUS_FVE_POLICY_USER_DISABLE_RDV_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210027, 'STATUS_FVE_POLICY_USER_DISABLE_RDV_NOT_ALLOWED')
    STATUS_FVE_CONV_RECOVERY_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210028, 'STATUS_FVE_CONV_RECOVERY_FAILED')
    STATUS_FVE_VIRTUALIZED_SPACE_TOO_BIG: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210029, 'STATUS_FVE_VIRTUALIZED_SPACE_TOO_BIG')
    STATUS_FVE_INVALID_DATUM_TYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC021002A, 'STATUS_FVE_INVALID_DATUM_TYPE')
    STATUS_FVE_VOLUME_TOO_SMALL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210030, 'STATUS_FVE_VOLUME_TOO_SMALL')
    STATUS_FVE_ENH_PIN_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210031, 'STATUS_FVE_ENH_PIN_INVALID')
    STATUS_FVE_FULL_ENCRYPTION_NOT_ALLOWED_ON_TP_STORAGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210032, 'STATUS_FVE_FULL_ENCRYPTION_NOT_ALLOWED_ON_TP_STORAGE')
    STATUS_FVE_WIPE_NOT_ALLOWED_ON_TP_STORAGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210033, 'STATUS_FVE_WIPE_NOT_ALLOWED_ON_TP_STORAGE')
    STATUS_FVE_NOT_ALLOWED_ON_CSV_STACK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210034, 'STATUS_FVE_NOT_ALLOWED_ON_CSV_STACK')
    STATUS_FVE_NOT_ALLOWED_ON_CLUSTER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210035, 'STATUS_FVE_NOT_ALLOWED_ON_CLUSTER')
    STATUS_FVE_NOT_ALLOWED_TO_UPGRADE_WHILE_CONVERTING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210036, 'STATUS_FVE_NOT_ALLOWED_TO_UPGRADE_WHILE_CONVERTING')
    STATUS_FVE_WIPE_CANCEL_NOT_APPLICABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210037, 'STATUS_FVE_WIPE_CANCEL_NOT_APPLICABLE')
    STATUS_FVE_EDRIVE_DRY_RUN_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210038, 'STATUS_FVE_EDRIVE_DRY_RUN_FAILED')
    STATUS_FVE_SECUREBOOT_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210039, 'STATUS_FVE_SECUREBOOT_DISABLED')
    STATUS_FVE_SECUREBOOT_CONFIG_CHANGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC021003A, 'STATUS_FVE_SECUREBOOT_CONFIG_CHANGE')
    STATUS_FVE_DEVICE_LOCKEDOUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC021003B, 'STATUS_FVE_DEVICE_LOCKEDOUT')
    STATUS_FVE_VOLUME_EXTEND_PREVENTS_EOW_DECRYPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC021003C, 'STATUS_FVE_VOLUME_EXTEND_PREVENTS_EOW_DECRYPT')
    STATUS_FVE_NOT_DE_VOLUME: Final[ThirtyTwoBits] = cached_ntstatus(0xC021003D, 'STATUS_FVE_NOT_DE_VOLUME')
    STATUS_FVE_PROTECTION_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC021003E, 'STATUS_FVE_PROTECTION_DISABLED')
    STATUS_FVE_PROTECTION_CANNOT_BE_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC021003F, 'STATUS_FVE_PROTECTION_CANNOT_BE_DISABLED')
    STATUS_FVE_OSV_KSR_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210040, 'STATUS_FVE_OSV_KSR_NOT_ALLOWED')
    STATUS_FVE_EDRIVE_BAND_ENUMERATION_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210041, 'STATUS_FVE_EDRIVE_BAND_ENUMERATION_FAILED')
    STATUS_FVE_POLICY_ON_RDV_EXCLUSION_LIST: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210042, 'STATUS_FVE_POLICY_ON_RDV_EXCLUSION_LIST')
    STATUS_FVE_DATASET_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210043, 'STATUS_FVE_DATASET_FULL')
    STATUS_FVE_METADATA_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0210044, 'STATUS_FVE_METADATA_FULL')
    STATUS_FWP_CALLOUT_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220001, 'STATUS_FWP_CALLOUT_NOT_FOUND')
    STATUS_FWP_CONDITION_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220002, 'STATUS_FWP_CONDITION_NOT_FOUND')
    STATUS_FWP_FILTER_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220003, 'STATUS_FWP_FILTER_NOT_FOUND')
    STATUS_FWP_LAYER_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220004, 'STATUS_FWP_LAYER_NOT_FOUND')
    STATUS_FWP_PROVIDER_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220005, 'STATUS_FWP_PROVIDER_NOT_FOUND')
    STATUS_FWP_PROVIDER_CONTEXT_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220006, 'STATUS_FWP_PROVIDER_CONTEXT_NOT_FOUND')
    STATUS_FWP_SUBLAYER_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220007, 'STATUS_FWP_SUBLAYER_NOT_FOUND')
    STATUS_FWP_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220008, 'STATUS_FWP_NOT_FOUND')
    STATUS_FWP_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220009, 'STATUS_FWP_ALREADY_EXISTS')
    STATUS_FWP_IN_USE: Final[ThirtyTwoBits] = cached_ntstatus(0xC022000A, 'STATUS_FWP_IN_USE')
    STATUS_FWP_DYNAMIC_SESSION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC022000B, 'STATUS_FWP_DYNAMIC_SESSION_IN_PROGRESS')
    STATUS_FWP_WRONG_SESSION: Final[ThirtyTwoBits] = cached_ntstatus(0xC022000C, 'STATUS_FWP_WRONG_SESSION')
    STATUS_FWP_NO_TXN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC022000D, 'STATUS_FWP_NO_TXN_IN_PROGRESS')
    STATUS_FWP_TXN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC022000E, 'STATUS_FWP_TXN_IN_PROGRESS')
    STATUS_FWP_TXN_ABORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC022000F, 'STATUS_FWP_TXN_ABORTED')
    STATUS_FWP_SESSION_ABORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220010, 'STATUS_FWP_SESSION_ABORTED')
    STATUS_FWP_INCOMPATIBLE_TXN: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220011, 'STATUS_FWP_INCOMPATIBLE_TXN')
    STATUS_FWP_TIMEOUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220012, 'STATUS_FWP_TIMEOUT')
    STATUS_FWP_NET_EVENTS_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220013, 'STATUS_FWP_NET_EVENTS_DISABLED')
    STATUS_FWP_INCOMPATIBLE_LAYER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220014, 'STATUS_FWP_INCOMPATIBLE_LAYER')
    STATUS_FWP_KM_CLIENTS_ONLY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220015, 'STATUS_FWP_KM_CLIENTS_ONLY')
    STATUS_FWP_LIFETIME_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220016, 'STATUS_FWP_LIFETIME_MISMATCH')
    STATUS_FWP_BUILTIN_OBJECT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220017, 'STATUS_FWP_BUILTIN_OBJECT')
    STATUS_FWP_TOO_MANY_CALLOUTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220018, 'STATUS_FWP_TOO_MANY_CALLOUTS')
    STATUS_FWP_NOTIFICATION_DROPPED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220019, 'STATUS_FWP_NOTIFICATION_DROPPED')
    STATUS_FWP_TRAFFIC_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC022001A, 'STATUS_FWP_TRAFFIC_MISMATCH')
    STATUS_FWP_INCOMPATIBLE_SA_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC022001B, 'STATUS_FWP_INCOMPATIBLE_SA_STATE')
    STATUS_FWP_NULL_POINTER: Final[ThirtyTwoBits] = cached_ntstatus(0xC022001C, 'STATUS_FWP_NULL_POINTER')
    STATUS_FWP_INVALID_ENUMERATOR: Final[ThirtyTwoBits] = cached_ntstatus(0xC022001D, 'STATUS_FWP_INVALID_ENUMERATOR')
    STATUS_FWP_INVALID_FLAGS: Final[ThirtyTwoBits] = cached_ntstatus(0xC022001E, 'STATUS_FWP_INVALID_FLAGS')
    STATUS_FWP_INVALID_NET_MASK: Final[ThirtyTwoBits] = cached_ntstatus(0xC022001F, 'STATUS_FWP_INVALID_NET_MASK')
    STATUS_FWP_INVALID_RANGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220020, 'STATUS_FWP_INVALID_RANGE')
    STATUS_FWP_INVALID_INTERVAL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220021, 'STATUS_FWP_INVALID_INTERVAL')
    STATUS_FWP_ZERO_LENGTH_ARRAY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220022, 'STATUS_FWP_ZERO_LENGTH_ARRAY')
    STATUS_FWP_NULL_DISPLAY_NAME: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220023, 'STATUS_FWP_NULL_DISPLAY_NAME')
    STATUS_FWP_INVALID_ACTION_TYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220024, 'STATUS_FWP_INVALID_ACTION_TYPE')
    STATUS_FWP_INVALID_WEIGHT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220025, 'STATUS_FWP_INVALID_WEIGHT')
    STATUS_FWP_MATCH_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220026, 'STATUS_FWP_MATCH_TYPE_MISMATCH')
    STATUS_FWP_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220027, 'STATUS_FWP_TYPE_MISMATCH')
    STATUS_FWP_OUT_OF_BOUNDS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220028, 'STATUS_FWP_OUT_OF_BOUNDS')
    STATUS_FWP_RESERVED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220029, 'STATUS_FWP_RESERVED')
    STATUS_FWP_DUPLICATE_CONDITION: Final[ThirtyTwoBits] = cached_ntstatus(0xC022002A, 'STATUS_FWP_DUPLICATE_CONDITION')
    STATUS_FWP_DUPLICATE_KEYMOD: Final[ThirtyTwoBits] = cached_ntstatus(0xC022002B, 'STATUS_FWP_DUPLICATE_KEYMOD')
    STATUS_FWP_ACTION_INCOMPATIBLE_WITH_LAYER: Final[ThirtyTwoBits] = cached_ntstatus(0xC022002C, 'STATUS_FWP_ACTION_INCOMPATIBLE_WITH_LAYER')
    STATUS_FWP_ACTION_INCOMPATIBLE_WITH_SUBLAYER: Final[ThirtyTwoBits] = cached_ntstatus(0xC022002D, 'STATUS_FWP_ACTION_INCOMPATIBLE_WITH_SUBLAYER')
    STATUS_FWP_CONTEXT_INCOMPATIBLE_WITH_LAYER: Final[ThirtyTwoBits] = cached_ntstatus(0xC022002E, 'STATUS_FWP_CONTEXT_INCOMPATIBLE_WITH_LAYER')
    STATUS_FWP_CONTEXT_INCOMPATIBLE_WITH_CALLOUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC022002F, 'STATUS_FWP_CONTEXT_INCOMPATIBLE_WITH_CALLOUT')
    STATUS_FWP_INCOMPATIBLE_AUTH_METHOD: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220030, 'STATUS_FWP_INCOMPATIBLE_AUTH_METHOD')
    STATUS_FWP_INCOMPATIBLE_DH_GROUP: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220031, 'STATUS_FWP_INCOMPATIBLE_DH_GROUP')
    STATUS_FWP_EM_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220032, 'STATUS_FWP_EM_NOT_SUPPORTED')
    STATUS_FWP_NEVER_MATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220033, 'STATUS_FWP_NEVER_MATCH')
    STATUS_FWP_PROVIDER_CONTEXT_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220034, 'STATUS_FWP_PROVIDER_CONTEXT_MISMATCH')
    STATUS_FWP_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220035, 'STATUS_FWP_INVALID_PARAMETER')
    STATUS_FWP_TOO_MANY_SUBLAYERS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220036, 'STATUS_FWP_TOO_MANY_SUBLAYERS')
    STATUS_FWP_CALLOUT_NOTIFICATION_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220037, 'STATUS_FWP_CALLOUT_NOTIFICATION_FAILED')
    STATUS_FWP_INVALID_AUTH_TRANSFORM: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220038, 'STATUS_FWP_INVALID_AUTH_TRANSFORM')
    STATUS_FWP_INVALID_CIPHER_TRANSFORM: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220039, 'STATUS_FWP_INVALID_CIPHER_TRANSFORM')
    STATUS_FWP_INCOMPATIBLE_CIPHER_TRANSFORM: Final[ThirtyTwoBits] = cached_ntstatus(0xC022003A, 'STATUS_FWP_INCOMPATIBLE_CIPHER_TRANSFORM')
    STATUS_FWP_INVALID_TRANSFORM_COMBINATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC022003B, 'STATUS_FWP_INVALID_TRANSFORM_COMBINATION')
    STATUS_FWP_DUPLICATE_AUTH_METHOD: Final[ThirtyTwoBits] = cached_ntstatus(0xC022003C, 'STATUS_FWP_DUPLICATE_AUTH_METHOD')
    STATUS_FWP_INVALID_TUNNEL_ENDPOINT: Final[ThirtyTwoBits] = cached_ntstatus(0xC022003D, 'STATUS_FWP_INVALID_TUNNEL_ENDPOINT')
    STATUS_FWP_L2_DRIVER_NOT_READY: Final[ThirtyTwoBits] = cached_ntstatus(0xC022003E, 'STATUS_FWP_L2_DRIVER_NOT_READY')
    STATUS_FWP_KEY_DICTATOR_ALREADY_REGISTERED: Final[ThirtyTwoBits] = cached_ntstatus(0xC022003F, 'STATUS_FWP_KEY_DICTATOR_ALREADY_REGISTERED')
    STATUS_FWP_KEY_DICTATION_INVALID_KEYING_MATERIAL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220040, 'STATUS_FWP_KEY_DICTATION_INVALID_KEYING_MATERIAL')
    STATUS_FWP_CONNECTIONS_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220041, 'STATUS_FWP_CONNECTIONS_DISABLED')
    STATUS_FWP_INVALID_DNS_NAME: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220042, 'STATUS_FWP_INVALID_DNS_NAME')
    STATUS_FWP_STILL_ON: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220043, 'STATUS_FWP_STILL_ON')
    STATUS_FWP_IKEEXT_NOT_RUNNING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220044, 'STATUS_FWP_IKEEXT_NOT_RUNNING')
    STATUS_FWP_TCPIP_NOT_READY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220100, 'STATUS_FWP_TCPIP_NOT_READY')
    STATUS_FWP_INJECT_HANDLE_CLOSING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220101, 'STATUS_FWP_INJECT_HANDLE_CLOSING')
    STATUS_FWP_INJECT_HANDLE_STALE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220102, 'STATUS_FWP_INJECT_HANDLE_STALE')
    STATUS_FWP_CANNOT_PEND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220103, 'STATUS_FWP_CANNOT_PEND')
    STATUS_FWP_DROP_NOICMP: Final[ThirtyTwoBits] = cached_ntstatus(0xC0220104, 'STATUS_FWP_DROP_NOICMP')
    STATUS_NDIS_CLOSING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0230002, 'STATUS_NDIS_CLOSING')
    STATUS_NDIS_BAD_VERSION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0230004, 'STATUS_NDIS_BAD_VERSION')
    STATUS_NDIS_BAD_CHARACTERISTICS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0230005, 'STATUS_NDIS_BAD_CHARACTERISTICS')
    STATUS_NDIS_ADAPTER_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0230006, 'STATUS_NDIS_ADAPTER_NOT_FOUND')
    STATUS_NDIS_OPEN_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0230007, 'STATUS_NDIS_OPEN_FAILED')
    STATUS_NDIS_DEVICE_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0230008, 'STATUS_NDIS_DEVICE_FAILED')
    STATUS_NDIS_MULTICAST_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0230009, 'STATUS_NDIS_MULTICAST_FULL')
    STATUS_NDIS_MULTICAST_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC023000A, 'STATUS_NDIS_MULTICAST_EXISTS')
    STATUS_NDIS_MULTICAST_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC023000B, 'STATUS_NDIS_MULTICAST_NOT_FOUND')
    STATUS_NDIS_REQUEST_ABORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC023000C, 'STATUS_NDIS_REQUEST_ABORTED')
    STATUS_NDIS_RESET_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC023000D, 'STATUS_NDIS_RESET_IN_PROGRESS')
    STATUS_NDIS_INVALID_PACKET: Final[ThirtyTwoBits] = cached_ntstatus(0xC023000F, 'STATUS_NDIS_INVALID_PACKET')
    STATUS_NDIS_INVALID_DEVICE_REQUEST: Final[ThirtyTwoBits] = cached_ntstatus(0xC0230010, 'STATUS_NDIS_INVALID_DEVICE_REQUEST')
    STATUS_NDIS_ADAPTER_NOT_READY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0230011, 'STATUS_NDIS_ADAPTER_NOT_READY')
    STATUS_NDIS_INVALID_LENGTH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0230014, 'STATUS_NDIS_INVALID_LENGTH')
    STATUS_NDIS_INVALID_DATA: Final[ThirtyTwoBits] = cached_ntstatus(0xC0230015, 'STATUS_NDIS_INVALID_DATA')
    STATUS_NDIS_BUFFER_TOO_SHORT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0230016, 'STATUS_NDIS_BUFFER_TOO_SHORT')
    STATUS_NDIS_INVALID_OID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0230017, 'STATUS_NDIS_INVALID_OID')
    STATUS_NDIS_ADAPTER_REMOVED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0230018, 'STATUS_NDIS_ADAPTER_REMOVED')
    STATUS_NDIS_UNSUPPORTED_MEDIA: Final[ThirtyTwoBits] = cached_ntstatus(0xC0230019, 'STATUS_NDIS_UNSUPPORTED_MEDIA')
    STATUS_NDIS_GROUP_ADDRESS_IN_USE: Final[ThirtyTwoBits] = cached_ntstatus(0xC023001A, 'STATUS_NDIS_GROUP_ADDRESS_IN_USE')
    STATUS_NDIS_FILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC023001B, 'STATUS_NDIS_FILE_NOT_FOUND')
    STATUS_NDIS_ERROR_READING_FILE: Final[ThirtyTwoBits] = cached_ntstatus(0xC023001C, 'STATUS_NDIS_ERROR_READING_FILE')
    STATUS_NDIS_ALREADY_MAPPED: Final[ThirtyTwoBits] = cached_ntstatus(0xC023001D, 'STATUS_NDIS_ALREADY_MAPPED')
    STATUS_NDIS_RESOURCE_CONFLICT: Final[ThirtyTwoBits] = cached_ntstatus(0xC023001E, 'STATUS_NDIS_RESOURCE_CONFLICT')
    STATUS_NDIS_MEDIA_DISCONNECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC023001F, 'STATUS_NDIS_MEDIA_DISCONNECTED')
    STATUS_NDIS_INVALID_ADDRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0230022, 'STATUS_NDIS_INVALID_ADDRESS')
    STATUS_NDIS_PAUSED: Final[ThirtyTwoBits] = cached_ntstatus(0xC023002A, 'STATUS_NDIS_PAUSED')
    STATUS_NDIS_INTERFACE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC023002B, 'STATUS_NDIS_INTERFACE_NOT_FOUND')
    STATUS_NDIS_UNSUPPORTED_REVISION: Final[ThirtyTwoBits] = cached_ntstatus(0xC023002C, 'STATUS_NDIS_UNSUPPORTED_REVISION')
    STATUS_NDIS_INVALID_PORT: Final[ThirtyTwoBits] = cached_ntstatus(0xC023002D, 'STATUS_NDIS_INVALID_PORT')
    STATUS_NDIS_INVALID_PORT_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC023002E, 'STATUS_NDIS_INVALID_PORT_STATE')
    STATUS_NDIS_LOW_POWER_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC023002F, 'STATUS_NDIS_LOW_POWER_STATE')
    STATUS_NDIS_REINIT_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0230030, 'STATUS_NDIS_REINIT_REQUIRED')
    STATUS_NDIS_NO_QUEUES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0230031, 'STATUS_NDIS_NO_QUEUES')
    STATUS_NDIS_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC02300BB, 'STATUS_NDIS_NOT_SUPPORTED')
    STATUS_NDIS_OFFLOAD_POLICY: Final[ThirtyTwoBits] = cached_ntstatus(0xC023100F, 'STATUS_NDIS_OFFLOAD_POLICY')
    STATUS_NDIS_OFFLOAD_CONNECTION_REJECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0231012, 'STATUS_NDIS_OFFLOAD_CONNECTION_REJECTED')
    STATUS_NDIS_OFFLOAD_PATH_REJECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0231013, 'STATUS_NDIS_OFFLOAD_PATH_REJECTED')
    STATUS_NDIS_DOT11_AUTO_CONFIG_ENABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0232000, 'STATUS_NDIS_DOT11_AUTO_CONFIG_ENABLED')
    STATUS_NDIS_DOT11_MEDIA_IN_USE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0232001, 'STATUS_NDIS_DOT11_MEDIA_IN_USE')
    STATUS_NDIS_DOT11_POWER_STATE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0232002, 'STATUS_NDIS_DOT11_POWER_STATE_INVALID')
    STATUS_NDIS_PM_WOL_PATTERN_LIST_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0232003, 'STATUS_NDIS_PM_WOL_PATTERN_LIST_FULL')
    STATUS_NDIS_PM_PROTOCOL_OFFLOAD_LIST_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0232004, 'STATUS_NDIS_PM_PROTOCOL_OFFLOAD_LIST_FULL')
    STATUS_NDIS_DOT11_AP_CHANNEL_CURRENTLY_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0232005, 'STATUS_NDIS_DOT11_AP_CHANNEL_CURRENTLY_NOT_AVAILABLE')
    STATUS_NDIS_DOT11_AP_BAND_CURRENTLY_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0232006, 'STATUS_NDIS_DOT11_AP_BAND_CURRENTLY_NOT_AVAILABLE')
    STATUS_NDIS_DOT11_AP_CHANNEL_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0232007, 'STATUS_NDIS_DOT11_AP_CHANNEL_NOT_ALLOWED')
    STATUS_NDIS_DOT11_AP_BAND_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0232008, 'STATUS_NDIS_DOT11_AP_BAND_NOT_ALLOWED')
    STATUS_QUIC_HANDSHAKE_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0240000, 'STATUS_QUIC_HANDSHAKE_FAILURE')
    STATUS_QUIC_VER_NEG_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0240001, 'STATUS_QUIC_VER_NEG_FAILURE')
    STATUS_QUIC_USER_CANCELED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0240002, 'STATUS_QUIC_USER_CANCELED')
    STATUS_QUIC_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0240003, 'STATUS_QUIC_INTERNAL_ERROR')
    STATUS_QUIC_PROTOCOL_VIOLATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0240004, 'STATUS_QUIC_PROTOCOL_VIOLATION')
    STATUS_QUIC_CONNECTION_IDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0240005, 'STATUS_QUIC_CONNECTION_IDLE')
    STATUS_QUIC_CONNECTION_TIMEOUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0240006, 'STATUS_QUIC_CONNECTION_TIMEOUT')
    STATUS_QUIC_ALPN_NEG_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0240007, 'STATUS_QUIC_ALPN_NEG_FAILURE')
    STATUS_TPM_ERROR_MASK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290000, 'STATUS_TPM_ERROR_MASK')
    STATUS_TPM_AUTHFAIL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290001, 'STATUS_TPM_AUTHFAIL')
    STATUS_TPM_BADINDEX: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290002, 'STATUS_TPM_BADINDEX')
    STATUS_TPM_BAD_PARAMETER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290003, 'STATUS_TPM_BAD_PARAMETER')
    STATUS_TPM_AUDITFAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290004, 'STATUS_TPM_AUDITFAILURE')
    STATUS_TPM_CLEAR_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290005, 'STATUS_TPM_CLEAR_DISABLED')
    STATUS_TPM_DEACTIVATED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290006, 'STATUS_TPM_DEACTIVATED')
    STATUS_TPM_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290007, 'STATUS_TPM_DISABLED')
    STATUS_TPM_DISABLED_CMD: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290008, 'STATUS_TPM_DISABLED_CMD')
    STATUS_TPM_FAIL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290009, 'STATUS_TPM_FAIL')
    STATUS_TPM_BAD_ORDINAL: Final[ThirtyTwoBits] = cached_ntstatus(0xC029000A, 'STATUS_TPM_BAD_ORDINAL')
    STATUS_TPM_INSTALL_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC029000B, 'STATUS_TPM_INSTALL_DISABLED')
    STATUS_TPM_INVALID_KEYHANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC029000C, 'STATUS_TPM_INVALID_KEYHANDLE')
    STATUS_TPM_KEYNOTFOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC029000D, 'STATUS_TPM_KEYNOTFOUND')
    STATUS_TPM_INAPPROPRIATE_ENC: Final[ThirtyTwoBits] = cached_ntstatus(0xC029000E, 'STATUS_TPM_INAPPROPRIATE_ENC')
    STATUS_TPM_MIGRATEFAIL: Final[ThirtyTwoBits] = cached_ntstatus(0xC029000F, 'STATUS_TPM_MIGRATEFAIL')
    STATUS_TPM_INVALID_PCR_INFO: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290010, 'STATUS_TPM_INVALID_PCR_INFO')
    STATUS_TPM_NOSPACE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290011, 'STATUS_TPM_NOSPACE')
    STATUS_TPM_NOSRK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290012, 'STATUS_TPM_NOSRK')
    STATUS_TPM_NOTSEALED_BLOB: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290013, 'STATUS_TPM_NOTSEALED_BLOB')
    STATUS_TPM_OWNER_SET: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290014, 'STATUS_TPM_OWNER_SET')
    STATUS_TPM_RESOURCES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290015, 'STATUS_TPM_RESOURCES')
    STATUS_TPM_SHORTRANDOM: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290016, 'STATUS_TPM_SHORTRANDOM')
    STATUS_TPM_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290017, 'STATUS_TPM_SIZE')
    STATUS_TPM_WRONGPCRVAL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290018, 'STATUS_TPM_WRONGPCRVAL')
    STATUS_TPM_BAD_PARAM_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290019, 'STATUS_TPM_BAD_PARAM_SIZE')
    STATUS_TPM_SHA_THREAD: Final[ThirtyTwoBits] = cached_ntstatus(0xC029001A, 'STATUS_TPM_SHA_THREAD')
    STATUS_TPM_SHA_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC029001B, 'STATUS_TPM_SHA_ERROR')
    STATUS_TPM_FAILEDSELFTEST: Final[ThirtyTwoBits] = cached_ntstatus(0xC029001C, 'STATUS_TPM_FAILEDSELFTEST')
    STATUS_TPM_AUTH2FAIL: Final[ThirtyTwoBits] = cached_ntstatus(0xC029001D, 'STATUS_TPM_AUTH2FAIL')
    STATUS_TPM_BADTAG: Final[ThirtyTwoBits] = cached_ntstatus(0xC029001E, 'STATUS_TPM_BADTAG')
    STATUS_TPM_IOERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC029001F, 'STATUS_TPM_IOERROR')
    STATUS_TPM_ENCRYPT_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290020, 'STATUS_TPM_ENCRYPT_ERROR')
    STATUS_TPM_DECRYPT_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290021, 'STATUS_TPM_DECRYPT_ERROR')
    STATUS_TPM_INVALID_AUTHHANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290022, 'STATUS_TPM_INVALID_AUTHHANDLE')
    STATUS_TPM_NO_ENDORSEMENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290023, 'STATUS_TPM_NO_ENDORSEMENT')
    STATUS_TPM_INVALID_KEYUSAGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290024, 'STATUS_TPM_INVALID_KEYUSAGE')
    STATUS_TPM_WRONG_ENTITYTYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290025, 'STATUS_TPM_WRONG_ENTITYTYPE')
    STATUS_TPM_INVALID_POSTINIT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290026, 'STATUS_TPM_INVALID_POSTINIT')
    STATUS_TPM_INAPPROPRIATE_SIG: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290027, 'STATUS_TPM_INAPPROPRIATE_SIG')
    STATUS_TPM_BAD_KEY_PROPERTY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290028, 'STATUS_TPM_BAD_KEY_PROPERTY')
    STATUS_TPM_BAD_MIGRATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290029, 'STATUS_TPM_BAD_MIGRATION')
    STATUS_TPM_BAD_SCHEME: Final[ThirtyTwoBits] = cached_ntstatus(0xC029002A, 'STATUS_TPM_BAD_SCHEME')
    STATUS_TPM_BAD_DATASIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC029002B, 'STATUS_TPM_BAD_DATASIZE')
    STATUS_TPM_BAD_MODE: Final[ThirtyTwoBits] = cached_ntstatus(0xC029002C, 'STATUS_TPM_BAD_MODE')
    STATUS_TPM_BAD_PRESENCE: Final[ThirtyTwoBits] = cached_ntstatus(0xC029002D, 'STATUS_TPM_BAD_PRESENCE')
    STATUS_TPM_BAD_VERSION: Final[ThirtyTwoBits] = cached_ntstatus(0xC029002E, 'STATUS_TPM_BAD_VERSION')
    STATUS_TPM_NO_WRAP_TRANSPORT: Final[ThirtyTwoBits] = cached_ntstatus(0xC029002F, 'STATUS_TPM_NO_WRAP_TRANSPORT')
    STATUS_TPM_AUDITFAIL_UNSUCCESSFUL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290030, 'STATUS_TPM_AUDITFAIL_UNSUCCESSFUL')
    STATUS_TPM_AUDITFAIL_SUCCESSFUL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290031, 'STATUS_TPM_AUDITFAIL_SUCCESSFUL')
    STATUS_TPM_NOTRESETABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290032, 'STATUS_TPM_NOTRESETABLE')
    STATUS_TPM_NOTLOCAL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290033, 'STATUS_TPM_NOTLOCAL')
    STATUS_TPM_BAD_TYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290034, 'STATUS_TPM_BAD_TYPE')
    STATUS_TPM_INVALID_RESOURCE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290035, 'STATUS_TPM_INVALID_RESOURCE')
    STATUS_TPM_NOTFIPS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290036, 'STATUS_TPM_NOTFIPS')
    STATUS_TPM_INVALID_FAMILY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290037, 'STATUS_TPM_INVALID_FAMILY')
    STATUS_TPM_NO_NV_PERMISSION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290038, 'STATUS_TPM_NO_NV_PERMISSION')
    STATUS_TPM_REQUIRES_SIGN: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290039, 'STATUS_TPM_REQUIRES_SIGN')
    STATUS_TPM_KEY_NOTSUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC029003A, 'STATUS_TPM_KEY_NOTSUPPORTED')
    STATUS_TPM_AUTH_CONFLICT: Final[ThirtyTwoBits] = cached_ntstatus(0xC029003B, 'STATUS_TPM_AUTH_CONFLICT')
    STATUS_TPM_AREA_LOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC029003C, 'STATUS_TPM_AREA_LOCKED')
    STATUS_TPM_BAD_LOCALITY: Final[ThirtyTwoBits] = cached_ntstatus(0xC029003D, 'STATUS_TPM_BAD_LOCALITY')
    STATUS_TPM_READ_ONLY: Final[ThirtyTwoBits] = cached_ntstatus(0xC029003E, 'STATUS_TPM_READ_ONLY')
    STATUS_TPM_PER_NOWRITE: Final[ThirtyTwoBits] = cached_ntstatus(0xC029003F, 'STATUS_TPM_PER_NOWRITE')
    STATUS_TPM_FAMILYCOUNT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290040, 'STATUS_TPM_FAMILYCOUNT')
    STATUS_TPM_WRITE_LOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290041, 'STATUS_TPM_WRITE_LOCKED')
    STATUS_TPM_BAD_ATTRIBUTES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290042, 'STATUS_TPM_BAD_ATTRIBUTES')
    STATUS_TPM_INVALID_STRUCTURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290043, 'STATUS_TPM_INVALID_STRUCTURE')
    STATUS_TPM_KEY_OWNER_CONTROL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290044, 'STATUS_TPM_KEY_OWNER_CONTROL')
    STATUS_TPM_BAD_COUNTER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290045, 'STATUS_TPM_BAD_COUNTER')
    STATUS_TPM_NOT_FULLWRITE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290046, 'STATUS_TPM_NOT_FULLWRITE')
    STATUS_TPM_CONTEXT_GAP: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290047, 'STATUS_TPM_CONTEXT_GAP')
    STATUS_TPM_MAXNVWRITES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290048, 'STATUS_TPM_MAXNVWRITES')
    STATUS_TPM_NOOPERATOR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290049, 'STATUS_TPM_NOOPERATOR')
    STATUS_TPM_RESOURCEMISSING: Final[ThirtyTwoBits] = cached_ntstatus(0xC029004A, 'STATUS_TPM_RESOURCEMISSING')
    STATUS_TPM_DELEGATE_LOCK: Final[ThirtyTwoBits] = cached_ntstatus(0xC029004B, 'STATUS_TPM_DELEGATE_LOCK')
    STATUS_TPM_DELEGATE_FAMILY: Final[ThirtyTwoBits] = cached_ntstatus(0xC029004C, 'STATUS_TPM_DELEGATE_FAMILY')
    STATUS_TPM_DELEGATE_ADMIN: Final[ThirtyTwoBits] = cached_ntstatus(0xC029004D, 'STATUS_TPM_DELEGATE_ADMIN')
    STATUS_TPM_TRANSPORT_NOTEXCLUSIVE: Final[ThirtyTwoBits] = cached_ntstatus(0xC029004E, 'STATUS_TPM_TRANSPORT_NOTEXCLUSIVE')
    STATUS_TPM_OWNER_CONTROL: Final[ThirtyTwoBits] = cached_ntstatus(0xC029004F, 'STATUS_TPM_OWNER_CONTROL')
    STATUS_TPM_DAA_RESOURCES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290050, 'STATUS_TPM_DAA_RESOURCES')
    STATUS_TPM_DAA_INPUT_DATA0: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290051, 'STATUS_TPM_DAA_INPUT_DATA0')
    STATUS_TPM_DAA_INPUT_DATA1: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290052, 'STATUS_TPM_DAA_INPUT_DATA1')
    STATUS_TPM_DAA_ISSUER_SETTINGS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290053, 'STATUS_TPM_DAA_ISSUER_SETTINGS')
    STATUS_TPM_DAA_TPM_SETTINGS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290054, 'STATUS_TPM_DAA_TPM_SETTINGS')
    STATUS_TPM_DAA_STAGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290055, 'STATUS_TPM_DAA_STAGE')
    STATUS_TPM_DAA_ISSUER_VALIDITY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290056, 'STATUS_TPM_DAA_ISSUER_VALIDITY')
    STATUS_TPM_DAA_WRONG_W: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290057, 'STATUS_TPM_DAA_WRONG_W')
    STATUS_TPM_BAD_HANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290058, 'STATUS_TPM_BAD_HANDLE')
    STATUS_TPM_BAD_DELEGATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290059, 'STATUS_TPM_BAD_DELEGATE')
    STATUS_TPM_BADCONTEXT: Final[ThirtyTwoBits] = cached_ntstatus(0xC029005A, 'STATUS_TPM_BADCONTEXT')
    STATUS_TPM_TOOMANYCONTEXTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC029005B, 'STATUS_TPM_TOOMANYCONTEXTS')
    STATUS_TPM_MA_TICKET_SIGNATURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC029005C, 'STATUS_TPM_MA_TICKET_SIGNATURE')
    STATUS_TPM_MA_DESTINATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC029005D, 'STATUS_TPM_MA_DESTINATION')
    STATUS_TPM_MA_SOURCE: Final[ThirtyTwoBits] = cached_ntstatus(0xC029005E, 'STATUS_TPM_MA_SOURCE')
    STATUS_TPM_MA_AUTHORITY: Final[ThirtyTwoBits] = cached_ntstatus(0xC029005F, 'STATUS_TPM_MA_AUTHORITY')
    STATUS_TPM_PERMANENTEK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290061, 'STATUS_TPM_PERMANENTEK')
    STATUS_TPM_BAD_SIGNATURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290062, 'STATUS_TPM_BAD_SIGNATURE')
    STATUS_TPM_NOCONTEXTSPACE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290063, 'STATUS_TPM_NOCONTEXTSPACE')
    STATUS_TPM_20_E_ASYMMETRIC: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290081, 'STATUS_TPM_20_E_ASYMMETRIC')
    STATUS_TPM_20_E_ATTRIBUTES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290082, 'STATUS_TPM_20_E_ATTRIBUTES')
    STATUS_TPM_20_E_HASH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290083, 'STATUS_TPM_20_E_HASH')
    STATUS_TPM_20_E_VALUE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290084, 'STATUS_TPM_20_E_VALUE')
    STATUS_TPM_20_E_HIERARCHY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290085, 'STATUS_TPM_20_E_HIERARCHY')
    STATUS_TPM_20_E_KEY_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290087, 'STATUS_TPM_20_E_KEY_SIZE')
    STATUS_TPM_20_E_MGF: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290088, 'STATUS_TPM_20_E_MGF')
    STATUS_TPM_20_E_MODE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290089, 'STATUS_TPM_20_E_MODE')
    STATUS_TPM_20_E_TYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC029008A, 'STATUS_TPM_20_E_TYPE')
    STATUS_TPM_20_E_HANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC029008B, 'STATUS_TPM_20_E_HANDLE')
    STATUS_TPM_20_E_KDF: Final[ThirtyTwoBits] = cached_ntstatus(0xC029008C, 'STATUS_TPM_20_E_KDF')
    STATUS_TPM_20_E_RANGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC029008D, 'STATUS_TPM_20_E_RANGE')
    STATUS_TPM_20_E_AUTH_FAIL: Final[ThirtyTwoBits] = cached_ntstatus(0xC029008E, 'STATUS_TPM_20_E_AUTH_FAIL')
    STATUS_TPM_20_E_NONCE: Final[ThirtyTwoBits] = cached_ntstatus(0xC029008F, 'STATUS_TPM_20_E_NONCE')
    STATUS_TPM_20_E_PP: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290090, 'STATUS_TPM_20_E_PP')
    STATUS_TPM_20_E_SCHEME: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290092, 'STATUS_TPM_20_E_SCHEME')
    STATUS_TPM_20_E_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290095, 'STATUS_TPM_20_E_SIZE')
    STATUS_TPM_20_E_SYMMETRIC: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290096, 'STATUS_TPM_20_E_SYMMETRIC')
    STATUS_TPM_20_E_TAG: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290097, 'STATUS_TPM_20_E_TAG')
    STATUS_TPM_20_E_SELECTOR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290098, 'STATUS_TPM_20_E_SELECTOR')
    STATUS_TPM_20_E_INSUFFICIENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC029009A, 'STATUS_TPM_20_E_INSUFFICIENT')
    STATUS_TPM_20_E_SIGNATURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC029009B, 'STATUS_TPM_20_E_SIGNATURE')
    STATUS_TPM_20_E_KEY: Final[ThirtyTwoBits] = cached_ntstatus(0xC029009C, 'STATUS_TPM_20_E_KEY')
    STATUS_TPM_20_E_POLICY_FAIL: Final[ThirtyTwoBits] = cached_ntstatus(0xC029009D, 'STATUS_TPM_20_E_POLICY_FAIL')
    STATUS_TPM_20_E_INTEGRITY: Final[ThirtyTwoBits] = cached_ntstatus(0xC029009F, 'STATUS_TPM_20_E_INTEGRITY')
    STATUS_TPM_20_E_TICKET: Final[ThirtyTwoBits] = cached_ntstatus(0xC02900A0, 'STATUS_TPM_20_E_TICKET')
    STATUS_TPM_20_E_RESERVED_BITS: Final[ThirtyTwoBits] = cached_ntstatus(0xC02900A1, 'STATUS_TPM_20_E_RESERVED_BITS')
    STATUS_TPM_20_E_BAD_AUTH: Final[ThirtyTwoBits] = cached_ntstatus(0xC02900A2, 'STATUS_TPM_20_E_BAD_AUTH')
    STATUS_TPM_20_E_EXPIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC02900A3, 'STATUS_TPM_20_E_EXPIRED')
    STATUS_TPM_20_E_POLICY_CC: Final[ThirtyTwoBits] = cached_ntstatus(0xC02900A4, 'STATUS_TPM_20_E_POLICY_CC')
    STATUS_TPM_20_E_BINDING: Final[ThirtyTwoBits] = cached_ntstatus(0xC02900A5, 'STATUS_TPM_20_E_BINDING')
    STATUS_TPM_20_E_CURVE: Final[ThirtyTwoBits] = cached_ntstatus(0xC02900A6, 'STATUS_TPM_20_E_CURVE')
    STATUS_TPM_20_E_ECC_POINT: Final[ThirtyTwoBits] = cached_ntstatus(0xC02900A7, 'STATUS_TPM_20_E_ECC_POINT')
    STATUS_TPM_20_E_INITIALIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290100, 'STATUS_TPM_20_E_INITIALIZE')
    STATUS_TPM_20_E_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290101, 'STATUS_TPM_20_E_FAILURE')
    STATUS_TPM_20_E_SEQUENCE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290103, 'STATUS_TPM_20_E_SEQUENCE')
    STATUS_TPM_20_E_PRIVATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC029010B, 'STATUS_TPM_20_E_PRIVATE')
    STATUS_TPM_20_E_HMAC: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290119, 'STATUS_TPM_20_E_HMAC')
    STATUS_TPM_20_E_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290120, 'STATUS_TPM_20_E_DISABLED')
    STATUS_TPM_20_E_EXCLUSIVE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290121, 'STATUS_TPM_20_E_EXCLUSIVE')
    STATUS_TPM_20_E_ECC_CURVE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290123, 'STATUS_TPM_20_E_ECC_CURVE')
    STATUS_TPM_20_E_AUTH_TYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290124, 'STATUS_TPM_20_E_AUTH_TYPE')
    STATUS_TPM_20_E_AUTH_MISSING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290125, 'STATUS_TPM_20_E_AUTH_MISSING')
    STATUS_TPM_20_E_POLICY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290126, 'STATUS_TPM_20_E_POLICY')
    STATUS_TPM_20_E_PCR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290127, 'STATUS_TPM_20_E_PCR')
    STATUS_TPM_20_E_PCR_CHANGED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290128, 'STATUS_TPM_20_E_PCR_CHANGED')
    STATUS_TPM_20_E_UPGRADE: Final[ThirtyTwoBits] = cached_ntstatus(0xC029012D, 'STATUS_TPM_20_E_UPGRADE')
    STATUS_TPM_20_E_TOO_MANY_CONTEXTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC029012E, 'STATUS_TPM_20_E_TOO_MANY_CONTEXTS')
    STATUS_TPM_20_E_AUTH_UNAVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC029012F, 'STATUS_TPM_20_E_AUTH_UNAVAILABLE')
    STATUS_TPM_20_E_REBOOT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290130, 'STATUS_TPM_20_E_REBOOT')
    STATUS_TPM_20_E_UNBALANCED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290131, 'STATUS_TPM_20_E_UNBALANCED')
    STATUS_TPM_20_E_COMMAND_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290142, 'STATUS_TPM_20_E_COMMAND_SIZE')
    STATUS_TPM_20_E_COMMAND_CODE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290143, 'STATUS_TPM_20_E_COMMAND_CODE')
    STATUS_TPM_20_E_AUTHSIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290144, 'STATUS_TPM_20_E_AUTHSIZE')
    STATUS_TPM_20_E_AUTH_CONTEXT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290145, 'STATUS_TPM_20_E_AUTH_CONTEXT')
    STATUS_TPM_20_E_NV_RANGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290146, 'STATUS_TPM_20_E_NV_RANGE')
    STATUS_TPM_20_E_NV_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290147, 'STATUS_TPM_20_E_NV_SIZE')
    STATUS_TPM_20_E_NV_LOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290148, 'STATUS_TPM_20_E_NV_LOCKED')
    STATUS_TPM_20_E_NV_AUTHORIZATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290149, 'STATUS_TPM_20_E_NV_AUTHORIZATION')
    STATUS_TPM_20_E_NV_UNINITIALIZED: Final[ThirtyTwoBits] = cached_ntstatus(0xC029014A, 'STATUS_TPM_20_E_NV_UNINITIALIZED')
    STATUS_TPM_20_E_NV_SPACE: Final[ThirtyTwoBits] = cached_ntstatus(0xC029014B, 'STATUS_TPM_20_E_NV_SPACE')
    STATUS_TPM_20_E_NV_DEFINED: Final[ThirtyTwoBits] = cached_ntstatus(0xC029014C, 'STATUS_TPM_20_E_NV_DEFINED')
    STATUS_TPM_20_E_BAD_CONTEXT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290150, 'STATUS_TPM_20_E_BAD_CONTEXT')
    STATUS_TPM_20_E_CPHASH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290151, 'STATUS_TPM_20_E_CPHASH')
    STATUS_TPM_20_E_PARENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290152, 'STATUS_TPM_20_E_PARENT')
    STATUS_TPM_20_E_NEEDS_TEST: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290153, 'STATUS_TPM_20_E_NEEDS_TEST')
    STATUS_TPM_20_E_NO_RESULT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290154, 'STATUS_TPM_20_E_NO_RESULT')
    STATUS_TPM_20_E_SENSITIVE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290155, 'STATUS_TPM_20_E_SENSITIVE')
    STATUS_TPM_COMMAND_BLOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290400, 'STATUS_TPM_COMMAND_BLOCKED')
    STATUS_TPM_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290401, 'STATUS_TPM_INVALID_HANDLE')
    STATUS_TPM_DUPLICATE_VHANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290402, 'STATUS_TPM_DUPLICATE_VHANDLE')
    STATUS_TPM_EMBEDDED_COMMAND_BLOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290403, 'STATUS_TPM_EMBEDDED_COMMAND_BLOCKED')
    STATUS_TPM_EMBEDDED_COMMAND_UNSUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290404, 'STATUS_TPM_EMBEDDED_COMMAND_UNSUPPORTED')
    STATUS_TPM_RETRY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290800, 'STATUS_TPM_RETRY')
    STATUS_TPM_NEEDS_SELFTEST: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290801, 'STATUS_TPM_NEEDS_SELFTEST')
    STATUS_TPM_DOING_SELFTEST: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290802, 'STATUS_TPM_DOING_SELFTEST')
    STATUS_TPM_DEFEND_LOCK_RUNNING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0290803, 'STATUS_TPM_DEFEND_LOCK_RUNNING')
    STATUS_TPM_COMMAND_CANCELED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0291001, 'STATUS_TPM_COMMAND_CANCELED')
    STATUS_TPM_TOO_MANY_CONTEXTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0291002, 'STATUS_TPM_TOO_MANY_CONTEXTS')
    STATUS_TPM_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0291003, 'STATUS_TPM_NOT_FOUND')
    STATUS_TPM_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0291004, 'STATUS_TPM_ACCESS_DENIED')
    STATUS_TPM_INSUFFICIENT_BUFFER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0291005, 'STATUS_TPM_INSUFFICIENT_BUFFER')
    STATUS_TPM_PPI_FUNCTION_UNSUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0291006, 'STATUS_TPM_PPI_FUNCTION_UNSUPPORTED')
    STATUS_PCP_ERROR_MASK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292000, 'STATUS_PCP_ERROR_MASK')
    STATUS_PCP_DEVICE_NOT_READY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292001, 'STATUS_PCP_DEVICE_NOT_READY')
    STATUS_PCP_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292002, 'STATUS_PCP_INVALID_HANDLE')
    STATUS_PCP_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292003, 'STATUS_PCP_INVALID_PARAMETER')
    STATUS_PCP_FLAG_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292004, 'STATUS_PCP_FLAG_NOT_SUPPORTED')
    STATUS_PCP_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292005, 'STATUS_PCP_NOT_SUPPORTED')
    STATUS_PCP_BUFFER_TOO_SMALL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292006, 'STATUS_PCP_BUFFER_TOO_SMALL')
    STATUS_PCP_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292007, 'STATUS_PCP_INTERNAL_ERROR')
    STATUS_PCP_AUTHENTICATION_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292008, 'STATUS_PCP_AUTHENTICATION_FAILED')
    STATUS_PCP_AUTHENTICATION_IGNORED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292009, 'STATUS_PCP_AUTHENTICATION_IGNORED')
    STATUS_PCP_POLICY_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC029200A, 'STATUS_PCP_POLICY_NOT_FOUND')
    STATUS_PCP_PROFILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC029200B, 'STATUS_PCP_PROFILE_NOT_FOUND')
    STATUS_PCP_VALIDATION_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC029200C, 'STATUS_PCP_VALIDATION_FAILED')
    STATUS_PCP_DEVICE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC029200D, 'STATUS_PCP_DEVICE_NOT_FOUND')
    STATUS_PCP_WRONG_PARENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC029200E, 'STATUS_PCP_WRONG_PARENT')
    STATUS_PCP_KEY_NOT_LOADED: Final[ThirtyTwoBits] = cached_ntstatus(0xC029200F, 'STATUS_PCP_KEY_NOT_LOADED')
    STATUS_PCP_NO_KEY_CERTIFICATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292010, 'STATUS_PCP_NO_KEY_CERTIFICATION')
    STATUS_PCP_KEY_NOT_FINALIZED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292011, 'STATUS_PCP_KEY_NOT_FINALIZED')
    STATUS_PCP_ATTESTATION_CHALLENGE_NOT_SET: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292012, 'STATUS_PCP_ATTESTATION_CHALLENGE_NOT_SET')
    STATUS_PCP_NOT_PCR_BOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292013, 'STATUS_PCP_NOT_PCR_BOUND')
    STATUS_PCP_KEY_ALREADY_FINALIZED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292014, 'STATUS_PCP_KEY_ALREADY_FINALIZED')
    STATUS_PCP_KEY_USAGE_POLICY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292015, 'STATUS_PCP_KEY_USAGE_POLICY_NOT_SUPPORTED')
    STATUS_PCP_KEY_USAGE_POLICY_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292016, 'STATUS_PCP_KEY_USAGE_POLICY_INVALID')
    STATUS_PCP_SOFT_KEY_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292017, 'STATUS_PCP_SOFT_KEY_ERROR')
    STATUS_PCP_KEY_NOT_AUTHENTICATED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292018, 'STATUS_PCP_KEY_NOT_AUTHENTICATED')
    STATUS_PCP_KEY_NOT_AIK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292019, 'STATUS_PCP_KEY_NOT_AIK')
    STATUS_PCP_KEY_NOT_SIGNING_KEY: Final[ThirtyTwoBits] = cached_ntstatus(0xC029201A, 'STATUS_PCP_KEY_NOT_SIGNING_KEY')
    STATUS_PCP_LOCKED_OUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC029201B, 'STATUS_PCP_LOCKED_OUT')
    STATUS_PCP_CLAIM_TYPE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC029201C, 'STATUS_PCP_CLAIM_TYPE_NOT_SUPPORTED')
    STATUS_PCP_TPM_VERSION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC029201D, 'STATUS_PCP_TPM_VERSION_NOT_SUPPORTED')
    STATUS_PCP_BUFFER_LENGTH_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC029201E, 'STATUS_PCP_BUFFER_LENGTH_MISMATCH')
    STATUS_PCP_IFX_RSA_KEY_CREATION_BLOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC029201F, 'STATUS_PCP_IFX_RSA_KEY_CREATION_BLOCKED')
    STATUS_PCP_TICKET_MISSING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292020, 'STATUS_PCP_TICKET_MISSING')
    STATUS_PCP_RAW_POLICY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292021, 'STATUS_PCP_RAW_POLICY_NOT_SUPPORTED')
    STATUS_PCP_KEY_HANDLE_INVALIDATED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0292022, 'STATUS_PCP_KEY_HANDLE_INVALIDATED')
    STATUS_RTPM_NO_RESULT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0293002, 'STATUS_RTPM_NO_RESULT')
    STATUS_RTPM_PCR_READ_INCOMPLETE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0293003, 'STATUS_RTPM_PCR_READ_INCOMPLETE')
    STATUS_RTPM_INVALID_CONTEXT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0293004, 'STATUS_RTPM_INVALID_CONTEXT')
    STATUS_RTPM_UNSUPPORTED_CMD: Final[ThirtyTwoBits] = cached_ntstatus(0xC0293005, 'STATUS_RTPM_UNSUPPORTED_CMD')
    STATUS_TPM_ZERO_EXHAUST_ENABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0294000, 'STATUS_TPM_ZERO_EXHAUST_ENABLED')
    STATUS_HV_INVALID_HYPERCALL_CODE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350002, 'STATUS_HV_INVALID_HYPERCALL_CODE')
    STATUS_HV_INVALID_HYPERCALL_INPUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350003, 'STATUS_HV_INVALID_HYPERCALL_INPUT')
    STATUS_HV_INVALID_ALIGNMENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350004, 'STATUS_HV_INVALID_ALIGNMENT')
    STATUS_HV_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350005, 'STATUS_HV_INVALID_PARAMETER')
    STATUS_HV_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350006, 'STATUS_HV_ACCESS_DENIED')
    STATUS_HV_INVALID_PARTITION_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350007, 'STATUS_HV_INVALID_PARTITION_STATE')
    STATUS_HV_OPERATION_DENIED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350008, 'STATUS_HV_OPERATION_DENIED')
    STATUS_HV_UNKNOWN_PROPERTY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350009, 'STATUS_HV_UNKNOWN_PROPERTY')
    STATUS_HV_PROPERTY_VALUE_OUT_OF_RANGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC035000A, 'STATUS_HV_PROPERTY_VALUE_OUT_OF_RANGE')
    STATUS_HV_INSUFFICIENT_MEMORY: Final[ThirtyTwoBits] = cached_ntstatus(0xC035000B, 'STATUS_HV_INSUFFICIENT_MEMORY')
    STATUS_HV_PARTITION_TOO_DEEP: Final[ThirtyTwoBits] = cached_ntstatus(0xC035000C, 'STATUS_HV_PARTITION_TOO_DEEP')
    STATUS_HV_INVALID_PARTITION_ID: Final[ThirtyTwoBits] = cached_ntstatus(0xC035000D, 'STATUS_HV_INVALID_PARTITION_ID')
    STATUS_HV_INVALID_VP_INDEX: Final[ThirtyTwoBits] = cached_ntstatus(0xC035000E, 'STATUS_HV_INVALID_VP_INDEX')
    STATUS_HV_INVALID_PORT_ID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350011, 'STATUS_HV_INVALID_PORT_ID')
    STATUS_HV_INVALID_CONNECTION_ID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350012, 'STATUS_HV_INVALID_CONNECTION_ID')
    STATUS_HV_INSUFFICIENT_BUFFERS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350013, 'STATUS_HV_INSUFFICIENT_BUFFERS')
    STATUS_HV_NOT_ACKNOWLEDGED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350014, 'STATUS_HV_NOT_ACKNOWLEDGED')
    STATUS_HV_INVALID_VP_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350015, 'STATUS_HV_INVALID_VP_STATE')
    STATUS_HV_ACKNOWLEDGED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350016, 'STATUS_HV_ACKNOWLEDGED')
    STATUS_HV_INVALID_SAVE_RESTORE_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350017, 'STATUS_HV_INVALID_SAVE_RESTORE_STATE')
    STATUS_HV_INVALID_SYNIC_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350018, 'STATUS_HV_INVALID_SYNIC_STATE')
    STATUS_HV_OBJECT_IN_USE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350019, 'STATUS_HV_OBJECT_IN_USE')
    STATUS_HV_INVALID_PROXIMITY_DOMAIN_INFO: Final[ThirtyTwoBits] = cached_ntstatus(0xC035001A, 'STATUS_HV_INVALID_PROXIMITY_DOMAIN_INFO')
    STATUS_HV_NO_DATA: Final[ThirtyTwoBits] = cached_ntstatus(0xC035001B, 'STATUS_HV_NO_DATA')
    STATUS_HV_INACTIVE: Final[ThirtyTwoBits] = cached_ntstatus(0xC035001C, 'STATUS_HV_INACTIVE')
    STATUS_HV_NO_RESOURCES: Final[ThirtyTwoBits] = cached_ntstatus(0xC035001D, 'STATUS_HV_NO_RESOURCES')
    STATUS_HV_FEATURE_UNAVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC035001E, 'STATUS_HV_FEATURE_UNAVAILABLE')
    STATUS_HV_INSUFFICIENT_BUFFER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350033, 'STATUS_HV_INSUFFICIENT_BUFFER')
    STATUS_HV_INSUFFICIENT_DEVICE_DOMAINS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350038, 'STATUS_HV_INSUFFICIENT_DEVICE_DOMAINS')
    STATUS_HV_CPUID_FEATURE_VALIDATION_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC035003C, 'STATUS_HV_CPUID_FEATURE_VALIDATION_ERROR')
    STATUS_HV_CPUID_XSAVE_FEATURE_VALIDATION_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC035003D, 'STATUS_HV_CPUID_XSAVE_FEATURE_VALIDATION_ERROR')
    STATUS_HV_PROCESSOR_STARTUP_TIMEOUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC035003E, 'STATUS_HV_PROCESSOR_STARTUP_TIMEOUT')
    STATUS_HV_SMX_ENABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC035003F, 'STATUS_HV_SMX_ENABLED')
    STATUS_HV_INVALID_LP_INDEX: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350041, 'STATUS_HV_INVALID_LP_INDEX')
    STATUS_HV_INVALID_REGISTER_VALUE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350050, 'STATUS_HV_INVALID_REGISTER_VALUE')
    STATUS_HV_INVALID_VTL_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350051, 'STATUS_HV_INVALID_VTL_STATE')
    STATUS_HV_NX_NOT_DETECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350055, 'STATUS_HV_NX_NOT_DETECTED')
    STATUS_HV_INVALID_DEVICE_ID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350057, 'STATUS_HV_INVALID_DEVICE_ID')
    STATUS_HV_INVALID_DEVICE_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350058, 'STATUS_HV_INVALID_DEVICE_STATE')
    STATUS_HV_PAGE_REQUEST_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350060, 'STATUS_HV_PAGE_REQUEST_INVALID')
    STATUS_HV_INVALID_CPU_GROUP_ID: Final[ThirtyTwoBits] = cached_ntstatus(0xC035006F, 'STATUS_HV_INVALID_CPU_GROUP_ID')
    STATUS_HV_INVALID_CPU_GROUP_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350070, 'STATUS_HV_INVALID_CPU_GROUP_STATE')
    STATUS_HV_OPERATION_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350071, 'STATUS_HV_OPERATION_FAILED')
    STATUS_HV_NOT_ALLOWED_WITH_NESTED_VIRT_ACTIVE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350072, 'STATUS_HV_NOT_ALLOWED_WITH_NESTED_VIRT_ACTIVE')
    STATUS_HV_INSUFFICIENT_ROOT_MEMORY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350073, 'STATUS_HV_INSUFFICIENT_ROOT_MEMORY')
    STATUS_HV_EVENT_BUFFER_ALREADY_FREED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350074, 'STATUS_HV_EVENT_BUFFER_ALREADY_FREED')
    STATUS_HV_INSUFFICIENT_CONTIGUOUS_MEMORY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350075, 'STATUS_HV_INSUFFICIENT_CONTIGUOUS_MEMORY')
    STATUS_HV_DEVICE_NOT_IN_DOMAIN: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350076, 'STATUS_HV_DEVICE_NOT_IN_DOMAIN')
    STATUS_HV_NESTED_VM_EXIT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350077, 'STATUS_HV_NESTED_VM_EXIT')
    STATUS_HV_CALL_PENDING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350079, 'STATUS_HV_CALL_PENDING')
    STATUS_HV_MSR_ACCESS_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350080, 'STATUS_HV_MSR_ACCESS_FAILED')
    STATUS_HV_INSUFFICIENT_MEMORY_MIRRORING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350081, 'STATUS_HV_INSUFFICIENT_MEMORY_MIRRORING')
    STATUS_HV_INSUFFICIENT_CONTIGUOUS_MEMORY_MIRRORING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350082, 'STATUS_HV_INSUFFICIENT_CONTIGUOUS_MEMORY_MIRRORING')
    STATUS_HV_INSUFFICIENT_CONTIGUOUS_ROOT_MEMORY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350083, 'STATUS_HV_INSUFFICIENT_CONTIGUOUS_ROOT_MEMORY')
    STATUS_HV_INSUFFICIENT_ROOT_MEMORY_MIRRORING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350084, 'STATUS_HV_INSUFFICIENT_ROOT_MEMORY_MIRRORING')
    STATUS_HV_INSUFFICIENT_CONTIGUOUS_ROOT_MEMORY_MIRRORING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0350085, 'STATUS_HV_INSUFFICIENT_CONTIGUOUS_ROOT_MEMORY_MIRRORING')
    STATUS_HV_NOT_PRESENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0351000, 'STATUS_HV_NOT_PRESENT')
    STATUS_IPSEC_BAD_SPI: Final[ThirtyTwoBits] = cached_ntstatus(0xC0360001, 'STATUS_IPSEC_BAD_SPI')
    STATUS_IPSEC_SA_LIFETIME_EXPIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0360002, 'STATUS_IPSEC_SA_LIFETIME_EXPIRED')
    STATUS_IPSEC_WRONG_SA: Final[ThirtyTwoBits] = cached_ntstatus(0xC0360003, 'STATUS_IPSEC_WRONG_SA')
    STATUS_IPSEC_REPLAY_CHECK_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0360004, 'STATUS_IPSEC_REPLAY_CHECK_FAILED')
    STATUS_IPSEC_INVALID_PACKET: Final[ThirtyTwoBits] = cached_ntstatus(0xC0360005, 'STATUS_IPSEC_INVALID_PACKET')
    STATUS_IPSEC_INTEGRITY_CHECK_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0360006, 'STATUS_IPSEC_INTEGRITY_CHECK_FAILED')
    STATUS_IPSEC_CLEAR_TEXT_DROP: Final[ThirtyTwoBits] = cached_ntstatus(0xC0360007, 'STATUS_IPSEC_CLEAR_TEXT_DROP')
    STATUS_IPSEC_AUTH_FIREWALL_DROP: Final[ThirtyTwoBits] = cached_ntstatus(0xC0360008, 'STATUS_IPSEC_AUTH_FIREWALL_DROP')
    STATUS_IPSEC_THROTTLE_DROP: Final[ThirtyTwoBits] = cached_ntstatus(0xC0360009, 'STATUS_IPSEC_THROTTLE_DROP')
    STATUS_IPSEC_DOSP_BLOCK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0368000, 'STATUS_IPSEC_DOSP_BLOCK')
    STATUS_IPSEC_DOSP_RECEIVED_MULTICAST: Final[ThirtyTwoBits] = cached_ntstatus(0xC0368001, 'STATUS_IPSEC_DOSP_RECEIVED_MULTICAST')
    STATUS_IPSEC_DOSP_INVALID_PACKET: Final[ThirtyTwoBits] = cached_ntstatus(0xC0368002, 'STATUS_IPSEC_DOSP_INVALID_PACKET')
    STATUS_IPSEC_DOSP_STATE_LOOKUP_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0368003, 'STATUS_IPSEC_DOSP_STATE_LOOKUP_FAILED')
    STATUS_IPSEC_DOSP_MAX_ENTRIES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0368004, 'STATUS_IPSEC_DOSP_MAX_ENTRIES')
    STATUS_IPSEC_DOSP_KEYMOD_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0368005, 'STATUS_IPSEC_DOSP_KEYMOD_NOT_ALLOWED')
    STATUS_IPSEC_DOSP_MAX_PER_IP_RATELIMIT_QUEUES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0368006, 'STATUS_IPSEC_DOSP_MAX_PER_IP_RATELIMIT_QUEUES')
    STATUS_VID_DUPLICATE_HANDLER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370001, 'STATUS_VID_DUPLICATE_HANDLER')
    STATUS_VID_TOO_MANY_HANDLERS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370002, 'STATUS_VID_TOO_MANY_HANDLERS')
    STATUS_VID_QUEUE_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370003, 'STATUS_VID_QUEUE_FULL')
    STATUS_VID_HANDLER_NOT_PRESENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370004, 'STATUS_VID_HANDLER_NOT_PRESENT')
    STATUS_VID_INVALID_OBJECT_NAME: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370005, 'STATUS_VID_INVALID_OBJECT_NAME')
    STATUS_VID_PARTITION_NAME_TOO_LONG: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370006, 'STATUS_VID_PARTITION_NAME_TOO_LONG')
    STATUS_VID_MESSAGE_QUEUE_NAME_TOO_LONG: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370007, 'STATUS_VID_MESSAGE_QUEUE_NAME_TOO_LONG')
    STATUS_VID_PARTITION_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370008, 'STATUS_VID_PARTITION_ALREADY_EXISTS')
    STATUS_VID_PARTITION_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370009, 'STATUS_VID_PARTITION_DOES_NOT_EXIST')
    STATUS_VID_PARTITION_NAME_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC037000A, 'STATUS_VID_PARTITION_NAME_NOT_FOUND')
    STATUS_VID_MESSAGE_QUEUE_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC037000B, 'STATUS_VID_MESSAGE_QUEUE_ALREADY_EXISTS')
    STATUS_VID_EXCEEDED_MBP_ENTRY_MAP_LIMIT: Final[ThirtyTwoBits] = cached_ntstatus(0xC037000C, 'STATUS_VID_EXCEEDED_MBP_ENTRY_MAP_LIMIT')
    STATUS_VID_MB_STILL_REFERENCED: Final[ThirtyTwoBits] = cached_ntstatus(0xC037000D, 'STATUS_VID_MB_STILL_REFERENCED')
    STATUS_VID_CHILD_GPA_PAGE_SET_CORRUPTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC037000E, 'STATUS_VID_CHILD_GPA_PAGE_SET_CORRUPTED')
    STATUS_VID_INVALID_NUMA_SETTINGS: Final[ThirtyTwoBits] = cached_ntstatus(0xC037000F, 'STATUS_VID_INVALID_NUMA_SETTINGS')
    STATUS_VID_INVALID_NUMA_NODE_INDEX: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370010, 'STATUS_VID_INVALID_NUMA_NODE_INDEX')
    STATUS_VID_NOTIFICATION_QUEUE_ALREADY_ASSOCIATED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370011, 'STATUS_VID_NOTIFICATION_QUEUE_ALREADY_ASSOCIATED')
    STATUS_VID_INVALID_MEMORY_BLOCK_HANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370012, 'STATUS_VID_INVALID_MEMORY_BLOCK_HANDLE')
    STATUS_VID_PAGE_RANGE_OVERFLOW: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370013, 'STATUS_VID_PAGE_RANGE_OVERFLOW')
    STATUS_VID_INVALID_MESSAGE_QUEUE_HANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370014, 'STATUS_VID_INVALID_MESSAGE_QUEUE_HANDLE')
    STATUS_VID_INVALID_GPA_RANGE_HANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370015, 'STATUS_VID_INVALID_GPA_RANGE_HANDLE')
    STATUS_VID_NO_MEMORY_BLOCK_NOTIFICATION_QUEUE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370016, 'STATUS_VID_NO_MEMORY_BLOCK_NOTIFICATION_QUEUE')
    STATUS_VID_MEMORY_BLOCK_LOCK_COUNT_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370017, 'STATUS_VID_MEMORY_BLOCK_LOCK_COUNT_EXCEEDED')
    STATUS_VID_INVALID_PPM_HANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370018, 'STATUS_VID_INVALID_PPM_HANDLE')
    STATUS_VID_MBPS_ARE_LOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370019, 'STATUS_VID_MBPS_ARE_LOCKED')
    STATUS_VID_MESSAGE_QUEUE_CLOSED: Final[ThirtyTwoBits] = cached_ntstatus(0xC037001A, 'STATUS_VID_MESSAGE_QUEUE_CLOSED')
    STATUS_VID_VIRTUAL_PROCESSOR_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC037001B, 'STATUS_VID_VIRTUAL_PROCESSOR_LIMIT_EXCEEDED')
    STATUS_VID_STOP_PENDING: Final[ThirtyTwoBits] = cached_ntstatus(0xC037001C, 'STATUS_VID_STOP_PENDING')
    STATUS_VID_INVALID_PROCESSOR_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC037001D, 'STATUS_VID_INVALID_PROCESSOR_STATE')
    STATUS_VID_EXCEEDED_KM_CONTEXT_COUNT_LIMIT: Final[ThirtyTwoBits] = cached_ntstatus(0xC037001E, 'STATUS_VID_EXCEEDED_KM_CONTEXT_COUNT_LIMIT')
    STATUS_VID_KM_INTERFACE_ALREADY_INITIALIZED: Final[ThirtyTwoBits] = cached_ntstatus(0xC037001F, 'STATUS_VID_KM_INTERFACE_ALREADY_INITIALIZED')
    STATUS_VID_MB_PROPERTY_ALREADY_SET_RESET: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370020, 'STATUS_VID_MB_PROPERTY_ALREADY_SET_RESET')
    STATUS_VID_MMIO_RANGE_DESTROYED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370021, 'STATUS_VID_MMIO_RANGE_DESTROYED')
    STATUS_VID_INVALID_CHILD_GPA_PAGE_SET: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370022, 'STATUS_VID_INVALID_CHILD_GPA_PAGE_SET')
    STATUS_VID_RESERVE_PAGE_SET_IS_BEING_USED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370023, 'STATUS_VID_RESERVE_PAGE_SET_IS_BEING_USED')
    STATUS_VID_RESERVE_PAGE_SET_TOO_SMALL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370024, 'STATUS_VID_RESERVE_PAGE_SET_TOO_SMALL')
    STATUS_VID_MBP_ALREADY_LOCKED_USING_RESERVED_PAGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370025, 'STATUS_VID_MBP_ALREADY_LOCKED_USING_RESERVED_PAGE')
    STATUS_VID_MBP_COUNT_EXCEEDED_LIMIT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370026, 'STATUS_VID_MBP_COUNT_EXCEEDED_LIMIT')
    STATUS_VID_SAVED_STATE_CORRUPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370027, 'STATUS_VID_SAVED_STATE_CORRUPT')
    STATUS_VID_SAVED_STATE_UNRECOGNIZED_ITEM: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370028, 'STATUS_VID_SAVED_STATE_UNRECOGNIZED_ITEM')
    STATUS_VID_SAVED_STATE_INCOMPATIBLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370029, 'STATUS_VID_SAVED_STATE_INCOMPATIBLE')
    STATUS_VID_VTL_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_ntstatus(0xC037002A, 'STATUS_VID_VTL_ACCESS_DENIED')
    STATUS_VID_INSUFFICIENT_RESOURCES_RESERVE: Final[ThirtyTwoBits] = cached_ntstatus(0xC037002B, 'STATUS_VID_INSUFFICIENT_RESOURCES_RESERVE')
    STATUS_VID_INSUFFICIENT_RESOURCES_PHYSICAL_BUFFER: Final[ThirtyTwoBits] = cached_ntstatus(0xC037002C, 'STATUS_VID_INSUFFICIENT_RESOURCES_PHYSICAL_BUFFER')
    STATUS_VID_INSUFFICIENT_RESOURCES_HV_DEPOSIT: Final[ThirtyTwoBits] = cached_ntstatus(0xC037002D, 'STATUS_VID_INSUFFICIENT_RESOURCES_HV_DEPOSIT')
    STATUS_VID_MEMORY_TYPE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC037002E, 'STATUS_VID_MEMORY_TYPE_NOT_SUPPORTED')
    STATUS_VID_INSUFFICIENT_RESOURCES_WITHDRAW: Final[ThirtyTwoBits] = cached_ntstatus(0xC037002F, 'STATUS_VID_INSUFFICIENT_RESOURCES_WITHDRAW')
    STATUS_VID_PROCESS_ALREADY_SET: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370030, 'STATUS_VID_PROCESS_ALREADY_SET')
    STATUS_DM_OPERATION_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0370600, 'STATUS_DM_OPERATION_LIMIT_EXCEEDED')
    STATUS_VOLMGR_DATABASE_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380001, 'STATUS_VOLMGR_DATABASE_FULL')
    STATUS_VOLMGR_DISK_CONFIGURATION_CORRUPTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380002, 'STATUS_VOLMGR_DISK_CONFIGURATION_CORRUPTED')
    STATUS_VOLMGR_DISK_CONFIGURATION_NOT_IN_SYNC: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380003, 'STATUS_VOLMGR_DISK_CONFIGURATION_NOT_IN_SYNC')
    STATUS_VOLMGR_PACK_CONFIG_UPDATE_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380004, 'STATUS_VOLMGR_PACK_CONFIG_UPDATE_FAILED')
    STATUS_VOLMGR_DISK_CONTAINS_NON_SIMPLE_VOLUME: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380005, 'STATUS_VOLMGR_DISK_CONTAINS_NON_SIMPLE_VOLUME')
    STATUS_VOLMGR_DISK_DUPLICATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380006, 'STATUS_VOLMGR_DISK_DUPLICATE')
    STATUS_VOLMGR_DISK_DYNAMIC: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380007, 'STATUS_VOLMGR_DISK_DYNAMIC')
    STATUS_VOLMGR_DISK_ID_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380008, 'STATUS_VOLMGR_DISK_ID_INVALID')
    STATUS_VOLMGR_DISK_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380009, 'STATUS_VOLMGR_DISK_INVALID')
    STATUS_VOLMGR_DISK_LAST_VOTER: Final[ThirtyTwoBits] = cached_ntstatus(0xC038000A, 'STATUS_VOLMGR_DISK_LAST_VOTER')
    STATUS_VOLMGR_DISK_LAYOUT_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC038000B, 'STATUS_VOLMGR_DISK_LAYOUT_INVALID')
    STATUS_VOLMGR_DISK_LAYOUT_NON_BASIC_BETWEEN_BASIC_PARTITIONS: Final[ThirtyTwoBits] = cached_ntstatus(0xC038000C, 'STATUS_VOLMGR_DISK_LAYOUT_NON_BASIC_BETWEEN_BASIC_PARTITIONS')
    STATUS_VOLMGR_DISK_LAYOUT_NOT_CYLINDER_ALIGNED: Final[ThirtyTwoBits] = cached_ntstatus(0xC038000D, 'STATUS_VOLMGR_DISK_LAYOUT_NOT_CYLINDER_ALIGNED')
    STATUS_VOLMGR_DISK_LAYOUT_PARTITIONS_TOO_SMALL: Final[ThirtyTwoBits] = cached_ntstatus(0xC038000E, 'STATUS_VOLMGR_DISK_LAYOUT_PARTITIONS_TOO_SMALL')
    STATUS_VOLMGR_DISK_LAYOUT_PRIMARY_BETWEEN_LOGICAL_PARTITIONS: Final[ThirtyTwoBits] = cached_ntstatus(0xC038000F, 'STATUS_VOLMGR_DISK_LAYOUT_PRIMARY_BETWEEN_LOGICAL_PARTITIONS')
    STATUS_VOLMGR_DISK_LAYOUT_TOO_MANY_PARTITIONS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380010, 'STATUS_VOLMGR_DISK_LAYOUT_TOO_MANY_PARTITIONS')
    STATUS_VOLMGR_DISK_MISSING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380011, 'STATUS_VOLMGR_DISK_MISSING')
    STATUS_VOLMGR_DISK_NOT_EMPTY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380012, 'STATUS_VOLMGR_DISK_NOT_EMPTY')
    STATUS_VOLMGR_DISK_NOT_ENOUGH_SPACE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380013, 'STATUS_VOLMGR_DISK_NOT_ENOUGH_SPACE')
    STATUS_VOLMGR_DISK_REVECTORING_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380014, 'STATUS_VOLMGR_DISK_REVECTORING_FAILED')
    STATUS_VOLMGR_DISK_SECTOR_SIZE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380015, 'STATUS_VOLMGR_DISK_SECTOR_SIZE_INVALID')
    STATUS_VOLMGR_DISK_SET_NOT_CONTAINED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380016, 'STATUS_VOLMGR_DISK_SET_NOT_CONTAINED')
    STATUS_VOLMGR_DISK_USED_BY_MULTIPLE_MEMBERS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380017, 'STATUS_VOLMGR_DISK_USED_BY_MULTIPLE_MEMBERS')
    STATUS_VOLMGR_DISK_USED_BY_MULTIPLE_PLEXES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380018, 'STATUS_VOLMGR_DISK_USED_BY_MULTIPLE_PLEXES')
    STATUS_VOLMGR_DYNAMIC_DISK_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380019, 'STATUS_VOLMGR_DYNAMIC_DISK_NOT_SUPPORTED')
    STATUS_VOLMGR_EXTENT_ALREADY_USED: Final[ThirtyTwoBits] = cached_ntstatus(0xC038001A, 'STATUS_VOLMGR_EXTENT_ALREADY_USED')
    STATUS_VOLMGR_EXTENT_NOT_CONTIGUOUS: Final[ThirtyTwoBits] = cached_ntstatus(0xC038001B, 'STATUS_VOLMGR_EXTENT_NOT_CONTIGUOUS')
    STATUS_VOLMGR_EXTENT_NOT_IN_PUBLIC_REGION: Final[ThirtyTwoBits] = cached_ntstatus(0xC038001C, 'STATUS_VOLMGR_EXTENT_NOT_IN_PUBLIC_REGION')
    STATUS_VOLMGR_EXTENT_NOT_SECTOR_ALIGNED: Final[ThirtyTwoBits] = cached_ntstatus(0xC038001D, 'STATUS_VOLMGR_EXTENT_NOT_SECTOR_ALIGNED')
    STATUS_VOLMGR_EXTENT_OVERLAPS_EBR_PARTITION: Final[ThirtyTwoBits] = cached_ntstatus(0xC038001E, 'STATUS_VOLMGR_EXTENT_OVERLAPS_EBR_PARTITION')
    STATUS_VOLMGR_EXTENT_VOLUME_LENGTHS_DO_NOT_MATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC038001F, 'STATUS_VOLMGR_EXTENT_VOLUME_LENGTHS_DO_NOT_MATCH')
    STATUS_VOLMGR_FAULT_TOLERANT_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380020, 'STATUS_VOLMGR_FAULT_TOLERANT_NOT_SUPPORTED')
    STATUS_VOLMGR_INTERLEAVE_LENGTH_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380021, 'STATUS_VOLMGR_INTERLEAVE_LENGTH_INVALID')
    STATUS_VOLMGR_MAXIMUM_REGISTERED_USERS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380022, 'STATUS_VOLMGR_MAXIMUM_REGISTERED_USERS')
    STATUS_VOLMGR_MEMBER_IN_SYNC: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380023, 'STATUS_VOLMGR_MEMBER_IN_SYNC')
    STATUS_VOLMGR_MEMBER_INDEX_DUPLICATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380024, 'STATUS_VOLMGR_MEMBER_INDEX_DUPLICATE')
    STATUS_VOLMGR_MEMBER_INDEX_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380025, 'STATUS_VOLMGR_MEMBER_INDEX_INVALID')
    STATUS_VOLMGR_MEMBER_MISSING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380026, 'STATUS_VOLMGR_MEMBER_MISSING')
    STATUS_VOLMGR_MEMBER_NOT_DETACHED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380027, 'STATUS_VOLMGR_MEMBER_NOT_DETACHED')
    STATUS_VOLMGR_MEMBER_REGENERATING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380028, 'STATUS_VOLMGR_MEMBER_REGENERATING')
    STATUS_VOLMGR_ALL_DISKS_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380029, 'STATUS_VOLMGR_ALL_DISKS_FAILED')
    STATUS_VOLMGR_NO_REGISTERED_USERS: Final[ThirtyTwoBits] = cached_ntstatus(0xC038002A, 'STATUS_VOLMGR_NO_REGISTERED_USERS')
    STATUS_VOLMGR_NO_SUCH_USER: Final[ThirtyTwoBits] = cached_ntstatus(0xC038002B, 'STATUS_VOLMGR_NO_SUCH_USER')
    STATUS_VOLMGR_NOTIFICATION_RESET: Final[ThirtyTwoBits] = cached_ntstatus(0xC038002C, 'STATUS_VOLMGR_NOTIFICATION_RESET')
    STATUS_VOLMGR_NUMBER_OF_MEMBERS_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC038002D, 'STATUS_VOLMGR_NUMBER_OF_MEMBERS_INVALID')
    STATUS_VOLMGR_NUMBER_OF_PLEXES_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC038002E, 'STATUS_VOLMGR_NUMBER_OF_PLEXES_INVALID')
    STATUS_VOLMGR_PACK_DUPLICATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC038002F, 'STATUS_VOLMGR_PACK_DUPLICATE')
    STATUS_VOLMGR_PACK_ID_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380030, 'STATUS_VOLMGR_PACK_ID_INVALID')
    STATUS_VOLMGR_PACK_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380031, 'STATUS_VOLMGR_PACK_INVALID')
    STATUS_VOLMGR_PACK_NAME_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380032, 'STATUS_VOLMGR_PACK_NAME_INVALID')
    STATUS_VOLMGR_PACK_OFFLINE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380033, 'STATUS_VOLMGR_PACK_OFFLINE')
    STATUS_VOLMGR_PACK_HAS_QUORUM: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380034, 'STATUS_VOLMGR_PACK_HAS_QUORUM')
    STATUS_VOLMGR_PACK_WITHOUT_QUORUM: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380035, 'STATUS_VOLMGR_PACK_WITHOUT_QUORUM')
    STATUS_VOLMGR_PARTITION_STYLE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380036, 'STATUS_VOLMGR_PARTITION_STYLE_INVALID')
    STATUS_VOLMGR_PARTITION_UPDATE_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380037, 'STATUS_VOLMGR_PARTITION_UPDATE_FAILED')
    STATUS_VOLMGR_PLEX_IN_SYNC: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380038, 'STATUS_VOLMGR_PLEX_IN_SYNC')
    STATUS_VOLMGR_PLEX_INDEX_DUPLICATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380039, 'STATUS_VOLMGR_PLEX_INDEX_DUPLICATE')
    STATUS_VOLMGR_PLEX_INDEX_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC038003A, 'STATUS_VOLMGR_PLEX_INDEX_INVALID')
    STATUS_VOLMGR_PLEX_LAST_ACTIVE: Final[ThirtyTwoBits] = cached_ntstatus(0xC038003B, 'STATUS_VOLMGR_PLEX_LAST_ACTIVE')
    STATUS_VOLMGR_PLEX_MISSING: Final[ThirtyTwoBits] = cached_ntstatus(0xC038003C, 'STATUS_VOLMGR_PLEX_MISSING')
    STATUS_VOLMGR_PLEX_REGENERATING: Final[ThirtyTwoBits] = cached_ntstatus(0xC038003D, 'STATUS_VOLMGR_PLEX_REGENERATING')
    STATUS_VOLMGR_PLEX_TYPE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC038003E, 'STATUS_VOLMGR_PLEX_TYPE_INVALID')
    STATUS_VOLMGR_PLEX_NOT_RAID5: Final[ThirtyTwoBits] = cached_ntstatus(0xC038003F, 'STATUS_VOLMGR_PLEX_NOT_RAID5')
    STATUS_VOLMGR_PLEX_NOT_SIMPLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380040, 'STATUS_VOLMGR_PLEX_NOT_SIMPLE')
    STATUS_VOLMGR_STRUCTURE_SIZE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380041, 'STATUS_VOLMGR_STRUCTURE_SIZE_INVALID')
    STATUS_VOLMGR_TOO_MANY_NOTIFICATION_REQUESTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380042, 'STATUS_VOLMGR_TOO_MANY_NOTIFICATION_REQUESTS')
    STATUS_VOLMGR_TRANSACTION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380043, 'STATUS_VOLMGR_TRANSACTION_IN_PROGRESS')
    STATUS_VOLMGR_UNEXPECTED_DISK_LAYOUT_CHANGE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380044, 'STATUS_VOLMGR_UNEXPECTED_DISK_LAYOUT_CHANGE')
    STATUS_VOLMGR_VOLUME_CONTAINS_MISSING_DISK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380045, 'STATUS_VOLMGR_VOLUME_CONTAINS_MISSING_DISK')
    STATUS_VOLMGR_VOLUME_ID_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380046, 'STATUS_VOLMGR_VOLUME_ID_INVALID')
    STATUS_VOLMGR_VOLUME_LENGTH_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380047, 'STATUS_VOLMGR_VOLUME_LENGTH_INVALID')
    STATUS_VOLMGR_VOLUME_LENGTH_NOT_SECTOR_SIZE_MULTIPLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380048, 'STATUS_VOLMGR_VOLUME_LENGTH_NOT_SECTOR_SIZE_MULTIPLE')
    STATUS_VOLMGR_VOLUME_NOT_MIRRORED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380049, 'STATUS_VOLMGR_VOLUME_NOT_MIRRORED')
    STATUS_VOLMGR_VOLUME_NOT_RETAINED: Final[ThirtyTwoBits] = cached_ntstatus(0xC038004A, 'STATUS_VOLMGR_VOLUME_NOT_RETAINED')
    STATUS_VOLMGR_VOLUME_OFFLINE: Final[ThirtyTwoBits] = cached_ntstatus(0xC038004B, 'STATUS_VOLMGR_VOLUME_OFFLINE')
    STATUS_VOLMGR_VOLUME_RETAINED: Final[ThirtyTwoBits] = cached_ntstatus(0xC038004C, 'STATUS_VOLMGR_VOLUME_RETAINED')
    STATUS_VOLMGR_NUMBER_OF_EXTENTS_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC038004D, 'STATUS_VOLMGR_NUMBER_OF_EXTENTS_INVALID')
    STATUS_VOLMGR_DIFFERENT_SECTOR_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC038004E, 'STATUS_VOLMGR_DIFFERENT_SECTOR_SIZE')
    STATUS_VOLMGR_BAD_BOOT_DISK: Final[ThirtyTwoBits] = cached_ntstatus(0xC038004F, 'STATUS_VOLMGR_BAD_BOOT_DISK')
    STATUS_VOLMGR_PACK_CONFIG_OFFLINE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380050, 'STATUS_VOLMGR_PACK_CONFIG_OFFLINE')
    STATUS_VOLMGR_PACK_CONFIG_ONLINE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380051, 'STATUS_VOLMGR_PACK_CONFIG_ONLINE')
    STATUS_VOLMGR_NOT_PRIMARY_PACK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380052, 'STATUS_VOLMGR_NOT_PRIMARY_PACK')
    STATUS_VOLMGR_PACK_LOG_UPDATE_FAILED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380053, 'STATUS_VOLMGR_PACK_LOG_UPDATE_FAILED')
    STATUS_VOLMGR_NUMBER_OF_DISKS_IN_PLEX_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380054, 'STATUS_VOLMGR_NUMBER_OF_DISKS_IN_PLEX_INVALID')
    STATUS_VOLMGR_NUMBER_OF_DISKS_IN_MEMBER_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380055, 'STATUS_VOLMGR_NUMBER_OF_DISKS_IN_MEMBER_INVALID')
    STATUS_VOLMGR_VOLUME_MIRRORED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380056, 'STATUS_VOLMGR_VOLUME_MIRRORED')
    STATUS_VOLMGR_PLEX_NOT_SIMPLE_SPANNED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380057, 'STATUS_VOLMGR_PLEX_NOT_SIMPLE_SPANNED')
    STATUS_VOLMGR_NO_VALID_LOG_COPIES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380058, 'STATUS_VOLMGR_NO_VALID_LOG_COPIES')
    STATUS_VOLMGR_PRIMARY_PACK_PRESENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0380059, 'STATUS_VOLMGR_PRIMARY_PACK_PRESENT')
    STATUS_VOLMGR_NUMBER_OF_DISKS_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC038005A, 'STATUS_VOLMGR_NUMBER_OF_DISKS_INVALID')
    STATUS_VOLMGR_MIRROR_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC038005B, 'STATUS_VOLMGR_MIRROR_NOT_SUPPORTED')
    STATUS_VOLMGR_RAID5_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC038005C, 'STATUS_VOLMGR_RAID5_NOT_SUPPORTED')
    STATUS_BCD_TOO_MANY_ELEMENTS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0390002, 'STATUS_BCD_TOO_MANY_ELEMENTS')
    STATUS_VHD_DRIVE_FOOTER_MISSING: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0001, 'STATUS_VHD_DRIVE_FOOTER_MISSING')
    STATUS_VHD_DRIVE_FOOTER_CHECKSUM_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0002, 'STATUS_VHD_DRIVE_FOOTER_CHECKSUM_MISMATCH')
    STATUS_VHD_DRIVE_FOOTER_CORRUPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0003, 'STATUS_VHD_DRIVE_FOOTER_CORRUPT')
    STATUS_VHD_FORMAT_UNKNOWN: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0004, 'STATUS_VHD_FORMAT_UNKNOWN')
    STATUS_VHD_FORMAT_UNSUPPORTED_VERSION: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0005, 'STATUS_VHD_FORMAT_UNSUPPORTED_VERSION')
    STATUS_VHD_SPARSE_HEADER_CHECKSUM_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0006, 'STATUS_VHD_SPARSE_HEADER_CHECKSUM_MISMATCH')
    STATUS_VHD_SPARSE_HEADER_UNSUPPORTED_VERSION: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0007, 'STATUS_VHD_SPARSE_HEADER_UNSUPPORTED_VERSION')
    STATUS_VHD_SPARSE_HEADER_CORRUPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0008, 'STATUS_VHD_SPARSE_HEADER_CORRUPT')
    STATUS_VHD_BLOCK_ALLOCATION_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0009, 'STATUS_VHD_BLOCK_ALLOCATION_FAILURE')
    STATUS_VHD_BLOCK_ALLOCATION_TABLE_CORRUPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A000A, 'STATUS_VHD_BLOCK_ALLOCATION_TABLE_CORRUPT')
    STATUS_VHD_INVALID_BLOCK_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A000B, 'STATUS_VHD_INVALID_BLOCK_SIZE')
    STATUS_VHD_BITMAP_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A000C, 'STATUS_VHD_BITMAP_MISMATCH')
    STATUS_VHD_PARENT_VHD_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A000D, 'STATUS_VHD_PARENT_VHD_NOT_FOUND')
    STATUS_VHD_CHILD_PARENT_ID_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A000E, 'STATUS_VHD_CHILD_PARENT_ID_MISMATCH')
    STATUS_VHD_CHILD_PARENT_TIMESTAMP_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A000F, 'STATUS_VHD_CHILD_PARENT_TIMESTAMP_MISMATCH')
    STATUS_VHD_METADATA_READ_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0010, 'STATUS_VHD_METADATA_READ_FAILURE')
    STATUS_VHD_METADATA_WRITE_FAILURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0011, 'STATUS_VHD_METADATA_WRITE_FAILURE')
    STATUS_VHD_INVALID_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0012, 'STATUS_VHD_INVALID_SIZE')
    STATUS_VHD_INVALID_FILE_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0013, 'STATUS_VHD_INVALID_FILE_SIZE')
    STATUS_VIRTDISK_PROVIDER_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0014, 'STATUS_VIRTDISK_PROVIDER_NOT_FOUND')
    STATUS_VIRTDISK_NOT_VIRTUAL_DISK: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0015, 'STATUS_VIRTDISK_NOT_VIRTUAL_DISK')
    STATUS_VHD_PARENT_VHD_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0016, 'STATUS_VHD_PARENT_VHD_ACCESS_DENIED')
    STATUS_VHD_CHILD_PARENT_SIZE_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0017, 'STATUS_VHD_CHILD_PARENT_SIZE_MISMATCH')
    STATUS_VHD_DIFFERENCING_CHAIN_CYCLE_DETECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0018, 'STATUS_VHD_DIFFERENCING_CHAIN_CYCLE_DETECTED')
    STATUS_VHD_DIFFERENCING_CHAIN_ERROR_IN_PARENT: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0019, 'STATUS_VHD_DIFFERENCING_CHAIN_ERROR_IN_PARENT')
    STATUS_VIRTUAL_DISK_LIMITATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A001A, 'STATUS_VIRTUAL_DISK_LIMITATION')
    STATUS_VHD_INVALID_TYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A001B, 'STATUS_VHD_INVALID_TYPE')
    STATUS_VHD_INVALID_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A001C, 'STATUS_VHD_INVALID_STATE')
    STATUS_VIRTDISK_UNSUPPORTED_DISK_SECTOR_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A001D, 'STATUS_VIRTDISK_UNSUPPORTED_DISK_SECTOR_SIZE')
    STATUS_VIRTDISK_DISK_ALREADY_OWNED: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A001E, 'STATUS_VIRTDISK_DISK_ALREADY_OWNED')
    STATUS_VIRTDISK_DISK_ONLINE_AND_WRITABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A001F, 'STATUS_VIRTDISK_DISK_ONLINE_AND_WRITABLE')
    STATUS_CTLOG_TRACKING_NOT_INITIALIZED: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0020, 'STATUS_CTLOG_TRACKING_NOT_INITIALIZED')
    STATUS_CTLOG_LOGFILE_SIZE_EXCEEDED_MAXSIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0021, 'STATUS_CTLOG_LOGFILE_SIZE_EXCEEDED_MAXSIZE')
    STATUS_CTLOG_VHD_CHANGED_OFFLINE: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0022, 'STATUS_CTLOG_VHD_CHANGED_OFFLINE')
    STATUS_CTLOG_INVALID_TRACKING_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0023, 'STATUS_CTLOG_INVALID_TRACKING_STATE')
    STATUS_CTLOG_INCONSISTENT_TRACKING_FILE: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0024, 'STATUS_CTLOG_INCONSISTENT_TRACKING_FILE')
    STATUS_VHD_METADATA_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0028, 'STATUS_VHD_METADATA_FULL')
    STATUS_VHD_INVALID_CHANGE_TRACKING_ID: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0029, 'STATUS_VHD_INVALID_CHANGE_TRACKING_ID')
    STATUS_VHD_CHANGE_TRACKING_DISABLED: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A002A, 'STATUS_VHD_CHANGE_TRACKING_DISABLED')
    STATUS_VHD_MISSING_CHANGE_TRACKING_INFORMATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0030, 'STATUS_VHD_MISSING_CHANGE_TRACKING_INFORMATION')
    STATUS_VHD_RESIZE_WOULD_TRUNCATE_DATA: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0031, 'STATUS_VHD_RESIZE_WOULD_TRUNCATE_DATA')
    STATUS_VHD_COULD_NOT_COMPUTE_MINIMUM_VIRTUAL_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0032, 'STATUS_VHD_COULD_NOT_COMPUTE_MINIMUM_VIRTUAL_SIZE')
    STATUS_VHD_ALREADY_AT_OR_BELOW_MINIMUM_VIRTUAL_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0033, 'STATUS_VHD_ALREADY_AT_OR_BELOW_MINIMUM_VIRTUAL_SIZE')
    STATUS_VHD_UNEXPECTED_ID: Final[ThirtyTwoBits] = cached_ntstatus(0xC03A0034, 'STATUS_VHD_UNEXPECTED_ID')
    STATUS_RKF_KEY_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0400001, 'STATUS_RKF_KEY_NOT_FOUND')
    STATUS_RKF_DUPLICATE_KEY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0400002, 'STATUS_RKF_DUPLICATE_KEY')
    STATUS_RKF_BLOB_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0400003, 'STATUS_RKF_BLOB_FULL')
    STATUS_RKF_STORE_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0400004, 'STATUS_RKF_STORE_FULL')
    STATUS_RKF_FILE_BLOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0400005, 'STATUS_RKF_FILE_BLOCKED')
    STATUS_RKF_ACTIVE_KEY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0400006, 'STATUS_RKF_ACTIVE_KEY')
    STATUS_RDBSS_RESTART_OPERATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0410001, 'STATUS_RDBSS_RESTART_OPERATION')
    STATUS_RDBSS_CONTINUE_OPERATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0410002, 'STATUS_RDBSS_CONTINUE_OPERATION')
    STATUS_RDBSS_POST_OPERATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0410003, 'STATUS_RDBSS_POST_OPERATION')
    STATUS_RDBSS_RETRY_LOOKUP: Final[ThirtyTwoBits] = cached_ntstatus(0xC0410004, 'STATUS_RDBSS_RETRY_LOOKUP')
    STATUS_BTH_ATT_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0420001, 'STATUS_BTH_ATT_INVALID_HANDLE')
    STATUS_BTH_ATT_READ_NOT_PERMITTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0420002, 'STATUS_BTH_ATT_READ_NOT_PERMITTED')
    STATUS_BTH_ATT_WRITE_NOT_PERMITTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0420003, 'STATUS_BTH_ATT_WRITE_NOT_PERMITTED')
    STATUS_BTH_ATT_INVALID_PDU: Final[ThirtyTwoBits] = cached_ntstatus(0xC0420004, 'STATUS_BTH_ATT_INVALID_PDU')
    STATUS_BTH_ATT_INSUFFICIENT_AUTHENTICATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0420005, 'STATUS_BTH_ATT_INSUFFICIENT_AUTHENTICATION')
    STATUS_BTH_ATT_REQUEST_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0420006, 'STATUS_BTH_ATT_REQUEST_NOT_SUPPORTED')
    STATUS_BTH_ATT_INVALID_OFFSET: Final[ThirtyTwoBits] = cached_ntstatus(0xC0420007, 'STATUS_BTH_ATT_INVALID_OFFSET')
    STATUS_BTH_ATT_INSUFFICIENT_AUTHORIZATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0420008, 'STATUS_BTH_ATT_INSUFFICIENT_AUTHORIZATION')
    STATUS_BTH_ATT_PREPARE_QUEUE_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0420009, 'STATUS_BTH_ATT_PREPARE_QUEUE_FULL')
    STATUS_BTH_ATT_ATTRIBUTE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC042000A, 'STATUS_BTH_ATT_ATTRIBUTE_NOT_FOUND')
    STATUS_BTH_ATT_ATTRIBUTE_NOT_LONG: Final[ThirtyTwoBits] = cached_ntstatus(0xC042000B, 'STATUS_BTH_ATT_ATTRIBUTE_NOT_LONG')
    STATUS_BTH_ATT_INSUFFICIENT_ENCRYPTION_KEY_SIZE: Final[ThirtyTwoBits] = cached_ntstatus(0xC042000C, 'STATUS_BTH_ATT_INSUFFICIENT_ENCRYPTION_KEY_SIZE')
    STATUS_BTH_ATT_INVALID_ATTRIBUTE_VALUE_LENGTH: Final[ThirtyTwoBits] = cached_ntstatus(0xC042000D, 'STATUS_BTH_ATT_INVALID_ATTRIBUTE_VALUE_LENGTH')
    STATUS_BTH_ATT_UNLIKELY: Final[ThirtyTwoBits] = cached_ntstatus(0xC042000E, 'STATUS_BTH_ATT_UNLIKELY')
    STATUS_BTH_ATT_INSUFFICIENT_ENCRYPTION: Final[ThirtyTwoBits] = cached_ntstatus(0xC042000F, 'STATUS_BTH_ATT_INSUFFICIENT_ENCRYPTION')
    STATUS_BTH_ATT_UNSUPPORTED_GROUP_TYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0420010, 'STATUS_BTH_ATT_UNSUPPORTED_GROUP_TYPE')
    STATUS_BTH_ATT_INSUFFICIENT_RESOURCES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0420011, 'STATUS_BTH_ATT_INSUFFICIENT_RESOURCES')
    STATUS_BTH_ATT_UNKNOWN_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0421000, 'STATUS_BTH_ATT_UNKNOWN_ERROR')
    STATUS_SECUREBOOT_ROLLBACK_DETECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0430001, 'STATUS_SECUREBOOT_ROLLBACK_DETECTED')
    STATUS_SECUREBOOT_POLICY_VIOLATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0430002, 'STATUS_SECUREBOOT_POLICY_VIOLATION')
    STATUS_SECUREBOOT_INVALID_POLICY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0430003, 'STATUS_SECUREBOOT_INVALID_POLICY')
    STATUS_SECUREBOOT_POLICY_PUBLISHER_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0430004, 'STATUS_SECUREBOOT_POLICY_PUBLISHER_NOT_FOUND')
    STATUS_SECUREBOOT_POLICY_NOT_SIGNED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0430005, 'STATUS_SECUREBOOT_POLICY_NOT_SIGNED')
    STATUS_SECUREBOOT_FILE_REPLACED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0430007, 'STATUS_SECUREBOOT_FILE_REPLACED')
    STATUS_SECUREBOOT_POLICY_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0430008, 'STATUS_SECUREBOOT_POLICY_NOT_AUTHORIZED')
    STATUS_SECUREBOOT_POLICY_UNKNOWN: Final[ThirtyTwoBits] = cached_ntstatus(0xC0430009, 'STATUS_SECUREBOOT_POLICY_UNKNOWN')
    STATUS_SECUREBOOT_POLICY_MISSING_ANTIROLLBACKVERSION: Final[ThirtyTwoBits] = cached_ntstatus(0xC043000A, 'STATUS_SECUREBOOT_POLICY_MISSING_ANTIROLLBACKVERSION')
    STATUS_SECUREBOOT_PLATFORM_ID_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC043000B, 'STATUS_SECUREBOOT_PLATFORM_ID_MISMATCH')
    STATUS_SECUREBOOT_POLICY_ROLLBACK_DETECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC043000C, 'STATUS_SECUREBOOT_POLICY_ROLLBACK_DETECTED')
    STATUS_SECUREBOOT_POLICY_UPGRADE_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC043000D, 'STATUS_SECUREBOOT_POLICY_UPGRADE_MISMATCH')
    STATUS_SECUREBOOT_REQUIRED_POLICY_FILE_MISSING: Final[ThirtyTwoBits] = cached_ntstatus(0xC043000E, 'STATUS_SECUREBOOT_REQUIRED_POLICY_FILE_MISSING')
    STATUS_SECUREBOOT_NOT_BASE_POLICY: Final[ThirtyTwoBits] = cached_ntstatus(0xC043000F, 'STATUS_SECUREBOOT_NOT_BASE_POLICY')
    STATUS_SECUREBOOT_NOT_SUPPLEMENTAL_POLICY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0430010, 'STATUS_SECUREBOOT_NOT_SUPPLEMENTAL_POLICY')
    STATUS_AUDIO_ENGINE_NODE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0440001, 'STATUS_AUDIO_ENGINE_NODE_NOT_FOUND')
    STATUS_HDAUDIO_EMPTY_CONNECTION_LIST: Final[ThirtyTwoBits] = cached_ntstatus(0xC0440002, 'STATUS_HDAUDIO_EMPTY_CONNECTION_LIST')
    STATUS_HDAUDIO_CONNECTION_LIST_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0440003, 'STATUS_HDAUDIO_CONNECTION_LIST_NOT_SUPPORTED')
    STATUS_HDAUDIO_NO_LOGICAL_DEVICES_CREATED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0440004, 'STATUS_HDAUDIO_NO_LOGICAL_DEVICES_CREATED')
    STATUS_HDAUDIO_NULL_LINKED_LIST_ENTRY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0440005, 'STATUS_HDAUDIO_NULL_LINKED_LIST_ENTRY')
    STATUS_VSM_NOT_INITIALIZED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0450000, 'STATUS_VSM_NOT_INITIALIZED')
    STATUS_VSM_DMA_PROTECTION_NOT_IN_USE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0450001, 'STATUS_VSM_DMA_PROTECTION_NOT_IN_USE')
    STATUS_IORING_REQUIRED_FLAG_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0460001, 'STATUS_IORING_REQUIRED_FLAG_NOT_SUPPORTED')
    STATUS_IORING_SUBMISSION_QUEUE_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0460002, 'STATUS_IORING_SUBMISSION_QUEUE_FULL')
    STATUS_IORING_VERSION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0460003, 'STATUS_IORING_VERSION_NOT_SUPPORTED')
    STATUS_IORING_SUBMISSION_QUEUE_TOO_BIG: Final[ThirtyTwoBits] = cached_ntstatus(0xC0460004, 'STATUS_IORING_SUBMISSION_QUEUE_TOO_BIG')
    STATUS_IORING_COMPLETION_QUEUE_TOO_BIG: Final[ThirtyTwoBits] = cached_ntstatus(0xC0460005, 'STATUS_IORING_COMPLETION_QUEUE_TOO_BIG')
    STATUS_IORING_SUBMIT_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0460006, 'STATUS_IORING_SUBMIT_IN_PROGRESS')
    STATUS_IORING_CORRUPT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0460007, 'STATUS_IORING_CORRUPT')
    STATUS_IORING_COMPLETION_QUEUE_TOO_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0460008, 'STATUS_IORING_COMPLETION_QUEUE_TOO_FULL')
    STATUS_VOLSNAP_BOOTFILE_NOT_VALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0500003, 'STATUS_VOLSNAP_BOOTFILE_NOT_VALID')
    STATUS_VOLSNAP_ACTIVATION_TIMEOUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0500004, 'STATUS_VOLSNAP_ACTIVATION_TIMEOUT')
    STATUS_VOLSNAP_NO_BYPASSIO_WITH_SNAPSHOT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0500005, 'STATUS_VOLSNAP_NO_BYPASSIO_WITH_SNAPSHOT')
    STATUS_IO_PREEMPTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0510001, 'STATUS_IO_PREEMPTED')
    STATUS_SVHDX_ERROR_STORED: Final[ThirtyTwoBits] = cached_ntstatus(0xC05C0000, 'STATUS_SVHDX_ERROR_STORED')
    STATUS_SVHDX_ERROR_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC05CFF00, 'STATUS_SVHDX_ERROR_NOT_AVAILABLE')
    STATUS_SVHDX_UNIT_ATTENTION_AVAILABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC05CFF01, 'STATUS_SVHDX_UNIT_ATTENTION_AVAILABLE')
    STATUS_SVHDX_UNIT_ATTENTION_CAPACITY_DATA_CHANGED: Final[ThirtyTwoBits] = cached_ntstatus(0xC05CFF02, 'STATUS_SVHDX_UNIT_ATTENTION_CAPACITY_DATA_CHANGED')
    STATUS_SVHDX_UNIT_ATTENTION_RESERVATIONS_PREEMPTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC05CFF03, 'STATUS_SVHDX_UNIT_ATTENTION_RESERVATIONS_PREEMPTED')
    STATUS_SVHDX_UNIT_ATTENTION_RESERVATIONS_RELEASED: Final[ThirtyTwoBits] = cached_ntstatus(0xC05CFF04, 'STATUS_SVHDX_UNIT_ATTENTION_RESERVATIONS_RELEASED')
    STATUS_SVHDX_UNIT_ATTENTION_REGISTRATIONS_PREEMPTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC05CFF05, 'STATUS_SVHDX_UNIT_ATTENTION_REGISTRATIONS_PREEMPTED')
    STATUS_SVHDX_UNIT_ATTENTION_OPERATING_DEFINITION_CHANGED: Final[ThirtyTwoBits] = cached_ntstatus(0xC05CFF06, 'STATUS_SVHDX_UNIT_ATTENTION_OPERATING_DEFINITION_CHANGED')
    STATUS_SVHDX_RESERVATION_CONFLICT: Final[ThirtyTwoBits] = cached_ntstatus(0xC05CFF07, 'STATUS_SVHDX_RESERVATION_CONFLICT')
    STATUS_SVHDX_WRONG_FILE_TYPE: Final[ThirtyTwoBits] = cached_ntstatus(0xC05CFF08, 'STATUS_SVHDX_WRONG_FILE_TYPE')
    STATUS_SVHDX_VERSION_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC05CFF09, 'STATUS_SVHDX_VERSION_MISMATCH')
    STATUS_VHD_SHARED: Final[ThirtyTwoBits] = cached_ntstatus(0xC05CFF0A, 'STATUS_VHD_SHARED')
    STATUS_SVHDX_NO_INITIATOR: Final[ThirtyTwoBits] = cached_ntstatus(0xC05CFF0B, 'STATUS_SVHDX_NO_INITIATOR')
    STATUS_VHDSET_BACKING_STORAGE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC05CFF0C, 'STATUS_VHDSET_BACKING_STORAGE_NOT_FOUND')
    STATUS_SMB_NO_PREAUTH_INTEGRITY_HASH_OVERLAP: Final[ThirtyTwoBits] = cached_ntstatus(0xC05D0000, 'STATUS_SMB_NO_PREAUTH_INTEGRITY_HASH_OVERLAP')
    STATUS_SMB_BAD_CLUSTER_DIALECT: Final[ThirtyTwoBits] = cached_ntstatus(0xC05D0001, 'STATUS_SMB_BAD_CLUSTER_DIALECT')
    STATUS_SMB_GUEST_LOGON_BLOCKED: Final[ThirtyTwoBits] = cached_ntstatus(0xC05D0002, 'STATUS_SMB_GUEST_LOGON_BLOCKED')
    STATUS_SMB_NO_SIGNING_ALGORITHM_OVERLAP: Final[ThirtyTwoBits] = cached_ntstatus(0xC05D0003, 'STATUS_SMB_NO_SIGNING_ALGORITHM_OVERLAP')
    STATUS_NETWORK_AUTHENTICATION_PROMPT_CANCELED: Final[ThirtyTwoBits] = cached_ntstatus(0xC05D0004, 'STATUS_NETWORK_AUTHENTICATION_PROMPT_CANCELED')
    STATUS_SPACES_FAULT_DOMAIN_TYPE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70001, 'STATUS_SPACES_FAULT_DOMAIN_TYPE_INVALID')
    STATUS_SPACES_RESILIENCY_TYPE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70003, 'STATUS_SPACES_RESILIENCY_TYPE_INVALID')
    STATUS_SPACES_DRIVE_SECTOR_SIZE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70004, 'STATUS_SPACES_DRIVE_SECTOR_SIZE_INVALID')
    STATUS_SPACES_DRIVE_REDUNDANCY_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70006, 'STATUS_SPACES_DRIVE_REDUNDANCY_INVALID')
    STATUS_SPACES_NUMBER_OF_DATA_COPIES_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70007, 'STATUS_SPACES_NUMBER_OF_DATA_COPIES_INVALID')
    STATUS_SPACES_INTERLEAVE_LENGTH_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70009, 'STATUS_SPACES_INTERLEAVE_LENGTH_INVALID')
    STATUS_SPACES_NUMBER_OF_COLUMNS_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E7000A, 'STATUS_SPACES_NUMBER_OF_COLUMNS_INVALID')
    STATUS_SPACES_NOT_ENOUGH_DRIVES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E7000B, 'STATUS_SPACES_NOT_ENOUGH_DRIVES')
    STATUS_SPACES_EXTENDED_ERROR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E7000C, 'STATUS_SPACES_EXTENDED_ERROR')
    STATUS_SPACES_PROVISIONING_TYPE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E7000D, 'STATUS_SPACES_PROVISIONING_TYPE_INVALID')
    STATUS_SPACES_ALLOCATION_SIZE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E7000E, 'STATUS_SPACES_ALLOCATION_SIZE_INVALID')
    STATUS_SPACES_ENCLOSURE_AWARE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E7000F, 'STATUS_SPACES_ENCLOSURE_AWARE_INVALID')
    STATUS_SPACES_WRITE_CACHE_SIZE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70010, 'STATUS_SPACES_WRITE_CACHE_SIZE_INVALID')
    STATUS_SPACES_NUMBER_OF_GROUPS_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70011, 'STATUS_SPACES_NUMBER_OF_GROUPS_INVALID')
    STATUS_SPACES_DRIVE_OPERATIONAL_STATE_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70012, 'STATUS_SPACES_DRIVE_OPERATIONAL_STATE_INVALID')
    STATUS_SPACES_UPDATE_COLUMN_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70013, 'STATUS_SPACES_UPDATE_COLUMN_STATE')
    STATUS_SPACES_MAP_REQUIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70014, 'STATUS_SPACES_MAP_REQUIRED')
    STATUS_SPACES_UNSUPPORTED_VERSION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70015, 'STATUS_SPACES_UNSUPPORTED_VERSION')
    STATUS_SPACES_CORRUPT_METADATA: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70016, 'STATUS_SPACES_CORRUPT_METADATA')
    STATUS_SPACES_DRT_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70017, 'STATUS_SPACES_DRT_FULL')
    STATUS_SPACES_INCONSISTENCY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70018, 'STATUS_SPACES_INCONSISTENCY')
    STATUS_SPACES_LOG_NOT_READY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70019, 'STATUS_SPACES_LOG_NOT_READY')
    STATUS_SPACES_NO_REDUNDANCY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E7001A, 'STATUS_SPACES_NO_REDUNDANCY')
    STATUS_SPACES_DRIVE_NOT_READY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E7001B, 'STATUS_SPACES_DRIVE_NOT_READY')
    STATUS_SPACES_DRIVE_SPLIT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E7001C, 'STATUS_SPACES_DRIVE_SPLIT')
    STATUS_SPACES_DRIVE_LOST_DATA: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E7001D, 'STATUS_SPACES_DRIVE_LOST_DATA')
    STATUS_SPACES_ENTRY_INCOMPLETE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E7001E, 'STATUS_SPACES_ENTRY_INCOMPLETE')
    STATUS_SPACES_ENTRY_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E7001F, 'STATUS_SPACES_ENTRY_INVALID')
    STATUS_SPACES_MARK_DIRTY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70020, 'STATUS_SPACES_MARK_DIRTY')
    STATUS_SPACES_PD_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70021, 'STATUS_SPACES_PD_NOT_FOUND')
    STATUS_SPACES_PD_LENGTH_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70022, 'STATUS_SPACES_PD_LENGTH_MISMATCH')
    STATUS_SPACES_PD_UNSUPPORTED_VERSION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70023, 'STATUS_SPACES_PD_UNSUPPORTED_VERSION')
    STATUS_SPACES_PD_INVALID_DATA: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70024, 'STATUS_SPACES_PD_INVALID_DATA')
    STATUS_SPACES_FLUSH_METADATA: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70025, 'STATUS_SPACES_FLUSH_METADATA')
    STATUS_SPACES_CACHE_FULL: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70026, 'STATUS_SPACES_CACHE_FULL')
    STATUS_SPACES_REPAIR_IN_PROGRESS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E70027, 'STATUS_SPACES_REPAIR_IN_PROGRESS')
    STATUS_SECCORE_INVALID_COMMAND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E80000, 'STATUS_SECCORE_INVALID_COMMAND')
    STATUS_SYSTEM_INTEGRITY_ROLLBACK_DETECTED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E90001, 'STATUS_SYSTEM_INTEGRITY_ROLLBACK_DETECTED')
    STATUS_SYSTEM_INTEGRITY_POLICY_VIOLATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E90002, 'STATUS_SYSTEM_INTEGRITY_POLICY_VIOLATION')
    STATUS_SYSTEM_INTEGRITY_INVALID_POLICY: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E90003, 'STATUS_SYSTEM_INTEGRITY_INVALID_POLICY')
    STATUS_SYSTEM_INTEGRITY_POLICY_NOT_SIGNED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E90004, 'STATUS_SYSTEM_INTEGRITY_POLICY_NOT_SIGNED')
    STATUS_SYSTEM_INTEGRITY_TOO_MANY_POLICIES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E90005, 'STATUS_SYSTEM_INTEGRITY_TOO_MANY_POLICIES')
    STATUS_SYSTEM_INTEGRITY_SUPPLEMENTAL_POLICY_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E90006, 'STATUS_SYSTEM_INTEGRITY_SUPPLEMENTAL_POLICY_NOT_AUTHORIZED')
    STATUS_SYSTEM_INTEGRITY_REPUTATION_MALICIOUS: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E90007, 'STATUS_SYSTEM_INTEGRITY_REPUTATION_MALICIOUS')
    STATUS_SYSTEM_INTEGRITY_REPUTATION_PUA: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E90008, 'STATUS_SYSTEM_INTEGRITY_REPUTATION_PUA')
    STATUS_SYSTEM_INTEGRITY_REPUTATION_DANGEROUS_EXT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E90009, 'STATUS_SYSTEM_INTEGRITY_REPUTATION_DANGEROUS_EXT')
    STATUS_SYSTEM_INTEGRITY_REPUTATION_OFFLINE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E9000A, 'STATUS_SYSTEM_INTEGRITY_REPUTATION_OFFLINE')
    STATUS_SYSTEM_INTEGRITY_REPUTATION_UNFRIENDLY_FILE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E9000B, 'STATUS_SYSTEM_INTEGRITY_REPUTATION_UNFRIENDLY_FILE')
    STATUS_SYSTEM_INTEGRITY_REPUTATION_UNATTAINABLE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E9000C, 'STATUS_SYSTEM_INTEGRITY_REPUTATION_UNATTAINABLE')
    STATUS_SYSTEM_INTEGRITY_REPUTATION_EXPLICIT_DENY_FILE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0E9000D, 'STATUS_SYSTEM_INTEGRITY_REPUTATION_EXPLICIT_DENY_FILE')
    STATUS_NO_APPLICABLE_APP_LICENSES_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EA0001, 'STATUS_NO_APPLICABLE_APP_LICENSES_FOUND')
    STATUS_CLIP_LICENSE_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EA0002, 'STATUS_CLIP_LICENSE_NOT_FOUND')
    STATUS_CLIP_DEVICE_LICENSE_MISSING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EA0003, 'STATUS_CLIP_DEVICE_LICENSE_MISSING')
    STATUS_CLIP_LICENSE_INVALID_SIGNATURE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EA0004, 'STATUS_CLIP_LICENSE_INVALID_SIGNATURE')
    STATUS_CLIP_KEYHOLDER_LICENSE_MISSING_OR_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EA0005, 'STATUS_CLIP_KEYHOLDER_LICENSE_MISSING_OR_INVALID')
    STATUS_CLIP_LICENSE_EXPIRED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EA0006, 'STATUS_CLIP_LICENSE_EXPIRED')
    STATUS_CLIP_LICENSE_SIGNED_BY_UNKNOWN_SOURCE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EA0007, 'STATUS_CLIP_LICENSE_SIGNED_BY_UNKNOWN_SOURCE')
    STATUS_CLIP_LICENSE_NOT_SIGNED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EA0008, 'STATUS_CLIP_LICENSE_NOT_SIGNED')
    STATUS_CLIP_LICENSE_HARDWARE_ID_OUT_OF_TOLERANCE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EA0009, 'STATUS_CLIP_LICENSE_HARDWARE_ID_OUT_OF_TOLERANCE')
    STATUS_CLIP_LICENSE_DEVICE_ID_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EA000A, 'STATUS_CLIP_LICENSE_DEVICE_ID_MISMATCH')
    STATUS_PLATFORM_MANIFEST_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EB0001, 'STATUS_PLATFORM_MANIFEST_NOT_AUTHORIZED')
    STATUS_PLATFORM_MANIFEST_INVALID: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EB0002, 'STATUS_PLATFORM_MANIFEST_INVALID')
    STATUS_PLATFORM_MANIFEST_FILE_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EB0003, 'STATUS_PLATFORM_MANIFEST_FILE_NOT_AUTHORIZED')
    STATUS_PLATFORM_MANIFEST_CATALOG_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EB0004, 'STATUS_PLATFORM_MANIFEST_CATALOG_NOT_AUTHORIZED')
    STATUS_PLATFORM_MANIFEST_BINARY_ID_NOT_FOUND: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EB0005, 'STATUS_PLATFORM_MANIFEST_BINARY_ID_NOT_FOUND')
    STATUS_PLATFORM_MANIFEST_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EB0006, 'STATUS_PLATFORM_MANIFEST_NOT_ACTIVE')
    STATUS_PLATFORM_MANIFEST_NOT_SIGNED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EB0007, 'STATUS_PLATFORM_MANIFEST_NOT_SIGNED')
    STATUS_APPEXEC_CONDITION_NOT_SATISFIED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EC0000, 'STATUS_APPEXEC_CONDITION_NOT_SATISFIED')
    STATUS_APPEXEC_HANDLE_INVALIDATED: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EC0001, 'STATUS_APPEXEC_HANDLE_INVALIDATED')
    STATUS_APPEXEC_INVALID_HOST_GENERATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EC0002, 'STATUS_APPEXEC_INVALID_HOST_GENERATION')
    STATUS_APPEXEC_UNEXPECTED_PROCESS_REGISTRATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EC0003, 'STATUS_APPEXEC_UNEXPECTED_PROCESS_REGISTRATION')
    STATUS_APPEXEC_INVALID_HOST_STATE: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EC0004, 'STATUS_APPEXEC_INVALID_HOST_STATE')
    STATUS_APPEXEC_NO_DONOR: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EC0005, 'STATUS_APPEXEC_NO_DONOR')
    STATUS_APPEXEC_HOST_ID_MISMATCH: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EC0006, 'STATUS_APPEXEC_HOST_ID_MISMATCH')
    STATUS_APPEXEC_UNKNOWN_USER: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EC0007, 'STATUS_APPEXEC_UNKNOWN_USER')
    STATUS_APPEXEC_APP_COMPAT_BLOCK: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EC0008, 'STATUS_APPEXEC_APP_COMPAT_BLOCK')
    STATUS_APPEXEC_CALLER_WAIT_TIMEOUT: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EC0009, 'STATUS_APPEXEC_CALLER_WAIT_TIMEOUT')
    STATUS_APPEXEC_CALLER_WAIT_TIMEOUT_TERMINATION: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EC000A, 'STATUS_APPEXEC_CALLER_WAIT_TIMEOUT_TERMINATION')
    STATUS_APPEXEC_CALLER_WAIT_TIMEOUT_LICENSING: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EC000B, 'STATUS_APPEXEC_CALLER_WAIT_TIMEOUT_LICENSING')
    STATUS_APPEXEC_CALLER_WAIT_TIMEOUT_RESOURCES: Final[ThirtyTwoBits] = cached_ntstatus(0xC0EC000C, 'STATUS_APPEXEC_CALLER_WAIT_TIMEOUT_RESOURCES')
