__all__ = [
    'Win32Error',
]

from enum import IntEnum
from typing import Final, Optional

from ntstatus._thirtytwobits import ThirtyTwoBits

WIN32ERROR_LOOKUP: dict[int, ThirtyTwoBits] = dict()

def cached_win32error(code: int, name: str) -> ThirtyTwoBits:
    if code in WIN32ERROR_LOOKUP:
        return WIN32ERROR_LOOKUP[code]

    bits = ThirtyTwoBits(code, name)
    WIN32ERROR_LOOKUP[code] = bits
    return bits

class Win32Error:
    """
    This class defines all named [WinAPI "Win32 Error Code" values](https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-erref/18d8fbe8-a967-4f1c-ae50-99ca8e491d2d) as attributes. While the name is similar, it's different from both `WindowsError` (an alias to `OSError`) and `ctypes.WinError` (a factory function for creating `OSError`s on Windows).

    Constants are represented using the `ntstatus.ThirtyTwoBits` class which can be equality-compared to integers; the comparision will return True for both the signed and unsigned interpretation of the underlying 32 bits. (Note that Win32 Error Codes are always in the 16-bit 0x000-0xFFFF range, so they can never be negative, but they're represented by `ntstatus.ThirtyTwoBits` for consistency.)

    Static methods:
    - `Win32Error.decode(code)`: Looks up an Win32 Error Code constant by its numeric value and returns it as an `ntstatus.ThirtyTwoBits` object, with its name set to the the name of the constant. Returns `None` if argument is not a known Win32 Error Code constant.
    - `Win32Error.decode_name(code, default='')`: Returns the name of the Win32 Error Code constant associated with the first argument if there is such a value. Otherwise, returns the second argument.
    - `Win32Error.make(code)`: Looks up an Win32 Error Code constant by its numeric value and returns it as an `ntstatus.ThirtyTwoBits` object, with its name set to the the name of the constant. Raises `ValueError` if argument is not a known Win32 Error Code constant.

    The complete list of Win32 Error Codes can be found [here](https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-erref/18d8fbe8-a967-4f1c-ae50-99ca8e491d2d) and [here](https://learn.microsoft.com/en-us/windows/win32/debug/system-error-codes). `Win32Error` encodes all of them defined in `winerror.h`.
    """

    @staticmethod
    def decode(code: int) -> Optional[ThirtyTwoBits]:
        """
        Looks up an Win32 Error Code constant by its numeric value and returns it as an `ntstatus.ThirtyTwoBits` object, with its name set to the the name of the constant. Returns `None` if argument is not a known Win32 Error Code constant.
        """

        if not isinstance(code, int):
            raise TypeError(f'Not a valid int value: {repr(code)}')

        return WIN32ERROR_LOOKUP.get(code, None)

    @staticmethod
    def decode_name(code: int, default: str = '') -> str:
        """
        Returns the name of the Win32 Error Code constant associated with the first argument if there is such a value. Otherwise, returns the second argument.
        """

        err = Win32Error.decode(code)
        if err is not None:
            return err.name
        else:
            return default

    @staticmethod
    def make(code: int) -> ThirtyTwoBits:
        """
        Looks up an Win32 Error Code constant by its numeric value and returns it as an `ntstatus.ThirtyTwoBits` object, with its name set to the the name of the constant. Raises `ValueError` if argument is not a known Win32 Error Code constant.
        """

        err = Win32Error.decode(code)
        if err is not None:
            return err
        else:
            raise ValueError(f'Unknown Win32 Error Code: {code}')

    # AUTOGENERATED FROM HERE, DO NOT TOUCH

    ERROR_SUCCESS: Final[ThirtyTwoBits] = cached_win32error(0x0000, 'ERROR_SUCCESS') # WinError 0 (0x0)
    ERROR_INVALID_FUNCTION: Final[ThirtyTwoBits] = cached_win32error(0x0001, 'ERROR_INVALID_FUNCTION') # WinError 1 (0x1)
    ERROR_FILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x0002, 'ERROR_FILE_NOT_FOUND') # WinError 2 (0x2)
    ERROR_PATH_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x0003, 'ERROR_PATH_NOT_FOUND') # WinError 3 (0x3)
    ERROR_TOO_MANY_OPEN_FILES: Final[ThirtyTwoBits] = cached_win32error(0x0004, 'ERROR_TOO_MANY_OPEN_FILES') # WinError 4 (0x4)
    ERROR_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_win32error(0x0005, 'ERROR_ACCESS_DENIED') # WinError 5 (0x5)
    ERROR_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_win32error(0x0006, 'ERROR_INVALID_HANDLE') # WinError 6 (0x6)
    ERROR_ARENA_TRASHED: Final[ThirtyTwoBits] = cached_win32error(0x0007, 'ERROR_ARENA_TRASHED') # WinError 7 (0x7)
    ERROR_NOT_ENOUGH_MEMORY: Final[ThirtyTwoBits] = cached_win32error(0x0008, 'ERROR_NOT_ENOUGH_MEMORY') # WinError 8 (0x8)
    ERROR_INVALID_BLOCK: Final[ThirtyTwoBits] = cached_win32error(0x0009, 'ERROR_INVALID_BLOCK') # WinError 9 (0x9)
    ERROR_BAD_ENVIRONMENT: Final[ThirtyTwoBits] = cached_win32error(0x000A, 'ERROR_BAD_ENVIRONMENT') # WinError 10 (0xA)
    ERROR_BAD_FORMAT: Final[ThirtyTwoBits] = cached_win32error(0x000B, 'ERROR_BAD_FORMAT') # WinError 11 (0xB)
    ERROR_INVALID_ACCESS: Final[ThirtyTwoBits] = cached_win32error(0x000C, 'ERROR_INVALID_ACCESS') # WinError 12 (0xC)
    ERROR_INVALID_DATA: Final[ThirtyTwoBits] = cached_win32error(0x000D, 'ERROR_INVALID_DATA') # WinError 13 (0xD)
    ERROR_OUTOFMEMORY: Final[ThirtyTwoBits] = cached_win32error(0x000E, 'ERROR_OUTOFMEMORY') # WinError 14 (0xE)
    ERROR_INVALID_DRIVE: Final[ThirtyTwoBits] = cached_win32error(0x000F, 'ERROR_INVALID_DRIVE') # WinError 15 (0xF)
    ERROR_CURRENT_DIRECTORY: Final[ThirtyTwoBits] = cached_win32error(0x0010, 'ERROR_CURRENT_DIRECTORY') # WinError 16 (0x10)
    ERROR_NOT_SAME_DEVICE: Final[ThirtyTwoBits] = cached_win32error(0x0011, 'ERROR_NOT_SAME_DEVICE') # WinError 17 (0x11)
    ERROR_NO_MORE_FILES: Final[ThirtyTwoBits] = cached_win32error(0x0012, 'ERROR_NO_MORE_FILES') # WinError 18 (0x12)
    ERROR_WRITE_PROTECT: Final[ThirtyTwoBits] = cached_win32error(0x0013, 'ERROR_WRITE_PROTECT') # WinError 19 (0x13)
    ERROR_BAD_UNIT: Final[ThirtyTwoBits] = cached_win32error(0x0014, 'ERROR_BAD_UNIT') # WinError 20 (0x14)
    ERROR_NOT_READY: Final[ThirtyTwoBits] = cached_win32error(0x0015, 'ERROR_NOT_READY') # WinError 21 (0x15)
    ERROR_BAD_COMMAND: Final[ThirtyTwoBits] = cached_win32error(0x0016, 'ERROR_BAD_COMMAND') # WinError 22 (0x16)
    ERROR_CRC: Final[ThirtyTwoBits] = cached_win32error(0x0017, 'ERROR_CRC') # WinError 23 (0x17)
    ERROR_BAD_LENGTH: Final[ThirtyTwoBits] = cached_win32error(0x0018, 'ERROR_BAD_LENGTH') # WinError 24 (0x18)
    ERROR_SEEK: Final[ThirtyTwoBits] = cached_win32error(0x0019, 'ERROR_SEEK') # WinError 25 (0x19)
    ERROR_NOT_DOS_DISK: Final[ThirtyTwoBits] = cached_win32error(0x001A, 'ERROR_NOT_DOS_DISK') # WinError 26 (0x1A)
    ERROR_SECTOR_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x001B, 'ERROR_SECTOR_NOT_FOUND') # WinError 27 (0x1B)
    ERROR_OUT_OF_PAPER: Final[ThirtyTwoBits] = cached_win32error(0x001C, 'ERROR_OUT_OF_PAPER') # WinError 28 (0x1C)
    ERROR_WRITE_FAULT: Final[ThirtyTwoBits] = cached_win32error(0x001D, 'ERROR_WRITE_FAULT') # WinError 29 (0x1D)
    ERROR_READ_FAULT: Final[ThirtyTwoBits] = cached_win32error(0x001E, 'ERROR_READ_FAULT') # WinError 30 (0x1E)
    ERROR_GEN_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x001F, 'ERROR_GEN_FAILURE') # WinError 31 (0x1F)
    ERROR_SHARING_VIOLATION: Final[ThirtyTwoBits] = cached_win32error(0x0020, 'ERROR_SHARING_VIOLATION') # WinError 32 (0x20)
    ERROR_LOCK_VIOLATION: Final[ThirtyTwoBits] = cached_win32error(0x0021, 'ERROR_LOCK_VIOLATION') # WinError 33 (0x21)
    ERROR_WRONG_DISK: Final[ThirtyTwoBits] = cached_win32error(0x0022, 'ERROR_WRONG_DISK') # WinError 34 (0x22)
    ERROR_SHARING_BUFFER_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x0024, 'ERROR_SHARING_BUFFER_EXCEEDED') # WinError 36 (0x24)
    ERROR_HANDLE_EOF: Final[ThirtyTwoBits] = cached_win32error(0x0026, 'ERROR_HANDLE_EOF') # WinError 38 (0x26)
    ERROR_HANDLE_DISK_FULL: Final[ThirtyTwoBits] = cached_win32error(0x0027, 'ERROR_HANDLE_DISK_FULL') # WinError 39 (0x27)
    ERROR_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x0032, 'ERROR_NOT_SUPPORTED') # WinError 50 (0x32)
    ERROR_REM_NOT_LIST: Final[ThirtyTwoBits] = cached_win32error(0x0033, 'ERROR_REM_NOT_LIST') # WinError 51 (0x33)
    ERROR_DUP_NAME: Final[ThirtyTwoBits] = cached_win32error(0x0034, 'ERROR_DUP_NAME') # WinError 52 (0x34)
    ERROR_BAD_NETPATH: Final[ThirtyTwoBits] = cached_win32error(0x0035, 'ERROR_BAD_NETPATH') # WinError 53 (0x35)
    ERROR_NETWORK_BUSY: Final[ThirtyTwoBits] = cached_win32error(0x0036, 'ERROR_NETWORK_BUSY') # WinError 54 (0x36)
    ERROR_DEV_NOT_EXIST: Final[ThirtyTwoBits] = cached_win32error(0x0037, 'ERROR_DEV_NOT_EXIST') # WinError 55 (0x37)
    ERROR_TOO_MANY_CMDS: Final[ThirtyTwoBits] = cached_win32error(0x0038, 'ERROR_TOO_MANY_CMDS') # WinError 56 (0x38)
    ERROR_ADAP_HDW_ERR: Final[ThirtyTwoBits] = cached_win32error(0x0039, 'ERROR_ADAP_HDW_ERR') # WinError 57 (0x39)
    ERROR_BAD_NET_RESP: Final[ThirtyTwoBits] = cached_win32error(0x003A, 'ERROR_BAD_NET_RESP') # WinError 58 (0x3A)
    ERROR_UNEXP_NET_ERR: Final[ThirtyTwoBits] = cached_win32error(0x003B, 'ERROR_UNEXP_NET_ERR') # WinError 59 (0x3B)
    ERROR_BAD_REM_ADAP: Final[ThirtyTwoBits] = cached_win32error(0x003C, 'ERROR_BAD_REM_ADAP') # WinError 60 (0x3C)
    ERROR_PRINTQ_FULL: Final[ThirtyTwoBits] = cached_win32error(0x003D, 'ERROR_PRINTQ_FULL') # WinError 61 (0x3D)
    ERROR_NO_SPOOL_SPACE: Final[ThirtyTwoBits] = cached_win32error(0x003E, 'ERROR_NO_SPOOL_SPACE') # WinError 62 (0x3E)
    ERROR_PRINT_CANCELLED: Final[ThirtyTwoBits] = cached_win32error(0x003F, 'ERROR_PRINT_CANCELLED') # WinError 63 (0x3F)
    ERROR_NETNAME_DELETED: Final[ThirtyTwoBits] = cached_win32error(0x0040, 'ERROR_NETNAME_DELETED') # WinError 64 (0x40)
    ERROR_NETWORK_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_win32error(0x0041, 'ERROR_NETWORK_ACCESS_DENIED') # WinError 65 (0x41)
    ERROR_BAD_DEV_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x0042, 'ERROR_BAD_DEV_TYPE') # WinError 66 (0x42)
    ERROR_BAD_NET_NAME: Final[ThirtyTwoBits] = cached_win32error(0x0043, 'ERROR_BAD_NET_NAME') # WinError 67 (0x43)
    ERROR_TOO_MANY_NAMES: Final[ThirtyTwoBits] = cached_win32error(0x0044, 'ERROR_TOO_MANY_NAMES') # WinError 68 (0x44)
    ERROR_TOO_MANY_SESS: Final[ThirtyTwoBits] = cached_win32error(0x0045, 'ERROR_TOO_MANY_SESS') # WinError 69 (0x45)
    ERROR_SHARING_PAUSED: Final[ThirtyTwoBits] = cached_win32error(0x0046, 'ERROR_SHARING_PAUSED') # WinError 70 (0x46)
    ERROR_REQ_NOT_ACCEP: Final[ThirtyTwoBits] = cached_win32error(0x0047, 'ERROR_REQ_NOT_ACCEP') # WinError 71 (0x47)
    ERROR_REDIR_PAUSED: Final[ThirtyTwoBits] = cached_win32error(0x0048, 'ERROR_REDIR_PAUSED') # WinError 72 (0x48)
    ERROR_FILE_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x0050, 'ERROR_FILE_EXISTS') # WinError 80 (0x50)
    ERROR_CANNOT_MAKE: Final[ThirtyTwoBits] = cached_win32error(0x0052, 'ERROR_CANNOT_MAKE') # WinError 82 (0x52)
    ERROR_FAIL_I24: Final[ThirtyTwoBits] = cached_win32error(0x0053, 'ERROR_FAIL_I24') # WinError 83 (0x53)
    ERROR_OUT_OF_STRUCTURES: Final[ThirtyTwoBits] = cached_win32error(0x0054, 'ERROR_OUT_OF_STRUCTURES') # WinError 84 (0x54)
    ERROR_ALREADY_ASSIGNED: Final[ThirtyTwoBits] = cached_win32error(0x0055, 'ERROR_ALREADY_ASSIGNED') # WinError 85 (0x55)
    ERROR_INVALID_PASSWORD: Final[ThirtyTwoBits] = cached_win32error(0x0056, 'ERROR_INVALID_PASSWORD') # WinError 86 (0x56)
    ERROR_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_win32error(0x0057, 'ERROR_INVALID_PARAMETER') # WinError 87 (0x57)
    ERROR_NET_WRITE_FAULT: Final[ThirtyTwoBits] = cached_win32error(0x0058, 'ERROR_NET_WRITE_FAULT') # WinError 88 (0x58)
    ERROR_NO_PROC_SLOTS: Final[ThirtyTwoBits] = cached_win32error(0x0059, 'ERROR_NO_PROC_SLOTS') # WinError 89 (0x59)
    ERROR_TOO_MANY_SEMAPHORES: Final[ThirtyTwoBits] = cached_win32error(0x0064, 'ERROR_TOO_MANY_SEMAPHORES') # WinError 100 (0x64)
    ERROR_EXCL_SEM_ALREADY_OWNED: Final[ThirtyTwoBits] = cached_win32error(0x0065, 'ERROR_EXCL_SEM_ALREADY_OWNED') # WinError 101 (0x65)
    ERROR_SEM_IS_SET: Final[ThirtyTwoBits] = cached_win32error(0x0066, 'ERROR_SEM_IS_SET') # WinError 102 (0x66)
    ERROR_TOO_MANY_SEM_REQUESTS: Final[ThirtyTwoBits] = cached_win32error(0x0067, 'ERROR_TOO_MANY_SEM_REQUESTS') # WinError 103 (0x67)
    ERROR_INVALID_AT_INTERRUPT_TIME: Final[ThirtyTwoBits] = cached_win32error(0x0068, 'ERROR_INVALID_AT_INTERRUPT_TIME') # WinError 104 (0x68)
    ERROR_SEM_OWNER_DIED: Final[ThirtyTwoBits] = cached_win32error(0x0069, 'ERROR_SEM_OWNER_DIED') # WinError 105 (0x69)
    ERROR_SEM_USER_LIMIT: Final[ThirtyTwoBits] = cached_win32error(0x006A, 'ERROR_SEM_USER_LIMIT') # WinError 106 (0x6A)
    ERROR_DISK_CHANGE: Final[ThirtyTwoBits] = cached_win32error(0x006B, 'ERROR_DISK_CHANGE') # WinError 107 (0x6B)
    ERROR_DRIVE_LOCKED: Final[ThirtyTwoBits] = cached_win32error(0x006C, 'ERROR_DRIVE_LOCKED') # WinError 108 (0x6C)
    ERROR_BROKEN_PIPE: Final[ThirtyTwoBits] = cached_win32error(0x006D, 'ERROR_BROKEN_PIPE') # WinError 109 (0x6D)
    ERROR_OPEN_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x006E, 'ERROR_OPEN_FAILED') # WinError 110 (0x6E)
    ERROR_BUFFER_OVERFLOW: Final[ThirtyTwoBits] = cached_win32error(0x006F, 'ERROR_BUFFER_OVERFLOW') # WinError 111 (0x6F)
    ERROR_DISK_FULL: Final[ThirtyTwoBits] = cached_win32error(0x0070, 'ERROR_DISK_FULL') # WinError 112 (0x70)
    ERROR_NO_MORE_SEARCH_HANDLES: Final[ThirtyTwoBits] = cached_win32error(0x0071, 'ERROR_NO_MORE_SEARCH_HANDLES') # WinError 113 (0x71)
    ERROR_INVALID_TARGET_HANDLE: Final[ThirtyTwoBits] = cached_win32error(0x0072, 'ERROR_INVALID_TARGET_HANDLE') # WinError 114 (0x72)
    ERROR_INVALID_CATEGORY: Final[ThirtyTwoBits] = cached_win32error(0x0075, 'ERROR_INVALID_CATEGORY') # WinError 117 (0x75)
    ERROR_INVALID_VERIFY_SWITCH: Final[ThirtyTwoBits] = cached_win32error(0x0076, 'ERROR_INVALID_VERIFY_SWITCH') # WinError 118 (0x76)
    ERROR_BAD_DRIVER_LEVEL: Final[ThirtyTwoBits] = cached_win32error(0x0077, 'ERROR_BAD_DRIVER_LEVEL') # WinError 119 (0x77)
    ERROR_CALL_NOT_IMPLEMENTED: Final[ThirtyTwoBits] = cached_win32error(0x0078, 'ERROR_CALL_NOT_IMPLEMENTED') # WinError 120 (0x78)
    ERROR_SEM_TIMEOUT: Final[ThirtyTwoBits] = cached_win32error(0x0079, 'ERROR_SEM_TIMEOUT') # WinError 121 (0x79)
    ERROR_INSUFFICIENT_BUFFER: Final[ThirtyTwoBits] = cached_win32error(0x007A, 'ERROR_INSUFFICIENT_BUFFER') # WinError 122 (0x7A)
    ERROR_INVALID_NAME: Final[ThirtyTwoBits] = cached_win32error(0x007B, 'ERROR_INVALID_NAME') # WinError 123 (0x7B)
    ERROR_INVALID_LEVEL: Final[ThirtyTwoBits] = cached_win32error(0x007C, 'ERROR_INVALID_LEVEL') # WinError 124 (0x7C)
    ERROR_NO_VOLUME_LABEL: Final[ThirtyTwoBits] = cached_win32error(0x007D, 'ERROR_NO_VOLUME_LABEL') # WinError 125 (0x7D)
    ERROR_MOD_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x007E, 'ERROR_MOD_NOT_FOUND') # WinError 126 (0x7E)
    ERROR_PROC_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x007F, 'ERROR_PROC_NOT_FOUND') # WinError 127 (0x7F)
    ERROR_WAIT_NO_CHILDREN: Final[ThirtyTwoBits] = cached_win32error(0x0080, 'ERROR_WAIT_NO_CHILDREN') # WinError 128 (0x80)
    ERROR_CHILD_NOT_COMPLETE: Final[ThirtyTwoBits] = cached_win32error(0x0081, 'ERROR_CHILD_NOT_COMPLETE') # WinError 129 (0x81)
    ERROR_DIRECT_ACCESS_HANDLE: Final[ThirtyTwoBits] = cached_win32error(0x0082, 'ERROR_DIRECT_ACCESS_HANDLE') # WinError 130 (0x82)
    ERROR_NEGATIVE_SEEK: Final[ThirtyTwoBits] = cached_win32error(0x0083, 'ERROR_NEGATIVE_SEEK') # WinError 131 (0x83)
    ERROR_SEEK_ON_DEVICE: Final[ThirtyTwoBits] = cached_win32error(0x0084, 'ERROR_SEEK_ON_DEVICE') # WinError 132 (0x84)
    ERROR_IS_JOIN_TARGET: Final[ThirtyTwoBits] = cached_win32error(0x0085, 'ERROR_IS_JOIN_TARGET') # WinError 133 (0x85)
    ERROR_IS_JOINED: Final[ThirtyTwoBits] = cached_win32error(0x0086, 'ERROR_IS_JOINED') # WinError 134 (0x86)
    ERROR_IS_SUBSTED: Final[ThirtyTwoBits] = cached_win32error(0x0087, 'ERROR_IS_SUBSTED') # WinError 135 (0x87)
    ERROR_NOT_JOINED: Final[ThirtyTwoBits] = cached_win32error(0x0088, 'ERROR_NOT_JOINED') # WinError 136 (0x88)
    ERROR_NOT_SUBSTED: Final[ThirtyTwoBits] = cached_win32error(0x0089, 'ERROR_NOT_SUBSTED') # WinError 137 (0x89)
    ERROR_JOIN_TO_JOIN: Final[ThirtyTwoBits] = cached_win32error(0x008A, 'ERROR_JOIN_TO_JOIN') # WinError 138 (0x8A)
    ERROR_SUBST_TO_SUBST: Final[ThirtyTwoBits] = cached_win32error(0x008B, 'ERROR_SUBST_TO_SUBST') # WinError 139 (0x8B)
    ERROR_JOIN_TO_SUBST: Final[ThirtyTwoBits] = cached_win32error(0x008C, 'ERROR_JOIN_TO_SUBST') # WinError 140 (0x8C)
    ERROR_SUBST_TO_JOIN: Final[ThirtyTwoBits] = cached_win32error(0x008D, 'ERROR_SUBST_TO_JOIN') # WinError 141 (0x8D)
    ERROR_BUSY_DRIVE: Final[ThirtyTwoBits] = cached_win32error(0x008E, 'ERROR_BUSY_DRIVE') # WinError 142 (0x8E)
    ERROR_SAME_DRIVE: Final[ThirtyTwoBits] = cached_win32error(0x008F, 'ERROR_SAME_DRIVE') # WinError 143 (0x8F)
    ERROR_DIR_NOT_ROOT: Final[ThirtyTwoBits] = cached_win32error(0x0090, 'ERROR_DIR_NOT_ROOT') # WinError 144 (0x90)
    ERROR_DIR_NOT_EMPTY: Final[ThirtyTwoBits] = cached_win32error(0x0091, 'ERROR_DIR_NOT_EMPTY') # WinError 145 (0x91)
    ERROR_IS_SUBST_PATH: Final[ThirtyTwoBits] = cached_win32error(0x0092, 'ERROR_IS_SUBST_PATH') # WinError 146 (0x92)
    ERROR_IS_JOIN_PATH: Final[ThirtyTwoBits] = cached_win32error(0x0093, 'ERROR_IS_JOIN_PATH') # WinError 147 (0x93)
    ERROR_PATH_BUSY: Final[ThirtyTwoBits] = cached_win32error(0x0094, 'ERROR_PATH_BUSY') # WinError 148 (0x94)
    ERROR_IS_SUBST_TARGET: Final[ThirtyTwoBits] = cached_win32error(0x0095, 'ERROR_IS_SUBST_TARGET') # WinError 149 (0x95)
    ERROR_SYSTEM_TRACE: Final[ThirtyTwoBits] = cached_win32error(0x0096, 'ERROR_SYSTEM_TRACE') # WinError 150 (0x96)
    ERROR_INVALID_EVENT_COUNT: Final[ThirtyTwoBits] = cached_win32error(0x0097, 'ERROR_INVALID_EVENT_COUNT') # WinError 151 (0x97)
    ERROR_TOO_MANY_MUXWAITERS: Final[ThirtyTwoBits] = cached_win32error(0x0098, 'ERROR_TOO_MANY_MUXWAITERS') # WinError 152 (0x98)
    ERROR_INVALID_LIST_FORMAT: Final[ThirtyTwoBits] = cached_win32error(0x0099, 'ERROR_INVALID_LIST_FORMAT') # WinError 153 (0x99)
    ERROR_LABEL_TOO_LONG: Final[ThirtyTwoBits] = cached_win32error(0x009A, 'ERROR_LABEL_TOO_LONG') # WinError 154 (0x9A)
    ERROR_TOO_MANY_TCBS: Final[ThirtyTwoBits] = cached_win32error(0x009B, 'ERROR_TOO_MANY_TCBS') # WinError 155 (0x9B)
    ERROR_SIGNAL_REFUSED: Final[ThirtyTwoBits] = cached_win32error(0x009C, 'ERROR_SIGNAL_REFUSED') # WinError 156 (0x9C)
    ERROR_DISCARDED: Final[ThirtyTwoBits] = cached_win32error(0x009D, 'ERROR_DISCARDED') # WinError 157 (0x9D)
    ERROR_NOT_LOCKED: Final[ThirtyTwoBits] = cached_win32error(0x009E, 'ERROR_NOT_LOCKED') # WinError 158 (0x9E)
    ERROR_BAD_THREADID_ADDR: Final[ThirtyTwoBits] = cached_win32error(0x009F, 'ERROR_BAD_THREADID_ADDR') # WinError 159 (0x9F)
    ERROR_BAD_ARGUMENTS: Final[ThirtyTwoBits] = cached_win32error(0x00A0, 'ERROR_BAD_ARGUMENTS') # WinError 160 (0xA0)
    ERROR_BAD_PATHNAME: Final[ThirtyTwoBits] = cached_win32error(0x00A1, 'ERROR_BAD_PATHNAME') # WinError 161 (0xA1)
    ERROR_SIGNAL_PENDING: Final[ThirtyTwoBits] = cached_win32error(0x00A2, 'ERROR_SIGNAL_PENDING') # WinError 162 (0xA2)
    ERROR_MAX_THRDS_REACHED: Final[ThirtyTwoBits] = cached_win32error(0x00A4, 'ERROR_MAX_THRDS_REACHED') # WinError 164 (0xA4)
    ERROR_LOCK_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x00A7, 'ERROR_LOCK_FAILED') # WinError 167 (0xA7)
    ERROR_BUSY: Final[ThirtyTwoBits] = cached_win32error(0x00AA, 'ERROR_BUSY') # WinError 170 (0xAA)
    ERROR_DEVICE_SUPPORT_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x00AB, 'ERROR_DEVICE_SUPPORT_IN_PROGRESS') # WinError 171 (0xAB)
    ERROR_CANCEL_VIOLATION: Final[ThirtyTwoBits] = cached_win32error(0x00AD, 'ERROR_CANCEL_VIOLATION') # WinError 173 (0xAD)
    ERROR_ATOMIC_LOCKS_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x00AE, 'ERROR_ATOMIC_LOCKS_NOT_SUPPORTED') # WinError 174 (0xAE)
    ERROR_INVALID_SEGMENT_NUMBER: Final[ThirtyTwoBits] = cached_win32error(0x00B4, 'ERROR_INVALID_SEGMENT_NUMBER') # WinError 180 (0xB4)
    ERROR_INVALID_ORDINAL: Final[ThirtyTwoBits] = cached_win32error(0x00B6, 'ERROR_INVALID_ORDINAL') # WinError 182 (0xB6)
    ERROR_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x00B7, 'ERROR_ALREADY_EXISTS') # WinError 183 (0xB7)
    ERROR_INVALID_FLAG_NUMBER: Final[ThirtyTwoBits] = cached_win32error(0x00BA, 'ERROR_INVALID_FLAG_NUMBER') # WinError 186 (0xBA)
    ERROR_SEM_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x00BB, 'ERROR_SEM_NOT_FOUND') # WinError 187 (0xBB)
    ERROR_INVALID_STARTING_CODESEG: Final[ThirtyTwoBits] = cached_win32error(0x00BC, 'ERROR_INVALID_STARTING_CODESEG') # WinError 188 (0xBC)
    ERROR_INVALID_STACKSEG: Final[ThirtyTwoBits] = cached_win32error(0x00BD, 'ERROR_INVALID_STACKSEG') # WinError 189 (0xBD)
    ERROR_INVALID_MODULETYPE: Final[ThirtyTwoBits] = cached_win32error(0x00BE, 'ERROR_INVALID_MODULETYPE') # WinError 190 (0xBE)
    ERROR_INVALID_EXE_SIGNATURE: Final[ThirtyTwoBits] = cached_win32error(0x00BF, 'ERROR_INVALID_EXE_SIGNATURE') # WinError 191 (0xBF)
    ERROR_EXE_MARKED_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x00C0, 'ERROR_EXE_MARKED_INVALID') # WinError 192 (0xC0)
    ERROR_BAD_EXE_FORMAT: Final[ThirtyTwoBits] = cached_win32error(0x00C1, 'ERROR_BAD_EXE_FORMAT') # WinError 193 (0xC1)
    ERROR_ITERATED_DATA_EXCEEDS_64k: Final[ThirtyTwoBits] = cached_win32error(0x00C2, 'ERROR_ITERATED_DATA_EXCEEDS_64k') # WinError 194 (0xC2)
    ERROR_INVALID_MINALLOCSIZE: Final[ThirtyTwoBits] = cached_win32error(0x00C3, 'ERROR_INVALID_MINALLOCSIZE') # WinError 195 (0xC3)
    ERROR_DYNLINK_FROM_INVALID_RING: Final[ThirtyTwoBits] = cached_win32error(0x00C4, 'ERROR_DYNLINK_FROM_INVALID_RING') # WinError 196 (0xC4)
    ERROR_IOPL_NOT_ENABLED: Final[ThirtyTwoBits] = cached_win32error(0x00C5, 'ERROR_IOPL_NOT_ENABLED') # WinError 197 (0xC5)
    ERROR_INVALID_SEGDPL: Final[ThirtyTwoBits] = cached_win32error(0x00C6, 'ERROR_INVALID_SEGDPL') # WinError 198 (0xC6)
    ERROR_AUTODATASEG_EXCEEDS_64k: Final[ThirtyTwoBits] = cached_win32error(0x00C7, 'ERROR_AUTODATASEG_EXCEEDS_64k') # WinError 199 (0xC7)
    ERROR_RING2SEG_MUST_BE_MOVABLE: Final[ThirtyTwoBits] = cached_win32error(0x00C8, 'ERROR_RING2SEG_MUST_BE_MOVABLE') # WinError 200 (0xC8)
    ERROR_RELOC_CHAIN_XEEDS_SEGLIM: Final[ThirtyTwoBits] = cached_win32error(0x00C9, 'ERROR_RELOC_CHAIN_XEEDS_SEGLIM') # WinError 201 (0xC9)
    ERROR_INFLOOP_IN_RELOC_CHAIN: Final[ThirtyTwoBits] = cached_win32error(0x00CA, 'ERROR_INFLOOP_IN_RELOC_CHAIN') # WinError 202 (0xCA)
    ERROR_ENVVAR_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x00CB, 'ERROR_ENVVAR_NOT_FOUND') # WinError 203 (0xCB)
    ERROR_NO_SIGNAL_SENT: Final[ThirtyTwoBits] = cached_win32error(0x00CD, 'ERROR_NO_SIGNAL_SENT') # WinError 205 (0xCD)
    ERROR_FILENAME_EXCED_RANGE: Final[ThirtyTwoBits] = cached_win32error(0x00CE, 'ERROR_FILENAME_EXCED_RANGE') # WinError 206 (0xCE)
    ERROR_RING2_STACK_IN_USE: Final[ThirtyTwoBits] = cached_win32error(0x00CF, 'ERROR_RING2_STACK_IN_USE') # WinError 207 (0xCF)
    ERROR_META_EXPANSION_TOO_LONG: Final[ThirtyTwoBits] = cached_win32error(0x00D0, 'ERROR_META_EXPANSION_TOO_LONG') # WinError 208 (0xD0)
    ERROR_INVALID_SIGNAL_NUMBER: Final[ThirtyTwoBits] = cached_win32error(0x00D1, 'ERROR_INVALID_SIGNAL_NUMBER') # WinError 209 (0xD1)
    ERROR_THREAD_1_INACTIVE: Final[ThirtyTwoBits] = cached_win32error(0x00D2, 'ERROR_THREAD_1_INACTIVE') # WinError 210 (0xD2)
    ERROR_LOCKED: Final[ThirtyTwoBits] = cached_win32error(0x00D4, 'ERROR_LOCKED') # WinError 212 (0xD4)
    ERROR_TOO_MANY_MODULES: Final[ThirtyTwoBits] = cached_win32error(0x00D6, 'ERROR_TOO_MANY_MODULES') # WinError 214 (0xD6)
    ERROR_NESTING_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x00D7, 'ERROR_NESTING_NOT_ALLOWED') # WinError 215 (0xD7)
    ERROR_EXE_MACHINE_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x00D8, 'ERROR_EXE_MACHINE_TYPE_MISMATCH') # WinError 216 (0xD8)
    ERROR_EXE_CANNOT_MODIFY_SIGNED_BINARY: Final[ThirtyTwoBits] = cached_win32error(0x00D9, 'ERROR_EXE_CANNOT_MODIFY_SIGNED_BINARY') # WinError 217 (0xD9)
    ERROR_EXE_CANNOT_MODIFY_STRONG_SIGNED_BINARY: Final[ThirtyTwoBits] = cached_win32error(0x00DA, 'ERROR_EXE_CANNOT_MODIFY_STRONG_SIGNED_BINARY') # WinError 218 (0xDA)
    ERROR_FILE_CHECKED_OUT: Final[ThirtyTwoBits] = cached_win32error(0x00DC, 'ERROR_FILE_CHECKED_OUT') # WinError 220 (0xDC)
    ERROR_CHECKOUT_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x00DD, 'ERROR_CHECKOUT_REQUIRED') # WinError 221 (0xDD)
    ERROR_BAD_FILE_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x00DE, 'ERROR_BAD_FILE_TYPE') # WinError 222 (0xDE)
    ERROR_FILE_TOO_LARGE: Final[ThirtyTwoBits] = cached_win32error(0x00DF, 'ERROR_FILE_TOO_LARGE') # WinError 223 (0xDF)
    ERROR_FORMS_AUTH_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x00E0, 'ERROR_FORMS_AUTH_REQUIRED') # WinError 224 (0xE0)
    ERROR_VIRUS_INFECTED: Final[ThirtyTwoBits] = cached_win32error(0x00E1, 'ERROR_VIRUS_INFECTED') # WinError 225 (0xE1)
    ERROR_VIRUS_DELETED: Final[ThirtyTwoBits] = cached_win32error(0x00E2, 'ERROR_VIRUS_DELETED') # WinError 226 (0xE2)
    ERROR_PIPE_LOCAL: Final[ThirtyTwoBits] = cached_win32error(0x00E5, 'ERROR_PIPE_LOCAL') # WinError 229 (0xE5)
    ERROR_BAD_PIPE: Final[ThirtyTwoBits] = cached_win32error(0x00E6, 'ERROR_BAD_PIPE') # WinError 230 (0xE6)
    ERROR_PIPE_BUSY: Final[ThirtyTwoBits] = cached_win32error(0x00E7, 'ERROR_PIPE_BUSY') # WinError 231 (0xE7)
    ERROR_NO_DATA: Final[ThirtyTwoBits] = cached_win32error(0x00E8, 'ERROR_NO_DATA') # WinError 232 (0xE8)
    ERROR_PIPE_NOT_CONNECTED: Final[ThirtyTwoBits] = cached_win32error(0x00E9, 'ERROR_PIPE_NOT_CONNECTED') # WinError 233 (0xE9)
    ERROR_MORE_DATA: Final[ThirtyTwoBits] = cached_win32error(0x00EA, 'ERROR_MORE_DATA') # WinError 234 (0xEA)
    ERROR_NO_WORK_DONE: Final[ThirtyTwoBits] = cached_win32error(0x00EB, 'ERROR_NO_WORK_DONE') # WinError 235 (0xEB)
    ERROR_VC_DISCONNECTED: Final[ThirtyTwoBits] = cached_win32error(0x00F0, 'ERROR_VC_DISCONNECTED') # WinError 240 (0xF0)
    ERROR_INVALID_EA_NAME: Final[ThirtyTwoBits] = cached_win32error(0x00FE, 'ERROR_INVALID_EA_NAME') # WinError 254 (0xFE)
    ERROR_EA_LIST_INCONSISTENT: Final[ThirtyTwoBits] = cached_win32error(0x00FF, 'ERROR_EA_LIST_INCONSISTENT') # WinError 255 (0xFF)
    WAIT_TIMEOUT: Final[ThirtyTwoBits] = cached_win32error(0x0102, 'WAIT_TIMEOUT') # WinError 258 (0x102)
    ERROR_NO_MORE_ITEMS: Final[ThirtyTwoBits] = cached_win32error(0x0103, 'ERROR_NO_MORE_ITEMS') # WinError 259 (0x103)
    ERROR_CANNOT_COPY: Final[ThirtyTwoBits] = cached_win32error(0x010A, 'ERROR_CANNOT_COPY') # WinError 266 (0x10A)
    ERROR_DIRECTORY: Final[ThirtyTwoBits] = cached_win32error(0x010B, 'ERROR_DIRECTORY') # WinError 267 (0x10B)
    ERROR_EAS_DIDNT_FIT: Final[ThirtyTwoBits] = cached_win32error(0x0113, 'ERROR_EAS_DIDNT_FIT') # WinError 275 (0x113)
    ERROR_EA_FILE_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x0114, 'ERROR_EA_FILE_CORRUPT') # WinError 276 (0x114)
    ERROR_EA_TABLE_FULL: Final[ThirtyTwoBits] = cached_win32error(0x0115, 'ERROR_EA_TABLE_FULL') # WinError 277 (0x115)
    ERROR_INVALID_EA_HANDLE: Final[ThirtyTwoBits] = cached_win32error(0x0116, 'ERROR_INVALID_EA_HANDLE') # WinError 278 (0x116)
    ERROR_EAS_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x011A, 'ERROR_EAS_NOT_SUPPORTED') # WinError 282 (0x11A)
    ERROR_NOT_OWNER: Final[ThirtyTwoBits] = cached_win32error(0x0120, 'ERROR_NOT_OWNER') # WinError 288 (0x120)
    ERROR_TOO_MANY_POSTS: Final[ThirtyTwoBits] = cached_win32error(0x012A, 'ERROR_TOO_MANY_POSTS') # WinError 298 (0x12A)
    ERROR_PARTIAL_COPY: Final[ThirtyTwoBits] = cached_win32error(0x012B, 'ERROR_PARTIAL_COPY') # WinError 299 (0x12B)
    ERROR_OPLOCK_NOT_GRANTED: Final[ThirtyTwoBits] = cached_win32error(0x012C, 'ERROR_OPLOCK_NOT_GRANTED') # WinError 300 (0x12C)
    ERROR_INVALID_OPLOCK_PROTOCOL: Final[ThirtyTwoBits] = cached_win32error(0x012D, 'ERROR_INVALID_OPLOCK_PROTOCOL') # WinError 301 (0x12D)
    ERROR_DISK_TOO_FRAGMENTED: Final[ThirtyTwoBits] = cached_win32error(0x012E, 'ERROR_DISK_TOO_FRAGMENTED') # WinError 302 (0x12E)
    ERROR_DELETE_PENDING: Final[ThirtyTwoBits] = cached_win32error(0x012F, 'ERROR_DELETE_PENDING') # WinError 303 (0x12F)
    ERROR_INCOMPATIBLE_WITH_GLOBAL_SHORT_NAME_REGISTRY_SETTING: Final[ThirtyTwoBits] = cached_win32error(0x0130, 'ERROR_INCOMPATIBLE_WITH_GLOBAL_SHORT_NAME_REGISTRY_SETTING') # WinError 304 (0x130)
    ERROR_SHORT_NAMES_NOT_ENABLED_ON_VOLUME: Final[ThirtyTwoBits] = cached_win32error(0x0131, 'ERROR_SHORT_NAMES_NOT_ENABLED_ON_VOLUME') # WinError 305 (0x131)
    ERROR_SECURITY_STREAM_IS_INCONSISTENT: Final[ThirtyTwoBits] = cached_win32error(0x0132, 'ERROR_SECURITY_STREAM_IS_INCONSISTENT') # WinError 306 (0x132)
    ERROR_INVALID_LOCK_RANGE: Final[ThirtyTwoBits] = cached_win32error(0x0133, 'ERROR_INVALID_LOCK_RANGE') # WinError 307 (0x133)
    ERROR_IMAGE_SUBSYSTEM_NOT_PRESENT: Final[ThirtyTwoBits] = cached_win32error(0x0134, 'ERROR_IMAGE_SUBSYSTEM_NOT_PRESENT') # WinError 308 (0x134)
    ERROR_NOTIFICATION_GUID_ALREADY_DEFINED: Final[ThirtyTwoBits] = cached_win32error(0x0135, 'ERROR_NOTIFICATION_GUID_ALREADY_DEFINED') # WinError 309 (0x135)
    ERROR_INVALID_EXCEPTION_HANDLER: Final[ThirtyTwoBits] = cached_win32error(0x0136, 'ERROR_INVALID_EXCEPTION_HANDLER') # WinError 310 (0x136)
    ERROR_DUPLICATE_PRIVILEGES: Final[ThirtyTwoBits] = cached_win32error(0x0137, 'ERROR_DUPLICATE_PRIVILEGES') # WinError 311 (0x137)
    ERROR_NO_RANGES_PROCESSED: Final[ThirtyTwoBits] = cached_win32error(0x0138, 'ERROR_NO_RANGES_PROCESSED') # WinError 312 (0x138)
    ERROR_NOT_ALLOWED_ON_SYSTEM_FILE: Final[ThirtyTwoBits] = cached_win32error(0x0139, 'ERROR_NOT_ALLOWED_ON_SYSTEM_FILE') # WinError 313 (0x139)
    ERROR_DISK_RESOURCES_EXHAUSTED: Final[ThirtyTwoBits] = cached_win32error(0x013A, 'ERROR_DISK_RESOURCES_EXHAUSTED') # WinError 314 (0x13A)
    ERROR_INVALID_TOKEN: Final[ThirtyTwoBits] = cached_win32error(0x013B, 'ERROR_INVALID_TOKEN') # WinError 315 (0x13B)
    ERROR_DEVICE_FEATURE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x013C, 'ERROR_DEVICE_FEATURE_NOT_SUPPORTED') # WinError 316 (0x13C)
    ERROR_MR_MID_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x013D, 'ERROR_MR_MID_NOT_FOUND') # WinError 317 (0x13D)
    ERROR_SCOPE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x013E, 'ERROR_SCOPE_NOT_FOUND') # WinError 318 (0x13E)
    ERROR_UNDEFINED_SCOPE: Final[ThirtyTwoBits] = cached_win32error(0x013F, 'ERROR_UNDEFINED_SCOPE') # WinError 319 (0x13F)
    ERROR_INVALID_CAP: Final[ThirtyTwoBits] = cached_win32error(0x0140, 'ERROR_INVALID_CAP') # WinError 320 (0x140)
    ERROR_DEVICE_UNREACHABLE: Final[ThirtyTwoBits] = cached_win32error(0x0141, 'ERROR_DEVICE_UNREACHABLE') # WinError 321 (0x141)
    ERROR_DEVICE_NO_RESOURCES: Final[ThirtyTwoBits] = cached_win32error(0x0142, 'ERROR_DEVICE_NO_RESOURCES') # WinError 322 (0x142)
    ERROR_DATA_CHECKSUM_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x0143, 'ERROR_DATA_CHECKSUM_ERROR') # WinError 323 (0x143)
    ERROR_INTERMIXED_KERNEL_EA_OPERATION: Final[ThirtyTwoBits] = cached_win32error(0x0144, 'ERROR_INTERMIXED_KERNEL_EA_OPERATION') # WinError 324 (0x144)
    ERROR_FILE_LEVEL_TRIM_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x0146, 'ERROR_FILE_LEVEL_TRIM_NOT_SUPPORTED') # WinError 326 (0x146)
    ERROR_OFFSET_ALIGNMENT_VIOLATION: Final[ThirtyTwoBits] = cached_win32error(0x0147, 'ERROR_OFFSET_ALIGNMENT_VIOLATION') # WinError 327 (0x147)
    ERROR_INVALID_FIELD_IN_PARAMETER_LIST: Final[ThirtyTwoBits] = cached_win32error(0x0148, 'ERROR_INVALID_FIELD_IN_PARAMETER_LIST') # WinError 328 (0x148)
    ERROR_OPERATION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x0149, 'ERROR_OPERATION_IN_PROGRESS') # WinError 329 (0x149)
    ERROR_BAD_DEVICE_PATH: Final[ThirtyTwoBits] = cached_win32error(0x014A, 'ERROR_BAD_DEVICE_PATH') # WinError 330 (0x14A)
    ERROR_TOO_MANY_DESCRIPTORS: Final[ThirtyTwoBits] = cached_win32error(0x014B, 'ERROR_TOO_MANY_DESCRIPTORS') # WinError 331 (0x14B)
    ERROR_SCRUB_DATA_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x014C, 'ERROR_SCRUB_DATA_DISABLED') # WinError 332 (0x14C)
    ERROR_NOT_REDUNDANT_STORAGE: Final[ThirtyTwoBits] = cached_win32error(0x014D, 'ERROR_NOT_REDUNDANT_STORAGE') # WinError 333 (0x14D)
    ERROR_RESIDENT_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x014E, 'ERROR_RESIDENT_FILE_NOT_SUPPORTED') # WinError 334 (0x14E)
    ERROR_COMPRESSED_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x014F, 'ERROR_COMPRESSED_FILE_NOT_SUPPORTED') # WinError 335 (0x14F)
    ERROR_DIRECTORY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x0150, 'ERROR_DIRECTORY_NOT_SUPPORTED') # WinError 336 (0x150)
    ERROR_NOT_READ_FROM_COPY: Final[ThirtyTwoBits] = cached_win32error(0x0151, 'ERROR_NOT_READ_FROM_COPY') # WinError 337 (0x151)
    ERROR_FT_WRITE_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x0152, 'ERROR_FT_WRITE_FAILURE') # WinError 338 (0x152)
    ERROR_FT_DI_SCAN_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x0153, 'ERROR_FT_DI_SCAN_REQUIRED') # WinError 339 (0x153)
    ERROR_INVALID_KERNEL_INFO_VERSION: Final[ThirtyTwoBits] = cached_win32error(0x0154, 'ERROR_INVALID_KERNEL_INFO_VERSION') # WinError 340 (0x154)
    ERROR_INVALID_PEP_INFO_VERSION: Final[ThirtyTwoBits] = cached_win32error(0x0155, 'ERROR_INVALID_PEP_INFO_VERSION') # WinError 341 (0x155)
    ERROR_OBJECT_NOT_EXTERNALLY_BACKED: Final[ThirtyTwoBits] = cached_win32error(0x0156, 'ERROR_OBJECT_NOT_EXTERNALLY_BACKED') # WinError 342 (0x156)
    ERROR_EXTERNAL_BACKING_PROVIDER_UNKNOWN: Final[ThirtyTwoBits] = cached_win32error(0x0157, 'ERROR_EXTERNAL_BACKING_PROVIDER_UNKNOWN') # WinError 343 (0x157)
    ERROR_COMPRESSION_NOT_BENEFICIAL: Final[ThirtyTwoBits] = cached_win32error(0x0158, 'ERROR_COMPRESSION_NOT_BENEFICIAL') # WinError 344 (0x158)
    ERROR_STORAGE_TOPOLOGY_ID_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x0159, 'ERROR_STORAGE_TOPOLOGY_ID_MISMATCH') # WinError 345 (0x159)
    ERROR_BLOCKED_BY_PARENTAL_CONTROLS: Final[ThirtyTwoBits] = cached_win32error(0x015A, 'ERROR_BLOCKED_BY_PARENTAL_CONTROLS') # WinError 346 (0x15A)
    ERROR_BLOCK_TOO_MANY_REFERENCES: Final[ThirtyTwoBits] = cached_win32error(0x015B, 'ERROR_BLOCK_TOO_MANY_REFERENCES') # WinError 347 (0x15B)
    ERROR_MARKED_TO_DISALLOW_WRITES: Final[ThirtyTwoBits] = cached_win32error(0x015C, 'ERROR_MARKED_TO_DISALLOW_WRITES') # WinError 348 (0x15C)
    ERROR_ENCLAVE_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x015D, 'ERROR_ENCLAVE_FAILURE') # WinError 349 (0x15D)
    ERROR_FAIL_NOACTION_REBOOT: Final[ThirtyTwoBits] = cached_win32error(0x015E, 'ERROR_FAIL_NOACTION_REBOOT') # WinError 350 (0x15E)
    ERROR_FAIL_SHUTDOWN: Final[ThirtyTwoBits] = cached_win32error(0x015F, 'ERROR_FAIL_SHUTDOWN') # WinError 351 (0x15F)
    ERROR_FAIL_RESTART: Final[ThirtyTwoBits] = cached_win32error(0x0160, 'ERROR_FAIL_RESTART') # WinError 352 (0x160)
    ERROR_MAX_SESSIONS_REACHED: Final[ThirtyTwoBits] = cached_win32error(0x0161, 'ERROR_MAX_SESSIONS_REACHED') # WinError 353 (0x161)
    ERROR_NETWORK_ACCESS_DENIED_EDP: Final[ThirtyTwoBits] = cached_win32error(0x0162, 'ERROR_NETWORK_ACCESS_DENIED_EDP') # WinError 354 (0x162)
    ERROR_DEVICE_HINT_NAME_BUFFER_TOO_SMALL: Final[ThirtyTwoBits] = cached_win32error(0x0163, 'ERROR_DEVICE_HINT_NAME_BUFFER_TOO_SMALL') # WinError 355 (0x163)
    ERROR_EDP_POLICY_DENIES_OPERATION: Final[ThirtyTwoBits] = cached_win32error(0x0164, 'ERROR_EDP_POLICY_DENIES_OPERATION') # WinError 356 (0x164)
    ERROR_EDP_DPL_POLICY_CANT_BE_SATISFIED: Final[ThirtyTwoBits] = cached_win32error(0x0165, 'ERROR_EDP_DPL_POLICY_CANT_BE_SATISFIED') # WinError 357 (0x165)
    ERROR_CLOUD_FILE_SYNC_ROOT_METADATA_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x0166, 'ERROR_CLOUD_FILE_SYNC_ROOT_METADATA_CORRUPT') # WinError 358 (0x166)
    ERROR_DEVICE_IN_MAINTENANCE: Final[ThirtyTwoBits] = cached_win32error(0x0167, 'ERROR_DEVICE_IN_MAINTENANCE') # WinError 359 (0x167)
    ERROR_NOT_SUPPORTED_ON_DAX: Final[ThirtyTwoBits] = cached_win32error(0x0168, 'ERROR_NOT_SUPPORTED_ON_DAX') # WinError 360 (0x168)
    ERROR_DAX_MAPPING_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x0169, 'ERROR_DAX_MAPPING_EXISTS') # WinError 361 (0x169)
    ERROR_CLOUD_FILE_PROVIDER_NOT_RUNNING: Final[ThirtyTwoBits] = cached_win32error(0x016A, 'ERROR_CLOUD_FILE_PROVIDER_NOT_RUNNING') # WinError 362 (0x16A)
    ERROR_CLOUD_FILE_METADATA_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x016B, 'ERROR_CLOUD_FILE_METADATA_CORRUPT') # WinError 363 (0x16B)
    ERROR_CLOUD_FILE_METADATA_TOO_LARGE: Final[ThirtyTwoBits] = cached_win32error(0x016C, 'ERROR_CLOUD_FILE_METADATA_TOO_LARGE') # WinError 364 (0x16C)
    ERROR_CLOUD_FILE_PROPERTY_BLOB_TOO_LARGE: Final[ThirtyTwoBits] = cached_win32error(0x016D, 'ERROR_CLOUD_FILE_PROPERTY_BLOB_TOO_LARGE') # WinError 365 (0x16D)
    ERROR_CLOUD_FILE_PROPERTY_BLOB_CHECKSUM_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x016E, 'ERROR_CLOUD_FILE_PROPERTY_BLOB_CHECKSUM_MISMATCH') # WinError 366 (0x16E)
    ERROR_CHILD_PROCESS_BLOCKED: Final[ThirtyTwoBits] = cached_win32error(0x016F, 'ERROR_CHILD_PROCESS_BLOCKED') # WinError 367 (0x16F)
    ERROR_STORAGE_LOST_DATA_PERSISTENCE: Final[ThirtyTwoBits] = cached_win32error(0x0170, 'ERROR_STORAGE_LOST_DATA_PERSISTENCE') # WinError 368 (0x170)
    ERROR_FILE_SYSTEM_VIRTUALIZATION_UNAVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x0171, 'ERROR_FILE_SYSTEM_VIRTUALIZATION_UNAVAILABLE') # WinError 369 (0x171)
    ERROR_FILE_SYSTEM_VIRTUALIZATION_METADATA_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x0172, 'ERROR_FILE_SYSTEM_VIRTUALIZATION_METADATA_CORRUPT') # WinError 370 (0x172)
    ERROR_FILE_SYSTEM_VIRTUALIZATION_BUSY: Final[ThirtyTwoBits] = cached_win32error(0x0173, 'ERROR_FILE_SYSTEM_VIRTUALIZATION_BUSY') # WinError 371 (0x173)
    ERROR_FILE_SYSTEM_VIRTUALIZATION_PROVIDER_UNKNOWN: Final[ThirtyTwoBits] = cached_win32error(0x0174, 'ERROR_FILE_SYSTEM_VIRTUALIZATION_PROVIDER_UNKNOWN') # WinError 372 (0x174)
    ERROR_GDI_HANDLE_LEAK: Final[ThirtyTwoBits] = cached_win32error(0x0175, 'ERROR_GDI_HANDLE_LEAK') # WinError 373 (0x175)
    ERROR_CLOUD_FILE_TOO_MANY_PROPERTY_BLOBS: Final[ThirtyTwoBits] = cached_win32error(0x0176, 'ERROR_CLOUD_FILE_TOO_MANY_PROPERTY_BLOBS') # WinError 374 (0x176)
    ERROR_CLOUD_FILE_PROPERTY_VERSION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x0177, 'ERROR_CLOUD_FILE_PROPERTY_VERSION_NOT_SUPPORTED') # WinError 375 (0x177)
    ERROR_NOT_A_CLOUD_FILE: Final[ThirtyTwoBits] = cached_win32error(0x0178, 'ERROR_NOT_A_CLOUD_FILE') # WinError 376 (0x178)
    ERROR_CLOUD_FILE_NOT_IN_SYNC: Final[ThirtyTwoBits] = cached_win32error(0x0179, 'ERROR_CLOUD_FILE_NOT_IN_SYNC') # WinError 377 (0x179)
    ERROR_CLOUD_FILE_ALREADY_CONNECTED: Final[ThirtyTwoBits] = cached_win32error(0x017A, 'ERROR_CLOUD_FILE_ALREADY_CONNECTED') # WinError 378 (0x17A)
    ERROR_CLOUD_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x017B, 'ERROR_CLOUD_FILE_NOT_SUPPORTED') # WinError 379 (0x17B)
    ERROR_CLOUD_FILE_INVALID_REQUEST: Final[ThirtyTwoBits] = cached_win32error(0x017C, 'ERROR_CLOUD_FILE_INVALID_REQUEST') # WinError 380 (0x17C)
    ERROR_CLOUD_FILE_READ_ONLY_VOLUME: Final[ThirtyTwoBits] = cached_win32error(0x017D, 'ERROR_CLOUD_FILE_READ_ONLY_VOLUME') # WinError 381 (0x17D)
    ERROR_CLOUD_FILE_CONNECTED_PROVIDER_ONLY: Final[ThirtyTwoBits] = cached_win32error(0x017E, 'ERROR_CLOUD_FILE_CONNECTED_PROVIDER_ONLY') # WinError 382 (0x17E)
    ERROR_CLOUD_FILE_VALIDATION_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x017F, 'ERROR_CLOUD_FILE_VALIDATION_FAILED') # WinError 383 (0x17F)
    ERROR_SMB1_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x0180, 'ERROR_SMB1_NOT_AVAILABLE') # WinError 384 (0x180)
    ERROR_FILE_SYSTEM_VIRTUALIZATION_INVALID_OPERATION: Final[ThirtyTwoBits] = cached_win32error(0x0181, 'ERROR_FILE_SYSTEM_VIRTUALIZATION_INVALID_OPERATION') # WinError 385 (0x181)
    ERROR_CLOUD_FILE_AUTHENTICATION_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x0182, 'ERROR_CLOUD_FILE_AUTHENTICATION_FAILED') # WinError 386 (0x182)
    ERROR_CLOUD_FILE_INSUFFICIENT_RESOURCES: Final[ThirtyTwoBits] = cached_win32error(0x0183, 'ERROR_CLOUD_FILE_INSUFFICIENT_RESOURCES') # WinError 387 (0x183)
    ERROR_CLOUD_FILE_NETWORK_UNAVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x0184, 'ERROR_CLOUD_FILE_NETWORK_UNAVAILABLE') # WinError 388 (0x184)
    ERROR_CLOUD_FILE_UNSUCCESSFUL: Final[ThirtyTwoBits] = cached_win32error(0x0185, 'ERROR_CLOUD_FILE_UNSUCCESSFUL') # WinError 389 (0x185)
    ERROR_CLOUD_FILE_NOT_UNDER_SYNC_ROOT: Final[ThirtyTwoBits] = cached_win32error(0x0186, 'ERROR_CLOUD_FILE_NOT_UNDER_SYNC_ROOT') # WinError 390 (0x186)
    ERROR_CLOUD_FILE_IN_USE: Final[ThirtyTwoBits] = cached_win32error(0x0187, 'ERROR_CLOUD_FILE_IN_USE') # WinError 391 (0x187)
    ERROR_CLOUD_FILE_PINNED: Final[ThirtyTwoBits] = cached_win32error(0x0188, 'ERROR_CLOUD_FILE_PINNED') # WinError 392 (0x188)
    ERROR_CLOUD_FILE_REQUEST_ABORTED: Final[ThirtyTwoBits] = cached_win32error(0x0189, 'ERROR_CLOUD_FILE_REQUEST_ABORTED') # WinError 393 (0x189)
    ERROR_CLOUD_FILE_PROPERTY_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x018A, 'ERROR_CLOUD_FILE_PROPERTY_CORRUPT') # WinError 394 (0x18A)
    ERROR_CLOUD_FILE_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_win32error(0x018B, 'ERROR_CLOUD_FILE_ACCESS_DENIED') # WinError 395 (0x18B)
    ERROR_CLOUD_FILE_INCOMPATIBLE_HARDLINKS: Final[ThirtyTwoBits] = cached_win32error(0x018C, 'ERROR_CLOUD_FILE_INCOMPATIBLE_HARDLINKS') # WinError 396 (0x18C)
    ERROR_CLOUD_FILE_PROPERTY_LOCK_CONFLICT: Final[ThirtyTwoBits] = cached_win32error(0x018D, 'ERROR_CLOUD_FILE_PROPERTY_LOCK_CONFLICT') # WinError 397 (0x18D)
    ERROR_CLOUD_FILE_REQUEST_CANCELED: Final[ThirtyTwoBits] = cached_win32error(0x018E, 'ERROR_CLOUD_FILE_REQUEST_CANCELED') # WinError 398 (0x18E)
    ERROR_EXTERNAL_SYSKEY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x018F, 'ERROR_EXTERNAL_SYSKEY_NOT_SUPPORTED') # WinError 399 (0x18F)
    ERROR_THREAD_MODE_ALREADY_BACKGROUND: Final[ThirtyTwoBits] = cached_win32error(0x0190, 'ERROR_THREAD_MODE_ALREADY_BACKGROUND') # WinError 400 (0x190)
    ERROR_THREAD_MODE_NOT_BACKGROUND: Final[ThirtyTwoBits] = cached_win32error(0x0191, 'ERROR_THREAD_MODE_NOT_BACKGROUND') # WinError 401 (0x191)
    ERROR_PROCESS_MODE_ALREADY_BACKGROUND: Final[ThirtyTwoBits] = cached_win32error(0x0192, 'ERROR_PROCESS_MODE_ALREADY_BACKGROUND') # WinError 402 (0x192)
    ERROR_PROCESS_MODE_NOT_BACKGROUND: Final[ThirtyTwoBits] = cached_win32error(0x0193, 'ERROR_PROCESS_MODE_NOT_BACKGROUND') # WinError 403 (0x193)
    ERROR_CLOUD_FILE_PROVIDER_TERMINATED: Final[ThirtyTwoBits] = cached_win32error(0x0194, 'ERROR_CLOUD_FILE_PROVIDER_TERMINATED') # WinError 404 (0x194)
    ERROR_NOT_A_CLOUD_SYNC_ROOT: Final[ThirtyTwoBits] = cached_win32error(0x0195, 'ERROR_NOT_A_CLOUD_SYNC_ROOT') # WinError 405 (0x195)
    ERROR_FILE_PROTECTED_UNDER_DPL: Final[ThirtyTwoBits] = cached_win32error(0x0196, 'ERROR_FILE_PROTECTED_UNDER_DPL') # WinError 406 (0x196)
    ERROR_VOLUME_NOT_CLUSTER_ALIGNED: Final[ThirtyTwoBits] = cached_win32error(0x0197, 'ERROR_VOLUME_NOT_CLUSTER_ALIGNED') # WinError 407 (0x197)
    ERROR_NO_PHYSICALLY_ALIGNED_FREE_SPACE_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x0198, 'ERROR_NO_PHYSICALLY_ALIGNED_FREE_SPACE_FOUND') # WinError 408 (0x198)
    ERROR_APPX_FILE_NOT_ENCRYPTED: Final[ThirtyTwoBits] = cached_win32error(0x0199, 'ERROR_APPX_FILE_NOT_ENCRYPTED') # WinError 409 (0x199)
    ERROR_RWRAW_ENCRYPTED_FILE_NOT_ENCRYPTED: Final[ThirtyTwoBits] = cached_win32error(0x019A, 'ERROR_RWRAW_ENCRYPTED_FILE_NOT_ENCRYPTED') # WinError 410 (0x19A)
    ERROR_RWRAW_ENCRYPTED_INVALID_EDATAINFO_FILEOFFSET: Final[ThirtyTwoBits] = cached_win32error(0x019B, 'ERROR_RWRAW_ENCRYPTED_INVALID_EDATAINFO_FILEOFFSET') # WinError 411 (0x19B)
    ERROR_RWRAW_ENCRYPTED_INVALID_EDATAINFO_FILERANGE: Final[ThirtyTwoBits] = cached_win32error(0x019C, 'ERROR_RWRAW_ENCRYPTED_INVALID_EDATAINFO_FILERANGE') # WinError 412 (0x19C)
    ERROR_RWRAW_ENCRYPTED_INVALID_EDATAINFO_PARAMETER: Final[ThirtyTwoBits] = cached_win32error(0x019D, 'ERROR_RWRAW_ENCRYPTED_INVALID_EDATAINFO_PARAMETER') # WinError 413 (0x19D)
    ERROR_LINUX_SUBSYSTEM_NOT_PRESENT: Final[ThirtyTwoBits] = cached_win32error(0x019E, 'ERROR_LINUX_SUBSYSTEM_NOT_PRESENT') # WinError 414 (0x19E)
    ERROR_FT_READ_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x019F, 'ERROR_FT_READ_FAILURE') # WinError 415 (0x19F)
    ERROR_STORAGE_RESERVE_ID_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x01A0, 'ERROR_STORAGE_RESERVE_ID_INVALID') # WinError 416 (0x1A0)
    ERROR_STORAGE_RESERVE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_win32error(0x01A1, 'ERROR_STORAGE_RESERVE_DOES_NOT_EXIST') # WinError 417 (0x1A1)
    ERROR_STORAGE_RESERVE_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x01A2, 'ERROR_STORAGE_RESERVE_ALREADY_EXISTS') # WinError 418 (0x1A2)
    ERROR_STORAGE_RESERVE_NOT_EMPTY: Final[ThirtyTwoBits] = cached_win32error(0x01A3, 'ERROR_STORAGE_RESERVE_NOT_EMPTY') # WinError 419 (0x1A3)
    ERROR_NOT_A_DAX_VOLUME: Final[ThirtyTwoBits] = cached_win32error(0x01A4, 'ERROR_NOT_A_DAX_VOLUME') # WinError 420 (0x1A4)
    ERROR_NOT_DAX_MAPPABLE: Final[ThirtyTwoBits] = cached_win32error(0x01A5, 'ERROR_NOT_DAX_MAPPABLE') # WinError 421 (0x1A5)
    ERROR_TIME_SENSITIVE_THREAD: Final[ThirtyTwoBits] = cached_win32error(0x01A6, 'ERROR_TIME_SENSITIVE_THREAD') # WinError 422 (0x1A6)
    ERROR_DPL_NOT_SUPPORTED_FOR_USER: Final[ThirtyTwoBits] = cached_win32error(0x01A7, 'ERROR_DPL_NOT_SUPPORTED_FOR_USER') # WinError 423 (0x1A7)
    ERROR_CASE_DIFFERING_NAMES_IN_DIR: Final[ThirtyTwoBits] = cached_win32error(0x01A8, 'ERROR_CASE_DIFFERING_NAMES_IN_DIR') # WinError 424 (0x1A8)
    ERROR_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x01A9, 'ERROR_FILE_NOT_SUPPORTED') # WinError 425 (0x1A9)
    ERROR_CLOUD_FILE_REQUEST_TIMEOUT: Final[ThirtyTwoBits] = cached_win32error(0x01AA, 'ERROR_CLOUD_FILE_REQUEST_TIMEOUT') # WinError 426 (0x1AA)
    ERROR_NO_TASK_QUEUE: Final[ThirtyTwoBits] = cached_win32error(0x01AB, 'ERROR_NO_TASK_QUEUE') # WinError 427 (0x1AB)
    ERROR_SRC_SRV_DLL_LOAD_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x01AC, 'ERROR_SRC_SRV_DLL_LOAD_FAILED') # WinError 428 (0x1AC)
    ERROR_NOT_SUPPORTED_WITH_BTT: Final[ThirtyTwoBits] = cached_win32error(0x01AD, 'ERROR_NOT_SUPPORTED_WITH_BTT') # WinError 429 (0x1AD)
    ERROR_ENCRYPTION_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x01AE, 'ERROR_ENCRYPTION_DISABLED') # WinError 430 (0x1AE)
    ERROR_ENCRYPTING_METADATA_DISALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x01AF, 'ERROR_ENCRYPTING_METADATA_DISALLOWED') # WinError 431 (0x1AF)
    ERROR_CANT_CLEAR_ENCRYPTION_FLAG: Final[ThirtyTwoBits] = cached_win32error(0x01B0, 'ERROR_CANT_CLEAR_ENCRYPTION_FLAG') # WinError 432 (0x1B0)
    ERROR_NO_SUCH_DEVICE: Final[ThirtyTwoBits] = cached_win32error(0x01B1, 'ERROR_NO_SUCH_DEVICE') # WinError 433 (0x1B1)
    ERROR_CLOUD_FILE_DEHYDRATION_DISALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x01B2, 'ERROR_CLOUD_FILE_DEHYDRATION_DISALLOWED') # WinError 434 (0x1B2)
    ERROR_FILE_SNAP_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x01B3, 'ERROR_FILE_SNAP_IN_PROGRESS') # WinError 435 (0x1B3)
    ERROR_FILE_SNAP_USER_SECTION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x01B4, 'ERROR_FILE_SNAP_USER_SECTION_NOT_SUPPORTED') # WinError 436 (0x1B4)
    ERROR_FILE_SNAP_MODIFY_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x01B5, 'ERROR_FILE_SNAP_MODIFY_NOT_SUPPORTED') # WinError 437 (0x1B5)
    ERROR_FILE_SNAP_IO_NOT_COORDINATED: Final[ThirtyTwoBits] = cached_win32error(0x01B6, 'ERROR_FILE_SNAP_IO_NOT_COORDINATED') # WinError 438 (0x1B6)
    ERROR_FILE_SNAP_UNEXPECTED_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x01B7, 'ERROR_FILE_SNAP_UNEXPECTED_ERROR') # WinError 439 (0x1B7)
    ERROR_FILE_SNAP_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_win32error(0x01B8, 'ERROR_FILE_SNAP_INVALID_PARAMETER') # WinError 440 (0x1B8)
    ERROR_UNSATISFIED_DEPENDENCIES: Final[ThirtyTwoBits] = cached_win32error(0x01B9, 'ERROR_UNSATISFIED_DEPENDENCIES') # WinError 441 (0x1B9)
    ERROR_CASE_SENSITIVE_PATH: Final[ThirtyTwoBits] = cached_win32error(0x01BA, 'ERROR_CASE_SENSITIVE_PATH') # WinError 442 (0x1BA)
    ERROR_UNEXPECTED_NTCACHEMANAGER_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x01BB, 'ERROR_UNEXPECTED_NTCACHEMANAGER_ERROR') # WinError 443 (0x1BB)
    ERROR_LINUX_SUBSYSTEM_UPDATE_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x01BC, 'ERROR_LINUX_SUBSYSTEM_UPDATE_REQUIRED') # WinError 444 (0x1BC)
    ERROR_DLP_POLICY_WARNS_AGAINST_OPERATION: Final[ThirtyTwoBits] = cached_win32error(0x01BD, 'ERROR_DLP_POLICY_WARNS_AGAINST_OPERATION') # WinError 445 (0x1BD)
    ERROR_DLP_POLICY_DENIES_OPERATION: Final[ThirtyTwoBits] = cached_win32error(0x01BE, 'ERROR_DLP_POLICY_DENIES_OPERATION') # WinError 446 (0x1BE)
    ERROR_SECURITY_DENIES_OPERATION: Final[ThirtyTwoBits] = cached_win32error(0x01BF, 'ERROR_SECURITY_DENIES_OPERATION') # WinError 447 (0x1BF)
    ERROR_UNTRUSTED_MOUNT_POINT: Final[ThirtyTwoBits] = cached_win32error(0x01C0, 'ERROR_UNTRUSTED_MOUNT_POINT') # WinError 448 (0x1C0)
    ERROR_DLP_POLICY_SILENTLY_FAIL: Final[ThirtyTwoBits] = cached_win32error(0x01C1, 'ERROR_DLP_POLICY_SILENTLY_FAIL') # WinError 449 (0x1C1)
    ERROR_CAPAUTHZ_NOT_DEVUNLOCKED: Final[ThirtyTwoBits] = cached_win32error(0x01C2, 'ERROR_CAPAUTHZ_NOT_DEVUNLOCKED') # WinError 450 (0x1C2)
    ERROR_CAPAUTHZ_CHANGE_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x01C3, 'ERROR_CAPAUTHZ_CHANGE_TYPE') # WinError 451 (0x1C3)
    ERROR_CAPAUTHZ_NOT_PROVISIONED: Final[ThirtyTwoBits] = cached_win32error(0x01C4, 'ERROR_CAPAUTHZ_NOT_PROVISIONED') # WinError 452 (0x1C4)
    ERROR_CAPAUTHZ_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_win32error(0x01C5, 'ERROR_CAPAUTHZ_NOT_AUTHORIZED') # WinError 453 (0x1C5)
    ERROR_CAPAUTHZ_NO_POLICY: Final[ThirtyTwoBits] = cached_win32error(0x01C6, 'ERROR_CAPAUTHZ_NO_POLICY') # WinError 454 (0x1C6)
    ERROR_CAPAUTHZ_DB_CORRUPTED: Final[ThirtyTwoBits] = cached_win32error(0x01C7, 'ERROR_CAPAUTHZ_DB_CORRUPTED') # WinError 455 (0x1C7)
    ERROR_CAPAUTHZ_SCCD_INVALID_CATALOG: Final[ThirtyTwoBits] = cached_win32error(0x01C8, 'ERROR_CAPAUTHZ_SCCD_INVALID_CATALOG') # WinError 456 (0x1C8)
    ERROR_CAPAUTHZ_SCCD_NO_AUTH_ENTITY: Final[ThirtyTwoBits] = cached_win32error(0x01C9, 'ERROR_CAPAUTHZ_SCCD_NO_AUTH_ENTITY') # WinError 457 (0x1C9)
    ERROR_CAPAUTHZ_SCCD_PARSE_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x01CA, 'ERROR_CAPAUTHZ_SCCD_PARSE_ERROR') # WinError 458 (0x1CA)
    ERROR_CAPAUTHZ_SCCD_DEV_MODE_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x01CB, 'ERROR_CAPAUTHZ_SCCD_DEV_MODE_REQUIRED') # WinError 459 (0x1CB)
    ERROR_CAPAUTHZ_SCCD_NO_CAPABILITY_MATCH: Final[ThirtyTwoBits] = cached_win32error(0x01CC, 'ERROR_CAPAUTHZ_SCCD_NO_CAPABILITY_MATCH') # WinError 460 (0x1CC)
    ERROR_CIMFS_IMAGE_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x01D6, 'ERROR_CIMFS_IMAGE_CORRUPT') # WinError 470 (0x1D6)
    ERROR_CIMFS_IMAGE_VERSION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x01D7, 'ERROR_CIMFS_IMAGE_VERSION_NOT_SUPPORTED') # WinError 471 (0x1D7)
    ERROR_STORAGE_STACK_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_win32error(0x01D8, 'ERROR_STORAGE_STACK_ACCESS_DENIED') # WinError 472 (0x1D8)
    ERROR_INSUFFICIENT_VIRTUAL_ADDR_RESOURCES: Final[ThirtyTwoBits] = cached_win32error(0x01D9, 'ERROR_INSUFFICIENT_VIRTUAL_ADDR_RESOURCES') # WinError 473 (0x1D9)
    ERROR_INDEX_OUT_OF_BOUNDS: Final[ThirtyTwoBits] = cached_win32error(0x01DA, 'ERROR_INDEX_OUT_OF_BOUNDS') # WinError 474 (0x1DA)
    ERROR_CLOUD_FILE_US_MESSAGE_TIMEOUT: Final[ThirtyTwoBits] = cached_win32error(0x01DB, 'ERROR_CLOUD_FILE_US_MESSAGE_TIMEOUT') # WinError 475 (0x1DB)
    ERROR_NOT_A_DEV_VOLUME: Final[ThirtyTwoBits] = cached_win32error(0x01DC, 'ERROR_NOT_A_DEV_VOLUME') # WinError 476 (0x1DC)
    ERROR_FS_GUID_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x01DD, 'ERROR_FS_GUID_MISMATCH') # WinError 477 (0x1DD)
    ERROR_CANT_ATTACH_TO_DEV_VOLUME: Final[ThirtyTwoBits] = cached_win32error(0x01DE, 'ERROR_CANT_ATTACH_TO_DEV_VOLUME') # WinError 478 (0x1DE)
    ERROR_INVALID_CONFIG_VALUE: Final[ThirtyTwoBits] = cached_win32error(0x01DF, 'ERROR_INVALID_CONFIG_VALUE') # WinError 479 (0x1DF)
    ERROR_PNP_QUERY_REMOVE_DEVICE_TIMEOUT: Final[ThirtyTwoBits] = cached_win32error(0x01E0, 'ERROR_PNP_QUERY_REMOVE_DEVICE_TIMEOUT') # WinError 480 (0x1E0)
    ERROR_PNP_QUERY_REMOVE_RELATED_DEVICE_TIMEOUT: Final[ThirtyTwoBits] = cached_win32error(0x01E1, 'ERROR_PNP_QUERY_REMOVE_RELATED_DEVICE_TIMEOUT') # WinError 481 (0x1E1)
    ERROR_PNP_QUERY_REMOVE_UNRELATED_DEVICE_TIMEOUT: Final[ThirtyTwoBits] = cached_win32error(0x01E2, 'ERROR_PNP_QUERY_REMOVE_UNRELATED_DEVICE_TIMEOUT') # WinError 482 (0x1E2)
    ERROR_DEVICE_HARDWARE_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x01E3, 'ERROR_DEVICE_HARDWARE_ERROR') # WinError 483 (0x1E3)
    ERROR_INVALID_ADDRESS: Final[ThirtyTwoBits] = cached_win32error(0x01E7, 'ERROR_INVALID_ADDRESS') # WinError 487 (0x1E7)
    ERROR_HAS_SYSTEM_CRITICAL_FILES: Final[ThirtyTwoBits] = cached_win32error(0x01E8, 'ERROR_HAS_SYSTEM_CRITICAL_FILES') # WinError 488 (0x1E8)
    ERROR_ENCRYPTED_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x01E9, 'ERROR_ENCRYPTED_FILE_NOT_SUPPORTED') # WinError 489 (0x1E9)
    ERROR_SPARSE_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x01EA, 'ERROR_SPARSE_FILE_NOT_SUPPORTED') # WinError 490 (0x1EA)
    ERROR_PAGEFILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x01EB, 'ERROR_PAGEFILE_NOT_SUPPORTED') # WinError 491 (0x1EB)
    ERROR_VOLUME_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x01EC, 'ERROR_VOLUME_NOT_SUPPORTED') # WinError 492 (0x1EC)
    ERROR_NOT_SUPPORTED_WITH_BYPASSIO: Final[ThirtyTwoBits] = cached_win32error(0x01ED, 'ERROR_NOT_SUPPORTED_WITH_BYPASSIO') # WinError 493 (0x1ED)
    ERROR_NO_BYPASSIO_DRIVER_SUPPORT: Final[ThirtyTwoBits] = cached_win32error(0x01EE, 'ERROR_NO_BYPASSIO_DRIVER_SUPPORT') # WinError 494 (0x1EE)
    ERROR_NOT_SUPPORTED_WITH_ENCRYPTION: Final[ThirtyTwoBits] = cached_win32error(0x01EF, 'ERROR_NOT_SUPPORTED_WITH_ENCRYPTION') # WinError 495 (0x1EF)
    ERROR_NOT_SUPPORTED_WITH_COMPRESSION: Final[ThirtyTwoBits] = cached_win32error(0x01F0, 'ERROR_NOT_SUPPORTED_WITH_COMPRESSION') # WinError 496 (0x1F0)
    ERROR_NOT_SUPPORTED_WITH_REPLICATION: Final[ThirtyTwoBits] = cached_win32error(0x01F1, 'ERROR_NOT_SUPPORTED_WITH_REPLICATION') # WinError 497 (0x1F1)
    ERROR_NOT_SUPPORTED_WITH_DEDUPLICATION: Final[ThirtyTwoBits] = cached_win32error(0x01F2, 'ERROR_NOT_SUPPORTED_WITH_DEDUPLICATION') # WinError 498 (0x1F2)
    ERROR_NOT_SUPPORTED_WITH_AUDITING: Final[ThirtyTwoBits] = cached_win32error(0x01F3, 'ERROR_NOT_SUPPORTED_WITH_AUDITING') # WinError 499 (0x1F3)
    ERROR_USER_PROFILE_LOAD: Final[ThirtyTwoBits] = cached_win32error(0x01F4, 'ERROR_USER_PROFILE_LOAD') # WinError 500 (0x1F4)
    ERROR_SESSION_KEY_TOO_SHORT: Final[ThirtyTwoBits] = cached_win32error(0x01F5, 'ERROR_SESSION_KEY_TOO_SHORT') # WinError 501 (0x1F5)
    ERROR_ACCESS_DENIED_APPDATA: Final[ThirtyTwoBits] = cached_win32error(0x01F6, 'ERROR_ACCESS_DENIED_APPDATA') # WinError 502 (0x1F6)
    ERROR_NOT_SUPPORTED_WITH_MONITORING: Final[ThirtyTwoBits] = cached_win32error(0x01F7, 'ERROR_NOT_SUPPORTED_WITH_MONITORING') # WinError 503 (0x1F7)
    ERROR_NOT_SUPPORTED_WITH_SNAPSHOT: Final[ThirtyTwoBits] = cached_win32error(0x01F8, 'ERROR_NOT_SUPPORTED_WITH_SNAPSHOT') # WinError 504 (0x1F8)
    ERROR_NOT_SUPPORTED_WITH_VIRTUALIZATION: Final[ThirtyTwoBits] = cached_win32error(0x01F9, 'ERROR_NOT_SUPPORTED_WITH_VIRTUALIZATION') # WinError 505 (0x1F9)
    ERROR_BYPASSIO_FLT_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x01FA, 'ERROR_BYPASSIO_FLT_NOT_SUPPORTED') # WinError 506 (0x1FA)
    ERROR_DEVICE_RESET_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x01FB, 'ERROR_DEVICE_RESET_REQUIRED') # WinError 507 (0x1FB)
    ERROR_VOLUME_WRITE_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_win32error(0x01FC, 'ERROR_VOLUME_WRITE_ACCESS_DENIED') # WinError 508 (0x1FC)
    ERROR_NOT_SUPPORTED_WITH_CACHED_HANDLE: Final[ThirtyTwoBits] = cached_win32error(0x01FD, 'ERROR_NOT_SUPPORTED_WITH_CACHED_HANDLE') # WinError 509 (0x1FD)
    ERROR_FS_METADATA_INCONSISTENT: Final[ThirtyTwoBits] = cached_win32error(0x01FE, 'ERROR_FS_METADATA_INCONSISTENT') # WinError 510 (0x1FE)
    ERROR_BLOCK_WEAK_REFERENCE_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x01FF, 'ERROR_BLOCK_WEAK_REFERENCE_INVALID') # WinError 511 (0x1FF)
    ERROR_BLOCK_SOURCE_WEAK_REFERENCE_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x0200, 'ERROR_BLOCK_SOURCE_WEAK_REFERENCE_INVALID') # WinError 512 (0x200)
    ERROR_BLOCK_TARGET_WEAK_REFERENCE_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x0201, 'ERROR_BLOCK_TARGET_WEAK_REFERENCE_INVALID') # WinError 513 (0x201)
    ERROR_BLOCK_SHARED: Final[ThirtyTwoBits] = cached_win32error(0x0202, 'ERROR_BLOCK_SHARED') # WinError 514 (0x202)
    ERROR_VOLUME_UPGRADE_NOT_NEEDED: Final[ThirtyTwoBits] = cached_win32error(0x0203, 'ERROR_VOLUME_UPGRADE_NOT_NEEDED') # WinError 515 (0x203)
    ERROR_VOLUME_UPGRADE_PENDING: Final[ThirtyTwoBits] = cached_win32error(0x0204, 'ERROR_VOLUME_UPGRADE_PENDING') # WinError 516 (0x204)
    ERROR_VOLUME_UPGRADE_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x0205, 'ERROR_VOLUME_UPGRADE_DISABLED') # WinError 517 (0x205)
    ERROR_VOLUME_UPGRADE_DISABLED_TILL_OS_DOWNGRADE_EXPIRED: Final[ThirtyTwoBits] = cached_win32error(0x0206, 'ERROR_VOLUME_UPGRADE_DISABLED_TILL_OS_DOWNGRADE_EXPIRED') # WinError 518 (0x206)
    ERROR_ARITHMETIC_OVERFLOW: Final[ThirtyTwoBits] = cached_win32error(0x0216, 'ERROR_ARITHMETIC_OVERFLOW') # WinError 534 (0x216)
    ERROR_PIPE_CONNECTED: Final[ThirtyTwoBits] = cached_win32error(0x0217, 'ERROR_PIPE_CONNECTED') # WinError 535 (0x217)
    ERROR_PIPE_LISTENING: Final[ThirtyTwoBits] = cached_win32error(0x0218, 'ERROR_PIPE_LISTENING') # WinError 536 (0x218)
    ERROR_VERIFIER_STOP: Final[ThirtyTwoBits] = cached_win32error(0x0219, 'ERROR_VERIFIER_STOP') # WinError 537 (0x219)
    ERROR_ABIOS_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x021A, 'ERROR_ABIOS_ERROR') # WinError 538 (0x21A)
    ERROR_WX86_WARNING: Final[ThirtyTwoBits] = cached_win32error(0x021B, 'ERROR_WX86_WARNING') # WinError 539 (0x21B)
    ERROR_WX86_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x021C, 'ERROR_WX86_ERROR') # WinError 540 (0x21C)
    ERROR_TIMER_NOT_CANCELED: Final[ThirtyTwoBits] = cached_win32error(0x021D, 'ERROR_TIMER_NOT_CANCELED') # WinError 541 (0x21D)
    ERROR_UNWIND: Final[ThirtyTwoBits] = cached_win32error(0x021E, 'ERROR_UNWIND') # WinError 542 (0x21E)
    ERROR_BAD_STACK: Final[ThirtyTwoBits] = cached_win32error(0x021F, 'ERROR_BAD_STACK') # WinError 543 (0x21F)
    ERROR_INVALID_UNWIND_TARGET: Final[ThirtyTwoBits] = cached_win32error(0x0220, 'ERROR_INVALID_UNWIND_TARGET') # WinError 544 (0x220)
    ERROR_INVALID_PORT_ATTRIBUTES: Final[ThirtyTwoBits] = cached_win32error(0x0221, 'ERROR_INVALID_PORT_ATTRIBUTES') # WinError 545 (0x221)
    ERROR_PORT_MESSAGE_TOO_LONG: Final[ThirtyTwoBits] = cached_win32error(0x0222, 'ERROR_PORT_MESSAGE_TOO_LONG') # WinError 546 (0x222)
    ERROR_INVALID_QUOTA_LOWER: Final[ThirtyTwoBits] = cached_win32error(0x0223, 'ERROR_INVALID_QUOTA_LOWER') # WinError 547 (0x223)
    ERROR_DEVICE_ALREADY_ATTACHED: Final[ThirtyTwoBits] = cached_win32error(0x0224, 'ERROR_DEVICE_ALREADY_ATTACHED') # WinError 548 (0x224)
    ERROR_INSTRUCTION_MISALIGNMENT: Final[ThirtyTwoBits] = cached_win32error(0x0225, 'ERROR_INSTRUCTION_MISALIGNMENT') # WinError 549 (0x225)
    ERROR_PROFILING_NOT_STARTED: Final[ThirtyTwoBits] = cached_win32error(0x0226, 'ERROR_PROFILING_NOT_STARTED') # WinError 550 (0x226)
    ERROR_PROFILING_NOT_STOPPED: Final[ThirtyTwoBits] = cached_win32error(0x0227, 'ERROR_PROFILING_NOT_STOPPED') # WinError 551 (0x227)
    ERROR_COULD_NOT_INTERPRET: Final[ThirtyTwoBits] = cached_win32error(0x0228, 'ERROR_COULD_NOT_INTERPRET') # WinError 552 (0x228)
    ERROR_PROFILING_AT_LIMIT: Final[ThirtyTwoBits] = cached_win32error(0x0229, 'ERROR_PROFILING_AT_LIMIT') # WinError 553 (0x229)
    ERROR_CANT_WAIT: Final[ThirtyTwoBits] = cached_win32error(0x022A, 'ERROR_CANT_WAIT') # WinError 554 (0x22A)
    ERROR_CANT_TERMINATE_SELF: Final[ThirtyTwoBits] = cached_win32error(0x022B, 'ERROR_CANT_TERMINATE_SELF') # WinError 555 (0x22B)
    ERROR_UNEXPECTED_MM_CREATE_ERR: Final[ThirtyTwoBits] = cached_win32error(0x022C, 'ERROR_UNEXPECTED_MM_CREATE_ERR') # WinError 556 (0x22C)
    ERROR_UNEXPECTED_MM_MAP_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x022D, 'ERROR_UNEXPECTED_MM_MAP_ERROR') # WinError 557 (0x22D)
    ERROR_UNEXPECTED_MM_EXTEND_ERR: Final[ThirtyTwoBits] = cached_win32error(0x022E, 'ERROR_UNEXPECTED_MM_EXTEND_ERR') # WinError 558 (0x22E)
    ERROR_BAD_FUNCTION_TABLE: Final[ThirtyTwoBits] = cached_win32error(0x022F, 'ERROR_BAD_FUNCTION_TABLE') # WinError 559 (0x22F)
    ERROR_NO_GUID_TRANSLATION: Final[ThirtyTwoBits] = cached_win32error(0x0230, 'ERROR_NO_GUID_TRANSLATION') # WinError 560 (0x230)
    ERROR_INVALID_LDT_SIZE: Final[ThirtyTwoBits] = cached_win32error(0x0231, 'ERROR_INVALID_LDT_SIZE') # WinError 561 (0x231)
    ERROR_INVALID_LDT_OFFSET: Final[ThirtyTwoBits] = cached_win32error(0x0233, 'ERROR_INVALID_LDT_OFFSET') # WinError 563 (0x233)
    ERROR_INVALID_LDT_DESCRIPTOR: Final[ThirtyTwoBits] = cached_win32error(0x0234, 'ERROR_INVALID_LDT_DESCRIPTOR') # WinError 564 (0x234)
    ERROR_TOO_MANY_THREADS: Final[ThirtyTwoBits] = cached_win32error(0x0235, 'ERROR_TOO_MANY_THREADS') # WinError 565 (0x235)
    ERROR_THREAD_NOT_IN_PROCESS: Final[ThirtyTwoBits] = cached_win32error(0x0236, 'ERROR_THREAD_NOT_IN_PROCESS') # WinError 566 (0x236)
    ERROR_PAGEFILE_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x0237, 'ERROR_PAGEFILE_QUOTA_EXCEEDED') # WinError 567 (0x237)
    ERROR_LOGON_SERVER_CONFLICT: Final[ThirtyTwoBits] = cached_win32error(0x0238, 'ERROR_LOGON_SERVER_CONFLICT') # WinError 568 (0x238)
    ERROR_SYNCHRONIZATION_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x0239, 'ERROR_SYNCHRONIZATION_REQUIRED') # WinError 569 (0x239)
    ERROR_NET_OPEN_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x023A, 'ERROR_NET_OPEN_FAILED') # WinError 570 (0x23A)
    ERROR_IO_PRIVILEGE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x023B, 'ERROR_IO_PRIVILEGE_FAILED') # WinError 571 (0x23B)
    ERROR_CONTROL_C_EXIT: Final[ThirtyTwoBits] = cached_win32error(0x023C, 'ERROR_CONTROL_C_EXIT') # WinError 572 (0x23C)
    ERROR_MISSING_SYSTEMFILE: Final[ThirtyTwoBits] = cached_win32error(0x023D, 'ERROR_MISSING_SYSTEMFILE') # WinError 573 (0x23D)
    ERROR_UNHANDLED_EXCEPTION: Final[ThirtyTwoBits] = cached_win32error(0x023E, 'ERROR_UNHANDLED_EXCEPTION') # WinError 574 (0x23E)
    ERROR_APP_INIT_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x023F, 'ERROR_APP_INIT_FAILURE') # WinError 575 (0x23F)
    ERROR_PAGEFILE_CREATE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x0240, 'ERROR_PAGEFILE_CREATE_FAILED') # WinError 576 (0x240)
    ERROR_INVALID_IMAGE_HASH: Final[ThirtyTwoBits] = cached_win32error(0x0241, 'ERROR_INVALID_IMAGE_HASH') # WinError 577 (0x241)
    ERROR_NO_PAGEFILE: Final[ThirtyTwoBits] = cached_win32error(0x0242, 'ERROR_NO_PAGEFILE') # WinError 578 (0x242)
    ERROR_ILLEGAL_FLOAT_CONTEXT: Final[ThirtyTwoBits] = cached_win32error(0x0243, 'ERROR_ILLEGAL_FLOAT_CONTEXT') # WinError 579 (0x243)
    ERROR_NO_EVENT_PAIR: Final[ThirtyTwoBits] = cached_win32error(0x0244, 'ERROR_NO_EVENT_PAIR') # WinError 580 (0x244)
    ERROR_DOMAIN_CTRLR_CONFIG_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x0245, 'ERROR_DOMAIN_CTRLR_CONFIG_ERROR') # WinError 581 (0x245)
    ERROR_ILLEGAL_CHARACTER: Final[ThirtyTwoBits] = cached_win32error(0x0246, 'ERROR_ILLEGAL_CHARACTER') # WinError 582 (0x246)
    ERROR_UNDEFINED_CHARACTER: Final[ThirtyTwoBits] = cached_win32error(0x0247, 'ERROR_UNDEFINED_CHARACTER') # WinError 583 (0x247)
    ERROR_FLOPPY_VOLUME: Final[ThirtyTwoBits] = cached_win32error(0x0248, 'ERROR_FLOPPY_VOLUME') # WinError 584 (0x248)
    ERROR_BIOS_FAILED_TO_CONNECT_INTERRUPT: Final[ThirtyTwoBits] = cached_win32error(0x0249, 'ERROR_BIOS_FAILED_TO_CONNECT_INTERRUPT') # WinError 585 (0x249)
    ERROR_BACKUP_CONTROLLER: Final[ThirtyTwoBits] = cached_win32error(0x024A, 'ERROR_BACKUP_CONTROLLER') # WinError 586 (0x24A)
    ERROR_MUTANT_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x024B, 'ERROR_MUTANT_LIMIT_EXCEEDED') # WinError 587 (0x24B)
    ERROR_FS_DRIVER_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x024C, 'ERROR_FS_DRIVER_REQUIRED') # WinError 588 (0x24C)
    ERROR_CANNOT_LOAD_REGISTRY_FILE: Final[ThirtyTwoBits] = cached_win32error(0x024D, 'ERROR_CANNOT_LOAD_REGISTRY_FILE') # WinError 589 (0x24D)
    ERROR_DEBUG_ATTACH_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x024E, 'ERROR_DEBUG_ATTACH_FAILED') # WinError 590 (0x24E)
    ERROR_SYSTEM_PROCESS_TERMINATED: Final[ThirtyTwoBits] = cached_win32error(0x024F, 'ERROR_SYSTEM_PROCESS_TERMINATED') # WinError 591 (0x24F)
    ERROR_DATA_NOT_ACCEPTED: Final[ThirtyTwoBits] = cached_win32error(0x0250, 'ERROR_DATA_NOT_ACCEPTED') # WinError 592 (0x250)
    ERROR_VDM_HARD_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x0251, 'ERROR_VDM_HARD_ERROR') # WinError 593 (0x251)
    ERROR_DRIVER_CANCEL_TIMEOUT: Final[ThirtyTwoBits] = cached_win32error(0x0252, 'ERROR_DRIVER_CANCEL_TIMEOUT') # WinError 594 (0x252)
    ERROR_REPLY_MESSAGE_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x0253, 'ERROR_REPLY_MESSAGE_MISMATCH') # WinError 595 (0x253)
    ERROR_LOST_WRITEBEHIND_DATA: Final[ThirtyTwoBits] = cached_win32error(0x0254, 'ERROR_LOST_WRITEBEHIND_DATA') # WinError 596 (0x254)
    ERROR_CLIENT_SERVER_PARAMETERS_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x0255, 'ERROR_CLIENT_SERVER_PARAMETERS_INVALID') # WinError 597 (0x255)
    ERROR_NOT_TINY_STREAM: Final[ThirtyTwoBits] = cached_win32error(0x0256, 'ERROR_NOT_TINY_STREAM') # WinError 598 (0x256)
    ERROR_STACK_OVERFLOW_READ: Final[ThirtyTwoBits] = cached_win32error(0x0257, 'ERROR_STACK_OVERFLOW_READ') # WinError 599 (0x257)
    ERROR_CONVERT_TO_LARGE: Final[ThirtyTwoBits] = cached_win32error(0x0258, 'ERROR_CONVERT_TO_LARGE') # WinError 600 (0x258)
    ERROR_FOUND_OUT_OF_SCOPE: Final[ThirtyTwoBits] = cached_win32error(0x0259, 'ERROR_FOUND_OUT_OF_SCOPE') # WinError 601 (0x259)
    ERROR_ALLOCATE_BUCKET: Final[ThirtyTwoBits] = cached_win32error(0x025A, 'ERROR_ALLOCATE_BUCKET') # WinError 602 (0x25A)
    ERROR_MARSHALL_OVERFLOW: Final[ThirtyTwoBits] = cached_win32error(0x025B, 'ERROR_MARSHALL_OVERFLOW') # WinError 603 (0x25B)
    ERROR_INVALID_VARIANT: Final[ThirtyTwoBits] = cached_win32error(0x025C, 'ERROR_INVALID_VARIANT') # WinError 604 (0x25C)
    ERROR_BAD_COMPRESSION_BUFFER: Final[ThirtyTwoBits] = cached_win32error(0x025D, 'ERROR_BAD_COMPRESSION_BUFFER') # WinError 605 (0x25D)
    ERROR_AUDIT_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x025E, 'ERROR_AUDIT_FAILED') # WinError 606 (0x25E)
    ERROR_TIMER_RESOLUTION_NOT_SET: Final[ThirtyTwoBits] = cached_win32error(0x025F, 'ERROR_TIMER_RESOLUTION_NOT_SET') # WinError 607 (0x25F)
    ERROR_INSUFFICIENT_LOGON_INFO: Final[ThirtyTwoBits] = cached_win32error(0x0260, 'ERROR_INSUFFICIENT_LOGON_INFO') # WinError 608 (0x260)
    ERROR_BAD_DLL_ENTRYPOINT: Final[ThirtyTwoBits] = cached_win32error(0x0261, 'ERROR_BAD_DLL_ENTRYPOINT') # WinError 609 (0x261)
    ERROR_BAD_SERVICE_ENTRYPOINT: Final[ThirtyTwoBits] = cached_win32error(0x0262, 'ERROR_BAD_SERVICE_ENTRYPOINT') # WinError 610 (0x262)
    ERROR_IP_ADDRESS_CONFLICT1: Final[ThirtyTwoBits] = cached_win32error(0x0263, 'ERROR_IP_ADDRESS_CONFLICT1') # WinError 611 (0x263)
    ERROR_IP_ADDRESS_CONFLICT2: Final[ThirtyTwoBits] = cached_win32error(0x0264, 'ERROR_IP_ADDRESS_CONFLICT2') # WinError 612 (0x264)
    ERROR_REGISTRY_QUOTA_LIMIT: Final[ThirtyTwoBits] = cached_win32error(0x0265, 'ERROR_REGISTRY_QUOTA_LIMIT') # WinError 613 (0x265)
    ERROR_NO_CALLBACK_ACTIVE: Final[ThirtyTwoBits] = cached_win32error(0x0266, 'ERROR_NO_CALLBACK_ACTIVE') # WinError 614 (0x266)
    ERROR_PWD_TOO_SHORT: Final[ThirtyTwoBits] = cached_win32error(0x0267, 'ERROR_PWD_TOO_SHORT') # WinError 615 (0x267)
    ERROR_PWD_TOO_RECENT: Final[ThirtyTwoBits] = cached_win32error(0x0268, 'ERROR_PWD_TOO_RECENT') # WinError 616 (0x268)
    ERROR_PWD_HISTORY_CONFLICT: Final[ThirtyTwoBits] = cached_win32error(0x0269, 'ERROR_PWD_HISTORY_CONFLICT') # WinError 617 (0x269)
    ERROR_UNSUPPORTED_COMPRESSION: Final[ThirtyTwoBits] = cached_win32error(0x026A, 'ERROR_UNSUPPORTED_COMPRESSION') # WinError 618 (0x26A)
    ERROR_INVALID_HW_PROFILE: Final[ThirtyTwoBits] = cached_win32error(0x026B, 'ERROR_INVALID_HW_PROFILE') # WinError 619 (0x26B)
    ERROR_INVALID_PLUGPLAY_DEVICE_PATH: Final[ThirtyTwoBits] = cached_win32error(0x026C, 'ERROR_INVALID_PLUGPLAY_DEVICE_PATH') # WinError 620 (0x26C)
    ERROR_QUOTA_LIST_INCONSISTENT: Final[ThirtyTwoBits] = cached_win32error(0x026D, 'ERROR_QUOTA_LIST_INCONSISTENT') # WinError 621 (0x26D)
    ERROR_EVALUATION_EXPIRATION: Final[ThirtyTwoBits] = cached_win32error(0x026E, 'ERROR_EVALUATION_EXPIRATION') # WinError 622 (0x26E)
    ERROR_ILLEGAL_DLL_RELOCATION: Final[ThirtyTwoBits] = cached_win32error(0x026F, 'ERROR_ILLEGAL_DLL_RELOCATION') # WinError 623 (0x26F)
    ERROR_DLL_INIT_FAILED_LOGOFF: Final[ThirtyTwoBits] = cached_win32error(0x0270, 'ERROR_DLL_INIT_FAILED_LOGOFF') # WinError 624 (0x270)
    ERROR_VALIDATE_CONTINUE: Final[ThirtyTwoBits] = cached_win32error(0x0271, 'ERROR_VALIDATE_CONTINUE') # WinError 625 (0x271)
    ERROR_NO_MORE_MATCHES: Final[ThirtyTwoBits] = cached_win32error(0x0272, 'ERROR_NO_MORE_MATCHES') # WinError 626 (0x272)
    ERROR_RANGE_LIST_CONFLICT: Final[ThirtyTwoBits] = cached_win32error(0x0273, 'ERROR_RANGE_LIST_CONFLICT') # WinError 627 (0x273)
    ERROR_SERVER_SID_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x0274, 'ERROR_SERVER_SID_MISMATCH') # WinError 628 (0x274)
    ERROR_CANT_ENABLE_DENY_ONLY: Final[ThirtyTwoBits] = cached_win32error(0x0275, 'ERROR_CANT_ENABLE_DENY_ONLY') # WinError 629 (0x275)
    ERROR_FLOAT_MULTIPLE_FAULTS: Final[ThirtyTwoBits] = cached_win32error(0x0276, 'ERROR_FLOAT_MULTIPLE_FAULTS') # WinError 630 (0x276)
    ERROR_FLOAT_MULTIPLE_TRAPS: Final[ThirtyTwoBits] = cached_win32error(0x0277, 'ERROR_FLOAT_MULTIPLE_TRAPS') # WinError 631 (0x277)
    ERROR_NOINTERFACE: Final[ThirtyTwoBits] = cached_win32error(0x0278, 'ERROR_NOINTERFACE') # WinError 632 (0x278)
    ERROR_DRIVER_FAILED_SLEEP: Final[ThirtyTwoBits] = cached_win32error(0x0279, 'ERROR_DRIVER_FAILED_SLEEP') # WinError 633 (0x279)
    ERROR_CORRUPT_SYSTEM_FILE: Final[ThirtyTwoBits] = cached_win32error(0x027A, 'ERROR_CORRUPT_SYSTEM_FILE') # WinError 634 (0x27A)
    ERROR_COMMITMENT_MINIMUM: Final[ThirtyTwoBits] = cached_win32error(0x027B, 'ERROR_COMMITMENT_MINIMUM') # WinError 635 (0x27B)
    ERROR_PNP_RESTART_ENUMERATION: Final[ThirtyTwoBits] = cached_win32error(0x027C, 'ERROR_PNP_RESTART_ENUMERATION') # WinError 636 (0x27C)
    ERROR_SYSTEM_IMAGE_BAD_SIGNATURE: Final[ThirtyTwoBits] = cached_win32error(0x027D, 'ERROR_SYSTEM_IMAGE_BAD_SIGNATURE') # WinError 637 (0x27D)
    ERROR_PNP_REBOOT_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x027E, 'ERROR_PNP_REBOOT_REQUIRED') # WinError 638 (0x27E)
    ERROR_INSUFFICIENT_POWER: Final[ThirtyTwoBits] = cached_win32error(0x027F, 'ERROR_INSUFFICIENT_POWER') # WinError 639 (0x27F)
    ERROR_MULTIPLE_FAULT_VIOLATION: Final[ThirtyTwoBits] = cached_win32error(0x0280, 'ERROR_MULTIPLE_FAULT_VIOLATION') # WinError 640 (0x280)
    ERROR_SYSTEM_SHUTDOWN: Final[ThirtyTwoBits] = cached_win32error(0x0281, 'ERROR_SYSTEM_SHUTDOWN') # WinError 641 (0x281)
    ERROR_PORT_NOT_SET: Final[ThirtyTwoBits] = cached_win32error(0x0282, 'ERROR_PORT_NOT_SET') # WinError 642 (0x282)
    ERROR_DS_VERSION_CHECK_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x0283, 'ERROR_DS_VERSION_CHECK_FAILURE') # WinError 643 (0x283)
    ERROR_RANGE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x0284, 'ERROR_RANGE_NOT_FOUND') # WinError 644 (0x284)
    ERROR_NOT_SAFE_MODE_DRIVER: Final[ThirtyTwoBits] = cached_win32error(0x0286, 'ERROR_NOT_SAFE_MODE_DRIVER') # WinError 646 (0x286)
    ERROR_FAILED_DRIVER_ENTRY: Final[ThirtyTwoBits] = cached_win32error(0x0287, 'ERROR_FAILED_DRIVER_ENTRY') # WinError 647 (0x287)
    ERROR_DEVICE_ENUMERATION_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x0288, 'ERROR_DEVICE_ENUMERATION_ERROR') # WinError 648 (0x288)
    ERROR_MOUNT_POINT_NOT_RESOLVED: Final[ThirtyTwoBits] = cached_win32error(0x0289, 'ERROR_MOUNT_POINT_NOT_RESOLVED') # WinError 649 (0x289)
    ERROR_INVALID_DEVICE_OBJECT_PARAMETER: Final[ThirtyTwoBits] = cached_win32error(0x028A, 'ERROR_INVALID_DEVICE_OBJECT_PARAMETER') # WinError 650 (0x28A)
    ERROR_MCA_OCCURED: Final[ThirtyTwoBits] = cached_win32error(0x028B, 'ERROR_MCA_OCCURED') # WinError 651 (0x28B)
    ERROR_DRIVER_DATABASE_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x028C, 'ERROR_DRIVER_DATABASE_ERROR') # WinError 652 (0x28C)
    ERROR_SYSTEM_HIVE_TOO_LARGE: Final[ThirtyTwoBits] = cached_win32error(0x028D, 'ERROR_SYSTEM_HIVE_TOO_LARGE') # WinError 653 (0x28D)
    ERROR_DRIVER_FAILED_PRIOR_UNLOAD: Final[ThirtyTwoBits] = cached_win32error(0x028E, 'ERROR_DRIVER_FAILED_PRIOR_UNLOAD') # WinError 654 (0x28E)
    ERROR_VOLSNAP_PREPARE_HIBERNATE: Final[ThirtyTwoBits] = cached_win32error(0x028F, 'ERROR_VOLSNAP_PREPARE_HIBERNATE') # WinError 655 (0x28F)
    ERROR_HIBERNATION_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x0290, 'ERROR_HIBERNATION_FAILURE') # WinError 656 (0x290)
    ERROR_PWD_TOO_LONG: Final[ThirtyTwoBits] = cached_win32error(0x0291, 'ERROR_PWD_TOO_LONG') # WinError 657 (0x291)
    ERROR_FILE_SYSTEM_LIMITATION: Final[ThirtyTwoBits] = cached_win32error(0x0299, 'ERROR_FILE_SYSTEM_LIMITATION') # WinError 665 (0x299)
    ERROR_ASSERTION_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x029C, 'ERROR_ASSERTION_FAILURE') # WinError 668 (0x29C)
    ERROR_ACPI_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x029D, 'ERROR_ACPI_ERROR') # WinError 669 (0x29D)
    ERROR_WOW_ASSERTION: Final[ThirtyTwoBits] = cached_win32error(0x029E, 'ERROR_WOW_ASSERTION') # WinError 670 (0x29E)
    ERROR_PNP_BAD_MPS_TABLE: Final[ThirtyTwoBits] = cached_win32error(0x029F, 'ERROR_PNP_BAD_MPS_TABLE') # WinError 671 (0x29F)
    ERROR_PNP_TRANSLATION_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x02A0, 'ERROR_PNP_TRANSLATION_FAILED') # WinError 672 (0x2A0)
    ERROR_PNP_IRQ_TRANSLATION_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x02A1, 'ERROR_PNP_IRQ_TRANSLATION_FAILED') # WinError 673 (0x2A1)
    ERROR_PNP_INVALID_ID: Final[ThirtyTwoBits] = cached_win32error(0x02A2, 'ERROR_PNP_INVALID_ID') # WinError 674 (0x2A2)
    ERROR_WAKE_SYSTEM_DEBUGGER: Final[ThirtyTwoBits] = cached_win32error(0x02A3, 'ERROR_WAKE_SYSTEM_DEBUGGER') # WinError 675 (0x2A3)
    ERROR_HANDLES_CLOSED: Final[ThirtyTwoBits] = cached_win32error(0x02A4, 'ERROR_HANDLES_CLOSED') # WinError 676 (0x2A4)
    ERROR_EXTRANEOUS_INFORMATION: Final[ThirtyTwoBits] = cached_win32error(0x02A5, 'ERROR_EXTRANEOUS_INFORMATION') # WinError 677 (0x2A5)
    ERROR_RXACT_COMMIT_NECESSARY: Final[ThirtyTwoBits] = cached_win32error(0x02A6, 'ERROR_RXACT_COMMIT_NECESSARY') # WinError 678 (0x2A6)
    ERROR_MEDIA_CHECK: Final[ThirtyTwoBits] = cached_win32error(0x02A7, 'ERROR_MEDIA_CHECK') # WinError 679 (0x2A7)
    ERROR_GUID_SUBSTITUTION_MADE: Final[ThirtyTwoBits] = cached_win32error(0x02A8, 'ERROR_GUID_SUBSTITUTION_MADE') # WinError 680 (0x2A8)
    ERROR_STOPPED_ON_SYMLINK: Final[ThirtyTwoBits] = cached_win32error(0x02A9, 'ERROR_STOPPED_ON_SYMLINK') # WinError 681 (0x2A9)
    ERROR_LONGJUMP: Final[ThirtyTwoBits] = cached_win32error(0x02AA, 'ERROR_LONGJUMP') # WinError 682 (0x2AA)
    ERROR_PLUGPLAY_QUERY_VETOED: Final[ThirtyTwoBits] = cached_win32error(0x02AB, 'ERROR_PLUGPLAY_QUERY_VETOED') # WinError 683 (0x2AB)
    ERROR_UNWIND_CONSOLIDATE: Final[ThirtyTwoBits] = cached_win32error(0x02AC, 'ERROR_UNWIND_CONSOLIDATE') # WinError 684 (0x2AC)
    ERROR_REGISTRY_HIVE_RECOVERED: Final[ThirtyTwoBits] = cached_win32error(0x02AD, 'ERROR_REGISTRY_HIVE_RECOVERED') # WinError 685 (0x2AD)
    ERROR_DLL_MIGHT_BE_INSECURE: Final[ThirtyTwoBits] = cached_win32error(0x02AE, 'ERROR_DLL_MIGHT_BE_INSECURE') # WinError 686 (0x2AE)
    ERROR_DLL_MIGHT_BE_INCOMPATIBLE: Final[ThirtyTwoBits] = cached_win32error(0x02AF, 'ERROR_DLL_MIGHT_BE_INCOMPATIBLE') # WinError 687 (0x2AF)
    ERROR_DBG_EXCEPTION_NOT_HANDLED: Final[ThirtyTwoBits] = cached_win32error(0x02B0, 'ERROR_DBG_EXCEPTION_NOT_HANDLED') # WinError 688 (0x2B0)
    ERROR_DBG_REPLY_LATER: Final[ThirtyTwoBits] = cached_win32error(0x02B1, 'ERROR_DBG_REPLY_LATER') # WinError 689 (0x2B1)
    ERROR_DBG_UNABLE_TO_PROVIDE_HANDLE: Final[ThirtyTwoBits] = cached_win32error(0x02B2, 'ERROR_DBG_UNABLE_TO_PROVIDE_HANDLE') # WinError 690 (0x2B2)
    ERROR_DBG_TERMINATE_THREAD: Final[ThirtyTwoBits] = cached_win32error(0x02B3, 'ERROR_DBG_TERMINATE_THREAD') # WinError 691 (0x2B3)
    ERROR_DBG_TERMINATE_PROCESS: Final[ThirtyTwoBits] = cached_win32error(0x02B4, 'ERROR_DBG_TERMINATE_PROCESS') # WinError 692 (0x2B4)
    ERROR_DBG_CONTROL_C: Final[ThirtyTwoBits] = cached_win32error(0x02B5, 'ERROR_DBG_CONTROL_C') # WinError 693 (0x2B5)
    ERROR_DBG_PRINTEXCEPTION_C: Final[ThirtyTwoBits] = cached_win32error(0x02B6, 'ERROR_DBG_PRINTEXCEPTION_C') # WinError 694 (0x2B6)
    ERROR_DBG_RIPEXCEPTION: Final[ThirtyTwoBits] = cached_win32error(0x02B7, 'ERROR_DBG_RIPEXCEPTION') # WinError 695 (0x2B7)
    ERROR_DBG_CONTROL_BREAK: Final[ThirtyTwoBits] = cached_win32error(0x02B8, 'ERROR_DBG_CONTROL_BREAK') # WinError 696 (0x2B8)
    ERROR_DBG_COMMAND_EXCEPTION: Final[ThirtyTwoBits] = cached_win32error(0x02B9, 'ERROR_DBG_COMMAND_EXCEPTION') # WinError 697 (0x2B9)
    ERROR_OBJECT_NAME_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x02BA, 'ERROR_OBJECT_NAME_EXISTS') # WinError 698 (0x2BA)
    ERROR_THREAD_WAS_SUSPENDED: Final[ThirtyTwoBits] = cached_win32error(0x02BB, 'ERROR_THREAD_WAS_SUSPENDED') # WinError 699 (0x2BB)
    ERROR_IMAGE_NOT_AT_BASE: Final[ThirtyTwoBits] = cached_win32error(0x02BC, 'ERROR_IMAGE_NOT_AT_BASE') # WinError 700 (0x2BC)
    ERROR_RXACT_STATE_CREATED: Final[ThirtyTwoBits] = cached_win32error(0x02BD, 'ERROR_RXACT_STATE_CREATED') # WinError 701 (0x2BD)
    ERROR_SEGMENT_NOTIFICATION: Final[ThirtyTwoBits] = cached_win32error(0x02BE, 'ERROR_SEGMENT_NOTIFICATION') # WinError 702 (0x2BE)
    ERROR_BAD_CURRENT_DIRECTORY: Final[ThirtyTwoBits] = cached_win32error(0x02BF, 'ERROR_BAD_CURRENT_DIRECTORY') # WinError 703 (0x2BF)
    ERROR_FT_READ_RECOVERY_FROM_BACKUP: Final[ThirtyTwoBits] = cached_win32error(0x02C0, 'ERROR_FT_READ_RECOVERY_FROM_BACKUP') # WinError 704 (0x2C0)
    ERROR_FT_WRITE_RECOVERY: Final[ThirtyTwoBits] = cached_win32error(0x02C1, 'ERROR_FT_WRITE_RECOVERY') # WinError 705 (0x2C1)
    ERROR_IMAGE_MACHINE_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x02C2, 'ERROR_IMAGE_MACHINE_TYPE_MISMATCH') # WinError 706 (0x2C2)
    ERROR_RECEIVE_PARTIAL: Final[ThirtyTwoBits] = cached_win32error(0x02C3, 'ERROR_RECEIVE_PARTIAL') # WinError 707 (0x2C3)
    ERROR_RECEIVE_EXPEDITED: Final[ThirtyTwoBits] = cached_win32error(0x02C4, 'ERROR_RECEIVE_EXPEDITED') # WinError 708 (0x2C4)
    ERROR_RECEIVE_PARTIAL_EXPEDITED: Final[ThirtyTwoBits] = cached_win32error(0x02C5, 'ERROR_RECEIVE_PARTIAL_EXPEDITED') # WinError 709 (0x2C5)
    ERROR_EVENT_DONE: Final[ThirtyTwoBits] = cached_win32error(0x02C6, 'ERROR_EVENT_DONE') # WinError 710 (0x2C6)
    ERROR_EVENT_PENDING: Final[ThirtyTwoBits] = cached_win32error(0x02C7, 'ERROR_EVENT_PENDING') # WinError 711 (0x2C7)
    ERROR_CHECKING_FILE_SYSTEM: Final[ThirtyTwoBits] = cached_win32error(0x02C8, 'ERROR_CHECKING_FILE_SYSTEM') # WinError 712 (0x2C8)
    ERROR_FATAL_APP_EXIT: Final[ThirtyTwoBits] = cached_win32error(0x02C9, 'ERROR_FATAL_APP_EXIT') # WinError 713 (0x2C9)
    ERROR_PREDEFINED_HANDLE: Final[ThirtyTwoBits] = cached_win32error(0x02CA, 'ERROR_PREDEFINED_HANDLE') # WinError 714 (0x2CA)
    ERROR_WAS_UNLOCKED: Final[ThirtyTwoBits] = cached_win32error(0x02CB, 'ERROR_WAS_UNLOCKED') # WinError 715 (0x2CB)
    ERROR_SERVICE_NOTIFICATION: Final[ThirtyTwoBits] = cached_win32error(0x02CC, 'ERROR_SERVICE_NOTIFICATION') # WinError 716 (0x2CC)
    ERROR_WAS_LOCKED: Final[ThirtyTwoBits] = cached_win32error(0x02CD, 'ERROR_WAS_LOCKED') # WinError 717 (0x2CD)
    ERROR_LOG_HARD_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x02CE, 'ERROR_LOG_HARD_ERROR') # WinError 718 (0x2CE)
    ERROR_ALREADY_WIN32: Final[ThirtyTwoBits] = cached_win32error(0x02CF, 'ERROR_ALREADY_WIN32') # WinError 719 (0x2CF)
    ERROR_IMAGE_MACHINE_TYPE_MISMATCH_EXE: Final[ThirtyTwoBits] = cached_win32error(0x02D0, 'ERROR_IMAGE_MACHINE_TYPE_MISMATCH_EXE') # WinError 720 (0x2D0)
    ERROR_NO_YIELD_PERFORMED: Final[ThirtyTwoBits] = cached_win32error(0x02D1, 'ERROR_NO_YIELD_PERFORMED') # WinError 721 (0x2D1)
    ERROR_TIMER_RESUME_IGNORED: Final[ThirtyTwoBits] = cached_win32error(0x02D2, 'ERROR_TIMER_RESUME_IGNORED') # WinError 722 (0x2D2)
    ERROR_ARBITRATION_UNHANDLED: Final[ThirtyTwoBits] = cached_win32error(0x02D3, 'ERROR_ARBITRATION_UNHANDLED') # WinError 723 (0x2D3)
    ERROR_CARDBUS_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x02D4, 'ERROR_CARDBUS_NOT_SUPPORTED') # WinError 724 (0x2D4)
    ERROR_MP_PROCESSOR_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x02D5, 'ERROR_MP_PROCESSOR_MISMATCH') # WinError 725 (0x2D5)
    ERROR_HIBERNATED: Final[ThirtyTwoBits] = cached_win32error(0x02D6, 'ERROR_HIBERNATED') # WinError 726 (0x2D6)
    ERROR_RESUME_HIBERNATION: Final[ThirtyTwoBits] = cached_win32error(0x02D7, 'ERROR_RESUME_HIBERNATION') # WinError 727 (0x2D7)
    ERROR_FIRMWARE_UPDATED: Final[ThirtyTwoBits] = cached_win32error(0x02D8, 'ERROR_FIRMWARE_UPDATED') # WinError 728 (0x2D8)
    ERROR_DRIVERS_LEAKING_LOCKED_PAGES: Final[ThirtyTwoBits] = cached_win32error(0x02D9, 'ERROR_DRIVERS_LEAKING_LOCKED_PAGES') # WinError 729 (0x2D9)
    ERROR_WAKE_SYSTEM: Final[ThirtyTwoBits] = cached_win32error(0x02DA, 'ERROR_WAKE_SYSTEM') # WinError 730 (0x2DA)
    ERROR_WAIT_1: Final[ThirtyTwoBits] = cached_win32error(0x02DB, 'ERROR_WAIT_1') # WinError 731 (0x2DB)
    ERROR_WAIT_2: Final[ThirtyTwoBits] = cached_win32error(0x02DC, 'ERROR_WAIT_2') # WinError 732 (0x2DC)
    ERROR_WAIT_3: Final[ThirtyTwoBits] = cached_win32error(0x02DD, 'ERROR_WAIT_3') # WinError 733 (0x2DD)
    ERROR_WAIT_63: Final[ThirtyTwoBits] = cached_win32error(0x02DE, 'ERROR_WAIT_63') # WinError 734 (0x2DE)
    ERROR_ABANDONED_WAIT_0: Final[ThirtyTwoBits] = cached_win32error(0x02DF, 'ERROR_ABANDONED_WAIT_0') # WinError 735 (0x2DF)
    ERROR_ABANDONED_WAIT_63: Final[ThirtyTwoBits] = cached_win32error(0x02E0, 'ERROR_ABANDONED_WAIT_63') # WinError 736 (0x2E0)
    ERROR_USER_APC: Final[ThirtyTwoBits] = cached_win32error(0x02E1, 'ERROR_USER_APC') # WinError 737 (0x2E1)
    ERROR_KERNEL_APC: Final[ThirtyTwoBits] = cached_win32error(0x02E2, 'ERROR_KERNEL_APC') # WinError 738 (0x2E2)
    ERROR_ALERTED: Final[ThirtyTwoBits] = cached_win32error(0x02E3, 'ERROR_ALERTED') # WinError 739 (0x2E3)
    ERROR_ELEVATION_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x02E4, 'ERROR_ELEVATION_REQUIRED') # WinError 740 (0x2E4)
    ERROR_REPARSE: Final[ThirtyTwoBits] = cached_win32error(0x02E5, 'ERROR_REPARSE') # WinError 741 (0x2E5)
    ERROR_OPLOCK_BREAK_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x02E6, 'ERROR_OPLOCK_BREAK_IN_PROGRESS') # WinError 742 (0x2E6)
    ERROR_VOLUME_MOUNTED: Final[ThirtyTwoBits] = cached_win32error(0x02E7, 'ERROR_VOLUME_MOUNTED') # WinError 743 (0x2E7)
    ERROR_RXACT_COMMITTED: Final[ThirtyTwoBits] = cached_win32error(0x02E8, 'ERROR_RXACT_COMMITTED') # WinError 744 (0x2E8)
    ERROR_NOTIFY_CLEANUP: Final[ThirtyTwoBits] = cached_win32error(0x02E9, 'ERROR_NOTIFY_CLEANUP') # WinError 745 (0x2E9)
    ERROR_PRIMARY_TRANSPORT_CONNECT_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x02EA, 'ERROR_PRIMARY_TRANSPORT_CONNECT_FAILED') # WinError 746 (0x2EA)
    ERROR_PAGE_FAULT_TRANSITION: Final[ThirtyTwoBits] = cached_win32error(0x02EB, 'ERROR_PAGE_FAULT_TRANSITION') # WinError 747 (0x2EB)
    ERROR_PAGE_FAULT_DEMAND_ZERO: Final[ThirtyTwoBits] = cached_win32error(0x02EC, 'ERROR_PAGE_FAULT_DEMAND_ZERO') # WinError 748 (0x2EC)
    ERROR_PAGE_FAULT_COPY_ON_WRITE: Final[ThirtyTwoBits] = cached_win32error(0x02ED, 'ERROR_PAGE_FAULT_COPY_ON_WRITE') # WinError 749 (0x2ED)
    ERROR_PAGE_FAULT_GUARD_PAGE: Final[ThirtyTwoBits] = cached_win32error(0x02EE, 'ERROR_PAGE_FAULT_GUARD_PAGE') # WinError 750 (0x2EE)
    ERROR_PAGE_FAULT_PAGING_FILE: Final[ThirtyTwoBits] = cached_win32error(0x02EF, 'ERROR_PAGE_FAULT_PAGING_FILE') # WinError 751 (0x2EF)
    ERROR_CACHE_PAGE_LOCKED: Final[ThirtyTwoBits] = cached_win32error(0x02F0, 'ERROR_CACHE_PAGE_LOCKED') # WinError 752 (0x2F0)
    ERROR_CRASH_DUMP: Final[ThirtyTwoBits] = cached_win32error(0x02F1, 'ERROR_CRASH_DUMP') # WinError 753 (0x2F1)
    ERROR_BUFFER_ALL_ZEROS: Final[ThirtyTwoBits] = cached_win32error(0x02F2, 'ERROR_BUFFER_ALL_ZEROS') # WinError 754 (0x2F2)
    ERROR_REPARSE_OBJECT: Final[ThirtyTwoBits] = cached_win32error(0x02F3, 'ERROR_REPARSE_OBJECT') # WinError 755 (0x2F3)
    ERROR_RESOURCE_REQUIREMENTS_CHANGED: Final[ThirtyTwoBits] = cached_win32error(0x02F4, 'ERROR_RESOURCE_REQUIREMENTS_CHANGED') # WinError 756 (0x2F4)
    ERROR_TRANSLATION_COMPLETE: Final[ThirtyTwoBits] = cached_win32error(0x02F5, 'ERROR_TRANSLATION_COMPLETE') # WinError 757 (0x2F5)
    ERROR_NOTHING_TO_TERMINATE: Final[ThirtyTwoBits] = cached_win32error(0x02F6, 'ERROR_NOTHING_TO_TERMINATE') # WinError 758 (0x2F6)
    ERROR_PROCESS_NOT_IN_JOB: Final[ThirtyTwoBits] = cached_win32error(0x02F7, 'ERROR_PROCESS_NOT_IN_JOB') # WinError 759 (0x2F7)
    ERROR_PROCESS_IN_JOB: Final[ThirtyTwoBits] = cached_win32error(0x02F8, 'ERROR_PROCESS_IN_JOB') # WinError 760 (0x2F8)
    ERROR_VOLSNAP_HIBERNATE_READY: Final[ThirtyTwoBits] = cached_win32error(0x02F9, 'ERROR_VOLSNAP_HIBERNATE_READY') # WinError 761 (0x2F9)
    ERROR_FSFILTER_OP_COMPLETED_SUCCESSFULLY: Final[ThirtyTwoBits] = cached_win32error(0x02FA, 'ERROR_FSFILTER_OP_COMPLETED_SUCCESSFULLY') # WinError 762 (0x2FA)
    ERROR_INTERRUPT_VECTOR_ALREADY_CONNECTED: Final[ThirtyTwoBits] = cached_win32error(0x02FB, 'ERROR_INTERRUPT_VECTOR_ALREADY_CONNECTED') # WinError 763 (0x2FB)
    ERROR_INTERRUPT_STILL_CONNECTED: Final[ThirtyTwoBits] = cached_win32error(0x02FC, 'ERROR_INTERRUPT_STILL_CONNECTED') # WinError 764 (0x2FC)
    ERROR_WAIT_FOR_OPLOCK: Final[ThirtyTwoBits] = cached_win32error(0x02FD, 'ERROR_WAIT_FOR_OPLOCK') # WinError 765 (0x2FD)
    ERROR_DBG_EXCEPTION_HANDLED: Final[ThirtyTwoBits] = cached_win32error(0x02FE, 'ERROR_DBG_EXCEPTION_HANDLED') # WinError 766 (0x2FE)
    ERROR_DBG_CONTINUE: Final[ThirtyTwoBits] = cached_win32error(0x02FF, 'ERROR_DBG_CONTINUE') # WinError 767 (0x2FF)
    ERROR_CALLBACK_POP_STACK: Final[ThirtyTwoBits] = cached_win32error(0x0300, 'ERROR_CALLBACK_POP_STACK') # WinError 768 (0x300)
    ERROR_COMPRESSION_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x0301, 'ERROR_COMPRESSION_DISABLED') # WinError 769 (0x301)
    ERROR_CANTFETCHBACKWARDS: Final[ThirtyTwoBits] = cached_win32error(0x0302, 'ERROR_CANTFETCHBACKWARDS') # WinError 770 (0x302)
    ERROR_CANTSCROLLBACKWARDS: Final[ThirtyTwoBits] = cached_win32error(0x0303, 'ERROR_CANTSCROLLBACKWARDS') # WinError 771 (0x303)
    ERROR_ROWSNOTRELEASED: Final[ThirtyTwoBits] = cached_win32error(0x0304, 'ERROR_ROWSNOTRELEASED') # WinError 772 (0x304)
    ERROR_BAD_ACCESSOR_FLAGS: Final[ThirtyTwoBits] = cached_win32error(0x0305, 'ERROR_BAD_ACCESSOR_FLAGS') # WinError 773 (0x305)
    ERROR_ERRORS_ENCOUNTERED: Final[ThirtyTwoBits] = cached_win32error(0x0306, 'ERROR_ERRORS_ENCOUNTERED') # WinError 774 (0x306)
    ERROR_NOT_CAPABLE: Final[ThirtyTwoBits] = cached_win32error(0x0307, 'ERROR_NOT_CAPABLE') # WinError 775 (0x307)
    ERROR_REQUEST_OUT_OF_SEQUENCE: Final[ThirtyTwoBits] = cached_win32error(0x0308, 'ERROR_REQUEST_OUT_OF_SEQUENCE') # WinError 776 (0x308)
    ERROR_VERSION_PARSE_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x0309, 'ERROR_VERSION_PARSE_ERROR') # WinError 777 (0x309)
    ERROR_BADSTARTPOSITION: Final[ThirtyTwoBits] = cached_win32error(0x030A, 'ERROR_BADSTARTPOSITION') # WinError 778 (0x30A)
    ERROR_MEMORY_HARDWARE: Final[ThirtyTwoBits] = cached_win32error(0x030B, 'ERROR_MEMORY_HARDWARE') # WinError 779 (0x30B)
    ERROR_DISK_REPAIR_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x030C, 'ERROR_DISK_REPAIR_DISABLED') # WinError 780 (0x30C)
    ERROR_INSUFFICIENT_RESOURCE_FOR_SPECIFIED_SHARED_SECTION_SIZE: Final[ThirtyTwoBits] = cached_win32error(0x030D, 'ERROR_INSUFFICIENT_RESOURCE_FOR_SPECIFIED_SHARED_SECTION_SIZE') # WinError 781 (0x30D)
    ERROR_SYSTEM_POWERSTATE_TRANSITION: Final[ThirtyTwoBits] = cached_win32error(0x030E, 'ERROR_SYSTEM_POWERSTATE_TRANSITION') # WinError 782 (0x30E)
    ERROR_SYSTEM_POWERSTATE_COMPLEX_TRANSITION: Final[ThirtyTwoBits] = cached_win32error(0x030F, 'ERROR_SYSTEM_POWERSTATE_COMPLEX_TRANSITION') # WinError 783 (0x30F)
    ERROR_MCA_EXCEPTION: Final[ThirtyTwoBits] = cached_win32error(0x0310, 'ERROR_MCA_EXCEPTION') # WinError 784 (0x310)
    ERROR_ACCESS_AUDIT_BY_POLICY: Final[ThirtyTwoBits] = cached_win32error(0x0311, 'ERROR_ACCESS_AUDIT_BY_POLICY') # WinError 785 (0x311)
    ERROR_ACCESS_DISABLED_NO_SAFER_UI_BY_POLICY: Final[ThirtyTwoBits] = cached_win32error(0x0312, 'ERROR_ACCESS_DISABLED_NO_SAFER_UI_BY_POLICY') # WinError 786 (0x312)
    ERROR_ABANDON_HIBERFILE: Final[ThirtyTwoBits] = cached_win32error(0x0313, 'ERROR_ABANDON_HIBERFILE') # WinError 787 (0x313)
    ERROR_LOST_WRITEBEHIND_DATA_NETWORK_DISCONNECTED: Final[ThirtyTwoBits] = cached_win32error(0x0314, 'ERROR_LOST_WRITEBEHIND_DATA_NETWORK_DISCONNECTED') # WinError 788 (0x314)
    ERROR_LOST_WRITEBEHIND_DATA_NETWORK_SERVER_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x0315, 'ERROR_LOST_WRITEBEHIND_DATA_NETWORK_SERVER_ERROR') # WinError 789 (0x315)
    ERROR_LOST_WRITEBEHIND_DATA_LOCAL_DISK_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x0316, 'ERROR_LOST_WRITEBEHIND_DATA_LOCAL_DISK_ERROR') # WinError 790 (0x316)
    ERROR_BAD_MCFG_TABLE: Final[ThirtyTwoBits] = cached_win32error(0x0317, 'ERROR_BAD_MCFG_TABLE') # WinError 791 (0x317)
    ERROR_DISK_REPAIR_REDIRECTED: Final[ThirtyTwoBits] = cached_win32error(0x0318, 'ERROR_DISK_REPAIR_REDIRECTED') # WinError 792 (0x318)
    ERROR_DISK_REPAIR_UNSUCCESSFUL: Final[ThirtyTwoBits] = cached_win32error(0x0319, 'ERROR_DISK_REPAIR_UNSUCCESSFUL') # WinError 793 (0x319)
    ERROR_CORRUPT_LOG_OVERFULL: Final[ThirtyTwoBits] = cached_win32error(0x031A, 'ERROR_CORRUPT_LOG_OVERFULL') # WinError 794 (0x31A)
    ERROR_CORRUPT_LOG_CORRUPTED: Final[ThirtyTwoBits] = cached_win32error(0x031B, 'ERROR_CORRUPT_LOG_CORRUPTED') # WinError 795 (0x31B)
    ERROR_CORRUPT_LOG_UNAVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x031C, 'ERROR_CORRUPT_LOG_UNAVAILABLE') # WinError 796 (0x31C)
    ERROR_CORRUPT_LOG_DELETED_FULL: Final[ThirtyTwoBits] = cached_win32error(0x031D, 'ERROR_CORRUPT_LOG_DELETED_FULL') # WinError 797 (0x31D)
    ERROR_CORRUPT_LOG_CLEARED: Final[ThirtyTwoBits] = cached_win32error(0x031E, 'ERROR_CORRUPT_LOG_CLEARED') # WinError 798 (0x31E)
    ERROR_ORPHAN_NAME_EXHAUSTED: Final[ThirtyTwoBits] = cached_win32error(0x031F, 'ERROR_ORPHAN_NAME_EXHAUSTED') # WinError 799 (0x31F)
    ERROR_OPLOCK_SWITCHED_TO_NEW_HANDLE: Final[ThirtyTwoBits] = cached_win32error(0x0320, 'ERROR_OPLOCK_SWITCHED_TO_NEW_HANDLE') # WinError 800 (0x320)
    ERROR_CANNOT_GRANT_REQUESTED_OPLOCK: Final[ThirtyTwoBits] = cached_win32error(0x0321, 'ERROR_CANNOT_GRANT_REQUESTED_OPLOCK') # WinError 801 (0x321)
    ERROR_CANNOT_BREAK_OPLOCK: Final[ThirtyTwoBits] = cached_win32error(0x0322, 'ERROR_CANNOT_BREAK_OPLOCK') # WinError 802 (0x322)
    ERROR_OPLOCK_HANDLE_CLOSED: Final[ThirtyTwoBits] = cached_win32error(0x0323, 'ERROR_OPLOCK_HANDLE_CLOSED') # WinError 803 (0x323)
    ERROR_NO_ACE_CONDITION: Final[ThirtyTwoBits] = cached_win32error(0x0324, 'ERROR_NO_ACE_CONDITION') # WinError 804 (0x324)
    ERROR_INVALID_ACE_CONDITION: Final[ThirtyTwoBits] = cached_win32error(0x0325, 'ERROR_INVALID_ACE_CONDITION') # WinError 805 (0x325)
    ERROR_FILE_HANDLE_REVOKED: Final[ThirtyTwoBits] = cached_win32error(0x0326, 'ERROR_FILE_HANDLE_REVOKED') # WinError 806 (0x326)
    ERROR_IMAGE_AT_DIFFERENT_BASE: Final[ThirtyTwoBits] = cached_win32error(0x0327, 'ERROR_IMAGE_AT_DIFFERENT_BASE') # WinError 807 (0x327)
    ERROR_ENCRYPTED_IO_NOT_POSSIBLE: Final[ThirtyTwoBits] = cached_win32error(0x0328, 'ERROR_ENCRYPTED_IO_NOT_POSSIBLE') # WinError 808 (0x328)
    ERROR_FILE_METADATA_OPTIMIZATION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x0329, 'ERROR_FILE_METADATA_OPTIMIZATION_IN_PROGRESS') # WinError 809 (0x329)
    ERROR_QUOTA_ACTIVITY: Final[ThirtyTwoBits] = cached_win32error(0x032A, 'ERROR_QUOTA_ACTIVITY') # WinError 810 (0x32A)
    ERROR_HANDLE_REVOKED: Final[ThirtyTwoBits] = cached_win32error(0x032B, 'ERROR_HANDLE_REVOKED') # WinError 811 (0x32B)
    ERROR_CALLBACK_INVOKE_INLINE: Final[ThirtyTwoBits] = cached_win32error(0x032C, 'ERROR_CALLBACK_INVOKE_INLINE') # WinError 812 (0x32C)
    ERROR_CPU_SET_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x032D, 'ERROR_CPU_SET_INVALID') # WinError 813 (0x32D)
    ERROR_ENCLAVE_NOT_TERMINATED: Final[ThirtyTwoBits] = cached_win32error(0x032E, 'ERROR_ENCLAVE_NOT_TERMINATED') # WinError 814 (0x32E)
    ERROR_ENCLAVE_VIOLATION: Final[ThirtyTwoBits] = cached_win32error(0x032F, 'ERROR_ENCLAVE_VIOLATION') # WinError 815 (0x32F)
    ERROR_SERVER_TRANSPORT_CONFLICT: Final[ThirtyTwoBits] = cached_win32error(0x0330, 'ERROR_SERVER_TRANSPORT_CONFLICT') # WinError 816 (0x330)
    ERROR_CERTIFICATE_VALIDATION_PREFERENCE_CONFLICT: Final[ThirtyTwoBits] = cached_win32error(0x0331, 'ERROR_CERTIFICATE_VALIDATION_PREFERENCE_CONFLICT') # WinError 817 (0x331)
    ERROR_FT_READ_FROM_COPY_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x0332, 'ERROR_FT_READ_FROM_COPY_FAILURE') # WinError 818 (0x332)
    ERROR_SECTION_DIRECT_MAP_ONLY: Final[ThirtyTwoBits] = cached_win32error(0x0333, 'ERROR_SECTION_DIRECT_MAP_ONLY') # WinError 819 (0x333)
    ERROR_EA_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_win32error(0x03E2, 'ERROR_EA_ACCESS_DENIED') # WinError 994 (0x3E2)
    ERROR_OPERATION_ABORTED: Final[ThirtyTwoBits] = cached_win32error(0x03E3, 'ERROR_OPERATION_ABORTED') # WinError 995 (0x3E3)
    ERROR_IO_INCOMPLETE: Final[ThirtyTwoBits] = cached_win32error(0x03E4, 'ERROR_IO_INCOMPLETE') # WinError 996 (0x3E4)
    ERROR_IO_PENDING: Final[ThirtyTwoBits] = cached_win32error(0x03E5, 'ERROR_IO_PENDING') # WinError 997 (0x3E5)
    ERROR_NOACCESS: Final[ThirtyTwoBits] = cached_win32error(0x03E6, 'ERROR_NOACCESS') # WinError 998 (0x3E6)
    ERROR_SWAPERROR: Final[ThirtyTwoBits] = cached_win32error(0x03E7, 'ERROR_SWAPERROR') # WinError 999 (0x3E7)
    ERROR_STACK_OVERFLOW: Final[ThirtyTwoBits] = cached_win32error(0x03E9, 'ERROR_STACK_OVERFLOW') # WinError 1001 (0x3E9)
    ERROR_INVALID_MESSAGE: Final[ThirtyTwoBits] = cached_win32error(0x03EA, 'ERROR_INVALID_MESSAGE') # WinError 1002 (0x3EA)
    ERROR_CAN_NOT_COMPLETE: Final[ThirtyTwoBits] = cached_win32error(0x03EB, 'ERROR_CAN_NOT_COMPLETE') # WinError 1003 (0x3EB)
    ERROR_INVALID_FLAGS: Final[ThirtyTwoBits] = cached_win32error(0x03EC, 'ERROR_INVALID_FLAGS') # WinError 1004 (0x3EC)
    ERROR_UNRECOGNIZED_VOLUME: Final[ThirtyTwoBits] = cached_win32error(0x03ED, 'ERROR_UNRECOGNIZED_VOLUME') # WinError 1005 (0x3ED)
    ERROR_FILE_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x03EE, 'ERROR_FILE_INVALID') # WinError 1006 (0x3EE)
    ERROR_FULLSCREEN_MODE: Final[ThirtyTwoBits] = cached_win32error(0x03EF, 'ERROR_FULLSCREEN_MODE') # WinError 1007 (0x3EF)
    ERROR_NO_TOKEN: Final[ThirtyTwoBits] = cached_win32error(0x03F0, 'ERROR_NO_TOKEN') # WinError 1008 (0x3F0)
    ERROR_BADDB: Final[ThirtyTwoBits] = cached_win32error(0x03F1, 'ERROR_BADDB') # WinError 1009 (0x3F1)
    ERROR_BADKEY: Final[ThirtyTwoBits] = cached_win32error(0x03F2, 'ERROR_BADKEY') # WinError 1010 (0x3F2)
    ERROR_CANTOPEN: Final[ThirtyTwoBits] = cached_win32error(0x03F3, 'ERROR_CANTOPEN') # WinError 1011 (0x3F3)
    ERROR_CANTREAD: Final[ThirtyTwoBits] = cached_win32error(0x03F4, 'ERROR_CANTREAD') # WinError 1012 (0x3F4)
    ERROR_CANTWRITE: Final[ThirtyTwoBits] = cached_win32error(0x03F5, 'ERROR_CANTWRITE') # WinError 1013 (0x3F5)
    ERROR_REGISTRY_RECOVERED: Final[ThirtyTwoBits] = cached_win32error(0x03F6, 'ERROR_REGISTRY_RECOVERED') # WinError 1014 (0x3F6)
    ERROR_REGISTRY_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x03F7, 'ERROR_REGISTRY_CORRUPT') # WinError 1015 (0x3F7)
    ERROR_REGISTRY_IO_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x03F8, 'ERROR_REGISTRY_IO_FAILED') # WinError 1016 (0x3F8)
    ERROR_NOT_REGISTRY_FILE: Final[ThirtyTwoBits] = cached_win32error(0x03F9, 'ERROR_NOT_REGISTRY_FILE') # WinError 1017 (0x3F9)
    ERROR_KEY_DELETED: Final[ThirtyTwoBits] = cached_win32error(0x03FA, 'ERROR_KEY_DELETED') # WinError 1018 (0x3FA)
    ERROR_NO_LOG_SPACE: Final[ThirtyTwoBits] = cached_win32error(0x03FB, 'ERROR_NO_LOG_SPACE') # WinError 1019 (0x3FB)
    ERROR_KEY_HAS_CHILDREN: Final[ThirtyTwoBits] = cached_win32error(0x03FC, 'ERROR_KEY_HAS_CHILDREN') # WinError 1020 (0x3FC)
    ERROR_CHILD_MUST_BE_VOLATILE: Final[ThirtyTwoBits] = cached_win32error(0x03FD, 'ERROR_CHILD_MUST_BE_VOLATILE') # WinError 1021 (0x3FD)
    ERROR_NOTIFY_ENUM_DIR: Final[ThirtyTwoBits] = cached_win32error(0x03FE, 'ERROR_NOTIFY_ENUM_DIR') # WinError 1022 (0x3FE)
    ERROR_DEPENDENT_SERVICES_RUNNING: Final[ThirtyTwoBits] = cached_win32error(0x041B, 'ERROR_DEPENDENT_SERVICES_RUNNING') # WinError 1051 (0x41B)
    ERROR_INVALID_SERVICE_CONTROL: Final[ThirtyTwoBits] = cached_win32error(0x041C, 'ERROR_INVALID_SERVICE_CONTROL') # WinError 1052 (0x41C)
    ERROR_SERVICE_REQUEST_TIMEOUT: Final[ThirtyTwoBits] = cached_win32error(0x041D, 'ERROR_SERVICE_REQUEST_TIMEOUT') # WinError 1053 (0x41D)
    ERROR_SERVICE_NO_THREAD: Final[ThirtyTwoBits] = cached_win32error(0x041E, 'ERROR_SERVICE_NO_THREAD') # WinError 1054 (0x41E)
    ERROR_SERVICE_DATABASE_LOCKED: Final[ThirtyTwoBits] = cached_win32error(0x041F, 'ERROR_SERVICE_DATABASE_LOCKED') # WinError 1055 (0x41F)
    ERROR_SERVICE_ALREADY_RUNNING: Final[ThirtyTwoBits] = cached_win32error(0x0420, 'ERROR_SERVICE_ALREADY_RUNNING') # WinError 1056 (0x420)
    ERROR_INVALID_SERVICE_ACCOUNT: Final[ThirtyTwoBits] = cached_win32error(0x0421, 'ERROR_INVALID_SERVICE_ACCOUNT') # WinError 1057 (0x421)
    ERROR_SERVICE_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x0422, 'ERROR_SERVICE_DISABLED') # WinError 1058 (0x422)
    ERROR_CIRCULAR_DEPENDENCY: Final[ThirtyTwoBits] = cached_win32error(0x0423, 'ERROR_CIRCULAR_DEPENDENCY') # WinError 1059 (0x423)
    ERROR_SERVICE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_win32error(0x0424, 'ERROR_SERVICE_DOES_NOT_EXIST') # WinError 1060 (0x424)
    ERROR_SERVICE_CANNOT_ACCEPT_CTRL: Final[ThirtyTwoBits] = cached_win32error(0x0425, 'ERROR_SERVICE_CANNOT_ACCEPT_CTRL') # WinError 1061 (0x425)
    ERROR_SERVICE_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_win32error(0x0426, 'ERROR_SERVICE_NOT_ACTIVE') # WinError 1062 (0x426)
    ERROR_FAILED_SERVICE_CONTROLLER_CONNECT: Final[ThirtyTwoBits] = cached_win32error(0x0427, 'ERROR_FAILED_SERVICE_CONTROLLER_CONNECT') # WinError 1063 (0x427)
    ERROR_EXCEPTION_IN_SERVICE: Final[ThirtyTwoBits] = cached_win32error(0x0428, 'ERROR_EXCEPTION_IN_SERVICE') # WinError 1064 (0x428)
    ERROR_DATABASE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_win32error(0x0429, 'ERROR_DATABASE_DOES_NOT_EXIST') # WinError 1065 (0x429)
    ERROR_SERVICE_SPECIFIC_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x042A, 'ERROR_SERVICE_SPECIFIC_ERROR') # WinError 1066 (0x42A)
    ERROR_PROCESS_ABORTED: Final[ThirtyTwoBits] = cached_win32error(0x042B, 'ERROR_PROCESS_ABORTED') # WinError 1067 (0x42B)
    ERROR_SERVICE_DEPENDENCY_FAIL: Final[ThirtyTwoBits] = cached_win32error(0x042C, 'ERROR_SERVICE_DEPENDENCY_FAIL') # WinError 1068 (0x42C)
    ERROR_SERVICE_LOGON_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x042D, 'ERROR_SERVICE_LOGON_FAILED') # WinError 1069 (0x42D)
    ERROR_SERVICE_START_HANG: Final[ThirtyTwoBits] = cached_win32error(0x042E, 'ERROR_SERVICE_START_HANG') # WinError 1070 (0x42E)
    ERROR_INVALID_SERVICE_LOCK: Final[ThirtyTwoBits] = cached_win32error(0x042F, 'ERROR_INVALID_SERVICE_LOCK') # WinError 1071 (0x42F)
    ERROR_SERVICE_MARKED_FOR_DELETE: Final[ThirtyTwoBits] = cached_win32error(0x0430, 'ERROR_SERVICE_MARKED_FOR_DELETE') # WinError 1072 (0x430)
    ERROR_SERVICE_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x0431, 'ERROR_SERVICE_EXISTS') # WinError 1073 (0x431)
    ERROR_ALREADY_RUNNING_LKG: Final[ThirtyTwoBits] = cached_win32error(0x0432, 'ERROR_ALREADY_RUNNING_LKG') # WinError 1074 (0x432)
    ERROR_SERVICE_DEPENDENCY_DELETED: Final[ThirtyTwoBits] = cached_win32error(0x0433, 'ERROR_SERVICE_DEPENDENCY_DELETED') # WinError 1075 (0x433)
    ERROR_BOOT_ALREADY_ACCEPTED: Final[ThirtyTwoBits] = cached_win32error(0x0434, 'ERROR_BOOT_ALREADY_ACCEPTED') # WinError 1076 (0x434)
    ERROR_SERVICE_NEVER_STARTED: Final[ThirtyTwoBits] = cached_win32error(0x0435, 'ERROR_SERVICE_NEVER_STARTED') # WinError 1077 (0x435)
    ERROR_DUPLICATE_SERVICE_NAME: Final[ThirtyTwoBits] = cached_win32error(0x0436, 'ERROR_DUPLICATE_SERVICE_NAME') # WinError 1078 (0x436)
    ERROR_DIFFERENT_SERVICE_ACCOUNT: Final[ThirtyTwoBits] = cached_win32error(0x0437, 'ERROR_DIFFERENT_SERVICE_ACCOUNT') # WinError 1079 (0x437)
    ERROR_CANNOT_DETECT_DRIVER_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x0438, 'ERROR_CANNOT_DETECT_DRIVER_FAILURE') # WinError 1080 (0x438)
    ERROR_CANNOT_DETECT_PROCESS_ABORT: Final[ThirtyTwoBits] = cached_win32error(0x0439, 'ERROR_CANNOT_DETECT_PROCESS_ABORT') # WinError 1081 (0x439)
    ERROR_NO_RECOVERY_PROGRAM: Final[ThirtyTwoBits] = cached_win32error(0x043A, 'ERROR_NO_RECOVERY_PROGRAM') # WinError 1082 (0x43A)
    ERROR_SERVICE_NOT_IN_EXE: Final[ThirtyTwoBits] = cached_win32error(0x043B, 'ERROR_SERVICE_NOT_IN_EXE') # WinError 1083 (0x43B)
    ERROR_NOT_SAFEBOOT_SERVICE: Final[ThirtyTwoBits] = cached_win32error(0x043C, 'ERROR_NOT_SAFEBOOT_SERVICE') # WinError 1084 (0x43C)
    ERROR_END_OF_MEDIA: Final[ThirtyTwoBits] = cached_win32error(0x044C, 'ERROR_END_OF_MEDIA') # WinError 1100 (0x44C)
    ERROR_FILEMARK_DETECTED: Final[ThirtyTwoBits] = cached_win32error(0x044D, 'ERROR_FILEMARK_DETECTED') # WinError 1101 (0x44D)
    ERROR_BEGINNING_OF_MEDIA: Final[ThirtyTwoBits] = cached_win32error(0x044E, 'ERROR_BEGINNING_OF_MEDIA') # WinError 1102 (0x44E)
    ERROR_SETMARK_DETECTED: Final[ThirtyTwoBits] = cached_win32error(0x044F, 'ERROR_SETMARK_DETECTED') # WinError 1103 (0x44F)
    ERROR_NO_DATA_DETECTED: Final[ThirtyTwoBits] = cached_win32error(0x0450, 'ERROR_NO_DATA_DETECTED') # WinError 1104 (0x450)
    ERROR_PARTITION_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x0451, 'ERROR_PARTITION_FAILURE') # WinError 1105 (0x451)
    ERROR_INVALID_BLOCK_LENGTH: Final[ThirtyTwoBits] = cached_win32error(0x0452, 'ERROR_INVALID_BLOCK_LENGTH') # WinError 1106 (0x452)
    ERROR_DEVICE_NOT_PARTITIONED: Final[ThirtyTwoBits] = cached_win32error(0x0453, 'ERROR_DEVICE_NOT_PARTITIONED') # WinError 1107 (0x453)
    ERROR_UNABLE_TO_LOCK_MEDIA: Final[ThirtyTwoBits] = cached_win32error(0x0454, 'ERROR_UNABLE_TO_LOCK_MEDIA') # WinError 1108 (0x454)
    ERROR_UNABLE_TO_UNLOAD_MEDIA: Final[ThirtyTwoBits] = cached_win32error(0x0455, 'ERROR_UNABLE_TO_UNLOAD_MEDIA') # WinError 1109 (0x455)
    ERROR_MEDIA_CHANGED: Final[ThirtyTwoBits] = cached_win32error(0x0456, 'ERROR_MEDIA_CHANGED') # WinError 1110 (0x456)
    ERROR_BUS_RESET: Final[ThirtyTwoBits] = cached_win32error(0x0457, 'ERROR_BUS_RESET') # WinError 1111 (0x457)
    ERROR_NO_MEDIA_IN_DRIVE: Final[ThirtyTwoBits] = cached_win32error(0x0458, 'ERROR_NO_MEDIA_IN_DRIVE') # WinError 1112 (0x458)
    ERROR_NO_UNICODE_TRANSLATION: Final[ThirtyTwoBits] = cached_win32error(0x0459, 'ERROR_NO_UNICODE_TRANSLATION') # WinError 1113 (0x459)
    ERROR_DLL_INIT_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x045A, 'ERROR_DLL_INIT_FAILED') # WinError 1114 (0x45A)
    ERROR_SHUTDOWN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x045B, 'ERROR_SHUTDOWN_IN_PROGRESS') # WinError 1115 (0x45B)
    ERROR_NO_SHUTDOWN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x045C, 'ERROR_NO_SHUTDOWN_IN_PROGRESS') # WinError 1116 (0x45C)
    ERROR_IO_DEVICE: Final[ThirtyTwoBits] = cached_win32error(0x045D, 'ERROR_IO_DEVICE') # WinError 1117 (0x45D)
    ERROR_SERIAL_NO_DEVICE: Final[ThirtyTwoBits] = cached_win32error(0x045E, 'ERROR_SERIAL_NO_DEVICE') # WinError 1118 (0x45E)
    ERROR_IRQ_BUSY: Final[ThirtyTwoBits] = cached_win32error(0x045F, 'ERROR_IRQ_BUSY') # WinError 1119 (0x45F)
    ERROR_MORE_WRITES: Final[ThirtyTwoBits] = cached_win32error(0x0460, 'ERROR_MORE_WRITES') # WinError 1120 (0x460)
    ERROR_COUNTER_TIMEOUT: Final[ThirtyTwoBits] = cached_win32error(0x0461, 'ERROR_COUNTER_TIMEOUT') # WinError 1121 (0x461)
    ERROR_FLOPPY_ID_MARK_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x0462, 'ERROR_FLOPPY_ID_MARK_NOT_FOUND') # WinError 1122 (0x462)
    ERROR_FLOPPY_WRONG_CYLINDER: Final[ThirtyTwoBits] = cached_win32error(0x0463, 'ERROR_FLOPPY_WRONG_CYLINDER') # WinError 1123 (0x463)
    ERROR_FLOPPY_UNKNOWN_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x0464, 'ERROR_FLOPPY_UNKNOWN_ERROR') # WinError 1124 (0x464)
    ERROR_FLOPPY_BAD_REGISTERS: Final[ThirtyTwoBits] = cached_win32error(0x0465, 'ERROR_FLOPPY_BAD_REGISTERS') # WinError 1125 (0x465)
    ERROR_DISK_RECALIBRATE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x0466, 'ERROR_DISK_RECALIBRATE_FAILED') # WinError 1126 (0x466)
    ERROR_DISK_OPERATION_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x0467, 'ERROR_DISK_OPERATION_FAILED') # WinError 1127 (0x467)
    ERROR_DISK_RESET_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x0468, 'ERROR_DISK_RESET_FAILED') # WinError 1128 (0x468)
    ERROR_EOM_OVERFLOW: Final[ThirtyTwoBits] = cached_win32error(0x0469, 'ERROR_EOM_OVERFLOW') # WinError 1129 (0x469)
    ERROR_NOT_ENOUGH_SERVER_MEMORY: Final[ThirtyTwoBits] = cached_win32error(0x046A, 'ERROR_NOT_ENOUGH_SERVER_MEMORY') # WinError 1130 (0x46A)
    ERROR_POSSIBLE_DEADLOCK: Final[ThirtyTwoBits] = cached_win32error(0x046B, 'ERROR_POSSIBLE_DEADLOCK') # WinError 1131 (0x46B)
    ERROR_MAPPED_ALIGNMENT: Final[ThirtyTwoBits] = cached_win32error(0x046C, 'ERROR_MAPPED_ALIGNMENT') # WinError 1132 (0x46C)
    ERROR_SET_POWER_STATE_VETOED: Final[ThirtyTwoBits] = cached_win32error(0x0474, 'ERROR_SET_POWER_STATE_VETOED') # WinError 1140 (0x474)
    ERROR_SET_POWER_STATE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x0475, 'ERROR_SET_POWER_STATE_FAILED') # WinError 1141 (0x475)
    ERROR_TOO_MANY_LINKS: Final[ThirtyTwoBits] = cached_win32error(0x0476, 'ERROR_TOO_MANY_LINKS') # WinError 1142 (0x476)
    ERROR_OLD_WIN_VERSION: Final[ThirtyTwoBits] = cached_win32error(0x047E, 'ERROR_OLD_WIN_VERSION') # WinError 1150 (0x47E)
    ERROR_APP_WRONG_OS: Final[ThirtyTwoBits] = cached_win32error(0x047F, 'ERROR_APP_WRONG_OS') # WinError 1151 (0x47F)
    ERROR_SINGLE_INSTANCE_APP: Final[ThirtyTwoBits] = cached_win32error(0x0480, 'ERROR_SINGLE_INSTANCE_APP') # WinError 1152 (0x480)
    ERROR_RMODE_APP: Final[ThirtyTwoBits] = cached_win32error(0x0481, 'ERROR_RMODE_APP') # WinError 1153 (0x481)
    ERROR_INVALID_DLL: Final[ThirtyTwoBits] = cached_win32error(0x0482, 'ERROR_INVALID_DLL') # WinError 1154 (0x482)
    ERROR_NO_ASSOCIATION: Final[ThirtyTwoBits] = cached_win32error(0x0483, 'ERROR_NO_ASSOCIATION') # WinError 1155 (0x483)
    ERROR_DDE_FAIL: Final[ThirtyTwoBits] = cached_win32error(0x0484, 'ERROR_DDE_FAIL') # WinError 1156 (0x484)
    ERROR_DLL_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x0485, 'ERROR_DLL_NOT_FOUND') # WinError 1157 (0x485)
    ERROR_NO_MORE_USER_HANDLES: Final[ThirtyTwoBits] = cached_win32error(0x0486, 'ERROR_NO_MORE_USER_HANDLES') # WinError 1158 (0x486)
    ERROR_MESSAGE_SYNC_ONLY: Final[ThirtyTwoBits] = cached_win32error(0x0487, 'ERROR_MESSAGE_SYNC_ONLY') # WinError 1159 (0x487)
    ERROR_SOURCE_ELEMENT_EMPTY: Final[ThirtyTwoBits] = cached_win32error(0x0488, 'ERROR_SOURCE_ELEMENT_EMPTY') # WinError 1160 (0x488)
    ERROR_DESTINATION_ELEMENT_FULL: Final[ThirtyTwoBits] = cached_win32error(0x0489, 'ERROR_DESTINATION_ELEMENT_FULL') # WinError 1161 (0x489)
    ERROR_ILLEGAL_ELEMENT_ADDRESS: Final[ThirtyTwoBits] = cached_win32error(0x048A, 'ERROR_ILLEGAL_ELEMENT_ADDRESS') # WinError 1162 (0x48A)
    ERROR_MAGAZINE_NOT_PRESENT: Final[ThirtyTwoBits] = cached_win32error(0x048B, 'ERROR_MAGAZINE_NOT_PRESENT') # WinError 1163 (0x48B)
    ERROR_DEVICE_REINITIALIZATION_NEEDED: Final[ThirtyTwoBits] = cached_win32error(0x048C, 'ERROR_DEVICE_REINITIALIZATION_NEEDED') # WinError 1164 (0x48C)
    ERROR_DEVICE_REQUIRES_CLEANING: Final[ThirtyTwoBits] = cached_win32error(0x048D, 'ERROR_DEVICE_REQUIRES_CLEANING') # WinError 1165 (0x48D)
    ERROR_DEVICE_DOOR_OPEN: Final[ThirtyTwoBits] = cached_win32error(0x048E, 'ERROR_DEVICE_DOOR_OPEN') # WinError 1166 (0x48E)
    ERROR_DEVICE_NOT_CONNECTED: Final[ThirtyTwoBits] = cached_win32error(0x048F, 'ERROR_DEVICE_NOT_CONNECTED') # WinError 1167 (0x48F)
    ERROR_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x0490, 'ERROR_NOT_FOUND') # WinError 1168 (0x490)
    ERROR_NO_MATCH: Final[ThirtyTwoBits] = cached_win32error(0x0491, 'ERROR_NO_MATCH') # WinError 1169 (0x491)
    ERROR_SET_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x0492, 'ERROR_SET_NOT_FOUND') # WinError 1170 (0x492)
    ERROR_POINT_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x0493, 'ERROR_POINT_NOT_FOUND') # WinError 1171 (0x493)
    ERROR_NO_TRACKING_SERVICE: Final[ThirtyTwoBits] = cached_win32error(0x0494, 'ERROR_NO_TRACKING_SERVICE') # WinError 1172 (0x494)
    ERROR_NO_VOLUME_ID: Final[ThirtyTwoBits] = cached_win32error(0x0495, 'ERROR_NO_VOLUME_ID') # WinError 1173 (0x495)
    ERROR_UNABLE_TO_REMOVE_REPLACED: Final[ThirtyTwoBits] = cached_win32error(0x0497, 'ERROR_UNABLE_TO_REMOVE_REPLACED') # WinError 1175 (0x497)
    ERROR_UNABLE_TO_MOVE_REPLACEMENT: Final[ThirtyTwoBits] = cached_win32error(0x0498, 'ERROR_UNABLE_TO_MOVE_REPLACEMENT') # WinError 1176 (0x498)
    ERROR_UNABLE_TO_MOVE_REPLACEMENT_2: Final[ThirtyTwoBits] = cached_win32error(0x0499, 'ERROR_UNABLE_TO_MOVE_REPLACEMENT_2') # WinError 1177 (0x499)
    ERROR_JOURNAL_DELETE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x049A, 'ERROR_JOURNAL_DELETE_IN_PROGRESS') # WinError 1178 (0x49A)
    ERROR_JOURNAL_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_win32error(0x049B, 'ERROR_JOURNAL_NOT_ACTIVE') # WinError 1179 (0x49B)
    ERROR_POTENTIAL_FILE_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x049C, 'ERROR_POTENTIAL_FILE_FOUND') # WinError 1180 (0x49C)
    ERROR_JOURNAL_ENTRY_DELETED: Final[ThirtyTwoBits] = cached_win32error(0x049D, 'ERROR_JOURNAL_ENTRY_DELETED') # WinError 1181 (0x49D)
    ERROR_PARTITION_TERMINATING: Final[ThirtyTwoBits] = cached_win32error(0x04A0, 'ERROR_PARTITION_TERMINATING') # WinError 1184 (0x4A0)
    ERROR_SHUTDOWN_IS_SCHEDULED: Final[ThirtyTwoBits] = cached_win32error(0x04A6, 'ERROR_SHUTDOWN_IS_SCHEDULED') # WinError 1190 (0x4A6)
    ERROR_SHUTDOWN_USERS_LOGGED_ON: Final[ThirtyTwoBits] = cached_win32error(0x04A7, 'ERROR_SHUTDOWN_USERS_LOGGED_ON') # WinError 1191 (0x4A7)
    ERROR_SHUTDOWN_DISKS_NOT_IN_MAINTENANCE_MODE: Final[ThirtyTwoBits] = cached_win32error(0x04A8, 'ERROR_SHUTDOWN_DISKS_NOT_IN_MAINTENANCE_MODE') # WinError 1192 (0x4A8)
    ERROR_BAD_DEVICE: Final[ThirtyTwoBits] = cached_win32error(0x04B0, 'ERROR_BAD_DEVICE') # WinError 1200 (0x4B0)
    ERROR_CONNECTION_UNAVAIL: Final[ThirtyTwoBits] = cached_win32error(0x04B1, 'ERROR_CONNECTION_UNAVAIL') # WinError 1201 (0x4B1)
    ERROR_DEVICE_ALREADY_REMEMBERED: Final[ThirtyTwoBits] = cached_win32error(0x04B2, 'ERROR_DEVICE_ALREADY_REMEMBERED') # WinError 1202 (0x4B2)
    ERROR_NO_NET_OR_BAD_PATH: Final[ThirtyTwoBits] = cached_win32error(0x04B3, 'ERROR_NO_NET_OR_BAD_PATH') # WinError 1203 (0x4B3)
    ERROR_BAD_PROVIDER: Final[ThirtyTwoBits] = cached_win32error(0x04B4, 'ERROR_BAD_PROVIDER') # WinError 1204 (0x4B4)
    ERROR_CANNOT_OPEN_PROFILE: Final[ThirtyTwoBits] = cached_win32error(0x04B5, 'ERROR_CANNOT_OPEN_PROFILE') # WinError 1205 (0x4B5)
    ERROR_BAD_PROFILE: Final[ThirtyTwoBits] = cached_win32error(0x04B6, 'ERROR_BAD_PROFILE') # WinError 1206 (0x4B6)
    ERROR_NOT_CONTAINER: Final[ThirtyTwoBits] = cached_win32error(0x04B7, 'ERROR_NOT_CONTAINER') # WinError 1207 (0x4B7)
    ERROR_EXTENDED_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x04B8, 'ERROR_EXTENDED_ERROR') # WinError 1208 (0x4B8)
    ERROR_INVALID_GROUPNAME: Final[ThirtyTwoBits] = cached_win32error(0x04B9, 'ERROR_INVALID_GROUPNAME') # WinError 1209 (0x4B9)
    ERROR_INVALID_COMPUTERNAME: Final[ThirtyTwoBits] = cached_win32error(0x04BA, 'ERROR_INVALID_COMPUTERNAME') # WinError 1210 (0x4BA)
    ERROR_INVALID_EVENTNAME: Final[ThirtyTwoBits] = cached_win32error(0x04BB, 'ERROR_INVALID_EVENTNAME') # WinError 1211 (0x4BB)
    ERROR_INVALID_DOMAINNAME: Final[ThirtyTwoBits] = cached_win32error(0x04BC, 'ERROR_INVALID_DOMAINNAME') # WinError 1212 (0x4BC)
    ERROR_INVALID_SERVICENAME: Final[ThirtyTwoBits] = cached_win32error(0x04BD, 'ERROR_INVALID_SERVICENAME') # WinError 1213 (0x4BD)
    ERROR_INVALID_NETNAME: Final[ThirtyTwoBits] = cached_win32error(0x04BE, 'ERROR_INVALID_NETNAME') # WinError 1214 (0x4BE)
    ERROR_INVALID_SHARENAME: Final[ThirtyTwoBits] = cached_win32error(0x04BF, 'ERROR_INVALID_SHARENAME') # WinError 1215 (0x4BF)
    ERROR_INVALID_PASSWORDNAME: Final[ThirtyTwoBits] = cached_win32error(0x04C0, 'ERROR_INVALID_PASSWORDNAME') # WinError 1216 (0x4C0)
    ERROR_INVALID_MESSAGENAME: Final[ThirtyTwoBits] = cached_win32error(0x04C1, 'ERROR_INVALID_MESSAGENAME') # WinError 1217 (0x4C1)
    ERROR_INVALID_MESSAGEDEST: Final[ThirtyTwoBits] = cached_win32error(0x04C2, 'ERROR_INVALID_MESSAGEDEST') # WinError 1218 (0x4C2)
    ERROR_SESSION_CREDENTIAL_CONFLICT: Final[ThirtyTwoBits] = cached_win32error(0x04C3, 'ERROR_SESSION_CREDENTIAL_CONFLICT') # WinError 1219 (0x4C3)
    ERROR_REMOTE_SESSION_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x04C4, 'ERROR_REMOTE_SESSION_LIMIT_EXCEEDED') # WinError 1220 (0x4C4)
    ERROR_DUP_DOMAINNAME: Final[ThirtyTwoBits] = cached_win32error(0x04C5, 'ERROR_DUP_DOMAINNAME') # WinError 1221 (0x4C5)
    ERROR_NO_NETWORK: Final[ThirtyTwoBits] = cached_win32error(0x04C6, 'ERROR_NO_NETWORK') # WinError 1222 (0x4C6)
    ERROR_CANCELLED: Final[ThirtyTwoBits] = cached_win32error(0x04C7, 'ERROR_CANCELLED') # WinError 1223 (0x4C7)
    ERROR_USER_MAPPED_FILE: Final[ThirtyTwoBits] = cached_win32error(0x04C8, 'ERROR_USER_MAPPED_FILE') # WinError 1224 (0x4C8)
    ERROR_CONNECTION_REFUSED: Final[ThirtyTwoBits] = cached_win32error(0x04C9, 'ERROR_CONNECTION_REFUSED') # WinError 1225 (0x4C9)
    ERROR_GRACEFUL_DISCONNECT: Final[ThirtyTwoBits] = cached_win32error(0x04CA, 'ERROR_GRACEFUL_DISCONNECT') # WinError 1226 (0x4CA)
    ERROR_ADDRESS_ALREADY_ASSOCIATED: Final[ThirtyTwoBits] = cached_win32error(0x04CB, 'ERROR_ADDRESS_ALREADY_ASSOCIATED') # WinError 1227 (0x4CB)
    ERROR_ADDRESS_NOT_ASSOCIATED: Final[ThirtyTwoBits] = cached_win32error(0x04CC, 'ERROR_ADDRESS_NOT_ASSOCIATED') # WinError 1228 (0x4CC)
    ERROR_CONNECTION_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x04CD, 'ERROR_CONNECTION_INVALID') # WinError 1229 (0x4CD)
    ERROR_CONNECTION_ACTIVE: Final[ThirtyTwoBits] = cached_win32error(0x04CE, 'ERROR_CONNECTION_ACTIVE') # WinError 1230 (0x4CE)
    ERROR_NETWORK_UNREACHABLE: Final[ThirtyTwoBits] = cached_win32error(0x04CF, 'ERROR_NETWORK_UNREACHABLE') # WinError 1231 (0x4CF)
    ERROR_HOST_UNREACHABLE: Final[ThirtyTwoBits] = cached_win32error(0x04D0, 'ERROR_HOST_UNREACHABLE') # WinError 1232 (0x4D0)
    ERROR_PROTOCOL_UNREACHABLE: Final[ThirtyTwoBits] = cached_win32error(0x04D1, 'ERROR_PROTOCOL_UNREACHABLE') # WinError 1233 (0x4D1)
    ERROR_PORT_UNREACHABLE: Final[ThirtyTwoBits] = cached_win32error(0x04D2, 'ERROR_PORT_UNREACHABLE') # WinError 1234 (0x4D2)
    ERROR_REQUEST_ABORTED: Final[ThirtyTwoBits] = cached_win32error(0x04D3, 'ERROR_REQUEST_ABORTED') # WinError 1235 (0x4D3)
    ERROR_CONNECTION_ABORTED: Final[ThirtyTwoBits] = cached_win32error(0x04D4, 'ERROR_CONNECTION_ABORTED') # WinError 1236 (0x4D4)
    ERROR_RETRY: Final[ThirtyTwoBits] = cached_win32error(0x04D5, 'ERROR_RETRY') # WinError 1237 (0x4D5)
    ERROR_CONNECTION_COUNT_LIMIT: Final[ThirtyTwoBits] = cached_win32error(0x04D6, 'ERROR_CONNECTION_COUNT_LIMIT') # WinError 1238 (0x4D6)
    ERROR_LOGIN_TIME_RESTRICTION: Final[ThirtyTwoBits] = cached_win32error(0x04D7, 'ERROR_LOGIN_TIME_RESTRICTION') # WinError 1239 (0x4D7)
    ERROR_LOGIN_WKSTA_RESTRICTION: Final[ThirtyTwoBits] = cached_win32error(0x04D8, 'ERROR_LOGIN_WKSTA_RESTRICTION') # WinError 1240 (0x4D8)
    ERROR_INCORRECT_ADDRESS: Final[ThirtyTwoBits] = cached_win32error(0x04D9, 'ERROR_INCORRECT_ADDRESS') # WinError 1241 (0x4D9)
    ERROR_ALREADY_REGISTERED: Final[ThirtyTwoBits] = cached_win32error(0x04DA, 'ERROR_ALREADY_REGISTERED') # WinError 1242 (0x4DA)
    ERROR_SERVICE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x04DB, 'ERROR_SERVICE_NOT_FOUND') # WinError 1243 (0x4DB)
    ERROR_NOT_AUTHENTICATED: Final[ThirtyTwoBits] = cached_win32error(0x04DC, 'ERROR_NOT_AUTHENTICATED') # WinError 1244 (0x4DC)
    ERROR_NOT_LOGGED_ON: Final[ThirtyTwoBits] = cached_win32error(0x04DD, 'ERROR_NOT_LOGGED_ON') # WinError 1245 (0x4DD)
    ERROR_CONTINUE: Final[ThirtyTwoBits] = cached_win32error(0x04DE, 'ERROR_CONTINUE') # WinError 1246 (0x4DE)
    ERROR_ALREADY_INITIALIZED: Final[ThirtyTwoBits] = cached_win32error(0x04DF, 'ERROR_ALREADY_INITIALIZED') # WinError 1247 (0x4DF)
    ERROR_NO_MORE_DEVICES: Final[ThirtyTwoBits] = cached_win32error(0x04E0, 'ERROR_NO_MORE_DEVICES') # WinError 1248 (0x4E0)
    ERROR_NO_SUCH_SITE: Final[ThirtyTwoBits] = cached_win32error(0x04E1, 'ERROR_NO_SUCH_SITE') # WinError 1249 (0x4E1)
    ERROR_DOMAIN_CONTROLLER_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x04E2, 'ERROR_DOMAIN_CONTROLLER_EXISTS') # WinError 1250 (0x4E2)
    ERROR_ONLY_IF_CONNECTED: Final[ThirtyTwoBits] = cached_win32error(0x04E3, 'ERROR_ONLY_IF_CONNECTED') # WinError 1251 (0x4E3)
    ERROR_OVERRIDE_NOCHANGES: Final[ThirtyTwoBits] = cached_win32error(0x04E4, 'ERROR_OVERRIDE_NOCHANGES') # WinError 1252 (0x4E4)
    ERROR_BAD_USER_PROFILE: Final[ThirtyTwoBits] = cached_win32error(0x04E5, 'ERROR_BAD_USER_PROFILE') # WinError 1253 (0x4E5)
    ERROR_NOT_SUPPORTED_ON_SBS: Final[ThirtyTwoBits] = cached_win32error(0x04E6, 'ERROR_NOT_SUPPORTED_ON_SBS') # WinError 1254 (0x4E6)
    ERROR_SERVER_SHUTDOWN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x04E7, 'ERROR_SERVER_SHUTDOWN_IN_PROGRESS') # WinError 1255 (0x4E7)
    ERROR_HOST_DOWN: Final[ThirtyTwoBits] = cached_win32error(0x04E8, 'ERROR_HOST_DOWN') # WinError 1256 (0x4E8)
    ERROR_NON_ACCOUNT_SID: Final[ThirtyTwoBits] = cached_win32error(0x04E9, 'ERROR_NON_ACCOUNT_SID') # WinError 1257 (0x4E9)
    ERROR_NON_DOMAIN_SID: Final[ThirtyTwoBits] = cached_win32error(0x04EA, 'ERROR_NON_DOMAIN_SID') # WinError 1258 (0x4EA)
    ERROR_APPHELP_BLOCK: Final[ThirtyTwoBits] = cached_win32error(0x04EB, 'ERROR_APPHELP_BLOCK') # WinError 1259 (0x4EB)
    ERROR_ACCESS_DISABLED_BY_POLICY: Final[ThirtyTwoBits] = cached_win32error(0x04EC, 'ERROR_ACCESS_DISABLED_BY_POLICY') # WinError 1260 (0x4EC)
    ERROR_REG_NAT_CONSUMPTION: Final[ThirtyTwoBits] = cached_win32error(0x04ED, 'ERROR_REG_NAT_CONSUMPTION') # WinError 1261 (0x4ED)
    ERROR_CSCSHARE_OFFLINE: Final[ThirtyTwoBits] = cached_win32error(0x04EE, 'ERROR_CSCSHARE_OFFLINE') # WinError 1262 (0x4EE)
    ERROR_PKINIT_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x04EF, 'ERROR_PKINIT_FAILURE') # WinError 1263 (0x4EF)
    ERROR_SMARTCARD_SUBSYSTEM_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x04F0, 'ERROR_SMARTCARD_SUBSYSTEM_FAILURE') # WinError 1264 (0x4F0)
    ERROR_DOWNGRADE_DETECTED: Final[ThirtyTwoBits] = cached_win32error(0x04F1, 'ERROR_DOWNGRADE_DETECTED') # WinError 1265 (0x4F1)
    ERROR_MACHINE_LOCKED: Final[ThirtyTwoBits] = cached_win32error(0x04F7, 'ERROR_MACHINE_LOCKED') # WinError 1271 (0x4F7)
    ERROR_SMB_GUEST_LOGON_BLOCKED: Final[ThirtyTwoBits] = cached_win32error(0x04F8, 'ERROR_SMB_GUEST_LOGON_BLOCKED') # WinError 1272 (0x4F8)
    ERROR_CALLBACK_SUPPLIED_INVALID_DATA: Final[ThirtyTwoBits] = cached_win32error(0x04F9, 'ERROR_CALLBACK_SUPPLIED_INVALID_DATA') # WinError 1273 (0x4F9)
    ERROR_SYNC_FOREGROUND_REFRESH_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x04FA, 'ERROR_SYNC_FOREGROUND_REFRESH_REQUIRED') # WinError 1274 (0x4FA)
    ERROR_DRIVER_BLOCKED: Final[ThirtyTwoBits] = cached_win32error(0x04FB, 'ERROR_DRIVER_BLOCKED') # WinError 1275 (0x4FB)
    ERROR_INVALID_IMPORT_OF_NON_DLL: Final[ThirtyTwoBits] = cached_win32error(0x04FC, 'ERROR_INVALID_IMPORT_OF_NON_DLL') # WinError 1276 (0x4FC)
    ERROR_ACCESS_DISABLED_WEBBLADE: Final[ThirtyTwoBits] = cached_win32error(0x04FD, 'ERROR_ACCESS_DISABLED_WEBBLADE') # WinError 1277 (0x4FD)
    ERROR_ACCESS_DISABLED_WEBBLADE_TAMPER: Final[ThirtyTwoBits] = cached_win32error(0x04FE, 'ERROR_ACCESS_DISABLED_WEBBLADE_TAMPER') # WinError 1278 (0x4FE)
    ERROR_RECOVERY_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x04FF, 'ERROR_RECOVERY_FAILURE') # WinError 1279 (0x4FF)
    ERROR_ALREADY_FIBER: Final[ThirtyTwoBits] = cached_win32error(0x0500, 'ERROR_ALREADY_FIBER') # WinError 1280 (0x500)
    ERROR_ALREADY_THREAD: Final[ThirtyTwoBits] = cached_win32error(0x0501, 'ERROR_ALREADY_THREAD') # WinError 1281 (0x501)
    ERROR_STACK_BUFFER_OVERRUN: Final[ThirtyTwoBits] = cached_win32error(0x0502, 'ERROR_STACK_BUFFER_OVERRUN') # WinError 1282 (0x502)
    ERROR_PARAMETER_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x0503, 'ERROR_PARAMETER_QUOTA_EXCEEDED') # WinError 1283 (0x503)
    ERROR_DEBUGGER_INACTIVE: Final[ThirtyTwoBits] = cached_win32error(0x0504, 'ERROR_DEBUGGER_INACTIVE') # WinError 1284 (0x504)
    ERROR_DELAY_LOAD_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x0505, 'ERROR_DELAY_LOAD_FAILED') # WinError 1285 (0x505)
    ERROR_VDM_DISALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x0506, 'ERROR_VDM_DISALLOWED') # WinError 1286 (0x506)
    ERROR_UNIDENTIFIED_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x0507, 'ERROR_UNIDENTIFIED_ERROR') # WinError 1287 (0x507)
    ERROR_INVALID_CRUNTIME_PARAMETER: Final[ThirtyTwoBits] = cached_win32error(0x0508, 'ERROR_INVALID_CRUNTIME_PARAMETER') # WinError 1288 (0x508)
    ERROR_BEYOND_VDL: Final[ThirtyTwoBits] = cached_win32error(0x0509, 'ERROR_BEYOND_VDL') # WinError 1289 (0x509)
    ERROR_INCOMPATIBLE_SERVICE_SID_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x050A, 'ERROR_INCOMPATIBLE_SERVICE_SID_TYPE') # WinError 1290 (0x50A)
    ERROR_DRIVER_PROCESS_TERMINATED: Final[ThirtyTwoBits] = cached_win32error(0x050B, 'ERROR_DRIVER_PROCESS_TERMINATED') # WinError 1291 (0x50B)
    ERROR_IMPLEMENTATION_LIMIT: Final[ThirtyTwoBits] = cached_win32error(0x050C, 'ERROR_IMPLEMENTATION_LIMIT') # WinError 1292 (0x50C)
    ERROR_PROCESS_IS_PROTECTED: Final[ThirtyTwoBits] = cached_win32error(0x050D, 'ERROR_PROCESS_IS_PROTECTED') # WinError 1293 (0x50D)
    ERROR_SERVICE_NOTIFY_CLIENT_LAGGING: Final[ThirtyTwoBits] = cached_win32error(0x050E, 'ERROR_SERVICE_NOTIFY_CLIENT_LAGGING') # WinError 1294 (0x50E)
    ERROR_DISK_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x050F, 'ERROR_DISK_QUOTA_EXCEEDED') # WinError 1295 (0x50F)
    ERROR_CONTENT_BLOCKED: Final[ThirtyTwoBits] = cached_win32error(0x0510, 'ERROR_CONTENT_BLOCKED') # WinError 1296 (0x510)
    ERROR_INCOMPATIBLE_SERVICE_PRIVILEGE: Final[ThirtyTwoBits] = cached_win32error(0x0511, 'ERROR_INCOMPATIBLE_SERVICE_PRIVILEGE') # WinError 1297 (0x511)
    ERROR_APP_HANG: Final[ThirtyTwoBits] = cached_win32error(0x0512, 'ERROR_APP_HANG') # WinError 1298 (0x512)
    ERROR_INVALID_LABEL: Final[ThirtyTwoBits] = cached_win32error(0x0513, 'ERROR_INVALID_LABEL') # WinError 1299 (0x513)
    ERROR_NOT_ALL_ASSIGNED: Final[ThirtyTwoBits] = cached_win32error(0x0514, 'ERROR_NOT_ALL_ASSIGNED') # WinError 1300 (0x514)
    ERROR_SOME_NOT_MAPPED: Final[ThirtyTwoBits] = cached_win32error(0x0515, 'ERROR_SOME_NOT_MAPPED') # WinError 1301 (0x515)
    ERROR_NO_QUOTAS_FOR_ACCOUNT: Final[ThirtyTwoBits] = cached_win32error(0x0516, 'ERROR_NO_QUOTAS_FOR_ACCOUNT') # WinError 1302 (0x516)
    ERROR_LOCAL_USER_SESSION_KEY: Final[ThirtyTwoBits] = cached_win32error(0x0517, 'ERROR_LOCAL_USER_SESSION_KEY') # WinError 1303 (0x517)
    ERROR_NULL_LM_PASSWORD: Final[ThirtyTwoBits] = cached_win32error(0x0518, 'ERROR_NULL_LM_PASSWORD') # WinError 1304 (0x518)
    ERROR_UNKNOWN_REVISION: Final[ThirtyTwoBits] = cached_win32error(0x0519, 'ERROR_UNKNOWN_REVISION') # WinError 1305 (0x519)
    ERROR_REVISION_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x051A, 'ERROR_REVISION_MISMATCH') # WinError 1306 (0x51A)
    ERROR_INVALID_OWNER: Final[ThirtyTwoBits] = cached_win32error(0x051B, 'ERROR_INVALID_OWNER') # WinError 1307 (0x51B)
    ERROR_INVALID_PRIMARY_GROUP: Final[ThirtyTwoBits] = cached_win32error(0x051C, 'ERROR_INVALID_PRIMARY_GROUP') # WinError 1308 (0x51C)
    ERROR_NO_IMPERSONATION_TOKEN: Final[ThirtyTwoBits] = cached_win32error(0x051D, 'ERROR_NO_IMPERSONATION_TOKEN') # WinError 1309 (0x51D)
    ERROR_CANT_DISABLE_MANDATORY: Final[ThirtyTwoBits] = cached_win32error(0x051E, 'ERROR_CANT_DISABLE_MANDATORY') # WinError 1310 (0x51E)
    ERROR_NO_LOGON_SERVERS: Final[ThirtyTwoBits] = cached_win32error(0x051F, 'ERROR_NO_LOGON_SERVERS') # WinError 1311 (0x51F)
    ERROR_NO_SUCH_LOGON_SESSION: Final[ThirtyTwoBits] = cached_win32error(0x0520, 'ERROR_NO_SUCH_LOGON_SESSION') # WinError 1312 (0x520)
    ERROR_NO_SUCH_PRIVILEGE: Final[ThirtyTwoBits] = cached_win32error(0x0521, 'ERROR_NO_SUCH_PRIVILEGE') # WinError 1313 (0x521)
    ERROR_PRIVILEGE_NOT_HELD: Final[ThirtyTwoBits] = cached_win32error(0x0522, 'ERROR_PRIVILEGE_NOT_HELD') # WinError 1314 (0x522)
    ERROR_INVALID_ACCOUNT_NAME: Final[ThirtyTwoBits] = cached_win32error(0x0523, 'ERROR_INVALID_ACCOUNT_NAME') # WinError 1315 (0x523)
    ERROR_USER_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x0524, 'ERROR_USER_EXISTS') # WinError 1316 (0x524)
    ERROR_NO_SUCH_USER: Final[ThirtyTwoBits] = cached_win32error(0x0525, 'ERROR_NO_SUCH_USER') # WinError 1317 (0x525)
    ERROR_GROUP_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x0526, 'ERROR_GROUP_EXISTS') # WinError 1318 (0x526)
    ERROR_NO_SUCH_GROUP: Final[ThirtyTwoBits] = cached_win32error(0x0527, 'ERROR_NO_SUCH_GROUP') # WinError 1319 (0x527)
    ERROR_MEMBER_IN_GROUP: Final[ThirtyTwoBits] = cached_win32error(0x0528, 'ERROR_MEMBER_IN_GROUP') # WinError 1320 (0x528)
    ERROR_MEMBER_NOT_IN_GROUP: Final[ThirtyTwoBits] = cached_win32error(0x0529, 'ERROR_MEMBER_NOT_IN_GROUP') # WinError 1321 (0x529)
    ERROR_LAST_ADMIN: Final[ThirtyTwoBits] = cached_win32error(0x052A, 'ERROR_LAST_ADMIN') # WinError 1322 (0x52A)
    ERROR_WRONG_PASSWORD: Final[ThirtyTwoBits] = cached_win32error(0x052B, 'ERROR_WRONG_PASSWORD') # WinError 1323 (0x52B)
    ERROR_ILL_FORMED_PASSWORD: Final[ThirtyTwoBits] = cached_win32error(0x052C, 'ERROR_ILL_FORMED_PASSWORD') # WinError 1324 (0x52C)
    ERROR_PASSWORD_RESTRICTION: Final[ThirtyTwoBits] = cached_win32error(0x052D, 'ERROR_PASSWORD_RESTRICTION') # WinError 1325 (0x52D)
    ERROR_LOGON_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x052E, 'ERROR_LOGON_FAILURE') # WinError 1326 (0x52E)
    ERROR_ACCOUNT_RESTRICTION: Final[ThirtyTwoBits] = cached_win32error(0x052F, 'ERROR_ACCOUNT_RESTRICTION') # WinError 1327 (0x52F)
    ERROR_INVALID_LOGON_HOURS: Final[ThirtyTwoBits] = cached_win32error(0x0530, 'ERROR_INVALID_LOGON_HOURS') # WinError 1328 (0x530)
    ERROR_INVALID_WORKSTATION: Final[ThirtyTwoBits] = cached_win32error(0x0531, 'ERROR_INVALID_WORKSTATION') # WinError 1329 (0x531)
    ERROR_PASSWORD_EXPIRED: Final[ThirtyTwoBits] = cached_win32error(0x0532, 'ERROR_PASSWORD_EXPIRED') # WinError 1330 (0x532)
    ERROR_ACCOUNT_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x0533, 'ERROR_ACCOUNT_DISABLED') # WinError 1331 (0x533)
    ERROR_NONE_MAPPED: Final[ThirtyTwoBits] = cached_win32error(0x0534, 'ERROR_NONE_MAPPED') # WinError 1332 (0x534)
    ERROR_TOO_MANY_LUIDS_REQUESTED: Final[ThirtyTwoBits] = cached_win32error(0x0535, 'ERROR_TOO_MANY_LUIDS_REQUESTED') # WinError 1333 (0x535)
    ERROR_LUIDS_EXHAUSTED: Final[ThirtyTwoBits] = cached_win32error(0x0536, 'ERROR_LUIDS_EXHAUSTED') # WinError 1334 (0x536)
    ERROR_INVALID_SUB_AUTHORITY: Final[ThirtyTwoBits] = cached_win32error(0x0537, 'ERROR_INVALID_SUB_AUTHORITY') # WinError 1335 (0x537)
    ERROR_INVALID_ACL: Final[ThirtyTwoBits] = cached_win32error(0x0538, 'ERROR_INVALID_ACL') # WinError 1336 (0x538)
    ERROR_INVALID_SID: Final[ThirtyTwoBits] = cached_win32error(0x0539, 'ERROR_INVALID_SID') # WinError 1337 (0x539)
    ERROR_INVALID_SECURITY_DESCR: Final[ThirtyTwoBits] = cached_win32error(0x053A, 'ERROR_INVALID_SECURITY_DESCR') # WinError 1338 (0x53A)
    ERROR_BAD_INHERITANCE_ACL: Final[ThirtyTwoBits] = cached_win32error(0x053C, 'ERROR_BAD_INHERITANCE_ACL') # WinError 1340 (0x53C)
    ERROR_SERVER_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x053D, 'ERROR_SERVER_DISABLED') # WinError 1341 (0x53D)
    ERROR_SERVER_NOT_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x053E, 'ERROR_SERVER_NOT_DISABLED') # WinError 1342 (0x53E)
    ERROR_INVALID_ID_AUTHORITY: Final[ThirtyTwoBits] = cached_win32error(0x053F, 'ERROR_INVALID_ID_AUTHORITY') # WinError 1343 (0x53F)
    ERROR_ALLOTTED_SPACE_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x0540, 'ERROR_ALLOTTED_SPACE_EXCEEDED') # WinError 1344 (0x540)
    ERROR_INVALID_GROUP_ATTRIBUTES: Final[ThirtyTwoBits] = cached_win32error(0x0541, 'ERROR_INVALID_GROUP_ATTRIBUTES') # WinError 1345 (0x541)
    ERROR_BAD_IMPERSONATION_LEVEL: Final[ThirtyTwoBits] = cached_win32error(0x0542, 'ERROR_BAD_IMPERSONATION_LEVEL') # WinError 1346 (0x542)
    ERROR_CANT_OPEN_ANONYMOUS: Final[ThirtyTwoBits] = cached_win32error(0x0543, 'ERROR_CANT_OPEN_ANONYMOUS') # WinError 1347 (0x543)
    ERROR_BAD_VALIDATION_CLASS: Final[ThirtyTwoBits] = cached_win32error(0x0544, 'ERROR_BAD_VALIDATION_CLASS') # WinError 1348 (0x544)
    ERROR_BAD_TOKEN_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x0545, 'ERROR_BAD_TOKEN_TYPE') # WinError 1349 (0x545)
    ERROR_NO_SECURITY_ON_OBJECT: Final[ThirtyTwoBits] = cached_win32error(0x0546, 'ERROR_NO_SECURITY_ON_OBJECT') # WinError 1350 (0x546)
    ERROR_CANT_ACCESS_DOMAIN_INFO: Final[ThirtyTwoBits] = cached_win32error(0x0547, 'ERROR_CANT_ACCESS_DOMAIN_INFO') # WinError 1351 (0x547)
    ERROR_INVALID_SERVER_STATE: Final[ThirtyTwoBits] = cached_win32error(0x0548, 'ERROR_INVALID_SERVER_STATE') # WinError 1352 (0x548)
    ERROR_INVALID_DOMAIN_STATE: Final[ThirtyTwoBits] = cached_win32error(0x0549, 'ERROR_INVALID_DOMAIN_STATE') # WinError 1353 (0x549)
    ERROR_INVALID_DOMAIN_ROLE: Final[ThirtyTwoBits] = cached_win32error(0x054A, 'ERROR_INVALID_DOMAIN_ROLE') # WinError 1354 (0x54A)
    ERROR_NO_SUCH_DOMAIN: Final[ThirtyTwoBits] = cached_win32error(0x054B, 'ERROR_NO_SUCH_DOMAIN') # WinError 1355 (0x54B)
    ERROR_DOMAIN_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x054C, 'ERROR_DOMAIN_EXISTS') # WinError 1356 (0x54C)
    ERROR_DOMAIN_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x054D, 'ERROR_DOMAIN_LIMIT_EXCEEDED') # WinError 1357 (0x54D)
    ERROR_INTERNAL_DB_CORRUPTION: Final[ThirtyTwoBits] = cached_win32error(0x054E, 'ERROR_INTERNAL_DB_CORRUPTION') # WinError 1358 (0x54E)
    ERROR_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x054F, 'ERROR_INTERNAL_ERROR') # WinError 1359 (0x54F)
    ERROR_GENERIC_NOT_MAPPED: Final[ThirtyTwoBits] = cached_win32error(0x0550, 'ERROR_GENERIC_NOT_MAPPED') # WinError 1360 (0x550)
    ERROR_BAD_DESCRIPTOR_FORMAT: Final[ThirtyTwoBits] = cached_win32error(0x0551, 'ERROR_BAD_DESCRIPTOR_FORMAT') # WinError 1361 (0x551)
    ERROR_NOT_LOGON_PROCESS: Final[ThirtyTwoBits] = cached_win32error(0x0552, 'ERROR_NOT_LOGON_PROCESS') # WinError 1362 (0x552)
    ERROR_LOGON_SESSION_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x0553, 'ERROR_LOGON_SESSION_EXISTS') # WinError 1363 (0x553)
    ERROR_NO_SUCH_PACKAGE: Final[ThirtyTwoBits] = cached_win32error(0x0554, 'ERROR_NO_SUCH_PACKAGE') # WinError 1364 (0x554)
    ERROR_BAD_LOGON_SESSION_STATE: Final[ThirtyTwoBits] = cached_win32error(0x0555, 'ERROR_BAD_LOGON_SESSION_STATE') # WinError 1365 (0x555)
    ERROR_LOGON_SESSION_COLLISION: Final[ThirtyTwoBits] = cached_win32error(0x0556, 'ERROR_LOGON_SESSION_COLLISION') # WinError 1366 (0x556)
    ERROR_INVALID_LOGON_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x0557, 'ERROR_INVALID_LOGON_TYPE') # WinError 1367 (0x557)
    ERROR_CANNOT_IMPERSONATE: Final[ThirtyTwoBits] = cached_win32error(0x0558, 'ERROR_CANNOT_IMPERSONATE') # WinError 1368 (0x558)
    ERROR_RXACT_INVALID_STATE: Final[ThirtyTwoBits] = cached_win32error(0x0559, 'ERROR_RXACT_INVALID_STATE') # WinError 1369 (0x559)
    ERROR_RXACT_COMMIT_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x055A, 'ERROR_RXACT_COMMIT_FAILURE') # WinError 1370 (0x55A)
    ERROR_SPECIAL_ACCOUNT: Final[ThirtyTwoBits] = cached_win32error(0x055B, 'ERROR_SPECIAL_ACCOUNT') # WinError 1371 (0x55B)
    ERROR_SPECIAL_GROUP: Final[ThirtyTwoBits] = cached_win32error(0x055C, 'ERROR_SPECIAL_GROUP') # WinError 1372 (0x55C)
    ERROR_SPECIAL_USER: Final[ThirtyTwoBits] = cached_win32error(0x055D, 'ERROR_SPECIAL_USER') # WinError 1373 (0x55D)
    ERROR_MEMBERS_PRIMARY_GROUP: Final[ThirtyTwoBits] = cached_win32error(0x055E, 'ERROR_MEMBERS_PRIMARY_GROUP') # WinError 1374 (0x55E)
    ERROR_TOKEN_ALREADY_IN_USE: Final[ThirtyTwoBits] = cached_win32error(0x055F, 'ERROR_TOKEN_ALREADY_IN_USE') # WinError 1375 (0x55F)
    ERROR_NO_SUCH_ALIAS: Final[ThirtyTwoBits] = cached_win32error(0x0560, 'ERROR_NO_SUCH_ALIAS') # WinError 1376 (0x560)
    ERROR_MEMBER_NOT_IN_ALIAS: Final[ThirtyTwoBits] = cached_win32error(0x0561, 'ERROR_MEMBER_NOT_IN_ALIAS') # WinError 1377 (0x561)
    ERROR_MEMBER_IN_ALIAS: Final[ThirtyTwoBits] = cached_win32error(0x0562, 'ERROR_MEMBER_IN_ALIAS') # WinError 1378 (0x562)
    ERROR_ALIAS_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x0563, 'ERROR_ALIAS_EXISTS') # WinError 1379 (0x563)
    ERROR_LOGON_NOT_GRANTED: Final[ThirtyTwoBits] = cached_win32error(0x0564, 'ERROR_LOGON_NOT_GRANTED') # WinError 1380 (0x564)
    ERROR_TOO_MANY_SECRETS: Final[ThirtyTwoBits] = cached_win32error(0x0565, 'ERROR_TOO_MANY_SECRETS') # WinError 1381 (0x565)
    ERROR_SECRET_TOO_LONG: Final[ThirtyTwoBits] = cached_win32error(0x0566, 'ERROR_SECRET_TOO_LONG') # WinError 1382 (0x566)
    ERROR_INTERNAL_DB_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x0567, 'ERROR_INTERNAL_DB_ERROR') # WinError 1383 (0x567)
    ERROR_TOO_MANY_CONTEXT_IDS: Final[ThirtyTwoBits] = cached_win32error(0x0568, 'ERROR_TOO_MANY_CONTEXT_IDS') # WinError 1384 (0x568)
    ERROR_LOGON_TYPE_NOT_GRANTED: Final[ThirtyTwoBits] = cached_win32error(0x0569, 'ERROR_LOGON_TYPE_NOT_GRANTED') # WinError 1385 (0x569)
    ERROR_NT_CROSS_ENCRYPTION_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x056A, 'ERROR_NT_CROSS_ENCRYPTION_REQUIRED') # WinError 1386 (0x56A)
    ERROR_NO_SUCH_MEMBER: Final[ThirtyTwoBits] = cached_win32error(0x056B, 'ERROR_NO_SUCH_MEMBER') # WinError 1387 (0x56B)
    ERROR_INVALID_MEMBER: Final[ThirtyTwoBits] = cached_win32error(0x056C, 'ERROR_INVALID_MEMBER') # WinError 1388 (0x56C)
    ERROR_TOO_MANY_SIDS: Final[ThirtyTwoBits] = cached_win32error(0x056D, 'ERROR_TOO_MANY_SIDS') # WinError 1389 (0x56D)
    ERROR_LM_CROSS_ENCRYPTION_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x056E, 'ERROR_LM_CROSS_ENCRYPTION_REQUIRED') # WinError 1390 (0x56E)
    ERROR_NO_INHERITANCE: Final[ThirtyTwoBits] = cached_win32error(0x056F, 'ERROR_NO_INHERITANCE') # WinError 1391 (0x56F)
    ERROR_FILE_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x0570, 'ERROR_FILE_CORRUPT') # WinError 1392 (0x570)
    ERROR_DISK_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x0571, 'ERROR_DISK_CORRUPT') # WinError 1393 (0x571)
    ERROR_NO_USER_SESSION_KEY: Final[ThirtyTwoBits] = cached_win32error(0x0572, 'ERROR_NO_USER_SESSION_KEY') # WinError 1394 (0x572)
    ERROR_LICENSE_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x0573, 'ERROR_LICENSE_QUOTA_EXCEEDED') # WinError 1395 (0x573)
    ERROR_WRONG_TARGET_NAME: Final[ThirtyTwoBits] = cached_win32error(0x0574, 'ERROR_WRONG_TARGET_NAME') # WinError 1396 (0x574)
    ERROR_MUTUAL_AUTH_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x0575, 'ERROR_MUTUAL_AUTH_FAILED') # WinError 1397 (0x575)
    ERROR_TIME_SKEW: Final[ThirtyTwoBits] = cached_win32error(0x0576, 'ERROR_TIME_SKEW') # WinError 1398 (0x576)
    ERROR_CURRENT_DOMAIN_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x0577, 'ERROR_CURRENT_DOMAIN_NOT_ALLOWED') # WinError 1399 (0x577)
    ERROR_INVALID_WINDOW_HANDLE: Final[ThirtyTwoBits] = cached_win32error(0x0578, 'ERROR_INVALID_WINDOW_HANDLE') # WinError 1400 (0x578)
    ERROR_INVALID_MENU_HANDLE: Final[ThirtyTwoBits] = cached_win32error(0x0579, 'ERROR_INVALID_MENU_HANDLE') # WinError 1401 (0x579)
    ERROR_INVALID_CURSOR_HANDLE: Final[ThirtyTwoBits] = cached_win32error(0x057A, 'ERROR_INVALID_CURSOR_HANDLE') # WinError 1402 (0x57A)
    ERROR_INVALID_ACCEL_HANDLE: Final[ThirtyTwoBits] = cached_win32error(0x057B, 'ERROR_INVALID_ACCEL_HANDLE') # WinError 1403 (0x57B)
    ERROR_INVALID_HOOK_HANDLE: Final[ThirtyTwoBits] = cached_win32error(0x057C, 'ERROR_INVALID_HOOK_HANDLE') # WinError 1404 (0x57C)
    ERROR_INVALID_DWP_HANDLE: Final[ThirtyTwoBits] = cached_win32error(0x057D, 'ERROR_INVALID_DWP_HANDLE') # WinError 1405 (0x57D)
    ERROR_TLW_WITH_WSCHILD: Final[ThirtyTwoBits] = cached_win32error(0x057E, 'ERROR_TLW_WITH_WSCHILD') # WinError 1406 (0x57E)
    ERROR_CANNOT_FIND_WND_CLASS: Final[ThirtyTwoBits] = cached_win32error(0x057F, 'ERROR_CANNOT_FIND_WND_CLASS') # WinError 1407 (0x57F)
    ERROR_WINDOW_OF_OTHER_THREAD: Final[ThirtyTwoBits] = cached_win32error(0x0580, 'ERROR_WINDOW_OF_OTHER_THREAD') # WinError 1408 (0x580)
    ERROR_HOTKEY_ALREADY_REGISTERED: Final[ThirtyTwoBits] = cached_win32error(0x0581, 'ERROR_HOTKEY_ALREADY_REGISTERED') # WinError 1409 (0x581)
    ERROR_CLASS_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x0582, 'ERROR_CLASS_ALREADY_EXISTS') # WinError 1410 (0x582)
    ERROR_CLASS_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_win32error(0x0583, 'ERROR_CLASS_DOES_NOT_EXIST') # WinError 1411 (0x583)
    ERROR_CLASS_HAS_WINDOWS: Final[ThirtyTwoBits] = cached_win32error(0x0584, 'ERROR_CLASS_HAS_WINDOWS') # WinError 1412 (0x584)
    ERROR_INVALID_INDEX: Final[ThirtyTwoBits] = cached_win32error(0x0585, 'ERROR_INVALID_INDEX') # WinError 1413 (0x585)
    ERROR_INVALID_ICON_HANDLE: Final[ThirtyTwoBits] = cached_win32error(0x0586, 'ERROR_INVALID_ICON_HANDLE') # WinError 1414 (0x586)
    ERROR_PRIVATE_DIALOG_INDEX: Final[ThirtyTwoBits] = cached_win32error(0x0587, 'ERROR_PRIVATE_DIALOG_INDEX') # WinError 1415 (0x587)
    ERROR_LISTBOX_ID_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x0588, 'ERROR_LISTBOX_ID_NOT_FOUND') # WinError 1416 (0x588)
    ERROR_NO_WILDCARD_CHARACTERS: Final[ThirtyTwoBits] = cached_win32error(0x0589, 'ERROR_NO_WILDCARD_CHARACTERS') # WinError 1417 (0x589)
    ERROR_CLIPBOARD_NOT_OPEN: Final[ThirtyTwoBits] = cached_win32error(0x058A, 'ERROR_CLIPBOARD_NOT_OPEN') # WinError 1418 (0x58A)
    ERROR_HOTKEY_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_win32error(0x058B, 'ERROR_HOTKEY_NOT_REGISTERED') # WinError 1419 (0x58B)
    ERROR_WINDOW_NOT_DIALOG: Final[ThirtyTwoBits] = cached_win32error(0x058C, 'ERROR_WINDOW_NOT_DIALOG') # WinError 1420 (0x58C)
    ERROR_CONTROL_ID_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x058D, 'ERROR_CONTROL_ID_NOT_FOUND') # WinError 1421 (0x58D)
    ERROR_INVALID_COMBOBOX_MESSAGE: Final[ThirtyTwoBits] = cached_win32error(0x058E, 'ERROR_INVALID_COMBOBOX_MESSAGE') # WinError 1422 (0x58E)
    ERROR_WINDOW_NOT_COMBOBOX: Final[ThirtyTwoBits] = cached_win32error(0x058F, 'ERROR_WINDOW_NOT_COMBOBOX') # WinError 1423 (0x58F)
    ERROR_INVALID_EDIT_HEIGHT: Final[ThirtyTwoBits] = cached_win32error(0x0590, 'ERROR_INVALID_EDIT_HEIGHT') # WinError 1424 (0x590)
    ERROR_DC_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x0591, 'ERROR_DC_NOT_FOUND') # WinError 1425 (0x591)
    ERROR_INVALID_HOOK_FILTER: Final[ThirtyTwoBits] = cached_win32error(0x0592, 'ERROR_INVALID_HOOK_FILTER') # WinError 1426 (0x592)
    ERROR_INVALID_FILTER_PROC: Final[ThirtyTwoBits] = cached_win32error(0x0593, 'ERROR_INVALID_FILTER_PROC') # WinError 1427 (0x593)
    ERROR_HOOK_NEEDS_HMOD: Final[ThirtyTwoBits] = cached_win32error(0x0594, 'ERROR_HOOK_NEEDS_HMOD') # WinError 1428 (0x594)
    ERROR_GLOBAL_ONLY_HOOK: Final[ThirtyTwoBits] = cached_win32error(0x0595, 'ERROR_GLOBAL_ONLY_HOOK') # WinError 1429 (0x595)
    ERROR_JOURNAL_HOOK_SET: Final[ThirtyTwoBits] = cached_win32error(0x0596, 'ERROR_JOURNAL_HOOK_SET') # WinError 1430 (0x596)
    ERROR_HOOK_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_win32error(0x0597, 'ERROR_HOOK_NOT_INSTALLED') # WinError 1431 (0x597)
    ERROR_INVALID_LB_MESSAGE: Final[ThirtyTwoBits] = cached_win32error(0x0598, 'ERROR_INVALID_LB_MESSAGE') # WinError 1432 (0x598)
    ERROR_SETCOUNT_ON_BAD_LB: Final[ThirtyTwoBits] = cached_win32error(0x0599, 'ERROR_SETCOUNT_ON_BAD_LB') # WinError 1433 (0x599)
    ERROR_LB_WITHOUT_TABSTOPS: Final[ThirtyTwoBits] = cached_win32error(0x059A, 'ERROR_LB_WITHOUT_TABSTOPS') # WinError 1434 (0x59A)
    ERROR_DESTROY_OBJECT_OF_OTHER_THREAD: Final[ThirtyTwoBits] = cached_win32error(0x059B, 'ERROR_DESTROY_OBJECT_OF_OTHER_THREAD') # WinError 1435 (0x59B)
    ERROR_CHILD_WINDOW_MENU: Final[ThirtyTwoBits] = cached_win32error(0x059C, 'ERROR_CHILD_WINDOW_MENU') # WinError 1436 (0x59C)
    ERROR_NO_SYSTEM_MENU: Final[ThirtyTwoBits] = cached_win32error(0x059D, 'ERROR_NO_SYSTEM_MENU') # WinError 1437 (0x59D)
    ERROR_INVALID_MSGBOX_STYLE: Final[ThirtyTwoBits] = cached_win32error(0x059E, 'ERROR_INVALID_MSGBOX_STYLE') # WinError 1438 (0x59E)
    ERROR_INVALID_SPI_VALUE: Final[ThirtyTwoBits] = cached_win32error(0x059F, 'ERROR_INVALID_SPI_VALUE') # WinError 1439 (0x59F)
    ERROR_SCREEN_ALREADY_LOCKED: Final[ThirtyTwoBits] = cached_win32error(0x05A0, 'ERROR_SCREEN_ALREADY_LOCKED') # WinError 1440 (0x5A0)
    ERROR_HWNDS_HAVE_DIFF_PARENT: Final[ThirtyTwoBits] = cached_win32error(0x05A1, 'ERROR_HWNDS_HAVE_DIFF_PARENT') # WinError 1441 (0x5A1)
    ERROR_NOT_CHILD_WINDOW: Final[ThirtyTwoBits] = cached_win32error(0x05A2, 'ERROR_NOT_CHILD_WINDOW') # WinError 1442 (0x5A2)
    ERROR_INVALID_GW_COMMAND: Final[ThirtyTwoBits] = cached_win32error(0x05A3, 'ERROR_INVALID_GW_COMMAND') # WinError 1443 (0x5A3)
    ERROR_INVALID_THREAD_ID: Final[ThirtyTwoBits] = cached_win32error(0x05A4, 'ERROR_INVALID_THREAD_ID') # WinError 1444 (0x5A4)
    ERROR_NON_MDICHILD_WINDOW: Final[ThirtyTwoBits] = cached_win32error(0x05A5, 'ERROR_NON_MDICHILD_WINDOW') # WinError 1445 (0x5A5)
    ERROR_POPUP_ALREADY_ACTIVE: Final[ThirtyTwoBits] = cached_win32error(0x05A6, 'ERROR_POPUP_ALREADY_ACTIVE') # WinError 1446 (0x5A6)
    ERROR_NO_SCROLLBARS: Final[ThirtyTwoBits] = cached_win32error(0x05A7, 'ERROR_NO_SCROLLBARS') # WinError 1447 (0x5A7)
    ERROR_INVALID_SCROLLBAR_RANGE: Final[ThirtyTwoBits] = cached_win32error(0x05A8, 'ERROR_INVALID_SCROLLBAR_RANGE') # WinError 1448 (0x5A8)
    ERROR_INVALID_SHOWWIN_COMMAND: Final[ThirtyTwoBits] = cached_win32error(0x05A9, 'ERROR_INVALID_SHOWWIN_COMMAND') # WinError 1449 (0x5A9)
    ERROR_NO_SYSTEM_RESOURCES: Final[ThirtyTwoBits] = cached_win32error(0x05AA, 'ERROR_NO_SYSTEM_RESOURCES') # WinError 1450 (0x5AA)
    ERROR_NONPAGED_SYSTEM_RESOURCES: Final[ThirtyTwoBits] = cached_win32error(0x05AB, 'ERROR_NONPAGED_SYSTEM_RESOURCES') # WinError 1451 (0x5AB)
    ERROR_PAGED_SYSTEM_RESOURCES: Final[ThirtyTwoBits] = cached_win32error(0x05AC, 'ERROR_PAGED_SYSTEM_RESOURCES') # WinError 1452 (0x5AC)
    ERROR_WORKING_SET_QUOTA: Final[ThirtyTwoBits] = cached_win32error(0x05AD, 'ERROR_WORKING_SET_QUOTA') # WinError 1453 (0x5AD)
    ERROR_PAGEFILE_QUOTA: Final[ThirtyTwoBits] = cached_win32error(0x05AE, 'ERROR_PAGEFILE_QUOTA') # WinError 1454 (0x5AE)
    ERROR_COMMITMENT_LIMIT: Final[ThirtyTwoBits] = cached_win32error(0x05AF, 'ERROR_COMMITMENT_LIMIT') # WinError 1455 (0x5AF)
    ERROR_MENU_ITEM_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x05B0, 'ERROR_MENU_ITEM_NOT_FOUND') # WinError 1456 (0x5B0)
    ERROR_INVALID_KEYBOARD_HANDLE: Final[ThirtyTwoBits] = cached_win32error(0x05B1, 'ERROR_INVALID_KEYBOARD_HANDLE') # WinError 1457 (0x5B1)
    ERROR_HOOK_TYPE_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x05B2, 'ERROR_HOOK_TYPE_NOT_ALLOWED') # WinError 1458 (0x5B2)
    ERROR_REQUIRES_INTERACTIVE_WINDOWSTATION: Final[ThirtyTwoBits] = cached_win32error(0x05B3, 'ERROR_REQUIRES_INTERACTIVE_WINDOWSTATION') # WinError 1459 (0x5B3)
    ERROR_TIMEOUT: Final[ThirtyTwoBits] = cached_win32error(0x05B4, 'ERROR_TIMEOUT') # WinError 1460 (0x5B4)
    ERROR_INVALID_MONITOR_HANDLE: Final[ThirtyTwoBits] = cached_win32error(0x05B5, 'ERROR_INVALID_MONITOR_HANDLE') # WinError 1461 (0x5B5)
    ERROR_INCORRECT_SIZE: Final[ThirtyTwoBits] = cached_win32error(0x05B6, 'ERROR_INCORRECT_SIZE') # WinError 1462 (0x5B6)
    ERROR_SYMLINK_CLASS_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x05B7, 'ERROR_SYMLINK_CLASS_DISABLED') # WinError 1463 (0x5B7)
    ERROR_SYMLINK_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x05B8, 'ERROR_SYMLINK_NOT_SUPPORTED') # WinError 1464 (0x5B8)
    ERROR_XML_PARSE_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x05B9, 'ERROR_XML_PARSE_ERROR') # WinError 1465 (0x5B9)
    ERROR_XMLDSIG_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x05BA, 'ERROR_XMLDSIG_ERROR') # WinError 1466 (0x5BA)
    ERROR_RESTART_APPLICATION: Final[ThirtyTwoBits] = cached_win32error(0x05BB, 'ERROR_RESTART_APPLICATION') # WinError 1467 (0x5BB)
    ERROR_WRONG_COMPARTMENT: Final[ThirtyTwoBits] = cached_win32error(0x05BC, 'ERROR_WRONG_COMPARTMENT') # WinError 1468 (0x5BC)
    ERROR_AUTHIP_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x05BD, 'ERROR_AUTHIP_FAILURE') # WinError 1469 (0x5BD)
    ERROR_NO_NVRAM_RESOURCES: Final[ThirtyTwoBits] = cached_win32error(0x05BE, 'ERROR_NO_NVRAM_RESOURCES') # WinError 1470 (0x5BE)
    ERROR_NOT_GUI_PROCESS: Final[ThirtyTwoBits] = cached_win32error(0x05BF, 'ERROR_NOT_GUI_PROCESS') # WinError 1471 (0x5BF)
    ERROR_EVENTLOG_FILE_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x05DC, 'ERROR_EVENTLOG_FILE_CORRUPT') # WinError 1500 (0x5DC)
    ERROR_EVENTLOG_CANT_START: Final[ThirtyTwoBits] = cached_win32error(0x05DD, 'ERROR_EVENTLOG_CANT_START') # WinError 1501 (0x5DD)
    ERROR_LOG_FILE_FULL: Final[ThirtyTwoBits] = cached_win32error(0x05DE, 'ERROR_LOG_FILE_FULL') # WinError 1502 (0x5DE)
    ERROR_EVENTLOG_FILE_CHANGED: Final[ThirtyTwoBits] = cached_win32error(0x05DF, 'ERROR_EVENTLOG_FILE_CHANGED') # WinError 1503 (0x5DF)
    ERROR_CONTAINER_ASSIGNED: Final[ThirtyTwoBits] = cached_win32error(0x05E0, 'ERROR_CONTAINER_ASSIGNED') # WinError 1504 (0x5E0)
    ERROR_JOB_NO_CONTAINER: Final[ThirtyTwoBits] = cached_win32error(0x05E1, 'ERROR_JOB_NO_CONTAINER') # WinError 1505 (0x5E1)
    ERROR_INVALID_TASK_NAME: Final[ThirtyTwoBits] = cached_win32error(0x060E, 'ERROR_INVALID_TASK_NAME') # WinError 1550 (0x60E)
    ERROR_INVALID_TASK_INDEX: Final[ThirtyTwoBits] = cached_win32error(0x060F, 'ERROR_INVALID_TASK_INDEX') # WinError 1551 (0x60F)
    ERROR_THREAD_ALREADY_IN_TASK: Final[ThirtyTwoBits] = cached_win32error(0x0610, 'ERROR_THREAD_ALREADY_IN_TASK') # WinError 1552 (0x610)
    ERROR_INSTALL_SERVICE_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x0641, 'ERROR_INSTALL_SERVICE_FAILURE') # WinError 1601 (0x641)
    ERROR_INSTALL_USEREXIT: Final[ThirtyTwoBits] = cached_win32error(0x0642, 'ERROR_INSTALL_USEREXIT') # WinError 1602 (0x642)
    ERROR_INSTALL_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x0643, 'ERROR_INSTALL_FAILURE') # WinError 1603 (0x643)
    ERROR_INSTALL_SUSPEND: Final[ThirtyTwoBits] = cached_win32error(0x0644, 'ERROR_INSTALL_SUSPEND') # WinError 1604 (0x644)
    ERROR_UNKNOWN_PRODUCT: Final[ThirtyTwoBits] = cached_win32error(0x0645, 'ERROR_UNKNOWN_PRODUCT') # WinError 1605 (0x645)
    ERROR_UNKNOWN_FEATURE: Final[ThirtyTwoBits] = cached_win32error(0x0646, 'ERROR_UNKNOWN_FEATURE') # WinError 1606 (0x646)
    ERROR_UNKNOWN_COMPONENT: Final[ThirtyTwoBits] = cached_win32error(0x0647, 'ERROR_UNKNOWN_COMPONENT') # WinError 1607 (0x647)
    ERROR_UNKNOWN_PROPERTY: Final[ThirtyTwoBits] = cached_win32error(0x0648, 'ERROR_UNKNOWN_PROPERTY') # WinError 1608 (0x648)
    ERROR_INVALID_HANDLE_STATE: Final[ThirtyTwoBits] = cached_win32error(0x0649, 'ERROR_INVALID_HANDLE_STATE') # WinError 1609 (0x649)
    ERROR_BAD_CONFIGURATION: Final[ThirtyTwoBits] = cached_win32error(0x064A, 'ERROR_BAD_CONFIGURATION') # WinError 1610 (0x64A)
    ERROR_INDEX_ABSENT: Final[ThirtyTwoBits] = cached_win32error(0x064B, 'ERROR_INDEX_ABSENT') # WinError 1611 (0x64B)
    ERROR_INSTALL_SOURCE_ABSENT: Final[ThirtyTwoBits] = cached_win32error(0x064C, 'ERROR_INSTALL_SOURCE_ABSENT') # WinError 1612 (0x64C)
    ERROR_INSTALL_PACKAGE_VERSION: Final[ThirtyTwoBits] = cached_win32error(0x064D, 'ERROR_INSTALL_PACKAGE_VERSION') # WinError 1613 (0x64D)
    ERROR_PRODUCT_UNINSTALLED: Final[ThirtyTwoBits] = cached_win32error(0x064E, 'ERROR_PRODUCT_UNINSTALLED') # WinError 1614 (0x64E)
    ERROR_BAD_QUERY_SYNTAX: Final[ThirtyTwoBits] = cached_win32error(0x064F, 'ERROR_BAD_QUERY_SYNTAX') # WinError 1615 (0x64F)
    ERROR_INVALID_FIELD: Final[ThirtyTwoBits] = cached_win32error(0x0650, 'ERROR_INVALID_FIELD') # WinError 1616 (0x650)
    ERROR_DEVICE_REMOVED: Final[ThirtyTwoBits] = cached_win32error(0x0651, 'ERROR_DEVICE_REMOVED') # WinError 1617 (0x651)
    ERROR_INSTALL_ALREADY_RUNNING: Final[ThirtyTwoBits] = cached_win32error(0x0652, 'ERROR_INSTALL_ALREADY_RUNNING') # WinError 1618 (0x652)
    ERROR_INSTALL_PACKAGE_OPEN_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x0653, 'ERROR_INSTALL_PACKAGE_OPEN_FAILED') # WinError 1619 (0x653)
    ERROR_INSTALL_PACKAGE_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x0654, 'ERROR_INSTALL_PACKAGE_INVALID') # WinError 1620 (0x654)
    ERROR_INSTALL_UI_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x0655, 'ERROR_INSTALL_UI_FAILURE') # WinError 1621 (0x655)
    ERROR_INSTALL_LOG_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x0656, 'ERROR_INSTALL_LOG_FAILURE') # WinError 1622 (0x656)
    ERROR_INSTALL_LANGUAGE_UNSUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x0657, 'ERROR_INSTALL_LANGUAGE_UNSUPPORTED') # WinError 1623 (0x657)
    ERROR_INSTALL_TRANSFORM_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x0658, 'ERROR_INSTALL_TRANSFORM_FAILURE') # WinError 1624 (0x658)
    ERROR_INSTALL_PACKAGE_REJECTED: Final[ThirtyTwoBits] = cached_win32error(0x0659, 'ERROR_INSTALL_PACKAGE_REJECTED') # WinError 1625 (0x659)
    ERROR_FUNCTION_NOT_CALLED: Final[ThirtyTwoBits] = cached_win32error(0x065A, 'ERROR_FUNCTION_NOT_CALLED') # WinError 1626 (0x65A)
    ERROR_FUNCTION_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x065B, 'ERROR_FUNCTION_FAILED') # WinError 1627 (0x65B)
    ERROR_INVALID_TABLE: Final[ThirtyTwoBits] = cached_win32error(0x065C, 'ERROR_INVALID_TABLE') # WinError 1628 (0x65C)
    ERROR_DATATYPE_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x065D, 'ERROR_DATATYPE_MISMATCH') # WinError 1629 (0x65D)
    ERROR_UNSUPPORTED_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x065E, 'ERROR_UNSUPPORTED_TYPE') # WinError 1630 (0x65E)
    ERROR_CREATE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x065F, 'ERROR_CREATE_FAILED') # WinError 1631 (0x65F)
    ERROR_INSTALL_TEMP_UNWRITABLE: Final[ThirtyTwoBits] = cached_win32error(0x0660, 'ERROR_INSTALL_TEMP_UNWRITABLE') # WinError 1632 (0x660)
    ERROR_INSTALL_PLATFORM_UNSUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x0661, 'ERROR_INSTALL_PLATFORM_UNSUPPORTED') # WinError 1633 (0x661)
    ERROR_INSTALL_NOTUSED: Final[ThirtyTwoBits] = cached_win32error(0x0662, 'ERROR_INSTALL_NOTUSED') # WinError 1634 (0x662)
    ERROR_PATCH_PACKAGE_OPEN_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x0663, 'ERROR_PATCH_PACKAGE_OPEN_FAILED') # WinError 1635 (0x663)
    ERROR_PATCH_PACKAGE_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x0664, 'ERROR_PATCH_PACKAGE_INVALID') # WinError 1636 (0x664)
    ERROR_PATCH_PACKAGE_UNSUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x0665, 'ERROR_PATCH_PACKAGE_UNSUPPORTED') # WinError 1637 (0x665)
    ERROR_PRODUCT_VERSION: Final[ThirtyTwoBits] = cached_win32error(0x0666, 'ERROR_PRODUCT_VERSION') # WinError 1638 (0x666)
    ERROR_INVALID_COMMAND_LINE: Final[ThirtyTwoBits] = cached_win32error(0x0667, 'ERROR_INVALID_COMMAND_LINE') # WinError 1639 (0x667)
    ERROR_INSTALL_REMOTE_DISALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x0668, 'ERROR_INSTALL_REMOTE_DISALLOWED') # WinError 1640 (0x668)
    ERROR_SUCCESS_REBOOT_INITIATED: Final[ThirtyTwoBits] = cached_win32error(0x0669, 'ERROR_SUCCESS_REBOOT_INITIATED') # WinError 1641 (0x669)
    ERROR_PATCH_TARGET_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x066A, 'ERROR_PATCH_TARGET_NOT_FOUND') # WinError 1642 (0x66A)
    ERROR_PATCH_PACKAGE_REJECTED: Final[ThirtyTwoBits] = cached_win32error(0x066B, 'ERROR_PATCH_PACKAGE_REJECTED') # WinError 1643 (0x66B)
    ERROR_INSTALL_TRANSFORM_REJECTED: Final[ThirtyTwoBits] = cached_win32error(0x066C, 'ERROR_INSTALL_TRANSFORM_REJECTED') # WinError 1644 (0x66C)
    ERROR_INSTALL_REMOTE_PROHIBITED: Final[ThirtyTwoBits] = cached_win32error(0x066D, 'ERROR_INSTALL_REMOTE_PROHIBITED') # WinError 1645 (0x66D)
    ERROR_PATCH_REMOVAL_UNSUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x066E, 'ERROR_PATCH_REMOVAL_UNSUPPORTED') # WinError 1646 (0x66E)
    ERROR_UNKNOWN_PATCH: Final[ThirtyTwoBits] = cached_win32error(0x066F, 'ERROR_UNKNOWN_PATCH') # WinError 1647 (0x66F)
    ERROR_PATCH_NO_SEQUENCE: Final[ThirtyTwoBits] = cached_win32error(0x0670, 'ERROR_PATCH_NO_SEQUENCE') # WinError 1648 (0x670)
    ERROR_PATCH_REMOVAL_DISALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x0671, 'ERROR_PATCH_REMOVAL_DISALLOWED') # WinError 1649 (0x671)
    ERROR_INVALID_PATCH_XML: Final[ThirtyTwoBits] = cached_win32error(0x0672, 'ERROR_INVALID_PATCH_XML') # WinError 1650 (0x672)
    ERROR_PATCH_MANAGED_ADVERTISED_PRODUCT: Final[ThirtyTwoBits] = cached_win32error(0x0673, 'ERROR_PATCH_MANAGED_ADVERTISED_PRODUCT') # WinError 1651 (0x673)
    ERROR_INSTALL_SERVICE_SAFEBOOT: Final[ThirtyTwoBits] = cached_win32error(0x0674, 'ERROR_INSTALL_SERVICE_SAFEBOOT') # WinError 1652 (0x674)
    ERROR_FAIL_FAST_EXCEPTION: Final[ThirtyTwoBits] = cached_win32error(0x0675, 'ERROR_FAIL_FAST_EXCEPTION') # WinError 1653 (0x675)
    ERROR_INSTALL_REJECTED: Final[ThirtyTwoBits] = cached_win32error(0x0676, 'ERROR_INSTALL_REJECTED') # WinError 1654 (0x676)
    ERROR_DYNAMIC_CODE_BLOCKED: Final[ThirtyTwoBits] = cached_win32error(0x0677, 'ERROR_DYNAMIC_CODE_BLOCKED') # WinError 1655 (0x677)
    ERROR_NOT_SAME_OBJECT: Final[ThirtyTwoBits] = cached_win32error(0x0678, 'ERROR_NOT_SAME_OBJECT') # WinError 1656 (0x678)
    ERROR_STRICT_CFG_VIOLATION: Final[ThirtyTwoBits] = cached_win32error(0x0679, 'ERROR_STRICT_CFG_VIOLATION') # WinError 1657 (0x679)
    ERROR_SET_CONTEXT_DENIED: Final[ThirtyTwoBits] = cached_win32error(0x067C, 'ERROR_SET_CONTEXT_DENIED') # WinError 1660 (0x67C)
    ERROR_CROSS_PARTITION_VIOLATION: Final[ThirtyTwoBits] = cached_win32error(0x067D, 'ERROR_CROSS_PARTITION_VIOLATION') # WinError 1661 (0x67D)
    ERROR_RETURN_ADDRESS_HIJACK_ATTEMPT: Final[ThirtyTwoBits] = cached_win32error(0x067E, 'ERROR_RETURN_ADDRESS_HIJACK_ATTEMPT') # WinError 1662 (0x67E)
    RPC_S_INVALID_STRING_BINDING: Final[ThirtyTwoBits] = cached_win32error(0x06A4, 'RPC_S_INVALID_STRING_BINDING') # WinError 1700 (0x6A4)
    RPC_S_WRONG_KIND_OF_BINDING: Final[ThirtyTwoBits] = cached_win32error(0x06A5, 'RPC_S_WRONG_KIND_OF_BINDING') # WinError 1701 (0x6A5)
    RPC_S_INVALID_BINDING: Final[ThirtyTwoBits] = cached_win32error(0x06A6, 'RPC_S_INVALID_BINDING') # WinError 1702 (0x6A6)
    RPC_S_PROTSEQ_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x06A7, 'RPC_S_PROTSEQ_NOT_SUPPORTED') # WinError 1703 (0x6A7)
    RPC_S_INVALID_RPC_PROTSEQ: Final[ThirtyTwoBits] = cached_win32error(0x06A8, 'RPC_S_INVALID_RPC_PROTSEQ') # WinError 1704 (0x6A8)
    RPC_S_INVALID_STRING_UUID: Final[ThirtyTwoBits] = cached_win32error(0x06A9, 'RPC_S_INVALID_STRING_UUID') # WinError 1705 (0x6A9)
    RPC_S_INVALID_ENDPOINT_FORMAT: Final[ThirtyTwoBits] = cached_win32error(0x06AA, 'RPC_S_INVALID_ENDPOINT_FORMAT') # WinError 1706 (0x6AA)
    RPC_S_INVALID_NET_ADDR: Final[ThirtyTwoBits] = cached_win32error(0x06AB, 'RPC_S_INVALID_NET_ADDR') # WinError 1707 (0x6AB)
    RPC_S_NO_ENDPOINT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x06AC, 'RPC_S_NO_ENDPOINT_FOUND') # WinError 1708 (0x6AC)
    RPC_S_INVALID_TIMEOUT: Final[ThirtyTwoBits] = cached_win32error(0x06AD, 'RPC_S_INVALID_TIMEOUT') # WinError 1709 (0x6AD)
    RPC_S_OBJECT_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x06AE, 'RPC_S_OBJECT_NOT_FOUND') # WinError 1710 (0x6AE)
    RPC_S_ALREADY_REGISTERED: Final[ThirtyTwoBits] = cached_win32error(0x06AF, 'RPC_S_ALREADY_REGISTERED') # WinError 1711 (0x6AF)
    RPC_S_TYPE_ALREADY_REGISTERED: Final[ThirtyTwoBits] = cached_win32error(0x06B0, 'RPC_S_TYPE_ALREADY_REGISTERED') # WinError 1712 (0x6B0)
    RPC_S_ALREADY_LISTENING: Final[ThirtyTwoBits] = cached_win32error(0x06B1, 'RPC_S_ALREADY_LISTENING') # WinError 1713 (0x6B1)
    RPC_S_NO_PROTSEQS_REGISTERED: Final[ThirtyTwoBits] = cached_win32error(0x06B2, 'RPC_S_NO_PROTSEQS_REGISTERED') # WinError 1714 (0x6B2)
    RPC_S_NOT_LISTENING: Final[ThirtyTwoBits] = cached_win32error(0x06B3, 'RPC_S_NOT_LISTENING') # WinError 1715 (0x6B3)
    RPC_S_UNKNOWN_MGR_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x06B4, 'RPC_S_UNKNOWN_MGR_TYPE') # WinError 1716 (0x6B4)
    RPC_S_UNKNOWN_IF: Final[ThirtyTwoBits] = cached_win32error(0x06B5, 'RPC_S_UNKNOWN_IF') # WinError 1717 (0x6B5)
    RPC_S_NO_BINDINGS: Final[ThirtyTwoBits] = cached_win32error(0x06B6, 'RPC_S_NO_BINDINGS') # WinError 1718 (0x6B6)
    RPC_S_NO_PROTSEQS: Final[ThirtyTwoBits] = cached_win32error(0x06B7, 'RPC_S_NO_PROTSEQS') # WinError 1719 (0x6B7)
    RPC_S_CANT_CREATE_ENDPOINT: Final[ThirtyTwoBits] = cached_win32error(0x06B8, 'RPC_S_CANT_CREATE_ENDPOINT') # WinError 1720 (0x6B8)
    RPC_S_OUT_OF_RESOURCES: Final[ThirtyTwoBits] = cached_win32error(0x06B9, 'RPC_S_OUT_OF_RESOURCES') # WinError 1721 (0x6B9)
    RPC_S_SERVER_UNAVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x06BA, 'RPC_S_SERVER_UNAVAILABLE') # WinError 1722 (0x6BA)
    RPC_S_SERVER_TOO_BUSY: Final[ThirtyTwoBits] = cached_win32error(0x06BB, 'RPC_S_SERVER_TOO_BUSY') # WinError 1723 (0x6BB)
    RPC_S_INVALID_NETWORK_OPTIONS: Final[ThirtyTwoBits] = cached_win32error(0x06BC, 'RPC_S_INVALID_NETWORK_OPTIONS') # WinError 1724 (0x6BC)
    RPC_S_NO_CALL_ACTIVE: Final[ThirtyTwoBits] = cached_win32error(0x06BD, 'RPC_S_NO_CALL_ACTIVE') # WinError 1725 (0x6BD)
    RPC_S_CALL_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x06BE, 'RPC_S_CALL_FAILED') # WinError 1726 (0x6BE)
    RPC_S_CALL_FAILED_DNE: Final[ThirtyTwoBits] = cached_win32error(0x06BF, 'RPC_S_CALL_FAILED_DNE') # WinError 1727 (0x6BF)
    RPC_S_PROTOCOL_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x06C0, 'RPC_S_PROTOCOL_ERROR') # WinError 1728 (0x6C0)
    RPC_S_PROXY_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_win32error(0x06C1, 'RPC_S_PROXY_ACCESS_DENIED') # WinError 1729 (0x6C1)
    RPC_S_UNSUPPORTED_TRANS_SYN: Final[ThirtyTwoBits] = cached_win32error(0x06C2, 'RPC_S_UNSUPPORTED_TRANS_SYN') # WinError 1730 (0x6C2)
    RPC_S_UNSUPPORTED_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x06C4, 'RPC_S_UNSUPPORTED_TYPE') # WinError 1732 (0x6C4)
    RPC_S_INVALID_TAG: Final[ThirtyTwoBits] = cached_win32error(0x06C5, 'RPC_S_INVALID_TAG') # WinError 1733 (0x6C5)
    RPC_S_INVALID_BOUND: Final[ThirtyTwoBits] = cached_win32error(0x06C6, 'RPC_S_INVALID_BOUND') # WinError 1734 (0x6C6)
    RPC_S_NO_ENTRY_NAME: Final[ThirtyTwoBits] = cached_win32error(0x06C7, 'RPC_S_NO_ENTRY_NAME') # WinError 1735 (0x6C7)
    RPC_S_INVALID_NAME_SYNTAX: Final[ThirtyTwoBits] = cached_win32error(0x06C8, 'RPC_S_INVALID_NAME_SYNTAX') # WinError 1736 (0x6C8)
    RPC_S_UNSUPPORTED_NAME_SYNTAX: Final[ThirtyTwoBits] = cached_win32error(0x06C9, 'RPC_S_UNSUPPORTED_NAME_SYNTAX') # WinError 1737 (0x6C9)
    RPC_S_UUID_NO_ADDRESS: Final[ThirtyTwoBits] = cached_win32error(0x06CB, 'RPC_S_UUID_NO_ADDRESS') # WinError 1739 (0x6CB)
    RPC_S_DUPLICATE_ENDPOINT: Final[ThirtyTwoBits] = cached_win32error(0x06CC, 'RPC_S_DUPLICATE_ENDPOINT') # WinError 1740 (0x6CC)
    RPC_S_UNKNOWN_AUTHN_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x06CD, 'RPC_S_UNKNOWN_AUTHN_TYPE') # WinError 1741 (0x6CD)
    RPC_S_MAX_CALLS_TOO_SMALL: Final[ThirtyTwoBits] = cached_win32error(0x06CE, 'RPC_S_MAX_CALLS_TOO_SMALL') # WinError 1742 (0x6CE)
    RPC_S_STRING_TOO_LONG: Final[ThirtyTwoBits] = cached_win32error(0x06CF, 'RPC_S_STRING_TOO_LONG') # WinError 1743 (0x6CF)
    RPC_S_PROTSEQ_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x06D0, 'RPC_S_PROTSEQ_NOT_FOUND') # WinError 1744 (0x6D0)
    RPC_S_PROCNUM_OUT_OF_RANGE: Final[ThirtyTwoBits] = cached_win32error(0x06D1, 'RPC_S_PROCNUM_OUT_OF_RANGE') # WinError 1745 (0x6D1)
    RPC_S_BINDING_HAS_NO_AUTH: Final[ThirtyTwoBits] = cached_win32error(0x06D2, 'RPC_S_BINDING_HAS_NO_AUTH') # WinError 1746 (0x6D2)
    RPC_S_UNKNOWN_AUTHN_SERVICE: Final[ThirtyTwoBits] = cached_win32error(0x06D3, 'RPC_S_UNKNOWN_AUTHN_SERVICE') # WinError 1747 (0x6D3)
    RPC_S_UNKNOWN_AUTHN_LEVEL: Final[ThirtyTwoBits] = cached_win32error(0x06D4, 'RPC_S_UNKNOWN_AUTHN_LEVEL') # WinError 1748 (0x6D4)
    RPC_S_INVALID_AUTH_IDENTITY: Final[ThirtyTwoBits] = cached_win32error(0x06D5, 'RPC_S_INVALID_AUTH_IDENTITY') # WinError 1749 (0x6D5)
    RPC_S_UNKNOWN_AUTHZ_SERVICE: Final[ThirtyTwoBits] = cached_win32error(0x06D6, 'RPC_S_UNKNOWN_AUTHZ_SERVICE') # WinError 1750 (0x6D6)
    EPT_S_INVALID_ENTRY: Final[ThirtyTwoBits] = cached_win32error(0x06D7, 'EPT_S_INVALID_ENTRY') # WinError 1751 (0x6D7)
    EPT_S_CANT_PERFORM_OP: Final[ThirtyTwoBits] = cached_win32error(0x06D8, 'EPT_S_CANT_PERFORM_OP') # WinError 1752 (0x6D8)
    EPT_S_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_win32error(0x06D9, 'EPT_S_NOT_REGISTERED') # WinError 1753 (0x6D9)
    RPC_S_NOTHING_TO_EXPORT: Final[ThirtyTwoBits] = cached_win32error(0x06DA, 'RPC_S_NOTHING_TO_EXPORT') # WinError 1754 (0x6DA)
    RPC_S_INCOMPLETE_NAME: Final[ThirtyTwoBits] = cached_win32error(0x06DB, 'RPC_S_INCOMPLETE_NAME') # WinError 1755 (0x6DB)
    RPC_S_INVALID_VERS_OPTION: Final[ThirtyTwoBits] = cached_win32error(0x06DC, 'RPC_S_INVALID_VERS_OPTION') # WinError 1756 (0x6DC)
    RPC_S_NO_MORE_MEMBERS: Final[ThirtyTwoBits] = cached_win32error(0x06DD, 'RPC_S_NO_MORE_MEMBERS') # WinError 1757 (0x6DD)
    RPC_S_NOT_ALL_OBJS_UNEXPORTED: Final[ThirtyTwoBits] = cached_win32error(0x06DE, 'RPC_S_NOT_ALL_OBJS_UNEXPORTED') # WinError 1758 (0x6DE)
    RPC_S_INTERFACE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x06DF, 'RPC_S_INTERFACE_NOT_FOUND') # WinError 1759 (0x6DF)
    RPC_S_ENTRY_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x06E0, 'RPC_S_ENTRY_ALREADY_EXISTS') # WinError 1760 (0x6E0)
    RPC_S_ENTRY_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x06E1, 'RPC_S_ENTRY_NOT_FOUND') # WinError 1761 (0x6E1)
    RPC_S_NAME_SERVICE_UNAVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x06E2, 'RPC_S_NAME_SERVICE_UNAVAILABLE') # WinError 1762 (0x6E2)
    RPC_S_INVALID_NAF_ID: Final[ThirtyTwoBits] = cached_win32error(0x06E3, 'RPC_S_INVALID_NAF_ID') # WinError 1763 (0x6E3)
    RPC_S_CANNOT_SUPPORT: Final[ThirtyTwoBits] = cached_win32error(0x06E4, 'RPC_S_CANNOT_SUPPORT') # WinError 1764 (0x6E4)
    RPC_S_NO_CONTEXT_AVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x06E5, 'RPC_S_NO_CONTEXT_AVAILABLE') # WinError 1765 (0x6E5)
    RPC_S_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x06E6, 'RPC_S_INTERNAL_ERROR') # WinError 1766 (0x6E6)
    RPC_S_ZERO_DIVIDE: Final[ThirtyTwoBits] = cached_win32error(0x06E7, 'RPC_S_ZERO_DIVIDE') # WinError 1767 (0x6E7)
    RPC_S_ADDRESS_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x06E8, 'RPC_S_ADDRESS_ERROR') # WinError 1768 (0x6E8)
    RPC_S_FP_DIV_ZERO: Final[ThirtyTwoBits] = cached_win32error(0x06E9, 'RPC_S_FP_DIV_ZERO') # WinError 1769 (0x6E9)
    RPC_S_FP_UNDERFLOW: Final[ThirtyTwoBits] = cached_win32error(0x06EA, 'RPC_S_FP_UNDERFLOW') # WinError 1770 (0x6EA)
    RPC_S_FP_OVERFLOW: Final[ThirtyTwoBits] = cached_win32error(0x06EB, 'RPC_S_FP_OVERFLOW') # WinError 1771 (0x6EB)
    RPC_X_NO_MORE_ENTRIES: Final[ThirtyTwoBits] = cached_win32error(0x06EC, 'RPC_X_NO_MORE_ENTRIES') # WinError 1772 (0x6EC)
    RPC_X_SS_CHAR_TRANS_OPEN_FAIL: Final[ThirtyTwoBits] = cached_win32error(0x06ED, 'RPC_X_SS_CHAR_TRANS_OPEN_FAIL') # WinError 1773 (0x6ED)
    RPC_X_SS_CHAR_TRANS_SHORT_FILE: Final[ThirtyTwoBits] = cached_win32error(0x06EE, 'RPC_X_SS_CHAR_TRANS_SHORT_FILE') # WinError 1774 (0x6EE)
    RPC_X_SS_IN_NULL_CONTEXT: Final[ThirtyTwoBits] = cached_win32error(0x06EF, 'RPC_X_SS_IN_NULL_CONTEXT') # WinError 1775 (0x6EF)
    RPC_X_SS_CONTEXT_DAMAGED: Final[ThirtyTwoBits] = cached_win32error(0x06F1, 'RPC_X_SS_CONTEXT_DAMAGED') # WinError 1777 (0x6F1)
    RPC_X_SS_HANDLES_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x06F2, 'RPC_X_SS_HANDLES_MISMATCH') # WinError 1778 (0x6F2)
    RPC_X_SS_CANNOT_GET_CALL_HANDLE: Final[ThirtyTwoBits] = cached_win32error(0x06F3, 'RPC_X_SS_CANNOT_GET_CALL_HANDLE') # WinError 1779 (0x6F3)
    RPC_X_NULL_REF_POINTER: Final[ThirtyTwoBits] = cached_win32error(0x06F4, 'RPC_X_NULL_REF_POINTER') # WinError 1780 (0x6F4)
    RPC_X_ENUM_VALUE_OUT_OF_RANGE: Final[ThirtyTwoBits] = cached_win32error(0x06F5, 'RPC_X_ENUM_VALUE_OUT_OF_RANGE') # WinError 1781 (0x6F5)
    RPC_X_BYTE_COUNT_TOO_SMALL: Final[ThirtyTwoBits] = cached_win32error(0x06F6, 'RPC_X_BYTE_COUNT_TOO_SMALL') # WinError 1782 (0x6F6)
    RPC_X_BAD_STUB_DATA: Final[ThirtyTwoBits] = cached_win32error(0x06F7, 'RPC_X_BAD_STUB_DATA') # WinError 1783 (0x6F7)
    ERROR_INVALID_USER_BUFFER: Final[ThirtyTwoBits] = cached_win32error(0x06F8, 'ERROR_INVALID_USER_BUFFER') # WinError 1784 (0x6F8)
    ERROR_UNRECOGNIZED_MEDIA: Final[ThirtyTwoBits] = cached_win32error(0x06F9, 'ERROR_UNRECOGNIZED_MEDIA') # WinError 1785 (0x6F9)
    ERROR_NO_TRUST_LSA_SECRET: Final[ThirtyTwoBits] = cached_win32error(0x06FA, 'ERROR_NO_TRUST_LSA_SECRET') # WinError 1786 (0x6FA)
    ERROR_NO_TRUST_SAM_ACCOUNT: Final[ThirtyTwoBits] = cached_win32error(0x06FB, 'ERROR_NO_TRUST_SAM_ACCOUNT') # WinError 1787 (0x6FB)
    ERROR_TRUSTED_DOMAIN_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x06FC, 'ERROR_TRUSTED_DOMAIN_FAILURE') # WinError 1788 (0x6FC)
    ERROR_TRUSTED_RELATIONSHIP_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x06FD, 'ERROR_TRUSTED_RELATIONSHIP_FAILURE') # WinError 1789 (0x6FD)
    ERROR_TRUST_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x06FE, 'ERROR_TRUST_FAILURE') # WinError 1790 (0x6FE)
    RPC_S_CALL_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x06FF, 'RPC_S_CALL_IN_PROGRESS') # WinError 1791 (0x6FF)
    ERROR_NETLOGON_NOT_STARTED: Final[ThirtyTwoBits] = cached_win32error(0x0700, 'ERROR_NETLOGON_NOT_STARTED') # WinError 1792 (0x700)
    ERROR_ACCOUNT_EXPIRED: Final[ThirtyTwoBits] = cached_win32error(0x0701, 'ERROR_ACCOUNT_EXPIRED') # WinError 1793 (0x701)
    ERROR_REDIRECTOR_HAS_OPEN_HANDLES: Final[ThirtyTwoBits] = cached_win32error(0x0702, 'ERROR_REDIRECTOR_HAS_OPEN_HANDLES') # WinError 1794 (0x702)
    ERROR_PRINTER_DRIVER_ALREADY_INSTALLED: Final[ThirtyTwoBits] = cached_win32error(0x0703, 'ERROR_PRINTER_DRIVER_ALREADY_INSTALLED') # WinError 1795 (0x703)
    ERROR_UNKNOWN_PORT: Final[ThirtyTwoBits] = cached_win32error(0x0704, 'ERROR_UNKNOWN_PORT') # WinError 1796 (0x704)
    ERROR_UNKNOWN_PRINTER_DRIVER: Final[ThirtyTwoBits] = cached_win32error(0x0705, 'ERROR_UNKNOWN_PRINTER_DRIVER') # WinError 1797 (0x705)
    ERROR_UNKNOWN_PRINTPROCESSOR: Final[ThirtyTwoBits] = cached_win32error(0x0706, 'ERROR_UNKNOWN_PRINTPROCESSOR') # WinError 1798 (0x706)
    ERROR_INVALID_SEPARATOR_FILE: Final[ThirtyTwoBits] = cached_win32error(0x0707, 'ERROR_INVALID_SEPARATOR_FILE') # WinError 1799 (0x707)
    ERROR_INVALID_PRIORITY: Final[ThirtyTwoBits] = cached_win32error(0x0708, 'ERROR_INVALID_PRIORITY') # WinError 1800 (0x708)
    ERROR_INVALID_PRINTER_NAME: Final[ThirtyTwoBits] = cached_win32error(0x0709, 'ERROR_INVALID_PRINTER_NAME') # WinError 1801 (0x709)
    ERROR_PRINTER_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x070A, 'ERROR_PRINTER_ALREADY_EXISTS') # WinError 1802 (0x70A)
    ERROR_INVALID_PRINTER_COMMAND: Final[ThirtyTwoBits] = cached_win32error(0x070B, 'ERROR_INVALID_PRINTER_COMMAND') # WinError 1803 (0x70B)
    ERROR_INVALID_DATATYPE: Final[ThirtyTwoBits] = cached_win32error(0x070C, 'ERROR_INVALID_DATATYPE') # WinError 1804 (0x70C)
    ERROR_INVALID_ENVIRONMENT: Final[ThirtyTwoBits] = cached_win32error(0x070D, 'ERROR_INVALID_ENVIRONMENT') # WinError 1805 (0x70D)
    RPC_S_NO_MORE_BINDINGS: Final[ThirtyTwoBits] = cached_win32error(0x070E, 'RPC_S_NO_MORE_BINDINGS') # WinError 1806 (0x70E)
    ERROR_NOLOGON_INTERDOMAIN_TRUST_ACCOUNT: Final[ThirtyTwoBits] = cached_win32error(0x070F, 'ERROR_NOLOGON_INTERDOMAIN_TRUST_ACCOUNT') # WinError 1807 (0x70F)
    ERROR_NOLOGON_WORKSTATION_TRUST_ACCOUNT: Final[ThirtyTwoBits] = cached_win32error(0x0710, 'ERROR_NOLOGON_WORKSTATION_TRUST_ACCOUNT') # WinError 1808 (0x710)
    ERROR_NOLOGON_SERVER_TRUST_ACCOUNT: Final[ThirtyTwoBits] = cached_win32error(0x0711, 'ERROR_NOLOGON_SERVER_TRUST_ACCOUNT') # WinError 1809 (0x711)
    ERROR_DOMAIN_TRUST_INCONSISTENT: Final[ThirtyTwoBits] = cached_win32error(0x0712, 'ERROR_DOMAIN_TRUST_INCONSISTENT') # WinError 1810 (0x712)
    ERROR_SERVER_HAS_OPEN_HANDLES: Final[ThirtyTwoBits] = cached_win32error(0x0713, 'ERROR_SERVER_HAS_OPEN_HANDLES') # WinError 1811 (0x713)
    ERROR_RESOURCE_DATA_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x0714, 'ERROR_RESOURCE_DATA_NOT_FOUND') # WinError 1812 (0x714)
    ERROR_RESOURCE_TYPE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x0715, 'ERROR_RESOURCE_TYPE_NOT_FOUND') # WinError 1813 (0x715)
    ERROR_RESOURCE_NAME_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x0716, 'ERROR_RESOURCE_NAME_NOT_FOUND') # WinError 1814 (0x716)
    ERROR_RESOURCE_LANG_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x0717, 'ERROR_RESOURCE_LANG_NOT_FOUND') # WinError 1815 (0x717)
    ERROR_NOT_ENOUGH_QUOTA: Final[ThirtyTwoBits] = cached_win32error(0x0718, 'ERROR_NOT_ENOUGH_QUOTA') # WinError 1816 (0x718)
    RPC_S_NO_INTERFACES: Final[ThirtyTwoBits] = cached_win32error(0x0719, 'RPC_S_NO_INTERFACES') # WinError 1817 (0x719)
    RPC_S_CALL_CANCELLED: Final[ThirtyTwoBits] = cached_win32error(0x071A, 'RPC_S_CALL_CANCELLED') # WinError 1818 (0x71A)
    RPC_S_BINDING_INCOMPLETE: Final[ThirtyTwoBits] = cached_win32error(0x071B, 'RPC_S_BINDING_INCOMPLETE') # WinError 1819 (0x71B)
    RPC_S_COMM_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x071C, 'RPC_S_COMM_FAILURE') # WinError 1820 (0x71C)
    RPC_S_UNSUPPORTED_AUTHN_LEVEL: Final[ThirtyTwoBits] = cached_win32error(0x071D, 'RPC_S_UNSUPPORTED_AUTHN_LEVEL') # WinError 1821 (0x71D)
    RPC_S_NO_PRINC_NAME: Final[ThirtyTwoBits] = cached_win32error(0x071E, 'RPC_S_NO_PRINC_NAME') # WinError 1822 (0x71E)
    RPC_S_NOT_RPC_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x071F, 'RPC_S_NOT_RPC_ERROR') # WinError 1823 (0x71F)
    RPC_S_UUID_LOCAL_ONLY: Final[ThirtyTwoBits] = cached_win32error(0x0720, 'RPC_S_UUID_LOCAL_ONLY') # WinError 1824 (0x720)
    RPC_S_SEC_PKG_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x0721, 'RPC_S_SEC_PKG_ERROR') # WinError 1825 (0x721)
    RPC_S_NOT_CANCELLED: Final[ThirtyTwoBits] = cached_win32error(0x0722, 'RPC_S_NOT_CANCELLED') # WinError 1826 (0x722)
    RPC_X_INVALID_ES_ACTION: Final[ThirtyTwoBits] = cached_win32error(0x0723, 'RPC_X_INVALID_ES_ACTION') # WinError 1827 (0x723)
    RPC_X_WRONG_ES_VERSION: Final[ThirtyTwoBits] = cached_win32error(0x0724, 'RPC_X_WRONG_ES_VERSION') # WinError 1828 (0x724)
    RPC_X_WRONG_STUB_VERSION: Final[ThirtyTwoBits] = cached_win32error(0x0725, 'RPC_X_WRONG_STUB_VERSION') # WinError 1829 (0x725)
    RPC_X_INVALID_PIPE_OBJECT: Final[ThirtyTwoBits] = cached_win32error(0x0726, 'RPC_X_INVALID_PIPE_OBJECT') # WinError 1830 (0x726)
    RPC_X_WRONG_PIPE_ORDER: Final[ThirtyTwoBits] = cached_win32error(0x0727, 'RPC_X_WRONG_PIPE_ORDER') # WinError 1831 (0x727)
    RPC_X_WRONG_PIPE_VERSION: Final[ThirtyTwoBits] = cached_win32error(0x0728, 'RPC_X_WRONG_PIPE_VERSION') # WinError 1832 (0x728)
    RPC_S_COOKIE_AUTH_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x0729, 'RPC_S_COOKIE_AUTH_FAILED') # WinError 1833 (0x729)
    RPC_S_DO_NOT_DISTURB: Final[ThirtyTwoBits] = cached_win32error(0x072A, 'RPC_S_DO_NOT_DISTURB') # WinError 1834 (0x72A)
    RPC_S_SYSTEM_HANDLE_COUNT_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x072B, 'RPC_S_SYSTEM_HANDLE_COUNT_EXCEEDED') # WinError 1835 (0x72B)
    RPC_S_SYSTEM_HANDLE_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x072C, 'RPC_S_SYSTEM_HANDLE_TYPE_MISMATCH') # WinError 1836 (0x72C)
    RPC_S_GROUP_MEMBER_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x076A, 'RPC_S_GROUP_MEMBER_NOT_FOUND') # WinError 1898 (0x76A)
    EPT_S_CANT_CREATE: Final[ThirtyTwoBits] = cached_win32error(0x076B, 'EPT_S_CANT_CREATE') # WinError 1899 (0x76B)
    RPC_S_INVALID_OBJECT: Final[ThirtyTwoBits] = cached_win32error(0x076C, 'RPC_S_INVALID_OBJECT') # WinError 1900 (0x76C)
    ERROR_INVALID_TIME: Final[ThirtyTwoBits] = cached_win32error(0x076D, 'ERROR_INVALID_TIME') # WinError 1901 (0x76D)
    ERROR_INVALID_FORM_NAME: Final[ThirtyTwoBits] = cached_win32error(0x076E, 'ERROR_INVALID_FORM_NAME') # WinError 1902 (0x76E)
    ERROR_INVALID_FORM_SIZE: Final[ThirtyTwoBits] = cached_win32error(0x076F, 'ERROR_INVALID_FORM_SIZE') # WinError 1903 (0x76F)
    ERROR_ALREADY_WAITING: Final[ThirtyTwoBits] = cached_win32error(0x0770, 'ERROR_ALREADY_WAITING') # WinError 1904 (0x770)
    ERROR_PRINTER_DELETED: Final[ThirtyTwoBits] = cached_win32error(0x0771, 'ERROR_PRINTER_DELETED') # WinError 1905 (0x771)
    ERROR_INVALID_PRINTER_STATE: Final[ThirtyTwoBits] = cached_win32error(0x0772, 'ERROR_INVALID_PRINTER_STATE') # WinError 1906 (0x772)
    ERROR_PASSWORD_MUST_CHANGE: Final[ThirtyTwoBits] = cached_win32error(0x0773, 'ERROR_PASSWORD_MUST_CHANGE') # WinError 1907 (0x773)
    ERROR_DOMAIN_CONTROLLER_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x0774, 'ERROR_DOMAIN_CONTROLLER_NOT_FOUND') # WinError 1908 (0x774)
    ERROR_ACCOUNT_LOCKED_OUT: Final[ThirtyTwoBits] = cached_win32error(0x0775, 'ERROR_ACCOUNT_LOCKED_OUT') # WinError 1909 (0x775)
    OR_INVALID_OXID: Final[ThirtyTwoBits] = cached_win32error(0x0776, 'OR_INVALID_OXID') # WinError 1910 (0x776)
    OR_INVALID_OID: Final[ThirtyTwoBits] = cached_win32error(0x0777, 'OR_INVALID_OID') # WinError 1911 (0x777)
    OR_INVALID_SET: Final[ThirtyTwoBits] = cached_win32error(0x0778, 'OR_INVALID_SET') # WinError 1912 (0x778)
    RPC_S_SEND_INCOMPLETE: Final[ThirtyTwoBits] = cached_win32error(0x0779, 'RPC_S_SEND_INCOMPLETE') # WinError 1913 (0x779)
    RPC_S_INVALID_ASYNC_HANDLE: Final[ThirtyTwoBits] = cached_win32error(0x077A, 'RPC_S_INVALID_ASYNC_HANDLE') # WinError 1914 (0x77A)
    RPC_S_INVALID_ASYNC_CALL: Final[ThirtyTwoBits] = cached_win32error(0x077B, 'RPC_S_INVALID_ASYNC_CALL') # WinError 1915 (0x77B)
    RPC_X_PIPE_CLOSED: Final[ThirtyTwoBits] = cached_win32error(0x077C, 'RPC_X_PIPE_CLOSED') # WinError 1916 (0x77C)
    RPC_X_PIPE_DISCIPLINE_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x077D, 'RPC_X_PIPE_DISCIPLINE_ERROR') # WinError 1917 (0x77D)
    RPC_X_PIPE_EMPTY: Final[ThirtyTwoBits] = cached_win32error(0x077E, 'RPC_X_PIPE_EMPTY') # WinError 1918 (0x77E)
    ERROR_NO_SITENAME: Final[ThirtyTwoBits] = cached_win32error(0x077F, 'ERROR_NO_SITENAME') # WinError 1919 (0x77F)
    ERROR_CANT_ACCESS_FILE: Final[ThirtyTwoBits] = cached_win32error(0x0780, 'ERROR_CANT_ACCESS_FILE') # WinError 1920 (0x780)
    ERROR_CANT_RESOLVE_FILENAME: Final[ThirtyTwoBits] = cached_win32error(0x0781, 'ERROR_CANT_RESOLVE_FILENAME') # WinError 1921 (0x781)
    RPC_S_ENTRY_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x0782, 'RPC_S_ENTRY_TYPE_MISMATCH') # WinError 1922 (0x782)
    RPC_S_NOT_ALL_OBJS_EXPORTED: Final[ThirtyTwoBits] = cached_win32error(0x0783, 'RPC_S_NOT_ALL_OBJS_EXPORTED') # WinError 1923 (0x783)
    RPC_S_INTERFACE_NOT_EXPORTED: Final[ThirtyTwoBits] = cached_win32error(0x0784, 'RPC_S_INTERFACE_NOT_EXPORTED') # WinError 1924 (0x784)
    RPC_S_PROFILE_NOT_ADDED: Final[ThirtyTwoBits] = cached_win32error(0x0785, 'RPC_S_PROFILE_NOT_ADDED') # WinError 1925 (0x785)
    RPC_S_PRF_ELT_NOT_ADDED: Final[ThirtyTwoBits] = cached_win32error(0x0786, 'RPC_S_PRF_ELT_NOT_ADDED') # WinError 1926 (0x786)
    RPC_S_PRF_ELT_NOT_REMOVED: Final[ThirtyTwoBits] = cached_win32error(0x0787, 'RPC_S_PRF_ELT_NOT_REMOVED') # WinError 1927 (0x787)
    RPC_S_GRP_ELT_NOT_ADDED: Final[ThirtyTwoBits] = cached_win32error(0x0788, 'RPC_S_GRP_ELT_NOT_ADDED') # WinError 1928 (0x788)
    RPC_S_GRP_ELT_NOT_REMOVED: Final[ThirtyTwoBits] = cached_win32error(0x0789, 'RPC_S_GRP_ELT_NOT_REMOVED') # WinError 1929 (0x789)
    ERROR_KM_DRIVER_BLOCKED: Final[ThirtyTwoBits] = cached_win32error(0x078A, 'ERROR_KM_DRIVER_BLOCKED') # WinError 1930 (0x78A)
    ERROR_CONTEXT_EXPIRED: Final[ThirtyTwoBits] = cached_win32error(0x078B, 'ERROR_CONTEXT_EXPIRED') # WinError 1931 (0x78B)
    ERROR_PER_USER_TRUST_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x078C, 'ERROR_PER_USER_TRUST_QUOTA_EXCEEDED') # WinError 1932 (0x78C)
    ERROR_ALL_USER_TRUST_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x078D, 'ERROR_ALL_USER_TRUST_QUOTA_EXCEEDED') # WinError 1933 (0x78D)
    ERROR_USER_DELETE_TRUST_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x078E, 'ERROR_USER_DELETE_TRUST_QUOTA_EXCEEDED') # WinError 1934 (0x78E)
    ERROR_AUTHENTICATION_FIREWALL_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x078F, 'ERROR_AUTHENTICATION_FIREWALL_FAILED') # WinError 1935 (0x78F)
    ERROR_REMOTE_PRINT_CONNECTIONS_BLOCKED: Final[ThirtyTwoBits] = cached_win32error(0x0790, 'ERROR_REMOTE_PRINT_CONNECTIONS_BLOCKED') # WinError 1936 (0x790)
    ERROR_NTLM_BLOCKED: Final[ThirtyTwoBits] = cached_win32error(0x0791, 'ERROR_NTLM_BLOCKED') # WinError 1937 (0x791)
    ERROR_PASSWORD_CHANGE_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x0792, 'ERROR_PASSWORD_CHANGE_REQUIRED') # WinError 1938 (0x792)
    ERROR_LOST_MODE_LOGON_RESTRICTION: Final[ThirtyTwoBits] = cached_win32error(0x0793, 'ERROR_LOST_MODE_LOGON_RESTRICTION') # WinError 1939 (0x793)
    ERROR_INVALID_PIXEL_FORMAT: Final[ThirtyTwoBits] = cached_win32error(0x07D0, 'ERROR_INVALID_PIXEL_FORMAT') # WinError 2000 (0x7D0)
    ERROR_BAD_DRIVER: Final[ThirtyTwoBits] = cached_win32error(0x07D1, 'ERROR_BAD_DRIVER') # WinError 2001 (0x7D1)
    ERROR_INVALID_WINDOW_STYLE: Final[ThirtyTwoBits] = cached_win32error(0x07D2, 'ERROR_INVALID_WINDOW_STYLE') # WinError 2002 (0x7D2)
    ERROR_METAFILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x07D3, 'ERROR_METAFILE_NOT_SUPPORTED') # WinError 2003 (0x7D3)
    ERROR_TRANSFORM_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x07D4, 'ERROR_TRANSFORM_NOT_SUPPORTED') # WinError 2004 (0x7D4)
    ERROR_CLIPPING_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x07D5, 'ERROR_CLIPPING_NOT_SUPPORTED') # WinError 2005 (0x7D5)
    ERROR_INVALID_CMM: Final[ThirtyTwoBits] = cached_win32error(0x07DA, 'ERROR_INVALID_CMM') # WinError 2010 (0x7DA)
    ERROR_INVALID_PROFILE: Final[ThirtyTwoBits] = cached_win32error(0x07DB, 'ERROR_INVALID_PROFILE') # WinError 2011 (0x7DB)
    ERROR_TAG_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x07DC, 'ERROR_TAG_NOT_FOUND') # WinError 2012 (0x7DC)
    ERROR_TAG_NOT_PRESENT: Final[ThirtyTwoBits] = cached_win32error(0x07DD, 'ERROR_TAG_NOT_PRESENT') # WinError 2013 (0x7DD)
    ERROR_DUPLICATE_TAG: Final[ThirtyTwoBits] = cached_win32error(0x07DE, 'ERROR_DUPLICATE_TAG') # WinError 2014 (0x7DE)
    ERROR_PROFILE_NOT_ASSOCIATED_WITH_DEVICE: Final[ThirtyTwoBits] = cached_win32error(0x07DF, 'ERROR_PROFILE_NOT_ASSOCIATED_WITH_DEVICE') # WinError 2015 (0x7DF)
    ERROR_PROFILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x07E0, 'ERROR_PROFILE_NOT_FOUND') # WinError 2016 (0x7E0)
    ERROR_INVALID_COLORSPACE: Final[ThirtyTwoBits] = cached_win32error(0x07E1, 'ERROR_INVALID_COLORSPACE') # WinError 2017 (0x7E1)
    ERROR_ICM_NOT_ENABLED: Final[ThirtyTwoBits] = cached_win32error(0x07E2, 'ERROR_ICM_NOT_ENABLED') # WinError 2018 (0x7E2)
    ERROR_DELETING_ICM_XFORM: Final[ThirtyTwoBits] = cached_win32error(0x07E3, 'ERROR_DELETING_ICM_XFORM') # WinError 2019 (0x7E3)
    ERROR_INVALID_TRANSFORM: Final[ThirtyTwoBits] = cached_win32error(0x07E4, 'ERROR_INVALID_TRANSFORM') # WinError 2020 (0x7E4)
    ERROR_COLORSPACE_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x07E5, 'ERROR_COLORSPACE_MISMATCH') # WinError 2021 (0x7E5)
    ERROR_INVALID_COLORINDEX: Final[ThirtyTwoBits] = cached_win32error(0x07E6, 'ERROR_INVALID_COLORINDEX') # WinError 2022 (0x7E6)
    ERROR_PROFILE_DOES_NOT_MATCH_DEVICE: Final[ThirtyTwoBits] = cached_win32error(0x07E7, 'ERROR_PROFILE_DOES_NOT_MATCH_DEVICE') # WinError 2023 (0x7E7)
    ERROR_CONNECTED_OTHER_PASSWORD: Final[ThirtyTwoBits] = cached_win32error(0x083C, 'ERROR_CONNECTED_OTHER_PASSWORD') # WinError 2108 (0x83C)
    ERROR_CONNECTED_OTHER_PASSWORD_DEFAULT: Final[ThirtyTwoBits] = cached_win32error(0x083D, 'ERROR_CONNECTED_OTHER_PASSWORD_DEFAULT') # WinError 2109 (0x83D)
    ERROR_BAD_USERNAME: Final[ThirtyTwoBits] = cached_win32error(0x089A, 'ERROR_BAD_USERNAME') # WinError 2202 (0x89A)
    ERROR_NOT_CONNECTED: Final[ThirtyTwoBits] = cached_win32error(0x08CA, 'ERROR_NOT_CONNECTED') # WinError 2250 (0x8CA)
    ERROR_OPEN_FILES: Final[ThirtyTwoBits] = cached_win32error(0x0961, 'ERROR_OPEN_FILES') # WinError 2401 (0x961)
    ERROR_ACTIVE_CONNECTIONS: Final[ThirtyTwoBits] = cached_win32error(0x0962, 'ERROR_ACTIVE_CONNECTIONS') # WinError 2402 (0x962)
    ERROR_DEVICE_IN_USE: Final[ThirtyTwoBits] = cached_win32error(0x0964, 'ERROR_DEVICE_IN_USE') # WinError 2404 (0x964)
    ERROR_UNKNOWN_PRINT_MONITOR: Final[ThirtyTwoBits] = cached_win32error(0x0BB8, 'ERROR_UNKNOWN_PRINT_MONITOR') # WinError 3000 (0xBB8)
    ERROR_PRINTER_DRIVER_IN_USE: Final[ThirtyTwoBits] = cached_win32error(0x0BB9, 'ERROR_PRINTER_DRIVER_IN_USE') # WinError 3001 (0xBB9)
    ERROR_SPOOL_FILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x0BBA, 'ERROR_SPOOL_FILE_NOT_FOUND') # WinError 3002 (0xBBA)
    ERROR_SPL_NO_STARTDOC: Final[ThirtyTwoBits] = cached_win32error(0x0BBB, 'ERROR_SPL_NO_STARTDOC') # WinError 3003 (0xBBB)
    ERROR_SPL_NO_ADDJOB: Final[ThirtyTwoBits] = cached_win32error(0x0BBC, 'ERROR_SPL_NO_ADDJOB') # WinError 3004 (0xBBC)
    ERROR_PRINT_PROCESSOR_ALREADY_INSTALLED: Final[ThirtyTwoBits] = cached_win32error(0x0BBD, 'ERROR_PRINT_PROCESSOR_ALREADY_INSTALLED') # WinError 3005 (0xBBD)
    ERROR_PRINT_MONITOR_ALREADY_INSTALLED: Final[ThirtyTwoBits] = cached_win32error(0x0BBE, 'ERROR_PRINT_MONITOR_ALREADY_INSTALLED') # WinError 3006 (0xBBE)
    ERROR_INVALID_PRINT_MONITOR: Final[ThirtyTwoBits] = cached_win32error(0x0BBF, 'ERROR_INVALID_PRINT_MONITOR') # WinError 3007 (0xBBF)
    ERROR_PRINT_MONITOR_IN_USE: Final[ThirtyTwoBits] = cached_win32error(0x0BC0, 'ERROR_PRINT_MONITOR_IN_USE') # WinError 3008 (0xBC0)
    ERROR_PRINTER_HAS_JOBS_QUEUED: Final[ThirtyTwoBits] = cached_win32error(0x0BC1, 'ERROR_PRINTER_HAS_JOBS_QUEUED') # WinError 3009 (0xBC1)
    ERROR_SUCCESS_REBOOT_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x0BC2, 'ERROR_SUCCESS_REBOOT_REQUIRED') # WinError 3010 (0xBC2)
    ERROR_SUCCESS_RESTART_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x0BC3, 'ERROR_SUCCESS_RESTART_REQUIRED') # WinError 3011 (0xBC3)
    ERROR_PRINTER_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x0BC4, 'ERROR_PRINTER_NOT_FOUND') # WinError 3012 (0xBC4)
    ERROR_PRINTER_DRIVER_WARNED: Final[ThirtyTwoBits] = cached_win32error(0x0BC5, 'ERROR_PRINTER_DRIVER_WARNED') # WinError 3013 (0xBC5)
    ERROR_PRINTER_DRIVER_BLOCKED: Final[ThirtyTwoBits] = cached_win32error(0x0BC6, 'ERROR_PRINTER_DRIVER_BLOCKED') # WinError 3014 (0xBC6)
    ERROR_PRINTER_DRIVER_PACKAGE_IN_USE: Final[ThirtyTwoBits] = cached_win32error(0x0BC7, 'ERROR_PRINTER_DRIVER_PACKAGE_IN_USE') # WinError 3015 (0xBC7)
    ERROR_CORE_DRIVER_PACKAGE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x0BC8, 'ERROR_CORE_DRIVER_PACKAGE_NOT_FOUND') # WinError 3016 (0xBC8)
    ERROR_FAIL_REBOOT_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x0BC9, 'ERROR_FAIL_REBOOT_REQUIRED') # WinError 3017 (0xBC9)
    ERROR_FAIL_REBOOT_INITIATED: Final[ThirtyTwoBits] = cached_win32error(0x0BCA, 'ERROR_FAIL_REBOOT_INITIATED') # WinError 3018 (0xBCA)
    ERROR_PRINTER_DRIVER_DOWNLOAD_NEEDED: Final[ThirtyTwoBits] = cached_win32error(0x0BCB, 'ERROR_PRINTER_DRIVER_DOWNLOAD_NEEDED') # WinError 3019 (0xBCB)
    ERROR_PRINT_JOB_RESTART_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x0BCC, 'ERROR_PRINT_JOB_RESTART_REQUIRED') # WinError 3020 (0xBCC)
    ERROR_INVALID_PRINTER_DRIVER_MANIFEST: Final[ThirtyTwoBits] = cached_win32error(0x0BCD, 'ERROR_INVALID_PRINTER_DRIVER_MANIFEST') # WinError 3021 (0xBCD)
    ERROR_PRINTER_NOT_SHAREABLE: Final[ThirtyTwoBits] = cached_win32error(0x0BCE, 'ERROR_PRINTER_NOT_SHAREABLE') # WinError 3022 (0xBCE)
    ERROR_SERVER_SERVICE_CALL_REQUIRES_SMB1: Final[ThirtyTwoBits] = cached_win32error(0x0BCF, 'ERROR_SERVER_SERVICE_CALL_REQUIRES_SMB1') # WinError 3023 (0xBCF)
    ERROR_NETWORK_AUTHENTICATION_PROMPT_CANCELED: Final[ThirtyTwoBits] = cached_win32error(0x0BD0, 'ERROR_NETWORK_AUTHENTICATION_PROMPT_CANCELED') # WinError 3024 (0xBD0)
    ERROR_REQUEST_PAUSED: Final[ThirtyTwoBits] = cached_win32error(0x0BEA, 'ERROR_REQUEST_PAUSED') # WinError 3050 (0xBEA)
    ERROR_APPEXEC_CONDITION_NOT_SATISFIED: Final[ThirtyTwoBits] = cached_win32error(0x0BF4, 'ERROR_APPEXEC_CONDITION_NOT_SATISFIED') # WinError 3060 (0xBF4)
    ERROR_APPEXEC_HANDLE_INVALIDATED: Final[ThirtyTwoBits] = cached_win32error(0x0BF5, 'ERROR_APPEXEC_HANDLE_INVALIDATED') # WinError 3061 (0xBF5)
    ERROR_APPEXEC_INVALID_HOST_GENERATION: Final[ThirtyTwoBits] = cached_win32error(0x0BF6, 'ERROR_APPEXEC_INVALID_HOST_GENERATION') # WinError 3062 (0xBF6)
    ERROR_APPEXEC_UNEXPECTED_PROCESS_REGISTRATION: Final[ThirtyTwoBits] = cached_win32error(0x0BF7, 'ERROR_APPEXEC_UNEXPECTED_PROCESS_REGISTRATION') # WinError 3063 (0xBF7)
    ERROR_APPEXEC_INVALID_HOST_STATE: Final[ThirtyTwoBits] = cached_win32error(0x0BF8, 'ERROR_APPEXEC_INVALID_HOST_STATE') # WinError 3064 (0xBF8)
    ERROR_APPEXEC_NO_DONOR: Final[ThirtyTwoBits] = cached_win32error(0x0BF9, 'ERROR_APPEXEC_NO_DONOR') # WinError 3065 (0xBF9)
    ERROR_APPEXEC_HOST_ID_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x0BFA, 'ERROR_APPEXEC_HOST_ID_MISMATCH') # WinError 3066 (0xBFA)
    ERROR_APPEXEC_UNKNOWN_USER: Final[ThirtyTwoBits] = cached_win32error(0x0BFB, 'ERROR_APPEXEC_UNKNOWN_USER') # WinError 3067 (0xBFB)
    ERROR_APPEXEC_APP_COMPAT_BLOCK: Final[ThirtyTwoBits] = cached_win32error(0x0BFC, 'ERROR_APPEXEC_APP_COMPAT_BLOCK') # WinError 3068 (0xBFC)
    ERROR_APPEXEC_CALLER_WAIT_TIMEOUT: Final[ThirtyTwoBits] = cached_win32error(0x0BFD, 'ERROR_APPEXEC_CALLER_WAIT_TIMEOUT') # WinError 3069 (0xBFD)
    ERROR_APPEXEC_CALLER_WAIT_TIMEOUT_TERMINATION: Final[ThirtyTwoBits] = cached_win32error(0x0BFE, 'ERROR_APPEXEC_CALLER_WAIT_TIMEOUT_TERMINATION') # WinError 3070 (0xBFE)
    ERROR_APPEXEC_CALLER_WAIT_TIMEOUT_LICENSING: Final[ThirtyTwoBits] = cached_win32error(0x0BFF, 'ERROR_APPEXEC_CALLER_WAIT_TIMEOUT_LICENSING') # WinError 3071 (0xBFF)
    ERROR_APPEXEC_CALLER_WAIT_TIMEOUT_RESOURCES: Final[ThirtyTwoBits] = cached_win32error(0x0C00, 'ERROR_APPEXEC_CALLER_WAIT_TIMEOUT_RESOURCES') # WinError 3072 (0xC00)
    ERROR_VRF_VOLATILE_CFG_AND_IO_ENABLED: Final[ThirtyTwoBits] = cached_win32error(0x0C08, 'ERROR_VRF_VOLATILE_CFG_AND_IO_ENABLED') # WinError 3080 (0xC08)
    ERROR_VRF_VOLATILE_NOT_STOPPABLE: Final[ThirtyTwoBits] = cached_win32error(0x0C09, 'ERROR_VRF_VOLATILE_NOT_STOPPABLE') # WinError 3081 (0xC09)
    ERROR_VRF_VOLATILE_SAFE_MODE: Final[ThirtyTwoBits] = cached_win32error(0x0C0A, 'ERROR_VRF_VOLATILE_SAFE_MODE') # WinError 3082 (0xC0A)
    ERROR_VRF_VOLATILE_NOT_RUNNABLE_SYSTEM: Final[ThirtyTwoBits] = cached_win32error(0x0C0B, 'ERROR_VRF_VOLATILE_NOT_RUNNABLE_SYSTEM') # WinError 3083 (0xC0B)
    ERROR_VRF_VOLATILE_NOT_SUPPORTED_RULECLASS: Final[ThirtyTwoBits] = cached_win32error(0x0C0C, 'ERROR_VRF_VOLATILE_NOT_SUPPORTED_RULECLASS') # WinError 3084 (0xC0C)
    ERROR_VRF_VOLATILE_PROTECTED_DRIVER: Final[ThirtyTwoBits] = cached_win32error(0x0C0D, 'ERROR_VRF_VOLATILE_PROTECTED_DRIVER') # WinError 3085 (0xC0D)
    ERROR_VRF_VOLATILE_NMI_REGISTERED: Final[ThirtyTwoBits] = cached_win32error(0x0C0E, 'ERROR_VRF_VOLATILE_NMI_REGISTERED') # WinError 3086 (0xC0E)
    ERROR_VRF_VOLATILE_SETTINGS_CONFLICT: Final[ThirtyTwoBits] = cached_win32error(0x0C0F, 'ERROR_VRF_VOLATILE_SETTINGS_CONFLICT') # WinError 3087 (0xC0F)
    ERROR_DIF_IOCALLBACK_NOT_REPLACED: Final[ThirtyTwoBits] = cached_win32error(0x0C76, 'ERROR_DIF_IOCALLBACK_NOT_REPLACED') # WinError 3190 (0xC76)
    ERROR_DIF_LIVEDUMP_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x0C77, 'ERROR_DIF_LIVEDUMP_LIMIT_EXCEEDED') # WinError 3191 (0xC77)
    ERROR_DIF_VOLATILE_SECTION_NOT_LOCKED: Final[ThirtyTwoBits] = cached_win32error(0x0C78, 'ERROR_DIF_VOLATILE_SECTION_NOT_LOCKED') # WinError 3192 (0xC78)
    ERROR_DIF_VOLATILE_DRIVER_HOTPATCHED: Final[ThirtyTwoBits] = cached_win32error(0x0C79, 'ERROR_DIF_VOLATILE_DRIVER_HOTPATCHED') # WinError 3193 (0xC79)
    ERROR_DIF_VOLATILE_INVALID_INFO: Final[ThirtyTwoBits] = cached_win32error(0x0C7A, 'ERROR_DIF_VOLATILE_INVALID_INFO') # WinError 3194 (0xC7A)
    ERROR_DIF_VOLATILE_DRIVER_IS_NOT_RUNNING: Final[ThirtyTwoBits] = cached_win32error(0x0C7B, 'ERROR_DIF_VOLATILE_DRIVER_IS_NOT_RUNNING') # WinError 3195 (0xC7B)
    ERROR_DIF_VOLATILE_PLUGIN_IS_NOT_RUNNING: Final[ThirtyTwoBits] = cached_win32error(0x0C7C, 'ERROR_DIF_VOLATILE_PLUGIN_IS_NOT_RUNNING') # WinError 3196 (0xC7C)
    ERROR_DIF_VOLATILE_PLUGIN_CHANGE_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x0C7D, 'ERROR_DIF_VOLATILE_PLUGIN_CHANGE_NOT_ALLOWED') # WinError 3197 (0xC7D)
    ERROR_DIF_VOLATILE_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x0C7E, 'ERROR_DIF_VOLATILE_NOT_ALLOWED') # WinError 3198 (0xC7E)
    ERROR_DIF_BINDING_API_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x0C7F, 'ERROR_DIF_BINDING_API_NOT_FOUND') # WinError 3199 (0xC7F)
    ERROR_IO_REISSUE_AS_CACHED: Final[ThirtyTwoBits] = cached_win32error(0x0F6E, 'ERROR_IO_REISSUE_AS_CACHED') # WinError 3950 (0xF6E)
    ERROR_WINS_INTERNAL: Final[ThirtyTwoBits] = cached_win32error(0x0FA0, 'ERROR_WINS_INTERNAL') # WinError 4000 (0xFA0)
    ERROR_CAN_NOT_DEL_LOCAL_WINS: Final[ThirtyTwoBits] = cached_win32error(0x0FA1, 'ERROR_CAN_NOT_DEL_LOCAL_WINS') # WinError 4001 (0xFA1)
    ERROR_STATIC_INIT: Final[ThirtyTwoBits] = cached_win32error(0x0FA2, 'ERROR_STATIC_INIT') # WinError 4002 (0xFA2)
    ERROR_INC_BACKUP: Final[ThirtyTwoBits] = cached_win32error(0x0FA3, 'ERROR_INC_BACKUP') # WinError 4003 (0xFA3)
    ERROR_FULL_BACKUP: Final[ThirtyTwoBits] = cached_win32error(0x0FA4, 'ERROR_FULL_BACKUP') # WinError 4004 (0xFA4)
    ERROR_REC_NON_EXISTENT: Final[ThirtyTwoBits] = cached_win32error(0x0FA5, 'ERROR_REC_NON_EXISTENT') # WinError 4005 (0xFA5)
    ERROR_RPL_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x0FA6, 'ERROR_RPL_NOT_ALLOWED') # WinError 4006 (0xFA6)
    PEERDIST_ERROR_CONTENTINFO_VERSION_UNSUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x0FD2, 'PEERDIST_ERROR_CONTENTINFO_VERSION_UNSUPPORTED') # WinError 4050 (0xFD2)
    PEERDIST_ERROR_CANNOT_PARSE_CONTENTINFO: Final[ThirtyTwoBits] = cached_win32error(0x0FD3, 'PEERDIST_ERROR_CANNOT_PARSE_CONTENTINFO') # WinError 4051 (0xFD3)
    PEERDIST_ERROR_MISSING_DATA: Final[ThirtyTwoBits] = cached_win32error(0x0FD4, 'PEERDIST_ERROR_MISSING_DATA') # WinError 4052 (0xFD4)
    PEERDIST_ERROR_NO_MORE: Final[ThirtyTwoBits] = cached_win32error(0x0FD5, 'PEERDIST_ERROR_NO_MORE') # WinError 4053 (0xFD5)
    PEERDIST_ERROR_NOT_INITIALIZED: Final[ThirtyTwoBits] = cached_win32error(0x0FD6, 'PEERDIST_ERROR_NOT_INITIALIZED') # WinError 4054 (0xFD6)
    PEERDIST_ERROR_ALREADY_INITIALIZED: Final[ThirtyTwoBits] = cached_win32error(0x0FD7, 'PEERDIST_ERROR_ALREADY_INITIALIZED') # WinError 4055 (0xFD7)
    PEERDIST_ERROR_SHUTDOWN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x0FD8, 'PEERDIST_ERROR_SHUTDOWN_IN_PROGRESS') # WinError 4056 (0xFD8)
    PEERDIST_ERROR_INVALIDATED: Final[ThirtyTwoBits] = cached_win32error(0x0FD9, 'PEERDIST_ERROR_INVALIDATED') # WinError 4057 (0xFD9)
    PEERDIST_ERROR_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x0FDA, 'PEERDIST_ERROR_ALREADY_EXISTS') # WinError 4058 (0xFDA)
    PEERDIST_ERROR_OPERATION_NOTFOUND: Final[ThirtyTwoBits] = cached_win32error(0x0FDB, 'PEERDIST_ERROR_OPERATION_NOTFOUND') # WinError 4059 (0xFDB)
    PEERDIST_ERROR_ALREADY_COMPLETED: Final[ThirtyTwoBits] = cached_win32error(0x0FDC, 'PEERDIST_ERROR_ALREADY_COMPLETED') # WinError 4060 (0xFDC)
    PEERDIST_ERROR_OUT_OF_BOUNDS: Final[ThirtyTwoBits] = cached_win32error(0x0FDD, 'PEERDIST_ERROR_OUT_OF_BOUNDS') # WinError 4061 (0xFDD)
    PEERDIST_ERROR_VERSION_UNSUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x0FDE, 'PEERDIST_ERROR_VERSION_UNSUPPORTED') # WinError 4062 (0xFDE)
    PEERDIST_ERROR_INVALID_CONFIGURATION: Final[ThirtyTwoBits] = cached_win32error(0x0FDF, 'PEERDIST_ERROR_INVALID_CONFIGURATION') # WinError 4063 (0xFDF)
    PEERDIST_ERROR_NOT_LICENSED: Final[ThirtyTwoBits] = cached_win32error(0x0FE0, 'PEERDIST_ERROR_NOT_LICENSED') # WinError 4064 (0xFE0)
    PEERDIST_ERROR_SERVICE_UNAVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x0FE1, 'PEERDIST_ERROR_SERVICE_UNAVAILABLE') # WinError 4065 (0xFE1)
    PEERDIST_ERROR_TRUST_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x0FE2, 'PEERDIST_ERROR_TRUST_FAILURE') # WinError 4066 (0xFE2)
    ERROR_DHCP_ADDRESS_CONFLICT: Final[ThirtyTwoBits] = cached_win32error(0x1004, 'ERROR_DHCP_ADDRESS_CONFLICT') # WinError 4100 (0x1004)
    ERROR_WMI_GUID_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x1068, 'ERROR_WMI_GUID_NOT_FOUND') # WinError 4200 (0x1068)
    ERROR_WMI_INSTANCE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x1069, 'ERROR_WMI_INSTANCE_NOT_FOUND') # WinError 4201 (0x1069)
    ERROR_WMI_ITEMID_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x106A, 'ERROR_WMI_ITEMID_NOT_FOUND') # WinError 4202 (0x106A)
    ERROR_WMI_TRY_AGAIN: Final[ThirtyTwoBits] = cached_win32error(0x106B, 'ERROR_WMI_TRY_AGAIN') # WinError 4203 (0x106B)
    ERROR_WMI_DP_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x106C, 'ERROR_WMI_DP_NOT_FOUND') # WinError 4204 (0x106C)
    ERROR_WMI_UNRESOLVED_INSTANCE_REF: Final[ThirtyTwoBits] = cached_win32error(0x106D, 'ERROR_WMI_UNRESOLVED_INSTANCE_REF') # WinError 4205 (0x106D)
    ERROR_WMI_ALREADY_ENABLED: Final[ThirtyTwoBits] = cached_win32error(0x106E, 'ERROR_WMI_ALREADY_ENABLED') # WinError 4206 (0x106E)
    ERROR_WMI_GUID_DISCONNECTED: Final[ThirtyTwoBits] = cached_win32error(0x106F, 'ERROR_WMI_GUID_DISCONNECTED') # WinError 4207 (0x106F)
    ERROR_WMI_SERVER_UNAVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x1070, 'ERROR_WMI_SERVER_UNAVAILABLE') # WinError 4208 (0x1070)
    ERROR_WMI_DP_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x1071, 'ERROR_WMI_DP_FAILED') # WinError 4209 (0x1071)
    ERROR_WMI_INVALID_MOF: Final[ThirtyTwoBits] = cached_win32error(0x1072, 'ERROR_WMI_INVALID_MOF') # WinError 4210 (0x1072)
    ERROR_WMI_INVALID_REGINFO: Final[ThirtyTwoBits] = cached_win32error(0x1073, 'ERROR_WMI_INVALID_REGINFO') # WinError 4211 (0x1073)
    ERROR_WMI_ALREADY_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x1074, 'ERROR_WMI_ALREADY_DISABLED') # WinError 4212 (0x1074)
    ERROR_WMI_READ_ONLY: Final[ThirtyTwoBits] = cached_win32error(0x1075, 'ERROR_WMI_READ_ONLY') # WinError 4213 (0x1075)
    ERROR_WMI_SET_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x1076, 'ERROR_WMI_SET_FAILURE') # WinError 4214 (0x1076)
    ERROR_NOT_APPCONTAINER: Final[ThirtyTwoBits] = cached_win32error(0x109A, 'ERROR_NOT_APPCONTAINER') # WinError 4250 (0x109A)
    ERROR_APPCONTAINER_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x109B, 'ERROR_APPCONTAINER_REQUIRED') # WinError 4251 (0x109B)
    ERROR_NOT_SUPPORTED_IN_APPCONTAINER: Final[ThirtyTwoBits] = cached_win32error(0x109C, 'ERROR_NOT_SUPPORTED_IN_APPCONTAINER') # WinError 4252 (0x109C)
    ERROR_INVALID_PACKAGE_SID_LENGTH: Final[ThirtyTwoBits] = cached_win32error(0x109D, 'ERROR_INVALID_PACKAGE_SID_LENGTH') # WinError 4253 (0x109D)
    ERROR_INVALID_MEDIA: Final[ThirtyTwoBits] = cached_win32error(0x10CC, 'ERROR_INVALID_MEDIA') # WinError 4300 (0x10CC)
    ERROR_INVALID_LIBRARY: Final[ThirtyTwoBits] = cached_win32error(0x10CD, 'ERROR_INVALID_LIBRARY') # WinError 4301 (0x10CD)
    ERROR_INVALID_MEDIA_POOL: Final[ThirtyTwoBits] = cached_win32error(0x10CE, 'ERROR_INVALID_MEDIA_POOL') # WinError 4302 (0x10CE)
    ERROR_DRIVE_MEDIA_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x10CF, 'ERROR_DRIVE_MEDIA_MISMATCH') # WinError 4303 (0x10CF)
    ERROR_MEDIA_OFFLINE: Final[ThirtyTwoBits] = cached_win32error(0x10D0, 'ERROR_MEDIA_OFFLINE') # WinError 4304 (0x10D0)
    ERROR_LIBRARY_OFFLINE: Final[ThirtyTwoBits] = cached_win32error(0x10D1, 'ERROR_LIBRARY_OFFLINE') # WinError 4305 (0x10D1)
    ERROR_EMPTY: Final[ThirtyTwoBits] = cached_win32error(0x10D2, 'ERROR_EMPTY') # WinError 4306 (0x10D2)
    ERROR_NOT_EMPTY: Final[ThirtyTwoBits] = cached_win32error(0x10D3, 'ERROR_NOT_EMPTY') # WinError 4307 (0x10D3)
    ERROR_MEDIA_UNAVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x10D4, 'ERROR_MEDIA_UNAVAILABLE') # WinError 4308 (0x10D4)
    ERROR_RESOURCE_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x10D5, 'ERROR_RESOURCE_DISABLED') # WinError 4309 (0x10D5)
    ERROR_INVALID_CLEANER: Final[ThirtyTwoBits] = cached_win32error(0x10D6, 'ERROR_INVALID_CLEANER') # WinError 4310 (0x10D6)
    ERROR_UNABLE_TO_CLEAN: Final[ThirtyTwoBits] = cached_win32error(0x10D7, 'ERROR_UNABLE_TO_CLEAN') # WinError 4311 (0x10D7)
    ERROR_OBJECT_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x10D8, 'ERROR_OBJECT_NOT_FOUND') # WinError 4312 (0x10D8)
    ERROR_DATABASE_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x10D9, 'ERROR_DATABASE_FAILURE') # WinError 4313 (0x10D9)
    ERROR_DATABASE_FULL: Final[ThirtyTwoBits] = cached_win32error(0x10DA, 'ERROR_DATABASE_FULL') # WinError 4314 (0x10DA)
    ERROR_MEDIA_INCOMPATIBLE: Final[ThirtyTwoBits] = cached_win32error(0x10DB, 'ERROR_MEDIA_INCOMPATIBLE') # WinError 4315 (0x10DB)
    ERROR_RESOURCE_NOT_PRESENT: Final[ThirtyTwoBits] = cached_win32error(0x10DC, 'ERROR_RESOURCE_NOT_PRESENT') # WinError 4316 (0x10DC)
    ERROR_INVALID_OPERATION: Final[ThirtyTwoBits] = cached_win32error(0x10DD, 'ERROR_INVALID_OPERATION') # WinError 4317 (0x10DD)
    ERROR_MEDIA_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x10DE, 'ERROR_MEDIA_NOT_AVAILABLE') # WinError 4318 (0x10DE)
    ERROR_DEVICE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x10DF, 'ERROR_DEVICE_NOT_AVAILABLE') # WinError 4319 (0x10DF)
    ERROR_REQUEST_REFUSED: Final[ThirtyTwoBits] = cached_win32error(0x10E0, 'ERROR_REQUEST_REFUSED') # WinError 4320 (0x10E0)
    ERROR_INVALID_DRIVE_OBJECT: Final[ThirtyTwoBits] = cached_win32error(0x10E1, 'ERROR_INVALID_DRIVE_OBJECT') # WinError 4321 (0x10E1)
    ERROR_LIBRARY_FULL: Final[ThirtyTwoBits] = cached_win32error(0x10E2, 'ERROR_LIBRARY_FULL') # WinError 4322 (0x10E2)
    ERROR_MEDIUM_NOT_ACCESSIBLE: Final[ThirtyTwoBits] = cached_win32error(0x10E3, 'ERROR_MEDIUM_NOT_ACCESSIBLE') # WinError 4323 (0x10E3)
    ERROR_UNABLE_TO_LOAD_MEDIUM: Final[ThirtyTwoBits] = cached_win32error(0x10E4, 'ERROR_UNABLE_TO_LOAD_MEDIUM') # WinError 4324 (0x10E4)
    ERROR_UNABLE_TO_INVENTORY_DRIVE: Final[ThirtyTwoBits] = cached_win32error(0x10E5, 'ERROR_UNABLE_TO_INVENTORY_DRIVE') # WinError 4325 (0x10E5)
    ERROR_UNABLE_TO_INVENTORY_SLOT: Final[ThirtyTwoBits] = cached_win32error(0x10E6, 'ERROR_UNABLE_TO_INVENTORY_SLOT') # WinError 4326 (0x10E6)
    ERROR_UNABLE_TO_INVENTORY_TRANSPORT: Final[ThirtyTwoBits] = cached_win32error(0x10E7, 'ERROR_UNABLE_TO_INVENTORY_TRANSPORT') # WinError 4327 (0x10E7)
    ERROR_TRANSPORT_FULL: Final[ThirtyTwoBits] = cached_win32error(0x10E8, 'ERROR_TRANSPORT_FULL') # WinError 4328 (0x10E8)
    ERROR_CONTROLLING_IEPORT: Final[ThirtyTwoBits] = cached_win32error(0x10E9, 'ERROR_CONTROLLING_IEPORT') # WinError 4329 (0x10E9)
    ERROR_UNABLE_TO_EJECT_MOUNTED_MEDIA: Final[ThirtyTwoBits] = cached_win32error(0x10EA, 'ERROR_UNABLE_TO_EJECT_MOUNTED_MEDIA') # WinError 4330 (0x10EA)
    ERROR_CLEANER_SLOT_SET: Final[ThirtyTwoBits] = cached_win32error(0x10EB, 'ERROR_CLEANER_SLOT_SET') # WinError 4331 (0x10EB)
    ERROR_CLEANER_SLOT_NOT_SET: Final[ThirtyTwoBits] = cached_win32error(0x10EC, 'ERROR_CLEANER_SLOT_NOT_SET') # WinError 4332 (0x10EC)
    ERROR_CLEANER_CARTRIDGE_SPENT: Final[ThirtyTwoBits] = cached_win32error(0x10ED, 'ERROR_CLEANER_CARTRIDGE_SPENT') # WinError 4333 (0x10ED)
    ERROR_UNEXPECTED_OMID: Final[ThirtyTwoBits] = cached_win32error(0x10EE, 'ERROR_UNEXPECTED_OMID') # WinError 4334 (0x10EE)
    ERROR_CANT_DELETE_LAST_ITEM: Final[ThirtyTwoBits] = cached_win32error(0x10EF, 'ERROR_CANT_DELETE_LAST_ITEM') # WinError 4335 (0x10EF)
    ERROR_MESSAGE_EXCEEDS_MAX_SIZE: Final[ThirtyTwoBits] = cached_win32error(0x10F0, 'ERROR_MESSAGE_EXCEEDS_MAX_SIZE') # WinError 4336 (0x10F0)
    ERROR_VOLUME_CONTAINS_SYS_FILES: Final[ThirtyTwoBits] = cached_win32error(0x10F1, 'ERROR_VOLUME_CONTAINS_SYS_FILES') # WinError 4337 (0x10F1)
    ERROR_INDIGENOUS_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x10F2, 'ERROR_INDIGENOUS_TYPE') # WinError 4338 (0x10F2)
    ERROR_NO_SUPPORTING_DRIVES: Final[ThirtyTwoBits] = cached_win32error(0x10F3, 'ERROR_NO_SUPPORTING_DRIVES') # WinError 4339 (0x10F3)
    ERROR_CLEANER_CARTRIDGE_INSTALLED: Final[ThirtyTwoBits] = cached_win32error(0x10F4, 'ERROR_CLEANER_CARTRIDGE_INSTALLED') # WinError 4340 (0x10F4)
    ERROR_IEPORT_FULL: Final[ThirtyTwoBits] = cached_win32error(0x10F5, 'ERROR_IEPORT_FULL') # WinError 4341 (0x10F5)
    ERROR_FILE_OFFLINE: Final[ThirtyTwoBits] = cached_win32error(0x10FE, 'ERROR_FILE_OFFLINE') # WinError 4350 (0x10FE)
    ERROR_REMOTE_STORAGE_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_win32error(0x10FF, 'ERROR_REMOTE_STORAGE_NOT_ACTIVE') # WinError 4351 (0x10FF)
    ERROR_REMOTE_STORAGE_MEDIA_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x1100, 'ERROR_REMOTE_STORAGE_MEDIA_ERROR') # WinError 4352 (0x1100)
    ERROR_NOT_A_REPARSE_POINT: Final[ThirtyTwoBits] = cached_win32error(0x1126, 'ERROR_NOT_A_REPARSE_POINT') # WinError 4390 (0x1126)
    ERROR_REPARSE_ATTRIBUTE_CONFLICT: Final[ThirtyTwoBits] = cached_win32error(0x1127, 'ERROR_REPARSE_ATTRIBUTE_CONFLICT') # WinError 4391 (0x1127)
    ERROR_INVALID_REPARSE_DATA: Final[ThirtyTwoBits] = cached_win32error(0x1128, 'ERROR_INVALID_REPARSE_DATA') # WinError 4392 (0x1128)
    ERROR_REPARSE_TAG_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x1129, 'ERROR_REPARSE_TAG_INVALID') # WinError 4393 (0x1129)
    ERROR_REPARSE_TAG_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x112A, 'ERROR_REPARSE_TAG_MISMATCH') # WinError 4394 (0x112A)
    ERROR_REPARSE_POINT_ENCOUNTERED: Final[ThirtyTwoBits] = cached_win32error(0x112B, 'ERROR_REPARSE_POINT_ENCOUNTERED') # WinError 4395 (0x112B)
    ERROR_APP_DATA_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x1130, 'ERROR_APP_DATA_NOT_FOUND') # WinError 4400 (0x1130)
    ERROR_APP_DATA_EXPIRED: Final[ThirtyTwoBits] = cached_win32error(0x1131, 'ERROR_APP_DATA_EXPIRED') # WinError 4401 (0x1131)
    ERROR_APP_DATA_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x1132, 'ERROR_APP_DATA_CORRUPT') # WinError 4402 (0x1132)
    ERROR_APP_DATA_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x1133, 'ERROR_APP_DATA_LIMIT_EXCEEDED') # WinError 4403 (0x1133)
    ERROR_APP_DATA_REBOOT_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x1134, 'ERROR_APP_DATA_REBOOT_REQUIRED') # WinError 4404 (0x1134)
    ERROR_SECUREBOOT_ROLLBACK_DETECTED: Final[ThirtyTwoBits] = cached_win32error(0x1144, 'ERROR_SECUREBOOT_ROLLBACK_DETECTED') # WinError 4420 (0x1144)
    ERROR_SECUREBOOT_POLICY_VIOLATION: Final[ThirtyTwoBits] = cached_win32error(0x1145, 'ERROR_SECUREBOOT_POLICY_VIOLATION') # WinError 4421 (0x1145)
    ERROR_SECUREBOOT_INVALID_POLICY: Final[ThirtyTwoBits] = cached_win32error(0x1146, 'ERROR_SECUREBOOT_INVALID_POLICY') # WinError 4422 (0x1146)
    ERROR_SECUREBOOT_POLICY_PUBLISHER_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x1147, 'ERROR_SECUREBOOT_POLICY_PUBLISHER_NOT_FOUND') # WinError 4423 (0x1147)
    ERROR_SECUREBOOT_POLICY_NOT_SIGNED: Final[ThirtyTwoBits] = cached_win32error(0x1148, 'ERROR_SECUREBOOT_POLICY_NOT_SIGNED') # WinError 4424 (0x1148)
    ERROR_SECUREBOOT_NOT_ENABLED: Final[ThirtyTwoBits] = cached_win32error(0x1149, 'ERROR_SECUREBOOT_NOT_ENABLED') # WinError 4425 (0x1149)
    ERROR_SECUREBOOT_FILE_REPLACED: Final[ThirtyTwoBits] = cached_win32error(0x114A, 'ERROR_SECUREBOOT_FILE_REPLACED') # WinError 4426 (0x114A)
    ERROR_SECUREBOOT_POLICY_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_win32error(0x114B, 'ERROR_SECUREBOOT_POLICY_NOT_AUTHORIZED') # WinError 4427 (0x114B)
    ERROR_SECUREBOOT_POLICY_UNKNOWN: Final[ThirtyTwoBits] = cached_win32error(0x114C, 'ERROR_SECUREBOOT_POLICY_UNKNOWN') # WinError 4428 (0x114C)
    ERROR_SECUREBOOT_POLICY_MISSING_ANTIROLLBACKVERSION: Final[ThirtyTwoBits] = cached_win32error(0x114D, 'ERROR_SECUREBOOT_POLICY_MISSING_ANTIROLLBACKVERSION') # WinError 4429 (0x114D)
    ERROR_SECUREBOOT_PLATFORM_ID_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x114E, 'ERROR_SECUREBOOT_PLATFORM_ID_MISMATCH') # WinError 4430 (0x114E)
    ERROR_SECUREBOOT_POLICY_ROLLBACK_DETECTED: Final[ThirtyTwoBits] = cached_win32error(0x114F, 'ERROR_SECUREBOOT_POLICY_ROLLBACK_DETECTED') # WinError 4431 (0x114F)
    ERROR_SECUREBOOT_POLICY_UPGRADE_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x1150, 'ERROR_SECUREBOOT_POLICY_UPGRADE_MISMATCH') # WinError 4432 (0x1150)
    ERROR_SECUREBOOT_REQUIRED_POLICY_FILE_MISSING: Final[ThirtyTwoBits] = cached_win32error(0x1151, 'ERROR_SECUREBOOT_REQUIRED_POLICY_FILE_MISSING') # WinError 4433 (0x1151)
    ERROR_SECUREBOOT_NOT_BASE_POLICY: Final[ThirtyTwoBits] = cached_win32error(0x1152, 'ERROR_SECUREBOOT_NOT_BASE_POLICY') # WinError 4434 (0x1152)
    ERROR_SECUREBOOT_NOT_SUPPLEMENTAL_POLICY: Final[ThirtyTwoBits] = cached_win32error(0x1153, 'ERROR_SECUREBOOT_NOT_SUPPLEMENTAL_POLICY') # WinError 4435 (0x1153)
    ERROR_OFFLOAD_READ_FLT_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x1158, 'ERROR_OFFLOAD_READ_FLT_NOT_SUPPORTED') # WinError 4440 (0x1158)
    ERROR_OFFLOAD_WRITE_FLT_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x1159, 'ERROR_OFFLOAD_WRITE_FLT_NOT_SUPPORTED') # WinError 4441 (0x1159)
    ERROR_OFFLOAD_READ_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x115A, 'ERROR_OFFLOAD_READ_FILE_NOT_SUPPORTED') # WinError 4442 (0x115A)
    ERROR_OFFLOAD_WRITE_FILE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x115B, 'ERROR_OFFLOAD_WRITE_FILE_NOT_SUPPORTED') # WinError 4443 (0x115B)
    ERROR_ALREADY_HAS_STREAM_ID: Final[ThirtyTwoBits] = cached_win32error(0x115C, 'ERROR_ALREADY_HAS_STREAM_ID') # WinError 4444 (0x115C)
    ERROR_SMR_GARBAGE_COLLECTION_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x115D, 'ERROR_SMR_GARBAGE_COLLECTION_REQUIRED') # WinError 4445 (0x115D)
    ERROR_WOF_WIM_HEADER_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x115E, 'ERROR_WOF_WIM_HEADER_CORRUPT') # WinError 4446 (0x115E)
    ERROR_WOF_WIM_RESOURCE_TABLE_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x115F, 'ERROR_WOF_WIM_RESOURCE_TABLE_CORRUPT') # WinError 4447 (0x115F)
    ERROR_WOF_FILE_RESOURCE_TABLE_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x1160, 'ERROR_WOF_FILE_RESOURCE_TABLE_CORRUPT') # WinError 4448 (0x1160)
    ERROR_OBJECT_IS_IMMUTABLE: Final[ThirtyTwoBits] = cached_win32error(0x1161, 'ERROR_OBJECT_IS_IMMUTABLE') # WinError 4449 (0x1161)
    ERROR_VOLUME_NOT_SIS_ENABLED: Final[ThirtyTwoBits] = cached_win32error(0x1194, 'ERROR_VOLUME_NOT_SIS_ENABLED') # WinError 4500 (0x1194)
    ERROR_SYSTEM_INTEGRITY_ROLLBACK_DETECTED: Final[ThirtyTwoBits] = cached_win32error(0x11C6, 'ERROR_SYSTEM_INTEGRITY_ROLLBACK_DETECTED') # WinError 4550 (0x11C6)
    ERROR_SYSTEM_INTEGRITY_POLICY_VIOLATION: Final[ThirtyTwoBits] = cached_win32error(0x11C7, 'ERROR_SYSTEM_INTEGRITY_POLICY_VIOLATION') # WinError 4551 (0x11C7)
    ERROR_SYSTEM_INTEGRITY_INVALID_POLICY: Final[ThirtyTwoBits] = cached_win32error(0x11C8, 'ERROR_SYSTEM_INTEGRITY_INVALID_POLICY') # WinError 4552 (0x11C8)
    ERROR_SYSTEM_INTEGRITY_POLICY_NOT_SIGNED: Final[ThirtyTwoBits] = cached_win32error(0x11C9, 'ERROR_SYSTEM_INTEGRITY_POLICY_NOT_SIGNED') # WinError 4553 (0x11C9)
    ERROR_SYSTEM_INTEGRITY_TOO_MANY_POLICIES: Final[ThirtyTwoBits] = cached_win32error(0x11CA, 'ERROR_SYSTEM_INTEGRITY_TOO_MANY_POLICIES') # WinError 4554 (0x11CA)
    ERROR_SYSTEM_INTEGRITY_SUPPLEMENTAL_POLICY_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_win32error(0x11CB, 'ERROR_SYSTEM_INTEGRITY_SUPPLEMENTAL_POLICY_NOT_AUTHORIZED') # WinError 4555 (0x11CB)
    ERROR_SYSTEM_INTEGRITY_REPUTATION_MALICIOUS: Final[ThirtyTwoBits] = cached_win32error(0x11CC, 'ERROR_SYSTEM_INTEGRITY_REPUTATION_MALICIOUS') # WinError 4556 (0x11CC)
    ERROR_SYSTEM_INTEGRITY_REPUTATION_PUA: Final[ThirtyTwoBits] = cached_win32error(0x11CD, 'ERROR_SYSTEM_INTEGRITY_REPUTATION_PUA') # WinError 4557 (0x11CD)
    ERROR_SYSTEM_INTEGRITY_REPUTATION_DANGEROUS_EXT: Final[ThirtyTwoBits] = cached_win32error(0x11CE, 'ERROR_SYSTEM_INTEGRITY_REPUTATION_DANGEROUS_EXT') # WinError 4558 (0x11CE)
    ERROR_SYSTEM_INTEGRITY_REPUTATION_OFFLINE: Final[ThirtyTwoBits] = cached_win32error(0x11CF, 'ERROR_SYSTEM_INTEGRITY_REPUTATION_OFFLINE') # WinError 4559 (0x11CF)
    ERROR_VSM_NOT_INITIALIZED: Final[ThirtyTwoBits] = cached_win32error(0x11D0, 'ERROR_VSM_NOT_INITIALIZED') # WinError 4560 (0x11D0)
    ERROR_VSM_DMA_PROTECTION_NOT_IN_USE: Final[ThirtyTwoBits] = cached_win32error(0x11D1, 'ERROR_VSM_DMA_PROTECTION_NOT_IN_USE') # WinError 4561 (0x11D1)
    ERROR_PLATFORM_MANIFEST_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_win32error(0x11DA, 'ERROR_PLATFORM_MANIFEST_NOT_AUTHORIZED') # WinError 4570 (0x11DA)
    ERROR_PLATFORM_MANIFEST_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x11DB, 'ERROR_PLATFORM_MANIFEST_INVALID') # WinError 4571 (0x11DB)
    ERROR_PLATFORM_MANIFEST_FILE_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_win32error(0x11DC, 'ERROR_PLATFORM_MANIFEST_FILE_NOT_AUTHORIZED') # WinError 4572 (0x11DC)
    ERROR_PLATFORM_MANIFEST_CATALOG_NOT_AUTHORIZED: Final[ThirtyTwoBits] = cached_win32error(0x11DD, 'ERROR_PLATFORM_MANIFEST_CATALOG_NOT_AUTHORIZED') # WinError 4573 (0x11DD)
    ERROR_PLATFORM_MANIFEST_BINARY_ID_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x11DE, 'ERROR_PLATFORM_MANIFEST_BINARY_ID_NOT_FOUND') # WinError 4574 (0x11DE)
    ERROR_PLATFORM_MANIFEST_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_win32error(0x11DF, 'ERROR_PLATFORM_MANIFEST_NOT_ACTIVE') # WinError 4575 (0x11DF)
    ERROR_PLATFORM_MANIFEST_NOT_SIGNED: Final[ThirtyTwoBits] = cached_win32error(0x11E0, 'ERROR_PLATFORM_MANIFEST_NOT_SIGNED') # WinError 4576 (0x11E0)
    ERROR_SYSTEM_INTEGRITY_REPUTATION_UNFRIENDLY_FILE: Final[ThirtyTwoBits] = cached_win32error(0x11E4, 'ERROR_SYSTEM_INTEGRITY_REPUTATION_UNFRIENDLY_FILE') # WinError 4580 (0x11E4)
    ERROR_SYSTEM_INTEGRITY_REPUTATION_UNATTAINABLE: Final[ThirtyTwoBits] = cached_win32error(0x11E5, 'ERROR_SYSTEM_INTEGRITY_REPUTATION_UNATTAINABLE') # WinError 4581 (0x11E5)
    ERROR_SYSTEM_INTEGRITY_REPUTATION_EXPLICIT_DENY_FILE: Final[ThirtyTwoBits] = cached_win32error(0x11E6, 'ERROR_SYSTEM_INTEGRITY_REPUTATION_EXPLICIT_DENY_FILE') # WinError 4582 (0x11E6)
    ERROR_DEPENDENT_RESOURCE_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x1389, 'ERROR_DEPENDENT_RESOURCE_EXISTS') # WinError 5001 (0x1389)
    ERROR_DEPENDENCY_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x138A, 'ERROR_DEPENDENCY_NOT_FOUND') # WinError 5002 (0x138A)
    ERROR_DEPENDENCY_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x138B, 'ERROR_DEPENDENCY_ALREADY_EXISTS') # WinError 5003 (0x138B)
    ERROR_RESOURCE_NOT_ONLINE: Final[ThirtyTwoBits] = cached_win32error(0x138C, 'ERROR_RESOURCE_NOT_ONLINE') # WinError 5004 (0x138C)
    ERROR_HOST_NODE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x138D, 'ERROR_HOST_NODE_NOT_AVAILABLE') # WinError 5005 (0x138D)
    ERROR_RESOURCE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x138E, 'ERROR_RESOURCE_NOT_AVAILABLE') # WinError 5006 (0x138E)
    ERROR_RESOURCE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x138F, 'ERROR_RESOURCE_NOT_FOUND') # WinError 5007 (0x138F)
    ERROR_SHUTDOWN_CLUSTER: Final[ThirtyTwoBits] = cached_win32error(0x1390, 'ERROR_SHUTDOWN_CLUSTER') # WinError 5008 (0x1390)
    ERROR_CANT_EVICT_ACTIVE_NODE: Final[ThirtyTwoBits] = cached_win32error(0x1391, 'ERROR_CANT_EVICT_ACTIVE_NODE') # WinError 5009 (0x1391)
    ERROR_OBJECT_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x1392, 'ERROR_OBJECT_ALREADY_EXISTS') # WinError 5010 (0x1392)
    ERROR_OBJECT_IN_LIST: Final[ThirtyTwoBits] = cached_win32error(0x1393, 'ERROR_OBJECT_IN_LIST') # WinError 5011 (0x1393)
    ERROR_GROUP_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x1394, 'ERROR_GROUP_NOT_AVAILABLE') # WinError 5012 (0x1394)
    ERROR_GROUP_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x1395, 'ERROR_GROUP_NOT_FOUND') # WinError 5013 (0x1395)
    ERROR_GROUP_NOT_ONLINE: Final[ThirtyTwoBits] = cached_win32error(0x1396, 'ERROR_GROUP_NOT_ONLINE') # WinError 5014 (0x1396)
    ERROR_HOST_NODE_NOT_RESOURCE_OWNER: Final[ThirtyTwoBits] = cached_win32error(0x1397, 'ERROR_HOST_NODE_NOT_RESOURCE_OWNER') # WinError 5015 (0x1397)
    ERROR_HOST_NODE_NOT_GROUP_OWNER: Final[ThirtyTwoBits] = cached_win32error(0x1398, 'ERROR_HOST_NODE_NOT_GROUP_OWNER') # WinError 5016 (0x1398)
    ERROR_RESMON_CREATE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x1399, 'ERROR_RESMON_CREATE_FAILED') # WinError 5017 (0x1399)
    ERROR_RESMON_ONLINE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x139A, 'ERROR_RESMON_ONLINE_FAILED') # WinError 5018 (0x139A)
    ERROR_RESOURCE_ONLINE: Final[ThirtyTwoBits] = cached_win32error(0x139B, 'ERROR_RESOURCE_ONLINE') # WinError 5019 (0x139B)
    ERROR_QUORUM_RESOURCE: Final[ThirtyTwoBits] = cached_win32error(0x139C, 'ERROR_QUORUM_RESOURCE') # WinError 5020 (0x139C)
    ERROR_NOT_QUORUM_CAPABLE: Final[ThirtyTwoBits] = cached_win32error(0x139D, 'ERROR_NOT_QUORUM_CAPABLE') # WinError 5021 (0x139D)
    ERROR_CLUSTER_SHUTTING_DOWN: Final[ThirtyTwoBits] = cached_win32error(0x139E, 'ERROR_CLUSTER_SHUTTING_DOWN') # WinError 5022 (0x139E)
    ERROR_INVALID_STATE: Final[ThirtyTwoBits] = cached_win32error(0x139F, 'ERROR_INVALID_STATE') # WinError 5023 (0x139F)
    ERROR_RESOURCE_PROPERTIES_STORED: Final[ThirtyTwoBits] = cached_win32error(0x13A0, 'ERROR_RESOURCE_PROPERTIES_STORED') # WinError 5024 (0x13A0)
    ERROR_NOT_QUORUM_CLASS: Final[ThirtyTwoBits] = cached_win32error(0x13A1, 'ERROR_NOT_QUORUM_CLASS') # WinError 5025 (0x13A1)
    ERROR_CORE_RESOURCE: Final[ThirtyTwoBits] = cached_win32error(0x13A2, 'ERROR_CORE_RESOURCE') # WinError 5026 (0x13A2)
    ERROR_QUORUM_RESOURCE_ONLINE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x13A3, 'ERROR_QUORUM_RESOURCE_ONLINE_FAILED') # WinError 5027 (0x13A3)
    ERROR_QUORUMLOG_OPEN_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x13A4, 'ERROR_QUORUMLOG_OPEN_FAILED') # WinError 5028 (0x13A4)
    ERROR_CLUSTERLOG_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x13A5, 'ERROR_CLUSTERLOG_CORRUPT') # WinError 5029 (0x13A5)
    ERROR_CLUSTERLOG_RECORD_EXCEEDS_MAXSIZE: Final[ThirtyTwoBits] = cached_win32error(0x13A6, 'ERROR_CLUSTERLOG_RECORD_EXCEEDS_MAXSIZE') # WinError 5030 (0x13A6)
    ERROR_CLUSTERLOG_EXCEEDS_MAXSIZE: Final[ThirtyTwoBits] = cached_win32error(0x13A7, 'ERROR_CLUSTERLOG_EXCEEDS_MAXSIZE') # WinError 5031 (0x13A7)
    ERROR_CLUSTERLOG_CHKPOINT_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x13A8, 'ERROR_CLUSTERLOG_CHKPOINT_NOT_FOUND') # WinError 5032 (0x13A8)
    ERROR_CLUSTERLOG_NOT_ENOUGH_SPACE: Final[ThirtyTwoBits] = cached_win32error(0x13A9, 'ERROR_CLUSTERLOG_NOT_ENOUGH_SPACE') # WinError 5033 (0x13A9)
    ERROR_QUORUM_OWNER_ALIVE: Final[ThirtyTwoBits] = cached_win32error(0x13AA, 'ERROR_QUORUM_OWNER_ALIVE') # WinError 5034 (0x13AA)
    ERROR_NETWORK_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x13AB, 'ERROR_NETWORK_NOT_AVAILABLE') # WinError 5035 (0x13AB)
    ERROR_NODE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x13AC, 'ERROR_NODE_NOT_AVAILABLE') # WinError 5036 (0x13AC)
    ERROR_ALL_NODES_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x13AD, 'ERROR_ALL_NODES_NOT_AVAILABLE') # WinError 5037 (0x13AD)
    ERROR_RESOURCE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x13AE, 'ERROR_RESOURCE_FAILED') # WinError 5038 (0x13AE)
    ERROR_CLUSTER_INVALID_NODE: Final[ThirtyTwoBits] = cached_win32error(0x13AF, 'ERROR_CLUSTER_INVALID_NODE') # WinError 5039 (0x13AF)
    ERROR_CLUSTER_NODE_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x13B0, 'ERROR_CLUSTER_NODE_EXISTS') # WinError 5040 (0x13B0)
    ERROR_CLUSTER_JOIN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x13B1, 'ERROR_CLUSTER_JOIN_IN_PROGRESS') # WinError 5041 (0x13B1)
    ERROR_CLUSTER_NODE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x13B2, 'ERROR_CLUSTER_NODE_NOT_FOUND') # WinError 5042 (0x13B2)
    ERROR_CLUSTER_LOCAL_NODE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x13B3, 'ERROR_CLUSTER_LOCAL_NODE_NOT_FOUND') # WinError 5043 (0x13B3)
    ERROR_CLUSTER_NETWORK_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x13B4, 'ERROR_CLUSTER_NETWORK_EXISTS') # WinError 5044 (0x13B4)
    ERROR_CLUSTER_NETWORK_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x13B5, 'ERROR_CLUSTER_NETWORK_NOT_FOUND') # WinError 5045 (0x13B5)
    ERROR_CLUSTER_NETINTERFACE_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x13B6, 'ERROR_CLUSTER_NETINTERFACE_EXISTS') # WinError 5046 (0x13B6)
    ERROR_CLUSTER_NETINTERFACE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x13B7, 'ERROR_CLUSTER_NETINTERFACE_NOT_FOUND') # WinError 5047 (0x13B7)
    ERROR_CLUSTER_INVALID_REQUEST: Final[ThirtyTwoBits] = cached_win32error(0x13B8, 'ERROR_CLUSTER_INVALID_REQUEST') # WinError 5048 (0x13B8)
    ERROR_CLUSTER_INVALID_NETWORK_PROVIDER: Final[ThirtyTwoBits] = cached_win32error(0x13B9, 'ERROR_CLUSTER_INVALID_NETWORK_PROVIDER') # WinError 5049 (0x13B9)
    ERROR_CLUSTER_NODE_DOWN: Final[ThirtyTwoBits] = cached_win32error(0x13BA, 'ERROR_CLUSTER_NODE_DOWN') # WinError 5050 (0x13BA)
    ERROR_CLUSTER_NODE_UNREACHABLE: Final[ThirtyTwoBits] = cached_win32error(0x13BB, 'ERROR_CLUSTER_NODE_UNREACHABLE') # WinError 5051 (0x13BB)
    ERROR_CLUSTER_NODE_NOT_MEMBER: Final[ThirtyTwoBits] = cached_win32error(0x13BC, 'ERROR_CLUSTER_NODE_NOT_MEMBER') # WinError 5052 (0x13BC)
    ERROR_CLUSTER_JOIN_NOT_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x13BD, 'ERROR_CLUSTER_JOIN_NOT_IN_PROGRESS') # WinError 5053 (0x13BD)
    ERROR_CLUSTER_INVALID_NETWORK: Final[ThirtyTwoBits] = cached_win32error(0x13BE, 'ERROR_CLUSTER_INVALID_NETWORK') # WinError 5054 (0x13BE)
    ERROR_CLUSTER_NODE_UP: Final[ThirtyTwoBits] = cached_win32error(0x13C0, 'ERROR_CLUSTER_NODE_UP') # WinError 5056 (0x13C0)
    ERROR_CLUSTER_IPADDR_IN_USE: Final[ThirtyTwoBits] = cached_win32error(0x13C1, 'ERROR_CLUSTER_IPADDR_IN_USE') # WinError 5057 (0x13C1)
    ERROR_CLUSTER_NODE_NOT_PAUSED: Final[ThirtyTwoBits] = cached_win32error(0x13C2, 'ERROR_CLUSTER_NODE_NOT_PAUSED') # WinError 5058 (0x13C2)
    ERROR_CLUSTER_NO_SECURITY_CONTEXT: Final[ThirtyTwoBits] = cached_win32error(0x13C3, 'ERROR_CLUSTER_NO_SECURITY_CONTEXT') # WinError 5059 (0x13C3)
    ERROR_CLUSTER_NETWORK_NOT_INTERNAL: Final[ThirtyTwoBits] = cached_win32error(0x13C4, 'ERROR_CLUSTER_NETWORK_NOT_INTERNAL') # WinError 5060 (0x13C4)
    ERROR_CLUSTER_NODE_ALREADY_UP: Final[ThirtyTwoBits] = cached_win32error(0x13C5, 'ERROR_CLUSTER_NODE_ALREADY_UP') # WinError 5061 (0x13C5)
    ERROR_CLUSTER_NODE_ALREADY_DOWN: Final[ThirtyTwoBits] = cached_win32error(0x13C6, 'ERROR_CLUSTER_NODE_ALREADY_DOWN') # WinError 5062 (0x13C6)
    ERROR_CLUSTER_NETWORK_ALREADY_ONLINE: Final[ThirtyTwoBits] = cached_win32error(0x13C7, 'ERROR_CLUSTER_NETWORK_ALREADY_ONLINE') # WinError 5063 (0x13C7)
    ERROR_CLUSTER_NETWORK_ALREADY_OFFLINE: Final[ThirtyTwoBits] = cached_win32error(0x13C8, 'ERROR_CLUSTER_NETWORK_ALREADY_OFFLINE') # WinError 5064 (0x13C8)
    ERROR_CLUSTER_NODE_ALREADY_MEMBER: Final[ThirtyTwoBits] = cached_win32error(0x13C9, 'ERROR_CLUSTER_NODE_ALREADY_MEMBER') # WinError 5065 (0x13C9)
    ERROR_CLUSTER_LAST_INTERNAL_NETWORK: Final[ThirtyTwoBits] = cached_win32error(0x13CA, 'ERROR_CLUSTER_LAST_INTERNAL_NETWORK') # WinError 5066 (0x13CA)
    ERROR_CLUSTER_NETWORK_HAS_DEPENDENTS: Final[ThirtyTwoBits] = cached_win32error(0x13CB, 'ERROR_CLUSTER_NETWORK_HAS_DEPENDENTS') # WinError 5067 (0x13CB)
    ERROR_INVALID_OPERATION_ON_QUORUM: Final[ThirtyTwoBits] = cached_win32error(0x13CC, 'ERROR_INVALID_OPERATION_ON_QUORUM') # WinError 5068 (0x13CC)
    ERROR_DEPENDENCY_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x13CD, 'ERROR_DEPENDENCY_NOT_ALLOWED') # WinError 5069 (0x13CD)
    ERROR_CLUSTER_NODE_PAUSED: Final[ThirtyTwoBits] = cached_win32error(0x13CE, 'ERROR_CLUSTER_NODE_PAUSED') # WinError 5070 (0x13CE)
    ERROR_NODE_CANT_HOST_RESOURCE: Final[ThirtyTwoBits] = cached_win32error(0x13CF, 'ERROR_NODE_CANT_HOST_RESOURCE') # WinError 5071 (0x13CF)
    ERROR_CLUSTER_NODE_NOT_READY: Final[ThirtyTwoBits] = cached_win32error(0x13D0, 'ERROR_CLUSTER_NODE_NOT_READY') # WinError 5072 (0x13D0)
    ERROR_CLUSTER_NODE_SHUTTING_DOWN: Final[ThirtyTwoBits] = cached_win32error(0x13D1, 'ERROR_CLUSTER_NODE_SHUTTING_DOWN') # WinError 5073 (0x13D1)
    ERROR_CLUSTER_JOIN_ABORTED: Final[ThirtyTwoBits] = cached_win32error(0x13D2, 'ERROR_CLUSTER_JOIN_ABORTED') # WinError 5074 (0x13D2)
    ERROR_CLUSTER_INCOMPATIBLE_VERSIONS: Final[ThirtyTwoBits] = cached_win32error(0x13D3, 'ERROR_CLUSTER_INCOMPATIBLE_VERSIONS') # WinError 5075 (0x13D3)
    ERROR_CLUSTER_MAXNUM_OF_RESOURCES_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x13D4, 'ERROR_CLUSTER_MAXNUM_OF_RESOURCES_EXCEEDED') # WinError 5076 (0x13D4)
    ERROR_CLUSTER_SYSTEM_CONFIG_CHANGED: Final[ThirtyTwoBits] = cached_win32error(0x13D5, 'ERROR_CLUSTER_SYSTEM_CONFIG_CHANGED') # WinError 5077 (0x13D5)
    ERROR_CLUSTER_RESOURCE_TYPE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x13D6, 'ERROR_CLUSTER_RESOURCE_TYPE_NOT_FOUND') # WinError 5078 (0x13D6)
    ERROR_CLUSTER_RESTYPE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x13D7, 'ERROR_CLUSTER_RESTYPE_NOT_SUPPORTED') # WinError 5079 (0x13D7)
    ERROR_CLUSTER_RESNAME_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x13D8, 'ERROR_CLUSTER_RESNAME_NOT_FOUND') # WinError 5080 (0x13D8)
    ERROR_CLUSTER_NO_RPC_PACKAGES_REGISTERED: Final[ThirtyTwoBits] = cached_win32error(0x13D9, 'ERROR_CLUSTER_NO_RPC_PACKAGES_REGISTERED') # WinError 5081 (0x13D9)
    ERROR_CLUSTER_OWNER_NOT_IN_PREFLIST: Final[ThirtyTwoBits] = cached_win32error(0x13DA, 'ERROR_CLUSTER_OWNER_NOT_IN_PREFLIST') # WinError 5082 (0x13DA)
    ERROR_CLUSTER_DATABASE_SEQMISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x13DB, 'ERROR_CLUSTER_DATABASE_SEQMISMATCH') # WinError 5083 (0x13DB)
    ERROR_RESMON_INVALID_STATE: Final[ThirtyTwoBits] = cached_win32error(0x13DC, 'ERROR_RESMON_INVALID_STATE') # WinError 5084 (0x13DC)
    ERROR_CLUSTER_GUM_NOT_LOCKER: Final[ThirtyTwoBits] = cached_win32error(0x13DD, 'ERROR_CLUSTER_GUM_NOT_LOCKER') # WinError 5085 (0x13DD)
    ERROR_QUORUM_DISK_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x13DE, 'ERROR_QUORUM_DISK_NOT_FOUND') # WinError 5086 (0x13DE)
    ERROR_DATABASE_BACKUP_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x13DF, 'ERROR_DATABASE_BACKUP_CORRUPT') # WinError 5087 (0x13DF)
    ERROR_CLUSTER_NODE_ALREADY_HAS_DFS_ROOT: Final[ThirtyTwoBits] = cached_win32error(0x13E0, 'ERROR_CLUSTER_NODE_ALREADY_HAS_DFS_ROOT') # WinError 5088 (0x13E0)
    ERROR_RESOURCE_PROPERTY_UNCHANGEABLE: Final[ThirtyTwoBits] = cached_win32error(0x13E1, 'ERROR_RESOURCE_PROPERTY_UNCHANGEABLE') # WinError 5089 (0x13E1)
    ERROR_NO_ADMIN_ACCESS_POINT: Final[ThirtyTwoBits] = cached_win32error(0x13E2, 'ERROR_NO_ADMIN_ACCESS_POINT') # WinError 5090 (0x13E2)
    ERROR_CLUSTER_MEMBERSHIP_INVALID_STATE: Final[ThirtyTwoBits] = cached_win32error(0x1702, 'ERROR_CLUSTER_MEMBERSHIP_INVALID_STATE') # WinError 5890 (0x1702)
    ERROR_CLUSTER_QUORUMLOG_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x1703, 'ERROR_CLUSTER_QUORUMLOG_NOT_FOUND') # WinError 5891 (0x1703)
    ERROR_CLUSTER_MEMBERSHIP_HALT: Final[ThirtyTwoBits] = cached_win32error(0x1704, 'ERROR_CLUSTER_MEMBERSHIP_HALT') # WinError 5892 (0x1704)
    ERROR_CLUSTER_INSTANCE_ID_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x1705, 'ERROR_CLUSTER_INSTANCE_ID_MISMATCH') # WinError 5893 (0x1705)
    ERROR_CLUSTER_NETWORK_NOT_FOUND_FOR_IP: Final[ThirtyTwoBits] = cached_win32error(0x1706, 'ERROR_CLUSTER_NETWORK_NOT_FOUND_FOR_IP') # WinError 5894 (0x1706)
    ERROR_CLUSTER_PROPERTY_DATA_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x1707, 'ERROR_CLUSTER_PROPERTY_DATA_TYPE_MISMATCH') # WinError 5895 (0x1707)
    ERROR_CLUSTER_EVICT_WITHOUT_CLEANUP: Final[ThirtyTwoBits] = cached_win32error(0x1708, 'ERROR_CLUSTER_EVICT_WITHOUT_CLEANUP') # WinError 5896 (0x1708)
    ERROR_CLUSTER_PARAMETER_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x1709, 'ERROR_CLUSTER_PARAMETER_MISMATCH') # WinError 5897 (0x1709)
    ERROR_NODE_CANNOT_BE_CLUSTERED: Final[ThirtyTwoBits] = cached_win32error(0x170A, 'ERROR_NODE_CANNOT_BE_CLUSTERED') # WinError 5898 (0x170A)
    ERROR_CLUSTER_WRONG_OS_VERSION: Final[ThirtyTwoBits] = cached_win32error(0x170B, 'ERROR_CLUSTER_WRONG_OS_VERSION') # WinError 5899 (0x170B)
    ERROR_CLUSTER_CANT_CREATE_DUP_CLUSTER_NAME: Final[ThirtyTwoBits] = cached_win32error(0x170C, 'ERROR_CLUSTER_CANT_CREATE_DUP_CLUSTER_NAME') # WinError 5900 (0x170C)
    ERROR_CLUSCFG_ALREADY_COMMITTED: Final[ThirtyTwoBits] = cached_win32error(0x170D, 'ERROR_CLUSCFG_ALREADY_COMMITTED') # WinError 5901 (0x170D)
    ERROR_CLUSCFG_ROLLBACK_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x170E, 'ERROR_CLUSCFG_ROLLBACK_FAILED') # WinError 5902 (0x170E)
    ERROR_CLUSCFG_SYSTEM_DISK_DRIVE_LETTER_CONFLICT: Final[ThirtyTwoBits] = cached_win32error(0x170F, 'ERROR_CLUSCFG_SYSTEM_DISK_DRIVE_LETTER_CONFLICT') # WinError 5903 (0x170F)
    ERROR_CLUSTER_OLD_VERSION: Final[ThirtyTwoBits] = cached_win32error(0x1710, 'ERROR_CLUSTER_OLD_VERSION') # WinError 5904 (0x1710)
    ERROR_CLUSTER_MISMATCHED_COMPUTER_ACCT_NAME: Final[ThirtyTwoBits] = cached_win32error(0x1711, 'ERROR_CLUSTER_MISMATCHED_COMPUTER_ACCT_NAME') # WinError 5905 (0x1711)
    ERROR_CLUSTER_NO_NET_ADAPTERS: Final[ThirtyTwoBits] = cached_win32error(0x1712, 'ERROR_CLUSTER_NO_NET_ADAPTERS') # WinError 5906 (0x1712)
    ERROR_CLUSTER_POISONED: Final[ThirtyTwoBits] = cached_win32error(0x1713, 'ERROR_CLUSTER_POISONED') # WinError 5907 (0x1713)
    ERROR_CLUSTER_GROUP_MOVING: Final[ThirtyTwoBits] = cached_win32error(0x1714, 'ERROR_CLUSTER_GROUP_MOVING') # WinError 5908 (0x1714)
    ERROR_CLUSTER_RESOURCE_TYPE_BUSY: Final[ThirtyTwoBits] = cached_win32error(0x1715, 'ERROR_CLUSTER_RESOURCE_TYPE_BUSY') # WinError 5909 (0x1715)
    ERROR_RESOURCE_CALL_TIMED_OUT: Final[ThirtyTwoBits] = cached_win32error(0x1716, 'ERROR_RESOURCE_CALL_TIMED_OUT') # WinError 5910 (0x1716)
    ERROR_INVALID_CLUSTER_IPV6_ADDRESS: Final[ThirtyTwoBits] = cached_win32error(0x1717, 'ERROR_INVALID_CLUSTER_IPV6_ADDRESS') # WinError 5911 (0x1717)
    ERROR_CLUSTER_INTERNAL_INVALID_FUNCTION: Final[ThirtyTwoBits] = cached_win32error(0x1718, 'ERROR_CLUSTER_INTERNAL_INVALID_FUNCTION') # WinError 5912 (0x1718)
    ERROR_CLUSTER_PARAMETER_OUT_OF_BOUNDS: Final[ThirtyTwoBits] = cached_win32error(0x1719, 'ERROR_CLUSTER_PARAMETER_OUT_OF_BOUNDS') # WinError 5913 (0x1719)
    ERROR_CLUSTER_PARTIAL_SEND: Final[ThirtyTwoBits] = cached_win32error(0x171A, 'ERROR_CLUSTER_PARTIAL_SEND') # WinError 5914 (0x171A)
    ERROR_CLUSTER_REGISTRY_INVALID_FUNCTION: Final[ThirtyTwoBits] = cached_win32error(0x171B, 'ERROR_CLUSTER_REGISTRY_INVALID_FUNCTION') # WinError 5915 (0x171B)
    ERROR_CLUSTER_INVALID_STRING_TERMINATION: Final[ThirtyTwoBits] = cached_win32error(0x171C, 'ERROR_CLUSTER_INVALID_STRING_TERMINATION') # WinError 5916 (0x171C)
    ERROR_CLUSTER_INVALID_STRING_FORMAT: Final[ThirtyTwoBits] = cached_win32error(0x171D, 'ERROR_CLUSTER_INVALID_STRING_FORMAT') # WinError 5917 (0x171D)
    ERROR_CLUSTER_DATABASE_TRANSACTION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x171E, 'ERROR_CLUSTER_DATABASE_TRANSACTION_IN_PROGRESS') # WinError 5918 (0x171E)
    ERROR_CLUSTER_DATABASE_TRANSACTION_NOT_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x171F, 'ERROR_CLUSTER_DATABASE_TRANSACTION_NOT_IN_PROGRESS') # WinError 5919 (0x171F)
    ERROR_CLUSTER_NULL_DATA: Final[ThirtyTwoBits] = cached_win32error(0x1720, 'ERROR_CLUSTER_NULL_DATA') # WinError 5920 (0x1720)
    ERROR_CLUSTER_PARTIAL_READ: Final[ThirtyTwoBits] = cached_win32error(0x1721, 'ERROR_CLUSTER_PARTIAL_READ') # WinError 5921 (0x1721)
    ERROR_CLUSTER_PARTIAL_WRITE: Final[ThirtyTwoBits] = cached_win32error(0x1722, 'ERROR_CLUSTER_PARTIAL_WRITE') # WinError 5922 (0x1722)
    ERROR_CLUSTER_CANT_DESERIALIZE_DATA: Final[ThirtyTwoBits] = cached_win32error(0x1723, 'ERROR_CLUSTER_CANT_DESERIALIZE_DATA') # WinError 5923 (0x1723)
    ERROR_DEPENDENT_RESOURCE_PROPERTY_CONFLICT: Final[ThirtyTwoBits] = cached_win32error(0x1724, 'ERROR_DEPENDENT_RESOURCE_PROPERTY_CONFLICT') # WinError 5924 (0x1724)
    ERROR_CLUSTER_NO_QUORUM: Final[ThirtyTwoBits] = cached_win32error(0x1725, 'ERROR_CLUSTER_NO_QUORUM') # WinError 5925 (0x1725)
    ERROR_CLUSTER_INVALID_IPV6_NETWORK: Final[ThirtyTwoBits] = cached_win32error(0x1726, 'ERROR_CLUSTER_INVALID_IPV6_NETWORK') # WinError 5926 (0x1726)
    ERROR_CLUSTER_INVALID_IPV6_TUNNEL_NETWORK: Final[ThirtyTwoBits] = cached_win32error(0x1727, 'ERROR_CLUSTER_INVALID_IPV6_TUNNEL_NETWORK') # WinError 5927 (0x1727)
    ERROR_QUORUM_NOT_ALLOWED_IN_THIS_GROUP: Final[ThirtyTwoBits] = cached_win32error(0x1728, 'ERROR_QUORUM_NOT_ALLOWED_IN_THIS_GROUP') # WinError 5928 (0x1728)
    ERROR_DEPENDENCY_TREE_TOO_COMPLEX: Final[ThirtyTwoBits] = cached_win32error(0x1729, 'ERROR_DEPENDENCY_TREE_TOO_COMPLEX') # WinError 5929 (0x1729)
    ERROR_EXCEPTION_IN_RESOURCE_CALL: Final[ThirtyTwoBits] = cached_win32error(0x172A, 'ERROR_EXCEPTION_IN_RESOURCE_CALL') # WinError 5930 (0x172A)
    ERROR_CLUSTER_RHS_FAILED_INITIALIZATION: Final[ThirtyTwoBits] = cached_win32error(0x172B, 'ERROR_CLUSTER_RHS_FAILED_INITIALIZATION') # WinError 5931 (0x172B)
    ERROR_CLUSTER_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_win32error(0x172C, 'ERROR_CLUSTER_NOT_INSTALLED') # WinError 5932 (0x172C)
    ERROR_CLUSTER_RESOURCES_MUST_BE_ONLINE_ON_THE_SAME_NODE: Final[ThirtyTwoBits] = cached_win32error(0x172D, 'ERROR_CLUSTER_RESOURCES_MUST_BE_ONLINE_ON_THE_SAME_NODE') # WinError 5933 (0x172D)
    ERROR_CLUSTER_MAX_NODES_IN_CLUSTER: Final[ThirtyTwoBits] = cached_win32error(0x172E, 'ERROR_CLUSTER_MAX_NODES_IN_CLUSTER') # WinError 5934 (0x172E)
    ERROR_CLUSTER_TOO_MANY_NODES: Final[ThirtyTwoBits] = cached_win32error(0x172F, 'ERROR_CLUSTER_TOO_MANY_NODES') # WinError 5935 (0x172F)
    ERROR_CLUSTER_OBJECT_ALREADY_USED: Final[ThirtyTwoBits] = cached_win32error(0x1730, 'ERROR_CLUSTER_OBJECT_ALREADY_USED') # WinError 5936 (0x1730)
    ERROR_NONCORE_GROUPS_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x1731, 'ERROR_NONCORE_GROUPS_FOUND') # WinError 5937 (0x1731)
    ERROR_FILE_SHARE_RESOURCE_CONFLICT: Final[ThirtyTwoBits] = cached_win32error(0x1732, 'ERROR_FILE_SHARE_RESOURCE_CONFLICT') # WinError 5938 (0x1732)
    ERROR_CLUSTER_EVICT_INVALID_REQUEST: Final[ThirtyTwoBits] = cached_win32error(0x1733, 'ERROR_CLUSTER_EVICT_INVALID_REQUEST') # WinError 5939 (0x1733)
    ERROR_CLUSTER_SINGLETON_RESOURCE: Final[ThirtyTwoBits] = cached_win32error(0x1734, 'ERROR_CLUSTER_SINGLETON_RESOURCE') # WinError 5940 (0x1734)
    ERROR_CLUSTER_GROUP_SINGLETON_RESOURCE: Final[ThirtyTwoBits] = cached_win32error(0x1735, 'ERROR_CLUSTER_GROUP_SINGLETON_RESOURCE') # WinError 5941 (0x1735)
    ERROR_CLUSTER_RESOURCE_PROVIDER_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x1736, 'ERROR_CLUSTER_RESOURCE_PROVIDER_FAILED') # WinError 5942 (0x1736)
    ERROR_CLUSTER_RESOURCE_CONFIGURATION_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x1737, 'ERROR_CLUSTER_RESOURCE_CONFIGURATION_ERROR') # WinError 5943 (0x1737)
    ERROR_CLUSTER_GROUP_BUSY: Final[ThirtyTwoBits] = cached_win32error(0x1738, 'ERROR_CLUSTER_GROUP_BUSY') # WinError 5944 (0x1738)
    ERROR_CLUSTER_NOT_SHARED_VOLUME: Final[ThirtyTwoBits] = cached_win32error(0x1739, 'ERROR_CLUSTER_NOT_SHARED_VOLUME') # WinError 5945 (0x1739)
    ERROR_CLUSTER_INVALID_SECURITY_DESCRIPTOR: Final[ThirtyTwoBits] = cached_win32error(0x173A, 'ERROR_CLUSTER_INVALID_SECURITY_DESCRIPTOR') # WinError 5946 (0x173A)
    ERROR_CLUSTER_SHARED_VOLUMES_IN_USE: Final[ThirtyTwoBits] = cached_win32error(0x173B, 'ERROR_CLUSTER_SHARED_VOLUMES_IN_USE') # WinError 5947 (0x173B)
    ERROR_CLUSTER_USE_SHARED_VOLUMES_API: Final[ThirtyTwoBits] = cached_win32error(0x173C, 'ERROR_CLUSTER_USE_SHARED_VOLUMES_API') # WinError 5948 (0x173C)
    ERROR_CLUSTER_BACKUP_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x173D, 'ERROR_CLUSTER_BACKUP_IN_PROGRESS') # WinError 5949 (0x173D)
    ERROR_NON_CSV_PATH: Final[ThirtyTwoBits] = cached_win32error(0x173E, 'ERROR_NON_CSV_PATH') # WinError 5950 (0x173E)
    ERROR_CSV_VOLUME_NOT_LOCAL: Final[ThirtyTwoBits] = cached_win32error(0x173F, 'ERROR_CSV_VOLUME_NOT_LOCAL') # WinError 5951 (0x173F)
    ERROR_CLUSTER_WATCHDOG_TERMINATING: Final[ThirtyTwoBits] = cached_win32error(0x1740, 'ERROR_CLUSTER_WATCHDOG_TERMINATING') # WinError 5952 (0x1740)
    ERROR_CLUSTER_RESOURCE_VETOED_MOVE_INCOMPATIBLE_NODES: Final[ThirtyTwoBits] = cached_win32error(0x1741, 'ERROR_CLUSTER_RESOURCE_VETOED_MOVE_INCOMPATIBLE_NODES') # WinError 5953 (0x1741)
    ERROR_CLUSTER_INVALID_NODE_WEIGHT: Final[ThirtyTwoBits] = cached_win32error(0x1742, 'ERROR_CLUSTER_INVALID_NODE_WEIGHT') # WinError 5954 (0x1742)
    ERROR_CLUSTER_RESOURCE_VETOED_CALL: Final[ThirtyTwoBits] = cached_win32error(0x1743, 'ERROR_CLUSTER_RESOURCE_VETOED_CALL') # WinError 5955 (0x1743)
    ERROR_RESMON_SYSTEM_RESOURCES_LACKING: Final[ThirtyTwoBits] = cached_win32error(0x1744, 'ERROR_RESMON_SYSTEM_RESOURCES_LACKING') # WinError 5956 (0x1744)
    ERROR_CLUSTER_RESOURCE_VETOED_MOVE_NOT_ENOUGH_RESOURCES_ON_DESTINATION: Final[ThirtyTwoBits] = cached_win32error(0x1745, 'ERROR_CLUSTER_RESOURCE_VETOED_MOVE_NOT_ENOUGH_RESOURCES_ON_DESTINATION') # WinError 5957 (0x1745)
    ERROR_CLUSTER_RESOURCE_VETOED_MOVE_NOT_ENOUGH_RESOURCES_ON_SOURCE: Final[ThirtyTwoBits] = cached_win32error(0x1746, 'ERROR_CLUSTER_RESOURCE_VETOED_MOVE_NOT_ENOUGH_RESOURCES_ON_SOURCE') # WinError 5958 (0x1746)
    ERROR_CLUSTER_GROUP_QUEUED: Final[ThirtyTwoBits] = cached_win32error(0x1747, 'ERROR_CLUSTER_GROUP_QUEUED') # WinError 5959 (0x1747)
    ERROR_CLUSTER_RESOURCE_LOCKED_STATUS: Final[ThirtyTwoBits] = cached_win32error(0x1748, 'ERROR_CLUSTER_RESOURCE_LOCKED_STATUS') # WinError 5960 (0x1748)
    ERROR_CLUSTER_SHARED_VOLUME_FAILOVER_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x1749, 'ERROR_CLUSTER_SHARED_VOLUME_FAILOVER_NOT_ALLOWED') # WinError 5961 (0x1749)
    ERROR_CLUSTER_NODE_DRAIN_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x174A, 'ERROR_CLUSTER_NODE_DRAIN_IN_PROGRESS') # WinError 5962 (0x174A)
    ERROR_CLUSTER_DISK_NOT_CONNECTED: Final[ThirtyTwoBits] = cached_win32error(0x174B, 'ERROR_CLUSTER_DISK_NOT_CONNECTED') # WinError 5963 (0x174B)
    ERROR_DISK_NOT_CSV_CAPABLE: Final[ThirtyTwoBits] = cached_win32error(0x174C, 'ERROR_DISK_NOT_CSV_CAPABLE') # WinError 5964 (0x174C)
    ERROR_RESOURCE_NOT_IN_AVAILABLE_STORAGE: Final[ThirtyTwoBits] = cached_win32error(0x174D, 'ERROR_RESOURCE_NOT_IN_AVAILABLE_STORAGE') # WinError 5965 (0x174D)
    ERROR_CLUSTER_SHARED_VOLUME_REDIRECTED: Final[ThirtyTwoBits] = cached_win32error(0x174E, 'ERROR_CLUSTER_SHARED_VOLUME_REDIRECTED') # WinError 5966 (0x174E)
    ERROR_CLUSTER_SHARED_VOLUME_NOT_REDIRECTED: Final[ThirtyTwoBits] = cached_win32error(0x174F, 'ERROR_CLUSTER_SHARED_VOLUME_NOT_REDIRECTED') # WinError 5967 (0x174F)
    ERROR_CLUSTER_CANNOT_RETURN_PROPERTIES: Final[ThirtyTwoBits] = cached_win32error(0x1750, 'ERROR_CLUSTER_CANNOT_RETURN_PROPERTIES') # WinError 5968 (0x1750)
    ERROR_CLUSTER_RESOURCE_CONTAINS_UNSUPPORTED_DIFF_AREA_FOR_SHARED_VOLUMES: Final[ThirtyTwoBits] = cached_win32error(0x1751, 'ERROR_CLUSTER_RESOURCE_CONTAINS_UNSUPPORTED_DIFF_AREA_FOR_SHARED_VOLUMES') # WinError 5969 (0x1751)
    ERROR_CLUSTER_RESOURCE_IS_IN_MAINTENANCE_MODE: Final[ThirtyTwoBits] = cached_win32error(0x1752, 'ERROR_CLUSTER_RESOURCE_IS_IN_MAINTENANCE_MODE') # WinError 5970 (0x1752)
    ERROR_CLUSTER_AFFINITY_CONFLICT: Final[ThirtyTwoBits] = cached_win32error(0x1753, 'ERROR_CLUSTER_AFFINITY_CONFLICT') # WinError 5971 (0x1753)
    ERROR_CLUSTER_RESOURCE_IS_REPLICA_VIRTUAL_MACHINE: Final[ThirtyTwoBits] = cached_win32error(0x1754, 'ERROR_CLUSTER_RESOURCE_IS_REPLICA_VIRTUAL_MACHINE') # WinError 5972 (0x1754)
    ERROR_CLUSTER_UPGRADE_INCOMPATIBLE_VERSIONS: Final[ThirtyTwoBits] = cached_win32error(0x1755, 'ERROR_CLUSTER_UPGRADE_INCOMPATIBLE_VERSIONS') # WinError 5973 (0x1755)
    ERROR_CLUSTER_UPGRADE_FIX_QUORUM_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x1756, 'ERROR_CLUSTER_UPGRADE_FIX_QUORUM_NOT_SUPPORTED') # WinError 5974 (0x1756)
    ERROR_CLUSTER_UPGRADE_RESTART_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x1757, 'ERROR_CLUSTER_UPGRADE_RESTART_REQUIRED') # WinError 5975 (0x1757)
    ERROR_CLUSTER_UPGRADE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x1758, 'ERROR_CLUSTER_UPGRADE_IN_PROGRESS') # WinError 5976 (0x1758)
    ERROR_CLUSTER_UPGRADE_INCOMPLETE: Final[ThirtyTwoBits] = cached_win32error(0x1759, 'ERROR_CLUSTER_UPGRADE_INCOMPLETE') # WinError 5977 (0x1759)
    ERROR_CLUSTER_NODE_IN_GRACE_PERIOD: Final[ThirtyTwoBits] = cached_win32error(0x175A, 'ERROR_CLUSTER_NODE_IN_GRACE_PERIOD') # WinError 5978 (0x175A)
    ERROR_CLUSTER_CSV_IO_PAUSE_TIMEOUT: Final[ThirtyTwoBits] = cached_win32error(0x175B, 'ERROR_CLUSTER_CSV_IO_PAUSE_TIMEOUT') # WinError 5979 (0x175B)
    ERROR_NODE_NOT_ACTIVE_CLUSTER_MEMBER: Final[ThirtyTwoBits] = cached_win32error(0x175C, 'ERROR_NODE_NOT_ACTIVE_CLUSTER_MEMBER') # WinError 5980 (0x175C)
    ERROR_CLUSTER_RESOURCE_NOT_MONITORED: Final[ThirtyTwoBits] = cached_win32error(0x175D, 'ERROR_CLUSTER_RESOURCE_NOT_MONITORED') # WinError 5981 (0x175D)
    ERROR_CLUSTER_RESOURCE_DOES_NOT_SUPPORT_UNMONITORED: Final[ThirtyTwoBits] = cached_win32error(0x175E, 'ERROR_CLUSTER_RESOURCE_DOES_NOT_SUPPORT_UNMONITORED') # WinError 5982 (0x175E)
    ERROR_CLUSTER_RESOURCE_IS_REPLICATED: Final[ThirtyTwoBits] = cached_win32error(0x175F, 'ERROR_CLUSTER_RESOURCE_IS_REPLICATED') # WinError 5983 (0x175F)
    ERROR_CLUSTER_NODE_ISOLATED: Final[ThirtyTwoBits] = cached_win32error(0x1760, 'ERROR_CLUSTER_NODE_ISOLATED') # WinError 5984 (0x1760)
    ERROR_CLUSTER_NODE_QUARANTINED: Final[ThirtyTwoBits] = cached_win32error(0x1761, 'ERROR_CLUSTER_NODE_QUARANTINED') # WinError 5985 (0x1761)
    ERROR_CLUSTER_DATABASE_UPDATE_CONDITION_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x1762, 'ERROR_CLUSTER_DATABASE_UPDATE_CONDITION_FAILED') # WinError 5986 (0x1762)
    ERROR_CLUSTER_SPACE_DEGRADED: Final[ThirtyTwoBits] = cached_win32error(0x1763, 'ERROR_CLUSTER_SPACE_DEGRADED') # WinError 5987 (0x1763)
    ERROR_CLUSTER_TOKEN_DELEGATION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x1764, 'ERROR_CLUSTER_TOKEN_DELEGATION_NOT_SUPPORTED') # WinError 5988 (0x1764)
    ERROR_CLUSTER_CSV_INVALID_HANDLE: Final[ThirtyTwoBits] = cached_win32error(0x1765, 'ERROR_CLUSTER_CSV_INVALID_HANDLE') # WinError 5989 (0x1765)
    ERROR_CLUSTER_CSV_SUPPORTED_ONLY_ON_COORDINATOR: Final[ThirtyTwoBits] = cached_win32error(0x1766, 'ERROR_CLUSTER_CSV_SUPPORTED_ONLY_ON_COORDINATOR') # WinError 5990 (0x1766)
    ERROR_GROUPSET_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x1767, 'ERROR_GROUPSET_NOT_AVAILABLE') # WinError 5991 (0x1767)
    ERROR_GROUPSET_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x1768, 'ERROR_GROUPSET_NOT_FOUND') # WinError 5992 (0x1768)
    ERROR_GROUPSET_CANT_PROVIDE: Final[ThirtyTwoBits] = cached_win32error(0x1769, 'ERROR_GROUPSET_CANT_PROVIDE') # WinError 5993 (0x1769)
    ERROR_CLUSTER_FAULT_DOMAIN_PARENT_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x176A, 'ERROR_CLUSTER_FAULT_DOMAIN_PARENT_NOT_FOUND') # WinError 5994 (0x176A)
    ERROR_CLUSTER_FAULT_DOMAIN_INVALID_HIERARCHY: Final[ThirtyTwoBits] = cached_win32error(0x176B, 'ERROR_CLUSTER_FAULT_DOMAIN_INVALID_HIERARCHY') # WinError 5995 (0x176B)
    ERROR_CLUSTER_FAULT_DOMAIN_FAILED_S2D_VALIDATION: Final[ThirtyTwoBits] = cached_win32error(0x176C, 'ERROR_CLUSTER_FAULT_DOMAIN_FAILED_S2D_VALIDATION') # WinError 5996 (0x176C)
    ERROR_CLUSTER_FAULT_DOMAIN_S2D_CONNECTIVITY_LOSS: Final[ThirtyTwoBits] = cached_win32error(0x176D, 'ERROR_CLUSTER_FAULT_DOMAIN_S2D_CONNECTIVITY_LOSS') # WinError 5997 (0x176D)
    ERROR_CLUSTER_INVALID_INFRASTRUCTURE_FILESERVER_NAME: Final[ThirtyTwoBits] = cached_win32error(0x176E, 'ERROR_CLUSTER_INVALID_INFRASTRUCTURE_FILESERVER_NAME') # WinError 5998 (0x176E)
    ERROR_CLUSTERSET_MANAGEMENT_CLUSTER_UNREACHABLE: Final[ThirtyTwoBits] = cached_win32error(0x176F, 'ERROR_CLUSTERSET_MANAGEMENT_CLUSTER_UNREACHABLE') # WinError 5999 (0x176F)
    ERROR_ENCRYPTION_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x1770, 'ERROR_ENCRYPTION_FAILED') # WinError 6000 (0x1770)
    ERROR_DECRYPTION_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x1771, 'ERROR_DECRYPTION_FAILED') # WinError 6001 (0x1771)
    ERROR_FILE_ENCRYPTED: Final[ThirtyTwoBits] = cached_win32error(0x1772, 'ERROR_FILE_ENCRYPTED') # WinError 6002 (0x1772)
    ERROR_NO_RECOVERY_POLICY: Final[ThirtyTwoBits] = cached_win32error(0x1773, 'ERROR_NO_RECOVERY_POLICY') # WinError 6003 (0x1773)
    ERROR_NO_EFS: Final[ThirtyTwoBits] = cached_win32error(0x1774, 'ERROR_NO_EFS') # WinError 6004 (0x1774)
    ERROR_WRONG_EFS: Final[ThirtyTwoBits] = cached_win32error(0x1775, 'ERROR_WRONG_EFS') # WinError 6005 (0x1775)
    ERROR_NO_USER_KEYS: Final[ThirtyTwoBits] = cached_win32error(0x1776, 'ERROR_NO_USER_KEYS') # WinError 6006 (0x1776)
    ERROR_FILE_NOT_ENCRYPTED: Final[ThirtyTwoBits] = cached_win32error(0x1777, 'ERROR_FILE_NOT_ENCRYPTED') # WinError 6007 (0x1777)
    ERROR_NOT_EXPORT_FORMAT: Final[ThirtyTwoBits] = cached_win32error(0x1778, 'ERROR_NOT_EXPORT_FORMAT') # WinError 6008 (0x1778)
    ERROR_FILE_READ_ONLY: Final[ThirtyTwoBits] = cached_win32error(0x1779, 'ERROR_FILE_READ_ONLY') # WinError 6009 (0x1779)
    ERROR_DIR_EFS_DISALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x177A, 'ERROR_DIR_EFS_DISALLOWED') # WinError 6010 (0x177A)
    ERROR_EFS_SERVER_NOT_TRUSTED: Final[ThirtyTwoBits] = cached_win32error(0x177B, 'ERROR_EFS_SERVER_NOT_TRUSTED') # WinError 6011 (0x177B)
    ERROR_BAD_RECOVERY_POLICY: Final[ThirtyTwoBits] = cached_win32error(0x177C, 'ERROR_BAD_RECOVERY_POLICY') # WinError 6012 (0x177C)
    ERROR_EFS_ALG_BLOB_TOO_BIG: Final[ThirtyTwoBits] = cached_win32error(0x177D, 'ERROR_EFS_ALG_BLOB_TOO_BIG') # WinError 6013 (0x177D)
    ERROR_VOLUME_NOT_SUPPORT_EFS: Final[ThirtyTwoBits] = cached_win32error(0x177E, 'ERROR_VOLUME_NOT_SUPPORT_EFS') # WinError 6014 (0x177E)
    ERROR_EFS_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x177F, 'ERROR_EFS_DISABLED') # WinError 6015 (0x177F)
    ERROR_EFS_VERSION_NOT_SUPPORT: Final[ThirtyTwoBits] = cached_win32error(0x1780, 'ERROR_EFS_VERSION_NOT_SUPPORT') # WinError 6016 (0x1780)
    ERROR_CS_ENCRYPTION_INVALID_SERVER_RESPONSE: Final[ThirtyTwoBits] = cached_win32error(0x1781, 'ERROR_CS_ENCRYPTION_INVALID_SERVER_RESPONSE') # WinError 6017 (0x1781)
    ERROR_CS_ENCRYPTION_UNSUPPORTED_SERVER: Final[ThirtyTwoBits] = cached_win32error(0x1782, 'ERROR_CS_ENCRYPTION_UNSUPPORTED_SERVER') # WinError 6018 (0x1782)
    ERROR_CS_ENCRYPTION_EXISTING_ENCRYPTED_FILE: Final[ThirtyTwoBits] = cached_win32error(0x1783, 'ERROR_CS_ENCRYPTION_EXISTING_ENCRYPTED_FILE') # WinError 6019 (0x1783)
    ERROR_CS_ENCRYPTION_NEW_ENCRYPTED_FILE: Final[ThirtyTwoBits] = cached_win32error(0x1784, 'ERROR_CS_ENCRYPTION_NEW_ENCRYPTED_FILE') # WinError 6020 (0x1784)
    ERROR_CS_ENCRYPTION_FILE_NOT_CSE: Final[ThirtyTwoBits] = cached_win32error(0x1785, 'ERROR_CS_ENCRYPTION_FILE_NOT_CSE') # WinError 6021 (0x1785)
    ERROR_ENCRYPTION_POLICY_DENIES_OPERATION: Final[ThirtyTwoBits] = cached_win32error(0x1786, 'ERROR_ENCRYPTION_POLICY_DENIES_OPERATION') # WinError 6022 (0x1786)
    ERROR_WIP_ENCRYPTION_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x1787, 'ERROR_WIP_ENCRYPTION_FAILED') # WinError 6023 (0x1787)
    ERROR_NO_BROWSER_SERVERS_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x17E6, 'ERROR_NO_BROWSER_SERVERS_FOUND') # WinError 6118 (0x17E6)
    SCHED_E_SERVICE_NOT_LOCALSYSTEM: Final[ThirtyTwoBits] = cached_win32error(0x1838, 'SCHED_E_SERVICE_NOT_LOCALSYSTEM') # WinError 6200 (0x1838)
    ERROR_CLUSTER_OBJECT_IS_CLUSTER_SET_VM: Final[ThirtyTwoBits] = cached_win32error(0x186A, 'ERROR_CLUSTER_OBJECT_IS_CLUSTER_SET_VM') # WinError 6250 (0x186A)
    ERROR_LOG_SECTOR_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x19C8, 'ERROR_LOG_SECTOR_INVALID') # WinError 6600 (0x19C8)
    ERROR_LOG_SECTOR_PARITY_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x19C9, 'ERROR_LOG_SECTOR_PARITY_INVALID') # WinError 6601 (0x19C9)
    ERROR_LOG_SECTOR_REMAPPED: Final[ThirtyTwoBits] = cached_win32error(0x19CA, 'ERROR_LOG_SECTOR_REMAPPED') # WinError 6602 (0x19CA)
    ERROR_LOG_BLOCK_INCOMPLETE: Final[ThirtyTwoBits] = cached_win32error(0x19CB, 'ERROR_LOG_BLOCK_INCOMPLETE') # WinError 6603 (0x19CB)
    ERROR_LOG_INVALID_RANGE: Final[ThirtyTwoBits] = cached_win32error(0x19CC, 'ERROR_LOG_INVALID_RANGE') # WinError 6604 (0x19CC)
    ERROR_LOG_BLOCKS_EXHAUSTED: Final[ThirtyTwoBits] = cached_win32error(0x19CD, 'ERROR_LOG_BLOCKS_EXHAUSTED') # WinError 6605 (0x19CD)
    ERROR_LOG_READ_CONTEXT_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x19CE, 'ERROR_LOG_READ_CONTEXT_INVALID') # WinError 6606 (0x19CE)
    ERROR_LOG_RESTART_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x19CF, 'ERROR_LOG_RESTART_INVALID') # WinError 6607 (0x19CF)
    ERROR_LOG_BLOCK_VERSION: Final[ThirtyTwoBits] = cached_win32error(0x19D0, 'ERROR_LOG_BLOCK_VERSION') # WinError 6608 (0x19D0)
    ERROR_LOG_BLOCK_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x19D1, 'ERROR_LOG_BLOCK_INVALID') # WinError 6609 (0x19D1)
    ERROR_LOG_READ_MODE_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x19D2, 'ERROR_LOG_READ_MODE_INVALID') # WinError 6610 (0x19D2)
    ERROR_LOG_NO_RESTART: Final[ThirtyTwoBits] = cached_win32error(0x19D3, 'ERROR_LOG_NO_RESTART') # WinError 6611 (0x19D3)
    ERROR_LOG_METADATA_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x19D4, 'ERROR_LOG_METADATA_CORRUPT') # WinError 6612 (0x19D4)
    ERROR_LOG_METADATA_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x19D5, 'ERROR_LOG_METADATA_INVALID') # WinError 6613 (0x19D5)
    ERROR_LOG_METADATA_INCONSISTENT: Final[ThirtyTwoBits] = cached_win32error(0x19D6, 'ERROR_LOG_METADATA_INCONSISTENT') # WinError 6614 (0x19D6)
    ERROR_LOG_RESERVATION_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x19D7, 'ERROR_LOG_RESERVATION_INVALID') # WinError 6615 (0x19D7)
    ERROR_LOG_CANT_DELETE: Final[ThirtyTwoBits] = cached_win32error(0x19D8, 'ERROR_LOG_CANT_DELETE') # WinError 6616 (0x19D8)
    ERROR_LOG_CONTAINER_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x19D9, 'ERROR_LOG_CONTAINER_LIMIT_EXCEEDED') # WinError 6617 (0x19D9)
    ERROR_LOG_START_OF_LOG: Final[ThirtyTwoBits] = cached_win32error(0x19DA, 'ERROR_LOG_START_OF_LOG') # WinError 6618 (0x19DA)
    ERROR_LOG_POLICY_ALREADY_INSTALLED: Final[ThirtyTwoBits] = cached_win32error(0x19DB, 'ERROR_LOG_POLICY_ALREADY_INSTALLED') # WinError 6619 (0x19DB)
    ERROR_LOG_POLICY_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_win32error(0x19DC, 'ERROR_LOG_POLICY_NOT_INSTALLED') # WinError 6620 (0x19DC)
    ERROR_LOG_POLICY_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x19DD, 'ERROR_LOG_POLICY_INVALID') # WinError 6621 (0x19DD)
    ERROR_LOG_POLICY_CONFLICT: Final[ThirtyTwoBits] = cached_win32error(0x19DE, 'ERROR_LOG_POLICY_CONFLICT') # WinError 6622 (0x19DE)
    ERROR_LOG_PINNED_ARCHIVE_TAIL: Final[ThirtyTwoBits] = cached_win32error(0x19DF, 'ERROR_LOG_PINNED_ARCHIVE_TAIL') # WinError 6623 (0x19DF)
    ERROR_LOG_RECORD_NONEXISTENT: Final[ThirtyTwoBits] = cached_win32error(0x19E0, 'ERROR_LOG_RECORD_NONEXISTENT') # WinError 6624 (0x19E0)
    ERROR_LOG_RECORDS_RESERVED_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x19E1, 'ERROR_LOG_RECORDS_RESERVED_INVALID') # WinError 6625 (0x19E1)
    ERROR_LOG_SPACE_RESERVED_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x19E2, 'ERROR_LOG_SPACE_RESERVED_INVALID') # WinError 6626 (0x19E2)
    ERROR_LOG_TAIL_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x19E3, 'ERROR_LOG_TAIL_INVALID') # WinError 6627 (0x19E3)
    ERROR_LOG_FULL: Final[ThirtyTwoBits] = cached_win32error(0x19E4, 'ERROR_LOG_FULL') # WinError 6628 (0x19E4)
    ERROR_COULD_NOT_RESIZE_LOG: Final[ThirtyTwoBits] = cached_win32error(0x19E5, 'ERROR_COULD_NOT_RESIZE_LOG') # WinError 6629 (0x19E5)
    ERROR_LOG_MULTIPLEXED: Final[ThirtyTwoBits] = cached_win32error(0x19E6, 'ERROR_LOG_MULTIPLEXED') # WinError 6630 (0x19E6)
    ERROR_LOG_DEDICATED: Final[ThirtyTwoBits] = cached_win32error(0x19E7, 'ERROR_LOG_DEDICATED') # WinError 6631 (0x19E7)
    ERROR_LOG_ARCHIVE_NOT_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x19E8, 'ERROR_LOG_ARCHIVE_NOT_IN_PROGRESS') # WinError 6632 (0x19E8)
    ERROR_LOG_ARCHIVE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x19E9, 'ERROR_LOG_ARCHIVE_IN_PROGRESS') # WinError 6633 (0x19E9)
    ERROR_LOG_EPHEMERAL: Final[ThirtyTwoBits] = cached_win32error(0x19EA, 'ERROR_LOG_EPHEMERAL') # WinError 6634 (0x19EA)
    ERROR_LOG_NOT_ENOUGH_CONTAINERS: Final[ThirtyTwoBits] = cached_win32error(0x19EB, 'ERROR_LOG_NOT_ENOUGH_CONTAINERS') # WinError 6635 (0x19EB)
    ERROR_LOG_CLIENT_ALREADY_REGISTERED: Final[ThirtyTwoBits] = cached_win32error(0x19EC, 'ERROR_LOG_CLIENT_ALREADY_REGISTERED') # WinError 6636 (0x19EC)
    ERROR_LOG_CLIENT_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_win32error(0x19ED, 'ERROR_LOG_CLIENT_NOT_REGISTERED') # WinError 6637 (0x19ED)
    ERROR_LOG_FULL_HANDLER_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x19EE, 'ERROR_LOG_FULL_HANDLER_IN_PROGRESS') # WinError 6638 (0x19EE)
    ERROR_LOG_CONTAINER_READ_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x19EF, 'ERROR_LOG_CONTAINER_READ_FAILED') # WinError 6639 (0x19EF)
    ERROR_LOG_CONTAINER_WRITE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x19F0, 'ERROR_LOG_CONTAINER_WRITE_FAILED') # WinError 6640 (0x19F0)
    ERROR_LOG_CONTAINER_OPEN_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x19F1, 'ERROR_LOG_CONTAINER_OPEN_FAILED') # WinError 6641 (0x19F1)
    ERROR_LOG_CONTAINER_STATE_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x19F2, 'ERROR_LOG_CONTAINER_STATE_INVALID') # WinError 6642 (0x19F2)
    ERROR_LOG_STATE_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x19F3, 'ERROR_LOG_STATE_INVALID') # WinError 6643 (0x19F3)
    ERROR_LOG_PINNED: Final[ThirtyTwoBits] = cached_win32error(0x19F4, 'ERROR_LOG_PINNED') # WinError 6644 (0x19F4)
    ERROR_LOG_METADATA_FLUSH_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x19F5, 'ERROR_LOG_METADATA_FLUSH_FAILED') # WinError 6645 (0x19F5)
    ERROR_LOG_INCONSISTENT_SECURITY: Final[ThirtyTwoBits] = cached_win32error(0x19F6, 'ERROR_LOG_INCONSISTENT_SECURITY') # WinError 6646 (0x19F6)
    ERROR_LOG_APPENDED_FLUSH_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x19F7, 'ERROR_LOG_APPENDED_FLUSH_FAILED') # WinError 6647 (0x19F7)
    ERROR_LOG_PINNED_RESERVATION: Final[ThirtyTwoBits] = cached_win32error(0x19F8, 'ERROR_LOG_PINNED_RESERVATION') # WinError 6648 (0x19F8)
    ERROR_INVALID_TRANSACTION: Final[ThirtyTwoBits] = cached_win32error(0x1A2C, 'ERROR_INVALID_TRANSACTION') # WinError 6700 (0x1A2C)
    ERROR_TRANSACTION_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_win32error(0x1A2D, 'ERROR_TRANSACTION_NOT_ACTIVE') # WinError 6701 (0x1A2D)
    ERROR_TRANSACTION_REQUEST_NOT_VALID: Final[ThirtyTwoBits] = cached_win32error(0x1A2E, 'ERROR_TRANSACTION_REQUEST_NOT_VALID') # WinError 6702 (0x1A2E)
    ERROR_TRANSACTION_NOT_REQUESTED: Final[ThirtyTwoBits] = cached_win32error(0x1A2F, 'ERROR_TRANSACTION_NOT_REQUESTED') # WinError 6703 (0x1A2F)
    ERROR_TRANSACTION_ALREADY_ABORTED: Final[ThirtyTwoBits] = cached_win32error(0x1A30, 'ERROR_TRANSACTION_ALREADY_ABORTED') # WinError 6704 (0x1A30)
    ERROR_TRANSACTION_ALREADY_COMMITTED: Final[ThirtyTwoBits] = cached_win32error(0x1A31, 'ERROR_TRANSACTION_ALREADY_COMMITTED') # WinError 6705 (0x1A31)
    ERROR_TM_INITIALIZATION_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x1A32, 'ERROR_TM_INITIALIZATION_FAILED') # WinError 6706 (0x1A32)
    ERROR_RESOURCEMANAGER_READ_ONLY: Final[ThirtyTwoBits] = cached_win32error(0x1A33, 'ERROR_RESOURCEMANAGER_READ_ONLY') # WinError 6707 (0x1A33)
    ERROR_TRANSACTION_NOT_JOINED: Final[ThirtyTwoBits] = cached_win32error(0x1A34, 'ERROR_TRANSACTION_NOT_JOINED') # WinError 6708 (0x1A34)
    ERROR_TRANSACTION_SUPERIOR_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x1A35, 'ERROR_TRANSACTION_SUPERIOR_EXISTS') # WinError 6709 (0x1A35)
    ERROR_CRM_PROTOCOL_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x1A36, 'ERROR_CRM_PROTOCOL_ALREADY_EXISTS') # WinError 6710 (0x1A36)
    ERROR_TRANSACTION_PROPAGATION_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x1A37, 'ERROR_TRANSACTION_PROPAGATION_FAILED') # WinError 6711 (0x1A37)
    ERROR_CRM_PROTOCOL_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x1A38, 'ERROR_CRM_PROTOCOL_NOT_FOUND') # WinError 6712 (0x1A38)
    ERROR_TRANSACTION_INVALID_MARSHALL_BUFFER: Final[ThirtyTwoBits] = cached_win32error(0x1A39, 'ERROR_TRANSACTION_INVALID_MARSHALL_BUFFER') # WinError 6713 (0x1A39)
    ERROR_CURRENT_TRANSACTION_NOT_VALID: Final[ThirtyTwoBits] = cached_win32error(0x1A3A, 'ERROR_CURRENT_TRANSACTION_NOT_VALID') # WinError 6714 (0x1A3A)
    ERROR_TRANSACTION_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x1A3B, 'ERROR_TRANSACTION_NOT_FOUND') # WinError 6715 (0x1A3B)
    ERROR_RESOURCEMANAGER_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x1A3C, 'ERROR_RESOURCEMANAGER_NOT_FOUND') # WinError 6716 (0x1A3C)
    ERROR_ENLISTMENT_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x1A3D, 'ERROR_ENLISTMENT_NOT_FOUND') # WinError 6717 (0x1A3D)
    ERROR_TRANSACTIONMANAGER_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x1A3E, 'ERROR_TRANSACTIONMANAGER_NOT_FOUND') # WinError 6718 (0x1A3E)
    ERROR_TRANSACTIONMANAGER_NOT_ONLINE: Final[ThirtyTwoBits] = cached_win32error(0x1A3F, 'ERROR_TRANSACTIONMANAGER_NOT_ONLINE') # WinError 6719 (0x1A3F)
    ERROR_TRANSACTIONMANAGER_RECOVERY_NAME_COLLISION: Final[ThirtyTwoBits] = cached_win32error(0x1A40, 'ERROR_TRANSACTIONMANAGER_RECOVERY_NAME_COLLISION') # WinError 6720 (0x1A40)
    ERROR_TRANSACTION_NOT_ROOT: Final[ThirtyTwoBits] = cached_win32error(0x1A41, 'ERROR_TRANSACTION_NOT_ROOT') # WinError 6721 (0x1A41)
    ERROR_TRANSACTION_OBJECT_EXPIRED: Final[ThirtyTwoBits] = cached_win32error(0x1A42, 'ERROR_TRANSACTION_OBJECT_EXPIRED') # WinError 6722 (0x1A42)
    ERROR_TRANSACTION_RESPONSE_NOT_ENLISTED: Final[ThirtyTwoBits] = cached_win32error(0x1A43, 'ERROR_TRANSACTION_RESPONSE_NOT_ENLISTED') # WinError 6723 (0x1A43)
    ERROR_TRANSACTION_RECORD_TOO_LONG: Final[ThirtyTwoBits] = cached_win32error(0x1A44, 'ERROR_TRANSACTION_RECORD_TOO_LONG') # WinError 6724 (0x1A44)
    ERROR_IMPLICIT_TRANSACTION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x1A45, 'ERROR_IMPLICIT_TRANSACTION_NOT_SUPPORTED') # WinError 6725 (0x1A45)
    ERROR_TRANSACTION_INTEGRITY_VIOLATED: Final[ThirtyTwoBits] = cached_win32error(0x1A46, 'ERROR_TRANSACTION_INTEGRITY_VIOLATED') # WinError 6726 (0x1A46)
    ERROR_TRANSACTIONMANAGER_IDENTITY_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x1A47, 'ERROR_TRANSACTIONMANAGER_IDENTITY_MISMATCH') # WinError 6727 (0x1A47)
    ERROR_RM_CANNOT_BE_FROZEN_FOR_SNAPSHOT: Final[ThirtyTwoBits] = cached_win32error(0x1A48, 'ERROR_RM_CANNOT_BE_FROZEN_FOR_SNAPSHOT') # WinError 6728 (0x1A48)
    ERROR_TRANSACTION_MUST_WRITETHROUGH: Final[ThirtyTwoBits] = cached_win32error(0x1A49, 'ERROR_TRANSACTION_MUST_WRITETHROUGH') # WinError 6729 (0x1A49)
    ERROR_TRANSACTION_NO_SUPERIOR: Final[ThirtyTwoBits] = cached_win32error(0x1A4A, 'ERROR_TRANSACTION_NO_SUPERIOR') # WinError 6730 (0x1A4A)
    ERROR_HEURISTIC_DAMAGE_POSSIBLE: Final[ThirtyTwoBits] = cached_win32error(0x1A4B, 'ERROR_HEURISTIC_DAMAGE_POSSIBLE') # WinError 6731 (0x1A4B)
    ERROR_TRANSACTIONAL_CONFLICT: Final[ThirtyTwoBits] = cached_win32error(0x1A90, 'ERROR_TRANSACTIONAL_CONFLICT') # WinError 6800 (0x1A90)
    ERROR_RM_NOT_ACTIVE: Final[ThirtyTwoBits] = cached_win32error(0x1A91, 'ERROR_RM_NOT_ACTIVE') # WinError 6801 (0x1A91)
    ERROR_RM_METADATA_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x1A92, 'ERROR_RM_METADATA_CORRUPT') # WinError 6802 (0x1A92)
    ERROR_DIRECTORY_NOT_RM: Final[ThirtyTwoBits] = cached_win32error(0x1A93, 'ERROR_DIRECTORY_NOT_RM') # WinError 6803 (0x1A93)
    ERROR_TRANSACTIONS_UNSUPPORTED_REMOTE: Final[ThirtyTwoBits] = cached_win32error(0x1A95, 'ERROR_TRANSACTIONS_UNSUPPORTED_REMOTE') # WinError 6805 (0x1A95)
    ERROR_LOG_RESIZE_INVALID_SIZE: Final[ThirtyTwoBits] = cached_win32error(0x1A96, 'ERROR_LOG_RESIZE_INVALID_SIZE') # WinError 6806 (0x1A96)
    ERROR_OBJECT_NO_LONGER_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x1A97, 'ERROR_OBJECT_NO_LONGER_EXISTS') # WinError 6807 (0x1A97)
    ERROR_STREAM_MINIVERSION_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x1A98, 'ERROR_STREAM_MINIVERSION_NOT_FOUND') # WinError 6808 (0x1A98)
    ERROR_STREAM_MINIVERSION_NOT_VALID: Final[ThirtyTwoBits] = cached_win32error(0x1A99, 'ERROR_STREAM_MINIVERSION_NOT_VALID') # WinError 6809 (0x1A99)
    ERROR_MINIVERSION_INACCESSIBLE_FROM_SPECIFIED_TRANSACTION: Final[ThirtyTwoBits] = cached_win32error(0x1A9A, 'ERROR_MINIVERSION_INACCESSIBLE_FROM_SPECIFIED_TRANSACTION') # WinError 6810 (0x1A9A)
    ERROR_CANT_OPEN_MINIVERSION_WITH_MODIFY_INTENT: Final[ThirtyTwoBits] = cached_win32error(0x1A9B, 'ERROR_CANT_OPEN_MINIVERSION_WITH_MODIFY_INTENT') # WinError 6811 (0x1A9B)
    ERROR_CANT_CREATE_MORE_STREAM_MINIVERSIONS: Final[ThirtyTwoBits] = cached_win32error(0x1A9C, 'ERROR_CANT_CREATE_MORE_STREAM_MINIVERSIONS') # WinError 6812 (0x1A9C)
    ERROR_REMOTE_FILE_VERSION_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x1A9E, 'ERROR_REMOTE_FILE_VERSION_MISMATCH') # WinError 6814 (0x1A9E)
    ERROR_HANDLE_NO_LONGER_VALID: Final[ThirtyTwoBits] = cached_win32error(0x1A9F, 'ERROR_HANDLE_NO_LONGER_VALID') # WinError 6815 (0x1A9F)
    ERROR_NO_TXF_METADATA: Final[ThirtyTwoBits] = cached_win32error(0x1AA0, 'ERROR_NO_TXF_METADATA') # WinError 6816 (0x1AA0)
    ERROR_LOG_CORRUPTION_DETECTED: Final[ThirtyTwoBits] = cached_win32error(0x1AA1, 'ERROR_LOG_CORRUPTION_DETECTED') # WinError 6817 (0x1AA1)
    ERROR_CANT_RECOVER_WITH_HANDLE_OPEN: Final[ThirtyTwoBits] = cached_win32error(0x1AA2, 'ERROR_CANT_RECOVER_WITH_HANDLE_OPEN') # WinError 6818 (0x1AA2)
    ERROR_RM_DISCONNECTED: Final[ThirtyTwoBits] = cached_win32error(0x1AA3, 'ERROR_RM_DISCONNECTED') # WinError 6819 (0x1AA3)
    ERROR_ENLISTMENT_NOT_SUPERIOR: Final[ThirtyTwoBits] = cached_win32error(0x1AA4, 'ERROR_ENLISTMENT_NOT_SUPERIOR') # WinError 6820 (0x1AA4)
    ERROR_RECOVERY_NOT_NEEDED: Final[ThirtyTwoBits] = cached_win32error(0x1AA5, 'ERROR_RECOVERY_NOT_NEEDED') # WinError 6821 (0x1AA5)
    ERROR_RM_ALREADY_STARTED: Final[ThirtyTwoBits] = cached_win32error(0x1AA6, 'ERROR_RM_ALREADY_STARTED') # WinError 6822 (0x1AA6)
    ERROR_FILE_IDENTITY_NOT_PERSISTENT: Final[ThirtyTwoBits] = cached_win32error(0x1AA7, 'ERROR_FILE_IDENTITY_NOT_PERSISTENT') # WinError 6823 (0x1AA7)
    ERROR_CANT_BREAK_TRANSACTIONAL_DEPENDENCY: Final[ThirtyTwoBits] = cached_win32error(0x1AA8, 'ERROR_CANT_BREAK_TRANSACTIONAL_DEPENDENCY') # WinError 6824 (0x1AA8)
    ERROR_CANT_CROSS_RM_BOUNDARY: Final[ThirtyTwoBits] = cached_win32error(0x1AA9, 'ERROR_CANT_CROSS_RM_BOUNDARY') # WinError 6825 (0x1AA9)
    ERROR_TXF_DIR_NOT_EMPTY: Final[ThirtyTwoBits] = cached_win32error(0x1AAA, 'ERROR_TXF_DIR_NOT_EMPTY') # WinError 6826 (0x1AAA)
    ERROR_INDOUBT_TRANSACTIONS_EXIST: Final[ThirtyTwoBits] = cached_win32error(0x1AAB, 'ERROR_INDOUBT_TRANSACTIONS_EXIST') # WinError 6827 (0x1AAB)
    ERROR_TM_VOLATILE: Final[ThirtyTwoBits] = cached_win32error(0x1AAC, 'ERROR_TM_VOLATILE') # WinError 6828 (0x1AAC)
    ERROR_ROLLBACK_TIMER_EXPIRED: Final[ThirtyTwoBits] = cached_win32error(0x1AAD, 'ERROR_ROLLBACK_TIMER_EXPIRED') # WinError 6829 (0x1AAD)
    ERROR_TXF_ATTRIBUTE_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x1AAE, 'ERROR_TXF_ATTRIBUTE_CORRUPT') # WinError 6830 (0x1AAE)
    ERROR_EFS_NOT_ALLOWED_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_win32error(0x1AAF, 'ERROR_EFS_NOT_ALLOWED_IN_TRANSACTION') # WinError 6831 (0x1AAF)
    ERROR_TRANSACTIONAL_OPEN_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x1AB0, 'ERROR_TRANSACTIONAL_OPEN_NOT_ALLOWED') # WinError 6832 (0x1AB0)
    ERROR_LOG_GROWTH_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x1AB1, 'ERROR_LOG_GROWTH_FAILED') # WinError 6833 (0x1AB1)
    ERROR_TRANSACTED_MAPPING_UNSUPPORTED_REMOTE: Final[ThirtyTwoBits] = cached_win32error(0x1AB2, 'ERROR_TRANSACTED_MAPPING_UNSUPPORTED_REMOTE') # WinError 6834 (0x1AB2)
    ERROR_TXF_METADATA_ALREADY_PRESENT: Final[ThirtyTwoBits] = cached_win32error(0x1AB3, 'ERROR_TXF_METADATA_ALREADY_PRESENT') # WinError 6835 (0x1AB3)
    ERROR_TRANSACTION_SCOPE_CALLBACKS_NOT_SET: Final[ThirtyTwoBits] = cached_win32error(0x1AB4, 'ERROR_TRANSACTION_SCOPE_CALLBACKS_NOT_SET') # WinError 6836 (0x1AB4)
    ERROR_TRANSACTION_REQUIRED_PROMOTION: Final[ThirtyTwoBits] = cached_win32error(0x1AB5, 'ERROR_TRANSACTION_REQUIRED_PROMOTION') # WinError 6837 (0x1AB5)
    ERROR_CANNOT_EXECUTE_FILE_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_win32error(0x1AB6, 'ERROR_CANNOT_EXECUTE_FILE_IN_TRANSACTION') # WinError 6838 (0x1AB6)
    ERROR_TRANSACTIONS_NOT_FROZEN: Final[ThirtyTwoBits] = cached_win32error(0x1AB7, 'ERROR_TRANSACTIONS_NOT_FROZEN') # WinError 6839 (0x1AB7)
    ERROR_TRANSACTION_FREEZE_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x1AB8, 'ERROR_TRANSACTION_FREEZE_IN_PROGRESS') # WinError 6840 (0x1AB8)
    ERROR_NOT_SNAPSHOT_VOLUME: Final[ThirtyTwoBits] = cached_win32error(0x1AB9, 'ERROR_NOT_SNAPSHOT_VOLUME') # WinError 6841 (0x1AB9)
    ERROR_NO_SAVEPOINT_WITH_OPEN_FILES: Final[ThirtyTwoBits] = cached_win32error(0x1ABA, 'ERROR_NO_SAVEPOINT_WITH_OPEN_FILES') # WinError 6842 (0x1ABA)
    ERROR_DATA_LOST_REPAIR: Final[ThirtyTwoBits] = cached_win32error(0x1ABB, 'ERROR_DATA_LOST_REPAIR') # WinError 6843 (0x1ABB)
    ERROR_SPARSE_NOT_ALLOWED_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_win32error(0x1ABC, 'ERROR_SPARSE_NOT_ALLOWED_IN_TRANSACTION') # WinError 6844 (0x1ABC)
    ERROR_TM_IDENTITY_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x1ABD, 'ERROR_TM_IDENTITY_MISMATCH') # WinError 6845 (0x1ABD)
    ERROR_FLOATED_SECTION: Final[ThirtyTwoBits] = cached_win32error(0x1ABE, 'ERROR_FLOATED_SECTION') # WinError 6846 (0x1ABE)
    ERROR_CANNOT_ACCEPT_TRANSACTED_WORK: Final[ThirtyTwoBits] = cached_win32error(0x1ABF, 'ERROR_CANNOT_ACCEPT_TRANSACTED_WORK') # WinError 6847 (0x1ABF)
    ERROR_CANNOT_ABORT_TRANSACTIONS: Final[ThirtyTwoBits] = cached_win32error(0x1AC0, 'ERROR_CANNOT_ABORT_TRANSACTIONS') # WinError 6848 (0x1AC0)
    ERROR_BAD_CLUSTERS: Final[ThirtyTwoBits] = cached_win32error(0x1AC1, 'ERROR_BAD_CLUSTERS') # WinError 6849 (0x1AC1)
    ERROR_COMPRESSION_NOT_ALLOWED_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_win32error(0x1AC2, 'ERROR_COMPRESSION_NOT_ALLOWED_IN_TRANSACTION') # WinError 6850 (0x1AC2)
    ERROR_VOLUME_DIRTY: Final[ThirtyTwoBits] = cached_win32error(0x1AC3, 'ERROR_VOLUME_DIRTY') # WinError 6851 (0x1AC3)
    ERROR_NO_LINK_TRACKING_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_win32error(0x1AC4, 'ERROR_NO_LINK_TRACKING_IN_TRANSACTION') # WinError 6852 (0x1AC4)
    ERROR_OPERATION_NOT_SUPPORTED_IN_TRANSACTION: Final[ThirtyTwoBits] = cached_win32error(0x1AC5, 'ERROR_OPERATION_NOT_SUPPORTED_IN_TRANSACTION') # WinError 6853 (0x1AC5)
    ERROR_EXPIRED_HANDLE: Final[ThirtyTwoBits] = cached_win32error(0x1AC6, 'ERROR_EXPIRED_HANDLE') # WinError 6854 (0x1AC6)
    ERROR_TRANSACTION_NOT_ENLISTED: Final[ThirtyTwoBits] = cached_win32error(0x1AC7, 'ERROR_TRANSACTION_NOT_ENLISTED') # WinError 6855 (0x1AC7)
    ERROR_CTX_WINSTATION_NAME_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x1B59, 'ERROR_CTX_WINSTATION_NAME_INVALID') # WinError 7001 (0x1B59)
    ERROR_CTX_INVALID_PD: Final[ThirtyTwoBits] = cached_win32error(0x1B5A, 'ERROR_CTX_INVALID_PD') # WinError 7002 (0x1B5A)
    ERROR_CTX_PD_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x1B5B, 'ERROR_CTX_PD_NOT_FOUND') # WinError 7003 (0x1B5B)
    ERROR_CTX_WD_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x1B5C, 'ERROR_CTX_WD_NOT_FOUND') # WinError 7004 (0x1B5C)
    ERROR_CTX_CANNOT_MAKE_EVENTLOG_ENTRY: Final[ThirtyTwoBits] = cached_win32error(0x1B5D, 'ERROR_CTX_CANNOT_MAKE_EVENTLOG_ENTRY') # WinError 7005 (0x1B5D)
    ERROR_CTX_SERVICE_NAME_COLLISION: Final[ThirtyTwoBits] = cached_win32error(0x1B5E, 'ERROR_CTX_SERVICE_NAME_COLLISION') # WinError 7006 (0x1B5E)
    ERROR_CTX_CLOSE_PENDING: Final[ThirtyTwoBits] = cached_win32error(0x1B5F, 'ERROR_CTX_CLOSE_PENDING') # WinError 7007 (0x1B5F)
    ERROR_CTX_NO_OUTBUF: Final[ThirtyTwoBits] = cached_win32error(0x1B60, 'ERROR_CTX_NO_OUTBUF') # WinError 7008 (0x1B60)
    ERROR_CTX_MODEM_INF_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x1B61, 'ERROR_CTX_MODEM_INF_NOT_FOUND') # WinError 7009 (0x1B61)
    ERROR_CTX_INVALID_MODEMNAME: Final[ThirtyTwoBits] = cached_win32error(0x1B62, 'ERROR_CTX_INVALID_MODEMNAME') # WinError 7010 (0x1B62)
    ERROR_CTX_MODEM_RESPONSE_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x1B63, 'ERROR_CTX_MODEM_RESPONSE_ERROR') # WinError 7011 (0x1B63)
    ERROR_CTX_MODEM_RESPONSE_TIMEOUT: Final[ThirtyTwoBits] = cached_win32error(0x1B64, 'ERROR_CTX_MODEM_RESPONSE_TIMEOUT') # WinError 7012 (0x1B64)
    ERROR_CTX_MODEM_RESPONSE_NO_CARRIER: Final[ThirtyTwoBits] = cached_win32error(0x1B65, 'ERROR_CTX_MODEM_RESPONSE_NO_CARRIER') # WinError 7013 (0x1B65)
    ERROR_CTX_MODEM_RESPONSE_NO_DIALTONE: Final[ThirtyTwoBits] = cached_win32error(0x1B66, 'ERROR_CTX_MODEM_RESPONSE_NO_DIALTONE') # WinError 7014 (0x1B66)
    ERROR_CTX_MODEM_RESPONSE_BUSY: Final[ThirtyTwoBits] = cached_win32error(0x1B67, 'ERROR_CTX_MODEM_RESPONSE_BUSY') # WinError 7015 (0x1B67)
    ERROR_CTX_MODEM_RESPONSE_VOICE: Final[ThirtyTwoBits] = cached_win32error(0x1B68, 'ERROR_CTX_MODEM_RESPONSE_VOICE') # WinError 7016 (0x1B68)
    ERROR_CTX_TD_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x1B69, 'ERROR_CTX_TD_ERROR') # WinError 7017 (0x1B69)
    ERROR_CTX_WINSTATION_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x1B6E, 'ERROR_CTX_WINSTATION_NOT_FOUND') # WinError 7022 (0x1B6E)
    ERROR_CTX_WINSTATION_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x1B6F, 'ERROR_CTX_WINSTATION_ALREADY_EXISTS') # WinError 7023 (0x1B6F)
    ERROR_CTX_WINSTATION_BUSY: Final[ThirtyTwoBits] = cached_win32error(0x1B70, 'ERROR_CTX_WINSTATION_BUSY') # WinError 7024 (0x1B70)
    ERROR_CTX_BAD_VIDEO_MODE: Final[ThirtyTwoBits] = cached_win32error(0x1B71, 'ERROR_CTX_BAD_VIDEO_MODE') # WinError 7025 (0x1B71)
    ERROR_CTX_GRAPHICS_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x1B7B, 'ERROR_CTX_GRAPHICS_INVALID') # WinError 7035 (0x1B7B)
    ERROR_CTX_LOGON_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x1B7D, 'ERROR_CTX_LOGON_DISABLED') # WinError 7037 (0x1B7D)
    ERROR_CTX_NOT_CONSOLE: Final[ThirtyTwoBits] = cached_win32error(0x1B7E, 'ERROR_CTX_NOT_CONSOLE') # WinError 7038 (0x1B7E)
    ERROR_CTX_CLIENT_QUERY_TIMEOUT: Final[ThirtyTwoBits] = cached_win32error(0x1B80, 'ERROR_CTX_CLIENT_QUERY_TIMEOUT') # WinError 7040 (0x1B80)
    ERROR_CTX_CONSOLE_DISCONNECT: Final[ThirtyTwoBits] = cached_win32error(0x1B81, 'ERROR_CTX_CONSOLE_DISCONNECT') # WinError 7041 (0x1B81)
    ERROR_CTX_CONSOLE_CONNECT: Final[ThirtyTwoBits] = cached_win32error(0x1B82, 'ERROR_CTX_CONSOLE_CONNECT') # WinError 7042 (0x1B82)
    ERROR_CTX_SHADOW_DENIED: Final[ThirtyTwoBits] = cached_win32error(0x1B84, 'ERROR_CTX_SHADOW_DENIED') # WinError 7044 (0x1B84)
    ERROR_CTX_WINSTATION_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_win32error(0x1B85, 'ERROR_CTX_WINSTATION_ACCESS_DENIED') # WinError 7045 (0x1B85)
    ERROR_CTX_INVALID_WD: Final[ThirtyTwoBits] = cached_win32error(0x1B89, 'ERROR_CTX_INVALID_WD') # WinError 7049 (0x1B89)
    ERROR_CTX_SHADOW_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x1B8A, 'ERROR_CTX_SHADOW_INVALID') # WinError 7050 (0x1B8A)
    ERROR_CTX_SHADOW_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x1B8B, 'ERROR_CTX_SHADOW_DISABLED') # WinError 7051 (0x1B8B)
    ERROR_CTX_CLIENT_LICENSE_IN_USE: Final[ThirtyTwoBits] = cached_win32error(0x1B8C, 'ERROR_CTX_CLIENT_LICENSE_IN_USE') # WinError 7052 (0x1B8C)
    ERROR_CTX_CLIENT_LICENSE_NOT_SET: Final[ThirtyTwoBits] = cached_win32error(0x1B8D, 'ERROR_CTX_CLIENT_LICENSE_NOT_SET') # WinError 7053 (0x1B8D)
    ERROR_CTX_LICENSE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x1B8E, 'ERROR_CTX_LICENSE_NOT_AVAILABLE') # WinError 7054 (0x1B8E)
    ERROR_CTX_LICENSE_CLIENT_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x1B8F, 'ERROR_CTX_LICENSE_CLIENT_INVALID') # WinError 7055 (0x1B8F)
    ERROR_CTX_LICENSE_EXPIRED: Final[ThirtyTwoBits] = cached_win32error(0x1B90, 'ERROR_CTX_LICENSE_EXPIRED') # WinError 7056 (0x1B90)
    ERROR_CTX_SHADOW_NOT_RUNNING: Final[ThirtyTwoBits] = cached_win32error(0x1B91, 'ERROR_CTX_SHADOW_NOT_RUNNING') # WinError 7057 (0x1B91)
    ERROR_CTX_SHADOW_ENDED_BY_MODE_CHANGE: Final[ThirtyTwoBits] = cached_win32error(0x1B92, 'ERROR_CTX_SHADOW_ENDED_BY_MODE_CHANGE') # WinError 7058 (0x1B92)
    ERROR_ACTIVATION_COUNT_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x1B93, 'ERROR_ACTIVATION_COUNT_EXCEEDED') # WinError 7059 (0x1B93)
    ERROR_CTX_WINSTATIONS_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x1B94, 'ERROR_CTX_WINSTATIONS_DISABLED') # WinError 7060 (0x1B94)
    ERROR_CTX_ENCRYPTION_LEVEL_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x1B95, 'ERROR_CTX_ENCRYPTION_LEVEL_REQUIRED') # WinError 7061 (0x1B95)
    ERROR_CTX_SESSION_IN_USE: Final[ThirtyTwoBits] = cached_win32error(0x1B96, 'ERROR_CTX_SESSION_IN_USE') # WinError 7062 (0x1B96)
    ERROR_CTX_NO_FORCE_LOGOFF: Final[ThirtyTwoBits] = cached_win32error(0x1B97, 'ERROR_CTX_NO_FORCE_LOGOFF') # WinError 7063 (0x1B97)
    ERROR_CTX_ACCOUNT_RESTRICTION: Final[ThirtyTwoBits] = cached_win32error(0x1B98, 'ERROR_CTX_ACCOUNT_RESTRICTION') # WinError 7064 (0x1B98)
    ERROR_RDP_PROTOCOL_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x1B99, 'ERROR_RDP_PROTOCOL_ERROR') # WinError 7065 (0x1B99)
    ERROR_CTX_CDM_CONNECT: Final[ThirtyTwoBits] = cached_win32error(0x1B9A, 'ERROR_CTX_CDM_CONNECT') # WinError 7066 (0x1B9A)
    ERROR_CTX_CDM_DISCONNECT: Final[ThirtyTwoBits] = cached_win32error(0x1B9B, 'ERROR_CTX_CDM_DISCONNECT') # WinError 7067 (0x1B9B)
    ERROR_CTX_SECURITY_LAYER_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x1B9C, 'ERROR_CTX_SECURITY_LAYER_ERROR') # WinError 7068 (0x1B9C)
    ERROR_TS_INCOMPATIBLE_SESSIONS: Final[ThirtyTwoBits] = cached_win32error(0x1B9D, 'ERROR_TS_INCOMPATIBLE_SESSIONS') # WinError 7069 (0x1B9D)
    ERROR_TS_VIDEO_SUBSYSTEM_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x1B9E, 'ERROR_TS_VIDEO_SUBSYSTEM_ERROR') # WinError 7070 (0x1B9E)
    FRS_ERR_INVALID_API_SEQUENCE: Final[ThirtyTwoBits] = cached_win32error(0x1F41, 'FRS_ERR_INVALID_API_SEQUENCE') # WinError 8001 (0x1F41)
    FRS_ERR_STARTING_SERVICE: Final[ThirtyTwoBits] = cached_win32error(0x1F42, 'FRS_ERR_STARTING_SERVICE') # WinError 8002 (0x1F42)
    FRS_ERR_STOPPING_SERVICE: Final[ThirtyTwoBits] = cached_win32error(0x1F43, 'FRS_ERR_STOPPING_SERVICE') # WinError 8003 (0x1F43)
    FRS_ERR_INTERNAL_API: Final[ThirtyTwoBits] = cached_win32error(0x1F44, 'FRS_ERR_INTERNAL_API') # WinError 8004 (0x1F44)
    FRS_ERR_INTERNAL: Final[ThirtyTwoBits] = cached_win32error(0x1F45, 'FRS_ERR_INTERNAL') # WinError 8005 (0x1F45)
    FRS_ERR_SERVICE_COMM: Final[ThirtyTwoBits] = cached_win32error(0x1F46, 'FRS_ERR_SERVICE_COMM') # WinError 8006 (0x1F46)
    FRS_ERR_INSUFFICIENT_PRIV: Final[ThirtyTwoBits] = cached_win32error(0x1F47, 'FRS_ERR_INSUFFICIENT_PRIV') # WinError 8007 (0x1F47)
    FRS_ERR_AUTHENTICATION: Final[ThirtyTwoBits] = cached_win32error(0x1F48, 'FRS_ERR_AUTHENTICATION') # WinError 8008 (0x1F48)
    FRS_ERR_PARENT_INSUFFICIENT_PRIV: Final[ThirtyTwoBits] = cached_win32error(0x1F49, 'FRS_ERR_PARENT_INSUFFICIENT_PRIV') # WinError 8009 (0x1F49)
    FRS_ERR_PARENT_AUTHENTICATION: Final[ThirtyTwoBits] = cached_win32error(0x1F4A, 'FRS_ERR_PARENT_AUTHENTICATION') # WinError 8010 (0x1F4A)
    FRS_ERR_CHILD_TO_PARENT_COMM: Final[ThirtyTwoBits] = cached_win32error(0x1F4B, 'FRS_ERR_CHILD_TO_PARENT_COMM') # WinError 8011 (0x1F4B)
    FRS_ERR_PARENT_TO_CHILD_COMM: Final[ThirtyTwoBits] = cached_win32error(0x1F4C, 'FRS_ERR_PARENT_TO_CHILD_COMM') # WinError 8012 (0x1F4C)
    FRS_ERR_SYSVOL_POPULATE: Final[ThirtyTwoBits] = cached_win32error(0x1F4D, 'FRS_ERR_SYSVOL_POPULATE') # WinError 8013 (0x1F4D)
    FRS_ERR_SYSVOL_POPULATE_TIMEOUT: Final[ThirtyTwoBits] = cached_win32error(0x1F4E, 'FRS_ERR_SYSVOL_POPULATE_TIMEOUT') # WinError 8014 (0x1F4E)
    FRS_ERR_SYSVOL_IS_BUSY: Final[ThirtyTwoBits] = cached_win32error(0x1F4F, 'FRS_ERR_SYSVOL_IS_BUSY') # WinError 8015 (0x1F4F)
    FRS_ERR_SYSVOL_DEMOTE: Final[ThirtyTwoBits] = cached_win32error(0x1F50, 'FRS_ERR_SYSVOL_DEMOTE') # WinError 8016 (0x1F50)
    FRS_ERR_INVALID_SERVICE_PARAMETER: Final[ThirtyTwoBits] = cached_win32error(0x1F51, 'FRS_ERR_INVALID_SERVICE_PARAMETER') # WinError 8017 (0x1F51)
    ERROR_DS_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_win32error(0x2008, 'ERROR_DS_NOT_INSTALLED') # WinError 8200 (0x2008)
    ERROR_DS_MEMBERSHIP_EVALUATED_LOCALLY: Final[ThirtyTwoBits] = cached_win32error(0x2009, 'ERROR_DS_MEMBERSHIP_EVALUATED_LOCALLY') # WinError 8201 (0x2009)
    ERROR_DS_NO_ATTRIBUTE_OR_VALUE: Final[ThirtyTwoBits] = cached_win32error(0x200A, 'ERROR_DS_NO_ATTRIBUTE_OR_VALUE') # WinError 8202 (0x200A)
    ERROR_DS_INVALID_ATTRIBUTE_SYNTAX: Final[ThirtyTwoBits] = cached_win32error(0x200B, 'ERROR_DS_INVALID_ATTRIBUTE_SYNTAX') # WinError 8203 (0x200B)
    ERROR_DS_ATTRIBUTE_TYPE_UNDEFINED: Final[ThirtyTwoBits] = cached_win32error(0x200C, 'ERROR_DS_ATTRIBUTE_TYPE_UNDEFINED') # WinError 8204 (0x200C)
    ERROR_DS_ATTRIBUTE_OR_VALUE_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x200D, 'ERROR_DS_ATTRIBUTE_OR_VALUE_EXISTS') # WinError 8205 (0x200D)
    ERROR_DS_BUSY: Final[ThirtyTwoBits] = cached_win32error(0x200E, 'ERROR_DS_BUSY') # WinError 8206 (0x200E)
    ERROR_DS_UNAVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x200F, 'ERROR_DS_UNAVAILABLE') # WinError 8207 (0x200F)
    ERROR_DS_NO_RIDS_ALLOCATED: Final[ThirtyTwoBits] = cached_win32error(0x2010, 'ERROR_DS_NO_RIDS_ALLOCATED') # WinError 8208 (0x2010)
    ERROR_DS_NO_MORE_RIDS: Final[ThirtyTwoBits] = cached_win32error(0x2011, 'ERROR_DS_NO_MORE_RIDS') # WinError 8209 (0x2011)
    ERROR_DS_INCORRECT_ROLE_OWNER: Final[ThirtyTwoBits] = cached_win32error(0x2012, 'ERROR_DS_INCORRECT_ROLE_OWNER') # WinError 8210 (0x2012)
    ERROR_DS_RIDMGR_INIT_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x2013, 'ERROR_DS_RIDMGR_INIT_ERROR') # WinError 8211 (0x2013)
    ERROR_DS_OBJ_CLASS_VIOLATION: Final[ThirtyTwoBits] = cached_win32error(0x2014, 'ERROR_DS_OBJ_CLASS_VIOLATION') # WinError 8212 (0x2014)
    ERROR_DS_CANT_ON_NON_LEAF: Final[ThirtyTwoBits] = cached_win32error(0x2015, 'ERROR_DS_CANT_ON_NON_LEAF') # WinError 8213 (0x2015)
    ERROR_DS_CANT_ON_RDN: Final[ThirtyTwoBits] = cached_win32error(0x2016, 'ERROR_DS_CANT_ON_RDN') # WinError 8214 (0x2016)
    ERROR_DS_CANT_MOD_OBJ_CLASS: Final[ThirtyTwoBits] = cached_win32error(0x2017, 'ERROR_DS_CANT_MOD_OBJ_CLASS') # WinError 8215 (0x2017)
    ERROR_DS_CROSS_DOM_MOVE_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x2018, 'ERROR_DS_CROSS_DOM_MOVE_ERROR') # WinError 8216 (0x2018)
    ERROR_DS_GC_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x2019, 'ERROR_DS_GC_NOT_AVAILABLE') # WinError 8217 (0x2019)
    ERROR_SHARED_POLICY: Final[ThirtyTwoBits] = cached_win32error(0x201A, 'ERROR_SHARED_POLICY') # WinError 8218 (0x201A)
    ERROR_POLICY_OBJECT_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x201B, 'ERROR_POLICY_OBJECT_NOT_FOUND') # WinError 8219 (0x201B)
    ERROR_POLICY_ONLY_IN_DS: Final[ThirtyTwoBits] = cached_win32error(0x201C, 'ERROR_POLICY_ONLY_IN_DS') # WinError 8220 (0x201C)
    ERROR_PROMOTION_ACTIVE: Final[ThirtyTwoBits] = cached_win32error(0x201D, 'ERROR_PROMOTION_ACTIVE') # WinError 8221 (0x201D)
    ERROR_NO_PROMOTION_ACTIVE: Final[ThirtyTwoBits] = cached_win32error(0x201E, 'ERROR_NO_PROMOTION_ACTIVE') # WinError 8222 (0x201E)
    ERROR_DS_OPERATIONS_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x2020, 'ERROR_DS_OPERATIONS_ERROR') # WinError 8224 (0x2020)
    ERROR_DS_PROTOCOL_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x2021, 'ERROR_DS_PROTOCOL_ERROR') # WinError 8225 (0x2021)
    ERROR_DS_TIMELIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x2022, 'ERROR_DS_TIMELIMIT_EXCEEDED') # WinError 8226 (0x2022)
    ERROR_DS_SIZELIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x2023, 'ERROR_DS_SIZELIMIT_EXCEEDED') # WinError 8227 (0x2023)
    ERROR_DS_ADMIN_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x2024, 'ERROR_DS_ADMIN_LIMIT_EXCEEDED') # WinError 8228 (0x2024)
    ERROR_DS_COMPARE_FALSE: Final[ThirtyTwoBits] = cached_win32error(0x2025, 'ERROR_DS_COMPARE_FALSE') # WinError 8229 (0x2025)
    ERROR_DS_COMPARE_TRUE: Final[ThirtyTwoBits] = cached_win32error(0x2026, 'ERROR_DS_COMPARE_TRUE') # WinError 8230 (0x2026)
    ERROR_DS_AUTH_METHOD_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x2027, 'ERROR_DS_AUTH_METHOD_NOT_SUPPORTED') # WinError 8231 (0x2027)
    ERROR_DS_STRONG_AUTH_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x2028, 'ERROR_DS_STRONG_AUTH_REQUIRED') # WinError 8232 (0x2028)
    ERROR_DS_INAPPROPRIATE_AUTH: Final[ThirtyTwoBits] = cached_win32error(0x2029, 'ERROR_DS_INAPPROPRIATE_AUTH') # WinError 8233 (0x2029)
    ERROR_DS_AUTH_UNKNOWN: Final[ThirtyTwoBits] = cached_win32error(0x202A, 'ERROR_DS_AUTH_UNKNOWN') # WinError 8234 (0x202A)
    ERROR_DS_REFERRAL: Final[ThirtyTwoBits] = cached_win32error(0x202B, 'ERROR_DS_REFERRAL') # WinError 8235 (0x202B)
    ERROR_DS_UNAVAILABLE_CRIT_EXTENSION: Final[ThirtyTwoBits] = cached_win32error(0x202C, 'ERROR_DS_UNAVAILABLE_CRIT_EXTENSION') # WinError 8236 (0x202C)
    ERROR_DS_CONFIDENTIALITY_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x202D, 'ERROR_DS_CONFIDENTIALITY_REQUIRED') # WinError 8237 (0x202D)
    ERROR_DS_INAPPROPRIATE_MATCHING: Final[ThirtyTwoBits] = cached_win32error(0x202E, 'ERROR_DS_INAPPROPRIATE_MATCHING') # WinError 8238 (0x202E)
    ERROR_DS_CONSTRAINT_VIOLATION: Final[ThirtyTwoBits] = cached_win32error(0x202F, 'ERROR_DS_CONSTRAINT_VIOLATION') # WinError 8239 (0x202F)
    ERROR_DS_NO_SUCH_OBJECT: Final[ThirtyTwoBits] = cached_win32error(0x2030, 'ERROR_DS_NO_SUCH_OBJECT') # WinError 8240 (0x2030)
    ERROR_DS_ALIAS_PROBLEM: Final[ThirtyTwoBits] = cached_win32error(0x2031, 'ERROR_DS_ALIAS_PROBLEM') # WinError 8241 (0x2031)
    ERROR_DS_INVALID_DN_SYNTAX: Final[ThirtyTwoBits] = cached_win32error(0x2032, 'ERROR_DS_INVALID_DN_SYNTAX') # WinError 8242 (0x2032)
    ERROR_DS_IS_LEAF: Final[ThirtyTwoBits] = cached_win32error(0x2033, 'ERROR_DS_IS_LEAF') # WinError 8243 (0x2033)
    ERROR_DS_ALIAS_DEREF_PROBLEM: Final[ThirtyTwoBits] = cached_win32error(0x2034, 'ERROR_DS_ALIAS_DEREF_PROBLEM') # WinError 8244 (0x2034)
    ERROR_DS_UNWILLING_TO_PERFORM: Final[ThirtyTwoBits] = cached_win32error(0x2035, 'ERROR_DS_UNWILLING_TO_PERFORM') # WinError 8245 (0x2035)
    ERROR_DS_LOOP_DETECT: Final[ThirtyTwoBits] = cached_win32error(0x2036, 'ERROR_DS_LOOP_DETECT') # WinError 8246 (0x2036)
    ERROR_DS_NAMING_VIOLATION: Final[ThirtyTwoBits] = cached_win32error(0x2037, 'ERROR_DS_NAMING_VIOLATION') # WinError 8247 (0x2037)
    ERROR_DS_OBJECT_RESULTS_TOO_LARGE: Final[ThirtyTwoBits] = cached_win32error(0x2038, 'ERROR_DS_OBJECT_RESULTS_TOO_LARGE') # WinError 8248 (0x2038)
    ERROR_DS_AFFECTS_MULTIPLE_DSAS: Final[ThirtyTwoBits] = cached_win32error(0x2039, 'ERROR_DS_AFFECTS_MULTIPLE_DSAS') # WinError 8249 (0x2039)
    ERROR_DS_SERVER_DOWN: Final[ThirtyTwoBits] = cached_win32error(0x203A, 'ERROR_DS_SERVER_DOWN') # WinError 8250 (0x203A)
    ERROR_DS_LOCAL_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x203B, 'ERROR_DS_LOCAL_ERROR') # WinError 8251 (0x203B)
    ERROR_DS_ENCODING_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x203C, 'ERROR_DS_ENCODING_ERROR') # WinError 8252 (0x203C)
    ERROR_DS_DECODING_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x203D, 'ERROR_DS_DECODING_ERROR') # WinError 8253 (0x203D)
    ERROR_DS_FILTER_UNKNOWN: Final[ThirtyTwoBits] = cached_win32error(0x203E, 'ERROR_DS_FILTER_UNKNOWN') # WinError 8254 (0x203E)
    ERROR_DS_PARAM_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x203F, 'ERROR_DS_PARAM_ERROR') # WinError 8255 (0x203F)
    ERROR_DS_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x2040, 'ERROR_DS_NOT_SUPPORTED') # WinError 8256 (0x2040)
    ERROR_DS_NO_RESULTS_RETURNED: Final[ThirtyTwoBits] = cached_win32error(0x2041, 'ERROR_DS_NO_RESULTS_RETURNED') # WinError 8257 (0x2041)
    ERROR_DS_CONTROL_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x2042, 'ERROR_DS_CONTROL_NOT_FOUND') # WinError 8258 (0x2042)
    ERROR_DS_CLIENT_LOOP: Final[ThirtyTwoBits] = cached_win32error(0x2043, 'ERROR_DS_CLIENT_LOOP') # WinError 8259 (0x2043)
    ERROR_DS_REFERRAL_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x2044, 'ERROR_DS_REFERRAL_LIMIT_EXCEEDED') # WinError 8260 (0x2044)
    ERROR_DS_SORT_CONTROL_MISSING: Final[ThirtyTwoBits] = cached_win32error(0x2045, 'ERROR_DS_SORT_CONTROL_MISSING') # WinError 8261 (0x2045)
    ERROR_DS_OFFSET_RANGE_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x2046, 'ERROR_DS_OFFSET_RANGE_ERROR') # WinError 8262 (0x2046)
    ERROR_DS_RIDMGR_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x2047, 'ERROR_DS_RIDMGR_DISABLED') # WinError 8263 (0x2047)
    ERROR_DS_ROOT_MUST_BE_NC: Final[ThirtyTwoBits] = cached_win32error(0x206D, 'ERROR_DS_ROOT_MUST_BE_NC') # WinError 8301 (0x206D)
    ERROR_DS_ADD_REPLICA_INHIBITED: Final[ThirtyTwoBits] = cached_win32error(0x206E, 'ERROR_DS_ADD_REPLICA_INHIBITED') # WinError 8302 (0x206E)
    ERROR_DS_ATT_NOT_DEF_IN_SCHEMA: Final[ThirtyTwoBits] = cached_win32error(0x206F, 'ERROR_DS_ATT_NOT_DEF_IN_SCHEMA') # WinError 8303 (0x206F)
    ERROR_DS_MAX_OBJ_SIZE_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x2070, 'ERROR_DS_MAX_OBJ_SIZE_EXCEEDED') # WinError 8304 (0x2070)
    ERROR_DS_OBJ_STRING_NAME_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x2071, 'ERROR_DS_OBJ_STRING_NAME_EXISTS') # WinError 8305 (0x2071)
    ERROR_DS_NO_RDN_DEFINED_IN_SCHEMA: Final[ThirtyTwoBits] = cached_win32error(0x2072, 'ERROR_DS_NO_RDN_DEFINED_IN_SCHEMA') # WinError 8306 (0x2072)
    ERROR_DS_RDN_DOESNT_MATCH_SCHEMA: Final[ThirtyTwoBits] = cached_win32error(0x2073, 'ERROR_DS_RDN_DOESNT_MATCH_SCHEMA') # WinError 8307 (0x2073)
    ERROR_DS_NO_REQUESTED_ATTS_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x2074, 'ERROR_DS_NO_REQUESTED_ATTS_FOUND') # WinError 8308 (0x2074)
    ERROR_DS_USER_BUFFER_TO_SMALL: Final[ThirtyTwoBits] = cached_win32error(0x2075, 'ERROR_DS_USER_BUFFER_TO_SMALL') # WinError 8309 (0x2075)
    ERROR_DS_ATT_IS_NOT_ON_OBJ: Final[ThirtyTwoBits] = cached_win32error(0x2076, 'ERROR_DS_ATT_IS_NOT_ON_OBJ') # WinError 8310 (0x2076)
    ERROR_DS_ILLEGAL_MOD_OPERATION: Final[ThirtyTwoBits] = cached_win32error(0x2077, 'ERROR_DS_ILLEGAL_MOD_OPERATION') # WinError 8311 (0x2077)
    ERROR_DS_OBJ_TOO_LARGE: Final[ThirtyTwoBits] = cached_win32error(0x2078, 'ERROR_DS_OBJ_TOO_LARGE') # WinError 8312 (0x2078)
    ERROR_DS_BAD_INSTANCE_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x2079, 'ERROR_DS_BAD_INSTANCE_TYPE') # WinError 8313 (0x2079)
    ERROR_DS_MASTERDSA_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x207A, 'ERROR_DS_MASTERDSA_REQUIRED') # WinError 8314 (0x207A)
    ERROR_DS_OBJECT_CLASS_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x207B, 'ERROR_DS_OBJECT_CLASS_REQUIRED') # WinError 8315 (0x207B)
    ERROR_DS_MISSING_REQUIRED_ATT: Final[ThirtyTwoBits] = cached_win32error(0x207C, 'ERROR_DS_MISSING_REQUIRED_ATT') # WinError 8316 (0x207C)
    ERROR_DS_ATT_NOT_DEF_FOR_CLASS: Final[ThirtyTwoBits] = cached_win32error(0x207D, 'ERROR_DS_ATT_NOT_DEF_FOR_CLASS') # WinError 8317 (0x207D)
    ERROR_DS_ATT_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x207E, 'ERROR_DS_ATT_ALREADY_EXISTS') # WinError 8318 (0x207E)
    ERROR_DS_CANT_ADD_ATT_VALUES: Final[ThirtyTwoBits] = cached_win32error(0x2080, 'ERROR_DS_CANT_ADD_ATT_VALUES') # WinError 8320 (0x2080)
    ERROR_DS_SINGLE_VALUE_CONSTRAINT: Final[ThirtyTwoBits] = cached_win32error(0x2081, 'ERROR_DS_SINGLE_VALUE_CONSTRAINT') # WinError 8321 (0x2081)
    ERROR_DS_RANGE_CONSTRAINT: Final[ThirtyTwoBits] = cached_win32error(0x2082, 'ERROR_DS_RANGE_CONSTRAINT') # WinError 8322 (0x2082)
    ERROR_DS_ATT_VAL_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x2083, 'ERROR_DS_ATT_VAL_ALREADY_EXISTS') # WinError 8323 (0x2083)
    ERROR_DS_CANT_REM_MISSING_ATT: Final[ThirtyTwoBits] = cached_win32error(0x2084, 'ERROR_DS_CANT_REM_MISSING_ATT') # WinError 8324 (0x2084)
    ERROR_DS_CANT_REM_MISSING_ATT_VAL: Final[ThirtyTwoBits] = cached_win32error(0x2085, 'ERROR_DS_CANT_REM_MISSING_ATT_VAL') # WinError 8325 (0x2085)
    ERROR_DS_ROOT_CANT_BE_SUBREF: Final[ThirtyTwoBits] = cached_win32error(0x2086, 'ERROR_DS_ROOT_CANT_BE_SUBREF') # WinError 8326 (0x2086)
    ERROR_DS_NO_CHAINING: Final[ThirtyTwoBits] = cached_win32error(0x2087, 'ERROR_DS_NO_CHAINING') # WinError 8327 (0x2087)
    ERROR_DS_NO_CHAINED_EVAL: Final[ThirtyTwoBits] = cached_win32error(0x2088, 'ERROR_DS_NO_CHAINED_EVAL') # WinError 8328 (0x2088)
    ERROR_DS_NO_PARENT_OBJECT: Final[ThirtyTwoBits] = cached_win32error(0x2089, 'ERROR_DS_NO_PARENT_OBJECT') # WinError 8329 (0x2089)
    ERROR_DS_PARENT_IS_AN_ALIAS: Final[ThirtyTwoBits] = cached_win32error(0x208A, 'ERROR_DS_PARENT_IS_AN_ALIAS') # WinError 8330 (0x208A)
    ERROR_DS_CANT_MIX_MASTER_AND_REPS: Final[ThirtyTwoBits] = cached_win32error(0x208B, 'ERROR_DS_CANT_MIX_MASTER_AND_REPS') # WinError 8331 (0x208B)
    ERROR_DS_CHILDREN_EXIST: Final[ThirtyTwoBits] = cached_win32error(0x208C, 'ERROR_DS_CHILDREN_EXIST') # WinError 8332 (0x208C)
    ERROR_DS_OBJ_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x208D, 'ERROR_DS_OBJ_NOT_FOUND') # WinError 8333 (0x208D)
    ERROR_DS_ALIASED_OBJ_MISSING: Final[ThirtyTwoBits] = cached_win32error(0x208E, 'ERROR_DS_ALIASED_OBJ_MISSING') # WinError 8334 (0x208E)
    ERROR_DS_BAD_NAME_SYNTAX: Final[ThirtyTwoBits] = cached_win32error(0x208F, 'ERROR_DS_BAD_NAME_SYNTAX') # WinError 8335 (0x208F)
    ERROR_DS_ALIAS_POINTS_TO_ALIAS: Final[ThirtyTwoBits] = cached_win32error(0x2090, 'ERROR_DS_ALIAS_POINTS_TO_ALIAS') # WinError 8336 (0x2090)
    ERROR_DS_CANT_DEREF_ALIAS: Final[ThirtyTwoBits] = cached_win32error(0x2091, 'ERROR_DS_CANT_DEREF_ALIAS') # WinError 8337 (0x2091)
    ERROR_DS_OUT_OF_SCOPE: Final[ThirtyTwoBits] = cached_win32error(0x2092, 'ERROR_DS_OUT_OF_SCOPE') # WinError 8338 (0x2092)
    ERROR_DS_OBJECT_BEING_REMOVED: Final[ThirtyTwoBits] = cached_win32error(0x2093, 'ERROR_DS_OBJECT_BEING_REMOVED') # WinError 8339 (0x2093)
    ERROR_DS_CANT_DELETE_DSA_OBJ: Final[ThirtyTwoBits] = cached_win32error(0x2094, 'ERROR_DS_CANT_DELETE_DSA_OBJ') # WinError 8340 (0x2094)
    ERROR_DS_GENERIC_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x2095, 'ERROR_DS_GENERIC_ERROR') # WinError 8341 (0x2095)
    ERROR_DS_DSA_MUST_BE_INT_MASTER: Final[ThirtyTwoBits] = cached_win32error(0x2096, 'ERROR_DS_DSA_MUST_BE_INT_MASTER') # WinError 8342 (0x2096)
    ERROR_DS_CLASS_NOT_DSA: Final[ThirtyTwoBits] = cached_win32error(0x2097, 'ERROR_DS_CLASS_NOT_DSA') # WinError 8343 (0x2097)
    ERROR_DS_INSUFF_ACCESS_RIGHTS: Final[ThirtyTwoBits] = cached_win32error(0x2098, 'ERROR_DS_INSUFF_ACCESS_RIGHTS') # WinError 8344 (0x2098)
    ERROR_DS_ILLEGAL_SUPERIOR: Final[ThirtyTwoBits] = cached_win32error(0x2099, 'ERROR_DS_ILLEGAL_SUPERIOR') # WinError 8345 (0x2099)
    ERROR_DS_ATTRIBUTE_OWNED_BY_SAM: Final[ThirtyTwoBits] = cached_win32error(0x209A, 'ERROR_DS_ATTRIBUTE_OWNED_BY_SAM') # WinError 8346 (0x209A)
    ERROR_DS_NAME_TOO_MANY_PARTS: Final[ThirtyTwoBits] = cached_win32error(0x209B, 'ERROR_DS_NAME_TOO_MANY_PARTS') # WinError 8347 (0x209B)
    ERROR_DS_NAME_TOO_LONG: Final[ThirtyTwoBits] = cached_win32error(0x209C, 'ERROR_DS_NAME_TOO_LONG') # WinError 8348 (0x209C)
    ERROR_DS_NAME_VALUE_TOO_LONG: Final[ThirtyTwoBits] = cached_win32error(0x209D, 'ERROR_DS_NAME_VALUE_TOO_LONG') # WinError 8349 (0x209D)
    ERROR_DS_NAME_UNPARSEABLE: Final[ThirtyTwoBits] = cached_win32error(0x209E, 'ERROR_DS_NAME_UNPARSEABLE') # WinError 8350 (0x209E)
    ERROR_DS_NAME_TYPE_UNKNOWN: Final[ThirtyTwoBits] = cached_win32error(0x209F, 'ERROR_DS_NAME_TYPE_UNKNOWN') # WinError 8351 (0x209F)
    ERROR_DS_NOT_AN_OBJECT: Final[ThirtyTwoBits] = cached_win32error(0x20A0, 'ERROR_DS_NOT_AN_OBJECT') # WinError 8352 (0x20A0)
    ERROR_DS_SEC_DESC_TOO_SHORT: Final[ThirtyTwoBits] = cached_win32error(0x20A1, 'ERROR_DS_SEC_DESC_TOO_SHORT') # WinError 8353 (0x20A1)
    ERROR_DS_SEC_DESC_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x20A2, 'ERROR_DS_SEC_DESC_INVALID') # WinError 8354 (0x20A2)
    ERROR_DS_NO_DELETED_NAME: Final[ThirtyTwoBits] = cached_win32error(0x20A3, 'ERROR_DS_NO_DELETED_NAME') # WinError 8355 (0x20A3)
    ERROR_DS_SUBREF_MUST_HAVE_PARENT: Final[ThirtyTwoBits] = cached_win32error(0x20A4, 'ERROR_DS_SUBREF_MUST_HAVE_PARENT') # WinError 8356 (0x20A4)
    ERROR_DS_NCNAME_MUST_BE_NC: Final[ThirtyTwoBits] = cached_win32error(0x20A5, 'ERROR_DS_NCNAME_MUST_BE_NC') # WinError 8357 (0x20A5)
    ERROR_DS_CANT_ADD_SYSTEM_ONLY: Final[ThirtyTwoBits] = cached_win32error(0x20A6, 'ERROR_DS_CANT_ADD_SYSTEM_ONLY') # WinError 8358 (0x20A6)
    ERROR_DS_CLASS_MUST_BE_CONCRETE: Final[ThirtyTwoBits] = cached_win32error(0x20A7, 'ERROR_DS_CLASS_MUST_BE_CONCRETE') # WinError 8359 (0x20A7)
    ERROR_DS_INVALID_DMD: Final[ThirtyTwoBits] = cached_win32error(0x20A8, 'ERROR_DS_INVALID_DMD') # WinError 8360 (0x20A8)
    ERROR_DS_OBJ_GUID_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x20A9, 'ERROR_DS_OBJ_GUID_EXISTS') # WinError 8361 (0x20A9)
    ERROR_DS_NOT_ON_BACKLINK: Final[ThirtyTwoBits] = cached_win32error(0x20AA, 'ERROR_DS_NOT_ON_BACKLINK') # WinError 8362 (0x20AA)
    ERROR_DS_NO_CROSSREF_FOR_NC: Final[ThirtyTwoBits] = cached_win32error(0x20AB, 'ERROR_DS_NO_CROSSREF_FOR_NC') # WinError 8363 (0x20AB)
    ERROR_DS_SHUTTING_DOWN: Final[ThirtyTwoBits] = cached_win32error(0x20AC, 'ERROR_DS_SHUTTING_DOWN') # WinError 8364 (0x20AC)
    ERROR_DS_UNKNOWN_OPERATION: Final[ThirtyTwoBits] = cached_win32error(0x20AD, 'ERROR_DS_UNKNOWN_OPERATION') # WinError 8365 (0x20AD)
    ERROR_DS_INVALID_ROLE_OWNER: Final[ThirtyTwoBits] = cached_win32error(0x20AE, 'ERROR_DS_INVALID_ROLE_OWNER') # WinError 8366 (0x20AE)
    ERROR_DS_COULDNT_CONTACT_FSMO: Final[ThirtyTwoBits] = cached_win32error(0x20AF, 'ERROR_DS_COULDNT_CONTACT_FSMO') # WinError 8367 (0x20AF)
    ERROR_DS_CROSS_NC_DN_RENAME: Final[ThirtyTwoBits] = cached_win32error(0x20B0, 'ERROR_DS_CROSS_NC_DN_RENAME') # WinError 8368 (0x20B0)
    ERROR_DS_CANT_MOD_SYSTEM_ONLY: Final[ThirtyTwoBits] = cached_win32error(0x20B1, 'ERROR_DS_CANT_MOD_SYSTEM_ONLY') # WinError 8369 (0x20B1)
    ERROR_DS_REPLICATOR_ONLY: Final[ThirtyTwoBits] = cached_win32error(0x20B2, 'ERROR_DS_REPLICATOR_ONLY') # WinError 8370 (0x20B2)
    ERROR_DS_OBJ_CLASS_NOT_DEFINED: Final[ThirtyTwoBits] = cached_win32error(0x20B3, 'ERROR_DS_OBJ_CLASS_NOT_DEFINED') # WinError 8371 (0x20B3)
    ERROR_DS_OBJ_CLASS_NOT_SUBCLASS: Final[ThirtyTwoBits] = cached_win32error(0x20B4, 'ERROR_DS_OBJ_CLASS_NOT_SUBCLASS') # WinError 8372 (0x20B4)
    ERROR_DS_NAME_REFERENCE_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x20B5, 'ERROR_DS_NAME_REFERENCE_INVALID') # WinError 8373 (0x20B5)
    ERROR_DS_CROSS_REF_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x20B6, 'ERROR_DS_CROSS_REF_EXISTS') # WinError 8374 (0x20B6)
    ERROR_DS_CANT_DEL_MASTER_CROSSREF: Final[ThirtyTwoBits] = cached_win32error(0x20B7, 'ERROR_DS_CANT_DEL_MASTER_CROSSREF') # WinError 8375 (0x20B7)
    ERROR_DS_SUBTREE_NOTIFY_NOT_NC_HEAD: Final[ThirtyTwoBits] = cached_win32error(0x20B8, 'ERROR_DS_SUBTREE_NOTIFY_NOT_NC_HEAD') # WinError 8376 (0x20B8)
    ERROR_DS_NOTIFY_FILTER_TOO_COMPLEX: Final[ThirtyTwoBits] = cached_win32error(0x20B9, 'ERROR_DS_NOTIFY_FILTER_TOO_COMPLEX') # WinError 8377 (0x20B9)
    ERROR_DS_DUP_RDN: Final[ThirtyTwoBits] = cached_win32error(0x20BA, 'ERROR_DS_DUP_RDN') # WinError 8378 (0x20BA)
    ERROR_DS_DUP_OID: Final[ThirtyTwoBits] = cached_win32error(0x20BB, 'ERROR_DS_DUP_OID') # WinError 8379 (0x20BB)
    ERROR_DS_DUP_MAPI_ID: Final[ThirtyTwoBits] = cached_win32error(0x20BC, 'ERROR_DS_DUP_MAPI_ID') # WinError 8380 (0x20BC)
    ERROR_DS_DUP_SCHEMA_ID_GUID: Final[ThirtyTwoBits] = cached_win32error(0x20BD, 'ERROR_DS_DUP_SCHEMA_ID_GUID') # WinError 8381 (0x20BD)
    ERROR_DS_DUP_LDAP_DISPLAY_NAME: Final[ThirtyTwoBits] = cached_win32error(0x20BE, 'ERROR_DS_DUP_LDAP_DISPLAY_NAME') # WinError 8382 (0x20BE)
    ERROR_DS_SEMANTIC_ATT_TEST: Final[ThirtyTwoBits] = cached_win32error(0x20BF, 'ERROR_DS_SEMANTIC_ATT_TEST') # WinError 8383 (0x20BF)
    ERROR_DS_SYNTAX_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x20C0, 'ERROR_DS_SYNTAX_MISMATCH') # WinError 8384 (0x20C0)
    ERROR_DS_EXISTS_IN_MUST_HAVE: Final[ThirtyTwoBits] = cached_win32error(0x20C1, 'ERROR_DS_EXISTS_IN_MUST_HAVE') # WinError 8385 (0x20C1)
    ERROR_DS_EXISTS_IN_MAY_HAVE: Final[ThirtyTwoBits] = cached_win32error(0x20C2, 'ERROR_DS_EXISTS_IN_MAY_HAVE') # WinError 8386 (0x20C2)
    ERROR_DS_NONEXISTENT_MAY_HAVE: Final[ThirtyTwoBits] = cached_win32error(0x20C3, 'ERROR_DS_NONEXISTENT_MAY_HAVE') # WinError 8387 (0x20C3)
    ERROR_DS_NONEXISTENT_MUST_HAVE: Final[ThirtyTwoBits] = cached_win32error(0x20C4, 'ERROR_DS_NONEXISTENT_MUST_HAVE') # WinError 8388 (0x20C4)
    ERROR_DS_AUX_CLS_TEST_FAIL: Final[ThirtyTwoBits] = cached_win32error(0x20C5, 'ERROR_DS_AUX_CLS_TEST_FAIL') # WinError 8389 (0x20C5)
    ERROR_DS_NONEXISTENT_POSS_SUP: Final[ThirtyTwoBits] = cached_win32error(0x20C6, 'ERROR_DS_NONEXISTENT_POSS_SUP') # WinError 8390 (0x20C6)
    ERROR_DS_SUB_CLS_TEST_FAIL: Final[ThirtyTwoBits] = cached_win32error(0x20C7, 'ERROR_DS_SUB_CLS_TEST_FAIL') # WinError 8391 (0x20C7)
    ERROR_DS_BAD_RDN_ATT_ID_SYNTAX: Final[ThirtyTwoBits] = cached_win32error(0x20C8, 'ERROR_DS_BAD_RDN_ATT_ID_SYNTAX') # WinError 8392 (0x20C8)
    ERROR_DS_EXISTS_IN_AUX_CLS: Final[ThirtyTwoBits] = cached_win32error(0x20C9, 'ERROR_DS_EXISTS_IN_AUX_CLS') # WinError 8393 (0x20C9)
    ERROR_DS_EXISTS_IN_SUB_CLS: Final[ThirtyTwoBits] = cached_win32error(0x20CA, 'ERROR_DS_EXISTS_IN_SUB_CLS') # WinError 8394 (0x20CA)
    ERROR_DS_EXISTS_IN_POSS_SUP: Final[ThirtyTwoBits] = cached_win32error(0x20CB, 'ERROR_DS_EXISTS_IN_POSS_SUP') # WinError 8395 (0x20CB)
    ERROR_DS_RECALCSCHEMA_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x20CC, 'ERROR_DS_RECALCSCHEMA_FAILED') # WinError 8396 (0x20CC)
    ERROR_DS_TREE_DELETE_NOT_FINISHED: Final[ThirtyTwoBits] = cached_win32error(0x20CD, 'ERROR_DS_TREE_DELETE_NOT_FINISHED') # WinError 8397 (0x20CD)
    ERROR_DS_CANT_DELETE: Final[ThirtyTwoBits] = cached_win32error(0x20CE, 'ERROR_DS_CANT_DELETE') # WinError 8398 (0x20CE)
    ERROR_DS_ATT_SCHEMA_REQ_ID: Final[ThirtyTwoBits] = cached_win32error(0x20CF, 'ERROR_DS_ATT_SCHEMA_REQ_ID') # WinError 8399 (0x20CF)
    ERROR_DS_BAD_ATT_SCHEMA_SYNTAX: Final[ThirtyTwoBits] = cached_win32error(0x20D0, 'ERROR_DS_BAD_ATT_SCHEMA_SYNTAX') # WinError 8400 (0x20D0)
    ERROR_DS_CANT_CACHE_ATT: Final[ThirtyTwoBits] = cached_win32error(0x20D1, 'ERROR_DS_CANT_CACHE_ATT') # WinError 8401 (0x20D1)
    ERROR_DS_CANT_CACHE_CLASS: Final[ThirtyTwoBits] = cached_win32error(0x20D2, 'ERROR_DS_CANT_CACHE_CLASS') # WinError 8402 (0x20D2)
    ERROR_DS_CANT_REMOVE_ATT_CACHE: Final[ThirtyTwoBits] = cached_win32error(0x20D3, 'ERROR_DS_CANT_REMOVE_ATT_CACHE') # WinError 8403 (0x20D3)
    ERROR_DS_CANT_REMOVE_CLASS_CACHE: Final[ThirtyTwoBits] = cached_win32error(0x20D4, 'ERROR_DS_CANT_REMOVE_CLASS_CACHE') # WinError 8404 (0x20D4)
    ERROR_DS_CANT_RETRIEVE_DN: Final[ThirtyTwoBits] = cached_win32error(0x20D5, 'ERROR_DS_CANT_RETRIEVE_DN') # WinError 8405 (0x20D5)
    ERROR_DS_MISSING_SUPREF: Final[ThirtyTwoBits] = cached_win32error(0x20D6, 'ERROR_DS_MISSING_SUPREF') # WinError 8406 (0x20D6)
    ERROR_DS_CANT_RETRIEVE_INSTANCE: Final[ThirtyTwoBits] = cached_win32error(0x20D7, 'ERROR_DS_CANT_RETRIEVE_INSTANCE') # WinError 8407 (0x20D7)
    ERROR_DS_CODE_INCONSISTENCY: Final[ThirtyTwoBits] = cached_win32error(0x20D8, 'ERROR_DS_CODE_INCONSISTENCY') # WinError 8408 (0x20D8)
    ERROR_DS_DATABASE_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x20D9, 'ERROR_DS_DATABASE_ERROR') # WinError 8409 (0x20D9)
    ERROR_DS_GOVERNSID_MISSING: Final[ThirtyTwoBits] = cached_win32error(0x20DA, 'ERROR_DS_GOVERNSID_MISSING') # WinError 8410 (0x20DA)
    ERROR_DS_MISSING_EXPECTED_ATT: Final[ThirtyTwoBits] = cached_win32error(0x20DB, 'ERROR_DS_MISSING_EXPECTED_ATT') # WinError 8411 (0x20DB)
    ERROR_DS_NCNAME_MISSING_CR_REF: Final[ThirtyTwoBits] = cached_win32error(0x20DC, 'ERROR_DS_NCNAME_MISSING_CR_REF') # WinError 8412 (0x20DC)
    ERROR_DS_SECURITY_CHECKING_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x20DD, 'ERROR_DS_SECURITY_CHECKING_ERROR') # WinError 8413 (0x20DD)
    ERROR_DS_SCHEMA_NOT_LOADED: Final[ThirtyTwoBits] = cached_win32error(0x20DE, 'ERROR_DS_SCHEMA_NOT_LOADED') # WinError 8414 (0x20DE)
    ERROR_DS_SCHEMA_ALLOC_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x20DF, 'ERROR_DS_SCHEMA_ALLOC_FAILED') # WinError 8415 (0x20DF)
    ERROR_DS_ATT_SCHEMA_REQ_SYNTAX: Final[ThirtyTwoBits] = cached_win32error(0x20E0, 'ERROR_DS_ATT_SCHEMA_REQ_SYNTAX') # WinError 8416 (0x20E0)
    ERROR_DS_GCVERIFY_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x20E1, 'ERROR_DS_GCVERIFY_ERROR') # WinError 8417 (0x20E1)
    ERROR_DS_DRA_SCHEMA_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x20E2, 'ERROR_DS_DRA_SCHEMA_MISMATCH') # WinError 8418 (0x20E2)
    ERROR_DS_CANT_FIND_DSA_OBJ: Final[ThirtyTwoBits] = cached_win32error(0x20E3, 'ERROR_DS_CANT_FIND_DSA_OBJ') # WinError 8419 (0x20E3)
    ERROR_DS_CANT_FIND_EXPECTED_NC: Final[ThirtyTwoBits] = cached_win32error(0x20E4, 'ERROR_DS_CANT_FIND_EXPECTED_NC') # WinError 8420 (0x20E4)
    ERROR_DS_CANT_FIND_NC_IN_CACHE: Final[ThirtyTwoBits] = cached_win32error(0x20E5, 'ERROR_DS_CANT_FIND_NC_IN_CACHE') # WinError 8421 (0x20E5)
    ERROR_DS_CANT_RETRIEVE_CHILD: Final[ThirtyTwoBits] = cached_win32error(0x20E6, 'ERROR_DS_CANT_RETRIEVE_CHILD') # WinError 8422 (0x20E6)
    ERROR_DS_SECURITY_ILLEGAL_MODIFY: Final[ThirtyTwoBits] = cached_win32error(0x20E7, 'ERROR_DS_SECURITY_ILLEGAL_MODIFY') # WinError 8423 (0x20E7)
    ERROR_DS_CANT_REPLACE_HIDDEN_REC: Final[ThirtyTwoBits] = cached_win32error(0x20E8, 'ERROR_DS_CANT_REPLACE_HIDDEN_REC') # WinError 8424 (0x20E8)
    ERROR_DS_BAD_HIERARCHY_FILE: Final[ThirtyTwoBits] = cached_win32error(0x20E9, 'ERROR_DS_BAD_HIERARCHY_FILE') # WinError 8425 (0x20E9)
    ERROR_DS_BUILD_HIERARCHY_TABLE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x20EA, 'ERROR_DS_BUILD_HIERARCHY_TABLE_FAILED') # WinError 8426 (0x20EA)
    ERROR_DS_CONFIG_PARAM_MISSING: Final[ThirtyTwoBits] = cached_win32error(0x20EB, 'ERROR_DS_CONFIG_PARAM_MISSING') # WinError 8427 (0x20EB)
    ERROR_DS_COUNTING_AB_INDICES_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x20EC, 'ERROR_DS_COUNTING_AB_INDICES_FAILED') # WinError 8428 (0x20EC)
    ERROR_DS_HIERARCHY_TABLE_MALLOC_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x20ED, 'ERROR_DS_HIERARCHY_TABLE_MALLOC_FAILED') # WinError 8429 (0x20ED)
    ERROR_DS_INTERNAL_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x20EE, 'ERROR_DS_INTERNAL_FAILURE') # WinError 8430 (0x20EE)
    ERROR_DS_UNKNOWN_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x20EF, 'ERROR_DS_UNKNOWN_ERROR') # WinError 8431 (0x20EF)
    ERROR_DS_ROOT_REQUIRES_CLASS_TOP: Final[ThirtyTwoBits] = cached_win32error(0x20F0, 'ERROR_DS_ROOT_REQUIRES_CLASS_TOP') # WinError 8432 (0x20F0)
    ERROR_DS_REFUSING_FSMO_ROLES: Final[ThirtyTwoBits] = cached_win32error(0x20F1, 'ERROR_DS_REFUSING_FSMO_ROLES') # WinError 8433 (0x20F1)
    ERROR_DS_MISSING_FSMO_SETTINGS: Final[ThirtyTwoBits] = cached_win32error(0x20F2, 'ERROR_DS_MISSING_FSMO_SETTINGS') # WinError 8434 (0x20F2)
    ERROR_DS_UNABLE_TO_SURRENDER_ROLES: Final[ThirtyTwoBits] = cached_win32error(0x20F3, 'ERROR_DS_UNABLE_TO_SURRENDER_ROLES') # WinError 8435 (0x20F3)
    ERROR_DS_DRA_GENERIC: Final[ThirtyTwoBits] = cached_win32error(0x20F4, 'ERROR_DS_DRA_GENERIC') # WinError 8436 (0x20F4)
    ERROR_DS_DRA_INVALID_PARAMETER: Final[ThirtyTwoBits] = cached_win32error(0x20F5, 'ERROR_DS_DRA_INVALID_PARAMETER') # WinError 8437 (0x20F5)
    ERROR_DS_DRA_BUSY: Final[ThirtyTwoBits] = cached_win32error(0x20F6, 'ERROR_DS_DRA_BUSY') # WinError 8438 (0x20F6)
    ERROR_DS_DRA_BAD_DN: Final[ThirtyTwoBits] = cached_win32error(0x20F7, 'ERROR_DS_DRA_BAD_DN') # WinError 8439 (0x20F7)
    ERROR_DS_DRA_BAD_NC: Final[ThirtyTwoBits] = cached_win32error(0x20F8, 'ERROR_DS_DRA_BAD_NC') # WinError 8440 (0x20F8)
    ERROR_DS_DRA_DN_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x20F9, 'ERROR_DS_DRA_DN_EXISTS') # WinError 8441 (0x20F9)
    ERROR_DS_DRA_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x20FA, 'ERROR_DS_DRA_INTERNAL_ERROR') # WinError 8442 (0x20FA)
    ERROR_DS_DRA_INCONSISTENT_DIT: Final[ThirtyTwoBits] = cached_win32error(0x20FB, 'ERROR_DS_DRA_INCONSISTENT_DIT') # WinError 8443 (0x20FB)
    ERROR_DS_DRA_CONNECTION_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x20FC, 'ERROR_DS_DRA_CONNECTION_FAILED') # WinError 8444 (0x20FC)
    ERROR_DS_DRA_BAD_INSTANCE_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x20FD, 'ERROR_DS_DRA_BAD_INSTANCE_TYPE') # WinError 8445 (0x20FD)
    ERROR_DS_DRA_OUT_OF_MEM: Final[ThirtyTwoBits] = cached_win32error(0x20FE, 'ERROR_DS_DRA_OUT_OF_MEM') # WinError 8446 (0x20FE)
    ERROR_DS_DRA_MAIL_PROBLEM: Final[ThirtyTwoBits] = cached_win32error(0x20FF, 'ERROR_DS_DRA_MAIL_PROBLEM') # WinError 8447 (0x20FF)
    ERROR_DS_DRA_REF_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x2100, 'ERROR_DS_DRA_REF_ALREADY_EXISTS') # WinError 8448 (0x2100)
    ERROR_DS_DRA_REF_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x2101, 'ERROR_DS_DRA_REF_NOT_FOUND') # WinError 8449 (0x2101)
    ERROR_DS_DRA_OBJ_IS_REP_SOURCE: Final[ThirtyTwoBits] = cached_win32error(0x2102, 'ERROR_DS_DRA_OBJ_IS_REP_SOURCE') # WinError 8450 (0x2102)
    ERROR_DS_DRA_DB_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x2103, 'ERROR_DS_DRA_DB_ERROR') # WinError 8451 (0x2103)
    ERROR_DS_DRA_NO_REPLICA: Final[ThirtyTwoBits] = cached_win32error(0x2104, 'ERROR_DS_DRA_NO_REPLICA') # WinError 8452 (0x2104)
    ERROR_DS_DRA_ACCESS_DENIED: Final[ThirtyTwoBits] = cached_win32error(0x2105, 'ERROR_DS_DRA_ACCESS_DENIED') # WinError 8453 (0x2105)
    ERROR_DS_DRA_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x2106, 'ERROR_DS_DRA_NOT_SUPPORTED') # WinError 8454 (0x2106)
    ERROR_DS_DRA_RPC_CANCELLED: Final[ThirtyTwoBits] = cached_win32error(0x2107, 'ERROR_DS_DRA_RPC_CANCELLED') # WinError 8455 (0x2107)
    ERROR_DS_DRA_SOURCE_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x2108, 'ERROR_DS_DRA_SOURCE_DISABLED') # WinError 8456 (0x2108)
    ERROR_DS_DRA_SINK_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x2109, 'ERROR_DS_DRA_SINK_DISABLED') # WinError 8457 (0x2109)
    ERROR_DS_DRA_NAME_COLLISION: Final[ThirtyTwoBits] = cached_win32error(0x210A, 'ERROR_DS_DRA_NAME_COLLISION') # WinError 8458 (0x210A)
    ERROR_DS_DRA_SOURCE_REINSTALLED: Final[ThirtyTwoBits] = cached_win32error(0x210B, 'ERROR_DS_DRA_SOURCE_REINSTALLED') # WinError 8459 (0x210B)
    ERROR_DS_DRA_MISSING_PARENT: Final[ThirtyTwoBits] = cached_win32error(0x210C, 'ERROR_DS_DRA_MISSING_PARENT') # WinError 8460 (0x210C)
    ERROR_DS_DRA_PREEMPTED: Final[ThirtyTwoBits] = cached_win32error(0x210D, 'ERROR_DS_DRA_PREEMPTED') # WinError 8461 (0x210D)
    ERROR_DS_DRA_ABANDON_SYNC: Final[ThirtyTwoBits] = cached_win32error(0x210E, 'ERROR_DS_DRA_ABANDON_SYNC') # WinError 8462 (0x210E)
    ERROR_DS_DRA_SHUTDOWN: Final[ThirtyTwoBits] = cached_win32error(0x210F, 'ERROR_DS_DRA_SHUTDOWN') # WinError 8463 (0x210F)
    ERROR_DS_DRA_INCOMPATIBLE_PARTIAL_SET: Final[ThirtyTwoBits] = cached_win32error(0x2110, 'ERROR_DS_DRA_INCOMPATIBLE_PARTIAL_SET') # WinError 8464 (0x2110)
    ERROR_DS_DRA_SOURCE_IS_PARTIAL_REPLICA: Final[ThirtyTwoBits] = cached_win32error(0x2111, 'ERROR_DS_DRA_SOURCE_IS_PARTIAL_REPLICA') # WinError 8465 (0x2111)
    ERROR_DS_DRA_EXTN_CONNECTION_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x2112, 'ERROR_DS_DRA_EXTN_CONNECTION_FAILED') # WinError 8466 (0x2112)
    ERROR_DS_INSTALL_SCHEMA_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x2113, 'ERROR_DS_INSTALL_SCHEMA_MISMATCH') # WinError 8467 (0x2113)
    ERROR_DS_DUP_LINK_ID: Final[ThirtyTwoBits] = cached_win32error(0x2114, 'ERROR_DS_DUP_LINK_ID') # WinError 8468 (0x2114)
    ERROR_DS_NAME_ERROR_RESOLVING: Final[ThirtyTwoBits] = cached_win32error(0x2115, 'ERROR_DS_NAME_ERROR_RESOLVING') # WinError 8469 (0x2115)
    ERROR_DS_NAME_ERROR_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x2116, 'ERROR_DS_NAME_ERROR_NOT_FOUND') # WinError 8470 (0x2116)
    ERROR_DS_NAME_ERROR_NOT_UNIQUE: Final[ThirtyTwoBits] = cached_win32error(0x2117, 'ERROR_DS_NAME_ERROR_NOT_UNIQUE') # WinError 8471 (0x2117)
    ERROR_DS_NAME_ERROR_NO_MAPPING: Final[ThirtyTwoBits] = cached_win32error(0x2118, 'ERROR_DS_NAME_ERROR_NO_MAPPING') # WinError 8472 (0x2118)
    ERROR_DS_NAME_ERROR_DOMAIN_ONLY: Final[ThirtyTwoBits] = cached_win32error(0x2119, 'ERROR_DS_NAME_ERROR_DOMAIN_ONLY') # WinError 8473 (0x2119)
    ERROR_DS_NAME_ERROR_NO_SYNTACTICAL_MAPPING: Final[ThirtyTwoBits] = cached_win32error(0x211A, 'ERROR_DS_NAME_ERROR_NO_SYNTACTICAL_MAPPING') # WinError 8474 (0x211A)
    ERROR_DS_CONSTRUCTED_ATT_MOD: Final[ThirtyTwoBits] = cached_win32error(0x211B, 'ERROR_DS_CONSTRUCTED_ATT_MOD') # WinError 8475 (0x211B)
    ERROR_DS_WRONG_OM_OBJ_CLASS: Final[ThirtyTwoBits] = cached_win32error(0x211C, 'ERROR_DS_WRONG_OM_OBJ_CLASS') # WinError 8476 (0x211C)
    ERROR_DS_DRA_REPL_PENDING: Final[ThirtyTwoBits] = cached_win32error(0x211D, 'ERROR_DS_DRA_REPL_PENDING') # WinError 8477 (0x211D)
    ERROR_DS_DS_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x211E, 'ERROR_DS_DS_REQUIRED') # WinError 8478 (0x211E)
    ERROR_DS_INVALID_LDAP_DISPLAY_NAME: Final[ThirtyTwoBits] = cached_win32error(0x211F, 'ERROR_DS_INVALID_LDAP_DISPLAY_NAME') # WinError 8479 (0x211F)
    ERROR_DS_NON_BASE_SEARCH: Final[ThirtyTwoBits] = cached_win32error(0x2120, 'ERROR_DS_NON_BASE_SEARCH') # WinError 8480 (0x2120)
    ERROR_DS_CANT_RETRIEVE_ATTS: Final[ThirtyTwoBits] = cached_win32error(0x2121, 'ERROR_DS_CANT_RETRIEVE_ATTS') # WinError 8481 (0x2121)
    ERROR_DS_BACKLINK_WITHOUT_LINK: Final[ThirtyTwoBits] = cached_win32error(0x2122, 'ERROR_DS_BACKLINK_WITHOUT_LINK') # WinError 8482 (0x2122)
    ERROR_DS_EPOCH_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x2123, 'ERROR_DS_EPOCH_MISMATCH') # WinError 8483 (0x2123)
    ERROR_DS_SRC_NAME_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x2124, 'ERROR_DS_SRC_NAME_MISMATCH') # WinError 8484 (0x2124)
    ERROR_DS_SRC_AND_DST_NC_IDENTICAL: Final[ThirtyTwoBits] = cached_win32error(0x2125, 'ERROR_DS_SRC_AND_DST_NC_IDENTICAL') # WinError 8485 (0x2125)
    ERROR_DS_DST_NC_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x2126, 'ERROR_DS_DST_NC_MISMATCH') # WinError 8486 (0x2126)
    ERROR_DS_NOT_AUTHORITIVE_FOR_DST_NC: Final[ThirtyTwoBits] = cached_win32error(0x2127, 'ERROR_DS_NOT_AUTHORITIVE_FOR_DST_NC') # WinError 8487 (0x2127)
    ERROR_DS_SRC_GUID_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x2128, 'ERROR_DS_SRC_GUID_MISMATCH') # WinError 8488 (0x2128)
    ERROR_DS_CANT_MOVE_DELETED_OBJECT: Final[ThirtyTwoBits] = cached_win32error(0x2129, 'ERROR_DS_CANT_MOVE_DELETED_OBJECT') # WinError 8489 (0x2129)
    ERROR_DS_PDC_OPERATION_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x212A, 'ERROR_DS_PDC_OPERATION_IN_PROGRESS') # WinError 8490 (0x212A)
    ERROR_DS_CROSS_DOMAIN_CLEANUP_REQD: Final[ThirtyTwoBits] = cached_win32error(0x212B, 'ERROR_DS_CROSS_DOMAIN_CLEANUP_REQD') # WinError 8491 (0x212B)
    ERROR_DS_ILLEGAL_XDOM_MOVE_OPERATION: Final[ThirtyTwoBits] = cached_win32error(0x212C, 'ERROR_DS_ILLEGAL_XDOM_MOVE_OPERATION') # WinError 8492 (0x212C)
    ERROR_DS_CANT_WITH_ACCT_GROUP_MEMBERSHPS: Final[ThirtyTwoBits] = cached_win32error(0x212D, 'ERROR_DS_CANT_WITH_ACCT_GROUP_MEMBERSHPS') # WinError 8493 (0x212D)
    ERROR_DS_NC_MUST_HAVE_NC_PARENT: Final[ThirtyTwoBits] = cached_win32error(0x212E, 'ERROR_DS_NC_MUST_HAVE_NC_PARENT') # WinError 8494 (0x212E)
    ERROR_DS_CR_IMPOSSIBLE_TO_VALIDATE: Final[ThirtyTwoBits] = cached_win32error(0x212F, 'ERROR_DS_CR_IMPOSSIBLE_TO_VALIDATE') # WinError 8495 (0x212F)
    ERROR_DS_DST_DOMAIN_NOT_NATIVE: Final[ThirtyTwoBits] = cached_win32error(0x2130, 'ERROR_DS_DST_DOMAIN_NOT_NATIVE') # WinError 8496 (0x2130)
    ERROR_DS_MISSING_INFRASTRUCTURE_CONTAINER: Final[ThirtyTwoBits] = cached_win32error(0x2131, 'ERROR_DS_MISSING_INFRASTRUCTURE_CONTAINER') # WinError 8497 (0x2131)
    ERROR_DS_CANT_MOVE_ACCOUNT_GROUP: Final[ThirtyTwoBits] = cached_win32error(0x2132, 'ERROR_DS_CANT_MOVE_ACCOUNT_GROUP') # WinError 8498 (0x2132)
    ERROR_DS_CANT_MOVE_RESOURCE_GROUP: Final[ThirtyTwoBits] = cached_win32error(0x2133, 'ERROR_DS_CANT_MOVE_RESOURCE_GROUP') # WinError 8499 (0x2133)
    ERROR_DS_INVALID_SEARCH_FLAG: Final[ThirtyTwoBits] = cached_win32error(0x2134, 'ERROR_DS_INVALID_SEARCH_FLAG') # WinError 8500 (0x2134)
    ERROR_DS_NO_TREE_DELETE_ABOVE_NC: Final[ThirtyTwoBits] = cached_win32error(0x2135, 'ERROR_DS_NO_TREE_DELETE_ABOVE_NC') # WinError 8501 (0x2135)
    ERROR_DS_COULDNT_LOCK_TREE_FOR_DELETE: Final[ThirtyTwoBits] = cached_win32error(0x2136, 'ERROR_DS_COULDNT_LOCK_TREE_FOR_DELETE') # WinError 8502 (0x2136)
    ERROR_DS_COULDNT_IDENTIFY_OBJECTS_FOR_TREE_DELETE: Final[ThirtyTwoBits] = cached_win32error(0x2137, 'ERROR_DS_COULDNT_IDENTIFY_OBJECTS_FOR_TREE_DELETE') # WinError 8503 (0x2137)
    ERROR_DS_SAM_INIT_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x2138, 'ERROR_DS_SAM_INIT_FAILURE') # WinError 8504 (0x2138)
    ERROR_DS_SENSITIVE_GROUP_VIOLATION: Final[ThirtyTwoBits] = cached_win32error(0x2139, 'ERROR_DS_SENSITIVE_GROUP_VIOLATION') # WinError 8505 (0x2139)
    ERROR_DS_CANT_MOD_PRIMARYGROUPID: Final[ThirtyTwoBits] = cached_win32error(0x213A, 'ERROR_DS_CANT_MOD_PRIMARYGROUPID') # WinError 8506 (0x213A)
    ERROR_DS_ILLEGAL_BASE_SCHEMA_MOD: Final[ThirtyTwoBits] = cached_win32error(0x213B, 'ERROR_DS_ILLEGAL_BASE_SCHEMA_MOD') # WinError 8507 (0x213B)
    ERROR_DS_NONSAFE_SCHEMA_CHANGE: Final[ThirtyTwoBits] = cached_win32error(0x213C, 'ERROR_DS_NONSAFE_SCHEMA_CHANGE') # WinError 8508 (0x213C)
    ERROR_DS_SCHEMA_UPDATE_DISALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x213D, 'ERROR_DS_SCHEMA_UPDATE_DISALLOWED') # WinError 8509 (0x213D)
    ERROR_DS_CANT_CREATE_UNDER_SCHEMA: Final[ThirtyTwoBits] = cached_win32error(0x213E, 'ERROR_DS_CANT_CREATE_UNDER_SCHEMA') # WinError 8510 (0x213E)
    ERROR_DS_INSTALL_NO_SRC_SCH_VERSION: Final[ThirtyTwoBits] = cached_win32error(0x213F, 'ERROR_DS_INSTALL_NO_SRC_SCH_VERSION') # WinError 8511 (0x213F)
    ERROR_DS_INSTALL_NO_SCH_VERSION_IN_INIFILE: Final[ThirtyTwoBits] = cached_win32error(0x2140, 'ERROR_DS_INSTALL_NO_SCH_VERSION_IN_INIFILE') # WinError 8512 (0x2140)
    ERROR_DS_INVALID_GROUP_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x2141, 'ERROR_DS_INVALID_GROUP_TYPE') # WinError 8513 (0x2141)
    ERROR_DS_NO_NEST_GLOBALGROUP_IN_MIXEDDOMAIN: Final[ThirtyTwoBits] = cached_win32error(0x2142, 'ERROR_DS_NO_NEST_GLOBALGROUP_IN_MIXEDDOMAIN') # WinError 8514 (0x2142)
    ERROR_DS_NO_NEST_LOCALGROUP_IN_MIXEDDOMAIN: Final[ThirtyTwoBits] = cached_win32error(0x2143, 'ERROR_DS_NO_NEST_LOCALGROUP_IN_MIXEDDOMAIN') # WinError 8515 (0x2143)
    ERROR_DS_GLOBAL_CANT_HAVE_LOCAL_MEMBER: Final[ThirtyTwoBits] = cached_win32error(0x2144, 'ERROR_DS_GLOBAL_CANT_HAVE_LOCAL_MEMBER') # WinError 8516 (0x2144)
    ERROR_DS_GLOBAL_CANT_HAVE_UNIVERSAL_MEMBER: Final[ThirtyTwoBits] = cached_win32error(0x2145, 'ERROR_DS_GLOBAL_CANT_HAVE_UNIVERSAL_MEMBER') # WinError 8517 (0x2145)
    ERROR_DS_UNIVERSAL_CANT_HAVE_LOCAL_MEMBER: Final[ThirtyTwoBits] = cached_win32error(0x2146, 'ERROR_DS_UNIVERSAL_CANT_HAVE_LOCAL_MEMBER') # WinError 8518 (0x2146)
    ERROR_DS_GLOBAL_CANT_HAVE_CROSSDOMAIN_MEMBER: Final[ThirtyTwoBits] = cached_win32error(0x2147, 'ERROR_DS_GLOBAL_CANT_HAVE_CROSSDOMAIN_MEMBER') # WinError 8519 (0x2147)
    ERROR_DS_LOCAL_CANT_HAVE_CROSSDOMAIN_LOCAL_MEMBER: Final[ThirtyTwoBits] = cached_win32error(0x2148, 'ERROR_DS_LOCAL_CANT_HAVE_CROSSDOMAIN_LOCAL_MEMBER') # WinError 8520 (0x2148)
    ERROR_DS_HAVE_PRIMARY_MEMBERS: Final[ThirtyTwoBits] = cached_win32error(0x2149, 'ERROR_DS_HAVE_PRIMARY_MEMBERS') # WinError 8521 (0x2149)
    ERROR_DS_STRING_SD_CONVERSION_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x214A, 'ERROR_DS_STRING_SD_CONVERSION_FAILED') # WinError 8522 (0x214A)
    ERROR_DS_NAMING_MASTER_GC: Final[ThirtyTwoBits] = cached_win32error(0x214B, 'ERROR_DS_NAMING_MASTER_GC') # WinError 8523 (0x214B)
    ERROR_DS_DNS_LOOKUP_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x214C, 'ERROR_DS_DNS_LOOKUP_FAILURE') # WinError 8524 (0x214C)
    ERROR_DS_COULDNT_UPDATE_SPNS: Final[ThirtyTwoBits] = cached_win32error(0x214D, 'ERROR_DS_COULDNT_UPDATE_SPNS') # WinError 8525 (0x214D)
    ERROR_DS_CANT_RETRIEVE_SD: Final[ThirtyTwoBits] = cached_win32error(0x214E, 'ERROR_DS_CANT_RETRIEVE_SD') # WinError 8526 (0x214E)
    ERROR_DS_KEY_NOT_UNIQUE: Final[ThirtyTwoBits] = cached_win32error(0x214F, 'ERROR_DS_KEY_NOT_UNIQUE') # WinError 8527 (0x214F)
    ERROR_DS_WRONG_LINKED_ATT_SYNTAX: Final[ThirtyTwoBits] = cached_win32error(0x2150, 'ERROR_DS_WRONG_LINKED_ATT_SYNTAX') # WinError 8528 (0x2150)
    ERROR_DS_SAM_NEED_BOOTKEY_PASSWORD: Final[ThirtyTwoBits] = cached_win32error(0x2151, 'ERROR_DS_SAM_NEED_BOOTKEY_PASSWORD') # WinError 8529 (0x2151)
    ERROR_DS_SAM_NEED_BOOTKEY_FLOPPY: Final[ThirtyTwoBits] = cached_win32error(0x2152, 'ERROR_DS_SAM_NEED_BOOTKEY_FLOPPY') # WinError 8530 (0x2152)
    ERROR_DS_CANT_START: Final[ThirtyTwoBits] = cached_win32error(0x2153, 'ERROR_DS_CANT_START') # WinError 8531 (0x2153)
    ERROR_DS_INIT_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x2154, 'ERROR_DS_INIT_FAILURE') # WinError 8532 (0x2154)
    ERROR_DS_NO_PKT_PRIVACY_ON_CONNECTION: Final[ThirtyTwoBits] = cached_win32error(0x2155, 'ERROR_DS_NO_PKT_PRIVACY_ON_CONNECTION') # WinError 8533 (0x2155)
    ERROR_DS_SOURCE_DOMAIN_IN_FOREST: Final[ThirtyTwoBits] = cached_win32error(0x2156, 'ERROR_DS_SOURCE_DOMAIN_IN_FOREST') # WinError 8534 (0x2156)
    ERROR_DS_DESTINATION_DOMAIN_NOT_IN_FOREST: Final[ThirtyTwoBits] = cached_win32error(0x2157, 'ERROR_DS_DESTINATION_DOMAIN_NOT_IN_FOREST') # WinError 8535 (0x2157)
    ERROR_DS_DESTINATION_AUDITING_NOT_ENABLED: Final[ThirtyTwoBits] = cached_win32error(0x2158, 'ERROR_DS_DESTINATION_AUDITING_NOT_ENABLED') # WinError 8536 (0x2158)
    ERROR_DS_CANT_FIND_DC_FOR_SRC_DOMAIN: Final[ThirtyTwoBits] = cached_win32error(0x2159, 'ERROR_DS_CANT_FIND_DC_FOR_SRC_DOMAIN') # WinError 8537 (0x2159)
    ERROR_DS_SRC_OBJ_NOT_GROUP_OR_USER: Final[ThirtyTwoBits] = cached_win32error(0x215A, 'ERROR_DS_SRC_OBJ_NOT_GROUP_OR_USER') # WinError 8538 (0x215A)
    ERROR_DS_SRC_SID_EXISTS_IN_FOREST: Final[ThirtyTwoBits] = cached_win32error(0x215B, 'ERROR_DS_SRC_SID_EXISTS_IN_FOREST') # WinError 8539 (0x215B)
    ERROR_DS_SRC_AND_DST_OBJECT_CLASS_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x215C, 'ERROR_DS_SRC_AND_DST_OBJECT_CLASS_MISMATCH') # WinError 8540 (0x215C)
    ERROR_SAM_INIT_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x215D, 'ERROR_SAM_INIT_FAILURE') # WinError 8541 (0x215D)
    ERROR_DS_DRA_SCHEMA_INFO_SHIP: Final[ThirtyTwoBits] = cached_win32error(0x215E, 'ERROR_DS_DRA_SCHEMA_INFO_SHIP') # WinError 8542 (0x215E)
    ERROR_DS_DRA_SCHEMA_CONFLICT: Final[ThirtyTwoBits] = cached_win32error(0x215F, 'ERROR_DS_DRA_SCHEMA_CONFLICT') # WinError 8543 (0x215F)
    ERROR_DS_DRA_EARLIER_SCHEMA_CONFLICT: Final[ThirtyTwoBits] = cached_win32error(0x2160, 'ERROR_DS_DRA_EARLIER_SCHEMA_CONFLICT') # WinError 8544 (0x2160)
    ERROR_DS_DRA_OBJ_NC_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x2161, 'ERROR_DS_DRA_OBJ_NC_MISMATCH') # WinError 8545 (0x2161)
    ERROR_DS_NC_STILL_HAS_DSAS: Final[ThirtyTwoBits] = cached_win32error(0x2162, 'ERROR_DS_NC_STILL_HAS_DSAS') # WinError 8546 (0x2162)
    ERROR_DS_GC_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x2163, 'ERROR_DS_GC_REQUIRED') # WinError 8547 (0x2163)
    ERROR_DS_LOCAL_MEMBER_OF_LOCAL_ONLY: Final[ThirtyTwoBits] = cached_win32error(0x2164, 'ERROR_DS_LOCAL_MEMBER_OF_LOCAL_ONLY') # WinError 8548 (0x2164)
    ERROR_DS_NO_FPO_IN_UNIVERSAL_GROUPS: Final[ThirtyTwoBits] = cached_win32error(0x2165, 'ERROR_DS_NO_FPO_IN_UNIVERSAL_GROUPS') # WinError 8549 (0x2165)
    ERROR_DS_CANT_ADD_TO_GC: Final[ThirtyTwoBits] = cached_win32error(0x2166, 'ERROR_DS_CANT_ADD_TO_GC') # WinError 8550 (0x2166)
    ERROR_DS_NO_CHECKPOINT_WITH_PDC: Final[ThirtyTwoBits] = cached_win32error(0x2167, 'ERROR_DS_NO_CHECKPOINT_WITH_PDC') # WinError 8551 (0x2167)
    ERROR_DS_SOURCE_AUDITING_NOT_ENABLED: Final[ThirtyTwoBits] = cached_win32error(0x2168, 'ERROR_DS_SOURCE_AUDITING_NOT_ENABLED') # WinError 8552 (0x2168)
    ERROR_DS_CANT_CREATE_IN_NONDOMAIN_NC: Final[ThirtyTwoBits] = cached_win32error(0x2169, 'ERROR_DS_CANT_CREATE_IN_NONDOMAIN_NC') # WinError 8553 (0x2169)
    ERROR_DS_INVALID_NAME_FOR_SPN: Final[ThirtyTwoBits] = cached_win32error(0x216A, 'ERROR_DS_INVALID_NAME_FOR_SPN') # WinError 8554 (0x216A)
    ERROR_DS_FILTER_USES_CONTRUCTED_ATTRS: Final[ThirtyTwoBits] = cached_win32error(0x216B, 'ERROR_DS_FILTER_USES_CONTRUCTED_ATTRS') # WinError 8555 (0x216B)
    ERROR_DS_UNICODEPWD_NOT_IN_QUOTES: Final[ThirtyTwoBits] = cached_win32error(0x216C, 'ERROR_DS_UNICODEPWD_NOT_IN_QUOTES') # WinError 8556 (0x216C)
    ERROR_DS_MACHINE_ACCOUNT_QUOTA_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x216D, 'ERROR_DS_MACHINE_ACCOUNT_QUOTA_EXCEEDED') # WinError 8557 (0x216D)
    ERROR_DS_MUST_BE_RUN_ON_DST_DC: Final[ThirtyTwoBits] = cached_win32error(0x216E, 'ERROR_DS_MUST_BE_RUN_ON_DST_DC') # WinError 8558 (0x216E)
    ERROR_DS_SRC_DC_MUST_BE_SP4_OR_GREATER: Final[ThirtyTwoBits] = cached_win32error(0x216F, 'ERROR_DS_SRC_DC_MUST_BE_SP4_OR_GREATER') # WinError 8559 (0x216F)
    ERROR_DS_CANT_TREE_DELETE_CRITICAL_OBJ: Final[ThirtyTwoBits] = cached_win32error(0x2170, 'ERROR_DS_CANT_TREE_DELETE_CRITICAL_OBJ') # WinError 8560 (0x2170)
    ERROR_DS_INIT_FAILURE_CONSOLE: Final[ThirtyTwoBits] = cached_win32error(0x2171, 'ERROR_DS_INIT_FAILURE_CONSOLE') # WinError 8561 (0x2171)
    ERROR_DS_SAM_INIT_FAILURE_CONSOLE: Final[ThirtyTwoBits] = cached_win32error(0x2172, 'ERROR_DS_SAM_INIT_FAILURE_CONSOLE') # WinError 8562 (0x2172)
    ERROR_DS_FOREST_VERSION_TOO_HIGH: Final[ThirtyTwoBits] = cached_win32error(0x2173, 'ERROR_DS_FOREST_VERSION_TOO_HIGH') # WinError 8563 (0x2173)
    ERROR_DS_DOMAIN_VERSION_TOO_HIGH: Final[ThirtyTwoBits] = cached_win32error(0x2174, 'ERROR_DS_DOMAIN_VERSION_TOO_HIGH') # WinError 8564 (0x2174)
    ERROR_DS_FOREST_VERSION_TOO_LOW: Final[ThirtyTwoBits] = cached_win32error(0x2175, 'ERROR_DS_FOREST_VERSION_TOO_LOW') # WinError 8565 (0x2175)
    ERROR_DS_DOMAIN_VERSION_TOO_LOW: Final[ThirtyTwoBits] = cached_win32error(0x2176, 'ERROR_DS_DOMAIN_VERSION_TOO_LOW') # WinError 8566 (0x2176)
    ERROR_DS_INCOMPATIBLE_VERSION: Final[ThirtyTwoBits] = cached_win32error(0x2177, 'ERROR_DS_INCOMPATIBLE_VERSION') # WinError 8567 (0x2177)
    ERROR_DS_LOW_DSA_VERSION: Final[ThirtyTwoBits] = cached_win32error(0x2178, 'ERROR_DS_LOW_DSA_VERSION') # WinError 8568 (0x2178)
    ERROR_DS_NO_BEHAVIOR_VERSION_IN_MIXEDDOMAIN: Final[ThirtyTwoBits] = cached_win32error(0x2179, 'ERROR_DS_NO_BEHAVIOR_VERSION_IN_MIXEDDOMAIN') # WinError 8569 (0x2179)
    ERROR_DS_NOT_SUPPORTED_SORT_ORDER: Final[ThirtyTwoBits] = cached_win32error(0x217A, 'ERROR_DS_NOT_SUPPORTED_SORT_ORDER') # WinError 8570 (0x217A)
    ERROR_DS_NAME_NOT_UNIQUE: Final[ThirtyTwoBits] = cached_win32error(0x217B, 'ERROR_DS_NAME_NOT_UNIQUE') # WinError 8571 (0x217B)
    ERROR_DS_MACHINE_ACCOUNT_CREATED_PRENT4: Final[ThirtyTwoBits] = cached_win32error(0x217C, 'ERROR_DS_MACHINE_ACCOUNT_CREATED_PRENT4') # WinError 8572 (0x217C)
    ERROR_DS_OUT_OF_VERSION_STORE: Final[ThirtyTwoBits] = cached_win32error(0x217D, 'ERROR_DS_OUT_OF_VERSION_STORE') # WinError 8573 (0x217D)
    ERROR_DS_INCOMPATIBLE_CONTROLS_USED: Final[ThirtyTwoBits] = cached_win32error(0x217E, 'ERROR_DS_INCOMPATIBLE_CONTROLS_USED') # WinError 8574 (0x217E)
    ERROR_DS_NO_REF_DOMAIN: Final[ThirtyTwoBits] = cached_win32error(0x217F, 'ERROR_DS_NO_REF_DOMAIN') # WinError 8575 (0x217F)
    ERROR_DS_RESERVED_LINK_ID: Final[ThirtyTwoBits] = cached_win32error(0x2180, 'ERROR_DS_RESERVED_LINK_ID') # WinError 8576 (0x2180)
    ERROR_DS_LINK_ID_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x2181, 'ERROR_DS_LINK_ID_NOT_AVAILABLE') # WinError 8577 (0x2181)
    ERROR_DS_AG_CANT_HAVE_UNIVERSAL_MEMBER: Final[ThirtyTwoBits] = cached_win32error(0x2182, 'ERROR_DS_AG_CANT_HAVE_UNIVERSAL_MEMBER') # WinError 8578 (0x2182)
    ERROR_DS_MODIFYDN_DISALLOWED_BY_INSTANCE_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x2183, 'ERROR_DS_MODIFYDN_DISALLOWED_BY_INSTANCE_TYPE') # WinError 8579 (0x2183)
    ERROR_DS_NO_OBJECT_MOVE_IN_SCHEMA_NC: Final[ThirtyTwoBits] = cached_win32error(0x2184, 'ERROR_DS_NO_OBJECT_MOVE_IN_SCHEMA_NC') # WinError 8580 (0x2184)
    ERROR_DS_MODIFYDN_DISALLOWED_BY_FLAG: Final[ThirtyTwoBits] = cached_win32error(0x2185, 'ERROR_DS_MODIFYDN_DISALLOWED_BY_FLAG') # WinError 8581 (0x2185)
    ERROR_DS_MODIFYDN_WRONG_GRANDPARENT: Final[ThirtyTwoBits] = cached_win32error(0x2186, 'ERROR_DS_MODIFYDN_WRONG_GRANDPARENT') # WinError 8582 (0x2186)
    ERROR_DS_NAME_ERROR_TRUST_REFERRAL: Final[ThirtyTwoBits] = cached_win32error(0x2187, 'ERROR_DS_NAME_ERROR_TRUST_REFERRAL') # WinError 8583 (0x2187)
    ERROR_NOT_SUPPORTED_ON_STANDARD_SERVER: Final[ThirtyTwoBits] = cached_win32error(0x2188, 'ERROR_NOT_SUPPORTED_ON_STANDARD_SERVER') # WinError 8584 (0x2188)
    ERROR_DS_CANT_ACCESS_REMOTE_PART_OF_AD: Final[ThirtyTwoBits] = cached_win32error(0x2189, 'ERROR_DS_CANT_ACCESS_REMOTE_PART_OF_AD') # WinError 8585 (0x2189)
    ERROR_DS_CR_IMPOSSIBLE_TO_VALIDATE_V2: Final[ThirtyTwoBits] = cached_win32error(0x218A, 'ERROR_DS_CR_IMPOSSIBLE_TO_VALIDATE_V2') # WinError 8586 (0x218A)
    ERROR_DS_THREAD_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x218B, 'ERROR_DS_THREAD_LIMIT_EXCEEDED') # WinError 8587 (0x218B)
    ERROR_DS_NOT_CLOSEST: Final[ThirtyTwoBits] = cached_win32error(0x218C, 'ERROR_DS_NOT_CLOSEST') # WinError 8588 (0x218C)
    ERROR_DS_CANT_DERIVE_SPN_WITHOUT_SERVER_REF: Final[ThirtyTwoBits] = cached_win32error(0x218D, 'ERROR_DS_CANT_DERIVE_SPN_WITHOUT_SERVER_REF') # WinError 8589 (0x218D)
    ERROR_DS_SINGLE_USER_MODE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x218E, 'ERROR_DS_SINGLE_USER_MODE_FAILED') # WinError 8590 (0x218E)
    ERROR_DS_NTDSCRIPT_SYNTAX_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x218F, 'ERROR_DS_NTDSCRIPT_SYNTAX_ERROR') # WinError 8591 (0x218F)
    ERROR_DS_NTDSCRIPT_PROCESS_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x2190, 'ERROR_DS_NTDSCRIPT_PROCESS_ERROR') # WinError 8592 (0x2190)
    ERROR_DS_DIFFERENT_REPL_EPOCHS: Final[ThirtyTwoBits] = cached_win32error(0x2191, 'ERROR_DS_DIFFERENT_REPL_EPOCHS') # WinError 8593 (0x2191)
    ERROR_DS_DRS_EXTENSIONS_CHANGED: Final[ThirtyTwoBits] = cached_win32error(0x2192, 'ERROR_DS_DRS_EXTENSIONS_CHANGED') # WinError 8594 (0x2192)
    ERROR_DS_REPLICA_SET_CHANGE_NOT_ALLOWED_ON_DISABLED_CR: Final[ThirtyTwoBits] = cached_win32error(0x2193, 'ERROR_DS_REPLICA_SET_CHANGE_NOT_ALLOWED_ON_DISABLED_CR') # WinError 8595 (0x2193)
    ERROR_DS_NO_MSDS_INTID: Final[ThirtyTwoBits] = cached_win32error(0x2194, 'ERROR_DS_NO_MSDS_INTID') # WinError 8596 (0x2194)
    ERROR_DS_DUP_MSDS_INTID: Final[ThirtyTwoBits] = cached_win32error(0x2195, 'ERROR_DS_DUP_MSDS_INTID') # WinError 8597 (0x2195)
    ERROR_DS_EXISTS_IN_RDNATTID: Final[ThirtyTwoBits] = cached_win32error(0x2196, 'ERROR_DS_EXISTS_IN_RDNATTID') # WinError 8598 (0x2196)
    ERROR_DS_AUTHORIZATION_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x2197, 'ERROR_DS_AUTHORIZATION_FAILED') # WinError 8599 (0x2197)
    ERROR_DS_INVALID_SCRIPT: Final[ThirtyTwoBits] = cached_win32error(0x2198, 'ERROR_DS_INVALID_SCRIPT') # WinError 8600 (0x2198)
    ERROR_DS_REMOTE_CROSSREF_OP_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x2199, 'ERROR_DS_REMOTE_CROSSREF_OP_FAILED') # WinError 8601 (0x2199)
    ERROR_DS_CROSS_REF_BUSY: Final[ThirtyTwoBits] = cached_win32error(0x219A, 'ERROR_DS_CROSS_REF_BUSY') # WinError 8602 (0x219A)
    ERROR_DS_CANT_DERIVE_SPN_FOR_DELETED_DOMAIN: Final[ThirtyTwoBits] = cached_win32error(0x219B, 'ERROR_DS_CANT_DERIVE_SPN_FOR_DELETED_DOMAIN') # WinError 8603 (0x219B)
    ERROR_DS_CANT_DEMOTE_WITH_WRITEABLE_NC: Final[ThirtyTwoBits] = cached_win32error(0x219C, 'ERROR_DS_CANT_DEMOTE_WITH_WRITEABLE_NC') # WinError 8604 (0x219C)
    ERROR_DS_DUPLICATE_ID_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x219D, 'ERROR_DS_DUPLICATE_ID_FOUND') # WinError 8605 (0x219D)
    ERROR_DS_INSUFFICIENT_ATTR_TO_CREATE_OBJECT: Final[ThirtyTwoBits] = cached_win32error(0x219E, 'ERROR_DS_INSUFFICIENT_ATTR_TO_CREATE_OBJECT') # WinError 8606 (0x219E)
    ERROR_DS_GROUP_CONVERSION_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x219F, 'ERROR_DS_GROUP_CONVERSION_ERROR') # WinError 8607 (0x219F)
    ERROR_DS_CANT_MOVE_APP_BASIC_GROUP: Final[ThirtyTwoBits] = cached_win32error(0x21A0, 'ERROR_DS_CANT_MOVE_APP_BASIC_GROUP') # WinError 8608 (0x21A0)
    ERROR_DS_CANT_MOVE_APP_QUERY_GROUP: Final[ThirtyTwoBits] = cached_win32error(0x21A1, 'ERROR_DS_CANT_MOVE_APP_QUERY_GROUP') # WinError 8609 (0x21A1)
    ERROR_DS_ROLE_NOT_VERIFIED: Final[ThirtyTwoBits] = cached_win32error(0x21A2, 'ERROR_DS_ROLE_NOT_VERIFIED') # WinError 8610 (0x21A2)
    ERROR_DS_WKO_CONTAINER_CANNOT_BE_SPECIAL: Final[ThirtyTwoBits] = cached_win32error(0x21A3, 'ERROR_DS_WKO_CONTAINER_CANNOT_BE_SPECIAL') # WinError 8611 (0x21A3)
    ERROR_DS_DOMAIN_RENAME_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x21A4, 'ERROR_DS_DOMAIN_RENAME_IN_PROGRESS') # WinError 8612 (0x21A4)
    ERROR_DS_EXISTING_AD_CHILD_NC: Final[ThirtyTwoBits] = cached_win32error(0x21A5, 'ERROR_DS_EXISTING_AD_CHILD_NC') # WinError 8613 (0x21A5)
    ERROR_DS_REPL_LIFETIME_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x21A6, 'ERROR_DS_REPL_LIFETIME_EXCEEDED') # WinError 8614 (0x21A6)
    ERROR_DS_DISALLOWED_IN_SYSTEM_CONTAINER: Final[ThirtyTwoBits] = cached_win32error(0x21A7, 'ERROR_DS_DISALLOWED_IN_SYSTEM_CONTAINER') # WinError 8615 (0x21A7)
    ERROR_DS_LDAP_SEND_QUEUE_FULL: Final[ThirtyTwoBits] = cached_win32error(0x21A8, 'ERROR_DS_LDAP_SEND_QUEUE_FULL') # WinError 8616 (0x21A8)
    ERROR_DS_DRA_OUT_SCHEDULE_WINDOW: Final[ThirtyTwoBits] = cached_win32error(0x21A9, 'ERROR_DS_DRA_OUT_SCHEDULE_WINDOW') # WinError 8617 (0x21A9)
    ERROR_DS_POLICY_NOT_KNOWN: Final[ThirtyTwoBits] = cached_win32error(0x21AA, 'ERROR_DS_POLICY_NOT_KNOWN') # WinError 8618 (0x21AA)
    ERROR_NO_SITE_SETTINGS_OBJECT: Final[ThirtyTwoBits] = cached_win32error(0x21AB, 'ERROR_NO_SITE_SETTINGS_OBJECT') # WinError 8619 (0x21AB)
    ERROR_NO_SECRETS: Final[ThirtyTwoBits] = cached_win32error(0x21AC, 'ERROR_NO_SECRETS') # WinError 8620 (0x21AC)
    ERROR_NO_WRITABLE_DC_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x21AD, 'ERROR_NO_WRITABLE_DC_FOUND') # WinError 8621 (0x21AD)
    ERROR_DS_NO_SERVER_OBJECT: Final[ThirtyTwoBits] = cached_win32error(0x21AE, 'ERROR_DS_NO_SERVER_OBJECT') # WinError 8622 (0x21AE)
    ERROR_DS_NO_NTDSA_OBJECT: Final[ThirtyTwoBits] = cached_win32error(0x21AF, 'ERROR_DS_NO_NTDSA_OBJECT') # WinError 8623 (0x21AF)
    ERROR_DS_NON_ASQ_SEARCH: Final[ThirtyTwoBits] = cached_win32error(0x21B0, 'ERROR_DS_NON_ASQ_SEARCH') # WinError 8624 (0x21B0)
    ERROR_DS_AUDIT_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x21B1, 'ERROR_DS_AUDIT_FAILURE') # WinError 8625 (0x21B1)
    ERROR_DS_INVALID_SEARCH_FLAG_SUBTREE: Final[ThirtyTwoBits] = cached_win32error(0x21B2, 'ERROR_DS_INVALID_SEARCH_FLAG_SUBTREE') # WinError 8626 (0x21B2)
    ERROR_DS_INVALID_SEARCH_FLAG_TUPLE: Final[ThirtyTwoBits] = cached_win32error(0x21B3, 'ERROR_DS_INVALID_SEARCH_FLAG_TUPLE') # WinError 8627 (0x21B3)
    ERROR_DS_HIERARCHY_TABLE_TOO_DEEP: Final[ThirtyTwoBits] = cached_win32error(0x21B4, 'ERROR_DS_HIERARCHY_TABLE_TOO_DEEP') # WinError 8628 (0x21B4)
    ERROR_DS_DRA_CORRUPT_UTD_VECTOR: Final[ThirtyTwoBits] = cached_win32error(0x21B5, 'ERROR_DS_DRA_CORRUPT_UTD_VECTOR') # WinError 8629 (0x21B5)
    ERROR_DS_DRA_SECRETS_DENIED: Final[ThirtyTwoBits] = cached_win32error(0x21B6, 'ERROR_DS_DRA_SECRETS_DENIED') # WinError 8630 (0x21B6)
    ERROR_DS_RESERVED_MAPI_ID: Final[ThirtyTwoBits] = cached_win32error(0x21B7, 'ERROR_DS_RESERVED_MAPI_ID') # WinError 8631 (0x21B7)
    ERROR_DS_MAPI_ID_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x21B8, 'ERROR_DS_MAPI_ID_NOT_AVAILABLE') # WinError 8632 (0x21B8)
    ERROR_DS_DRA_MISSING_KRBTGT_SECRET: Final[ThirtyTwoBits] = cached_win32error(0x21B9, 'ERROR_DS_DRA_MISSING_KRBTGT_SECRET') # WinError 8633 (0x21B9)
    ERROR_DS_DOMAIN_NAME_EXISTS_IN_FOREST: Final[ThirtyTwoBits] = cached_win32error(0x21BA, 'ERROR_DS_DOMAIN_NAME_EXISTS_IN_FOREST') # WinError 8634 (0x21BA)
    ERROR_DS_FLAT_NAME_EXISTS_IN_FOREST: Final[ThirtyTwoBits] = cached_win32error(0x21BB, 'ERROR_DS_FLAT_NAME_EXISTS_IN_FOREST') # WinError 8635 (0x21BB)
    ERROR_INVALID_USER_PRINCIPAL_NAME: Final[ThirtyTwoBits] = cached_win32error(0x21BC, 'ERROR_INVALID_USER_PRINCIPAL_NAME') # WinError 8636 (0x21BC)
    ERROR_DS_OID_MAPPED_GROUP_CANT_HAVE_MEMBERS: Final[ThirtyTwoBits] = cached_win32error(0x21BD, 'ERROR_DS_OID_MAPPED_GROUP_CANT_HAVE_MEMBERS') # WinError 8637 (0x21BD)
    ERROR_DS_OID_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x21BE, 'ERROR_DS_OID_NOT_FOUND') # WinError 8638 (0x21BE)
    ERROR_DS_DRA_RECYCLED_TARGET: Final[ThirtyTwoBits] = cached_win32error(0x21BF, 'ERROR_DS_DRA_RECYCLED_TARGET') # WinError 8639 (0x21BF)
    ERROR_DS_DISALLOWED_NC_REDIRECT: Final[ThirtyTwoBits] = cached_win32error(0x21C0, 'ERROR_DS_DISALLOWED_NC_REDIRECT') # WinError 8640 (0x21C0)
    ERROR_DS_HIGH_ADLDS_FFL: Final[ThirtyTwoBits] = cached_win32error(0x21C1, 'ERROR_DS_HIGH_ADLDS_FFL') # WinError 8641 (0x21C1)
    ERROR_DS_HIGH_DSA_VERSION: Final[ThirtyTwoBits] = cached_win32error(0x21C2, 'ERROR_DS_HIGH_DSA_VERSION') # WinError 8642 (0x21C2)
    ERROR_DS_LOW_ADLDS_FFL: Final[ThirtyTwoBits] = cached_win32error(0x21C3, 'ERROR_DS_LOW_ADLDS_FFL') # WinError 8643 (0x21C3)
    ERROR_DOMAIN_SID_SAME_AS_LOCAL_WORKSTATION: Final[ThirtyTwoBits] = cached_win32error(0x21C4, 'ERROR_DOMAIN_SID_SAME_AS_LOCAL_WORKSTATION') # WinError 8644 (0x21C4)
    ERROR_DS_UNDELETE_SAM_VALIDATION_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x21C5, 'ERROR_DS_UNDELETE_SAM_VALIDATION_FAILED') # WinError 8645 (0x21C5)
    ERROR_INCORRECT_ACCOUNT_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x21C6, 'ERROR_INCORRECT_ACCOUNT_TYPE') # WinError 8646 (0x21C6)
    ERROR_DS_SPN_VALUE_NOT_UNIQUE_IN_FOREST: Final[ThirtyTwoBits] = cached_win32error(0x21C7, 'ERROR_DS_SPN_VALUE_NOT_UNIQUE_IN_FOREST') # WinError 8647 (0x21C7)
    ERROR_DS_UPN_VALUE_NOT_UNIQUE_IN_FOREST: Final[ThirtyTwoBits] = cached_win32error(0x21C8, 'ERROR_DS_UPN_VALUE_NOT_UNIQUE_IN_FOREST') # WinError 8648 (0x21C8)
    ERROR_DS_MISSING_FOREST_TRUST: Final[ThirtyTwoBits] = cached_win32error(0x21C9, 'ERROR_DS_MISSING_FOREST_TRUST') # WinError 8649 (0x21C9)
    ERROR_DS_VALUE_KEY_NOT_UNIQUE: Final[ThirtyTwoBits] = cached_win32error(0x21CA, 'ERROR_DS_VALUE_KEY_NOT_UNIQUE') # WinError 8650 (0x21CA)
    ERROR_WEAK_WHFBKEY_BLOCKED: Final[ThirtyTwoBits] = cached_win32error(0x21CB, 'ERROR_WEAK_WHFBKEY_BLOCKED') # WinError 8651 (0x21CB)
    ERROR_DS_PER_ATTRIBUTE_AUTHZ_FAILED_DURING_ADD: Final[ThirtyTwoBits] = cached_win32error(0x21CC, 'ERROR_DS_PER_ATTRIBUTE_AUTHZ_FAILED_DURING_ADD') # WinError 8652 (0x21CC)
    ERROR_LOCAL_POLICY_MODIFICATION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x21CD, 'ERROR_LOCAL_POLICY_MODIFICATION_NOT_SUPPORTED') # WinError 8653 (0x21CD)
    ERROR_POLICY_CONTROLLED_ACCOUNT: Final[ThirtyTwoBits] = cached_win32error(0x21CE, 'ERROR_POLICY_CONTROLLED_ACCOUNT') # WinError 8654 (0x21CE)
    ERROR_LAPS_LEGACY_SCHEMA_MISSING: Final[ThirtyTwoBits] = cached_win32error(0x21CF, 'ERROR_LAPS_LEGACY_SCHEMA_MISSING') # WinError 8655 (0x21CF)
    ERROR_LAPS_SCHEMA_MISSING: Final[ThirtyTwoBits] = cached_win32error(0x21D0, 'ERROR_LAPS_SCHEMA_MISSING') # WinError 8656 (0x21D0)
    ERROR_LAPS_ENCRYPTION_REQUIRES_2016_DFL: Final[ThirtyTwoBits] = cached_win32error(0x21D1, 'ERROR_LAPS_ENCRYPTION_REQUIRES_2016_DFL') # WinError 8657 (0x21D1)
    DNS_ERROR_RCODE_FORMAT_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x2329, 'DNS_ERROR_RCODE_FORMAT_ERROR') # WinError 9001 (0x2329)
    DNS_ERROR_RCODE_SERVER_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x232A, 'DNS_ERROR_RCODE_SERVER_FAILURE') # WinError 9002 (0x232A)
    DNS_ERROR_RCODE_NAME_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x232B, 'DNS_ERROR_RCODE_NAME_ERROR') # WinError 9003 (0x232B)
    DNS_ERROR_RCODE_NOT_IMPLEMENTED: Final[ThirtyTwoBits] = cached_win32error(0x232C, 'DNS_ERROR_RCODE_NOT_IMPLEMENTED') # WinError 9004 (0x232C)
    DNS_ERROR_RCODE_REFUSED: Final[ThirtyTwoBits] = cached_win32error(0x232D, 'DNS_ERROR_RCODE_REFUSED') # WinError 9005 (0x232D)
    DNS_ERROR_RCODE_YXDOMAIN: Final[ThirtyTwoBits] = cached_win32error(0x232E, 'DNS_ERROR_RCODE_YXDOMAIN') # WinError 9006 (0x232E)
    DNS_ERROR_RCODE_YXRRSET: Final[ThirtyTwoBits] = cached_win32error(0x232F, 'DNS_ERROR_RCODE_YXRRSET') # WinError 9007 (0x232F)
    DNS_ERROR_RCODE_NXRRSET: Final[ThirtyTwoBits] = cached_win32error(0x2330, 'DNS_ERROR_RCODE_NXRRSET') # WinError 9008 (0x2330)
    DNS_ERROR_RCODE_NOTAUTH: Final[ThirtyTwoBits] = cached_win32error(0x2331, 'DNS_ERROR_RCODE_NOTAUTH') # WinError 9009 (0x2331)
    DNS_ERROR_RCODE_NOTZONE: Final[ThirtyTwoBits] = cached_win32error(0x2332, 'DNS_ERROR_RCODE_NOTZONE') # WinError 9010 (0x2332)
    DNS_ERROR_RCODE_BADSIG: Final[ThirtyTwoBits] = cached_win32error(0x2338, 'DNS_ERROR_RCODE_BADSIG') # WinError 9016 (0x2338)
    DNS_ERROR_RCODE_BADKEY: Final[ThirtyTwoBits] = cached_win32error(0x2339, 'DNS_ERROR_RCODE_BADKEY') # WinError 9017 (0x2339)
    DNS_ERROR_RCODE_BADTIME: Final[ThirtyTwoBits] = cached_win32error(0x233A, 'DNS_ERROR_RCODE_BADTIME') # WinError 9018 (0x233A)
    DNS_ERROR_KEYMASTER_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x238D, 'DNS_ERROR_KEYMASTER_REQUIRED') # WinError 9101 (0x238D)
    DNS_ERROR_NOT_ALLOWED_ON_SIGNED_ZONE: Final[ThirtyTwoBits] = cached_win32error(0x238E, 'DNS_ERROR_NOT_ALLOWED_ON_SIGNED_ZONE') # WinError 9102 (0x238E)
    DNS_ERROR_NSEC3_INCOMPATIBLE_WITH_RSA_SHA1: Final[ThirtyTwoBits] = cached_win32error(0x238F, 'DNS_ERROR_NSEC3_INCOMPATIBLE_WITH_RSA_SHA1') # WinError 9103 (0x238F)
    DNS_ERROR_NOT_ENOUGH_SIGNING_KEY_DESCRIPTORS: Final[ThirtyTwoBits] = cached_win32error(0x2390, 'DNS_ERROR_NOT_ENOUGH_SIGNING_KEY_DESCRIPTORS') # WinError 9104 (0x2390)
    DNS_ERROR_UNSUPPORTED_ALGORITHM: Final[ThirtyTwoBits] = cached_win32error(0x2391, 'DNS_ERROR_UNSUPPORTED_ALGORITHM') # WinError 9105 (0x2391)
    DNS_ERROR_INVALID_KEY_SIZE: Final[ThirtyTwoBits] = cached_win32error(0x2392, 'DNS_ERROR_INVALID_KEY_SIZE') # WinError 9106 (0x2392)
    DNS_ERROR_SIGNING_KEY_NOT_ACCESSIBLE: Final[ThirtyTwoBits] = cached_win32error(0x2393, 'DNS_ERROR_SIGNING_KEY_NOT_ACCESSIBLE') # WinError 9107 (0x2393)
    DNS_ERROR_KSP_DOES_NOT_SUPPORT_PROTECTION: Final[ThirtyTwoBits] = cached_win32error(0x2394, 'DNS_ERROR_KSP_DOES_NOT_SUPPORT_PROTECTION') # WinError 9108 (0x2394)
    DNS_ERROR_UNEXPECTED_DATA_PROTECTION_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x2395, 'DNS_ERROR_UNEXPECTED_DATA_PROTECTION_ERROR') # WinError 9109 (0x2395)
    DNS_ERROR_UNEXPECTED_CNG_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x2396, 'DNS_ERROR_UNEXPECTED_CNG_ERROR') # WinError 9110 (0x2396)
    DNS_ERROR_UNKNOWN_SIGNING_PARAMETER_VERSION: Final[ThirtyTwoBits] = cached_win32error(0x2397, 'DNS_ERROR_UNKNOWN_SIGNING_PARAMETER_VERSION') # WinError 9111 (0x2397)
    DNS_ERROR_KSP_NOT_ACCESSIBLE: Final[ThirtyTwoBits] = cached_win32error(0x2398, 'DNS_ERROR_KSP_NOT_ACCESSIBLE') # WinError 9112 (0x2398)
    DNS_ERROR_TOO_MANY_SKDS: Final[ThirtyTwoBits] = cached_win32error(0x2399, 'DNS_ERROR_TOO_MANY_SKDS') # WinError 9113 (0x2399)
    DNS_ERROR_INVALID_ROLLOVER_PERIOD: Final[ThirtyTwoBits] = cached_win32error(0x239A, 'DNS_ERROR_INVALID_ROLLOVER_PERIOD') # WinError 9114 (0x239A)
    DNS_ERROR_INVALID_INITIAL_ROLLOVER_OFFSET: Final[ThirtyTwoBits] = cached_win32error(0x239B, 'DNS_ERROR_INVALID_INITIAL_ROLLOVER_OFFSET') # WinError 9115 (0x239B)
    DNS_ERROR_ROLLOVER_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x239C, 'DNS_ERROR_ROLLOVER_IN_PROGRESS') # WinError 9116 (0x239C)
    DNS_ERROR_STANDBY_KEY_NOT_PRESENT: Final[ThirtyTwoBits] = cached_win32error(0x239D, 'DNS_ERROR_STANDBY_KEY_NOT_PRESENT') # WinError 9117 (0x239D)
    DNS_ERROR_NOT_ALLOWED_ON_ZSK: Final[ThirtyTwoBits] = cached_win32error(0x239E, 'DNS_ERROR_NOT_ALLOWED_ON_ZSK') # WinError 9118 (0x239E)
    DNS_ERROR_NOT_ALLOWED_ON_ACTIVE_SKD: Final[ThirtyTwoBits] = cached_win32error(0x239F, 'DNS_ERROR_NOT_ALLOWED_ON_ACTIVE_SKD') # WinError 9119 (0x239F)
    DNS_ERROR_ROLLOVER_ALREADY_QUEUED: Final[ThirtyTwoBits] = cached_win32error(0x23A0, 'DNS_ERROR_ROLLOVER_ALREADY_QUEUED') # WinError 9120 (0x23A0)
    DNS_ERROR_NOT_ALLOWED_ON_UNSIGNED_ZONE: Final[ThirtyTwoBits] = cached_win32error(0x23A1, 'DNS_ERROR_NOT_ALLOWED_ON_UNSIGNED_ZONE') # WinError 9121 (0x23A1)
    DNS_ERROR_BAD_KEYMASTER: Final[ThirtyTwoBits] = cached_win32error(0x23A2, 'DNS_ERROR_BAD_KEYMASTER') # WinError 9122 (0x23A2)
    DNS_ERROR_INVALID_SIGNATURE_VALIDITY_PERIOD: Final[ThirtyTwoBits] = cached_win32error(0x23A3, 'DNS_ERROR_INVALID_SIGNATURE_VALIDITY_PERIOD') # WinError 9123 (0x23A3)
    DNS_ERROR_INVALID_NSEC3_ITERATION_COUNT: Final[ThirtyTwoBits] = cached_win32error(0x23A4, 'DNS_ERROR_INVALID_NSEC3_ITERATION_COUNT') # WinError 9124 (0x23A4)
    DNS_ERROR_DNSSEC_IS_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x23A5, 'DNS_ERROR_DNSSEC_IS_DISABLED') # WinError 9125 (0x23A5)
    DNS_ERROR_INVALID_XML: Final[ThirtyTwoBits] = cached_win32error(0x23A6, 'DNS_ERROR_INVALID_XML') # WinError 9126 (0x23A6)
    DNS_ERROR_NO_VALID_TRUST_ANCHORS: Final[ThirtyTwoBits] = cached_win32error(0x23A7, 'DNS_ERROR_NO_VALID_TRUST_ANCHORS') # WinError 9127 (0x23A7)
    DNS_ERROR_ROLLOVER_NOT_POKEABLE: Final[ThirtyTwoBits] = cached_win32error(0x23A8, 'DNS_ERROR_ROLLOVER_NOT_POKEABLE') # WinError 9128 (0x23A8)
    DNS_ERROR_NSEC3_NAME_COLLISION: Final[ThirtyTwoBits] = cached_win32error(0x23A9, 'DNS_ERROR_NSEC3_NAME_COLLISION') # WinError 9129 (0x23A9)
    DNS_ERROR_NSEC_INCOMPATIBLE_WITH_NSEC3_RSA_SHA1: Final[ThirtyTwoBits] = cached_win32error(0x23AA, 'DNS_ERROR_NSEC_INCOMPATIBLE_WITH_NSEC3_RSA_SHA1') # WinError 9130 (0x23AA)
    DNS_INFO_NO_RECORDS: Final[ThirtyTwoBits] = cached_win32error(0x251D, 'DNS_INFO_NO_RECORDS') # WinError 9501 (0x251D)
    DNS_ERROR_BAD_PACKET: Final[ThirtyTwoBits] = cached_win32error(0x251E, 'DNS_ERROR_BAD_PACKET') # WinError 9502 (0x251E)
    DNS_ERROR_NO_PACKET: Final[ThirtyTwoBits] = cached_win32error(0x251F, 'DNS_ERROR_NO_PACKET') # WinError 9503 (0x251F)
    DNS_ERROR_RCODE: Final[ThirtyTwoBits] = cached_win32error(0x2520, 'DNS_ERROR_RCODE') # WinError 9504 (0x2520)
    DNS_ERROR_UNSECURE_PACKET: Final[ThirtyTwoBits] = cached_win32error(0x2521, 'DNS_ERROR_UNSECURE_PACKET') # WinError 9505 (0x2521)
    DNS_REQUEST_PENDING: Final[ThirtyTwoBits] = cached_win32error(0x2522, 'DNS_REQUEST_PENDING') # WinError 9506 (0x2522)
    DNS_ERROR_INVALID_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x254F, 'DNS_ERROR_INVALID_TYPE') # WinError 9551 (0x254F)
    DNS_ERROR_INVALID_IP_ADDRESS: Final[ThirtyTwoBits] = cached_win32error(0x2550, 'DNS_ERROR_INVALID_IP_ADDRESS') # WinError 9552 (0x2550)
    DNS_ERROR_INVALID_PROPERTY: Final[ThirtyTwoBits] = cached_win32error(0x2551, 'DNS_ERROR_INVALID_PROPERTY') # WinError 9553 (0x2551)
    DNS_ERROR_TRY_AGAIN_LATER: Final[ThirtyTwoBits] = cached_win32error(0x2552, 'DNS_ERROR_TRY_AGAIN_LATER') # WinError 9554 (0x2552)
    DNS_ERROR_NOT_UNIQUE: Final[ThirtyTwoBits] = cached_win32error(0x2553, 'DNS_ERROR_NOT_UNIQUE') # WinError 9555 (0x2553)
    DNS_ERROR_NON_RFC_NAME: Final[ThirtyTwoBits] = cached_win32error(0x2554, 'DNS_ERROR_NON_RFC_NAME') # WinError 9556 (0x2554)
    DNS_STATUS_FQDN: Final[ThirtyTwoBits] = cached_win32error(0x2555, 'DNS_STATUS_FQDN') # WinError 9557 (0x2555)
    DNS_STATUS_DOTTED_NAME: Final[ThirtyTwoBits] = cached_win32error(0x2556, 'DNS_STATUS_DOTTED_NAME') # WinError 9558 (0x2556)
    DNS_STATUS_SINGLE_PART_NAME: Final[ThirtyTwoBits] = cached_win32error(0x2557, 'DNS_STATUS_SINGLE_PART_NAME') # WinError 9559 (0x2557)
    DNS_ERROR_INVALID_NAME_CHAR: Final[ThirtyTwoBits] = cached_win32error(0x2558, 'DNS_ERROR_INVALID_NAME_CHAR') # WinError 9560 (0x2558)
    DNS_ERROR_NUMERIC_NAME: Final[ThirtyTwoBits] = cached_win32error(0x2559, 'DNS_ERROR_NUMERIC_NAME') # WinError 9561 (0x2559)
    DNS_ERROR_NOT_ALLOWED_ON_ROOT_SERVER: Final[ThirtyTwoBits] = cached_win32error(0x255A, 'DNS_ERROR_NOT_ALLOWED_ON_ROOT_SERVER') # WinError 9562 (0x255A)
    DNS_ERROR_NOT_ALLOWED_UNDER_DELEGATION: Final[ThirtyTwoBits] = cached_win32error(0x255B, 'DNS_ERROR_NOT_ALLOWED_UNDER_DELEGATION') # WinError 9563 (0x255B)
    DNS_ERROR_CANNOT_FIND_ROOT_HINTS: Final[ThirtyTwoBits] = cached_win32error(0x255C, 'DNS_ERROR_CANNOT_FIND_ROOT_HINTS') # WinError 9564 (0x255C)
    DNS_ERROR_INCONSISTENT_ROOT_HINTS: Final[ThirtyTwoBits] = cached_win32error(0x255D, 'DNS_ERROR_INCONSISTENT_ROOT_HINTS') # WinError 9565 (0x255D)
    DNS_ERROR_DWORD_VALUE_TOO_SMALL: Final[ThirtyTwoBits] = cached_win32error(0x255E, 'DNS_ERROR_DWORD_VALUE_TOO_SMALL') # WinError 9566 (0x255E)
    DNS_ERROR_DWORD_VALUE_TOO_LARGE: Final[ThirtyTwoBits] = cached_win32error(0x255F, 'DNS_ERROR_DWORD_VALUE_TOO_LARGE') # WinError 9567 (0x255F)
    DNS_ERROR_BACKGROUND_LOADING: Final[ThirtyTwoBits] = cached_win32error(0x2560, 'DNS_ERROR_BACKGROUND_LOADING') # WinError 9568 (0x2560)
    DNS_ERROR_NOT_ALLOWED_ON_RODC: Final[ThirtyTwoBits] = cached_win32error(0x2561, 'DNS_ERROR_NOT_ALLOWED_ON_RODC') # WinError 9569 (0x2561)
    DNS_ERROR_NOT_ALLOWED_UNDER_DNAME: Final[ThirtyTwoBits] = cached_win32error(0x2562, 'DNS_ERROR_NOT_ALLOWED_UNDER_DNAME') # WinError 9570 (0x2562)
    DNS_ERROR_DELEGATION_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x2563, 'DNS_ERROR_DELEGATION_REQUIRED') # WinError 9571 (0x2563)
    DNS_ERROR_INVALID_POLICY_TABLE: Final[ThirtyTwoBits] = cached_win32error(0x2564, 'DNS_ERROR_INVALID_POLICY_TABLE') # WinError 9572 (0x2564)
    DNS_ERROR_ADDRESS_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x2565, 'DNS_ERROR_ADDRESS_REQUIRED') # WinError 9573 (0x2565)
    DNS_ERROR_ZONE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_win32error(0x2581, 'DNS_ERROR_ZONE_DOES_NOT_EXIST') # WinError 9601 (0x2581)
    DNS_ERROR_NO_ZONE_INFO: Final[ThirtyTwoBits] = cached_win32error(0x2582, 'DNS_ERROR_NO_ZONE_INFO') # WinError 9602 (0x2582)
    DNS_ERROR_INVALID_ZONE_OPERATION: Final[ThirtyTwoBits] = cached_win32error(0x2583, 'DNS_ERROR_INVALID_ZONE_OPERATION') # WinError 9603 (0x2583)
    DNS_ERROR_ZONE_CONFIGURATION_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x2584, 'DNS_ERROR_ZONE_CONFIGURATION_ERROR') # WinError 9604 (0x2584)
    DNS_ERROR_ZONE_HAS_NO_SOA_RECORD: Final[ThirtyTwoBits] = cached_win32error(0x2585, 'DNS_ERROR_ZONE_HAS_NO_SOA_RECORD') # WinError 9605 (0x2585)
    DNS_ERROR_ZONE_HAS_NO_NS_RECORDS: Final[ThirtyTwoBits] = cached_win32error(0x2586, 'DNS_ERROR_ZONE_HAS_NO_NS_RECORDS') # WinError 9606 (0x2586)
    DNS_ERROR_ZONE_LOCKED: Final[ThirtyTwoBits] = cached_win32error(0x2587, 'DNS_ERROR_ZONE_LOCKED') # WinError 9607 (0x2587)
    DNS_ERROR_ZONE_CREATION_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x2588, 'DNS_ERROR_ZONE_CREATION_FAILED') # WinError 9608 (0x2588)
    DNS_ERROR_ZONE_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x2589, 'DNS_ERROR_ZONE_ALREADY_EXISTS') # WinError 9609 (0x2589)
    DNS_ERROR_AUTOZONE_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x258A, 'DNS_ERROR_AUTOZONE_ALREADY_EXISTS') # WinError 9610 (0x258A)
    DNS_ERROR_INVALID_ZONE_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x258B, 'DNS_ERROR_INVALID_ZONE_TYPE') # WinError 9611 (0x258B)
    DNS_ERROR_SECONDARY_REQUIRES_MASTER_IP: Final[ThirtyTwoBits] = cached_win32error(0x258C, 'DNS_ERROR_SECONDARY_REQUIRES_MASTER_IP') # WinError 9612 (0x258C)
    DNS_ERROR_ZONE_NOT_SECONDARY: Final[ThirtyTwoBits] = cached_win32error(0x258D, 'DNS_ERROR_ZONE_NOT_SECONDARY') # WinError 9613 (0x258D)
    DNS_ERROR_NEED_SECONDARY_ADDRESSES: Final[ThirtyTwoBits] = cached_win32error(0x258E, 'DNS_ERROR_NEED_SECONDARY_ADDRESSES') # WinError 9614 (0x258E)
    DNS_ERROR_WINS_INIT_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x258F, 'DNS_ERROR_WINS_INIT_FAILED') # WinError 9615 (0x258F)
    DNS_ERROR_NEED_WINS_SERVERS: Final[ThirtyTwoBits] = cached_win32error(0x2590, 'DNS_ERROR_NEED_WINS_SERVERS') # WinError 9616 (0x2590)
    DNS_ERROR_NBSTAT_INIT_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x2591, 'DNS_ERROR_NBSTAT_INIT_FAILED') # WinError 9617 (0x2591)
    DNS_ERROR_SOA_DELETE_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x2592, 'DNS_ERROR_SOA_DELETE_INVALID') # WinError 9618 (0x2592)
    DNS_ERROR_FORWARDER_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x2593, 'DNS_ERROR_FORWARDER_ALREADY_EXISTS') # WinError 9619 (0x2593)
    DNS_ERROR_ZONE_REQUIRES_MASTER_IP: Final[ThirtyTwoBits] = cached_win32error(0x2594, 'DNS_ERROR_ZONE_REQUIRES_MASTER_IP') # WinError 9620 (0x2594)
    DNS_ERROR_ZONE_IS_SHUTDOWN: Final[ThirtyTwoBits] = cached_win32error(0x2595, 'DNS_ERROR_ZONE_IS_SHUTDOWN') # WinError 9621 (0x2595)
    DNS_ERROR_ZONE_LOCKED_FOR_SIGNING: Final[ThirtyTwoBits] = cached_win32error(0x2596, 'DNS_ERROR_ZONE_LOCKED_FOR_SIGNING') # WinError 9622 (0x2596)
    DNS_ERROR_PRIMARY_REQUIRES_DATAFILE: Final[ThirtyTwoBits] = cached_win32error(0x25B3, 'DNS_ERROR_PRIMARY_REQUIRES_DATAFILE') # WinError 9651 (0x25B3)
    DNS_ERROR_INVALID_DATAFILE_NAME: Final[ThirtyTwoBits] = cached_win32error(0x25B4, 'DNS_ERROR_INVALID_DATAFILE_NAME') # WinError 9652 (0x25B4)
    DNS_ERROR_DATAFILE_OPEN_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x25B5, 'DNS_ERROR_DATAFILE_OPEN_FAILURE') # WinError 9653 (0x25B5)
    DNS_ERROR_FILE_WRITEBACK_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x25B6, 'DNS_ERROR_FILE_WRITEBACK_FAILED') # WinError 9654 (0x25B6)
    DNS_ERROR_DATAFILE_PARSING: Final[ThirtyTwoBits] = cached_win32error(0x25B7, 'DNS_ERROR_DATAFILE_PARSING') # WinError 9655 (0x25B7)
    DNS_ERROR_RECORD_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_win32error(0x25E5, 'DNS_ERROR_RECORD_DOES_NOT_EXIST') # WinError 9701 (0x25E5)
    DNS_ERROR_RECORD_FORMAT: Final[ThirtyTwoBits] = cached_win32error(0x25E6, 'DNS_ERROR_RECORD_FORMAT') # WinError 9702 (0x25E6)
    DNS_ERROR_NODE_CREATION_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x25E7, 'DNS_ERROR_NODE_CREATION_FAILED') # WinError 9703 (0x25E7)
    DNS_ERROR_UNKNOWN_RECORD_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x25E8, 'DNS_ERROR_UNKNOWN_RECORD_TYPE') # WinError 9704 (0x25E8)
    DNS_ERROR_RECORD_TIMED_OUT: Final[ThirtyTwoBits] = cached_win32error(0x25E9, 'DNS_ERROR_RECORD_TIMED_OUT') # WinError 9705 (0x25E9)
    DNS_ERROR_NAME_NOT_IN_ZONE: Final[ThirtyTwoBits] = cached_win32error(0x25EA, 'DNS_ERROR_NAME_NOT_IN_ZONE') # WinError 9706 (0x25EA)
    DNS_ERROR_CNAME_LOOP: Final[ThirtyTwoBits] = cached_win32error(0x25EB, 'DNS_ERROR_CNAME_LOOP') # WinError 9707 (0x25EB)
    DNS_ERROR_NODE_IS_CNAME: Final[ThirtyTwoBits] = cached_win32error(0x25EC, 'DNS_ERROR_NODE_IS_CNAME') # WinError 9708 (0x25EC)
    DNS_ERROR_CNAME_COLLISION: Final[ThirtyTwoBits] = cached_win32error(0x25ED, 'DNS_ERROR_CNAME_COLLISION') # WinError 9709 (0x25ED)
    DNS_ERROR_RECORD_ONLY_AT_ZONE_ROOT: Final[ThirtyTwoBits] = cached_win32error(0x25EE, 'DNS_ERROR_RECORD_ONLY_AT_ZONE_ROOT') # WinError 9710 (0x25EE)
    DNS_ERROR_RECORD_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x25EF, 'DNS_ERROR_RECORD_ALREADY_EXISTS') # WinError 9711 (0x25EF)
    DNS_ERROR_SECONDARY_DATA: Final[ThirtyTwoBits] = cached_win32error(0x25F0, 'DNS_ERROR_SECONDARY_DATA') # WinError 9712 (0x25F0)
    DNS_ERROR_NO_CREATE_CACHE_DATA: Final[ThirtyTwoBits] = cached_win32error(0x25F1, 'DNS_ERROR_NO_CREATE_CACHE_DATA') # WinError 9713 (0x25F1)
    DNS_ERROR_NAME_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_win32error(0x25F2, 'DNS_ERROR_NAME_DOES_NOT_EXIST') # WinError 9714 (0x25F2)
    DNS_WARNING_PTR_CREATE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x25F3, 'DNS_WARNING_PTR_CREATE_FAILED') # WinError 9715 (0x25F3)
    DNS_WARNING_DOMAIN_UNDELETED: Final[ThirtyTwoBits] = cached_win32error(0x25F4, 'DNS_WARNING_DOMAIN_UNDELETED') # WinError 9716 (0x25F4)
    DNS_ERROR_DS_UNAVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x25F5, 'DNS_ERROR_DS_UNAVAILABLE') # WinError 9717 (0x25F5)
    DNS_ERROR_DS_ZONE_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x25F6, 'DNS_ERROR_DS_ZONE_ALREADY_EXISTS') # WinError 9718 (0x25F6)
    DNS_ERROR_NO_BOOTFILE_IF_DS_ZONE: Final[ThirtyTwoBits] = cached_win32error(0x25F7, 'DNS_ERROR_NO_BOOTFILE_IF_DS_ZONE') # WinError 9719 (0x25F7)
    DNS_ERROR_NODE_IS_DNAME: Final[ThirtyTwoBits] = cached_win32error(0x25F8, 'DNS_ERROR_NODE_IS_DNAME') # WinError 9720 (0x25F8)
    DNS_ERROR_DNAME_COLLISION: Final[ThirtyTwoBits] = cached_win32error(0x25F9, 'DNS_ERROR_DNAME_COLLISION') # WinError 9721 (0x25F9)
    DNS_ERROR_ALIAS_LOOP: Final[ThirtyTwoBits] = cached_win32error(0x25FA, 'DNS_ERROR_ALIAS_LOOP') # WinError 9722 (0x25FA)
    DNS_INFO_AXFR_COMPLETE: Final[ThirtyTwoBits] = cached_win32error(0x2617, 'DNS_INFO_AXFR_COMPLETE') # WinError 9751 (0x2617)
    DNS_ERROR_AXFR: Final[ThirtyTwoBits] = cached_win32error(0x2618, 'DNS_ERROR_AXFR') # WinError 9752 (0x2618)
    DNS_INFO_ADDED_LOCAL_WINS: Final[ThirtyTwoBits] = cached_win32error(0x2619, 'DNS_INFO_ADDED_LOCAL_WINS') # WinError 9753 (0x2619)
    DNS_STATUS_CONTINUE_NEEDED: Final[ThirtyTwoBits] = cached_win32error(0x2649, 'DNS_STATUS_CONTINUE_NEEDED') # WinError 9801 (0x2649)
    DNS_ERROR_NO_TCPIP: Final[ThirtyTwoBits] = cached_win32error(0x267B, 'DNS_ERROR_NO_TCPIP') # WinError 9851 (0x267B)
    DNS_ERROR_NO_DNS_SERVERS: Final[ThirtyTwoBits] = cached_win32error(0x267C, 'DNS_ERROR_NO_DNS_SERVERS') # WinError 9852 (0x267C)
    DNS_ERROR_DP_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_win32error(0x26AD, 'DNS_ERROR_DP_DOES_NOT_EXIST') # WinError 9901 (0x26AD)
    DNS_ERROR_DP_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x26AE, 'DNS_ERROR_DP_ALREADY_EXISTS') # WinError 9902 (0x26AE)
    DNS_ERROR_DP_NOT_ENLISTED: Final[ThirtyTwoBits] = cached_win32error(0x26AF, 'DNS_ERROR_DP_NOT_ENLISTED') # WinError 9903 (0x26AF)
    DNS_ERROR_DP_ALREADY_ENLISTED: Final[ThirtyTwoBits] = cached_win32error(0x26B0, 'DNS_ERROR_DP_ALREADY_ENLISTED') # WinError 9904 (0x26B0)
    DNS_ERROR_DP_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x26B1, 'DNS_ERROR_DP_NOT_AVAILABLE') # WinError 9905 (0x26B1)
    DNS_ERROR_DP_FSMO_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x26B2, 'DNS_ERROR_DP_FSMO_ERROR') # WinError 9906 (0x26B2)
    DNS_ERROR_RRL_NOT_ENABLED: Final[ThirtyTwoBits] = cached_win32error(0x26B7, 'DNS_ERROR_RRL_NOT_ENABLED') # WinError 9911 (0x26B7)
    DNS_ERROR_RRL_INVALID_WINDOW_SIZE: Final[ThirtyTwoBits] = cached_win32error(0x26B8, 'DNS_ERROR_RRL_INVALID_WINDOW_SIZE') # WinError 9912 (0x26B8)
    DNS_ERROR_RRL_INVALID_IPV4_PREFIX: Final[ThirtyTwoBits] = cached_win32error(0x26B9, 'DNS_ERROR_RRL_INVALID_IPV4_PREFIX') # WinError 9913 (0x26B9)
    DNS_ERROR_RRL_INVALID_IPV6_PREFIX: Final[ThirtyTwoBits] = cached_win32error(0x26BA, 'DNS_ERROR_RRL_INVALID_IPV6_PREFIX') # WinError 9914 (0x26BA)
    DNS_ERROR_RRL_INVALID_TC_RATE: Final[ThirtyTwoBits] = cached_win32error(0x26BB, 'DNS_ERROR_RRL_INVALID_TC_RATE') # WinError 9915 (0x26BB)
    DNS_ERROR_RRL_INVALID_LEAK_RATE: Final[ThirtyTwoBits] = cached_win32error(0x26BC, 'DNS_ERROR_RRL_INVALID_LEAK_RATE') # WinError 9916 (0x26BC)
    DNS_ERROR_RRL_LEAK_RATE_LESSTHAN_TC_RATE: Final[ThirtyTwoBits] = cached_win32error(0x26BD, 'DNS_ERROR_RRL_LEAK_RATE_LESSTHAN_TC_RATE') # WinError 9917 (0x26BD)
    DNS_ERROR_VIRTUALIZATION_INSTANCE_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x26C1, 'DNS_ERROR_VIRTUALIZATION_INSTANCE_ALREADY_EXISTS') # WinError 9921 (0x26C1)
    DNS_ERROR_VIRTUALIZATION_INSTANCE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_win32error(0x26C2, 'DNS_ERROR_VIRTUALIZATION_INSTANCE_DOES_NOT_EXIST') # WinError 9922 (0x26C2)
    DNS_ERROR_VIRTUALIZATION_TREE_LOCKED: Final[ThirtyTwoBits] = cached_win32error(0x26C3, 'DNS_ERROR_VIRTUALIZATION_TREE_LOCKED') # WinError 9923 (0x26C3)
    DNS_ERROR_INVAILD_VIRTUALIZATION_INSTANCE_NAME: Final[ThirtyTwoBits] = cached_win32error(0x26C4, 'DNS_ERROR_INVAILD_VIRTUALIZATION_INSTANCE_NAME') # WinError 9924 (0x26C4)
    DNS_ERROR_DEFAULT_VIRTUALIZATION_INSTANCE: Final[ThirtyTwoBits] = cached_win32error(0x26C5, 'DNS_ERROR_DEFAULT_VIRTUALIZATION_INSTANCE') # WinError 9925 (0x26C5)
    DNS_ERROR_ZONESCOPE_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x26DF, 'DNS_ERROR_ZONESCOPE_ALREADY_EXISTS') # WinError 9951 (0x26DF)
    DNS_ERROR_ZONESCOPE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_win32error(0x26E0, 'DNS_ERROR_ZONESCOPE_DOES_NOT_EXIST') # WinError 9952 (0x26E0)
    DNS_ERROR_DEFAULT_ZONESCOPE: Final[ThirtyTwoBits] = cached_win32error(0x26E1, 'DNS_ERROR_DEFAULT_ZONESCOPE') # WinError 9953 (0x26E1)
    DNS_ERROR_INVALID_ZONESCOPE_NAME: Final[ThirtyTwoBits] = cached_win32error(0x26E2, 'DNS_ERROR_INVALID_ZONESCOPE_NAME') # WinError 9954 (0x26E2)
    DNS_ERROR_NOT_ALLOWED_WITH_ZONESCOPES: Final[ThirtyTwoBits] = cached_win32error(0x26E3, 'DNS_ERROR_NOT_ALLOWED_WITH_ZONESCOPES') # WinError 9955 (0x26E3)
    DNS_ERROR_LOAD_ZONESCOPE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x26E4, 'DNS_ERROR_LOAD_ZONESCOPE_FAILED') # WinError 9956 (0x26E4)
    DNS_ERROR_ZONESCOPE_FILE_WRITEBACK_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x26E5, 'DNS_ERROR_ZONESCOPE_FILE_WRITEBACK_FAILED') # WinError 9957 (0x26E5)
    DNS_ERROR_INVALID_SCOPE_NAME: Final[ThirtyTwoBits] = cached_win32error(0x26E6, 'DNS_ERROR_INVALID_SCOPE_NAME') # WinError 9958 (0x26E6)
    DNS_ERROR_SCOPE_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_win32error(0x26E7, 'DNS_ERROR_SCOPE_DOES_NOT_EXIST') # WinError 9959 (0x26E7)
    DNS_ERROR_DEFAULT_SCOPE: Final[ThirtyTwoBits] = cached_win32error(0x26E8, 'DNS_ERROR_DEFAULT_SCOPE') # WinError 9960 (0x26E8)
    DNS_ERROR_INVALID_SCOPE_OPERATION: Final[ThirtyTwoBits] = cached_win32error(0x26E9, 'DNS_ERROR_INVALID_SCOPE_OPERATION') # WinError 9961 (0x26E9)
    DNS_ERROR_SCOPE_LOCKED: Final[ThirtyTwoBits] = cached_win32error(0x26EA, 'DNS_ERROR_SCOPE_LOCKED') # WinError 9962 (0x26EA)
    DNS_ERROR_SCOPE_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x26EB, 'DNS_ERROR_SCOPE_ALREADY_EXISTS') # WinError 9963 (0x26EB)
    DNS_ERROR_POLICY_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x26F3, 'DNS_ERROR_POLICY_ALREADY_EXISTS') # WinError 9971 (0x26F3)
    DNS_ERROR_POLICY_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_win32error(0x26F4, 'DNS_ERROR_POLICY_DOES_NOT_EXIST') # WinError 9972 (0x26F4)
    DNS_ERROR_POLICY_INVALID_CRITERIA: Final[ThirtyTwoBits] = cached_win32error(0x26F5, 'DNS_ERROR_POLICY_INVALID_CRITERIA') # WinError 9973 (0x26F5)
    DNS_ERROR_POLICY_INVALID_SETTINGS: Final[ThirtyTwoBits] = cached_win32error(0x26F6, 'DNS_ERROR_POLICY_INVALID_SETTINGS') # WinError 9974 (0x26F6)
    DNS_ERROR_CLIENT_SUBNET_IS_ACCESSED: Final[ThirtyTwoBits] = cached_win32error(0x26F7, 'DNS_ERROR_CLIENT_SUBNET_IS_ACCESSED') # WinError 9975 (0x26F7)
    DNS_ERROR_CLIENT_SUBNET_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_win32error(0x26F8, 'DNS_ERROR_CLIENT_SUBNET_DOES_NOT_EXIST') # WinError 9976 (0x26F8)
    DNS_ERROR_CLIENT_SUBNET_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x26F9, 'DNS_ERROR_CLIENT_SUBNET_ALREADY_EXISTS') # WinError 9977 (0x26F9)
    DNS_ERROR_SUBNET_DOES_NOT_EXIST: Final[ThirtyTwoBits] = cached_win32error(0x26FA, 'DNS_ERROR_SUBNET_DOES_NOT_EXIST') # WinError 9978 (0x26FA)
    DNS_ERROR_SUBNET_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x26FB, 'DNS_ERROR_SUBNET_ALREADY_EXISTS') # WinError 9979 (0x26FB)
    DNS_ERROR_POLICY_LOCKED: Final[ThirtyTwoBits] = cached_win32error(0x26FC, 'DNS_ERROR_POLICY_LOCKED') # WinError 9980 (0x26FC)
    DNS_ERROR_POLICY_INVALID_WEIGHT: Final[ThirtyTwoBits] = cached_win32error(0x26FD, 'DNS_ERROR_POLICY_INVALID_WEIGHT') # WinError 9981 (0x26FD)
    DNS_ERROR_POLICY_INVALID_NAME: Final[ThirtyTwoBits] = cached_win32error(0x26FE, 'DNS_ERROR_POLICY_INVALID_NAME') # WinError 9982 (0x26FE)
    DNS_ERROR_POLICY_MISSING_CRITERIA: Final[ThirtyTwoBits] = cached_win32error(0x26FF, 'DNS_ERROR_POLICY_MISSING_CRITERIA') # WinError 9983 (0x26FF)
    DNS_ERROR_INVALID_CLIENT_SUBNET_NAME: Final[ThirtyTwoBits] = cached_win32error(0x2700, 'DNS_ERROR_INVALID_CLIENT_SUBNET_NAME') # WinError 9984 (0x2700)
    DNS_ERROR_POLICY_PROCESSING_ORDER_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x2701, 'DNS_ERROR_POLICY_PROCESSING_ORDER_INVALID') # WinError 9985 (0x2701)
    DNS_ERROR_POLICY_SCOPE_MISSING: Final[ThirtyTwoBits] = cached_win32error(0x2702, 'DNS_ERROR_POLICY_SCOPE_MISSING') # WinError 9986 (0x2702)
    DNS_ERROR_POLICY_SCOPE_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x2703, 'DNS_ERROR_POLICY_SCOPE_NOT_ALLOWED') # WinError 9987 (0x2703)
    DNS_ERROR_SERVERSCOPE_IS_REFERENCED: Final[ThirtyTwoBits] = cached_win32error(0x2704, 'DNS_ERROR_SERVERSCOPE_IS_REFERENCED') # WinError 9988 (0x2704)
    DNS_ERROR_ZONESCOPE_IS_REFERENCED: Final[ThirtyTwoBits] = cached_win32error(0x2705, 'DNS_ERROR_ZONESCOPE_IS_REFERENCED') # WinError 9989 (0x2705)
    DNS_ERROR_POLICY_INVALID_CRITERIA_CLIENT_SUBNET: Final[ThirtyTwoBits] = cached_win32error(0x2706, 'DNS_ERROR_POLICY_INVALID_CRITERIA_CLIENT_SUBNET') # WinError 9990 (0x2706)
    DNS_ERROR_POLICY_INVALID_CRITERIA_TRANSPORT_PROTOCOL: Final[ThirtyTwoBits] = cached_win32error(0x2707, 'DNS_ERROR_POLICY_INVALID_CRITERIA_TRANSPORT_PROTOCOL') # WinError 9991 (0x2707)
    DNS_ERROR_POLICY_INVALID_CRITERIA_NETWORK_PROTOCOL: Final[ThirtyTwoBits] = cached_win32error(0x2708, 'DNS_ERROR_POLICY_INVALID_CRITERIA_NETWORK_PROTOCOL') # WinError 9992 (0x2708)
    DNS_ERROR_POLICY_INVALID_CRITERIA_INTERFACE: Final[ThirtyTwoBits] = cached_win32error(0x2709, 'DNS_ERROR_POLICY_INVALID_CRITERIA_INTERFACE') # WinError 9993 (0x2709)
    DNS_ERROR_POLICY_INVALID_CRITERIA_FQDN: Final[ThirtyTwoBits] = cached_win32error(0x270A, 'DNS_ERROR_POLICY_INVALID_CRITERIA_FQDN') # WinError 9994 (0x270A)
    DNS_ERROR_POLICY_INVALID_CRITERIA_QUERY_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x270B, 'DNS_ERROR_POLICY_INVALID_CRITERIA_QUERY_TYPE') # WinError 9995 (0x270B)
    DNS_ERROR_POLICY_INVALID_CRITERIA_TIME_OF_DAY: Final[ThirtyTwoBits] = cached_win32error(0x270C, 'DNS_ERROR_POLICY_INVALID_CRITERIA_TIME_OF_DAY') # WinError 9996 (0x270C)
    WSAEINTR: Final[ThirtyTwoBits] = cached_win32error(0x2714, 'WSAEINTR') # WinError 10004 (0x2714)
    WSAEBADF: Final[ThirtyTwoBits] = cached_win32error(0x2719, 'WSAEBADF') # WinError 10009 (0x2719)
    WSAEACCES: Final[ThirtyTwoBits] = cached_win32error(0x271D, 'WSAEACCES') # WinError 10013 (0x271D)
    WSAEFAULT: Final[ThirtyTwoBits] = cached_win32error(0x271E, 'WSAEFAULT') # WinError 10014 (0x271E)
    WSAEINVAL: Final[ThirtyTwoBits] = cached_win32error(0x2726, 'WSAEINVAL') # WinError 10022 (0x2726)
    WSAEMFILE: Final[ThirtyTwoBits] = cached_win32error(0x2728, 'WSAEMFILE') # WinError 10024 (0x2728)
    WSAEWOULDBLOCK: Final[ThirtyTwoBits] = cached_win32error(0x2733, 'WSAEWOULDBLOCK') # WinError 10035 (0x2733)
    WSAEINPROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x2734, 'WSAEINPROGRESS') # WinError 10036 (0x2734)
    WSAEALREADY: Final[ThirtyTwoBits] = cached_win32error(0x2735, 'WSAEALREADY') # WinError 10037 (0x2735)
    WSAENOTSOCK: Final[ThirtyTwoBits] = cached_win32error(0x2736, 'WSAENOTSOCK') # WinError 10038 (0x2736)
    WSAEDESTADDRREQ: Final[ThirtyTwoBits] = cached_win32error(0x2737, 'WSAEDESTADDRREQ') # WinError 10039 (0x2737)
    WSAEMSGSIZE: Final[ThirtyTwoBits] = cached_win32error(0x2738, 'WSAEMSGSIZE') # WinError 10040 (0x2738)
    WSAEPROTOTYPE: Final[ThirtyTwoBits] = cached_win32error(0x2739, 'WSAEPROTOTYPE') # WinError 10041 (0x2739)
    WSAENOPROTOOPT: Final[ThirtyTwoBits] = cached_win32error(0x273A, 'WSAENOPROTOOPT') # WinError 10042 (0x273A)
    WSAEPROTONOSUPPORT: Final[ThirtyTwoBits] = cached_win32error(0x273B, 'WSAEPROTONOSUPPORT') # WinError 10043 (0x273B)
    WSAESOCKTNOSUPPORT: Final[ThirtyTwoBits] = cached_win32error(0x273C, 'WSAESOCKTNOSUPPORT') # WinError 10044 (0x273C)
    WSAEOPNOTSUPP: Final[ThirtyTwoBits] = cached_win32error(0x273D, 'WSAEOPNOTSUPP') # WinError 10045 (0x273D)
    WSAEPFNOSUPPORT: Final[ThirtyTwoBits] = cached_win32error(0x273E, 'WSAEPFNOSUPPORT') # WinError 10046 (0x273E)
    WSAEAFNOSUPPORT: Final[ThirtyTwoBits] = cached_win32error(0x273F, 'WSAEAFNOSUPPORT') # WinError 10047 (0x273F)
    WSAEADDRINUSE: Final[ThirtyTwoBits] = cached_win32error(0x2740, 'WSAEADDRINUSE') # WinError 10048 (0x2740)
    WSAEADDRNOTAVAIL: Final[ThirtyTwoBits] = cached_win32error(0x2741, 'WSAEADDRNOTAVAIL') # WinError 10049 (0x2741)
    WSAENETDOWN: Final[ThirtyTwoBits] = cached_win32error(0x2742, 'WSAENETDOWN') # WinError 10050 (0x2742)
    WSAENETUNREACH: Final[ThirtyTwoBits] = cached_win32error(0x2743, 'WSAENETUNREACH') # WinError 10051 (0x2743)
    WSAENETRESET: Final[ThirtyTwoBits] = cached_win32error(0x2744, 'WSAENETRESET') # WinError 10052 (0x2744)
    WSAECONNABORTED: Final[ThirtyTwoBits] = cached_win32error(0x2745, 'WSAECONNABORTED') # WinError 10053 (0x2745)
    WSAECONNRESET: Final[ThirtyTwoBits] = cached_win32error(0x2746, 'WSAECONNRESET') # WinError 10054 (0x2746)
    WSAENOBUFS: Final[ThirtyTwoBits] = cached_win32error(0x2747, 'WSAENOBUFS') # WinError 10055 (0x2747)
    WSAEISCONN: Final[ThirtyTwoBits] = cached_win32error(0x2748, 'WSAEISCONN') # WinError 10056 (0x2748)
    WSAENOTCONN: Final[ThirtyTwoBits] = cached_win32error(0x2749, 'WSAENOTCONN') # WinError 10057 (0x2749)
    WSAESHUTDOWN: Final[ThirtyTwoBits] = cached_win32error(0x274A, 'WSAESHUTDOWN') # WinError 10058 (0x274A)
    WSAETOOMANYREFS: Final[ThirtyTwoBits] = cached_win32error(0x274B, 'WSAETOOMANYREFS') # WinError 10059 (0x274B)
    WSAETIMEDOUT: Final[ThirtyTwoBits] = cached_win32error(0x274C, 'WSAETIMEDOUT') # WinError 10060 (0x274C)
    WSAECONNREFUSED: Final[ThirtyTwoBits] = cached_win32error(0x274D, 'WSAECONNREFUSED') # WinError 10061 (0x274D)
    WSAELOOP: Final[ThirtyTwoBits] = cached_win32error(0x274E, 'WSAELOOP') # WinError 10062 (0x274E)
    WSAENAMETOOLONG: Final[ThirtyTwoBits] = cached_win32error(0x274F, 'WSAENAMETOOLONG') # WinError 10063 (0x274F)
    WSAEHOSTDOWN: Final[ThirtyTwoBits] = cached_win32error(0x2750, 'WSAEHOSTDOWN') # WinError 10064 (0x2750)
    WSAEHOSTUNREACH: Final[ThirtyTwoBits] = cached_win32error(0x2751, 'WSAEHOSTUNREACH') # WinError 10065 (0x2751)
    WSAENOTEMPTY: Final[ThirtyTwoBits] = cached_win32error(0x2752, 'WSAENOTEMPTY') # WinError 10066 (0x2752)
    WSAEPROCLIM: Final[ThirtyTwoBits] = cached_win32error(0x2753, 'WSAEPROCLIM') # WinError 10067 (0x2753)
    WSAEUSERS: Final[ThirtyTwoBits] = cached_win32error(0x2754, 'WSAEUSERS') # WinError 10068 (0x2754)
    WSAEDQUOT: Final[ThirtyTwoBits] = cached_win32error(0x2755, 'WSAEDQUOT') # WinError 10069 (0x2755)
    WSAESTALE: Final[ThirtyTwoBits] = cached_win32error(0x2756, 'WSAESTALE') # WinError 10070 (0x2756)
    WSAEREMOTE: Final[ThirtyTwoBits] = cached_win32error(0x2757, 'WSAEREMOTE') # WinError 10071 (0x2757)
    WSASYSNOTREADY: Final[ThirtyTwoBits] = cached_win32error(0x276B, 'WSASYSNOTREADY') # WinError 10091 (0x276B)
    WSAVERNOTSUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x276C, 'WSAVERNOTSUPPORTED') # WinError 10092 (0x276C)
    WSANOTINITIALISED: Final[ThirtyTwoBits] = cached_win32error(0x276D, 'WSANOTINITIALISED') # WinError 10093 (0x276D)
    WSAEDISCON: Final[ThirtyTwoBits] = cached_win32error(0x2775, 'WSAEDISCON') # WinError 10101 (0x2775)
    WSAENOMORE: Final[ThirtyTwoBits] = cached_win32error(0x2776, 'WSAENOMORE') # WinError 10102 (0x2776)
    WSAECANCELLED: Final[ThirtyTwoBits] = cached_win32error(0x2777, 'WSAECANCELLED') # WinError 10103 (0x2777)
    WSAEINVALIDPROCTABLE: Final[ThirtyTwoBits] = cached_win32error(0x2778, 'WSAEINVALIDPROCTABLE') # WinError 10104 (0x2778)
    WSAEINVALIDPROVIDER: Final[ThirtyTwoBits] = cached_win32error(0x2779, 'WSAEINVALIDPROVIDER') # WinError 10105 (0x2779)
    WSAEPROVIDERFAILEDINIT: Final[ThirtyTwoBits] = cached_win32error(0x277A, 'WSAEPROVIDERFAILEDINIT') # WinError 10106 (0x277A)
    WSASYSCALLFAILURE: Final[ThirtyTwoBits] = cached_win32error(0x277B, 'WSASYSCALLFAILURE') # WinError 10107 (0x277B)
    WSASERVICE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x277C, 'WSASERVICE_NOT_FOUND') # WinError 10108 (0x277C)
    WSATYPE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x277D, 'WSATYPE_NOT_FOUND') # WinError 10109 (0x277D)
    WSA_E_NO_MORE: Final[ThirtyTwoBits] = cached_win32error(0x277E, 'WSA_E_NO_MORE') # WinError 10110 (0x277E)
    WSA_E_CANCELLED: Final[ThirtyTwoBits] = cached_win32error(0x277F, 'WSA_E_CANCELLED') # WinError 10111 (0x277F)
    WSAEREFUSED: Final[ThirtyTwoBits] = cached_win32error(0x2780, 'WSAEREFUSED') # WinError 10112 (0x2780)
    WSAHOST_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x2AF9, 'WSAHOST_NOT_FOUND') # WinError 11001 (0x2AF9)
    WSATRY_AGAIN: Final[ThirtyTwoBits] = cached_win32error(0x2AFA, 'WSATRY_AGAIN') # WinError 11002 (0x2AFA)
    WSANO_RECOVERY: Final[ThirtyTwoBits] = cached_win32error(0x2AFB, 'WSANO_RECOVERY') # WinError 11003 (0x2AFB)
    WSANO_DATA: Final[ThirtyTwoBits] = cached_win32error(0x2AFC, 'WSANO_DATA') # WinError 11004 (0x2AFC)
    WSA_QOS_RECEIVERS: Final[ThirtyTwoBits] = cached_win32error(0x2AFD, 'WSA_QOS_RECEIVERS') # WinError 11005 (0x2AFD)
    WSA_QOS_SENDERS: Final[ThirtyTwoBits] = cached_win32error(0x2AFE, 'WSA_QOS_SENDERS') # WinError 11006 (0x2AFE)
    WSA_QOS_NO_SENDERS: Final[ThirtyTwoBits] = cached_win32error(0x2AFF, 'WSA_QOS_NO_SENDERS') # WinError 11007 (0x2AFF)
    WSA_QOS_NO_RECEIVERS: Final[ThirtyTwoBits] = cached_win32error(0x2B00, 'WSA_QOS_NO_RECEIVERS') # WinError 11008 (0x2B00)
    WSA_QOS_REQUEST_CONFIRMED: Final[ThirtyTwoBits] = cached_win32error(0x2B01, 'WSA_QOS_REQUEST_CONFIRMED') # WinError 11009 (0x2B01)
    WSA_QOS_ADMISSION_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x2B02, 'WSA_QOS_ADMISSION_FAILURE') # WinError 11010 (0x2B02)
    WSA_QOS_POLICY_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x2B03, 'WSA_QOS_POLICY_FAILURE') # WinError 11011 (0x2B03)
    WSA_QOS_BAD_STYLE: Final[ThirtyTwoBits] = cached_win32error(0x2B04, 'WSA_QOS_BAD_STYLE') # WinError 11012 (0x2B04)
    WSA_QOS_BAD_OBJECT: Final[ThirtyTwoBits] = cached_win32error(0x2B05, 'WSA_QOS_BAD_OBJECT') # WinError 11013 (0x2B05)
    WSA_QOS_TRAFFIC_CTRL_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x2B06, 'WSA_QOS_TRAFFIC_CTRL_ERROR') # WinError 11014 (0x2B06)
    WSA_QOS_GENERIC_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x2B07, 'WSA_QOS_GENERIC_ERROR') # WinError 11015 (0x2B07)
    WSA_QOS_ESERVICETYPE: Final[ThirtyTwoBits] = cached_win32error(0x2B08, 'WSA_QOS_ESERVICETYPE') # WinError 11016 (0x2B08)
    WSA_QOS_EFLOWSPEC: Final[ThirtyTwoBits] = cached_win32error(0x2B09, 'WSA_QOS_EFLOWSPEC') # WinError 11017 (0x2B09)
    WSA_QOS_EPROVSPECBUF: Final[ThirtyTwoBits] = cached_win32error(0x2B0A, 'WSA_QOS_EPROVSPECBUF') # WinError 11018 (0x2B0A)
    WSA_QOS_EFILTERSTYLE: Final[ThirtyTwoBits] = cached_win32error(0x2B0B, 'WSA_QOS_EFILTERSTYLE') # WinError 11019 (0x2B0B)
    WSA_QOS_EFILTERTYPE: Final[ThirtyTwoBits] = cached_win32error(0x2B0C, 'WSA_QOS_EFILTERTYPE') # WinError 11020 (0x2B0C)
    WSA_QOS_EFILTERCOUNT: Final[ThirtyTwoBits] = cached_win32error(0x2B0D, 'WSA_QOS_EFILTERCOUNT') # WinError 11021 (0x2B0D)
    WSA_QOS_EOBJLENGTH: Final[ThirtyTwoBits] = cached_win32error(0x2B0E, 'WSA_QOS_EOBJLENGTH') # WinError 11022 (0x2B0E)
    WSA_QOS_EFLOWCOUNT: Final[ThirtyTwoBits] = cached_win32error(0x2B0F, 'WSA_QOS_EFLOWCOUNT') # WinError 11023 (0x2B0F)
    WSA_QOS_EUNKOWNPSOBJ: Final[ThirtyTwoBits] = cached_win32error(0x2B10, 'WSA_QOS_EUNKOWNPSOBJ') # WinError 11024 (0x2B10)
    WSA_QOS_EPOLICYOBJ: Final[ThirtyTwoBits] = cached_win32error(0x2B11, 'WSA_QOS_EPOLICYOBJ') # WinError 11025 (0x2B11)
    WSA_QOS_EFLOWDESC: Final[ThirtyTwoBits] = cached_win32error(0x2B12, 'WSA_QOS_EFLOWDESC') # WinError 11026 (0x2B12)
    WSA_QOS_EPSFLOWSPEC: Final[ThirtyTwoBits] = cached_win32error(0x2B13, 'WSA_QOS_EPSFLOWSPEC') # WinError 11027 (0x2B13)
    WSA_QOS_EPSFILTERSPEC: Final[ThirtyTwoBits] = cached_win32error(0x2B14, 'WSA_QOS_EPSFILTERSPEC') # WinError 11028 (0x2B14)
    WSA_QOS_ESDMODEOBJ: Final[ThirtyTwoBits] = cached_win32error(0x2B15, 'WSA_QOS_ESDMODEOBJ') # WinError 11029 (0x2B15)
    WSA_QOS_ESHAPERATEOBJ: Final[ThirtyTwoBits] = cached_win32error(0x2B16, 'WSA_QOS_ESHAPERATEOBJ') # WinError 11030 (0x2B16)
    WSA_QOS_RESERVED_PETYPE: Final[ThirtyTwoBits] = cached_win32error(0x2B17, 'WSA_QOS_RESERVED_PETYPE') # WinError 11031 (0x2B17)
    WSA_SECURE_HOST_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x2B18, 'WSA_SECURE_HOST_NOT_FOUND') # WinError 11032 (0x2B18)
    WSA_IPSEC_NAME_POLICY_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x2B19, 'WSA_IPSEC_NAME_POLICY_ERROR') # WinError 11033 (0x2B19)
    ERROR_IPSEC_QM_POLICY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x32C8, 'ERROR_IPSEC_QM_POLICY_EXISTS') # WinError 13000 (0x32C8)
    ERROR_IPSEC_QM_POLICY_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x32C9, 'ERROR_IPSEC_QM_POLICY_NOT_FOUND') # WinError 13001 (0x32C9)
    ERROR_IPSEC_QM_POLICY_IN_USE: Final[ThirtyTwoBits] = cached_win32error(0x32CA, 'ERROR_IPSEC_QM_POLICY_IN_USE') # WinError 13002 (0x32CA)
    ERROR_IPSEC_MM_POLICY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x32CB, 'ERROR_IPSEC_MM_POLICY_EXISTS') # WinError 13003 (0x32CB)
    ERROR_IPSEC_MM_POLICY_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x32CC, 'ERROR_IPSEC_MM_POLICY_NOT_FOUND') # WinError 13004 (0x32CC)
    ERROR_IPSEC_MM_POLICY_IN_USE: Final[ThirtyTwoBits] = cached_win32error(0x32CD, 'ERROR_IPSEC_MM_POLICY_IN_USE') # WinError 13005 (0x32CD)
    ERROR_IPSEC_MM_FILTER_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x32CE, 'ERROR_IPSEC_MM_FILTER_EXISTS') # WinError 13006 (0x32CE)
    ERROR_IPSEC_MM_FILTER_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x32CF, 'ERROR_IPSEC_MM_FILTER_NOT_FOUND') # WinError 13007 (0x32CF)
    ERROR_IPSEC_TRANSPORT_FILTER_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x32D0, 'ERROR_IPSEC_TRANSPORT_FILTER_EXISTS') # WinError 13008 (0x32D0)
    ERROR_IPSEC_TRANSPORT_FILTER_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x32D1, 'ERROR_IPSEC_TRANSPORT_FILTER_NOT_FOUND') # WinError 13009 (0x32D1)
    ERROR_IPSEC_MM_AUTH_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x32D2, 'ERROR_IPSEC_MM_AUTH_EXISTS') # WinError 13010 (0x32D2)
    ERROR_IPSEC_MM_AUTH_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x32D3, 'ERROR_IPSEC_MM_AUTH_NOT_FOUND') # WinError 13011 (0x32D3)
    ERROR_IPSEC_MM_AUTH_IN_USE: Final[ThirtyTwoBits] = cached_win32error(0x32D4, 'ERROR_IPSEC_MM_AUTH_IN_USE') # WinError 13012 (0x32D4)
    ERROR_IPSEC_DEFAULT_MM_POLICY_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x32D5, 'ERROR_IPSEC_DEFAULT_MM_POLICY_NOT_FOUND') # WinError 13013 (0x32D5)
    ERROR_IPSEC_DEFAULT_MM_AUTH_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x32D6, 'ERROR_IPSEC_DEFAULT_MM_AUTH_NOT_FOUND') # WinError 13014 (0x32D6)
    ERROR_IPSEC_DEFAULT_QM_POLICY_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x32D7, 'ERROR_IPSEC_DEFAULT_QM_POLICY_NOT_FOUND') # WinError 13015 (0x32D7)
    ERROR_IPSEC_TUNNEL_FILTER_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x32D8, 'ERROR_IPSEC_TUNNEL_FILTER_EXISTS') # WinError 13016 (0x32D8)
    ERROR_IPSEC_TUNNEL_FILTER_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x32D9, 'ERROR_IPSEC_TUNNEL_FILTER_NOT_FOUND') # WinError 13017 (0x32D9)
    ERROR_IPSEC_MM_FILTER_PENDING_DELETION: Final[ThirtyTwoBits] = cached_win32error(0x32DA, 'ERROR_IPSEC_MM_FILTER_PENDING_DELETION') # WinError 13018 (0x32DA)
    ERROR_IPSEC_TRANSPORT_FILTER_PENDING_DELETION: Final[ThirtyTwoBits] = cached_win32error(0x32DB, 'ERROR_IPSEC_TRANSPORT_FILTER_PENDING_DELETION') # WinError 13019 (0x32DB)
    ERROR_IPSEC_TUNNEL_FILTER_PENDING_DELETION: Final[ThirtyTwoBits] = cached_win32error(0x32DC, 'ERROR_IPSEC_TUNNEL_FILTER_PENDING_DELETION') # WinError 13020 (0x32DC)
    ERROR_IPSEC_MM_POLICY_PENDING_DELETION: Final[ThirtyTwoBits] = cached_win32error(0x32DD, 'ERROR_IPSEC_MM_POLICY_PENDING_DELETION') # WinError 13021 (0x32DD)
    ERROR_IPSEC_MM_AUTH_PENDING_DELETION: Final[ThirtyTwoBits] = cached_win32error(0x32DE, 'ERROR_IPSEC_MM_AUTH_PENDING_DELETION') # WinError 13022 (0x32DE)
    ERROR_IPSEC_QM_POLICY_PENDING_DELETION: Final[ThirtyTwoBits] = cached_win32error(0x32DF, 'ERROR_IPSEC_QM_POLICY_PENDING_DELETION') # WinError 13023 (0x32DF)
    WARNING_IPSEC_MM_POLICY_PRUNED: Final[ThirtyTwoBits] = cached_win32error(0x32E0, 'WARNING_IPSEC_MM_POLICY_PRUNED') # WinError 13024 (0x32E0)
    WARNING_IPSEC_QM_POLICY_PRUNED: Final[ThirtyTwoBits] = cached_win32error(0x32E1, 'WARNING_IPSEC_QM_POLICY_PRUNED') # WinError 13025 (0x32E1)
    ERROR_IPSEC_IKE_NEG_STATUS_BEGIN: Final[ThirtyTwoBits] = cached_win32error(0x35E8, 'ERROR_IPSEC_IKE_NEG_STATUS_BEGIN') # WinError 13800 (0x35E8)
    ERROR_IPSEC_IKE_AUTH_FAIL: Final[ThirtyTwoBits] = cached_win32error(0x35E9, 'ERROR_IPSEC_IKE_AUTH_FAIL') # WinError 13801 (0x35E9)
    ERROR_IPSEC_IKE_ATTRIB_FAIL: Final[ThirtyTwoBits] = cached_win32error(0x35EA, 'ERROR_IPSEC_IKE_ATTRIB_FAIL') # WinError 13802 (0x35EA)
    ERROR_IPSEC_IKE_NEGOTIATION_PENDING: Final[ThirtyTwoBits] = cached_win32error(0x35EB, 'ERROR_IPSEC_IKE_NEGOTIATION_PENDING') # WinError 13803 (0x35EB)
    ERROR_IPSEC_IKE_GENERAL_PROCESSING_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x35EC, 'ERROR_IPSEC_IKE_GENERAL_PROCESSING_ERROR') # WinError 13804 (0x35EC)
    ERROR_IPSEC_IKE_TIMED_OUT: Final[ThirtyTwoBits] = cached_win32error(0x35ED, 'ERROR_IPSEC_IKE_TIMED_OUT') # WinError 13805 (0x35ED)
    ERROR_IPSEC_IKE_NO_CERT: Final[ThirtyTwoBits] = cached_win32error(0x35EE, 'ERROR_IPSEC_IKE_NO_CERT') # WinError 13806 (0x35EE)
    ERROR_IPSEC_IKE_SA_DELETED: Final[ThirtyTwoBits] = cached_win32error(0x35EF, 'ERROR_IPSEC_IKE_SA_DELETED') # WinError 13807 (0x35EF)
    ERROR_IPSEC_IKE_SA_REAPED: Final[ThirtyTwoBits] = cached_win32error(0x35F0, 'ERROR_IPSEC_IKE_SA_REAPED') # WinError 13808 (0x35F0)
    ERROR_IPSEC_IKE_MM_ACQUIRE_DROP: Final[ThirtyTwoBits] = cached_win32error(0x35F1, 'ERROR_IPSEC_IKE_MM_ACQUIRE_DROP') # WinError 13809 (0x35F1)
    ERROR_IPSEC_IKE_QM_ACQUIRE_DROP: Final[ThirtyTwoBits] = cached_win32error(0x35F2, 'ERROR_IPSEC_IKE_QM_ACQUIRE_DROP') # WinError 13810 (0x35F2)
    ERROR_IPSEC_IKE_QUEUE_DROP_MM: Final[ThirtyTwoBits] = cached_win32error(0x35F3, 'ERROR_IPSEC_IKE_QUEUE_DROP_MM') # WinError 13811 (0x35F3)
    ERROR_IPSEC_IKE_QUEUE_DROP_NO_MM: Final[ThirtyTwoBits] = cached_win32error(0x35F4, 'ERROR_IPSEC_IKE_QUEUE_DROP_NO_MM') # WinError 13812 (0x35F4)
    ERROR_IPSEC_IKE_DROP_NO_RESPONSE: Final[ThirtyTwoBits] = cached_win32error(0x35F5, 'ERROR_IPSEC_IKE_DROP_NO_RESPONSE') # WinError 13813 (0x35F5)
    ERROR_IPSEC_IKE_MM_DELAY_DROP: Final[ThirtyTwoBits] = cached_win32error(0x35F6, 'ERROR_IPSEC_IKE_MM_DELAY_DROP') # WinError 13814 (0x35F6)
    ERROR_IPSEC_IKE_QM_DELAY_DROP: Final[ThirtyTwoBits] = cached_win32error(0x35F7, 'ERROR_IPSEC_IKE_QM_DELAY_DROP') # WinError 13815 (0x35F7)
    ERROR_IPSEC_IKE_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x35F8, 'ERROR_IPSEC_IKE_ERROR') # WinError 13816 (0x35F8)
    ERROR_IPSEC_IKE_CRL_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x35F9, 'ERROR_IPSEC_IKE_CRL_FAILED') # WinError 13817 (0x35F9)
    ERROR_IPSEC_IKE_INVALID_KEY_USAGE: Final[ThirtyTwoBits] = cached_win32error(0x35FA, 'ERROR_IPSEC_IKE_INVALID_KEY_USAGE') # WinError 13818 (0x35FA)
    ERROR_IPSEC_IKE_INVALID_CERT_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x35FB, 'ERROR_IPSEC_IKE_INVALID_CERT_TYPE') # WinError 13819 (0x35FB)
    ERROR_IPSEC_IKE_NO_PRIVATE_KEY: Final[ThirtyTwoBits] = cached_win32error(0x35FC, 'ERROR_IPSEC_IKE_NO_PRIVATE_KEY') # WinError 13820 (0x35FC)
    ERROR_IPSEC_IKE_SIMULTANEOUS_REKEY: Final[ThirtyTwoBits] = cached_win32error(0x35FD, 'ERROR_IPSEC_IKE_SIMULTANEOUS_REKEY') # WinError 13821 (0x35FD)
    ERROR_IPSEC_IKE_DH_FAIL: Final[ThirtyTwoBits] = cached_win32error(0x35FE, 'ERROR_IPSEC_IKE_DH_FAIL') # WinError 13822 (0x35FE)
    ERROR_IPSEC_IKE_CRITICAL_PAYLOAD_NOT_RECOGNIZED: Final[ThirtyTwoBits] = cached_win32error(0x35FF, 'ERROR_IPSEC_IKE_CRITICAL_PAYLOAD_NOT_RECOGNIZED') # WinError 13823 (0x35FF)
    ERROR_IPSEC_IKE_INVALID_HEADER: Final[ThirtyTwoBits] = cached_win32error(0x3600, 'ERROR_IPSEC_IKE_INVALID_HEADER') # WinError 13824 (0x3600)
    ERROR_IPSEC_IKE_NO_POLICY: Final[ThirtyTwoBits] = cached_win32error(0x3601, 'ERROR_IPSEC_IKE_NO_POLICY') # WinError 13825 (0x3601)
    ERROR_IPSEC_IKE_INVALID_SIGNATURE: Final[ThirtyTwoBits] = cached_win32error(0x3602, 'ERROR_IPSEC_IKE_INVALID_SIGNATURE') # WinError 13826 (0x3602)
    ERROR_IPSEC_IKE_KERBEROS_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x3603, 'ERROR_IPSEC_IKE_KERBEROS_ERROR') # WinError 13827 (0x3603)
    ERROR_IPSEC_IKE_NO_PUBLIC_KEY: Final[ThirtyTwoBits] = cached_win32error(0x3604, 'ERROR_IPSEC_IKE_NO_PUBLIC_KEY') # WinError 13828 (0x3604)
    ERROR_IPSEC_IKE_PROCESS_ERR: Final[ThirtyTwoBits] = cached_win32error(0x3605, 'ERROR_IPSEC_IKE_PROCESS_ERR') # WinError 13829 (0x3605)
    ERROR_IPSEC_IKE_PROCESS_ERR_SA: Final[ThirtyTwoBits] = cached_win32error(0x3606, 'ERROR_IPSEC_IKE_PROCESS_ERR_SA') # WinError 13830 (0x3606)
    ERROR_IPSEC_IKE_PROCESS_ERR_PROP: Final[ThirtyTwoBits] = cached_win32error(0x3607, 'ERROR_IPSEC_IKE_PROCESS_ERR_PROP') # WinError 13831 (0x3607)
    ERROR_IPSEC_IKE_PROCESS_ERR_TRANS: Final[ThirtyTwoBits] = cached_win32error(0x3608, 'ERROR_IPSEC_IKE_PROCESS_ERR_TRANS') # WinError 13832 (0x3608)
    ERROR_IPSEC_IKE_PROCESS_ERR_KE: Final[ThirtyTwoBits] = cached_win32error(0x3609, 'ERROR_IPSEC_IKE_PROCESS_ERR_KE') # WinError 13833 (0x3609)
    ERROR_IPSEC_IKE_PROCESS_ERR_ID: Final[ThirtyTwoBits] = cached_win32error(0x360A, 'ERROR_IPSEC_IKE_PROCESS_ERR_ID') # WinError 13834 (0x360A)
    ERROR_IPSEC_IKE_PROCESS_ERR_CERT: Final[ThirtyTwoBits] = cached_win32error(0x360B, 'ERROR_IPSEC_IKE_PROCESS_ERR_CERT') # WinError 13835 (0x360B)
    ERROR_IPSEC_IKE_PROCESS_ERR_CERT_REQ: Final[ThirtyTwoBits] = cached_win32error(0x360C, 'ERROR_IPSEC_IKE_PROCESS_ERR_CERT_REQ') # WinError 13836 (0x360C)
    ERROR_IPSEC_IKE_PROCESS_ERR_HASH: Final[ThirtyTwoBits] = cached_win32error(0x360D, 'ERROR_IPSEC_IKE_PROCESS_ERR_HASH') # WinError 13837 (0x360D)
    ERROR_IPSEC_IKE_PROCESS_ERR_SIG: Final[ThirtyTwoBits] = cached_win32error(0x360E, 'ERROR_IPSEC_IKE_PROCESS_ERR_SIG') # WinError 13838 (0x360E)
    ERROR_IPSEC_IKE_PROCESS_ERR_NONCE: Final[ThirtyTwoBits] = cached_win32error(0x360F, 'ERROR_IPSEC_IKE_PROCESS_ERR_NONCE') # WinError 13839 (0x360F)
    ERROR_IPSEC_IKE_PROCESS_ERR_NOTIFY: Final[ThirtyTwoBits] = cached_win32error(0x3610, 'ERROR_IPSEC_IKE_PROCESS_ERR_NOTIFY') # WinError 13840 (0x3610)
    ERROR_IPSEC_IKE_PROCESS_ERR_DELETE: Final[ThirtyTwoBits] = cached_win32error(0x3611, 'ERROR_IPSEC_IKE_PROCESS_ERR_DELETE') # WinError 13841 (0x3611)
    ERROR_IPSEC_IKE_PROCESS_ERR_VENDOR: Final[ThirtyTwoBits] = cached_win32error(0x3612, 'ERROR_IPSEC_IKE_PROCESS_ERR_VENDOR') # WinError 13842 (0x3612)
    ERROR_IPSEC_IKE_INVALID_PAYLOAD: Final[ThirtyTwoBits] = cached_win32error(0x3613, 'ERROR_IPSEC_IKE_INVALID_PAYLOAD') # WinError 13843 (0x3613)
    ERROR_IPSEC_IKE_LOAD_SOFT_SA: Final[ThirtyTwoBits] = cached_win32error(0x3614, 'ERROR_IPSEC_IKE_LOAD_SOFT_SA') # WinError 13844 (0x3614)
    ERROR_IPSEC_IKE_SOFT_SA_TORN_DOWN: Final[ThirtyTwoBits] = cached_win32error(0x3615, 'ERROR_IPSEC_IKE_SOFT_SA_TORN_DOWN') # WinError 13845 (0x3615)
    ERROR_IPSEC_IKE_INVALID_COOKIE: Final[ThirtyTwoBits] = cached_win32error(0x3616, 'ERROR_IPSEC_IKE_INVALID_COOKIE') # WinError 13846 (0x3616)
    ERROR_IPSEC_IKE_NO_PEER_CERT: Final[ThirtyTwoBits] = cached_win32error(0x3617, 'ERROR_IPSEC_IKE_NO_PEER_CERT') # WinError 13847 (0x3617)
    ERROR_IPSEC_IKE_PEER_CRL_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3618, 'ERROR_IPSEC_IKE_PEER_CRL_FAILED') # WinError 13848 (0x3618)
    ERROR_IPSEC_IKE_POLICY_CHANGE: Final[ThirtyTwoBits] = cached_win32error(0x3619, 'ERROR_IPSEC_IKE_POLICY_CHANGE') # WinError 13849 (0x3619)
    ERROR_IPSEC_IKE_NO_MM_POLICY: Final[ThirtyTwoBits] = cached_win32error(0x361A, 'ERROR_IPSEC_IKE_NO_MM_POLICY') # WinError 13850 (0x361A)
    ERROR_IPSEC_IKE_NOTCBPRIV: Final[ThirtyTwoBits] = cached_win32error(0x361B, 'ERROR_IPSEC_IKE_NOTCBPRIV') # WinError 13851 (0x361B)
    ERROR_IPSEC_IKE_SECLOADFAIL: Final[ThirtyTwoBits] = cached_win32error(0x361C, 'ERROR_IPSEC_IKE_SECLOADFAIL') # WinError 13852 (0x361C)
    ERROR_IPSEC_IKE_FAILSSPINIT: Final[ThirtyTwoBits] = cached_win32error(0x361D, 'ERROR_IPSEC_IKE_FAILSSPINIT') # WinError 13853 (0x361D)
    ERROR_IPSEC_IKE_FAILQUERYSSP: Final[ThirtyTwoBits] = cached_win32error(0x361E, 'ERROR_IPSEC_IKE_FAILQUERYSSP') # WinError 13854 (0x361E)
    ERROR_IPSEC_IKE_SRVACQFAIL: Final[ThirtyTwoBits] = cached_win32error(0x361F, 'ERROR_IPSEC_IKE_SRVACQFAIL') # WinError 13855 (0x361F)
    ERROR_IPSEC_IKE_SRVQUERYCRED: Final[ThirtyTwoBits] = cached_win32error(0x3620, 'ERROR_IPSEC_IKE_SRVQUERYCRED') # WinError 13856 (0x3620)
    ERROR_IPSEC_IKE_GETSPIFAIL: Final[ThirtyTwoBits] = cached_win32error(0x3621, 'ERROR_IPSEC_IKE_GETSPIFAIL') # WinError 13857 (0x3621)
    ERROR_IPSEC_IKE_INVALID_FILTER: Final[ThirtyTwoBits] = cached_win32error(0x3622, 'ERROR_IPSEC_IKE_INVALID_FILTER') # WinError 13858 (0x3622)
    ERROR_IPSEC_IKE_OUT_OF_MEMORY: Final[ThirtyTwoBits] = cached_win32error(0x3623, 'ERROR_IPSEC_IKE_OUT_OF_MEMORY') # WinError 13859 (0x3623)
    ERROR_IPSEC_IKE_ADD_UPDATE_KEY_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3624, 'ERROR_IPSEC_IKE_ADD_UPDATE_KEY_FAILED') # WinError 13860 (0x3624)
    ERROR_IPSEC_IKE_INVALID_POLICY: Final[ThirtyTwoBits] = cached_win32error(0x3625, 'ERROR_IPSEC_IKE_INVALID_POLICY') # WinError 13861 (0x3625)
    ERROR_IPSEC_IKE_UNKNOWN_DOI: Final[ThirtyTwoBits] = cached_win32error(0x3626, 'ERROR_IPSEC_IKE_UNKNOWN_DOI') # WinError 13862 (0x3626)
    ERROR_IPSEC_IKE_INVALID_SITUATION: Final[ThirtyTwoBits] = cached_win32error(0x3627, 'ERROR_IPSEC_IKE_INVALID_SITUATION') # WinError 13863 (0x3627)
    ERROR_IPSEC_IKE_DH_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x3628, 'ERROR_IPSEC_IKE_DH_FAILURE') # WinError 13864 (0x3628)
    ERROR_IPSEC_IKE_INVALID_GROUP: Final[ThirtyTwoBits] = cached_win32error(0x3629, 'ERROR_IPSEC_IKE_INVALID_GROUP') # WinError 13865 (0x3629)
    ERROR_IPSEC_IKE_ENCRYPT: Final[ThirtyTwoBits] = cached_win32error(0x362A, 'ERROR_IPSEC_IKE_ENCRYPT') # WinError 13866 (0x362A)
    ERROR_IPSEC_IKE_DECRYPT: Final[ThirtyTwoBits] = cached_win32error(0x362B, 'ERROR_IPSEC_IKE_DECRYPT') # WinError 13867 (0x362B)
    ERROR_IPSEC_IKE_POLICY_MATCH: Final[ThirtyTwoBits] = cached_win32error(0x362C, 'ERROR_IPSEC_IKE_POLICY_MATCH') # WinError 13868 (0x362C)
    ERROR_IPSEC_IKE_UNSUPPORTED_ID: Final[ThirtyTwoBits] = cached_win32error(0x362D, 'ERROR_IPSEC_IKE_UNSUPPORTED_ID') # WinError 13869 (0x362D)
    ERROR_IPSEC_IKE_INVALID_HASH: Final[ThirtyTwoBits] = cached_win32error(0x362E, 'ERROR_IPSEC_IKE_INVALID_HASH') # WinError 13870 (0x362E)
    ERROR_IPSEC_IKE_INVALID_HASH_ALG: Final[ThirtyTwoBits] = cached_win32error(0x362F, 'ERROR_IPSEC_IKE_INVALID_HASH_ALG') # WinError 13871 (0x362F)
    ERROR_IPSEC_IKE_INVALID_HASH_SIZE: Final[ThirtyTwoBits] = cached_win32error(0x3630, 'ERROR_IPSEC_IKE_INVALID_HASH_SIZE') # WinError 13872 (0x3630)
    ERROR_IPSEC_IKE_INVALID_ENCRYPT_ALG: Final[ThirtyTwoBits] = cached_win32error(0x3631, 'ERROR_IPSEC_IKE_INVALID_ENCRYPT_ALG') # WinError 13873 (0x3631)
    ERROR_IPSEC_IKE_INVALID_AUTH_ALG: Final[ThirtyTwoBits] = cached_win32error(0x3632, 'ERROR_IPSEC_IKE_INVALID_AUTH_ALG') # WinError 13874 (0x3632)
    ERROR_IPSEC_IKE_INVALID_SIG: Final[ThirtyTwoBits] = cached_win32error(0x3633, 'ERROR_IPSEC_IKE_INVALID_SIG') # WinError 13875 (0x3633)
    ERROR_IPSEC_IKE_LOAD_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3634, 'ERROR_IPSEC_IKE_LOAD_FAILED') # WinError 13876 (0x3634)
    ERROR_IPSEC_IKE_RPC_DELETE: Final[ThirtyTwoBits] = cached_win32error(0x3635, 'ERROR_IPSEC_IKE_RPC_DELETE') # WinError 13877 (0x3635)
    ERROR_IPSEC_IKE_BENIGN_REINIT: Final[ThirtyTwoBits] = cached_win32error(0x3636, 'ERROR_IPSEC_IKE_BENIGN_REINIT') # WinError 13878 (0x3636)
    ERROR_IPSEC_IKE_INVALID_RESPONDER_LIFETIME_NOTIFY: Final[ThirtyTwoBits] = cached_win32error(0x3637, 'ERROR_IPSEC_IKE_INVALID_RESPONDER_LIFETIME_NOTIFY') # WinError 13879 (0x3637)
    ERROR_IPSEC_IKE_INVALID_MAJOR_VERSION: Final[ThirtyTwoBits] = cached_win32error(0x3638, 'ERROR_IPSEC_IKE_INVALID_MAJOR_VERSION') # WinError 13880 (0x3638)
    ERROR_IPSEC_IKE_INVALID_CERT_KEYLEN: Final[ThirtyTwoBits] = cached_win32error(0x3639, 'ERROR_IPSEC_IKE_INVALID_CERT_KEYLEN') # WinError 13881 (0x3639)
    ERROR_IPSEC_IKE_MM_LIMIT: Final[ThirtyTwoBits] = cached_win32error(0x363A, 'ERROR_IPSEC_IKE_MM_LIMIT') # WinError 13882 (0x363A)
    ERROR_IPSEC_IKE_NEGOTIATION_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x363B, 'ERROR_IPSEC_IKE_NEGOTIATION_DISABLED') # WinError 13883 (0x363B)
    ERROR_IPSEC_IKE_QM_LIMIT: Final[ThirtyTwoBits] = cached_win32error(0x363C, 'ERROR_IPSEC_IKE_QM_LIMIT') # WinError 13884 (0x363C)
    ERROR_IPSEC_IKE_MM_EXPIRED: Final[ThirtyTwoBits] = cached_win32error(0x363D, 'ERROR_IPSEC_IKE_MM_EXPIRED') # WinError 13885 (0x363D)
    ERROR_IPSEC_IKE_PEER_MM_ASSUMED_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x363E, 'ERROR_IPSEC_IKE_PEER_MM_ASSUMED_INVALID') # WinError 13886 (0x363E)
    ERROR_IPSEC_IKE_CERT_CHAIN_POLICY_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x363F, 'ERROR_IPSEC_IKE_CERT_CHAIN_POLICY_MISMATCH') # WinError 13887 (0x363F)
    ERROR_IPSEC_IKE_UNEXPECTED_MESSAGE_ID: Final[ThirtyTwoBits] = cached_win32error(0x3640, 'ERROR_IPSEC_IKE_UNEXPECTED_MESSAGE_ID') # WinError 13888 (0x3640)
    ERROR_IPSEC_IKE_INVALID_AUTH_PAYLOAD: Final[ThirtyTwoBits] = cached_win32error(0x3641, 'ERROR_IPSEC_IKE_INVALID_AUTH_PAYLOAD') # WinError 13889 (0x3641)
    ERROR_IPSEC_IKE_DOS_COOKIE_SENT: Final[ThirtyTwoBits] = cached_win32error(0x3642, 'ERROR_IPSEC_IKE_DOS_COOKIE_SENT') # WinError 13890 (0x3642)
    ERROR_IPSEC_IKE_SHUTTING_DOWN: Final[ThirtyTwoBits] = cached_win32error(0x3643, 'ERROR_IPSEC_IKE_SHUTTING_DOWN') # WinError 13891 (0x3643)
    ERROR_IPSEC_IKE_CGA_AUTH_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3644, 'ERROR_IPSEC_IKE_CGA_AUTH_FAILED') # WinError 13892 (0x3644)
    ERROR_IPSEC_IKE_PROCESS_ERR_NATOA: Final[ThirtyTwoBits] = cached_win32error(0x3645, 'ERROR_IPSEC_IKE_PROCESS_ERR_NATOA') # WinError 13893 (0x3645)
    ERROR_IPSEC_IKE_INVALID_MM_FOR_QM: Final[ThirtyTwoBits] = cached_win32error(0x3646, 'ERROR_IPSEC_IKE_INVALID_MM_FOR_QM') # WinError 13894 (0x3646)
    ERROR_IPSEC_IKE_QM_EXPIRED: Final[ThirtyTwoBits] = cached_win32error(0x3647, 'ERROR_IPSEC_IKE_QM_EXPIRED') # WinError 13895 (0x3647)
    ERROR_IPSEC_IKE_TOO_MANY_FILTERS: Final[ThirtyTwoBits] = cached_win32error(0x3648, 'ERROR_IPSEC_IKE_TOO_MANY_FILTERS') # WinError 13896 (0x3648)
    ERROR_IPSEC_IKE_NEG_STATUS_END: Final[ThirtyTwoBits] = cached_win32error(0x3649, 'ERROR_IPSEC_IKE_NEG_STATUS_END') # WinError 13897 (0x3649)
    ERROR_IPSEC_IKE_KILL_DUMMY_NAP_TUNNEL: Final[ThirtyTwoBits] = cached_win32error(0x364A, 'ERROR_IPSEC_IKE_KILL_DUMMY_NAP_TUNNEL') # WinError 13898 (0x364A)
    ERROR_IPSEC_IKE_INNER_IP_ASSIGNMENT_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x364B, 'ERROR_IPSEC_IKE_INNER_IP_ASSIGNMENT_FAILURE') # WinError 13899 (0x364B)
    ERROR_IPSEC_IKE_REQUIRE_CP_PAYLOAD_MISSING: Final[ThirtyTwoBits] = cached_win32error(0x364C, 'ERROR_IPSEC_IKE_REQUIRE_CP_PAYLOAD_MISSING') # WinError 13900 (0x364C)
    ERROR_IPSEC_KEY_MODULE_IMPERSONATION_NEGOTIATION_PENDING: Final[ThirtyTwoBits] = cached_win32error(0x364D, 'ERROR_IPSEC_KEY_MODULE_IMPERSONATION_NEGOTIATION_PENDING') # WinError 13901 (0x364D)
    ERROR_IPSEC_IKE_COEXISTENCE_SUPPRESS: Final[ThirtyTwoBits] = cached_win32error(0x364E, 'ERROR_IPSEC_IKE_COEXISTENCE_SUPPRESS') # WinError 13902 (0x364E)
    ERROR_IPSEC_IKE_RATELIMIT_DROP: Final[ThirtyTwoBits] = cached_win32error(0x364F, 'ERROR_IPSEC_IKE_RATELIMIT_DROP') # WinError 13903 (0x364F)
    ERROR_IPSEC_IKE_PEER_DOESNT_SUPPORT_MOBIKE: Final[ThirtyTwoBits] = cached_win32error(0x3650, 'ERROR_IPSEC_IKE_PEER_DOESNT_SUPPORT_MOBIKE') # WinError 13904 (0x3650)
    ERROR_IPSEC_IKE_AUTHORIZATION_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x3651, 'ERROR_IPSEC_IKE_AUTHORIZATION_FAILURE') # WinError 13905 (0x3651)
    ERROR_IPSEC_IKE_STRONG_CRED_AUTHORIZATION_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x3652, 'ERROR_IPSEC_IKE_STRONG_CRED_AUTHORIZATION_FAILURE') # WinError 13906 (0x3652)
    ERROR_IPSEC_IKE_AUTHORIZATION_FAILURE_WITH_OPTIONAL_RETRY: Final[ThirtyTwoBits] = cached_win32error(0x3653, 'ERROR_IPSEC_IKE_AUTHORIZATION_FAILURE_WITH_OPTIONAL_RETRY') # WinError 13907 (0x3653)
    ERROR_IPSEC_IKE_STRONG_CRED_AUTHORIZATION_AND_CERTMAP_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x3654, 'ERROR_IPSEC_IKE_STRONG_CRED_AUTHORIZATION_AND_CERTMAP_FAILURE') # WinError 13908 (0x3654)
    ERROR_IPSEC_IKE_NEG_STATUS_EXTENDED_END: Final[ThirtyTwoBits] = cached_win32error(0x3655, 'ERROR_IPSEC_IKE_NEG_STATUS_EXTENDED_END') # WinError 13909 (0x3655)
    ERROR_IPSEC_BAD_SPI: Final[ThirtyTwoBits] = cached_win32error(0x3656, 'ERROR_IPSEC_BAD_SPI') # WinError 13910 (0x3656)
    ERROR_IPSEC_SA_LIFETIME_EXPIRED: Final[ThirtyTwoBits] = cached_win32error(0x3657, 'ERROR_IPSEC_SA_LIFETIME_EXPIRED') # WinError 13911 (0x3657)
    ERROR_IPSEC_WRONG_SA: Final[ThirtyTwoBits] = cached_win32error(0x3658, 'ERROR_IPSEC_WRONG_SA') # WinError 13912 (0x3658)
    ERROR_IPSEC_REPLAY_CHECK_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3659, 'ERROR_IPSEC_REPLAY_CHECK_FAILED') # WinError 13913 (0x3659)
    ERROR_IPSEC_INVALID_PACKET: Final[ThirtyTwoBits] = cached_win32error(0x365A, 'ERROR_IPSEC_INVALID_PACKET') # WinError 13914 (0x365A)
    ERROR_IPSEC_INTEGRITY_CHECK_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x365B, 'ERROR_IPSEC_INTEGRITY_CHECK_FAILED') # WinError 13915 (0x365B)
    ERROR_IPSEC_CLEAR_TEXT_DROP: Final[ThirtyTwoBits] = cached_win32error(0x365C, 'ERROR_IPSEC_CLEAR_TEXT_DROP') # WinError 13916 (0x365C)
    ERROR_IPSEC_AUTH_FIREWALL_DROP: Final[ThirtyTwoBits] = cached_win32error(0x365D, 'ERROR_IPSEC_AUTH_FIREWALL_DROP') # WinError 13917 (0x365D)
    ERROR_IPSEC_THROTTLE_DROP: Final[ThirtyTwoBits] = cached_win32error(0x365E, 'ERROR_IPSEC_THROTTLE_DROP') # WinError 13918 (0x365E)
    ERROR_IPSEC_DOSP_BLOCK: Final[ThirtyTwoBits] = cached_win32error(0x3665, 'ERROR_IPSEC_DOSP_BLOCK') # WinError 13925 (0x3665)
    ERROR_IPSEC_DOSP_RECEIVED_MULTICAST: Final[ThirtyTwoBits] = cached_win32error(0x3666, 'ERROR_IPSEC_DOSP_RECEIVED_MULTICAST') # WinError 13926 (0x3666)
    ERROR_IPSEC_DOSP_INVALID_PACKET: Final[ThirtyTwoBits] = cached_win32error(0x3667, 'ERROR_IPSEC_DOSP_INVALID_PACKET') # WinError 13927 (0x3667)
    ERROR_IPSEC_DOSP_STATE_LOOKUP_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3668, 'ERROR_IPSEC_DOSP_STATE_LOOKUP_FAILED') # WinError 13928 (0x3668)
    ERROR_IPSEC_DOSP_MAX_ENTRIES: Final[ThirtyTwoBits] = cached_win32error(0x3669, 'ERROR_IPSEC_DOSP_MAX_ENTRIES') # WinError 13929 (0x3669)
    ERROR_IPSEC_DOSP_KEYMOD_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x366A, 'ERROR_IPSEC_DOSP_KEYMOD_NOT_ALLOWED') # WinError 13930 (0x366A)
    ERROR_IPSEC_DOSP_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_win32error(0x366B, 'ERROR_IPSEC_DOSP_NOT_INSTALLED') # WinError 13931 (0x366B)
    ERROR_IPSEC_DOSP_MAX_PER_IP_RATELIMIT_QUEUES: Final[ThirtyTwoBits] = cached_win32error(0x366C, 'ERROR_IPSEC_DOSP_MAX_PER_IP_RATELIMIT_QUEUES') # WinError 13932 (0x366C)
    ERROR_SXS_SECTION_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x36B0, 'ERROR_SXS_SECTION_NOT_FOUND') # WinError 14000 (0x36B0)
    ERROR_SXS_CANT_GEN_ACTCTX: Final[ThirtyTwoBits] = cached_win32error(0x36B1, 'ERROR_SXS_CANT_GEN_ACTCTX') # WinError 14001 (0x36B1)
    ERROR_SXS_INVALID_ACTCTXDATA_FORMAT: Final[ThirtyTwoBits] = cached_win32error(0x36B2, 'ERROR_SXS_INVALID_ACTCTXDATA_FORMAT') # WinError 14002 (0x36B2)
    ERROR_SXS_ASSEMBLY_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x36B3, 'ERROR_SXS_ASSEMBLY_NOT_FOUND') # WinError 14003 (0x36B3)
    ERROR_SXS_MANIFEST_FORMAT_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x36B4, 'ERROR_SXS_MANIFEST_FORMAT_ERROR') # WinError 14004 (0x36B4)
    ERROR_SXS_MANIFEST_PARSE_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x36B5, 'ERROR_SXS_MANIFEST_PARSE_ERROR') # WinError 14005 (0x36B5)
    ERROR_SXS_ACTIVATION_CONTEXT_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x36B6, 'ERROR_SXS_ACTIVATION_CONTEXT_DISABLED') # WinError 14006 (0x36B6)
    ERROR_SXS_KEY_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x36B7, 'ERROR_SXS_KEY_NOT_FOUND') # WinError 14007 (0x36B7)
    ERROR_SXS_VERSION_CONFLICT: Final[ThirtyTwoBits] = cached_win32error(0x36B8, 'ERROR_SXS_VERSION_CONFLICT') # WinError 14008 (0x36B8)
    ERROR_SXS_WRONG_SECTION_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x36B9, 'ERROR_SXS_WRONG_SECTION_TYPE') # WinError 14009 (0x36B9)
    ERROR_SXS_THREAD_QUERIES_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x36BA, 'ERROR_SXS_THREAD_QUERIES_DISABLED') # WinError 14010 (0x36BA)
    ERROR_SXS_PROCESS_DEFAULT_ALREADY_SET: Final[ThirtyTwoBits] = cached_win32error(0x36BB, 'ERROR_SXS_PROCESS_DEFAULT_ALREADY_SET') # WinError 14011 (0x36BB)
    ERROR_SXS_UNKNOWN_ENCODING_GROUP: Final[ThirtyTwoBits] = cached_win32error(0x36BC, 'ERROR_SXS_UNKNOWN_ENCODING_GROUP') # WinError 14012 (0x36BC)
    ERROR_SXS_UNKNOWN_ENCODING: Final[ThirtyTwoBits] = cached_win32error(0x36BD, 'ERROR_SXS_UNKNOWN_ENCODING') # WinError 14013 (0x36BD)
    ERROR_SXS_INVALID_XML_NAMESPACE_URI: Final[ThirtyTwoBits] = cached_win32error(0x36BE, 'ERROR_SXS_INVALID_XML_NAMESPACE_URI') # WinError 14014 (0x36BE)
    ERROR_SXS_ROOT_MANIFEST_DEPENDENCY_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_win32error(0x36BF, 'ERROR_SXS_ROOT_MANIFEST_DEPENDENCY_NOT_INSTALLED') # WinError 14015 (0x36BF)
    ERROR_SXS_LEAF_MANIFEST_DEPENDENCY_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_win32error(0x36C0, 'ERROR_SXS_LEAF_MANIFEST_DEPENDENCY_NOT_INSTALLED') # WinError 14016 (0x36C0)
    ERROR_SXS_INVALID_ASSEMBLY_IDENTITY_ATTRIBUTE: Final[ThirtyTwoBits] = cached_win32error(0x36C1, 'ERROR_SXS_INVALID_ASSEMBLY_IDENTITY_ATTRIBUTE') # WinError 14017 (0x36C1)
    ERROR_SXS_MANIFEST_MISSING_REQUIRED_DEFAULT_NAMESPACE: Final[ThirtyTwoBits] = cached_win32error(0x36C2, 'ERROR_SXS_MANIFEST_MISSING_REQUIRED_DEFAULT_NAMESPACE') # WinError 14018 (0x36C2)
    ERROR_SXS_MANIFEST_INVALID_REQUIRED_DEFAULT_NAMESPACE: Final[ThirtyTwoBits] = cached_win32error(0x36C3, 'ERROR_SXS_MANIFEST_INVALID_REQUIRED_DEFAULT_NAMESPACE') # WinError 14019 (0x36C3)
    ERROR_SXS_PRIVATE_MANIFEST_CROSS_PATH_WITH_REPARSE_POINT: Final[ThirtyTwoBits] = cached_win32error(0x36C4, 'ERROR_SXS_PRIVATE_MANIFEST_CROSS_PATH_WITH_REPARSE_POINT') # WinError 14020 (0x36C4)
    ERROR_SXS_DUPLICATE_DLL_NAME: Final[ThirtyTwoBits] = cached_win32error(0x36C5, 'ERROR_SXS_DUPLICATE_DLL_NAME') # WinError 14021 (0x36C5)
    ERROR_SXS_DUPLICATE_WINDOWCLASS_NAME: Final[ThirtyTwoBits] = cached_win32error(0x36C6, 'ERROR_SXS_DUPLICATE_WINDOWCLASS_NAME') # WinError 14022 (0x36C6)
    ERROR_SXS_DUPLICATE_CLSID: Final[ThirtyTwoBits] = cached_win32error(0x36C7, 'ERROR_SXS_DUPLICATE_CLSID') # WinError 14023 (0x36C7)
    ERROR_SXS_DUPLICATE_IID: Final[ThirtyTwoBits] = cached_win32error(0x36C8, 'ERROR_SXS_DUPLICATE_IID') # WinError 14024 (0x36C8)
    ERROR_SXS_DUPLICATE_TLBID: Final[ThirtyTwoBits] = cached_win32error(0x36C9, 'ERROR_SXS_DUPLICATE_TLBID') # WinError 14025 (0x36C9)
    ERROR_SXS_DUPLICATE_PROGID: Final[ThirtyTwoBits] = cached_win32error(0x36CA, 'ERROR_SXS_DUPLICATE_PROGID') # WinError 14026 (0x36CA)
    ERROR_SXS_DUPLICATE_ASSEMBLY_NAME: Final[ThirtyTwoBits] = cached_win32error(0x36CB, 'ERROR_SXS_DUPLICATE_ASSEMBLY_NAME') # WinError 14027 (0x36CB)
    ERROR_SXS_FILE_HASH_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x36CC, 'ERROR_SXS_FILE_HASH_MISMATCH') # WinError 14028 (0x36CC)
    ERROR_SXS_POLICY_PARSE_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x36CD, 'ERROR_SXS_POLICY_PARSE_ERROR') # WinError 14029 (0x36CD)
    ERROR_SXS_XML_E_MISSINGQUOTE: Final[ThirtyTwoBits] = cached_win32error(0x36CE, 'ERROR_SXS_XML_E_MISSINGQUOTE') # WinError 14030 (0x36CE)
    ERROR_SXS_XML_E_COMMENTSYNTAX: Final[ThirtyTwoBits] = cached_win32error(0x36CF, 'ERROR_SXS_XML_E_COMMENTSYNTAX') # WinError 14031 (0x36CF)
    ERROR_SXS_XML_E_BADSTARTNAMECHAR: Final[ThirtyTwoBits] = cached_win32error(0x36D0, 'ERROR_SXS_XML_E_BADSTARTNAMECHAR') # WinError 14032 (0x36D0)
    ERROR_SXS_XML_E_BADNAMECHAR: Final[ThirtyTwoBits] = cached_win32error(0x36D1, 'ERROR_SXS_XML_E_BADNAMECHAR') # WinError 14033 (0x36D1)
    ERROR_SXS_XML_E_BADCHARINSTRING: Final[ThirtyTwoBits] = cached_win32error(0x36D2, 'ERROR_SXS_XML_E_BADCHARINSTRING') # WinError 14034 (0x36D2)
    ERROR_SXS_XML_E_XMLDECLSYNTAX: Final[ThirtyTwoBits] = cached_win32error(0x36D3, 'ERROR_SXS_XML_E_XMLDECLSYNTAX') # WinError 14035 (0x36D3)
    ERROR_SXS_XML_E_BADCHARDATA: Final[ThirtyTwoBits] = cached_win32error(0x36D4, 'ERROR_SXS_XML_E_BADCHARDATA') # WinError 14036 (0x36D4)
    ERROR_SXS_XML_E_MISSINGWHITESPACE: Final[ThirtyTwoBits] = cached_win32error(0x36D5, 'ERROR_SXS_XML_E_MISSINGWHITESPACE') # WinError 14037 (0x36D5)
    ERROR_SXS_XML_E_EXPECTINGTAGEND: Final[ThirtyTwoBits] = cached_win32error(0x36D6, 'ERROR_SXS_XML_E_EXPECTINGTAGEND') # WinError 14038 (0x36D6)
    ERROR_SXS_XML_E_MISSINGSEMICOLON: Final[ThirtyTwoBits] = cached_win32error(0x36D7, 'ERROR_SXS_XML_E_MISSINGSEMICOLON') # WinError 14039 (0x36D7)
    ERROR_SXS_XML_E_UNBALANCEDPAREN: Final[ThirtyTwoBits] = cached_win32error(0x36D8, 'ERROR_SXS_XML_E_UNBALANCEDPAREN') # WinError 14040 (0x36D8)
    ERROR_SXS_XML_E_INTERNALERROR: Final[ThirtyTwoBits] = cached_win32error(0x36D9, 'ERROR_SXS_XML_E_INTERNALERROR') # WinError 14041 (0x36D9)
    ERROR_SXS_XML_E_UNEXPECTED_WHITESPACE: Final[ThirtyTwoBits] = cached_win32error(0x36DA, 'ERROR_SXS_XML_E_UNEXPECTED_WHITESPACE') # WinError 14042 (0x36DA)
    ERROR_SXS_XML_E_INCOMPLETE_ENCODING: Final[ThirtyTwoBits] = cached_win32error(0x36DB, 'ERROR_SXS_XML_E_INCOMPLETE_ENCODING') # WinError 14043 (0x36DB)
    ERROR_SXS_XML_E_MISSING_PAREN: Final[ThirtyTwoBits] = cached_win32error(0x36DC, 'ERROR_SXS_XML_E_MISSING_PAREN') # WinError 14044 (0x36DC)
    ERROR_SXS_XML_E_EXPECTINGCLOSEQUOTE: Final[ThirtyTwoBits] = cached_win32error(0x36DD, 'ERROR_SXS_XML_E_EXPECTINGCLOSEQUOTE') # WinError 14045 (0x36DD)
    ERROR_SXS_XML_E_MULTIPLE_COLONS: Final[ThirtyTwoBits] = cached_win32error(0x36DE, 'ERROR_SXS_XML_E_MULTIPLE_COLONS') # WinError 14046 (0x36DE)
    ERROR_SXS_XML_E_INVALID_DECIMAL: Final[ThirtyTwoBits] = cached_win32error(0x36DF, 'ERROR_SXS_XML_E_INVALID_DECIMAL') # WinError 14047 (0x36DF)
    ERROR_SXS_XML_E_INVALID_HEXIDECIMAL: Final[ThirtyTwoBits] = cached_win32error(0x36E0, 'ERROR_SXS_XML_E_INVALID_HEXIDECIMAL') # WinError 14048 (0x36E0)
    ERROR_SXS_XML_E_INVALID_UNICODE: Final[ThirtyTwoBits] = cached_win32error(0x36E1, 'ERROR_SXS_XML_E_INVALID_UNICODE') # WinError 14049 (0x36E1)
    ERROR_SXS_XML_E_WHITESPACEORQUESTIONMARK: Final[ThirtyTwoBits] = cached_win32error(0x36E2, 'ERROR_SXS_XML_E_WHITESPACEORQUESTIONMARK') # WinError 14050 (0x36E2)
    ERROR_SXS_XML_E_UNEXPECTEDENDTAG: Final[ThirtyTwoBits] = cached_win32error(0x36E3, 'ERROR_SXS_XML_E_UNEXPECTEDENDTAG') # WinError 14051 (0x36E3)
    ERROR_SXS_XML_E_UNCLOSEDTAG: Final[ThirtyTwoBits] = cached_win32error(0x36E4, 'ERROR_SXS_XML_E_UNCLOSEDTAG') # WinError 14052 (0x36E4)
    ERROR_SXS_XML_E_DUPLICATEATTRIBUTE: Final[ThirtyTwoBits] = cached_win32error(0x36E5, 'ERROR_SXS_XML_E_DUPLICATEATTRIBUTE') # WinError 14053 (0x36E5)
    ERROR_SXS_XML_E_MULTIPLEROOTS: Final[ThirtyTwoBits] = cached_win32error(0x36E6, 'ERROR_SXS_XML_E_MULTIPLEROOTS') # WinError 14054 (0x36E6)
    ERROR_SXS_XML_E_INVALIDATROOTLEVEL: Final[ThirtyTwoBits] = cached_win32error(0x36E7, 'ERROR_SXS_XML_E_INVALIDATROOTLEVEL') # WinError 14055 (0x36E7)
    ERROR_SXS_XML_E_BADXMLDECL: Final[ThirtyTwoBits] = cached_win32error(0x36E8, 'ERROR_SXS_XML_E_BADXMLDECL') # WinError 14056 (0x36E8)
    ERROR_SXS_XML_E_MISSINGROOT: Final[ThirtyTwoBits] = cached_win32error(0x36E9, 'ERROR_SXS_XML_E_MISSINGROOT') # WinError 14057 (0x36E9)
    ERROR_SXS_XML_E_UNEXPECTEDEOF: Final[ThirtyTwoBits] = cached_win32error(0x36EA, 'ERROR_SXS_XML_E_UNEXPECTEDEOF') # WinError 14058 (0x36EA)
    ERROR_SXS_XML_E_BADPEREFINSUBSET: Final[ThirtyTwoBits] = cached_win32error(0x36EB, 'ERROR_SXS_XML_E_BADPEREFINSUBSET') # WinError 14059 (0x36EB)
    ERROR_SXS_XML_E_UNCLOSEDSTARTTAG: Final[ThirtyTwoBits] = cached_win32error(0x36EC, 'ERROR_SXS_XML_E_UNCLOSEDSTARTTAG') # WinError 14060 (0x36EC)
    ERROR_SXS_XML_E_UNCLOSEDENDTAG: Final[ThirtyTwoBits] = cached_win32error(0x36ED, 'ERROR_SXS_XML_E_UNCLOSEDENDTAG') # WinError 14061 (0x36ED)
    ERROR_SXS_XML_E_UNCLOSEDSTRING: Final[ThirtyTwoBits] = cached_win32error(0x36EE, 'ERROR_SXS_XML_E_UNCLOSEDSTRING') # WinError 14062 (0x36EE)
    ERROR_SXS_XML_E_UNCLOSEDCOMMENT: Final[ThirtyTwoBits] = cached_win32error(0x36EF, 'ERROR_SXS_XML_E_UNCLOSEDCOMMENT') # WinError 14063 (0x36EF)
    ERROR_SXS_XML_E_UNCLOSEDDECL: Final[ThirtyTwoBits] = cached_win32error(0x36F0, 'ERROR_SXS_XML_E_UNCLOSEDDECL') # WinError 14064 (0x36F0)
    ERROR_SXS_XML_E_UNCLOSEDCDATA: Final[ThirtyTwoBits] = cached_win32error(0x36F1, 'ERROR_SXS_XML_E_UNCLOSEDCDATA') # WinError 14065 (0x36F1)
    ERROR_SXS_XML_E_RESERVEDNAMESPACE: Final[ThirtyTwoBits] = cached_win32error(0x36F2, 'ERROR_SXS_XML_E_RESERVEDNAMESPACE') # WinError 14066 (0x36F2)
    ERROR_SXS_XML_E_INVALIDENCODING: Final[ThirtyTwoBits] = cached_win32error(0x36F3, 'ERROR_SXS_XML_E_INVALIDENCODING') # WinError 14067 (0x36F3)
    ERROR_SXS_XML_E_INVALIDSWITCH: Final[ThirtyTwoBits] = cached_win32error(0x36F4, 'ERROR_SXS_XML_E_INVALIDSWITCH') # WinError 14068 (0x36F4)
    ERROR_SXS_XML_E_BADXMLCASE: Final[ThirtyTwoBits] = cached_win32error(0x36F5, 'ERROR_SXS_XML_E_BADXMLCASE') # WinError 14069 (0x36F5)
    ERROR_SXS_XML_E_INVALID_STANDALONE: Final[ThirtyTwoBits] = cached_win32error(0x36F6, 'ERROR_SXS_XML_E_INVALID_STANDALONE') # WinError 14070 (0x36F6)
    ERROR_SXS_XML_E_UNEXPECTED_STANDALONE: Final[ThirtyTwoBits] = cached_win32error(0x36F7, 'ERROR_SXS_XML_E_UNEXPECTED_STANDALONE') # WinError 14071 (0x36F7)
    ERROR_SXS_XML_E_INVALID_VERSION: Final[ThirtyTwoBits] = cached_win32error(0x36F8, 'ERROR_SXS_XML_E_INVALID_VERSION') # WinError 14072 (0x36F8)
    ERROR_SXS_XML_E_MISSINGEQUALS: Final[ThirtyTwoBits] = cached_win32error(0x36F9, 'ERROR_SXS_XML_E_MISSINGEQUALS') # WinError 14073 (0x36F9)
    ERROR_SXS_PROTECTION_RECOVERY_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x36FA, 'ERROR_SXS_PROTECTION_RECOVERY_FAILED') # WinError 14074 (0x36FA)
    ERROR_SXS_PROTECTION_PUBLIC_KEY_TOO_SHORT: Final[ThirtyTwoBits] = cached_win32error(0x36FB, 'ERROR_SXS_PROTECTION_PUBLIC_KEY_TOO_SHORT') # WinError 14075 (0x36FB)
    ERROR_SXS_PROTECTION_CATALOG_NOT_VALID: Final[ThirtyTwoBits] = cached_win32error(0x36FC, 'ERROR_SXS_PROTECTION_CATALOG_NOT_VALID') # WinError 14076 (0x36FC)
    ERROR_SXS_UNTRANSLATABLE_HRESULT: Final[ThirtyTwoBits] = cached_win32error(0x36FD, 'ERROR_SXS_UNTRANSLATABLE_HRESULT') # WinError 14077 (0x36FD)
    ERROR_SXS_PROTECTION_CATALOG_FILE_MISSING: Final[ThirtyTwoBits] = cached_win32error(0x36FE, 'ERROR_SXS_PROTECTION_CATALOG_FILE_MISSING') # WinError 14078 (0x36FE)
    ERROR_SXS_MISSING_ASSEMBLY_IDENTITY_ATTRIBUTE: Final[ThirtyTwoBits] = cached_win32error(0x36FF, 'ERROR_SXS_MISSING_ASSEMBLY_IDENTITY_ATTRIBUTE') # WinError 14079 (0x36FF)
    ERROR_SXS_INVALID_ASSEMBLY_IDENTITY_ATTRIBUTE_NAME: Final[ThirtyTwoBits] = cached_win32error(0x3700, 'ERROR_SXS_INVALID_ASSEMBLY_IDENTITY_ATTRIBUTE_NAME') # WinError 14080 (0x3700)
    ERROR_SXS_ASSEMBLY_MISSING: Final[ThirtyTwoBits] = cached_win32error(0x3701, 'ERROR_SXS_ASSEMBLY_MISSING') # WinError 14081 (0x3701)
    ERROR_SXS_CORRUPT_ACTIVATION_STACK: Final[ThirtyTwoBits] = cached_win32error(0x3702, 'ERROR_SXS_CORRUPT_ACTIVATION_STACK') # WinError 14082 (0x3702)
    ERROR_SXS_CORRUPTION: Final[ThirtyTwoBits] = cached_win32error(0x3703, 'ERROR_SXS_CORRUPTION') # WinError 14083 (0x3703)
    ERROR_SXS_EARLY_DEACTIVATION: Final[ThirtyTwoBits] = cached_win32error(0x3704, 'ERROR_SXS_EARLY_DEACTIVATION') # WinError 14084 (0x3704)
    ERROR_SXS_INVALID_DEACTIVATION: Final[ThirtyTwoBits] = cached_win32error(0x3705, 'ERROR_SXS_INVALID_DEACTIVATION') # WinError 14085 (0x3705)
    ERROR_SXS_MULTIPLE_DEACTIVATION: Final[ThirtyTwoBits] = cached_win32error(0x3706, 'ERROR_SXS_MULTIPLE_DEACTIVATION') # WinError 14086 (0x3706)
    ERROR_SXS_PROCESS_TERMINATION_REQUESTED: Final[ThirtyTwoBits] = cached_win32error(0x3707, 'ERROR_SXS_PROCESS_TERMINATION_REQUESTED') # WinError 14087 (0x3707)
    ERROR_SXS_RELEASE_ACTIVATION_CONTEXT: Final[ThirtyTwoBits] = cached_win32error(0x3708, 'ERROR_SXS_RELEASE_ACTIVATION_CONTEXT') # WinError 14088 (0x3708)
    ERROR_SXS_SYSTEM_DEFAULT_ACTIVATION_CONTEXT_EMPTY: Final[ThirtyTwoBits] = cached_win32error(0x3709, 'ERROR_SXS_SYSTEM_DEFAULT_ACTIVATION_CONTEXT_EMPTY') # WinError 14089 (0x3709)
    ERROR_SXS_INVALID_IDENTITY_ATTRIBUTE_VALUE: Final[ThirtyTwoBits] = cached_win32error(0x370A, 'ERROR_SXS_INVALID_IDENTITY_ATTRIBUTE_VALUE') # WinError 14090 (0x370A)
    ERROR_SXS_INVALID_IDENTITY_ATTRIBUTE_NAME: Final[ThirtyTwoBits] = cached_win32error(0x370B, 'ERROR_SXS_INVALID_IDENTITY_ATTRIBUTE_NAME') # WinError 14091 (0x370B)
    ERROR_SXS_IDENTITY_DUPLICATE_ATTRIBUTE: Final[ThirtyTwoBits] = cached_win32error(0x370C, 'ERROR_SXS_IDENTITY_DUPLICATE_ATTRIBUTE') # WinError 14092 (0x370C)
    ERROR_SXS_IDENTITY_PARSE_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x370D, 'ERROR_SXS_IDENTITY_PARSE_ERROR') # WinError 14093 (0x370D)
    ERROR_MALFORMED_SUBSTITUTION_STRING: Final[ThirtyTwoBits] = cached_win32error(0x370E, 'ERROR_MALFORMED_SUBSTITUTION_STRING') # WinError 14094 (0x370E)
    ERROR_SXS_INCORRECT_PUBLIC_KEY_TOKEN: Final[ThirtyTwoBits] = cached_win32error(0x370F, 'ERROR_SXS_INCORRECT_PUBLIC_KEY_TOKEN') # WinError 14095 (0x370F)
    ERROR_UNMAPPED_SUBSTITUTION_STRING: Final[ThirtyTwoBits] = cached_win32error(0x3710, 'ERROR_UNMAPPED_SUBSTITUTION_STRING') # WinError 14096 (0x3710)
    ERROR_SXS_ASSEMBLY_NOT_LOCKED: Final[ThirtyTwoBits] = cached_win32error(0x3711, 'ERROR_SXS_ASSEMBLY_NOT_LOCKED') # WinError 14097 (0x3711)
    ERROR_SXS_COMPONENT_STORE_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x3712, 'ERROR_SXS_COMPONENT_STORE_CORRUPT') # WinError 14098 (0x3712)
    ERROR_ADVANCED_INSTALLER_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3713, 'ERROR_ADVANCED_INSTALLER_FAILED') # WinError 14099 (0x3713)
    ERROR_XML_ENCODING_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x3714, 'ERROR_XML_ENCODING_MISMATCH') # WinError 14100 (0x3714)
    ERROR_SXS_MANIFEST_IDENTITY_SAME_BUT_CONTENTS_DIFFERENT: Final[ThirtyTwoBits] = cached_win32error(0x3715, 'ERROR_SXS_MANIFEST_IDENTITY_SAME_BUT_CONTENTS_DIFFERENT') # WinError 14101 (0x3715)
    ERROR_SXS_IDENTITIES_DIFFERENT: Final[ThirtyTwoBits] = cached_win32error(0x3716, 'ERROR_SXS_IDENTITIES_DIFFERENT') # WinError 14102 (0x3716)
    ERROR_SXS_ASSEMBLY_IS_NOT_A_DEPLOYMENT: Final[ThirtyTwoBits] = cached_win32error(0x3717, 'ERROR_SXS_ASSEMBLY_IS_NOT_A_DEPLOYMENT') # WinError 14103 (0x3717)
    ERROR_SXS_FILE_NOT_PART_OF_ASSEMBLY: Final[ThirtyTwoBits] = cached_win32error(0x3718, 'ERROR_SXS_FILE_NOT_PART_OF_ASSEMBLY') # WinError 14104 (0x3718)
    ERROR_SXS_MANIFEST_TOO_BIG: Final[ThirtyTwoBits] = cached_win32error(0x3719, 'ERROR_SXS_MANIFEST_TOO_BIG') # WinError 14105 (0x3719)
    ERROR_SXS_SETTING_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_win32error(0x371A, 'ERROR_SXS_SETTING_NOT_REGISTERED') # WinError 14106 (0x371A)
    ERROR_SXS_TRANSACTION_CLOSURE_INCOMPLETE: Final[ThirtyTwoBits] = cached_win32error(0x371B, 'ERROR_SXS_TRANSACTION_CLOSURE_INCOMPLETE') # WinError 14107 (0x371B)
    ERROR_SMI_PRIMITIVE_INSTALLER_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x371C, 'ERROR_SMI_PRIMITIVE_INSTALLER_FAILED') # WinError 14108 (0x371C)
    ERROR_GENERIC_COMMAND_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x371D, 'ERROR_GENERIC_COMMAND_FAILED') # WinError 14109 (0x371D)
    ERROR_SXS_FILE_HASH_MISSING: Final[ThirtyTwoBits] = cached_win32error(0x371E, 'ERROR_SXS_FILE_HASH_MISSING') # WinError 14110 (0x371E)
    ERROR_SXS_DUPLICATE_ACTIVATABLE_CLASS: Final[ThirtyTwoBits] = cached_win32error(0x371F, 'ERROR_SXS_DUPLICATE_ACTIVATABLE_CLASS') # WinError 14111 (0x371F)
    ERROR_EVT_INVALID_CHANNEL_PATH: Final[ThirtyTwoBits] = cached_win32error(0x3A98, 'ERROR_EVT_INVALID_CHANNEL_PATH') # WinError 15000 (0x3A98)
    ERROR_EVT_INVALID_QUERY: Final[ThirtyTwoBits] = cached_win32error(0x3A99, 'ERROR_EVT_INVALID_QUERY') # WinError 15001 (0x3A99)
    ERROR_EVT_PUBLISHER_METADATA_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x3A9A, 'ERROR_EVT_PUBLISHER_METADATA_NOT_FOUND') # WinError 15002 (0x3A9A)
    ERROR_EVT_EVENT_TEMPLATE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x3A9B, 'ERROR_EVT_EVENT_TEMPLATE_NOT_FOUND') # WinError 15003 (0x3A9B)
    ERROR_EVT_INVALID_PUBLISHER_NAME: Final[ThirtyTwoBits] = cached_win32error(0x3A9C, 'ERROR_EVT_INVALID_PUBLISHER_NAME') # WinError 15004 (0x3A9C)
    ERROR_EVT_INVALID_EVENT_DATA: Final[ThirtyTwoBits] = cached_win32error(0x3A9D, 'ERROR_EVT_INVALID_EVENT_DATA') # WinError 15005 (0x3A9D)
    ERROR_EVT_CHANNEL_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x3A9F, 'ERROR_EVT_CHANNEL_NOT_FOUND') # WinError 15007 (0x3A9F)
    ERROR_EVT_MALFORMED_XML_TEXT: Final[ThirtyTwoBits] = cached_win32error(0x3AA0, 'ERROR_EVT_MALFORMED_XML_TEXT') # WinError 15008 (0x3AA0)
    ERROR_EVT_SUBSCRIPTION_TO_DIRECT_CHANNEL: Final[ThirtyTwoBits] = cached_win32error(0x3AA1, 'ERROR_EVT_SUBSCRIPTION_TO_DIRECT_CHANNEL') # WinError 15009 (0x3AA1)
    ERROR_EVT_CONFIGURATION_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x3AA2, 'ERROR_EVT_CONFIGURATION_ERROR') # WinError 15010 (0x3AA2)
    ERROR_EVT_QUERY_RESULT_STALE: Final[ThirtyTwoBits] = cached_win32error(0x3AA3, 'ERROR_EVT_QUERY_RESULT_STALE') # WinError 15011 (0x3AA3)
    ERROR_EVT_QUERY_RESULT_INVALID_POSITION: Final[ThirtyTwoBits] = cached_win32error(0x3AA4, 'ERROR_EVT_QUERY_RESULT_INVALID_POSITION') # WinError 15012 (0x3AA4)
    ERROR_EVT_NON_VALIDATING_MSXML: Final[ThirtyTwoBits] = cached_win32error(0x3AA5, 'ERROR_EVT_NON_VALIDATING_MSXML') # WinError 15013 (0x3AA5)
    ERROR_EVT_FILTER_ALREADYSCOPED: Final[ThirtyTwoBits] = cached_win32error(0x3AA6, 'ERROR_EVT_FILTER_ALREADYSCOPED') # WinError 15014 (0x3AA6)
    ERROR_EVT_FILTER_NOTELTSET: Final[ThirtyTwoBits] = cached_win32error(0x3AA7, 'ERROR_EVT_FILTER_NOTELTSET') # WinError 15015 (0x3AA7)
    ERROR_EVT_FILTER_INVARG: Final[ThirtyTwoBits] = cached_win32error(0x3AA8, 'ERROR_EVT_FILTER_INVARG') # WinError 15016 (0x3AA8)
    ERROR_EVT_FILTER_INVTEST: Final[ThirtyTwoBits] = cached_win32error(0x3AA9, 'ERROR_EVT_FILTER_INVTEST') # WinError 15017 (0x3AA9)
    ERROR_EVT_FILTER_INVTYPE: Final[ThirtyTwoBits] = cached_win32error(0x3AAA, 'ERROR_EVT_FILTER_INVTYPE') # WinError 15018 (0x3AAA)
    ERROR_EVT_FILTER_PARSEERR: Final[ThirtyTwoBits] = cached_win32error(0x3AAB, 'ERROR_EVT_FILTER_PARSEERR') # WinError 15019 (0x3AAB)
    ERROR_EVT_FILTER_UNSUPPORTEDOP: Final[ThirtyTwoBits] = cached_win32error(0x3AAC, 'ERROR_EVT_FILTER_UNSUPPORTEDOP') # WinError 15020 (0x3AAC)
    ERROR_EVT_FILTER_UNEXPECTEDTOKEN: Final[ThirtyTwoBits] = cached_win32error(0x3AAD, 'ERROR_EVT_FILTER_UNEXPECTEDTOKEN') # WinError 15021 (0x3AAD)
    ERROR_EVT_INVALID_OPERATION_OVER_ENABLED_DIRECT_CHANNEL: Final[ThirtyTwoBits] = cached_win32error(0x3AAE, 'ERROR_EVT_INVALID_OPERATION_OVER_ENABLED_DIRECT_CHANNEL') # WinError 15022 (0x3AAE)
    ERROR_EVT_INVALID_CHANNEL_PROPERTY_VALUE: Final[ThirtyTwoBits] = cached_win32error(0x3AAF, 'ERROR_EVT_INVALID_CHANNEL_PROPERTY_VALUE') # WinError 15023 (0x3AAF)
    ERROR_EVT_INVALID_PUBLISHER_PROPERTY_VALUE: Final[ThirtyTwoBits] = cached_win32error(0x3AB0, 'ERROR_EVT_INVALID_PUBLISHER_PROPERTY_VALUE') # WinError 15024 (0x3AB0)
    ERROR_EVT_CHANNEL_CANNOT_ACTIVATE: Final[ThirtyTwoBits] = cached_win32error(0x3AB1, 'ERROR_EVT_CHANNEL_CANNOT_ACTIVATE') # WinError 15025 (0x3AB1)
    ERROR_EVT_FILTER_TOO_COMPLEX: Final[ThirtyTwoBits] = cached_win32error(0x3AB2, 'ERROR_EVT_FILTER_TOO_COMPLEX') # WinError 15026 (0x3AB2)
    ERROR_EVT_MESSAGE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x3AB3, 'ERROR_EVT_MESSAGE_NOT_FOUND') # WinError 15027 (0x3AB3)
    ERROR_EVT_MESSAGE_ID_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x3AB4, 'ERROR_EVT_MESSAGE_ID_NOT_FOUND') # WinError 15028 (0x3AB4)
    ERROR_EVT_UNRESOLVED_VALUE_INSERT: Final[ThirtyTwoBits] = cached_win32error(0x3AB5, 'ERROR_EVT_UNRESOLVED_VALUE_INSERT') # WinError 15029 (0x3AB5)
    ERROR_EVT_UNRESOLVED_PARAMETER_INSERT: Final[ThirtyTwoBits] = cached_win32error(0x3AB6, 'ERROR_EVT_UNRESOLVED_PARAMETER_INSERT') # WinError 15030 (0x3AB6)
    ERROR_EVT_MAX_INSERTS_REACHED: Final[ThirtyTwoBits] = cached_win32error(0x3AB7, 'ERROR_EVT_MAX_INSERTS_REACHED') # WinError 15031 (0x3AB7)
    ERROR_EVT_EVENT_DEFINITION_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x3AB8, 'ERROR_EVT_EVENT_DEFINITION_NOT_FOUND') # WinError 15032 (0x3AB8)
    ERROR_EVT_MESSAGE_LOCALE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x3AB9, 'ERROR_EVT_MESSAGE_LOCALE_NOT_FOUND') # WinError 15033 (0x3AB9)
    ERROR_EVT_VERSION_TOO_OLD: Final[ThirtyTwoBits] = cached_win32error(0x3ABA, 'ERROR_EVT_VERSION_TOO_OLD') # WinError 15034 (0x3ABA)
    ERROR_EVT_VERSION_TOO_NEW: Final[ThirtyTwoBits] = cached_win32error(0x3ABB, 'ERROR_EVT_VERSION_TOO_NEW') # WinError 15035 (0x3ABB)
    ERROR_EVT_CANNOT_OPEN_CHANNEL_OF_QUERY: Final[ThirtyTwoBits] = cached_win32error(0x3ABC, 'ERROR_EVT_CANNOT_OPEN_CHANNEL_OF_QUERY') # WinError 15036 (0x3ABC)
    ERROR_EVT_PUBLISHER_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x3ABD, 'ERROR_EVT_PUBLISHER_DISABLED') # WinError 15037 (0x3ABD)
    ERROR_EVT_FILTER_OUT_OF_RANGE: Final[ThirtyTwoBits] = cached_win32error(0x3ABE, 'ERROR_EVT_FILTER_OUT_OF_RANGE') # WinError 15038 (0x3ABE)
    ERROR_EC_SUBSCRIPTION_CANNOT_ACTIVATE: Final[ThirtyTwoBits] = cached_win32error(0x3AE8, 'ERROR_EC_SUBSCRIPTION_CANNOT_ACTIVATE') # WinError 15080 (0x3AE8)
    ERROR_EC_LOG_DISABLED: Final[ThirtyTwoBits] = cached_win32error(0x3AE9, 'ERROR_EC_LOG_DISABLED') # WinError 15081 (0x3AE9)
    ERROR_EC_CIRCULAR_FORWARDING: Final[ThirtyTwoBits] = cached_win32error(0x3AEA, 'ERROR_EC_CIRCULAR_FORWARDING') # WinError 15082 (0x3AEA)
    ERROR_EC_CREDSTORE_FULL: Final[ThirtyTwoBits] = cached_win32error(0x3AEB, 'ERROR_EC_CREDSTORE_FULL') # WinError 15083 (0x3AEB)
    ERROR_EC_CRED_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x3AEC, 'ERROR_EC_CRED_NOT_FOUND') # WinError 15084 (0x3AEC)
    ERROR_EC_NO_ACTIVE_CHANNEL: Final[ThirtyTwoBits] = cached_win32error(0x3AED, 'ERROR_EC_NO_ACTIVE_CHANNEL') # WinError 15085 (0x3AED)
    ERROR_MUI_FILE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x3AFC, 'ERROR_MUI_FILE_NOT_FOUND') # WinError 15100 (0x3AFC)
    ERROR_MUI_INVALID_FILE: Final[ThirtyTwoBits] = cached_win32error(0x3AFD, 'ERROR_MUI_INVALID_FILE') # WinError 15101 (0x3AFD)
    ERROR_MUI_INVALID_RC_CONFIG: Final[ThirtyTwoBits] = cached_win32error(0x3AFE, 'ERROR_MUI_INVALID_RC_CONFIG') # WinError 15102 (0x3AFE)
    ERROR_MUI_INVALID_LOCALE_NAME: Final[ThirtyTwoBits] = cached_win32error(0x3AFF, 'ERROR_MUI_INVALID_LOCALE_NAME') # WinError 15103 (0x3AFF)
    ERROR_MUI_INVALID_ULTIMATEFALLBACK_NAME: Final[ThirtyTwoBits] = cached_win32error(0x3B00, 'ERROR_MUI_INVALID_ULTIMATEFALLBACK_NAME') # WinError 15104 (0x3B00)
    ERROR_MUI_FILE_NOT_LOADED: Final[ThirtyTwoBits] = cached_win32error(0x3B01, 'ERROR_MUI_FILE_NOT_LOADED') # WinError 15105 (0x3B01)
    ERROR_RESOURCE_ENUM_USER_STOP: Final[ThirtyTwoBits] = cached_win32error(0x3B02, 'ERROR_RESOURCE_ENUM_USER_STOP') # WinError 15106 (0x3B02)
    ERROR_MUI_INTLSETTINGS_UILANG_NOT_INSTALLED: Final[ThirtyTwoBits] = cached_win32error(0x3B03, 'ERROR_MUI_INTLSETTINGS_UILANG_NOT_INSTALLED') # WinError 15107 (0x3B03)
    ERROR_MUI_INTLSETTINGS_INVALID_LOCALE_NAME: Final[ThirtyTwoBits] = cached_win32error(0x3B04, 'ERROR_MUI_INTLSETTINGS_INVALID_LOCALE_NAME') # WinError 15108 (0x3B04)
    ERROR_MRM_RUNTIME_NO_DEFAULT_OR_NEUTRAL_RESOURCE: Final[ThirtyTwoBits] = cached_win32error(0x3B06, 'ERROR_MRM_RUNTIME_NO_DEFAULT_OR_NEUTRAL_RESOURCE') # WinError 15110 (0x3B06)
    ERROR_MRM_INVALID_PRICONFIG: Final[ThirtyTwoBits] = cached_win32error(0x3B07, 'ERROR_MRM_INVALID_PRICONFIG') # WinError 15111 (0x3B07)
    ERROR_MRM_INVALID_FILE_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x3B08, 'ERROR_MRM_INVALID_FILE_TYPE') # WinError 15112 (0x3B08)
    ERROR_MRM_UNKNOWN_QUALIFIER: Final[ThirtyTwoBits] = cached_win32error(0x3B09, 'ERROR_MRM_UNKNOWN_QUALIFIER') # WinError 15113 (0x3B09)
    ERROR_MRM_INVALID_QUALIFIER_VALUE: Final[ThirtyTwoBits] = cached_win32error(0x3B0A, 'ERROR_MRM_INVALID_QUALIFIER_VALUE') # WinError 15114 (0x3B0A)
    ERROR_MRM_NO_CANDIDATE: Final[ThirtyTwoBits] = cached_win32error(0x3B0B, 'ERROR_MRM_NO_CANDIDATE') # WinError 15115 (0x3B0B)
    ERROR_MRM_NO_MATCH_OR_DEFAULT_CANDIDATE: Final[ThirtyTwoBits] = cached_win32error(0x3B0C, 'ERROR_MRM_NO_MATCH_OR_DEFAULT_CANDIDATE') # WinError 15116 (0x3B0C)
    ERROR_MRM_RESOURCE_TYPE_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x3B0D, 'ERROR_MRM_RESOURCE_TYPE_MISMATCH') # WinError 15117 (0x3B0D)
    ERROR_MRM_DUPLICATE_MAP_NAME: Final[ThirtyTwoBits] = cached_win32error(0x3B0E, 'ERROR_MRM_DUPLICATE_MAP_NAME') # WinError 15118 (0x3B0E)
    ERROR_MRM_DUPLICATE_ENTRY: Final[ThirtyTwoBits] = cached_win32error(0x3B0F, 'ERROR_MRM_DUPLICATE_ENTRY') # WinError 15119 (0x3B0F)
    ERROR_MRM_INVALID_RESOURCE_IDENTIFIER: Final[ThirtyTwoBits] = cached_win32error(0x3B10, 'ERROR_MRM_INVALID_RESOURCE_IDENTIFIER') # WinError 15120 (0x3B10)
    ERROR_MRM_FILEPATH_TOO_LONG: Final[ThirtyTwoBits] = cached_win32error(0x3B11, 'ERROR_MRM_FILEPATH_TOO_LONG') # WinError 15121 (0x3B11)
    ERROR_MRM_UNSUPPORTED_DIRECTORY_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x3B12, 'ERROR_MRM_UNSUPPORTED_DIRECTORY_TYPE') # WinError 15122 (0x3B12)
    ERROR_MRM_INVALID_PRI_FILE: Final[ThirtyTwoBits] = cached_win32error(0x3B16, 'ERROR_MRM_INVALID_PRI_FILE') # WinError 15126 (0x3B16)
    ERROR_MRM_NAMED_RESOURCE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x3B17, 'ERROR_MRM_NAMED_RESOURCE_NOT_FOUND') # WinError 15127 (0x3B17)
    ERROR_MRM_MAP_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x3B1F, 'ERROR_MRM_MAP_NOT_FOUND') # WinError 15135 (0x3B1F)
    ERROR_MRM_UNSUPPORTED_PROFILE_TYPE: Final[ThirtyTwoBits] = cached_win32error(0x3B20, 'ERROR_MRM_UNSUPPORTED_PROFILE_TYPE') # WinError 15136 (0x3B20)
    ERROR_MRM_INVALID_QUALIFIER_OPERATOR: Final[ThirtyTwoBits] = cached_win32error(0x3B21, 'ERROR_MRM_INVALID_QUALIFIER_OPERATOR') # WinError 15137 (0x3B21)
    ERROR_MRM_INDETERMINATE_QUALIFIER_VALUE: Final[ThirtyTwoBits] = cached_win32error(0x3B22, 'ERROR_MRM_INDETERMINATE_QUALIFIER_VALUE') # WinError 15138 (0x3B22)
    ERROR_MRM_AUTOMERGE_ENABLED: Final[ThirtyTwoBits] = cached_win32error(0x3B23, 'ERROR_MRM_AUTOMERGE_ENABLED') # WinError 15139 (0x3B23)
    ERROR_MRM_TOO_MANY_RESOURCES: Final[ThirtyTwoBits] = cached_win32error(0x3B24, 'ERROR_MRM_TOO_MANY_RESOURCES') # WinError 15140 (0x3B24)
    ERROR_MRM_UNSUPPORTED_FILE_TYPE_FOR_MERGE: Final[ThirtyTwoBits] = cached_win32error(0x3B25, 'ERROR_MRM_UNSUPPORTED_FILE_TYPE_FOR_MERGE') # WinError 15141 (0x3B25)
    ERROR_MRM_UNSUPPORTED_FILE_TYPE_FOR_LOAD_UNLOAD_PRI_FILE: Final[ThirtyTwoBits] = cached_win32error(0x3B26, 'ERROR_MRM_UNSUPPORTED_FILE_TYPE_FOR_LOAD_UNLOAD_PRI_FILE') # WinError 15142 (0x3B26)
    ERROR_MRM_NO_CURRENT_VIEW_ON_THREAD: Final[ThirtyTwoBits] = cached_win32error(0x3B27, 'ERROR_MRM_NO_CURRENT_VIEW_ON_THREAD') # WinError 15143 (0x3B27)
    ERROR_DIFFERENT_PROFILE_RESOURCE_MANAGER_EXIST: Final[ThirtyTwoBits] = cached_win32error(0x3B28, 'ERROR_DIFFERENT_PROFILE_RESOURCE_MANAGER_EXIST') # WinError 15144 (0x3B28)
    ERROR_OPERATION_NOT_ALLOWED_FROM_SYSTEM_COMPONENT: Final[ThirtyTwoBits] = cached_win32error(0x3B29, 'ERROR_OPERATION_NOT_ALLOWED_FROM_SYSTEM_COMPONENT') # WinError 15145 (0x3B29)
    ERROR_MRM_DIRECT_REF_TO_NON_DEFAULT_RESOURCE: Final[ThirtyTwoBits] = cached_win32error(0x3B2A, 'ERROR_MRM_DIRECT_REF_TO_NON_DEFAULT_RESOURCE') # WinError 15146 (0x3B2A)
    ERROR_MRM_GENERATION_COUNT_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x3B2B, 'ERROR_MRM_GENERATION_COUNT_MISMATCH') # WinError 15147 (0x3B2B)
    ERROR_PRI_MERGE_VERSION_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x3B2C, 'ERROR_PRI_MERGE_VERSION_MISMATCH') # WinError 15148 (0x3B2C)
    ERROR_PRI_MERGE_MISSING_SCHEMA: Final[ThirtyTwoBits] = cached_win32error(0x3B2D, 'ERROR_PRI_MERGE_MISSING_SCHEMA') # WinError 15149 (0x3B2D)
    ERROR_PRI_MERGE_LOAD_FILE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3B2E, 'ERROR_PRI_MERGE_LOAD_FILE_FAILED') # WinError 15150 (0x3B2E)
    ERROR_PRI_MERGE_ADD_FILE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3B2F, 'ERROR_PRI_MERGE_ADD_FILE_FAILED') # WinError 15151 (0x3B2F)
    ERROR_PRI_MERGE_WRITE_FILE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3B30, 'ERROR_PRI_MERGE_WRITE_FILE_FAILED') # WinError 15152 (0x3B30)
    ERROR_PRI_MERGE_MULTIPLE_PACKAGE_FAMILIES_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x3B31, 'ERROR_PRI_MERGE_MULTIPLE_PACKAGE_FAMILIES_NOT_ALLOWED') # WinError 15153 (0x3B31)
    ERROR_PRI_MERGE_MULTIPLE_MAIN_PACKAGES_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x3B32, 'ERROR_PRI_MERGE_MULTIPLE_MAIN_PACKAGES_NOT_ALLOWED') # WinError 15154 (0x3B32)
    ERROR_PRI_MERGE_BUNDLE_PACKAGES_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x3B33, 'ERROR_PRI_MERGE_BUNDLE_PACKAGES_NOT_ALLOWED') # WinError 15155 (0x3B33)
    ERROR_PRI_MERGE_MAIN_PACKAGE_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x3B34, 'ERROR_PRI_MERGE_MAIN_PACKAGE_REQUIRED') # WinError 15156 (0x3B34)
    ERROR_PRI_MERGE_RESOURCE_PACKAGE_REQUIRED: Final[ThirtyTwoBits] = cached_win32error(0x3B35, 'ERROR_PRI_MERGE_RESOURCE_PACKAGE_REQUIRED') # WinError 15157 (0x3B35)
    ERROR_PRI_MERGE_INVALID_FILE_NAME: Final[ThirtyTwoBits] = cached_win32error(0x3B36, 'ERROR_PRI_MERGE_INVALID_FILE_NAME') # WinError 15158 (0x3B36)
    ERROR_MRM_PACKAGE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x3B37, 'ERROR_MRM_PACKAGE_NOT_FOUND') # WinError 15159 (0x3B37)
    ERROR_MRM_MISSING_DEFAULT_LANGUAGE: Final[ThirtyTwoBits] = cached_win32error(0x3B38, 'ERROR_MRM_MISSING_DEFAULT_LANGUAGE') # WinError 15160 (0x3B38)
    ERROR_MRM_SCOPE_ITEM_CONFLICT: Final[ThirtyTwoBits] = cached_win32error(0x3B39, 'ERROR_MRM_SCOPE_ITEM_CONFLICT') # WinError 15161 (0x3B39)
    ERROR_MCA_INVALID_CAPABILITIES_STRING: Final[ThirtyTwoBits] = cached_win32error(0x3B60, 'ERROR_MCA_INVALID_CAPABILITIES_STRING') # WinError 15200 (0x3B60)
    ERROR_MCA_INVALID_VCP_VERSION: Final[ThirtyTwoBits] = cached_win32error(0x3B61, 'ERROR_MCA_INVALID_VCP_VERSION') # WinError 15201 (0x3B61)
    ERROR_MCA_MONITOR_VIOLATES_MCCS_SPECIFICATION: Final[ThirtyTwoBits] = cached_win32error(0x3B62, 'ERROR_MCA_MONITOR_VIOLATES_MCCS_SPECIFICATION') # WinError 15202 (0x3B62)
    ERROR_MCA_MCCS_VERSION_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x3B63, 'ERROR_MCA_MCCS_VERSION_MISMATCH') # WinError 15203 (0x3B63)
    ERROR_MCA_UNSUPPORTED_MCCS_VERSION: Final[ThirtyTwoBits] = cached_win32error(0x3B64, 'ERROR_MCA_UNSUPPORTED_MCCS_VERSION') # WinError 15204 (0x3B64)
    ERROR_MCA_INTERNAL_ERROR: Final[ThirtyTwoBits] = cached_win32error(0x3B65, 'ERROR_MCA_INTERNAL_ERROR') # WinError 15205 (0x3B65)
    ERROR_MCA_INVALID_TECHNOLOGY_TYPE_RETURNED: Final[ThirtyTwoBits] = cached_win32error(0x3B66, 'ERROR_MCA_INVALID_TECHNOLOGY_TYPE_RETURNED') # WinError 15206 (0x3B66)
    ERROR_MCA_UNSUPPORTED_COLOR_TEMPERATURE: Final[ThirtyTwoBits] = cached_win32error(0x3B67, 'ERROR_MCA_UNSUPPORTED_COLOR_TEMPERATURE') # WinError 15207 (0x3B67)
    ERROR_AMBIGUOUS_SYSTEM_DEVICE: Final[ThirtyTwoBits] = cached_win32error(0x3B92, 'ERROR_AMBIGUOUS_SYSTEM_DEVICE') # WinError 15250 (0x3B92)
    ERROR_SYSTEM_DEVICE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x3BC3, 'ERROR_SYSTEM_DEVICE_NOT_FOUND') # WinError 15299 (0x3BC3)
    ERROR_HASH_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x3BC4, 'ERROR_HASH_NOT_SUPPORTED') # WinError 15300 (0x3BC4)
    ERROR_HASH_NOT_PRESENT: Final[ThirtyTwoBits] = cached_win32error(0x3BC5, 'ERROR_HASH_NOT_PRESENT') # WinError 15301 (0x3BC5)
    ERROR_SECONDARY_IC_PROVIDER_NOT_REGISTERED: Final[ThirtyTwoBits] = cached_win32error(0x3BD9, 'ERROR_SECONDARY_IC_PROVIDER_NOT_REGISTERED') # WinError 15321 (0x3BD9)
    ERROR_GPIO_CLIENT_INFORMATION_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x3BDA, 'ERROR_GPIO_CLIENT_INFORMATION_INVALID') # WinError 15322 (0x3BDA)
    ERROR_GPIO_VERSION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x3BDB, 'ERROR_GPIO_VERSION_NOT_SUPPORTED') # WinError 15323 (0x3BDB)
    ERROR_GPIO_INVALID_REGISTRATION_PACKET: Final[ThirtyTwoBits] = cached_win32error(0x3BDC, 'ERROR_GPIO_INVALID_REGISTRATION_PACKET') # WinError 15324 (0x3BDC)
    ERROR_GPIO_OPERATION_DENIED: Final[ThirtyTwoBits] = cached_win32error(0x3BDD, 'ERROR_GPIO_OPERATION_DENIED') # WinError 15325 (0x3BDD)
    ERROR_GPIO_INCOMPATIBLE_CONNECT_MODE: Final[ThirtyTwoBits] = cached_win32error(0x3BDE, 'ERROR_GPIO_INCOMPATIBLE_CONNECT_MODE') # WinError 15326 (0x3BDE)
    ERROR_GPIO_INTERRUPT_ALREADY_UNMASKED: Final[ThirtyTwoBits] = cached_win32error(0x3BDF, 'ERROR_GPIO_INTERRUPT_ALREADY_UNMASKED') # WinError 15327 (0x3BDF)
    ERROR_CANNOT_SWITCH_RUNLEVEL: Final[ThirtyTwoBits] = cached_win32error(0x3C28, 'ERROR_CANNOT_SWITCH_RUNLEVEL') # WinError 15400 (0x3C28)
    ERROR_INVALID_RUNLEVEL_SETTING: Final[ThirtyTwoBits] = cached_win32error(0x3C29, 'ERROR_INVALID_RUNLEVEL_SETTING') # WinError 15401 (0x3C29)
    ERROR_RUNLEVEL_SWITCH_TIMEOUT: Final[ThirtyTwoBits] = cached_win32error(0x3C2A, 'ERROR_RUNLEVEL_SWITCH_TIMEOUT') # WinError 15402 (0x3C2A)
    ERROR_RUNLEVEL_SWITCH_AGENT_TIMEOUT: Final[ThirtyTwoBits] = cached_win32error(0x3C2B, 'ERROR_RUNLEVEL_SWITCH_AGENT_TIMEOUT') # WinError 15403 (0x3C2B)
    ERROR_RUNLEVEL_SWITCH_IN_PROGRESS: Final[ThirtyTwoBits] = cached_win32error(0x3C2C, 'ERROR_RUNLEVEL_SWITCH_IN_PROGRESS') # WinError 15404 (0x3C2C)
    ERROR_SERVICES_FAILED_AUTOSTART: Final[ThirtyTwoBits] = cached_win32error(0x3C2D, 'ERROR_SERVICES_FAILED_AUTOSTART') # WinError 15405 (0x3C2D)
    ERROR_COM_TASK_STOP_PENDING: Final[ThirtyTwoBits] = cached_win32error(0x3C8D, 'ERROR_COM_TASK_STOP_PENDING') # WinError 15501 (0x3C8D)
    ERROR_INSTALL_OPEN_PACKAGE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3CF0, 'ERROR_INSTALL_OPEN_PACKAGE_FAILED') # WinError 15600 (0x3CF0)
    ERROR_INSTALL_PACKAGE_NOT_FOUND: Final[ThirtyTwoBits] = cached_win32error(0x3CF1, 'ERROR_INSTALL_PACKAGE_NOT_FOUND') # WinError 15601 (0x3CF1)
    ERROR_INSTALL_INVALID_PACKAGE: Final[ThirtyTwoBits] = cached_win32error(0x3CF2, 'ERROR_INSTALL_INVALID_PACKAGE') # WinError 15602 (0x3CF2)
    ERROR_INSTALL_RESOLVE_DEPENDENCY_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3CF3, 'ERROR_INSTALL_RESOLVE_DEPENDENCY_FAILED') # WinError 15603 (0x3CF3)
    ERROR_INSTALL_OUT_OF_DISK_SPACE: Final[ThirtyTwoBits] = cached_win32error(0x3CF4, 'ERROR_INSTALL_OUT_OF_DISK_SPACE') # WinError 15604 (0x3CF4)
    ERROR_INSTALL_NETWORK_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x3CF5, 'ERROR_INSTALL_NETWORK_FAILURE') # WinError 15605 (0x3CF5)
    ERROR_INSTALL_REGISTRATION_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x3CF6, 'ERROR_INSTALL_REGISTRATION_FAILURE') # WinError 15606 (0x3CF6)
    ERROR_INSTALL_DEREGISTRATION_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x3CF7, 'ERROR_INSTALL_DEREGISTRATION_FAILURE') # WinError 15607 (0x3CF7)
    ERROR_INSTALL_CANCEL: Final[ThirtyTwoBits] = cached_win32error(0x3CF8, 'ERROR_INSTALL_CANCEL') # WinError 15608 (0x3CF8)
    ERROR_INSTALL_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3CF9, 'ERROR_INSTALL_FAILED') # WinError 15609 (0x3CF9)
    ERROR_REMOVE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3CFA, 'ERROR_REMOVE_FAILED') # WinError 15610 (0x3CFA)
    ERROR_PACKAGE_ALREADY_EXISTS: Final[ThirtyTwoBits] = cached_win32error(0x3CFB, 'ERROR_PACKAGE_ALREADY_EXISTS') # WinError 15611 (0x3CFB)
    ERROR_NEEDS_REMEDIATION: Final[ThirtyTwoBits] = cached_win32error(0x3CFC, 'ERROR_NEEDS_REMEDIATION') # WinError 15612 (0x3CFC)
    ERROR_INSTALL_PREREQUISITE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3CFD, 'ERROR_INSTALL_PREREQUISITE_FAILED') # WinError 15613 (0x3CFD)
    ERROR_PACKAGE_REPOSITORY_CORRUPTED: Final[ThirtyTwoBits] = cached_win32error(0x3CFE, 'ERROR_PACKAGE_REPOSITORY_CORRUPTED') # WinError 15614 (0x3CFE)
    ERROR_INSTALL_POLICY_FAILURE: Final[ThirtyTwoBits] = cached_win32error(0x3CFF, 'ERROR_INSTALL_POLICY_FAILURE') # WinError 15615 (0x3CFF)
    ERROR_PACKAGE_UPDATING: Final[ThirtyTwoBits] = cached_win32error(0x3D00, 'ERROR_PACKAGE_UPDATING') # WinError 15616 (0x3D00)
    ERROR_DEPLOYMENT_BLOCKED_BY_POLICY: Final[ThirtyTwoBits] = cached_win32error(0x3D01, 'ERROR_DEPLOYMENT_BLOCKED_BY_POLICY') # WinError 15617 (0x3D01)
    ERROR_PACKAGES_IN_USE: Final[ThirtyTwoBits] = cached_win32error(0x3D02, 'ERROR_PACKAGES_IN_USE') # WinError 15618 (0x3D02)
    ERROR_RECOVERY_FILE_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x3D03, 'ERROR_RECOVERY_FILE_CORRUPT') # WinError 15619 (0x3D03)
    ERROR_INVALID_STAGED_SIGNATURE: Final[ThirtyTwoBits] = cached_win32error(0x3D04, 'ERROR_INVALID_STAGED_SIGNATURE') # WinError 15620 (0x3D04)
    ERROR_DELETING_EXISTING_APPLICATIONDATA_STORE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3D05, 'ERROR_DELETING_EXISTING_APPLICATIONDATA_STORE_FAILED') # WinError 15621 (0x3D05)
    ERROR_INSTALL_PACKAGE_DOWNGRADE: Final[ThirtyTwoBits] = cached_win32error(0x3D06, 'ERROR_INSTALL_PACKAGE_DOWNGRADE') # WinError 15622 (0x3D06)
    ERROR_SYSTEM_NEEDS_REMEDIATION: Final[ThirtyTwoBits] = cached_win32error(0x3D07, 'ERROR_SYSTEM_NEEDS_REMEDIATION') # WinError 15623 (0x3D07)
    ERROR_APPX_INTEGRITY_FAILURE_CLR_NGEN: Final[ThirtyTwoBits] = cached_win32error(0x3D08, 'ERROR_APPX_INTEGRITY_FAILURE_CLR_NGEN') # WinError 15624 (0x3D08)
    ERROR_RESILIENCY_FILE_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x3D09, 'ERROR_RESILIENCY_FILE_CORRUPT') # WinError 15625 (0x3D09)
    ERROR_INSTALL_FIREWALL_SERVICE_NOT_RUNNING: Final[ThirtyTwoBits] = cached_win32error(0x3D0A, 'ERROR_INSTALL_FIREWALL_SERVICE_NOT_RUNNING') # WinError 15626 (0x3D0A)
    ERROR_PACKAGE_MOVE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3D0B, 'ERROR_PACKAGE_MOVE_FAILED') # WinError 15627 (0x3D0B)
    ERROR_INSTALL_VOLUME_NOT_EMPTY: Final[ThirtyTwoBits] = cached_win32error(0x3D0C, 'ERROR_INSTALL_VOLUME_NOT_EMPTY') # WinError 15628 (0x3D0C)
    ERROR_INSTALL_VOLUME_OFFLINE: Final[ThirtyTwoBits] = cached_win32error(0x3D0D, 'ERROR_INSTALL_VOLUME_OFFLINE') # WinError 15629 (0x3D0D)
    ERROR_INSTALL_VOLUME_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x3D0E, 'ERROR_INSTALL_VOLUME_CORRUPT') # WinError 15630 (0x3D0E)
    ERROR_NEEDS_REGISTRATION: Final[ThirtyTwoBits] = cached_win32error(0x3D0F, 'ERROR_NEEDS_REGISTRATION') # WinError 15631 (0x3D0F)
    ERROR_INSTALL_WRONG_PROCESSOR_ARCHITECTURE: Final[ThirtyTwoBits] = cached_win32error(0x3D10, 'ERROR_INSTALL_WRONG_PROCESSOR_ARCHITECTURE') # WinError 15632 (0x3D10)
    ERROR_DEV_SIDELOAD_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x3D11, 'ERROR_DEV_SIDELOAD_LIMIT_EXCEEDED') # WinError 15633 (0x3D11)
    ERROR_INSTALL_OPTIONAL_PACKAGE_REQUIRES_MAIN_PACKAGE: Final[ThirtyTwoBits] = cached_win32error(0x3D12, 'ERROR_INSTALL_OPTIONAL_PACKAGE_REQUIRES_MAIN_PACKAGE') # WinError 15634 (0x3D12)
    ERROR_PACKAGE_NOT_SUPPORTED_ON_FILESYSTEM: Final[ThirtyTwoBits] = cached_win32error(0x3D13, 'ERROR_PACKAGE_NOT_SUPPORTED_ON_FILESYSTEM') # WinError 15635 (0x3D13)
    ERROR_PACKAGE_MOVE_BLOCKED_BY_STREAMING: Final[ThirtyTwoBits] = cached_win32error(0x3D14, 'ERROR_PACKAGE_MOVE_BLOCKED_BY_STREAMING') # WinError 15636 (0x3D14)
    ERROR_INSTALL_OPTIONAL_PACKAGE_APPLICATIONID_NOT_UNIQUE: Final[ThirtyTwoBits] = cached_win32error(0x3D15, 'ERROR_INSTALL_OPTIONAL_PACKAGE_APPLICATIONID_NOT_UNIQUE') # WinError 15637 (0x3D15)
    ERROR_PACKAGE_STAGING_ONHOLD: Final[ThirtyTwoBits] = cached_win32error(0x3D16, 'ERROR_PACKAGE_STAGING_ONHOLD') # WinError 15638 (0x3D16)
    ERROR_INSTALL_INVALID_RELATED_SET_UPDATE: Final[ThirtyTwoBits] = cached_win32error(0x3D17, 'ERROR_INSTALL_INVALID_RELATED_SET_UPDATE') # WinError 15639 (0x3D17)
    ERROR_INSTALL_OPTIONAL_PACKAGE_REQUIRES_MAIN_PACKAGE_FULLTRUST_CAPABILITY: Final[ThirtyTwoBits] = cached_win32error(0x3D18, 'ERROR_INSTALL_OPTIONAL_PACKAGE_REQUIRES_MAIN_PACKAGE_FULLTRUST_CAPABILITY') # WinError 15640 (0x3D18)
    ERROR_DEPLOYMENT_BLOCKED_BY_USER_LOG_OFF: Final[ThirtyTwoBits] = cached_win32error(0x3D19, 'ERROR_DEPLOYMENT_BLOCKED_BY_USER_LOG_OFF') # WinError 15641 (0x3D19)
    ERROR_PROVISION_OPTIONAL_PACKAGE_REQUIRES_MAIN_PACKAGE_PROVISIONED: Final[ThirtyTwoBits] = cached_win32error(0x3D1A, 'ERROR_PROVISION_OPTIONAL_PACKAGE_REQUIRES_MAIN_PACKAGE_PROVISIONED') # WinError 15642 (0x3D1A)
    ERROR_PACKAGES_REPUTATION_CHECK_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3D1B, 'ERROR_PACKAGES_REPUTATION_CHECK_FAILED') # WinError 15643 (0x3D1B)
    ERROR_PACKAGES_REPUTATION_CHECK_TIMEDOUT: Final[ThirtyTwoBits] = cached_win32error(0x3D1C, 'ERROR_PACKAGES_REPUTATION_CHECK_TIMEDOUT') # WinError 15644 (0x3D1C)
    ERROR_DEPLOYMENT_OPTION_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x3D1D, 'ERROR_DEPLOYMENT_OPTION_NOT_SUPPORTED') # WinError 15645 (0x3D1D)
    ERROR_APPINSTALLER_ACTIVATION_BLOCKED: Final[ThirtyTwoBits] = cached_win32error(0x3D1E, 'ERROR_APPINSTALLER_ACTIVATION_BLOCKED') # WinError 15646 (0x3D1E)
    ERROR_REGISTRATION_FROM_REMOTE_DRIVE_NOT_SUPPORTED: Final[ThirtyTwoBits] = cached_win32error(0x3D1F, 'ERROR_REGISTRATION_FROM_REMOTE_DRIVE_NOT_SUPPORTED') # WinError 15647 (0x3D1F)
    ERROR_APPX_RAW_DATA_WRITE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3D20, 'ERROR_APPX_RAW_DATA_WRITE_FAILED') # WinError 15648 (0x3D20)
    ERROR_DEPLOYMENT_BLOCKED_BY_VOLUME_POLICY_PACKAGE: Final[ThirtyTwoBits] = cached_win32error(0x3D21, 'ERROR_DEPLOYMENT_BLOCKED_BY_VOLUME_POLICY_PACKAGE') # WinError 15649 (0x3D21)
    ERROR_DEPLOYMENT_BLOCKED_BY_VOLUME_POLICY_MACHINE: Final[ThirtyTwoBits] = cached_win32error(0x3D22, 'ERROR_DEPLOYMENT_BLOCKED_BY_VOLUME_POLICY_MACHINE') # WinError 15650 (0x3D22)
    ERROR_DEPLOYMENT_BLOCKED_BY_PROFILE_POLICY: Final[ThirtyTwoBits] = cached_win32error(0x3D23, 'ERROR_DEPLOYMENT_BLOCKED_BY_PROFILE_POLICY') # WinError 15651 (0x3D23)
    ERROR_DEPLOYMENT_FAILED_CONFLICTING_MUTABLE_PACKAGE_DIRECTORY: Final[ThirtyTwoBits] = cached_win32error(0x3D24, 'ERROR_DEPLOYMENT_FAILED_CONFLICTING_MUTABLE_PACKAGE_DIRECTORY') # WinError 15652 (0x3D24)
    ERROR_SINGLETON_RESOURCE_INSTALLED_IN_ACTIVE_USER: Final[ThirtyTwoBits] = cached_win32error(0x3D25, 'ERROR_SINGLETON_RESOURCE_INSTALLED_IN_ACTIVE_USER') # WinError 15653 (0x3D25)
    ERROR_DIFFERENT_VERSION_OF_PACKAGED_SERVICE_INSTALLED: Final[ThirtyTwoBits] = cached_win32error(0x3D26, 'ERROR_DIFFERENT_VERSION_OF_PACKAGED_SERVICE_INSTALLED') # WinError 15654 (0x3D26)
    ERROR_SERVICE_EXISTS_AS_NON_PACKAGED_SERVICE: Final[ThirtyTwoBits] = cached_win32error(0x3D27, 'ERROR_SERVICE_EXISTS_AS_NON_PACKAGED_SERVICE') # WinError 15655 (0x3D27)
    ERROR_PACKAGED_SERVICE_REQUIRES_ADMIN_PRIVILEGES: Final[ThirtyTwoBits] = cached_win32error(0x3D28, 'ERROR_PACKAGED_SERVICE_REQUIRES_ADMIN_PRIVILEGES') # WinError 15656 (0x3D28)
    ERROR_REDIRECTION_TO_DEFAULT_ACCOUNT_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x3D29, 'ERROR_REDIRECTION_TO_DEFAULT_ACCOUNT_NOT_ALLOWED') # WinError 15657 (0x3D29)
    ERROR_PACKAGE_LACKS_CAPABILITY_TO_DEPLOY_ON_HOST: Final[ThirtyTwoBits] = cached_win32error(0x3D2A, 'ERROR_PACKAGE_LACKS_CAPABILITY_TO_DEPLOY_ON_HOST') # WinError 15658 (0x3D2A)
    ERROR_UNSIGNED_PACKAGE_INVALID_CONTENT: Final[ThirtyTwoBits] = cached_win32error(0x3D2B, 'ERROR_UNSIGNED_PACKAGE_INVALID_CONTENT') # WinError 15659 (0x3D2B)
    ERROR_UNSIGNED_PACKAGE_INVALID_PUBLISHER_NAMESPACE: Final[ThirtyTwoBits] = cached_win32error(0x3D2C, 'ERROR_UNSIGNED_PACKAGE_INVALID_PUBLISHER_NAMESPACE') # WinError 15660 (0x3D2C)
    ERROR_SIGNED_PACKAGE_INVALID_PUBLISHER_NAMESPACE: Final[ThirtyTwoBits] = cached_win32error(0x3D2D, 'ERROR_SIGNED_PACKAGE_INVALID_PUBLISHER_NAMESPACE') # WinError 15661 (0x3D2D)
    ERROR_PACKAGE_EXTERNAL_LOCATION_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x3D2E, 'ERROR_PACKAGE_EXTERNAL_LOCATION_NOT_ALLOWED') # WinError 15662 (0x3D2E)
    ERROR_INSTALL_FULLTRUST_HOSTRUNTIME_REQUIRES_MAIN_PACKAGE_FULLTRUST_CAPABILITY: Final[ThirtyTwoBits] = cached_win32error(0x3D2F, 'ERROR_INSTALL_FULLTRUST_HOSTRUNTIME_REQUIRES_MAIN_PACKAGE_FULLTRUST_CAPABILITY') # WinError 15663 (0x3D2F)
    ERROR_PACKAGE_LACKS_CAPABILITY_FOR_MANDATORY_STARTUPTASKS: Final[ThirtyTwoBits] = cached_win32error(0x3D30, 'ERROR_PACKAGE_LACKS_CAPABILITY_FOR_MANDATORY_STARTUPTASKS') # WinError 15664 (0x3D30)
    ERROR_INSTALL_RESOLVE_HOSTRUNTIME_DEPENDENCY_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3D31, 'ERROR_INSTALL_RESOLVE_HOSTRUNTIME_DEPENDENCY_FAILED') # WinError 15665 (0x3D31)
    ERROR_MACHINE_SCOPE_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x3D32, 'ERROR_MACHINE_SCOPE_NOT_ALLOWED') # WinError 15666 (0x3D32)
    ERROR_CLASSIC_COMPAT_MODE_NOT_ALLOWED: Final[ThirtyTwoBits] = cached_win32error(0x3D33, 'ERROR_CLASSIC_COMPAT_MODE_NOT_ALLOWED') # WinError 15667 (0x3D33)
    ERROR_STAGEFROMUPDATEAGENT_PACKAGE_NOT_APPLICABLE: Final[ThirtyTwoBits] = cached_win32error(0x3D34, 'ERROR_STAGEFROMUPDATEAGENT_PACKAGE_NOT_APPLICABLE') # WinError 15668 (0x3D34)
    ERROR_PACKAGE_NOT_REGISTERED_FOR_USER: Final[ThirtyTwoBits] = cached_win32error(0x3D35, 'ERROR_PACKAGE_NOT_REGISTERED_FOR_USER') # WinError 15669 (0x3D35)
    ERROR_PACKAGE_NAME_MISMATCH: Final[ThirtyTwoBits] = cached_win32error(0x3D36, 'ERROR_PACKAGE_NAME_MISMATCH') # WinError 15670 (0x3D36)
    ERROR_APPINSTALLER_URI_IN_USE: Final[ThirtyTwoBits] = cached_win32error(0x3D37, 'ERROR_APPINSTALLER_URI_IN_USE') # WinError 15671 (0x3D37)
    ERROR_APPINSTALLER_IS_MANAGED_BY_SYSTEM: Final[ThirtyTwoBits] = cached_win32error(0x3D38, 'ERROR_APPINSTALLER_IS_MANAGED_BY_SYSTEM') # WinError 15672 (0x3D38)
    APPMODEL_ERROR_NO_PACKAGE: Final[ThirtyTwoBits] = cached_win32error(0x3D54, 'APPMODEL_ERROR_NO_PACKAGE') # WinError 15700 (0x3D54)
    APPMODEL_ERROR_PACKAGE_RUNTIME_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x3D55, 'APPMODEL_ERROR_PACKAGE_RUNTIME_CORRUPT') # WinError 15701 (0x3D55)
    APPMODEL_ERROR_PACKAGE_IDENTITY_CORRUPT: Final[ThirtyTwoBits] = cached_win32error(0x3D56, 'APPMODEL_ERROR_PACKAGE_IDENTITY_CORRUPT') # WinError 15702 (0x3D56)
    APPMODEL_ERROR_NO_APPLICATION: Final[ThirtyTwoBits] = cached_win32error(0x3D57, 'APPMODEL_ERROR_NO_APPLICATION') # WinError 15703 (0x3D57)
    APPMODEL_ERROR_DYNAMIC_PROPERTY_READ_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3D58, 'APPMODEL_ERROR_DYNAMIC_PROPERTY_READ_FAILED') # WinError 15704 (0x3D58)
    APPMODEL_ERROR_DYNAMIC_PROPERTY_INVALID: Final[ThirtyTwoBits] = cached_win32error(0x3D59, 'APPMODEL_ERROR_DYNAMIC_PROPERTY_INVALID') # WinError 15705 (0x3D59)
    APPMODEL_ERROR_PACKAGE_NOT_AVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x3D5A, 'APPMODEL_ERROR_PACKAGE_NOT_AVAILABLE') # WinError 15706 (0x3D5A)
    APPMODEL_ERROR_NO_MUTABLE_DIRECTORY: Final[ThirtyTwoBits] = cached_win32error(0x3D5B, 'APPMODEL_ERROR_NO_MUTABLE_DIRECTORY') # WinError 15707 (0x3D5B)
    ERROR_STATE_LOAD_STORE_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3DB8, 'ERROR_STATE_LOAD_STORE_FAILED') # WinError 15800 (0x3DB8)
    ERROR_STATE_GET_VERSION_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3DB9, 'ERROR_STATE_GET_VERSION_FAILED') # WinError 15801 (0x3DB9)
    ERROR_STATE_SET_VERSION_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3DBA, 'ERROR_STATE_SET_VERSION_FAILED') # WinError 15802 (0x3DBA)
    ERROR_STATE_STRUCTURED_RESET_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3DBB, 'ERROR_STATE_STRUCTURED_RESET_FAILED') # WinError 15803 (0x3DBB)
    ERROR_STATE_OPEN_CONTAINER_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3DBC, 'ERROR_STATE_OPEN_CONTAINER_FAILED') # WinError 15804 (0x3DBC)
    ERROR_STATE_CREATE_CONTAINER_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3DBD, 'ERROR_STATE_CREATE_CONTAINER_FAILED') # WinError 15805 (0x3DBD)
    ERROR_STATE_DELETE_CONTAINER_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3DBE, 'ERROR_STATE_DELETE_CONTAINER_FAILED') # WinError 15806 (0x3DBE)
    ERROR_STATE_READ_SETTING_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3DBF, 'ERROR_STATE_READ_SETTING_FAILED') # WinError 15807 (0x3DBF)
    ERROR_STATE_WRITE_SETTING_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3DC0, 'ERROR_STATE_WRITE_SETTING_FAILED') # WinError 15808 (0x3DC0)
    ERROR_STATE_DELETE_SETTING_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3DC1, 'ERROR_STATE_DELETE_SETTING_FAILED') # WinError 15809 (0x3DC1)
    ERROR_STATE_QUERY_SETTING_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3DC2, 'ERROR_STATE_QUERY_SETTING_FAILED') # WinError 15810 (0x3DC2)
    ERROR_STATE_READ_COMPOSITE_SETTING_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3DC3, 'ERROR_STATE_READ_COMPOSITE_SETTING_FAILED') # WinError 15811 (0x3DC3)
    ERROR_STATE_WRITE_COMPOSITE_SETTING_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3DC4, 'ERROR_STATE_WRITE_COMPOSITE_SETTING_FAILED') # WinError 15812 (0x3DC4)
    ERROR_STATE_ENUMERATE_CONTAINER_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3DC5, 'ERROR_STATE_ENUMERATE_CONTAINER_FAILED') # WinError 15813 (0x3DC5)
    ERROR_STATE_ENUMERATE_SETTINGS_FAILED: Final[ThirtyTwoBits] = cached_win32error(0x3DC6, 'ERROR_STATE_ENUMERATE_SETTINGS_FAILED') # WinError 15814 (0x3DC6)
    ERROR_STATE_COMPOSITE_SETTING_VALUE_SIZE_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x3DC7, 'ERROR_STATE_COMPOSITE_SETTING_VALUE_SIZE_LIMIT_EXCEEDED') # WinError 15815 (0x3DC7)
    ERROR_STATE_SETTING_VALUE_SIZE_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x3DC8, 'ERROR_STATE_SETTING_VALUE_SIZE_LIMIT_EXCEEDED') # WinError 15816 (0x3DC8)
    ERROR_STATE_SETTING_NAME_SIZE_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x3DC9, 'ERROR_STATE_SETTING_NAME_SIZE_LIMIT_EXCEEDED') # WinError 15817 (0x3DC9)
    ERROR_STATE_CONTAINER_NAME_SIZE_LIMIT_EXCEEDED: Final[ThirtyTwoBits] = cached_win32error(0x3DCA, 'ERROR_STATE_CONTAINER_NAME_SIZE_LIMIT_EXCEEDED') # WinError 15818 (0x3DCA)
    ERROR_API_UNAVAILABLE: Final[ThirtyTwoBits] = cached_win32error(0x3DE1, 'ERROR_API_UNAVAILABLE') # WinError 15841 (0x3DE1)
    STORE_ERROR_UNLICENSED: Final[ThirtyTwoBits] = cached_win32error(0x3DF5, 'STORE_ERROR_UNLICENSED') # WinError 15861 (0x3DF5)
    STORE_ERROR_UNLICENSED_USER: Final[ThirtyTwoBits] = cached_win32error(0x3DF6, 'STORE_ERROR_UNLICENSED_USER') # WinError 15862 (0x3DF6)
    STORE_ERROR_PENDING_COM_TRANSACTION: Final[ThirtyTwoBits] = cached_win32error(0x3DF7, 'STORE_ERROR_PENDING_COM_TRANSACTION') # WinError 15863 (0x3DF7)
    STORE_ERROR_LICENSE_REVOKED: Final[ThirtyTwoBits] = cached_win32error(0x3DF8, 'STORE_ERROR_LICENSE_REVOKED') # WinError 15864 (0x3DF8)
